

SET(DEFAULT_OPTIONS "-tra -c -force")

FUNCTION(GET_OPTIONS_TEST VOPTIONS)
  SET(VOPTIONS ${DEFAULT_OPTIONS} PARENT_SCOPE)
ENDFUNCTION(GET_OPTIONS_TEST VOPTIONS)

FUNCTION(GET_DIR_TEST EXAMPLE_NAME DIR_TEST)
  SET(XX ${CMAKE_CURRENT_SOURCE_DIR}/XTEST)
  IF(DEFINED MULTI)
    SET(XX ${XX}/${EXAMPLE_NAME})
  ENDIF(DEFINED MULTI)
  SET(DIR_TEST ${XX} PARENT_SCOPE)
ENDFUNCTION(GET_DIR_TEST DIR_TEST)

FUNCTION(COMPILE_OPTIONS EXAMPLE_NAME OPTIONS)

  IF(NOT EXISTS "$ENV{Signal_bin}")
    MESSAGE("============================================================================")
    MESSAGE("        The variable Signal_bin is not defined...exiting")
    MESSAGE("============================================================================")
    RETURN()
  ENDIF(NOT EXISTS "$ENV{Signal_bin}")


  SET(SIGNAL_COMPILER $ENV{pK_bin}/signal)
  SET(REF_DIR ${EXAMPLE_NAME})
  SET(SRC ${CMAKE_CURRENT_SOURCE_DIR}/${EXAMPLE_NAME}.SIG)
  SET(PAR ${CMAKE_CURRENT_SOURCE_DIR}/${EXAMPLE_NAME}.PAR)
  
  IF(NOT EXISTS ${PAR})
    SET(PARAMETERS " ")
  ELSE(NOT EXISTS ${PAR})
    SET(PARAMETERS "-par=${PAR}")
  ENDIF(NOT EXISTS ${PAR})


  SET(LANG) 
  IF ("${OPTIONS}" MATCHES "-c ")
    SET(LANG C)
  ELSEIF ("${OPTIONS}" MATCHES "-java ")
    SET(LANG Java)
  ENDIF ("${OPTIONS}" MATCHES "-c ")
  
  SET(MOD ${DIR_TEST}/${EXAMPLE_NAME}_body.c)
  
  SET_SOURCE_FILES_PROPERTIES(${MOD} 
    PROPERTIES OBJECT_DEPENDS "${SRC}"
    GENERATED TRUE
    )

  GET_DIR_TEST(${EXAMPLE_NAME} DIR_TEST)
  MESSAGE("      COMMAND ${CMAKE_COMMAND} -E chdir ${DIR_TEST} cmake . -DFOO=${EXAMPLE_NAME} -DLANG=${LANG}")

  IF(DEFINED LANG)
    ADD_CUSTOM_COMMAND(
      OUTPUT ${MOD}
      COMMAND ${CMAKE_COMMAND} -E echo "Compiling ${SRC}  ${PARAMETERS} ${OPTIONS} -D=${DIR_TEST}"
      COMMAND ${CMAKE_COMMAND} -E copy_directory ${REF_DIR} ${DIR_TEST}
      COMMAND ${CMAKE_COMMAND} -E remove -f ${MOD} ${DIR_TEST}/CMakeCache.txt
      COMMAND ${CMAKE_COMMAND} -E remove_directory ${DIR_TEST}/cmake
      COMMAND ${CMAKE_COMMAND} -E remove -f ${MOD} ${DIR_TEST}/${EXAMPLE_NAME}
      COMMAND ${CMAKE_COMMAND} -E remove -f ${MOD} ${DIR_TEST}/*.class
      COMMAND ${CMAKE_COMMAND} -E remove -f ${MOD} ${DIR_TEST}/*.o
      COMMAND ${CMAKE_COMMAND} -E make_directory ${DIR_TEST}/cmake
#      COMMAND ${CMAKE_COMMAND} -E chdir ${DIR_TEST} pre_compile
      COMMAND ${SIGNAL_COMPILER} ARGS ${SRC} ${PARAMETERS} ${OPTIONS} -D=${DIR_TEST}
      COMMAND ${CMAKE_COMMAND} -E chdir ${DIR_TEST} cmake . -DFOO=${EXAMPLE_NAME} -DLANG=${LANG}
      COMMAND ${CMAKE_COMMAND} -E chdir ${DIR_TEST} make
      COMMAND ${CMAKE_COMMAND} -E echo "Done."
      WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}
      DEPENDS ${SRC}
      )
  ELSE(DEFINED LANG)
    ADD_CUSTOM_COMMAND(
      OUTPUT ${MOD}
      COMMAND ${CMAKE_COMMAND} -E echo "Compiling ${SRC}  ${PARAMETERS} ${OPTIONS} -D=${DIR_TEST}"
      COMMAND ${SIGNAL_COMPILER} ARGS ${SRC} ${PARAMETERS} ${OPTIONS} -D=${DIR_TEST}
      COMMAND ${CMAKE_COMMAND} -E echo "Done."
      WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR}
      DEPENDS ${SRC}
      )
    
  ENDIF(DEFINED LANG)  
  MESSAGE("Target Name= T_${EXAMPLE_NAME}")
  ADD_CUSTOM_TARGET(T_${EXAMPLE_NAME} ALL DEPENDS "${MOD}")
  
ENDFUNCTION(COMPILE_OPTIONS EXAMPLE_NAME OPTIONS)

FUNCTION(COMPILE EXAMPLE_NAME)
  SET(MULTI)
  GET_OPTIONS_TEST(VOPTIONS)
  COMPILE_OPTIONS(${EXAMPLE_NAME} ${VOPTIONS})
ENDFUNCTION(COMPILE EXAMPLE_NAME)


FUNCTION(COMPILE_MULTI EXAMPLE_NAME)
  SET(MULTI TRUE)
  GET_OPTIONS_TEST(VOPTIONS)
  COMPILE_OPTIONS(${EXAMPLE_NAME} ${VOPTIONS})
ENDFUNCTION(COMPILE_MULTI EXAMPLE_NAME)


# Defining the CTEST.

ENABLE_TESTING()

# Defining a test by running EXAMPLE_NAME (C/C++ code generation).
FUNCTION(ADD_STD_TEST_LOC EXAMPLE_NAME)
  GET_DIR_TEST(${EXAMPLE_NAME} DIR_TEST)
  SET(COMMAND_TEST ${EXAMPLE_NAME})
  SET(NAME_TEST ${CMAKE_CURRENT_SOURCE_DIR}_TEST_${EXAMPLE_NAME})
  ADD_TEST(NAME ${NAME_TEST}
      WORKING_DIRECTORY "${DIR_TEST}" 
      COMMAND ${COMMAND_TEST}
      )
ENDFUNCTION(ADD_STD_TEST_LOC EXAMPLE_NAME)


# Defining a test by running EXAMPLE_NAME (C/C++ code generation).
FUNCTION(ADD_STD_TEST EXAMPLE_NAME)
  SET(MULTI)
  ADD_STD_TEST_LOC(${EXAMPLE_NAME})
ENDFUNCTION(ADD_STD_TEST EXAMPLE_NAME)


# Defining a test by running EXAMPLE_NAME (C/C++ code generation).
FUNCTION(ADD_STD_TEST_MULTI EXAMPLE_NAME)
  SET(MULTI TRUE)
  ADD_STD_TEST_LOC(${EXAMPLE_NAME})
ENDFUNCTION(ADD_STD_TEST_MULTI EXAMPLE_NAME)

# Defining a test by running 
#   java -classpath .:$ENV{SignalLibJava_graphical}/SignalGraphicalJavaStd.jar ${MAINCLASS}
# 
FUNCTION(ADD_JAVA_STD_GRAPHICAL_TEST EXAMPLE_NAME MAINCLASS)
  
  GET_DIR_TEST(${EXAMPLE_NAME} DIR_TEST)
  SET(COMMAND_TEST java -classpath .:$ENV{SignalLibJava_graphical}/SignalGraphicalJavaStd.jar ${MAINCLASS})
  SET(NAME_TEST ${CMAKE_CURRENT_SOURCE_DIR}_TEST_${EXAMPLE_NAME})
  ADD_TEST(NAME ${NAME_TEST}
    WORKING_DIRECTORY "${DIR_TEST}" 
    COMMAND ${COMMAND_TEST}
    )
ENDFUNCTION(ADD_JAVA_STD_GRAPHICAL_TEST EXAMPLE_NAME MAINCLASS)


FUNCTION(ADD_GREP_TEST EXAMPLE_NAME VSTR)
  
  GET_DIR_TEST(${EXAMPLE_NAME} DIR_TEST)
  SET(COMMAND_TEST "grep" "${VSTR}" "${DIR_TEST}/${EXAMPLE_NAME}_body.c")
  MESSAGE("COMMAND = ${COMMAND_TEST}")
  SET(NAME_TEST ${CMAKE_CURRENT_SOURCE_DIR}_TEST_${EXAMPLE_NAME})
  ADD_TEST(NAME ${NAME_TEST}
    WORKING_DIRECTORY "${DIR_TEST}" 
    COMMAND ${COMMAND_TEST}
    )
ENDFUNCTION(ADD_GREP_TEST EXAMPLE_NAME VTSR)

