% Chapter 5, section 5.4: A railway level crossing%
process RAILWAY = % Two-tracks railway level crossing %
  ( ? event A1, A2, A3, B1, B2, B3; % wheels on tracks A and B %
      boolean PRESENCE;
    ! boolean CLOSE;)
  (| CLOSE_A := TRACK (A1, A2, A3, ^PRESENCE)
   | CLOSE_B := TRACK (B1, B2, B3, ^PRESENCE)
   | CLOSE := BARRIER (CLOSE_A, CLOSE_B, PRESENCE)
   |)
  where
    integer CLOSE_A, CLOSE_B;
    process TRACK =
      ( ? event WHEEL1, WHEEL2, WHEEL3, TENTHS;
        ! integer CLOSE; % 1=crossing in 45s, 2=in 30s, 3=gone %
          )
      (| S ^= CNT ^= DETECT1^+ DETECT2^+ DETECT3^+ TENTHS
% States of the automaton %
       | S := -1 when DETECT1
            default 0 when DETECT2
            default 1 when ZS = 0 and CNT <= 450
            default 2 when ZS = 1 and CNT <= 300
            default 3 when DETECT3
            default ZS
       | ZS := S$ init 3
% counting of time, unit assumed: 0.1 sec %
       | CNT := 0 when DETECT1
             default 240*ZCNT when DETECT2
             default ZCNT+1 when ZS = -1
             default ZCNT-1 when ZCNT > 0
             default 0
       | ZCNT := CNT$ init 0
       | CLOSE := S when S /= ZS and S > 0
       | BUSY := WHEEL1 default not WHEEL3
       | ZBUSY := BUSY$ init false
       | DETECT1 := DETECT (WHEEL1, when ZBUSY and not BUSY)
       | DETECT2 := DETECT (WHEEL2, when ZBUSY and not BUSY)
       | DETECT3 := DETECT (WHEEL3, when BUSY and not ZBUSY)
       |)
      where
        integer S, ZS, CNT, ZCNT;
        event DETECT1, DETECT2, DETECT3;
        boolean BUSY, ZBUSY;
        
        process DETECT =
          ( ? event WHEEL, RESET;
            ! event COMING;)
          (| READY := RESET default not WHEEL
           | COMING := WHEEL when READY$ init true
           |)
          where
            boolean READY;
          end % DETECT %;
        
      end % TRACK %;
    process BARRIER =
      ( ? integer CLOSE_A, CLOSE_B; % closing orders from tracks A or B %
          boolean PRESENCE; % sampled every 0.1 s %
        ! boolean CLOSE; % effective action on barriers %
        )
        (| COMMON := ^CLOSE_A default ^CLOSE_B default ^PRESENCE
         | C_A := CLOSE_A cell COMMON
         | C_B := CLOSE_B cell COMMON
         | IS_CLOSED := PRESENCE cell COMMON
                     or C_A = 2 or C_B = 2 % imperative closing %
                     or WAS_CLOSED and (C_A = 1 or C_B = 1)
         | WAS_CLOSED := IS_CLOSED$ init false
         | CLOSE := IS_CLOSED when IS_CLOSED /= WAS_CLOSED
         |)
        where
          event COMMON;
          integer C_A, C_B;
          boolean IS_CLOSED, WAS_CLOSED;
        end % BARRIER %;
  end % RAILWAY %;
    
