/*
   Authors: Loïc Besnard, François Fabre, Thierry Gautier.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - EPL License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony SSME Platform (website: http://www.irisa.fr/espresso/polychrony).

   Polychrony SSME Platform is free software; you can redistribute it and/or modify it under
   the terms of the EPL as published by the Eclipse Foundation.

   Polychrony SSME Platform are distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the Eclipse Public License - v 1.0 for more details.

   You should have received a copy of the Eclipse Public License - v 1.0
   along with Polychrony SSME Platform; see the file COPYING. If not, consult the url
   http://www.eclipse.org/org/documents/epl-v10.php, or write to the Eclipse Foundation, Inc.
   102 Centrepointe Drive. Ottawa, Ontario,Canada, K2G 6B1.
*/

#define ERROR_NODE -1

static ptDCG vgraph;

static ptpKServicesAPI vpKServices;
static ptOptionManager OM;

static const char* strGraphValueClass="org/eclipse/pop/ssme/polychrony/GraphValue";

static char* copyString(const char *str, int size)
{
	char * c_name;
	int i;
	c_name = (char*)malloc (size+1);
	for (i =0; i < size; i++)
		c_name[i] = str[i];
	c_name[size] = '\0';
	return c_name;
}

static char* jstring2CString(jstring javaString, JNIEnv *env)
{
	const char *str = (env)->GetStringUTFChars(javaString, 0);
	int size = (env)->GetStringUTFLength(javaString);
	char* c_name= copyString(str, size);

	env->ReleaseStringUTFChars(javaString, str);
	env->DeleteLocalRef(javaString);
	return c_name;
}

static jobject mkReturnGraphValue(JNIEnv * env, jint ret, jlong vgraph)
{
  jclass cls = env->FindClass(strGraphValueClass);
  if (env->ExceptionOccurred() || cls == NULL) return NULL; /* exception thrown */

  /* Get the method ID for the GraphValue constructor */
  jmethodID cid = env->GetMethodID(cls, "<init>", "(IJ)V");
  if (env->ExceptionOccurred() || cid == NULL) return NULL; /* exception thrown */
  
  /* Construct a GraphValue object */
  jobject result = env->NewObject(cls, cid, ret, (jlong) vgraph);
  /* Free local references */
  env->DeleteLocalRef(cls);  // ????(Loïc)

  return result;
}


static boolean DCGPoly2DCGBool(ptDCG_poly graphNode,ptDCG_bool *gBool, boolean btraceALL)
{
  ptDCG_endo gEndo=NULL;
  boolean b;

  if (! vpKServices->DCGPoly2DCGEndo((ptDCG_poly) graphNode, &gEndo)) return pk_false;
  pK_messenger->announce_ENDO();
  b = vpKServices->DCGEndo2DCGBool(gEndo, gBool);
  pK_messenger->announce_BOOL();
  return b;

}

static void announce_POLYCHRONY()
{
  pK_messenger->gmessload(systmess);
  pK_messenger->announce_POLYCHRONY();
}
