/*
   Authors: Loïc Besnard, François Fabre, Thierry Gautier.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - EPL License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony SSME Platform (website: http://www.irisa.fr/espresso/polychrony).

   Polychrony SSME Platform is free software; you can redistribute it and/or modify it under
   the terms of the EPL as published by the Eclipse Foundation.

   Polychrony SSME Platform are distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the Eclipse Public License - v 1.0 for more details.

   You should have received a copy of the Eclipse Public License - v 1.0
   along with Polychrony SSME Platform; see the file COPYING. If not, consult the url
   http://www.eclipse.org/org/documents/epl-v10.php, or write to the Eclipse Foundation, Inc.
   102 Centrepointe Drive. Ottawa, Ontario,Canada, K2G 6B1.
*/

#include <string.h>
#define ERROR_NODE -1

#define PARSER_MESSAGE_SIZE 1024
static char gStringErrorParser[PARSER_MESSAGE_SIZE];

static char* parserFileName;

static char* copyString(const char *str, int size)
{
	char * c_name;
	int i;
	c_name = allocate_string(size+1);
	for (i =0; i < size; i++)
		c_name[i] = str[i];
	c_name[size] = '\0';
	return c_name;
}

static char* jstring2CString(jstring javaString, JNIEnv *env)
{
	const char *str = (env)->GetStringUTFChars(javaString, 0);
	int size = (env)->GetStringUTFLength(javaString);
	char* c_name= copyString(str, size);

	env->ReleaseStringUTFChars(javaString, str);
	env->DeleteLocalRef(javaString);
	return c_name;
}

static jstring str2jstring(ptrepstr vstr, JNIEnv *env)
{
	return env->NewStringUTF(repstr2string(vstr));
}

static ptrepstr jstring2str(jstring js, JNIEnv *env)
{
	return string2repstr(jstring2CString(js, env));
}

/**
 * Saves the string content (bytes) in the parserFileName file
 * This file will be given to the syntaxic analyser
 */
static boolean mkParserFile(JNIEnv *env, jstring expression)
{
	FILE *fp = fopen(parserFileName, "w");
	if (fp == NULL)
		return pk_false;

	char * str = jstring2CString(expression, env);

	fwrite(str, 1, strlen(str), fp);
	fprintf(fp, "\n");
	fclose(fp);
	return pk_true;
}

static void announce_POLYCHRONY()
{
  pK_messenger->gmessload(systmess);
  pK_messenger->announce_POLYCHRONY();
}


/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseString
 * Signature: (ILjava/lang/String;)I
 */
jlong JNICALL parse(JNIEnv * env, jobject obj,jstring expression, jint phylum)
{
	ntrees result;

	setJNIEnv(env);
	if (mkParserFile(env, expression) && sysparser((integer)phylum, parserFileName, &result)) return (jlong) result;
	return (jlong) ERROR_NODE;
}
