/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor */
  GraphManager::GraphManager ()
{
  database = NILLPOINTER;
  number = 0;
}

/** Destructor */
GraphManager::~GraphManager ()
{
  database = NILLPOINTER;
}

/** @defgroup  graphDataBaseSet GraphManager: updating graph database. */

/** @ingroup graphDataBaseSet
    It adds a graph (g) in the database.
    @return the number of the graph. This number is not yet assigned with it.
*/
integer
GraphManager::addGraph (ptDCG g)
{
  ptInfoGraph elem;

  elem = new InfoGraph (NILPOINTER, g);
  ptlPointer_insert_elem (&database, elem);
  number++;
  return number;
}

/**  @ingroup graphDataBaseSet
     It associates an abstract tree (vAbstractTree) with a graph (g) in the database.
*/
void
GraphManager::setLinkGraph (ptDCG g, pointer vAbstractTree)
{
  ptInfoGraph infoGraph;
  if (getEntry (g, &infoGraph))
    infoGraph->setLink (vAbstractTree);
}

/**  @ingroup graphDataBaseSet
     It increases the number of references associated with a graph (g) in the database.
*/
void
GraphManager::incrNbReferences (ptDCG g)
{
  ptInfoGraph infoGraph;
  integer val;

  if (getEntry (g, &infoGraph))
    {
      val = infoGraph->getNbReferences ();
      infoGraph->setNbReferences (val++);
    }
}

/** @defgroup  graphDataBaseGet GraphManager: access to the graph database. */

/** @ingroup graphDataBaseGet
    @return the number of references associated with a graph (g) in the database.
*/
integer
GraphManager::getNbReferences (ptDCG g)
{
  ptInfoGraph infoGraph;

  if (getEntry (g, &infoGraph))
    return infoGraph->getNbReferences ();
  return -1;
}

/** @ingroup graphDataBaseGet
    @return "true" if it exists a graph (g) associated with an abstract tree (vAbstractTree) in the database, "false" otherwise.
*/
boolean
GraphManager::getDCGOfTree (pointer vAbstractTree, ptDCG * g)
{
  ptInfoGraph infoGraph;

  if (!getEntry (vAbstractTree, &infoGraph))
    return pk_false;
  (*g) = infoGraph->getGraph ();
  return pk_true;
}

/**  @ingroup graphDataBaseGet
     @return "true" if it exists an abstract tree (vtree) associated with a graph (g) in the database, "false" otherwise.
*/
boolean
GraphManager::getTreeOfIGraph (ptDCG g, pointer * vtree)
{
  ptInfoGraph infoGraph;

  if (!getEntry (g, &infoGraph))
    return pk_false;
  (*vtree) = infoGraph->getProduceFrom ();
  return pk_true;
}


/**  @ingroup graphDataBaseGet
     @return "true" if it exists a graph (g) associated with a process with id as name, "false" otherwise.   
boolean GraphManager::xxgetGraph(ptident id, ptDCG *g)
{
  ptlPointer li;
  pointer elem;
  ptDCG g1;
  ptInfoGraph info;
 
  li = database;
  while (pk_true)
    {
      if (!ptlPointer_get_elemNext(&li, &elem)) return pk_false;
      info = (ptInfoGraph) elem;
      g1 = info->getGraph();
      if (eq_idents(g1->getName(),id)) {*g=g1; return pk_true;}
    }
  return pk_false;
}
*/


/** @ingroup graphDataBaseSet
    It removes a graph (g) from the database.
*/
void
GraphManager::removeGraph (ptDCG g)
{
  ptInfoGraph infoGraph;

  if (getEntry (g, &infoGraph))
    ptlPointer_remove_elem (&database, infoGraph);

}
