/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor of an element of an associative table composed of 
    - the abstract tree produced by the parser for a model (vproduceFrom),
    - the graph associated with the abstract tree (vg) ,
    - the number of references of the graph in the environment (set to 0).
*/
  InfoGraph::InfoGraph (pointer vproduceFrom, ptDCG vg)
{
  nbReferences = 0;
  produceFrom = vproduceFrom;
  vgraph = vg;
}

/** @return the graph of the element.*/
ptDCG
InfoGraph::getGraph ()
{
  return vgraph;
}

/** @return the abstract tree of the element.*/
pointer
InfoGraph::getProduceFrom ()
{
  return produceFrom;
}

/** @return the number of references of the element.*/
integer
InfoGraph::getNbReferences ()
{
  return nbReferences;
}

/** It sets to val the the number of references of the element. */
void
InfoGraph::setNbReferences (integer val)
{
  nbReferences = val;
}

/** It sets to vp the abstract tree of the element. */
void
InfoGraph::setLink (pointer vp)
{
  produceFrom = vp;
}
