/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   The functionality is applied on a graph (g).
*/
  booleans2events::booleans2events (ptDCG g)
{
  AGRAPH = g;
  EM = g->getExpressionManager ();
  CKM = g->getClockManager ();
  NM = g->getNodeManager ();
  lsigrew = NILLINSTANCES;
  metNodes = NILLINODE;
  CKM->initSystemForModification ();
  CKM->hclmklequiv (0);
  the_nullClock_B2E_copy = CKM->getNullClock ();
  initrew (pk_false, pk_false, pk_false);
  vmark = pKbool2eventMark;	// pKbool2eventMark is a predefined mark
}

/**
   Destructor.
*/
booleans2events::~booleans2events ()
{
}

/**
  It normalizes boolean expressions:
     - logical expreesion if logicalOp is pk_true
     - relational expression if relatOp is pk_true
     - default expression if DefaultExp is pk_true
     @return true if an expression is rewritten, false otherwise.
*/
boolean booleans2events::apply (boolean logicalOp, boolean relatOp, boolean DefaultExp)
{
  boolean bresult, somecondrewrite;
  ptinsth vh;
  ptlinsth list;

  initrew (logicalOp, relatOp, DefaultExp);
  bresult = pk_false;
  somecondrewrite = pk_true;
  while (somecondrewrite)
    {
      somecondrewrite = pk_false;
      list = CKM->getFormulaList (sub);
      while (ptlinsth_get_next (&list, &vh))
	if (applyPositiveExtraction (vh))
	  {
	    CKM->hclmklequiv (0);
	    somecondrewrite = pk_true;
	    bresult = pk_true;
	  }
    }
  EM->resetMarkAll (vmark);
  return bresult;
}

