/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" is x1 and x2 are the same clock (formula in the the same class), "false" otherwise.
*/
  _VIRTUAL_ boolean booleans2events::satisfiesClockEq (ptinsth x1, ptinsth x2)
{
  x1 = CKM->getRepresentativeFormula (x1);
  x2 = CKM->getRepresentativeFormula (x2);
  return (x1 == x2);
}

/** @return the clock [vexpcond] ^* vhexpcond
    @param vexpcond: a Signal expression.
    @param vhexpcond: the clock of  the expression vexpcond.
    @param vhns (out): the clock [not vexpcond] ^* vhexpcond.
*/
ptinsth
booleans2events::mksubnosub (ptexpcal vexpcond, ptinsth vhexpcond, ptinsth * vhns)
{
  ptinsth vhresult;
  noperators vop;
  ptinsth vh, vhc, vhcnot;
  boolean bsw;

  vhexpcond = CKM->getRepresentativeFormula (vhexpcond);

  if (EM->getOperator (vexpcond) == opnon)
    {
      bsw = pk_true;
      vexpcond = EM->getUnaryArg (vexpcond);
    }
  else
    bsw = pk_false;

  if (EM->getSynchronousExprHorl (vexpcond, &vh))
    {
      if (EM->IsEvent (vexpcond))
	{
	  vhc = vhexpcond;
	  vhcnot = the_nullClock_B2E_copy;
	}
      else
	{
	  vhc = CKM->makeSub (vh, vexpcond, EM->getPredecessors (vexpcond));
	  vhcnot = CKM->makeNoSub (vexpcond);
	}

      if (!satisfiesClockEq (vh, vhexpcond))
	{
	  vhresult = CKM->makeInf (vhc, vhexpcond);
	  (*vhns) = CKM->makeInf (vhcnot, vhexpcond);
	}
      else
	{
	  vhresult = vhc;
	  (*vhns) = vhcnot;
	}
    }
  else
    {
      vop = EM->getOperator (vexpcond);
      if (vop == optrue)
	{
	  vhresult = vhexpcond;
	  *vhns = the_nullClock_B2E_copy;
	}
      else if (vop == opfalse)
	{
	  vhresult = the_nullClock_B2E_copy;
	  *vhns = vhexpcond;
	}
      else
	{
	  vhresult = CKM->makeSub (vhexpcond, vexpcond, EM->getPredecessors (vexpcond));
	  *vhns = CKM->makeNoSub (vexpcond);
	}
    }
  if (bsw)
    {
      vh = vhresult;
      vhresult = *vhns;
      *vhns = vh;
    }
  return vhresult;
}


/**
  @return the result of the rewritting of when (vexpcond).
  @param vexpcond  (in) : a default expression (E1 default E2... default En)
  @param vhns (out):  result of the rewritting of when (not vexpcond).
*/
ptinsth
booleans2events::mksubnosubmele (ptexpcal vexpcond, ptinsth * vhns)
{
  ptinsth vhresult, vhsubg, vhsubd, vhnosubg, vhnosubd, hcontr1, hcontr2;
  ptexpcal vhe1, vhe2, vexp1, vexp2;

  EM->getDefaultArgs (vexpcond, &vhe1, &vhe2);
  vexp1 = EM->getWhenOrCellArg (vhe1);
  if (EM->getOperator (vexp1) == opmele)
    vhsubg = mksubnosubmele (vexp1, &vhnosubg);
  else
    {
      hcontr1 = EM->getContributionClock (vhe1);
      vhsubg = mksubnosub (vexp1, hcontr1, &vhnosubg);
    }
  vexp2 = EM->getWhenOrCellArg (vhe2);
  if (EM->getOperator (vexp2) == opmele)
    vhsubd = mksubnosubmele (vexp2, &vhnosubd);
  else
    {
      hcontr2 = EM->getContributionClock (vhe2);
      vhsubd = mksubnosub (vexp2, hcontr2, &vhnosubd);
    }

  vhresult = CKM->makeSup (vhsubg, vhsubd);
  *vhns = CKM->makeSup (vhnosubg, vhnosubd);
  return vhresult;
}

/** @return "true" if the operand expression exp is a recursive one.
   @param exp  (in): assumed to be a synchronous expression.
*/
boolean
booleans2events::isexpsyncrec (ptexpcal exp)
{
  noperators ope;

  ope = EM->getOperator (exp);
  switch (ope)
    {
    case opident:
      return is_equationrec (exp);

    case opou:
    case opouex:
    case opet:
      if (isexpsyncrec (EM->getBinaryArgstAt (exp, 1)))
	return pk_true;
      return isexpsyncrec (EM->getBinaryArgstAt (exp, 2));

    case opnon:
      return isexpsyncrec (EM->getUnaryArg (exp));
    }
  return pk_false;
}

/** @return "true" if the operand expression exp is a recursive one.
   @param exp  (in): assumed to be a DEFAULT expression.
*/
boolean
booleans2events::isexpmelerec (ptexpcal exp)
{
  boolean b;
  ptexpcal vhe1, vhe2, vexp1, vexp2;

  EM->getDefaultArgs (exp, &vhe1, &vhe2);
  vexp1 = EM->getWhenOrCellArg (vhe1);
  if (EM->getOperator (vexp1) == opmele)
    b = isexpmelerec (vexp1);
  else
    b = isexpsyncrec (vexp1);
  if (b)
    return pk_true;
  vexp2 = EM->getWhenOrCellArg (vhe2);
  if (EM->getOperator (vexp2) == opmele)
    return isexpmelerec (vexp2);
  return isexpsyncrec (vexp2);
}

/** @return "true" if the signal vsig is a recursive one.
   @param vsig  (in): assumed to be a variable expression (signal).
*/
boolean
booleans2events::is_equationrec (ptexpcal vsig)
{
  ptINode vinstr;
  ptexpcal vhe, exp;
  boolean b;

  if (ptlinstances_isin_exp (lsigrew, vsig))
    return pk_false;

  vinstr = EM->getNode (vsig);
  if (ptlINode_isIn (metNodes, vinstr))
    return pk_true;

  ptlINode_insert_elem (&metNodes, vinstr);
  b = pk_false;
  if (NM->is_equation (vinstr))
    if (NM->get_rightOperand (vinstr, &vhe))
      if (EM->IsCellExpr (vhe))
	{
	  exp = EM->getWhenOrCellArg (vhe);
	  if (EM->getOperator (exp) != opmele)
	    b = isexpsyncrec (exp);
	  else if (rewmele)
	    b = isexpmelerec (exp);
	}
  if (!b)
    lsigrew = ptlinstances_push_front (lsigrew, vsig);

  ptlINode_remove_elem (&metNodes, vinstr);
  return b;
}

/** @return "true" if the operand expression vsig is a default expression (direct or it is 
   a signal defined by a default expression).
   @param vsig (in): a signal expression.
*/
boolean
booleans2events::isSignalDefault (ptexpcal vsig)
{
  ptINode vinstr;
  ptexpcal expd;

  if (!EM->IsSignal (vsig))
    return (EM->getOperator (vsig) == opmele);
  vinstr = EM->getNode (vsig);
  if (!NM->is_equation (vinstr))
    return pk_false;
  NM->get_rightOperand (vinstr, &expd);
  expd = EM->getWhenOrCellArg (expd);
  if (EM->IsWhenOrCellExpr (expd))
    expd = EM->getWhenOrCellArg (expd);
  return (EM->getOperator (expd) == opmele);
}


/** Rewriting.
   It replaces the extraction vhsub by vhsubn and 
   the extraction vhnosub by vhnosubn in the clock system.
   It assumes that vhsub and vhnosubn are build on the same expression (one is positive, the other negative extraction).
*/
void
booleans2events::replaceSubs (ptinsth vhsub, ptinsth vhsubn, ptinsth vhnosub, ptinsth vhnosubn)
{
  boolean keep;
  ptexpcal e;
  boolean becomesvari;

  e = CKM->getConditionSub (vhsub);
  if (EM->IsEvent (e))
    keep = pk_false;
  else if ((vhsubn == the_nullClock_B2E_copy) || (vhnosubn == the_nullClock_B2E_copy))
    keep = pk_false;
  else
    keep = !isSignalDefault (e);

  //  CKM->replaceSub (vhsub, vhsubn, keep); CKM->replaceSub (vhnosub, vhnosubn, keep); replaced by : LBesnard Oct 2015.  
  if (vhsub == vhsubn)
    CKM->removeSub (vhsub, true, &becomesvari);
  else
    CKM->replaceSub (vhsub, vhsubn, keep);
  if (vhnosub == vhnosubn)
    CKM->removeSub (vhnosub, true, &becomesvari);
  else
    CKM->replaceSub (vhnosub, vhnosubn, keep);
}

/**
  @return 
   - pk_true if the expression EXP with VHEXP as clock is a relationnal expression in normal form
   ( <= or = operator and an order on expressions are used for the normal form).
   - false othewise, the rewritting of the expression is done (in E) and 
      - vhsub is set to the clock [E];
      - vhnosub is set to the clock [not E].
*/
boolean
booleans2events::isRelationalExprNormalForm (ptexpcal exp, ptinsth vhexp, ptinsth * vhsub, ptinsth * vhnosub)
{
  ptexpcal newexp;

  if (EM->normFormRelationalExpr (exp, &newexp))
    return pk_true;
  *vhsub = mksubnosub (newexp, vhexp, vhnosub);
  return pk_false;
}

/** Rewriting OR expression. 
    It applies the rules:
      - [e1 or e2] --> [e1] ^+ [e2]
      - [not(e1 or e2)]  --> [not e1] ^* [not e2]

   @param exp  (in) an OR  expression (e1 or e2).
   @param vhexp  (in) the clock of exp.
   @param vhn  (out) the clock [e1] ^+ [e2]
   @param vhnosubn  (out) the clock  [not e1] ^* [not e2]
*/
void
booleans2events::rewor (ptexpcal exp, ptinsth vhexp, ptinsth * vhn, ptinsth * vhnosubn)
{
  ptexpcal vopg, vopd;
  ptinsth vhsubg, vhnosubg, vhsubd, vhnosubd;

  EM->getBinaryArgs (exp, &vopg, &vopd);
  vhsubg = mksubnosub (vopg, vhexp, &vhnosubg);
  vhsubd = mksubnosub (vopd, vhexp, &vhnosubd);
  *vhn = CKM->makeSup (vhsubg, vhsubd);
  *vhnosubn = CKM->makeInf (vhnosubg, vhnosubd);
}


/** Rewriting EXCLUSIVE-OR expression.
   It applies the rules:
     - [e1 eor e2] --> ([e1] ^*[not e2]) ^+ ([e2] ^* [not e1])
     - [not(e1 or e2)]  --> ([not e1] ^* [not e2]) ^+ ( [e1] ^* [e2])

   @param exp  (in) an EXCLUSIVE-OR  expression (e1 eor e2). (assumed)
   @param vhexp  (in) the clock of exp.
   @param vhn  (out) the clock ([e1] ^*[not e2]) ^+ ([e2] ^* [not e1]).
   @param vhnosubn  (out) the clock ([not e1] ^* [not e2]) ^+ ( [e1] ^* [e2]). 
*/
void
booleans2events::rewexclusiveor (ptexpcal exp, ptinsth vhexp, ptinsth * vhn, ptinsth * vhnosubn)
{
  ptexpcal vopg, vopd;
  ptinsth vhsubg, vhnosubg, vhsubd, vhnosubd, t1, t2;

  EM->getBinaryArgs (exp, &vopg, &vopd);
  vhsubg = mksubnosub (vopg, vhexp, &vhnosubg);
  vhsubd = mksubnosub (vopd, vhexp, &vhnosubd);

  t1 = CKM->makeInf (vhnosubg, vhsubd);
  t2 = CKM->makeInf (vhnosubd, vhsubg);
  *vhn = CKM->makeSup (t1, t2);

  t1 = CKM->makeInf (vhnosubg, vhnosubd);
  t2 = CKM->makeInf (vhsubd, vhsubg);
  *vhnosubn = CKM->makeSup (t1, t2);

}



/** Rewriting AND expression.
   It applies the rules:
    - [e1 and e2] -->[e1] ^* [e2] 
    - [not(e1 and e2)]  -->[not e1] ^+ [not e2]

   @param exp  (in) an AND  expression (e1 and e2). (assumed)
   @param vhexp  (in) the clock of exp.
   @param vhn  (out) the clock [e1] ^* [e2]
   @param vhnosubn  (out) the clock [not e1] ^+ [not e2]
*/
void
booleans2events::rewand (ptexpcal exp, ptinsth vhexp, ptinsth * vhn, ptinsth * vhnosubn)
{
  ptexpcal vopg, vopd;
  ptinsth vhsubg, vhnosubg, vhsubd, vhnosubd;

  EM->getBinaryArgs (exp, &vopg, &vopd);
  vhsubg = mksubnosub (vopg, vhexp, &vhnosubg);
  vhsubd = mksubnosub (vopd, vhexp, &vhnosubd);
  *vhn = CKM->makeInf (vhsubg, vhsubd);
  *vhnosubn = CKM->makeSup (vhnosubg, vhnosubd);
}

/** Rewriting NOT expression. Applying the rule:
       - [not(not e)]  --> [e]

   @param exp  (in) an (assumed) NOT  expression (not e). (assumed)
   @param  vhexp  (in) the clock of exp.
   @param vhsubn the clock when (e).
   @param vhnosubn  (out) the clock when (not e). 
*/
void
booleans2events::rewnot (ptexpcal exp, ptinsth vhexp, ptinsth * vhsubn, ptinsth * vhnosubn)
{
  ptinsth x;

  x = mksubnosub (EM->getUnaryArg (exp), vhexp, vhnosubn);
  *vhsubn = *vhnosubn;
  *vhnosubn = x;
}

/**
  Conditionnal rewriting boolean signal.
  @return
  - pk_true if a signal (vsig with vhsig as clock) is "free" (input, delay, produced by an external call), or rewritings are not allowed.
  - false otherwise, in this case (where E denotes the expression of the definition of the signal vsig)
        - vhsubn is the clock "when E"
	- vhnosubn the clock "when (not E)". 
*/
boolean
booleans2events::isFreeSignal (ptexpcal vsig, ptinsth vhsig, ptinsth * vhsubn, ptinsth * vhnosubn)
{
  ptINode vinstr;
  ptexpcal vhe;
  ptexpcal exp;
  ptinsth vhexp;

  if (!EM->IsSignal (vsig)) return pk_true;
  if (EM->isStateVar (vsig)) return pk_true;

  if (EM->IsEvent (vsig))
    {
      *vhsubn = vhsig;
      *vhnosubn = the_nullClock_B2E_copy;
      return pk_false;
    }

  if (is_equationrec (vsig)) return pk_true;
  vinstr = EM->getNode (vsig);
  if (!NM->is_equation (vinstr)) return pk_true;

  if (!NM->get_rightOperand (vinstr, &vhe)) return pk_true;
  if (EM->IsCellExpr (vhe))
    {
      exp = EM->getWhenOrCellArg (vhe);
      if (EM->getContributionClock (exp) != EM->getContributionClock (vhe))
	return pk_true;
    }
  exp = EM->getWhenOrCellArg (vhe);
  if (EM->getOperator (exp) == opmele)	/* default */
    {
      if (!rewmele)
	return pk_true;
      *vhsubn = mksubnosubmele (exp, vhnosubn);
      return pk_false;
    }
  /* when */
  vhexp = EM->getContributionClock (vhe);
  if (CKM->satisfiesInitClock (vhexp)) vhexp = vhsig;
  *vhsubn = mksubnosub (exp, vhexp, vhnosubn);
  return pk_false;
}


/** @return 
   - pk_true if the expression exp is an event expression, 
   then vhsub and vhnsub are respectivly set to vhexp and the_hnullClock_B2E_copy.
   false otherwise.
*/
boolean
booleans2events::isConstantExtract (ptexpcal exp, ptinsth vhexp, ptinsth * vhsub, ptinsth * vhnsub)
{
  if (EM->IsEvent (exp) || EM->eq_trueExpcal (exp))
    {
      *vhsub = vhexp;
      *vhnsub = the_nullClock_B2E_copy;
      return pk_true;
    }
  if (EM->eq_falseExpcal (exp))
    {
      *vhsub = the_nullClock_B2E_copy;
      *vhnsub = vhexp;
      return pk_true;
    }
  return pk_false;
}

/**
  Conditionnal rewriting of an expression.
  @return "true" if an expression (exp) with vhexp as clock is free.
  "false" otherwise, in this case the rewriting of the expression is done (depending on some global parameters)
    - nvhsub is the clock that replaces [exp] 
    - nvhnosub the clock that repalces [not exp]

*/
boolean
booleans2events::isFreeCondition (ptexpcal exp, ptinsth vhexp, ptinsth * nvhsub, ptinsth * nvhnosub)
{
  noperators ope;

  if (isConstantExtract (exp, vhexp, nvhsub, nvhnosub))
    return pk_false;

  if (rewcond)
    {
      ope = EM->getOperator (exp);
      switch (ope)
	{
	case opident:
	  return isFreeSignal (exp, vhexp, nvhsub, nvhnosub);
	case opou:
	  rewor (exp, vhexp, nvhsub, nvhnosub);
	  return pk_false;
	case opouex:
	  rewexclusiveor (exp, vhexp, nvhsub, nvhnosub);
	  return pk_false;
	case opet:
	  rewand (exp, vhexp, nvhsub, nvhnosub);
	  return pk_false;
	case opnon:
	  rewnot (exp, vhexp, nvhsub, nvhnosub);
	  return pk_false;
	}
    }
  if (rewexprel)
    return isRelationalExprNormalForm (exp, vhexp, nvhsub, nvhnosub);
  return pk_true;
}


/**
  @return 
    - pk_true if the expression (e) assigned to an (assumed) positive extraction clock (vhsub) is not a free condition.
      In this case, the clock [e] and [not e] are replaced in the system;
    - false otherwise, e is a free condition or rewritings are not allowed.
*/
boolean booleans2events::applyPositiveExtraction (ptinsth vhsub)
{
  boolean somecondrewrite;
  ptinsth vhcond, vhnosub, nvhsub, nvhnosub;
  ptexpcal exp;

  // if (vhsub != CKM->getRepresentativeFormula (vhsub)) return pk_false;  remove Oct 2015 LBesnard.
  if (CKM->satisfiesIsAnAssumption (vhsub)) return pk_false;
  if (CKM->isAnAffineClock (vhsub)) return pk_false;
  vhcond = CKM->getClockSub (vhsub);
  CKM->getComplInFather (vhsub, &vhnosub);
  somecondrewrite = pk_false;
  if (vhcond == the_nullClock_B2E_copy)
    {
      vhsub = CKM->makeSynchro (vhsub, the_nullClock_B2E_copy);
      vhnosub = CKM->makeSynchro (vhnosub, the_nullClock_B2E_copy);
    }
  else
    {
      exp = CKM->getConditionSub (vhsub);
      if (!EM->isMarked (exp, vmark))
	{
	  EM->addMark (exp, vmark);
	  if (!isFreeCondition (exp, vhcond, &nvhsub, &nvhnosub))
	    {
	      replaceSubs (vhsub, nvhsub, vhnosub, nvhnosub);
	      somecondrewrite = pk_true;
	    }
	}
    }
  CKM->hclmklequiv (0);
  return somecondrewrite;
}

/**
  Initializing rewrittings.
  @param logicalOp  (in) when it is pk_true, logical expressions  are rewritten;
  @param relatOp  (in) when it is pk_true, relational expressions are rewritten;
  @param defaultOp  (in) when it is pk_true, default expressions are rewritten;
*/
void
booleans2events::initrew (boolean logicalOp, boolean relatOp, boolean defaultOp)
{
  rewcond = logicalOp;		/* logical operator */
  rewexprel = relatOp;		/* relational operator */
  rewmele = defaultOp;		/* default expression */
}
