/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.\n
   The functionality is applied on a graph (g).\n
    A \ref GraphTravel "travel object" is created and associated with the current object.
*/
  CircuitCalculus::CircuitCalculus (ptDCG g)
{
  DEBUG_MODE = pk_false;
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  PARC = new TravelDepManager (AGRAPH);
  PARC->setDebugMode (DEBUG_MODE);
  lireached = NILLINODE;
  lhreached = NILLINSTH;
}

/** Destructor.
 */
CircuitCalculus::~CircuitCalculus ()
{
}

/**
   @return "true" if the graph (parameter of the constructor) contains cycles, "false" otherwise.
   The list of the cycles can be obtained by getCycles() method.
*/
boolean CircuitCalculus::circuit_detection ()
{
  lireached = NILLINODE;
  lhreached = NILLINSTH;
  return circuitDetection ();
}

/**
   @return the list of the static cycles of the graph (ie it includes the false cycles).
   The cycles are available after the circuit_detection() method.
*/
ptlPointer CircuitCalculus::getCycles ()
{
  cycles = NILLPOINTER;
  visitReached ();
  return cycles;
}
