/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
/**
   Misc: tracing operation for a node (node).
*/
  void
ClosureCalculus::TRACE (ptINode node)
{
  ptrepstr vstr;
  integer num;

  return;
  num = NM->get_num (node);
  makeRepstr_string_int (" node: ", num, &vstr);	/* not a name: non standard use */
  pK_messenger->printStringDebug (vstr);
}

/**
   Closure calculus for a node (node) in the current mode (the successors are not informed).
*/
void
ClosureCalculus::nodeClosure (ptINode node)
{
  switch (NM->get_oper (node))
    {
    case opdefpart:
    case opdefsig:
      equClosure (node);
      break;

    case opmoti:
      arrayClosure (node);
      break;

    case opret:
      delayClosure (node);
      break;

    case opappe:
      procCallClosure (node);
      break;

    case opsubgraph:
    case opldefpart:
    case opbund:		// Symbolise struct/bundle
    case opnodecall:		// Rien a faire : traitement reportés sur les sorties du "nodeCall"
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}

/**
    If the clock (clk) has not been explored, it is marked and 
       - its successors are visited (recursively)
       - then, the closure calculus is applied for the clock (clockClosure() method)  
*/
_VIRTUAL_ void
ClosureCalculus::visitClock (ptinsth clk)
{
  if (!CKM->satisfiesMarked (clk, vmarkpushed))
    {
      CKM->addMark (clk, vmarkpushed);
      visitClocks (CKM->getClockReferencesTo (clk));
      visitNodes (CKM->getNodeReferencesTo (clk));
      clockClosure (clk);
      CKM->removeMark (clk, vmarkpushed);
    }
}

/**
   If a node (node) has not been explored, it is marked and 
   - its successors are visited (recursively)
   - then, the closure calculus is applied for the node (nodeClosure() method).
*/
_VIRTUAL_ void
ClosureCalculus::visitNode (ptINode node)
{
  ptINode imoti;

  if (!NM->isMarked (node, vmarkpushed))
    {
      TRACE (node);
      NM->addMark (node, vmarkpushed);
      visitDependences (NM->get_successors (node));
      visitClocks (NM->get_clockSuccessors (node));
      if (NM->is_partialDefinition (node))
	{
	  imoti = EM->getNode (NM->get_leftOperand (node));
	  visitNode (imoti);
	}
      else
	imoti = node;
      while (NM->get_inComposedNode (imoti, &imoti))
	visitNode (imoti);

      nodeClosure (node);
      NM->removeMark (node, vmarkpushed);
    }
}
