/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return a new \ref lclosure descriptor composed of a list of signals predecessors (iclos) 
    and  a list of clock predecessors (hclos).
 */
extern ptlclosure
ptlclosure_RIEN_create (pticlosure iclos, pthclosure hclos)
{
  ptlclosure newlclos;

  newlclos = new lclosure ();
  newlclos->iclos = iclos;
  newlclos->hclos = hclos;
  return newlclos;
}


/**
   @return a new \ref iclosure descriptor for a signal (exp) and the dependence is conditionned by a clock (hdep).
*/
extern pticlosure
pticlosure_RIEN_create (ptinsth hdep, ptexpcal exp, pticlosure nextclos)
{
  pticlosure newiclos;

  newiclos = new iclosure ();
  newiclos->hdep = hdep;
  newiclos->from = exp;
  newiclos->next = nextclos;
  return newiclos;
}

/**
   @return a new \ref hclosure descriptor for a clock (h) and the dependence is conditionned by a clock (hdep).
*/
extern pthclosure
pthclosure_RIEN_create (ptinsth hdep, ptinsth h, pthclosure nextclos)
{
  pthclosure newhclos;

  newhclos = new hclosure ();
  newhclos->hdep = hdep;
  newhclos->from = h;
  newhclos->next = nextclos;
  return newhclos;
}

/** 
    @return the list of signal-predecessors of a closure descriptor.
*/
extern pticlosure
ptlclosure_get_iclos (ptlclosure lclos)
{
  if (lclos == NILLCLOSURE)
    return NILICLOSURE;
  return lclos->iclos;
}



/** It assigns the list of signal-predecessors (iclos) of a closure descriptor (lclos).
    The closure descriptor is created if it is undefined.
*/
extern void
ptlclosure_set_iclos (ptlclosure * lclos, pticlosure iclos)
{
  if ((*lclos) != NILLCLOSURE)
    (*lclos)->iclos = iclos;
  else if (iclos != NILICLOSURE)
    (*lclos) = ptlclosure_RIEN_create (iclos, NILHCLOSURE);
}

/** 
    @return the list of clock-predecessors of a closure descriptor.
*/
extern pthclosure
ptlclosure_get_hclos (ptlclosure lclos)
{
  if (lclos == NILLCLOSURE)
    return NILHCLOSURE;
  return lclos->hclos;
}

/** It assigns the list of clock-predecessors (hclos)  of a closure descriptor(lclos).
    The closure descriptor is created if it is undefined.
*/
extern void
ptlclosure_set_hclos (ptlclosure * lclos, pthclosure hclos)
{
  if ((*lclos) != NILLCLOSURE)
    (*lclos)->hclos = hclos;
  else if (hclos != NILHCLOSURE)
    *lclos = ptlclosure_RIEN_create (NILICLOSURE, hclos);
}

/** @return the signal source of a dependence (iclos).
 */
extern ptexpcal
pticlosure_get_from (pticlosure iclos)
{
  return iclos->from;
}

/**
   @return the clock of a dependence (iclos).
*/
extern ptinsth
pticlosure_get_hdep (pticlosure iclos)
{
  return iclos->hdep;
}

/**
   It assigns the clock (h) of a dependence (iclos).
*/
extern void
pticlosure_set_hdep (pticlosure iclos, ptinsth h)
{
  iclos->hdep = h;
}

/** 
    @return the next element of a iclosure dependence.
*/
extern pticlosure
pticlosure_get_next (pticlosure iclos)
{
  return iclos->next;
}

/**
   The successor of iclos becomes iclosnext.
 */
extern void
pticlosure_set_next (pticlosure iclos, pticlosure iclosnext)
{
  iclos->next = iclosnext;
}

/** 
    @return the head of the list (iclos) and sets this list (iclos) to the 
    tail of the list.
    Assumed : iclos is not empty.
*/
extern pticlosure
pticlosure_removehead_elem (pticlosure * iclos)
{
  pticlosure vres;

  vres = *iclos;
  *iclos = (*iclos)->next;
  return vres;
}

/**
   @return the clock of a dependence (hclos).
*/
extern ptinsth
pthclosure_get_from (pthclosure hclos)
{
  return hclos->from;
}

/** Horloge  de dependance de ?HCLOS */
extern ptinsth
pthclosure_get_hdep (pthclosure hclos)
{
  return hclos->hdep;
}

/**
   @return the clock of a dependence (iclos).
*/
extern void
pthclosure_set_hdep (pthclosure hclos, ptinsth h)
{
  hclos->hdep = h;
}

/** 
    @return the next element of a hclosure dependence.
*/
extern pthclosure
pthclosure_get_next (pthclosure hclos)
{
  return hclos->next;
}

/**
   The successor of hclos becomes hclosnext.
 */
extern void
pthclosure_set_next (pthclosure hclos, pthclosure hclosnext)
{
  hclos->next = hclosnext;
}

/** 
    @return the head of the list (hclos) and sets this list (hclos) to the 
    tail of the list.
    Assumed : hclos is not empty.
*/
extern pthclosure
pthclosure_removehead_elem (pthclosure * hclos)
{
  pthclosure vres;

  vres = *hclos;
  *hclos = (*hclos)->next;
  return vres;
}
