/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/** 
    Representation of the dependences associated with a signal or a clock.
    It is used only for cycle detection and closure calculus.
    Two lists are associated with each signal (\b x), 
     - a list of (p1 when h1,..., pn when hn) where pi are signals, hi are clocks
     - a list of (k1 when v1,..., km when vm where ki are clocks, and vi are clocks
     It specifies that 
      - x depends on pj when hj is available (transitively)
      - x depends on kj when vj is available (transitively)
*/
typedef struct lclosure
{
  pticlosure iclos;		///< list of signals, predecessors with conditionned dependence.
  pthclosure hclos;		///< list of clocks, predecessors with conditionned dependence.
}
tlclosure;

/**
   List of dependences (from when hdep) where from is a signal.
*/
typedef struct iclosure
{
  ptinsth hdep;			///<  condition associated with the depedence 
  ptexpcal from;		///< signal predecessor.
  pticlosure next;		///< next element of the list.
}
ticlosure;

/**
   List of dependences (from when hdep) where from is a clock.
*/
typedef struct hclosure
{
  ptinsth hdep;			///<  condition associated with the depedence
  ptinsth from;			///< signal predecessor.
  pthclosure next;		///< next element of the list.
}
thclosure;
