/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

typedef ptlMark ptlrepstr;
#define NILLREPSTR    (ptlrepstr)pK_NIL  


typedef class cGenerator  *ptcGenerator; ///< Pointer to a C generator objet.

#include <list>
#include <map> // STL

class cGenerator
{
  typedef std::map<TypeType, ptident> TypeTypeIdentTableType;
  
#include "cGenerator_m.h"
#include "cGeneratorSimulator_i.h"
#include "cGeneratorParametrizedCode_i.h"
#include "cGeneratorPrefixManagment_i.h"
#include "cGeneratorHeaderBody_i.h"
#include "cGeneratorTypes_i.h"
#include "cGeneratorExternFile_i.h"
#include "cGeneratorIO_i.h"
#include "cGeneratorExpr_i.h"
#include "cGeneratorClock_i.h"
#include "cGeneratorDeclaration_i.h"
#include "cGeneratorInit_i.h"
#include "cGeneratorNode_i.h"
#include "cGeneratorSpatialProcessing_i.h"
#include "cGeneratorCmp_i.h"

#include "cGenerator_g.h"

protected:
  ptDCG_sch AGRAPH;    ///< the functionality is applied to a DCG_sch graph.
  ptExprManager EM;   ///< operand graph info: expression manager.
  ptClockManager CKM; ///< operand graph info: clock manager
  ptNodeManager NM;   ///< operand graph info: node manager.
  ptNameManager nameManager;  ///< Name manager of the graph AGRAPH
  ptTraceabilityManager TM; ///< Traceability manager assigned to the AGRAPH graph.

  boolean bEndochronousWithInternalClockRoot; ///< true when the graph is an endochronous with an internal clock root.
  ptlinstances Inputs,     ///< Graph infos: Inputs
    Outputs,               ///< Graph infos: Outputs
    Locals;                ///< Graph infos: locals (all the local signals are merged in this list).
  ptlinstances Constants,  ///< Graph infos: constants
    Parameters,            ///< Graph infos: parameters
    Indexes;               ///< Graph infos: indexes
  ptlinstances Types;      ///< Graph infos: user types.
  
  ptlINode ExternalsProcs;     ///< Graph infos: list of externals calls
  ptlinstances ExternalsFuncs; ///< Graph infos: list of externals calls used in expressions (only functions)

  ptident progName;           ///< Graph infos: name of the program.
  ptinsth tick;             ///< Graph infos (sbDC+): the tick clock of the program (always exists in sbDCplus)
  ptinsth vhComplete;       ///< Graph infos: the Complete clock
  
  ptelemTableGenCode *SubGraphTable; /**<  Graph infos (sbDC+): set of sub-graphs. 
					- SubGraphTable[i] (i>0) is the set of ordered nodes of the i-th cluster; 
					- SubGraphTable[0] is the sequential scheduler
				     */
  int nbSets;  ///< Graph infos (sbDC+): size of the SubGraphTable table
  ptlINode lNodesHinit;  ///< Graph infos (sbDC+): constant signals
  ptlinstances lClocksNotToBeInitiliazed;  ///< Graph infos (sbDC+): Clocks built on memories ; they must NOT be initialized.

  // temporary attributes
  typeMark AttributeNameGencode; 
  typeMark NOTGREYBOXBLACKBOXREFERENCE;
  typeMark travelIndexAttribute, IteratorAttribute;
  typeMark OPTIMIZED_INDEX;
  typeMark INIT_DONE;
  typeMark affineSampleAttribute;
  typeMark alreadyVisisted;  ///< to avoid  code duplication.
  typeMark vcdmark; ///< For VCD simulation.

  ptlINode lAffineSample;
  ptlinstances indexStack;       ///< Spatial managing: stack of indexes.
  int nbMaxIndex;                ///< Spatial managing.
  ptlinstances InternalIndexLst; ///< Spatial managing.
  noperators contextOp;          ///< Spatial managing: current operation set to NULLOP (used only for sequential definition (next operator))

  TypeType tprec;        ///< current type (declaration)
  ptident nomtypprec;    ///< current type name (declaration)
  ptinsth currentClock;  ///< current clock.
  boolean okForElse;     ///< Set to "true" when the "else" statement is accepted. It allows to generate "if (c) {statements} else {statements}" instead of "if (c) {statements}; if (!c) {statements}".
  boolean SEPARATE, 
    bCallEntrypointStepInitialize, 
    INTERNAL_BLOCK,
    ARRAY_CONTEXT;

  ptexpcal One, Zero;

  ptlinstances lIncrDelay;  ///< Management of the nodes of delay trees. List of updated indexes.

  ptlpragmas tableOpCode;
  ptrepstr repstrBlackBox, repstrGreyBox;
  ptlrepstr lBlackGreyBoxesDeclarations;
  // ptident idIterBlackBox;
  boolean functionDeclaration, someProblem, debutdcl, sameLine;
  ptrepstr PragmaGencodeLanguage, PragmaImportLanguage, PragmaDeclareLanguage;
  ptrepstr prefixString;

  TypeTypeIdentTableType EqualEqualTable;  //< == operator
  TypeTypeIdentTableType LessLessEqualTable;  //< <<= operator
  ptident eqeqId, ltlteqId; ///<  == identifier and <<= identifier.
  TypeType undefinedType;
  boolean extensionInhibit; ///<  context for extension of the operators.
  boolean externalObjects; ///< set to true when external objects are referenced.

  ptinteprintc TextForm; ///< text shaping object
};



