
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ CCodeGenerator::CCodeGenerator ()
{
  lids = NILTREE;
}

CCodeGenerator::CCodeGenerator (ptDCG_sch g):CodeGenerator (g)
{
  lids = NILTREE;
}


/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    This is the entry point of the C code generation in SEPARATED COMPILING mode.
    It is applied to a graph which must be a sbDC+ one. The graph is with or without clusters.
    It generates the following files (for each graph of a process)
    
    -  foo_externals.h which contains the definition of the external objects (constants, types,...);
    -  foo_types.h which contains the definition of the non-external types
    -  foo_body_separate.c which constains the entry points to call the process as a separated C module.
    
  \see{DCplus2bDCplus, bDCplus2sbDCplus, IGraphInputsTrain}
*/
void
CCodeGenerator::separate ()
{

  for_all_graphs_do (separateOneLevel ());
  pK_messenger->setcputime ();
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    This is the entry point of the C code generation in NON-SEPARATED COMPILING mode.
    It is applied to a graph which must be a sbDC+ one. The graph is with or without clusters.
    It generates the following files (for each graph of a process)
    
    -  foo_externals.h which contains the definition of the external objects (constants, types,...);
    -  foo_types.h which contains the definition of the non-external types
    -  then optionally, foo_main.c, foo_body.c, foo_body.h foo_io.c.
    
    @param mainFile (in) at "true", the file  foo_main.c is generated.
    @param bodyFile  (in) at "true", the file  foo_body.c is generated.
    @param ioFile  (in) at "true", the file foo_io.c is generated.

    \see{DCplus2bDCplus, bDCplus2sbDCplus, IGraphInputsTrain}
*/
void
CCodeGenerator::seq (boolean mainFile, boolean bodyFile, boolean ioFile)
{
  for_all_graphs_do (seqOneLevel (mainFile, bodyFile, ioFile, getGraph ()));
  pK_messenger->setcputime ();
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    This is the entry point of the C code generation with tasks.
    It is applied to a graph which must be a sbDC+ one. The graph is with or without clusters.
    It generates the following files (for each graph of a process)
    
    -  foo_externals.h which contains the definition of the external objects (constants, types,...);
    -  foo_types.h which contains the definition of the non-external types
    -  foo_tasks.h which contains the declarations of the tasks
    -  foo_body.c which constains the entry points to call the process as a separated C module.
    
  \see{DCplus2bDCplus, bDCplus2sbDCplus, IGraphInputsTrain}
*/
void
CCodeGenerator::tasks ()
{
  for_all_graphs_do (tasksOneLevel ());
  pK_messenger->setcputime ();
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    This is the entry point of the Distibuted C code generation with tasks.
    It is applied to a graph which must be a Scheduled one. The graph is with clusters.
    It generates the following files (for each graph of a process)
    
    -  foo_externals.h which contains the definition of the external objects (constants, types,...);
    -  foo_types.h which contains the definition of the non-external types
    -  foo_tasks.h which contains the declarations of the tasks
    -  foo_body.c which constains the entry points to call the process as a separated C module.
    -  runConfig.h file defining codification of the processors for communications.
  \see{DCplus2bDCplus, bDCplus2sbDCplus, IGraphInputsTrain}
*/
void
CCodeGenerator::dist ()
{
  for_all_graphs_do (distOneLevel ());
  pK_messenger->setcputime ();
}



_VIRTUAL_ void
CCodeGenerator::generateConfigFileDist (ptlDCG lsOfGraphs)
{
  int i, n, cpt;
  ptcGeneratorDist gen;

  gen = new cGeneratorDist ();
  gen->generateConfigFileDist (lsOfGraphs);
  delete gen;
}
