
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include "CodeGenerator_Macros.c"

/**
   It prints the announce of the code generator and sets the "External Implementation Language" to the current graph.
 */
void
CCodeGenerator::intro ()
{
  ptrepstr v;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  v = string2repstr (cstCGen);
  gw->setImplementationLanguage (STree_mkExternalImplLanguage (v));
  pK_messenger->messagestrIdent (generatorMess, v, gw->getName ());
}

boolean CCodeGenerator::isSeparateModel (ptDCG_sch gw)
{
  return existPredefinedPragma (gw->getPragmas (), PRAGMA_SEPARATE) || existPredefinedPragma (gw->getPragmas (), PRAGMA_UNEXPANDED);
}

/** C Code generation for a model.
   Assume : the graph is a sbDC+ one.
   @param mainFile (in) at pk_true when main (step iterate) file must be generated.
   @param bodyFile (in) at pk_true when body file (definition of a step) must be generated.
   @param ioFile (in) at pk_true when code must be generated for inputs/outputs.
   @param mGraph (in) the main graph for a graph of a module or the current otherwise.

   \ref CodeGenerators "For more details..."
 */
void
CCodeGenerator::seqOneLevel (boolean mainFile, boolean bodyFile, boolean ioFile, ptDCG_sch mGraph)
{
  ptcGenerator gen;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    {
      intro ();
      gen = new cGenerator ();
      if (isSeparateModel (gw))
	gen->generateSeparate (gw);
      else
	gen->generate (gw, mainFile, bodyFile, ioFile, mGraph);
      delete gen;
    }
}

void
CCodeGenerator::separateOneLevel (ptDCG_sch gw)
{
  ptcGenerator gen;

  intro ();
  gen = new cGenerator ();
  gen->generateSeparate (gw);
  delete gen;
}

/** C Code generation for a model for separated compiling.
   Assume : the graph is a sbDC+ one.
   "black boxes" code or "grey boxes" code is produced according to "cluster" status.

   \ref CodeGenerators "For more details..."
 */
void
CCodeGenerator::separateOneLevel ()
{
  ptDCG_sch gw;
  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    separateOneLevel (getCurrentGraph ());
}

/** 
    Generating C tasks (threads) for the current model.
    Assumed: the graph is a sbDC+ one and the clusters are already built.
*/
void
CCodeGenerator::tasksOneLevel ()
{
  ptcGeneratorTasks gen;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    {
      if (isSeparateModel (gw))
	separateOneLevel (gw);
      else
	{
	  intro ();
	  gen = new cGeneratorTasks ();
	  gen->generate (gw);
	  delete gen;
	}
    }
}

/** 
    Generating Distributed C code with tasks for the current model.
    Assumed: the graph is a sbDC+ one and the clusters are already built.
*/
void
CCodeGenerator::distOneLevel ()
{
  ptcGeneratorDist gen;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    {
      if (isSeparateModel (gw))
	separateOneLevel (gw);
      else
	{
	  intro ();
	  gen = new cGeneratorDist ();
	  gen->generate (gw);
	  delete gen;
	}
    }
}
