/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" 
   - if EqualEqualTable[t] is defined for op=opegeg,
   - if LessLessEqualTable[t] is defined for op=opiieg,
   - "false" otherwise.
*/
  boolean cGenerator::isDefinedCmp (TypeType t, noperators op)
{
  if (op == opegeg)
    return (EqualEqualTable.find (t) != EqualEqualTable.end ());
  return (LessLessEqualTable.find (t) != LessLessEqualTable.end ());
}

/**
   @return EqualEqualTable[t] when op=opegeg, LessLessEqualTable[t] otherwise.
*/
ptident cGenerator::getIdCmpTable (TypeType t, noperators op)
{
  if (op == opegeg) return EqualEqualTable[t];
  return LessLessEqualTable[t];
}

/**
   It defines EqualEqualTable[t] (resp.LessLessEqualTable[t])  by id for  op=opegeg (resp. otherwise).
*/
void
cGenerator::addCmpOperator (TypeType t, ptident id, noperators op)
{
  if (op == opegeg) EqualEqualTable[t] = id; else LessLessEqualTable[t] = id;
}

/**
   It adds (addCmpOperator()) the name of the function associated with a type (t), 
   for the comparison operator (op).
   The name is composed of a prefix depending on the comparison operator, suffixed by a string (vid).
   */
_VIRTUAL_ void
cGenerator::genDeclCmpOperatorName (TypeType t, ptrepstr vid, noperators op)
{
  ptident id;
  if (op == opegeg) id = eqeqId; else id = ltlteqId;
  id = createSuffIdent (id, vid);
  addCmpOperator (t, id, op);
}

/**
   @return the name of the comparison function (op: the comparison operator) associated with a type (t).
   This name is set (addCmpOperator()) in the global table.
*/
ptident cGenerator::mkIdentCmp (noperators op, TypeType t)
{
  ptident id;
  // assumed : conversion TypeType/int is possible.
  if (op == opegeg) id = eqeqId; else id = ltlteqId;
  id = makeIdent_ident_string_int (id, "_Type", (int) t);
  addCmpOperator (t, id, op);
  return id;
}

/**
   It generates the comparison functions for an enumeration type (t).

   @param t (in) an enumeration type (assumed).
   @param op (in) the comparison operator (opegeg or opiieg)
*/
_VIRTUAL_ void
cGenerator::genDeclCmpOperatorEnumerationClass (TypeType t, noperators op, TypeType Nt)
{
  genDeclCmpOperatorName (t, string2repstr ("predef"), op);
}

/**
   It generates the comparison functions for an external type (t).

   @param t (in) an extern type (assumed).
   @param op (in) the comparison operator (opegeg or opiieg)
*/
_VIRTUAL_ void
cGenerator::genDeclCmpOperatorExternClass (TypeType t, noperators op)
{
  genDeclCmpOperatorName (t, ExternTypeGetName (t), op);
}


/**
   It generates the comparison functions of the sub-types of a named type (t).
   The same name function is associated with t.

   @param t (in) a named type (assumed).
   @param op (in) the comparison operator (opegeg or opiieg)
*/
_VIRTUAL_ void
cGenerator::genDeclCmpOperatorNamedClass (TypeType t, noperators op)
{
  TypeType tt;

  tt = NamedTypeGetTerm (t);
  genDeclCmpOperator (tt, op, t);
  addCmpOperator (t, getIdCmpTable (tt, op), op);
}


/**
   It generates the comparison functions of the sub-types of a array type (t), 
   before the generation of the code of the function associated with this type.

   @param t (in) an array type (assumed).
   @param op (in) the comparison operator (opegeg or opiieg)
   @param nt (in) the named type of t (if it exists, t otherwise).
*/
_VIRTUAL_ void
cGenerator::genDeclCmpOperatorArrayClass (TypeType t, noperators op, TypeType nt)
{
  TypeType tt;
  ptexpcal vdim;
  ptident id1, id2, id, idtt;
  boolean First;
  int n;

  // Generation of the cmp of the type of the elements
  tt = ArrayMatrixTypeGetTerm (t);
  genDeclCmpOperator (tt, op, tt);
  // Gene of the cmp.
  id = mkIdentCmp (op, t);
  id1 = string2ident ("X");
  id2 = string2ident ("Y");
  vdim = ArrayMatrixTypeGetDimLst (t);
  //Header
  TextForm->printHeaderFunction (id, BooleanType, pk_false);
  TextForm->printLP ();
  TextForm->printtype (nt, pk_true);
  TextForm->printident (id1);
  if (nt == t)
    genDclDimensions (t);
  TextForm->printComma ();
  TextForm->printtype (nt, pk_true);
  TextForm->printident (id2);
  if (nt == t)
    genDclDimensions (t);
  TextForm->printRP ();
  TextForm->printLine ();
  // Body
  TextForm->blockIndentOpening ();
  n = genOpenLoopForDimLst (vdim);
  idtt = getIdCmpTable (tt, op);

  TextForm->printkwif ();
  TextForm->printLP ();
  TextForm->printkwnot ();

  TextForm->printident (idtt);
  TextForm->printLP ();
  TextForm->printident (id1);
  genLindexes ();		// pushed in genOpenLoopForDimLst
  TextForm->printComma ();
  TextForm->printident (id2);
  genLindexes ();		// pushed in genOpenLoopForDimLst
  TextForm->printRP ();
  TextForm->printRP ();
  TextForm->printSpace ();
  TextForm->printreturn (pk_false);

  backIndent ();
  genDeclarationOfIndexes ();
  genCloseIteration (n);
  TextForm->printtruefuncend ();
}

/**
   It generates the comparison functions of the sub-types of a structured type (t), 
   before the generation of the code of the function associated with this type.
   @param t (in) a structured type (assumed).
   @param op (in) the comparison operator (opegeg or opiieg)
   @param nt (in) the named type of t (if it exists, t otherwise).
*/
_VIRTUAL_ void
cGenerator::genDeclCmpOperatorStructureClass (TypeType t, noperators op, TypeType nt)
{
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon;
  NameSubList NSubLst;
  TypeType FieldType;
  NameClass FieldName;

  // Gen. decl of the referenced types in t.
  NL = StructureTypeGetNameLst (t);	// field Name
  NSubLst = NameListCast2SubList (NL);
  fieldsType = StructureTypeGetTypeLst (t);
  NSubLstType = TypeListCast2SubList (fieldsType);
  Goon = pk_true;
  while (Goon)
    {
      FieldType = TypeSubListGetHead (NSubLstType);
      genDeclCmpOperator (FieldType, op, FieldType);
      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail (NSubLst);
	  NSubLstType = TypeSubListGetTail (NSubLstType);
	}
    }
  // Generation of the associated definition.
  ptident id, id1, id2, idcmp;
  boolean First;

  id = mkIdentCmp (op, t);
  id1 = string2ident ("X");
  id2 = string2ident ("Y");
  TextForm->printHeaderCmp (id, nt, id1, id2, pk_true);

  NL = StructureTypeGetNameLst (t);	// field Name
  NSubLst = NameListCast2SubList (NL);
  fieldsType = StructureTypeGetTypeLst (t);
  NSubLstType = TypeListCast2SubList (fieldsType);
  Goon = pk_true;
  First = pk_true;
  TextForm->printkwreturn ();
  TextForm->printLP ();
  while (Goon)
    {
      FieldName = NameSubListGetHead (NSubLst);
      FieldType = TypeSubListGetHead (NSubLstType);
      if (First)
	First = pk_false;
      else
	{
	  TextForm->printSpace ();
	  TextForm->printkwand ();
	  TextForm->printSpace ();
	}
      idcmp = getIdCmpTable (FieldType, op);
      TextForm->printident (idcmp);
      TextForm->printLP ();
      TextForm->printident (id1);
      TextForm->printDot ();
      TextForm->printstr (FieldName);
      TextForm->printComma ();
      TextForm->printident (id2);
      TextForm->printDot ();
      TextForm->printstr (FieldName);
      TextForm->printRP ();
      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail (NSubLst);
	  NSubLstType = TypeSubListGetTail (NSubLstType);
	}
    }
  TextForm->printRPSemiColonLine ();
  TextForm->backindent (1);
  TextForm->printBlockClosingBackIndent ();
  TextForm->printLine ();
}


/** 
    Generation of the code of the expression exp1 op exp2 when op is equal to (==) or smaller/equal to (<<=) operator.
    The code is generated 
    - for predefined type, extern type, enumeration type by genCmpOperator(TypeType, noperators , ptexpcal, ptexpcal)
    - for array type. It is generated as an external call. 
    A library that defines these operations is available in the Polychrony distribution (lib directory).
    Example: pK_ArrayEqint(x, y, sizeOf(T)) is generated for x == y where T is the type of the expression.

    It is not implemented for the other types.
    @return "true" when the code is generated, "false" otherwise.
*/
_VIRTUAL_ void
cGenerator::genDeclCmpOperator (TypeType t, noperators op, TypeType NType)
{
  if (!isDefinedCmp (t, op))
    {
      // if (TypeTestPredef(t)) done, initialisation.
      if (TypeTestEnumerationClass (t))
	genDeclCmpOperatorEnumerationClass (t, op, NType);
      else if (TypeTestExternClass (t))
	genDeclCmpOperatorExternClass (t, op);
      else if (TypeTestNamedClass (t))
	genDeclCmpOperatorNamedClass (t, op);
      else if (TypeTestArrayClass (t))
	genDeclCmpOperatorArrayClass (t, op, NType);
      else if (TypeTestStructureClass (t))
	genDeclCmpOperatorStructureClass (t, op, NType);
      // else if (TypeTestMetaClass(t)) genDeclCmpOperatorMetaClass(t, op);
      else
	{
	  TextForm->printnotimplemented ();
	  setSomeProblem ();
	}
    }
}

/**
   It initialises the global tables (EqualEqualTable, LessLessEqualTable)  of the names of 
   the functions associated with the predefined types.
*/
_VIRTUAL_ void
cGenerator::genCmpOperatorinit ()
{
  ptident id;

  eqeqId = string2ident ("pK_eqeq");	// for == operator
  ltlteqId = string2ident ("pK_ltlteq");	// for <<= operator

  id = newIdent_stringSuffixed (eqeqId, "predef");
  EqualEqualTable[BooleanType] = id;
  EqualEqualTable[PureType] = id;	// no occ.
  EqualEqualTable[IntegerType] = id;
  EqualEqualTable[LongType] = id;
  EqualEqualTable[ShortType] = id;
  EqualEqualTable[RealType] = id;
  EqualEqualTable[DRealType] = id;
  EqualEqualTable[CharType] = id;

  EqualEqualTable[StringType] = string2ident ("pK_StringEQ");
  EqualEqualTable[ComplexType] = string2ident ("pK_complexEq");
  EqualEqualTable[DComplexType] = string2ident ("pK_complexdEq");

  id = newIdent_stringSuffixed (ltlteqId, "predef");
  LessLessEqualTable[BooleanType] = id;
  LessLessEqualTable[PureType] = id;
  LessLessEqualTable[IntegerType] = id;
  LessLessEqualTable[LongType] = id;
  LessLessEqualTable[ShortType] = id;
  LessLessEqualTable[RealType] = id;
  LessLessEqualTable[DRealType] = id;
  LessLessEqualTable[CharType] = id;

  LessLessEqualTable[StringType] = string2ident ("pK_StringLTE");
  LessLessEqualTable[ComplexType] = string2ident ("pK_complexLEq");
  LessLessEqualTable[DComplexType] = string2ident ("pK_complexdLEq");
}

/**
   == and <<= operators.
   @param lexp list of expressions with == or <<= as operator.
   @param op (in) the operator of the element of lexp assumed to opegeg or opiieg.

   For such an expression E1 op E2, a function is generated for the type of E1.
   \see genDeclCmpOperator().
*/
_VIRTUAL_ void
cGenerator::genCmpList (ptlinstances lexp, noperators op)
{
  ptexpcal exp, exp1;
  TypeType t;
  while (ptlinstances_get_next (&lexp, &exp))
    {
      exp1 = EM->getBinaryArgstAt (exp, 1);
      t = EM->getType (exp1);
      genDeclCmpOperator (t, op, t);
    }
}

/**
   Generating the code for the == and <<= operators.
   For each type referenced in these operators, a function is generated.
   After the initial step (genCmpOperatorinit()), it calls genCmpList() for the list of == expressions, and <<= expressions.

   \code
   for the following SIGNAL type, T := struct(integer aa,bb);
   it generates for ==, the following function:

   static logical pK_eqeq_Type_4(T X, T Y)
   {
        return(pK_eqeq_predef(X.aa,Y.aa) && pK_eqeq_predef(X.bb,Y.bb));
   }

   \endcode
*/
_VIRTUAL_ void
cGenerator::genCmpOperators ()
{
  ptlinstances lexp1, lexp2;
  boolean b1, b2;

  b1 = EM->getExprList (opegeg, &lexp1);
  b2 = EM->getExprList (opiieg, &lexp2);
  if (b1 || b2)
    {
      genCmpOperatorinit ();
      if (b1) genCmpList (lexp1, opegeg);
      if (b2) genCmpList (lexp2, opiieg);
    }
}
