/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor.  */
  cGeneratorDist::cGeneratorDist ()
{
}

/** Destructor */
_VIRTUAL_ cGeneratorDist::~cGeneratorDist ()
{
}

/**
   It generates POSIX threads using semaphores (if the clusters are available).
   A thread is associated with
   - each cluster of the application 
   - each input/output signal of the application,
   - the "iterate" entry point (scheduler).
   
   The code is generated in the same file than the classical code generation. 
   A new file is generared, FOO_tasks.h that contains the headers of the tasks.

   @return "true" if the operation succeeds, "false" otherwise: graph without clusters,...

   \note Keep in mind that the dependences are inverted for the delays.
   \see initGenCodeTasks()
*/
_VIRTUAL_
  boolean
cGeneratorDist::generate (ptDCG_sch vGraph)
{
  boolean b;

  simpleComm = pk_true;		// Will be generalized for fifo(m) m>1 (L.besnard)
  b = cGeneratorTasks::generate (vGraph);
  //  if (b) genComm();
  return b;
}


_VIRTUAL_ void
cGeneratorDist::generateConfigFileDist (ptlDCG lsOfGraphs)
{
  ptDCG localGraph;
  int cpt;

  TextForm = new inteprintc ();
  TextForm->openConfigFile ();

  cpt = 0;

  while (ptlPointer_get_elemNext (&lsOfGraphs, (pointer *) & localGraph))
    {
      if (existPredefinedPragma (localGraph->getPragmas (), PRAGMA_RUN_ON))
	{
	  TextForm->printDirectiveDefined (ident2repstr (localGraph->getName ()), pk_false);
	  TextForm->printSpace ();
	  TextForm->printint (cpt);
	  TextForm->printLine ();

	  cpt++;
	}
    }
  TextForm->closeConfigFile ();
}
