/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   It generates the declaration of the entry points: foo_initialize, foo_STEP_initialize, foo_STEP_finalize,
   where foo is the name of the current graph.
 */
  _VIRTUAL_ void
cGenerator::genHeaderInitialize ()
{
  TextForm->printheaderinitialize (progName, SEPARATE);
  TextForm->printheaderentrypointstepinitialize (progName, SEPARATE);
  TextForm->printheaderentrypointstepfinalize (progName, SEPARATE);
}

/** 
    It generates the declaration of the header "iterate" entry associated with the name of the current graph.
*/
_VIRTUAL_ void
cGenerator::genHeaderSchedulerIsolate ()
{
  TextForm->printheaderstepcallmodule (progName);
}

/** 
   It generates the declarations of the headers of clusters.
   For each cluster, it calls genEntryPointSubGraphHeader().
*/
_VIRTUAL_ void
cGenerator::genHeaderSubGraphs ()
{
  int i;

  ptINode aSubGraph;

  for (i = 1; i < nbSets; i++)
    {
      aSubGraph = SubGraphTable[i]->aSubGraph;
      genEntryPointSubGraphHeader (aSubGraph, pk_true);
    }
}

/** It generates the header of the scheduler according to the generating mode (separated or not).
    It calls: genHeaderSchedulerIsolate(), genEntryPointDeclarationSchedulerBlackBox().
*/
_VIRTUAL_ void
cGenerator::genHeaderScheduler ()
{
  if (!SEPARATE)
    genHeaderSchedulerIsolate ();
  else if (nbSets == 1)
    genEntryPointDeclarationSchedulerBlackBox (pk_true);
}

/**
   Entry point of the  generation of the headers for the "body" module.
   The headers are generated in the file 
   - foo_body_separate.h (in separated compiling mode), 
   -  foo_body.h otherwise.

   It calls:
   - printdclbodyfileheader()
   - gendeclarationsBody()
   - genHeaderInitialize()
   - genHeaderSubGraphs() 
   - genHeaderScheduler()
   - closefile().
   
*/
_VIRTUAL_ void
cGenerator::genBodyHeader ()
{
  TextForm->printdclbodyfileheader (progName, SEPARATE);	// opening <progName>_body.h
  if (SEPARATE)
    {
      TextForm->printClassBeginDeclaration ();
      gendeclarationsBody ();
      TextForm->genClassEnd ();
    }
  genHeaderInitialize ();
  genHeaderSubGraphs ();
  genHeaderScheduler ();
  TextForm->closefile ();	// closing <progName>_body.h 
}
