/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PROTECTED_
/**
   Setting a prefix (vprefix).
   It will be generated before each reference to a signal.
   The predefined "nilstr" string must be used to disable prefix generation.

   \note Such a prefix is used for the generation of C code in separated compiling mode.
   All the variable declarations are done in a structure (A). So, the reference of a signal (x) in this 
   context will be A->x. 
   It allows to have multi-references to such a model in another program by allocating of 
   several A structure (one by reference, reentrant code).
*/
  _VIRTUAL_ void
cGenerator::setPrefix (ptrepstr vprefix)
{
  prefixString = vprefix;
}

/** @return the current prefix value.
*/
_VIRTUAL_ ptrepstr
cGenerator::getprefix ()
{
  return prefixString;
}

/** 
    It generates the prefix string.
*/
_VIRTUAL_ void
cGenerator::genPrefix ()
{
  TextForm->printstr (prefixString);
}

/**
   It generates the ident id prefixed if the global prefix (prefixString) is set.
   (Useful for separate mode).
*/
_VIRTUAL_ void
cGenerator::printPrefixedIdent (ptident id)
{
  genPrefix ();
  TextForm->printident (id);
}
