/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    @return the number of dimensions of a type (TT)
*/
  int
cGenerator::getNbDimFlattenedType (TypeType TT)
{
  int n = 0;
  
  if (TypeTestNamedClass (TT)) TT = NamedTypeGetTerm (TT);
  while (TypeTestArrayClass (TT))
    {
      n = n + EM->getSize (ArrayMatrixTypeGetDimLst (TT));
      TT = ArrayMatrixTypeGetTerm (TT);
      if (TypeTestNamedClass (TT))
	TT = NamedTypeGetTerm (TT);
    }
  return n;
}

/**
   @return the predefined type (internal representation of the types) for a type identifier (id).
   For an unknown predefined type name, it returns an errored type. 
   It is restricted to the names of the predefined types.
*/
TypeType cGenerator::identType2TypeType (ptrepstr id)
{
  noperators ope;
  ope = identType2SignalType (id);
  if ( ope == opevent) return get_PureType ();
  if ( ope == oplogical) return get_BooleanType ();
  if ( ope == opshort) return get_ShortType ();
  if ( ope == opinteger) return get_IntegerType ();
  if ( ope == oplong) return get_LongType ();
  if ( ope == opreal) return get_RealType ();
  if ( ope == opdpreal) return get_DRealType ();
  if ( ope == opcomplex) return get_ComplexType ();
  if ( ope == opdcomplex) return get_DComplexType ();
  if ( ope == opchar) return get_CharType ();
  if ( ope == opstring) return get_StringType ();
  return get_ErrorType ();
}


/**
   @return "true" if a type (TT) is an element of l marked by m, "false" otherwise.
   TT is in l iff it has already been genrated.
*/
_VIRTUAL_
  boolean
cGenerator::TestTypeHasBeenGenerated (TypeType TT, ptlinstances l, typeMark m)
{
  ptexpcal elem;

  if (TypeTestPredef (TT))
    return pk_true;
  while (ptlinstances_get_next (&l, &elem))
    if (EM->getOperator (elem) == optype)
      if (EM->getExpType (elem) == TT)
	if (EM->isMarked (elem, m))
	  return pk_true;
  return pk_false;
}

/**
   @return "true" if the types referenced by a type (tt) are elements of l marked by m, "false" otherwise.
   The marked elements are already generated.
*/
_VIRTUAL_ boolean
cGenerator::TestSubTypesHaveBeenGenerated (TypeType t, ptlinstances l, typeMark m)
{
  TypeList l1;
  TypeType t1;
  TypeSubList SubLst1;

  if (TypeTestPredef (t))
    return pk_true;
  if (TypeTestArrayClass (t))
    {
      // Verifier que les constantes des ArrayMatrixTypeGetDimLst(t) sont generees
      return TestSubTypesHaveBeenGenerated (ArrayMatrixTypeGetTerm (t), l, m);
    }
  if (TypeTestExternClass (t))
    return TestTypeHasBeenGenerated (t, l, m);
  if (TypeTestEnumerationClass (t))
    return pk_true;
  if (TypeTestMetaClass (t))
    return TestTypeHasBeenGenerated (t, l, m);
  if (TypeTestNamedClass (t))
    return TestSubTypesHaveBeenGenerated (NamedTypeGetTerm (t), l, m);
  if (TypeTestStructureClass (t))
    return TestSubTypesHaveBeenGenerated (StructureTypeGetArgument (t), l, m);
  if (TypeTestBundleClass (t))
    return TestSubTypesHaveBeenGenerated (StructureTypeGetArgument (BundleTypeGetTerm (t)), l, m);
  if (TypeTestTupleClass (t))
    {
      l1 = TupleTypeGetTypeLst (t);
      SubLst1 = TypeListCast2SubList (l1);
      while (pk_true)
	{
	  t1 = TypeSubListGetHead (SubLst1);
	  if (!TestTypeHasBeenGenerated (t1, l, m))
	    return pk_false;
	  if (TypeSubListTestLast (SubLst1))
	    return pk_true;
	  SubLst1 = TypeSubListGetTail (SubLst1);
	}
    }
  return pk_false;
  /*
     if (TypeTestSubrangeClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestBitsClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestLambdaClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestGenericNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestMetaNodeClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestMatrixClass ( t ))  return STree_mkNOTIMPLEMENTED();;
     if (TypeTestTupleClass ( t ))  return STree_mkNOTIMPLEMENTED();
   */

}

/**
   It generates the definition of a type (t), t is a structure type.
*/
_VIRTUAL_ void
cGenerator::GenTypeStructureClass (TypeType t)
{
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon;
  NameSubList NSubLst;

  NL = StructureTypeGetNameLst (t);	// field Name

  NSubLst = NameListCast2SubList (NL);

  fieldsType = StructureTypeGetTypeLst (t);
  NSubLstType = TypeListCast2SubList (fieldsType);
  TextForm->printBeginStructure ();
  Goon = pk_true;
  while (Goon)
    {
      GenTypeDefinition (NameSubListGetHead (NSubLst), TypeSubListGetHead (NSubLstType));
      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail (NSubLst);
	  NSubLstType = TypeSubListGetTail (NSubLstType);
	}
    }
  TextForm->printEndStructure ();
}

/**
   It generates the definition of a base type of a type (t): for an array type, it ignores the dimensions.
   When it is a structure/bundle type, it calls GenTypeStructureClass(), printtype() otherwise.
*/
_VIRTUAL_ void
cGenerator::genTypeBase (TypeType t)
{
  if (TypeTestStructureClass (t))
    GenTypeStructureClass (t);
  else if (TypeTestBundleClass (t))
    GenTypeStructureClass (BundleTypeGetTerm (t));
  else if (!TextForm->printtype (t, pk_true))
    setSomeProblem ();
}


/**
   It generates the definition of a type (T) with id as name.
   It calls 
   - genTypeBase() for the definition of the base-type of T (ignoring the dimensions)
   - genDclDimensions() for the definition of the dimensions.
*/
_VIRTUAL_ void
cGenerator::GenTypeDefinition (NameClass id, TypeType T)
{
  genTypeBase (T);
  TextForm->printstr (id);
  genDclDimensions (T);
  TextForm->printSemiColonLine ();
}

/**
   @return the number of indexes useful to travel a type(t).
*/
_VIRTUAL_ int
cGenerator::computeIndexes (TypeType t)
{
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon;
  NameSubList NSubLst;
  ptexpcal vdim, aDim;
  int m, n;

  n = 0;
  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  if (TypeTestArrayClass (t))
    {
      vdim = ArrayMatrixTypeGetDimLst (t);
      while (EM->getNaryArgs (&vdim, &aDim))
	n++;
      n = n + computeIndexes (ArrayMatrixTypeGetTerm (t));
    }
  else if (TypeTestStructureClass (t))
    {
      NL = StructureTypeGetNameLst (t);	// field Name
      NSubLst = NameListCast2SubList (NL);
      fieldsType = StructureTypeGetTypeLst (t);
      NSubLstType = TypeListCast2SubList (fieldsType);
      Goon = pk_true;
      while (Goon)
	{
	  m = computeIndexes (TypeSubListGetHead (NSubLstType));
	  if (m > n)
	    n = m;
	  if (NameSubListTestLast (NSubLst))
	    Goon = pk_false;
	  else
	    {
	      NSubLst = NameSubListGetTail (NSubLst);
	      NSubLstType = TypeSubListGetTail (NSubLstType);
	    }
	}
    }
  else;				//TypeTestBundleClass(t)
  return n;
}

/**
   It generates the declaration of a type (T) with id as name.
   (See GenTypeDefinition() for the definition of a type).
   
   \code
   Example: for the Signal declaration 
        type T := struct(real aa,bb);
   the following code is generated
   typedef struct
    {
      float aa;
      float bb;
    }T;
   \endcode
*/
_VIRTUAL_ void
cGenerator::GenTypedefDefinition (NameClass id, TypeType T)
{
  TextForm->printLine ();
  TextForm->printkwTypedef ();
  GenTypeDefinition (id, T);
  setMaxIndex (computeIndexes (T));
}

/** 
    Generating of the declaration of a non predefined type (t).
    The following types are implemented:
    - external, 
    - enumeration,
    - meta and named types (in particular structure/bundle type, array type).
*/
_VIRTUAL_ void
cGenerator::genTypeDeclaration (TypeType t)
{
  ptrepstr vstr;
  boolean b;

  b = pk_true;
  if (TypeTestExternClass (t))
    vstr = ExternTypeGetName (t);
  else if (TypeTestMetaClass (t))
    vstr = MetaTypeGetName (t);
  else
    b = pk_false;

  if (b)
    TextForm->printdclexterntype (vstr);
  else if (TypeTestEnumerationClass (t))
    TextForm->printdclenumtype (t);
  else if (TypeTestNamedClass (t))
    GenTypedefDefinition (NamedTypeGetName (t), NamedTypeGetTerm (t));

  // Not occured : TypeTestStructureClass --> nameTyped for a structure type declaration.
  // Not occured : TypeTestBundleClass --> nameTyped for a structure type declaration.
  /*
     if(TypeTestSubrangeClass(t)) return cNOTIMPLEMENTED();
     if(TypeTestBitsClass(t)) return cNOTIMPLEMENTED();
     if(TypeTestLambdaClass(t))  return cNOTIMPLEMENTED();
     if(TypeTestNodeClass(t))  return cNOTIMPLEMENTED();
     if(TypeTestGenericNodeClass(t))  return cNOTIMPLEMENTED();
     if(TypeTestMetaNodeClass(t)) return cNOTIMPLEMENTED();
     if(TypeTestMatrixClass(t))  return cNOTIMPLEMENTED();;
     if(TypeTestTupleClass(t))  return cNOTIMPLEMENTED();
   */
}



/** 
    Generates the declaration of non predefined types and non external types referenced in the current graph.
    A type is generated (genTypeDeclaration()) if all the types it references 
    are generated (TestSubTypesHaveBeenGenerated()). 
    When a type is generated, it is marked.
    
*/
_VIRTUAL_ void
cGenerator::genNonExternalTypes ()
{
  ptlinstances ltypes;
  TypeType TT;
  ptexpcal elem;
  boolean b;
  typeMark aMark;

  if (Types != NILLINSTANCES)
    {
      aMark = getMark ();
      do
	{
	  b = pk_false;
	  ltypes = Types;
	  while (ptlinstances_get_next (&ltypes, &elem))
	    if (!EM->isMarked (elem, aMark))
	      {
		TT = EM->getExpType (elem);
		if (TypeTestExternClass (TT) || TypeTestMetaClass (TT))
		  EM->addMark (elem, aMark);
		else
		  {
		    if (TestSubTypesHaveBeenGenerated (TT, Types, aMark))
		      {
			genTypeDeclaration (TT);
			EM->addMark (elem, aMark);
			b = pk_true;
		      }
		  }
	      }
	}
      while (b);
      EM->resetMark (Types, aMark);
    }
}


/**
   Generating the declaration of types referenced in the graph in the file foo_types.h, excluding the 
   external types (see the note below).
   It calls: 
   - printdcltypefile() opening the file
   - genNonExternalTypes() genrating
   - closefile() closing the file.
 
   \note the external types are generated in a specific file with other "external undefined" objects
   ( foo_externalsUNDEF_LIS.h where foo is the name of the current graph).
*/
_VIRTUAL_ void
cGenerator::dclTypes ()
{
  TextForm->printdcltypefile (progName);	// opening the file, generating standard types in the C directive #ifndef ...
  genNonExternalTypes ();	// generating
  TextForm->printDirectiveEndIf (nilstr);	// closing the C directive #ifndef
  TextForm->closefile ();	// closing the file
}
