/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor.  */
  cGenerator::cGenerator ()
{
  repstrBlackBox = getPredefinedPragma (PRAGMA_BLACK_BOX);
  repstrGreyBox = getPredefinedPragma (PRAGMA_GREY_BOX);
  // idIterBlackBox = string2ident(STRING_ITERATE);
}

/** Destructor */
_VIRTUAL_ cGenerator::~cGenerator ()
{
}

/** 
    @return "true" if the operation succeeds, "false" othewise.
    Entry point of the C code generation in non-separated compiling mode.
    It is applied to the graph vGraph which must be a sbDC+ one. The graph is with or without clusters. 
    It can be a graph of a model of a module (mGraph).
    It generates the following files :
    - foo_externals.h which contains the declaration of the external objects defined in the two following files.      
    - foo_externalsProc.h contains the declaration of the external functions/procedures
     -foo_externalsUNDEF_LIS.h contains the declaration of the undefined objects (external types, external constants)
    - foo_types.h which contains the definition of the non-external types
    - then optionally, foo_main.c, foo_body.c, foo_body.h foo_io.c.
       - the file foo_io.c contains the i/o operations.
       - the file foo_main.c contains : 
         - the initialization of the execution environment defined in the file foo_io.c; 
	 - the loop that calls to the "iterate" entry point defined in the file foo_body.c.
       - the code of the file foo_body.c references i/o operations defined in the file foo_io and 
         contains the definitions of a step.
       - the file foo_body.h is the header file of foo_body.c

    @param vGraph (in) is a sbDC+ graph.
    @param mainFile (in) when pk_true, the foo_main.c is generated.
    @param bodyFile (in) when pk_true, the foo_body.c is generated.
    @param ioFile (in) when pk_true, the foo_io.c is generated.
    @param mGraph (in) is the graph of a module (ie vGraph is a graph of a model of a module represented by mGraph graph).
    
    It calls:
    - inteprintc() it initializes the shaping of a particular language
    - initialize() it initializes the generator
    - addImportsFrom() when vGraph is a internal graph of a module (mGraph)
    - gencode() it generates the different files described above.
    - finalize() it finalizes the code generation.
    
    For the C code generation in separated compiling mode: see generateSeparate().
*/
_VIRTUAL_
  boolean
cGenerator::generate (ptDCG_sch vGraph, boolean mainFile, boolean bodyFile, boolean ioFile, ptDCG_sch mGraph)
{
  TextForm = new inteprintc ();
  initialize (vGraph, pk_false, nilstr);
  if (vGraph != mGraph)
    addImportsFrom (mGraph);
  gencode (mainFile, bodyFile, ioFile, pk_false);
  genSimulatorCode ();
  return finalize ();
}

/**
     @return "true" if the operation succeeds, "false" othewise.

     Entry point of the C code generation in separated compiling mode.
     It is applied to the graph vGraph which must be a sbDC+ one. The graph is with or without clusters.

     It generates the following files :
     - foo_externals.h which contains the declaration of the external objects defined in the two following files.      
     - foo_externalsProc.h contains the declaration of the external functions/procedures
     - foo_externalsUNDEF_LIS.h contains the declaration of the undefined objects (external types, external constants)
     - foo_types.h which contains the definition of the non-external types,
     - foo_body_separate.c which contains the definition of several entries:
           - one for the initialisation of the module;
	   - one by cluster; the signals are defined in such entry points.
     - foo_body_separate.h, the header file of foo_body_separate.c.

     @param  vGraph (in) is a sbDC+ graph.
     
     It calls:
     - inteprintc() it initializes the shaping of a particular language
     - initialize() it initializes the generator
     - genCodeSeparate() it generates the different files described above.
     - finalize() it finalizes the code generation.

     For the C code generation: see generate().
*/
_VIRTUAL_ boolean
cGenerator::generateSeparate (ptDCG_sch vGraph)
{
  TextForm = new inteprintc ();
  initialize (vGraph, pk_true, string2repstr (DATA_AREA_SEPARATE_PREFIX));
  genCodeSeparate ();
  return finalize ();
}
