/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
#define SIMULATOR "_simulator"
  _VIRTUAL_ ptrepstr inteprintc::printTracefilebegin (ptident processName)
{
  ptrepstr
    name,
    Filename;

  makeRepstr_repstr_string (ident2repstr (processName), SIMULATOR, &name);
  Filename = mkNameWithLangsuffix (name);
  openOut_file (Filename, 0);
  printinclude (string2repstr ("simulator_defs.h"), '"', '"');
  cwriextCvoid ();
  printLinedclfunchead (string2ident ("simulator_INIT"));
  printEndCallIterate (inifuncname);
  blockIndentClosing ();

  cwriextCvoid ();
  printLinedclfunchead (string2ident ("simulator_STEP"));
  printEndCallIterate (iterfuncname);
  blockIndentClosing ();

  cwriextCvoid ();
  printLinedclfunchead (string2ident ("simulator_TERM"));
  blockIndentClosing ();
  return name;
}

/**
   It generates the code
   "int instanciationPathTable [nbPaths] [sizePath] = {"
*/
_VIRTUAL_ void
inteprintc::printBeginDeclPathTable (int nbPaths, int sizePath)
{
  printstr (string2repstr ("int instanciationPathTable"));

  printLeftSquareBracket ();
  printint (nbPaths);
  printRightSquareBracket ();

  printLeftSquareBracket ();
  printint (sizePath);
  printRightSquareBracket ();
  printassign ();
  printLBracket ();
  printLine ();
}


_VIRTUAL_ void
inteprintc::printEndDeclPathTable ()
{
  printRBracket ();
  printSemiColonLine ();
}


/**
   It generates the code
   void getValueByIndices(messageRTB *pM)
   {
   int i;
   pM->buffer[0] = 0.0; // default;
   if (getPath(pM, &i)
   {
   switch (i)
	{
*/
_VIRTUAL_ void
inteprintc::printBeginGetValue ()
{
  cwriextCvoid ();
  printident (string2ident ("getValueByIndices"));
  printLP ();
  printstr (string2repstr ("messageRTB *pM"));
  printRP ();
  printLine ();
  blockIndentOpening ();
  printkwIntSpace ();
  printstr (string2repstr ("i;"));
  printLine ();
  printBufferValueEqual (0);
  printstr (string2repstr ("0.0;"));
  printLine ();
  printstr (string2repstr ("if (getIndexPath(pM, &i))"));
  printLine ();
  printstr (string2repstr ("switch (i)"));
  printLine ();
  printLBracket ();
  printLine ();
}

_VIRTUAL_ void
inteprintc::printEndGetValue ()
{
  printstr (string2repstr ("default:;"));
  printLine ();
  printRBracket ();
  printLine ();
  blockIndentClosing ();
}

_VIRTUAL_ void
inteprintc::printBufferValue (int index)
{
  printstr (string2repstr ("pM->buffer"));
  printLeftSquareBracket ();
  printint (index);
  printRightSquareBracket ();
}

_VIRTUAL_ void
inteprintc::printBufferValueEqual (int index)
{
  printBufferValue (index);
  printassign ();
}


_VIRTUAL_ void
inteprintc::printCase (int i)
{
  printstr (string2repstr ("case "));
  printint (i);
  printColon ();
  printLine ();
}

_VIRTUAL_ void
inteprintc::printBreakSemiColonLine ()
{
  printstr (string2repstr ("break"));
  printSemiColon ();
  printLine ();
}


_VIRTUAL_ void
inteprintc::printGetPathFunction (int n, int maxDepthCall)
{
  printHeaderFunction (string2ident ("getIndexPath"), BooleanType, pk_false);
  printLP ();
  printstr (string2repstr ("messageRTB *pM"));
  printComma ();
  printkwIntSpace ();
  printstar ();
  printchar ('i');
  printRP ();
  printLine ();
  blockIndentOpening ();

  printkwIntSpace ();
  printchar ('j');
  printSemiColonLine ();

  printstr (string2repstr ("(*i) = 0;"));
  printLine ();
  printkwWhile ();
  printLP ();
  printstr (string2repstr ("(*i) < "));
  printint (n);
  printRP ();
  printLine ();
  blockIndentOpening ();
  printstr (string2repstr ("j = 0;"));
  printLine ();
  printkwWhile ();
  printLP ();
  printstr (string2repstr ("instanciationPathTable[*i][j] == pM->indices[j]"));
  printRP ();
  printLine ();
  blockIndentOpening ();
  printstr (string2repstr ("if (pM->indices[j] ==0)"));
  printreturn (pk_true);
  printstr (string2repstr ("j++;"));
  printLine ();
  blockIndentClosing ();
  printstr (string2repstr ("(*i)++;"));
  printLine ();
  printblockclosing ();

  printreturn (pk_false);
  blockIndentClosing ();
}


_VIRTUAL_ void
inteprintc::printBeginSetValue ()
{
  // A "partager" avec printBeginGetValue
  cwriextCvoid ();
  printident (string2ident ("setValueByIndices"));
  printLP ();
  printstr (string2repstr ("messageRTB *pM"));
  printRP ();
  printLine ();
  blockIndentOpening ();
  printkwIntSpace ();
  printstr (string2repstr ("i;"));
  printLine ();
  printstr (string2repstr ("if (getIndexPath(pM, &i))"));
  printLine ();
  printstr (string2repstr ("switch (i)"));
  printLine ();
  printLBracket ();
  printLine ();
}


_VIRTUAL_ void
inteprintc::printEndSetValue ()
{
  printEndGetValue ();
}

/**
   It generates the reference to the include of <processName>_simulator.h
*/
_VIRTUAL_ void
inteprintc::printRefSimulator (ptident processName)
{
  ptrepstr name = makeStr_fileName (processName, SIMULATOR, C_SUFFIX);
  printinclude (name, '"', '"');
}
