/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   Generation of the hearder of the global/external procedure or function for C++.
   @return "true" if the header is generated, "false" for cast function.
   @param name (in) name of the procedure/function
   @param asFunc (in) at pk_true for a function, pk_false for a procedure.
   @param inputs (in) input formal parameters.
   @param outputs (in) output formal parameters.
   @param clusterEntry (in) at pk_false, the name of formal parameters are the names of the signals of the inputs/outputs lists,
           otherwise for a signal X of inputs/outputs, the name of the formal parameter is _X_.
   @param declaration (in) at pk_true, the header is completed by the separator of declarations.
*/
  _VIRTUAL_ boolean
  cppGenerator::genEntryPoint (ptident name, ptexpcal inputs, ptexpcal outputs, boolean asFunc, boolean clusterEntry, boolean declaration)
{
  integer n;

  if (TextForm->isCastFunction (name))
    return pk_false;
  functionDeclaration = pk_true;

  ((inteprintcpp *) TextForm)->printEntryPointBegin (name, declaration);

  n = genDclFormalParameters (inputs, (outputs != NILEXPCAL), pk_false, clusterEntry, 0);
  n = genDclFormalParameters (outputs, pk_false, pk_true, clusterEntry, n);

  if (declaration)
    TextForm->printRPSemiColonLine ();
  else
    TextForm->printRPline ();
  functionDeclaration = pk_false;
  return pk_true;
}


/** 
    Without effect for C++. The declarations are defined in the class.
 */
_VIRTUAL_ void
cppGenerator::gendeclarationsBody ()
{
}
