/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
  Entry point of the  generation of the headers for the "body" module.
  The headers are generated in the file foo_body_separate.h in separated mode generating, 
  otherwise in the file foo_body.h.
  It calls 
  - printdclbodyfileheader()
  - printClassBeginDeclaration()
  - cppGendeclarationsBody()
  - genHeaderInitialize()
  - genHeaderSubGraphs()
  - genHeaderScheduler()
  - genClassEnd()
  - closefile()
  \code
  Such method generates the following code for a process named foo
  class foo
  {
    private:
    //  ==> parameters and indexes
    //  ==> input signals
    float a, b, c;
    // ==> output signals
    float x2, x1;
    logical error;
    ....

    public:logical foo_initialize();
    private:void foo_STEP_initialize();
    public:logical foo_STEP_finalize();
    public:logical foo_iterate();
    };
   \endcode
*/
_VIRTUAL_ void
cppGenerator::genBodyHeader ()
{
  TextForm->printdclbodyfileheader (progName, SEPARATE);	/* opening file */
  TextForm->printClassBeginDeclaration ();
  cppGendeclarationsBody ();
  genHeaderInitialize ();
  genHeaderSubGraphs ();
  genHeaderScheduler ();
  TextForm->genClassEnd ();
  TextForm->closefile ();	/* closing file <progName>_body.h */
}
