/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor */
  cppGenerator::cppGenerator ()
{
}

/** Destructor */
_VIRTUAL_ cppGenerator::~cppGenerator ()
{
}

/** 
   @return "true" if the operation succeeds, "false" othewise.

  Entry point of the C++ code generation in non-separated compiling mode.
  It is applied to the graph vGraph which must be a sbDC+ one. The graph is with or without clusters.
  It generates the following files :
  - foo_externals.h which contains the definition of the external objects (constants, types,...);
  - foo_types.h which contains the definition of the non-external types
  - (optionally) foo_main.c, foo_body.c, foo_body.h foo_io.c.
     - The file foo_main contains : 
       - the initialization of the execution environment defined in the file foo_io; 
       - the loop that calls to the "iterate" entry point defined in the file foo_body.
     - The code of the file foo_body references i/o operations defined in the file foo_io 
     and contains the definitions of the signals.

    @param vGraph (in) is a sbDC+ graph.
    @param mainFile (in) when pk_true, the foo_main.c is generated.
    @param bodyFile (in) when pk_true, the foo_body.c is generated.
    @param ioFile (in) when pk_true, the foo_io.c is generated.

    Operations:
    - it creates a C++ shaper (inteprintcpp()).
    - it initializes the generator (cppGenerator::initialize())
    - it calls cppGenerator::gencode()
    - then cppGenerator::finalize()

  \see{generateSeparate}
*/
_VIRTUAL_
  boolean
cppGenerator::generate (ptDCG_sch vGraph, boolean mainFile, boolean bodyFile, boolean ioFile)
{
  // printf("DEBUG entering cppGenerator::generate\n"); fflush(stdout);
  TextForm = new inteprintcpp ();
  initialize (vGraph, pk_false, nilstr);
  gencode (mainFile, bodyFile, ioFile, pk_false);
  // printf("DEBUG exiting cppGenerator::generate is finalize()\n"); fflush(stdout);
  return finalize ();
}

/**
    @return "true" if the operation succeeds, "false" othewise.
  Entry point of the C++ code generation in separated compiling mode.

  It is applied to the graph ?vGraph which must be a sbDC+ one. The graph is with or without clusters.
  It generates the following files :
  - foo_externals.h which contains the definition of the external objects (constants, types,...);
  - foo_types.h which contains the definition of the non-external types,
  - foo_body_separate.c which contains the definition of several entries; the signals are defined in such entry points.
  - foo_body_separate.h which contains the declaration of the entries of the file foo_body_separate.c

  @param  vGraph (in) is a sbDC+ graph (assumed).
  
  Operations:
    - it creates a C++ shaper (inteprintcpp()).
    - it initializes the generator (cppGenerator::initialize())
    - it calls cppGenerator::genCodeSeparate()
    - then cppGenerator::finalize()

  \see{generate}
*/
_VIRTUAL_ boolean
cppGenerator::generateSeparate (ptDCG_sch vGraph)
{
  TextForm = new inteprintcpp ();
  initialize (vGraph, pk_true, nilstr);
  genCodeSeparate ();
  return finalize ();
}
