/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Shaper constructor for C++ language.
*/
  inteprintcpp::inteprintcpp ()
{
}

/**
   Destructor.
*/
_VIRTUAL_ inteprintcpp::~inteprintcpp ()
{
}

/**
   \spe of inteprintc::initialize() for C++ language.
   Moreover, it defines the object associated with the current graph: foo_obj_ is an object of the class foo, where vname=foo.
   This object is returned by getObjectName() method..
*/
_VIRTUAL_ void
inteprintcpp::initialize (ptrepstr vname)
{
  gInitialize (vname);
  makeRepstr_repstr_string (getProcessName (), OBJECT_SUFFIX, &objectName);
  bGenHeader = pk_false;
}

/**
   Specialization of inteprintc::printkwpublic()   for C++ language.
   It generates the key-word "public:"
*/
_VIRTUAL_ void
inteprintcpp::printkwpublic ()
{
  printstr (string2repstr (kwpublic));
}

/**
   It generates the key-word "private:"
*/
_VIRTUAL_ void
inteprintcpp::printkwprivate ()
{
  printstr (string2repstr (kwprivate));
}

/**
   It generates the key-word "protected:"
*/
_VIRTUAL_ void
inteprintcpp::printkwprotected ()
{
  printstr (string2repstr (kwprotected));
}

/**
   Specialization of inteprintc::printkwstatic() for C++ language.
   It generates a empty string.
*/
_VIRTUAL_ void
inteprintcpp::printkwstatic ()
{
  printstrspace (string2repstr (kwstatic));
}

/**
   Specialization of inteprintc::printkwClass()for C++ language.
   It generates the string "class".
*/
_VIRTUAL_ void
inteprintcpp::printkwClass ()
{
  printstrspace (string2repstr (kwclass));
}

/**
   It generates the beginning of the declaration of the class associated with the current model.
   The symbol _INHERIT_foo may be defined by the user when inheritance is useful.
   \code
   It generates, with \b foo is the name of the current model
   class foo  _INHERIT_foo
   {
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printClassBeginDeclaration ()
{
  printkwClass ();
  printstr (getClassName ());
  printInherit (getClassName ());
  printLine ();
  blockIndentOpening ();
  bGenHeader = pk_true;
}

/**
   Specialization of inteprintc::printheaderstepcallmodule() for C++ language.
   It calls printBeginDeclMethodName(), printstepcallmodule();
   \code
   With progName=foo, it generates 
   public: logical foo_iterate();
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printheaderstepcallmodule (ptident progName)
{
  printBeginDeclMethodName (0, pk_true);
  printstepcallmodule (progName);
}

/**
   Specialization of inteprintc::printheaderinitialize()for C++ language.
   The sep parameter is here unused.
   It calls printBeginDeclMethodName(), printinitcallmodule();
   \code
   With progName=foo, it generates 
   public: logical foo_initialize();
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printheaderinitialize (ptident progName, boolean sep)
{
  printBeginDeclMethodName (0, pk_true);
  printinitcallmodule (progName, pk_false);
}

/**
   It generates the declaration of the iterate entry + the opening of the body statement.
   \code 
   For C, with foo the name of the current model, it generates
   logical foo_iterate() {
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printiterfunchead ()
{
  printBeginDeclMethodName (-1, pk_true);
  printLinedclfunchead (iterfuncname);
}

/**
   Specialization of inteprintc::printheaderentrypointstepfinalize() for C++ language.
   The sep parameter is here unused.
   It calls printBeginDeclMethodName(), printLineidemptycall();
   \code
   With progName=foo, it generates 
   public: logical foo_STEP_finalize();
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printheaderentrypointstepfinalize (ptident progName, boolean bsep)
{
  printBeginDeclMethodName (0, pk_true);
  printLineidemptycall (make_stepFinalIdent (progName));
}

/**
   Specialization of inteprintc::printentrypointstepinitialize() for C++ language.
   The sep parameter is here unused.
    \code
    With progName=foo, it generates 
    void foo_STEP_initialize();
    \endcode
*/
_VIRTUAL_ void
inteprintcpp::printentrypointstepinitialize (ptident progName, boolean sep)
{
  printkwvoid ();
  printClassNameSeparator ();
  printLinedclfunchead (make_stepInitIdent (progName));
}

/**
   \code
   It generates the following code with vname=foo and
   - declaration=true
          public: void foo(
   - declaration=false 
   void foo(
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printEntryPointBegin (ptident vname, boolean declaration)
{
  if (declaration)
    printBeginDeclMethodName (0, pk_false);
  else
    printTypeMethodName (pk_false);
  printident (vname);
  printchar ('(');
}

/**
   Specialization of inteprintc::printdclfile()for C++ language.
   It calls printSystemIOIncludes(), printfilevarname().
   \code
   - with id=a, bInput=true, first=true, it generates
      fstream fra

    - with id=x1, bInput=false, first=false, it generates
       , fwx1
  
    \endcode
*/
_VIRTUAL_ void
inteprintcpp::printdclfile (ptident id, boolean bInput, boolean * bfirst)
{
  if (*bfirst)
    {
      printSystemIOIncludes ();
      printkwfile ();
      *bfirst = pk_false;
    }
  else
    {
      printComma ();
      printLine ();
    }
  printfilevarname (id, bInput);
}

/**
   Specialization of inteprintc::printfileclose()for C++ language.
   \code
    For C++, 
    - with nomfich=a, bInput=true, it generates
           fra.close();
    - with nomfich=x1, bInput=false, it generates
           fwx1.fclose();
    \endcode
*/
_VIRTUAL_ void
inteprintcpp::printfileclose (ptident nomfich, boolean bInput)
{
  printfilevarname (nomfich, bInput);
  printSeparatorSelectMethod ();
  printstr (string2repstr (kwfclose));
  printSemiColonLine ();
}

/**
   Specialization of inteprintc::printfileopen()for C++ language.
   \code
   For C++, it generates 
   - with nomfich=a, bInput=true
      fra.open("Ra.dat",ios::in);
      if (!fra)
        {
         cerr << "Can not open file :" << "Ra.dat" << endl ;
         exit(1);
        }

   - with nomfich=x1, bInput=false (it is an output.)
      fwx1.open("Wx1.dat",ios::out);
      if (!fwx1)
        {
         cerr << "Can not open file :" << "Wx1.dat" << endl ;
         exit(1);
        }
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printfileopen (ptident nomfich, boolean bInput)
{
  ptrepstr openMode;

  printfilevarname (nomfich, bInput);
  printSeparatorSelectMethod ();
  printstr (string2repstr (kwfopen));
  printchar ('(');
  printsysfilename (nomfich, bInput);
  printchar (',');
  if (bInput)
    openMode = string2repstr (kwOpenIn);
  else
    openMode = string2repstr (kwOpenOut);
  printstr (openMode);
  printRPSemiColonLine ();

  printkwif ();
  buf->writeChar ('(');
  printkwnot ();
  printfilevarname (nomfich, bInput);
  printRPlineindent ();
  printblockopening ();
  printstrspace (string2repstr (kwstderr));
  cppPrintWriteOperator ();
  printstr (string2repstr (strcannotopen));
  cppPrintWriteOperator ();
  printsysfilename (nomfich, bInput);
  cppPrintWriteOperator ();
  printstr (string2repstr (strendl));
  printSemiColonLine ();
  printstr (string2repstr (strexit1));
  printSemiColonLine ();
  printBlockClosingBackIndent ();
  printLine ();
}

/**
   Specialization of inteprintc::printbodyreadhead()for C++ language.
   Here, the parameter idt is unused. It calls printfilevarname(), cppPrintReadOperator().
   \code
   For C++, with idarg=a, it generates the folowing sequence
   fra >> 
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printbodyreadhead (ptident idarg, TypeType idt)
{
  // printblockopening();
  printfilevarname (idarg, pk_true);
  cppPrintReadOperator ();
}

/**
   Specialization of inteprintc::printdclreadtail() for C++ language.
   \code
   For C++, with idarg=a, it generates the folowing sequence
   if (fra.eof()) return FALSE;
   return TRUE;
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printdclreadtail (ptident idarg)
{
  printSemiColonLine ();

  printkwif ();
  printSpace ();
  printLP ();
  printfilevarname (idarg, pk_true);
  printSeparatorSelectMethod ();
  printstr (string2repstr (kweof));
  printRP ();
  printreturn (pk_false);
}



/**
   Specialization of inteprintc::printbodywritehead()for C++ language.
    Here, the parameter idt is unused. It calls printfilevarname(), cppPrintWriteOperator().
    \code 
    For C, it generates with id=error,
    ferror <<
    \endcode 
*/
_VIRTUAL_ void
inteprintcpp::printbodywritehead (ptident id, TypeType idt)
{
  printfilevarname (id, pk_false);
  cppPrintWriteOperator ();
}


/**
   Specialization of inteprintc::printbodywritetail()for C++ language.
   Here, the parameter id is unused.
   \code
   For C++, it generates
   << endl;
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printbodywritetail (ptident id)
{
  printSemiColonLine ();
}


_VIRTUAL_ void
inteprintcpp::printbodywritefinal (ptident ids)
{
  printfilevarname (ids, pk_false);
  cppPrintWriteOperator ();
  printstr (string2repstr (kwendline));
  printSemiColonLine ();
}

/**
   Specialization of inteprintc::printstaticfunction() for C++ language.
   Here, the parameter bsep is unused.
   \code
   It generates 
   - when bdef=true, it generates with name=foo_Cluster_2
        private: void foo_Cluster_2()
   - when bdef=false, it generates with name= foo_Cluster_2
        void foo_Cluster_2()
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printstaticfunction (ptident name, boolean bdef, boolean bsep)
{
  if (bdef)
    printTypeMethodName (pk_false);
  else
    printBeginDeclMethodName (1, pk_false);
  printident (name);
  printchar ('(');
  printchar (')');
  if (bdef)
    printLine ();
}

/**
   Specialization of inteprintc::genClassEnd()for C++ language.
   It calls printBlockClosingBackIndent(), printEndDeclaration().
   \code
   It generates the sequence
   } ;
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::genClassEnd ()
{
  printBlockClosingBackIndent ();
  printEndDeclaration ();
  bGenHeader = pk_false;
}

/**
   Specialization of inteprintc::printdclinitializefunchead()for C++ language.
   The parameter bsep is unused here. It calls printdcltypefunchead().
   \code
   logical foo::foo_initialize() {
   \endcode   
*/
_VIRTUAL_ void
inteprintcpp::printdclinitializefunchead (boolean bsep)
{
  printdcltypefunchead (inifuncname, BooleanType);
}

/**
   Specialization of inteprintc::printentrypointstepfinalize().for C++ language.
   The parameter bsep is unused here.
   \code
   logical foo::foo_STEP_finalize()
   \endcode   
*/
_VIRTUAL_ void
inteprintcpp::printentrypointstepfinalize (ptident graphName, boolean bsep)
{
  ptident id;

  id = make_stepFinalIdent (graphName);
  printdcltypefunchead (id, BooleanType);
}

/**
   It prints idpar->field if field is defined, idpar otherwise.
*/
_VIRTUAL_ void
inteprintcpp::printReadRefIdField (ptident idpar, ptrepstr field, boolean bref, boolean inCall)
{
  if (field != nilstr)
    {
      if (inCall)
	printkwref ();
      printLP ();
      printident (idpar);
      printObjectReference ();
      printstr (field);
      printRP ();
    }
  else
    {
      if (inCall || !bref)
	{
	  if (!inCall)
	    printstar ();
	  if (bref)
	    printkwref ();
	}
      printident (idpar);
    }
}

/**
   It prints the header of the declaration of an I/O operation
   --- TO BE COMPLETED---
*/
_VIRTUAL_ void
inteprintcpp::printHeaderFunctionIOType (TypeType t, ptident fname, ptident id, boolean binput)
{
  printHeaderFunction (fname, (binput ? BooleanType : UndefinedType), pk_false);
  printLP ();
  printkwfile ();
  printSpace ();
  // printstar();
  printkwref ();
  printfilevarname (id, binput);
  printComma ();
  printtype (t, pk_true);
  printidentstar (id, binput);
  printRP ();
  printLine ();
}


_VIRTUAL_ void
inteprintcpp::printkwfile ()
{
  printstrspace (string2repstr (kwfile));
}


_VIRTUAL_ void
inteprintcpp::printValueSeparator ()
{
  cppPrintWriteOperator ();
  printchar ('"');
  printchar (' ');
  printchar ('"');
}
