/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define codeReadConst  'r'
#define codeWriteConst 'w'
#define OBJECT_SUFFIX "_obj_"
#define kwstatic        ""
#define kwprivate       "private:"
#define kwpublic        "public:"
#define kwprotected     "protected:"
#define kwclass         "class "
#define kwfile          "fstream"
#define kwpKio          "pKio.h"
#define strcannotopen   "\"Can not open file :\""
#define strexit1        "exit(1)"
#define kwstderr        "cerr "
#define strendl         "endl "
#define kwfopen         "open"
#define kwfclose        "close()"
#define kweof           "eof()"
#define kwOpenIn        "ios::in"
#define kwOpenOut       "ios::out"
#define kwendline       "endl"
#define strstdio        "iostream"
/**
   @return the name of the object associated with the current graph, initialized in inteprintcpp::initialize().
*/
  _VIRTUAL_ ptrepstr inteprintcpp::getObjectName ()
{
  return objectName;
}

/**
   This method generates the declarations of the objects of the main program.
*/
_VIRTUAL_ void
inteprintcpp::printMainObjectDeclarations ()
{
  printstrspace (getClassName ());
  printstr (getObjectName ());
  printEndDeclaration ();
  printLine ();
}

/**
   This method generates the declarations of the objects of the main program.
*/
_VIRTUAL_ void
inteprintcpp::printMainDeclaration ()
{
  inteprintc::printMainDeclaration ();
  printMainObjectDeclarations ();
}

/**
   Specialization of inteprintc::printMainDeclaration().
   \code
   It generates, for foo the name of the current graph:
   logical code;
   foo foo_obj_;
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printRefObject ()
{
  printstr (getObjectName ());
  printSeparatorSelectMethod ();
}

/**
   Specialization of inteprintc::printEndCallIterate()
   It generates a call to name without any parameters (printLineidemptycall()).
   \code
   With name=foo, it generates
   foo_obj_.foo_iterate();
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printEndCallIterate (ptident name)
{
  printRefObject ();
  printLineidemptycall (name);
}


/**
   It generates the string "::".
*/
_VIRTUAL_ void
inteprintcpp::printSepClassNameMethodClass ()
{
  printColon ();
  printColon ();
}

/**
   It generates  thee scope {public, private, protected} depending on the value of vscope (resp. 0, 1,2)
*/
_VIRTUAL_ void
inteprintcpp::printscope (integer vscope)
{
  switch (vscope)
    {
    case 0:
      printkwpublic ();
      break;
    case 1:
      printkwprivate ();
      break;
    case 2:
      printkwprotected ();
      break;
    }
}

/**
   It generates the name of the current class (name of the graph) followed by "::".
*/
_VIRTUAL_ void
inteprintcpp::printClassNameSeparator ()
{
  printstr (getClassName ());
  printSepClassNameMethodClass ();
}

/**
   It generates the boolean type when boolResult, or void type otherwise, followed by 
   printClassNameSeparator() when the generator is not in the declaration part of the class associated 
   with the current graph.
*/
_VIRTUAL_ void
inteprintcpp::printTypeMethodName (boolean boolResult)
{
  if (boolResult)
    printkwlogical ();
  else
    printkwvoid ();
  printSpace ();
  if (!bGenHeader)
    printClassNameSeparator ();
}

/**
   It generates the beginning of the declaration of a method, characterized by its scope (vscope) and 
   its returned type (boolResult, boolean or void).
   It calls printscope() followed by printTypeMethodName().
*/
_VIRTUAL_ void
inteprintcpp::printBeginDeclMethodName (integer vscope, boolean boolResult)
{
  printscope (vscope);
  printTypeMethodName (boolResult);
}

/**
   Specialization of inteprintc::printdcltypefunchead().
   \code
   Example: with vname=name and t=BooleanType, foo is the name of the current graph, it generates
   logical foo::name() {
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printdcltypefunchead (ptident vname, TypeType t)
{
  printtype (t, pk_true);
  printClassNameSeparator ();
  printLinedclfunchead (vname);
}


/**
   Specialization of inteprintc::printMainFunct().
   \code
   Example: It generates
   int main() {
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printMainFunct ()
{
  printtype (IntegerType, pk_true);
  printLinedclfunchead (mainname);
}


/**
   It generates the C++ wrties operator <<.
*/
_VIRTUAL_ void
inteprintcpp::cppPrintWriteOperator ()
{
  printSpace ();
  printchar ('<');
  printchar ('<');
  printSpace ();
}



/**
   It generates the C++ wrties operator >> .
*/
_VIRTUAL_ void
inteprintcpp::cppPrintReadOperator ()
{
  printSpace ();
  printchar ('>');
  printchar ('>');
  printSpace ();
}

/**
   It generates the separator used to select a method (dot).
*/
_VIRTUAL_ void
inteprintcpp::printSeparatorSelectMethod ()
{
  printchar ('.');
}


/**
   It generates the inclusion of the Polychrony IO library.
   \code
   It generates
   #include "pKio.h"
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printSystemIOIncludes ()
{
  printinclude (string2repstr (kwpKio), '"', '"');
}

/**
    Specialization of inteprintc::printIOIncludefileSystem()
    \code
   It generates
   #include <cstdio>
   #include <iostream>
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printIOIncludefileSystem ()
{
  printinclude (string2repstr ("cstdio"), '<', '>');
  printinclude (string2repstr (strstdio), '<', '>');
}

/**
   Specialization of inteprintc::printStdIncludefileSystem()
   \code
   It generates
   #include <cstdlib>
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printStdIncludefileSystem ()
{
  printinclude (string2repstr ("cstdlib"), '<', '>');
}

/**
   Specialization of inteprintc::printPredefinedMathInclude()
   \code
   It generates
   #include <cmath>
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::printPredefinedMathInclude ()
{
  printinclude (string2repstr ("cmath"), '<', '>');
}


/**
   Redefinition of inteprintc::cgenCallSeparate().
   \code
   It generates with id=foo
   foo();
   \endcode
*/
_VIRTUAL_ void
inteprintcpp::cgenCallSeparate (ptident id)
{
  printidemptycall (id);	// printident(id); printchar('('); printchar(')');
  printchar (';');
  printLine ();
}

/*
  It generates the reference to the symbol _INHERIT_<className>
*/
_VIRTUAL_ void
inteprintcpp::printInherit (ptrepstr className)
{
  printSpace ();
  printstr (string2repstr ("_INHERIT_"));
  printstr (className);
}
