

/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


#include "CodeGenerator_Macros.c"

void
javaCodeGenerator::intro ()
{
  ptrepstr v;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  v = string2repstr (cstJavaGen);
  gw->setImplementationLanguage (STree_mkExternalImplLanguage (v));
  pK_messenger->messagestrIdent (generatorMess, v, gw->getName ());
}

//----------------------------------------------------------------------------------------
//PLG: IGraph2JavaOneLevel is a DCG_sch function 

/** Java Code generation for a model.
   Assume : the graph is a sbDC+ one.
   @param mainFile (in) at pk_true when main (step iterate) file must be generated.
   @param bodyFile (in) at pk_true when body file (definition of a step) must be generated.
   @param ioFile (in) at pk_true when code must be generated for inputs/outputs.

   \ref CodeGenerators "For more details..."
 */
void
javaCodeGenerator::seqOneLevel (boolean mainFile, boolean bodyFile, boolean ioFile)
{
  ptjavaGenerator gen;
  boolean bsep;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    {
      intro ();
      bsep = existPredefinedPragma (gw->getPragmas (), PRAGMA_SEPARATE) || existPredefinedPragma (gw->getPragmas (), PRAGMA_UNEXPANDED);	// WHY??
      gen = new javaGenerator ();
      if (bsep)
	gen->generateSeparate (gw);
      else
	gen->generate (gw, mainFile, bodyFile, ioFile);
      delete gen;
    }
}

//----------------------------------------------------------------------------------------
//PLG: IGraph2JavaSeparateOneLevel is a DCG_sch function 
void
javaCodeGenerator::separateOneLevel ()
{
  ptjavaGenerator gen;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    {
      intro ();
      gen = new javaGenerator ();
      gen->generateSeparate (gw);
      delete gen;
    }
}

//----------------------------------------------------------------------------------------
//PLG: WHY this assumption ?
//PLG: clusters are DCG_sch objects

/** Threads Java Code generation for a model.
   A thread id generated for each cluster.
   Assume : the graph is a sbDC+ one and "clusters" have been produced. 
   @param mainFile (in) at pk_true when main (step iterate) file must be generated.
   @param bodyFile (in) at pk_true when body file (definition of a step) must be generated.
   @param ioFile (in) at pk_true when code must be generated for inputs/outputs.

   \ref CodeGenerators "For more details..."
 */
void
javaCodeGenerator::tasksOneLevel (boolean mainFile, boolean bodyFile, boolean ioFile)
{
  ptthreadsJavaGenerator gen;
  ptrepstr v;
  ptDCG_sch gw;

  gw = getCurrentGraph ();
  if (!gw->isGraphOfExternalProcess ())
    {
      intro ();
      gen = new threadsJavaGenerator ();
      gen->generate (gw, mainFile, bodyFile, ioFile);
      delete gen;
    }
}


// ==================================================================================================
