/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Shaper constructor for Java language.
*/
  inteprintjava::inteprintjava ()
{
}

/**
   Destructor.
*/
_VIRTUAL_ inteprintjava::~inteprintjava ()
{
}

/**
   \spe of inteprintcpp::initialize() for Java language.
*/
_VIRTUAL_ void
inteprintjava::initialize (ptrepstr nom)
{
  gInitialize (nom);
  FILE_SUFFIX = string2repstr (".java");
  HEADER_SUFFIX = FILE_SUFFIX;
}


/**
   \spe of inteprintcpp::printmainfile() for Java language.
   \code
   With foo the name of the current graph, it generates
   public class foo_main{
    public static void main(String[] args)  throws Exception {
        boolean code ;
        foo foo_obj_=new foo();
        foo_obj_.openIO();
        code =foo_obj_.foo_initialize();
        while(code)code =foo_obj_.foo_step();
        foo_obj_.closeIO();
	}
    }
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printmainfile (ptident ProcessName)
{
  ptrepstr vmain;

  vmain = getMainFileName (ProcessName);
  openOut_file (makeNameWithLangSuffix (vmain), mess_gen_main);
  printLine ();
  genClassBegin (vmain, 1, pk_false);
  javaPrintdclmainfunc (ProcessName);
  genClassEnd ();
  closefile ();
}


/**
   \spe Without effect.
*/
_VIRTUAL_ void
inteprintjava::printkwref ()
{
}

/**
   \spe It generates  "private"
*/
_VIRTUAL_ void
inteprintjava::printkwstatic ()
{
  buf->printAlfa (PRIVATE, strlen (PRIVATE));
  buf->space ();
}

/**
   It prints the key word interface.
*/
_VIRTUAL_ void
inteprintjava::printkwInterface ()
{
  javaPrintString (kwInterface);
}

/**
   It prints the key word interface.
*/
_VIRTUAL_ void
inteprintjava::printkwOrdinal ()
{
  javaPrintString (kwOrdinal);
}

/** 
    It generates the beginning of the declaration of a class in the file associated with the shaper.
    The scope of the class is given by  codeAccess:
    - 1 for public
    - 2 for private
    - protected, otherwise
    \code
    For vName="foo", codeAccess=1, it generates 
    public class foo
    {
    \endcode
*/
_VIRTUAL_ void
inteprintjava::genClassBegin (ptrepstr vName, integer codeAccess, boolean babstract)
{
  genScope (codeAccess);
  if (babstract)
    printkwInterface ();
  else
    printkwClass ();
  printstr (vName);
  printLBracket ();
  printLine ();
  indent ();
}

/**
   \spe It generates the end of the declaration of a class.
   It signals a back indentation before a block closing assumed previously in printClassBeginDeclaration().
*/
_VIRTUAL_ void
inteprintjava::genClassEnd ()
{
  backindent (1);
  printblockclosing ();
}

/**
   It generates the end of the declaration of a class(genClassEnd()) 
   then, when bCloseFile, it closes the file associated with the shaper.
*/
_VIRTUAL_ void
inteprintjava::genClassEnd (boolean bCloseFile)
{
  printLine ();
  genClassEnd ();
  printLine ();
  if (bCloseFile)
    closefile ();
}

/**
   It generates the string "=new "
*/
_VIRTUAL_ void
inteprintjava::printEqAllocate ()
{
  printequal ();
  javaPrintString (ALLOCATE);
}

/**
   It generates the access to the predefined object for IO.
*/
_VIRTUAL_ void
inteprintjava::printAccessIOObject ()
{
  genIOObjectParameterName ();
  printDot ();
}

/**
   It generates the declarations for IO in the body class.
   \code
   With foo the name of the current graph, it generates
   private foo_io _io_;
   private KeyboardReader KR = new KeyboardReader();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printJavaDeclarationIO ()
{
  javaPrintString (PRIVATE);
  genIOClassName ();
  genIOObjectName ();
  printSemiColon ();
  printLine ();
  genKeyboard (2);
}

_VIRTUAL_ void
inteprintjava::printJavaExternalObjectName ()
{
  printstr (getStringWithProcessTag (EXTERNAL_OBJECT));
}

/**
   It generates the declarations for the external objects in the body class.
   \code
   With foo the name of the current graph, it generates
   private foo_externalsProc foo_eo_;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printJavaDeclarationExternalObject ()
{
  javaPrintString (PRIVATE);
  printstr (getExternalsProcName (progName));
  printSpace ();
  printJavaExternalObjectName ();
  printSemiColon ();
  printLine ();
}

/**
   It generates where foo is the name of the current graph.
   \code
   _io_=new foo_io();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printJavaCreateIOObject ()
{
  genIOObjectName ();
  printEqAllocate ();
  genIOClassName ();
  genLPRPSemiColonLine ();
}

/**
   It generates the code of the IO opening/closing  methods.
   \code
   It generates for foo the name of the current graph.
   public void openIO(){
        _io_.foo_OpenIO();
    }
   public void closeIO(){
        _io_.foo_CloseIO();
    }
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printJavaIOMethods ()
{
  javaPrintString ("public void openIO(){");
  printLine ();
  indent ();
  genIORefObjectName ();
  printstr (progName);
  javaPrintString ("_OpenIO();");
  printLine ();
  backindent (1);
  printRBracket ();
  printLine ();
  printLine ();

  javaPrintString ("public void closeIO(){");
  printLine ();
  indent ();
  genIORefObjectName ();
  printstr (progName);
  javaPrintString ("_CloseIO();");
  printLine ();
  backindent (1);
  printRBracket ();
  printLine ();
  printLine ();
}

/**
   It generates the method that allows to initialize the external object.
   \code
   foo denotes the current name of the application.
   public void setExternalObj(foo_externalsProc obj)
    { 
      foo_eo_ = obj; 
    }
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printJavaSetExternalObject ()
{
  printkwpublic ();
  printkwvoid ();
  javaPrintString (" setExternalObj ");
  printLP ();
  printstr (getExternalsProcName (progName));
  javaPrintString (" obj ");
  printRP ();
  printLBracket ();
  printLine ();
  indent ();
  printJavaExternalObjectName ();
  printequal ();
  javaPrintString (" obj ");
  printSemiColonLine ();
  backindent (1);
  printRBracket ();
  printLine ();
}


/**
   It opens the file to generate the io module(foo_io.java, with processName=foo).
*/
_VIRTUAL_ void
inteprintjava::printdcliofilehead (ptident processName)
{
  openOut_file (makeNameWithLangSuffix (getIOFileName (processName)), mess_gen_io);
  addpKStdImports ();
}

/**
   It generates the beginning of the declaration of the io class.
   \code
   It generates with foo the name of the current graph.
    class foo_io
    {
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintIOClassBeginDeclaration ()
{
  printstrspace (string2repstr (CLASS));
  genIOClassName ();
  printLine ();
  blockIndentOpening ();
}

/**
   It generates the declaration of an input file descriptor associated with id.
   \code
   For id="a", it generates
   private InputReader fra;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaprintdclfileInput (ptident id)
{
  javaPrintString (PRIVATE);
  javaPrintString (IO_INPUT_READER);
  printfilevarname (id, pk_true);
}

/**
   It generates the declaration of an output file descriptor associated with id.
   \code
   For id="x", it generates
   private OutputWriter fwx;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaprintdclfileOutput (ptident id)
{
  javaPrintString (PRIVATE);
  javaPrintString (IO_OUTPUT_WRITER);
  printfilevarname (id, pk_false);
}

/**
   It generates the allocation of the object with id, declared in javaprintdclfileInput().
   \code
   For id="a", it generates
   fra=new InputReader();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::genAllocateFileDescriptorInput (ptident id)
{
  printfilevarname (id, pk_true);
  printEqAllocate ();
  javaPrintString (IO_INPUT_READER);
  genLPRPSemiColonLine ();
}

/**
   It generates the allocation of the object with id, declared in javaprintdclfileOutput().
   \code
   For id="x", it generates
   fwx=new OutputWriter();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::genAllocateFileDescriptorOutput (ptident id)
{
  printfilevarname (id, pk_false);
  printEqAllocate ();
  javaPrintString (IO_OUTPUT_WRITER);
  genLPRPSemiColonLine ();
}

/**
   It generates the opening of a file associated with an  input identifier(id)  when bInput, an output otherwise.
  \code
  - For id="a", an input, it generates
     fra.open("Ra.dat"); 

  - For id="x", an output, it generates
     fwx.open("Wx.dat");
  \endcode
*/
_VIRTUAL_ void
inteprintjava::printfileopen (ptident id, boolean bInput)
{
  printfilevarname (id, bInput);
  printSeparatorSelectMethod ();
  printstr (string2repstr (kwfopen));
  printLP ();
  printsysfilename (id, bInput);
  printRPSemiColonLine ();
}

/**
   \spe of nteprintc::printReadFunctionLP().
   It calls inteprintc::printIOname()
*/
_VIRTUAL_ void
inteprintjava::printReadFunctionLP (ptident id, ptident processName)
{
  inteprintc::printIOname (id, processName, pk_true);
}

/**
   \spe of nteprintc::printWriteFunctionLP().
   It calls inteprintc::printIOname()
*/
_VIRTUAL_ void
inteprintjava::printWriteFunctionLP (ptident id, ptident processName)
{
  inteprintc::printIOname (id, processName, pk_false);
}

/**
   \spe Specialization of inteprintcpp::printbodyreadhead().
   \code
   It generates for a signal idarg="a" of real type(idt)
   fra.readFloat();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printbodyreadhead (ptident idarg, TypeType idt)
{
  printfilevarname (idarg, pk_true);
  javaPrintStringWithoutSpace (IO_READ);
  printformat (idt);
  printLPRP ();
  printSemiColonLine ();
}

/**
   \spe Specialization of inteprintcpp::printdclreadtailRPRP().
   \code
   It generates for a signal idarg="a" 
   if(fra.isEof())return pk_false;
   return pk_true;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printdclreadtail (ptident idarg)
{
  printkwif ();
  printSpace ();
  printLP ();
  printfilevarname (idarg, pk_true);
  printSeparatorSelectMethod ();
  printstr (string2repstr (kweof));
  printRP ();
  printreturn (pk_false);
}

/**
    \spe Specialization of inteprintcpp::printbodywritetail()
*/
_VIRTUAL_ void
inteprintjava::printbodywritetail (integer nbck, ptident id)
{
  printSemiColonLine ();
  printchar ('}');
  printLine ();
  printBlockClosingBackIndent (nbck);
}

/**
    \spe Specialization of inteprintcpp::printbodywritehead()
*/
_VIRTUAL_ void
inteprintjava::printbodywritehead (ptident id, TypeType idt)
{
  printfilevarname (id, pk_false);
  javaPrintString (IO_WRITE);
  printLP ();
}

/**
   It generates the code of  a "new line" in the file associated with an identifier.
   \code
   For id= x, it generates
   fwx.ps.println();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printbodywriteln (ptident id)
{
  printfilevarname (id, pk_false);
  javaPrintString (IO_WRITELN);
  printLPRP ();
  printchar (';');
  printLine ();
  printdclfuncend ();
}

/** 
    It generates the keyword "this".
 */
_VIRTUAL_ void
inteprintjava::printkwthis ()
{
  printstr (string2repstr (kwThis));
}

/**
   It generates the keyword "public".
*/
_VIRTUAL_ void
inteprintjava::printkwpublic ()
{
  printstrspace (string2repstr (PUBLIC));
}


/**
   It generates the keyword "private".
*/
_VIRTUAL_ void
inteprintjava::printkwprivate ()
{
  printstrspace (string2repstr (PRIVATE));
}


/**
   It generates the keyword "protected".
*/
_VIRTUAL_ void
inteprintjava::printkwprotected ()
{
  printstrspace (string2repstr (PROTECTED));
}

/**
   It generates the definition of header of the constructor associated with the IO class of the current graph.
   \code
   For "foo" the name of t he current graph, it generates
   public foo_io(){
   
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printIOConstructorHeader ()
{
  printkwpublic ();
  genIOClassName ();
  printLP ();
  printRP ();
  blockIndentOpening ();
}

/**
   \spe of inteprintc::printdclexternfile
   For  Java, the external references are generated in foo_Externals.java.
*/
_VIRTUAL_ void
inteprintjava::printdclexternfile (ptident procName)
{
  openOut_file (getExternalsFileName (procName), mess_gen_types);
}

/**
   It links the shaper to a file when bOpenFile, then it generates(genClassBegin()) if bclass.
   The beginning of the declaration of a class taking into account the scope(codeAccess).
   
   The associated file name is defined by makeNameWithLangSuffix() applied to vName(foo.java when vName=foo).
*/
_VIRTUAL_ void
inteprintjava::javaPrintClassBegin (ptrepstr vName, integer codeAccess, boolean bOpenFile, boolean bclass)
{
  if (bOpenFile)
    openOut_file (makeNameWithLangSuffix (vName), mess_gen_types);
  if (bclass)
    {
      addpKStdImports ();
      genClassBegin (vName, codeAccess, pk_false);
    }
}


/**
   It generates the specification of a Java constant:
   - "public static final" when bpublic, 
   - "static final" otherwise.
*/
_VIRTUAL_ void
inteprintjava::printConstantSpec (boolean bpublic)
{
  printPublicStatic (bpublic);
  printFinal ();
}

/**
   It generates
   - "public static" when bpublic
   - static otherwise
*/
void
inteprintjava::printPublicStatic (boolean bpublic)
{
  if (bpublic)
    printkwpublic ();
  printStatic ();
}

/**
   It generates "final" keyword.
*/
void
inteprintjava::printFinal ()
{
  javaPrintString ("final");
}

/**
   It generates "static" keyword.
*/
void
inteprintjava::printStatic ()
{
  javaPrintString (STATIC);
}

/**
   It generates the boolean constant "true" when cb, "false" otherwise.
*/
_VIRTUAL_ void
inteprintjava::printkwlogicons (boolean cb)
{
  if (cb)
    buf->printAlfa (kwTrue, lgkwTrue);
  else
    buf->printAlfa (kwFalse, lgkwFalse);
}

/**
   It generates the pow operator("(int)java.lang.Math.pow").
*/
_VIRTUAL_ void
inteprintjava::printPow ()
{
  printstr (string2repstr ("(int)java.lang.Math.pow"));
}

/**
   It generates the run method(" public void run() ").
*/
_VIRTUAL_ void
inteprintjava::printRun ()
{
  printstr (string2repstr (" public void run() "));
}

/**
   Misc. Must be tested.
   It generates id=id;
*/
_VIRTUAL_ void
inteprintjava::printSetVariable (ptident id, ptident className)
{
  // Very strange x=x, it should be x= className.x 
  printident (id);
  printassign ();
  printident (id);
  printSemiColon ();
  printLine ();
}

/**
   It generates the definition of an identifier(velem) by the value i(used to implement the enumerated types).
   \code
   For a value(velem=SOFT) of a enumerated type(rank i=0), it generates
   public static final int SOFT=0;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printEnumValue (ptrepstr velem, integer i)
{
  printConstantSpec (pk_true);
  printkwIntSpace ();
  printstr (velem);
  printequal ();
  printint (i);
  printchar (';');
  printLine ();
}

/**
   It generates the call to the READ operation of an identifier(id) of a process(processName).
   \code
   It generates for id="H" and processName="foo"

   if(!_io_.r_foo_H(this)) return pk_false;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printReadEnv (ptident id, ptident processName)
{
  printkwif ();
  printLP ();
  printkwnot ();
  genIOObjectName ();
  printDot ();
  printReadFunctionLP (id, processName);
  printkwthis ();
  printRP ();
  printRP ();
  printSpace ();
  printreturn (pk_false);
}

/**
   It generates the call to the WRITE operation of an identifier(id) of a process(processName).
   \code
   It generates for id="x" and processName="foo"
   _io_.w_foo_x(this);
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printWriteEnv (ptident id, ptident processName)
{
  genIOObjectName ();
  printDot ();
  printWriteFunctionLP (id, processName);
  printkwthis ();
  printRPSemiColonLine ();
}

/**
   It generates the header of a READ operation of an identifier(id) of a process(processName).
   \code
   For processName=foo, id=a,  
   - bdef=true, it generates
         public void r_foo_a(foo _bobj) { 
   - bdef=false, it generates:
         public void r_foo_a(foo _bobj) ;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printReadHeaderExterncall (ptident id, ptident processName, boolean bdef)
{
  printDclEnvReadFuncBegin (id, processName);
  printidentspace (processName);
  genIOObjectParameterName ();
  if (bdef)
    printRPline ();
  else
    printRPSemiColonLine ();
}


/**
   It generates the header of a WRITE operation of an identifier(id) of a process(processName).
   \code
   For processName=foo, id=x,  
   - bdef=true, it generates
           public void w_foo_x(foo _bobj) { 
   - bdef=false, it generates:
           public void w_foo_x(foo _bobj) ;
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printWriteHeaderExterncall (ptident id, ptident processName, boolean bdef)
{
  printDclEnvWriteFuncBegin (id, processName);
  printidentspace (processName);
  genIOObjectParameterName ();
  if (bdef)
    printRPline ();
  else
    printRPSemiColonLine ();
}



/**
   It generates the declaration of the class of the external references.
*/
_VIRTUAL_ void
inteprintjava::printdclexternfileProcBegin (ptident progName)
{
  genClassBegin (getExternalsProcName (ident2repstr (progName)), 1, pk_true);
}


/**
   It finalizes the generation of the file that contains the external references of an application.
   After this call, the file closed.
   This file is previously opened by the method printdclexternfileProc().

   @param progName (in) the name of the application.
*/
_VIRTUAL_ void
inteprintjava::printdclexternfileProcEnd (ptident progName)
{
  genClassEnd ();
  inteprintc::printdclexternfileProcEnd (progName);
}


/**
   It prints the declaration of an assumed enumerated type (t).
   From Java5.0, it is similar to C.
*/
_VIRTUAL_ void
inteprintjava::printdclenumtype (TypeType t)
{
  printkwEnum ();
  printSpace ();
  cwristr (EnumerationTypeGetName (t));
  printListEnum (t);
  printreleasedcl ();
}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::genIOOpeningVCDFile ()
{

}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::genIOClosingVCDFile ()
{
}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::printVCDDeclarations ()
{

}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::printstepiobeginbody ()
{
}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::printStrVCDfile (ptident vcdfile, ptrepstr vstr, boolean beol)
{
}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::printdclVCDvar (TypeType t, ptident id, ptident vcdname)
{
}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::printVCDValue (TypeType t, ptident ids, ptident vcdname, boolean bread)
{
}

/**
   The VCD file is not managed in Java.
*/
_VIRTUAL_ void
inteprintjava::printVCDHeaderEnd ()
{
}

/**
   It generates the Signal standard imports.
   See the definitions of pkStdJavaLib and libTable
*/
_VIRTUAL_ void
inteprintjava::addpKStdImports ()
{
  ptrepstr vstr;
  integer i;

  vstr = string2repstr (pkStdJavaLib);
  for (int i = 0; i < 7; i++)
    {
      printstr (vstr);
      printstr (string2repstr (libTable[i]));
      printSemiColonLine ();
    }
  printLine ();
}
