/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define kwSignalStringType "String "
#define kwThis  "this"
#define kwfopen "open"
#define kweof   "isEof()"
#define kwInterface "interface"
#define kwOrdinal "ordinal"
#define kwTrue "true"
#define lgkwTrue 4
#define kwFalse "false"
#define lgkwFalse 5
#define STR_SEPARATE     "_separate"
#define IO_INPUT_READER  "InputReader "
#define IO_OUTPUT_WRITER "OutputWriter "
#define IO_WRITE         ".ps.print"
#define IO_WRITELN       ".ps.println"
#define IO_READ          ".read"
#define  PUBLIC    "public"
#define  PRIVATE   "private"
#define  PROTECTED "protected"
#define  STATIC    "static"
#define  VOID     "void"
#define  INT      "int"
#define  CHAR     "char"
#define  FLOAT    "float"
#define  DOUBLE   "double"
#define  BOOLEAN  "boolean"
#define  CLASS  "class"
#define  IF       "if"
#define  ELSE     "else"
#define  WHILE    "while"
#define  FOR      "for"
#define  BREAK    "break"
#define  CASE     "case"
#define  SWITCH   "switch"
#define  ALLOCATE "new "
#define  MAIN_INITIALIZE "_initialize"
#define  MAIN_OBJECT     "_obj_"
#define  MAIN_OPEN_IO    "openIO"
#define  MAIN_CLOSE_IO   "closeIO"
#define  MAIN_CODE       "code"
#define  _MAIN           "_main"
#define  MAIN            "main"
#define  BODY_STATIC_IO  "_io_"
#define  OBJECT_PARAMETER_FOR_IO "_bobj"
#define  EXTERNAL_OBJECT "_eo_"
#define  BODY_ITERATE "_step"
#define  JAVA_BODY_SUFFIX  ""
#define  JAVA_SUFFIX   ".java"
#define  STR_SEPARATE  "_separate"
#define  IO_SUFFIX    "_io"
#define KEYBOARD_INSTANCE  "KeyboardReader KR = new KeyboardReader();"
#define EXTERNAL_SUFFIX "_Externals"
#define pkStdJavaLib "import SignalLibraries.Java."
static const char *libTable[7] = { "InputReader", "KeyboardReader", "OutputWriter", "pK_complex", "pK_complexd", "pK_Obj", "pK_String" };



_VIRTUAL_ char *
inteprintjava::getHeaderSuffix ()
{
  return (char *) JAVA_SUFFIX;
}

/** It generates the C-string s. */
_VIRTUAL_ void
inteprintjava::javaPrintStringWithoutSpace (const char *s)
{
  int l = strlen (s);
  for (int i = 0; i < l; i++)
    if (s[i] == '\n')
      printLine ();
    else
      printchar (s[i]);
}

/** It generates the C-string s (javaPrintStringWithoutSpace()) followed by a space. */
_VIRTUAL_ void
inteprintjava::javaPrintString (const char *s)
{
  javaPrintStringWithoutSpace (s);
  printSpace ();
}

/** @return the name suffixed with the Java file suffix.*/
_VIRTUAL_ ptrepstr
inteprintjava::makeNameWithLangSuffix (ptrepstr name)
{
  return makeStr_fileName (name, JAVA_SUFFIX);
}

/**
   \code
   It generates
   public static void main (String[] args)  throws Exception {
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintMainFunct ()
{
  javaPrintString (PUBLIC);
  javaPrintString (STATIC);
  javaPrintString (VOID);
  javaPrintString (MAIN);
  javaPrintString ("(String[] args) ");
  printException ();
  printLBracket ();
  indent ();
}

/**
   @return the Polychrony-string obtained by the concatenation of the current graph name and vstring. 
*/
_VIRTUAL_ ptrepstr
inteprintjava::getStringWithProcessTag (const char *vstring)
{
  ptrepstr vstr;
  if (makeRepstr_repstr_string (progName, vstring, &vstr))
    return vstr;
  return progName;
}

/**
   @return the main class associated with the current graph.
*/
_VIRTUAL_ ptrepstr
inteprintjava::getMainClassName ()
{
  return getStringWithProcessTag (_MAIN);
}

/**
   \code 
   For progName=foo, it generates
        boolean code ;
        foo foo_obj_=new foo();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintMainDeclaration (ptident progName)
{
  javaPrintString (BOOLEAN);
  javaPrintString (MAIN_CODE);
  printSemiColon ();
  printLine ();
  printidentspace (progName);
  printstr (getStringWithProcessTag (MAIN_OBJECT));
  printEqAllocate ();
  printident (progName);
  genLPRPSemiColonLine ();
}

/**
   \code
   For foo the name of the graph, it generates
        foo_obj_.openIO ();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintOpenIO ()
{
  printstr (getStringWithProcessTag (MAIN_OBJECT));
  printDot ();
  javaPrintString (MAIN_OPEN_IO);
  printLPRP ();
  printSemiColon ();
  printLine ();
}

/**
   \code
   For foo the name of the graph, it generates
   foo_obj_.closeIO ();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintCloseIO ()
{
  printstr (getStringWithProcessTag (MAIN_OBJECT));
  printDot ();
  javaPrintString (MAIN_CLOSE_IO);
  printLPRP ();
  printSemiColon ();
  printLine ();
}

/**
   \code
   For foo the name of the graph, it generates
       code =foo_obj_.foo_initialize();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintInitialize ()
{
  javaPrintString (MAIN_CODE);
  printequal ();
  printstr (getStringWithProcessTag (MAIN_OBJECT));
  printDot ();
  printstr (getStringWithProcessTag (MAIN_INITIALIZE));
  printLPRP ();
  printSemiColon ();
  printLine ();
}

/**
   \code
   For foo the name of the graph, it generates
       code =foo_obj_.foo_step();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintCallIterate ()
{
  javaPrintString (MAIN_CODE);
  printequal ();
  printstr (getStringWithProcessTag (MAIN_OBJECT));
  printDot ();
  printstr (getStringWithProcessTag (BODY_ITERATE));
  printLPRP ();
  printSemiColon ();
  printLine ();
}

/**
   \spe of inteprintcpp::printkwlogical()
   It generates the java keyword for boolean.
*/
_VIRTUAL_ void
inteprintjava::printkwlogical ()
{
  buf->printAlfa (BOOLEAN, 7);
  buf->space ();
}

/**
   \spe of inteprintcpp::printdcltypefunchead().
*/
_VIRTUAL_ void
inteprintjava::printdcltypefunchead (ptident nom, TypeType t, boolean bext)
{
  inteprintc::printFunctionHeadNoParameters (nom, t, bext);
}

/**
   \spe of inteprintcpp::cwrikwexternC()
   It generates the public java keyword.
*/
_VIRTUAL_ void
inteprintjava::cwrikwexternC ()
{
  buf->writeAlfa (PUBLIC, strlen (PUBLIC));
  buf->space ();
}

/**
   \spe of inteprintcpp::getHeaderName_Body()
*/
_VIRTUAL_ ptrepstr
inteprintjava::getHeaderName_Body (ptident id)
{
  return makeStr_fileName (id, JAVA_BODY_SUFFIX, JAVA_SUFFIX);
}

/**
   \spe of inteprintcpp::getHeaderName_BodySeparate()
*/
_VIRTUAL_ ptrepstr
inteprintjava::getHeaderName_BodySeparate (ptident id)
{
  return makeStr_fileName (id, STR_SEPARATE, JAVA_SUFFIX);
}

/**
   \spe of inteprintcpp::printSystemIOIncludes()
   Without effect in Java.
*/
_VIRTUAL_ void
inteprintjava::printSystemIOIncludes ()
{
}

/**
   \code
   It generates 
   - for id=foo_step, 
         foo_step() throws Exception {
   - for id=foo,
         foo() {
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printLinedclfunchead (ptident id)
{
  ptrepstr st;
  st = ident2repstr (id);
  printidemptycall (id);
  if (strstr (repstr2string (st), BODY_ITERATE) != NULL)
    printException ();		// id_iterate 
  blockIndentOpening ();
}

/**
   \spe of inteprintc::printkwstringtype().
   The keyword "String" is used.
*/
_VIRTUAL_ void
inteprintjava::printkwstringtype ()
{
  printstr (string2repstr (kwSignalStringType));
}

/**
   \spe of inteprintc::printformat().
   The keyword Bool, Short, int, Long, Float, Double are used in Java for predefined types.
*/
_VIRTUAL_ void
inteprintjava::printformat (TypeType t)
{
  if (TypeTestArrayClass (t))
    t = ArrayMatrixTypeGetTerm (t);
  switch ((int) t)
    {
    case BooleanType:
    case PureType:
      printstr (string2repstr ("Bool"));
      break;

    case ShortType:
      printstr (string2repstr ("Short"));
      break;

    case IntegerType:
      printstr (string2repstr ("int"));
      break;

    case LongType:
      printstr (string2repstr ("Long"));
      break;

    case RealType:
      printstr (string2repstr ("Float"));
      break;
    case DRealType:
      printstr (string2repstr ("Double"));
      break;

    case StringType:
      printkwstringtype ();
      break;
    case CharType:
      printkwchartype ();
      break;
    default:
      printnotimplemented ();	/*pour les types externes ou nommes ?? */
      break;
    }
}


/**
   \spe of  inteprintcpp::printincludefileSystem()
   Without effect for Java.
*/
_VIRTUAL_ void
inteprintjava::printincludefileSystem ()
{
}


/**
   \spe of  inteprintcpp::printClassNameSeparator().
   Without effect for Java.
*/
_VIRTUAL_ void
inteprintjava::printClassNameSeparator ()
{
}

/**
   It generates the predefined name of the IO object used as parameter of a IO method.
*/
_VIRTUAL_ void
inteprintjava::genIOObjectParameterName ()
{
  javaPrintStringWithoutSpace (OBJECT_PARAMETER_FOR_IO);
}


/**
   It generates the predefined name ("_io_") of the IO object used in the body class for IO operations.
*/
_VIRTUAL_ void
inteprintjava::genIOObjectName ()
{
  javaPrintStringWithoutSpace (BODY_STATIC_IO);
}

/**
   It generates the name of the class used for IO (foo_io is foo is the name of the current graph).
*/
_VIRTUAL_ void
inteprintjava::genIOClassName ()
{
  printstrspace (getStringWithProcessTag (IO_SUFFIX));
}

/**
   It generates the reference ("_io_.") to the object used for IO.
*/
_VIRTUAL_ void
inteprintjava::genIORefObjectName ()
{
  genIOObjectName ();
  printDot ();
}

/**
   @return foo_Externals.java, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintjava::getExternalsFileName (ptident processName)
{
  return makeStr_fileName (processName, EXTERNAL_SUFFIX, JAVA_SUFFIX);
}

/**
   It generates the exception operator ("throws Exception ")
*/
_VIRTUAL_ void
inteprintjava::printException ()
{
  printstr (string2repstr (" throws Exception "));
}

/**
   It calls javaPrintMainFunct(), printbodyrun()  genClassEnd() to generate the main method for a graph named progName.
   \code 
   For progName=foo, it generates
        boolean code ;
        foo foo_obj_=new foo();
        foo_obj_.openIO ();
        code =foo_obj_.foo_initialize();
        while (code) code =foo_obj_.foo_step();
        foo_obj_.closeIO ();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::printbodyrun (ptident progName)
{
  javaPrintMainDeclaration (progName);
  javaPrintOpenIO ();
  javaPrintInitialize ();
  javaPrintString (WHILE);
  printLP ();
  javaPrintString (MAIN_CODE);
  printRP ();
  javaPrintCallIterate ();
  javaPrintCloseIO ();
}

/**
   It calls javaPrintMainFunct(), printbodyrun()  genClassEnd() to generate the main method for a graph named progName.
   \code 
   For progName=foo, it generates
   public static void main ( String[] args )  throws Exception {
        boolean code ;
        foo foo_obj_=new foo();
        foo_obj_.openIO ();
        code =foo_obj_.foo_initialize();
        while (code) code =foo_obj_.foo_step();
        foo_obj_.closeIO ();
    }
   \endcode
*/
_VIRTUAL_ void
inteprintjava::javaPrintdclmainfunc (ptident progName)
{
  javaPrintMainFunct ();
  printbodyrun (progName);
  genClassEnd ();
}


/**
   It generates the keyword scope given by codeAccess
   - 1 for public
   - 2 for private
   - protected, otherwise
*/
_VIRTUAL_ void
inteprintjava::genScope (integer codeAccess)
{
  if (codeAccess == 1)
    printkwpublic ();
  else if (codeAccess == 2)
    printkwprivate ();
  else
    printkwprotected ();
}

/**
   It generates the declarations for IO in the body class, the scope (genScope()) is given by code.
   \code
   public|private|protected KeyboardReader KR = new KeyboardReader();
   \endcode
*/
_VIRTUAL_ void
inteprintjava::genKeyboard (integer code)
{
  genScope (code);
  printstr (string2repstr (KEYBOARD_INSTANCE));
  printLine ();
}




/**
   @return the type to be generated for a type (t).
   For Java, the aliases of the predefined type are not defined.
*/
_VIRTUAL_ TypeType inteprintjava::getTypeToGenerate (TypeType t)
{
  TypeType
    TT;

  if (t == UndefinedType)
    return t;
  if (TypeTestNamedClass (t))
    {
      TT = NamedTypeGetTerm (t);
      if (TypeTestPredef (TT))
	return TT;
    }
  return t;
}

/*
  No effect for Java.
  Specialisation of inteprintcpp::printInherit()
*/
_VIRTUAL_ void
inteprintjava::printInherit (ptrepstr className)
{
}
