/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   Shaper constructor for Java language, code generation with Java threads.
*/
  inteprintjavathread::inteprintjavathread (ptident gn)
{
  makeRepstr_repstr_string (ident2repstr (gn), "_Area", &areaName);
}

/**
   Destructor.
*/
_VIRTUAL_ inteprintjavathread::~inteprintjavathread ()
{
}

/**
   It generates the "run" method for a graph with progName as name.
   \code 
   For progName=foo, it generates
   public void run() throws Exception {
    boolean code ;
    foo foo_obj_=new foo();
    foo_obj_.openIO ();
    code =foo_obj_.foo_initialize();
    while (code )code =foo_obj_.foo_step();
    foo_obj_.closeIO ();
    }
   \endcode
*/
_VIRTUAL_ void
inteprintjavathread::genRunFunction (ptident progName)
{
  printRun ();
  printException ();
  blockIndentOpening ();
  printLine ();
  printbodyrun (progName);
  genClassEnd ();
}


/**
   It generates the constructor of the body class associated with a program (foo as name)
   \code
   public foo(foo_Area Xa){
         XArea = Xa;}

   \endcode
*/
_VIRTUAL_ void
inteprintjavathread::printJavaAreaConstructor (ptident name)
{
  printLine ();
  printkwpublic ();
  printident (name);
  printLP ();
  genAreaName ();
  printstr (string2repstr (" Xa"));
  printRP ();
  blockIndentOpening ();
  printstr (string2repstr (" XArea = Xa;"));
  printBlockClosingBackIndent ();
  printLine ();
}


/**
   It generates the beginning of the foo_Area class (declarations for IO).
   \code
   class foo_Area{
   public  KeyboardReader KR = new KeyboardReader();
   \endcode
*/
_VIRTUAL_ void
inteprintjavathread::genClassAreaBegin ()
{
  printkwClass ();
  genAreaName ();
  blockIndentOpening ();
  printLine ();
  genKeyboard (1);
}

/**
   It generates the tail of the foo_Area class (foo is the name of the currnet graph).
*/
_VIRTUAL_ void
inteprintjavathread::genClassAreaEnd ()
{
  printBlockClosingBackIndent ();
  printLine ();
}

/**
   It generates the entry point for a cluster called name. It is its declaration when declaration, its definition otherwise.
   \code
   For name="foo_Cluster_1", declaration=false, it generates
   class foo_Cluster_1 extends Thread 
   {
   \endcode

*/
_VIRTUAL_ void
inteprintjavathread::genEntryPointDeclaration (ptident name, boolean declaration)
{
  printkwClass ();
  printident (name);
  printstr (string2repstr (" extends Thread "));
  printLine ();
  if (declaration)
    printSemiColonLine ();
}


/**
   It generates the head of the code for a cluster called name.
   \code
   For name="foo_Cluster_1", it generates

   public foo_Area XArea ;

   public foo_Cluster_1(foo_Area Xa){
   XArea = Xa;}
    
     public void run() {
   \endcode
 */
_VIRTUAL_ void
inteprintjavathread::genBeginCluster (ptident name)
{
  blockIndentOpening ();
  genHeaderArea ();
  printSemiColonLine ();
  printJavaAreaConstructor (name);
  printRun ();
  printLBracket ();
  printLine ();
}

/**
   It generates the head of the code for a cluster called name.
*/
_VIRTUAL_ void
inteprintjavathread::genEndCluster ()
{
  printchar ('}');
  printLine ();
  printdclfuncend ();
}

/**
   It generates the following code for id=foo_Cluster_2
   \code
    foo_Cluster_2 Th_foo_Cluster_2 = new foo_Cluster_2(XArea);
        Th_foo_Cluster_2.start();
         try { 
        Th_foo_Cluster_2.join();}
        catch(InterruptedException ie) {} 
   \endcode
*/
_VIRTUAL_ void
inteprintjavathread::genRefSubGraphNode (ptident id)
{
  ptrepstr name_thread;

  printident (id);
  printSpace ();
  strconc2str (string2repstr ("Th_"), ident2repstr (id), pk_false, &name_thread);
  printstrspace (name_thread);
  printassign ();
  printstr (string2repstr ("new "));
  printident (id);
  printstr (string2repstr ("(XArea);"));
  printLine ();
  printstr (name_thread);
  printstr (string2repstr (".start();"));
  printLine ();
  genJoinFunction (name_thread);
}

/**
   It generates the delcaration of the "common" area.
   \code
   public foo_Area XArea =new foo_Area();
   \endcode
*/
_VIRTUAL_ void
inteprintjavathread::genDclrArea ()
{
  genHeaderArea ();
  printEqAllocate ();
  genAreaName ();
  genLPRPSemiColonLine ();
}

/**
   It generates  the beginning of the class associated with a graph with progName as name.
   \code
   For progName=foo, it generates 
   class foo
    {
      private foo_io _io_;
   \endcode
*/
_VIRTUAL_ void
inteprintjavathread::genBodyClassBegin (ptident progName)
{
  printdclbodyfileheader (progName, pk_false);
  printClassBeginDeclaration ();
  printJavaDeclarationIO ();
}
