/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/** Constructor */
  javaGenerator::javaGenerator ()
{
  isIOmodule = pk_false;
}

/** Destructor */
_VIRTUAL_ javaGenerator::~javaGenerator ()
{
}

/** 
  @return "true" if the operation succeeds, "false" othewise.

  Entry point of the Java code generation in non-separated compiling mode.
  It is applied to the graph vGraph which must be a sbDC+ one. The graph is with or without clusters.
  It generates the following files (optionally) foo_main.java, foo.java, foo_io.java.
  The file foo_main contains : 
     - the initialization of the execution environment defined in the file foo_io; 
     - the loop that calls to the "iterate" entry point defined in the file foo.java.

  The code of the file foo,java references i/o operations defined in the file foo_io.java and 
     contains the definitions of the signals.

    @param vGraph (in) is a sbDC+ graph.
    @param mainFile (in) when pk_true, the foo_main.java is generated.
    @param bodyFile (in) when pk_true, the foo.java is generated.
    @param ioFile (in) when pk_true, the foo_io.java is generated.

    Operations:
    - it creates a java shape (inteprintjava()).
    - it initializes the generator (javaGenerator::initialize())
    - it calls javaGenerator::gencode()
    - then javaGenerator::finalize()

    \see{generateSeparate}
*/

_VIRTUAL_
  boolean
javaGenerator::generate (ptDCG_sch vGraph, boolean mainFile, boolean bodyFile, boolean ioFile)
{
  TextForm = new inteprintjava ();
  initialize (vGraph, pk_false, nilstr);
  gencode (mainFile, bodyFile, ioFile, pk_false);
  return finalize ();
}

/**
   @return "true" if the operation succeeds, "false" othewise.

   Entry point of the Java code generation in separated compiling mode.

  It is applied to the graph ?vGraph which must be a sbDC+ one. The graph is with or without clusters.
  It generates the following files, for a foo Signal model, a file foo.java
  @param  vGraph (in) is a sbDC+ graph.

    Operations:
    - it creates a java shaper (inteprintjava).
    - it initializes the generator (javaGenerator::initialize())
    - it calls javaGenerator::genCodeSeparate()
    - then javaGenerator::finalize()

  \see{generate}
*/
_VIRTUAL_ boolean
javaGenerator::generateSeparate (ptDCG_sch vGraph)
{
  TextForm = new inteprintjava ();
  initialize (vGraph, pk_true, nilstr);
  genCodeSeparate ();
  return finalize ();
}
