/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define isEventClockedFormat(t) ((t == DCG_BASIC) ||(t == DCG_POLY) || (t == DCG_ENDO))

_VIRTUAL_  boolean Colouring::filteredClock(ptinsth vh)
{
  return pk_false;
}


/**
   All the formula of a clock (vh) and the assertions are marked.
   When b is "true" the proved equivalent formula are not marked.
*/
  _VIRTUAL_ void
Colouring::markClass (ptinsth vh, boolean b)
{
  ptinsth vharg1;

  vh = CKM->getRepresentativeFormula (vh);
  vharg1 = CKM->getNextFormula (vh, b);
  while (vharg1 != vh)
    {
      markClock (vharg1);
      vharg1 = CKM->getNextFormula (vharg1, b);
    }

  // Assertions.
  vharg1 = vh;
  while (CKM->getNextAssertion (vharg1, &vharg1))
    markClock (vharg1);
}



/**
   It marks a clock (h) and for a bDC+ graph it marks also the branching clock of h with the "currentClock".
*/
_VIRTUAL_ void
Colouring::markClockExpression (ptinsth h)
{
  ptinsth hAt;

  h = CKM->getRepresentativeFormula (h);
  if (!DCplusGraph)
    {
      /*if (AGRAPH->getFormatLevel() == DC) h1 = CKM->getTickClock(); else h1= currentClock; */
      if (CKM->getBooleanClockAt (h, CKM->getBranching (h, currentClock), &hAt))
	markClock (hAt);
    }
  markClock (h);
}


/**
   It marks, recursively, the operands of an expression (vexp).
*/
_VIRTUAL_ void
Colouring::markExpression (ptexpcal vexp)
{
  ptexpcal e1, e2;
  ptident id;

  if (vexp == NILEXPCAL)
    return;
  switch ((int) (EM->getOperatorClass (vexp)))
    {
    case constant:
    case opertyp:
      break;

    case variabl:
      if (EM->IsSignal (vexp))
	marksig (vexp);
      if (EM->IsConstant (vexp))
	markConstantSignal (vexp);
      break;

    case oper1:
    case oper1tp:
      markExpression (EM->getUnaryArg (vexp));
      break;
    case oper2:
    case oper2tp:
      if ((EM->getOperator (vexp) != opelem) && EM->getSignalAssignedTo (vexp, &e1))
	marksig (e1);		// access to a field of a bundle x.y 
      else
	{
	  markExpression (EM->getBinaryArgstAt (vexp, 1));
	  markExpression (EM->getBinaryArgstAt (vexp, 2));
	}
      break;
    case oper3:
    case oper3tp:
      markExpression (EM->getTernaryArgstAt (vexp, 1));
      markExpression (EM->getTernaryArgstAt (vexp, 2));
      markExpression (EM->getTernaryArgstAt (vexp, 3));
      break;
    case opereh:
      markExpression (EM->getWhenOrCellArg (vexp));
      markClockExpression (EM->getContributionClock (vexp));
      break;

    case call:			// function call
      EM->getFunctionCall (vexp, &id, &e1, &e2);
      markExpression (e1);
      markExpression (e2);
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}

/**
   The constraints and assertions, associated with a clock (vh) are marked, but not the equivalent formula.
*/
_VIRTUAL_ void
Colouring::markConstraintsAndProvedEquiv (ptinsth vh)
{
  markClass (vh, pk_true);
}

/**
   It marks a default expression (vexp), left argument of a cell if bcell.
*/
_VIRTUAL_ void
Colouring::markDefaultExpr (ptexpcal vexp, boolean bcell)
{
  ptexpcal vexpg, vhorlexpcal, vexpLast;

  markClock (EM->getContributionClock (vexp));
  vexpLast = EM->getDefaultArgstAt (vexp, 2);
  if (!bcell && !bMarkLastContributionClockDefault)
    vexpLast = EM->getWhenOrCellArg (vexpLast);
  markExpression (vexpLast);

  vhorlexpcal = EM->getDefaultArgstAt (vexp, 1);
  vexpg = vhorlexpcal;
  while (EM->getOperator (vexpg) == opmele)
    {
      markClockExpression (EM->getContributionClock (vexpg));
      markExpression (EM->getDefaultArgstAt (vexpg, 2));
      vhorlexpcal = EM->getDefaultArgstAt (vexpg, 1);
      vexpg = EM->getWhenOrCellArg (vhorlexpcal);
    }
  markClockExpression (EM->getContributionClock (vhorlexpcal));
  markExpression (vhorlexpcal);
}

/** 
    @return "true" if the colouring is applied on a node (vnode), "false" otherwise.
    It marks a node (vnode) if it is defined with a default expression.
*/
_VIRTUAL_ boolean
Colouring::specialNodeProcessing (ptINode vnode)
{
  ptexpcal vexp, vhorlexpcal;
  boolean bcell;

  if (!NM->is_equation (vnode))
    return pk_false;
  if (!NM->get_rightOperand (vnode, &vhorlexpcal))
    return pk_false;
  bcell = EM->IsCellExpr (vhorlexpcal);
  vexp = EM->getBasicExpr (vhorlexpcal);
  if (EM->getOperator (vexp) != opmele)
    {
      if (bcell)
	markClockExpression (EM->getContributionClock (EM->getWhenOrCellArg (vhorlexpcal)));
      return pk_false;
    }
  markDefaultExpr (vexp, bcell);
  return pk_true;
}

/** -- UNUSED but KEPT--\n
    Previously, useful for DC+ code generation.\n
    It marks all contribution clock referenced in a node (vnode).
*/
_VIRTUAL_ void
Colouring::marklContribClock (ptINode vnode)
{
  ptinsth clkinstr, h;
  ptexpcal vexp, vexpg, vdedef;
  boolean bcontinue;

  if (!NM->is_equation (vnode))
    return;
  if (!NM->get_rightOperand (vnode, &vdedef))
    return;
  clkinstr = NM->get_nodeClock (vnode);
  if (EM->IsCellExpr (vdedef))
    vdedef = EM->getWhenOrCellArg (vdedef);
  vexp = EM->getWhenOrCellArg (vdedef);
  if (EM->getOperator (vexp) != opmele)
    return;
  markClock (EM->getContributionClock (vexp));
  vexpg = EM->getDefaultArgstAt (vexp, 1);	/* vexpg = E1 when h1; */
  bcontinue = pk_true;
  while (bcontinue)
    {
      vexp = EM->getWhenOrCellArg (vexpg);
      if (EM->getOperator (vexp) != opmele)
	bcontinue = pk_false;
      else
	{
	  h = EM->getContributionClock (vexpg);
	  markClock (h);
	}
      vexpg = EM->getDefaultArgstAt (vexpg, 1);
    }
}

/**
   @return "true" if a node (node) defines a list of fields (associated with a structure), "false" otherwise.
   It marks the node associated with each signal of the structure.
*/
_VIRTUAL_ boolean
Colouring::marklfields (ptINode node)
{
  if (!NM->is_lfields (node))
    return pk_false;
  marklsig (NM->get_lfields_arg (node));
  return pk_true;
}

/**
   It marks the node associated with each signal defined in a node (vnode).
*/
_VIRTUAL_ void
Colouring::markDefinedSignalsIn (ptINode vnode)
{
  marklsig (NM->getInodelOutputFlows (vnode));
}

/**
   It marks the node associated with each signal referenced in a node (vnode).
*/
_VIRTUAL_ void
Colouring::markReferencedSignalsIn (ptINode vnode)
{
  marklsig (NM->getInodelInputFlows (vnode));
}

/**
   It marks the node associated with each signal referenced in a node (vnode).
*/
_VIRTUAL_ void
Colouring::markReferencedSignalsInOut (ptINode vnode)
{
  marklsig (NM->getInodelInOutFlows (vnode));
}

/**
   It marks the node associated with each constant referenced in a node (vnode).
*/
_VIRTUAL_ void
Colouring::markReferencedConstantsIn (ptINode vnode)
{
  marklsig (NM->getInodelConstants (vnode));
}

/**
   It marks the node associated with each clock referenced in a node (vnode).
*/
_VIRTUAL_ void
Colouring::markReferencedClocksIn (ptINode vnode)
{
  markClocksList (NM->getInodelInputClocks (vnode));
}

/**
   It marks the set of the nodes of a patial node (vnode).
*/
_VIRTUAL_ void
Colouring::markSetOfPartialDefinitions (ptINode vnode)
{
  marklNodes (NM->get_setOfPartialDefinitions_arg (vnode));
}

/**
   It marks a node (vnode) without filtering "special" node (see specialNodeProcessing()).
   It marks
      - the signals defined in (markDefinedSignalsIn()) the node,
      - the signals referenced in (markReferencedSignalsIn()) the node ,
      - the constants referenced in (markReferencedConstantsIn()) the node,
      - the constants referenced in (markReferencedClocksIn()) the node.
*/
_VIRTUAL_ void
Colouring::generalNodeProcessing (ptINode vnode)
{
  ptINode nodeCall;

  markDefinedSignalsIn (vnode);
  if (NM->is_setOfPartialDefinitions (vnode))
    markSetOfPartialDefinitions (vnode);
  else
    {
      markReferencedSignalsIn (vnode);
      markReferencedSignalsInOut (vnode);
      markReferencedConstantsIn (vnode);
      markReferencedClocksIn (vnode);
      if (NM->is_outputNodeCall (vnode, &nodeCall))
	markNode (nodeCall);
      else
	marklfields (vnode);
    }
}

/**
   It marks a node (vnode) by applying first the specialNodeProcessing() method to filter some 
   particular nodes and, in case of failure, it applies generalNodeProcessing() method.
   The operations are applied in a "context clock" (currentclock attribute member) defined by the clock of a operand node (vnode).
 */
_VIRTUAL_ void
Colouring::markNodeCase (ptINode vnode)
{
  ptinsth save;
  save = currentClock;
  currentClock = NM->get_nodeClock (vnode);
  if (!specialNodeProcessing (vnode))
    generalNodeProcessing (vnode);
  currentClock = save;
}

/**
   @return "true" if a node (node) is unselected for the operation, "false" otherwise.
*/
_VIRTUAL_ boolean
Colouring::filteredNode (ptINode node)
{
  return pk_false;
}

/** 
    When a node (node) is selected for the colouring operation (filteredNode()), it marks as useful (recursively)
    the objects (signal/clock) of the graph useful for its production.
*/
_VIRTUAL_ void
Colouring::markNode (ptINode node)
{
  ptinsth h;
  ptINode nBlock;

  if (!filteredNode (node))
    {
      h = NM->get_nodeClock (node);
      if (!CKM->satisfiesNullClock (h))
	if (!NM->isMarked (node, aMark))
	  {
	    NM->addMark (node, aMark);
	    if (!ptlINode_isIn (lboundary, node))
	      {
		markClock (h);
		markNodeCase (node);
	      }
	    if (NM->get_inComposedNode (node, &nBlock))
	      markNode (nBlock);
	  }
    }
}


/**
   It marks the expressions referenced in a type (t).
*/
_VIRTUAL_ void
Colouring::markType (TypeType t)
{
  TypeList l1;
  TypeType t1;
  TypeSubList SubLst1;

  if (TypeTestPredef (t))
    return;
  if (TypeTestExternClass (t))
    return;
  if (TypeTestEnumerationClass (t))
    return;
  if (TypeTestMetaClass (t))
    return;
  /* if (MetaTypeTestDefined(t)) markType(MetaTypeGetCurrent(t)); */

  if (TypeTestArrayClass (t))
    markExpression (ArrayMatrixTypeGetDimLst (t));
  if (TypeTestNamedClass (t))
    markType (NamedTypeGetTerm (t));
  if (TypeTestStructureClass (t))
    markType (StructureTypeGetArgument (t));
  if (TypeTestTupleClass (t))
    {
      l1 = TupleTypeGetTypeLst (t);
      SubLst1 = TypeListCast2SubList (l1);
      while (pk_true)
	{
	  t1 = TypeSubListGetHead (SubLst1);
	  markType (t1);
	  if (TypeSubListTestLast (SubLst1))
	    return;
	  SubLst1 = TypeSubListGetTail (SubLst1);
	}
    }
  /* Not implemented --
     if (TypeTestBundleClass ( t ))  return TypeTypeBundle2ntrees(t);
     if (TypeTestSubrangeClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestBitsClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestLambdaClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestGenericNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestMetaNodeClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestMatrixClass ( t ))  return STree_mkNOTIMPLEMENTED();;
     if (TypeTestTupleClass ( t ))  return STree_mkNOTIMPLEMENTED();
   */
}


/**
   It marks the type of an epxression (exp).
*/
_VIRTUAL_ void
Colouring::markType (ptexpcal exp)
{
  markType (EM->getUnamedType (exp));
}



/**
   It marks definition of a signal (vsig) and its clones (x[i], x[i][j]... are clones of x).
*/
void
Colouring::markDefinition (ptexpcal vsig)
{
  ptexpcal vClone;
  boolean stop;

  vsig = EM->getCloneClassHead (vsig);
  markNode (EM->getNode (vsig));
  do
    if (!EM->getClone (vsig, &vClone))
      stop = pk_true;
    else if (EM->isCloneClassHead (vClone))
      stop = pk_true;
    else
      {
	markNode (EM->getNode (vClone));
	stop = pk_false;
	vsig = vClone;
      }
  while (!stop);
}

_VIRTUAL_ void
Colouring::markBundle (ptexpcal vsig)
{
  ptlinstances lsig;

  if (TypeTestBundleClass (EM->getUnamedType (vsig)))
    if (EM->getFieldsOf (vsig, &lsig))
      marklsig (lsig);
}

/**
   It marks the objects useful to the production of a signal (vsig).
*/
_VIRTUAL_ void
Colouring::marksig (ptexpcal x)
{
  cloper eClass;
  ptinsth h;

  eClass = EM->getOperatorClass (x);
  if (eClass == oper2)
    if (EM->getSignalAssignedTo (x, &x))
      eClass = EM->getOperatorClass (x);
  if (eClass == variabl)
    {
      h = EM->getClock (x);
      if (!CKM->satisfiesNullClock (h))
	{
	  EM->addMark (x, aMark);
	  markDefinition (x);
	  if (!EM->IsIndex (x))
	    markClock (h);
	  markExpression (EM->getInitialValue (x));
	  markType (x);
	  markBundle (x);
	}
    }
}

/**
   It marks the objects useful to the definition of a named constant (x).
 */
_VIRTUAL_ void
Colouring::markConstantSignal (ptexpcal x)
{
  markNode (EM->getNode (x));
  markType (x);
}

/** 
    It marks a list of signals (lsig).
*/
_VIRTUAL_ void
Colouring::marklsig (ptlinstances lsig)
{
  ptexpcal aSig;
  while (ptlinstances_get_next (&lsig, &aSig))
    marksig (aSig);
}


/**
   It marks the operands of a formula (vh).
*/
_VIRTUAL_ void
Colouring::markFormula (ptinsth vh)
{
  ptlinstances lsig;
  ptinsth vharg1, vharg2;

  switch (CKM->getOperator (vh))
    {
    case super:
      if (CKM->getSupArgs (vh, &vharg1, &vharg2))
	{
	  markClock (CKM->getRepresentativeFormula (vharg1));
	  markClock (CKM->getRepresentativeFormula (vharg2));
	};
      break;
    case infer:
      if (CKM->getInfArgs (vh, &vharg1, &vharg2))
	{
	  markClock (CKM->getRepresentativeFormula (vharg1));
	  markClock (CKM->getRepresentativeFormula (vharg2));
	};
      break;
    case sub:
    case nosub:
      // lsig = CKM->getSignalsOfCondition(vh);
      lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vh));
      marklsig (lsig);
      vharg1 = CKM->getRepresentativeFormula (CKM->getClockSub (vh));
      markClock (vharg1);
      ptlinstances_free (lsig);
      break;
    case hnot:
      CKM->getUnaryArg (vh, &vharg1);
      markClock (CKM->getRepresentativeFormula (vharg1));
      break;
    case hcomp:
      CKM->getUnaryArg (vh, &vharg1);
      markClock (CKM->getRepresentativeFormula (vharg1));
      markClock (CKM->getFather (vh));
      break;
    case vari:
    case predef:
    case troot:
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}



/** It marsk a list of clock (lh).
 */
_VIRTUAL_ void
Colouring::markClocksList (ptlinsth lh)
{
  ptinsth vh;
  while (ptlinsth_get_next (&lh, &vh))
    markClock (CKM->getRepresentativeFormula (vh));
}


/** It marks, if it is not done, a clock (vh) (class of formulas) and recursively its operands.
 */
_VIRTUAL_ void
Colouring::markClock (ptinsth vh)
{
  if (!CKM->satisfiesMarked (vh, aMark))
    if (! filteredClock(vh))
    {
      CKM->addMark (vh, aMark);
      markFormula (vh);
      //      if (!CKM->satisfiesIsAnEquivalentForm (vh) && !CKM->satisfiesIsAnAssumption (vh))
      if (!CKM->satisfiesIsAnEquivalentForm (vh) )
	markConstraintsAndProvedEquiv (vh);
    }
}


/** It marks the node associated with an output (exp). 
    It is without effect if it is not an output of the graph.
 */
_VIRTUAL_ void
Colouring::marksigout (ptexpcal vexp)
{
  ptINode vnode, vsucc, vfrom;
  ptldependence ldep;
  ptdependence adep;

  vnode = EM->getNode (vexp);
  ldep = NM->get_successors (vnode);
  while (ldep->get_next (&adep, &ldep))
    {
      adep->getDependence (&vfrom, &vsucc);
      if (NM->is_writeNode (vsucc))
	{
	  markNode (vsucc);
	  markClock (NM->get_nodeClock (vsucc));
	}
    }
}


/**
   It marks the list of outputs of the graph (lsig).
*/
_VIRTUAL_ void
Colouring::marklsigout (ptlinstances lsig)
{
  ptexpcal velem;
  while (ptlinstances_get_next (&lsig, &velem))
    marksigout (velem);
}


/**
   It marks the "unsafe" nodes of a list of nodes (lNodes).
*/
_VIRTUAL_ void
Colouring::marklNodesUnSafe (ptlINode lNodes)
{
  ptINode n;
  while (ptlINode_get_elem (&lNodes, &n))
    if (NM->isMarked (n, DontOptimize))
      markNode (n);
    else
      if (NM->get_safetyProperty (n) != opsafe)
	markNode (n);
}

/**
   It marks a list of nodes (lNodes).
*/
_VIRTUAL_ void
Colouring::marklNodes (ptlINode lNodes)
{
  ptINode vi;
  while (ptlINode_get_elem (&lNodes, &vi))
    markNode (vi);
}



/**
   Colouring explicit dependences.\n
   For each explicit dependence of the graph, it marks the clock associated with the dependence
   if the nodes referenced in the dependence are marked.
*/
_VIRTUAL_ void
Colouring::markDependences ()
{
  ptinsth vh1;
  ptlinsth lh1;
  ptldependence lexplicitDep;
  ptdependence aDep;
  ptINode vFrom, vTo;

  lh1 = AGRAPH->getIODependencesClocks ();
  markClocksList (lh1);

  /* Explicit dependences */
  lexplicitDep = AGRAPH->getExplicitDependences ();
  while (lexplicitDep->get_next (&aDep, &lexplicitDep))
    {
      vh1 = aDep->getDependence (&vFrom, &vTo);
      if (NM->isMarked (vFrom, aMark) && NM->isMarked (vTo, aMark))
	markClock (vh1);
    }
}


/**
   It marks, if they are differents, the clocks h and k, for each (h,k) elements of the lists (lh1, lh2).
 */
void
Colouring::markCoupledLists (ptlinsth lh1, ptlinsth lh2)
{
  ptinsth vh1, vh2;

  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getRepresentativeFormula (vh2);
      if (vh1 != vh2)
	{
	  markClock (vh1);
	  markClock (vh2);
	}
    }
}

/**
   It marks the formula equivalent (but not proved) to the null clock.
*/
_VIRTUAL_ void
Colouring::markNullClockConstraints ()
{
  ptlinsth lh1;
  if (AGRAPH->getNullClockConstraints (&lh1))
    {
      markClock (CKM->getNullClock ());
      markClocksList (lh1);
    }
}

/**
   It marks the assertions of the null clock.
*/
_VIRTUAL_ void
Colouring::markNullClockAssertions ()
{
  ptlinsth lh1;
  if (CKM->getNullClockAssertions (&lh1))
    {
      markClock (CKM->getNullClock ());
      markClocksList (lh1);
    }
}


/**
   It marks the constraints on placed clocks: two nodes of trees of clocks can be equivalent
   but not (yet) proved.
*/
_VIRTUAL_ void
Colouring::markConstraintsOnPlacedClocks ()
{
  ptlinsth lh1, lh2;
  if (CKM->getConstraintsOnPlacedClocks (&lh1, &lh2))
    markCoupledLists (lh1, lh2);
}


/**
  It marks the assertions when the current graph is a DC+ one.
  The assertions are translated in "assert" nodes in bDC+ format, they are marked
  in markNodeClasses() method.
*/
_VIRTUAL_ void
Colouring::markAssertionsOnPlacedClocks ()
{
  ptlinsth lh1, lh2;
  if (DCplusGraph)
    if (CKM->getAssertionsOnPlacedClocks (&lh1, &lh2))
      markCoupledLists (lh1, lh2);
}

/**
   It marks a clock (vh) if it is not a class reduced to a singleton.
*/
_VIRTUAL_ void
Colouring::markConstraintedClock (ptinsth vh)
{
  if (!CKM->satisfiesSingleton (vh))
    markClock (CKM->getRepresentativeFormula (vh));
}

/**
   It marks the clocks with several definitions.
   when a clock is defined only for adding a constraint in the program, it could be not marked by the colouring algorithm. 
   For example, the constraint ( u ^= u ^* x ) and u not used in an another formula was pruned, 
   but this constraint written x ^= u ^+ x was not pruned (because of the reference to x)

   \note It fixes the bug #8546.
*/
_VIRTUAL_ void
Colouring::markConstraintsClass ()
{
  ptlinsth lh;
  ptinsth vh;

  lh = CKM->getlhroots ();
  while (ptlinsth_get_next (&lh, &vh))
    markConstraintedClock (vh);

  lh = CKM->getlhplaced ();
  while (ptlinsth_get_next (&lh, &vh))
    markConstraintedClock (vh);
}

/**
   It marks the following constraints of the graph:
   - null clock equivalence (markNullClockConstraints()),
   - placed formula equivalence (markConstraintsOnPlacedClocks()),
   - assertions formula equivalence (markAssertionsOnPlacedClocks(), markNullClockAssertions()),
   - constraints on the tick clock (markTickClock()).
*/
_VIRTUAL_ void
Colouring::markConstraints ()
{
  /* Constraints */
  markNullClockConstraints ();
  markConstraintsOnPlacedClocks ();
  markConstraintsClass ();
  /* Assertions */
  markNullClockAssertions ();
  markAssertionsOnPlacedClocks ();

  markConstraintedClock (CKM->getTickClock ());
}

/**
   It marks the list of subgraphs (lNodes).
*/
_VIRTUAL_ void
Colouring::colouringSubGraph (ptlINode lNodes)
{
  ptlINode ln;
  ptINode vi;
  ptlinsth lh;
  ptexpcal vlab;

  while (ptlINode_get_elem (&lNodes, &vi))
    {
      if (NM->get_subGraphDefinitions (vi, &ln, &lh))
	marklNodes (ln);
      if (NM->get_label (vi, &vlab))
	marksig (vlab);
    }
}


/**
   It marks the list of iterates (lNodes).
*/
_VIRTUAL_ void
Colouring::colouringArrays (ptlINode lNodes)
{
  ptlINode ln;
  ptINode vi;
  ptlinsth lh;

  while (ptlINode_get_elem (&lNodes, &vi))
    if (NM->get_array_bodyDefinitions (vi, &ln, &lh))
      {
	ln = ptlINode_remove_unmarked (ln, aMark);
	NM->arrayBodySetDef (vi, ln, lh);
	if (ln != NILLINODE)
	  {
	    markNode (vi);
	    if (NM->get_array_initDefinitions (vi, &ln, &lh))
	      marklNodes (ln);
	  }
      }
}

/**
   @return "true" if one signal of a list (lsig) is marked as useful.
*/
_VIRTUAL_ boolean
Colouring::existOneMarked (ptlinstances ls)
{
  ptexpcal s;

  while (ptlinstances_get_next (&ls, &s))
    if (NM->isMarked (EM->getNode (s), aMark))
      return pk_true;
  return pk_false;
}

/**
   It marks the list of fields nodes(lNodes).
*/
_VIRTUAL_ void
Colouring::colouringFieldNodes (ptlINode lNodes)
{
  ptINode vi;
  ptlinstances ls;

  while (ptlINode_get_elem (&lNodes, &vi))
    {
      ls = NM->get_lfields_arg (vi);
      if (NM->isMarked (vi, aMark))
	marklsig (ls); // perhaps already done ?
      else
	// if one field is marked, the node and all the fields are marked.
      if (existOneMarked (ls))
	markNode (vi);
    }
}

/**
   It marks the parameters of the graph.
*/
_VIRTUAL_ void
Colouring::markParameters ()
{
  marklsig (AGRAPH->getParameters ());
}

/**
   It marks the inputs of the graph.
*/
_VIRTUAL_ void
Colouring::markInputs ()
{
  marklsig (AGRAPH->getInputs ());
  markClocksList (AGRAPH->getImportedClocks ());
}

/**
   It marks the outputs of the graph.
*/
_VIRTUAL_ void
Colouring::markOutputs ()
{
  ptlinstances l;

  l = AGRAPH->getOutputs ();
  marklsig (l);
  marklsigout (l);
  markClocksList (AGRAPH->getExportedClocks ());
  // The in-out signals (shared signals by several graphs)
  marklsig (AGRAPH->getInputsOutputs ());
}

/**
   It marks the signals of the interface of the graph
   - parameters (markParameters()),
   - inputs (markInputs()),
   - outputs (markOutputs()).
*/
_VIRTUAL_ void
Colouring::markInterface ()
{
  markParameters ();
  markInputs ();
  markOutputs ();
}

/**
   It marks the computed control clocks (IO step, Step,...).
*/
_VIRTUAL_ void
Colouring::markControlClocks ()
{
  ptinsth save;

  save = currentClock;
  currentClock = CKM->getTickClock ();
  markClockExpression (AGRAPH->getIOstep ());
  markClockExpression (AGRAPH->getStep ());
  currentClock = save;
}



/**
   It marks external call , node call,  sub-graphs, arrays, set of fields nodes of the graph.
*/
_VIRTUAL_ void
Colouring::markNodeClasses ()
{
  ptlINode lNodes;

  if (NM->getProcCalls (&lNodes))
    marklNodesUnSafe (lNodes);				    /** External call */
  if (NM->getNodeCalls (&lNodes))
    marklNodesUnSafe (lNodes);				    /** Node call */
  // if (NM->getSubGraphs(&lNodes)) colouringSubGraph(lNodes); /** Sub-graphs */   // Removed Sept 2014 Lbesnard.
  if (NM->getIterates (&lNodes))
    colouringArrays (lNodes);				   /** Arrays */
  if (NM->getBundles (&lNodes))
    colouringFieldNodes (lNodes);	/* Structure/Bundle (set of fields) */
}


_VIRTUAL_ void
Colouring::markSharedSignalMultiGraphs ()
{
  ptlinstances lidents;
  ptexpcal vs;

  // Added JAn 2015 for he generation of automata.
  if (EM->getExprList (opident, &lidents))
    while (ptlinstances_get_next (&lidents, &vs))
      if (EM->isSharedSignalBySeveralGraphs (vs))
	marksig (vs);
}


/**
   It marks the signals/clocks of the graph useful to the production of the outputs and the reading of the inputs.
   Moreover, constraints are aslo marked and composed nodes (array,...).

   \see {markInterface(), markNodeClasses(), markConstraints(), markDependences().}
   \todo { don't always mark the composed nodes, but only the useful one.}
*/
_VIRTUAL_ void
Colouring::colouring ()
{
  markInterface ();
  markSharedSignalMultiGraphs ();
  markNodeClasses ();
  markConstraints ();
  markDependences ();
  markControlClocks ();
}
