/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   Use init() to set the graph.
*/
  Containers::Containers ()
{
}

/**
   Constructor.
   The functionality is applied to a graph (g).
   A predefined object of this class, called pK_UtilityClock, is available.
*/
Containers::Containers (ptDCG_poly g)
{
  init (g);
}


/** Destructor. */
Containers::~Containers ()
{
}


/**
   It initializes the utility clock functionality for the operand graph g.
*/
void
Containers::init (ptDCG_poly g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
}

/**
   It applies the reduceComm(ptINode, ptexpcal ) method for each inputs of the operand subgraph(aSubGraph).
   @return "true" if the operation succeeds, "false" otherwise(internal error)
*/
boolean
Containers::reduceComm (ptDCG_poly g, ptINode aSubGraph, boolean bOnlyAssignmentClock)
{
  ptlinstances ls;
  ptexpcal vsig;
  boolean b;

  init (g);
  b = pk_true;
  ls = NM->getInodelInputFlows (aSubGraph);
  while (b && ptlinstances_get_next (&ls, &vsig))
    if (!reduceComm (aSubGraph, vsig, bOnlyAssignmentClock))
      b = pk_false;

  if (b)
    NM->set_interface (aSubGraph, NILINTERFACE);	// to rebuilt the interface.
  return b;
}
