
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/// PREDEFINED OBJECT OF THE VIRTUAL MACHINE.
Containers *pK_Containers = new Containers ();

_PROTECTED_
/**
   @return true if the node of the operand signal (vsig) is in a subgraph (sg), "false" otherwise.
*/
  boolean Containers::getSubGraphOf (ptexpcal vsig, ptINode * sg)
{
  ptINode
    node;

  node = EM->getNode (vsig);
  *sg = NILINODE;
  if (!NM->get_inComposedNode (node, sg))
    return pk_false;
  return NM->is_subgraph (*sg);
}

/**

*/
void
Containers::insertNodeInSubGraph (ptINode n, ptINode sg)
{
  NM->set_inComposedNode (n, sg);
  NM->subGraphInsertNode (sg, n);
}

boolean
Containers::getRunOnPragma (ptINode node, ptpragma * aPrag)
{
  return getPragmaKeyword (NM->getPragmas (node), PRAGMA_RUN_ON, aPrag);
}

/**
   Reducing the communication using the  utility/assignment clocks.

   For x sent from SGi to SGj
   - x_for_SGj := x when H is added to the graph: when bOnlyAssignmentClock, H is the assignment clock otherwise is the lower bound of the utility clock and the assignment clock.
   - x (receiver) is replaced, in SGj, by x_r := x_for_SGi default x_r $1.
   @return "true" if the operation succeeds, "false" otherwise(internal error)
*/
boolean
Containers::reduceComm (ptINode aSubGraph, ptexpcal vsig, boolean bOnlyAssignmentClock)
{
  ptexpcal vsig_Receiver, vsig_Sender, vdelay, vDefaultConstantValue;
  ptINode node_delay, sgSender;
  ptinsth hsig, h1, h2, h_vsig_Sender;
  boolean bDefaultConstantValue, b, bapply;
  ptINode node_vsig_Receiver, node_vsig_Sender;
  ptpragma aPrag;

  if (EM->IsInputFlow (vsig))
    return pk_true;

  if (!getSubGraphOf (vsig, &sgSender))	// A signal not assigned to a subgraph!
    {
      printf ("DEBUG in Containers::reduceComm(): the subgraph of the sender is not set! \n");
      return pk_false;
    }
  hsig = EM->getClock (vsig);
  bapply = pk_true;
  bDefaultConstantValue = pk_false;
  if (EM->getUtilityClockInSubGraph (vsig, aSubGraph, &h1))	/* if (EM->getUtilityClock(vsig, &h1)) */
    {
      if (EM->getAssignmentClock (vsig, &h2))
	b = CKM->eqClock (h2, hsig);
      else
	b = pk_true;
      if (b)
	{ 
	  if (!EM->getDefaultConstantValueInfos (vsig, &h2, &vDefaultConstantValue))
	    h2 = hsig;
	  else
	    bDefaultConstantValue = pk_true;
	}
    }
  else
    {
      h1 = hsig;
      if (EM->getAssignmentClock (vsig, &h2))
	b = CKM->eqClock (h2, hsig);
      else
	b = pk_true;
      if (b)
	{
	  if (!EM->getDefaultConstantValueInfos (vsig, &h2, &vDefaultConstantValue))
	    bapply = pk_false;
	  else
	    bDefaultConstantValue = pk_true;
	}
    }
  if (bapply)
    {
      // New signal in the subgraph in which the current signal is produced: the utility clock may introduce cycles.
      if (bOnlyAssignmentClock)
	h_vsig_Sender = h2;
      else
	h_vsig_Sender = CKM->makeInf (h1, h2);
      if (!CKM->eqClock (hsig, h_vsig_Sender))
	{
	  // Sender.
	  vsig_Sender = EM->mkIdentCopy (vsig, NM->mk_node_undefined (), h_vsig_Sender);
	  node_vsig_Sender = NM->insertEquation (vsig_Sender, vsig, h_vsig_Sender, h_vsig_Sender);
	  insertNodeInSubGraph (node_vsig_Sender, sgSender);
	  if (getRunOnPragma (sgSender, &aPrag))
	    NM->addPragma (node_vsig_Sender, aPrag);
	  NM->set_inComposedNode (node_vsig_Sender, sgSender);
	  NM->subGraphInsertNode (sgSender, node_vsig_Sender);

	  // Receiver  a new signal is created in the subgraph in which the current signal is received. 
	  //           vsig_Receiver := (vsig_Sender when h_vsig_Sender) default E, with E the delayed signal (vsig_Receiver$1) or the default value.
	  vsig_Receiver = EM->mkIdentCopy (vsig, NM->mk_node_undefined (), hsig);

	  NM->replace (vsig, vsig_Receiver, aSubGraph);
	  node_vsig_Receiver =
	    NM->mk_node_equation (vsig_Receiver, vsig_Sender, h_vsig_Sender, ptlinstances_insert (NILLINSTANCES, vsig_Sender),
				  EM->getBackAnnotation (vsig));
	  if (bDefaultConstantValue)
	    node_vsig_Receiver =
	      NM->mk_node_equation (vsig_Receiver, vDefaultConstantValue, CKM->makeMinus (hsig, h_vsig_Sender), NILLINSTANCES,
				    EM->getBackAnnotation (vsig));
	  else
	    {
	      vdelay = NM->mkDelayOf (vsig_Receiver);
	      node_delay = EM->getNode (vdelay);
	      insertNodeInSubGraph (node_delay, aSubGraph);
	      if (getRunOnPragma (aSubGraph, &aPrag))
		NM->addPragma (node_delay, aPrag);
	      node_vsig_Receiver =
		NM->mk_node_equation (vsig_Receiver, vdelay, CKM->makeMinus (hsig, h_vsig_Sender), ptlinstances_insert (NILLINSTANCES, vdelay),
				      EM->getBackAnnotation (vsig));
	    }

	  if (getRunOnPragma (aSubGraph, &aPrag))
	    NM->addPragma (node_vsig_Receiver, aPrag);
	  NM->insertNode (node_vsig_Receiver, hsig);
	  insertNodeInSubGraph (node_vsig_Receiver, aSubGraph);
	  // printf(" ::DEBUG:: optimizing comm for %s, Sender = %s, receiver= %s \n", ident2string( EM->getName(vsig)),  ident2string( EM->getName(vsig_Sender)), ident2string( EM->getName(vsig_Receiver)));
	}
    }
  return pk_true;
}
