/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    Constructor. 
*/
    CommManager::CommManager(ptDCG g)
{
  AGRAPH = g;
  EM = g->getExpressionManager();
  CKM = g->getClockManager();
  NM = g->getNodeManager();
  inputs = g->getInputs();
  outputs = g->getOutputs();
  vtrEnv = getPredefinedPragma(PRAGMA_COMM_ENV);
  vtrInput = getPredefinedPragma(PRAGMA_COMM_RECEIVE);
  vtrOutput = getPredefinedPragma(PRAGMA_COMM_SEND);
}

/** 
    Destructor. 
*/
_VIRTUAL_ CommManager::~CommManager()
{
}

/**
   Goal : Building the containers for communications around the processes.

   It adds the communication nodes in the extracted graphs of sub-graphs.
   For two subgraphs SG1,SG2 elements of lSubGraphs, and the associated extracted graph (G1, G2) elements of lsg,
   for each signal, output of SG1 and input of SG2:
   - a communication node (Send) is inserted in G1,
   - a communication node (Receive) is inserted in G2.

  Assumed: lsg is an ordered list of the graphs, extractions of the subgraphs of lsubGraphs: the i-th 
  element of lsg is the graph associated with the i-th element of lSubGraphs.

   \see addCommunicationsFromTo
*/
boolean CommManager::AddCommContainers(ptlINode lSubGraphs, ptlDCG lsg)
{
  ptINode aSubGraph1, aSubGraph2;
  ptlINode lSubGraphs_head, l;
  integer stamp;
  integer *markg1, *markg2;
  ptDCG g1, g2;
  ptlDCG lsg_head, lg;

  stamp = 0;
  if (!addCommIO(lSubGraphs, lsg, &stamp))
    return pk_false;

  lSubGraphs_head = lSubGraphs;
  lsg_head = lsg;
  while (ptlINode_get_elem(&lSubGraphs, &aSubGraph1))
    {
      if (!ptlPointer_get_elemNext(&lsg, (pointer *) & g1))
	return pk_false;
      l = lSubGraphs_head;
      lg = lsg_head;
      while (ptlINode_get_elem(&l, &aSubGraph2))
	{
	  if (!ptlPointer_get_elemNext(&lg, (pointer *) & g2))
	    return pk_false;
	  if (aSubGraph2 != aSubGraph1)
	    if (!addCommunicationsFromTo(aSubGraph1, g1, aSubGraph2, g2, &stamp))
	      return pk_false;
	}
    }
  return pk_true;
}
