/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor of a DCG2dot object (printer for dotty).
 */
  DCG2dot::DCG2dot ()
{

}

/**
   Destructor.
*/
_VIRTUAL_ DCG2dot::~DCG2dot ()
{

}

/**
   It prints for a graph (g)
   - its hierarchy printHierarchy()
   - the hierarchy + the nodes of the graph (printGraph()
   - the hierarchy + for each node of the hierarchy the associated sub-graph (printPartGraphClock()).
*/
void
DCG2dot::print (ptDCG g)
{
  printHierarchy (g);
  printGraph (g);
  printPartGraphClock (g);
}


/**
   It prints in foo_ckm.dot file, where foo is the name of the graph (g)
   the hierarchy of the graph (g). The definitions are not produced only a representation of the nodes.
   It gets a marks (dotAttribute) and  calls the following methods:
   init(),
   printGraphClocks(), ClockHierachyInfo(), printConstraints(), 
   final()
*/
void
DCG2dot::printHierarchy (ptDCG g)
{
  init(g, "_ckm");
  printGraphClocks ();
  ClockHierachyInfo ();
  printConstraints ();
  final();
}


/**
   It prints in foo_Graph.dot file, where foo is the name of the graph (g) the hierarchy + all the nodes of the graphs + the dependences.
   It gets a marks (dotAttribute) and  calls the following methods:
   init(),
   printGraphClocks(), ClockHierachyInfo(),
   printGraphNodes(), printDependences(), printGraphEnd(),
   final()
*/
void
DCG2dot::printGraph (ptDCG g)
{
  init(g, "_Graph");
  if (isGraphOfClusters())
    {
      GraphOfClusters= pk_true;
      printGraphOfClusters ();
    }
  else
    {
      printGraphClocks ();
      ClockHierachyInfo ();
      printGraphNodes ();
      printDependences ();
    }
    
  final();
}


/**
   It prints in foo_gfd.dot file, where foo is the name of the graph (g) the hierarchy  + for each node of the hierarchy, 
   the associated sub-graph and the dependences.

   It  calls the following methods:
   init(),
   printGraphClocks(), ClockHierachyInfo(),
   printGraphAssignedToClocks() for root and placed clocks,
   printSubGraphs(), 
   final()
*/
void
DCG2dot::printPartGraphClock (ptDCG g)
{
  init(g, "_gfd");
  printGraphClocks ();
  ClockHierachyInfo ();
  printGraphAssignedToClocks (lhroots);
  printGraphAssignedToClocks (lhplaced); 
  printSubGraphs ();
  resetMark (); // why here ?
  printDependences ();
  final();
}
