/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor.*/
  FiacreService::FiacreService (ptDCG_poly g)
{
  vDCG = g;
  vDCGFiacre = NILDCG_FIACRE;
  bErrCode = pk_false;
  codeOp = string2repstr ("Fiacre Generator");
}

/** Destructor.*/
FiacreService::~FiacreService ()
{
}

/**
   @return "true" if the operation succeeds, "false" otherwise (operator not yet implemented,...).

   <em> This  method can be applied to a graph of a process, not on a graph of a module.</em>\n
   This is the entry point of the code generation for Fiacres format.
  
  The code is generated in the file \e foo., where  \e foo is the name of the process.
  xxxx
*/
boolean FiacreService::translation ()
{
  ptINode
    selectedTypesSubGraph,
    othersSubGraph;
  ptDCG_basic
    g1;
  ptDCG_poly
    gpoly;

  pK_messenger->messagestr (mbeginOperName, codeOp);
  vDCGFiacre = NILDCG_FIACRE;

  vDCG->expressionExpand ();
  vDCG->separateSelectedTypesOthers (FiacreTypes, &selectedTypesSubGraph, &othersSubGraph);
  g1 = vDCG->cloneSubgraph2Graph (selectedTypesSubGraph, pk_true, pk_true);
  g1->setName (newIdent_stringSuffixed (vDCG->getName (), "FiacrePart"));
  gpoly = new DCG_poly (g1);
  gpoly->setIntrinsicClocks ();
  bErrCode = !DCG_Fiacre::mkDCGFiacre (gpoly, &vDCGFiacre);
  if (bErrCode)
    pK_messenger->printString ("  FiacreService: Program not exported to Fiacre format");
  return !bErrCode;
}

/**
   @return "true" if the operation succeeds, "false" otherwise (error);
   if there is no detected before this calling, it prints 
   - the properties to prove (foo_CMD.z3z file, see printPropertiesToProve())
   - the expression of the synchronization in z3z (foo.z3z file, see printz3zfile())
*/
boolean FiacreService::print ()
{
  if (vDCGFiacre == NILDCG_FIACRE)
    if (bErrCode)
      {
	pK_messenger->
	  printString ("  FiacreService::print: the graph cannot be printed; you must take into account the result of FiacreService::translation()");
	pK_messenger->error (errcompilerbug);
      }
    else
      pK_messenger->printString ("  FiacreService::print: the graph cannot be printed; you must call FiacreService::translation() before.");
  else
    {
      vDCGFiacre->unparse ();
      if (initPrinter ())
	{
	  declareProcesses ();
	  declareComponents ();
	  finalizePrinter ();
	}
    }
  if (bErrCode)
    pK_messenger->printString ("  FiacreService::print::not complete!");
  pK_messenger->messagestr (mendOperName, codeOp);
  return !bErrCode;
}
