/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
#define kwint           "int"
#define lgkwint         3
#define kwlogical       "bool"
#define lgkwlogical     4
// sync --> none for cadp
#define kwevent       "none"
#define lgkwevent     4
/** Constructor.*/
  FiacreShaper::FiacreShaper ()
{
}

/** Destructor.*/
FiacreShaper::~FiacreShaper ()
{
}


void
FiacreShaper::initialize (ptrepstr vname)
{
  progName = vname;
  strtypeint = string2repstr (kwint);
  strtypelogical = string2repstr (kwlogical);
  strtypeevent = string2repstr (kwevent);
}

/**
    @return "true" if the opening of file with unsuffixedFilePathName.z3z as name sucesses, "false" otherwise.
*/
boolean FiacreShaper::openOut_langFile (ptrepstr unsuffixedFileName)
{
  ptrepstr
    filePathName;
  return open_newBuffer (unsuffixedFileName, FIACRESUFFIX, 40, 160, &filePathName);
}

void
FiacreShaper::printCompoment (ptrepstr name)
{
  printString ("component");
  printSpace ();
  printString (name);
}


void
FiacreShaper::printBeginDefinition ()
{
  printLine ();
  printString ("is");
  printLine ();
}


void
FiacreShaper::printEndComposition ()
{
  printLine ();
  printString ("end");
  printLine ();
}


void
FiacreShaper::printBeginComposition ()
{
  printLine ();
  printString ("par");
  printLine ();
}


/**
   It generates the representation of a (assumed) predefined type (ts).
*/
void
FiacreShaper::printtypePredef (TypeType ts)
{
  switch ((int) ts)
    {
    case BooleanType:
      printkwlogical ();
      break;

    case PureType:
      printkwevent ();
      break;

    case IntegerType:
    case LongType:
    case ShortType:
      printkwint ();
      break;

    default:
      printkwnotimpl ();
    }
}

/**
   It generates the key-word for an int.
*/
void
FiacreShaper::printkwint ()
{
  buf->fit (lgkwint);
  buf->writeRepstr (strtypeint);
}

/**
   It generates the key-word for a boolean
*/
void
FiacreShaper::printkwlogical ()
{
  buf->fit (lgkwlogical);
  buf->writeRepstr (strtypelogical);
}

/**
   It generates the key-word for an event.
*/
void
FiacreShaper::printkwevent ()
{
  buf->fit (lgkwevent);
  buf->writeRepstr (strtypeevent);
}

void
FiacreShaper::printPortAtribute (int v)
{
  if (v == 1)
    printString (" in ");
  else if (v == 2)
    printString (" out ");
  else if (v == 3)
    printString (" in out ");
}


/**
   It generates the composition operator ||
*/
void
FiacreShaper::printComposition ()
{
  printBar ();
  printBar ();
}
