/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   In Lustre, the clock of a signal must be declared before the signal.
*/
  _VIRTUAL_ ptlinstances DCG2LustreTree::mkDeclarationsForLustre (ptlinstances l)
{
  ptlinstances
    l1,
    l0;
  ptexpcal
    vsig,
    cond;
  ptinsth
    vh;

  l1 = NILLINSTANCES;
  l0 = l;
  while (ptlinstances_get_next (&l, &vsig))
    {
      if (!ptlinstances_isin_exp (l1, vsig))
	l1 = ptlinstances_push_end (l1, vsig);
      vh = EM->getClock (vsig);
      if (!CKM->satisfiesTickClock (vh))
	if (CKM->satisfiesSub (vh))
	  if (CKM->getOperator (vh) == sub)
	    {
	      cond = CKM->getConditionSub (vh);
	      if (!EM->IsInputFlow (cond))
		if (!EM->IsOutputFlow (cond))
		  l1 = ptlinstances_add_pre (l1, cond, vsig);
	    }
    }
  ptlinstances_free (l0);
  return l1;
}

/** 
    @return vtdecl in which a declaration (aDecl) of a signal/type (vexp) is added.
    @param loc (in): must be "true" for a list of local declaration, "false" otherwise.
*/
_VIRTUAL_ ntrees
DCG2LustreTree::addSignalDeclaration (ntrees vldecl, ntrees aDecl, boolean loc)
{
  if (loc)
    return STree_addEndDeclarationToList (vldecl, aDecl);
  return STree_addEndDeclarationToList (vldecl, aDecl);
}

/**
   In "flat mode", it adds the declarations of the signals (G_LocSig) and the clocks (G_Clocks) 
   in the global model (G_TreeGraph).
   The signals, clocks are stored in the list during the graph travel.
*/
_VIRTUAL_ void
DCG2LustreTree::insertDeclarationsFlatMode ()
{
  if (flatMode)
    {
      // The clock of a signal MUST be declared before the signal.
      appdeclsig (G_TreeGraph, mkDeclarationsForLustre (G_LocSig), 0, pk_false);
      // if (G_Clocks!= NILLINSTH) STree_addLocalDeclaration(G_TreeGraph ,mkDeclarationOfClocks(G_Clocks));
    }
}

_VIRTUAL_ boolean
DCG2LustreTree::isSelectedNode (ptINode inode)
{
  if (inode == TickNode)
    return pk_false;
  return DCG2SignalTree::isSelectedNode (inode);
}

/** 
    @return the name of the current graph 
*/
_VIRTUAL_ ptident
DCG2LustreTree::mkGraphNameWithSubFormat ()
{
  return AGRAPH->getName ();
}


/**
   @copydoc DCG2SignalTree::tick2Tree
*/
_VIRTUAL_ ntrees
DCG2LustreTree::tick2Tree (ptinsth h)
{
  return NILTREE;
}

/**
   @copydoc DCG2SignalTree::clock2Tree
*/
_VIRTUAL_ ntrees
DCG2LustreTree::clock2Tree (ptinsth h)
{
  return NILTREE;
}

/**
   @copydoc DCG2SignalTree::mktreelsigsync
*/
_VIRTUAL_ void
DCG2LustreTree::mktreelsigsync (ptinsth vh, ntrees * vtli, ntrees * vtlni /*, ntrees * vsigtree */ )
{
  // SET_NILTREE (*vsigtree);
  SET_NILTREE (*vtli);
  SET_NILTREE (*vtlni);
}

/**  We override this function to ensure that in Lustre models, all process calls are flattened.

     In Lustre, we always want to flatten the hiearchy of process
     calls.  With a pk_true value, we remove every 'ACT_Tick{}' and other
     similar process calls from the generated output.
 */
_VIRTUAL_ boolean
DCG2LustreTree::getHierarchyModePrinting ()
{
  return pk_true;
}

/**
   @copydoc DCG2SignalTree::mkdecsig
 */
_VIRTUAL_ ntrees
DCG2LustreTree::mkdecsig (ptexpcal vexp, boolean withInit)
{
  ntrees vlsig, t, t1;
  noperators op;
  ptinsth h;

  if (vexp == TickSignal)
    return NILTREE;		// added For Lustre
  withInit = pk_false;		// added For Lustre
  vlsig = LL_make_lsign_list (mkIdentWithInit (vexp, withInit));
  t = typeExp2ntrees (vexp);
  if (STree_isImplicitTypeTree (t))
    pK_messenger->warningaccForce (warUntypedSignal, vlsig);
  if (EM->isStateVar (vexp))
    op = opdecvar;
  else
    op = opdecsig;
  // The clock is added for Lustre in the type.
  h = EM->getClock (vexp);
  if (!CKM->satisfiesTickClock (h))
    t = LL_make_extr_tree (t, mkhonametree (h));
  // ------
  t1 = nmakebinary (op, t, vlsig);
  return t1;
}

/**
   @copydoc DCG2SignalTree::SpecificTranslation.
*/
_VIRTUAL_ ptexpcal
DCG2LustreTree::SpecificTranslation (ptexpcal expr, ptexpcal upExpr, ptinsth hUpExpr)
{
  ptexpcal lp, li;
  ptident id;

  if (hUpExpr == TheTick)
    return expr;
  if (EM->getOperator (expr) == opcond)
    return expr;
  if (EM->getOperatorClass (expr) == call)
    {
      EM->getFunctionCall (expr, &id, &lp, &li);
      if (eq_idents (id, idWhen))
	return expr;
      if (eq_idents (id, idCurrent))
	return expr;
      if (eq_idents (id, idPre))
	return expr;
    }
  return DCG2SignalTree::SpecificTranslation (expr, upExpr, hUpExpr);
}


/**
   Specialisation of DCG2SignalTree::specificTranslatorInit() method.
   It initialises attribute members about the root clock: 
   - TheTick: the master clock of the program
   - TickSignal: the boolean signal that represents the TheTick Clock
   - TickNode: the node associated with the TickSignal signal.
*/
_VIRTUAL_ void
DCG2LustreTree::specificTranslatorInit ()
{
  TheTick = CKM->getTickClock ();
  TickSignal = CKM->getBooleanSignal (TheTick);
  TickNode = EM->getNode (TickSignal);
}
