/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define X1_indent 5
#define smallindent 3
#define actindent 2

#define comBeg "/*"
#define comEnd "*/"
#define comincom "*"
#define comincomend "%------<"
#define errSymbol "~"
#define cstSymbolHolo "..."

_PROTECTED_ _VIRTUAL_ void
LustrePrinter::postlude_root (ntrees p)
{
}

_VIRTUAL_ boolean
LustrePrinter::isprimsig (noperators op)
{
  // clock is encoded in the delcaration
  if (op == opimplicit)
    return pk_true;
  if (op == opinteger)
    return pk_true;
  if (op == oplogical)
    return pk_true;
  if (op == oplong)
    return pk_true;
  if (op == opreal)
    return pk_true;
  if (op == opshort)
    return pk_true;
  if (op == opstring)
    return pk_true;
  if (op == opchar)
    return pk_true;
  if (op == opcomplex)
    return pk_true;
  if (op == opevent)
    return pk_true;
  if (op == opdpreal)
    return pk_true;
  return SignalPrinter::isprimsig (op);
}

_VIRTUAL_ void
LustrePrinter::decta1_opstru ()
{
  //  this_intePrinter->printalphakeysp(kwstruct);
  if (ischildprint (1))
    markkeyunparsekey (kwcouv, kwcfer, vholoph);
}

_VIRTUAL_ void
LustrePrinter::decta2_opappe ()
{
  ptident id;
  ntrees t;

  id = STree_getNameCalledModel (treeunparseTree);
  if (eq_idents (id, idWhen))
    {
      treeunparseTree = nchild (2, treeunparseTree);
      if (ischildprint (2))
	{
	  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
	  if (ischildprint (1))
	    embedarg (!(isprimsig (voperator)));
	  this_intePrinter->printspalphakey (kwwhen);
	  if (ischildprint (2))
	    embedarg (!(isprimsig (voperator)));
	  this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
	}
    }
  else if (eq_idents (id, idCurrent))
    {
      this_intePrinter->printkey (kwpouv);
      SignalPrinter::unalphasigexp (kwcurrent, 2);
      this_intePrinter->printkey (kwpfer);
    }
  else if (eq_idents (id, idPre))
    {
      t = nchild (2, treeunparseTree);
      unparse_subtree (nchild (1, t), vholoph);
      this_intePrinter->printkey (kwfollowedby);
      this_intePrinter->printkey (kwpre);
      this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
      unparse_subtree (nchild (2, t), vholoph);
      this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
    }
  else
    SignalPrinter::decta2_opappe ();
}

_VIRTUAL_ void
LustrePrinter::decta2_opdeccon ()
{
  ntrees vtype, lids, x, id, v;
  int i, n;

  // this_intePrinter->printalphakeysp(kwconstant);
  vtype = nchild (1, treeunparseTree);
  lids = nchild (2, treeunparseTree);
  n = ntreearity (lids);
  for (i = 1; i <= n; i++)
    {
      x = nchild (i, lids);
      id = nchild (1, x);
      v = nchild (2, x);
      if (IS_NILTREE (v))
	this_intePrinter->printkey (kwimported);
      unparse_subtree (id, vholoph);
      this_intePrinter->printalphakeysp (kwdpin);
      unparse_subtree (vtype, vholoph);
      if (ISNOT_NILTREE (v))
	{
	  this_intePrinter->printalphakeysp (kwequl);
	  unparse_subtree (nchild (2, x), vholoph);
	}
    }
}

_VIRTUAL_ void
LustrePrinter::decta2_opdecsig ()
{
  ntrees vtype, lids, x;
  int i, n;

  // this_intePrinter->printalphakeysp(kwconstant);
  vtype = nchild (1, treeunparseTree);
  lids = nchild (2, treeunparseTree);
  n = ntreearity (lids);
  for (i = 1; i <= n; i++)
    {
      x = nchild (i, lids);
      // id = value.
      if (noper (x) == opinit)
	x = nchild (1, x);
      if (i > 1)
	{
	  this_intePrinter->printkeysp (kwvirg);
	}
      unparse_subtree (x, vholoph);
    }
  this_intePrinter->printalphakeysp (kwdpin);
  unparse_subtree (vtype, vholoph);
}

_VIRTUAL_ void
LustrePrinter::declfeedLustre (ntrees p)
{
  int nbfils, i;
  ntrees t;

  nbfils = ntreearity (p);
  if (nbfils == 0)
    {
      this_intePrinter->printerror ();
      this_intePrinter->printalphakeysp (kwpvir);
    }
  else
    {
      for (i = 1; i <= nbfils; i++)
	{
	  t = nchild (i, p);
	  if (noper (t) == oplfeed)
	    declfeedLustre (t);
	  else
	    {
	      if (ISNOT_NILTREE (t))
		unparse_subtree (t, vholoph);
	      this_intePrinter->printalphakeysp (kwpvir);
	      wline (vholoph - actindent);
	    }
	}
    }
}

// if then else
_VIRTUAL_ void
LustrePrinter::decta3_opcond ()
{
  snmark (vholoph - actindent);
  this_intePrinter->printalphakeysp (kwif);
  recsondecunparse (1);
  wline (vholoph - actindent);
  this_intePrinter->printspalphakey (kwthen);
  recsondecunparse (2);
  wline (vholoph - actindent);
  unalphasigexp (kwelse, 3);
  snrelease (vholoph - actindent);
}


/**
   Printing a list of composition with seplst as separator.
*/
_VIRTUAL_ void
LustrePrinter::declfeed (int seplst)
{
  this_intePrinter->printgluekeymark (kwlet, vholoph - actindent);
  wline (vholoph - actindent);
  declfeedLustre (treeunparseTree);
  wline (vholoph - actindent);
  this_intePrinter->printgluekeyrelease (kwtel, vholoph - actindent);
}


_VIRTUAL_ void
LustrePrinter::SelectUnparsingDeclaration (ntrees p, noperators vop, int tk)
{
  ntrees decl, x;
  int nb, i;
  boolean b;

  b = pk_false;
  decl = STreeModel_getLocalDeclarations (p);
  nb = ntreearity (decl);
  for (i = 1; i <= nb; i++)
    {
      x = nchild (i, decl);
      if (noper (x) == vop)
	{
	  if (!b)
	    {
	      b = pk_true;
	      if (tk > 0)
		{
		  this_intePrinter->printgluekeymark (tk, vholoph - actindent);
		  wline (vholoph - actindent);
		}
	    }
	  unparse_subtree (x, vholoph);
	  this_intePrinter->printalphakeysp (kwpvir);
	  if (i < nb)
	    {
	      wline (vholoph - actindent);
	    }
	}
    }
  if (b)
    {
      if (tk > 0)
	snrelease (vholoph - actindent);
      wline (vholoph - actindent);
    }
  treeunparseTree = p;
}

_VIRTUAL_ void
LustrePrinter::declist_opldecs ()
{
  // separate with ";" and add a newline
  decl0 (kwpvir, -3);
}

_VIRTUAL_ void
LustrePrinter::constantUnparsing (ntrees p)
{
  SelectUnparsingDeclaration (p, opdeccon, kwConstantLustre);
  SelectUnparsingDeclaration (p, opdecconp, kwConstantLustre);
}

_VIRTUAL_ void
LustrePrinter::typeUnparsing (ntrees p)
{
  SelectUnparsingDeclaration (p, opltype, -1);
  // SelectUnparsingDeclaration(p, opdectypep, kwtype );
}

_VIRTUAL_ void
LustrePrinter::localNodesUnparsing (ntrees p)
{
  SelectUnparsingDeclaration (p, opprocess, -1);
  SelectUnparsingDeclaration (p, opnode, -1);
  SelectUnparsingDeclaration (p, opfunc, -1);
  // and more
}


_VIRTUAL_ void
LustrePrinter::localDeclUnparsing (ntrees p)
{
  SelectUnparsingDeclaration (p, opdecsig, kwvar);
}


_VIRTUAL_ ntrees
LustrePrinter::processNameUnparsing (ntrees p)
{
  ntrees nomproc;

  treeunparseTree = p;
  if (ischildprint (1))
    {
      nomproc = operande;
      unparse_subtree (operande, vholoph);
    }
  else
    {
      this_intePrinter->printerror ();
      SET_NILTREE (nomproc);
    }
  wline (vholoph - actindent);
  return nomproc;
}

_VIRTUAL_ void
LustrePrinter::BodyUnparsing (ntrees p)
{
  if (ISNOT_NILTREE (p))
    {
      unparse_subtree (STreeModel_getBody (p), vholoph);
      wline (vholoph - actindent);
    }
}


_VIRTUAL_ void
LustrePrinter::InterfaceUnparsing (ntrees p)
{
  ntrees t, externalGraph;

  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
  t = STree_getParameters (p);
  if (ISNOT_NILTREE (t))
    unparse_subtree (t, vholoph);
  t = STree_getInputs (p);
  if (ISNOT_NILTREE (t))
    unparse_subtree (t, vholoph);
  this_intePrinter->printkeyrelease (kwpfer, vholoph - actindent);
  wline (vholoph - actindent);
  this_intePrinter->printalphakeysp (kwreturns);
  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
  t = STree_getOutputs (p);
  if (ISNOT_NILTREE (t))
    unparse_subtree (t, vholoph);

  this_intePrinter->printkeyrelease (kwpfer, vholoph - actindent);
  this_intePrinter->printalphakeysp (kwpvir);
  wline (vholoph - actindent);
  externalGraph = STree_getSpec (p);
}

/**
   @copydoc LustrePrinter::processUnparsing.
*/
_VIRTUAL_ void
LustrePrinter::processUnparsing ()
{
  ntrees nomproc, p;

  p = treeunparseTree;

  constantUnparsing (p);
  typeUnparsing (p);
  localNodesUnparsing (p);
  this_intePrinter->printalphakeysp (kwprocess);	// node.
  nomproc = processNameUnparsing (p);	// name 
  InterfaceUnparsing (p);	// Interface
  // body
  localDeclUnparsing (p);
  BodyUnparsing (p);
  if ((ISNOT_NILTREE (nomproc)) && viscomm)
    {
      this_intePrinter->intedebcomprint ();
      identornameprint (nomproc);
      this_intePrinter->intefincomprint ();
    }
  this_intePrinter->printkeysp (kwpvir);
  treeunparseTree = p;
}

_VIRTUAL_ void
LustrePrinter::decta3_opprocess ()
{
  // Normaly, it should be a node model!!!.
  processUnparsing ();
}


_VIRTUAL_ void
LustrePrinter::decta3_opprocessp ()
{
  // I don't kown if it exists the similar thing in Lustre
  // this_intePrinter->printalphakeysp(kwprivate);
  // this_intePrinter->printkeysp(kwprocess);
  processUnparsing ();
}

_VIRTUAL_ void
LustrePrinter::decta3_opfunc ()
{
  processUnparsing ();
}

_VIRTUAL_ void
LustrePrinter::decta3_opfuncp ()
{
  // I don't kown if it exists the similar thing in Lustre
  // this_intePrinter->printalphakeysp(kwprivate); 
  processUnparsing ();
}

_VIRTUAL_ void
LustrePrinter::decta3_opnode ()
{
  // this_intePrinter->printalphakeysp(kwnode);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opnodep operator.
*/
_VIRTUAL_ void
LustrePrinter::decta3_opnodep ()
{
  // I don't kown if it exists the similar thing in Lustre
  // this_intePrinter->printalphakeysp(kwprivate);
  // this_intePrinter->printkeysp(kwnode);
  processUnparsing ();
}

_VIRTUAL_ boolean
LustrePrinter::initPrinter ()
{
  if (notannounced)
    {
      notannounced = pk_false;
      this_intePrinter = new intePrinter (string2repstr (comBeg), string2repstr (comEnd),
					  string2repstr (comincom), string2repstr (comincomend),
					  string2repstr (errSymbol), string2repstr (cstSymbolHolo));
      if (!this_intePrinter->load_KWLangTable ("LUSTRE"))
	return pk_false;		// LUSTRE.kw file is missing.
    }
  return pk_true;
}
