
_PUBLIC_
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
  _PUBLIC_
/** Constructor.*/
  LustreService::LustreService (ptDCG_bool g)
{
  vDCG = g;
  vDCGLustre = NILDCG_LUSTRE;
  SET_NILTREE (ASTLustre);
  gw = new DCG_Lustre ();
}

/** Destructor.*/
LustreService::~LustreService ()
{
}

/** 
    xxxx
*/
boolean
LustreService::translation ()
{
  vDCGLustre = new DCG_Lustre (vDCG);

  for_all_graphs_do_br (translationOneLevel (), breturn);
  if (breturn)
    vDCGLustre->buildClockProcessInterface ();
  else
    {
      delete vDCGLustre;
      vDCGLustre = NILDCG_LUSTRE;
      setOnError ();
    }
  setTranslationApplied ();
  pK_messenger->setcputime ();
  return breturn;
}

/**
   It produces the Abstract Lustre tree of the current graph. It can be applied to a graph of a module.
*/
boolean
LustreService::produce_tree ()
{
  if (!isTranslationApplied ())
    return pk_false;
  if (isOnError ())
    return pk_false;

  for_all_graphs_do_br (produceOneLevel (), b);
  pK_messenger->setcputime ();
  return b;
}

/**
   It prints the AST built by the produce_tree() method. 
*/
void
LustreService::print ()
{
  ptLustrePrinter oLustrePrinter;
  ptrepstr vstr;

  // Must be verified for a module
  if (ISNOT_NILTREE (ASTLustre))
    {
      vstr = ident2repstr (vDCG->getName ());
      pK_messenger->globalMessageStr (infoLustreMess, vstr);
      oLustrePrinter = new LustrePrinter ();
      oLustrePrinter->fileUnparse (ASTLustre, getPrinterBuffer (), vstr);
      delete oLustrePrinter;
      ntrees_memoryRestoreIndex ();
    }
  pK_messenger->setcputime ();
}
