

/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define ptDCG_OriginalGraphType ptDCG_bool
#include "Service_Macros.c"
  ptDCG_bool LustreService::getGraph ()
{
  return vDCG;
}

ptDCG_Lustre LustreService::getCurrentGraph ()
{
  NM = gw->getNodeManager ();
  EM = gw->getExpressionManager ();
  CKM = gw->getClockManager ();
  return gw;
}

void
LustreService::setCurrentDCG (ptDCG g)
{
  gw->setDCG (g);
}

_PROTECTED_
/**
   @return the operand expression (exp) in which the last contribution clock of a default is removed.
   \verbatim
   For (E1 when h1) default (E2 when h2), 
   it returns (E1 when h1) default (E2).
   \endverbatim
*/
  ptexpcal LustreService::removeLastContributionClockDefault (ptexpcal exp)
{
  ptexpcal
    exp_save,
    e2,
    e1,
    e;
  boolean
    b;
  ptinsth
    hc;

  exp_save = exp;
  if (!EM->IsWhenOrCellExpr (exp))
    b = pk_false;
  else
    {
      b = pk_true;
      hc = EM->getContributionClock (exp);
      exp = EM->getBasicExpr (exp);
    }

  if (!EM->isDefaultExpression (exp))
    return exp_save;
  e2 = EM->getDefaultArgstAt (exp, 2);
  if (!EM->IsWhenOrCellExpr (e2))
    return exp_save;
  e1 = EM->getDefaultArgstAt (exp, 1);
  e2 = EM->getWhenOrCellArg (e2);
  e = EM->mkBinaryExpTp (opmele, e1, e2, EM->getBackAnnotation (exp));
  EM->setContributionClock (e, EM->getContributionClock (exp));
  if (!b)
    return e;
  return EM->mkWhenOrCellExp (EM->getOperator (exp_save), e, hc);
}

/**
   For a equation or a partial definition (node), it removes the last contribution clock of the default expression.
*/
void
LustreService::removeLastContributionClockDefault (ptINode node)
{
  ptexpcal vdedef, e;

  if (NM->is_partialDefinition (node) || NM->is_equation (node))
    if (NM->get_rightOperand (node, &vdedef))
      {
	e = removeLastContributionClockDefault (vdedef);
	if (e != vdedef)
	  NM->update (node, vdedef, e);
      }
}


/**
   It applies removeLastContributionClockDefault(ptINode) for each node of lnodes.
*/
void
LustreService::removeLastContributionClockDefault (ptlINode lnodes)
{
  ptINode node;
  while (ptlINode_get_elem (&lnodes, &node))
    removeLastContributionClockDefault (node);
}

/**
   It applies removeLastContributionClockDefault(ptlINode) for the set 
   of equation and the set of partial definitions of the node manager.
*/
void
LustreService::removeLastContributionClockDefault ()
{
  ptlINode lnodes;
  ptINode node;

  if (NM->getEquations (&lnodes))
    removeLastContributionClockDefault (lnodes);
  if (NM->getGroupsOfPartialEquations (&lnodes))
    while (ptlINode_get_elem (&lnodes, &node))
      removeLastContributionClockDefault (NM->get_setOfPartialDefinitions_arg (node));
}


//----------------------------------------------------------------------------------------
// SPLIT  SPLIT  SPLIT  SPLIT  SPLIT  SPLIT  SPLIT  SPLIT  SPLIT  SPLIT  SPLIT
// DCG2LustreOneLevel(1) ===>> buildIterativeFunction is a DCG_bool function
//     that transforms the DCG into a function the clocks of which are defined "Ã  la Lustre"
// DCG2LustreOneLevel(2) ===>> LustreTranslation(a DCG_bool iterative function) is a Lustre SERVICE
//     that translates a Signal iterative function DCG into a DCG_Lustre DCG
// DCG2LustreOneLevel(3) ===>> export_Lustre(a DCG_Lustre) is a Lustre SERVICE
//     that produces the Lustre AST associated with a Lustre DCG and unparse this AST
//     SPLIT DCG_Lustre::Unparse into DCG_Lustre::produce followed by AST::export_Lustre ?
// DCG2LustreOneLevel ===>> export_LustreTranslation(a DCG_bool) is a Lustre SERVICE
//     that calls buildIterativeFunction, LustreTranslation and then export_Lustre?? 
//----------------------------------------------------------------------------------------

/**
 // PLG SPLIT produceIterativeFunction is a DCG_bool function that generates a function 
 // that has an input signal with the fastest clock
 // PLG: LustreGenerator
*/
/**
   Not fully implemented (in development).
   
   \ref LustreGenerator "For more details..."
*/
boolean LustreService::translationOneLevel ()
{
  ptSignal2Lustre
    o;
  boolean
    interModified,
    b;
  ptDCG_Lustre
    gw;

  gw = getCurrentGraph ();
  // The last contribution clock of default is not used for Lustre.
  removeLastContributionClockDefault ();
  o = new Signal2Lustre (gw->getDCG ());
  b = o->apply (&interModified);
  delete
    o;
  return b;
}


/**
   It produces the Abstract Lustre tree of the current graph.
*/
boolean LustreService::produceOneLevel ()
{
  ptDCG2LustreTree
    LustreTreeBuilder;

  ntrees_memoryStoreIndex ();
  LustreTreeBuilder = new DCG2LustreTree ();
  ASTLustre = LustreTreeBuilder->mkLustreTree (gw);
  delete
    LustreTreeBuilder;
  return pk_true;
}
