/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   Constructor of Lustre translator.
   The functionality is applied to a graph (g), the attribute members (graph, managers,...) are imported from it.
   Assumed: the graph is a DCG_bool one.
*/
  Signal2Lustre::Signal2Lustre (ptDCG g)
{
  AGRAPH = g;
  CKM = AGRAPH->getClockManager ();
  EM = AGRAPH->getExpressionManager ();
  NM = AGRAPH->getNodeManager ();
  DEBUG = pk_false;
  error = pk_false;
  unusedNodes = g->getUnused (&unusedClocks);
  inputs = AGRAPH->getInputs ();
  outputs = AGRAPH->getOutputs ();
  theTick = CKM->getTickClock ();

  // Lustre Inputs/outputs
  NewInputs = NILLINSTANCES;
  NewOutputs = NILLINSTANCES;

  idWhenLustre = string2ident (ID_WHEN_LUSTRE);
  idCurrentLustre = string2ident (ID_CURRENT_LUSTRE);
  idPreLustre = string2ident (ID_PRE_LUSTRE);
}


/**
   Destructor.
*/
_VIRTUAL_ Signal2Lustre::~Signal2Lustre ()
{
}

/** 
    It transforms a bDC+ graph into a Lustre one. 
    The interface may be modified (interModified).
    In Lustre, the clock of an input/output X must be declared before the input/output.
    So, if the clock of an output is a local signal, it becomes an output.
    @return "true" if the operation succeeds, "false" otherwise.
    When it returns "true", interModified is set to "true" if the interface has been modified, "false" otherwise.
*/
boolean Signal2Lustre::apply (boolean * interModified)
{
  ptlINode
    ln;

  initialize ();
  clockOfInputs ();
  if (!error)
    clockOfOutputs ();
  if (!error)
    {
      if (NM->getAllNodes (&ln))
	translateNodes (ptlINode_Copy (ln));
      translateConstrainst ();
    }
  (*interModified) = finalyze ();
  return !error;
}
