/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/



_PROTECTED_
/** 
    @return "true" if a node (node) is a selected node for the operation, "false" otherwise.
*/
  boolean Signal2Lustre::isSelectedNode (ptINode node)
{
  return !ptlINode_isIn (unusedNodes, node);
}


ptlinstances
Signal2Lustre::addClockBeforeSignal (ptlinstances l, ptexpcal vclock, ptexpcal vsig)
{
  return ptlinstances_add_pre (l, vclock, vsig);
  //  if (!ptlinstances_isin_exp(l, vclock)) l = ptlinstances_push_end(l, vclock);
  // return l;
}

/** @return "true" if the clock of an input signal (vsig) is (or added) in the interface of the model respecting Lustre rules, "false" otherwise. */
boolean
Signal2Lustre::clockOfAnInput (ptexpcal vsig)
{
  ptinsth vh;
  ptexpcal cond;

  if (!ptlinstances_isin_exp (NewInputs, vsig))
    NewInputs = ptlinstances_push_end (NewInputs, vsig);
  vh = EM->getClock (vsig);
  if (!CKM->satisfiesTickClock (vh))
    {
      if (!CKM->satisfiesSub (vh))
	return pk_false;
      if (CKM->getOperator (vh) == nosub)
	return pk_false;	//  an output signal cannot be added to define the clock of an input.
      cond = CKM->getConditionSub (vh);
      if (!EM->IsInputFlow (cond))
	return pk_false;
      // EM->belongsInFlow(cond);
      // It must be added before the current input (vsig).
      NewInputs = addClockBeforeSignal (NewInputs, cond, vsig);
    }
  return pk_true;
}

/** @return "true" if the clock of an output signal (vsig) is (or added) in the interface of the model respecting Lustre rules, "false" otherwise. */
boolean
Signal2Lustre::clockOfAnOutput (ptexpcal vsig)
{
  ptinsth vh;
  ptexpcal cond;

  if (!ptlinstances_isin_exp (NewOutputs, vsig))
    NewOutputs = ptlinstances_push_end (NewOutputs, vsig);
  vh = EM->getClock (vsig);
  if (!CKM->satisfiesTickClock (vh))
    {
      if (!CKM->satisfiesSub (vh))
	return pk_false;
      if (CKM->getOperator (vh) == nosub)
	return pk_false;	// a signal must be added.
      cond = CKM->getConditionSub (vh);
      if (EM->IsInputFlow (cond))
	return pk_true;
      // It must be added before the current output (vsig).
      EM->belongsOutFlow (cond);
      NewOutputs = addClockBeforeSignal (NewOutputs, cond, vsig);
    }
  return pk_true;
}

void
Signal2Lustre::clockOfInputs ()
{
  ptlinstances lsig;
  ptexpcal vsig;

  lsig = inputs;
  while (ptlinstances_get_next (&lsig, &vsig))
    if (!clockOfAnInput (vsig))
      {
	pK_messenger->errorIdent (ELustreMess_1, EM->getName (vsig));
	error = pk_true;
      }
}

void
Signal2Lustre::clockOfOutputs ()
{
  ptlinstances lsig;
  ptexpcal vsig;
  integer nsize, osize;

  lsig = outputs;
  if (lsig == NILLINSTANCES)
    {
      pK_messenger->errorIdent (ELustreMess_3, AGRAPH->getName ());
      error = pk_true;
      return;
    }
  nsize = ptlinstances_get_size (lsig);
  do
    {
      osize = nsize;
      while (ptlinstances_get_next (&lsig, &vsig))
	if (!clockOfAnOutput (vsig))
	  {
	    pK_messenger->errorIdent (ELustreMess_2, EM->getName (vsig));
	    error = pk_true;
	  }
      lsig = NewOutputs;
      nsize = ptlinstances_get_size (NewOutputs);
    }
  while (osize != nsize);
}

void
Signal2Lustre::initialize ()
{
  previousTableOfMessages = pK_messenger->gmessload (LustreMess);
  pK_messenger->message (MLustreMess_1);
  EM->setModeTypeChecking (pk_false);
}

boolean
Signal2Lustre::finalyze ()
{
  pK_messenger->gmessrestore (previousTableOfMessages);
  if (error)
    return pk_false;
  // return pk_true if the interface is modified, result of comparing inputs/NewInputs and Outputs/NewOutputs.
  AGRAPH->setInputs (NewInputs);
  AGRAPH->setOutputs (NewOutputs);
  return pk_true;
}

/**
   @return a call expression id{}(e1,e2) at the clock h, the result type is TT.
*/
ptexpcal
Signal2Lustre::mkCall (ptident id, ptexpcal e1, ptexpcal e2, ptinsth h, TypeType TT)
{
  ptexpcal e, call;
  if (e2 != NILEXPCAL)
    e = EM->mkNaryExp (opleflot, e2, NILEXPCAL, NILBACKANNOTATION);
  else
    e = NILEXPCAL;
  if (e1 != NILEXPCAL)
    e = EM->mkNaryExp (opleflot, e1, e, NILBACKANNOTATION);
  call = EM->mkCallExp (id, NILEXPCAL, e, TT, h);

  if (id == idWhenLustre || id == idCurrentLustre || id == idPreLustre)
    {
      EM->addMark (call, DONTDECLARE);
    }

  return call;
}

/**
   @param mvexp (in) the cell signal.
   @param vexp (in) is assumed to be a cell expression (ie "mvexp := (E when CE) cell C")
   @param vinit (in) initial value of the cell
   @param hn (in)clock of the cell
   @return ((if CE then current(E) else (vinit-> pre(E)) at hn, using current operator.
*/
ptexpcal
Signal2Lustre::translateCell (ptexpcal mvexp, ptexpcal vexp, ptexpcal vinit, ptinsth hn)
{
  ptexpcal EwhenCE, E, ClockCE, ifExpr, thenPart, elsePart;
  ptinsth ClockC, clockE, clock_init, clock_CCE;
  TypeType TT;

  EwhenCE = EM->getWhenOrCellArg (vexp);
  ClockC = EM->getContributionClock (vexp);
  E = EM->getWhenOrCellArg (EwhenCE);
  TT = EM->getType (E);
  clockE = EM->getContributionClock (EwhenCE);
  ClockCE = CKM->getBooleanSignal (CKM->getBooleanClock (clockE));

  if (!EM->getExprClock (ClockCE, &clock_CCE))
    clock_CCE = theTick;
  ClockCE = getExprAt (ClockCE, clock_CCE, hn);

  if (!EM->getExprClock (vinit, &clock_init))
    clock_init = theTick;
  vinit = getExprAt (vinit, clock_init, hn);

  E = translateSynchronousExpr (E, clockE);	// For constant translation.

  thenPart = getExprAt (E, clockE, hn);
  elsePart = mkCall (idPreLustre, vinit, mvexp, clockE, TT);
  ifExpr = EM->mkTernaryExp (opcond, ClockCE, thenPart, elsePart, NILBACKANNOTATION);
  return ifExpr;
}

boolean
Signal2Lustre::existCurrentNodeAt (ptexpcal e, ptinsth h, ptexpcal * currentE)
{
  ptlINode ln;
  ptINode n;
  ptident id;
  ptexpcal params, OrderedInputs, OrderedOutputs;

  ln = CKM->getNodesAssignedTo (h);
  while (ptlINode_get_elem (&ln, &n))
    if (NM->isProcessCall (n))
      {
	NM->get_procCall_arg (n, &id, &params, &OrderedInputs, &OrderedOutputs);
	if (eq_idents (id, idCurrentLustre))
	  if (EM->getNaryArgs (&OrderedInputs) == e)
	    {
	      (*currentE) = EM->getNaryArgs (&OrderedOutputs);
	      return pk_true;
	    }
      }
  return pk_false;
}

/**  */
ptexpcal
Signal2Lustre::mkCurrentCall (ptexpcal e, ptinsth h, TypeType TT)
{
  ptexpcal x, ins, outs;
  ptINode node;

  // Lustre Compiler does not accept current(current(x)), so it is necessary to introduce some auxillary signals (LBesnard, 2006 Nov)
  if (!existCurrentNodeAt (e, h, &x))
    {
      ins = EM->mkNaryExp (oplesign, e, NILEXPCAL, NILBACKANNOTATION);
      x = EM->mkIdentExp (getIdentNil (), tsignal, TT, NM->mk_node_undefined (), h, NILBACKANNOTATION);
      outs = EM->mkNaryExp (oplesign, x, NILEXPCAL, NILBACKANNOTATION);
      node = NM->mk_procCall (idCurrentLustre, NILEXPCAL, ins, outs, NILBACKANNOTATION);
      NM->insertNode (node, h);
    }
  return x;
}

/** @return the expression of the memory, in Lustre (Current operator) of an expression (vexp) at a clock (hTo).
    The expression is built from hFrom assumed to be under hTo in the hierarchy.
*/
ptexpcal
Signal2Lustre::makeCurrentLustreFromTo (ptinsth hFrom, ptinsth hTo, ptexpcal vexp)
{
  ptexpcal e;
  ptinsth k;
  TypeType TT;

  if (hFrom == hTo)
    return vexp;
  e = vexp;
  TT = EM->getType (e);
  while (hFrom != hTo)
    {
      k = CKM->getFather (hFrom);
      if (k == hTo)
	e = mkCall (idCurrentLustre, e, NILEXPCAL, hFrom, TT);
      else
	e = mkCurrentCall (e, k, TT);
      hFrom = k;
    }
  return e;
}

ptexpcal
Signal2Lustre::mkWhenLustre (ptexpcal e, ptinsth h, boolean b)
{
  ptexpcal c;
  TypeType TT;

  c = CKM->getBooleanSignal (h);
  TT = EM->getType (e);
  if (b)
    return mkCall (idWhenLustre, c, e, h, TT);
  return mkCall (idWhenLustre, e, c, h, TT);
}

/** 
    @return the expression (vexp when hTo) when h1....when hn when hFrom, where h1,...,hn are clocks on the branch hFrom, hTo, with
    hFrom under hTo is the hierachy (assumed).
*/
ptexpcal
Signal2Lustre::makeWhenLustreFromTo (ptinsth hFrom, ptinsth hTo, ptexpcal vexp, boolean fson)
{
  ptexpcal e1;
  ptinsth fClk;

  if (hFrom == hTo)
    return vexp;
  fClk = CKM->getFather (hFrom);
  if (!fson)
    if (fClk == hTo)
      return vexp;
  e1 = makeWhenLustreFromTo (fClk, hTo, vexp, fson);
  return mkWhenLustre (e1, hFrom, pk_false);
}

ptexpcal
Signal2Lustre::getClockAt (ptinsth vClk, ptinsth hTo)
{
  ptexpcal e;
  vClk = CKM->getRepresentativeFormula (vClk);
  e = CKM->getBooleanSignal (CKM->getBooleanClock (vClk));
  return getExprAt (e, CKM->getFather (vClk), hTo);
}


ptexpcal
Signal2Lustre::getExprAt (ptexpcal e, ptinsth vClk, ptinsth hTo)
{
  ptinsth branchingClock;
  ptexpcal e1;

  if (vClk == hTo)
    return e;
  branchingClock = CKM->getBranching (vClk, hTo);
  if (branchingClock == hTo)
    return makeCurrentLustreFromTo (vClk, branchingClock, e);
  if (branchingClock == vClk)
    return makeWhenLustreFromTo (hTo, branchingClock, e, pk_true);
  e1 = makeCurrentLustreFromTo (vClk, branchingClock, e);
  return makeWhenLustreFromTo (hTo, branchingClock, e1, pk_true);
}

ptexpcal
Signal2Lustre::translateSynchronousExpr (ptexpcal vexp, ptinsth hAt)
{
  ptexpcal e;
  e = CKM->getBooleanSignal (hAt);
  if (hAt != theTick)
    {
      e = makeWhenLustreFromTo (CKM->getFather (hAt), theTick, e, pk_true);
    }
  return translateSynchronousExpr (vexp, hAt, e);
}

ptexpcal
Signal2Lustre::translateWhenInDefaultExpr (ptexpcal e, ptinsth vClk, ptinsth nodeClock)
{
  ptexpcal nexp;
  boolean bConst;
  ptinsth eClk;

  bConst = EM->IsConstantExpr (e);
  nexp = translateSynchronousExpr (e, nodeClock);
  if (bConst)
    return nexp;
  if (!EM->getExprClock (e, &eClk))
    eClk = vClk;
  nexp = translateWhenExpr (nexp, eClk, vClk);
  if (!CKM->satisfiesTickClock (nodeClock))
    if (vClk != nodeClock)
      nexp = translateWhenExpr (nexp, vClk, nodeClock);
  return nexp;
}

ptexpcal
Signal2Lustre::translateDefaultExpr (ptexpcal vexp, ptinsth hn, boolean * b)
{
  ptexpcal exp, exp1, exp2, nexp1, nexp2, e1, vcondition;
  ptinsth vClk;

  exp = EM->getWhenOrCellArg (vexp);
  if (EM->getOperator (exp) != opmele)
    {
      // Last part of the default.
      e1 = EM->getWhenOrCellArg (exp);
      return translateWhenInDefaultExpr (e1, hn, hn);
    }

  EM->getDefaultArgs (exp, &exp1, &exp2);
  nexp2 = translateDefaultExpr (exp2, hn, b);

  e1 = EM->getWhenOrCellArg (exp1);
  vClk = EM->getContributionClock (exp1);
  nexp1 = translateWhenInDefaultExpr (e1, vClk, hn);
  vcondition = getClockAt (vClk, hn);
  (*b) = (hn == CKM->getBranching (hn, vClk));
  return EM->mkTernaryExp (opcond, vcondition, nexp1, nexp2,  NILBACKANNOTATION);
}


ptexpcal
Signal2Lustre::translateDefault (ptexpcal vexp, ptinsth hn)
{
  boolean bUp;
  return translateDefaultExpr (vexp, hn, &bUp);
}


/**
   For vexp defined by left_tt(e1,e2), hn clock of the vexp, 
   it buids the Lustre memorisation of e1, using the current operator, at hn.
 */
ptexpcal
Signal2Lustre::translateLeftttExpr (ptexpcal vexp, ptinsth hn)
{
  ptinsth e1Clk;
  ptexpcal e1;

  e1 = EM->getBinaryArgstAt (vexp, 1);
  if (!EM->getExprClock (e1, &e1Clk))
    e1Clk = hn;
  return makeCurrentLustreFromTo (e1Clk, hn, e1);
}

ptexpcal
Signal2Lustre::translateCounterExpr (ptexpcal vexp, ptinsth hn)
{
  pK_messenger->errorString (ELustreMess_4, "counters");
  error = pk_true;
  return vexp;
}

ptexpcal
Signal2Lustre::translateSynchronousExpr (ptexpcal vexp, ptinsth hAt, ptexpcal samplingExpr)
{
  ptexpcal simexp, basicExpr, vexpcal1, vexpcal2, v1, v2;
  ptident id;
  ptinsth h;

  if (vexp == NILEXPCAL)
    return vexp;
  EM->getIndirection (vexp, &vexp);	// useful ?

  basicExpr = EM->getBasicExpr (vexp);
  simexp = simulateSynchronousExpr (basicExpr, hAt);
  if (simexp)
    {
      // vexp is an expression which cannot be expressed
      // directly in Lustre and must be interpreted
      return simexp;
    }

  if (hAt == theTick)
    return vexp;

  // Simple translations for remaining operators
  switch (EM->getOperatorClass (vexp))
    {
    case variabl:
      {
	switch (EM->getClassIdentifier (vexp))
	  {
	  case tconstant:
	  case tparam:
	    // Static parameters and constant identifiers must be sampled in Lustre
	    return getExprAt (vexp, theTick, hAt);
	  case tsignal:
	  case tindex:
	  case tfield:
	  case tlabel:
	    return vexp;
	  case ttyp:
	    // unexpected case
	    pK_messenger->Caseerror (currentLine, currentFile);
	  }
	// run-time check for bad enum coverage / bad value
	pK_messenger->Caseerror (currentLine, currentFile);
      }
    case constant:
      return getExprAt (vexp, theTick, hAt);

    case oper1:
      vexpcal1 = EM->getUnaryArg (vexp);
      v1 = translateSynchronousExpr (vexpcal1, hAt, samplingExpr);
      if (v1 != vexpcal1)
	return EM->mkUnaryExp (EM->getOperator (vexp), v1, NILBACKANNOTATION);
      return vexp;

    case oper2:
    case oper2tp:
      // voperator = EM->getOperator (vexp);
      vexpcal1 = EM->getBinaryArgstAt (vexp, 1);
      v1 = translateSynchronousExpr (vexpcal1, hAt, samplingExpr);
      vexpcal2 = EM->getBinaryArgstAt (vexp, 2);
      v2 = translateSynchronousExpr (vexpcal2, hAt, samplingExpr);
      if ((v1 != vexpcal1) || (v2 != vexpcal2))
	return EM->mkBinaryExp (EM->getOperator (vexp), v1, v2, NILBACKANNOTATION);
      return vexp;

    case call:
      EM->getFunctionCall (vexp, &id, &vexpcal1, &vexpcal2);
      v2 = translateSynchronousExpr (vexpcal2, hAt, samplingExpr);
      if (v2 == vexpcal2)
	return vexp;
      return EM->mkCallExp (id, vexpcal1, v2, EM->getType (vexp), EM->getContributionClock (vexp));

    case opertyp:
      return vexp;

    case opereh:
      vexpcal1 = EM->getWhenOrCellArg (vexp);
      vexpcal1 = translateSynchronousExpr (vexpcal1, hAt, samplingExpr);
      h = CKM->getRepresentativeFormula (EM->getContributionClock (vexp));
      return translateWhenExpr (vexpcal1, h, h);
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vexp;
}

/** 
    It translates (vexp when hWhen, with hexp the clock of vexp).
*/
ptexpcal
Signal2Lustre::translateWhenExpr (ptexpcal vexp, ptinsth hexp, ptinsth hWhen)
{
  return getExprAt (vexp, hexp, hWhen);
}

/**
   Translate Signal expressions that need to be simulated in Lustre
   (default, ...)
   @param basicExpr: expression coming from some
          call to ExprManager::getBasicExpr
   @param clock : the required clock of the resulting expression
   @returns non-null iff the given expression is translated (no further processing needed)
*/
ptexpcal
Signal2Lustre::simulateSynchronousExpr (ptexpcal basicExpr, ptinsth clock)
{
  ptexpcal nexpr;
  ptinsth vh;
  noperators vop;

  switch (EM->getOperator (basicExpr))
    {
    case opmele:
      return translateDefault (basicExpr, clock);
    case opleft_tt:
      return translateLeftttExpr (basicExpr, clock);
    case opnombm:
      return translateCounterExpr (EM->getBinaryArgstAt (basicExpr, 2), clock);
    default:
      return NILEXPCAL;
    }
}

/**  
     @return "true" if the operation succeeds, "false" otherwise.
*/
void
Signal2Lustre::translateEquation (ptINode vnode)
{
  ptexpcal vopg, vopd, basicExpr, nexpr;
  ptinsth hg, hd, vh;

  if (NM->get_rightOperand (vnode, &vopd))
    {
      vopg = NM->get_leftOperand (vnode);
      hg = EM->getClock (vopg);
      nexpr = vopd;

      if (EM->IsCellExpr (vopd))
	nexpr = translateCell (vopg, vopd, EM->getInitialValue (vopg), hg);
      else
	{
	  basicExpr = EM->getBasicExpr (vopd);
	  if (EM->getSynchronousExprHorl (basicExpr, &vh))
	    vh = CKM->getRepresentativeFormula (vh);
	  else
	    vh = hg;
	  nexpr = translateSynchronousExpr (basicExpr, hg);
	  if (vh != hg)
	    nexpr = translateWhenExpr (nexpr, vh, hg);
	}
      if (vopd != nexpr)
	NM->modifiedDefinition (vopg, nexpr, hg, hg);
    }
}


/**  
     xxxxxxxxxxxxxxx
*/
void
Signal2Lustre::translateArray (ptINode vnode)
{
  pK_messenger->errorString (ELustreMess_4, "arrays");
  error = pk_true;
}

ptexpcal
Signal2Lustre::getithExprOf (ptexpcal expr, integer n)
{
  integer i;
  ptexpcal e;

  if (!EM->satisfiesList (expr))
    return expr;
  if (n <= 0)
    return NILEXPCAL;
  i = 0;
  while (i != n)
    {
      if (!EM->getNaryArgs (&expr, &e))
	return NILEXPCAL;
      i++;
    }
  return e;
}


/**  
    Delay translation. For zx := x $N init {V0,V1,..VN-1} it produces zx = V0 -> pre (V1 -> ... VN-1->pre(x))
   
    To be completed for: zx := x windows N.
*/
void
Signal2Lustre::translateDelay (ptINode vnode)
{
  ptexpcal vdef, delayedSignal, vdelay, vwin, vinit, nexp;
  ptinsth h;
  integer i, valret;
  TypeType TT;

  /* <vdef> = <delayedSignal> $ <vdelay> window <vwin> */
  // zx := x $1 init V0, zx = V0 -> pre(x)
  // zx := x $N init {V0,V1,..VN-1}, zx = V0 -> pre (V1 -> ... VN-1->pre(x))
  // zx := x windows N, 
  NM->get_delay_arg (vnode, &vdef, &delayedSignal, &vdelay, &vwin, &vinit);
  h = EM->getClock (vdef);
  if (vwin != NILEXPCAL)
    {
      pK_messenger->errorString (ELustreMess_4, "Window");
      error = pk_true;
      return;
    }
  if (!EM->IsAssigned (vdelay))
    {
      pK_messenger->error (ELustreMess_5);
      error = pk_true;
      return;
    }
  valret = EM->getValue (vdelay);
  if (valret == 0)
    {
      // no occurrence here, I hope!
      pK_messenger->error (ELustreMess_5);
      error = pk_true;
      return;
    }
  nexp = NILEXPCAL;
  TT = EM->getType (delayedSignal);
  nexp = mkCall (idPreLustre, translateSynchronousExpr (getithExprOf (vinit, valret), h), delayedSignal, h, TT);
  i = valret - 1;
  while (i > 0)
    {
      nexp = mkCall (idPreLustre, translateSynchronousExpr (getithExprOf (vinit, valret - i), h), nexp, h, TT);
      i--;
    }
  NM->modifiedDefinition (vdef, nexp, h, h);
}

/**  
     xxxxxxxxxx
*/
void
Signal2Lustre::translateProcCall (ptINode vnode)
{
  ptident id;

  if (!NM->is_ioNode (vnode))
    if (EM->isNILEXPCAL (NM->get_leftOperand (vnode)))
      if (NM->get_nameCall (vnode, &id))
	{
	  pK_messenger->errorIdent (ELustreMess_3, id);
	  error = pk_true;
	}
}

/**  
     xxxxxxxxxx
*/
void
Signal2Lustre::translateNodeCall (ptINode vnode)
{
  ptident id;

  NM->get_nameCall (vnode, &id);
  if (!EM->isNILEXPCAL (NM->get_leftOperand (vnode)))
    {
      pK_messenger->errorString (ELustreMess_4, "node call");
      error = pk_true;
    }
  else
    {
      pK_messenger->errorIdent (ELustreMess_3, id);
      error = pk_true;
    }
}

/**  
     xxxxxxxxxxxxxxx
*/
void
Signal2Lustre::translateReferenceToSubGraph (ptINode vnode)
{
  pK_messenger->errorString (ELustreMess_4, "subGraph");
  error = pk_true;
}

/**  
     xxxxxxxxx
*/
void
Signal2Lustre::translateNode (ptINode vnode)
{
  switch ((int) (NM->get_oper (vnode)))
    {
    case opdefsig:
      translateEquation (vnode);
      break;
    case opmoti:
      translateArray (vnode);
      break;
    case opret:
      translateDelay (vnode);
      break;
    case opappe:
      translateProcCall (vnode);
      break;
    case opnodecall:
      translateNodeCall (vnode);
      break;
    case opsubgraph:
      translateReferenceToSubGraph (vnode);
      break;
    case opnil:
      break;

    case opdefpart:
      // Assumed: the partial equations have been translated in "default" node before
      break;
    case opldefpart:
      // Assumed: the partial equations have been translated in "default" node before
      break;

    case opbund:
      // Not yet implemented in Polychrony.
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
      error = pk_true;
    }
}

_VIRTUAL_ void
Signal2Lustre::translateNodes (ptlINode ln)
{
  ptINode n;

  // When synchrone, Pas de default, cell, compteurs, left_tt, 
  // Utilisation de current et du when synchrone de Lustre.
  while (ptlINode_get_elem (&ln, &n))
    if (!error)
      if (isSelectedNode (n))
	translateNode (n);
  // free(ln);
}


_VIRTUAL_ void
Signal2Lustre::translateConstrainst (ptinsth vh1, ptinsth vh2)
{
  ptinsth h;
  ptexpcal c, evh1, evh2, e;

  h = CKM->getBranching (vh1, vh2);
  evh1 = getClockAt (vh1, h);
  evh2 = getClockAt (vh2, h);
  c = EM->mkIdentExp (getIdentNil (), tsignal, BooleanType, NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION);
  e = EM->mkBinaryExp (opegal, evh1, evh2, NILBACKANNOTATION);
  NM->insertEquation (c, e, h, h);
}


/**
   For each clock of a list (lh), 
   it adds, in Lustre, the expression of the synchonisation of the clock with the null clock.
   @todo {TO BE COMPLETED.}
*/
_VIRTUAL_ void
Signal2Lustre::addConstraintsNullClock (ptlinsth lh)
{
  ptinsth vh, hnull;

  if (lh == NILLINSTH)
    return;
  hnull = CKM->getNullClock ();
  while (ptlinsth_get_next (&lh, &vh))
    {

    }
}

_VIRTUAL_ void
Signal2Lustre::translateConstrainst ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  // Null clock
  if (AGRAPH->getNullClockConstraints (&lh1))
    addConstraintsNullClock (lh1);
  // Other constraints on placed clocks.
  CKM->getConstraintsOnPlacedClocks (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getRepresentativeFormula (vh2);
      if (vh1 != vh2)
	translateConstrainst (vh1, vh2);
    }
}
