/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define TRACE(s)

#define tab "    "
#define genBeginComment() ofile << ";"

#define genBeginAssign() ofile << "(= "
#define genImplies() ofile << "=> "
#define genSpace() ofile << " "

// ite : if then else
#define genITEClock(vclk)  ofile << "(ite "; genClockId(vclk);

ptident SMTPrinter::getClockId(ptinsth vh)
{
  ptexpcal vsig;
  ptident id;
  ptrepstr repstrResult;

  id = CKM->getName(vh);
  if (! CKM->isClockRepresentedBySignal (vh, &vsig) && ! CKM->isClockUnifiedWithSignal(vh, &vsig) ) return id;
  makeRepstr_string_ident ("^", id, &repstrResult);
  return repstr2ident(repstrResult);
}

ptident SMTPrinter::getSigId(ptexpcal e)
{
  return EM->getName(e);
}


ptident SMTPrinter::getIdentCondition(ptexpcal e)
{
  ptident id;

  if (EM->IsEvent(e)) return id_true;		// ???
  if (EM->IsSignal(e)) return getSigId(e);
  if (EM->getAttribute(e, idconditionattr, (attributeValue *) &id)) return id;
  return string2ident("_UNDEF_");
}

void SMTPrinter::genIdent(ptident id)
{
  genString(ident2string(id));
}


void SMTPrinter::genOperator(noperators vop)
{
  // if (vop != opvvarin)

  if (vop == opfalse) genString("false");
  else if (vop == optrue) genString("true");
  else if (vop == opplus) genString("+");
  else if (vop == opdivi) genString("/");
  else if (vop == opmod)  genString("modulo");
  else if (vop == opmult) genString("*");
  else if (vop == opmoin) genString("-");
  else if (vop == oppuis) genString("pow");
  else if (vop == opdiff) genString("!=");
  else if (vop == opegal) genString("==");
  else if (vop == opegeg)  genString("==");
  else if (vop == opiegal) genString("<=");
  else if (vop == opinfe)  genString("<");
  else if (vop == opsegal) genString(">=");
  else if (vop == opsupe) genString(">");
  else if (vop == opiieg) genString("<=");
  else if (vop == opet)   genString("and");
  else if (vop == opou)   genString("or");
  else if (vop == opouex) genString("xor");
  else if (vop == opnega) genString("-");
  else if (vop == opnon) genString("not");
  else genString("NYI_OPERATOR");
  genSpace();
}


void SMTPrinter::genString(string vstr)
{
  ofile << vstr.c_str();
}


void SMTPrinter::genIdentifier(ptexpcal vsig)
{
  genIdent(getSigId(vsig));
}

/**
   It generates a 'when expression' (exp).
*/
void SMTPrinter::genWhen(ptexpcal exp)
{
  genITEClock(EM->getContributionClock (exp));
  genSpace();
  genExpression (EM->getWhenOrCellArg (exp));
  genSpace();
}

/**
   @return the translation of a constant expression (vexp).
   \see constantExpr2SyndexTree().
*/
void SMTPrinter::genConstantExpression(ptexpcal vexp)
{
  noperators vop;

  vop = EM->getOperator(vexp);
  if ((vop == optrue) || (vop == opfalse)) 
    genOperator(vop);
  else
    genString(repstr2string(EM->getConstant (vexp)));
}

/**
   It generates an unary expression (vexp).
*/
void SMTPrinter::genUnaryExpression(ptexpcal vexp)
{
  noperators vop;

  vop = EM->getOperator(vexp);
  genOpenBlock();
  genOperator(vop);
  genExpression(EM->getUnaryArg(vexp));
  genCloseBlock();
}

/**
   It generates a binary expression (vexp).
*/
void SMTPrinter::genBinaryExpression(ptexpcal vexp)
{
  ptexpcal arg1, arg2;
  noperators vop;
  ptinsth vh;

  vop = EM->getOperator(vexp);
  arg1 = EM->getBinaryArgstAt(vexp, 1);
  arg2 = EM->getBinaryArgstAt(vexp, 2);
  genOpenBlock();
  genOperator(vop);
  genExpression(arg1);
  genSpace();
  genExpression(arg2);
  genCloseBlock();
}

/**
   @returns the number of terms in the default expression(exp)
*/
int SMTPrinter::genDefaultExpression (ptexpcal exp)
{
  int n;

  if (EM->getOperator (exp) == opmele)
    {
      n = genDefaultExpression (EM->getDefaultArgstAt (exp, 1));
      genWhen(EM->getDefaultArgstAt (exp, 2));
      n++;
    }
  else
    {
      genWhen(exp);
      n = 1;
    }
  return n;
}

/**
   It generates a default expression (e).
*/
void SMTPrinter::genDefault(ptexpcal e)
{
  int nbblocks = genDefaultExpression (EM->getDefaultArgstAt (e, 1));
  // The last one is generated without the clock condition.
  genExpression (EM->getWhenOrCellArg (EM->getDefaultArgstAt (e, 2)));
  for (int i=1; i <= nbblocks; i++) genCloseBlock();
}

/**
   It generates a ternary expression (vexp).
*/
void SMTPrinter::genTernaryExpression(ptexpcal vexp)
{
  genString("(genTernaryExpression xxxx)");
  /*
  ptexpcal arg1, arg2, arg3;
  arg1 = EM->getTernaryArgstAt(vexp, 1);
  arg2 = EM->getTernaryArgstAt(vexp, 2);
  arg3 = EM->getTernaryArgstAt(vexp, 3);

  genExpression(arg1);
  genSpace();
  genExpression(arg2);
  genSpace();
  genExpression(arg3);
  */
}

/**
   It generates a function call expression (vexp).
*/
void SMTPrinter::genFunctionCallExpression(ptexpcal vexp)
{
  ptident id;
  ptexpcal vexpcal1, vexpcal2;
  ptexpcal elem;
  int i;

  EM->getFunctionCall(vexp, &id, &vexpcal2, &vexpcal1);	// vexpcal2 : parameters; vexpcal1: inputs
  i = 1;
  while (EM->getNaryArgs(&vexpcal1, &elem))
    {
      // Separator ???
      genExpression(elem);
      i++;
    }
}

/**
   It generates a list expression (vexp).
*/
void SMTPrinter::genListExpression(ptexpcal vexp)
{
  ptexpcal elem;
  while (EM->getNaryArgs(&vexp, &elem))
    {
      // Separator ???
      genExpression(elem);
    }
}

/**
   It generates a constant signal (vexp).
   \see genEqu()
*/
void SMTPrinter::genSignal(ptexpcal vexp)
{
  genIdentifier(vexp);
}

void SMTPrinter::genExpression(ptexpcal vexp)
{
  ptexpcal nexp;

  if (vexp == NILEXPCAL)
    return;
  if (EM->getIndirection(vexp, &nexp)) vexp = nexp; // useful ?
  switch ((int)(EM->getOperatorClass(vexp)))
    {
    case variabl:
      genSignal(vexp);
      break;

      // reference to a constant expression.
    case constant:
      genConstantExpression(vexp);
      break;

      // Unary expression : not , -
    case oper1:
      genUnaryExpression(vexp);
      break;

      // Binary Expression : +,*,...
    case oper2:
      if (EM->satisfiesList(vexp))
	genListExpression(vexp);
      else
	genBinaryExpression(vexp);
      break;

      // Delay/Window (perhaps filtered before)
    case oper3:
      genTernaryExpression(vexp);
      break;

      // function call id( i1,...., in)
    case call:
      genFunctionCallExpression(vexp);
      break;

      // E1 when clk ( E1 cell clk has been filtered before)
    case opereh:
      genWhen(vexp);
      break;

    case oper2tp:		// default operator
      genDefault(vexp);
      break;

      // no occurrence
    case oper1tp:
    case oper3tp:
    case opertyp:
      break;

    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
}

void SMTPrinter::genClockId(ptinsth vclk)
{
  genIdent(getClockId(vclk));
}


void SMTPrinter::genIfClock(ptinsth vclk)
{
  ofile << "(=> ";
  genClockId(vclk);
}

void SMTPrinter::genOpenBlock()
{
  ofile << "(";
}

void SMTPrinter::genCloseBlock()
{
  ofile << ")";
}

/*
  It generates the definition: (=> vclk (= id E)).
  Signal: id := E , id ^= vclk
 */
void SMTPrinter::genDefinition(ptinsth vclk, ptident id, ptexpcal e)
{
  boolean b;
  b = indent;
  indent = true;
  genIfClock(vclk);
  genString("(= ");
  genString(ident2string(id));
  genString(" ");
  genExpression(e);
  genCloseBlock();
  indent = b;
}

void SMTPrinter::genSubFormula(ptinsth vh)
{
  ptinsth hcond, vharg1;
  ptident idcond;
  ptexpcal vcond;

  hcond = CKM->getClockSub(vh);
  CKM->getComplInFather(vh, &vharg1); // vharg1 = when not C, vh = when C
  vcond = CKM->getConditionSub(vh);
  idcond = getIdentCondition(vcond);
  // -- sub
  genBeginAssign();
  genClockId(vh);
  genSpace();
  genOpenBlock();
  genString("and");
  genSpace();
  genClockId(hcond);
  genSpace();
  genIdent(idcond);
  genString("))");
  genNewLine();
  // -- nosub
  genBeginAssign();
  genClockId(vharg1);
  genSpace();
  genOpenBlock();
  genString("and");
  genSpace();
  genClockId(hcond);
  genSpace();
  genString("(not ");
  genIdent(idcond);
  genString(")))");

  // -- when (f(x1,...)), f(x1,...x2) is replaced by a signal
 if (EM->getAttribute(vcond, idconditionattr, (attributeValue *) & idcond))
    {
      genNewLine();
      genDefinition(hcond, idcond, vcond);
    }
}

boolean SMTPrinter::genFormula(ptinsth vh)
{
  ptinsth vharg1, vharg2;
  boolean b;

  if (CKM->satisfiesMarked(vh, vMarkSeen)) return false;
  b = true;
  CKM->addMark(vh, vMarkSeen);
  switch ((int)(CKM->getOperator(vh)))
    {
    case hcomp:
      ofile << " hcomp detected !!!! ";
      break;

    case hnot:
      ofile << " hnot detected !!!! ";
      break;

    case sub:
      genSubFormula(vh);
      break;

    case infer: 
      // (= vh (and vharg1 vharg2)) || (= vh (and vharg1 (not vharg2)))
      genBeginAssign();
      genClockId(vh);
      genSpace();
      genOpenBlock();
      genString("and");
      genSpace();
      
      if (CKM->getMinusArgs(vh, &vharg1, &vharg2))
	{
	  vharg1 = CKM->getRepresentativeFormula(vharg1);
	  vharg2 = CKM->getRepresentativeFormula(vharg2);
	  genClockId(vharg2);
	  genSpace();
	  genString("(not ");
	  genClockId(vharg1);
	  genString(")))");
	}
      else
	{
	  CKM->getInfArgs(vh, &vharg1, &vharg2);
	  vharg1 = CKM->getRepresentativeFormula(vharg1);
	  vharg2 = CKM->getRepresentativeFormula(vharg2);
	  genClockId(vharg1);
	  genSpace();
	  genClockId(vharg2);
	  genString("))");
	}
      break;

    case super:
      CKM->getSupArgs(vh, &vharg1, &vharg2);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      vharg2 = CKM->getRepresentativeFormula(vharg2);

      genBeginAssign();
      genClockId(vh);
      genSpace();
      genOpenBlock();
      genString("or");
      genSpace();
      genClockId(vharg1);
      genSpace();
      genClockId(vharg2);
      genString("))");
      
      break;

    default:
      b = false;
    }
  if (b) genNewLine();
  return true;
}

void SMTPrinter::genClockDef(ptinsth clk)
{
  ptinsth clk1;

  clk = CKM->getRepresentativeFormula(clk);
  if (genFormula(clk))
    {
      clk1 = CKM->getNextFormula(clk, pk_true);
      while (clk1 != clk)
	{
	  genFormula(clk1);
	  clk1 = CKM->getNextFormula(clk1, pk_true);
	}

      clk1 = clk;
      while (CKM->getNextAssertion(clk1, &clk1))
	{
	  // an assertion. TODO
	}
    }
}

void SMTPrinter::genNewLine()
{
  ofile << endl;
  if (indent) ofile << tab;
  if (inComment) genBeginComment();
}

void SMTPrinter::genFuncInit(ptident id)
{
  genNewLine();
  genOpenBlock();
  genString("define-fun");
  genSpace();
  genIdent(id);
  genSpace();
  genString("()");
  genType(BooleanType);
  genNewLine();
  genOpenBlock();
  genString("and");
  indent = true;
  genNewLine();
}

void SMTPrinter::genFuncFinal()
{
  indent = false;
  genCloseBlock(); // closing "( and"
  genNewLine();
  genSpace();
  genCloseBlock(); // closing define-fun
  genNewLine();
}

void SMTPrinter::genType(TypeType TT)
{
  // TODO : to be completed
  genSpace();
  if (TT == BooleanType) genString("Bool");
  else if (TT == PureType) genString("Bool");
  else if (TT == ShortType) genString("Int");
  else if (TT == IntegerType) genString("Int");
  else if (TT == LongType) genString("Int");
  else genString("TT_NYI");
}

void SMTPrinter::genIdentDecl(string id, TypeType TT, boolean bline)
{
  genSpace();
  genOpenBlock();
  genString("declare-fun");
  genSpace();
  genString(id);
  genSpace();
  genString("()");
  genType(TT);
  genCloseBlock();
  if (bline) genNewLine();
}

void SMTPrinter::genClockIdentDecl(ptinsth vh)
{
  printSignals(vh); // print the synchronous signals with vh.
  genIdentDecl(ident2string(getClockId(vh)), BooleanType, pk_false);
  genNewLine();
}

/**
   Assumed : e is a boolean expression.
   Generates a new signal when the expression e is not an event or a signal.
*/
void SMTPrinter::genConditionDecl(ptexpcal e)
{
  ptident id;
  if (!EM->IsEvent(e))
    {
      if (!EM->IsSignal(e))
	{
	  if (!EM->getAttribute(e, idconditionattr, (attributeValue *) &id))
	    {
	      id = NAMES->exclusiveIdent(conditionId);
	      EM->addAttribute(e, idconditionattr, (attributeValue) id);
	    }
	}
      else 
	id = EM->getName(e);
      genIdentDecl(ident2string(id), BooleanType, pk_true);
    }
}

boolean SMTPrinter::declareFormula(ptinsth vh)
{
  ptinsth vharg1, vharg2;

  if (CKM->satisfiesMarked(vh, vMarkSeen))
    return false;
  CKM->addMark(vh, vMarkSeen);
  genClockIdentDecl(vh);
  switch ((int)(CKM->getOperator(vh)))
    {
    case sub:
      genConditionDecl(CKM->getConditionSub(vh));
      CKM->getComplInFather(vh, &vharg1);
      declareFormula(vharg1);
      declareFormula(CKM->getClockSub(vh));
      break;

    case infer:
      if (!CKM->getMinusArgs(vh, &vharg1, &vharg2))
	CKM->getInfArgs(vh, &vharg1, &vharg2);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      vharg2 = CKM->getRepresentativeFormula(vharg2);
      declareFormula(vharg1);
      declareFormula(vharg2);
      break;

    case super:
      CKM->getSupArgs(vh, &vharg1, &vharg2);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      vharg2 = CKM->getRepresentativeFormula(vharg2);
      declareFormula(vharg1);
      declareFormula(vharg2);
      break;
    default:;
    }
  return true;
}

void SMTPrinter::printSignals(ptinsth vclk)
{
  ptexpcal vsig;
  ptlinstances lsig;

  lsig = CKM->getSynchronousSignals(vclk);
  if (lsig != NILLINSTANCES)
    {
      genString("; Clock of ");
      while (ptlinstances_get_next(&lsig, &vsig))
	{
	  genIdent(getSigId(vsig));
	  if (lsig != NILLINSTANCES)
	    genString(", ");
	}
      genNewLine();
    }
}

void SMTPrinter::genClockDecl(ptinsth clk)
{
  ptinsth clk1;

  clk = CKM->getRepresentativeFormula(clk);
  declareFormula(clk);

  clk1 = CKM->getNextFormula(clk, pk_true);
  while (clk1 != clk)
    {
      declareFormula(clk1);
      clk1 = CKM->getNextFormula(clk1, pk_true);
    }

  clk1 = clk;
  while (CKM->getNextAssertion(clk1, &clk1))
    {
      // an assertion. TODO
    }
}

void SMTPrinter::genClockDecls(ptlinsth list)
{
  ptinsth vh;

  TRACE(cout << " SMTPrinter::declareClocks lh = " << list << endl);
  while (ptlinsth_get_next(&list, &vh))
    if (!ptlinsth_isin_elem(unusedClocks, vh))
      genClockDecl(vh);
}

void SMTPrinter::genClockDefs(ptlinsth list)
{
  ptinsth vh;

  while (ptlinsth_get_next(&list, &vh))
    if (!ptlinsth_isin_elem(unusedClocks, vh))
      genClockDef(vh);
}

void SMTPrinter::genClassConstraint(ptinsth clk)
{
  ptinsth clk1;

  clk = CKM->getRepresentativeFormula(clk);
  clk1 = CKM->getNextFormula(clk, pk_true);
  while (clk1 != clk)
    {
      addConstraint(clk1, clk);
      clk1 = CKM->getNextFormula(clk1, pk_true);
    }

  // Assertions 
  clk1 = clk;
  while (CKM->getNextAssertion(clk1, &clk1))
    {
      // TODO
    }
}

void SMTPrinter::genConstraints(ptlinsth list)
{
  ptinsth vh;

  while (ptlinsth_get_next(&list, &vh))
    if (!ptlinsth_isin_elem(unusedClocks, vh))
      genClassConstraint(vh);
}

void SMTPrinter::addConstraint(ptinsth vh1, ptinsth vh2)
{
  if (addFuncForConstraints && ! prog_with_constraints)
    {
      genFuncInit(idGraphConstraints);
      prog_with_constraints = true;
    }

  genBeginAssign();
  genClockId(vh1);
  genSpace();
  genClockId(vh2);
  genCloseBlock();
  genNewLine();
}

void SMTPrinter::genConstraintsOnPlacedClocks()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  CKM->getConstraintsOnPlacedClocks(&lh1, &lh2);
  while (ptlinsth_get_next(&lh1, &vh1))
    {
      ptlinsth_get_next(&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula(vh1);
      vh2 = CKM->getRepresentativeFormula(vh2);
      if (vh1 != vh2)
	addConstraint(vh1, vh2);
    }
  /*    TODO Later Loic 
     if (isEventClockedFormat (graph_type))
     {
     CKM->getAssertionsOnPlacedClocks (&lh1, &lh2);
     while (ptlinsth_get_next (&lh1, &vh1))
     {
     ptlinsth_get_next (&lh2, &vh2);
     vh1 = CKM->getRepresentativeFormula (vh1);
     vh2 = CKM->getRepresentativeFormula (vh2);
     if (vh1 != vh2)
     addAssertion (vh1, vh2);
     }
     } */
}

ptident SMTPrinter::getMemoryId(ptident idSig, int num)
{
 ptident id;
 id = makeIdent_ident_ident(idMem,  idSig);
 return id;
 // id = makeIdent_ident_string(id, "!");
 // return makeIdent_ident_int (id, num);
}

ptident SMTPrinter::getMemoryInitId(ptident id)
{
  id = makeIdent_ident_string(id, "!");
  return makeIdent_ident_int (id, 0);
}

ptident SMTPrinter::getMemoryNextId(ptident id)
{
  id = makeIdent_ident_string(id, "!");
  return makeIdent_ident_string (id, "next");
}

void SMTPrinter::genSigDeclDelay(ptexpcal vsig)
{
  ptINode aNode;
  ptident id;
  ptexpcal resret, sigret,  vret, vfen, vinit;

  aNode = EM->getNode(vsig);
  if (NM->is_delay(aNode))
    {
      NM->get_delay_arg (aNode, &resret, &sigret, &vret, &vfen, &vinit);
      id = getMemoryId(getSigId(sigret), NM->get_num(aNode));
      TypeType TT =  EM->getType(sigret);
      genIdentDecl(ident2string(id), TT, pk_true); // M.x
      genIdentDecl(ident2string(getMemoryInitId(id)), TT, pk_true); // M.x0
      genIdentDecl(ident2string(getMemoryNextId(id)), TT, pk_true);  // M.x_next
    }
}
/**
   @return true if the signal must be declared, "false" otherwise. 
*/
boolean SMTPrinter::mustBeDeclared(ptexpcal vsig)
{
  // ptinsth h;
  //  if (EM->getExtractionLink(vsig, &h)) return false;
  //  h = EM->getClock(vsig);
  return true; 
  // return ! eq_idents( EM->getName(vsig), CKM->getName(h));
}


void SMTPrinter::genSigDecl(ptexpcal vsig)
{
  if (isSelectedType(vsig))
    {
      if (mustBeDeclared(vsig))
	genIdentDecl(ident2string(getSigId(vsig)), EM->getType(vsig), pk_true);
      else; // already declared
      genSigDeclDelay(vsig);
    }
}

void SMTPrinter::genSigDecls(ptlinstances lsig)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&lsig, &vsig))
    genSigDecl(vsig);
}

boolean SMTPrinter::isSelectedNode(ptINode n)
{
  return ! ptlINode_isIn(unusedNodes, n);
}

boolean SMTPrinter::isSelectedType(ptexpcal vsig)
{
  if (! restrictedToBooleans) return pk_true;
  TypeType TT = EM->getType(vsig);
  return (( TT == BooleanType) || ( TT == PureType ));
}

void SMTPrinter::genEqu(ptexpcal leftPart, ptexpcal rightPart)
{
  genSpace();
  genBeginAssign();
  genExpression(leftPart);
  genSpace();
  genExpression(rightPart);
  genCloseBlock();
}

void SMTPrinter::genCell(ptexpcal leftPart, ptexpcal rightPart)
{
  genSpace();
  genBeginAssign();
  genExpression(leftPart);
  genSpace();
  genCloseBlock();
}

void SMTPrinter::genEqu(ptinsth vh, ptINode aNode)
{
  ptexpcal leftPart, rightPart, basicExpr;

  if (NM->get_rightOperand(aNode, &rightPart))
    {
      genIfClock(vh);
      leftPart = NM->get_leftOperand(aNode);
      if (EM->IsCellExpr(rightPart))	// Cell
	{
	  genCell(leftPart, rightPart);
	}
      else
	{
	  basicExpr = EM->getBasicExpr(rightPart);
	  genEqu(leftPart, basicExpr);
	}
      genCloseBlock();
      genNewLine();
    }

}

void SMTPrinter::genDelay(ptinsth vh, ptINode node)
{
  ptident idMemory, idInitialValue, idNextValue;
  ptexpcal resret, sigret,  vret, vfen, vinit;
  
  NM->get_delay_arg (node, &resret, &sigret, &vret, &vfen, &vinit);
  idMemory = getMemoryId(getSigId(sigret), NM->get_num(node));
  idNextValue = getMemoryNextId(idMemory);
  
  // initial value
  idInitialValue = getMemoryInitId(idMemory);
  genBeginAssign();  
  genIdent(idInitialValue);
  genSpace();
  genExpression(vinit);
  genCloseBlock();
  genNewLine();

  // zx := mem(x)
  genIfClock(vh);
  genSpace();
  genBeginAssign();
  genIdentifier(resret);
  genSpace();
  genIdent(idMemory);
  genCloseBlock();
  genCloseBlock();
  genNewLine();

 // next 
  genIfClock(vh);
  genSpace();
  genBeginAssign();
  genIdent(idNextValue);
  genSpace();
  genIdentifier(sigret);
  genCloseBlock();
  genCloseBlock();
  genNewLine();

}

void SMTPrinter::genNotImpl(ptinsth vh, ptexpcal vsig)
{
  ofile << "Definition of ";
  genIdentifier(vsig);
  ofile << " is not yet implemented";
  genNewLine();
}

void SMTPrinter::genSigDef(ptexpcal vsig)
{
  ptinsth vh;
  ptINode aNode;

  aNode = EM->getNode(vsig);
  if (isSelectedType(vsig) && isSelectedNode(aNode))
    {
      vh = EM->getClock(vsig);
      switch ((int)(NM->get_oper(aNode)))
	{
	case opdefsig:
	  genEqu(vh, aNode);
	  break;

	case opret:
	  genDelay(vh, aNode);
	  break;

	case opldefpart:
	  genNotImpl(vh, vsig);
	  break;

	case opdefpart:
	  genNotImpl(vh, vsig);
	  break;

	case opsubgraph:
	  genNotImpl(vh, vsig);
	  break;

	case opmoti:
	  genNotImpl(vh, vsig);
	  break;

	case opbund:
	  genNotImpl(vh, vsig);
	  break;

	case opappe:
	case opnodecall:
	case opnil:
	  break;

	default:
	  pK_messenger->Caseerror(currentLine, currentFile);
	}
    }
}

void SMTPrinter::genSigDefs(ptlinstances lsig)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&lsig, &vsig))
    genSigDef(vsig);
}

boolean SMTPrinter::existIntegerSignals(ptlinstances lsig)
{
  TypeType TT;
  ptexpcal vsig;

  if (restrictedToBooleans) return pk_false;
  while (ptlinstances_get_next(&lsig, &vsig))
    {
      TT = EM->getType(vsig);
      if (TT == ShortType)
	return pk_true;
      if (TT == IntegerType)
	return pk_true;
      if (TT == LongType)
	return pk_true;
    }
  return pk_false;
}


void SMTPrinter::genComment(string vcomment)
{
  genBeginComment();
  genString(vcomment);
}

boolean SMTPrinter::getPreviousName(ptinsth vh, ptident *idprev)
{
  ptinsth vhelem;
  ptlinsth lh;
  ptlPointer lids;
  pointer x;

  lids = lPreviousNames;
  lh = lhWithNewNames;
  while (ptlinsth_get_next(&lh, &vhelem))
    {
      ptlPointer_get_elemNext (&lids, &x);
      if (vhelem == vh)
	{
	  *idprev = (ptident&)x;
	  return pk_true;
	}
    }
  return pk_false;
}


void SMTPrinter::genMappingOfVariables()
{
  ptinsth vh, vhnew;
  ptident id_vhnew, id_vh;
  ptlinsth lh;

  lh = lClocksPreviousSystem;
  if (lh != NILLINSTH)
    {
      ofile << "; ---- Mapping of variables";
      genNewLine();
      while (ptlinsth_get_next (&lh, &vh))
	{
	  vhnew = CKM->getRepresentativeFormula(vh);
	  // vh has been rewritten in vhnew ?
	  if (vhnew != vh )
	    {
	      id_vh = CKM->getName(vh);
	      id_vhnew = CKM->getName(vhnew);
	      if ( ! eq_idents(id_vh, id_vhnew))
		{
		  ofile << "(= " << ident2string(getClockId(vh)) << " " << ident2string(getClockId(vhnew)) << ")";
		  genNewLine();
		}
	    }
	  else
	    if (getPreviousName(vh, &id_vh))
	      {
		id_vhnew = CKM->getName(vhnew);
		ofile << "(= " << ident2string(id_vh) << " " << ident2string(getClockId(vhnew)) << ")";
		genNewLine();
	      }
	}
    }
}


void SMTPrinter::final()
{
  CKM->resetMarkAll(vMarkSeen);
  EM->resetMarkAll(idconditionattr);
  EM->resetMarkAll(vMarkSeen);
}


void SMTPrinter::init()
{
  idconditionattr = getMark();
  vMarkSeen = getMark();
}

/**
   Open a file foo_decls.smt (when bdecl)  or foo_defs.smt otherwise
 */
boolean SMTPrinter::openFile(boolean bdecl)
{
  ptrepstr filename;
  char * vfile;
  ptrepstr idfile;

  idfile = ident2repstr(makeIdent_ident_string(idGraph,  (bdecl ? "Decls" : "Defs")));
  if (!sys_suffixedFilePathInUserOutputDir(idfile, "smt", &filename)) return pk_false;
  vfile = repstr2string(filename);
  ofile.open (vfile);
  pK_messenger->globalMessageStrStr(messGeneratorFile,string2repstr("SMT declaration") ,filename);
  // cout << "     * SMT code generated in " << vfile << endl;
  return pk_true;
}

void SMTPrinter::closeFile()
{
  ofile.close();
}

void SMTPrinter::genClockDecls()
{
  genComment(" ---------------- Clocks");
  genNewLine();
  genClockDecls(lroot_CurrentDCG);
  genClockDecls(lplaced_CurrentDCG);
  CKM->resetMarkAll(vMarkSeen);
}


void SMTPrinter::genClockDefs()
{
  genClockDefs(lroot_CurrentDCG);
  genClockDefs(lplaced_CurrentDCG);
}

void SMTPrinter::genConstraints(boolean b)
{
  addFuncForConstraints = b;
  genConstraints(lroot_CurrentDCG);
  genConstraints(lplaced_CurrentDCG);
  genConstraintsOnPlacedClocks();
}

/* void SMTPrinter::genFuncConstraints()
{
  prog_with_constraints = false;
  genConstraints(pk_true);
  if (prog_with_constraints)
    {
      genFuncFinal();
      genValidating(idGraph, idGraphConstraints, false);
    }
}
*/

void SMTPrinter::genSigDecls()
{
  if (lsig_CurrentDCG != NILLINSTANCES)
    {
      genComment(" ---------------- Signals");
      genNewLine();
      genSigDecls(lsig_CurrentDCG);
    }
}

void SMTPrinter::genSigDefs()
{
  if (lsig_CurrentDCG != NILLINSTANCES)
    {
      genSigDefs(lsig_CurrentDCG);
    }
}

/**
   Generate the declarations.
*/
boolean SMTPrinter::genDeclarations()
{
  if (! openFile(true)) return pk_false;
  genClockDecls();
  genSigDecls();
  closeFile();
  return pk_true;
}

/**
   Generate the definitions in a function.
   (define-fun P () Bool
    (and  eq1 eq2 ...eqn)
    )
 */
boolean SMTPrinter::genDefinitions(boolean bWithConstraints)
{
  if (!openFile(false)) return pk_false;
  genFuncInit(idGraph);
  genClockDefs();
  if (bWithConstraints) genConstraints(pk_false);
  genSigDefs();
  genMappingOfVariables();
  genFuncFinal();
  indent = true;
  closeFile();
  return pk_true;
}


void SMTPrinter::setCurrentGraph(ptDCG vDCG)
{
  CurrentDCG =  vDCG;
  CKM = CurrentDCG->getClockManager ();
  EM =CurrentDCG->getExpressionManager ();
  NM = CurrentDCG->getNodeManager ();
  NAMES = CurrentDCG->getNameManager ();
  unusedNodes =  CurrentDCG->getUnused(&unusedClocks);

  lplaced_CurrentDCG = CKM->getlhplaced ();
  lroot_CurrentDCG = CKM->getlhroots ();
  idGraph = CurrentDCG-> getLevelName ();
  idGraphConstraints = makeIdent_ident_string (idGraph, "Constraints" );
  lsig_CurrentDCG = NILLINSTANCES;
  EM->getExprList(opident, &lsig_CurrentDCG);
}
