/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor.*/
SMTService::SMTService(ptDCG_basic g)
{
  vDCG = g;
  BERROR = pk_false;
}

/** Destructor.*/
SMTService::~SMTService()
{
}

/**
   @return "true" if the operation succeeds, "false" otherwise (operator not yet implemented,...).
*/
boolean SMTService::translation(boolean trace)
{
  ptDCG_poly gPoly;
  ptDCG_basic G;

  BERROR = pk_false;

  // vDCG->groupPartialDefinitions();
  vDCG->undersamplingOnSignal(pk_false, pk_false);
  if (vDCG->buildSMTAbstraction(&G))
    {
      G->setSubstituableClocks ();
      exportedDCG = G -> getDCG();
      G->setFormatLevel(vDCG->getFormatLevel ());
      if (trace)
	{
	  gPoly = DCG_poly::new_prePoly (exportedDCG);
	  gPoly-> setIntrinsicClocks ();
	  gPoly -> unparse ();
	  gPoly ->initGraphForModifications ();
	}

      return pk_true;
    }
  pK_messenger->printString("ERROR during buildSMTAbstraction");
  pK_messenger->printString("SMTService: Program not exported to SMT.");
  BERROR = pk_true;
  return pk_false;
}

/**
   @return "true" if the operation succeeds, "false" otherwise (error);
   if there is no detected before this calling, it prints 
   - the properties to prove (foo_CMD.z3z file, see printPropertiesToProve())
   - the expression of the synchronization in z3z (foo.z3z file, see unparse_all())
*/
boolean SMTService::print()
{
  boolean b;
  ptlinsth lhUnused, lhPreviousSystem, lhModifiedNames;
  ptlPointer lprevNames;
  ptTraceabilityManager tm = vDCG->getTraceabilityManager ();

  ptSMTPrinter printer = new SMTPrinter();
  lhPreviousSystem =  tm ->getClocksPreviousSystem();
  vDCG->getUnused(&lhUnused);
  lhPreviousSystem = ptlinsth_list_remove (lhPreviousSystem, lhUnused);
  tm ->getClocksWithNewName(lhPreviousSystem, &lhModifiedNames, &lprevNames);

  b = printer->genSMT(exportedDCG, lhPreviousSystem, lhModifiedNames, lprevNames);
  delete printer;
  return b;
}
