

/* 
   Authors: LoÃÂ¯c Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_ boolean DCG_Sigali::verifDelayOfOne (ptlINode ln)
{
  ptNodeManager NM;
  ptExprManager EM;
  ptINode node;
  ptexpcal e, x, vret, vfen, vinit;
  compilerInteger val;

  NM = getNodeManager ();
  EM = getExpressionManager ();
  while (ptlINode_get_elem (&ln, &node))
    {
      NM->get_delay_arg (node, &x, &e, &vret, &vfen, &vinit);
      // Restriction: delay 1 assumed.
      if (!EM->getIntegerValue (vret, &val)) return pk_false;
      if (val != 1) return pk_false;
    }
  return pk_true;
}


/**  State variables.
 */
boolean
DCG_Sigali::verifStateVariables ()
{
  ptINode node;
  ptlINode ln;
  ptNodeManager NM;
  ptExprManager EM;
  ptexpcal id;
  ptlinstances lids;

  // a REVOIR: les variables d'etat doivent etre exportées. 
  NM = getNodeManager ();
  EM = getExpressionManager ();
  if (EM->getExprList (opident, &lids))
    while (ptlinstances_get_next (&lids, &id))
      if (EM->isStateVar (id))
	{
	  node = EM->getNode (id);
	  if (!NM->is_setOfPartialDefinitions (node))
	    return pk_false;
	  ln = NM->get_setOfPartialDefinitions_arg (node);
	  if (ptlINode_get_size (ln) > 1)
	    return pk_false;
	}
  return pk_true;
}

/**
   @return "true" if all the nodes of ln are marked by the predefined SIGALI pragma, "false" otherwise.
*/
boolean
DCG_Sigali::allTaggedSigali (ptlINode ln)
{
  ptNodeManager NM;
  ptINode node;

  NM = getNodeManager ();
  while (ptlINode_get_elem (&ln, &node))
    if (!existPredefinedPragma (NM->getPragmas (node), PRAGMA_SIGALI))
      return pk_false;
  return pk_true;
}

boolean
DCG_Sigali::isTranslatable ()
{
  ptlINode lNodes;
  ptNodeManager NM;

  NM = getNodeManager ();
  if (NM->getDelays (&lNodes))
    if (!verifDelayOfOne (lNodes))
      {
	pK_messenger->printString ("SigaliService::translation: program with delay > 1.");
	return pk_false;
      }

  if (NM->getProcCalls (&lNodes))
    if (!allTaggedSigali (lNodes))
      {
	pK_messenger->printString ("SigaliService::translation: program with proc call.");
	return pk_false;
      }
  if (NM->getNodeCalls (&lNodes))
    if (!allTaggedSigali (lNodes))
      {
	pK_messenger->printString ("SigaliService::translation: program with node call.");
	return pk_false;
      }

  if (NM->getSubGraphs (&lNodes))
    {
      pK_messenger->printString ("SigaliService::translation:program with sub-Graph references. ");
      return pk_false;
    }


  if (NM->getGroupsOfPartialEquations (&lNodes))
    {
      pK_messenger->printString ("SigaliService::translation: program with partial definitions. ");
      return pk_false;
    }

  if (NM->getBundles (&lNodes))
    {
      pK_messenger->printString ("SigaliService::translation: program with bundle definitions.");
      return pk_false;
    }

  if (NM->getIterates (&lNodes))
    {
      pK_messenger->printString ("SigaliService::translation: program with array/iterate definitions.");
      return pk_false;
    }

  if (!verifStateVariables ())
    {
      pK_messenger->printString ("SigaliService::translation: program with state var definitions.");
      return pk_false;
    }
  return pk_true;
}

/** 
    The name of a clock (vhx) is changed when it is unified or represented by a signal (x).
    @return true if the name of the clock is changed, false otherwise.
 */
boolean DCG_Sigali::changeName(ptinsth vhx, ptexpcal x)
{
  ptexpcal aSig;
  ptClockManager CKM;
  ptExprManager EM;
  ptident id, idclk;;

  CKM = getClockManager ();
  EM = getExpressionManager ();
  if (EM->IsSignal(x))
    {
      idclk = CKM -> getName(vhx);
      id = EM->getName(x);
      if (eq_idents(id, idclk))
	{
	  id = getNameManager()->new_pureClockIdent (CKM->getNum (vhx));
	  CKM -> setName(vhx, id);
	  CKM ->setClockRepresentedBySignal(vhx, NILEXPCAL);
	  return pk_true;
	}
    }
  return pk_false;
}

/** 
    Modify the name of the clocks for Sigali. If a clock is represented by a signal
    - and this signal is a delay, it must be changed
    - and this signal is an input signal, it must be changed.
 */
void DCG_Sigali::modifyClockNames()
{
  // Added Jan 2017 (lbesnard)
  ptlINode lNodes;
  ptNodeManager NM;
  ptExprManager EM;
  ptINode node;
  ptexpcal zx, x, vret, vfen, vinit;

  NM = getNodeManager ();
  EM = getExpressionManager ();

  if (NM->getDelays (&lNodes))
    {
      while (ptlINode_get_elem (&lNodes, &node))
	{
	  NM->get_delay_arg (node, &zx, &x, &vret, &vfen, &vinit);
	  changeName(EM->getClock(zx), x);
	}
    }
  ptlinstances ls;
  ls = getInputs();
  while (ptlinstances_get_next (&ls, &x))
    {
      changeName(EM->getClock(x), x);
    }
}
