/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
#define cstrcommand "_CMD"
/**
   @return the string associated with the first element of an assumed list of expressions (lexp).
   It assumed that the fisrt element is a string.
*/
  ptrepstr SigaliService::getstrOperand (ptexpcal lexp)
{
  ptINode
    n;
  ptexpcal
    vname,
    exp2,
    expd;

  vname = EM->getNaryArgs (&lexp);
  if (EM->getOperator (vname) != opchai)
    {
      n = EM->getNode (vname);
      NM->get_rightOperand (n, &expd);
      expd = EM->getWhenOrCellArg (expd);
      EM->getBinaryArgs (expd, &vname, &exp2);
    }
  return ExtractString (EM->getConstant (vname));
}

/**
   It generates Sigali orders for a node (aNode) if it is not yet done. 
   It recalls recursively this function on the input operand of the node.
*/
void
SigaliService::translateOrder (ptINode aNode)
{
  ptident name;
  ptexpcal vsig, OrderedParams, OrderedInputs, OrderedOutputs, e;
  ptDCG aGraph;
  ptINode n;
  ptpragma valPragma;

  if (!NM->isMarked (aNode, attributeNumber))
    {
      NM->addMark (aNode, attributeNumber);
      if (NM->isProcessCall (aNode))
	{
	  NM->get_procCall_arg (aNode, &name, &OrderedParams, &OrderedInputs, &OrderedOutputs);	// Proc Call
	  if (!getPragmaStrKeyword (NM->getPragmas (aNode), SigaliPragma, &valPragma))
	    valPragma = NILPRAGMA;
	}
      else
	{
	  if (!NM->isNodeCall (aNode))
	    return;
	  NM->get_nodeCall_arg (aNode, &aGraph, &OrderedParams, &OrderedInputs, &OrderedOutputs);	// node Call
	  name = aGraph->getName ();
	  if (!getPragmaStrKeyword (aGraph->getPragmas (), SigaliPragma, &valPragma))
	    valPragma = NILPRAGMA;
	}
      e = OrderedInputs;
      while (EM->getNaryArgs (&e, &vsig))
	{
	  if (EM->IsSignal (vsig))
	    {
	      n = EM->getNode (vsig);
	      if (!NM->is_outputNodeCall (n, &aNode))
		aNode = n;
	      translateOrder (aNode);
	    }
	}
      printSigaliorder (name, OrderedParams, OrderedInputs, OrderedOutputs, valPragma);
    }
}

/**
   It generates the Sigali orders for the nodes of properties (lproperties).
   These nodes are references to SIGALI intrinsics. The operand expresses the property.
*/
void
SigaliService::mkSigaliCheckfile (ptlINode lproperties)
{
  ptINode aProperties;

  if (lproperties != NILLINODE)
    {
      while (ptlINode_get_elem (&lproperties, &aProperties))
	translateOrder (aProperties);

      NM->resetMarkAll (attributeNumber);
    }
}

/** @return the list of the nodes that reference the "SIGALI" intrinsic funtion.
    Only SIGALI nodes are selected (root of a property). For the following example,
    <em> SIGALI(Never(B_And(B_False(CSTART),B_True(CRELAX))))</em>,
    Only the node SIGALI() will be added in the lproperties list.
    The others Sigali functions will be take into account at generation step.
*/
ptlINode SigaliService::getldynamicproperties ()
{
  ptlINode
    ls;
  ptINode
    aNode;
  ptident
    nom;
  ptlINode
    lproperties;

  lproperties = NILLINODE;
  if (NM->getProcCalls (&ls))
    while (ptlINode_get_elem (&ls, &aNode))
      if (existPredefinedPragma (NM->getPragmas (aNode), PRAGMA_SIGALI))
	{
	  NM->get_nameCall (aNode, &nom);
	  if (isfuncsigali (nom))
	    ptlINode_insert_elem (&lproperties, aNode);
	}

  if (NM->getNodeCalls (&ls))
    while (ptlINode_get_elem (&ls, &aNode))
      if (existPredefinedPragma (NM->getPragmas (aNode), PRAGMA_SIGALI))
	{
	  ptDCG
	    g = NM->get_nodeCall_graph (aNode);
	  nom = g->getName ();
	  if (isfuncsigali (nom))
	    ptlINode_insert_elem (&lproperties, aNode);
	}
  return lproperties;
}


/**
   @return "true" if there are some properties to prove in the current graph, "false" otherwise.
   When there are some properties to prove, a file (foo_CMD.z3z) is generated.
   \see mkSigaliCheckfile.
*/
boolean SigaliService::printPropertiesToProve (ptlINode lproperties)
{
  ptrepstr
    fname;

  if (lproperties == NILLINODE)
    return pk_false;
  if (!makeRepstr_repstr_string (ProcessNameStr, cstrcommand, &fname))
    fname = ProcessNameStr;
  if (!sys_add_typeSuffix2repstr (fname, Z3ZSUFF, &fname))
    fname = ProcessNameStr;
  if (!sys_filePathInUserOutputDir (fname, &fname))
    return pk_false;
  shaper->initSigaliCommandFile (fname, ProcessNameStr);	// it shaper->opens CMD file and generates the 
  mkSigaliCheckfile (lproperties);	// generates the orders for directives.
  shaper->endSigaliCommandFile ();	// shaper finalizing.
  return pk_true;
}

/**
   It marks the signals referenced in a reference (aNode) to the Sigali command.
*/
void
SigaliService::markSignalsProperty (ptINode aNode)
{
  ptexpcal vsig, OrderedInputs, OrderedOutputs;
  ptINode n;

  NM->get_rightOperand (aNode, &OrderedInputs);

  if (NM->isProcessCall (aNode) || NM->isNodeCall (aNode))
    while (EM->getNaryArgs (&OrderedInputs, &vsig))
      {
	if (EM->IsSignal (vsig))
	  {
	    n = EM->getNode (vsig);
	    if (!NM->is_outputNodeCall (n, &aNode))
	      aNode = n;
	    markSignalsProperty (aNode);
	  }
      }
  else
    {
      OrderedOutputs = NM->get_leftOperand (aNode);
      if (EM->satisfiesList (OrderedOutputs))
	while (EM->getNaryArgs (&OrderedInputs, &vsig))
	  EM->addMark (vsig, refInSigaliCmd);
      else
	EM->addMark (OrderedOutputs, refInSigaliCmd);
    }
}

/**
   It marks the signals referenced in the properties (lproperties) to be proved.
   The list is the set of SIGALI commands.
   It calls markSignalsProperty().
*/
void
SigaliService::markSignalsProperties (ptlINode lproperties)
{
  ptINode aProperties;

  while (ptlINode_get_elem (&lproperties, &aProperties))
    markSignalsProperty (aProperties);
}
