/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   Access to the operands of a definition (d) defined by <em> id = p </em>.
*/
extern void
getdefpoly (ptdefpoly d, ptident * id, ptpolynomial * p)
{
  *id = d->nom;
  *p = d->def;
}


/** @return a new empty list of definitions. */
extern ptldefpoly
makeDefinitionList (void)
{
  ptldefpoly list;

  list = new ldefpoly ();
  list->blist = NILDEFPOLY;
  list->elist = NILDEFPOLY;
  list->current = NILDEFPOLY;
  return list;
}

/** It adds the defintion <em> id = p </em> at the end of the \e list. */
extern void
adddefpoly (ptldefpoly list, ptident id, ptpolynomial p)
{
  ptdefpoly d;

  d = new defpoly ();
  d->nom = id;
  d->def = p;
  d->next = NILDEFPOLY;
  if (list->elist == NILDEFPOLY)
    list->blist = d;
  else
    list->elist->next = d;
  list->elist = d;
}



/** It initialises the iterator of the list (list) to the first element. */
extern void
initIteratorldefpoly (ptldefpoly list)
{
  list->current = list->blist;
}

/**
   @return "true" if the travel is finished, "false" otherwise.
   An iterator is previously initialised by initIteratorldefpoly().
*/
extern boolean
isEndldefpoly (ptldefpoly list)
{
  return (boolean) (list->current == NILDEFPOLY);
}

/** 
    @return the current element of a list (list), and the iterator is set to the next element.
    An iterator is previously initialised by initIteratorldefpoly().
 */
extern ptdefpoly
nextdefpoly (ptldefpoly list)
{
  ptdefpoly p;

  p = list->current;
  list->current = list->current->next;
  return p;
}
