/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return the statevar associated with a delay (r).
*/
extern ptpolynomial
getDelaystatevar (ptrepDelay r)
{
  return r->statevar;
}



/**  @return the delayed signal associated with a delay (r).
 */
extern ptexpcal
getDelaySig (ptrepDelay r)
{
  return r->sig;
}


/**
   @return the clock of the delayed signal associated with a delay (r).
*/
extern ptinsth
getDelayClock (ptrepDelay r)
{
  return r->clock;
}

/** 
    @return the initial value associated with a delay (r). 
*/
extern ptpolynomial
getDelayInit (ptrepDelay r)
{
  return r->init;
}

/** @return a new empty list of delays.  */
extern ptlrepDelay
makeDelayList (void)
{
  ptlrepDelay list;

  list = new lrepDelay ();
  list->blist = NILREPDELAY;
  list->elist = NILREPDELAY;
  list->current = NILREPDELAY;
  return list;
}

/**
   It adds a delay at the end of the \e list.
   The delay is characterized by
     - its state variable (v), 
     - the delayed signal (s),
     - the clock of the signal (hs) 
     - and the initial value (i).
*/
extern void
addDelay (ptlrepDelay list, ptpolynomial v, ptexpcal s, ptinsth hs, ptpolynomial i)
{
  ptrepDelay ret;

  ret = new repDelay ();
  ret->statevar = v;
  ret->sig = s;
  ret->clock = hs;
  ret->init = i;
  ret->next = NILREPDELAY;
  if (list->elist == NILREPDELAY)
    list->blist = ret;
  else
    list->elist->next = ret;
  list->elist = ret;
}

/** It initialises the iterator of the list (list) to the first element. */
extern void
initIteratorlDelay (ptlrepDelay list)
{
  list->current = list->blist;
}



/**
   @return "true" if the travel is finished, "false" otherwise.
   An iterator is previously initialised by initIteratorlDelay().
*/
extern boolean
isEndlDelay (ptlrepDelay list)
{
  return (boolean) (list->current == NILREPDELAY);
}


/** 
    @return the current element of a list (list), and the iterator is set to the next element.
    An iterator is previously initialised by initIteratorlDelay().
 */
extern ptrepDelay
nextDelay (ptlrepDelay list)
{
  ptrepDelay r;

  r = list->current;
  list->current = list->current->next;
  return r;
}
