/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

///< polynomial representation.
typedef struct polynomial
{
  typolynomial nature;		///< type of polynomial
  union
  {
    int value;			///< for a constant
    ptident name;		///< for a variable
    ptpolynomial op;		///< for an unary
    struct
    {
      ptpolynomial op1, op2;	///< operands of a binary
    } binary;			///< for a binary
  } U;				///< depending on nature field.
} polynomial;


typedef struct elemlist
{
  ptpolynomial elem;
  ptelemlist next;
} elemlist;

///<  representation of a list of polynomials (head, current, end). Current is used an iterator.
typedef struct lpolynomial
{
  ptelemlist blist, elist, current;
} lpolynomial;


typedef struct repDelay
{
  ptpolynomial statevar;
  ptexpcal sig;
  ptinsth clock;
  ptpolynomial init;
  ptrepDelay next;
} repDelay;

///<  representation of a list of delays (head, current, end). Current is used an iterator.
typedef struct lrepDelay
{
  ptrepDelay blist, elist, current;
} lrepDelay;

typedef struct defpoly
{
  ptident nom;
  ptpolynomial def;
  ptdefpoly next;
} defpoly;

///<  representation of a list of definitions (head, current, end). Current is used an iterator.
typedef struct ldefpoly
{
  ptdefpoly blist, elist, current;
} ldefpoly;

static ptpolynomial un, polyMinusun, zero;
static ptpolynomial tabConstants[5];

/**
  @return the constant polynomial c.
*/
static ptpolynomial
makeconstant (int c)
{
  ptpolynomial poly;

  poly = new polynomial ();
  poly->nature = constpoly;
  poly->U.value = c;
  return poly;
}



/** @return "true" if a polynomial (p) is a not one or a minus one, false otherwise.
 */
static boolean
setpolyNotpolyMinus (ptpolynomial p)
{
  typolynomial t;

  t = getpolytype (p);
  return ((t == polyNot) || (t == polyMinus));
}

/** @return "true" if p1 reference p2.
 */
static boolean
polyrecursif (ptpolynomial p1, ptpolynomial p2)
{
  boolean resu;

  if (p1 == p2)
    return pk_true;

  switch ((int) (p1->nature))
    {
    case constpoly:
    case variable:
      resu = pk_false;
      break;
    case polySum:
    case polyMult:
    case polyDiff:
    case polyDefault:
    case polyWhen:
    case polyOr:
    case polyAnd:
      resu = polyrecursif (p1->U.binary.op1, p2);
      if (!resu)
	resu = polyrecursif (p1->U.binary.op2, p2);
      break;
    case polyWhen1:
    case polyNot:
    case polyMinus:
    case polySquare:
      resu = polyrecursif (p1->U.op, p2);
      break;
    default:;
    }
  return resu;
}
