/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/** Constructor.*/
  z3zShaper::z3zShaper ()
{
}

/** Destructor.*/
z3zShaper::~z3zShaper ()
{
}

/** It writes the buffer with overflow (tabulation) managing. */
void
z3zShaper::printline (void)
{
  bufferOut->empty ();
  if (over)
    {
      over = pk_false;
      bufferOut->backTabulation ();
    }
}


/** It writes an identifier (id) */
void
z3zShaper::printident (ptident id)
{
  ptrepstr vstr;

  vstr = ident2repstr (id);
  if (bufferOut->outRepstr (vstr));
  else
    {
      depasspoly ();
      bufferOut->outRepstr (vstr);
    }
}


/** It writes a space. */
void
z3zShaper::printespace (void)
{
  if (bufferOut->outChar (' '));
  else
    depasspoly ();
}



/** It writes a char (c). */
void
z3zShaper::printchar (char c)
{
  if (bufferOut->outChar (c));
  else
    {
      depasspoly ();
      bufferOut->outChar (c);
    }
}



/** It writes a string (ch). */
void
z3zShaper::printString (ptrepstr ch)
{
  if (bufferOut->outRepstr (ch));
  else
    {
      depasspoly ();
      bufferOut->outRepstr (ch);
    }
}


/** 
    create a buffer bufferOut, open filePathName in writing mode and link bufferOut to filePathName, 
    @return "true" if the operation succeeds
*/
boolean z3zShaper::open (ptrepstr filePathName)
{
  bufferOut = new buffer (12, 128);
  return bufferOut->buffer_open (filePathName, 0);
}


/** Close the file previously opened by open() function. */
void
z3zShaper::close (void)
{
  bufferOut->empty ();
  bufferOut->buffer_close ();
}




/** It prints a polynomial (p).  */
void
z3zShaper::printpolynomial (ptpolynomial p)
{
  printpoly (p, 0);
}

/** It prints a list of polynomials introduced by a string (listName). */
void
z3zShaper::unparse_listpoly (ptlpolynomial list, ptrepstr listName)
{
  ptpolynomial p;

  printdeblist (listName);
  initIteratorListPoly (list);
  if (!isEndlistPoly (list))
    {
      p = nextpolynomial (list);
      while (!isEndlistPoly (list))
	{
	  printpoly (p, 0);
	  printseparateur ();
	  p = nextpolynomial (list);
	}
      printpoly (p, 0);
    }
  printfinlist ();
}



/**
   It prints a list of variable declarations (list) introduced by a string (chdeclare).
*/
void
z3zShaper::unparse_declaration (ptlpolynomial list, ptrepstr chdeclare)
{
  ptpolynomial p;

  printdebfonction (chdeclare);
  initIteratorListPoly (list);
  if (!isEndlistPoly (list))
    {
      p = nextpolynomial (list);
      while (!isEndlistPoly (list))
	{
	  printpoly (p, 0);
	  printseparateur ();
	  p = nextpolynomial (list);
	}
      printpoly (p, 0);
    }
  printfinfonction ();
}



/** It prints a definition (d) as VAR = POLY.  */
void
z3zShaper::printdefinition (ptdefpoly d)
{
  ptpolynomial p;
  ptident ident;

  getdefpoly (d, &ident, &p);
  printident (ident);
  bufferOut->space ();
  printchar (':');
  bufferOut->space ();
  printpoly (p, 0);
  printchar (';');
  printline ();
  printline ();
}


/** It prints a list of definitions (list).*/
void
z3zShaper::unparse_listdef (ptldefpoly list)
{
  printline ();
  printline ();
  initIteratorldefpoly (list);
  while (!isEndldefpoly (list))
    printdefinition (nextdefpoly (list));
  printline ();
  printline ();
}


/**
   It initializes the "command" code generation (processname_CMD.z3z file): 
   - some global strings (vstrcommand, vstrpredefs, vstrmkp) are created,
   - the file processname_CMD.z3z is opened,
   - then, the printreadfilesz3z() "initialisations" of the system (processname) are generated.
*/
void
z3zShaper::initSigaliCommandFile (ptrepstr fname, ptrepstr processname)
{
  vstrpredefs = string2repstr (cstrpredefs);
  vstrmkp = string2repstr (cstrmkp);
  open (fname);
  printreadfilesz3z (processname);
}

/** It writes the string: read("chaine.z3z").
 */
void
z3zShaper::printreadfile (ptrepstr chaine)
{
  printdebcommand (vreadstr);
  printfilecommandsuff (chaine, string2repstr (Z3ZSUFF));
}

/**
   It finalyzes the "command" code generation (foo_CMD.z3z file).
   previously begun by initSigaliCommandFile() function.
*/
void
z3zShaper::endSigaliCommandFile (void)
{
  close ();
}




/** DEBUG: Not documented.*/
void
z3zShaper::printdebugpoly (ptdefpoly d)
{
  ptpolynomial p;
  ptident ident;

  getdefpoly (d, &ident, &p);
  printf (" %d ,", getpolytype (p));
}

/** DEBUG: Not documented.*/
void
z3zShaper::printdebugdef (ptldefpoly list)
{
  initIteratorldefpoly (list);
  while (!isEndldefpoly (list))
    printdebugpoly (nextdefpoly (list));
}

void
z3zShaper::printkwnotimplemented ()
{
  printString (string2repstr ("NOT_IMPLEMENTED"));
}

void
z3zShaper::printkwUndefined ()
{
  printchar ('"');
  printString (string2repstr ("UNDEFINED"));
  printchar ('"');
}

/** component initializing: the priority of the operators is stored in a global array (priorityTable) */
void
z3zShaper::initprintpoly (void)
{
  priorityTable[(int) constpoly] = 10;
  priorityTable[(int) variable] = 10;
  priorityTable[(int) polySum] = 6;
  priorityTable[(int) polyMult] = 7;
  priorityTable[(int) polyDiff] = 6;
  priorityTable[(int) polyDefault] = 1;
  priorityTable[(int) polyWhen] = 2;
  priorityTable[(int) polyOr] = 3;
  priorityTable[(int) polyAnd] = 4;
  priorityTable[(int) equation] = 0;
  priorityTable[(int) polyWhen1] = 2;
  priorityTable[(int) polyNot] = 5;
  priorityTable[(int) polyMinus] = 8;
  priorityTable[(int) polySquare] = 9;

  opolydef = string2repstr ("default");
  opolywhen = string2repstr ("when");
  opolyand = string2repstr ("and");
  opolyor = string2repstr ("or");
  opolynot = string2repstr ("not");
  vreadstr = string2repstr ("read");
}
