/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define libsuff "lib"
#define cstrpredefs "Bibli"
#define cstrmkp     "Creat_SDP"
/** 
    It empties the buffer and indents.
*/
  void
z3zShaper::depasspoly (void)
{
  bufferOut->empty ();
  if (!over)
    {
      over = pk_true;
      bufferOut->tabulation (4);
    }
}


/** It writes an int (v).
    v is assumed to be equal to 0 or 1. 
*/
void
z3zShaper::printinteger (int v)
{
  if (v == 0)
    printchar ('0');
  else
    printchar ('1');
}


/** It writes an operator (op)*/
void
z3zShaper::printopoly (ptrepstr op)
{
  if (bufferOut->isOver (strsize (op)))
    depasspoly ();
  bufferOut->printString (op);
  printespace ();
}



/** It writes an operator(op). */
void
z3zShaper::printoper (typolynomial op)
{
  switch ((int) op)
    {
    case constpoly:
    case variable:
      break;
    case polySum:
      printespace ();
      printchar ('+');
      printespace ();
      break;
    case polyDiff:
      printespace ();
      printchar ('-');
      printespace ();
      break;
    case polyMult:
      printespace ();
      printchar ('*');
      printespace ();
      break;
    case polyDefault:
      printespace ();
      printopoly (opolydef);
      break;
    case polyWhen:
      printespace ();
      printopoly (opolywhen);
      break;
    case polyOr:
      printespace ();
      printopoly (opolyor);
      break;
    case polyAnd:
      printespace ();
      printopoly (opolyand);
      break;
    case equation:
      printespace ();
      printchar ('=');
      printespace ();
      break;
    case polyWhen1:
      printopoly (opolywhen);
      break;
    case polyNot:
      printopoly (opolynot);
      break;
    case polyMinus:
      printchar ('-');
      printespace ();
      break;
    case polySquare:
      printespace ();
      if (bufferOut->isOver (2))
	depasspoly ();
      bufferOut->outChar ('^');
      bufferOut->outChar ('2');
      break;
    default:;
    }
}



/** It writes the string ' str [ ' */
void
z3zShaper::printdeblist (ptrepstr str)
{
  over = pk_false;
  bufferOut->empty ();
  bufferOut->printString (str);
  printespace ();
  bufferOut->outChar ('[');
  bufferOut->empty ();
  bufferOut->tabulation (8);
}



/**
   It writes the string ' str( '
*/
void
z3zShaper::printdebfonction (ptrepstr str)
{
  over = pk_false;
  bufferOut->empty ();
  bufferOut->printString (str);
  bufferOut->outChar ('(');
  bufferOut->tabulation (8);
}



/** It writes the ',' separator.  */
void
z3zShaper::printseparateur (void)
{
  if (!bufferOut->outChar (','))
    {
      bufferOut->empty ();
      bufferOut->outChar (',');
    }
  bufferOut->empty ();
  if (over)
    {
      over = pk_false;
      bufferOut->backTabulation ();
    }
}

/**
   It writes the string '];' and empties the buffer.
*/
void
z3zShaper::printfinlist (void)
{
  if (over)
    bufferOut->backTabulation ();
  if (bufferOut->isOver (2))
    bufferOut->empty ();
  bufferOut->outChar (']');
  bufferOut->outChar (';');
  bufferOut->empty ();
  bufferOut->backTabulation ();
  bufferOut->empty ();
}



/** It writes the string ' ); ' and empties the buffer.
 */
void
z3zShaper::printfinfonction (void)
{
  if (over)
    bufferOut->backTabulation ();
  if (bufferOut->isOver (2))
    bufferOut->empty ();
  bufferOut->outChar (')');
  bufferOut->outChar (';');
  bufferOut->empty ();
  bufferOut->backTabulation ();
  bufferOut->empty ();
}



/**
   It writes a polynomial (p) taking account the priority (prio) of the father operator.
*/
void
z3zShaper::printpoly (ptpolynomial p, int prio)
{
  typolynomial t;
  int priogauche, priodroit, c;

  t = getpolytype (p);
  switch ((int) t)
    {
    case constpoly:
      c = getpolyvalue (p);
      if (c < 0)
	{
	  priogauche = priorityTable[(int) polyMinus];
	  if (priogauche <= prio)
	    printchar ('(');
	  printoper (polyMinus);
	  printinteger (-c);
	  if (priogauche <= prio)
	    printchar (')');
	}
      else
	printinteger (c);
      break;
    case variable:
      printident (getpolyname (p));
      break;
    case polySum:
    case polyMult:
    case polyDiff:
    case polyDefault:
    case polyWhen:
    case polyOr:
    case polyAnd:
    case equation:
      priogauche = priorityTable[(int) t];
      if (t == polyDiff)
	{
	  priodroit = priorityTable[(int) polyMult];
	}
      else
	priodroit = priogauche;
      if (prio > priogauche)
	printchar ('(');
      printpoly (getop1 (p), priogauche);
      printoper (t);
      printpoly (getop2 (p), priodroit);
      if (prio > priogauche)
	printchar (')');
      break;
    case polyWhen1:
    case polyNot:
    case polyMinus:
      priogauche = priorityTable[(int) t];
      if (priogauche <= prio)
	printchar ('(');
      printoper (t);
      printpoly (getop (p), priogauche);
      if (priogauche <= prio)
	printchar (')');
      break;
    case polySquare:
      priogauche = priorityTable[(int) polySquare];
      if (priogauche <= prio)
	printchar ('(');
      printpoly (getop (p), priogauche);
      printoper (polySquare);
      if (priogauche <= prio)
	printchar (')');
      break;
    default:;
    }
}

/**   It writes the string  ' vstr ( ' */
void
z3zShaper::printdebcommand (ptrepstr vstr)
{
  bufferOut->writeRepstr (vstr);
  bufferOut->printChar ('(');
  bufferOut->printChar ('"');
}

/**
   It writes the string "chaine.suff );" and empties the buffer.
*/
void
z3zShaper::printfilecommandsuff (ptrepstr chaine, ptrepstr suff)
{
  bufferOut->writeRepstr (chaine);
  bufferOut->printChar ('.');
  bufferOut->writeRepstr (suff);
  bufferOut->printChar ('"');
  bufferOut->printChar (')');
  bufferOut->printChar (';');
  bufferOut->empty ();
}


/** It writes  the Sigali command: <em> read("chaine.suff");  </em> 
 */
void
z3zShaper::printreadfileSuffix (ptrepstr chaine, ptrepstr suff)
{
  printdebcommand (vreadstr);
  printfilecommandsuff (chaine, suff);
}

/**
   It writes the Sigali commands:
   - <em> read("<processname>.z3z");  </em>
   - <em> read("Creat_SDP.lib");  </em>
   - <em> read("Bibli.lib");  </em>
*/
void
z3zShaper::printreadfilesz3z (ptrepstr processname)
{
  ptrepstr libstr;

  libstr = string2repstr (libsuff);
  // printtracefile(processname); // trace file <processname>.log
  printreadfile (processname);	// read("<processname>.z3z");
  printreadfileSuffix (vstrmkp, libstr);	// read("Creat_SDP.lib");
  printreadfileSuffix (vstrpredefs, libstr);	// read("Bibli.lib");   
}

/** 
    UNUSED NOW (lbesnard).

    It writes the Sigali command:\n
    SIMUL(S,"processname.sim","processname.res");
*/
void
z3zShaper::printsimufilesigali (ptrepstr processname)
{
  bufferOut->writeAlfa ("simul(S,", 8);
  bufferOut->printChar ('"');
  bufferOut->writeRepstr (processname);
  bufferOut->printChar ('.');
  bufferOut->writeAlfa ("res    ", 3);
  bufferOut->printChar ('"');
  bufferOut->printChar (',');
  bufferOut->printChar ('"');
  bufferOut->writeRepstr (processname);
  bufferOut->printChar ('.');
  bufferOut->writeAlfa ("sim    ", 3);
  bufferOut->printChar ('"');
  bufferOut->printChar (')');
  bufferOut->printChar (';');
  bufferOut->empty ();
}
