/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <string> //STL to use string class
using std::string;
#include <list>
#include <map>

typedef class DCG2Syndex *ptDCG2Syndex;

typedef std::map<ptrepstr, boolean> SyndexTableOperationType;
class DCG2Syndex
{
#include "DCG2Syndex_m.h"
#include "DCG2SyndexDeclarations_i.h"
#include "DCG2SyndexClocks_i.h"
#include "DCG2SyndexNodes_i.h"
#include "DCG2SyndexNodesRef_i.h"
#include "DCG2SyndexExpressions_i.h"
#include "DCG2SyndexExprRef_i.h"
#include "DCG2SyndexPartitionning_i.h"
#include "DCG2SyndexArchitecture_i.h"
#include "DCG2Syndex_g.h"

protected:
  ptDCG AGRAPH; ///< Current Graph
  ptClockManager CKM; ///< Clock manager
  ptExprManager EM;  ///< Expression manager
  ptNodeManager NM;  ///< Node Manager

  ptinsth TheTick; ///< Master Clock

  ptident graphName;   ///< name of the Graph
  ptlINode unNodes; ///< Unused Nodes
  ptlinsth unClocks; ///< Unused Clocks

  GraphLevelType graph_type; ///< Graph Level in {DC+, bDC+, STS, DC}
  ptlinstances inputs, outputs, parameters,locals; ///< Inputs/outputs/parameters/locals of a graph or a subgraph

  ptlINode lwrites; 
  ptinteTreeBuilder TBI;  ///< Tree Builder Interface object
    typeMark vSyndexAttribute, equivAttribute, vConnectionMark,
      setOperationsMark; // sets to a node, it is the set of operations defined in Syndex for the translation. Used for the definition of the groups.
  ptINode ArrayBlock;
  int CurrentSubGraph;
  ptINode TheCurrentNode;
  boolean addClockTreePost;
  boolean addPrefixForAlgorithm;
  boolean  Restriction_DRealType, asMain, returnCode, isRunOnPartionning;
  ptrepstr MyStdOutput, ///< Standard output
    StrUndefNode;

  std::map<ptrepstr, integer> ReferenceCountersTable;  ///< Associative table for reference counters of operators
  std::map<noperators, ptrepstr> OperatorNamesTable;  ///< Associative table for the name of operators
  std::map<noperators, boolean> implementedForVector;  ///< Associative table for array implementation
  std::map<TypeType, ptrepstr> TypeTable;  ///< Associative table for types
  std::list<ptSyndexAttribute> lattr; ///< for cleaning
  SyndexTableOperationType SyndexOperations; ///< Managed only when isRunOnPartionning is true, SyndexOperations[op] is true if the syndex operation op is assigned to a group.
};

