/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

typedef class SyndexAttribute  *ptSyndexAttribute;

/** 
    It defines a Syndex attribute. It is used to link internal structures of the Polychrony graph to the Syndex objects.

    Example: For the equation x:=a*b in the subgraph 56, the following attribute (Arit_mul_1, .o, 56) is created and associated with x,
    \c Arit_mul_1 is an instance of a predefined operator declared Signal_float/Arit_mul<1> Arit_mul_1; in the Syndex \c references: 
    This attribute is then used to connect syndex operators: a reference to x is replaced by "Arit_mul_1.o".
*/
class SyndexAttribute
{
#include "SyndexAttribute_m.h"
#include "SyndexAttribute_g.h"

protected:
  ptrepstr node; ///<a Syndex node name  (reference in the Syndex terminology).
  ptrepstr name; ///< the name of an output of the node.
  integer  SubGraphNumber; ///< the number of the associated Signal subgraph.
};
