/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_ ptrepstr DCG2Syndex::getReferenceName (ntrees t)
{
  ptrepstr
    operation;

  operation = TBI->getReferenceName (t);
  if (isRunOnPartionning)
    SyndexOperations[operation] = pk_false;
  return operation;
}

/*
  @retrun the type of an expression.
  \note {the dreal type is replaced by the real type.}
 */
TypeType DCG2Syndex::getType (ptexpcal e)
{
  TypeType
    TT;
  TT = EM->getType (e);
  return TT;
}

/**
   TYPES translating: to be completed; Currently, only for predefined type.
   @return the Signal operator associated with a type (TT) when TT is a predefined one, opnil otherwise.
   
   \note { the dreal is replaced by real.}
*/
noperators DCG2Syndex::predefTypeType2SyndexType (TypeType TT)
{
  if (TypeTestPredef (TT))
    {
      switch ((int) TT)
	{
	case ErrorType:
	  return opnil;
	case LabelType:
	  return opnil;
	case PureType:
	  return opevent;
	case BooleanType:
	  return oplogical;
	case ShortType:
	  return opshort;
	case IntegerType:
	  return opinteger;
	case LongType:
	  return oplong;
	case RealType:
	  return opreal;
	case DRealType:
	  return opdpreal;	// restriction(1); return opreal;

	case ComplexType:
	  return opcomplex;
	case DComplexType:
	  return opdcomplex;
	case CharType:
	  return opchar;
	case StringType:
	  return opstring;
	case ByteType:
	  return opnil;
	case WordType:
	  return opnil;
	}
    }
  return opnil;
}

/**
   TYPES translating: to be completed; Currently, only for predefined type.
   @return the Syndex representation of a type (t) when t is a predefined one, NILTREE otherwise.
*/
ntrees
DCG2Syndex::TypeType2ntrees (TypeType t)
{
  // here i only consider the predef type now, haven't the complex types.
  if (TypeTestPredef (t))
    return TBI->mkNullaryTree (predefTypeType2SyndexType (t));

  /* 
     if (TypeTestArrayClass (t)) 
     {
     treeTypeBase = TypeType2ntrees(ArrayMatrixTypeGetTerm(t));
     return LL_make_tvect_tree( translateConstantExpression(ArrayMatrixTypeGetDimLst(t)), treeTypeBase);
     }
     if (TypeTestExternClass (t))  return repstr2ident(ExternTypeGetName(t));
     if (TypeTestEnumerationClass (t)) return repstr2ident(EnumerationTypeGetName(t));// do nothing for enum.
     if (TypeTestMetaClass(t)) return repstr2ident(MetaTypeGetName(t));
     if (TypeTestStructureClass ( t ))  return TypeTypeStructure2ntrees(t);
     if (TypeTestBundleClass ( t ))  return TypeTypeBundle2ntrees(t);
     if (TypeTestNamedClass (t))     return repstr2ident(NamedTypeGetName(t));

     if (TypeTestSubrangeClass ( t )) return mkNOTIMPLEMENTED();
     if (TypeTestBitsClass ( t )) return mkNOTIMPLEMENTED();
     if (TypeTestLambdaClass ( t ))  return mkNOTIMPLEMENTED();
     if (TypeTestNodeClass ( t ))  return mkNOTIMPLEMENTED();
     if (TypeTestGenericNodeClass ( t ))  return mkNOTIMPLEMENTED();
     if (TypeTestMetaNodeClass ( t )) return mkNOTIMPLEMENTED();
     if (TypeTestMatrixClass ( t ))  return mkNOTIMPLEMENTED();;
     if (TypeTestTupleClass ( t ))  return mkNOTIMPLEMENTED();

     return TBI->mkNullaryTree((opimplicit);
   */
  return NILTREE;
}

/**
   @return the Syndex representation of the type of an expression (vexp).
*/
ntrees
DCG2Syndex::typeExp2ntrees (ptexpcal vexp)
{
  return TypeType2ntrees (getType (vexp));
}

/** 
    Constants translating.
    It declares the constants (addTreeWithCst()) of the current program in the Syndex program.
*/
_VIRTUAL_ void
DCG2Syndex::translateConstantsDeclarations ()
{
  ptlinstances lCst;

  lCst = AGRAPH->getConstants ();
  addTreeWithCst (lCst);
}

/**
   Constants and parameters translating.
   It adds in the Syndex program the declaration of a constant (aConst).
*/
void
DCG2Syndex::mkCstDlc (ptexpcal aConst)
{
  ptexpcal vopg, vopd;
  ptINode vinstr;
  ntrees vtdef;

  if (EM->IsExternalConstant (aConst))
    {
      vopg = aConst;
      SET_NILTREE (vtdef);
    }
  else
    {				// not external, the value is treated as parameter.     
      vinstr = EM->getNode (aConst);
      NM->get_equation_arg (vinstr, &vopg, &vopd);
      vtdef = translateConstantExpression (vopd);
    }
  TBI->mkCstDlc (typeExp2ntrees (vopg), getRepstrIdName (vopg), vtdef);
}

/**
   @param exp (in) an expression
   @param dim (out) the representation of the dimensions of the type of exp,
   @param t (out) the name of the type of exp, or the predefined Syndex type for a predefined type (Signal_float.sdx for example).
   @return the name of the type of exp.
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::getTypeLibrary (ptexpcal exp, ntrees * dim, ptrepstr * t)
{
  TypeType TT;
  boolean bLibrary;
  ptrepstr tname;

  TT = getType (exp);
  TT = getInfosType (-1, TT, dim, &bLibrary);
  tname = getTypeName (TT);
  if (TypeTestPredef (TT))
    *t = TBI->getPredefinedLibraryName (tname);
  else
    *t = tname;
  return tname;
}

/**
   It makes (getTypeLibrary()) the declaration of the Syndex type library associated with the type of an expression (exp).
   This declaration is added to the Syndex program.
   @param dim (out) the tree that represents the dimensions of the type.
   @param exp an Signal expression.
   @return the the Syndex type library.
*/
_VIRTUAL_ ptrepstr
DCG2Syndex::addTypeLibrary (ptexpcal exp, ntrees * dim)
{
  ptrepstr tname, tToDeclare;

  tname = getTypeLibrary (exp, dim, &tToDeclare);
  TBI->addInclude (tToDeclare);
  return tname;
}

/**
  Declaring the constants (lCst) of the algorithm. 
  For each constant, if it is the predefined type, 
  add the include tree. Otherwise, make a new constant algorithm. At the end, add the reference to mainalgorithm.
*/
_VIRTUAL_ void
DCG2Syndex::addTreeWithCst (ptlinstances lCst)
{
  ptexpcal aConst;
  ntrees vdim;
  while (ptlinstances_get_next (&lCst, &aConst))
    {
      TRACE (cout << " in constant" << endl);
      addTypeLibrary (aConst, &vdim);
      mkCstDlc (aConst);
    }
}

/** Adding a declaration of a i/o signal in the current algorithm.
   ? <type(elem)> <name(elem)>
   Assumes : TBI current tree is valid.
   The declaration is added in the "current" tree of the TBI object and following the w parameter.
   @param w (in) area where the signal must be declared (see typsignal enum type defined in sign_type.h )
   @param elem a signal declaration.
   A link from the signal to its tree representation can be kept using addAttribute method.

*/
_VIRTUAL_ ptrepstr
DCG2Syndex::translateDeclarationOfSignal (typsignal w, ptexpcal elem)
{
  ptrepstr t1, name;
  ntrees vdim;

  t1 = addTypeLibrary (elem, &vdim);
  name = getRepstrIdName (elem);
  if (w == InSignal)
    TBI->addInput (t1, vdim, name);
  else
    TBI->addOutput (t1, vdim, name);
  return name;
}


/**
   Adding a declaration of a i/O signal. A sensor/actuator is created and returned. 
   The name of the operation is the name of the input/output signal.
 */
_VIRTUAL_ ptrepstr
DCG2Syndex::addIODeclaration (typsignal w, ptexpcal elem)
{
  noperators op;
  ntrees vdim, t, dim;
  int theNthOperation;
  ptrepstr nameOperSyn, typeSyn;
  boolean bLibrary;

  addTypeLibrary (elem, &vdim);
  if (w == InSignal)
    op = opread;
  else
    op = opwrite;
  //  return getLanguageOperator(op,getType(elem)); replaced by (to have the name of the input/output as the name of the operation).
  nameOperSyn = getInfosForReferenceDeclaration (op, getType (elem), &theNthOperation, &typeSyn, &dim, &bLibrary);
  // Here, the theNthOperation value is not used, the name of the Signal i/o  is used.
  t = TBI->addReference (bLibrary, typeSyn, nameOperSyn, ident2repstr (EM->getName (elem)), TBI->mkArgsValues (NILTREE, NILTREE, NILTREE, dim));
  return getReferenceName (t);
}

/**
   \note --UNUSED. Only referenced in pushParameters() --

   Adding a declaration of a signal.
   Assumes : TBI current tree is valid.
   The declaration is added in the "current" tree of the TBI object and following the w parameter.
   @param elem a signal declaration.
   A link from the signal to its tree representation can be kept using addAttribute method.
*/
_VIRTUAL_ void
DCG2Syndex::translateDeclarationOfParameter (ptexpcal elem)
{
  ptrepstr t1;
  TypeType type;

  type = getType (elem);
  t1 = TypeTable[type];		// Type in syndex
  //  TBI-> addParameter(t1, getRepstrIdName(elem));
}


/** 
    Adding Input signal declarations (translateDeclarationOfSignal()).
    @param list (in) a list of signals.
*/
_VIRTUAL_ void
DCG2Syndex::pushInputDeclarations (ptlinstances list)
{
  ptexpcal elem;
  ptrepstr referenceName;

  while (ptlinstances_get_next (&list, &elem))
    {
      referenceName = translateDeclarationOfSignal (InSignal, elem);
      setReferenceTo (elem, mkSyndexAttribute (nilstr, referenceName));
    }
}

/** 
    Adding Output signal declarations (translateDeclarationOfSignal()).
    @param list (in) a list of signals.
*/
_VIRTUAL_ void
DCG2Syndex::pushOutputDeclarations (ptlinstances list)
{
  ptexpcal elem;
  while (ptlinstances_get_next (&list, &elem))
    translateDeclarationOfSignal (OutSignal, elem);
}

/**
   \note --UNUSED.--

  Adding parameters declarations.
  @param list (in) : a list of signals.
*/
_VIRTUAL_ void
DCG2Syndex::pushParameters (ptlinstances list)
{
  ptexpcal elem;
  while (ptlinstances_get_next (&list, &elem))
    translateDeclarationOfParameter (elem);
}


/** Push input/outputs signal declarations for the current clock.
*/
_VIRTUAL_ void
DCG2Syndex::PushInterfaceDeclarations ()
{
  pushInputDeclarations (inputs);
  pushOutputDeclarations (outputs);
  // pushParameters(parameters);
}

/**
   It pops (popReference()) each element of a list of signals (list).
*/
_VIRTUAL_ void
DCG2Syndex::popDeclarations (ptlinstances list)
{
  ptexpcal elem;
  while (ptlinstances_get_next (&list, &elem))
    popReference (elem);
}

/** 
    It pops input/output signal declarations (popDeclarations()) for the current clock.
*/
_VIRTUAL_ void
DCG2Syndex::PopInterfaceDeclarations ()
{
  popDeclarations (inputs);
  popDeclarations (outputs);
  // popDeclarations(parameters);
}

/**
   @return the Syndex declarations of the parameters of the current graph.
*/
_VIRTUAL_ ntrees
DCG2Syndex::mkProgramParameters ()
{
  ptlinstances lparams;
  ptexpcal elem;
  ntrees vt, vdim;

  lparams = AGRAPH->getParameters ();
  if (lparams == NILLINSTANCES)
    return NILTREE;
  vt = TBI->make_lparameters_emptyList ();
  while (ptlinstances_get_next (&lparams, &elem))
    if (EM->getOperatorClass (elem) == variabl)	// Filtering types declarations.
      {
	addTypeLibrary (elem, &vdim);
	vt = TBI->AddList (vt, TBI->makeIdentifier (getRepstrIdName (elem)));
      }
  return vt;
}
