/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/**
   @return the string associated with a constant expression (vexp).
   A double constant is replaced by a float. (xxdyy replaced by xxeyy).
*/
ptrepstr DCG2Syndex::getConstant (ptexpcal vexp)
{
  ptrepstr
    vstr;
  vstr = EM->getConstant (vexp);
  if (EM->getType (vexp) == DRealType)
    {
      vstr = strdoubletoCdouble (vstr);	/* restriction(1); */
    }
  return vstr;
}


/**
  returns pk_true if the expression is a evaluated expression.
  In this case it returns in t the value representation tree.
*/
_VIRTUAL_
  boolean
DCG2Syndex::genValue (ptexpcal exp, ntrees * t)
{
  TypeType TT;
  long int val;

  if (EM->IsIndex (exp))
    return pk_false;
  if (!EM->IsAssigned (exp))
    return pk_false;
  TT = EM->getUnamedType (exp);
  if (!TypeTestPredef (TT))
    return pk_false;
  val = EM->getValue (exp);
  *t = TBI->makeAtomInteger (val);
  return pk_true;
}


/** Returns the abstract tree of a CONSTANT expression.
   Assumes : the expression references parameters, constants and   +, -, /, *  operators.
*/
ntrees DCG2Syndex::constantExpr2SyndexTree (ptexpcal vexp, ntrees * vt)
{
  noperators
    operateur,
    op;
  ptexpcal
    vexpcal1,
    vexpcal2,
    nexp;
  ptrepstr
    vstr;
  int
    val;

  if (vexp == NILEXPCAL)
    return NILTREE;
  if (EM->getIndirection (vexp, &nexp))
    vexp = nexp;		// A virer mais urgence
  //if (genValue(vexp, &gvt)) return gvt;

  switch ((int) (EM->getOperatorClass (vexp)))
    {
    case variabl:
      if (NM->get_rightOperand (EM->getNode (vexp), &nexp))
	return constantExpr2SyndexTree (EM->getBasicExpr (nexp), vt);
      return TBI->makeIdentifier (getRepstrIdName (vexp));

    case constant:		//2
      operateur = EM->getOperator (vexp);
      if (operateur != opcstenum)
	{
	  if (operateur == optrue)
	    return TBI->makeConstantInt (1);
	  if (operateur == opfalse)
	    return TBI->makeConstantInt (0);
	  return TBI->makeConstant (getConstant (vexp));
	}
      pK_messenger->Caseerror (currentLine, currentFile);
      break;

    case oper1:
      operateur = EM->getOperator (vexp);
      vexpcal1 = EM->getUnaryArg (vexp);
      if (operateur == opnega)
	return TBI->make_SyndexUminus_tree (constantExpr2SyndexTree (vexpcal1, vt));
      pK_messenger->Caseerror (currentLine, currentFile);
      break;
    case oper2:		//5
      operateur = EM->getOperator (vexp);
      vexpcal1 = EM->getBinaryArgstAt (vexp, 1);
      vexpcal2 = EM->getBinaryArgstAt (vexp, 2);
      if (operateur == opvenum)
	{
	  vstr = EM->getConstant (vexpcal2);
	  EnumTypeOrdinal (getType (vexp), vstr, &val);
	  return TBI->makeConstantInt (val);
	}
      if (LL_is_listOp (operateur))
	{			/* list */
	  if (IS_NILTREE (*vt))
	    *vt = TBI->make_Syndexlexpr_emptyList ();
	  while (vexp != NILEXPCAL)
	    *vt = ntreeList_appendLast (*vt, constantExpr2SyndexTree (EM->getNaryArgs (&vexp), vt));
	  return *vt;
	}
      switch ((int) (operateur))
	{
	case opplus:
	case opmoin:
	case opdivi:
	case opmult:
	  return TBI->makeBinary ((int) (operateur), constantExpr2SyndexTree (vexpcal1, vt), constantExpr2SyndexTree (vexpcal2, vt));
	case opelevec:
	  return TBI->make_SyndexElevec_tree (constantExpr2SyndexTree (vexpcal2, vt));
	default:
	  op = -1;
	}
      pK_messenger->printString ("Access to an expression : operator not yet implemented!! ");
      return NILTREE;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return NILTREE;
}

/**
   With the attribute (nodeName, outputNumber) associated with an expression (arg), it adds the connection of the outputNumber-th output
   of the nodeName node to the n-th input of a node (vattr).
*/
void
DCG2Syndex::connectionFromTo (ptexpcal arg, ptrepstr vattr, int n)
{
  ptSyndexAttribute argAttr;

  if (getSyndexAttributeOf (arg, &argAttr))
    TBI->addConnectionFromToAt (argAttr->getSyndexNode (), argAttr->getSyndexOutput (), vattr, n);
  else
    TRACE (cout << getStringIdName (arg) << " ERROR addConnection 2" << endl);
}

/**
   With the attribute (nodeName, outputNumber) associated with an expression (arg), it adds the connection of the outputNumber-th output
   of the nodeName node to the input (inputName) of a node (vattr).
*/
void
DCG2Syndex::connectionFromTo2 (ptexpcal arg, ptrepstr vattr, ptrepstr inputName)
{
  ptSyndexAttribute argAttr;

  if (getSyndexAttributeOf (arg, &argAttr))
    TBI->addConnectionFromToAt2 (argAttr->getSyndexNode (), argAttr->getSyndexOutput (), vattr, inputName);
  else
    TRACE (cout << getStringIdName (arg) << " ERROR addConnection 2" << endl);
}

/**
   It adds the connection from the output of arg to the n-th input of vexp.
*/
void
DCG2Syndex::addConnection (ptexpcal vexp, ptexpcal arg, int n)
{
  ptrepstr vattr;

  if (EM->isNILEXPCAL (arg))
    return;
  if (EM->isNILEXPCAL (vexp))
    return;
  if (EM->getOperator (arg) == opvvarin)
    arg = EM->getUnaryArg (arg);
  if (getReferenceTo (vexp, &vattr))
    connectionFromTo (arg, vattr, n);
  else
    TRACE (cout << "addConnection 1,getReferenceTo ERROR,(I2SExpressions_i.c)" << endl);
}

/** 
    Syndex Cell translation. 
    @param leftPart (in) the defined signal of the memory.
    @param rightPart (in) rigth part of the definition.

*/
_VIRTUAL_ void
DCG2Syndex::translateCell (ptexpcal leftPart, ptexpcal rightPart)
{
  ptexpcal arg1, arg3, e;
  ptexpcal initExpr;

  initExpr = EM->getInitialValue (leftPart);
  e = EM->getWhenOrCellArg (rightPart);
  arg1 = EM->getWhenOrCellArg (e);
  //arg2 = initExpr;
  arg3 = CKM->getBooleanSignal (EM->getContributionClock (e));
  translateExpression (arg1);
  //translateExpression(arg2);
  translateExpression (arg3);
  // addConnection(rightPart,arg2,1); // init : is set in the declaration of the Syndex cell operator

  // it is neccessary to use the attribute of the leftpart. Otherwise X:= A cell C, Y:= A cell C will produce a 
  // code with only one cell syndex operator (and multiple connection to this operator). This is induced by the unicity of the expressions
  // and the fact that the init expression is set to x (resp. y).

  addConnection (leftPart, arg1, 1);	// value
  addConnection (leftPart, arg3, 2);	// clock
}

/**
   It translates a 'when expression' (e).
   \see translateExpression(), addConnection()
*/
_VIRTUAL_ void
DCG2Syndex::translateWhen (ptexpcal e)
{
  translateExpression (EM->getWhenOrCellArg (e));	// e1
}


/**
   @return the translation of a constant expression (vexp).
   \see constantExpr2SyndexTree().
*/
_VIRTUAL_ ntrees
DCG2Syndex::translateConstantExpression (ptexpcal vexp)
{
  ntrees vt;

  SET_NILTREE (vt);
  return constantExpr2SyndexTree (vexp, &vt);
}

/**
   It translates an unary expression (vexp).
   \see translateExpression(), addConnection()
*/
_VIRTUAL_ void
DCG2Syndex::translateUnaryExpression (ptexpcal vexp)
{
  ptexpcal arg1;
  if (EM->getOperator (vexp) != opvvarin)
    {
      arg1 = EM->getUnaryArg (vexp);
      translateExpression (arg1);
      addConnection (vexp, arg1, 1);
    }
}

/**
   @return the second operand of an assumed binary expression (vexp).
   For an acces to an array (x[i1,i2,...]) andthe size of [i1,i2,...] is reduced to one; it returns the element.
*/
ptexpcal DCG2Syndex::getSecondArgBinaryExpression (ptexpcal vexp, noperators vop)
{
  ptexpcal
    arg2;

  // I don't well understand the particular case (lb).
  arg2 = EM->getBinaryArgstAt (vexp, 2);
  if (vop != opelem)
    return arg2;
  if (EM->getSize (arg2) != 1)
    return arg2;
  return EM->getNaryArgs (&arg2);
}

/**
   It translates a binary expression (vexp).
   \see translateExpression(), addConnection()
*/
void
DCG2Syndex::translateBinaryExpression (ptexpcal vexp)
{
  ptexpcal arg1, arg2;
  noperators vop;
  ptinsth vh;

  vop = EM->getOperator (vexp);
  if (vop != opvenum)		// nothing to do for T#blue (enumerate type)
    {
      arg1 = EM->getBinaryArgstAt (vexp, 1);
      if (vop == opleft_tt)
	{
	  EM->getSynchronousExprHorl (arg1, &vh);
	  arg2 = CKM->getBooleanSignal (vh);
	}
      else
	arg2 = getSecondArgBinaryExpression (vexp, vop);
      translateExpression (arg1);
      translateExpression (arg2);
      addConnection (vexp, arg1, 1);
      addConnection (vexp, arg2, 2);
    }
}

/**
   It translates a default expression (e).
   \see translateExpression(), addConnection()
*/
void
DCG2Syndex::translateDefault (ptexpcal e)
{
  ptexpcal arg1, arg2, arg3, x;
  ptinsth h;

  x = EM->getBinaryArgstAt (e, 1);	// a When
  arg1 = EM->getWhenOrCellArg (x);
  h = EM->getContributionClock (x);

  arg2 = CKM->getBooleanSignal (h);
  arg3 = EM->getBinaryArgstAt (e, 2);
  arg3 = EM->getWhenOrCellArg (arg3);

  translateExpression (arg1);
  translateExpression (arg2);
  translateExpression (arg3);

  addConnection (e, arg1, 1);
  addConnection (e, arg2, 2);
  addConnection (e, arg3, 3);

}

/**
   It translates a ternary expression (vexp).
   \see translateExpression(), addConnection()
*/
void
DCG2Syndex::translateTernaryExpression (ptexpcal vexp)
{
  ptexpcal arg1, arg2, arg3;

  arg1 = EM->getTernaryArgstAt (vexp, 1);
  arg2 = EM->getTernaryArgstAt (vexp, 2);
  arg3 = EM->getTernaryArgstAt (vexp, 3);

  translateExpression (arg1);
  translateExpression (arg2);
  translateExpression (arg3);

  addConnection (vexp, arg1, 1);
  addConnection (vexp, arg2, 2);
  addConnection (vexp, arg3, 3);
}

/**
   It translates a function call expression (vexp).
   \see translateExpression(), addConnection()
*/
void
DCG2Syndex::translateFunctionCallExpression (ptexpcal vexp)
{
  ptident id;
  ptexpcal vexpcal1, vexpcal2;
  ptexpcal elem;
  int i;

  EM->getFunctionCall (vexp, &id, &vexpcal2, &vexpcal1);	// vexpcal2 : parameters; vexpcal1: inputs
  i = 1;
  while (EM->getNaryArgs (&vexpcal1, &elem))
    {
      translateExpression (elem);
      addConnection (vexp, elem, i);
      i++;
    }
}


/**
   It translates a list expression (vexp).
   \see translateExpression(), addConnection()
*/
void
DCG2Syndex::translateListExpression (ptexpcal vexp)
{
  ptexpcal elem;
  while (EM->getNaryArgs (&vexp, &elem))
    translateExpression (elem);
}

/**
   It translates a constant signal (vexp).
   \see translateEqu()
*/
void
DCG2Syndex::translateConstantSignal (ptexpcal vexp)
{
  ptINode n;
  if (EM->IsConstant (vexp) || EM->IsParameter (vexp))
    {
      n = EM->getNode (vexp);
      if (!NM->is_undefined (n))
	translateEqu (n);
    }
}

/**
   It translates an expression (vexp).
   It realises a switch on the operator of the expression and calls:
   translateConstantSignal(), translateConstantExpression(), translateUnaryExpression(), translateListExpression(),
   translateBinaryExpression(), translateFunctionCallExpression(), 
   translateTernaryExpression(), translateWhen(), translateDefault().
*/
void
DCG2Syndex::translateExpression (ptexpcal vexp)
{
  ptexpcal nexp;

  if (vexp == NILEXPCAL)
    return;
  if (EM->getIndirection (vexp, &nexp))
    vexp = nexp;
  if (isAlreadyConnected (vexp))
    return;

  TRACE (cout << EM->getOperatorClass (vexp) << " the operator class" << endl);
  TRACE (cout << EM->getOperator (vexp) << " the operator" << endl);

  belongConnected (vexp);
  switch ((int) (EM->getOperatorClass (vexp)))
    {
      // reference to an identifier (nothing to do for connections)
    case variabl:		//operatorClass:1
      TRACE (cout << getStringIdName (vexp) << " the name of expression" << endl);
      translateConstantSignal (vexp);
      break;

      // reference to a constant expression.
    case constant:
      TRACE (cout << "after translatevariable constant" << endl);
      translateConstantExpression (vexp);
      break;

      // Unary expression : not , -
    case oper1:
      TRACE (cout << "after translatevariable unary" << endl);
      translateUnaryExpression (vexp);
      break;

      // Binary Expression : +,*,...
    case oper2:
      TRACE (cout << "after translatevariable binary" << endl);
      if (EM->satisfiesList (vexp))
	translateListExpression (vexp);
      else
	translateBinaryExpression (vexp);
      break;

      // Delay/Window (perhaps filtered before)
    case oper3:
      TRACE (cout << "after translatevariable ternary" << endl);
      translateTernaryExpression (vexp);
      break;

      // function call id( i1,...., in)
    case call:
      TRACE (cout << "translatevariable function " << endl);
      translateFunctionCallExpression (vexp);
      break;

      // E1 when clk ( E1 cell clk has been filtered before)
    case opereh:
      TRACE (cout << "translatevariable when" << endl);
      translateWhen (vexp);
      break;

    case oper2tp:		// default operator
      translateDefault (vexp);
      break;

      // no occurrence
    case oper1tp:
    case oper3tp:
      break;

    case opertyp:
      break;

      // Internal error
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}
