/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" if a node (vinstr) is selected for the current translation, "false" otherwise.
 */
  boolean DCG2Syndex::isSelectedNode (ptINode vinstr)
{
  return !ptlINode_isIn (unNodes, vinstr);
}

/** Array managment.
   @return true if the node is internal node of current array block.
*/
boolean DCG2Syndex::isNodeCurrentBlock (ptINode vinstr)
{
  ptINode
    cn;
  // if (NM->is_readNode(vinstr)) return pk_false;
  // if (NM->is_writeNode(vinstr)) return pk_false; 
  if (!isSelectedNode (vinstr))
    return pk_false;
  NM->get_inComposedNode (vinstr, &cn);
  return (cn == ArrayBlock);
}


/** Array managment.
 */
_VIRTUAL_
  ptINode
DCG2Syndex::DCG2SignalTreepusharray (ptINode vinstr)
{
  ptINode vpushinstr;

  vpushinstr = ArrayBlock;
  ArrayBlock = vinstr;
  return vpushinstr;
}

/** Array managment.
 */
_VIRTUAL_ void
DCG2Syndex::DCG2SignalTreepoparray (ptINode vpushinstr)
{
  ArrayBlock = vpushinstr;
}

/**
   It adds connections, in the Syndex program, for a status node (aNode).
   "zmemvx := status{vinit} (zvx)"
   Recall that for Syndex, the state variable are replaced by status intrinsic funtion.
   This is done before this translation.
*/
_VIRTUAL_ void
DCG2Syndex::translateStatusNode (ptINode aNode)
{
  ptexpcal defPart, leftPart, parameters, statemem, arg2, arg3, x;
  ptident name;

  NM->get_procCall_arg (aNode, &name, &parameters, &defPart, &leftPart);
  EM->getNaryArgs (&leftPart, &statemem);
  EM->getNaryArgs (&defPart, &x);
  arg2 = EM->getWhenOrCellArg (x);
  arg3 = CKM->getBooleanSignal (EM->getContributionClock (x));
  translateExpression (arg2);
  translateExpression (arg3);
  addConnection (statemem, arg2, 1);
  addConnection (statemem, arg3, 2);
}

/**
   It adds connections, in the Syndex program, for an equation node (aNode).
   For x:= E, and E a cell expression, it call translateCell(), otherwise translateExpression().
*/
_VIRTUAL_ void
DCG2Syndex::translateEqu (ptINode aNode)
{
  ptexpcal leftPart, rightPart, basicExpr;
  boolean b;
  ptSyndexAttribute def;
  // --LB-- Warning << opleft_tt: opnombm: operators >> ???
  if (NM->get_rightOperand (aNode, &rightPart))
    {
      leftPart = NM->get_leftOperand (aNode);
      if (EM->IsCellExpr (rightPart))	// Cell
	{
	  translateCell (leftPart, rightPart);
	  if (getSyndexAttributeOf (rightPart, &def))
	    setNodeReference (aNode, def->getSyndexNode ());
	}
      else
	{
	  basicExpr = EM->getBasicExpr (rightPart);
	  translateExpression (basicExpr);
	  if (getSameLevelSyndexAttributeOf (basicExpr, &def))
	    b = pk_true;
	  else
	    b = getSyndexAttributeOf (basicExpr, &def);
	  if (b)
	    setNodeReference (aNode, def->getSyndexNode ());
	  else
	    TRACE (cout << "getSyndexAttributeOf not defined (case 2) " << endl);
	}
    }
}

/**
   It adds connections, in the Syndex program, for a delay node (aNode).
*/
_VIRTUAL_ void
DCG2Syndex::translateDelay (ptINode aNode)
{
  ptexpcal vexpsigres, vexpsigret, vexpret, vexpfen, vexpInit;
  ptSyndexAttribute def;

  //eg: ZX:= X $m window n, vexpfen:n,vexpret:m, vexpsigret:X, vexpsigres:ZX
  NM->get_delay_arg (aNode, &vexpsigres, &vexpsigret, &vexpret, &vexpfen, &vexpInit);
  if (getSyndexAttributeOf (vexpsigres, &def))
    setNodeReference (aNode, def->getSyndexNode ());
  translateExpression (vexpsigret);
  addConnection (vexpsigres, vexpsigret, 1);	//there is only one input for delay for version6.6.0
  TRACE (cout << "end of translate delay" << endl);
}


/**
   It adds connections, in the Syndex program, for each input of list (inputs)  to a syndex node (SyndexOperator).
   ouput of the Syndex Node that produces the i-th input to the i-th input of the SyndexOperator node.
   
   \note parameters is unused.
*/
_VIRTUAL_ void
DCG2Syndex::mkCallConnection (ptexpcal parameters, ptexpcal inputs, ptrepstr SyndexOperator)
{
  int i;
  ptexpcal p;

  translateExpression (inputs);
  i = 1;
  while (EM->getNaryArgs (&inputs, &p))
    {
      if (EM->getOperatorClass (p) == opereh)
	p = EM->getWhenOrCellArg (p);
      connectionFromTo (p, SyndexOperator, i);
      i++;
    }
}

/**
   It adds connections, in the Syndex program, (mkCallConnection()) for an external call (aNode).
   Some particular cases: translateStatusNode().
*/
_VIRTUAL_ void
DCG2Syndex::translateExternalCall (ptINode aNode)
{
  ptident name;
  boolean toDo;
  ptexpcal expg, expd, par;

  if (NM->is_intrinsicfunc (aNode))
    {
      toDo = pk_false;
      if (NM->get_nameCall (aNode, &name))
	{
	  if (isStatusId (name))
	    translateStatusNode (aNode);
	  else
	    toDo = !eq_idents (name, string2ident ("writeS")) && !eq_idents (name, string2ident ("readS"));
	}
    }
  else
    toDo = pk_true;

  if (toDo)
    {
      NM->get_procCall_arg (aNode, &name, &par, &expd, &expg);
      mkCallConnection (par, expd, getNodeReference (aNode));
    }
}

/**
   It adds connections, in the Syndex program, for a node call (aNode).
   Similar to mkCallConnection() method.
*/
void
DCG2Syndex::translateNodeCall (ptINode aNode)
{
  ptexpcal parameters, inputs, outputs, p, vs;
  ptDCG aGraph;
  ptrepstr SyndexOperator;
  ptlinstances l;

  NM->get_nodeCall_arg (aNode, &aGraph, &parameters, &inputs, &outputs);
  if (aGraph->isGraphOfExternalProcess ())
    mkCallConnection (parameters, inputs, getNodeReference (aNode));
  else
    {
      SyndexOperator = getNodeReference (aNode);
      translateExpression (inputs);
      l = aGraph->getInputs ();
      while (EM->getNaryArgs (&inputs, &p))
	{
	  ptlinstances_get_next (&l, &vs);
	  if (EM->getOperatorClass (p) == opereh)
	    p = EM->getWhenOrCellArg (p);
	  connectionFromTo2 (p, SyndexOperator, getRepstrIdName (vs));
	}
    }
}

/**
   It adds connections, in the Syndex program, for a reference to a sub-graph (aNode).
   \see mkCallConnection().
*/
void
DCG2Syndex::translateRefSubGraph (ptINode aNode)
{
  ptexpcal lpars, lins, louts;
  getInterface (aNode, &lpars, &lins, &louts);
  mkCallConnection (lpars, lins, getNodeReference (aNode));
}

/**
   It adds connections, in the Syndex program, for an array (aNode).
   \note NOT YET IMPLEMENTED.
*/
void
DCG2Syndex::translateArray (ptINode aNode)
{
  // later.
}

/** 
    Translating a node.
    @param aNode (in) node to be translated.
*/
void
DCG2Syndex::mkConnectionsNode (ptINode aNode)
{
  if (!isSelectedNode (aNode))
    return;

  TRACE (cout << endl << " In Node connection, a new node" << endl);
  TRACE (cout << NM->get_oper (aNode) << " In Node connection, the operator class\n");

  switch ((int) (NM->get_oper (aNode)))
    {
    case opdefsig:
    case opdefpart:
      translateEqu (aNode);	/* X:= E */
      break;

    case opret:
      translateDelay (aNode);	/* x:= E1 $ n window m  */
      break;

    case opappe:
      if (!NM->is_ioNode (aNode))
	translateExternalCall (aNode);
      break;

    case opnodecall:
      translateNodeCall (aNode);
      break;

    case opsubgraph:
      translateRefSubGraph (aNode);
      break;

    case opmoti:		/* array structure. */
      translateArray (aNode);	// later
      break;

    case opbund:		// not yet implemented???
    case opldefpart:		// nothing to do
    case opnil:
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}


/**
   Translating a set of nodes : connections (second Step)
   @param lnodes (in) list of nodes to be translateed.
*/
void
DCG2Syndex::mkConnectionsNodes (ptlINode lnodes)
{
  ptINode aNode;

  while (ptlINode_get_elem (&lnodes, &aNode))
    mkConnectionsNode (aNode);
}
