/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   It specifies that the syndex operation (vstr) is assigned to a group.
*/
  void
DCG2Syndex::setOperationAssigneToGroup (ptrepstr vstr)
{
  SyndexOperations[vstr] = pk_true;
}


/**
   @return true if a node (aNode) is associated with a Syndex node, false otherwise (optimized).
   When a  Syndex node is associated with the node, vt is the Syndex tree representation.
   (ie  represented as a binary tree with 
       - opSyndexHierarchy as operator
       - the name of the Syndex main algoritm, for the first child
       - 
    The textual form is \\first_child\second_child

*/
boolean
DCG2Syndex::mkRefNodeSyndexTree (ptINode aNode, ntrees * vt)
{
  ptrepstr vstr;
  ntrees n;

  vstr = getNodeReference (aNode);
  if (vstr == nilstr)
    return pk_false;		// exemple: x:= y when c, y calculÃ© dans un autre algo (entrÃ©e de l'algo courant)
  setOperationAssigneToGroup (vstr);
  *vt = TBI->mkHierarchyreference (vstr, CurrentSubGraph == 0);
  // h = get_nodeClock(); hfather() not useful here, it is a STS graph.
  return pk_true;
}


void
DCG2Syndex::assignNodeUndefinedRunOn (ptINode aNode)
{
  ntrees vt;
  if (mkRefNodeSyndexTree (aNode, &vt))
    TBI->addElemUndefinedGroup (vt);
}


void
DCG2Syndex::assignNode (ptINode aNode, ptpragma aPragma)
{
  ptrepstr EnonceValue;
  integer i, nbEnonce;
  ntrees tEnonce, vt;

  tEnonce = getPragmaStatement (aPragma);
  if (ISNOT_NILTREE (tEnonce))
    {
      // if (NM->isNodeCall(aNode)) louts = NM->getInodelOutputFlows(aNode); else louts = NILLINSTANCES;
      nbEnonce = ntreearity (tEnonce);
      // I don't know if Syndex accepts code dupplication.(LB, Nov 2011)
      for (i = 1; i <= nbEnonce; i++)
	{
	  EnonceValue = nstringof (nchild (i, tEnonce));
	  if (mkRefNodeSyndexTree (aNode, &vt))
	    TBI->addElemGroup (EnonceValue, vt);
	  // lsig = louts; while (ptlinstances_get_next(&lsig, &vsig)) ptINodeAttr_append_mark( EM->getNode(vsig), Mark); 
	}
    }
}


/**
   When the predefined pragma (RunOn, or Binding or BindingIO) is associated with a node (aNode), 
   the constraint of the partitionning is exported to synDEx.
 */
void
DCG2Syndex::genRunOnPartionning (ptINode aNode)
{
  ptpragma aPragma;

  if (isSelectedNode (aNode))
    {
      if (getPragmaKeyword (NM->getPragmas (aNode), PRAGMA_RUN_ON, &aPragma) || getPragmaKeyword (NM->getPragmas (aNode), PRAGMA_BINDINGIO, &aPragma))
	assignNode (aNode, aPragma);
      else
	assignNodeUndefinedRunOn (aNode);
    }
}


/**
   For each element of the operand list, it applies the method genRunOnPartionning(ptINode).
   @param lnodes (in) list of nodes to visit.
*/
void
DCG2Syndex::genRunOnPartionning (ptlINode lnodes)
{
  ptINode aNode;

  while (ptlINode_get_elem (&lnodes, &aNode))
    genRunOnPartionning (aNode);

  ntrees vt;
  SyndexTableOperationType::iterator it;
  ptrepstr vsyndexoperation;

  for (it = SyndexOperations.begin (); it != SyndexOperations.end (); it++)
    {
      vsyndexoperation = (*it).first;
      if (!SyndexOperations[vsyndexoperation])
	{
	  vt = TBI->mkHierarchyreference (vsyndexoperation, CurrentSubGraph == 0);
	  TBI->addElemUndefinedGroup (vt);
	}
    }
}
