/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/** Constructor.
   Initializing the translator for a root graph.
   It uses a Tree Builder Interface (TBI) object to make trees.
*/
  DCG2Syndex::DCG2Syndex (ptDCG g)
{
  init ();
  TBI = new SyndexTreeInterface ();
  initializeForGraph (g);
}

/**
   Destructor.
*/
_VIRTUAL_ DCG2Syndex::~DCG2Syndex ()
{
  std::list < ptSyndexAttribute >::iterator i;
  // Attribute cleaning.
  EM->resetMarkAll (vSyndexAttribute);
  EM->resetMarkAll (equivAttribute);
  EM->resetMarkAll (vConnectionMark);
  EM->resetMarkAll (setOperationsMark);
  // To do Cleaning lattr list
  for (i = lattr.begin (); i != lattr.end (); i++)
    delete *
      i;
  delete
    TBI;
}

/** Entry point of the translation.
    @param MainProg (in) specifies code generation mode :
      - at "true" when the program is a main one (using actuator/sensor for Syndex),
      - at "false" it will be used as a separate algorithm in syndex.
   @param bPrefix (in) when the name of the subgraph assigned to a clock must be prefixed by the 
   name of the graph (to avoid the same name for several algorithms).
   @param  AlgoSyndexTree (out): result of the translation (abstract tree).
   This tree must be printed outside.
   @return true when translation succeeds (AlgoSyndexTree is available), false otherwise.
 */
_VIRTUAL_
  boolean
DCG2Syndex::translate (boolean MainProg, boolean bPrefix, ntrees * AlgoSyndexTree)
{
  asMain = MainProg;
  if (bPrefix)
    addPrefixForAlgorithm = bPrefix;

  TBI->initialize ();
  translateConstantsDeclarations ();
  //For Syndex, the local variables are not translated.
  TBI->setMainAlgoName (getStrClockName (TheTick));
  translateloc ();
  translateArchitecture ();
  if (asMain)
    TBI->mkMainAlgorithmDeclaration (mkProgramParameters ());
  finalize ();
  (*AlgoSyndexTree) = TBI->getResult ();

  printInfosRestrictions ();
  return returnCode;
}
