/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define TRACE(s)
#define getRepstrIdName(s) ident2repstr(EM->getName(s))	//to repstr
#define getStringIdName(s) ident2string(EM->getName(s))	//to string
#define ISRELATIONNALOP(op) (op == opdiff) || (op == opegal) \
|| (op == opegeg) || (op == opiegal ) || (op == opinfe ) \
|| (op == opsegal ) || (op ==  opsupe) || (op == opiieg)
/**
   Interface conversion.
   
   @param aNode (in) a Signal Node.
   @param lpars (out) the combed expression of the parameters referenced in aNode.
   @param lins (out) the combed expression of the signal referenced in aNode.
   @param louts (out) the combed expression of the signal defined in aNode.
 */
  void
DCG2Syndex::getInterface (ptINode aNode, ptexpcal * lpars, ptexpcal * lins, ptexpcal * louts)
{
  ptinterface vinterface;

  vinterface = NM->get_interface (aNode);
  *lpars = EM->convertLinstances2ListExpcal (ptinterface_get_parameters (vinterface));
  *lins = EM->convertLinstances2ListExpcal (ptinterface_get_inputSignals (vinterface));
  *louts = EM->convertLinstances2ListExpcal (ptinterface_get_outputSignals (vinterface));
}

/**
   @return the explicit conversion to a type (given by it name, id) for the type of an expression (exp), when the list of pragmas (lp) contains the 
   pragma CONVERSION, id otherwise.

   example : dreal(x) with type(x)=real is replaced by real_2_dreal(x).
 */
ptident DCG2Syndex::explicitConversion (ptlpragmas lp, ptexpcal vexp, ptident id)
{
  TypeType
    vtype;
  ptident
    id1;

  if (!existPredefinedPragma (lp, PRAGMA_CONVERSION))
    return id;
  if (EM->satisfiesList (vexp))
    vexp = EM->getNaryArgs (&vexp);
  vtype = EM->getType (vexp);
  return makeIdent_string_repstr_string_ident ("pK_", predefinedType2ident (predefTypeType2SignalType (vtype)), "2_", id);
}

/**
   @return the name of a clock (h) prexifed by the name of current graph if addPrefixForAlgorithm (attribute member of the class).
*/
ptrepstr DCG2Syndex::getStrClockName (ptinsth h)
{
  ptrepstr
    id;

  id = ident2repstr (CKM->getName (h));
  if (addPrefixForAlgorithm)
    makeRepstr_ident_repstr (graphName, id, &id);
  return id;
}

/** @return true if the expression has been travelled during the connection step.
*/
boolean DCG2Syndex::isAlreadyConnected (ptexpcal s)
{
  return EM->isMarked (s, vConnectionMark);
}

/** @return true if the expression has been travelled during the connection step.
*/
void
DCG2Syndex::belongConnected (ptexpcal s)
{
  EM->addMark (s, vConnectionMark);
}

/** Setting an equivalence form leftPart to basicExpr. */
void
DCG2Syndex::setIndirectionTo (ptexpcal leftPart, ptexpcal basicExpr)
{
  EM->addAttribute (leftPart, equivAttribute, (attributeValue) basicExpr);
}

/** Setting an equivalence form leftPart to basicExpr. */
ptexpcal DCG2Syndex::getIndirection (ptexpcal v)
{
  ptexpcal
    indi;
  if (!EM->getAttribute (v, equivAttribute, (attributeValue *) & indi))
    return v;
  return getIndirection (indi);
}

/**
   @return an attribute (node, n CurrentSubGraph), where CurrentSubGraph is the number of the current subgraph.
*/
ptSyndexAttribute DCG2Syndex::mkSyndexAttribute (ptrepstr node, ptrepstr n)
{
  ptSyndexAttribute
    x;
  // To do : build a list of mkSyndexAttribute for cleaning.
  x = new SyndexAttribute (node, n, CurrentSubGraph);
  lattr.push_front (x);
  return x;
}


/** It assignes an attribute to a Signal expression.
   This attribute is the name of the Syndex reference.
*/
void
DCG2Syndex::setReferenceTo (ptexpcal s, ptSyndexAttribute vattr)
{
  ptStack S;

  if (!EM->getAttribute (s, vSyndexAttribute, (attributeValue *) & S))
    S = NILSTACK;
  S = stackPush (S, vattr);
  EM->addAttribute (s, vSyndexAttribute, (attributeValue) S);
  // else (input, or output of the current subgraph)
}



/** @return the name of the Syndex node assigned to a Signal expression (s).
*/
boolean DCG2Syndex::getReferenceTo (ptexpcal s, ptrepstr * vattr)
{
  ptStack
    S;
  ptSyndexAttribute
    x;
  ptElemStack
    y;

  s = getIndirection (s);
  if (!EM->getAttribute (s, vSyndexAttribute, (attributeValue *) & S))
    return pk_false;
  if (!stackTop (S, &y))
    return pk_false;
  x = (ptSyndexAttribute) y;
  *vattr = x->getSyndexNode ();
  return pk_true;
}

/** @return the Syndex attribute assigned to a Signal expression.
   This attribute is the name of the Syndex reference.
*/
boolean DCG2Syndex::isAlreadyReferenced (ptexpcal s)
{
  ptStack
    S;
  ptSyndexAttribute
    x;
  ptElemStack
    y;

  s = getIndirection (s);
  if (!EM->getAttribute (s, vSyndexAttribute, (attributeValue *) & S))
    return pk_false;
  if (!stackTop (S, &y))
    return pk_false;
  x = (ptSyndexAttribute) y;
  return (x->getSubGraphNumber () == CurrentSubGraph);
}


/**
   @return "true" if the attribute (vattr) assigned to an expression (s) has the same graph number than the current one (CurrentSubGraph),
   "false" otherwise. vattr is always returned if it exists.
*/
boolean DCG2Syndex::getSameLevelSyndexAttributeOf (ptexpcal s, ptSyndexAttribute * vattr)
{
  if (!getSyndexAttributeOf (s, vattr))
    return pk_false;
  return ((*vattr)->getSubGraphNumber () == CurrentSubGraph);
}


/** @return the Syndex attribute assigned to a Signal expression.
   This attribute is the name of the Syndex reference.
*/
boolean DCG2Syndex::getSyndexAttributeOf (ptexpcal s, ptSyndexAttribute * vattr)
{
  ptStack
    S;

  s = getIndirection (s);
  if (!EM->getAttribute (s, vSyndexAttribute, (attributeValue *) & S))
    return pk_false;
  if (!stackTop (S, (pointer *) vattr))
    return pk_false;
  return pk_true;
}

/** It pops attribute assigned to an expression. */
void
DCG2Syndex::popReference (ptexpcal s)
{
  ptStack S;

  if (EM->getAttribute (s, vSyndexAttribute, (attributeValue *) & S))
    {
      S = stackPop (S);
      EM->addAttribute (s, vSyndexAttribute, (attributeValue) S);
    }
}

/**
   It associates the Syndex node (SyndexOperator) to the Signal node (n).
*/
void
DCG2Syndex::setNodeReference (ptINode n, ptrepstr SyndexOperator)
{
  NM->addAttribute (n, vSyndexAttribute, (attributeValue) SyndexOperator);
}

/**
   @return the Syndex node associated with a Signal node (n), previously set by setNodeReference(), 
   or UNDEF_NODE if it does not exist.
*/
ptrepstr DCG2Syndex::getNodeReference (ptINode n)
{
  ptrepstr
    SyndexOperator;
  if (!NM->getAttribute (n, vSyndexAttribute, (attributeValue *) & SyndexOperator))
    return StrUndefNode;
  return SyndexOperator;
}

/**
   It declares the Syndex algorithm associated with a clock (aClock), for example "def algorithm CLK_err2_1", 
   where CLK_err2_1 is the name of the clock.
   Moreover, the inputs of the sub-graph associated with the clock (set in translateGraphClockInit) are 
   connected to the nodes that produce them:
   For an input, called Y, produced in a node (N) by the output called O, the connection\n
         strong_precedence_data N.O -> CLK_err2_1_1.Y \n
       is added.
*/
void
DCG2Syndex::declareSubGraphOf (ptinsth aClock)
{
  ptrepstr currentName, dn, on, name;
  ptlinstances l;
  ptexpcal anInput;
  ptSyndexAttribute argAttr;

  currentName = getStrClockName (aClock);
  name = TBI->declareAlgorithmInMain (currentName);
  l = inputs;
  while (ptlinstances_get_next (&l, &anInput))
    if (getSyndexAttributeOf (anInput, &argAttr))
      {
	dn = argAttr->getSyndexNode ();
	on = argAttr->getSyndexOutput ();
	TBI->addConnectionFromToAtInMain (dn, on, name, getRepstrIdName (anInput));
      }
    else;			// ??????????
}


/** Translating recursively a hierarchical graph.
   @param aClock (in) clock root of the hierarchy.
*/
void
DCG2Syndex::translateHierarchy (ptinsth aClock)
{
  ptlinsth lch;			// Children of the clock aClock.
  ptinsth aChild;

  if (isSelectedClock (aClock))
    {
      translateMainGraphClock (aClock);

      lch = CKM->getChildren (aClock);
      // In STS : One level
      while (ptlinsth_get_next (&lch, &aChild))
	if (isSelectedClock (aChild))
	  if (translateGraphClock (aChild))
	    declareSubGraphOf (aChild);

      TBI->addCurrentAlgoTree (TBI->getCurrentClockAlgo ());	//add the current clock Algorithm at the end of the tree.
    }
}

/** Translating recursively a hierarchical graph.
   @param aClock (in) clock root of the hierarchy.
*/
void
DCG2Syndex::setDefinedInSubGraphs (ptinsth aClock)
{
  ptlinsth lch;			// Children of the clock aClock.
  ptinsth aChild;

  if (isSelectedClock (aClock))
    {
      if (!asMain)
	setDefinedInSubGraph (aClock);
      // Master clock. : nothing to do
      lch = CKM->getChildren (aClock);
      // In STS : One level
      while (ptlinsth_get_next (&lch, &aChild))
	if (isSelectedClock (aChild))
	  if (gfd_isSelected (CKM->getClockSubGraph (aChild)))
	    setDefinedInSubGraph (aChild);
    }
}


/**
   It translates a hierachy with TheTick as root. TheTick is an attribute member of the class.
   It calls:
   - setDefinedInSubGraphs() with the current TheTick clock, 
   - then translateHierarchy() with the current TheTick clock.
*/
void
DCG2Syndex::translateloc ()
{
  // translateInternalGraphs();
  setDefinedInSubGraphs (TheTick);
  translateHierarchy (TheTick);
}


/**
   It translates the Node call nodes of the current program.
   \note Currently, not referenced.
*/
void
DCG2Syndex::translateInternalGraphs ()
{
  ptINode n;
  ptlINode list;
  ptexpcal expg, expd, par;
  ptDCG aGraph;
  ptDCG2Syndex gen;

  if (NM->getNodeCalls (&list))
    while (ptlINode_get_elem (&list, &n))
      {
	NM->get_nodeCall_arg (n, &aGraph, &par, &expd, &expg);
	if (!aGraph->isGraphOfExternalProcess ())
	  {
	    gen = new DCG2Syndex (aGraph);
	    gen->translateloc ();
	    delete gen;
	  }
      }
}




/** Initializing.
   Three tables are used in this translator.
   ReferenceCountersTable : Associative table for reference counter of each operators(operator+external calls) in Syndex.
   OperatorNamesTable : Associative table for the name of operators. It links the Signal operator to Syndex operator.
   TypeTable : Associative table for types,(between Signal and Syndex).
*/
void
DCG2Syndex::TablesInitialize ()
{
  initTypeTable ();
  //here the operators are all signal operators.

  /*read & write */
  OperatorNamesTable[opread] = string2repstr ("input");
  implementedForVector[opread] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opread]] = 0;

  OperatorNamesTable[opwrite] = string2repstr ("output");
  implementedForVector[opwrite] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opwrite]] = 0;

  /* constants */
  OperatorNamesTable[opcstent] = string2repstr ("cst");
  implementedForVector[opcstent] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[opcstent]] = 0;
  OperatorNamesTable[opcstcar] = string2repstr ("cst");
  implementedForVector[opcstcar] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[opcstcar]] = 0;
  OperatorNamesTable[opcstcha] = string2repstr ("cst");
  implementedForVector[opcstcha] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[opcstcha]] = 0;
  OperatorNamesTable[opcstreel] = string2repstr ("cst");
  implementedForVector[opcstreel] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[opcstreel]] = 0;
  OperatorNamesTable[opcstdp] = string2repstr ("cst");
  implementedForVector[opcstdp] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[opcstdp]] = 0;
  OperatorNamesTable[opfalse] = string2repstr ("cst");
  implementedForVector[opfalse] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[opfalse]] = 0;
  OperatorNamesTable[optrue] = string2repstr ("cst");
  implementedForVector[optrue] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[optrue]] = 0;

  //constant list definition
  OperatorNamesTable[oplelevec] = string2repstr ("n_cst");	//this one shouldn't appear in version6.6
  implementedForVector[oplelevec] = pk_false;
  ReferenceCountersTable[OperatorNamesTable[oplelevec]] = 0;
  //constant def end

  /* Binary operators */
  OperatorNamesTable[opplus] = string2repstr ("Arit_add");
  implementedForVector[opplus] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opplus]] = 0;

  OperatorNamesTable[opdivi] = string2repstr ("Arit_div");
  implementedForVector[opdivi] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opdivi]] = 0;

  OperatorNamesTable[opmod] = string2repstr ("Arit_mod");
  implementedForVector[opmod] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opmod]] = 0;

  OperatorNamesTable[opmult] = string2repstr ("Arit_mul");
  implementedForVector[opmult] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opmult]] = 0;

  OperatorNamesTable[opmoin] = string2repstr ("Arit_sub");
  implementedForVector[opmoin] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opmoin]] = 0;

  OperatorNamesTable[oppuis] = string2repstr ("pow");	//x**n, To see: give defs in libs later
  implementedForVector[oppuis] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[oppuis]] = 0;

  OperatorNamesTable[opdiff] = string2repstr ("Log_ne");	// != 
  implementedForVector[opdiff] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opdiff]] = 0;

  OperatorNamesTable[opegal] = string2repstr ("Log_eq");	//To see: for syndex, ==, = are same?
  implementedForVector[opegal] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opegal]] = 0;

  OperatorNamesTable[opegeg] = string2repstr ("Log_eq");	// ==
  implementedForVector[opegeg] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opegeg]] = 0;

  OperatorNamesTable[opiegal] = string2repstr ("Log_ngt");	// <= 
  implementedForVector[opiegal] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opiegal]] = 0;

  OperatorNamesTable[opinfe] = string2repstr ("Log_lt");	// <
  implementedForVector[opinfe] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opinfe]] = 0;

  OperatorNamesTable[opsegal] = string2repstr ("Log_nlt");	// >=
  implementedForVector[opsegal] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opsegal]] = 0;

  OperatorNamesTable[opsupe] = string2repstr ("Log_gt");	// >
  implementedForVector[opsupe] = pk_false;	//true;
  ReferenceCountersTable[OperatorNamesTable[opsupe]] = 0;

  OperatorNamesTable[opiieg] = string2repstr ("lessEqEq");	// <== To see: give defs in libs later
  implementedForVector[opiieg] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opiieg]] = 0;

  OperatorNamesTable[opelem] = string2repstr ("AccessValue");
  implementedForVector[opelem] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opelem]] = 0;

  /*  delay and window operators */
  //simple delay is treated as n-delay<1>, reference to IGr..NodeRef_i.c
  //OperatorNamesTable[opret] = string2repstr("delay"); // simple delay alone. ($1)
  //  implementedForVector[opret] = string2repstr("delay"); // simple delay alone. ($1)
  //ReferenceCountersTable[OperatorNamesTable[opret]] = 0;

  OperatorNamesTable[opretN] = string2repstr ("delay");	// delay ($ n)
  implementedForVector[opretN] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opretN]] = 0;

  OperatorNamesTable[opfen] = string2repstr ("window");	// window alone.
  implementedForVector[opfen] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opfen]] = 0;

  OperatorNamesTable[opretfen] = string2repstr ("delayed_window");	// delay+window.
  implementedForVector[opretfen] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opretfen]] = 0;

  OperatorNamesTable[opet] = string2repstr ("Log_and");
  implementedForVector[opet] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opet]] = 0;

  OperatorNamesTable[opou] = string2repstr ("Log_or");
  implementedForVector[opou] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opou]] = 0;

  OperatorNamesTable[opouex] = string2repstr ("Log_xor");	// Exclusive Or//it's not same with syndex libs, must write m4 definition for this
  implementedForVector[opouex] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opouex]] = 0;

  /* unary */
  OperatorNamesTable[opnega] = string2repstr ("Arit_minus");	/* Unary Minus */
  implementedForVector[opnega] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opnega]] = 0;

  OperatorNamesTable[opnon] = string2repstr ("Log_not");	//it's not same with syndex libs, must write m4 definition for this
  implementedForVector[opnon] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opnon]] = 0;

  /*********the operators of SIGNAL*********/
  //this is for left_tt
  OperatorNamesTable[opleft_tt] = string2repstr ("left_tt");
  implementedForVector[opleft_tt] = pk_false;	// pk_true;
  ReferenceCountersTable[OperatorNamesTable[opleft_tt]] = 0;

  //for cell.
  OperatorNamesTable[opmemo] = string2repstr ("cell");
  implementedForVector[opmemo] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opmemo]] = 0;

  //Use for coding state variable ***Warning***Warning***Warning
  OperatorNamesTable[opvvarin] = string2repstr ("stateVar");
  implementedForVector[opvvarin] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opvvarin]] = 0;


  //this is for default
  OperatorNamesTable[opmele] = string2repstr ("default");
  implementedForVector[opmele] = pk_true;
  ReferenceCountersTable[OperatorNamesTable[opmele]] = 0;
}

void
DCG2Syndex::restriction (int codeRestriction)
{
  // if (codeRestriction == 1) Restriction_DRealType = pk_true;
}

void
DCG2Syndex::printInfosRestrictions ()
{
  if (Restriction_DRealType)
    pK_messenger->printString ("***** WARNING Syndex restriction : the dreal type has been replaced by the real type.******");
}


/** Initializing. */
void
DCG2Syndex::initTypeTable ()
{
  TypeTable[BooleanType] = string2repstr ("bool");
  TypeTable[PureType] = string2repstr ("bool");
  TypeTable[ShortType] = string2repstr ("ushort");
  TypeTable[IntegerType] = string2repstr ("int");
  TypeTable[LongType] = string2repstr ("lint");
  TypeTable[RealType] = string2repstr ("float");
  TypeTable[DRealType] = string2repstr ("double");	// never referenced (double is replaced by float).

  TypeTable[ComplexType] = string2repstr ("Complex");
  TypeTable[DComplexType] = string2repstr ("DComplex");
  TypeTable[CharType] = string2repstr ("char");
  TypeTable[StringType] = string2repstr ("String");

  TypeTable[ErrorType] = string2repstr ("UNDEF");
  TypeTable[LabelType] = string2repstr ("UNDEF");
  TypeTable[ByteType] = string2repstr ("UNDEF");
  TypeTable[WordType] = string2repstr ("UNDEF");

  //in Syndex, should have uchar and FILE type. think again later.
}


boolean DCG2Syndex::setRunOnPartitionningMode ()
{
  ptlpragmas
    lprag;
  ptpragma
    vpragma;

  lprag = AGRAPH->getPragmas ();
  while (ptlpragmas_get_next (&lprag, &vpragma))
    {
      if (isPragmaKeyword (vpragma, PRAGMA_RUN_ON))
	return pk_true;
      if (isPragmaKeyword (vpragma, PRAGMA_BINDINGIO))
	return pk_true;
    }
  return pk_false;
}



/**
   It initialises the translator for a graph (g).
   The attributes member (See DCG2Syndex class) are set:
   AGRAPH, CKM, EM, NM, graph_type, unNodes,  unClocks, inputs, outputs, graphName, TheTick, CurrentSubGraph (-1).
*/
void
DCG2Syndex::initializeForGraph (ptDCG g)
{
  AGRAPH = g;
  CKM = AGRAPH->getClockManager ();
  EM = AGRAPH->getExpressionManager ();
  NM = AGRAPH->getNodeManager ();
  graph_type = AGRAPH->getFormatLevel ();
  unNodes = AGRAPH->getUnused (&unClocks);
  inputs = AGRAPH->getInputs ();
  outputs = AGRAPH->getOutputs ();
  graphName = AGRAPH->getName ();
  //Master clock:it exists for bDC+, STS, DC format(For Syndex it is STS level.)
  TheTick = CKM->getTickClock ();
  CurrentSubGraph = -1;
  isRunOnPartionning = setRunOnPartitionningMode ();
}


void
DCG2Syndex::init ()
{
  TablesInitialize ();
  isRunOnPartionning = pk_false;
  vSyndexAttribute = getMark ();	// link Signal -> Syndex
  equivAttribute = getMark ();	// signal equivalence (x:=y or x:= y when c)
  vConnectionMark = getMark ();	// sets to expression when it is "connected"
  setOperationsMark = getMark ();	// sets to a node, it is the set  of operations defined in Syndex for the translation.
  MyStdOutput = string2repstr ("o");
  StrUndefNode = string2repstr ("UNDEF_NODE");
  returnCode = pk_true;
  Restriction_DRealType = pk_false;	// Double replaced by float.
}

void
DCG2Syndex::finalize ()
{
  ptlpragmas lprag;
  ptpragma vpragma;
  ntrees tEnonce;

  // getting the partitionning info for the undefined group.
  lprag = AGRAPH->getPragmas ();
  if (getPragmaStrKeyword (lprag, string2repstr ("OBinding"), &vpragma))
    tEnonce = getPragmaStatement (vpragma);
  else
    SET_NILTREE (tEnonce);
  TBI->finalize (tEnonce);
}
