/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
#define extension ".sdx"
#define vernumber "7.0.6"
#define GenByPolychrony "Generated by Polychrony "
#define mainDeclaration "main algorithm "
#define mainArchitectureComment "main architecture must be defined and called IMA"
#define commentImportDuration "The following line must be replaced by the user!!!"
#define beginComment "# "
#define sepLine         "#-------------------------"
/********define the keywords of syndex**********/
#define version         "syndex_version : "
#define leftbracket     "<"
#define rightbracket    ">"
#define leftbracket1     "["
#define rightbracket1    "]"
#define aslash          "\\"
#define quotation       "\""
#define incl            "include "
#define defi            "def "
#define input           "? "
#define output          "! "
#define semicolon       ";"
#define colon           ":"
#define dot           "."
#define equal           " = "
#define dpdsymbol       " -> "
#define cds             "conditions: "
#define rfs             "references: \n"
#define dpd             "dependences:\n"
#define leftbigbra      "{"
#define rightbigbra     "}"
#define comma           ","	//for link of initial value of delay,{1,2,3}
#define tkGroup "operation_group"
#define tkGroupConstraint "absolute constraint"
#define kwOn "on"
#define tkMinus "-"
#define tkMult  "*"
#define tkDivi  "/"
#define tkAdd   "+"
#define tkArchi "architecture"
#define tkOperators "operators:"
#define tkOperator "operator"
#define tkMedias "medias:"
#define tkMedia "media"
#define tkConnections "connections:"
#define tkCodePhases "code_phases:"
#define kwGate "gate"
#define kwimport "--importDuration"
#define kwwarning "         Warning: "
/**
   It opens the file (name) for writing, using the attribute member of the class called fs.
   The effective name is given by the sys_filePathInUserOutputDir() method, the suffix ".sdx" is added to the name of the file.
   To close it, use closeFile().
*/
  void
SyndexPrinter::openFile (ptrepstr name)
{
  boolean err;
  char *filename;
  ptrepstr processName;

  processName = name;
  sys_filePathInUserOutputDir (name, &name);
  if (!makeRepstr_repstr_string (name, extension, &name))
    err = pk_true;
  else
    {
      filename = repstr2string (name);
      fs.open (filename, ios::out);
      err = fs.fail ();
    }
  if (err)
    pK_messenger->globalMessageStr (error_openingFile, name);
  else
    pK_messenger->globalMessageStrStr (messGeneratorFile, string2repstr ("SynDEx"), name);
}

/**
   It closes the file (fs: attribute member of the class) previously opened by openFile().
*/
void
SyndexPrinter::closeFile ()
{
  fs.close ();
}

/**
   @return true if the operation succeeds, false otherwise.
   If the file (filnam.sdx) exists it is inserted in the current the current generated file.
*/
boolean SyndexPrinter::includeFile (ptrepstr filnam)
{
  char *
    filename;

  makeRepstr_repstr_string (filnam, extension, &filnam);
  sys_filePathInUserOutputDir (filnam, &filnam);
  filename = repstr2string (filnam);
  fs << "# Imported file:" << filename << endl;
  fduration.open (filename, ios::in);
  if (fduration.fail ())
    return pk_false;
  fs << fduration.rdbuf ();
  fduration.close ();
  bPrintListSeparator = pk_false;
  return pk_true;
}

/**
   @return "true" is the n-th child (operand) of a tree (t) is not the null tree (NILTREE), "false" otherwise.
   
*/
boolean SyndexPrinter::getChild (ntrees t, int n, ntrees * operand)
{
  *operand = nchild (n, t);
  return (ISNOT_NILTREE (*operand));
}

/**
   @return "true" is the n-th child of a tree (t)x is not the null tree (NILTREE), "false" otherwise.
*/
boolean SyndexPrinter::isChild (ntrees t, int n)
{
  ntrees
    operand;
  operand = nchild (n, t);
  return (ISNOT_NILTREE (operand));
}

/**
   It prints the Polychrony version use to produce the Syndex code.
*/
_VIRTUAL_ void
SyndexPrinter::printPolychronyVersion ()
{
  printKw (beginComment);
  printKw (GenByPolychrony);
  printString (sysGetSignalVersion ());
  printBL ();
  printBL ();
}

/**
   It prints the Syndex version used to define the Syndex code.
*/
void
SyndexPrinter::printSyndexVersion ()
{
  printBL ();
  printKw (version);
  printKw (quotation);
  printString (vernumber);
  printKw (quotation);
  printBL ();
}

/**
   It prints a tabulation.
*/
void
SyndexPrinter::printTabul ()
{
  fs << "  ";
}


/**
   It prints an int (c).
*/
void
SyndexPrinter::printInt (int c)
{
  fs << c;
}

/**
   It prints a character (c).
*/
void
SyndexPrinter::printChar (char c)
{
  fs << c;
}


/**
   It prints a string (kw).
   Equivalent to printKw().
*/
void
SyndexPrinter::printString (const char *s)
{
  fs << s;
}

/**
   It prints a string (e), internal string representation.
*/
void
SyndexPrinter::printString (ptrepstr e)
{
  printString (repstr2string (e));
}

/**
   It prints a string (kw).
*/
void
SyndexPrinter::printKw (const char *kw)
{
  fs << kw;
}

/**
   It prints a string (kw) followed by a new line.
*/
void
SyndexPrinter::printKwLine (const char *kw)
{
  fs << kw << endl;
}

/**
   It prints a beginning of a comment followed by a string (kw) and e new line.
*/
void
SyndexPrinter::printCommentKw (const char *kw)
{
  fs << beginComment << kw << endl;
}

/**
   It prints a new line.
*/
void
SyndexPrinter::printBL ()
{
  fs << endl;
}

/**
   It prints a space character.
*/
void
SyndexPrinter::printSpc ()
{
  printString (" ");
}

/**
   It prints the "error" string.
*/
void
SyndexPrinter::printError ()
{
  printString ("~error~");
}

/**
   It prints an atomic tree (t) that represents a string (assumed).
*/
void
SyndexPrinter::printStrTree (ntrees t)
{
  printString (repstr2string (nstringof (t)));
}

/**
   It prints a Syndex algorithm type (t) according to the value associated with t.
   The possible values are: constant, sensor, actuator, memory, algorithm.
*/
void
SyndexPrinter::printAlgoType (ntrees t)
{
  printKw (defi);
  switch (nintof (t))
    {
    case altyconstant:		//constant
      printString ("constant");
      break;
    case altysensor:		//sensor
      printString ("sensor");
      break;
    case altyactuator:		//actuator
      printString ("actuator");
      break;
    case altymemory:		//memory
      printString ("memory");
      break;
    case altyalgorithm:	//algorithm
      printString ("algorithm");
      break;
    default:
      printString ("error in algotype convertion");
    }
  printSpc ();
}

/**
   It prints a Syndex dependence (t) according to the value associated with t.
   The possible dependencies are: strong_precedence_data, weak_precedence_data, precedence, data.
*/
void
SyndexPrinter::printDpdType (ntrees t)
{
  switch (nintof (t))
    {
    case strong:
      printString ("strong_precedence_data ");
      break;
    case weak:
      printString ("weak_precedence_data ");
      break;
    case precedence:
      printString ("precedence ");
      break;
    case data:
      printString ("data ");
      break;
    default:
      printString ("error in dpdtype convertion");
    }
}

/**
   @return the arity of a Syndex operator (voperator), 4 for a list, -1 for an out of bounds.
 */
int
SyndexPrinter::getOperatorArity (noperators voperator)
{
  if ((voperator > BaseSyndexOperator) && (voperator <= SYN_NB_OPERATORS))
    {
      if (voperator < opSyndexBinary)
	{
	  if (voperator < opSyndexUnary)
	    return 0;
	  return 1;
	}
      if (voperator >= opSyndexTernary)
	return 3;
      if (voperator < opSyndexList)
	return 2;
      return 4;			// list
    }
  return -1;			// out of bounds --> -1
}

/**
   It prints the n-th child of a tree (t), if it exists.
*/
void
SyndexPrinter::printNthChild (ntrees t, int n)
{
  ntrees o;
  if (getChild (t, n, &o))
    unparse_subtree (o);
}


/**
   It prints the n-th child of a tree (t), if it exists.
*/
void
SyndexPrinter::printNthChild (ntrees t, int n, boolean bglue)
{
  ntrees o;
  if (getChild (t, n, &o))
    unparse_subtree (o, bglue);
}

/**
   It prints an atomic tree (t).
*/
void
SyndexPrinter::printAtomic (ntrees t)
{
  switch ((int) noper (t))
    {
    case opincltype:		// "Signal_<t>.sdx";
      printKw (quotation);
      printStrTree (t);
      printKw (extension);
      printKw (quotation);
      break;

    case opalgotype:
      printAlgoType (t);
      break;

    case opalgoname:
      printStrTree (t);
      printSpc ();
      break;


    case opintype:		//? int 
      printKw (input);
      printStrTree (t);
      printSpc ();
      break;

    case opoutype:		// ! 
      printKw (output);
      printStrTree (t);
      printSpc ();
      break;

    case opinterfacename:	//A;
      printSpc ();
      printStrTree (t);
      break;

    case opvalue:
      printInt (nintof (t));
      break;

    case oprefname:
      if (!glue)
	printSpc ();
      printStrTree (t);
      break;

    case opdpdtype:
      printDpdType (t);
      break;

    case opcdsname:
    case opreftype:
    case opleftdpd:
    case oprightdpd:
    case opSyndexConstant:
    case opSyndexId:
      printStrTree (t);
      break;

    default:
      printBL ();
    }

}

/**
   If a file named durations_<t> exists in the target directory, it is included in the produced file
   otherwise the non correct Syndex line (--importDuration durations_<t>;)is generated. In this last case, the user must modify the
   produced file: a warning is send to the user.
*/
void
SyndexPrinter::printOrIncludeDuration (ntrees t)
{
  ntrees name;
  if (getChild (t, 1, &name))
    if (!includeFile (nstringof (name)))
      {
	printCommentKw (commentImportDuration);
	printKw (kwimport);
	unparse_subtree (name);
	bimportDuration = pk_true;
      }
}

/**
   It prints an unary tree(t).
*/
void
SyndexPrinter::printUnary (ntrees t)
{
  switch ((int) noper (t))
    {
    case opSyndexUminus:
      printKw (tkMinus);
      printNthChild (t, 1);
      break;

    case opimportduration:
      printOrIncludeDuration (t);
      break;

    case opinclude:
      printKw (incl);
      printNthChild (t, 1);
      break;

    case opinitvalue:
      if (noper (nchild (1, t)) == opSyndexlexpr)
	printParameters (nchild (1, t));
      else
	{
	  printKw (leftbigbra);
	  printNthChild (t, 1);
	  printKw (rightbigbra);
	}
      break;

    case opSyndexElevec:
    case opparameter:
    case oppararef:
    case opinterface:
      printNthChild (t, 1);
      break;

    }
}

/**
   It prints a list of parameters (t).
*/
void
SyndexPrinter::printParameters (ntrees t)
{
  int n, i;

  n = ntreearity (t);

  if (n != 0)
    {
      if (noper (t) == opSyndexlexpr)
	printKw (leftbigbra);
      else
	printKw (leftbracket);
      for (i = 1; i <= n; i++)
	{
	  printNthChild (t, i);
	  if (i != n)
	    {
	      if (noper (t) == opSyndexlexpr)
		printKw (comma);
	      else
		printKw (semicolon);
	    }
	}
      if (noper (t) == opSyndexlexpr)
	printKw (rightbigbra);
      else
	printKw (rightbracket);
    }
}

/**
   It prints thearchitecture only if it is defined, otherwise it generates the 
   imports of the architecture (#include architecture_filenamestr.sdx)
*/
void
SyndexPrinter::printArchitecture (ntrees t)
{
  /* no operator : imcomplete description */
  if (nisemptyList (nchild (1, nchild (2, t))))
    {
      pK_messenger->message (infoArchiNotDefined);
      printCommentKw (mainArchitectureComment);
    }
  printKw (defi);
  printKw (tkArchi);
  printNthChild (t, 1);
  printKw (colon);
  printNthChild (t, 2);		// operators, medias, connections
  printKwLine ("main architecture IMA;");
}

/**
   It prints a binary Syndex tree (t).
*/
void
SyndexPrinter::printBinary (ntrees t)
{
  ntrees operand;

  /* opinport  opoutport  opconds  opinterface  
     opMainDeclaration 
     opSyndexPlus opSyndexMinus  opSyndexMult opSyndexDivi 
     opSyndexGroup opSyndexConstraint opRefArchiNode 
     opArchitecture opSyndexDefMedia 
   */
  switch ((int) noper (t))
    {
    case opconds:
      printBL ();
      printKw (cds);
      printNthChild (t, 1);
      if (isChild (t, 2))
	{
	  printKw (equal);
	  printNthChild (t, 2);
	}
      printKw (semicolon);
      printBL ();
      break;

    case opMainDeclaration:
      printCommentKw (mainDeclaration);
      printKw (mainDeclaration);
      printNthChild (t, 1);
      if (getChild (t, 2, &operand))
	printParameters (operand);
      printKw (semicolon);
      printBL ();
      break;

    case opSyndexPlus:
      printNthChild (t, 1);
      printKw (tkAdd);
      printNthChild (t, 2);
      break;

    case opSyndexMinus:
      printNthChild (t, 1);
      printKw (tkMinus);
      printNthChild (t, 2);
      break;

    case opSyndexDivi:
      printNthChild (t, 1);
      printKw (tkDivi);
      printNthChild (t, 2);
      break;

    case opSyndexMult:
      printNthChild (t, 1);
      printKw (tkDivi);
      printNthChild (t, 2);
      break;

    case opTypeAndDimension:
      printNthChild (t, 1);
      if (ISNOT_NILTREE (nchild (2, t)))
	{
	  printKw (leftbracket1);
	  printNthChild (t, 2);
	  printKw (rightbracket1);
	}
      break;

    case opSyndexGroup:
      printKw (tkGroup);
      printNthChild (t, 1);
      printKw (colon);
      printBL ();
      if (ISNOT_NILTREE (nchild (2, t)))
	{
	  printNthChild (t, 2);
	}
      break;

    case opSyndexConstraint:	// example: "absolute constraint : g_cpiom1 on IMA.CPIOM1"
      printKw (tkGroupConstraint);
      printKw (colon);
      printNthChild (t, 1);
      printSpc ();
      printKw (kwOn);
      printNthChild (t, 2);
      break;


    case opRefArchiNode:	// example: IMA.CPIOM1 "
      printNthChild (t, 1);
      printKw (dot);
      printNthChild (t, 2, pk_true);
      break;

    case opArchitecture:	// def architecture FOO:
      printArchitecture (t);
      break;

    case opSyndexDefMedia:	// def media FOO: sammp;
      printKw (defi);
      printKw (tkMedia);
      printNthChild (t, 1);
      printKw (colon);
      printBL ();
      printNthChild (t, 2);
      break;

    default:
      printNthChild (t, 1);
      printNthChild (t, 2);
    }
}

/**
   It prints a ternary Syndex tree (t).
*/
void
SyndexPrinter::printTernary (ntrees t)
{
  ntrees operand;

  /* opalgo  optypename  opbody  opSyndexref  opdpd  optop opDefArchitecture opSyndexDefOperator */
  switch ((int) noper (t))
    {
    case opdpd:		// dependence
      printNthChild (t, 1);
      printNthChild (t, 2);
      printKw (dpdsymbol);
      printNthChild (t, 3);
      break;
    case optypename:
      printNthChild (t, 1);	// algotype
      printNthChild (t, 2);	// name
      if (getChild (t, 3, &operand))
	printParameters (operand);
      printKw (colon);
      printBL ();
      break;

    case opSyndexHierarchy:
      printKw (aslash);
      printKw (aslash);
      printNthChild (t, 1, pk_true);
      if (getChild (t, 2, &operand))
	{
	  printKw (aslash);
	  unparse_subtree (operand, pk_true);
	}
      printKw (aslash);
      printNthChild (t, 3, pk_true);
      break;

    case opDefArchitecture:
      printBL ();
      printKw (tkOperators);
      printBL ();
      printNthChild (t, 1);	// operators
      printKw (tkMedias);
      printBL ();
      printNthChild (t, 2);	// medias
      printKw (tkConnections);
      printBL ();
      printNthChild (t, 3);	// connections
      break;

    case opSyndexDefOperator:
      printKw (defi);		// def operator FOO:
      printKw (tkOperator);
      printNthChild (t, 1);
      printKw (colon);
      printBL ();
      printNthChild (t, 2);	// list of gates (opSyndexGates)
      printKw (tkCodePhases);
      printNthChild (t, 3);	// list of code phases (opSyndexlAttrOperator)
      //      printBL();
      break;

    case opalgo:
      printKw (sepLine);
      printBL ();

    default:
      printNthChild (t, 1);
      printNthChild (t, 2);
      printNthChild (t, 3);
    }
}

/**
   @return "true" if it exists a header marker associated with a list operator (op), "false" otherwise.
   When it exists, the header marker is printed.
 */
boolean SyndexPrinter::printListBegin (noperators op)
{
  bPrintListSeparator = pk_true;
  switch ((int) op)
    {
    case oplrefs:
      printKw (rfs);
      return pk_true;
    case opldpds:
      printKw (dpd);
      return pk_true;
    default:
      return pk_false;
    }
}

/**
   It prints the list separator associated with an operator (op).
 */
void
SyndexPrinter::printListSeparator (noperators op)
{
  if (op == opSyndexElemGroup)
    {
      printKw (", attach_all");	// pour test LBesnard
      printKw (rightbracket1);
    }
  else if (bPrintListSeparator)
    if ((op != opSyndexProgram) && (op != oplalgo) && (op != oplbody) && (op != opSyndexlAttrOperator))
      printKw (semicolon);

  if (op != opSyndexlAttrOperator)
    printBL ();
}

/**
   Postlude for a list.
*/
void
SyndexPrinter::printListEnd (noperators op)
{
}

/**
   Printing a list of arguments (t), according to the Syndex syntax.
   
   oper(t) = oplargval, printed as <arg1, ..., argn>.
*/
void
SyndexPrinter::printArgsNames (ntrees t)
{
  boolean bPrinted;
  int i, n;

  n = ntreearity (t);
  if (n > 0)
    {
      bPrinted = pk_false;
      printKw (leftbracket);
      for (i = 1; i <= n; i++)
	if (ISNOT_NILTREE (nchild (i, t)))
	  {
	    if (bPrinted)
	      printKw (semicolon);
	    printNthChild (t, i);
	    bPrinted = pk_true;
	  }
      printKw (rightbracket);
    }
}

void
SyndexPrinter::printPreElemList (noperators op)
{
  if (op == opSyndexElemGroup)
    printKw (leftbracket1);
  else if (op == opSyndexGates)
    printKw (kwGate);
}

/**
   Printing a list tree (t), according to the Syndex syntax.
*/
void
SyndexPrinter::printList (ntrees t)
{
  int n, i;
  noperators op;
  boolean b;

  op = noper (t);
  if (op == oplargval)
    printArgsNames (t);
  else
    {
      b = printListBegin (op);
      n = ntreearity (t);
      if (n > 0)
	for (i = 1; i <= n; i++)
	  {
	    if (b)
	      printTabul ();
	    printPreElemList (op);
	    printNthChild (t, i);
	    printListSeparator (op);
	  }
      printListEnd (op);
    }
}

/**
   Main entry point for printing a syndex tree (trees).
   According to the arity of this tree, it calls:
   - printAtomic() a leaf
   - printUnary() an unary tree
   - printBinary() a binary tree
   - printTernary() a ternary tree
   - printList() a list tree

   The Syndex operators  are declared in opersyndex.h file.
*/
void
SyndexPrinter::unparse_subtree (ntrees trees)
{
  switch (getOperatorArity (noper (trees)))
    {
    case -1:
      printError ();
      break;

    case 0:
      printAtomic (trees);
      break;
    case 1:
      printUnary (trees);
      break;

    case 2:
      printBinary (trees);
      break;

    case 3:
      printTernary (trees);
      break;

    case 4:
      printList (trees);
      break;
    }
}

void
SyndexPrinter::unparse_subtree (ntrees trees, boolean bglue)
{
  boolean sglue;

  sglue = glue;
  glue = pk_true;
  unparse_subtree (trees);
  glue = sglue;
}


/**
   Initializing the printer.
*/
void
SyndexPrinter::init ()
{
  glue = pk_false;
  bimportDuration = pk_false;
  bPrintListSeparator = pk_true;
}

void
SyndexPrinter::finalize ()
{
  if (bimportDuration)
    cout << kwwarning << "some durations must be provided in the produced file(see lines with " << kwimport << ")" << endl;

}
