
_PUBLIC_
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
  _PUBLIC_
/** Constructor.*/
  SyndexService::SyndexService (ptDCG_flat g, boolean asMain)
{
  mainProg = asMain;
  vDCG = g;
  vDCGSyndex = NILDCG_SYNDEX;
  SET_NILTREE (ASTSyndex);
  gw = new DCG_Syndex ();
}

/** Destructor.*/
SyndexService::~SyndexService ()
{
}

/** 
   This is the entry point of the code generation for SYnDEx tool.
   This  method can be applied to a graph of a module (set of processes) or a process.
   It applies \ref IGraph2SynDExOneLevel for each graph.
   @return true if the operation succeeds.
   \see {http://www-rocq.inria.fr/syndex}
*/
boolean SyndexService::translation ()
{
  ptrepstr
    codeOp;

//  codeOp = string2repstr(cstSyndexExport);
//  pK_messenger->messagestr(mbeginOperName,codeOp);
  vDCGSyndex = new DCG_Syndex (vDCG);

  for_all_graphs_do_br (translationOneLevel (), breturn);
  if (breturn)
    {
      vDCGSyndex->removeSignalEqualities ();
      vDCGSyndex->NodesInterfaceCalculus ();
      vDCGSyndex->nonProductiveSignalElimination ();
      vDCGSyndex->buildClockProcessInterface ();
    }
  else
    {
      delete
	vDCGSyndex;
      vDCGSyndex = NILDCG_SYNDEX;
      setOnError ();
    }
  setTranslationApplied ();
//  pK_messenger->messagestr(mendOperName, codeOp);
  pK_messenger->setcputime ();
  return breturn;
}

/**
   It produces the Abstract Syndex tree of the current graph. It can be applied to a graph of a module.
*/
boolean SyndexService::produce_tree ()
{
  if (!isTranslationApplied ())
    return pk_false;
  if (isOnError ())
    return pk_false;

  for_all_graphs_do_br (produceOneLevel (), b);
  pK_messenger->setcputime ();
  return b;
}

/**
   It prints the AST built by the produce_tree() method. 
*/
void
SyndexService::print ()
{
  int i, n;
  ntrees p;
  ptident id;
  ptSyndexPrinter printer;

  if (ISNOT_NILTREE (ASTSyndex))
    {
      printer = new SyndexPrinter ();
      n = ntreearity (ASTSyndex);
      for (i = 1; i <= n; i++)
	{
	  p = nchild (i, ASTSyndex);
	  id = nattrntrees (p, SyndexService_ATTRIBUTENUMBER);
	  printer->fileUnparse (p, ident2repstr (id));
	}
      delete printer;
    }
  pK_messenger->setcputime ();
}
