/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   It initializes the current object for the attribute members
   - tCurrentInclude=nil, current include
   - tCurrentAlgo=nil, current algorithm
   - tlInclude=empty list, list of includes
   - tlAlgo = empty list, list of algorithms
   - tSyndex =  the main tree
   - stdInputName="i"
   - stdOutputName= "o"

   It is necessary to call finalize () after the translation and call getresult() to have the result abstract tree.
*/
  void
SyndexTreeInterface::initialize ()
{
  archiName = string2repstr ("IMA");
  vstrNoBroadcast = string2repstr ("no_broadcast");
  vstrDefaultMediaCommType = string2repstr ("sammp");
  vstrDefaultCodePhase = string2repstr ("loopseq");
  StrUndefRunOnGroup = string2repstr ("undefGroup");

  tArchi = LL_make_Architecture_tree (mkSyndexIdent (archiName),
				      LL_make_DefArchitecture_tree (LL_make_SyndexOperators_emptyList (), LL_make_SyndexMedias_emptyList (),
								    LL_make_SyndexArchiConnections_emptyList ()));
  SET_NILTREE (tCurrentInclude);
  SET_NILTREE (tCurrentAlgo);
  tlInclude = LL_make_linclude_emptyList ();
  tlAlgo = LL_make_lalgo_emptyList ();
  tlGroup = LL_make_lgroup_emptyList ();
  tlConstraint = LL_make_lSyndexConstraint_emptyList ();
  tlIncludeDurations = LL_make_linclude_emptyList ();
  tDefProcessors = LL_make_ldefoperator_emptyList ();
  tDefMedia = LL_make_ldefmedia_emptyList ();
  // defining the resulted tree : tSyndex
  tSyndex = LL_make_SyndexProgram_list (tlInclude);
  tSyndex = ntreeList_appendLast (tSyndex, tlAlgo);
  tSyndex = ntreeList_appendLast (tSyndex, tlGroup);
  tSyndex = ntreeList_appendLast (tSyndex, tlConstraint);
  tSyndex = ntreeList_appendLast (tSyndex, tDefProcessors);
  tSyndex = ntreeList_appendLast (tSyndex, tDefMedia);
  tSyndex = ntreeList_appendLast (tSyndex, tArchi);
  tSyndex = ntreeList_appendLast (tSyndex, tlIncludeDurations);
  stdInputName = string2repstr ("i");
  stdOutputName = string2repstr ("o");
}

/**
   It finalizes the resulting Syndex program.
   To have the Syndex abtract tree, call getresult() method.
*/
void
SyndexTreeInterface::finalize (ntrees undefGroupConstraint)
{
  partitionFinalize (undefGroupConstraint);
  architectureFinalize ();
}

/**
   @return the representation of a Syndex identifier named vstr.
*/
ntrees
SyndexTreeInterface::mkSyndexIdent (ptrepstr vstr)
{
  return LL_make_refname_atom (vstr);
}

/**
   @return the n-th input/output when binput/not binput.
*/
ptrepstr SyndexTreeInterface::mkOName (int n)
{
  ptrepstr
    vstr1;
  makeRepstr_char_int ('o', n, &vstr1);
  return vstr1;
}

/**
   @return the n-th input/output when binput/not binput.
*/
ptrepstr SyndexTreeInterface::mkIName (int n)
{
  ptrepstr
    vstr1;
  makeRepstr_char_int ('i', n, &vstr1);
  return vstr1;
}

/**
   It adds the condition, clockName = value, to the current algorithm.
   \see setCondition
*/
void
SyndexTreeInterface::addCondition (ptrepstr clockName, int value)
{
  setCondition (clockName, LL_make_value_atom (value));
  /*To do: If there are many conditions in an algorithm,eg: name =2, where such 2 from and How to do,later */
}

/**
   It adds the pk_true condition to the current algorithm.
   \see setCondition
*/
void
SyndexTreeInterface::addTrueCondition ()
{
  setCondition (string2repstr ("true"), NILTREE);
}

/**
   It adds the connection, <b> FromNode.outputname -> ToNode.n </b>, to the current alogorithm.
   \see addConnectionSimple
*/
void
SyndexTreeInterface::addConnectionFromToAt (ptrepstr FromNode, ptrepstr outputname, ptrepstr ToNode, int n)
{
  ptrepstr ToInputNode, FromOutputNode;

  ToInputNode = mkInputName (ToNode, n);
  FromOutputNode = mkOutputName (FromNode, outputname);
  addConnectionSimple (FromOutputNode, ToInputNode);
}

/**
   It adds the connection, <b> FromNode.outputname -> ToNode.inputname </b>, to the current alogorithm.
   \see addConnectionSimple
*/
void
SyndexTreeInterface::addConnectionFromToAt2 (ptrepstr FromNode, ptrepstr outputname, ptrepstr ToNode, ptrepstr inputname)
{
  ptrepstr ToInputNode, FromOutputNode;

  ptrepstr vstr;

  makeRepstr_repstr_string_repstr (ToNode, dot_string, inputname, &ToInputNode);
  FromOutputNode = mkOutputName (FromNode, outputname);
  addConnectionSimple (FromOutputNode, ToInputNode);
}


/**
   It adds the connection, <b> FromNode.outputname1 -> ToNode.outputname2 </b>, to the current alogorithm.
   \see addConnectionInMain
*/
void
SyndexTreeInterface::addConnectionFromToAtInMain (ptrepstr FromNode, ptrepstr outputname1, ptrepstr ToNode, ptrepstr outputname2)
{
  ptrepstr ToInputNode, FromOutputNode;

  FromOutputNode = mkOutputName (FromNode, outputname1);
  ToInputNode = mkOutputName (ToNode, outputname2);
  addConnectionInMain (FromOutputNode, ToInputNode);
}



/**
   It adds the current algorithm (tCurrentAlgo: attribute memebr) to the list of aglorithms (tlAlgo: attribute member).
   It is added at the tail if addPostClock, at the head otherwise.
*/
void
SyndexTreeInterface::addCurrentClock (boolean addPostClock)
{
  //for mastClock, or the clock which has external call, add to the last.
  if (addPostClock)
    ntreeList_appendLast (tlAlgo, tCurrentAlgo);
  else
    ntreeList_insertFirst (tCurrentAlgo, tlAlgo);
}

/**
   It adds, at the tail, an algorithm (mainAlgoTree) to the list of aglorithms (tlAlgo: attribute member).
*/
void
SyndexTreeInterface::addCurrentAlgoTree (ntrees mainAlgoTree)
{
  ntreeList_appendLast (tlAlgo, mainAlgoTree);
}

/**
   @return the predefined libbrary name, given by the concatenation of 
   SignalIdentPrefix_string name and type.
*/
ptrepstr SyndexTreeInterface::getPredefinedLibraryName (ptrepstr type)
{
  makeRepstr_string_repstr (SignalIdentPrefix_string, type, &type);
  return type;
}

/**
   It adds a library (type) to the list of includes (tlInclude). Unicity is guaranteed.
*/
void
SyndexTreeInterface::addInclude (ptrepstr type)
{
  boolean haveadded;
  ptrepstr havename;

  haveadded = pk_false;
  for (int i = 1; i <= ntreearity (tlInclude); i++)
    {
      havename = nstringof (nchild (1, nchild (i, tlInclude)));
      if (type == havename)
	{
	  haveadded = pk_true;
	  break;
	}
    }
  if (!haveadded)
    ntreeList_appendLast (tlInclude, mkincludename (type));
}


/**  
     It declares an input signal characterized by its base type (type) with some possible dimensions (vdim) and its name (name).
*/
void
SyndexTreeInterface::addInput (ptrepstr type, ntrees vdim, ptrepstr name)
{
  ntrees tree, list, vtype;

  vtype = LL_make_TypeAndDimension_tree (LL_make_intype_atom (type), vdim);
  tree = LL_make_inport_tree (vtype, LL_make_interfacename_atom (name));
  //add to the interface list
  if (getOperatorTree (tCurrentAlgo, oplimports, &list))
    ntreeList_appendLast (list, tree);
}


/**  
   It declares an output signal characterized by its base type (type) with some possible dimensions (vdim) and its name (name).
*/
void
SyndexTreeInterface::addOutput (ptrepstr type, ntrees vdim, ptrepstr name)
{
  ntrees tree, list, vtype;

  vtype = LL_make_TypeAndDimension_tree (LL_make_outype_atom (type), vdim);
  tree = LL_make_outport_tree (vtype, LL_make_interfacename_atom (name));
  //add to the interface list
  if (getOperatorTree (tCurrentAlgo, oploutports, &list))
    ntreeList_appendLast (list, tree);
}


/**
   @return a new reference of an algorithm (addAlgoName), this reference is declared in the current algorithm.
   \see makeInstSubGraphName, makeTreeReference.
*/
ptrepstr SyndexTreeInterface::declareAlgorithmInMain (ptrepstr addAlgoName)
{
  ntrees
    tr,
    listref;
  ptrepstr
    instOfaddAlgoName;

  instOfaddAlgoName = makeInstSubGraphName (addAlgoName);
  listref = getAlgorithmReferences (getCurrentClockAlgo ());
  //  tr = makeTreeReference ( makeReferenceType(pk_false, nilstr, addAlgoName), instOfaddAlgoName);
  tr = LL_make_Syndexref_tree (makeReferenceType (pk_false, nilstr, addAlgoName), LL_make_largval_emptyList (), mkSyndexIdent (instOfaddAlgoName));
  ntreeList_appendLast (listref, tr);
  return instOfaddAlgoName;
}


/**
   @return the name of a reference (t) defined when the operator of t is opSyndexref, opinport, opoutport.
   Otherwise it returns a predefined error name.
   \verbatim
   Example: for the following reference declaration (t)
      Signal_float/delay<{1.0};1> delay_2;
      it returns delay_2
   \endverbatim
*/
ptrepstr
SyndexTreeInterface::getReferenceName (ntrees t)
{
  if (noper (t) == opSyndexref)
    return nstringof (nchild (3, t));
  if ((noper (t) == opinport) || (noper (t) == opoutport))
    return nstringof (nchild (2, t));
  return string2repstr ("errortogainNameNumber");
}

/**
   @return the name of the instance associated with a modelName (modelName).
*/
ptrepstr SyndexTreeInterface::makeInstSubGraphName (ptrepstr modelName)
{
  return nameAndNumber (modelName, 1);
}

/**
   It declares a constant in the current algorithm, characterized by its type (treeType), 
   its name (nameCst) and its definition (vtdef).
*/
void
SyndexTreeInterface::mkCstDlc (ntrees treeType, ptrepstr nameCst, ntrees vtdef)
{
  ntrees theListResult, theListRef;

  if (getOperatorTree (tCurrentAlgo, oplparameters, &theListResult))
    ntreeList_appendLast (theListResult, LL_make_parameter_tree (makeIdentifier (nameCst)));
  if (getOperatorTree (tCurrentAlgo, oplrefs, &theListRef))
    if (ISNOT_NILTREE (vtdef))
      ntreeList_appendLast (theListRef, LL_make_Syndexref_tree (treeType, vtdef, mkSyndexIdent (nameCst)));
}

/**
   @return a \ref mkSkeletonAlgo() "skeleton abstract tree",  characterized by its name (name) and some parameters (lparam).
*/
ntrees
SyndexTreeInterface::mkSubGraphSkeletonAlgorithm (ptrepstr name, ntrees lparam)
{
  currentAlgoName = name;
  return mkSkeletonAlgo (altyalgorithm, name, lparam);
}

/** It sets the main algorithm declaration with params as parameters */
void
SyndexTreeInterface::mkMainAlgorithmDeclaration (ntrees params)
{
  ntrees mainnametree;
  mainnametree = LL_make_MainDeclaration_tree (mkSyndexIdent (mainAlgorithm), params);
  ntreeList_appendLast (tSyndex, mainnametree);
}

/** create and return an empty list of SynDEx expressions */
ntrees SyndexTreeInterface::make_Syndexlexpr_emptyList ()
{
  return (LL_make_Syndexlexpr_emptyList ());
}

/** create and return an empty list of SynDEx parameters */
ntrees SyndexTreeInterface::make_lparameters_emptyList ()
{
  return (LL_make_lparameters_emptyList ());
}

/**
   @return an argument value (sub-term of a declaration of a reference), composed of 
   a initial value (initvaltree), the number of delays (nbdelay), the size for a window (vwindow)
   and a size (vsize) associated with the declared reference.
   The built argument ignores the NILTREE parameters.

   \verbatim
   Example: <{1.0};1> is an argument value in the following reference declaration.
      Signal_float/delay<{1.0};1> delay_2;
   \endverbatim
*/
ntrees SyndexTreeInterface::mkArgsValues (ntrees initvaltree, ntrees nbdelay, ntrees vwindow, ntrees vsize)
{
  ntrees
    x,
    y;

  x = LL_make_largval_emptyList ();
  if (ISNOT_NILTREE (initvaltree))
    x = ntreeList_appendLast (x, LL_make_initvalue_tree (initvaltree));
  if (ISNOT_NILTREE (nbdelay))
    if (ISNOT_NILTREE (vwindow))
      {
	y = LL_make_Syndexlexpr_emptyList ();
	y = AddList (y, nbdelay);
	y = AddList (y, vwindow);
	x = ntreeList_appendLast (x, LL_make_pararef_tree (y));
      }
    else
      x = ntreeList_appendLast (x, LL_make_pararef_tree (nbdelay));
  else if (ISNOT_NILTREE (vwindow))
    x = ntreeList_appendLast (x, LL_make_pararef_tree (vwindow));

  if (ISNOT_NILTREE (vsize))
    x = ntreeList_appendLast (x, LL_make_pararef_tree (vsize));
  return x;
}

/**
   @return a new reference to a Syndex algorithm. The reference is the abstract tree of the 
   theNthOperation-th reference of a model characterized by its name (name), 
   its type (type), possibly predefined (predefined), and some specific arguments (argsname).
   
   This new reference is added to the current algorithhm (tCurrentBody, attribute member of the class).
   \see makeReferenceType, nameAndNumber.

    \verbatim
    Example: With predefined=true, type = float, name=delay, theNthOperation=2, argsname=<{1.0};1>
    This method declares in the following reference
      Signal_float/delay<{1.0};1> delay_2;
    in the current algorithm.
   \endverbatim
*/
ntrees SyndexTreeInterface::addReference (boolean predefined, ptrepstr type, ptrepstr name, ptrepstr nameOperation, ntrees argsname)
{
  ntrees
    vt,
    refLTree,
    typeSynTree;

  typeSynTree = makeReferenceType (predefined, type, name);
  vt = LL_make_Syndexref_tree (typeSynTree, argsname, mkSyndexIdent (nameOperation));
  if (!getOperatorTree (tCurrentBody, oplrefs, &refLTree))
    refLTree = LL_make_lrefs_emptyList ();
  ntreeList_appendLast (refLTree, vt);
  return vt;
}

ntrees SyndexTreeInterface::addReference (boolean predefined, ptrepstr type, ptrepstr name, int theNthOperation, ntrees argsname)
{
  ptrepstr
    nameAndNth;

  nameAndNth = nameAndNumber (name, theNthOperation);
  return addReference (predefined, type, name, nameAndNth, argsname);
}

/**
   It adds a reference without any specific argument.

   @copydoc addReference().
*/
ntrees SyndexTreeInterface::addReference (boolean predefinedType, ptrepstr type, ptrepstr name, int theNthOperation)
{
  return addReference (predefinedType, type, name, theNthOperation, NILTREE);
}


/**
   @return the predefined name for the intrinsic library when  bintrinsic,
   or the predefined name for the User library otherwise.
*/
ptrepstr SyndexTreeInterface::getLibraryName (boolean bintrinsic)
{
  if (bintrinsic)
    return string2repstr ("IntrinsicLibrary");
  return string2repstr ("UserLibrary");
}



/** 
    @return the tree that represents an identifier with vname as name.
 */
ntrees SyndexTreeInterface::makeIdentifier (ptrepstr vname)
{
  return LL_make_SyndexId_atom (vname);
}

/** 
    @return the tree that represents a constant with v as value.
 */
ntrees SyndexTreeInterface::makeConstant (ptrepstr v)
{
  return LL_make_SyndexConstant_atom (v);
}

/** 
    @return the tree that represents a constant with v as value.
 */
ntrees SyndexTreeInterface::makeConstantInt (int anInt)
{
  return makeConstant (integer2repstr (anInt));
}

/** @return the abstract Syndex atomic tree with op as operator. */
ntrees
SyndexTreeInterface::mkNullaryTree (noperators op)
{
  return nmakenullatom_syndex (op);
}

/**
   @return the abstract Syndex binary tree op(t1,t2).
*/
ntrees SyndexTreeInterface::make_SyndexUminus_tree (ntrees t)
{
  return LL_make_SyndexUminus_tree (t);
}

/**
   @return the abstract Syndex binary tree op(t1,t2).
*/
ntrees SyndexTreeInterface::make_SyndexElevec_tree (ntrees t)
{
  return LL_make_SyndexElevec_tree (t);
}

/**
   @return the abstract Syndex binary tree op(t1,t2).
*/
ntrees SyndexTreeInterface::makeBinary (noperators op, ntrees t1, ntrees t2)
{
  return nmakebinary (op, t1, t2);
}

/**
   @return the abstract Syndex binary tree op(t1,t2).
ntrees
SyndexTreeInterface::make_tvect_tree(ntrees t1, ntrees t2)
{
  return LL_make_tvect_tree(t1, t2);
}
*/

/**
   @return the abstract Syndex list tree p in which elem is added at the tail.
*/
ntrees SyndexTreeInterface::AddList (ntrees p, ntrees elem)
{
  return ntreeList_appendLast (p, elem);
}

/**
   @return the abstract Syndex tree that repretents an int value (val).
*/
ntrees SyndexTreeInterface::makeAtomInteger (long int val)
{
  return LL_make_value_atom (val);
}


/**
   It adds a new element in a group.
   @param grpname (in) the name of the group.
   @param elemGroup (in) the elem group to insert.
*/
void
SyndexTreeInterface::addElemGroup (ptrepstr grpname, ntrees elemGroup)
{
  ntreeList_appendLast (getGroupNamed (grpname), elemGroup);
}

/**
   It adds a new element in the undefined group (object without any binding).
   @param grpname (in) the name of the group.
   @param elemGroup (in) the elem group to insert.
*/
void
SyndexTreeInterface::addElemUndefinedGroup (ntrees elemGroup)
{
  addElemGroup (StrUndefRunOnGroup, elemGroup);
}

ntrees SyndexTreeInterface::mkHierarchyreference (ptrepstr operationName, boolean isInMainClock)
{
  ntrees
    subClock;

  // makeInstSubGraphName returns the name of the instance of the current clock.
  if (isInMainClock)
    SET_NILTREE (subClock);
  else
    subClock = mkSyndexIdent (makeInstSubGraphName (currentAlgoName));
  return LL_make_SyndexHierarchy_tree (mkSyndexIdent (mainAlgorithm), subClock, mkSyndexIdent (operationName));
}

void
SyndexTreeInterface::addArchiOperator (ptrepstr procModel, ptrepstr nameProc)
{
  ntrees t;
  t = nchild (1, nchild (2, tArchi));
  ntreeList_appendLast (t,
			LL_make_SyndexOperator_tree (mkSyndexIdent (procModel), mkSyndexIdent (nameProc), LL_make_SyndexlAttrOperator_emptyList ()));
  if (ProcessorTable.find (procModel) == ProcessorTable.end ())
    ProcessorTable[procModel] = LL_make_SyndexDefOperator_tree (mkSyndexIdent (procModel), LL_make_SyndexGates_emptyList (), mkDefaultCodePhases ());
  if (ProcessorInstanceTable.find (nameProc) == ProcessorInstanceTable.end ())
    ProcessorInstanceTable[nameProc] = procModel;
}


void
SyndexTreeInterface::addMediaOperator (ptrepstr mediaModel, ptrepstr nameMedia)
{
  ntrees t;
  t = nchild (2, nchild (2, tArchi));
  ntreeList_appendLast (t,
			LL_make_SyndexOperator_tree (mkSyndexIdent (mediaModel), mkSyndexIdent (nameMedia),
						     LL_make_SyndexlAttrOperator_list (mkDefaultMediaType ())));
  if (MediaTable.find (mediaModel) == MediaTable.end ())
    MediaTable[mediaModel] = LL_make_SyndexDefMedia_tree (mkSyndexIdent (mediaModel), mkDefaultMediaCommType ());	// def media FOO: sammp;
  if (MediaInstanceTable.find (nameMedia) == MediaInstanceTable.end ())
    MediaInstanceTable[nameMedia] = mediaModel;
  else;				//error double declaration?
}


void
SyndexTreeInterface::addArchiConnection (ptrepstr vproc, ptrepstr vmedium)
{
  ntrees vt, t;
  ptrepstr vstr;

  t = nchild (3, nchild (2, tArchi));	// connections
  makeRepstr_repstr_string_repstr (vproc, dot_string, mkGateName (getIndexMedia (vmedium)), &vstr);
  vt = LL_make_SyndexlAttrOperator_list (mkSyndexIdent (vstr));
  vt = ntreeList_appendLast (vt, mkSyndexIdent (vmedium));
  t = ntreeList_appendLast (t, vt);
}
