/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define inSuffix_string      ".i"	//inputstring in connections, this is used for the sensor&actuator&lib
#define outSuffix_string     ".o"	//connectOutputString//also in referenceForDefault()
#define dot_string      "."	//for connection,
#define defaultNameSep_char '_'	//for default name, eg default_4
#define gatePrefix_string "io"
#define SignalIdentPrefix_string "Signal_"
#define TRACE(s) s
#define OSDependentFilePath(DIR, FILE) makeRepstr_repstr_string_repstr (DIR, "/", FILE, &FILE);	// System dependent
/**
   @return the connection  <b> type opleftdpd -> oprightdpd </b>, where type is assumed to be 
   the ordinal of one element of the enum  \ref dpdType type.
*/
  ntrees SyndexTreeInterface::makeConnection (int type, ptrepstr leftstr, ptrepstr rightstr)
{
  return LL_make_dpd_tree (LL_make_dpdtype_atom (type), LL_make_leftdpd_atom (leftstr), LL_make_rightdpd_atom (rightstr));
}

/** 
   It adds the \em strong connection  <b> def.o -> vattr.iN </b>  to the current algorithm,
   where N is optional, equal to theNthInterface if theNthInterface !=-1.
*/
boolean SyndexTreeInterface::addConnection (ptrepstr vattr, ptrepstr def, int theNthInterface)
{
  ptrepstr
    leftstr,
    rightstr;

  if (makeRepstr_repstr_string (def, outSuffix_string, &leftstr))
    {
      ptrepstr
	dotinputN;
      if (theNthInterface != -1)
	{
	  if (!makeRepstr_repstr_string_int (vattr, inSuffix_string, theNthInterface, &rightstr))
	    return pk_false;
	}
      else if (!makeRepstr_repstr_string (vattr, inSuffix_string, &rightstr))
	return pk_false;
      addConnectionSimple (leftstr, rightstr);
      return pk_true;
    }
  return pk_false;
}

/**
   It adds the \em strong connection  <b> leftstr -> rightstr  </b>, to the current algorithm.
*/
void
SyndexTreeInterface::addConnectionSimple (ptrepstr leftstr, ptrepstr rightstr)
{
  addConnectionInBodyOf (leftstr, rightstr, tCurrentBody);
}

/**
   It adds the \em strong connection  <b> FromOutputNode -> ToInputNode  </b>, to the current algorithm.
*/
void
SyndexTreeInterface::addConnectionInMain (ptrepstr ToInputNode, ptrepstr FromOutputNode)
{
  addConnectionInBodyOf (ToInputNode, FromOutputNode, getAlgorithmBody (getCurrentClockAlgo ()));
}

/**
   It adds the \em strong connection  <b> leftstr -> rightstr  </b>, to a body of an algorithm (body).
*/
void
SyndexTreeInterface::addConnectionInBodyOf (ptrepstr leftstr, ptrepstr rightstr, ntrees body)
{
  ntrees vt, dpdLtree;
  dpdType dpd;

  dpd = strong;			//Here, i think all the type is strong. until now, i don't know where weak dpd comes.To do: later
  vt = makeConnection ((int) dpd, leftstr, rightstr);
  getOperatorTree (body, opldpds, &dpdLtree);
  ntreeList_appendLast (dpdLtree, vt);
}


/**
   @return the declarations of references of an algoritm (algo).
*/
ntrees SyndexTreeInterface::getAlgorithmReferences (ntrees algo)
{
  return nchild (2, nchild (1, nchild (3, algo)));
}


/**
   @return the body of an algoritm (algo).
*/
ntrees SyndexTreeInterface::getAlgorithmBody (ntrees algo)
{
  return nchild (3, nchild (1, nchild (3, algo)));
}


/**
   @return the header of a declaration of an object characterized by its class (atype), its name (name) and some parameters (lparam).
*/
ntrees SyndexTreeInterface::mkSkeletonTypeName (algoType atype, ptrepstr name, ntrees lparam)
{
  if (IS_NILTREE (lparam))
    lparam = LL_make_lparameters_emptyList ();
  return LL_make_typename_tree (LL_make_algotype_atom ((int) atype), LL_make_algoname_atom (name), lparam);
}

/**
   @return the representation of an empty interface (empty list of inputs/outputs).
*/
ntrees
SyndexTreeInterface::mkSkeletonInterface ()
{
  return LL_make_interface_tree (LL_make_limports_emptyList (), LL_make_loutports_emptyList ());
}

/**
   @return the representation of a body composed of an undefined condition part, an empty list of references, an empty list of dependences.
   The condition tree is composed of two NILTREE children.
   See Syndex Syntax for more details.
*/
ntrees
SyndexTreeInterface::mkSkeletonBodyList ()
{
  tCurrentBody = LL_make_body_tree (LL_make_conds_tree (NILTREE, NILTREE), LL_make_lrefs_emptyList (), LL_make_ldpds_emptyList ());
  tlBody = LL_make_lbody_list (tCurrentBody);
  return tlBody;
}

/**
   @return the complete type declaration of a reference RTYPE/name, where Rtype=type when it is not a predefined one (predefinedType parameter is false),
   or type prefixed by a predefined name otherwise.
   \verbatim
   For the predefined type float, and name=Arit_div, it produces the tree that represents the following text
   Signal_float/Arit_div
   \endverbatim
*/
ntrees SyndexTreeInterface::makeReferenceType (boolean predefinedType, ptrepstr type, ptrepstr name)
{
  if (predefinedType)
    type = getPredefinedLibraryName (type);
  if (!isNilString (type))
    {
      addInclude (type);
      OSDependentFilePath (type, name);	// PLG System dependent should use sysMake_strPath_repstr_repstr in sysfile_g
    }
  return LL_make_reftype_atom (name);
}

/**
   @return name if theNthOperation<0, name_N otherwise.
*/
ptrepstr SyndexTreeInterface::nameAndNumber (ptrepstr name, int theNthOperation)
{
  if (theNthOperation < 0)
    return name;
  if (makeRepstr_repstr_char_int (name, defaultNameSep_char, theNthOperation, &name))
    return name;
  return nilstr;
}


/**
   @return the string 
   - node.i<n> if n >= 0, 
   - node.i for n=-1,
   - node if n=-2.
*/
ptrepstr SyndexTreeInterface::mkInputName (ptrepstr node, int n)
{
  ptrepstr
    vstr;
  // -1 means don't add the number => return node.i
  // -2 means without suffix => return node
  if (n == -2)
    return node;
  if (n < 0)
    vstr = stdInputName;
  else
    vstr = mkIName (n);
  makeRepstr_repstr_string_repstr (node, dot_string, vstr, &vstr);
  return vstr;
}


/**
   @return the string node.outputname
*/
ptrepstr SyndexTreeInterface::mkOutputName (ptrepstr node, ptrepstr outputname)
{
  ptrepstr
    vstr;

  if (node == nilstr)
    return outputname;
  makeRepstr_repstr_string_repstr (node, dot_string, outputname, &vstr);
  return vstr;
}

/**
   It makes the skeleton of different algo (algorithm, constant, memory,sensor...
   It is memorized in tCurrentAlgo (attribute member) and so becomes the current algorithm.

  @param atype (in) the class of the definition.
  @param name (in) the name of the algorithm (constant...)
  @param lparam (in) the list of the parameters of  the algorithm.

  @return the skeleton tree.
  \see mkSkeletonTypeName
*/
ntrees SyndexTreeInterface::mkSkeletonAlgo (algoType atype, ptrepstr name, ntrees lparam)
{
  return tCurrentAlgo = LL_make_algo_tree (mkSkeletonTypeName (atype, name, lparam), mkSkeletonInterface (), mkSkeletonBodyList ());
}

/**
   It adds the condition :c=v , or condition: c to the current algorithm (tCurrentBody).
*/
void
SyndexTreeInterface::setCondition (ptrepstr c, ntrees v)
{
  ntrees conditionTree;

  if (getOperatorTree (tCurrentBody, opconds, &conditionTree))
    {
      nchangchild (conditionTree, LL_make_cdsname_atom (c), 1);
      if (ISNOT_NILTREE (v))
	nchangchild (conditionTree, v, 2);
    }
}


/**
   @return the named group grpname. The group is created
*/
ntrees SyndexTreeInterface::getGroupNamed (ptrepstr grpname)
{
  if (PartitionTable.find (grpname) == PartitionTable.end ())
    PartitionTable[grpname] = LL_make_SyndexElemGroup_emptyList ();
  return PartitionTable[grpname];
}

/**
   It defines
   - the groups induced by the RunOn/BindingIO partitonning of the application. The groups are managed in the tlGroup member of the class, 
   - the "absolute constraints" induced,managed in the tlConstraint member of the class.

   These trees are built from the entries of the PartitionTable.
*/
void
SyndexTreeInterface::partitionFinalize (ntrees undefGroupConstraint)
{
  ptrepstr vstr, vstr1, EnonceValue;
  integer i, nbEnonce;
  ntrees vlnodearchi;

  SyndexTableType1::iterator it;
  for (it = PartitionTable.begin (); it != PartitionTable.end (); it++)
    {
      vstr = ExtractString ((*it).first);	// "foo" --> foo.
      makeRepstr_string_repstr ("g_", vstr, &vstr1);
      tlGroup = ntreeList_appendLast (tlGroup, LL_make_SyndexGroup_tree (mkSyndexIdent (vstr1), (*it).second));
      // Adding the constraint 
      if ((vstr == StrUndefRunOnGroup) && (ISNOT_NILTREE (undefGroupConstraint)))
	{
	  vlnodearchi = LL_make_SyndexlAttrOperator_emptyList ();
	  nbEnonce = ntreearity (undefGroupConstraint);
	  for (i = 1; i <= nbEnonce; i++)
	    vlnodearchi =
	      ntreeList_appendLast (vlnodearchi,
				    LL_make_RefArchiNode_tree (mkSyndexIdent (archiName), mkSyndexIdent (getPragmaTerm (undefGroupConstraint, i))));
	}
      else
	vlnodearchi = LL_make_SyndexlAttrOperator_list (LL_make_RefArchiNode_tree (mkSyndexIdent (archiName), mkSyndexIdent (vstr)));
      tlConstraint = ntreeList_appendLast (tlConstraint, LL_make_SyndexConstraint_tree (mkSyndexIdent (vstr1), vlnodearchi));
    }
}

ntrees
SyndexTreeInterface::mkincludename (ptrepstr name)
{
  return LL_make_include_tree (LL_make_incltype_atom (name));
}

ntrees SyndexTreeInterface::mkDurationFileName (ptrepstr vname)
{
  ptrepstr
    vstr;
  //example: include durations_FOO.sdx, with vname=FOO the file suffix is added at printing time.
  makeRepstr_string_repstr ("durations_", vname, &vstr);
  return LL_make_importduration_tree (mkSyndexIdent (vstr));	// opimportduration : user must modify the generated file. 
}

ptrepstr
SyndexTreeInterface::mkGateName (int i)
{
  ptrepstr vstr;
  makeRepstr_string_int (gatePrefix_string, i, &vstr);
  return vstr;
}


void
SyndexTreeInterface::addGatesDef (ntrees lgates)
{
  SyndexTableType1::iterator it;
  int i;
  ntrees vt;

  i = 1;
  for (it = MediaTable.begin (); it != MediaTable.end (); it++)
    {
      vt = LL_make_SyndexlAttrOperator_list (mkSyndexIdent ((*it).first));
      vt = ntreeList_appendLast (vt, mkSyndexIdent (mkGateName (i)));
      lgates = ntreeList_appendLast (lgates, vt);
      i++;
    }
}

void
SyndexTreeInterface::architectureFinalize ()
{
  SyndexTableType1::iterator it;
  ntrees vt;

  for (it = ProcessorTable.begin (); it != ProcessorTable.end (); it++)
    {
      vt = (*it).second;
      addGatesDef (nchild (2, vt));
      tDefProcessors = ntreeList_appendLast (tDefProcessors, vt);
      tlIncludeDurations = ntreeList_appendLast (tlIncludeDurations, mkDurationFileName ((*it).first));
    }
  for (it = MediaTable.begin (); it != MediaTable.end (); it++)
    {
      tDefMedia = ntreeList_appendLast (tDefMedia, (*it).second);
      tlIncludeDurations = ntreeList_appendLast (tlIncludeDurations, mkDurationFileName ((*it).first));
    }
}

ntrees SyndexTreeInterface::mkDefaultMediaType ()
{
  return mkSyndexIdent (vstrNoBroadcast);
}

ntrees SyndexTreeInterface::mkDefaultMediaCommType ()
{
  return LL_make_SyndexlAttrOperator_list (mkSyndexIdent (vstrDefaultMediaCommType));
}

ntrees SyndexTreeInterface::mkDefaultCodePhases ()
{
  return LL_make_SyndexlAttrOperator_list (mkSyndexIdent (vstrDefaultCodePhase));
}

int
SyndexTreeInterface::getIndexTable (ptrepstr vname, SyndexTableType2 & T)
{
  SyndexTableType2::iterator it;
  int i;

  i = 1;
  for (it = T.begin (); it != T.end (); it++)
    {
      if ((*it).first == vname)
	return i;
      i++;
    }
  return -1;			// the name does not exist.
}

int
SyndexTreeInterface::getIndexProc (ptrepstr vproc)
{
  return getIndexTable (vproc, ProcessorInstanceTable);
}


int
SyndexTreeInterface::getIndexMedia (ptrepstr vmedium)
{
  return getIndexTable (vmedium, MediaInstanceTable);
}
