/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return the Syndex program root.
*/
  ntrees SyndexTreeInterface::getResult ()
{
  return tSyndex;
}

/**
  It sets the current algorithm to t.
*/
void
SyndexTreeInterface::setCurrentAlgo (ntrees t)
{
  tCurrentAlgo = t;
}

/**
   @return the current algorithm.
*/
ntrees SyndexTreeInterface::getCurrentAlgo ()
{
  return tCurrentAlgo;
}

/**
   It sets the main algorithm to name. 
   It is the algorithm associated with the tick clock.
*/
void
SyndexTreeInterface::setMainAlgoName (ptrepstr name)
{
  mainAlgorithm = name;
}

/**
   @return the main algorithm (associated with the tick clock).
*/
ptrepstr SyndexTreeInterface::getMainAlgoName ()
{
  return mainAlgorithm;
}

/**
   @return the current algorithm associated with the current clock.
*/
ntrees SyndexTreeInterface::getCurrentClockAlgo ()
{
  return tMainAlgorithm;
}

/**
   It sets the current algorithm (associated with the current clock) to t.
*/
void
SyndexTreeInterface::setCurrentClockAlgo (ntrees t)
{
  tMainAlgorithm = t;
}

/**
   Getting the child (r) of a tree (t) with operator equals to op.
   If it is a list, r = first tree of the list.
   if the operator of t is equal to op, r = t.
   @return true when succeed, otherwise false.
*/
boolean SyndexTreeInterface::getOperatorTree (ntrees t, noperators op, ntrees * r)
{
  ntrees
    v;
  noperators
    optree;

  if (IS_NILTREE (t))
    return pk_false;

  optree = noper (t);
  if (optree == op)
    {
      *r = t;
      return pk_true;
    }

  for (int i = 1; i <= ((getOperatorArity (optree) < 4) ? getOperatorArity (optree) : 1); i++)
    {
      //oplist is treated as 1.
      v = nchild (i, t);
      if (ISNOT_NILTREE (v) && !isleaf (v))
	{
	  if (getOperatorTree (v, op, r))
	    return pk_true;
	}
      //getOperatorTree(v,op,&r),To do:again
    }
  return pk_false;
}

/**
   @return true if a tree(t) is a leaf (atomic tree), false otherwise.
*/
boolean SyndexTreeInterface::isleaf (ntrees t)
{
  if (getOperatorArity (noper (t)) == 0)
    return pk_true;
  return pk_false;
}


/**
   @return the arity of a Syndex operator (voperator), 4 for a list, -1 for an out of bounds.
*/
int
SyndexTreeInterface::getOperatorArity (noperators voperator)
{
  if ((voperator > BaseSyndexOperator) && (voperator <= SYN_NB_OPERATORS))
    {
      if (voperator < opSyndexBinary)
	{
	  if (voperator < opSyndexUnary)
	    return 0;
	  return 1;
	}
      if (voperator >= opSyndexTernary)
	return 3;
      if (voperator < opSyndexList)
	return 2;
      return 4;			// liste
    }
  return -1;			// out of bounds --> -1
}
