/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   Constructor: the functionality is applied for a graph (g).
*/
  DCG2KernelTranslator::DCG2KernelTranslator (ptDCG_basic g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  hinit = CKM->getInitClock ();
  currentNode = NILINODE;
  bachieved = pk_true;
}

/**
   Desctructor.
 */
DCG2KernelTranslator::~DCG2KernelTranslator ()
{
}

/**
   It applies the reduction to Basic Signal, \ref node2Kernel(), for all the nodes of the current graph (AGRAPH,n selected in the contructor):  
   - The cell is not rewritten in Basic Signal. Only structural rules are applied
   - The partial definition are not considered. It is assumed that the method groupPartialDefinitions() has been applied before this transformation
   - The structure/bundle are not considered. It is assumed that the method splitDataStructure() has been applied before this transformation.
   - Restrictions: the delay >1, the array of processus are not yet implemented. 

   @return "true" if the graph is in the kernel representation, "false" otherwise.
   
   For example,
   \code
   x:= (a+b) when h1 default (e+t) when h2
   is rewritten into
   (| x1:= a+b 
    | x2:= (e+t) 
    | x:= (x1 when h1) default (x2  when h2)
    |)
   \endcode
   Note : The normal form of the expression, default(when(synchronous expr, clock)), is kept.
*/
boolean
DCG2KernelTranslator::apply ()
{
  ptlINode l;
  ptINode n;

  bachieved = pk_true;
  if (NM->getAllNodes (&l))
    while (ptlINode_get_elem (&l, &n))
      node2Kernel (n);
  return bachieved;
}
