/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   Constructor. The object is linked with a graph (g).
   The provided methods work on this graph.
*/
  arrayClockUtl::arrayClockUtl (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  if (!NM->getIterates (&lArrayNodes))
    lArrayNodes = NILLINODE;
}

/**
   Destructor.
*/
arrayClockUtl::~arrayClockUtl ()
{
}

/**
   It fixes the clock of the arrays of processes for the selected graph (constructor).
   The following operations are sequentialy applied:
   - some initialisations (initialize()),
   - the operation is applied using the extractions built on indexes as seeds (setInternalClocksArrays()),
   - some verifications (arrayVerification())
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean arrayClockUtl::fixArrayClocks ()
{
  boolean
    b;

  if (lArrayNodes == NILLINODE)
    return pk_true;
  aMark = getMark ();
  initialize ();
  setInternalClocksArrays ();
  b = arrayVerification ();
  CKM->resetMarkAll (aMark);
  if (!b)
    pK_messenger->message (infosigmotif);
  finalyze ();
  return b;
}
