/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/** Linking a clock with an array-node */
  void
arrayClockUtl::linkClockToArray (ptinsth vclk, ptINode n)
{
  NM->arrayBodyInsertClock (n, vclk);
  CKM->setInComposedNode (vclk, n);
}

/**
  @return "true" if the clock arguments (vharg1, vharg2) are in an array structure, "false" otherwise.
  @param vharg1 (in) a clock.
  @param vharg2 (in) a clock.
  @param x (out) the most inner array node that contains one of the operand clock (it is set only when it returns "true").
*/
boolean arrayClockUtl::maxlevelarray (ptinsth vharg1, ptinsth vharg2, ptINode * x)
{
  ptINode
    vblockarray1,
    vblockarray2;

  CKM->getInComposedNode (vharg1, &vblockarray1);
  CKM->getInComposedNode (vharg2, &vblockarray2);
  if (NM->arrayGetLevelBlock (vblockarray1) > NM->arrayGetLevelBlock (vblockarray2))
    *x = vblockarray1;
  else
    *x = vblockarray2;
  return ((*x) != NILINODE);
}

/**
   Defining the array block for the references to an operand clock (vh).
*/
void
arrayClockUtl::linkusestoarray (ptinsth vh)
{
  ptinsth vharg1, vharg2, vhuse;
  ptlinsth luse;
  ptINode arrayNode, cn;

  if (!CKM->satisfiesMarked (vh, aMark))
    {
      CKM->addMark (vh, aMark);
      luse = CKM->getClockReferencesTo (vh);
      while (ptlinsth_get_next (&luse, &vhuse))
	{
	  switch ((int) (CKM->getOperator (vhuse)))
	    {
	    case super:
	      CKM->getSupArgs (vhuse, &vharg1, &vharg2);
	      if (maxlevelarray (vharg1, vharg2, &arrayNode))
		linkClockToArray (vhuse, arrayNode);
	      break;
	    case infer:
	      if (!CKM->getMinusArgs (vhuse, &vharg1, &vharg2))
		CKM->getInfArgs (vhuse, &vharg1, &vharg2);
	      if (maxlevelarray (vharg1, vharg2, &arrayNode))
		linkClockToArray (vhuse, arrayNode);
	      break;
	    case hcomp:
	      CKM->getInComposedNode (vh, &cn);
	      linkClockToArray (vhuse, cn);
	      break;
	    case sub:
	    case nosub:
	    case hnot:
	    case vari:
	    case predef:
	      break;
	    default:
	      pK_messenger->Caseerror (currentLine, currentFile);
	    }
	  linkusestoarray (vhuse);
	}
    }
}

/**
   Linking clock with an array structure.
   If the operand clock (vh) is not an undersampling-clock, this method is without effect.

   An undersampling-clock is assigned to the most inner block where the signals/indexes it references are defined.
   Then this information is propagated to its references (linkusestoarray()). It can fix the array block of the other clocks (upper, lower bounds,...)
*/
void
arrayClockUtl::linkDownSamplingClockToArray (ptinsth vh)
{
  ptlinstances lindexes, ls, l;
  ptINode cnode, indexNode, nodeArrayIndex, cn, n;
  ptexpcal indice, c, velem, vexp;
  ptinsth vhnosub;

  vh = CKM->getRepresentativeFormula (vh);
  if (CKM->getOperator (vh) == sub)
    {
      CKM->removeMark (vh, aMark);	//added May 2011 the clock may have been set during a linkusestoarray (clock of the condition).
      c = CKM->getConditionSub (vh);
      lindexes = EM->getIndexesReferencedIn (c);
      ls = EM->getSignalsReferencedIn (c);

      l = ls;
      while (ptlinstances_get_next (&l, &velem))
	{
	  if (EM->getAttribute (velem, AccessToArrayElement, (Pointer *) & n))
	    {
	      if (NM->get_rightOperand (n, &vexp))
		lindexes = ptlinstances_sortedfusion (EM->getIndexesReferencedIn (vexp), lindexes);
	      else;
	      // added May 2011
	      if (NM->get_inComposedNode (n, &cnode))
		{
		  indexNode = NM->get_indexNode (cnode);
		  indice = NM->get_index (indexNode);
		  lindexes = ptlinstances_insert (lindexes, indice);
		}		// end-added May 2011 
	    }
	}
      if (NM->arrayGetMaxLevelIndex (ls, lindexes, &nodeArrayIndex))
	{
	  NM->get_inComposedNode (nodeArrayIndex, &cn);
	  linkClockToArray (vh, cn);
	  linkusestoarray (vh);
	  CKM->getComplExtraction (vh, &vhnosub);
	  linkClockToArray (vhnosub, cn);
	  linkusestoarray (vhnosub);
	}
      ptlinstances_free (lindexes);
    }
}

/**
   Defining internal clocks of arrays: it calls linkDownSamplingClockToArray() method for
   each clock of the system.
*/
void
arrayClockUtl::setInternalClocksArrays ()
{
  ptlinsth lcour;
  ptinsth vh;

  lcour = CKM->getlhroots ();
  while (ptlinsth_get_next (&lcour, &vh))
    linkDownSamplingClockToArray (vh);

  lcour = CKM->getlhplaced ();
  while (ptlinsth_get_next (&lcour, &vh))
    linkDownSamplingClockToArray (vh);
}

/** Defining internal clocks of arrays.
    For a basic graph, errors are not detected (an array may have a clock that depends on internal indexes).
*/
boolean arrayClockUtl::arrayVerification ()
{
  ptINode
    n,
    v1,
    v2;
  ptlINode
    ln;
  boolean
    bBasicGraph,
    b;
  ptinsth
    h1,
    h;
  ptexpcal
    vindex,
    ins,
    outs,
    x;
  ptrepstr
    indexStr;

  bBasicGraph = (AGRAPH->getFormatLevel () == DCG_BASIC);
  //if (AGRAPH->getFormatLevel() == DCG_BASIC) return pk_true;
  b = pk_true;
  ln = lArrayNodes;
  while (ptlINode_get_elem (&ln, &n))
    {
      h = NM->get_nodeClock (n);
      CKM->getInComposedNode (h, &v1);
      NM->get_inComposedNode (n, &v2);
      if (NM->arrayGetLevelBlock (v1) > NM->arrayGetLevelBlock (v2))
	{
	  NM->arrayBodyRemoveClock (v1, h);
	  if (CKM->getComplExtraction (h, &h1))
	    NM->arrayBodyRemoveClock (v1, h1);
	  linkClockToArray (h, v2);
	  if (!bBasicGraph)
	    {
	      vindex = NM->get_index (NM->get_indexNode (n));
	      indexStr = ident2repstr (EM->getName (vindex));
	      pK_messenger->errorStr (errArrayClock, indexStr);
	      b = pk_false;
	    }
	}
      else if (!bBasicGraph)
	{			// The clock of the outputs must not be internal to the array.
	  NM->get_array_io (n, &ins, &outs);
	  vindex = NM->get_index (NM->get_indexNode (n));
	  indexStr = ident2repstr (EM->getName (vindex));
	  while (ptexpcal_get_Nary (&outs, &x))
	    {
	      h = EM->getClock (x);
	      CKM->getInComposedNode (h, &v1);
	      if (NM->arrayGetLevelBlock (v1) > NM->arrayGetLevelBlock (v2))
		{
		  pK_messenger->errorIdentStr (errArrayClockSig, EM->getName (x), indexStr);
		  b = pk_false;
		}
	    }
	}
    }
  return b;
}


/**
   It removes from an assumed array node (aSubGraph) the clocks associated with it. 
   Moreover, the dependence from the index of the array to these clocks is removed.
*/
void
arrayClockUtl::removingClocksOfArray (ptINode aSubGraph)
{
  ptINode internalArrayNode, IndexNode;
  ptlINode ln;
  ptlinsth lh;
  ptinsth h;

  IndexNode = NM->get_indexNode (aSubGraph);
  NM->get_array_inputNode (aSubGraph, &internalArrayNode);
  NM->get_array_bodyDefinitions (aSubGraph, &ln, &lh);
  while (ptlinsth_get_next (&lh, &h))
    {
      CKM->setInComposedNode (h, NILINODE);
      NM->removeDependence (IndexNode, h);
      NM->removeDependence (internalArrayNode, h);
    }
  NM->arrayBodySetDef (aSubGraph, ln, NILLINSTH);
  NM->set_interface (aSubGraph, NILINTERFACE);	// March 2014 18 2014
}


/** 
    It initializes the operation.
    It applies removingClocksOfArray() method for each array node of the graph.
*/
void
arrayClockUtl::initialize ()
{
  ptINode n;
  ptlINode l, ln;

  ln = lArrayNodes;
  if (ln != NILLINODE)
    {
      // Cleaning the current clocks ()
      while (ptlINode_get_elem (&ln, &n))
	removingClocksOfArray (n);
      // The nodes of an array are linked with this array.
      ln = lArrayNodes;
      while (ptlINode_get_elem (&ln, &n))
	NM->arrayLinkNodes (n, pk_false);
    }
}

void
arrayClockUtl::resetArrayNodesLink ()
{
  ptlINode ln;
  ptINode arrayNode;

  ln = lArrayNodes;
  while (ptlINode_get_elem (&ln, &arrayNode))
    NM->arrayUnlinkNodes (arrayNode, pk_false);
  // code mort Sep 2014 LBesnard reportÃ© dans arrayUnlinkNodes
  /*  ptlINode l, ln;
     ptlinsth lh;
     ptINode arrayNode, aNode, n;  
     ptinsth vh;

     while (ptlINode_get_elem(&ln, &arrayNode))
     {  
     if (NM->get_array_bodyDefinitions(arrayNode, &ln, &lh))
     {
     while (ptlINode_get_elem(&ln, &aNode)) NM->set_inComposedNode(aNode, NILINODE);
     while (ptlinsth_get_next(&lh, &vh)) CKM->setInComposedNode(vh, NILINODE);

     if (NM->get_inComposedNode(arrayNode, &n))
     if (NM->get_array_initDefinitions(arrayNode, &ln, &lh))
     while (ptlINode_get_elem(&ln, &aNode)) NM->set_inComposedNode(aNode, NILINODE);
     }
     }
   */
}

/** 
    It finalyzes the operation by
    - cleaning (method resetArrayNodesLink()) the link from an internal node to the array node.
*/
void
arrayClockUtl::finalyze ()
{
  resetArrayNodesLink ();

}
