/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/****************************************************************************/
extern clock_type
ck_create (void)
{
  clock_type result;
  result = MALLOC (clock_type, sizeof (clock_struct));
  ASSERT_MALLOC_OK (result);
  result->children = list_create ();
  result->bdd = NILBDD;
  result->horl = NILINSTH;
  result->parent = NILCLOCK;
  result->head = NILCLOCK;
  result->id = ERR_ID;
  result->support = NIL_bddSupportSet_type;
  result->vars = NIL_bddSupportSet_type;
  result->form = NILBDDFORMULA;
  return result;
}

extern void
ck_clone (clock_type ck1, clock_type ck2)
{
  ck1->id = ck2->id;
  ck1->children = list_create ();
  ck1->bdd = ck2->bdd;
  // ck1->horl = NILINSTH;
  ck1->parent = NILCLOCK;
  ck1->head = NILCLOCK;
  ck1->support = ck2->support;
  ck1->vars = ck2->vars;
  ck1->form = ck2->form;
}

/****************************************************************************/
extern void
ck_destroy (clock_type ck)
{
  list_destroy (ck->children);
  if (ck->bdd != NILBDD)
    pK_bdd_free (ck->bdd);
  FREE (ck);
}

/****************************************************************************/
extern void
ck_set_id (clock_type ck, bdd_id_boolean_variable_type id)
{
  ck->id = id;
}

/****************************************************************************/
extern void
ck_set_children (clock_type ck, list_type list)
{
  ck->children = list;
}

/****************************************************************************/
extern void
ck_set_bdd (clock_type ck, bdd_type bdd)
{
  ck->bdd = bdd;
  ck->support = supp_create (bdd);
  if (ACCEPT_ASSIGN (ck))
    set_form__ (ck);
}


/****************************************************************************/
extern void
ck_set_horl (clock_type ck, ptinsth horl)
{
  ck->horl = horl;
  if (ACCEPT_ASSIGN (ck))
    set_form__ (ck);
}

/****************************************************************************/
extern void
ck_set_parent (clock_type ck, clock_type parent)
{
  ck->parent = parent;
  if (ACCEPT_ASSIGN (ck))
    set_form__ (ck);
}

/****************************************************************************/
extern void
ck_set_head (clock_type ck, clock_type head)
{
  ck->head = head;
}

/****************************************************************************/
extern clock_type
ck_get_head (clock_type ck)
{
  return ck->head;
}

/****************************************************************************/
extern clock_type
ck_get_parent (clock_type ck)
{
  return ck->parent;
}

/****************************************************************************/
extern ptinsth
ck_get_horl (clock_type ck)
{
  return ck->horl;
}

/****************************************************************************/
extern bdd_type
ck_get_bdd (clock_type ck)
{
  return ck->bdd;
}

/****************************************************************************/
extern list_type
ck_get_children (clock_type ck)
{
  return ck->children;
}

/****************************************************************************/
extern bdd_id_boolean_variable_type
ck_get_id (clock_type ck)
{
  return ck->id;
}

/****************************************************************************/
extern form_type
ck_get_form (clock_type ck)
{
  return ck->form;
}




/****************************************************************************/
extern void
ck_update_horl (clock_type ck, ptinsth h)
{
  ck->horl = h;
}
