/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/****************************************************************************/
/*** Ce fichier contient un ensemble de fonctions d'interface             ***/
/*** pour interagir avec une horloge. Il presente une horloge             ***/
/*** comme etant un n-uplet (id,children_list,bdd,root)                   ***/
/*** Note: root est la racine qui a servi de reference a la               ***/
/*** construction de bdd. Lorsqu'on se rend compte qu'il n'est            ***/
/*** plus la racine effective, bdd doit etre recalcule avant              ***/
/***  d'etre utilise                                                      ***/
/****************************************************************************/
/*
  modifs Loic Besnard : uint remplace par bdd_id_boolean_variable_type, ou bdd_rank_in_order_type selon le contexte.
*/

#define ERR_ID 0xFFFFFFFF

typedef struct __clock
{
  bdd_id_boolean_variable_type id;
  list_type children;
  bdd_type bdd;
  ptinsth horl;
  clock_type parent;
  clock_type head;
  supp_type support;
  supp_type vars;
  form_type form;
} clock_struct;

#define ACCEPT_ASSIGN(ck) (ck->parent != NILCLOCK) && (ck->bdd != NILBDD) && (ck->form == NILBDDFORMULA) && (ck->horl != NILINSTH)

static boolean is_subst_able (clock_type root, supp_type depend_vars, bdd_id_boolean_variable_type id);
static void recompute_vars (clock_type ck);
static clock_type retrieve_by_id (clock_type root, bdd_id_boolean_variable_type id);
static form_type ck_get_form__ (clock_type ck);
static clock_type retrieve_by_insth (clock_type ck, ptinsth h);
static void get_partition (bdd_id_boolean_variable_type id, clock_type ck, clock_type * positive_ck, clock_type * negative_ck);
static void reduce (form_type * form);
static form_type make_lit_form (clock_type ck);

/****************************************************************************/
static clock_type
get_root (clock_type ck)
{
  if (ck->parent == NILCLOCK)
    return ck;
  return get_root (ck->parent);
}

/****************************************************************************/
static void
set_form__ (clock_type ck)
{
  int op;
  clock_type ck1, ck2, root;
  ptinsth h1, h2;
  form_type arg1, arg2;
  bdd_type bddone;

  root = ck;
  while (root->parent != NILCLOCK)
    root = root->parent;
  ck->form = form_create ();
  op = mygetop (ck->horl);
  switch (op)
    {
    case OP_AND:
    case OP_OR:
    case OP_DIFF:
      {
	/* ck is a clock defined by some formula */
	mygetargs (ck->horl, &h1, &h2);
	ck1 = retrieve_by_insth (root, h1);
	ck2 = retrieve_by_insth (root, h2);
	/* quick and dirty trick to convert insth-ops into form-ops */
	/* this conversion assumes that  AND,OR, and DIFF have the  */
	/* same order in both worlds                                */
	/* it's classical but dirty                                 */
	form_set_op (ck->form, (op_type) (op - OP_AND + (int) FORM_AND));
	arg1 = form_create ();
	arg2 = form_create ();
	form_set_op (arg1, FORM_LIT);
	form_set_op (arg2, FORM_LIT);
	form_set_literal (arg1, (literal_type) ck1);
	form_set_literal (arg2, (literal_type) ck2);
	form_set_args (ck->form, arg1, arg2);
	/* begin by creating an empty support set */
	bddone = pK_bdd_one ();
	ck->vars = supp_create (bddone);
	pK_bdd_free (bddone);
	/* then add the vars of its operands */
	/* and you get its vars              */
	if (ck1->vars != NIL_bddSupportSet_type)
	  supp_union (ck->vars, ck1->vars);
	if (ck2->vars != NIL_bddSupportSet_type)
	  supp_union (ck->vars, ck2->vars);
	ck->form = ck_get_form__ (ck);
	break;
      }
    default:
      {
	/* ck is a \hsub{C} or a \hnosub{C} */
	form_set_op (ck->form, FORM_LIT);
	form_set_literal (ck->form, (literal_type) ck);
	/* just say that  its vars is    */
	/* \hsub{C} which is the support */
	ck->vars = ck->support;
	break;
      }
    }
}

static boolean substitute (form_type * form, boolean inck, form_type old, form_type newf);
/****************************************************************************/
static form_type
ck_get_form__ (clock_type ck)
{
  bdd_id_boolean_variable_type id;
  supp_type diff_vars, depend_vars;
  boolean dummy_bool;
  clock_type positive_ck, negative_ck, root;
  form_type zero_form;

  zero_form = form_zero ();

  switch (form_get_op (ck->form))
    {
    case FORM_AND:
    case FORM_OR:
    case FORM_DIFF:
      {
	root = get_root (ck);
	diff_vars = supp_clone (ck->vars);
	depend_vars = supp_clone (ck->support);
	supp_diff (diff_vars, depend_vars);
	supp_set_to_max (diff_vars);

	while (!supp_is_at_end (diff_vars))
	  {
	    id = supp_get_current (diff_vars);
	    if (is_subst_able (root, depend_vars, id))
	      {
		get_partition (id, ck, &positive_ck, &negative_ck);
		/* dummy assignment */
		dummy_bool = substitute (&ck->form, pk_false, positive_ck->form, zero_form);
		reduce (&(ck->form));
		dummy_bool = substitute (&ck->form, pk_false, negative_ck->form, make_lit_form (negative_ck->parent));
		recompute_vars (ck);
		supp_destroy (diff_vars);
		diff_vars = supp_clone (ck->vars);
		supp_diff (diff_vars, depend_vars);
		supp_set_to_max (diff_vars);
	      }
	    else
	      {
		supp_insert (depend_vars, id);
		supp_remove_current (diff_vars);
		/* after this element has been removed, the current */
		/* element is the greatest                         */
	      }
	  }
	break;
      }
    default:
      /* no need to perform a substution on a \hsub{C} or a \hnosub{C} */
      break;
    }
  return ck->form;
}

/****************************************************************************/
/*** la facon dont la fonction suivante est e'crite suppose que           ***/
/*** {\bf les deux sous-e'ch existent}                                    ***/
static void
get_partition (bdd_id_boolean_variable_type id, clock_type ck, clock_type * positive_ck, clock_type * negative_ck)
{
  list_type siblings;
  ptinsth vnosub;
  boolean stop;
  ck = get_root (ck);
  *positive_ck = retrieve_by_id (ck, id);
  vnosub = mygethnosub ((*positive_ck)->horl);
  siblings = (*positive_ck)->parent->children;
  list_set_to_first (siblings);
  stop = pk_false;
  while (!stop)
    {
      *negative_ck = (clock_type) list_get_current (siblings);
      stop = (vnosub == (*negative_ck)->horl);
      list_forward (siblings);
    }
}

/****************************************************************************/
static boolean
subst__ (form_type * form, boolean inck, form_type old, form_type newf)
{
  form_type f1, f2, aux;
  boolean result;

  if (*form == form_one ())
    return pk_false;
  if (*form == form_zero ())
    return pk_false;

  if (form_equal (old, *form))
    {
      *form = newf;
      return pk_true;
    }
  switch (form_get_op (*form))
    {
    case FORM_LIT:
      {
	aux = ((clock_type) form_get_literal (*form))->form;
	if (aux == *form)
	  return pk_false;
	if (subst__ (&aux, pk_true, old, newf))
	  {
	    *form = aux;
	    return pk_true;
	  }
	return pk_false;
      }
    case FORM_AND:
    case FORM_OR:
    case FORM_DIFF:
      form_get_args (*form, &f1, &f2);
      /* result = subs__(f1) || subst__(f2) */
      result = subst__ (&f1, inck, old, newf);
      if (subst__ (&f2, inck, old, newf))
	result = pk_true;
      if (result)
	{
	  if (inck)
	    {
	      /* if faut creer une autre formule */
	      aux = form_create ();
	      form_set_op (aux, form_get_op (*form));
	      *form = aux;
	    }
	  form_set_args (*form, f1, f2);
	}
      return result;

    default:
      return pk_false;
    }
}

/****************************************************************************/
static boolean
substitute (form_type * form, boolean inck, form_type old, form_type newf)
{
  return subst__ (form, inck, old, newf);
}

/****************************************************************************/
static void
reduce__ (form_type * form, op_type op, form_type f1, form_type f2)
{
  form_type one_form, zero_form;
  one_form = form_one ();
  zero_form = form_zero ();
  switch (op)
    {
    case FORM_AND:
      {
	if (f1 == zero_form)
	  {
	    *form = zero_form;
	    return;
	  }
	if (f1 == one_form)
	  {
	    *form = f2;
	    return;
	  }
	if (f2 == zero_form)
	  {
	    *form = zero_form;
	    return;
	  }
	if (f2 == one_form)
	  {
	    *form = f1;
	    return;
	  }
	break;
      }
    case FORM_OR:
      {
	if (f1 == zero_form)
	  {
	    *form = f2;
	    return;
	  }
	if (f1 == one_form)
	  {
	    *form = one_form;
	    return;
	  }
	if (f2 == zero_form)
	  {
	    *form = f1;
	    return;
	  }
	if (f2 == one_form)
	  {
	    *form = one_form;
	    return;
	  }
	break;
      }
    case FORM_DIFF:
      {
	if (f1 == zero_form)
	  {
	    *form = zero_form;
	    return;
	  }
	if (f2 == zero_form)
	  {
	    *form = f1;
	    return;
	  }
	if (f2 == one_form)
	  {
	    *form = zero_form;
	    return;
	  }
	break;
      }
    default:
      fprintf (stderr, "error: unauthorized value for op\n");
    }
}

/****************************************************************************/
static void
reduce (form_type * form)
{
  form_type f1, f2;
  op_type op;

  op = form_get_op (*form);
  switch (op)
    {
    case FORM_LIT:
      return;

    case FORM_AND:
    case FORM_OR:
    case FORM_DIFF:
      {
	form_get_args (*form, &f1, &f2);
	reduce (&f1);
	reduce (&f2);
	form_set_args (*form, f1, f2);
	reduce__ (form, op, f1, f2);
	return;
      }
    default:
      return;
    }
}

/****************************************************************************/
static clock_type
recur_retr (clock_type root, ptinsth h)
{
  clock_type result;
  list_type children;

  result = NILCLOCK;
  if (h == root->horl)
    return root;
  children = ck_get_children (root);
  list_set_to_last (children);
  while (!list_is_at_end (children))
    {
      result = recur_retr ((clock_type) list_get_current (children), h);
      if (result != NILCLOCK)
	return result;
      list_backward (children);
    }
  return NILCLOCK;
}

/*************************************************************************/
static clock_type
retrieve_by_insth (clock_type ck, ptinsth h)
{
  while (ck->parent != NILCLOCK)
    ck = ck->parent;
  return recur_retr (ck, h);
}



/****************************************************************************/
static clock_type
retrieve_by_id (clock_type root, bdd_id_boolean_variable_type id)
{
  clock_type result;
  list_type children;
  result = NILCLOCK;
  if (id == root->id)
    return root;
  children = ck_get_children (root);
  list_set_to_last (children);
  while (!list_is_at_end (children))
    {
      result = retrieve_by_id ((clock_type) list_get_current (children), id);
      if (result != NILCLOCK)
	return result;
      list_backward (children);
    }
  return NILCLOCK;
}

/****************************************************************************/
static form_type
make_lit_form (clock_type ck)
{
  form_type result;
  result = form_create ();
  form_set_op (result, FORM_LIT);
  form_set_literal (result, (literal_type) ck);
  return result;
}



/****************************************************************************/
static void
recur_vars (form_type form, supp_type vars)
{
  op_type op;
  form_type f1, f2;
  clock_type aux;

  op = form_get_op (form);
  switch (op)
    {
    case FORM_AND:
    case FORM_OR:
    case FORM_DIFF:
      {
	form_get_args (form, &f1, &f2);
	recur_vars (f1, vars);
	recur_vars (f2, vars);
	break;
      }
    case FORM_LIT:
      {
	aux = (clock_type) form_get_literal (form);
	supp_union (vars, aux->vars);
	break;
      }
    default:
      {
	fprintf (stderr, "big bug!!!\nUnknown op for formula\n");
	// exit(1);
      }
    }
}

/****************************************************************************/
static boolean
is_in_l (clock_type ck, bdd_id_boolean_variable_type id)
{
  while (ck != NILCLOCK)
    {
      if (ck->support == NIL_bddSupportSet_type)
	return pk_false;
      if (supp_is_member (ck->support, id))
	return pk_true;
      ck = ck->parent;
    }
  return pk_false;
}

/****************************************************************************/
static void
recompute_vars (clock_type ck)
{
  bdd_type one;
  supp_destroy (ck->vars);
  one = pK_bdd_one ();
  ck->vars = supp_create (one);
  pK_bdd_free (one);
  recur_vars (ck->form, ck->vars);
}

/****************************************************************************/
static boolean
is_subst_able (clock_type root, supp_type depend_vars, bdd_id_boolean_variable_type id)
{
  boolean result;
  clock_type ck;
  bdd_id_boolean_variable_type dep_var;

  result = pk_true;
  supp_set_to_max (depend_vars);
  while (result && (!supp_is_at_end (depend_vars)))
    {
      dep_var = supp_get_current (depend_vars);
      ck = retrieve_by_id (root, dep_var);
      result = result && (!is_in_l (ck, id));
      supp_backward (depend_vars);
    }
  return result;
}
