/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*
  Placement bdd de l'horloge H dont la formule est ?H1 ?OP ?H2. H1 et H2
  sont senses etre dans le meme arbre. 
  !PARENT denote le pere. !ALREADY permet de renvoyer un fils de !PARENT qui aurait le meme bdd; si !ALREADY
  est different de ?H, c'est qu'il y en a un.
  -------------------
  Modif Loic Besnard : 
  cette fonction rend 
  -la valeur "vrai" en cas d'un placement 
  -la valeur "faux" (pour une unification).
*/
extern boolean
findparent (ptinsth h, ptinsth h1, ptinsth h2, int op, ptinsth * already, clock_type * associated)
{
  bdd_type bdd;
  clock_type ck, parentck;
  clock_type ck1, ck2, head;

  ck1 = get_ck (Table, h1);
  ck2 = get_ck (Table, h2);
  *associated = NILCLOCK;
  findparent_ck (Table, ck1, ck2, op, &parentck, &head, &bdd);
  if ((head == NILCLOCK) && (parentck != NILCLOCK) && (bdd != NILBDD))
    {
      /* cas d'un placement */
      ck = get_ck (Table, h);
      *associated = ck;
      ck_set_parent (ck, parentck);
      list_insert_last (ck_get_children (parentck), ck);
      ck_set_bdd (ck, bdd);
      return pk_true;
    };
  if (head == NILCLOCK)
    return pk_true;
  /* cas d'une unification */
  *already = ck_get_horl (head);
  return pk_false;
}

/**
   @return 
   - "true" if the clock h has a asociated bdd in the global table. In this case, the same bdd is associated with the other clock (hc)
   -"false" otherwise.
*/
extern boolean
bdd_clone (ptinsth h, ptinsth hc, ptinsth fhc)
{
  clock_type ck, ckclone, parent;

  ck = retrieve_clock (Table, h);
  if (ck == NILCLOCK)
    return pk_false;
  ckclone = retrieve_clock (Table, hc);
  if (ckclone == NILCLOCK)
    ckclone = mk_new_clock_type (Table, hc);
  ck_clone (ckclone, ck);
  parent = retrieve_clock (Table, fhc);
  if (parent == NILCLOCK)
    parent = mk_new_clock_type (Table, fhc);
  ck_set_parent (ckclone, parent);
  list_insert_last (ck_get_children (parent), ckclone);
  return pk_true;
}

/** BDDHierarchy is initialized from ClockHierarchy2BDD
    It initializes BDD2ClockHierarchy
    @param aClockSystem is assumed to be initialized with the null clock defined
 */
extern void
BDDhierarchy_init (ptClockSystem aClockSystem)
{
  if (Table == NIL_list_type)
    {
      Table = list_create ();
      pK_bdd_init (0);
    }
  if (the_nullClock_BDD_copy == NILINSTH)
    {
      the_nullClock_BDD_copy = BDD2ClockHierarchy_init (aClockSystem);
      the_null_clock_type = mk_new_clock_type (Table, the_nullClock_BDD_copy);
      ck_set_bdd (the_null_clock_type, pK_bdd_zero ());
    }
}


extern void
print_all_bdd ()
{
  /*  
     list_type table;
     clock_type current_clock;
     bdd_type current_bdd;

     table = Table;
     list_set_to_first(table);
     while (!list_is_at_end(table))
     {
     current_clock = (clock_type) list_get_current(table);
     current_bdd = ck_get_bdd(current_clock);
     bdd_print(current_bdd);
     list_forward(table);
     }
   */
}


/**
   Removing the link between a clock and its bdd.
*/
extern boolean
bdd_remove (ptinsth h)
{
  list_type table;
  clock_type current;

  table = Table;
  list_set_to_first (table);
  while (!list_is_at_end (table))
    {
      current = (clock_type) list_get_current (table);
      if (isClockEq (ck_get_horl (current), h))
	{
	  list_remove_current (table);
	  return pk_true;
	}
      list_forward (table);
    }
  return pk_false;
}
