/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*
  modifs Loic Besnard :
  uint remplace par bdd_id_boolean_variable_type, ou bdd_rank_in_order_type selon le contexte.
  La gestion de l'ordre se fait NECESSAIREMENT en passant par le module d'interfacage 
  => ajout de pK_bdd_get_order_first(), pK_bdd_get_order_next().
*/

/* Codage des operations booleenes */
#define OPER_AND 0
#define OPER_OR 1
#define OPER_DIFF 2

#define ARG2_NOT_FOUND 3	/*** ces cas ne sont pas cense's arriver   ***/
#define ARG1_NOT_FOUND 2	/*** mais bon; on les pre'voit quand-me^me ***/
static list_type Table = NIL_list_type;
static ptinsth the_nullClock_BDD_copy = NILINSTH;
static clock_type the_null_clock_type;

static clock_type update (list_type table, clock_type ck);	/* refs en avant (revoir
								   graphe des appels) */
static clock_type retrieve_ck_by_id (clock_type ck, bdd_id_boolean_variable_type id);
static clock_type treat_hnull (list_type table, ptinsth h);
static clock_type treat_sub (list_type table, ptinsth h);
static clock_type treat_nosub (list_type table, ptinsth h);
static clock_type treat_racine (list_type table, ptinsth h);
static void recur_recomp (clock_type ck, bdd_rank_in_order_type * ord);
static void unify (clock_type repr, clock_type ck);
static bdd_type compute_gprime (bdd_type g, bdd_type pot, boolean * rapid_exclusion);
static clock_type get_ck (list_type table, ptinsth h);


/***                                                                      ***/
static clock_type
get_junction (list_type table, clock_type ck1, clock_type ck2)
{
  list_type asc1;
  clock_type ck;

  asc1 = list_create ();
  ck = ck1;
  while (ck != NILCLOCK)
    {
      ck = update (table, ck);
      list_insert_last (asc1, ck);
      ck = ck_get_parent (ck);
    }
  ck = ck2;
  while (ck != NILCLOCK)
    {
      ck = update (table, ck);
      list_set_to_first (asc1);
      while (!list_is_at_end (asc1))
	{
	  if (ck == (clock_type) list_get_current (asc1))
	    {
	      list_destroy (asc1);
	      return ck;
	    }
	  list_forward (asc1);
	}
      ck = ck_get_parent (ck);
    }
  fprintf (stderr, "severe error\nnot in the same tree\n");
  return NILCLOCK;
}

/***                                                                      ***/
static void
recompute_ord (clock_type root)
{
  bdd_rank_in_order_type ord;
  ord = pK_bdd_get_order_first ();	/* modif Lbesnard */
  recur_recomp (root, &ord);
}

/***                                                                      ***/
static bdd_type
compute_l (clock_type junction, clock_type ck)
{
  bdd_type result;
  bdd_type aux;
  result = pK_bdd_one ();
  while (ck != junction)
    {
      aux = result;
      result = pK_bdd_and (result, ck_get_bdd (ck));
      pK_bdd_free (aux);
      ck = ck_get_parent (ck);
    }
  return result;
}

/****************************************************************************/
static bdd_type
bdd_compute (bdd_type b1, bdd_type b2, int op)
{
  bdd_type result, aux;

  switch (op)
    {
    case OPER_AND:
      {
	result = pK_bdd_and (b1, b2);
	break;
      }
    case OPER_OR:
      {
	result = pK_bdd_or (b1, b2);
	break;
      }
    case OPER_DIFF:
      {
	aux = pK_bdd_not (b2);
	result = pK_bdd_and (b1, aux);
	pK_bdd_free (aux);
	break;
      }
    default:
      {
	fprintf (stderr, "fatal error\nunexpected bdd operator %d\n", op);
	result = b1;
	//      exit(1);
      }
    }
  return result;
}

static boolean is_pot_parent (clock_type pot, bdd_type g, bdd_type * gp);
/***                                                                      ***/
static void
bdd_parent (list_type table, clock_type junction, bdd_type bdd, clock_type * parent, bdd_type * gp)
{
  clock_type pot;
  boolean stop;
  bdd_id_boolean_variable_type id;
  bdd_type gprime;
  list_type asc;

  id = pK_bdd_lastvar_id (bdd);
  pot = retrieve_ck_by_id (junction, id);
  pot = ck_get_parent (pot);
  asc = list_create ();
  while (pot != junction)
    {
      list_insert_first (asc, (itemValue_type) pot);
      pot = ck_get_parent (pot);
    }
  *gp = pK_bdd_dup (bdd);
  *parent = junction;
  list_set_to_first (asc);
  stop = pk_false;
  while (!stop)
    if (list_is_at_end (asc))
      stop = pk_true;
    else
      {
	pot = (clock_type) list_get_current (asc);
	if (is_pot_parent (pot, *gp, &gprime))
	  {
	    list_forward (asc);
	    pK_bdd_free (*gp);
	    *gp = gprime;
	    *parent = pot;
	  }
	else
	  stop = pk_true;
      }
  list_destroy (asc);
  return;
}

/*****************************************************************************/
static clock_type
search_head (list_type children, bdd_type bdd)
{
  clock_type sibling;
  list_set_to_first (children);
  while (!list_is_at_end (children))
    {
      sibling = (clock_type) list_get_current (children);
      if (pK_bdd_equal (bdd, ck_get_bdd (sibling)))
	return sibling;
      list_forward (children);
    }
  return NILCLOCK;
}

static boolean
isClockEq (ptinsth hcurrent, ptinsth h)
{
  return (hcurrent == h);
}

static clock_type
mk_new_clock_type (list_type table, ptinsth h)
{
  clock_type ck;

  ck = ck_create ();
  ck_set_horl (ck, h);
  list_insert_last (table, ck);
  return ck;
}

/***                                                                      ***/
static clock_type
retrieve_clock (list_type table, ptinsth h)
{
  clock_type current;
  list_set_to_first (table);
  while (!list_is_at_end (table))
    {
      current = (clock_type) list_get_current (table);
      if (isClockEq (ck_get_horl (current), h))
	return current;
      list_forward (table);
    }
  return NILCLOCK;
}

/***                                                                      ***/
static clock_type
treat_sub (list_type table, ptinsth h)
{
  clock_type ck, parent, ck2, parent_head;

  ck = mk_new_clock_type (table, h);

  ck_set_id (ck, pK_bdd_new_id ());
  ck_set_bdd (ck, pK_bdd_get (ck_get_id (ck)));

  ck2 = mk_new_clock_type (table, mygethnosub (h));
  ck_set_bdd (ck2, pK_bdd_not (ck_get_bdd (ck)));
  if (h != mygetracine (h))
    {
      parent = get_ck (table, mygetpere (h));
      parent_head = ck_get_head (parent);
      /* le test suivant est rendu necessaire par le le fait */
      /* que l'on veut utiliser findparent_ck pour le parcours */
      /* apres clockcalculus; */
      /* si findparent_ck est utilise  au cours du calcul, tous */
      /* tous noeuds conserves dans la table ont leur head nil */
      if (parent_head != NILCLOCK)
	parent = parent_head;
      ck_set_parent (ck, parent);
      list_insert_last (ck_get_children (parent), ck);
      ck_set_parent (ck2, parent);
      list_insert_last (ck_get_children (parent), ck2);
    }
  return ck;
}

/****************************************************************************/
static clock_type
treat_nosub (list_type table, ptinsth h)
{
  clock_type ck;
  ck = get_ck (table, mygethsub (h));
  ck = get_ck (table, h);	/* la deuxieme fois est force'ment la bonne */
  if (ck == NILCLOCK)		/* neanmoins on s'en assure */
    fprintf (stderr, "nosub not found\n");
  return ck;
}

/***                                                                      ***/
static clock_type
treat_racine (list_type table, ptinsth h)
{
  return mk_new_clock_type (table, h);
}

/****************************************************************************/
/***                                                                      ***/
/***                                                                      ***/
static clock_type
get_ck (list_type table, ptinsth h)
{
  clock_type result;

  if (h == the_nullClock_BDD_copy)
    return the_null_clock_type;
  result = retrieve_clock (table, h);
  if (result == NILCLOCK)
    {
      if (myishsub (h))
	return treat_sub (table, h);
      if (myishnosub (h))
	return treat_nosub (table, h);
      if (h == mygetracine (h))
	return treat_racine (table, h);
      return mk_new_clock_type (table, h);
    }
  return result;
}

/***                                                                      ***/
static clock_type
update (list_type table, clock_type ck)
{
  ptinsth h, repr;
  clock_type reprck, aux, aux2;

  reprck = ck_get_head (ck);
  if (reprck != NILCLOCK)
    {
      unify (reprck, ck);
      return reprck;
    }

  h = ck_get_horl (ck);
  repr = mygetrepr (h);
  if (h != repr)
    {
      reprck = get_ck (table, repr);
      unify (reprck, ck);
      ck = reprck;
    }
  aux = ck_get_parent (ck);
  if (aux == NILCLOCK)
    {
      h = ck_get_horl (ck);
      if (h != mygetracine (h))
	{
	  aux = get_ck (table, mygetpere (h));
	  ck_set_parent (ck, aux);
	  list_insert_last (ck_get_children (aux), ck);
	  if (myishsub (h))
	    {
	      aux2 = get_ck (table, mygethnosub (h));
	      if (ck_get_parent (aux2) == NILCLOCK)
		{
		  ck_set_parent (aux2, aux);
		  list_insert_last (ck_get_children (aux), aux2);
		}
	    }
	  else if (myishnosub (h))
	    {
	      aux2 = get_ck (table, mygethsub (h));
	      if (ck_get_parent (aux2) == NILCLOCK)
		{
		  ck_set_parent (aux2, aux);
		  list_insert_last (ck_get_children (aux), aux2);
		}
	    }
	}
    }
  return ck;
}

/***                                                                      ***/
static void
recur_recomp (clock_type ck, bdd_rank_in_order_type * ord)
{
  list_type children;

  pK_bdd_set_ord (ck_get_id (ck), *ord);
  *ord = pK_bdd_get_order_next (*ord);	/* Modif L. Besnard :  *ord = *ord + 1; remplace par  */
  children = ck_get_children (ck);
  list_set_to_first (children);
  while (!list_is_at_end (children))
    {
      recur_recomp ((clock_type) list_get_current (children), ord);
      list_forward (children);
    }
}

/*****************************************************************************/
static clock_type
retrieve_ck_by_id (clock_type ck, bdd_id_boolean_variable_type id)
{
  clock_type result;
  list_type children;

  result = NILCLOCK;
  if (id == ck_get_id (ck))
    return ck;
  children = ck_get_children (ck);
  list_set_to_last (children);
  while (!list_is_at_end (children))
    {
      result = retrieve_ck_by_id ((clock_type) list_get_current (children), id);
      if (result != NILCLOCK)
	return result;
      list_backward (children);
    }
  return NILCLOCK;
}

/***                                                                      ***/
static boolean
is_pot_parent (clock_type pot, bdd_type g, bdd_type * gp)
{
  boolean rapid_exclusion;
  bdd_type aux;
  boolean result;

  *gp = compute_gprime (g, ck_get_bdd (pot), &rapid_exclusion);
  if (rapid_exclusion)
    return pk_false;
  aux = pK_bdd_and (ck_get_bdd (pot), *gp);
  result = (pK_bdd_equal (aux, g) && (retrieve_ck_by_id (pot, pK_bdd_topvar_id (*gp)) != NILCLOCK));
  pK_bdd_free (aux);
  if (!result)
    {
      pK_bdd_free (*gp);	/* on cherche a liberer le moindre espace memoire */
      *gp = NIL_bdd_type;
      return pk_false;
    }
  return pk_true;
}


/***                                                                      ***/
static void
unify (clock_type repr, clock_type ck)
{
  list_type ch1, ch2;
  clock_type current;
  ch1 = ck_get_children (repr);
  ch2 = ck_get_children (ck);
  list_set_to_first (ch2);
  while (!list_is_at_end (ch2))
    {
      current = (clock_type) list_get_current (ch2);
      list_insert_last (ch1, (itemValue_type) current);
      ck_set_parent (current, repr);
      list_remove_current (ch2);
    }
  ck_set_head (ck, repr);
}

static boolean has_1path (bdd_type bdd, list_type path);
/***                                                                      ***/
static bdd_type
compute_gprime (bdd_type g, bdd_type pot, boolean * rapid_exclusion)
{
  list_type path;
  bdd_type result;
  bdd_type aux, aux2;

  path = list_create ();
  has_1path (pot, path);
  /* toujours vrai car non nul */
  list_set_to_first (path);
  result = pK_bdd_dup (g);
  /* on va faire les manip sur la copie */
  /* ce pour ne pas faire un free sur g */
  *rapid_exclusion = pk_false;
  while (!list_is_at_end (path) && !(*rapid_exclusion))
    {
      aux = (bdd_type) list_get_current (path);
      if (pK_bdd_topvar_id (aux) != pK_bdd_topvar_id (result))
	{
	  *rapid_exclusion = pk_true;
	  pK_bdd_free (result);
	  result = NIL_bdd_type;
	}
      else
	{
	  aux2 = pK_bdd_cofactor (result, aux);
	  pK_bdd_free (result);
	  result = aux2;
	  list_forward (path);
	}
    }
  /* nous allons liberer les bdds qui sont dedans */
  list_set_to_first (path);
  while (!list_is_at_end (path))
    {
      pK_bdd_free ((bdd_type) list_get_current (path));
      list_remove_current (path);
    }
  list_destroy (path);
  return result;
}

/***                                                                      ***/
static boolean
has_1path (bdd_type bdd, list_type path)
{
  bdd_type aux;
  aux = pK_bdd_one ();
  if (pK_bdd_equal (bdd, aux))
    {
      pK_bdd_free (aux);
      return pk_true;
    }
  pK_bdd_free (aux);
  aux = pK_bdd_zero ();
  if (pK_bdd_equal (bdd, aux))
    {
      pK_bdd_free (aux);
      return pk_false;
    }
  pK_bdd_free (aux);
  aux = pK_bdd_then (bdd);
  if (has_1path (aux, path))
    {
      list_insert_first (path, (itemValue_type) pK_bdd_topvar (bdd));
      pK_bdd_free (aux);
      return pk_true;
    }
  pK_bdd_free (aux);
  aux = pK_bdd_else (bdd);
  if (has_1path (aux, path))
    {
      pK_bdd_free (aux);
      aux = pK_bdd_topvar (bdd);
      list_insert_first (path, (itemValue_type) pK_bdd_not (aux));
      pK_bdd_free (aux);
      return pk_true;
    }
  return pk_false;
}



/****************************************************************************/
/***                                                                      ***/
/***                                                                      ***/
static void
findparent_ck (list_type table, clock_type ck1, clock_type ck2, int op, clock_type * parentck, clock_type * head, bdd_type * bdd_result)
{
  clock_type junction;
  bdd_type bdd, b1, b2;
  list_type children;

  *head = NILCLOCK;
  *parentck = NILCLOCK;
  *bdd_result = NIL_bdd_type;
  junction = get_junction (table, ck1, ck2);
  recompute_ord (junction);
  b1 = compute_l (junction, ck1);
  b2 = compute_l (junction, ck2);
  bdd = bdd_compute (b1, b2, op);
  pK_bdd_free (b1);
  pK_bdd_free (b2);
  b1 = pK_bdd_one ();
  if (pK_bdd_equal (bdd, b1))
    {
      pK_bdd_free (b1);
      pK_bdd_free (bdd);
      *head = junction;
      return;
    }
  pK_bdd_free (b1);
  b1 = pK_bdd_zero ();
  if (pK_bdd_equal (bdd, b1))
    {
      pK_bdd_free (b1);
      pK_bdd_free (bdd);
      *head = the_null_clock_type;
      return;
    }
  pK_bdd_free (b1);
  bdd_parent (table, junction, bdd, parentck, &b1);
  children = ck_get_children (*parentck);
  *head = search_head (children, b1);
  if (*head != NILCLOCK)
    {
      pK_bdd_free (b1);
      pK_bdd_free (bdd);
      return;
    }
  pK_bdd_free (bdd);
  *bdd_result = b1;
}
