/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Formula Attributes.
   @param clkv (in) any signal expression.
   @param num (in) a mark.
   Removes the mark num to the expression clkv.
   @return node.
*/
  void
ClockFormula::removeMark (ptinsth clkv, typeMark num)
{
  ptinsth_removeMark (clkv, num);
}

/** Formula Attributes.
   @returns the pragmas assigned to the clock vh.
*/
ptlpragmas ClockFormula::getPragmas (ptinsth vh)
{
  return ptinsth_get_pragmas (vh);
}

/** Formula Attributes.
   @return l'attribut de numéro ?NUM affecte a l'expression ?CLK.
*/
boolean ClockFormula::getAttribute (ptinsth clk, typeMark num, attributeValue * valAttribute)
{
  return ptinsth_getAttribute (clk, num, valAttribute);
}

/** Formula Attributes.
  @param clkv (in) any clock.
  @param num (in) an attribute number.
  @return true if the expression clkv is annotated by the num-th mark, false otherwise.
*/
boolean ClockFormula::satisfiesMarked (ptinsth clkv, typeMark num)
{
  return ptinsth_satisfies_marked (clkv, num);
}

/** Formula Attributes.
  @param clkv (in) any signal expression.
  @param num (in) a mark.
  Adds the mark num to the expression clkv.  
*/
void
ClockFormula::addMark (ptinsth clkv, typeMark num)
{
  ptinsth_addMark (clkv, num);
}

/** Attribute.
   Adds the pragma aPragma to the clock vh */
void
ClockFormula::addPragma (ptinsth vh, ptpragma aPragma)
{
  ptinsth_addPragma (vh, aPragma);
}

/** Attribute.
   Affecte l'attribut ?VATTR de numéro ?NUM à l'expression ?CLK */
void
ClockFormula::addAttribute (ptinsth clk, typeMark num, attributeValue vattr)
{
  ptinsth_addAttribute (clk, num, vattr);
}

/** Attribute.
   Supprime l'attribut de  numéro ?NUM s'il est asscocié à l'expression ?CLK. */
void
ClockFormula::removeAttribute (ptinsth clk, typeMark num)
{
  ptinsth_removeAttribute (clk, num);
}

/**
   It marks by v the clocks of list.
 */
void
ClockFormula::marklClocks (ptlinsth list, typeMark v)
{
  ptinsth vh;

  while (ptlinsth_get_next (&list, &vh))
    ptinsth_addMark (vh, v);
}

/**
   It unmarks the clocks of list marked by v.
 */
void
ClockFormula::resetMarklClocks (ptlinsth list, typeMark v)
{
  ptinsth vh, vh1;

  while (ptlinsth_get_next (&list, &vh))
    {
      vh = ptinsth_get_clhead (vh);
      ptinsth_removeMark (vh, v);
      vh1 = ptinsth_get_clsucc (vh, pk_false);
      while (vh1 != vh)
	{
	  ptinsth_removeMark (vh1, v);
	  vh1 = ptinsth_get_clsucc (vh1, pk_false);
	}
    }
}
