/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return the subgraph assigned to a clock (vclk).
*/
  ptgfd ClockFormula::getClockSubGraph (ptinsth vclk)
{
  // gfd should be an internal representation. (or a subgraph node)
  return ptinsth_get_gfd (ptinsth_get_clhead (vclk));
}

/**  ---- Sub-graph assigned to a clock ---.
     It resets the interface of the graph assigned to the operand clock (clk).
*/
void
ClockFormula::resetSubGraphInterface (ptinsth clk)
{
  ptinsth_reset_gfd_interface (ptinsth_get_clhead (clk));
}

/**
   @return the graph associated with a clock (vh).
*/
ptGraphInfos ClockFormula::getGraphInfos (ptinsth vh)
{
  return ptinsth_get_GraphInfos (vh);
}

/**
   It set the graph (vg) associated with a clock (vh).
*/
void
ClockFormula::setGraphInfos (ptinsth vh, ptGraphInfos vg)
{
  ptinsth_set_GraphInfos (vh, vg);
}

/**
   @return the nodes assigned to a clock (vclk).
*/
ptlINode ClockFormula::getNodesAssignedTo (ptinsth vclk)
{
  return ptinsth_get_nodes (ptinsth_get_clhead (vclk));
}

/**
   @return the clocks assigned to a clock (vclk)
*/
ptlinsth ClockFormula::getClocksAssignedTo (ptinsth vclk)
{
  return ptinsth_get_clocks (ptinsth_get_clhead (vclk));
}

/** 
    @return the set of nodes that references a clock (vh).
    Dependence clock to node can be managed.
*/
ptlINode ClockFormula::getNodeReferencesTo (ptinsth vh)
{
  return ptinsth_get_lsucci (ptinsth_get_clhead (vh));
}

/** 
    It assigns the set of nodes (l) that references a clock (vh).
    Dependence clock to node can be managed.
*/
void
ClockFormula::setNodeReferencesTo (ptinsth vh, ptlINode l)
{
  ptinsth_set_lsucci (vh, l);
}

/** ---- Sub-graph dominated by a clock ---.
    It assigns the interface (v) to the clock operand clock (clk).
    v is The interface of the subhierachy dominated by the operand clock (clk).
*/
void
ClockFormula::setInterfaceSubHierarchy (ptinsth clk, ptinterface v)
{
  ptinsth_set_interface (clk, v);
}

/** ---- Sub-graph dominated by a clock ---.
    @returns the interface of the subhierachy dominated by the operand clock (clk).
    The structure is created if it does not exist.
*/
ptinterface ClockFormula::getInterfaceSubHierarchy (ptinsth clk)
{
  return ptinsth_getset_interface (ptinsth_get_clhead (clk));
}

/** ---- Composed node management --
    @param clk (in) a clock.
    @param cNode (out) the composed node in which the clock clk is.
    @return true if a clock is defined in a composed node, false otherwise.
*/
boolean ClockFormula::getInComposedNode (ptinsth clk, ptINode * cNode)
{
  return ptinsth_get_inComposedNode (clk, cNode);
}

/** ---- Composed node management
    A clock (clk) becomes defined in a composed node (cNode). 
*/
void
ClockFormula::setInComposedNode (ptinsth clk, ptINode cNode)
{
  ptinsth_set_inComposedNode (clk, cNode);
}

/** 
   It sets dependencies for a not clock (vh).
*/
void
ClockFormula::modifyDependenceComplementaryClock (ptinsth vh)
{
  ptinsth_modify_dep_hnot (vh);
}

/**
   It merges in vh2 (side effect) the graphs of the two operand clocks (vh1,vh2).
   The name of the old object (vh1) is kept if it is the clock is a clock of a signal and not vh2.
*/
void
ClockFormula::fusionInfosGraph (ptinsth vh1, ptinsth vh2)
{
  ptident id;
  ptexpcal aSig;

  if (ptinsth_get_lsig (vh2) == NILLINSTANCES)
    if (ptinsth_get_lsig (vh1) != NILLINSTANCES)
      if (ptinsth_get_name (vh1, &id))
	{
	  if (isClockRepresentedBySignal (vh1, &aSig))
	    setClockRepresentedBySignal (vh2, aSig);
	  ptinsth_set_name (vh2, id);
	}
  ptinsth_fusion_infosGraph (vh1, vh2);
}


