/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_ _PROTECTED_
/**
   \note {this feature is still experimental and incomplete.}
   @return a new condition associated with an affine relation (affr).
*/
 ptcondition ClockSystem::new_affineExtractions(ptinsth vhorl, ptAffineRelation affr)
{
  insthClock vinsth, vinsthns;
  ptcondition aux;
  AffineRelation affrCompl;

  vinsth = ptinsth_newClock(sub);
  insertlhroots(vinsth);
  setAffineRelation(vinsth, affr);

  vinsthns = ptinsth_newClock(nosub);
  insertlhroots(vinsthns);

  if (ptAffineRelation_complementary(affr, &affrCompl))
    setAffineRelation(vinsthns, ptAffineRelation_duplicate(&affrCompl));

  aux = new condition(vinsth, vinsthns, vhorl, NILEXPCAL, NILLINSTANCES);
  nbConditions++;
  setArgSub(vinsth, aux);
  setReferenceToFrom(vinsth, vhorl);

  setArgSub(vinsthns, aux);
  setReferenceToFrom(vinsthns, vhorl);

  lssub = insertlcond(lssub, vinsth);

  setComplementary(vinsth, vinsthns);
  return aux;
}

/**
   \note {this feature is still experimental and incomplete.}
   Assumed: vhemb, when it is defined, is over vharg in the hierarchy.
   @return 
   - "true" if a formula (vharg) is an affine clock on vhemb. 
   It applies the composition (ptAffineRelation_simple_composition()) on affine clocks from vharg until vhemb.
   It returns the parameters of the result of the composition (affr).
   "false" otherwise
*/
boolean ClockSystem::getAffineRelationAt(ptinsth vharg, ptinsth vhemb, ptAffineRelation affr)
{
  ptinsth vh1;
  ptAffineRelation affropl1, affropr1, affrres1;

  if (ptinsth_is_hundef(vhemb))
    return pk_false;
  if (vharg == vhemb)
    return pk_false;
  vh1 = vharg;
  if (!getAffineRelation(vh1, &affrres1))
    return pk_false;
  affrres1 = ptAffineRelation_duplicate(affrres1);
  vh1 = getFather(vh1);
  while (vh1 != vhemb)
    {
      affropl1 = affrres1;
      if (!getAffineRelation(vh1, &affropr1))
	return pk_false;
      ptAffineRelation_simple_composition(affropr1, affropl1, affrres1);
      vh1 = getFather(vh1);
    }
  *affr = *affrres1;
  return pk_true;
}

/**
   \note {this feature is still experimental and incomplete.}
   Assumed: The operand vharg1 and vharg2 are in the same tree with vhemb as root and are not in the same branch.

   @return 
   - "true" are affine undersamplings ptaffr1 = (1, phi1, d1),  ptaffr2 = (1, phi2, d2) of vhemb.
   - "false" otherwise, ptaffr1, ptaffr2 are not modified.
*/
boolean ClockSystem::affineBranching(ptinsth vharg1, ptinsth vharg2, ptinsth vhemb, ptAffineRelation ptaffr1, ptAffineRelation ptaffr2)
{
  ptinsth vh1, vh2, vhe;
  AffineRelation affrres1, affrres2;

  vh1 = vharg1;
  vh2 = vharg2;
  vhe = vhemb;

  if (!getAffineRelationAt(vh1, vhe, &affrres1))
    return pk_false;
  if (!getAffineRelationAt(vh2, vhe, &affrres2))
    return pk_false;
  *ptaffr1 = affrres1;
  *ptaffr2 = affrres2;
  return pk_true;
}

/**
   \note {this feature is still experimental and incomplete.}
   @return "true" if vharg1 and vharg2 are equivalent affine undersamplings, "false" otherwise.
*/
boolean ClockSystem::isEquivAffineFormTree(ptinsth vharg1, ptinsth vharg2, ptinsth vhemb)
{
  AffineRelation affr1, affr2;

  if (!affineBranching(vharg1, vharg2, vhemb, &affr1, &affr2))
    return pk_false;
  if (!ptAffineRelation_isEquivalent(&affr1, &affr2))
    return pk_false;
  return pk_true;
}

/**
   \note {this feature is still experimental and incomplete.}
   @return "true" if vhsub is an affine undersampling, with vclk as father clock, and equivalent to affr, "false" otherwise.
   (see ptAffineRelation_isEquivalent()).
*/
boolean ClockSystem::isEquivalentAffineClock(ptinsth vhsub, ptinsth vclk, ptAffineRelation affr)
{
  ptAffineRelation ptaffraux;
  ptinsth vhnosub;

  if (!getAffineRelation(vhsub, &ptaffraux)) return pk_false;
  if (vclk == getClockSub(vhsub)) return pk_false;
  if (/* (vclk == getClockSub(vhsub)) && */ ptAffineRelation_isEquivalent(affr, ptaffraux)) return pk_true;
  getComplExtraction(vhsub, &vhnosub);
  if (getAffineRelation(vhnosub, &ptaffraux) && ptAffineRelation_isEquivalent(affr, ptaffraux)) return pk_true;
  return pk_false;
}

/**
   \note {this feature is still experimental and incomplete.}
   @return 
   -"true" if it exists an affine undersampling in the list (list), with affr as parameters and
   vclk as father clock. Then velem is this formula.
   - "false" otherwise, velem is not modified.
*/
boolean ClockSystem::searchAffineClock(ptlinsth list, ptAffineRelation affr, ptinsth vclk, ptcondition * velem)
{
  ptinsth vhsub;
  ptlinsth aux;

  aux = list;
  while (ptlinsth_get_next(&aux, &vhsub))
    if (isEquivalentAffineClock(vhsub, vclk, affr))
      {
	*velem = getArgSub(vhsub);
	return pk_true;
      }
  return pk_false;
}

/**
   \note {this feature is still experimental and incomplete.}
   @return 
   - "true" if  the affine undersampling, with affr as parameters and vh as father clock already exists in the system, 
   Then vhresult is this formula.
   - "false" otherwise, velem is not modified.
*/
boolean ClockSystem::rchlsubaff(ptinsth vh, ptAffineRelation affr, ptinsth * vhresult)
{
  boolean trouve;
  ptcondition velem;
  ptAffineRelation affraux;
  ptinsth vhaux;

  // Voir These Irina à partir de la page 162 
  trouve = searchAffineClock(lssub, affr, vh, &velem);
  if (!trouve)
    trouve = searchAffineClock(lsextract, affr, vh, &velem);
  if (!trouve)
    return pk_false;
  vhaux = velem->get_clksub();
  getAffineRelation(vhaux, &affraux);
  if (ptAffineRelation_compare(affraux, affr))
    {
      *vhresult = vhaux;
      *vhresult = getRepresentativeFormula(*vhresult);
      return pk_true;
    }
  vhaux = velem->get_clknosub();
  getAffineRelation(vhaux, &affraux);
  if (!ptAffineRelation_compare(affraux, affr))
    return pk_false;
  *vhresult = vhaux;
  *vhresult = getRepresentativeFormula(*vhresult);
  return pk_true;
}

/**
   \note {this feature is still experimental and incomplete.}
   @return 
   - "true" if the complement of vharg in vhemb is an affine clock on vhemb. Then, ptaffr is defined by the affine relation (1, phi, d) such that 
   complement(vharg) = sample{ptaffr}(vhemb)
   - "false" otherwise.
*/
boolean ClockSystem::isComplementAffineClock(ptinsth vharg, ptinsth vhemb, ptAffineRelation ptaffr)
{
  AffineRelation affr1, affr2, affrcompl, affrcomp /*, affrsup */ ;
  pkClockType ophorl;
  ptinsth vhsub, vFather;
  ptAffineRelation affrsup, affrsub, affrpere;

  if (vharg == vhemb)
    return pk_false;
  if (getAffineRelationAt(vharg, vhemb, &affr1))
    {
      if (!ptAffineRelation_complementary(&affr1, &affr2))
	return pk_false;
      ptAffineRelation_copy(&affr2, ptaffr);
      return pk_true;
    }

  ophorl = getOperator(vharg);
  if (ophorl != nosub)
    return pk_false;

  vhsub = (getArgSub(vharg))->get_clksub();
  if (!getAffineRelation(vhsub, &affrsub))
    return pk_false;
  vFather = getFather(vhsub);

  if (vFather == vhemb)
    {
      ptAffineRelation_copy(affrsub, ptaffr);
      return pk_true;
    }
  if (!getAffineRelation(vFather, &affrpere))
    return pk_false;
  ptAffineRelation_simple_composition(affrpere, affrsub, &affrcomp);
  while (vFather != vhemb)
    {
      if (!ptAffineRelation_complementary(affrpere, &affrcompl))
	return pk_false;
      if (!ptAffineRelation_comp_sup_is_affine(&affrcomp, &affrcompl, &affrsup))
	return pk_false;
      ptAffineRelation_copy(affrsup, &affrcomp);
      vFather = getFather(vFather);
      if (vFather != vhemb)
	if (!getAffineRelation(vFather, &affrpere))
	  return pk_false;
    }
  ptAffineRelation_copy(&affrcomp, ptaffr);
  return pk_true;
}

/** 
    \note {this feature is still experimental and incomplete.}
    Assumed: otherterm is the result of the placing of the formula <em> vharg1 ophorl vharg2 </em> at the branching clock vhemb.
    It assigns the affine relations to otherterm from the affine relations (for affine clocks) assigned to the operands (vharg1, vharg2).
*/
void ClockSystem::treeSetAffineRelation(ptinsth otherterm, ptinsth vharg1, ptinsth vharg2, ptinsth vhemb, pkClockType ophorl)
{
  AffineRelation affr1, affr2, affr3;
  ptAffineRelation ptaffr, affr;

  if (!getAffineRelation(otherterm, &affr))
    /* dans ce cas la relation affine de OTHERTERM obtenue a partir de 
       VHARG1 et VHARG2 doit être egale a PTAFFR ===> on ne recalcule pas */
    if (getAffineRelationAt(vharg1, vhemb, &affr1))
      switch (ophorl)
	{
	case hnot:
	  if (isComplementAffineClock(vharg2, vhemb, &affr3))
	    if (ptAffineRelation_comp_inf_is_affine(&affr1, &affr3, &ptaffr))
	      setAffineRelation(otherterm, ptaffr);
	  break;
	case infer:
	case super:
	  if (getAffineRelationAt(vharg2, vhemb, &affr2))
	    switch (ophorl)
	      {
	      case infer:
		if (ptAffineRelation_comp_inf_is_affine(&affr1, &affr2, &ptaffr))
		  setAffineRelation(otherterm, ptaffr);
		break;
	      case super:
		if (ptAffineRelation_comp_sup_is_affine(&affr1, &affr2, &ptaffr))
		  setAffineRelation(otherterm, ptaffr);
		break;
	      default:
		break;
	      }
	default:
	  break;
	}
}

/**
   \note {this feature is still experimental and incomplete.}
   @return "true" if the operand clocks (vh1, vh2) are equivalent affine clocks.
   The test of equivalence is defined in isEquivalentAffineClock().
*/
boolean ClockSystem::clockEqAffine(ptinsth vh1, ptinsth vh2)
{
  ptAffineRelation affr;
  if (vh1 == vh2) return pk_false;
  if (!getAffineRelation(vh1, &affr)) return pk_false;
  return isEquivalentAffineClock(vh2, getClockSub(vh1), affr);
}
