/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    @param vhemb (in) : assumed to be the branching clock of a set of clocks (lh);
    @param lh (in) : a set of clocks;
    @return the input list (lh) completed by the clocks of the branch vhi until vhemb, for vhi a clock of lh.
*/
  ptlinsth ClockSystem::CompleteUpstreamParametrization (ptlinsth lh, ptinsth vhemb)
{
  ptinsth vh;
  ptlinsth lres, lh1;

  lres = ptlinsth_duplicate (lh);
  while (ptlinsth_get_next (&lh, &vh))
    {
      lh1 = getClocksUntil (ptinsth_get_clhead (vh), vhemb);
      lres = ptlinsth_sorted_fusion (lres, lh1);
    }
  return lres;
}

/**
   @return the branching clock of a set of clocks (lh).
   If the list lh is empty
*/
ptinsth ClockSystem::getBranchingClocks (ptlinsth lh)
{
  ptinsth vh, vhemb;

  if (!ptlinsth_get_next (&lh, &vh))
    return hundef;
  vhemb = vh;
  while (ptlinsth_get_next (&lh, &vh))
    {
      vhemb = getBranching (vhemb, vh);
      if (ptinsth_is_hundef (vhemb))
	return getTickClock ();
    }
  return vhemb;
}

/**  @return true
     - if one of the clocks vh1 or vh2 is placed in the sub-tree of the other.
     - the two clocks are placed in different trees.
*/
boolean ClockSystem::existHierarchyConstraintsOn (ptinsth vharg1, ptinsth vharg2)
{
  int depth1, depth2;
  ptinsth vhson1, vhson2, vhemb;

  if (!satisfiesIsAnAssumption (vharg1))
    if (satisfiesSub (vharg1))
      {
	if (getClockSub (vharg1) == vharg2) return pk_true;
	// if (!satisfiesIsAnAssumption (vharg2) && satisfiesSub (vharg2)) return pk_true; // added then removed ! Aug 2016 LBesnard (Polyendochrounous ...)
      }

  if (!satisfiesIsAnAssumption (vharg2))
    if (satisfiesSub (vharg2))
      if (getClockSub (vharg2) == vharg1)
	return pk_true;

  if (getCompleteBranching (&vharg1, &vharg2, &vhson1, &vhson2, &vhemb))
    return !isEquivAffineFormTree (vharg1, vharg2, vhemb);

  if (!ptinsth_is_hundef (vhemb)) return pk_true;

  vhson1 = getRoot (vharg1, &depth1);
  vhson2 = getRoot (vharg2, &depth2);
  return (boolean) ((depth1 > 0) && (depth2 > 0));
}
