/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ void
ClockSystem::DEBUG_traceClock (char *methodName, insthFormula anInsthFormula)
{
  ptinsth_DEBUG_traceClock (methodName, anInsthFormula);
}

ClockSystem::ClockSystem ()
{
  assumedWithoutConstraints = pk_false;
  ModeSynchroSolving = pk_false;

  the_nullClock_CKS_copy = getNullClock ();
  hinit = getInitClock ();
  hundef = getUndefClock ();

  lvariables = NILLINSTH;
  lvariablesEnd = NILLINSTH;
  lssub = NILLINSTH;
  lsinfer = NILLINSTH;
  lssuper = NILLINSTH;
  lsnot = NILLINSTH;

  lsroots = NILLINSTH;
  lsrootsEnd = NILLINSTH;
  lsplaced = NILLINSTH;
  lsplacedEnd = NILLINSTH;

  lsextract = NILLINSTH;
  lssupintree = NILLINSTH;
  lssupintreeEnd = NILLINSTH;
  lsinfintree = NILLINSTH;
  lsinfintreeEnd = NILLINSTH;

  lterm1equpl = NILLINSTH;
  lterm2equpl = NILLINSTH;
  lterm1assertpl = NILLINSTH;
  lterm2assertpl = NILLINSTH;

  lterm1equ = NILLINSTH;
  lterm2equ = NILLINSTH;
  this_nullClockList = NILLINSTH;
  lsnullpl = NILLINSTH;
  lsmodified = NILLINSTH;

//  vinsth = mkClockObject(troot);
  initHierarchy (ptinsth_newClock (troot));
  nbConditions = 0;

  // for projection
  ExportedClocksIOClocks = NILLINSTH;
  ExportedClocksOthers = NILLINSTH;	// not built on IO clocks.
  ExportedSignalsInSpec = NILLINSTANCES;
  mark3 = getMark ();
  IsInSimplifyMode = true;
}

/** Destructor. */
ClockSystem::~ClockSystem ()
{
  // Freeing only the lists, the clocks/formula are freed in the ClockManager class.
  ptlinsth_free (lvariables);
  ptlinsth_free (lssub);
  ptlinsth_free (lsinfer);
  ptlinsth_free (lssuper);
  ptlinsth_free (lsnot);

  ptlinsth_free (lsroots);
  ptlinsth_free (lsplaced);

  ptlinsth_free (lsextract);
  ptlinsth_free (lssupintree);
  ptlinsth_free (lsinfintree);

  ptlinsth_free (lterm1equpl);
  ptlinsth_free (lterm2equpl);
  ptlinsth_free (lterm1assertpl);
  ptlinsth_free (lterm2assertpl);

  ptlinsth_free (lterm1equ);
  ptlinsth_free (lterm2equ);
  ptlinsth_free (this_nullClockList);
  ptlinsth_free (lsnullpl);
  ptlinsth_free (lsmodified);

  ptlinsth_free (ExportedClocksIOClocks);
  ptlinsth_free (ExportedClocksOthers);
  ptlinstances_free (ExportedSignalsInSpec);
}

/**
  Initializing a system.
  @param aExprManager (in) the expression manager assigned to the current graph.
  @return none.
*/
void
ClockSystem::initialize (ptExprManager aExprManager)
{
  EM = aExprManager;
}

/** 
    It marks, by v, all the clocks of the system.
    The universal clocks are not considered.
*/
void
ClockSystem::markAll (typeMark v)
{
  marklClocks (getlhroots (), v);
  marklClocks (getlhplaced (), v);
}

/** 
    It unmarks all the clocks of the system, marked by v.
*/
void
ClockSystem::resetMarkAll (typeMark v)
{
  resetMarklClocks (getlhroots (), v);
  resetMarklClocks (getlhplaced (), v);
}

/**
   It assigns the informations about the projection of the system on the 
   clocks of the inputs/outputs signals of the graph and the clocks of the 
   input/output data dependencies. It is coumputed outside the clock system.

   @param lh (in) set of clocks to be exported
   @param lConditions (in) set of booleans signals to be exported
*/
void
ClockSystem::setProjectedSystem (ptlinsth ClocksBuiltOnIOClock, ptlinsth ClocksNotBuiltOnIOClock, ptlinstances lConditions)
{
  //setProjectedSystem(ptlinsth lh, ptlinstances lConditions)
  ExportedClocksIOClocks = ClocksBuiltOnIOClock;
  ExportedClocksOthers = ClocksNotBuiltOnIOClock;
  ExportedSignalsInSpec = lConditions;
}

/**
  @param b (in) : at "true", it specifies that the system is assumed to be without constraints.
  @return the previous value of the tag.
*/
boolean ClockSystem::setConstraintsMode (boolean b)
{
  boolean x;

  x = assumedWithoutConstraints;
  assumedWithoutConstraints = b;
  return x;
}

/**
   It authorises (when b)/suspends (when not b) the merge of classes.
*/
void
ClockSystem::setModeSynchroSolving (boolean b)
{
  ModeSynchroSolving = b;
}

/**
   It authorises (when b)/suspends (when not b) the elementary simplifications.
*/
void
ClockSystem::setSimplifyMode (boolean b)
{
  IsInSimplifyMode = b;
}
