/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/** Formula.
   @return true if h is a equivalent formula (ie proved but kept in its class), false otherwise.
 */
  boolean ClockFormula::satisfiesIsAnEquivalentForm (ptinsth h)
{
  return ptinsth_IsAnEquivalentForm (h);
}

/** Formula.
   @return true if h is a assumption (ie not proved but assumed equivalent), false otherwise.
 */
boolean ClockFormula::satisfiesIsAnAssumption (ptinsth h)
{
  return ptinsth_IsAnAssumption (h);
}

/** Formula.
   @return the identifier assigned to the clock vh (NILIDENT if unkown)*/
ptident ClockFormula::getName (ptinsth vh)
{
  ptident
    id;
  ptinsth_get_name (vh, &id);
  return id;
}

/** Formula.
   @return the operator of the formula ?vh. */
pkClockType ClockFormula::getOperator (ptinsth vh)
{
  return ptinsth_get_operator (vh);
}

/** Formula.
   @param  vh (in) : an assumed binary formula.
   @param vharg1 (out) first operand of vh-formula.
   @param vharg2 (out) second operand of vh-formula.
   @return node.
*/
void
ClockFormula::getBinaryArgs (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  ptinsth_get_2args (vh, vharg1, vharg2);
}

/** Formula.
   @param vh (in) a clock formula.
   @param vharg1 (out) first operand of vh.Set only for vh a inf-formula.
   @param vharg2 (out) second operand of vh.Set only for vh a inf-formula.
   @return true if ?vh is a inf-formula
*/
boolean ClockFormula::getInfArgs (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  return ptinsth_get_infargs (vh, vharg1, vharg2);
}

/** Formula.
   @param vh (in) a clock formula.
   @param vharg1 (out) first operand of vh. Set only for vh a sup-formula.
   @param vharg2 (out) second operand of vh.Set only for vh a sup-formula.
   @return true if ?vh is a sup-formula
*/
boolean ClockFormula::getSupArgs (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  return ptinsth_get_supargs (vh, vharg1, vharg2);
}

/** Formula.
   @return true if the formula ?vh is a minus one(defined by !vharg2 - !vharg1.)
*/
boolean ClockFormula::getMinusArgs (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  return ptinsth_get_minusargs (vh, vharg1, vharg2);
}

/** Formula.
    @param vh (in) : a formula 
    @param vharg (out) argument of a unary formula (complementary , not operators)
    @return true if the formula vh is a unary formula, false otherwise.
*/
boolean ClockFormula::getUnaryArg (ptinsth vh, ptinsth * vharg)
{
  return ptinsth_get_1arg (vh, vharg);
}

/** Formula.
   @param vh  (in) a clock formula.
   @return "true" if vh is an undersampling one (sub, nosub operator)
*/
boolean ClockFormula::satisfiesSub (ptinsth vh)
{
  return ptinsth_satisfies_extraction (vh);
}

/** Formula.
   @param vh  (in) a clock formula.
   @return the condition for vh an undersampling, nil expression otherwise */
ptexpcal ClockFormula::getConditionSub (ptinsth vh)
{
  return ptinsth_get_condition (vh);
}

/** Formula.
    It assigns an expression (exp) and the list of the signals referenced in this expression (ls) to an undersampling formula (vh).
    No effects if the vformula is not an undersampling one.  
*/
void
ClockFormula::setConditionSub (ptinsth vh, ptexpcal exp, ptlinstances ls)
{
  ptinsth_set_condition (vh, exp, ls);
}

/**
    Formula.
    @return
      - pk_true if vh is an extraction, then vhc (out) is its complement (remind that: [c] ^+ [not c] = ^c).
      - false otherwise
*/
boolean ClockFormula::getComplExtraction (ptinsth vh, ptinsth * vhc)
{
  return ptinsth_get_complExtraction (vh, vhc);
}

/** Formula.
   @param vh  (in) a clock formula.Assumed to be a undersampling ([c], [not(c)]) formula (sub,nosub operators)
   @return the clock of the condition (^c)
*/
ptinsth ClockFormula::getClockSub (ptinsth vh)
{
  return ptinsth_get_clkCondition (vh);
}

/** Formula.
   @param vh  (in) a clock formula.
   @return the set of referenced signals in the condition of an sub-formula, nil list for
   other clock formula.
*/
ptlinstances ClockFormula::getSignalsOfCondition (ptinsth vh)
{
  return ptinsth_get_SignalOfcondition (vh);
}

/** Formula.
    @return the number assigned to a formula (vh).
 */
int
ClockFormula::getNum (ptinsth vh)
{
  return ptinsth_get_num (vh);
}

/** Formula.
    @return the condition associated with the undersampling formula (vh). 
    Assumed : vh is a undersampling.
*/
ptcondition ClockFormula::getArgSub (ptinsth vh)
{
  return ptinsth_get_argsub (vh);
}
