/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    It assigns to a clock (vh) the list of the expressions (l) that reference it.
*/
  void
ClockFormula::setExprReferencesTo (ptinsth vh, ptlinstances l)
{
  ptinsth_set_luse (vh, l);
}

/** Formula.
   @param vh  (in) a clock formula.
   @param l  (in) a list of clocks.
   @return none.
   It assigns the set of clocks (l) that references vh.
*/
void
ClockFormula::setReferencesAt (ptinsth vh, ptlinsth l)
{
  ptinsth_set_lsucch (vh, l);
}

/** Formula.
   @param vh  (in) a clock formula.
   @return the set of clocks that reference the clock operand.
*/
ptlinsth ClockFormula::getClockReferencesTo (ptinsth vh)
{
  return ptinsth_get_lsucch (vh);
}

/** Formula.
   @param vh  (in) a clock formula.
   @return the set of expressions that reference vh.
*/
ptlinstances ClockFormula::getExprReferencesTo (ptinsth vh)
{
  return ptinsth_get_luse (vh);
}

/** Formula.
    It deletes the reference to the formula vh for its operands.
*/
void
ClockFormula::removeLinksTo (ptinsth vh)
{
  ptinsth_remove_linksTo (vh);
}
