/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return "true" if the two affine parameters are equal, "false" otherwise.*/
extern boolean
AffineParam_compare (AffineParam ap1, AffineParam ap2)
{
  return (ap1 == ap2);
}

/** @return the max(app1, app2). */
extern AffineParam
AffineParam_max (AffineParam app1, AffineParam app2)
{
  if (app1 >= app2)
    return app1;
  return app2;
}

/** @return the min(app1, app2). */
extern AffineParam
AffineParam_min (AffineParam app1, AffineParam app2)
{
  if (app1 < app2)
    return app1;
  return app2;
}

/** @return the gcd(ap1,ap2) (greatest common divisor)
    \note not implemented.
*/
extern AffineParam
AffineParam_gcd_simple (AffineParam ap1, AffineParam ap2)
{
  // fprintf(stderr, "We are in function   AffineParam_gcd_simple   now!!!\n");
  // fprintf(stderr, "Function   AffineParam_gcd_simple   has not been implemented\n");
  return ap1;
}

/**
   @return n=gcd(ap1,ap2).
   - app1=ap1/n, 
   - app2=ap2/n
   \note not implemented.
*/
extern AffineParam
AffineParam_gcd_quotient (AffineParam ap1, AffineParam ap2, AffineParam * app1, AffineParam * app2)
{
  //  fprintf(stderr, "We are in function AffineParam_gcd_quotient     now!!!\n");
  //  fprintf(stderr, "Function  AffineParam_gcd_quotient    has not been implemented\n");
  return ap1;
}

/** 
   @return |ap|
*/
extern AffineParam
AffineParam_mod (AffineParam ap)
{
  if (ap < 0)
    return -ap;
  return ap;
}

/** 
    For a, b integers, the decomposition a = b*q + r is computed with q, r integers.
*/
extern void
AffineParam_div_and_mod (AffineParam a, AffineParam b, AffineParam * q, AffineParam * r)
{
  AffineParam aaux, baux, qaux, raux, qtemp, rtemp, signed_b_one;	/* signed_b_one =  1 si b > 0
									   signed_b_one = -1 si b < 0 */

  /**
     Methode de calcul : Soit qaux, raux tels que |a| = |b|*qaux + raux, 0 <= raux < |b|.
     Si a < 0 :
          Si raux = 0       ==>  q = -sign(b)*qaux, r = raux
	  Sinon (raux <> 0) ==>  q = -sign(b)*(qaux+1), r = sign(b)*b - raux
     Si a >= 0 : q = sign(b)*qaux, r = raux
  */
  aaux = AffineParam_mod (a);
  baux = AffineParam_mod (b);
  qaux = 0;
  while (aaux >= baux)
    {
      aaux = aaux - baux;
      qaux = qaux + 1;
    }
  raux = aaux;
  if (b > 0)
    signed_b_one = 1;
  else
    signed_b_one = -1;
  if (a >= 0)
    {
      qtemp = signed_b_one * qaux;
      rtemp = raux;
    }
  else
    {
      if (raux == 0)
	{
	  qtemp = -signed_b_one * qaux;
	  rtemp = raux;
	}
      else
	{
	  qtemp = -signed_b_one * (qaux + 1);
	  rtemp = signed_b_one * b - raux;
	}
    }
  *q = qtemp;
  *r = rtemp;
}
