/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return the n parameter of an affine relation pt=(n,phi,d).
*/
static AffineParam
ptAffineRelation_get_periodN (ptAffineRelation pt)
{
  return pt->period_n;
}

/**
   @return the phi parameter of an affine relation pt=(n,phi,d).
*/
static AffineParam
ptAffineRelation_get_phasis (ptAffineRelation pt)
{
  return pt->phasis;
}

/**
   @return the d parameter of an affine relation pt=(n,phi,d).
*/
static AffineParam
ptAffineRelation_get_periodD (ptAffineRelation pt)
{
  return pt->period_d;
}

/**
   It defines n=period_n for an affine relation pt=(n,phi,d).
*/
static void
ptAffineRelation_set_periodN (ptAffineRelation pt, AffineParam period_n)
{
  pt->period_n = period_n;
}

/**
   It defines phi=phasis for an affine relation pt=(period_n,phi,d).
*/
static void
ptAffineRelation_set_phasis (ptAffineRelation pt, AffineParam phasis)
{
  pt->phasis = phasis;
}

/**
   It defines d=period_d for an affine relation pt=(period_n,phi,d).
*/
static void
ptAffineRelation_set_periodD (ptAffineRelation pt, AffineParam period_d)
{
  pt->period_d = period_d;
}

/**
    @return 
    - "true" for ptin=(n, phi, d), (we consider the following decomposition (n, max(0, -phi), 1) with (1, max(0, phi), d)).
       - if phi=d and n = q*d, then ptout=(1, 0, d)
       - if n=d and n = q*phi, then ptout=(1, 0, d)
       - if d=q1*n  and phi= q2*n, then pout=(1, max(0,-phi), n)
   - "false" otherwise.
*/
static boolean
ptAffineRelation_sup_is_affine (ptAffineRelation ptin, ptAffineRelation * ptout)
{
  AffineParam q, r, period_n, phasis, period_d, phasisout;
  boolean result;

  result = pk_false;
  ptAffineRelation_get_elems (ptin, &period_n, &phasis, &period_d);
  if (phasis <= 0)
    phasis = -phasis;
  if (phasis == period_d)
    {
      AffineParam_div_and_mod (period_n, period_d, &q, &r);
      if (r == 0)
	result = pk_true;
    }
  else if (period_n == period_d)
    {
      AffineParam_div_and_mod (period_n, phasis, &q, &r);
      if ((r == 0) && (q == 2))
	result = pk_true;
    }
  if (result)
    *ptout = mkAffineRelation (1, 0, period_d);
  else
    {
      AffineParam_div_and_mod (period_d, period_n, &q, &r);
      if (r == 0)
	{
	  AffineParam_div_and_mod (phasis, period_n, &q, &r);
	  if (r == 0)
	    {
	      result = pk_true;
	      phasisout = AffineParam_max (0, -phasis);
	      *ptout = mkAffineRelation (1, phasisout, period_n);
	    }
	}
    }
  return result;
}

/**
   Operation not yet implemented.

   @return always "false".

*/
static boolean
ptAffineRelation_inf_is_affine (ptAffineRelation ptin, ptAffineRelation * ptout)
{
  /*
     ---> True si ?PTIN = (n, phi, d) est une relation affine dont l'intersection des horloges
     affines operandes est une horloge en relation affine avec chaque operande.
     Dans ce cas, !PTOUT contient la relation affine (1, phi, d) entre l'embranchement commun des 
     horloges operandes et l'horloge intersection (on considere la decomposition 
     (n, phi, d) = (n, max(0, -phi), 1) * (1, max(0, phi), d)).
   */
  /*il faut faire un ptout = mkAffineRelation(...) */
  //  fprintf(stderr, "We are in function ptAffineRelation_inf_is_affine now!!!\n");
  //  fprintf(stderr, "Function ptAffineRelation_inf_is_affine has not been implemented\n");
  return pk_false;
}
