/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   Setting the number of a clock (used for sorts).
*/
extern void
ptinsth_set_num (ptinsth vh, int v)
{
  vh->setnum (v);
}

/**
   @return the number of a clock.
*/
extern int
ptinsth_get_num (ptinsth vh)
{
  return vh->getnum ();
}

/**
   @return the pragmas assigned to a clock.
*/
extern ptlpragmas
ptinsth_get_pragmas (ptinsth vh)
{
  return vh->getPragmas ();
}

/**
   Adding a pragma to a clock.
*/
extern void
ptinsth_addPragma (ptinsth vh, ptpragma aPragma)
{
  vh->addPragma (aPragma);
}

/**
   Adding a list of pragmas to a clock.
*/
extern void
ptinsth_addPragmas (ptinsth vh, ptlpragmas pragmas)
{
  vh->addPragmas (pragmas);
}

/** 
    Adding a mark to a clock.
*/
extern void
ptinsth_addMark (ptinsth vh, int val)
{
  vh->addMark (val);
}

/** 
    Removing a mark assigned to clock.
*/
extern void
ptinsth_removeMark (ptinsth vh, int val)
{
  vh->removeMark (val);
}

/** 
    @return true if a mark (val) is assigned to a clock (vh).
*/
extern boolean
ptinsth_satisfies_marked (ptinsth vh, int val)
{
  return vh->satisfiesMarked (val);
}

/** 
    @return 
    - pk_true if the num-th attribute is assigned to a clock (vexp), in this case,
    valAttribute (out) is the value of this attribute.
    - false otherwise.
*/
extern boolean
ptinsth_getAttribute (ptinsth vexp, int num, attributeValue * valAttribute)
{
  return vexp->getAttribute (num, valAttribute);
}

/** 
    It assigns to a clock (vexp) the num-th attribute with vattr as value.
*/
extern void
ptinsth_addAttribute (ptinsth vexp, int num, attributeValue vattr)
{
  vexp->addAttribute (num, vattr);
}

/**
   Removing the  num-th attribute assigned to clock.
*/
extern void
ptinsth_removeAttribute (ptinsth vexp, int num)
{
  vexp->removeAttribute (num);
}
