/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
  It creates and initialises a descriptor giving the access to the graph associated to a clock.
*/
extern ptGraphInfos
GraphInfos_create ()
{
  ptGraphInfos vig;

  vig = new GraphInfos ();
  GraphInfos_init (vig);
  return vig;
}

/**
   It marks a clock (vh) as an element of a composed node (vnode).
*/
extern void
ptinsth_set_inComposedNode (ptinsth vh, ptINode vnode)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vh);
  vig->ptINodebc = vnode;
}

/**
   @return the composed node in which the clock (vclk) is an element.
*/
extern boolean
ptinsth_get_inComposedNode (ptinsth vclk, ptINode * cNode)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vclk);
  *cNode = vig->ptINodebc;
  return ((*cNode) != NILINODE);
}

/**
   It assigns the list of the synchronous signals (vlist) to a clock (vclk). 
*/
extern void
ptinsth_set_lsig (ptinsth vclk, ptlinstances vlist)
{
  ptGraphInfos vig;
  vig = ptinsth_get_GraphInfos (vclk);
  vig->lsig = vlist;
}

extern void
ptinsth_remove_sig (ptinsth vh, ptexpcal sig)
{
  ptinsth_set_lsig (vh, ptlinstances_remove_exp (ptinsth_get_lsig (vh), sig));
}

/**
    @return the list of the synchronous signals (vlist) assigned to the clock (vclk). 
*/
extern ptlinstances
ptinsth_get_lsig (ptinsth vclk)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vclk);
  return vig->lsig;
}

/**
   It assigns the list of the synchronous signals (vlist) to a clock (vclk) removed from the
   graph, but kept for tracability.
*/
extern void
ptinsth_set_lsigkept (ptinsth vclk, ptlinstances vlist)
{
  ptGraphInfos vig;
  vig = ptinsth_get_GraphInfos (vclk);
  vig->lsigkept = vlist;
}

/**
    @return the list of the synchronous signals (vlist) assigned to the clock (vclk), removed from the
    graph, but kept for tracability. 
*/
extern ptlinstances
ptinsth_get_lsigkept (ptinsth vclk)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vclk);
  return vig->lsigkept;
}

/**
   It initialises the graph associated to a clock (vform) from the informations of another one (vh).
*/
extern void
ptinsth_reset_gfd_init (ptinsth vform, ptinsth vh)
{
  ptGraphInfos vigvform, vigvh;
  ptgfd gfdvh;

  vigvform = ptinsth_get_GraphInfos (vform);
  vigvh = ptinsth_get_GraphInfos (vh);
  gfdvh = vigvh->gfd;
  vigvform->gfd = ptgfd_copy (vigvform->gfd, gfdvh);
  vigvform->lsig = vigvh->lsig;
  vigvform->lsigkept = vigvh->lsigkept;
}

/**
   It merges in vhnew (side effect) the graphs of the two operand cloks (vhold, vhnew).
*/
extern void
ptinsth_fusion_infosGraph (ptinsth vhold, ptinsth vhnew)
{
  ptlINode lnOld, lnNew;
  ptINode onode, nnode;

  lnOld = ptinsth_get_lsucci (vhold);
  lnNew = ptinsth_get_lsucci (vhnew);
  lnNew = ptlINode_fusion (lnOld, lnNew);
  ptinsth_set_lsucci (vhnew, lnNew);
  ptinsth_fusion_gfd (vhold, vhnew);
  ptinsth_set_lsig (vhnew, ptlinstances_sortedfusion (ptinsth_get_lsig (vhold), ptinsth_get_lsig (vhnew)));
  ptinsth_set_lsigkept (vhnew, ptlinstances_sortedfusion (ptinsth_get_lsigkept (vhold), ptinsth_get_lsigkept (vhnew)));
  ptinsth_get_inComposedNode (vhold, &onode);
  ptinsth_get_inComposedNode (vhnew, &nnode);
  //  if (nnode == NILINODE) ptinsth_set_inComposedNode(vhnew, onode); 
}

/**
   @return the list of the nodes that reference a clock (vclk).
*/
extern ptlsucc
ptinsth_get_lsucci (ptinsth vclk)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vclk);
  return vig->lsucci;
}

/**
   It resets the list of the nodes that reference a clock (vclk).
*/
extern void
ptinsth_set_lsucci_init (ptinsth vclk)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vclk);
  vig->lsucci = NILLINODE;
}

/**
  It assigns the list of the nodes (vln) that reference a clock (vclk).
*/
extern void
ptinsth_set_lsucci (ptinsth vclk, ptlsucc vln)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (vclk);
  vig->lsucci = vln;
}

/**
   @return the sub-graph assigned to a clock.
*/
extern ptgfd
ptinsth_get_gfd (ptinsth clk)
{
  ptGraphInfos vig;

  vig = ptinsth_get_GraphInfos (clk);
  return vig->gfd;
}

/**
   It associates a sub-graph (vgfd) to a clock (vh).
*/
extern void
ptinsth_set_gfd (ptinsth vh, ptgfd vgfd)
{
  ptGraphInfos vig;
  /* Non referencees */
  vig = ptinsth_get_GraphInfos (vh);
  vig->gfd = vgfd;
}

/**
   @return the interface of the subtree of the operand clock (clk).
*/
extern ptinterface
//ptinsth_getInterfaceSubHierarchy(ptinsth clk)
ptinsth_get_interface (ptinsth clk)
{
  ptGraphInfos vig;
  vig = ptinsth_get_GraphInfos (clk);
  return vig->interface;
}

/** ---- Sub-graph assigned to a clock ---.
    It returns the interface (profile) of the subgraph with ?clk as root.
*/
extern ptinterface
ptinsth_getset_interface (ptinsth clk)
{
  ptinterface vinter;
  vinter = ptinsth_get_interface (clk);
  if (vinter == NILINTERFACE)
    {
      vinter = ptinterface_create ();
      ptinsth_set_interface (clk, vinter);
    }
  return vinter;
}

/**
   It assigns the interface (vinterface) of the graph with clk as root.
*/
extern void
ptinsth_set_interface (ptinsth clk, ptinterface vinterface)
{
  ptGraphInfos vig;
  vig = ptinsth_get_GraphInfos (clk);
  vig->interface = vinterface;
}

/**
   @return the nodes assigned to a clock (clk).
*/
extern ptlINode
ptinsth_get_nodes (ptinsth clk)
{
  return ptgfd_get_nodes (ptinsth_get_gfd (clk));
}

/**
   @return the clocks assigned to a clock (clk).
*/
extern ptlinsth
ptinsth_get_clocks (ptinsth clk)
{
  return ptgfd_get_clocks (ptinsth_get_gfd (clk));
}

/**
   It resets the interface of the graph assigned to a clock(clk).
*/
extern void
ptinsth_reset_gfd_interface (ptinsth clk)
{
  ptgfd_reset_interface (ptinsth_get_gfd (clk));
}

/**
   @return the graph associated with a clock (vh).
*/
extern ptGraphInfos
ptinsth_get_GraphInfos (ptinsth vh)
{
  return vh->vGraphInfos;
}

/**
   It set the graph (vg) associated with a clock (vh).
*/
extern void
ptinsth_set_GraphInfos (ptinsth vh, ptGraphInfos vg)
{
  vh->vGraphInfos = vg;
}
