/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** print a resume of the formula vform 
    return false if vform is NILINSTH
 */
extern void
ptinsth_printFormula_number (insthFormula anInsthFormula)
{
  if (anInsthFormula == NILINSTH)
    printf ("#");
  else
    printf ("%i", anInsthFormula->getnum ());
}

/** print a resume of the formula vform 
    return false if vform is NILINSTH
 */
extern void
ptinsth_printBinaryFormla (insthFormula anInsthFormula, const char *stringOp)
{
  printf ("(");
  if (anInsthFormula == NILINSTH)
    printf ("#");
  else
    {
      ptinsth_printFormula_number (INSTHFORM_GETBINARYARG1 (anInsthFormula));
      printf ("%s", stringOp);
      ptinsth_printFormula_number (INSTHFORM_GETBINARYARG2 (anInsthFormula));
    }
  printf (")");
}

/** print a resume of the formula vform 
    return false if vform is NILINSTH
 */
extern boolean
ptinsth_DEBUG_printFormula (insthFormula anInsthFormula)
{
  printf ("[");
  if (anInsthFormula == NILINSTH)
    {
      printf ("#]");
      return pk_false;
    }
  if (INSTHFORM_ISAREFERENCE (anInsthFormula))
    printf ("@");
  if (INSTHFORM_ISHEADER (anInsthFormula))
    printf ("^");
  printf ("%i:", anInsthFormula->getnum ());
  if (INSTHCLOCK_ISMETAROOT (anInsthFormula))
    printf ("Â°");
  else if (INSTHFORM_ISFREEFORMULA (anInsthFormula))
    printf ("?");
  else if (INSTHFORM_ISSUB (anInsthFormula))
    {
      printf ("wT");
    }
  else if (INSTHFORM_ISNOSUB (anInsthFormula))
    {
      printf ("wF");
    }
  else if (INSTHFORM_ISNOT (anInsthFormula))
    {
      printf ("N");
    }
  else if (INSTHFORM_ISCOMP (anInsthFormula))
    {
      printf ("C");
    }
  else if (INSTHFORM_ISINFER (anInsthFormula))
    ptinsth_printBinaryFormla (anInsthFormula, "*");
  else if (INSTHFORM_ISSUPER (anInsthFormula))
    ptinsth_printBinaryFormla (anInsthFormula, "+");
  if (INSTHFORM_ISSATISFIED (anInsthFormula))
    {
      printf (",=");
    }
  printf (":");
  printf ("%li", (long) anInsthFormula);
  printf ("]");
  return pk_true;
}

/** trace the content of the anInsthFormula clock class
    return pk_true is the current class is well formed
 */
extern boolean
ptinsth_DEBUG_printIndentedClock (insthFormula anInsthFormula, int indent)
{
  insthFormula nextFormula;
  insthFormula headerFormula;
  int i;

  for (i = 1; i <= indent; i++)
    printf (" ");
  if (anInsthFormula == NILINSTH)
    {
      printf ("[~]\n");
      fflush (stdout);
      return pk_false;
    }
  if (INSTHFORM_ISAREFERENCE (anInsthFormula))
    {
      nextFormula = ptinsth_DEBUG_printReferenceFormulaUntiClassFormula (anInsthFormula);
      if (nextFormula == NILINSTH)
	return pk_false;
    }
  else
    nextFormula = anInsthFormula;
  if (INSTHFORM_ISHEADER (nextFormula))
    headerFormula = nextFormula;
  else
    {
      anInsthFormula = nextFormula;
      headerFormula = ptinsth_DEBUG_printFormulaUntilHeader (anInsthFormula, "-");
      if (headerFormula == NILINSTH)
	return pk_false;
      if (headerFormula == anInsthFormula)
	{
	  printf ("[CYCLIC]\n");
	  fflush (stdout);
	  return pk_false;
	}
    }
  printf ("{");
  ptinsth_DEBUG_printFormula (headerFormula);
  printf ("-->");
  nextFormula = ptinsth_DEBUG_printFormulaUntilHeader (INSTHFORM_GETSUCC (headerFormula), "-->");
  if (nextFormula == NILINSTH)
    {
      printf ("[~]");
      return pk_false;
    }
  else if (nextFormula == headerFormula)
    printf ("[.]");
  else
    ptinsth_DEBUG_printFormula (nextFormula);
  printf ("}\n");
  fflush (stdout);
  return pk_true;
}

/** print a formula resume and its header formula resume
    @param anInsthFormula is a reference or a member of a clock
   @return false: anInsthFormula is a header formula; anInsthFormula is printed
   @return true: anInsthFormula is not a header formula; anInsthFormula followed by its header formula are printed
 */
extern boolean
ptinsth_DEBUG_printClockResume (insthFormula anInsthFormula)
{
  if (anInsthFormula == NILINSTH)
    {
      ptinsth_DEBUG_printFormula (anInsthFormula);
      return pk_false;
    }
  if (INSTHFORM_ISHEADER (anInsthFormula))
    ptinsth_DEBUG_printFormula (anInsthFormula);
  else
    {
      printf ("{");
      ptinsth_DEBUG_printFormula (anInsthFormula);
      printf ("-->");
      ptinsth_DEBUG_printFormula (ptinsth_get_clhead (anInsthFormula));
      printf ("}");
    }
  return pk_true;
}

extern void
ptinsth_DEBUG_traceClock (char *methodName, insthFormula anInsthFormula)
{
  printf ("DEBUG %s TRACE CLASS content is\n", methodName);
  (void) ptinsth_DEBUG_printIndentedClock (anInsthFormula, 5);
}

/** Construtor of a clock node.
    @param formulaOperator (in) : the associated formula type (^+, ^*, ..).
*/
extern insthClock
ptinsth_newClock (pkClockType formulaOperator)
{
  insthClock vinsth;

  vinsth = new insth ();
  setNumClock (vinsth);
  vinsth->luse = NILLINSTANCES;
  vinsth->vGraphInfos = GraphInfos_create ();
  if (formulaOperator == predef)
    vinsth->father = vinsth;
  else
    vinsth->father = NILINSTH;
  vinsth->root = NILINSTH;
  vinsth->vdepth = 0;
  vinsth->hocompl = NILINSTH;
  vinsth->identh = NILIDENT;
  vinsth->eventSignal = NILIDENT;
  vinsth->lsucch = NILLINSTH;
  vinsth->children = NILLINSTH;
  vinsth->isAnEquivalentForm = pk_false;
  vinsth->isAnAssumption = pk_false;
  ptinsth_set_clform (vinsth, formulaOperator, pk_true);
  vinsth->affr = NILAFFINERELATION;
  vinsth->setwhoami (clockObject);
  return vinsth;
}

/** @return "true" if a clock (pt) is a free one (variable without associated formula), "false" otherwise.  */
extern boolean
ptinsth_satisfies_vari (ptinsth pt)
{
  if (ptinsth_satisfies_opvari (pt))
    return ptinsth_satisfies_clsingleton (pt);
  return pk_false;
}

/**
   @return "true" if a clock (pt) is a fictitious root one
   (ie its children are the real roots of the hierarchy), "false" otherwise.
*/
extern boolean
ptinsth_isFictitiousRootFormula (ptinsth pt)
{
  return (ptinsth_get_operator (pt) == troot);
}

/**
   It merges  two clasees (vh1 and vh2) in vh1.
*/
extern void
ptinsth_fusionclass (ptinsth vh1, ptinsth vh2)
{
  ptinsth vh4;

  vh4 = vh1->clsucc;
  if (ptinsth_satisfies_opvari (vh2))
    {
      if (vh2->clsucc != vh2)
	{
	  vh1->clsucc = vh2->clsucc;
	  setpredvh2 (vh2, vh4);
	};
      vh2->clsucc = vh1;
    }
  else
    {
      vh1->clsucc = vh2;
      setpredvh2 (vh2, vh4);
    }
  vh2->clhead = pk_false;
}

/** @return the next fomula of the class pointed by vh.
    When constraints is pk_true, it takes into account only the synchronisation constraints,
    otherwise all formula (equivalent formula, assumption, constraint) is returned.  
*/
extern ptinsth
ptinsth_get_clsucc (ptinsth vh, boolean constraints)
{
  ptinsth vh1;

  vh1 = vh->clsucc;
  if (!constraints)
    return vh1;
  while (!vh1->clhead)
    {
      if (!vh1->isAnEquivalentForm && !vh1->isAnAssumption)
	return vh1;
      vh1 = vh1->clsucc;
    }
  return vh1;
}

/** @return "true" when it exists an assumption (vassert) in a class (vh).
    It searches from vh and stops if the representative formula is found.
*/
extern boolean
ptinsth_get_nextassertion (ptinsth vh, ptinsth * vassert)
{
  ptinsth vh1;

  vh1 = vh->clsucc;
  while (!vh1->clhead)
    {
      if (vh1->isAnAssumption)
	{
	  *vassert = vh1;
	  return pk_true;
	}
      vh1 = vh1->clsucc;
    }
  return pk_false;
}

/** @return the representative formula of a class (vh). */
extern ptinsth
ptinsth_get_clhead (ptinsth vh)
{
  while (!vh->clhead)
    vh = vh->clsucc;
  return vh;
}

/** @return "true" if a class (vform) is reduced to one element, "false" otherwise. */
extern boolean
ptinsth_satisfies_clsingleton (ptinsth vform)
{
  vform = ptinsth_get_clhead (vform);
  return (ptinsth_get_clsucc (vform, pk_true) == vform);
}

/** @return "true" if a formula (vh) is the representative formula of its class, "false" otherwise. */
extern boolean
ptinsth_satisfies_clhead (ptinsth vh)
{
  return vh->clhead;
}

/** @return "true" if a formula (vh) is normalized, "false" otherwise. */
extern boolean
ptinsth_satisfies_normalized (ptinsth vh)
{
  vh = ptinsth_get_clhead (vh);
  return vh->normform;
}

/** It defines a formula (vh) as a normalized one. */
extern void
ptinsth_belongs_normalized (ptinsth vh)
{
  vh->normform = pk_true;
}

/**
  It inserts the formula of a class (vfom) in another one (vhcl).
  The representative formula of the result is the representative formula of vhcl.
*/
extern void
ptinsth_insert_form (ptinsth vhcl, ptinsth vform)
{
  vform->clhead = pk_false;
  vform->clsucc = vhcl->clsucc;
  vhcl->clsucc = vform;
}

/**
   It removes a formula (vform) if it is not the representative formula of its class.
   @return the representative formula of the class of the formula.
*/
extern ptinsth
ptinsth_remove_form_belongs_class (ptinsth vform)
{
  ptinsth result;
  ptinsth vh;

  if (vform->clhead)
    return vform;
  vh = ptinsth_get_clhead (vform);
  result = vh;
  while (vh->clsucc != vform)
    vh = vh->clsucc;
  vh->clsucc = vform->clsucc;
  vform->clsucc = vh;
  return result;
}

/**
   It removes a formula (vform) if it is a formula of a clock (vhcl) and if it is not the representative formula of this class.
   When it is removed, the formula (vform) becomes a new classes of formula (reduced to vform).
   @return "true" if the operation succeeds, "false" otherwise (it is already a representative formula).
*/
extern boolean
ptinsth_remove_form (ptinsth vhcl, ptinsth vform)
{
  ptinsth vh;
  boolean found, stop;

  if (vform->clhead)
    return pk_false;

  vh = ptinsth_get_clhead (vhcl);
  found = pk_false;
  stop = pk_false;
  while (!found && !stop)
    if (vh->clsucc == vform)
      found = pk_true;
    else if (vh->clsucc->clhead)
      stop = pk_true;
    else
      vh = vh->clsucc;
  if (found)
    {
      vh->clsucc = vform->clsucc;
      vform->clsucc = vform;
      vform->clhead = pk_true;
      vform->vGraphInfos = GraphInfos_create ();
    }
  return pk_true;
}

/**
   The formula (vform) becomes a free variable.
   The formula (vform) is removed from its class if it is not the representative formula of its class.
   @return the representative formula of the class after the operation.
*/
extern ptinsth
ptinsth_belongs_vari (ptinsth vform)
{
  ptinsth result;

  result = ptinsth_remove_form_belongs_class (vform);
  vform->ophorl = vari;
  return result;
}

/**
   The formula (vh) looses the status of representative element. If such a formula is a free variable
   not reduced to one element, it is removed form the class.
*/
extern void
ptinsth_belongs_notclhead (ptinsth vh)
{
  ptinsth_set_clhead (vh, pk_false);
  if (ptinsth_satisfies_opvari (vh))
    ptinsth_belongs_vari (vh);
}

/** @return the operator (^+, ^*,...) of a formula (vh). */
extern pkClockType
ptinsth_get_operator (ptinsth vh)
{
  return vh->ophorl;
}

/** It sets the operator (op) of a formula (clk). */
extern void
ptinsth_set_operator (ptinsth clk, pkClockType op)
{
  clk->ophorl = op;
}

/**
   @param vh (in) : vh is a upper-bound or lower-bound formula (assumed).
   @param vharg1 (out) : first operand of vh.
   @param vharg2 (out) : second operand of vh.
*/
extern void
ptinsth_get_2args (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  *vharg1 = ptinsth_get_left_arg (vh);
  *vharg2 = ptinsth_get_right_arg (vh);
}

/**
   @return 
     - "true" if a formula (vh) is a lower-bound formula: vh = vharg1 ^* vharg2;
     - false otherwise; in this case, vh is returned in vharg1 parameter.
*/
extern boolean
ptinsth_get_infargs (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  if (vh->ophorl == infer)
    {
      ptinsth_get_2args (vh, vharg1, vharg2);
      return pk_true;
    }
  *vharg1 = vh;
  return pk_false;
}

/**
   @return 
       - "true" if a formula (vh) is a upper-bound formula : vh = vharg1 ^* vharg2;
       - false otherwise; in this case, vh is returned in vharg1 parameter.
*/
extern boolean
ptinsth_get_supargs (ptinsth vh, ptinsth * vharg1, ptinsth * vharg2)
{
  if (vh->ophorl == super)
    {
      ptinsth_get_2args (vh, vharg1, vharg2);
      return pk_true;
    }
  *vharg1 = vh;
  return pk_false;
}

/**
   @return 
     - "true" if a formula (vh) is a complementary formula (vh = vhother ^- vharghnot).
       The implentation is vh = vhother ^* hnot(vharghnot).
     - false otherwise; 
*/
extern boolean
ptinsth_get_minusargs (ptinsth pt, ptinsth * vharghnot, ptinsth * vhother)
{
  return locisinfhnot (pt, vharghnot, vhother);
}

/** 
   @return true
     - pk_true if a formula (pt) is a not(vh) formula (used to represent the minus operator) or a hcomp(vh)
     (used to represent the minus operator relative to its father in a hierarchy)
     - false otherwise, then vh is equal pt.
*/
extern boolean
ptinsth_get_1arg (ptinsth pt, ptinsth * vh)
{
  if ((pt->ophorl == hnot) || (pt->ophorl == hcomp))
    {
      *vh = pt->Operand.op1;
      return pk_true;
    }
  *vh = pt;
  return pk_false;
}

/**
   It assigns the operand (vharg) to a not-formula (vform).
   The references are managed.
*/
extern void
ptinsth_set_1arg (ptinsth vform, ptinsth vharg)
{
  setNotArg (vform, vharg);
  ptinsth_insert_dep (vform, vharg);
}

/** Assumed : vh is a binary formula.
    @return the other operand than a given operand (vhs1).
*/
extern ptinsth
ptinsth_get_otherarg (ptinsth vh, ptinsth vhs1)
{
  if (vh->Operand.binOp.op21 == vhs1)
    return vh->Operand.binOp.op22;
  return vh->Operand.binOp.op21;
}

/**
  @return the condition associated with the undersampling formula (vh). 
  Assumed : vh is a undersampling.
*/
extern ptcondition
ptinsth_get_argsub (ptinsth vh)
{
  return vh->Operand.opsub1;
}

/**
   It assigns a condition (vcond) to an undersampling formula (vh). 
   Assumed : vh is a undersampling.
*/
extern void
ptinsth_set_argsub (ptinsth vh, ptcondition vcond)
{
  vh->Operand.opsub1 = vcond;
}

/**
   Assumed : pth is an undersampling formula.
   It replaces the clock of the condition by another one (argh).
   The references are managed.*
*/
extern void
ptinsth_replace_argsub (ptinsth pth, ptinsth argh)
{
  ptinsth vhused, vhnot;
  ptcondition vcond;
  boolean b;

  vcond = ptinsth_get_argsub (pth);
  if (vcond->set_condsub (argh, &vhnot, &vhused))
    {
      vhused->lsucch = ptlinsth_remove_elem (vhused->lsucch, pth, &b);
      vhused->lsucch = ptlinsth_remove_elem (vhused->lsucch, vhnot, &b);
      argh->lsucch = ptlinsth_insert_elem (argh->lsucch, pth);
      argh->lsucch = ptlinsth_insert_elem (argh->lsucch, vhnot);
    }
}

/** The operand formula (vh) becomes the representative formula of the class when ishead is "true"
    otherwise the formula (vh) is not the representative one.
*/
extern void
ptinsth_set_clhead (ptinsth vh, boolean ishead)
{
  vh->clhead = ishead;
}

/**
   It removes a formula (vform) from its class and updates the references.
   @param   vform (in): a binary formula (assumed).
   @param becomesvari (out): set to "true" value if the new class is reduced to a free variable, "false" otherwise.
   @return the representative formula of the new class.
*/
extern ptinsth
ptinsth_remove_binaryformula (ptinsth vform, boolean * becomesvari)
{
  ptinsth result;

  ptinsth_remove_binaryformula_link (vform);
  result = ptinsth_belongs_vari (vform);
  *becomesvari = ptinsth_satisfies_clsingleton (vform);
  return result;
}

/**
  It removes the references to an assumed binary formula (vform) for its operands.
*/
extern void
ptinsth_remove_binaryformula_link (ptinsth vform)
{
  ptinsth vh1, vh2;
  boolean b;
  ptinsth_get_2args (vform, &vh1, &vh2);
  vh1->lsucch = ptlinsth_remove_elem (vh1->lsucch, vform, &b);
  vh2->lsucch = ptlinsth_remove_elem (vh2->lsucch, vform, &b);
}

/**
   when b is "true", it specifies that a formula (vh) is an equivalent one.
*/
extern void
ptinsth_set_isAnEquivalentForm (ptinsth vh, boolean b)
{
  vh->isAnEquivalentForm = b;
}

/**
   @return "true" when a formula (vh) is declared as an equivalent one, "false" otherwise.
*/
extern boolean
ptinsth_IsAnEquivalentForm (ptinsth vh)
{
  return vh->isAnEquivalentForm;
}

/**
   When b is pk_true, it specifies that a formula (vh) is an assumption.
*/
extern void
ptinsth_set_AnAssumption (ptinsth vh, boolean b)
{
  vh->isAnAssumption = b;
}

/**
   @return "true" when a formula (vh) is an  assumption, "false" otherwise.
*/
extern boolean
ptinsth_IsAnAssumption (ptinsth vh)
{
  return vh->isAnAssumption;
}

/**
   It removes a formula (vform) from its class and updates the references.
   @param vform (in): an undersampling formula (assumed).
   @param vcond (in): the condition associated with vform.
   @param becomesvari (out): set to "true" value if the new class is reduced to a free variable, "false" otherwise.
   @return the representative formula of the new class.
*/
extern ptinsth
ptinsth_remove_sub (ptinsth vform, ptcondition vcond, boolean * becomesvari)
{
  ptinsth result;

  ptinsth_remove_sub_link (vform, vcond);
  result = ptinsth_belongs_vari (vform);
  *becomesvari = ptinsth_satisfies_clsingleton (vform);
  ptinsth_set_affinerelation (vform, NILAFFINERELATION);
  return result;
}

/**
   It removes the references to an assumed undersampling formula (vform) based on the vcond condition.
*/
extern void
ptinsth_remove_sub_link (ptinsth vform, ptcondition vcond)
{
  ptinsth vhused;
  boolean b;

  vhused = vcond->get_clkcond ();
  vhused->lsucch = ptlinsth_remove_elem (vhused->lsucch, vform, &b);
}

/**
   It removes a formula (vform) from its class and updates the references.
   @param vform (in): a not-formula (assumed).
   @param becomesvari (out): set to "true" value if the new class is reduced to a free variable, "false" otherwise.
   @return the representative formula of the new class.
*/
extern ptinsth
ptinsth_remove_not (ptinsth vform, boolean * becomesvari)
{
  ptinsth res;
  ptinsth vhused;
  boolean b;

  ptinsth_get_1arg (vform, &vhused);
  vhused->lsucch = ptlinsth_remove_elem (vhused->lsucch, vform, &b);
  res = ptinsth_belongs_vari (vform);
  *becomesvari = ptinsth_satisfies_clsingleton (vform);
  return res;
}

/**
   @return the list of the signals referenced in the expression associated with an undersampling formula (vh).
   An empty list is returned when the formula is not an undersampling one.
   \See {ptinsth_set_condition}
 */
extern ptlinstances
ptinsth_get_SignalOfcondition (ptinsth vh)
{
  pkClockType vop;

  vop = ptinsth_get_operator (vh);
  if ((vop == sub) || (vop == nosub))
    return ptinsth_get_argsub (vh)->get_lsign ();
  return NILLINSTANCES;
}

/**
   @return the expression associated with the condition of  an undersampling formula (vh). An undefined expression is returned 
   when the formula is not an undersampling one.
*/
extern ptexpcal
ptinsth_get_condition (ptinsth vh)
{
  pkClockType vop;

  vop = ptinsth_get_operator (vh);
  if ((vop == sub) || (vop == nosub))
    return ptinsth_get_argsub (vh)->get_condsub ();
  return NILEXPCAL;
}

/**
   It assigns an expression (vexp) and the list of the signals referenced in this expression to an undersampling formula (vh).
   No effects if the vformula is not an undersampling one.   
*/
extern void
ptinsth_set_condition (ptinsth vh, ptexpcal vexp, ptlinstances ls)
{
  pkClockType vop;

  vop = ptinsth_get_operator (vh);
  if ((vop == sub) || (vop == nosub))
    ptinsth_get_argsub (vh)->linkExpression (vexp, ls);
}

/** 
    It assigns the clock of the condition (argh) to an undersampling formula (vh).
    No effects if the vformula is not an undersampling one. 
*/
extern void
ptinsth_set_clkCondition (ptinsth vh, ptinsth argh)
{
  pkClockType vop;

  vop = ptinsth_get_operator (vh);
  if ((vop == sub) || (vop == nosub))
    ptinsth_get_argsub (vh)->set_clkcond (argh);
}

/**
   @return the clock of the condition assigned to an undersampling formula (vh).
   If the vformula is not an undersampling one, it returns vh.
*/
extern ptinsth
ptinsth_get_clkCondition (ptinsth vh)
{
  pkClockType vop;

  vop = ptinsth_get_operator (vh);
  if ((vop == sub) || (vop == nosub))
    vh = ptinsth_get_argsub (vh)->get_clkcond ();
  return ptinsth_get_clhead (vh);	// (assumption managing)
  // return vh ; 
}

/**
   @return "true" if a formula is an undersampling one, "false" otherwise.
*/
extern boolean
ptinsth_satisfies_extraction (ptinsth vharg)
{
  pkClockType ope;
  ope = ptinsth_get_operator (vharg);
  return ((boolean) ((ope == sub) || (ope == nosub)));
}

/**
  @return the class in which the representative formula is a given formula (vform).
  @param vnewhead (out): is "true" if the representative formula has been changed, "false" otherwise.
*/
extern ptinsth
ptinsth_resetclass (ptinsth vform, boolean * vnewhead)
{
  ptinsth vh;

  if (ptinsth_satisfies_clhead (vform))
    *vnewhead = pk_false;
  else
    {
      vh = ptinsth_get_clhead (vform);

      ptinsth_set_clhead (vform, pk_true);
      ptinsth_move_references (vh, vform);

      ptinsth_reset_gfd_init (vform, vh);
      ptinsth_belongs_notclhead (vh);
      vform->father = vform;
      vform->root = vform;
      vform->vdepth = 0;
      *vnewhead = pk_true;
    }
  return vform;
}

/** It assigns a name(id) to a formula (vh). */
extern void
ptinsth_set_name (ptinsth vh, ptident id)
{
  vh->identh = id;
}

/**
   @return true if the name (nam) of a formula (vh) is known, false otherwise.
*/
extern boolean
ptinsth_get_name (ptinsth vh, ptident * id)
{
  *id = vh->identh;
  return ((*id) != NILIDENT);
}

/** It specifies that vhcompl is the complementary clock of vharg in their father clock.*/
extern void
ptinsth_set_compl (ptinsth vharg, ptinsth vhcompl)
{
  vharg->hocompl = vhcompl;
}

/**
    Formula.
    @return
      - pk_true if vharg is an extraction, then vhcompl (out) is its complement (remind that: [c] ^+ [not c] = ^c).
      - false otherwise
*/
extern boolean
ptinsth_get_complExtraction (ptinsth vharg, ptinsth * vhcompl)
{
  pkClockType ope;

  ope = ptinsth_get_operator (vharg);
  if (ope == sub)
    {
      ptinsth_set_compl (vharg, ptinsth_get_argsub (vharg)->get_clknosub ());
      *vhcompl = vharg->hocompl;
      return pk_true;
    }
  if (ope == nosub)
    {
      ptinsth_set_compl (vharg, ptinsth_get_argsub (vharg)->get_clksub ());
      *vhcompl = vharg->hocompl;
      return pk_true;
    }
  return pk_false;
}

/** @return "true" if a clock (pt) is a predefined one, "false" otherwise.
 */
extern boolean
ptinsth_satisfies_predef (ptinsth pt)
{
  return (pt->ophorl == predef);
}

/* -- REFERENCES MANAGING-- */

/**
   @return the list of the expressions that reference a clock (vh).
*/
extern ptlinstances
ptinsth_get_luse (ptinsth vh)
{
  return vh->luse;
}

/**
   It assigns to a clock (vh) the list of the expressions that reference it.
*/

extern ptlinstances
ptinsth_set_luse (ptinsth vh, ptlinstances l)
{
  vh->luse = l;
  return l;
}

/** Set a reference from a clock (vhpred) to an another one (vinsth).
 */
extern void
ptinsth_insert_dep (ptinsth vinsth, ptinsth vhpred)
{
  vhpred->lsucch = ptlinsth_insert_elem (vhpred->lsucch, vinsth);
}

/** It assigns to a clock (vhpred) the list of the clocks (vl) that reference it.  */
extern void
ptinsth_set_lsucch (ptinsth vhpred, ptlinsth vl)
{
  vhpred->lsucch = vl;
}

/** It sets a reference from two clocks (vhpred1,vhpred2)  to another one (vinsth).
 */
extern void
ptinsth_set_2args (ptinsth vinsth, ptinsth vhpred1, ptinsth vhpred2)
{
  vhpred1->lsucch = ptlinsth_insert_elem (vhpred1->lsucch, vinsth);
  vhpred2->lsucch = ptlinsth_insert_elem (vhpred2->lsucch, vinsth);
  setform2 (vinsth, vhpred1, vhpred2);
}

/** @return the list of the clock that reference a clock (vclk). */
extern ptlinsth
ptinsth_get_lsucch (ptinsth vclk)
{
  return vclk->lsucch;
}

/** -- misc operation--
    Assumed : vh is a lower-bound formula with vh1 as operand.
    @return "true" if a formula (vh1) is a not-formula, "false" otherwise.
    When it is a not-formula not(vharg), the reference from vharg to vh is removed, and the reference vharg to vh1 is set.
    \See ptinsth_modify_dep_hnot
*/
extern boolean
ptinsth_switch_referenceshnot (ptinsth vh, ptinsth vh1)
{
  ptinsth vharg;
  boolean b;

  if (vh1->ophorl != hnot)
    return pk_false;
  ptinsth_get_1arg (vh1, &vharg);
  vharg->lsucch = ptlinsth_remove_elem (vharg->lsucch, vh, &b);
  if (b)
    {
      if (!ptlinsth_isin_elem (vharg->lsucch, vh1))
	vharg->lsucch = ptlinsth_insert_elem (vharg->lsucch, vh1);
      if (!ptlinsth_isin_elem (vh1->lsucch, vh))
	vh1->lsucch = ptlinsth_insert_elem (vh1->lsucch, vh);
    }
  return pk_true;
}

/** -- misc operation--
  Assumed : vh is a not-formula (ie vh = not(vhused))
  It removes the reference vhused to vh and adds the reference vhused to vh_i for all lower-bound formula vh_i 
  such that vh_i = ^* (X, vh).
  \See ptinsth_switch_referenceshnot
*/
extern void
ptinsth_modify_dep_hnot (ptinsth vh)
{
  ptinsth vhused, vh1;
  ptlinsth luse;
  boolean bdummy;

  ptinsth_get_1arg (vh, &vhused);
  vhused->lsucch = ptlinsth_remove_elem (vhused->lsucch, vh, &bdummy);
  vhused = ptinsth_get_clhead (vhused);
  setNotArg (vh, vhused);
  luse = vh->lsucch;
  while (ptlinsth_get_next (&luse, &vh1))
    if (!ptlinsth_isin_elem (vhused->lsucch, vh1))
      vhused->lsucch = ptlinsth_insert_elem (vhused->lsucch, vh1);
}

/**  It removes the reference from clk_from to clk_to. */
extern void
ptinsth_remove_ref (ptinsth clk_from, ptinsth clk_to)
{
  boolean bdummy;
  clk_from->lsucch = ptlinsth_remove_elem (clk_from->lsucch, clk_to, &bdummy);
}

/** 
    It removes the references from the operands of a formula (vh) to this formula.
    @return "true" if vh is an undersampling formula, "false" otherwise.
*/
extern boolean
ptinsth_remove_linksTo (ptinsth vh)
{
  ptinsth vharg1, vharg2;

  switch (ptinsth_get_operator (vh))
    {
    case super:
      if (ptinsth_get_supargs (vh, &vharg1, &vharg2))
	{
	  ptinsth_remove_ref (vharg1, vh);
	  ptinsth_remove_ref (vharg2, vh);
	};
      break;
    case infer:
      if (ptinsth_get_minusargs (vh, &vharg1, &vharg2))
	ptinsth_remove_ref (vharg1, vh);
      if (ptinsth_get_infargs (vh, &vharg1, &vharg2))
	{
	  ptinsth_remove_ref (vharg1, vh);
	  ptinsth_remove_ref (vharg2, vh);
	};
      break;
    case sub:
    case nosub:
      vharg1 = ptinsth_get_clkCondition (vh);
      ptinsth_remove_ref (vharg1, vh);
      return pk_true;
      break;
    case hcomp:
      ptinsth_get_1arg (vh, &vharg1);
      ptinsth_remove_ref (vharg1, vh);
      break;
    default:;
    }

  return pk_false;
}

/**
   Affine clock.
   @param vh (in) : a clock.
   @param ar (out) : an affine relation, if it exists.
   @return "true" if an affine relation is associated with the operand clock, "false" otherwise.
  */
extern boolean
ptinsth_get_affinerelation (ptinsth vh, ptAffineRelation * ar)
{
  *ar = vh->affr;
  return ((*ar) != NILAFFINERELATION);
}

/**
   Affine clock.
   It associates with a clock (vh) an affine relation (ar).
  */
extern void
ptinsth_set_affinerelation (ptinsth vh, ptAffineRelation ar)
{
  vh->affr = ar;
}

/**
   Step of changing a representative element of a class. 
   The current representative element is vh, the future is vform.
extern ptinsth
ptinsth_substitute_clhead_init(ptinsth vh, ptinsth vform)
{
  ptinsth vhaux;
  ptlinsth lsucc;
  
  ptinsth_set_clhead(vform, pk_true);
  ptinsth_move_references(vh, vform);
  return vh;
}
*/

/**
   Step of changing a representative element of a class. 
   The current representative element is vh, the future is vform.
   The references to vh are replaced by references to vfom.
*/
extern ptinsth
ptinsth_move_references (ptinsth vh, ptinsth vform)
{
  ptinsth vhaux;
  ptlinsth lsucc;

  lsucc = vh->lsucch;
  vform->lsucch = lsucc;
  while (lsucc != NILLINSTH)
    {
      vhaux = lsucc->elem;
      if (vh != vhaux)
	resetform (vh, vhaux, vform);
      lsucc = lsucc->next;
    }
  return vh;
}

/**
   @return the number of created formula.
 */
extern int
ptinsth_getClockCounter ()
{
  return numhorl;
}

/**
   @return the predefined null universal clock.
*/
extern ptinsth
ptinsth_get_hnull ()
{
  return the_nullClock;
}

/**
   @return the predefined init universal clock.
*/
extern ptinsth
ptinsth_get_hinit ()
{
  return hinit;
}

/**
   @return the predefined undefined universal clock.
*/
extern ptinsth
ptinsth_get_hundef ()
{
  return hundef;
}

/**
   @return true is vh is the predefined universal null clock, false otherwise.
*/
extern boolean
ptinsth_is_hnull (ptinsth vh)
{
  return (vh == the_nullClock);
}

/**
   @return true is vh is the predefined universal init clock, false otherwise.
*/
extern boolean
ptinsth_is_hinit (ptinsth vh)
{
  return (vh == hinit);
}

/**
   @return true is vh is the predefined universal undefined clock, false otherwise.
*/
extern boolean
ptinsth_is_hundef (ptinsth vh)
{
  return (vh == hundef);
}

/**
   Global initializing of the module.
*/
extern void
INIT_CLOCKS ()
{
  if (numhorl == -1)
    {
      numhorl = 0;
      // universal Null Clock.
      the_nullClock = ptinsth_newClock (predef);
      // universal Undefined Clock.
      hundef = ptinsth_newClock (predef);
      // universal Init Clock.
      hinit = ptinsth_newClock (predef);
      numhorl = 2;
    }
}

/**
   It frees a clock  descriptor (vinsth)
*/
extern void
ptinsth_free (insthClock vinsth)
{
  delete vinsth;
}
