/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <stdio.h>
static int numhorl = -1;

// Universal clocks.
static ptinsth hundef, the_nullClock, hinit;

/**
   The argument of a not formula (vform) becomes vharg.
*/
static void
setNotArg (ptinsth vform, ptinsth vharg)
{
  vform->Operand.op1 = vharg;
}

/**
   trace indirection list from vform to a class formula
   the indirection mark is the character "*"
   the separator is "..>"
   assume a correct insthFormula vform
*/
/** print a sequence of reference clocks
    @return the clock following the last printed one or NILINSTH
 */
static insthFormula
ptinsth_DEBUG_printReferenceFormulaUntiClassFormula (insthFormula vform)
{
  insthFormula nextFormula = vform;
  if (nextFormula == NILINSTH)
    return NILINSTH;
  while (nextFormula != NILINSTH)
    {
      if (!(INSTHFORM_ISAREFERENCE (nextFormula)))
	return nextFormula;
      if (ptinsth_DEBUG_printFormula (nextFormula));
      printf ("..>");
      if (INSTHFORM_ISALONE (nextFormula))
	{
	  printf ("#%i", nextFormula->getnum ());
	  return NILINSTH;
	}
      INSTHFORM_NEXT (nextFormula);
      if (nextFormula == vform)
	{
	  printf ("#%i", vform->getnum ());
	  return NILINSTH;
	}
    }
  return NILINSTH;
}

static insthFormula
ptinsth_DEBUG_printFormulaUntilHeader (insthFormula vform, const char *separator)
{
  insthFormula resFormula = vform;
  while (!(INSTHFORM_ISHEADER (resFormula)))
    {
      ptinsth_DEBUG_printFormula (resFormula);
      printf ("%s", separator);
      INSTHFORM_NEXT (resFormula);
      if (resFormula == NILINSTH) return NILINSTH;
      if (resFormula == vform)	return vform;
    }
  return resFormula;
}

/**
   Modification de la definition d'une horloge. 
  vh     : horloge de substitution 
  vinsth : horloge consideree 
  vform  : horloge de substitution

  If vinsth is
  - an extraction, vform becomes the clock of the condition.
  - is a not formula, vform becomes ist argument.
  - a lower/super bound, the references to vh are replaced by vform
*/
static void
resetform (ptinsth vh, ptinsth vinsth, ptinsth vform)
{
  switch (vinsth->ophorl)
    {
    case sub:
    case nosub:
      vinsth->Operand.opsub1->set_clkcond (vform);
      break;
    case hnot:
    case hcomp:
      vinsth->Operand.op1 = vform;
      break;
    case infer:
    case super:
      if (vinsth->Operand.binOp.op21 == vh)
	vinsth->Operand.binOp.op21 = vform;
      else
	vinsth->Operand.binOp.op22 = vform;
      break;
    default:;
    }
}

/**
   @return "true" is a formula (pt) is a complemetary formula, pt = (hnot(vhargnot), vhother) 
  -"false" otherwise, vharghnot and vhother parameters are not modified
*/
static boolean
locisinfhnot (ptinsth pt, ptinsth * vharghnot, ptinsth * vhother)
{
  ptinsth vhop1, vhop2;

  if (pt->ophorl != infer)
    return pk_false;
  vhop1 = pt->Operand.binOp.op21;
  if (vhop1->ophorl == hnot)
    {
      *vharghnot = vhop1->Operand.op1;
      *vhother = pt->Operand.binOp.op22;
      return pk_true;
    }
  vhop2 = pt->Operand.binOp.op22;
  if (vhop2->ophorl == hnot)
    {
      *vharghnot = vhop2->Operand.op1;
      *vhother = vhop1;
      return pk_true;
    }
  return pk_false;
}

/**
   It assigns the operands (vhpred1, vhpred2 ) to an assumed binary formula (vinsth).
*/
static void
setform2 (ptinsth vinsth, ptinsth vhpred1, ptinsth vhpred2)
{
  vinsth->Operand.binOp.op21 = vhpred1;
  vinsth->Operand.binOp.op22 = vhpred2;
}

/**
   It inserts a formula (vh2) in a class (vh1). vh1 is assumed to be the head of the class (representative element).
*/
static void
setpredvh2 (ptinsth vh1, ptinsth vh2)
{
  ptinsth vh3;

  vh3 = vh1->clsucc;
  while (vh3->clsucc != vh1)
    vh3 = vh3->clsucc;
  vh3->clsucc = vh2;
}

/**
   @return the first operand of a binary formula (assumed).
*/
static ptinsth
ptinsth_get_left_arg (ptinsth pt)
{
  return pt->Operand.binOp.op21;
}

/**
   @return the second operand of a binary formula (assumed).
*/
static ptinsth
ptinsth_get_right_arg (ptinsth pt)
{
  return pt->Operand.binOp.op22;
}

/**
   @return "true" if the operator of a formula (pt) is the "variable" operator.
 */
static boolean
ptinsth_satisfies_opvari (ptinsth pt)
{
  return (pt->ophorl == vari);
}

/**
   It assigns a number (unique) to the formula (vh).
*/
static void
setNumClock (ptinsth vh)
{
  numhorl = numhorl + 1;
  vh->setnum (numhorl);
}

/**
  It creates a new class of formula with only vinsth as element, characterized by its operator (codhorl).
  Its becomes tre representative element (if visaclas). The associated graph is (re)initialized.
  \note it is possible to have a class without any representative element.
*/
static void
ptinsth_set_clform (ptinsth vinsth, pkClockType codhorl, boolean visaclas)
{
  vinsth->clhead = visaclas;
  vinsth->ophorl = codhorl;
  vinsth->clsucc = vinsth;
  vinsth->normform = pk_false;
  ptinsth_set_lsucci_init (vinsth);
}
