/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   It associates the boolean vdef as the boolean clock of the clock clk 
   relatively to the clock father.
   vdef references the signals of the list lrefs.
*/
  ptinsth ClockManager::setClockAsBoolean (ptinsth clk, ptexpcal vdef, ptlinstances lrefs, ptinsth father)
{
  ptinsth vhEquiv;

  vhEquiv = makeEquivForm (clk, vdef, lrefs, father);
  if (clk != vhEquiv)
    if (!satisfiesTickClock (clk))
      {
	switchFormula (clk, vhEquiv, father);
	cloneClocks (clk, vhEquiv);	// Jan 2014 Lbesnard. xxxxxxxxxxxxxxxx TEST
	/** 2 lines added Nov 2013 LBesnard */
	removelhplaced (clk, pk_true);
	insertlhplaced (vhEquiv);
	setRoot (vhEquiv, getTickClock ());
      }
  return vhEquiv;
}

/** @return the boolean form of a clock (clk) relatively to its mother clock,  or 
    the undefined clock when it is not defined.
*/
ptinsth ClockManager::getBooleanClock (ptinsth clk)
{
  ptinsth
    vhresult;

  if (getBooleanClockAt (clk, getFather (clk), &vhresult))
    return vhresult;
  return hundef;
}

/** @return the booleen that represents a clock (clk),  relatively to its mother clock, or NIL when it is not 
    defined.
    In bDC+ sub-format, such information will be available.
 */
ptexpcal ClockManager::getBooleanSignal (ptinsth clk)
{
  ptinsth
    vh;
  pkClockType
    th;

  vh = getBooleanClock (clk);
  th = getOperator (vh);
  if (th == sub)
    return getConditionSub (vh);
  /* blindage */
  return NILEXPCAL;
}

/** @return true when it exists a clock (vhresult) an equivalent extraction to clk at the clock hAT, false otherwise.
 */
boolean ClockManager::getBooleanClockAt (ptinsth clk, ptinsth hAT, ptinsth * vhresult)
{
  boolean
    found,
    btick;
  ptinsth
    vh1;

  clk = getRepresentativeFormula (clk);
  hAT = getRepresentativeFormula (hAT);
  btick = satisfiesTickClock (clk);
  if ((clk == hAT) && (!btick))
    {				/* cas bizare de ^- (avec bdd) xxxxx A REVOIR */
      found = satisfiesSub (clk);
      vh1 = clk;
    }
  else
    {
      vh1 = clk;
      found = pk_false;
      do
	{
	  if (satisfiesSub (vh1))
	    found = (getClockSub (vh1) == hAT) || (getFather (vh1) == hAT) || btick;
	  if (!found)
	    vh1 = getNextFormula (vh1, pk_false);
	}
      while ((vh1 != clk) && !found);
    }
  if (found)
    *vhresult = vh1;
  else
    *vhresult = hundef;
  return found;
}

/** @return 
    - 0 if it exists an extraction clock (vh) that represents a clock (clk) at another clock (hAt).
    - 1 if it exists an equivalent (vh) placed above hAt (hAt is included in such a clock) 
    - 2 otherwise, in this case vh is the son of hAt on the branch (hAT,clk).
      This (boolean) clock can be used to create a signal defined by a left_tt that represent clk at hAt.

  \pre clk and hAT sare clocks such that the uper placed bound (in the hierarchy) is hAt.
*/
int
ClockManager::getBooleanClockScopeExtandAt (ptinsth clk, ptinsth hAT, ptinsth * vh)
{
  ptinsth vhson1, vhson2, vhemb;

  if (clk == hAT)
    {
      *vh = hAT;
      return 0;
    }
  if (getBooleanClockAt (clk, hAT, vh))
    return 0;
  if (getBooleanClockUp (clk, hAT))
    {
      *vh = hAT;
      return -1;
    }
  getCompleteBranching (&hAT, &clk, &vhson1, &vhson2, &vhemb);
  *vh = vhson1;
  return 2;
}

/** @return true if it exists an extraction clock (undersampling by an expression) equivalent to
    clk above the clock hAt (hAt is included in such a clock).
    \pre Assume : clk is a undersampling by an expression.
*/
boolean ClockManager::getBooleanClockUp (ptinsth clk, ptinsth hAT)
{
  boolean
    found;
  ptinsth
    vh1,
    vh,
    vhcond;
  ptexpcal
    cond;

  clk = getRepresentativeFormula (clk);
  cond = getConditionSub (clk);
  vhcond = getClockSub (clk);
  if (satisfiesFictitiousRoot (vhcond))
    return pk_true;
  found = pk_false;
  hAT = getRepresentativeFormula (hAT);
  vh1 = clk;
  do
    {
      if (getOperator (vh1) == sub)
	{
	  vh = getClockSub (vh1);
	  if (getBranching (getRepresentativeFormula (vh), getRepresentativeFormula (hAT)) == vh)
	    return pk_true;
	}
      vh1 = getNextFormula (vh1, pk_false);
    }
  while (vh1 != clk);
  return pk_false;
}
