/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
  @return 
  - the placed formula (vharg1 ^*vharg2) if the clocks vharg1 and  vharg2 are in the same tree 
  - normalized formula (sum of products of extractions/variables) of vharg1 ^*vharg2, otherwise.
*/
  ptinsth ClockManager::normalizeInf (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    otherterm;
  if (treeReduceInf (vharg1, vharg2, &otherterm))
    return otherterm;
  return infBooleanNormForm (vharg1, vharg2);
}

/**
  @return 
  - the placed formula (vharg1 ^+ vharg2) if the clocks vharg1 and  vharg2 are in the same tree 
  - normalized formula (sum of products of extractions/variables) of vharg1 ^+ vharg2, otherwise.
*/
ptinsth
ClockManager::normalizeSup (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth otherterm;

  if (treeReduceSup (vharg1, vharg2, &otherterm))
    return otherterm;
  return supBooleanNormForm (vharg1, vharg2);
}

/**
  It holoph-simplifies a complementary formula.
  @param vharg1 is the term of a not-formula (assumed),
  @param vharg2 a formula;
  @param  newterm result of the simplification
  @return 
  - "true" if the not(vharg1) ^* vharg2 is holoph-simplified in newterm,
  - "false" otherwise (newterm is not assigned).
*/
boolean
ClockManager::infReducehnot (int holoph, ptinsth vharg1, ptinsth vharg2, ptinsth * newterm)
{
  ptinsth vh1, vh2;
  noperators vop;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  if (isIncluded (vharg2, vharg1))
    {
      *newterm = the_nullClock_CKM_copy;
      return pk_true;
    }
  if (holoph == 0)
    return pk_false;
  vharg2 = getRepresentativeFormula (vharg2);
  vop = getOperator (vharg2);
  if (vop == infer)
    {
      if (getMinusArgs (vharg2, &vh1, &vh2))
	return pk_false;
      getInfArgs (vharg2, &vh1, &vh2);
      if (infReducehnot (holoph - 1, vharg1, vh1, newterm))
	{
	  if (!satisfiesNullClock (*newterm))
	    *newterm = booleanReducedInf (holoph - 1, hundef, *newterm, vh2, pk_true);
	  return pk_true;
	}
      if (!infReducehnot (holoph - 1, vharg1, vh2, newterm))
	return pk_false;
      if (!satisfiesNullClock (*newterm))
	*newterm = booleanReducedInf (holoph - 1, hundef, *newterm, vh1, pk_true);
      return pk_true;
    }
  if (vop == super)
    {
      getSupArgs (vharg2, &vh1, &vh2);
      if (infReducehnot (holoph - 1, vharg1, vh1, newterm))
	{
	  if (satisfiesNullClock (*newterm))
	    *newterm = vh2;
	  else
	    *newterm = booleanReducedSup (holoph - 1, hundef, *newterm, vh2, pk_true);
	  return pk_true;
	}
      if (infReducehnot (holoph - 1, vharg1, vh2, newterm))
	{
	  if (satisfiesNullClock (*newterm))
	    *newterm = vh1;
	  else
	    *newterm = booleanReducedSup (holoph - 1, hundef, *newterm, vh1, pk_true);
	  return pk_true;
	}
      return pk_false;
    }
  return pk_false;
}

/**
   It holoph-simplifies vterm2 by removing the terms T of vterm2 such that T*vharg1 = ^0.

   @param vharg1 (in) is an extraction (assumed),
   @param vterm2 (in/out) a formula, vterm2 (out) the holoph-simplification of vterm2 (in) by vharg1.

   @return 
   - "true" if the vharg1 ^* vharg2 is holoph-simplified in null clock,
   - "false" otherwise (vterm2 can be modified).
*/
boolean
ClockManager::infReduceNullClockExtract (int holoph, ptinsth vharg1, ptinsth * vterm2)
{
  ptinsth vharg2, vh1, vh2, svh1, svh2;
  noperators vop;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (*vterm2);
  vop = getOperator (vharg2);
  if  ((vop == sub) || (vop == nosub)) return TestExclusiveConditions(vharg1, vharg2);
  if (holoph == 0) return pk_false;
  if (vop == infer)
    {
      if (getMinusArgs (vharg2, &vh1, &vh2)) return pk_false;
      getInfArgs (vharg2, &vh1, &vh2);
      svh1 = vh1;
      if (infReduceNullClockExtract (holoph - 1, vharg1, &vh1)) return pk_true;
      svh2 = vh2;
      if (infReduceNullClockExtract (holoph - 1, vharg1, &vh2)) return pk_true;
      if ((vh1 != svh1) || (vh2 != svh2))
	{
	  *vterm2 = booleanReducedInf (holoph - 1, hundef, vh1, vh2, pk_true);
	  return satisfiesNullClock(*vterm2);
	}
      return pk_false;
    }
  if (vop == super)
    {
      getSupArgs (vharg2, &vh1, &vh2);
      // restriction: no recursive terms.
      if ( getRepresentativeFormula (vh1) == vharg2) return  pk_false;
      if ( getRepresentativeFormula (vh2) == vharg2) return  pk_false;

      svh1 = vh1;
      if (infReduceNullClockExtract (holoph - 1, vharg1, &vh1))
	{
	  if (infReduceNullClockExtract (holoph - 1, vharg1, &vh2)) return pk_true;
	  *vterm2 = vh2;
	  return pk_false;
	}
      svh2 = vh2;
      if (infReduceNullClockExtract (holoph - 1, vharg1, &vh2)) 
	*vterm2 = vh1;
      else if ((vh1 != svh1) || (vh2 != svh2))
	{
	  *vterm2 = booleanReducedSup (holoph - 1, hundef, vh1, vh2, pk_true);
	  return satisfiesNullClock(*vterm2);
	}
      return pk_false;
    }
  return  pk_false;
}

/**
   It holoph-simplifies vterm2 by removing the terms T of vterm2 such that T*vharg1 = ^0.

   @param vharg1 (in) is an extraction (assumed),
   @param vterm2 (in/out) a formula, vterm2 (out) the holoph-simplification of vterm2 (in) by vharg1.

   It calls infReduceNullClockExtract() for vharg1 an extraction.
   
   @return 
   - "true" if the vharg1 ^* vharg2 is holoph-simplified in null clock,
   - "false" otherwise (vterm2 can be modified).
*/
boolean
ClockManager::infReduceNullClock (int holoph, ptinsth vharg1, ptinsth * vharg2)
{
  ptinsth vh1, vh2;
  noperators vop;

  if (holoph == 0) return pk_false;
  vop = getOperator (vharg1);
  switch ((int) vop)
    {
    case sub:
    case nosub:
      return infReduceNullClockExtract (holoph, vharg1, vharg2);

    case infer:
      if (getMinusArgs (vharg1, &vh1, &vh2))
	return pk_false;

      getInfArgs (vharg1, &vh1, &vh2);
      if (infReduceNullClock (holoph - 1, vh1, vharg2))
	return pk_true;
      return infReduceNullClock (holoph - 1, vh2, vharg2);

    case super:
      return pk_false;

    default:
      return pk_false;
    }
}

/**
   It holoph-simplifies a term (T) vharg1 ^* vharg2 without decomposition of the term vharg2.
   @return "true" if the term T is holoph-simplified in vharg2, false otherwise.
   if breduce, vharg1 (out) is vharg1 (in) - the elements greater than vharg2 (in).
*/
boolean
ClockManager::infRemainder (int holoph, boolean breduce, ptinsth * vharg1, ptinsth * vharg2)
{
  boolean bresult;
  ptinsth vh, vh1, vh2, vhs1, vhs2;
  boolean tester;

  vh = getRepresentativeFormula (*vharg1);
  if (isIncluded (*vharg2, vh))
    return pk_true;
  if (holoph == 0)
    return pk_false;

  switch ((int) (getOperator (vh)))
    {
    case sub:
    case nosub:
      vh2 = getClockSub (vh);
      if (vh2 == (*vharg2))
	*vharg2 = the_nullClock_CKM_copy;
      else
	infRemainder (holoph - 1, pk_false, &vh2, vharg2);
      return pk_false;

    case infer:
      if (!getMinusArgs (vh, &vh1, &vh2))
	{
	  getInfArgs (vh, &vh1, &vh2);
	  vh = *vharg2;
	  vhs1 = vh1;
	  if (infRemainder (holoph - 1, pk_true, &vh1, vharg2))
	    if (infRemainder (holoph - 1, pk_true, &vh2, vharg2))
	      bresult = pk_true;
	    else
	      {
		*vharg1 = vh2;
		bresult = pk_false;
	      }
	  else
	    {
	      vhs2 = vh2;
	      if (satisfiesNullClock (*vharg2))
		{
		  *vharg2 = vh;
		  tester = pk_true;
		}
	      else
		tester = pk_false;
	      bresult = pk_false;
	      if (infRemainder (holoph - 1, pk_true, &vh2, vharg2))
		{
		  if (satisfiesNullClock (*vharg2))
		    {
		      bresult = pk_true;
		      tester = pk_false;
		    }
		  else
		    *vharg1 = vh1;
		}
	      else if (vh1 == vh2)
		*vharg1 = vh1;
	      else if (breduce && ((vhs1 != vh1) || (vhs2 != vh2)))
		{
		  *vharg1 = booleanReducedInf (holoph - 1, hundef, vh1, vh2, pk_true);
		  if ((*vharg1) == vh)
		    {
		      tester = pk_false;
		      *vharg2 = vh;
		      bresult = pk_true;
		    }
		}
	      if (tester)
		*vharg2 = the_nullClock_CKM_copy;
	    }
	}
      else			// vharg1 = vh2 ^- vh1
	{
	  bresult = pk_false;
	  if (isIncluded (*vharg2, vh1))
	    {
	      *vharg2 = the_nullClock_CKM_copy;
	      bresult = pk_true;
	    }
	  else if (isIncluded (vh2, *vharg2))	// rule : H *(K \ Z) with K < H  @return  (K\Z)
	    {
	      *vharg2 = vh;
	      bresult = pk_true;
	    }
	  /*
	   *vhs2 = *vharg2; vhs1 = vh2; if (infRemainder(holoph - 1, pk_false, vharg2, &vh1)) { 
	   *vharg2  = vh; bresult = pk_true;} else { *vharg2 = vhs2; bresult = pk_false; } */
	}
      return bresult;

    case super:
      getBinaryArgs (vh, &vh1, &vh2);
      vh = *vharg2;
      vhs1 = vh1;
      if (infRemainder (holoph - 1, pk_true, &vh1, vharg2))
	bresult = pk_true;
      else
	{
	  if (satisfiesNullClock (*vharg2))
	    {
	      *vharg2 = vh;
	      tester = pk_false;
	    }
	  else
	    tester = pk_true;
	  vhs2 = vh2;
	  if (infRemainder (holoph - 1, pk_true, &vh2, vharg2))
	    bresult = pk_true;
	  else
	    {
	      bresult = pk_false;
	      if (vh1 == vh2)
		*vharg1 = vh1;
	      else if (breduce && ((vhs1 != vh1) || (vhs2 != vh2)))
		{
		  *vharg1 = booleanReducedSup (holoph - 1, hundef, vh1, vh2, pk_true);
		  if ((*vharg1) == vh)
		    {
		      tester = pk_true;
		      bresult = pk_true;
		    }
		}
	      if (tester)
		*vharg2 = vh;
	    }
	}
      return bresult;

    case hcomp:
    case hnot:
    case vari:
    case predef:
      return pk_false;

    default:
      pK_messenger->Caseerror (__LINE__, __FILE__);
    }
  return pk_false;
}

/**
   @return "true" if the term vharg1 ^*vharg2 is holoph-simplified in vharg2, false otherwise.
*/
boolean
ClockManager::infWithoutRemainder (int holoph, ptinsth * vharg1, ptinsth * vharg2)
{
  ptinsth newterm, vh;

  if (hredhnot (*vharg1, *vharg2)) return pk_true;
  if (hredhnot (*vharg2, *vharg1)) { *vharg2 = *vharg1; return pk_true; }
  if (hredhnotNullClock (*vharg1, *vharg2))
    {
      *vharg2 = the_nullClock_CKM_copy;
      return pk_true;
    }
  if (getUnaryArg (*vharg1, &vh))	// vharg1 = NOT(vh)
    {
      if (!infReducehnot (holoph, vh, *vharg2, &newterm))
	return pk_false;
      *vharg2 = newterm;
      return satisfiesNullClock (*vharg2);
    }
  if (infReduceNullClock (holoph, *vharg1, vharg2))
    {
      *vharg2 = the_nullClock_CKM_copy;
      return pk_true;
    }
  if (infRemainder (holoph, pk_true, vharg1, vharg2))
    return pk_true;
  if (!satisfiesNullClock (*vharg2))
    return pk_false;
  *vharg2 = *vharg1;
  return pk_true;
}

/**
   It holoph-simplifies a term vharg1 ^+ vharg2 without decomposition of the term vharg2.
   @return 
   - "true" if the term vharg1 ^+ vharg2 is holoph-simplified in vharg2 (out) 
   - "false" otherwise.
   vharg1(out) and vharg2 (out) are reduced if breduce.
*/
boolean
ClockManager::supRemainder (int holoph, boolean breduce, ptinsth * vharg1, ptinsth * vharg2)
{
  boolean bresult;
  ptinsth vh, vh1, vh2, vhs1, vhs2;
  boolean tester;

  vh = getRepresentativeFormula (*vharg1);
  if (isIncluded (vh, *vharg2))
    return pk_true;
  if (holoph == 0)
    return pk_false;
  switch ((int) (getOperator (vh)))
    {
    case vari:
    case predef:
    case hcomp:
    case hnot:
      return pk_false;

    case sub:
    case nosub:
      vh1 = getClockSub (vh);
      vh = *vharg2;
      return supRemainder (holoph - 1, pk_false, &vh1, &vh);
      break;

    case infer:
      getBinaryArgs (vh, &vh1, &vh2);
      vh = *vharg2;
      vhs1 = vh1;
      if (supRemainder (holoph - 1, pk_true, &vh1, vharg2))
	return pk_true;
      if (satisfiesNullClock (*vharg2))
	{
	  *vharg2 = vh;
	  tester = pk_false;
	}
      else
	tester = pk_true;
      vhs2 = vh2;
      if (supRemainder (holoph - 1, pk_true, &vh2, vharg2))
	return pk_true;
      bresult = pk_false;
      if (vh1 == vh2)
	*vharg1 = vh1;
      else if (breduce && ((vhs1 != vh1) || (vhs2 != vh2)))
	{
	  *vharg1 = booleanReducedInf (holoph - 1, hundef, vh1, vh2, pk_true);
	  if ((*vharg1) == vh)
	    {
	      tester = pk_true;
	      bresult = pk_true;
	    }
	}
      if (tester)
	*vharg2 = vh;
      break;

    case super:
      getBinaryArgs (vh, &vh1, &vh2);
      vh = *vharg2;
      vhs1 = vh1;
      if (supRemainder (holoph - 1, pk_true, &vh1, vharg2))
	{
	  if (supRemainder (holoph - 1, pk_true, &vh2, vharg2))
	    return pk_true;
	  *vharg1 = vh2;
	  return pk_false;
	}
      if (satisfiesNullClock (*vharg2))
	{
	  *vharg2 = vh;
	  tester = pk_true;
	}
      else
	tester = pk_false;

      vhs2 = vh2;
      bresult = pk_false;
      if (supRemainder (holoph - 1, pk_true, &vh2, vharg2))
	*vharg1 = vh1;
      else if (vh1 == vh2)
	*vharg1 = vh1;
      else if (breduce && ((vhs1 != vh1) || (vhs2 != vh2)))
	{
	  *vharg1 = booleanReducedSup (holoph - 1, hundef, vh1, vh2, pk_true);
	  if ((*vharg1) == vh)
	    {
	      tester = pk_false;
	      *vharg2 = vh;
	      bresult = pk_true;
	    }
	}
      if (tester)
	*vharg2 = the_nullClock_CKM_copy;
      break;
    default:
      pK_messenger->Caseerror (__LINE__, __FILE__);
    }
  return bresult;
}

/**
  It holoph-simplifies a term (T) of the form (x ^* y) ^+ vharg2.
  @param vharg1 a complementary formula (assumed),
  @param vharg2 a formula;
  @param newterm result of the simplification
  @return 
  - "true" if T is simplified in newterm,
  - "false" otherwise (newterm is not assigned).
*/
boolean
ClockManager::supReducehnot (int holoph, ptinsth vharg1, ptinsth vharg2, ptinsth * newterm)
{
  ptinsth vhnot, vh1, vh2;
  noperators vop;

  if (holoph == 0)
    return pk_false;

  vharg2 = getRepresentativeFormula (vharg2);
  vop = getOperator (vharg2);
  if (vop == infer)
    {
      if (getMinusArgs (vharg2, &vhnot, newterm))
	{
	  if (vhnot == vharg1)
	    return pk_true;
	  if (getInfArgs (vhnot, &vh1, &vh2))
	    return ((vharg1 == vh2) || (vharg1 == vh1));
	  return pk_false;
	}
      getInfArgs (vharg2, &vh1, &vh2);
      if (supReducehnot (holoph - 1, vharg1, vh1, newterm))
	{
	  *newterm = booleanReducedInf (holoph - 1, hundef, *newterm, vh2, pk_true);
	  return pk_true;
	}
      if (!supReducehnot (holoph - 1, vharg1, vh2, newterm))
	return pk_false;
      *newterm = booleanReducedInf (holoph - 1, hundef, *newterm, vh1, pk_true);
      return pk_true;
    }
  if (vop == super)
    {
      getSupArgs (vharg2, &vh1, &vh2);
      if (supReducehnot (holoph - 1, vharg1, vh1, newterm))
	{
	  *newterm = booleanReducedSup (holoph - 1, hundef, *newterm, vh2, pk_true);
	  return pk_true;
	}
      if (!supReducehnot (holoph - 1, vharg1, vh2, newterm))
	return pk_false;
      *newterm = booleanReducedSup (holoph - 1, hundef, *newterm, vh1, pk_true);
      return pk_true;
    }
  return pk_false;
}

/**
   It holoph-simplifies a term (T) vharg1 ^+ vharg2 without decomposition of the term vharg2.
   vharg1 and vharg2 can be modified.
   @return "true" if the result of the simpliciation is vharg2 (out), "false" otherwise.
*/
boolean
ClockManager::supWithoutRemainder (int holoph, ptinsth * vharg1, ptinsth * vharg2)
{
  ptinsth newterm;

  if (supReducehnot (holoph, *vharg1, *vharg2, &newterm))
    {
      *vharg2 = newterm;
      return pk_false;
    }
  if (supRemainder (holoph, pk_true, vharg1, vharg2))
    return pk_true;
  if (satisfiesNullClock (*vharg2))
    {
      *vharg2 = *vharg1;
      return pk_true;
    }
  return pk_false;
}

/**
   It holoph-simplifies the term (T) vharg1 ^*vharg2 using 
    - (A) classical boolean rules if the operands are not in the same tree of clocks;
    - (B) tree reduction technique otherwise.
   @return the result of the simplification of T.
   @param vcommu (in) at pk_true for applying commutativity, otherwise vharg2 is not decomposed in (A).
   @param vfinf  (in) : original formula (vharg1 ^*vharg2) or undefined clock. When it is undefined, formula are
   created in the system, otherwise the original formula is replaced or rewritten.
   @param holoph (in) : depth of solving used to limit the rewritings in (A).
*/
ptinsth
ClockManager::booleanReducedInf (int holoph, ptinsth vfinf, ptinsth vharg1, ptinsth vharg2, boolean vcommu)
{
  ptinsth vhresult;
  ptinsth otherterm, vhs1, vhs2;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  // begin added Oct 2015 LBesnard.
  vhs2 = vharg2;
  if (infReduceNullClock (holoph, vharg1, &vhs2)) return replaceInf (vfinf, the_nullClock_CKM_copy);
  // if ( vhs2 != vharg2 ) return rewriteInf (vfinf, vharg1, vhs2);
  if (vcommu)
    {
      vhs1 = vharg1;
      if (infReduceNullClock (holoph, vharg2, &vhs1)) return replaceInf (vfinf, the_nullClock_CKM_copy);
      // if ( vhs1 != vharg1) return rewriteInf (vfinf, vhs1, vharg2 );
    }
  // -- end added Oct 2015
  if (treeReduceInf (vharg1, vharg2, &otherterm)) return replaceInf (vfinf, otherterm);

  if (holoph == 0) return rewriteInf (vfinf, vharg1, vharg2);

  vhs1 = vharg1;
  if (infWithoutRemainder (holoph, &vharg1, &vharg2)) return replaceInf (vfinf, vharg2);
  if (vcommu && (vharg1 == vhs1))
    {
      vhs2 = vharg2;
      if (infWithoutRemainder (holoph, &vharg2, &vhs1)) return replaceInf (vfinf, vhs1);
      if (vhs2 == vharg2)
	{
	  if (vhs1 == vharg1)
	    vhresult = rewriteInf (vfinf, vharg1, vharg2);
	  else
	    vhresult = booleanReducedInf (holoph - 1, vfinf, vharg2, vharg1, pk_false);
	}
      else 
	if (vhs1 == vharg1)
	  vhresult = booleanReducedInf (holoph - 1, vfinf, vharg1, vharg2, pk_false);
	else
	  vhresult = rewriteInf (vfinf, vharg1, vharg2);
    }
  else
    vhresult = rewriteInf (vfinf, vharg1, vharg2);
  return vhresult;
}

/**
   It holoph-simplifies the term (T) vharg1 ^+ vharg2 using 
    - (A) classical boolean rules if the operands are not in the same tree of clocks;
    - (B) tree reduction technique otherwise.
   @return the result of the simplification of T.
   @param vcommu (in) at pk_true for applying commutativity, otherwise vharg2 is not decomposed in (A).
   @param vfsup (in) : original formula (vharg1 ^*vharg2) or undefined clock. When it is undefined, formula are
   created in the system, otherwise the original formula is replaced or rewritten in the system.
   @param holoph (in) : depth of solving used to limit the rewritings in (A).
*/
ptinsth
ClockManager::booleanReducedSup (int holoph, ptinsth vfsup, ptinsth vharg1, ptinsth vharg2, boolean vcommu)
{
  ptinsth vhresult;
  ptinsth otherterm, vhs1, vhs2;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  if (treeReduceSup (vharg1, vharg2, &otherterm))
    return replaceSup (vfsup, otherterm);

  if (holoph == 0)
    return rewriteSup (vfsup, vharg1, vharg2);
  vhs1 = vharg1;
  if (supWithoutRemainder (holoph, &vharg1, &vharg2))
    return replaceSup (vfsup, vharg2);
  if (vcommu && (vharg1 == vhs1))
    {
      vhs2 = vharg2;
      if (supWithoutRemainder (holoph, &vharg2, &vhs1))
	vhresult = replaceSup (vfsup, vhs1);
      else if (vhs2 == vharg2)
	if (vhs1 == vharg1)
	  {
	    if (!supConditionCompl (&vharg1, &vharg2))
	      supConditionCompl (&vharg2, &vharg1);
	    vhresult = rewriteSup (vfsup, vharg1, vharg2);
	  }
	else
	  vhresult = booleanReducedSup (holoph - 1, vfsup, vharg2, vharg1, pk_false);
      else if (vhs1 == vharg1)
	vhresult = booleanReducedSup (holoph - 1, vfsup, vharg1, vharg2, pk_false);
      else
	{
	  if (!supConditionCompl (&vharg1, &vharg2))
	    supConditionCompl (&vharg2, &vharg1);
	  vhresult = rewriteSup (vfsup, vharg1, vharg2);
	}
    }
  else
    {
      supConditionCompl (&vharg1, &vharg2);
      vhresult = rewriteSup (vfsup, vharg1, vharg2);
    }
  return vhresult;
}

/**
   It applies the rule : ( F ^+ [c] ) ^+ [not(c)]  -> F ^+  ^c
   If vharg1 (in) is a term of the form (F ^+ [c]), and if vharg2 a term of the form [not(c)] 
   it returns 
    - "true", and vharg1 (out) becomes F, vharg2 (out) becomes ^c
    - "false" otherwise.
*/
boolean
ClockManager::supConditionCompl (ptinsth * vharg1, ptinsth * vharg2)
{
  ptinsth vh1, vh1c, vh2;
  ptinsth vhson1, vhson2, vhemb;

  if (!getSupArgs (*vharg1, &vh1, &vh2))
    return pk_false;
  if (satisfiesSub (*vharg2))
    {
      if (getComplExtraction (vh1, &vh1c))
	if (vh1c == *vharg2)
	  {
	    *vharg1 = vh2;
	    *vharg2 = getClockSub (vh1);
	    return pk_true;
	  }
      if (getComplExtraction (vh2, &vh1c))
	if (vh1c == *vharg2)
	  {
	    *vharg1 = vh1;
	    *vharg2 = getClockSub (vh2);
	    return pk_true;
	  }
      // return pk_false; removed AUG 2016
    }
  // Added Feb 2013 (LBesnard)
  if (!hoclembr (vh1, *vharg2, &vhson1, &vhson2, &vhemb))
    if (!ptinsth_is_hundef (vhemb))
      {
	*vharg1 = vhemb;
	*vharg2 = vh2;
	return pk_true;
      }
  if (!hoclembr (vh2, *vharg2, &vhson1, &vhson2, &vhemb))
    if (!ptinsth_is_hundef (vhemb))
      {
	*vharg1 = vhemb;
	*vharg2 = vh1;
	return pk_true;
      }
  return pk_false;
}

/**
   @return 
     - false if it does not exists some equivalences to do, 
     - true otherwise, in this case it builds this equivalences and the induced equivalences.
*/
boolean
ClockManager::hclmklequiv (int holoph)
{
  ptinsth vh1, vh2;

  if (!ModeSynchroSolving)
    return pk_false;
  if (!hclsomesynchro (&vh1, &vh2))
    return pk_false;
  do
    {
      vh1 = hclmkunion (holoph, vh1, vh2);
    }
  while (hclsomesynchro (&vh1, &vh2));
  return pk_true;
}

/** 
    @return "true" if vharg1 < vharg2 using the hierarchy, "false" otherwise.
    \remarks The inclusion test depends on the global variable strongInclusionTest.
 */
boolean
ClockManager::isIncluded (ptinsth vharg1, ptinsth vharg2)
{
  if (isFictitiousRootFormula (vharg1))
    return (getOperator (vharg2) != hnot);
  if (strongInclusionTest)
    return isIncludedIn (vharg1, vharg2);
  return testfsupfintree (vharg2, vharg1);
}

boolean ClockManager::hredhnotNullClock (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vh, vharghnot;

  if (getOperator (vharg1) != hnot) return pk_false;
  getUnaryArg (vharg1, &vh);
  if (!getMinusArgs (vharg2, &vharghnot, &vharg2)) return pk_false;
  while (getMinusArgs (vharg2, &vharghnot, &vharg2));
  return isIncluded (vharg2, vh);
}
