/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include "atformula_clock_g.h"
_PRIVATE_
/**
  @return true if vharg1 ^* vharg2 is the null clock.
  Tested for vharg1 an extraction.
*/
  boolean ClockManager::isInfNullTermC (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    vh;
  if (!getComplExtraction (vharg1, &vh))
    return pk_false;
  return isIncludedIn (vharg2, vh);
}

/**
   It tries to reduce VHARG1 ^+ VHARG2.
   @return true if one of the operands is placed.
*/
boolean
ClockManager::horeducesup (ptinsth * vharg1, ptinsth * vharg2, boolean * breduce)
{
  boolean breturn;
  ptinsth vrac1, vrac2;
  int depth1, depth2;

  breturn = pk_true;
  *vharg1 = getRepresentativeFormula (*vharg1);
  *vharg2 = getRepresentativeFormula (*vharg2);
  vrac1 = getRoot (*vharg1, &depth1);
  vrac2 = getRoot (*vharg2, &depth2);
  if (vrac1 == vrac2)
    {
      if (*breduce)
	treeReduceSup (*vharg1, *vharg2, vharg1);
      else;
    }
  else if ((depth1 != 0) || (depth2 != 0))
    if (isIncludedIn (vrac1, *vharg2))
      if (*breduce)
	{
	  *vharg1 = *vharg2;
	  *vharg2 = the_nullClock_CKM_copy;
	}
      else
	*vharg1 = the_nullClock_CKM_copy;
    else if (isIncludedIn (vrac2, *vharg1))
      *vharg2 = the_nullClock_CKM_copy;
    else
      {
	if (*breduce)
	  *vharg1 = buildnormformsup (*vharg1, *vharg2);
      }
  else
    breturn = pk_false;
  return breturn;
}

/**
  Applying boolean reductions for builting normal form for root-formula.
  For vharg1, vharg2 monomials, and  breduce (in):
    - pk_true: it returns the normal form of vharg1 ^+ vharg2 
    - false: it returns the reduced form of vharg1 by vharg2, and (vharg2 (out)) the reduced 
    form of vharg2 by vharg1. 
  breduce (out): at pk_true if applied reductions can induce new reductions on the current rewriting term.
*/
ptinsth
ClockManager::normalizeSupmn (ptinsth vharg1, ptinsth * vharg2, boolean * breduce)
{
  boolean breduce1;

  breduce1 = *breduce;
  if (horeducesup (&vharg1, vharg2, &breduce1))
    {
      *breduce = breduce1;
      return vharg1;
    }
  if (isIncludedIn (vharg1, *vharg2))
    {
      if (*breduce)
	return *vharg2;
      return the_nullClock_CKM_copy;
    }
  if (isIncludedIn (*vharg2, vharg1))
    {
      *vharg2 = the_nullClock_CKM_copy;
      return vharg1;
    }
  if (*breduce)
    return buildnormformsup (vharg1, *vharg2);
  return vharg1;
}

/**
  Applying boolean reductions for builting normal form for root-formula.
  For vharg1 a sum, vharg2 a monomial, and  breduce (in):
    - pk_true: it returns the normal form of vharg1 ^+ vharg2 
    - false: it returns the reduced form of vharg1 by vharg2, and (vharg2 (out)) the reduced 
    form of vharg2 by vharg1. 
  breduce (out): at pk_true if applied reductions can induce new reductions on the current rewriting term.
*/
ptinsth
ClockManager::normalizeSupsn (ptinsth vharg1, ptinsth * vharg2, boolean * breduce)
{
  ptinsth vhsmon, vhmon, vhsarg2, vharg12, vhsarg12;
  boolean breduce1, breduce2;
  pkClockType vop1;

  breduce1 = *breduce;
  if (horeducesup (&vharg1, vharg2, &breduce1))
    {
      *breduce = breduce1;
      return vharg1;
    }

  getSupArgs (vharg1, &vhmon, &vharg12);
  breduce1 = pk_false;
  vhsarg2 = *vharg2;
  vhsmon = normalizeSupmn (vhmon, vharg2, &breduce1);
  if ((*vharg2) == the_nullClock_CKM_copy)
    return vharg1;

  breduce2 = pk_false;
  vop1 = getOperator (vharg12);
  if (vop1 == super)
    vhsarg12 = normalizeSupsn (vharg12, vharg2, &breduce2);
  else
    vhsarg12 = normalizeSupmn (vharg12, vharg2, &breduce2);

  breduce1 = (boolean) (breduce1 || breduce2);
  if (vhsmon == the_nullClock_CKM_copy)
    vharg1 = vhsarg12;
  else if (vhsarg12 == the_nullClock_CKM_copy)
    vharg1 = vhsmon;
  else if ((vhsmon != vhmon) || (vhsarg12 != vharg12))
    if (breduce1)
      if (getOperator (vhsarg12) == super)
	vharg1 = normalizeSupsn (vhsarg12, &vhsmon, &breduce1);
      else
	vharg1 = normalizeSupmn (vhsarg12, &vhsmon, &breduce1);
    else
      vharg1 = buildnormformsup (vhsmon, vharg12);

  if (*breduce)
    {
      if (vharg1 == the_nullClock_CKM_copy)
	return *vharg2;
      if ((*vharg2) == the_nullClock_CKM_copy)
	return vharg1;
      return buildnormformsup (vharg1, *vharg2);
    }
  *breduce = breduce1;
  return vharg1;
}

/**
  Applying boolean reductions for builting normal form for root-formula.
  For vharg1 and vahrg2 sums, and  breduce (in):
    - pk_true: it returns the normal form of vharg1 ^+ vharg2 
    - false: it returns the reduced form of vharg1 by vharg2, and (vharg2 (out)) the reduced 
    form of vharg2 by vharg1. 
  breduce (out): at pk_true if applied reductions can induce new reductions on the current rewriting term.
*/
ptinsth
ClockManager::normalizeSupss (ptinsth vharg1, ptinsth * vharg2, boolean * breduce)
{
  ptinsth vhsmon, vhmon, vhsarg1, vharg22, vhsarg22;
  boolean breduce1, breduce2;
  pkClockType vop1, vop2;

  breduce1 = *breduce;
  if (horeducesup (&vharg1, vharg2, &breduce1))
    {
      *breduce = breduce1;
      return vharg1;
    }
  breduce1 = pk_false;
  getSupArgs (*vharg2, &vhmon, &vharg22);
  vhsmon = vhmon;
  vhsarg1 = normalizeSupsn (vharg1, &vhmon, &breduce1);
  if (vhsarg1 == the_nullClock_CKM_copy)
    {
      if (*breduce)
	return *vharg2;
      return the_nullClock_CKM_copy;
    }

  breduce2 = pk_false;
  vharg1 = vhsarg1;
  vhsarg22 = vharg22;
  vop1 = getOperator (vharg1);
  vop2 = getOperator (vharg22);
  if (vop2 == super)
    if (vop1 == super)
      vhsarg1 = normalizeSupss (vharg1, &vhsarg22, &breduce2);
    else
      vhsarg22 = normalizeSupsn (vharg22, &vhsarg1, &breduce2);
  else if (vop1 == super)
    vhsarg1 = normalizeSupsn (vharg1, &vhsarg22, &breduce2);
  else
    vhsarg1 = normalizeSupmn (vharg1, &vhsarg22, &breduce2);
  breduce1 = (boolean) (breduce1 || breduce2);
  if (vhmon == the_nullClock_CKM_copy)
    *vharg2 = vhsarg22;
  else if (vhsarg22 == the_nullClock_CKM_copy)
    *vharg2 = vhmon;
  else if ((vhsmon != vhmon) || (vhsarg22 != vharg22))
    if (breduce1)
      if (getOperator (vhsarg22) == super)
	*vharg2 = normalizeSupsn (vhsarg22, &vhmon, &breduce1);
      else
	*vharg2 = normalizeSupmn (vhsarg22, &vhmon, &breduce1);
    else
      *vharg2 = buildnormformsup (vhmon, vhsarg22);

  if (*breduce)
    {
      if (vhsarg1 == the_nullClock_CKM_copy)
	return *vharg2;
      if ((*vharg2) == the_nullClock_CKM_copy)
	return vhsarg1;
      return buildnormformsup (vhsarg1, *vharg2);
    }
  *breduce = breduce1;
  return vhsarg1;
}

/**
   Assumed : vharg1 and vharg2 are formula not in the same tree.
   @return the normal form (sum of products of conditions/free variables) of vharg1 ^+ vharg2.
*/
ptinsth
ClockManager::supBooleanNormForm (ptinsth vharg1, ptinsth vharg2)
{
  boolean breduce;
  pkClockType vop1, vop2;

  breduce = pk_true;
  vop1 = getOperator (vharg1);
  vop2 = getOperator (vharg2);
  if (vop1 == super)
    {
      if (vop2 == super)
	return normalizeSupss (vharg1, &vharg2, &breduce);
      return normalizeSupsn (vharg1, &vharg2, &breduce);
    }
  if (vop2 == super)
    return normalizeSupsn (vharg2, &vharg1, &breduce);
  return normalizeSupmn (vharg1, &vharg2, &breduce);
}

/**
   @return the term VHARG1 ^* VHARG2 taking account the null clock.
*/
ptinsth
ClockManager::insertin (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vh1, vharg22, vhmon;

  if (!getInfArgs (vharg2, &vhmon, &vharg22))
    return rewriteInfnf (vharg1, vharg2);

  if (treeReduceInf (vharg1, vhmon, &vh1))
    {
      if (vh1 == the_nullClock_CKM_copy)
	return the_nullClock_CKM_copy;
      return rewriteInfnf (vh1, vharg22);
    }

  if (isbefore (vharg1, vhmon))
    return rewriteInfnf (vharg1, vharg2);
  vh1 = insertin (vharg1, vharg22);
  if (vh1 == the_nullClock_CKM_copy)
    return the_nullClock_CKM_copy;
  return rewriteInfnf (vhmon, vh1);
}

/**
   @return the term VHARG1 ^* VHARG2 taking account the null clock.
   Assumed : VHARG1, VHARG2 are monomials
   Local to \ref buildnormformmn.
*/
ptinsth
ClockManager::reducemn (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vhmon;
  boolean b;

  if (getOperator (vharg1) == infer)
    {
      if (getOperator (vharg2) != infer)
	return insertin (vharg2, vharg1);
      b = pk_true;
      while (b)
	{
	  b = getInfArgs (vharg1, &vhmon, &vharg1);
	  if (isInfNullTermC (vhmon, vharg2))
	    {
	      vharg2 = the_nullClock_CKM_copy;
	      b = pk_false;
	    }
	  else
	    {
	      vharg2 = insertin (vhmon, vharg2);
	      if (vharg2 == the_nullClock_CKM_copy)
		b = pk_false;
	    }
	}
      if (vharg2 == the_nullClock_CKM_copy)
	return vharg2;
      return insertin (vharg1, vharg2);
    }
  if (getOperator (vharg2) == infer)
    return insertin (vharg1, vharg2);
  if (isbefore (vharg1, vharg2))
    return rewriteInfnf (vharg1, vharg2);
  return rewriteInfnf (vharg2, vharg1);
}

/**
   @return the term VHARG1 ^* VHARG2 taking account the null clock.
   Assumed : VHARG1, VHARG2 are monomials
*/
ptinsth
ClockManager::buildnormformmn (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vrac1, vrac2;
  int depth1, depth2;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  if (treeReduceInf (vharg1, vharg2, &vrac1))
    return vrac1;

  vrac1 = getRoot (vharg1, &depth1);
  vrac2 = getRoot (vharg2, &depth2);
  if (vrac1 != vharg1)
    {
      if (vrac2 == vharg2)
	return insertin (vharg1, vharg2);
      if (isbefore (vharg1, vharg2))
	return rewriteInfnf (vharg1, vharg2);
      return rewriteInfnf (vharg2, vharg1);
    }
  if (vrac2 != vharg2)
    return insertin (vharg2, vharg1);
  return reducemn (vharg1, vharg2);
}

/**
  Applying boolean reductions for builting normal form for root-formula.
  Assumed : vharg1, vharg2 are monomials.
  @return normal form (sum of products of conditions/free variables) of vharg1 ^* vharg2.
*/
ptinsth
ClockManager::normalizeInfmn (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vrac1, vrac2;
  int depth1, depth2;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  vrac1 = getRoot (vharg1, &depth1);
  vrac2 = getRoot (vharg2, &depth2);
  if (vrac1 != vrac2)
    {
      if (isIncludedIn (vrac1, vharg2))
	return vharg1;
      if (isIncludedIn (vrac2, vharg1))
	return vharg2;
      return buildnormformmn (vharg1, vharg2);
    }
  if (treeReduceInf (vharg1, vharg2, &vrac1))
    return vrac1;
  return NILINSTH;
}

/**
  Applying boolean reductions for builting normal form for root-formula.
  Assumed : vharg1 is a polynomial, vharg2 a monomial.
  @return normal form (sum of products of conditions/free variables) of vharg1 ^* vharg2.
*/
ptinsth
ClockManager::normalizeInfsn (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vh1, vh2, vhmon, vharg12;

  if (!getSupArgs (vharg1, &vhmon, &vharg12))
    return NILINSTH;
  vh1 = normalizeInfmn (vhmon, vharg2);
  if (vh1 == vharg2)
    return vharg2;
  if (getOperator (vharg12) == super)
    vh2 = normalizeInfsn (vharg12, vharg2);
  else
    vh2 = normalizeInfmn (vharg12, vharg2);
  if (vh2 == vharg2)
    return vh2;
  if (vh2 == the_nullClock_CKM_copy)
    return vh1;
  if (vh1 == the_nullClock_CKM_copy)
    return vh2;
  return normalizeSup (vh1, vh2);
}

/**
  Applying boolean reductions for builting normal form for root-formula.
  Assumed : vharg1, vharg2 are polynomials.
  @return normal form (sum of products of conditions/free variables) of vharg1 ^* vharg2.
*/
ptinsth
ClockManager::normalizeInfss (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vh1, vh2, vhmon, vharg22;

  if (!getSupArgs (vharg2, &vhmon, &vharg22))
    return NILINSTH;
  vh1 = normalizeInfsn (vharg1, vhmon);
  if (vh1 == vharg1)
    return vh1;
  if (getOperator (vharg22) == super)
    vh2 = normalizeInfss (vharg1, vharg22);
  else
    vh2 = normalizeInfsn (vharg1, vharg22);
  if (vh2 == vharg1)
    return vharg1;
  if (vh1 == the_nullClock_CKM_copy)
    return vh2;
  if (vh2 == the_nullClock_CKM_copy)
    return vh1;
  return normalizeSup (vh1, vh2);
}

/**
   Assumed : vharg1, vharg2 are formula not in the same tree.
  @return the normal form (sum of products of conditions/free variables) of vharg1 ^* vharg2.
*/
ptinsth
ClockManager::infBooleanNormForm (ptinsth vharg1, ptinsth vharg2)
{
  pkClockType vop1, vop2;

  vop1 = getOperator (vharg1);
  vop2 = getOperator (vharg2);
  if (vop1 == super)
    {
      if (vop2 == super)
	return normalizeInfss (vharg1, vharg2);
      return normalizeInfsn (vharg1, vharg2);
    }
  if (vop2 == super)
    return normalizeInfsn (vharg2, vharg1);
  return normalizeInfmn (vharg1, vharg2);
}

/**
   Assumed : vharg1 (polynomial), vharg2 (monomial) are formula not in the same tree.
   @return the normal form (sum of products of conditions/free variables) of vharg1 ^+ vharg2.
*/
ptinsth
ClockManager::buildnormformsupsn (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth otherterm, vhmon, vharg12;

  if (!getSupArgs (vharg1, &vhmon, &vharg12))
    return NILINSTH;
  if (treeReduceSup (vhmon, vharg2, &otherterm))
    return rewriteSupnf (otherterm, vharg12);
  if (isbefore (vharg2, vhmon))
    return rewriteSupnf (vharg2, vharg1);
  if (getOperator (vharg12) == super)
    return rewriteSupnf (vhmon, buildnormformsupsn (vharg12, vharg2));
  if (isbefore (vharg12, vharg2))
    return rewriteSupnf (vhmon, rewriteSupnf (vharg12, vharg2));
  return rewriteSupnf (vhmon, rewriteSupnf (vharg2, vharg12));
}

/**
   Assumed : vharg1 (polynomial), vharg2 (polynomial) are formula not in the same tree.
   @return the normal form (sum of products of conditions/free variables) of vharg1 ^+ vharg2.
*/
ptinsth
ClockManager::buildnormformsupss (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vh1, vhmon, vharg22;

  if (!getSupArgs (vharg2, &vhmon, &vharg22))
    return NILINSTH;
  vh1 = buildnormformsupsn (vharg1, vhmon);
  if (getOperator (vharg22) == super)
    return buildnormformsupss (vh1, vharg22);
  return buildnormformsupsn (vh1, vharg22);
}

/**
  Assumed : vharg1 (placed), vharg2 (polynomial) are formula not in the same tree.
  @return the normal form (sum of products of conditions/free variables) of vharg1 ^+ vharg2.
*/
ptinsth
ClockManager::buildnormformsuppls (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth otherterm, vhmon, vharg22;

  if (!getSupArgs (vharg2, &vhmon, &vharg22))
    return NILINSTH;
  if (getDepth (vhmon) == 0)
    return rewriteSupnf (vharg1, vharg2);
  if (treeReduceSup (vharg1, vhmon, &otherterm))
    return rewriteSupnf (otherterm, vharg22);
  if (isbefore (vharg1, vhmon))
    return rewriteSupnf (vharg1, vharg2);
  if (getOperator (vharg22) == super)
    return buildnormformsuppls (vharg1, vharg22);
  if (isbefore (vharg1, vharg22))
    otherterm = rewriteSupnf (vharg1, vharg22);
  else
    otherterm = rewriteSupnf (vharg22, vharg1);
  return rewriteSupnf (vhmon, otherterm);
}

/**
   Assumed : vharg1, vharg2 are formula not in the same tree.   
   @return the normal form (sum of products of conditions/free variables) of vharg1 ^+ vharg2.
*/
ptinsth
ClockManager::buildnormformsup (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth otherterm;
  int depth1, depth2;

  if (treeReduceSup (vharg1, vharg2, &otherterm))
    return otherterm;
  depth1 = getDepth (vharg1);
  depth2 = getDepth (vharg2);
  if (depth1 > 0)
    {
      if (depth2 > 0)
	{
	  if (isbefore (vharg1, vharg2))
	    return rewriteSupnf (vharg1, vharg2);
	  return rewriteSupnf (vharg2, vharg1);
	}
      if (getOperator (vharg2) == super)
	return buildnormformsuppls (vharg1, vharg2);
      return rewriteSupnf (vharg1, vharg2);
    }
  if (depth2 > 0)
    {
      if (getOperator (vharg1) == super)
	return buildnormformsuppls (vharg2, vharg1);
      return rewriteSupnf (vharg2, vharg1);
    }
  if (getOperator (vharg1) == super)
    {
      if (getOperator (vharg2) == super)
	return buildnormformsupss (vharg1, vharg2);
      return buildnormformsupsn (vharg1, vharg2);
    }
  if (getOperator (vharg2) == super)
    return buildnormformsupsn (vharg2, vharg1);
  if (isbefore (vharg1, vharg2))
    return rewriteSupnf (vharg1, vharg2);
  return rewriteSupnf (vharg2, vharg1);
}

/**
  @return the class of the normalized formula vharg1 ^+ vharg2.
  Assumed : vharg1 and vharg2 are formula not in the same tree.
*/
ptinsth
ClockManager::rewriteSupnf (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vfsup;

  vfsup = getUndefClock ();
  vfsup = rewriteSup (vfsup, vharg1, vharg2);
  setNormalized (vfsup);
  return vfsup;
}

/**
  @return the class of the normalized formula (VHARG1 ^* VHARG2) 
  Assumed : vharg1 and vharg2 are formula not in the same tree.
*/
ptinsth
ClockManager::rewriteInfnf (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth vfinf;

  vfinf = getUndefClock ();
  vfinf = rewriteInf (vfinf, vharg1, vharg2);
  setNormalized (vfinf);
  return vfinf;
}

/**
   @return the class result of the merge of two classes (x1 and x2).
   It realizes recursively the induced merging.
   \see mergingClasses, hclunionwithhnull.
*/
ptinsth
ClockManager::hclmkunion (int holoph, ptinsth x1, ptinsth x2)
{
  ptinsth vhs1, vh1, vh2;

  // ptinsth_DEBUG_traceClock("entering ClockSystem::hclmkunion x1", x1);
  // ptinsth_DEBUG_traceClock("         ClockSystem::hclmkunion x2", x2);
  vh1 = getRepresentativeFormula (x1);
  vh2 = getRepresentativeFormula (x2);
  if (vh1 == vh2) return vh1;

  if (satisfiesIsAnAssumption (x1))
    {
      setAssumption (x2, pk_true);
      setAssumption (vh2, pk_true);
    }
  else if (satisfiesIsAnAssumption (x2))
    {
      setAssumption (x1, pk_true);
      setAssumption (vh1, pk_true);
    }

  if (satisfiesNullClock (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, the_nullClock_CKM_copy, pk_false);
      hclunionwithhnull (vh2);
      return the_nullClock_CKM_copy;
    }
  if (satisfiesNullClock (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, the_nullClock_CKM_copy, pk_false);
      hclunionwithhnull (vh1);
      return the_nullClock_CKM_copy;
    }

  if (satisfiesIsAnEquivalentForm (vh1))
    {
      if (satisfiesVari (vh2))
	{
	  PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
	  hclfusion (vh2, vh1);
	  removeVari (vh2);
	  return vh1;
	}
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      hclfusion (vh1, vh2);
      return vh2;
    }
  if (satisfiesIsAnEquivalentForm (vh2))
    {
      if (satisfiesVari (vh1))
	{
	  PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
	  hclfusion (vh1, vh2);
	  removeVari (vh1);
	  return vh2;
	}
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      hclfusion (vh2, vh1);
      return vh1;
    }
  // if [C] ^= [not C] then ^0 ^= ^C
  if (satisfiesSub (vh1))
    if (getComplExtraction (vh1, &vhs1))
      if (vh2 == vhs1)
	{
	  hclunionwithhnull (getClockSub (vh1));
	  return the_nullClock_CKM_copy;
	}
  // ------
  if (existHierarchyConstraintsOn (vh2, vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_true);
      if (!assumedWithoutConstraints)
	addConstraintPlacedClocks (vh1, vh2);
      return vh1;
    }
  // Assumption managing.
  if (satisfiesIsAnAssumption (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      if (!satisfiesNullDepth (vh1))
	{
	  mergingClasses (vh1, vh2, holoph);
	  return vh1;
	}
      mergingClasses (vh2, vh1, holoph);
      return vh2;

    }

  if (satisfiesIsAnAssumption (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      if (!satisfiesNullDepth (vh2))
	{
	  mergingClasses (vh2, vh1, holoph);
	  return vh2;
	}
      mergingClasses (vh1, vh2, holoph);
      return vh1;
    }

  if (isFictitiousRootFormula (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      mergingClasses (vh1, vh2, holoph);
      return vh1;
    }

  if (isFictitiousRootFormula (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      mergingClasses (vh2, vh1, holoph);
      return vh2;
    }
  if (satisfiesVari (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      mergingClasses (vh1, vh2, holoph);
      removeVari (vh2);
      return vh1;
    }
  if (satisfiesVari (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      mergingClasses (vh2, vh1, holoph);
      removeVari (vh1);
      return vh2;
    }

  if (!ModeSynchroSolving)
    {
      hcladdsynchro (vh1, vh2);
      return vh1;
    }

  if (isFormulaBeforeFormula (vh1, vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      mergingClasses (vh1, vh2, holoph);
      return vh1;
    }
  PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
  mergingClasses (vh2, vh1, holoph);
  return vh2;
}

/**
   @return the result of the merge of two clocks (classes of formula), vh1 and vh2.
   It realizes recursively the induced merges.
   \see hclmkunion.
*/
void
ClockManager::mergingClasses (ptinsth vh1, ptinsth vh2, int holoph)
{
  ptlinsth luse;
  ptinsth use;

  luse = getClockReferencesTo (vh2);
  hclfusion (vh2, vh1);
  while (luse != NILLINSTH)
    {
      ptlinsth_get_next (&luse, &use);
      switch ((int) (getOperator (use)))
	{
	case sub:
	  rewriteSub (use, vh1);
	  break;
	case nosub:
	  break;
	case hcomp:
	  break;
	case vari:
	case predef:
	  break;

	case hnot:
	  rewriteNot (use, vh1);
	  break;
	case infer:
	  use = booleanReducedInf (holoph, use, vh1, otherArgOf (use, vh2), pk_true);
	  break;
	case super:
	  use = booleanReducedSup (holoph, use, vh1, otherArgOf (use, vh2), pk_true);
	  break;

	default:
	  pK_messenger->Caseerror (__LINE__, __FILE__);
	}
    }
}
