/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return the reformulation, in "insth" (compiler type), of the placed, in BDD-tree, clock ck.
   form (in) is the formula-expression of the clock ck.
*/
  ptinsth ClockManager::recur_subterms (clock_type ck, form_type form)
{
  form_type f1, f2;
  ptinsth h1, h2, par;
  ptinsth result;

  par = CH2BDD_ck_get_horl (CH2BDD_ck_get_parent (ck));
  if (CH2BDD_form_get_op (form) == FORM_LIT)
    return CH2BDD_ck_get_horl ((clock_type) CH2BDD_form_get_literal (form));
  if (CH2BDD_form_get_op (form) == FORM_AND)
    {
      CH2BDD_form_get_args (form, &f1, &f2);
      h1 = recur_subterms (ck, f1);
      h2 = recur_subterms (ck, f2);
      result = new_binaryFormulaClock (h1, h2, infer);
      return addPlaced (result, par);
    }
  if (CH2BDD_form_get_op (form) == FORM_OR)
    {
      CH2BDD_form_get_args (form, &f1, &f2);
      h1 = recur_subterms (ck, f1);
      h2 = recur_subterms (ck, f2);
      // Ajout LB : Il faut reduire h1 + h2 sinon introduction de cycle dans 
      // h1 = ( F1 + [c]) +  h2 = [not(c)] par exemple
      // Il faudra generaliser supConditionCompl
      {
	if (!supConditionCompl (&h1, &h2))
	  supConditionCompl (&h2, &h1);
	result = new_binaryFormulaClock (h1, h2, super);
      }
      return addPlaced (result, par);
    }
  if (CH2BDD_form_get_op (form) == FORM_DIFF)
    {
      CH2BDD_form_get_args (form, &f1, &f2);
      h2 = recur_subterms (ck, f2);
      if (f1 == CH2BDD_form_one ())
	{
	  /* ceci est faux */
	  /* mais comme a` l'heure actuelle Loic n'autorise pas */
	  /* de diff dans la generation de code                 */
	  /* on genere un hcomp pour passer.                    */
	  /* de toute facon, le seul objectif est de compter le */
	  /* nombre de noeuds qu'on obtient.                    */
	  result = new_compClockInTree (h2);
	}
      else
	{
	  h1 = recur_subterms (ck, f1);
	  if (h1 == par)
	    result = new_compClockInTree (h2);
	  else
	    result = new_binaryFormulaClock (h1, getForce_notClock (h2), infer);
	}
      return addPlaced (result, par);
    }
  return NILINSTH;
}

/**
   @return the reformulation, in "insth" (compiler type), of the placed, in BDD-tree, clock ck.
*/
ptinsth ClockManager::subtermsplace (clock_type ck)
{
  ptinsth
    result;

  result = recur_subterms (ck, CH2BDD_ck_get_form (ck));
  CH2BDD_ck_update_horl (ck, result);
  return result;
}

/**
   Placing a term (h = h1 op h2) by a arborescent canonical form calculus.
   otherterm (out) is the placed formula.
   @return true if the placing has been done, false for an unification.
*/
boolean ClockManager::bddplacing (ptinsth h, ptinsth h1, ptinsth h2, int op, ptinsth * otherterm)
{
  clock_type
    ck;
  boolean
    placement;

  placement = CH2BDD_findparent (h, h1, h2, op, otherterm, &ck);
  if (placement)
    {
      (*otherterm) = subtermsplace (ck);
      h = removeFormulaFromClass (h);
      /* Sans la reformulation en insth addPlaced(h, parent); otherterm = h; */
    }
  if (h == (*otherterm))
    return pk_true;
  /* unification */
  removeLinksTo (h);
  return pk_false;
}

/**
   @return the placed formula (vharg1 ^* vharg2)  by a arborescent canonical form calculus.
   The operands vharg1 and vharg2 are assumed to be in the same tree with vhemb as branching, and not on the same branch.
*/
ptinsth ClockManager::inftreenormalize (ptinsth vharg1, ptinsth vharg2, ptinsth vhemb)
{
  ptinsth
    h,
    otherterm;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  h = new_binaryFormulaClock (vharg1, vharg2, infer);
  bddplacing (h, vharg1, vharg2, 0, &otherterm);
  treeSetAffineRelation (otherterm, vharg1, vharg2, vhemb, infer);
  return otherterm;
}

/**
   @return the placed formula (vharg1 ^+ vharg2)  by a arborescent canonical form calculus.
   The operands vharg1 and vharg2 are assumed to be in the same tree with vhemb as branching, and not on the same branch.
*/
ptinsth ClockManager::suptreenormalize (ptinsth vharg1, ptinsth vharg2, ptinsth vhemb)
{
  ptinsth
    h,
    otherterm;

  vharg1 = getRepresentativeFormula (vharg1);
  vharg2 = getRepresentativeFormula (vharg2);
  h = new_binaryFormulaClock (vharg1, vharg2, super);
  bddplacing (h, vharg1, vharg2, 1, &otherterm);
  treeSetAffineRelation (otherterm, vharg1, vharg2, vhemb, super);
  return otherterm;
}

/**
   @return the placed formula (vharginf ^- vharghnot) by a arborescent canonical form calculus.
   The operands vharg1 and vharg2 are assumed to be in the same tree and not on the same branch.
   The branching of vharginf and vharghnot is vhemb.
   vfinf is the not normalized original formula.
*/
ptinsth ClockManager::compltreenormalize (ptinsth vfinf, ptinsth vharghnot, ptinsth vharginf, ptinsth vhemb)
{
  ptinsth
    newterm,
    h,
    aux1,
    aux2;

  getInfArgs (vfinf, &aux1, &aux2);
  aux1 = getRepresentativeFormula (aux1);
  aux2 = getRepresentativeFormula (aux2);
  h = new_binaryFormulaClock (aux1, aux2, infer);
  vharginf = getRepresentativeFormula (vharginf);
  vharghnot = getRepresentativeFormula (vharghnot);
  bddplacing (h, vharginf, vharghnot, 2, &newterm);
  treeSetAffineRelation (newterm, vharginf, vharghnot, vhemb, hnot);
  return newterm;
}
