/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : an any-formula
   - vharg2 : Polynomial formula.
*/
  boolean ClockManager::testfinfs (ptinsth vharg1, ptinsth vharg2)
{
  switch ((int) (getOperator (vharg1)))
    {
    case super:
      return testsinfs (vharg1, vharg2);
    case infer:
      return testminfs (vharg1, vharg2);
    case sub:
    case nosub:
      return testvinfs (vharg1, vharg2);
    case predef:
    case hcomp:
      return pk_false;
    default:
      return pk_false;
    }
}

/**
   @return "true" if all the monomials of vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : polynomial formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testallminfs (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg1nonvide,
    okinf;
  ptinsth
    vharg11;

  do
    {
      vharg1nonvide = getSupArgs (vharg1, &vharg11, &vharg1);
      okinf = testminfs (vharg11, vharg2);
    }
  while (!(!vharg1nonvide || !okinf));
  return okinf;
}

/**
   @return "true" if it exists an extraction of  vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : monomial formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testexistvinfs (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg1nonvide;
  ptinsth
    vharg11;

  do
    {
      vharg1nonvide = getInfArgs (vharg1, &vharg11, &vharg1);
      if (testvinfs (vharg11, vharg2))
	return pk_true;
    }
  while (vharg1nonvide);
  return pk_false;
}

/**
   @return "true" if vharg1 < m where m is a monomial of vharg2, "false" otherwise.
   Assumed:
   - vharg1 : monomial formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testminfoneofm (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg2nonvide;
  ptinsth
    vharg21;

  do
    {
      vharg2nonvide = getSupArgs (vharg2, &vharg21, &vharg2);
      if (testminfm (vharg1, vharg21))
	return pk_true;
    }
  while (vharg2nonvide);
  return pk_false;
}

/**
   @return "true" if vharg1 < m where m is a monomial of vharg2, "false" otherwise.
   Assumed:
   - vharg1 : extraction formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testvinfoneofm (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg2nonvide;
  ptinsth
    vharg21;

  do
    {
      vharg2nonvide = getSupArgs (vharg2, &vharg21, &vharg2);
      if (vharg1 == vharg21)
	return pk_true;
    }
  while (vharg2nonvide);
  return pk_false;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : any formula
   - vharg2 : extraction formula.
*/
boolean ClockManager::testfinfv (ptinsth vharg1, ptinsth vharg2)
{
  switch (getOperator (vharg1))
    {
    case super:
      return testsinfv (vharg1, vharg2);
    case infer:
      return testminfv (vharg1, vharg2);
    case nosub:
    case sub:
      return testvinfv (vharg1, vharg2);
    case predef:
    case hcomp:
      return pk_false;
    default:
      return pk_false;
    }
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : any formula
   - vharg2 : monomial formula.
*/
boolean ClockManager::testfinfm (ptinsth vharg1, ptinsth vharg2)
{
  switch (getOperator (vharg1))
    {
    case super:
      return testsinfm (vharg1, vharg2);
    case infer:
      return testminfm (vharg1, vharg2);
    case sub:
    case nosub:
      return testvinfm (vharg1, vharg2);
    case predef:
    case hcomp:
      return pk_false;
    default:
      return pk_false;
    }
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg2 < one of args of vharg1)
   Assumed:
   - vharg1 : polynomial formula
   - vharg2 : monomial formula.
*/
boolean ClockManager::testsinfm (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg1nonvide;
  ptinsth
    vharg11,
    vhemb,
    vhson1,
    vhson2;

  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (vhemb == vharg2);
  if (vhson1 != vharg1)
    return testfinfm (vhson1, vharg2);
  vharg1nonvide = getSupArgs (vharg1, &vharg11, &vharg1);
  if (!testminfm (vharg11, vharg2))
    return pk_false;
  if (vharg1nonvide)
    return testsinfm (vharg1, vharg2);
  return pk_true;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg1 < one of args of vharg2)
   Assumed:
   - vharg1 : polynomial formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testsinfs (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    vhemb,
    vhson1,
    vhson2;

  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (vhemb == vharg2);
  if (vharg1 == vhson1)
    {
      if (vharg2 == vhson2)
	return testallminfs (vharg1, vharg2);
      return pk_false;
    }
  if (vharg2 == vhson2)
    {
      if (testallminfs (vharg1, vharg2))
	return pk_true;
      return testfinfs (getFather (vharg1), vharg2);
    }
  return testfinfs (vhson1, vharg2);
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg2 < one of args of vharg1)
   Assumed:
   - vharg1 : polynomial formula
   - vharg2 : extraction formula.
*/
boolean ClockManager::testsinfv (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg1nonvide;
  ptinsth
    vharg11,
    vhemb,
    vhson1,
    vhson2;

  if (getOperator (vharg2) == hcomp)
    return pk_false;
  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (boolean) (vhemb == vharg2);
  if (vhson1 != vharg1)
    return testfinfv (vhson1, vharg2);
  vharg1nonvide = getSupArgs (vharg1, &vharg11, &vharg1);
  if (!testminfv (vharg11, vharg2))
    return pk_false;
  if (vharg1nonvide)
    return testsinfv (vharg1, vharg2);
  return pk_true;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg1 < one of monomials of vharg2)
   Assumed:
   - vharg1 : monomial formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testminfs (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    bres;
  ptinsth
    vhson1,
    vhson2,
    vhemb;

  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (vhemb == vharg2);
  bres = pk_false;
  if (vharg1 == vhson1)
    {
      if (vharg2 == vhson2)
	{
	  if (testminfoneofm (vharg1, vharg2))
	    bres = pk_true;
	  else
	    bres = testexistvinfs (vharg1, vharg2);
	}
      else
	bres = testexistvinfs (vharg1, vharg2);
    }
  else if (vharg2 == vhson2)
    {
      if (testminfoneofm (vharg1, vharg2))
	bres = pk_true;
      else
	bres = testfinfs (getFather (vharg1), vharg2);
    }
  else
    bres = testfinfs (vhson1, vharg2);
  return bres;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg2 < one of args of vharg1)
   Assumed:
   - vharg1 : monomial formula
   - vharg2 : monomial formula.
*/
boolean ClockManager::testminfm (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg2nonvide;
  ptinsth
    vharg21,
    vhemb,
    vhson1,
    vhson2;

  if (getOperator (vharg1) == hcomp)
    return pk_false;
  if (getOperator (vharg2) == hcomp)
    return pk_false;
  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (vhemb == vharg2);
  if (vhson1 != vharg1)
    return testfinfm (vhson1, vharg2);
  vharg2nonvide = getInfArgs (vharg2, &vharg21, &vharg2);
  if (!testminfv (vharg1, vharg21))
    return pk_false;
  if (vharg2nonvide)
    return testminfm (vharg1, vharg2);
  return pk_true;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg2 < one of args of vharg1)
   Assumed:
   - vharg1 : monomial formula
   - vharg2 : extraction formula.
*/
boolean ClockManager::testminfv (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    vharg1nonvide;
  ptinsth
    vharg11,
    vhemb,
    vhson1,
    vhson2,
    vhaux;

  if (getOperator (vharg2) == hcomp)
    return pk_false;
  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (boolean) (vhemb == vharg2);
  if (vhson1 == vharg1)
    {
      vharg1nonvide = getInfArgs (vharg1, &vharg11, &vhaux);
      if (testvinfv (vharg11, vharg2))
	return pk_true;
      if (vhaux == vharg1)
	return pk_false;
      if (vharg1nonvide)
	return testminfv (vhaux, vharg2);
      return pk_false;
    }
  return testfinfv (vhson1, vharg2);
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg1 < one of monomials of vharg2)
   Assumed:
   - vharg1 : extraction formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testvinfs (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    vhson1,
    vhson2,
    vhemb;

  if (getOperator (vharg1) == hcomp)
    return pk_false;
  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (vhemb == vharg2);
  if (vhson1 == vharg1)
    {
      if (vharg2 == vhson2)
	return testvinfoneofm (vharg1, vharg2);
      return pk_false;
    }
  if (vharg2 == vhson2)
    {
      if (testvinfoneofm (vharg1, vharg2))
	return pk_true;
      return testfinfs (getFather (vharg1), vharg2);
    }
  return testfinfs (vhson1, vharg2);
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg1 < one of args of vharg2)
   Assumed:
   - vharg1 : extraction formula
   - vharg2 : monomial formula.
*/
boolean ClockManager::testvinfm (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    vhemb,
    vhson1,
    vhson2;

  if (getOperator (vharg1) == hcomp)
    return pk_false;
  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (vhemb == vharg2);
  if (vhson1 == vharg1)
    return pk_false;
  return testfinfm (vhson1, vharg2);
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   (same branch and vharg2 above vharg1, or vharg2 < one of args of vharg1)
   Assumed:
   - vharg1 : extraction formula
   - vharg2 : extraction formula.
*/
boolean ClockManager::testvinfv (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    vhemb,
    vhson1,
    vhson2;

  if (!hoclembr (vharg1, vharg2, &vhson1, &vhson2, &vhemb))
    return (boolean) (vhemb == vharg2);
  if (vhson1 == vharg1)
    return pk_false;
  else
    return testfinfv (vhson1, vharg2);
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : any formula with vop1 as clock type.
   - vharg2 : any formula with vop2 as clock type.
*/
boolean ClockManager::testfinff (ptinsth vharg1, ptinsth vharg2, pkClockType vop1, pkClockType vop2)
{
  if (vop1 == super)
    {
      if (vop2 == super)
	return testsinfs (vharg1, vharg2);
      if (vop2 == infer)
	return testsinfm (vharg1, vharg2);
      if (vop2 == hcomp)
	return pk_false;
      return testsinfv (vharg1, vharg2);
    }
  if (vop1 == infer)
    {
      if (vop2 == super)
	return testminfs (vharg1, vharg2);
      if (vop2 == infer)
	return testminfm (vharg1, vharg2);
      if (vop2 == hcomp)
	return pk_false;
      return testminfv (vharg1, vharg2);
    }
  if (vop1 == hcomp)
    return pk_false;

  /* vharg1 denote une extraction */
  {
    if (vop2 == super)
      return testvinfs (vharg1, vharg2);
    if (vop2 == infer)
      return testvinfm (vharg1, vharg2);
    if (vop2 == hcomp)
      return pk_false;
    return testvinfv (vharg1, vharg2);
  }
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : extraction formula
   - vharg2 : any formula.
*/
boolean ClockManager::testvinff (ptinsth vharg1, ptinsth vharg2)
{
  ptinsth
    vh1,
    vh2;
  boolean
    b;
  noperators
    vop2;

  vop2 = getOperator (vharg2);
  b = pk_false;
  if (vop2 == super)
    if (!satisfiesMarked (vharg2, mark2))
      {
	addMark (vharg2, mark2);
	getSupArgs (vharg2, &vh1, &vh2);
	if (isIncludedIn (vharg1, vh1))
	  b = pk_true;
	else
	  b = isIncludedIn (vharg1, vh2);
	removeMark (vharg2, mark2);
      }
  if (!b)
    return isIncludedIn (getClockSub (vharg1), vharg2);
  return pk_true;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : monomial formula
   - vharg2 : any formula.
*/
boolean ClockManager::testminff (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    bres;
  ptinsth
    vh1,
    vh2;

  bres = pk_false;
  switch ((int) (getOperator (vharg2)))
    {
    case predef:
    case hcomp:
    case hnot:
      bres = pk_false;
      break;
    case vari:
    case sub:
    case nosub:
      addMark (vharg1, mark1);
      getInfArgs (vharg1, &vh1, &vh2);
      if (isIncludedIn (vh1, vharg2))
	bres = pk_true;
      else
	bres = isIncludedIn (vh2, vharg2);
      removeMark (vharg1, mark1);
      break;
    case infer:
      if (satisfiesMarked (vharg2, mark2))
	bres = pk_false;
      else
	{
	  addMark (vharg2, mark2);
	  getInfArgs (vharg2, &vh1, &vh2);
	  if (isIncludedIn (vharg1, vh1))
	    bres = isIncludedIn (vharg1, vh2);
	  else
	    bres = pk_false;
	  removeMark (vharg2, mark2);
	}
      break;
    case super:
      bres = testfinfSup (vharg1, vharg2);
      break;
    default:;
    }
  return bres;
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : polynomial formula 
   - vharg2 : any formula.
*/
boolean ClockManager::testsinff (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    bres;
  ptinsth
    vh1,
    vh2;

  addMark (vharg1, mark1);
  getSupArgs (vharg1, &vh1, &vh2);
  if (isIncludedIn (vh1, vharg2))
    bres = isIncludedIn (vh2, vharg2);
  else
    bres = pk_false;
  removeMark (vharg1, mark1);
  return bres;
}

/** 
    Inclusion test (partial implementation). It depends on the global parameter treeOperationAuthorized.
    @return "true" if vharg1 is included in vharg2 (vharg1 < vharg2), "false" otherwise.
*/
boolean ClockManager::isIncludedIn (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    b;

  if (treeOperationAuthorized)
    b = testfsupfintree (vharg2, vharg1);
  else
    b = (vharg1 == vharg2);
  if (b)
    return pk_true;

  switch (getOperator (vharg1))
    {
    case predef:
    case hcomp:
    case hnot:
      return pk_false;

    case vari:
      return testxinff (vharg1, vharg2);

    case sub:
    case nosub:
      return testvinff (vharg1, vharg2);

    case infer:
      if (satisfiesMarked (vharg1, mark1))
	return pk_false;
      return testminff (vharg1, vharg2);

    case super:
      if (!satisfiesMarked (vharg1, mark1))
	return testsinff (vharg1, vharg2);
      if (satisfiesMarked (vharg1, mark2))
	return pk_false;
      if (getOperator (vharg2) != super)
	return pk_false;
      return testfinfSup (vharg1, vharg2);
    default:
      pK_messenger->Caseerror (__LINE__, __FILE__);
      return pk_false;
    }

}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
   - vharg1 : any formula
   - vharg2 : polynomial formula.
*/
boolean ClockManager::testfinfSup (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    bres;
  ptinsth
    vh1,
    vh2;

  if (satisfiesMarked (vharg2, mark2))
    return pk_false;
  addMark (vharg2, mark2);
  getSupArgs (vharg2, &vh1, &vh2);
  if (isIncludedIn (vharg1, vh1))
    bres = pk_true;
  else
    bres = isIncludedIn (vharg1, vh2);
  removeMark (vharg2, mark2);
  return bres;
}

/** @return "true" if vh1 and vh2 are in the same tree and VH1 > VH2, "false" otherwise. */
boolean ClockManager::testfsupfintree (ptinsth vh1, ptinsth vh2)
{
  ptinsth
    svh1,
    vhson1,
    vhson2,
    vhemb;

  svh1 = vh1;
  if (getCompleteBranching (&vh1, &vh2, &vhson1, &vhson2, &vhemb))
    {
      if (svh1 == vh1)
	return testfinff (vhson2, vh1, getOperator (vhson2), getOperator (vh1));
      return testfinff (vhson1, vh2, getOperator (vhson1), getOperator (vh2));
    }
  if (satisfiesUndefClock (vhemb))
    return pk_false;
  return (vhemb == svh1);
}

/**
   @return "true" if vharg1 < vharg2, "false" otherwise.
   Assumed:
     - vharg1: a free variable.
     - vharg2: any formula.
*/
boolean ClockManager::testxinff (ptinsth vharg1, ptinsth vharg2)
{
  boolean
    Bresult;
  ptinsth
    vh1,
    vh2;
  pkClockType
    vop;

  if (satisfiesMarked (vharg2, mark2))
    return pk_false;
  vop = getOperator (vharg2);
  if (vop == infer)
    {
      addMark (vharg2, mark2);
      getInfArgs (vharg2, &vh1, &vh2);
      if (isIncludedIn (vharg1, vh1))
	Bresult = isIncludedIn (vharg1, vh2);
      else
	Bresult = pk_false;
      removeMark (vharg2, mark2);
    }
  else if (vop == super)
    {
      addMark (vharg2, mark2);
      getSupArgs (vharg2, &vh1, &vh2);
      if (isIncludedIn (vharg1, vh1))
	Bresult = pk_true;
      else
	Bresult = isIncludedIn (vharg1, vh2);
      removeMark (vharg2, mark2);
    }
  else
    Bresult = pk_false;
  return Bresult;
}
