/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
  Construtor of a condition descriptor.
  vinsth   : Positive extraction formula of the condition (when vexpcal)
  vinsthns : Negative extraction formula of the condition (when not vexpcal)
  vhorl    : Clock of the condtion (^expcal)
  vexpcal  : expression of the condition 
  linst    : Signals referenced in the condtion.
*/
  condition::condition (ptinsth vinsth, ptinsth vinsthns, ptinsth vhorl, ptexpcal vexpcal, ptlinstances linst)
{
  ophcond = vhorl;
  ophsub = vinsth;
  ophnosub = vinsthns;
  linkExpression (vexpcal, linst);
}

condition::~condition ()
{

}

/**
   It assigns to the current condition the expression and the signals list referenced in the condtion.
*/
void
condition::linkExpression (ptexpcal vexpcal, ptlinstances linst)
{
  expcond = vexpcal;
  lsign = linst;
}

/**
  Modifying a condition.
  @param argh  (in)    : new clock of the condition 
  @param vhnot (out) : negative extraction clock of the condition 
  @param vhused (out) : old clock of the condition
  @return "true" if the current clock of the condition is not argh, "false" otherwise (modified).
*/
boolean condition::set_condsub (ptinsth argh, ptinsth * vhnot, ptinsth * vhused)
{
  (*vhused) = ophcond;
  if ((*vhused) == argh)
    return pk_false;
  set_clkcond (argh);
  (*vhnot) = ophnosub;
  return pk_true;
}

/** It assigns the clock argh as the clock of the condition.
 */
void
condition::set_clkcond (ptinsth argh)
{
  this->ophcond = argh;
}

/** @return the clock of the condition. */
ptinsth condition::get_clkcond ()
{
  return ophcond;
}

/** @return the sets of signals referenced in the condition.*/
ptlinstances condition::get_lsign ()
{
  return lsign;
}

/**
   @return the signal expression associated  to the condition.
*/
ptexpcal condition::get_condsub ()
{
  return expcond;
}

/** @return the positive extraction associated  to the condition.  */
ptinsth condition::get_clksub ()
{
  return ophsub;
}

/** @return the negative extraction associated  to the condition. */
ptinsth condition::get_clknosub ()
{
  return ophnosub;
}
