/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

struct GraphInfos 
{
  ptlinstances lsig;    ///< Set of synchronous signals.
  ptlinstances lsigkept;    ///< Set of synchronous signals, removed from the graph kept for tracability.
  ptgfd  gfd;           ///< graph assigned to the clock.
  ptlsucc  lsucci;      ///< dependences : set of successors of the clock
  ptINode  ptINodebc;   ///< Array of process link
  ptinterface interface;  
};

struct insth : public ObjectCommonInfos
{
  ptlinstances luse;          ///< List of signal-expressions that reference this clock.
  ptGraphInfos  vGraphInfos;  ///< Acces to the graph-informations (see below)
  /*-- Access hierachy -- */
  ptinsth father, root;    ///< mother clock, root clock
  int  vdepth;           /* depth of the clock in its tree. */ 
  ptinsth  hocompl;        /* complementary clock (not always available) */
  ptlinsth  children;      /* set of clock children */

  ptident  identh;         /* Identifier */
  ptident  eventSignal;    /* A user name of a signal of event type ( used to unify clock and signal ) */ 
  /*-- class access -- */
  ptinsth  clsucc;         /* Chain of formula of the class (circular chain) */ 
  bool isAReference;    /* true when this formula has been removed from the class list  */ 
  boolean clhead;          /* Representative element of the classs (when it is true)  */ 
  boolean normform;        /* At true, it specifies that the formula is normalized. */ 
  ptlinsth  lsucch;        /* List of the uses of the formula. */ 
  ptlinsth  orderInfos;    /* liste des horloges plus petites (non transitivement) */ 

  /*-- Formula Access --*/
  boolean isAnEquivalentForm; /* To keep the equivalences of formulae */
  boolean isAnAssumption;  /* true for assertions. */

  ptAffineRelation affr;  /* Affine relation (1, phi, d) with phi >= 0  */
  pkClockType  ophorl;    /* type of the formula (cf enumerated type pkClockType) */ 
  union
  {
    ptcondition  opsub1; /* extraction on condition */
    ptinsth  op1;        /* hnot, hcomp (complement in its mother clock) */
    struct      /* upper/lower bound, difference */
    {
      ptinsth  op21;     
      ptinsth  op22;
    } binOp;

  }Operand;
};

#define INSTH_GETGRAPHINFOS(ANINSTH) ((ANINSTH)->vGraphInfos)
#define INSTH_SETGRAPHINFOS(ANINSTH, AGRAPH) INSTH_GETGRAPHINFOS(ANINSTH) = (AGRAPH)

#define INSTH_GETIDENT(ANINSTH) ((ANINSTH)->identh)
#define INSTH_SETIDENT(ANINSTH, ANIDENT) INSTH_GETIDENT(ANINSTH) = (ANIDENT)
#define INSTH_SETNILIDENT(ANINSTH) INSTH_SETIDENT(ANINSTH, NILIDENT)

#define INSTHFORM_ISHEADER(ANINSTH) ( (ANINSTH)->clhead )
#define INSTHFORM_SETHEADERSTATUS(ANINSTH, ISHEAD) ((ANINSTH)->clhead) = ISHEAD
#define INSTHFORM_SETHEADER(ANINSTH) INSTHFORM_SETHEADERSTATUS(ANINSTH, true)
#define INSTHFORM_SETCOMMON(ANINSTH) INSTHFORM_SETHEADERSTATUS(ANINSTH, false)

#define INSTHFORM_GETSUCC(ANINSTH) ( (ANINSTH)->clsucc )
#define INSTHFORM_SETSUCC(ANINSTH, NEXT) INSTHFORM_GETSUCC(ANINSTH) = (NEXT)
#define INSTHFORM_SETSUCC_SUCCOF(ANINSTH, PREDOFNEXT) INSTHFORM_SETSUCC(ANINSTH, INSTHFORM_GETSUCC(PREDOFNEXT))
#define INSTHFORM_SUCC_IS(ANINSTH, TARGETINSTH) ( (INSTHFORM_GETSUCC(ANINSTH)) == TARGETINSTH )
#define INSTHFORM_ISALONE(ANINSTH) ( INSTHFORM_SUCC_IS(ANINSTH, ANINSTH) )
#define INSTHFORM_SETALONE(ANINSTH) INSTHFORM_SETSUCC(ANINSTH, ANINSTH)
#define INSTHFORM_NEXT(ANINSTH) (ANINSTH) = INSTHFORM_GETSUCC(ANINSTH)

#define INSTHFORM_ISAREFERENCE(ANINSTH) ( ((ANINSTH)->isAReference) )
#define INSTHFORM_SETASREFERENCE(ANINSTH) ((ANINSTH)->isAReference) = true
#define INSTHFORM_SETASINCLASSLIST(ANINSTH) ((ANINSTH)->isAReference) = false

#define INSTHFORM_ISSATISFIED(ANINSTH) ((ANINSTH)->isAnEquivalentForm)
#define INSTHFORM_SETSATISFIED(ANINSTH) ((ANINSTH)->isAnEquivalentForm)=true
#define INSTHFORM_SETTOBESATISFIED(ANINSTH) ((ANINSTH)->isAnEquivalentForm)=false

#define INSTHFORM_GETOPERATOR(AFORMULA) ((AFORMULA)->ophorl)
#define INSTHFORM_SETOPERATOR(AFORMULA, ANOPERATOR) INSTHFORM_GETOPERATOR(AFORMULA) = ANOPERATOR
#define INSTHFORM_HASOPERATOR(AFORMULA, ANOPERATOR) (INSTHFORM_GETOPERATOR(AFORMULA) == ANOPERATOR)
#define INSTHFORM_SETFREEFORMULA(ANINSTH) INSTHFORM_GETOPERATOR(ANINSTH) = vari
#define INSTHFORM_ISFREEFORMULA(ANINSTH) (INSTHFORM_HASOPERATOR(ANINSTH, vari))
#define INSTHCLOCK_SETMETAROOT(ACLOCK) INSTHFORM_SETOPERATOR(ACLOCK, troot)
#define INSTHCLOCK_ISMETAROOT(ACLOCK) ( INSTHFORM_HASOPERATOR(ACLOCK, troot))
#define INSTHFORM_ISMETAROOT(AFORM) ( INSTHCLOCK_ISMETAROOT(ptinsth_get_clhead(AFORM)) )
#define INSTHFORM_ISSUB(AFORMULA) ( INSTHFORM_HASOPERATOR(AFORMULA, sub) )
#define INSTHFORM_ISNOSUB(AFORMULA) ( INSTHFORM_HASOPERATOR(AFORMULA,  nosub) )
#define INSTHFORM_ISNOT(AFORMULA) ( INSTHFORM_HASOPERATOR(AFORMULA, hnot) )
#define INSTHFORM_ISCOMP(AFORMULA) ( INSTHFORM_HASOPERATOR(AFORMULA, hcomp) )
#define INSTHFORM_ISINFER(AFORMULA) ( INSTHFORM_HASOPERATOR(AFORMULA, infer) )
#define INSTHFORM_ISSUPER(AFORMULA) ( INSTHFORM_HASOPERATOR(AFORMULA, super) )

#define INSTHFORM_GETUNARYARG(AFORMULA) ((AFORMULA)->Operand.op1)
#define INSTHFORM_SETUNARYARG(AFORMULA,ACLOCK) INSTHFORM_GETUNARYARG(AFORMULA) = (ACLOCK)
#define INSTHFORM_GETBINARYARG1(AFORMULA) ((AFORMULA)->Operand.binOp.op21)
#define INSTHFORM_SETBINARYARG1(AFORMULA,ACLOCK) INSTHFORM_GETBINARYARG1(AFORMULA) = (ACLOCK)
#define INSTHFORM_GETBINARYARG2(AFORMULA) ((AFORMULA)->Operand.binOp.op22)
#define INSTHFORM_SETBINARYARG2(AFORMULA,ACLOCK) INSTHFORM_GETBINARYARG2(AFORMULA) = (ACLOCK)

#define INSTHFORM_GETCONDITION(AFORMULA) ((AFORMULA)->Operand.opsub1)
#define INSTHFORM_SETCONDITION(AFORMULA, ACONDITION) INSTHFORM_GETCONDITION(AFORMULA)=ACONDITION

#define INSTHCLOCK_GETFATHER(ACLOCK) ((ACLOCK)->father)
#define INSTHCLOCK_SETFATHER(ACLOCK, AFATHERCLOCK) (INSTHCLOCK_GETFATHER(ACLOCK)) = (AFATHERCLOCK)
#define INSTHCLOCK_SETNILFATHER(ACLOCK) INSTHCLOCK_SETFATHER(ACLOCK,NILINSTH)
#define INSTHCLOCK_HASNILFATHER(ACLOCK) (INSTHCLOCK_GETFATHER(ACLOCK) == NILINSTH)
#define INSTHCLOCK_SETSELFFATHER(ACLOCK) INSTHCLOCK_SETFATHER(ACLOCK,ACLOCK)
#define INSTHCLOCK_HASSELFFATHER(ACLOCK) ( INSTHCLOCK_GETFATHER(ACLOCK) == ACLOCK )
#define INSTHCLOCK_SETROOT(OPER, ACLOCK)   if (OPER == predef) INSTHCLOCK_SETSELFFATHER(ACLOCK); else INSTHCLOCK_SETNILFATHER(ACLOCK);

#define INSTHCLOCK_ISROOT(ACLOCK) ( INSTHCLOCK_HASNILFATHER(ACLOCK) || INSTHCLOCK_HASSELFFATHER(ACLOCK) )
#define INSTHCLOCK_HASSELFFATHERFORM(ACLOCK) ( ptinsth_get_clhead(INSTHCLOCK_GETFATHER(ACLOCK)) == ACLOCK )
#define INSTHFORM_ISROOT(ACLOCK) ( INSTHCLOCK_HASNILFATHER(ACLOCK) || INSTHCLOCK_HASSELFFATHER(ACLOCK) )

/* Clock list structure */
typedef struct linsth
{
  ptinsth  elem;  /* list element */
  ptlinsth  next; /* next element */
}  tlinsth;



