/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   It initializes
   - the set of pragmas to NILLPRAGMAS
   - the set of marks to NILLMARK
   - the number of the object to 0.
   - the user data to  NILLATTRIBUTES.
   - the "whoami" member  to undefObject
*/
  ObjectCommonInfos::ObjectCommonInfos ()
{
  num = 0;
  clear ();
}


void
ObjectCommonInfos::clear ()
{
  vpragmas = NILLPRAGMAS;
  lmark = NILLMARK;
  userData = NILLATTRIBUTES;
  whoami = undefObject;
}

/**
   Destructor.
*/
ObjectCommonInfos::~ObjectCommonInfos ()
{
}


/**
   @return the list of pragmas of the object.
*/
ptlpragmas
ObjectCommonInfos::getPragmas ()
{
  return vpragmas;
}

/**
   It defines by l the pragmas member of an ObjectCommonInfos object.
*/
void
ObjectCommonInfos::setPragmas (ptlpragmas l)
{
  vpragmas = l;
}

/**
   It adds aPragma to the list of pragmas of an ObjectCommonInfos object.
   @return "true" if the pragma is added, "false" otherwise (i.e. it is already associated with the object).
*/
boolean
ObjectCommonInfos::addPragma (ptpragma aPragma)
{
  if (existPragmaLike (vpragmas, aPragma))
    return pk_false;
  ptlPointer_insert_elem (&vpragmas, (pointer) aPragma);
  return pk_true;
}

/**
   It removes the pragma called name from the list of pragmas of an ObjectCommonInfos object.
   @return "true" if the pragma is in the list of the pragmas, "false" otherwise.
*/
boolean
ObjectCommonInfos::removePragma (ptrepstr name)
{
  ptpragma aPragma;
  boolean b;

  b = pk_false;
  while (getPragmaStrKeyword (vpragmas, name, &aPragma))
    {
      ptlPointer_remove_elem (&vpragmas, (pointer) aPragma);
      b = pk_true;
    }

  return b;
}

/**
   It adds each pragma of pragmas to the list of pragmas of an ObjectCommonInfos object.
*/
void
ObjectCommonInfos::addPragmas (ptlpragmas pragmas)
{
  ptpragma apragma;
  while (ptlpragmas_get_next (&pragmas, &apragma))
    addPragma (apragma);
}

/**
   It adds a mark (val) to the list of pragmas of an ObjectCommonInfos object.
*/
void
ObjectCommonInfos::addMark (typeMark val)
{
  ptlMark_insert_elem (&lmark, val);
}


/**
   @return "true" if the obejct is marked by a mark (val), "false" otherwise.
*/
boolean ObjectCommonInfos::satisfiesMarked (typeMark val)
{
  return ptlMark_IsIn (lmark, val);
}

/**
   It removes a mark (val) from the list of marks of an ObjectCommonInfos object.
*/
void
ObjectCommonInfos::removeMark (typeMark val)
{
  ptlMark_removeMark (&lmark, val);
}

/**
   It sets to val the numer of the object (used to sort).
*/
void
ObjectCommonInfos::setnum (int val)
{
  num = val;
}

/**
   @return the number associated with the current object.
*/
int
ObjectCommonInfos::getnum ()
{
  return num;
}

/**
   It defines by val the kind of the object.
*/
void
ObjectCommonInfos::setwhoami (objectType val)
{
  whoami = val;
}

/**
   @return the kind of the object.
*/
objectType ObjectCommonInfos::getwhoami ()
{
  return whoami;
}


/* @return 
   - "true" if the object if attributed with the attribute number num, valAttribute is the value of this attribute,
   - "false" otherwise, valAttribute is not modified
*/
boolean ObjectCommonInfos::getAttribute (int num, attributeValue * valAttribute)
{
  return ptlAttributes_get (userData, num, valAttribute);
}

/**
   It adds an attribute (num) with vattr as value to the object.
*/
void
ObjectCommonInfos::addAttribute (int num, attributeValue vattr)
{
  userData = ptlAttributes_insert (userData, num, vattr);
}

/**
   It removes the num-th attribute from the set of attributes of the object.
*/
void
ObjectCommonInfos::removeAttribute (int num)
{
  userData = ptlAttributes_remove (userData, num);
}
