/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** 
    @return "true" if an interface (vinterface) denotes the undefined interface, "false" otherwise.
 */
extern boolean
ptinterface_is_undefined (ptinterface vinterface)
{
  return (vinterface == NILINTERFACE);
}

/** 
    @return "true" if a clock (vh) is in the set of the output clocks of an interface(vinterface), "false" otherwise.
*/
extern boolean
ptinterface_is_outputClock (ptinterface vinterface, ptinsth vh)
{
  if (ptinterface_is_undefined (vinterface))
    return pk_false;
  return ptlinsth_sorted_isin_elem (vinterface->outClocks, vh);
}

/** 
    @return the set of the local clocks assigned to an interface(vinterface).
*/
extern ptlinsth
ptinterface_get_localClocks (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTH;
  return vinterface->localClocks;
}

/** 
    @return "true" if a clock (vh) is in the set of the local clocks of an interface(vinterface), "false" otherwise.
*/
extern boolean
ptinterface_is_localClock (ptinterface vinterface, ptinsth vh)
{
  if (ptinterface_is_undefined (vinterface))
    return pk_false;
  return ptlinsth_sorted_isin_elem (vinterface->localClocks, vh);
}

/** 
    @return the set of the output clocks assigned to an interface(vinterface).
*/
extern ptlinsth
ptinterface_get_outputClocks (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTH;
  return vinterface->outClocks;
}



/** 
    @return the set of the input clocks assigned to an interface(vinterface).
*/
extern ptlinsth
ptinterface_get_inputClocks (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTH;
  return vinterface->inClocks;
}



/** 
    @return the set of the parameters assigned to an interface(vinterface).
*/
extern ptlinstances
ptinterface_get_parameters (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTANCES;
  return vinterface->lpar;
}


/** 
    @return the set of the constants assigned to an interface(vinterface).
*/
extern ptlinstances
ptinterface_get_constants (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTANCES;
  return vinterface->lconstants;
}

/** 
    @return the set of the input signals assigned to an interface(vinterface).
*/
extern ptlinstances
ptinterface_get_inputSignals (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTANCES;
  return vinterface->inSignals;
}

/** 
    @return the set of the output signals assigned to an interface(vinterface).
*/
extern ptlinstances
ptinterface_get_outputSignals (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTANCES;
  return vinterface->outSignals;
}

/** 
    @return the set of the input-output signals (shared signals) assigned to an interface(vinterface).
*/
extern ptlinstances
ptinterface_get_inOutSignals (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTANCES;
  return vinterface->inOutSignals;
}


/** 
    @return the set of the local signals assigned to an interface(vinterface).
*/
extern ptlinstances
ptinterface_get_localSignals (ptinterface vinterface)
{
  if (ptinterface_is_undefined (vinterface))
    return NILLINSTANCES;
  return vinterface->localSignals;
}
