/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* @name Expression Manager Class.
   An Expression Manager is assigned to each Graph.
*/
#include "common_type.h"
class ExprManager
{ 
#include "ExprMake_m.h"
#include "ExprMake_g.h"
#include "SpatialProcessing_i.h"
#include "TypeChecking_i.h"
#include "mkEvaluations_i.h"
#include "ExprRewrite_i.h"
#include "ExprRewrite_g.h"
#include "ExprAccessSet_g.h"
#include "ExprAccessGet_g.h"
#include "ExprAttributes_g.h"
#include "SpatialProcessing_g.h"
#include "ExprMisc_g.h"
#include "TypeChecking_g.h"
#include "ExprManager_g.h"

private:
  ptlexpcal table[SIG_NB_OPERATORS];    ///< table of expressions classified by operators.
  ptlexpcal tableEnd[SIG_NB_OPERATORS]; ///< table of the last element of the previous tabel (table).
    ptlinstances lIndexForInit;  ///> list of index used for initialisations.

  ptinsth hundef; ///< undefined clock (from clockmanager).
  ptlexpcal lexptotest, lexpmodified, lequ1exp, lequ2exp; ///< Internal lists used for rewritings.

  ptNameManager aNameManager;  ///< link to a name manager.
  ptNodeManager NM;            ///< link to a node manager.
  ptClockManager CKM;          ///< link to a clock manager.
  ptTraceabilityManager TM;  ///< link to a traceability manager.

  static ptexpcal ptexpcal_FALSE;    ///< internal representation of "false" expression
  static ptexpcal ptexpcal_TRUE;    ///< internal representation of "true" expression
  static ptexpcal ptexpcal_0;    ///< internal representation of "0" expression
  static ptexpcal ptexpcal_1;   ///< internal representation of "1" expression
  int nbIdent;      ///< Internal counter (number of created identifiers).

  boolean VerifForEachDimensionOf; ///< global for type checking.
  int lastTypeCheckingelem;    ///< global for type checking (see setTypeCheckingOutOfBounds())
  boolean bunifexp, modeEval, modeNoEvents, modeArithmRules, modeTypeCheckingOutOfBounds; ///< Internal modes of the manager.
  ptlINode lModifiedNodes; ///< list of the modified nodes during the rewriting of the expressions.
};

