/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Getting attribute. 
   @param vexp (in) any signal expression. 
   @param num (in) an attribute number. 
   @param valAttribute (out) the value of the ?num-th attribute if it exists. 
   @return "true" if teh expression vexp is annotated by the ?num-th attribute, "false" otherwise. 
*/
boolean ExprManager::getAttribute (ptexpcal vexp, int num, attributeValue * valAttribute)
{
  return ptexpcal_getAttribute (vexp, num, valAttribute);
}

/** Adding an attribute. 
   @param vexp (in) any signal expression. 
   @param num (in) an attribute number. 
   @param vattr (in) the value (void*) of the attribute. 
   It adds the attribute (num, vattr) to the expression vexp. 
*/
void
ExprManager::addAttribute (ptexpcal vexp, int num, attributeValue vattr)
{
  ptexpcal_addAttribute (vexp, num, vattr);
}

/**
   Removing an attribute. 
   @param vexp (in) any signal expression. 
   @param num (in) an attribute number. 
   It removes the num-th attribute assigned to the expression vexp. 
*/
void
ExprManager::removeAttribute (ptexpcal vexp, int num)
{
  ptexpcal_removeAttribute (vexp, num);
}


/** Testing flag. 
  @param vexp (in) any signal expression. 
  @param num (in) an attribute number. 
  @return "true" if the expression vexp is annotated by the num-th mark, "false" otherwise. 
*/
boolean ExprManager::isMarked (ptexpcal vexp, int num)
{
  return ptexpcal_satisfies_marked (vexp, num);
}

/**
   Adding flag. 
   @param vexp (in) any signal expression. 
   @param num (in) a mark. 
   It adds the mark num to the expression vexp.   
*/
void
ExprManager::addMark (ptexpcal vexp, int num)
{
  ptexpcal_addMark (vexp, num);
}

/**
   Removing flag. 
   @param vexp (in) any signal expression. 
   @param num (in) a mark. 
   It removes the mark num to the expression vexp.   
*/
void
ExprManager::removeMark (ptexpcal vexp, int num)
{
  ptexpcal_removeMark (vexp, num);
}


/** Getting pragmas. 
   @param vexp (in) any signal expression. 
   @return the pragmas assigned to the expression. 
*/
ptlpragmas ExprManager::getPragmas (ptexpcal vexp)
{
  return ptexpcal_get_pragmas (vexp);
}

/**
  Adding a pragma. 
  @param vexp (in) any signal expression. 
  @param aPragma (in) a pragma. 
  It adds the pragma aPragma to the pragmas assigned to the vexp expression. 
*/
void
ExprManager::addPragma (ptexpcal vexp, ptpragma aPragma)
{
  ptexpcal_addPragma (vexp, aPragma);
}

/**
  Removing a pragma. 
  @param vexp (in) any signal expression. 
  @param name (in) assumed to be a name of a pragma. 
  It removes the pragma named name from the pragmas assigned to the vexp expression.
  @return "true" if the operation succeeds, "false" otherwise.
*/
boolean ExprManager::removePragma (ptexpcal vexp, ptrepstr name)
{
  return ptexpcal_removePragma (vexp, name);
}

/** Removing a mark/attribute for a list of expressions.
   @param list (in) list of expressions.
   @param val (in) number of the attribute
*/
void
ExprManager::resetMark (ptlinstances list, typeMark val)
{
  ptexpcal exp;

  while (ptlinstances_get_next (&list, &exp))
    {
      ptexpcal_removeAttribute (exp, val);
      ptexpcal_removeMark (exp, val);
    }
}

/** 
   Removing a mark/attribute for all the expressions of the expression manager.
   @param val (in) number of the attribute
*/
void
ExprManager::resetMarkAll (int val)
{
  int i;
  for (i = 0; i < SIG_NB_OPERATORS; i++)
    resetMark (table[i], val);
}
