/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <string.h>
#include <iostream>
#include <fstream>
using namespace std;


_PRIVATE_
/** Adding an expression in set of expressions (classified by operator). */
  void
ExprManager::addintabexp (noperators op, ptexpcal vexp)
{
  ptlexpcal x;
  int n, m;

  x = tableEnd[op];
  if (x == NILLEXPCAL)
    {
      table[op] = ptlinstances_insert (table[op], vexp);
      tableEnd[op] = table[op];
    }
  else
    {
      n = ptexpcal_get_num (vexp);
      m = ptexpcal_get_num (ptlinstances_element (x));
      if (n != m)
	{
	  if (n > m)
	    {
	      x = ptlinstances_push_end (x, vexp);
	      tableEnd[op] = ptlinstances_next (x);
	    }
	  else
	    table[op] = ptlinstances_insert (table[op], vexp);
	}
    }
}

/**
   @return "true" if it exists an index (vindex) at rank vrank in the global list lIndexForInit.
*/
boolean ExprManager::getIndexForInit (ptexpcal * vindex, int vrank)
{
  ptlexpcal
    lids;
  int
    i;

  lids = lIndexForInit;
  i = 0;
  while (i != vrank)
    {
      if (!ptlinstances_get_next (&lids, vindex))
	return pk_false;
      i++;
    }
  return pk_true;
}

/**
   It inserts at the end of the global list lIndexForInit the assumed index vind.
 */
void
ExprManager::addIndexForInit (ptexpcal vind)
{
  lIndexForInit = ptlinstances_push_end (lIndexForInit, vind);
}

/** 
    It removes an expression (vexp), characterized by its operator (op), from the manager.
*/
boolean
ExprManager::rmintabexp (noperators op, ptexpcal vexp)
{
  ptlexpcal l, last;
  int n, m;

  last = tableEnd[op];
  l = table[op];
  if (last != NILLEXPCAL)
    {
      n = ptexpcal_get_num (vexp);
      m = ptexpcal_get_num (ptlinstances_element (last));
      if (n == m)
	{
	  table[op] = ptlinstances_remove_last_elem (l, &last);
	  tableEnd[op] = last;
	  return pk_true;
	}
    }

  if (!ptlinstances_isin_remove_exp (&l, vexp))
    return pk_false;
  table[op] = l;
  return pk_true;
}

/**
   @return "true" if it exists an expression with vop as operator in the a list tableexp, "false" otherwise.
   @param vop (in) a Signal operator.
   @param vexp (out) the founded expression, if it exists.
   @param tableexp (in/out) a list of expressions, set to the founded element.
*/
boolean
ExprManager::getNextExprWithOperator (noperators vop, ptlexpcal * tableexp, ptexpcal * vexp)
{
  while (ptlinstances_get_next (tableexp, vexp))
    if (ptexpcal_get_operator (*vexp) == vop)
      return pk_true;
  return pk_false;
}



/**
   @return "true" if the binary temporal expression, vexp = vop(x1, x2), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinBinTpExprSet (noperators vop, ptexpcal x1, ptexpcal x2, ptexpcal * vexp)
{
  ptlexpcal lcour;

  lcour = ptexpcal_get_luse (x1);
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (vop, &lcour, vexp))
      if (ptexpcal_get_defaultArgstAt (*vexp, 1) == x1)
	if (ptexpcal_get_defaultArgstAt (*vexp, 2) == x2)
	  return pk_true;
  return pk_false;
}



/**
   @return "true" if the constant, vexp = vop(x1), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinConstantExprSet (noperators vop, ptrepstr x1, ptexpcal * vexp)
{
  ptlexpcal lcour;

  lcour = table[vop];
  while (lcour != NILLEXPCAL)
    {
      if (getNextExprWithOperator (vop, &lcour, vexp))
	{
	  if (ptexpcal_get_constant (*vexp) == x1)
	    {
	      return pk_true;
	    }
	}
    }
  return pk_false;
}



/**
   @return "true" if the unary expression, vexp = vop(x1), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinUnaryExprSet (noperators vop, ptexpcal x1, ptexpcal * vexp)
{
  ptlexpcal lcour;

  lcour = ptexpcal_get_luse (x1);
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (vop, &lcour, vexp))
      if (ptexpcal_get_unaryArg (*vexp) == x1)
	return pk_true;
  return pk_false;
}



/**
   @return "true" if the synchronous binary expression, vexp = vop(x1,x2), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinBinExprSet (noperators vop, ptexpcal x1, ptexpcal x2, ptexpcal * vexp)
{
  ptlexpcal lcour;

  lcour = ptexpcal_get_luse (x1);
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (vop, &lcour, vexp))
      if (ptexpcal_get_binaryArgstAt (*vexp, 1) == x1)
	if (ptexpcal_get_binaryArgstAt (*vexp, 2) == x2)
	  return pk_true;
  return pk_false;
}



/**
   @return "true" if the ternary expression, vexp = vop(x1,x2,x3), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinTerExprSet (noperators vop, ptexpcal x1, ptexpcal x2, ptexpcal x3, ptexpcal * vexp)
{
  ptlexpcal lcour;

  lcour = ptexpcal_get_luse (x1);
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (vop, &lcour, vexp))
      if (ptexpcal_get_ternaryArgstAt (*vexp, 1) == x1)
	if (ptexpcal_get_ternaryArgstAt (*vexp, 2) == x2)
	  if (ptexpcal_get_ternaryArgstAt (*vexp, 3) == x3)
	    return pk_true;
  return pk_false;
}



/**
   @return "true" if the list expression, vexp = vop(x1,x2), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinListSet (noperators vop, ptexpcal x1, ptexpcal x2, ptexpcal * vexp)
{
  ptlexpcal lcour;
  ptexpcal exp1;

  lcour = ptexpcal_get_luse (x1);
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (vop, &lcour, vexp))
      {
	exp1 = *vexp;
	if (ptexpcal_getNaryArgs (&exp1) == x1)
	  {
	    if (exp1 == NILEXPCAL)
	      if (x2 == NILEXPCAL)
		return pk_true;
	      else;
	    else if (ptexpcal_getNaryArgs (&exp1) == x2)
	      return pk_true;
	  }
      }
  return pk_false;
}



/**
   @return "true" if the when/cell expression, vexp = vop(expr, h), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinWhenCellSet (noperators vop, ptexpcal expr, ptinsth vh, ptexpcal * vexp)
{
  ptlexpcal lcour;

  lcour = ptexpcal_get_luse (expr);
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (vop, &lcour, vexp))
      if (ptexpcal_get_expWhen (*vexp) == expr)
	if (ptexpcal_getContributionClock (*vexp) == vh)
	  return pk_true;
  return pk_false;
}

/**
   @return "true" if the type expression (reference to a type), vexp = optype(t), already exists in the manager, "false" otherwise.
*/
boolean
ExprManager::isinTypeExprSet (TypeType t, ptexpcal * vexp)
{
  ptlexpcal lcour;
  noperators op;

  op = optype;
  lcour = table[op];
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (op, &lcour, vexp))
      if (getExpType (*vexp) == t)
	return pk_true;
  return pk_false;
}


/**
   @return "true" if the function call expression, vexp = id {params}(ins), already exists in the manager, "false" otherwise.
   where 
   - params : list of parameters, 
   - ins : list of inputs.
*/
boolean
ExprManager::isinCallExprSet (ptident id, ptexpcal params, ptexpcal ins, ptexpcal * vexp)
{
  ptlexpcal lcour;
  noperators op;
  ptident currentId;
  ptexpcal currentIns, lp;

  op = opappe;
  lcour = table[op];
  while (lcour != NILLEXPCAL)
    if (getNextExprWithOperator (op, &lcour, vexp))
      if (getFunctionCall (*vexp, &currentId, &lp, &currentIns))
	if ((currentId == id) && (currentIns == ins) && (lp == params))
	  return pk_true;
  return pk_false;
}

/**
   It signals that vexp references the clock vh.
*/
void
ExprManager::addReferenceClock2Expr (ptinsth vh, ptexpcal vexp)
{
  ptlinstances b;
  ptlinstances l;

  l = CKM->getExprReferencesTo (vh);
  if (!ptlinstances_isin_insert_exp (&l, &b, vexp))
    CKM->setExprReferencesTo (vh, l);
}

/**
  It removes the reference from vh to vexp.
*/
void
ExprManager::removeReferenceClock2Expr (ptinsth vh, ptexpcal vexp)
{
  ptlinstances l;
  l = CKM->getExprReferencesTo (vh);
  if (!ptlinstances_isin_remove_exp (&l, vexp))
    CKM->setExprReferencesTo (vh, l);
}


/** 
    It applies the following rules :
    - e1 when e1  -> pk_true when e1
    - e1 when not e1  -> false when e1
    - not e1 when e1  -> false when e1
    - not e1 when not e1  -> pk_true when e1
*/
ptexpcal ExprManager::applyWhenCellConstantRules (ptexpcal e1, ptinsth vh)
{
  ptexpcal
    vcond;

  if (CKM->satisfiesSub (vh))
    {
      vcond = CKM->getConditionSub (vh);
      if (vcond == e1)
	{
	  if (CKM->getOperator (vh) == sub)
	    e1 = ptexpcal_getTrue ();
	  else
	    e1 = ptexpcal_getFalse ();
	}
      else 
	if ((ptexpcal_get_operator (e1) == opnon) && (vcond == getUnaryArg (e1)))
	  {
	    if (CKM->getOperator (vh) == sub)
	      e1 = ptexpcal_getFalse ();
	    else
	      e1 = ptexpcal_getTrue ();
	  }
    }
  return e1;
}

/**
   @return a new when/cell expression op(e1, vh).
*/
ptexpcal
ExprManager::mkExtractOrMemory (noperators op, ptexpcal e1, ptinsth vh)
{
  ptexpcal vcond, vexp;

  e1 = applyWhenCellConstantRules (e1, vh);
  vexp = ptexpcal_RIEN_create_whenExp (op, e1, vh);
  TypeCheckingExtractOrMemory (op, vexp, e1, NILBACKANNOTATION);
  addReferenceClock2Expr (vh, vexp);
  addintabexp (op, vexp);
  return vexp;
}


/**
   Assumed e1 a constant expression.
   @return "true" if e1 << e2 according to the ordrer on expressions (see e1beforee2()), "false" otherwise.
*/
boolean
ExprManager::e1beforee2Const (ptexpcal e1, ptexpcal e2, cloper tyexp2)
{
  noperators vop1, vop2;

  if (tyexp2 != constant)
    return pk_true;
  vop1 = ptexpcal_get_operator (e1);
  vop2 = ptexpcal_get_operator (e2);
  if (vop1 == vop2)
    return (boolean) (ptexpcal_get_constant (e1) < ptexpcal_get_constant (e2));
  return (boolean) (vop1 < vop2);
}


/**
   Assumed, e1 is an identifier.
   @return "true" if e1 << e2 according to the ordrer on expressions (see e1beforee2()), "false" otherwise.
*/
boolean
ExprManager::e1beforee2Id (ptexpcal e1, ptexpcal e2, cloper tyexp2)
{
  if (tyexp2 == constant)
    return pk_false;
  if (tyexp2 == variabl)
    return (boolean) (ptexpcal_get_num (e1) < ptexpcal_get_num (e2));
  return pk_true;
}


/**
   Assumed: e1 is an unary expression.
   @return "true" if e1 << e2 according to the ordrer on expressions (see e1beforee2()), "false" otherwise.
*/
boolean
ExprManager::e1beforee2Oper1 (ptexpcal e1, ptexpcal e2, cloper tyexp2)
{
  noperators vop1, vop2;

  if (tyexp2 == constant)
    return pk_false;
  if (tyexp2 == variabl)
    return pk_false;
  if (tyexp2 == oper1)
    {
      vop1 = ptexpcal_get_operator (e1);
      vop2 = ptexpcal_get_operator (e2);
      if (vop1 == vop2)
	return e1beforee2 (ptexpcal_get_unaryArg (e1), ptexpcal_get_unaryArg (e2));
      return (boolean) (vop1 < vop2);
    }
  return pk_true;
}


/**
   Assumed: e1 is an binary expression.
   @return "true" if e1 << e2 according to the ordrer on expressions (see e1beforee2()), "false" otherwise.
*/
boolean
ExprManager::e1beforee2Oper2 (ptexpcal e1, ptexpcal e2, cloper tyexp2)
{
  noperators vop1, vop2;
  ptexpcal e11, e21;

  if (tyexp2 == constant)
    return pk_false;
  if (tyexp2 == variabl)
    return pk_false;
  if (tyexp2 == oper1)
    return pk_false;
  if (tyexp2 == oper2)
    {
      vop1 = ptexpcal_get_operator (e1);
      vop2 = ptexpcal_get_operator (e2);
      if (vop1 == vop2)
	{
	  e11 = ptexpcal_get_binaryArgstAt (e1, 1);
	  e21 = ptexpcal_get_binaryArgstAt (e2, 1);
	  if (e11 == e21)
	    return e1beforee2 (ptexpcal_get_binaryArgstAt (e1, 2), ptexpcal_get_binaryArgstAt (e2, 2));
	  return e1beforee2 (e11, e21);
	}
      return (boolean) (vop1 < vop2);
    }
  return pk_true;
}


/**
   Assumed: e1 is an ternary expression.
   @return "true" if e1 << e2 according to the ordrer on expressions (see e1beforee2()), "false" otherwise.
*/
boolean
ExprManager::e1beforee2Oper3 (ptexpcal e1, ptexpcal e2, cloper tyexp2)
{
  noperators vop1, vop2;
  ptexpcal e11, e21;

  if (tyexp2 == constant)
    return pk_false;
  if (tyexp2 == variabl)
    return pk_false;
  if (tyexp2 == oper1)
    return pk_false;
  if (tyexp2 == oper2)
    return pk_false;
  if (tyexp2 == oper3)
    {
      vop1 = ptexpcal_get_operator (e1);
      vop2 = ptexpcal_get_operator (e2);
      if (vop1 == vop2)
	{
	  e11 = ptexpcal_get_ternaryArgstAt (e1, 1);
	  e21 = ptexpcal_get_ternaryArgstAt (e2, 1);
	  if (e11 == e21)
	    {
	      e11 = ptexpcal_get_ternaryArgstAt (e1, 2);
	      e21 = ptexpcal_get_ternaryArgstAt (e2, 2);
	      if (e11 == e21)
		return e1beforee2 (ptexpcal_get_ternaryArgstAt (e1, 3), ptexpcal_get_ternaryArgstAt (e2, 3));
	      return e1beforee2 (e11, e21);
	    }
	  return e1beforee2 (e11, e21);
	}
      return (boolean) (vop1 < vop2);
    }
  return pk_true;
}


/**
   Assumed : E1, E2 are synchronous expressions. Used in normFormRelationalExpr() method.

   @return "true" if  E1 << E2 , << internal order on the expressions.
   constant << identifier << unary << binary << ternary.
   - two constants: e1 << e2 if operator(e1) < operator (e2), or operator(e1) = operator (e2) and value(e1) < value(e2).
   - two identifiers: e1 << e2 if number(e1) < number(e2).
   - two unary expressions: e1 << e2 if operator(e1) < operator(e2), or operator(e1) = operator(e2) and operand(e1) << operand(e2)
   - two binary expressions: e1 << e2 if operator(e1) < operator(e2), or 
          operator(e1) = operator(e2), e1 != e2,  and first operand(e1) << first operand(e2),
          operator(e1) = operator(e2), e1 = e2,  and second operand(e1) << second operand(e2).

   - two ternary expressions: similar to the previous case.

   It calls e1beforee2Const(), e1beforee2Id(), e1beforee2Oper1(), e1beforee2Oper2(), e1beforee2Oper3().
 */
boolean
ExprManager::e1beforee2 (ptexpcal e1, ptexpcal e2)
{
  cloper tyexp1, tyexp2;

  tyexp1 = ptexpcal_get_operatorClass (e1);
  tyexp2 = ptexpcal_get_operatorClass (e2);
  if (tyexp1 == constant)
    return e1beforee2Const (e1, e2, tyexp2);
  if (tyexp1 == variabl)
    return e1beforee2Id (e1, e2, tyexp2);
  if (tyexp1 == oper1)
    return e1beforee2Oper1 (e1, e2, tyexp2);
  if (tyexp1 == oper2)
    return e1beforee2Oper2 (e1, e2, tyexp2);
  if (tyexp1 == oper3)
    return e1beforee2Oper3 (e1, e2, tyexp2);
  return pk_true;
}

/**
   @return "true" if the constant expression vexp=op(cst) is not already in the manager, in this case it is created;
   "false" otherwise.
*/
boolean
ExprManager::mkesigconst (noperators vop, ptrepstr cst, ptexpcal * vexp)
{
  if (isinConstantExprSet (vop, cst, vexp))
    {
      return pk_false;
    }
  *vexp = ptexpcal_RIEN_create_constant (vop, cst);
  addintabexp (vop, *vexp);
  return pk_true;
}



/**
  @return "true" if the unary expression vexp=op(cst) of the pcl class, is not already in the manager, in this case it is created;
  "false" otherwise.
  
  pcl specifies the class temporal/synchronous class of the vop operator (must be oper1 or oper1tp).
*/
ptexpcal
ExprManager::mkesigexp1 (noperators vop, cloper pcl, ptexpcal vexpcal1)
{
  ptexpcal vexp;

  vexp = ptexpcal_RIEN_create_unaryExp (vop, pcl, vexpcal1);
  addintabexp (vop, vexp);
  return vexp;
}


/**
  @return "true" if the binary expression vexp=op(vexpcal1, vexpcal2) of the pcl class is not already in the manager, 
  in this case it is created;  "false" otherwise.
  
  pcl specifies the class temporal/synchronous class of the vop operator (must be oper2 or oper2tp).
*/
ptexpcal
ExprManager::mkesigexp2 (noperators vop, cloper pcl, ptexpcal vexpcal1, ptexpcal vexpcal2)
{
  ptexpcal vexp;
  vexp = ptexpcal_RIEN_create_binaryExp (vop, pcl, vexpcal1, vexpcal2);
  addintabexp (vop, vexp);
  return vexp;
}


/**
  @return "true" if the list expression vexp=op(vexpcal1, vexpcal2) is not already in the manager, 
  in this case it is created;  "false" otherwise.

  A list is managed as binary expressions.
*/
ptexpcal
ExprManager::mkesigexplist (noperators vop, cloper pcl, ptexpcal vexpcal1, ptexpcal vexpcal2)
{
  ptexpcal vexp;
  vexp = ptexpcal_RIEN_create_binaryListExp (vop, oper2, vexpcal1, vexpcal2);
  addintabexp (vop, vexp);
  return vexp;
}


/**
   @return the complementary boolean expression of vexpcal1.
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkNot (ptexpcal vexpcal1, BackAnnotationType refTree)
{
  ptexpcal vexp;

  if (!modeEval)
    return mkesigexp1 (opnon, oper1, vexpcal1);
  if (!evalNot (vexpcal1, refTree, &vexp))
    {
      vexp = mkesigexp1 (opnon, oper1, vexpcal1);
      ptexpcal_set_type (vexp, get_BooleanType ());
    }
  return vexp;
}


/**
   @return the conjonction (and) of boolean expressions vexpcal1, vexpcal2.
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkAnd (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  ptexpcal vres;

  if (vexpcal1 == vexpcal2)
    return vexpcal1;
  if (!modeEval)
    return mkesigexp2 (opet, oper2, vexpcal1, vexpcal2);
  if (!evalAnd (vexpcal1, vexpcal2, refTree, &vres))
    {
      return mkbinbool (opet, vexpcal1, vexpcal2, refTree);
    }
  return vres;
}

/**
   @return the disjunction (or) of boolean expressions vexpcal1, vexpcal2.
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkOr (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  ptexpcal vres;

  if (vexpcal1 == vexpcal2) return vexpcal1;
  if (!modeEval) return mkesigexp2 (opou, oper2, vexpcal1, vexpcal2);
  if (!evalOr (vexpcal1, vexpcal2, refTree, &vres)) return mkbinbool (opou, vexpcal1, vexpcal2, refTree);
  return vres;
}

/**
   @return the exclusive disjunction (xor) of boolean expressions vexpcal1, vexpcal2.
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkExclOr (ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  ptexpcal vres;

  if (!modeEval)
    return mkesigexp2 (opouex, oper2, vexpcal1, vexpcal2);
  if (evalXor (vexpcal1, vexpcal2, refTree, &vres))
    return vres;
  return mkbinbool (opouex, vexpcal1, vexpcal2, refTree);
}


/**
   @return the boolean expression oper( vexpcal1, vexpcal2). Called by mkExclOr(), mkOr(), mkAnd() method.
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkbinbool (noperators oper, ptexpcal vexpcal1, ptexpcal vexpcal2,  BackAnnotationType refTree)
{
  ptexpcal vexp;

  if (!TypeCheckingSameSize (vexpcal1, vexpcal2, refTree))
    return ptexpcal_getTrue ();
  vexp = mkesigexp2 (oper, oper2, vexpcal1, vexpcal2);
  //ptexpcal_set_type(vexp, getType(vexpcal1)); // Nov 2013- No, (true xor pk_true) without simplification, the type is not event but boolean.
  ptexpcal_set_type (vexp, BooleanType);
  return vexp;
}


/**
   @return the new temporal binary expression (default) expression op(vexpcal1, vexpcal2).
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkExpTp2 (noperators op, ptexpcal vexp1, ptexpcal vexp2, BackAnnotationType refTree)
{
  ptexpcal vexp;

  vexp = ptexpcal_RIEN_create_defaultExp (op, vexp1, vexp2);
  addintabexp (op, vexp);
  if (modeEval)
    TypeCheckingBinaryTemporal (op, vexp, vexp1, vexp2, refTree);
  return vexp;
}


/**
   @return the unary expression vop(vexpcal1).
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkExp1 (noperators vop, ptexpcal vexpcal1, BackAnnotationType refTree)
{
  ptexpcal vexp;
  ptexpcal nb, r, d1, d2;
  TypeType Tr;

  if (vop == opnega)
    if (ptexpcal_get_operator (vexpcal1) == opnega)
      return getUnaryArg (vexpcal1);

  if (!modeEval)
    return mkesigexp1 (vop, oper1, vexpcal1);
  if (vop == opnon) return mkNot (vexpcal1, refTree);
  vexp = mkesigexp1 (vop, oper1, vexpcal1);
  if (modeEval)
    if (TypeCheckingUnary (vop, vexp, vexpcal1, refTree))
      if (vop == optran)	// matrix transposition.
	{
	  getDimAndType (vexpcal1, &r, &Tr);
	  d1 = getNaryArgs (&r);
	  nb = mkNaryExp (oplesign, d1, NILEXPCAL, NILBACKANNOTATION);
	  d2 = getNaryArgs (&r);
	  nb = mkNaryExp (oplesign, d2, nb, NILBACKANNOTATION);
	  ptexpcal_set_type (vexp, mkArrayType (nb, Tr));
	}
  return vexp;
}



/**
   @return the binary expression vop(vexpcal1, vexpcal2).
   The type checking is applied: on error, a message is hooked to refTree.
   It calls mkAnd(), mkOr(), mkExclOr() for boolean expressions.
   It applies applyExprRelaRules() for relational expressions.
*/
ptexpcal
ExprManager::mkExp2 (noperators vop, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  ptexpcal vexp, sexp;

  if (!modeEval)
    return mkesigexp2 (vop, oper2, vexpcal1, vexpcal2);
  switch ((int) vop)
    {
    case opet:
      return mkAnd (vexpcal1, vexpcal2, refTree);
    case opou:
      return mkOr (vexpcal1, vexpcal2, refTree);
    case opouex:
      return mkExclOr (vexpcal1, vexpcal2, refTree);

    default:
      // Ajouter des traitements elementaires pour les vops 
      // relationnels (c = pk_true) -> c, (c=false) -> not c.
      if (! applyExprRelaRules (vop, vexpcal1, vexpcal2, refTree, &vexp))
	if (! applyExprArithmRules (vop, vexpcal1, vexpcal2, refTree, &vexp))
	  {
	    vexp = mkesigexp2 (vop, oper2, vexpcal1, vexpcal2);
	    if (modeEval)
	      {
		sexp = vexp;
		vexp = TypeCheckingBinary (vop, vexp, vexpcal1, vexpcal2, refTree);
		if (sexp != vexp)
		  rmintabexp (vop, sexp);
	      }
	  }
    }
  return vexp;
}

/**
   @return the new list expression vop(vexpcal1, vexpcal2).
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkExpN (noperators vop, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree)
{
  ptexpcal vexp;
  ptexpcal nb, r, d1, d2;
  TypeType Tr;

  vexp = mkesigexplist (vop, oper2, vexpcal1, vexpcal2);
  if (!modeEval)
    return vexp;
  if (TypeCheckingList (vop, vexp, vexpcal1, vexpcal2, refTree))
    if (vop == oplelevec)
      {
	getDimAndType (vexpcal1, &r, &Tr);
	nb = mkNaryExp (oplesign, mkIntegerConstant (getSize (vexp)), r, NILBACKANNOTATION);
	ptexpcal_set_type (vexp, mkArrayType (nb, Tr));
      }
  return vexp;
}


/**
   @return the ternary expression vop(vexpcal1, vexpcal2, vexpcal3).
   The type checking is applied: on error, a message is hooked to refTree.
*/
ptexpcal
ExprManager::mkExp3 (noperators vop, ptexpcal vexpcal1, ptexpcal vexpcal2, ptexpcal vexpcal3, BackAnnotationType refTree)
{
  ptexpcal vexp, dim;
  compilerInteger v, v1, v2, n;
  TypeType t;

  vexp = ptexpcal_RIEN_create_ternaryExp (vop, oper3, vexpcal1, vexpcal2, vexpcal3);
  if (!modeEval) return vexp;
  switch (vop)
    {
    case opret:
      TypeCheckingDelay (vop, vexp, vexpcal1, vexpcal2, vexpcal3, refTree);
      break;
    case opborn:
      // controle effectue sur l'vop opparc
      break;
    case openum:		// e1 .. e2 step e3
      if (TypeCheckingIndex (vexp, vexpcal1, vexpcal2, vexpcal3, refTree))
	{
	  if (IsBounded (vexp))
	    {
	      v = getValue (vexpcal3);
	      ptexpcal_get_integerBounds (vexp, &v1, &v2);
	      n = v2 - v1;
	      if (v > 0) n = n / v + 1; else n = n / -v + 1;
	      dim = mkIntegerConstant (n);
	      dim = mkNaryExp (oplesign, dim, NILEXPCAL, NILBACKANNOTATION);
	      t = mkArrayType (dim, IntegerType);
	      ptexpcal_set_type (vexp, t);
	    }
	  else;
	}
      break;
    }
  addintabexp (vop, vexp);
  return vexp;
}




/** 
    @return "true" if one of the following rules on vop(vexpcal1, vexpcal2)
    - ==(vexpcal1, vexpcal1) -> pk_true 
    - <<= (vexpcal1, vexpcal1) -> pk_true
    - = (vexpcal1, vexpcal1) -> pk_true 
    - =  (pk_true, vexpcal2) -> vexpcal2
    - = (vexpcal1, pk_true) -> vexpcal1
    - = (pk_false, vexpcal2) -> not vexpcal2
    - = (vexpcal1, pk_false) -> not vexpcal1
    - != (vexpcal1, vexpcal1) -> false
    - !=  (pk_true, vexpcal2) -> not vexpcal2
    - != (vexpcal1, pk_true) -> not vexpcal1
    - != (pk_false, vexpcal2) -> vexpcal2
    - != (vexpcal1, pk_false) -> vexpcal1

    - <= (vexpcal1, vexpcal1) -> pk_true 
    - < (vexpcal1, vexpcal1) -> false

    - >= (vexpcal1, vexpcal1) -> pk_true
    - > (vexpcal1, vexpcal1) -> false

   is applied ; in this case, vexp is the result.
   "false" otherwise, vexp is not modified.
   The type checking is applied: on error, a message is hooked to refTree.
*/
boolean
ExprManager::applyExprRelaRules (noperators vop, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree, ptexpcal * vexp)
{
  if (isNILEXPCAL (vexpcal1)) return pk_false;
  if (isNILEXPCAL (vexpcal2)) return pk_false;
  switch ((int) (vop))
    {
    case opegeg:		// ==
    case opiieg:		//  <<=    case opegal:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getTrue ();
	  return pk_true;
	}
      break;

    case opegal:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getTrue ();
	  return pk_true;
	}
      if (!ptexpcal_IsBoolean (vexpcal1))
	return pk_false;
      if (!ptexpcal_IsBoolean (vexpcal2))
	return pk_false;
      if (ptexpcal_IsAssigned (vexpcal1) && ptexpcal_IsAssigned (vexpcal2))
	return pk_false;	// cas traité ailleurs

      if (ptexpcal_equalTrue (vexpcal1))
	{
	  *vexp = vexpcal2;
	  return pk_true;
	}
      if (ptexpcal_equalTrue (vexpcal2))
	{
	  *vexp = vexpcal1;
	  return pk_true;
	}
      if (ptexpcal_equalFalse (vexpcal1))
	{
	  *vexp = mkUnaryExp (opnon, vexpcal2, refTree);
	  return pk_true;
	}
      if (ptexpcal_equalFalse (vexpcal2))
	{
	  *vexp = mkUnaryExp (opnon, vexpcal1, refTree);
	  return pk_true;
	}
      break;

    case opdiff:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getFalse ();
	  return pk_true;
	}
      if (!ptexpcal_IsBoolean (vexpcal1))
	return pk_false;
      if (!ptexpcal_IsBoolean (vexpcal2))
	return pk_false;
      if (ptexpcal_IsAssigned (vexpcal1) && ptexpcal_IsAssigned (vexpcal2))
	return pk_false; // cas traité ailleurs

      if (ptexpcal_equalTrue (vexpcal1))
	{
	  *vexp = mkUnaryExp (opnon, vexpcal2, refTree);
	  return pk_true;
	}
      if (ptexpcal_equalTrue (vexpcal2))
	{
	  *vexp = mkUnaryExp (opnon, vexpcal1, refTree);
	  return pk_true;
	}
      if (ptexpcal_equalFalse (vexpcal1))
	{
	  *vexp = vexpcal2;
	  return pk_true;
	}
      if (ptexpcal_equalFalse (vexpcal2))
	{
	  *vexp = vexpcal1;
	  return pk_true;
	}
      break;

    case opiegal:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getTrue ();
	  return pk_true;
	}
      return pk_false;
      break;

    case opinfe:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getFalse ();
	  return pk_true;
	}
      return pk_false;
      break;
    case opsegal:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getTrue ();
	  return pk_true;
	}
      return pk_false;
      break;

    case opsupe:
      if (vexpcal1 == vexpcal2)
	{
	  *vexp = ptexpcal_getFalse ();
	  return pk_true;
	}
      return pk_false;
      break;
    }
  return pk_false;
}

boolean
ExprManager::reduceArithmExpr (noperators vop, ptexpcal vexpcal1, ptexpcal vexpcal2, ptexpcal * vexp)
{
  boolean apply;
  ptexpcal vexpcal11, vexpcal12;
  noperators vop1;

  if (vop == opplus)
    apply = (ptexpcal_get_operator (vexpcal1) == opmoin);
  else if (vop == opmoin)
    apply = (ptexpcal_get_operator (vexpcal1) == opplus);
  else
    apply = pk_false;

  if (apply)
    {
      // (E11 +/- E12) -/+ E12 -> E11,
      // (E11+/- E12) -/+ E11 -> E12
      vexpcal11 = ptexpcal_get_binaryArgstAt (vexpcal1, 1);
      vexpcal12 = ptexpcal_get_binaryArgstAt (vexpcal1, 2);
      if (vexpcal11 == vexpcal2)
	{
	  *vexp = vexpcal12;
	  return pk_true;
	}
      if (vexpcal12 == vexpcal2)
	{
	  *vexp = vexpcal11;
	  return pk_true;
	}
    }
  return pk_false;
}

/** 
    @return 
    - "true" if one of the following rules on vop(vexpcal1, vexpcal2)
        - (E11+ E12) - E12 -> E11, where vexpcal1 is "E11+ E12", vexpcal2 is "E12"
        - (E11+ E12) - E11 -> E12  where vexpcal1 is "E11+ E12", vexpcal2 is "E11"
	is applied ; in this case, vexp is the result.
    - "false" otherwise, vexp is not modified.
*/
boolean
ExprManager::applyExprArithmRules (noperators vop, ptexpcal vexpcal1, ptexpcal vexpcal2, BackAnnotationType refTree, ptexpcal * vexp)
{
  if (!modeArithmRules) return pk_false;
  if (isNILEXPCAL (vexpcal1)) return pk_false;
  if (isNILEXPCAL (vexpcal2)) return pk_false;
  if (reduceArithmExpr (vop, vexpcal1, vexpcal2, vexp))    return pk_true;
  return (reduceArithmExpr (vop, vexpcal2, vexpcal1, vexp));
}



/**
   @return the default initial value for a given type (TT) with some restrictions (NILEXPCAL is returned in this case).
*/
ptexpcal
ExprManager::getDefaultInit (TypeType TT)
{
  ptrepstr enumValueFirst;

  if (TypeTestPredef (TT))
    {
      switch ((int) TT)
	{
	case PureType:
	  return ptexpcal_getTrue ();
	case BooleanType:
	  return ptexpcal_getFalse ();

	case LongType:
	case ShortType:
	case IntegerType:
	  return mkIntegerConstant (0);

	case RealType:
	  return mk_realPtexpcal (string2repstr ("0.0e0"));
	case DRealType:
	  return mk_dprealPtexpcal (string2repstr ("0.0d0"));
	case StringType:
	  return mkNaryExp (opchai, mk_stringPtexpcal (string2repstr ("\"\"")), NILEXPCAL, NILBACKANNOTATION);
	case CharType:
	  return mk_charPtexpcal (string2repstr ("'\\0'"));
	case ComplexType:
	  return mkBinaryExp (opcplx, getDefaultInit (RealType), getDefaultInit (RealType), NILBACKANNOTATION);
	case DComplexType:
	  return mkBinaryExp (opcplx, getDefaultInit (DRealType), getDefaultInit (DRealType), NILBACKANNOTATION);
	}
      return NILEXPCAL;
    }
  if (TypeTestEnumerationClass (TT))
    {
      // Default value : first element
      enumValueFirst = NameSubListGetHead (NameListCast2SubList (EnumerationTypeGetNameLst (TT)));
      return mkBinaryExp (opvenum, mkExpType (TT), mk_enumPtexpcal (enumValueFirst), NILBACKANNOTATION);
    }
  if (TypeTestNamedClass (TT))
    return getDefaultInit (NamedTypeGetTerm (TT));

  return NILEXPCAL;
}

/**
   @return a copy of vdim excluding the last element.
   Assumed: vdim is a list of expressions.
*/
ptexpcal
ExprManager::DuplicateWithoutLast (ptexpcal vdim)
{
  ptexpcal velem;
  if (!getNaryArgs (&vdim, &velem))
    return NILEXPCAL;
  if (isNILEXPCAL (vdim))
    return NILEXPCAL;
  return mkNaryExp (ptexpcal_get_operator (vdim), velem, DuplicateWithoutLast (vdim), NILBACKANNOTATION);
}

/**
   It appends a type (TT) in a list of types (lTypes) excluding the predefined types.
   For composed types, the inner types are also inserted.
*/
void
ExprManager::appendType (TypeType TT, ptlinstances * lTypes)
{
  ptlinstances vdummy;
  boolean bstop;
  TypeList l1;
  TypeType t1;
  TypeSubList SubLst1;
  
  if (TypeTestPredef (TT)) return;

  if (ptlinstances_isin_insert_exp (lTypes, &vdummy, mkExpType (TT)))
    {
      if (TypeTestArrayClass (TT))
	appendType (ArrayMatrixTypeGetTerm (TT), lTypes);
      else if (TypeTestNamedClass (TT))
	appendType (NamedTypeGetTerm (TT), lTypes);
      else
	{
	  if (TypeTestStructureClass (TT)) 
	    TT = StructureTypeGetArgument (TT);
	  if (TypeTestTupleClass (TT))
	    {
	      l1 = TupleTypeGetTypeLst (TT);
	      SubLst1 = TypeListCast2SubList (l1);
	      bstop = pk_false;
	      while (!bstop)
		{
		  t1 = TypeSubListGetHead (SubLst1);
		  appendType (t1, lTypes);
		  if (TypeSubListTestLast (SubLst1))
		    bstop = pk_true;
		  else
		    SubLst1 = TypeSubListGetTail (SubLst1);
		}
	    }
	}
    }

  /* if (TypeTestMetaClass(t);
     if (TypeTestEnumerationClass(t)); if (TypeTestExternClass(t)); if (TypeTestSubrangeClass(t)) ;
     if (TypeTestBitsClass(t)) ; if (TypeTestLambdaClass(t)); if (TypeTestNodeClass(t)) ;
     if (TypeTestGenericNodeClass(t)); if (TypeTestMetaNodeClass(t)) ; if (TypeTestMatrixClass(t)) ;
     if (TypeTestTupleClass(t)) ; 
     if (TypeTestStructureClass(t)); 
   */
}



/** @return true if e depends on "current" indexes, "false" otherwise. */
boolean
ExprManager::analyzeIndexDependence (ptexpcal e)
{
  if (IsIndex (e))
    return pk_true;
  if (ptexpcal_satisfies_ConstantExpr (e))
    return pk_true;
  // if (ptexpcal_get_operatorClass(e) == constant) return pk_true;
  return pk_false;
}


/**
   @return true if a reference to a type (TT) is already in a list  (l), "false" otherwise.
*/
_VIRTUAL_ boolean ExprManager::TypeTypeIsIn (TypeType TT, ptlinstances l)
{
  ptexpcal
    elem;
  while (ptlinstances_get_next (&l, &elem))
    {
      if (ptexpcal_get_operatorClass (elem) == opertyp)
	if (getExpType (elem) == TT)
	  {
	    return pk_true;
	  }
    }
  return pk_false;
}


/**
   @return a new identifier characterized by
   @param r (in) tree-identifier used as prefix to build the name of the instance.
   @param ar (in) abstract tree of the declaration of the flow(parser information).
   @param ti (in) specialisation of the identifier as a signal, constant, ...
   @param TT  (in) type of the identifier.
   @param node (in) node of the signal (can be undefined node)
   @param clk clock of the signal (it can be the undefined value)
*/
ptexpcal
ExprManager::newIdent (ptident r, ntrees ar, typinst ti, TypeType TT, ptINode node, ptinsth clk)
{
  ptexpcal vexpcal;
  ptident id;
  vexpcal = ptexpcal_newIdentifier (r, ar, ti, TT, node, clk);
  NM->set_leftOperand (node, vexpcal);
  if (!CKM->satisfiesUndefClock (clk))
    setClock (vexpcal, clk);
  id = mkSignalName (vexpcal, r, pk_true);
  addintabexp (opident, vexpcal);
  return vexpcal;
}


/** 
    @return "true" if an expression (exp) is an expression on constants, "false" otherwise.
*/
boolean
ExprManager::ptexpcal_satisfies_ConstantExpr (ptexpcal exp)
{
  noperators op;
  cloper clop;
  ptexpcal e1, e2;

  if (exp == NILEXPCAL)
    return pk_false;
  clop = ptexpcal_get_operatorClass (exp);
  switch ((int) clop)
    {
    case variabl:
      return (ptexpcal_IsParameter (exp) || ptexpcal_IsConstant (exp));

    case constant:
      return pk_true;

    case oper1:
    case oper1tp:
      return ptexpcal_satisfies_ConstantExpr (ptexpcal_get_unaryArg (exp));

    case oper2:
    case oper2tp:
      op = ptexpcal_get_operator (exp);
      if (op == opvenum)
	return pk_true;
      if (op != opparc)
	{
	  e1 = ptexpcal_get_binaryArgstAt (exp, 1);
	  if (!ptexpcal_satisfies_ConstantExpr (e1))
	    return pk_false;
	}
      e2 = ptexpcal_get_binaryArgstAt (exp, 2);
      return ptexpcal_satisfies_ConstantExpr (e2);

    case oper3:
    case oper3tp:
      if (!ptexpcal_satisfies_ConstantExpr (ptexpcal_get_ternaryArgstAt (exp, 1)))
	return pk_false;
      if (!ptexpcal_satisfies_ConstantExpr (ptexpcal_get_ternaryArgstAt (exp, 2)))
	return pk_false;
      return ptexpcal_satisfies_ConstantExpr (ptexpcal_get_ternaryArgstAt (exp, 3));

    case opereh:
      return ptexpcal_satisfies_ConstantExpr (ptexpcal_get_expWhen (exp));

    case call:
      return ptexpcal_satisfies_ConstantExpr (ptexpcal_GetInputsFunctionCall (exp));
    case opertyp:
    default:
      return pk_false;
    }
}

/** @return "true" if all elements of an expression (exp) are defined by the same expression (vce), "false" otherwise.
*/
boolean
ExprManager::ptexpcal_satisfies_DefinedBySameExpression (ptexpcal exp, ptexpcal * vce)
{
  ptexpcal vce1, vce2, vce3;
  noperators op;

  if (exp == NILEXPCAL)
    return pk_true;
  op = ptexpcal_get_operator (exp);
  switch ((int) (ptexpcal_get_operatorClass (exp)))
    {
    case variabl:
      *vce = exp;
      if (ptexpcal_IsParameter (exp))
	return pk_true;
      if (ptexpcal_IsConstant (exp))
	return pk_true;
      if (!NM->is_equation (getNode (exp)))
	return pk_false;
      return getAttribute (exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) vce);

    case constant:
      *vce = exp;
      return pk_true;

    case oper1:
      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_unaryArg (exp), &vce1))
	return pk_false;
      if (op == optran)
	*vce = vce1;
      else
	*vce = exp;
      return pk_true;

    case oper2:

      if (satisfiesList (exp))
	{
	  vce1 = exp;
	  while (getNaryArgs (&vce1, &vce2))
	    if (!ptexpcal_satisfies_DefinedBySameExpression (vce2, &vce3))
	      return pk_false;
	  *vce = exp;
	  return pk_true;
	}
      if (op == opelevec)
	return pk_false;
      if (op == opparc)
	return pk_false;
      if (op == opensvec)
	return pk_false;
      if (op == oprest)
	return pk_false;

      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_binaryArgstAt (exp, 1), &vce1))
	return pk_false;
      if (op == oprepe)
	{
	  *vce = vce1;
	  return pk_true;
	}
      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_binaryArgstAt (exp, 2), &vce2))
	return pk_false;
      *vce = exp;
      return pk_true;


    case oper3:
      if (op == openum)
	return pk_false;
      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_ternaryArgstAt (exp, 1), &vce1))
	return pk_false;
      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_ternaryArgstAt (exp, 2), &vce2))
	return pk_false;
      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_ternaryArgstAt (exp, 3), &vce3))
	return pk_false;
      *vce = exp;
      return pk_true;

    case opereh:
      return ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_get_expWhen (exp), vce);

    case call:
      if (!ptexpcal_satisfies_DefinedBySameExpression (ptexpcal_GetInputsFunctionCall (exp), &vce1))
	return pk_false;
      *vce = exp;
      return pk_true;

    case opertyp:
    case oper2tp:
    case oper1tp:
    case oper3tp:
    default:
      return pk_false;
    }
}


/**
  @param e1,e2 (in) any expression.
  @return "true" if e1 is equal to e2: e1=e2 or, if e1 and e2 are valued 
  expression (int) and they have the same value, "false" otherwise.
*/
boolean
ExprManager::isExprEqualLookingValue (ptexpcal e1, ptexpcal e2)
{
  compilerInteger i, j;
  if (e1 == e2)
    return pk_true;
  if (getIntegerValue (e1, &i))
    if (getIntegerValue (e2, &j))
      return (i == j);
  return pk_false;
}

/**
   Concatenating two lists of expressions: the elements of a list (l1) are added before the elements of a second list (l2).
   When l2 is empty, l1 is duplicated !!!.
   @param opl (in) code operator of the list.
   @param l1 (in) a list of expressions (assumed)
   @param l2 (in) a list of expressions (assumed)
   @param ba (in) back annotation to the abstract tree.(traceability)
*/
ptexpcal
ExprManager::fusionlist (noperators opl, ptexpcal l1, ptexpcal l2, BackAnnotationType ba)
{
  ptexpcal vexpcal;

  vexpcal = getNaryArgs (&l1);
  if (l1 == NILEXPCAL) return mkNaryExp (opl, vexpcal, l2, ba);
  return mkNaryExp (opl, vexpcal, fusionlist (opl, l1, l2, ba), ba);
}

/**
   @return the default of the 2 operands (arg1, arg2) assumed to be normalized.
   The result is normalized ( i.e (E1 when h1) default ...default (En when hn), Ei are synchronous expressions).
*/
ptexpcal
ExprManager::mergeExpressions (ptexpcal arg1, ptexpcal arg2)
{
  if (getOperator (arg2) == opmele)
    {
      arg1 = mergeExpressions (arg1, getDefaultArgstAt (arg2, 1));
      return mkClockedDefault (arg1, getDefaultArgstAt (arg2, 2));
    }
  return mkClockedDefault (arg1, arg2);
}

/**
   UNUSED.
   @return the root string of an boolean expression (exp).
*/
ptident ExprManager::getRootStringBooleanExpression (ptexpcal exp, int vholoph)
{
  ptident id;
  noperators op;
  cloper texp;

  if (exp == NILEXPCAL)
    return getPrefixBooleanClock ();
  //  op = getOperator(exp);
  texp = getOperatorClass (exp);
  if (texp == variabl)
    return getName (exp);
  if (texp == opertyp)
    return getPrefixBooleanClock ();
  if (texp == constant)
    return getPrefixBooleanClock ();
  if (texp == call)
    return ptexpcal_GetIdentFunctionCall (exp);
  if (vholoph < 0)
    return getPrefixBooleanClock ();
  vholoph--;
  if ((texp == oper1) || (texp == oper1tp))
    return getRootStringBooleanExpression (ptexpcal_get_unaryArg (exp), vholoph);
  if ((texp == oper2) || (texp == oper2tp))
    return getRootStringBooleanExpression (ptexpcal_get_binaryArgstAt (exp, 1), vholoph);
  if ((texp == oper3) || (texp == oper3tp))
    return getRootStringBooleanExpression (ptexpcal_get_ternaryArgstAt (exp, 1), vholoph);
  if (texp == opereh)
    return getRootStringBooleanExpression (ptexpcal_get_expWhen (exp), vholoph);
  pK_messenger->Caseerror (currentLine, currentFile);
  return getPrefixBooleanClock ();
}

