/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   It resets an expression (e). The operands are set to null expressions and the references to e are updated.   
*/
    void
 ExprManager::resetReferenceTo(ptexpcal e)
{
  ptinsth h;
  noperators op;

  ptexpcal_reset_luse(e);
  op = ptexpcal_get_operator(e);
  if (op == opextr)
    {
      h = ptexpcal_getContributionClock(e);
      removeReferenceClock2Expr(h, e);
    }
}

/** 
    It realizes the rewriting of [e1] in [e2]
    It is assumed that the clock [e1] is not placed in the hierarchy.
*/
void ExprManager::updateClockManager(ptexpcal e1, ptexpcal e2)
{
  ptinsth vhSub, vh1, vhs, vhns, vhNoSub, vhcond;
  boolean sw, breplace;
  noperators vop;

  if (ptexpcal_get_extraction(e1, &vhSub))
    {
      if (CKM->satisfiesIsAnEquivalentForm(vhSub))
	{
	  CKM->getComplExtraction(vhSub, &vhNoSub);
	  CKM->removeFormulaFromClass(vhSub);
	  CKM->removeFormulaFromClass(vhNoSub);
	}
      else
	{
	  CKM->getComplExtraction(vhSub, &vhNoSub);
	  vhcond = CKM->getClockSub(vhSub);
	  // vhSub : sub[e1], vhNoSub : nosub[e1]
	  breplace = pk_true;
	  vop = ptexpcal_get_operator(e2);
	  if (vop == optrue)
	    {
	      vhs = vhcond;
	      vhns = CKM->getNullClock();
	    }
	  else if (vop == opfalse)
	    {
	      vhns = vhcond;
	      vhs = CKM->getNullClock();
	    }
	  else
	    {
	      sw = (vop == opnon);
	      if (sw)
		e2 = ptexpcal_get_unaryArg(e2);
	      if (ptexpcal_get_extraction(e2, &vhs))
		{
		  CKM->getComplExtraction(vhs, &vhns);
		  if (sw)
		    {
		      vh1 = vhs;
		      vhs = vhns;
		      vhns = vh1;
		    }
		}
	      else if (sw)
		{
		  vhns = CKM->makeSub(vhcond, e2, getSignalsReferencedIn(e2));
		  vhs = CKM->makeNoSub(e2);
		}
	      else
		{
		  breplace = pk_false;
		  CKM->rewriteCondSub(vhSub, e2, getSignalsReferencedIn(e2));
		}
	    }
	  if (breplace)
	    {
	      CKM->replaceSub(vhSub, vhs, pk_false);
	      CKM->replaceSub(vhNoSub, vhns, pk_false);
	    }
	  CKM->hclmklequiv(0);
	}
    }
}

/** @return "true" if there is an equivalence to be treated, exp1 must be replaced by exp2,  "false" otherwise.
*/
boolean ExprManager::expsomeequiv(ptexpcal * exp1, ptexpcal * exp2)
{
  if (!ptlinstances_get_next(&lequ1exp, exp1))
    return pk_false;
  return ptlinstances_get_next(&lequ2exp, exp2);

}

/** @return "true" if there is an expression (exp) that has been modified, "false" otherwise. */
boolean ExprManager::somemodified(ptexpcal * exp)
{
  //WHY?  return pk_false;
  return ptlinstances_get_next(&lexpmodified, exp);
}

/** It signals that an expression (vexp) has been modified; It must be, and its uses, revisited for rewriting.
*/
void ExprManager::addModified(ptexpcal vexp)
{
  lexpmodified = ptlinstances_push_front(lexpmodified, vexp);
}

/**
   It signals the equivalence of oexp with nexp, i.e. oexp is replaced by nexp.
*/
void ExprManager::addEquivalence(ptexpcal oexp, ptexpcal nexp)
{
  if (oexp == nexp)
    addModified(oexp);
  else
    {
      if (remove(oexp))
	resetReferenceTo(oexp);
      lequ1exp = ptlinstances_push_front(lequ1exp, oexp);
      lequ2exp = ptlinstances_push_front(lequ2exp, nexp);
      if (ptexpcal_get_operatorClass(oexp) != variabl)
	setIndirection(oexp, nexp);
    }
}

/**
   It substitues e1 by e2 in an expression (use); use is an unary expression.
   Local to substituteIdByIn().
*/
void ExprManager::rewriteOper1(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  ptexpcal exp1, nexp;
  noperators ope;

  exp1 = ptexpcal_get_unaryArg(use);
  if (exp1 != e1)
    return;			// Security
  ope = ptexpcal_get_operator(use);
  nexp = mkUnaryExp(ope, e2, NILBACKANNOTATION);
  addEquivalence(use, nexp);
}

/**
   It substitues e1 by e2 in an expression (use); use is a binary expression.
   Local to rewrite().
*/
void ExprManager::rewriteOper2(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  ptexpcal exp1, exp2, nexp;
  noperators ope;

  ope = ptexpcal_get_operator(use);
  ptexpcal_get_binaryArgs(use, &exp1, &exp2);
  if (e1 == exp1)
    exp1 = e2;
  if (e1 == exp2)
    exp2 = e2;
  nexp = mkBinaryExp(ope, exp1, exp2, NILBACKANNOTATION);
  addEquivalence(use, nexp);
}

/**
   It substitues e1 by e2 in an expression (use); use is a ternary expression.
   Local to rewrite().
*/
void ExprManager::rewriteOper3(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  ptexpcal exp1, exp2, exp3, nexp;
  noperators ope;

  ope = ptexpcal_get_operator(use);
  ptexpcal_get_ternaryArgs(use, &exp1, &exp2, &exp3);
  if (e1 == exp1)
    exp1 = e2;
  if (e1 == exp2)
    exp2 = e2;
  if (e1 == exp3)
    exp3 = e2;
  if (isinTerExprSet(ope, exp1, exp2, exp3, &nexp))
    addEquivalence(use, nexp);
  else
    {
      ptexpcal_set_ternaryArgstAt(use, 1, exp1);
      ptexpcal_set_ternaryArgstAt(use, 2, exp2);
      ptexpcal_set_ternaryArgstAt(use, 3, exp3);
      addModified(use);
    }
}

/**
   It substitues e1 by e2 in an expression (use); use is a when/cell expression.
   Local to rewrite().
*/
void ExprManager::rewriteOpereh(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  ptexpcal nexp, exp1;
  ptinsth h;
  noperators ope, ope2;

  exp1 = ptexpcal_get_expWhen(use);
  if (exp1 != e1)
    return;			// Blindage
  ope = ptexpcal_get_operator(use);	// opextr, opmemo

  if (ope == opmemo)
    {
      ope2 = ptexpcal_get_operator(e2);
      if ((ope2 != opextr) && (ope2 != opmele))
	return;
    }
  h = ptexpcal_getContributionClock(use);
  nexp = mkWhenOrCellExp(ope, e2, h);
  addEquivalence(use, nexp);
}

/**
   It substitues e1 by e2 in an expression (use); use is a binary temporal expression (default expression).
   Local to rewrite().
*/
void ExprManager::rewriteOper2tp(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  ptexpcal nexp, exp1, exp2;
  noperators ope;

  ptexpcal_get_defaultArgs(use, &exp1, &exp2);
  if (exp1 == e1)
    exp1 = e2;
  if (exp2 == e1)
    exp2 = e2;
  ope = ptexpcal_get_operator(use);	// opmele
  if (isinBinTpExprSet(ope, exp1, exp2, &nexp))
    addEquivalence(use, nexp);
  else
    {
      ptexpcal_set_defaultArgstAt(use, 1, exp1);
      ptexpcal_set_defaultArgstAt(use, 2, exp2);
      addModified(use);
    }

}

void ExprManager::rewriteCall(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  ptexpcal_updateFunctionCall(use, e1, e2), addModified(use);
}

/**
   It substitues e1 by e2 in an expression (use).
   @return "true" if the operation is realized, "false" otherwise (use doesn't reference e1).
*/
boolean ExprManager::rewrite(ptexpcal e1, ptexpcal e2, ptexpcal use)
{
  cloper ope;
  boolean b;

  ope = ptexpcal_get_operatorClass(use);
  b = pk_true;
  switch ((int)ope)
    {
    case oper1:
      rewriteOper1(e1, e2, use);
      break;
    case oper2:
      rewriteOper2(e1, e2, use);
      break;
    case oper3:
      rewriteOper3(e1, e2, use);
      break;
    case opereh:
      rewriteOpereh(e1, e2, use);
      break;
    case oper2tp:
      rewriteOper2tp(e1, e2, use);
      break;
    case call:
      rewriteCall(e1, e2, use);
      break;

    case variabl:
    case constant:
    case oper1tp:		// event 
    case oper3tp:
      b = pk_false;
      break;
    default:
      b = pk_false;
    }
  return b;
}

/**
   To update the interface of the nodes connected to the expression, the modification is propagated.
*/
void ExprManager::updateModifiedExpr(ptexpcal exp)
{
  // Added (before it was empty) Aug 2014. May it induce an infinite loop ?
  addModified(exp);
}

/**
   It calls rewriting(ptexpcal) for each use of an expression (exp).
*/
void ExprManager::updateUses(ptexpcal exp)
{
  ptexpcal use;
  ptlinstances luse;

  luse = ptexpcal_get_luse(exp);
  while (ptlinstances_get_next(&luse, &use))
    updateModifiedExpr(use);
}

/** 
    It solves,recursively, all the unresolved equivalences, and the expressions an argument of which is modified.
    It calls somemodified(), updateUses(), updateNodes(), expmklequiv().
*/
void ExprManager::rewriting()
{
  boolean arreter;
  ptexpcal vexp;

  arreter = pk_false;
  while (!arreter)
    {
      if (somemodified(&vexp))
	{
	  updateUses(vexp);
	  updateNodes(vexp, vexp);
	  arreter = pk_false;
	}
      else
	arreter = pk_true;
      if (expmklequiv())
	arreter = pk_false;
    }
}

/**
  @return "true" if it exists at least one equivalence to solved, "false" otherwise.
  It builds the unresolved and induced equivalences.
*/
boolean ExprManager::expmklequiv()
{
  ptexpcal exp1, exp2;
  boolean sortir;

  if (!expsomeequiv(&exp1, &exp2))
    return pk_false;
  sortir = pk_false;
  while (!sortir)
    {
      expReplace(exp1, exp2, hundef);
      sortir = !expsomeequiv(&exp1, &exp2);
    }
  return pk_true;
}

/**
   It rewrites an expression (use) assuming that an expression (vexp) is the null expression.
   Local to nullTermDelete().
*/
void ExprManager::nullTermRewriteIn(ptexpcal vexp, ptexpcal use)
{
  cloper clop;
  ptexpcal vexp1, vexp2;
  noperators ope;
  ptINode node;
  ptlINode ln;

  clop = ptexpcal_get_operatorClass(use);
  if (clop == oper2tp)
    {
      // default, mais toujours (e1 when h1) default (e2 when h2)
      ptexpcal_get_defaultArgs(use, &vexp1, &vexp2);
      ptexpcal_remove_luse(vexp1, use);
      ptexpcal_remove_luse(vexp2, use);
      if (vexp1 == vexp)
	if (vexp2 == vexp)
	  nullTermDelete(use);
	else
	  substitute(use, vexp2, hundef);
      else if (vexp2 == vexp)
	substitute(use, vexp1, hundef);
      return;
    }
  if (clop != opereh)
    return;
  ope = ptexpcal_get_operator(use);
  if (ope == opextr)
    {
      vexp1 = ptexpcal_get_expWhen(use);
      ptexpcal_remove_luse(vexp1, use);
      substitute(use, vexp1, hundef);
    }
  else				// cell
    {
      ln = ptexpcal_get_NodesFeedBy(use);
      while (ptlINode_get_elem(&ln, &node))
	{
	  vexp1 = NM->get_leftOperand(node);
	  vexp1 = getInitialValue(vexp1);
	  if (vexp1 != NILEXPCAL)
	    vexp1 = mkWhenOrCellExp(opextr, vexp1, CKM->getInitClock());
	  else
	    pK_messenger->Caseerror(currentLine, currentFile);
	  NM->update(node, use, vexp1);
	}
    }
}

/**
   It updates the manager for a null expression (vexp). The references to this term are updated.
*/
void ExprManager::nullTerm(ptexpcal vexp)
{
  ptexpcal x;

  if (satisfiesList(vexp))
    while (getNaryArgs(&vexp, &x))
      nullTerm(x);
  else
    {
      if (IsSignal(vexp))
	CKM->makeSynchro(getClock(vexp), CKM->getNullClock());
      nullTermDelete(vexp);
    }
}

/**
   @return "true" if an expression (vexp) is removed from the manager, "false" otherwise.
   It tries to delete an expression (vexp) from the manager (it is not removed if it is a list of expressions).
   It applied nullTermRewriteIn() for each use of the expression, and updates each node (NodeManager::update()) that references
   the expression.
*/
boolean ExprManager::nullTermDelete(ptexpcal vexp)
{
  ptlinstances luse;
  ptexpcal x, use;
  ptlINode ln;
  ptINode node;
  boolean toBeRemoved;

  toBeRemoved = pk_true;
  luse = ptexpcal_get_luse(vexp);
  while (ptlinstances_get_next(&luse, &use))
    if (!satisfiesList(use))
      nullTermRewriteIn(vexp, use);
    else
      toBeRemoved = pk_false;

  ln = ptexpcal_get_NodesFeedBy(vexp);
  while (ptlINode_get_elem(&ln, &node))
    {
      x = NM->get_leftOperand(node);
      NM->update(node, vexp, NILEXPCAL);
      if (!NM->is_partialDefinition(node))
	nullTerm(x);
      else
	toBeRemoved = NM->removeDefPart(x, node);
    }

  if (toBeRemoved)
    if (remove(vexp))
      resetReferenceTo(vexp);
  return toBeRemoved;
}

/**
   It replaces a clock (h1) by an another one (h2) in an expression (vexp).
   This substitution induces some rewritings (rewriting()).
*/
void ExprManager::clockReplace(ptinsth h1, ptinsth h2, ptexpcal vexp)
{
  ptexpcal nexp, nexp1, exp1;
  boolean nc;
  ptexpcal use;
  noperators op;

  nc = CKM->satisfiesNullClock(h2);
  if ((h1 != h2) || nc)
    {
      use = vexp;
      if (nc)
	{
	  if (nullTermDelete(use))
	    return;
	}
      exp1 = ptexpcal_get_expWhen(use);

      nexp1 = applyWhenCellConstantRules(exp1, h2);
      op = ptexpcal_get_operator(use);
      if (nexp1 == exp1)
	{
	  if (isinWhenCellSet(op, exp1, h2, &nexp))
	    addEquivalence(use, nexp);
	  else
	    {
	      removeReferenceClock2Expr(h1, use);
	      ptexpcal_set_ContributionClock(use, h2);
	      addReferenceClock2Expr(h2, use);
	      if (nc)
		addModified(use);
	    }
	}
      else
	addEquivalence(use, mkExtractOrMemory(op, nexp1, h2));
      rewriting();
    }
}

/**
   It updates all the When expressions of the manager.
*/
void ExprManager::rewritingWhenExpr()
{
  ptexpcal vexp;
  ptinsth h, hhead;
  ptlinstances liste;

  liste = table[opextr];
  while (ptlinstances_get_next(&liste, &vexp))
    if (ptexpcal_get_operator(vexp) == opextr)	// warning : indirection
      {
	h = ptexpcal_getContributionClock(vexp);
	hhead = CKM->get_insthClock(h);	// via ClockGetSet_g
	clockReplace(h, hhead, vexp);
      }
}

/**
   It updates all the Cell expressions of the manager.
*/
void ExprManager::rewritingCellExpr()
{
  ptexpcal exp1, vexp;
  ptinsth h, hhead, hcexp1;
  ptlinstances liste;

  liste = table[opmemo];
  while (ptlinstances_get_next(&liste, &vexp))
    if (ptexpcal_get_operator(vexp) == opmemo)	// warning : indirection
      {
	h = ptexpcal_getContributionClock(vexp);
	hhead = CKM->get_insthClock(h);	// via ClockGetSet_g
	if (CKM->satisfiesNullClock(hhead))
	  nullTermDelete(vexp);
	else
	  {
	    exp1 = ptexpcal_get_expWhen(vexp);
	    hcexp1 = CKM->get_insthClock(ptexpcal_getContributionClock(exp1));	// via ClockGetSet_g
	    if (hcexp1 == hhead)
	      {
		if (ptexpcal_get_operator(exp1) != opextr)
		  exp1 = mkWhenOrCellExp(opextr, exp1, hhead);
		substitute(vexp, exp1, hundef);
	      }
	    else
	      clockReplace(h, hhead, vexp);
	  }
      }
}

void ExprManager::propagatePragmas(ptexpcal e1, ptexpcal e2)
{
  ptlpragmas lp;
  ptpragma v;

  if (e1 != e2)
    {
      lp = getPragmas(e1);
      while (ptlpragmas_get_next(&lp, &v))
	addPragma(e2, v);
    }
}

/** Without effect. */
void ExprManager::updateClockManager()
{
  ptinsth vhSub, vhNoSub;
  ptINode node;
  ptlINode ln;
  ptexpcal exp;

  return;			// A revoir 
  // ------------------------------------
  ln = lModifiedNodes;

  while (ptlINode_get_elem(&ln, &node))
    if (NM->is_delay(node) || NM->is_equation(node))
      {
	exp = NM->get_leftOperand(node);
	if (IsEvent(exp))
	  if (ptexpcal_get_extraction(exp, &vhSub))
	    if (CKM->satisfiesRoot(vhSub))	// restriction. a lever.
	      {
		CKM->getComplExtraction(vhSub, &vhNoSub);
		// vhSub : sub[exp], vhNoSub : nosub[exp]
		CKM->replaceSub(vhSub, CKM->getClockSub(vhSub), pk_false);
		CKM->replaceSub_null(vhNoSub);
		CKM->hclmklequiv(0);
	      }
      }
}

/*
  xxxx
*/
void ExprManager::updateInterfaceModifiedNodes()
{
  NM->updateInterfaceNodes(lModifiedNodes);
  updateClockManager();
  ptlINode_free(lModifiedNodes);
  lModifiedNodes = NILLINODE;
}

/**
   The dependences from the node of a signal (vexp) are removed if
   this signal is not used in any expression and it is not an undersampling.
*/
void ExprManager::cleanDependences(ptexpcal vexp)
{
  ptinsth vh;

  if (IsSignal(vexp))
    if ((ptexpcal_get_luse(vexp) == NILLEXPCAL) && !ptexpcal_get_extraction(vexp, &vh))
      NM->removeDependenceFrom(getNode(vexp));
}

/**
   @return "true" if the name of e2 is changed by the name of e2, "false" otherwise.
*/
boolean ExprManager::changeNameSignal(ptexpcal e1, ptexpcal  e2)
{
  boolean b;
  ptident id1, id2;
  if (ptexpcal_get_operator(e1) != opident) return pk_false;
  if (ptexpcal_get_operator(e2) != opident) return pk_false;
  id1 = ptexpcal_get_identifier(e1);
  if (ptexpcal_IsOutputFlow(e1))
    {
      ptexpcal_set_identifier(e2, id1);
      ptexpcal_copy_num(e2, e1);
      return pk_true;
    }
  if (IsAuxiliarySignal(e2) && !IsAuxiliarySignal(e1))
    b = pk_true;
  else
    {
      id2 = ptexpcal_get_identifier(e2);
      b = (identGet_componentNumber(id1) == 1) && (identGet_componentNumber(id2) != 1);
    }
  if (b) ptexpcal_set_identifier(e2, id1);
  return b;
}

/**
   If the clock of e1 is represented by the e1 (ie e1 is a signal), it resets it and e2 represented the clock if is a signal and e1 ^= e2.
   @return "true" if e2 represents now the clock of e1,e2, "false" otherwise.
 */
boolean ExprManager::updateClockRepresentedBySignal(ptexpcal e1, ptexpcal  e2)
{
  ptinsth clk_e1;
  ptexpcal aSig;
  if (!IsSignal(e1)) return pk_false;
  clk_e1 = getClock(e1);
  if (! CKM->isClockRepresentedBySignal(clk_e1, &aSig)) return pk_false;
  CKM->setClockRepresentedBySignal(clk_e1, NILEXPCAL);
  if (e1 != aSig)  return pk_false;
  if (!IsSignal(e2)) return pk_false;
  if (! CKM->eqClock(getClock(e2), clk_e1)) return pk_false;
  CKM->setClockRepresentedBySignal(clk_e1, e2);
  return pk_true;
}
