/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/**
   @return a new identifier characterized by 
   - its name (r), 
   - its tree declaration (ar) (link to the abstract tree), 
   - its kind of identifier (ti): signal/constant/parameter/label...
   - its node definition (vnode), 
   - its clock (vclk) 
*/
extern ptlinstid
ptlinstid_new2 (ptident r, ntrees ar, typinst ti, ptINode vnode, ptinsth vclk, TypeType TT)
{
  ptlinstid new_vid;

  new_vid = ptlinstid_new (vnode, vclk);
  new_vid->identorig = r;

  new_vid->tinst = ti;
  new_vid->InOutLocAssert = LocSignal;

  new_vid->IdCategory = undefCategory;
  new_vid->vali = NILEXPCAL;
  new_vid->decl = ar;
  new_vid->vType = TT; // Delcared type.

  new_vid->attribut = NILLINSTIDATTR;
  return new_vid;
}

/** 
    @return the original name (without sufix) of an identifier (vid).
*/
extern void
ptlinstid_set_originalident (ptlinstid vid, ptident id)
{
  vid->identorig = id;
}

/** 
    @return the original name (without sufix) of an identifier (vid).
*/
extern ptident
ptlinstid_get_originalident (ptlinstid vid)
{
  return vid->identorig;
}

/**
   It sets the the name of the operand identifier(vid) to vname.
 */
extern void
ptlinstid_set_identvar (ptlinstid vid, ptident vname)
{
  vid->identvar = vname;
}

/** @return the name of the operand identifier(vid)
 */
extern ptident
ptlinstid_get_identvar (ptlinstid vid)
{
  return vid->identvar;
}

/*
  The identifier (vid) is declared as (parameter, index, signal, constant, label...).
 */
extern void
ptlinstid_set_tinst (ptlinstid vid, typinst vtypinst)
{
  vid->tinst = vtypinst;
}

/** 
    @return the class attribute (signal, constant, index, parameter,,... ) of an identifier (vid).
*/
extern typinst
ptlinstid_get_tinst (ptlinstid vid)
{
  return vid->tinst;
}

/** 
    @return the class attribute (signal, constant, index, parameter,,... ) of an identifier (vid).
*/
extern boolean
ptlinstid_has_tinst (ptlinstid vid, typinst tyId)
{
  return (vid->tinst) == tyId;
}

/**
   @return "true" if an identifier (vid) is a parameter, "false" otherwise.
 */
extern boolean
ptlinstid_is_parameter (ptlinstid vid)
{
  return (boolean) (vid->tinst == tparam);
}

/**
   @return "true" if an identifier (vid) is a constant, "false" otherwise.
 */
extern boolean
ptlinstid_is_constant (ptlinstid vid)
{
  return (boolean) (vid->tinst == tconstant);
}


/**
   @return "true" if an identifier (vid) is a signal, "false" otherwise.
 */
extern boolean
ptlinstid_is_signal (ptlinstid vid)
{
  return (boolean) (vid->tinst == tsignal);
}

/**
   @return "true" if an identifier (vid) is a field, "false" otherwise.
 */
extern boolean
ptlinstid_is_SignalField (ptlinstid vid)
{
  return (boolean) (vid->tinst == tfield);
}

/**
   @return "true" if an identifier (vid) is an index, "false" otherwise.
 */
extern boolean
ptlinstid_is_index (ptlinstid vid)
{
  return (boolean) (vid->tinst == tindex);
}


/**
   @return "true" if an identifier (vid) is a label, "false" otherwise.
 */
extern boolean
ptlinstid_is_label (ptlinstid vid)
{
  return (boolean) (vid->tinst == tlabel);
}

/** @return true if vid is an input signal
 */
extern void
ptlinstid_setInput (ptlinstid vid)
{
  vid->InOutLocAssert = InSignal;
}

/** @return true if vid is an output signal
 */
extern void
ptlinstid_setOutput (ptlinstid vid)
{
  vid->InOutLocAssert = OutSignal;
}

/** @return true if vid is a local signal
 */
extern void
ptlinstid_setLocal (ptlinstid vid)
{
  vid->InOutLocAssert = LocSignal;
}

/** @return true if vid is an input signal
 */
extern boolean
ptlinstid_isInput (ptlinstid vid)
{
  return (vid->InOutLocAssert) == InSignal;
}

/** @return true if vid is an output signal
 */
extern boolean
ptlinstid_isOutput (ptlinstid vid)
{
  return (vid->InOutLocAssert) == OutSignal;
}

/** @return true if vid is a local signal
 */
extern boolean
ptlinstid_isLocal (ptlinstid vid)
{
  return (vid->InOutLocAssert) == LocSignal;
}

/*  vType */
/*  vali */
/** Its sets the initial value of an identifier (vid) to vinit*/
extern void
ptlinstid_set_vali (ptlinstid vid, ptexpcal vinit)
{
  vid->vali = vinit;
}

/** 
    @return the initial value assigned with an identifier (vid).
 */
extern ptexpcal
ptlinstid_get_vali (ptlinstid vid)
{
  return vid->vali;
}

/* decl  */
/** It sets the link to the declaration of the signal operand (vid), to vdecl ( the adress of its declaration in the abstact tree).
 */
extern void
ptlinstid_set_decl (ptlinstid vid, ntrees vdecl)
{
  vid->decl = vdecl;
}


/** @return the declared type of a signal (vid).
*/
extern TypeType
ptlinstid_get_declared_type (ptlinstid vid)
{
  return vid->vType;
}


/** @return the link to the declaration of the signal operand (vid).
    It is the adress of its declaration in the abstact tree.
*/
extern ntrees
ptlinstid_get_decl (ptlinstid vid)
{
  return vid->decl;
}

/* leval  */
/** Its sets the node of an identifier (vid) to vnode.
 */
extern void
ptlinstid_set_leval (ptlinstid vid, ptINode vnode)
{
  vid->leval = vnode;
}

/* vhorl  */
/** 
    Its sets the clock of an identifier (vid) to vclk.
 */
extern void
ptlinstid_set_vhorl (ptlinstid vid, ptinsth vclk)
{
  vid->vhorl = vclk;
}

/**
   @return the clock of an identifier (vid)
*/
extern ptinsth
ptlinstid_get_vhorl (ptlinstid vid)
{
  if (vid == NILLINSTID)
    return NILINSTH;
  return vid->vhorl;
}

/*  expcalass  */
/** Its sets the expression representing the identifier (vid) to vexp*/
extern void
ptlinstid_set_expcalass (ptlinstid vid, ptexpcal vexp)
{
  vid->expcalass = vexp;
}

/** 
    @return the expression assigned with an identifier (vid). 
*/
extern ptexpcal
ptlinstid_get_expcalass (ptlinstid vid)
{
  if (vid != NILLINSTID)
    return vid->expcalass;
  return NILEXPCAL;
}

/* attribut   */

/* stateVariable   */

/*
  An identifier(vid) is declared as a state variable.
*/
extern void
ptlinstid_set_stateTag (ptlinstid vid, boolean b)
{
  if (b)
    vid->IdCategory = statevarSignal;
  else
    vid->IdCategory = undefCategory;
}

/** 
    @return "true" if the identifier (vid) is a state variable, "false" otherwise.
 */
extern boolean
ptlinstid_get_stateTag (ptlinstid vid)
{
  return (vid->IdCategory == statevarSignal);
}

/*
  An identifier(vid) is declared as a shared signal (global signal) if b.
*/
extern void
linstid_set_sharedTag (ptlinstid vid, boolean b)
{
  if (b)
    vid->IdCategory = sharedSignal;
  else
    vid->IdCategory = undefCategory;
}

/** 
    @return "true" if the identifier (vid) is a state variable, "false" otherwise.
 */
extern boolean
linstid_get_sharedTag (ptlinstid vid)
{
  return (vid->IdCategory == sharedSignal);
}

/* ================== compound actions */

/**
   @return a new identifier, a copy of the operand vid, excepted for its node definition, set to vnode.
   The attributes are not copied.
*/
extern ptlinstid
ptlinstid_copy (ptlinstid vid, ptINode vnode)
{
  ptlinstid new_vid;

  new_vid = ptlinstid_new (vnode, vid->vhorl);
  new_vid->identorig = vid->identorig;
  // identvar undef
  new_vid->tinst = vid->tinst;
  new_vid->InOutLocAssert = vid->InOutLocAssert;
  new_vid->vType = vid->vType;
  new_vid->vali = vid->vali;
  new_vid->decl = vid->decl;
  new_vid->IdCategory = vid->IdCategory;
  // leval undef
  // new_vid->vhorl = vid->vhorl;
  // expcalass undef
  new_vid->attribut = NILLINSTIDATTR;
  return new_vid;
}

/**
   @return a new identifier, a copy of the operand vid, excepted for its node definition, set to vnode, and its clock, set to vh.
   The attributes are not copied.   
*/
extern ptlinstid
ptlinstid_copy2 (ptlinstid vid, ptINode vnode, ptinsth vh)
{
  ptlinstid new_vid;

  new_vid = ptlinstid_copy (vid, vnode);
  new_vid->identorig = copy_ident (vid->identorig);	// A copy, why??
  // identvar undef
  // leval undef
  new_vid->vhorl = vh;
  // expcalass undef
  return new_vid;
}

/*  
    Array managing: renaming of the input signals of the iterations of processes.
    @return the signal (seen outside of the iteration) assigned with the operand (vid).
    The indexed signals are chained.
*/
extern ptlinstid
ptlinstid_get_instid_ref (ptlinstid vid)
{
  ptlinstid vaux;
  ptINode vnode;

  vnode = GET_LINSTIDLEVAL (vid);
  if (ptINode_get_instidref (vnode, &vaux))
    {
      if (!(IS_NILLINSTID (vaux)))
	return ptlinstid_get_instid_ref (vaux);
    }
  return vid;
}


/** 
    @return the node that defines an identifier (vid). 
*/
extern ptINode
ptlinstid_get_node (ptlinstid vid)
{
  return (GET_LINSTIDLEVAL (ptlinstid_get_instid_ref (vid)));
}

extern ptmemnode
ptlinstid_get_node_tree_delay (ptlinstid vinst)
{
  ptinstidattr vattr;

  vattr = ptlinstid_getAttribute (vinst);
  return ptlinstidattr_getNodeTreeDelay (vattr);
}

/** Attache ?VMEM au signal ?VSIG */
extern void
ptlinstid_set_node_tree_delay (ptlinstid vinst, ptmemnode vmem)
{
  vinst = ptlinstid_get_instid_ref (vinst);
  SET_LINSTIDATRIBUTE (vinst, ptlinstidattr_setNodeTreeDelay (GET_LINSTIDATRIBUTE (vinst), vinst, vmem));
}


/**
   It frees a signal identifier  descriptor (vintid).
*/
void
ptlinstid_free (ptlinstid vinstid)
{
  delete vinstid;
}
