/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return the size of the list of expressions (lsig). */
extern int
ptlinstances_get_size (ptlinstances lsig)
{
  int nb;
  ptexpcal x;

  nb = 0;
  while (ptlinstances_get_next (&lsig, &x))
    nb++;
  return nb;
}


/**
   @return "true" if the in-list of expressions is not empty, false otherwise.
   @param vlist  (out) : the tail of the in-list if it was not empty.
   @param aSig  (out) : the element of the head of the in-list if it was not empty.
*/
extern boolean
ptlinstances_get_next (ptlinstances * vlist, ptexpcal * aSig)
{
  if ((*vlist) == NILLINSTANCES)
    return pk_false;
  (*aSig) = ptlinstances_gethead (vlist);
  return pk_true;
}

/** @return the list lexp without the element exp (side-effect).  */
extern ptlinstances
ptlinstances_remove_exp (ptlinstances lexp, ptexpcal exp)
{
  ptlinstances linst, pred;
  boolean found;

  if (lexp == NILLINSTANCES)
    return NILLINSTANCES;
  if (lexp->elem == exp)
    return lexp->next;

  pred = lexp;
  linst = lexp;
  found = pk_false;
  lexp = lexp->next;
  while ((lexp != NILLINSTANCES) && !found)
    if (lexp->elem == exp)
      {
	found = pk_true;
	pred->next = lexp->next;
      }
    else
      {
	pred = lexp;
	lexp = lexp->next;
      }
  return linst;
}



/** It adds x at the head of the list lx. 
    @return the new list. */
extern ptlinstances
ptlinstances_push_front (ptlinstances lx, ptexpcal x)
{
  ptlinstances vlist;

  vlist = ptlinstances_create ();
  vlist->elem = x;
  vlist->next = lx;
  return vlist;
}

/** It adds x at the tail of the list lx. 
    @return the new list. */
extern ptlinstances
ptlinstances_push_end (ptlinstances lx, ptexpcal x)
{
  ptlinstances vlist, l, l1;
  ptexpcal n1;

  vlist = ptlinstances_create ();
  vlist->elem = x;
  vlist->next = NILLINSTANCES;
  if (lx == NILLINSTANCES)
    return vlist;
  l1 = lx;
  l = l1;
  while (ptlinstances_get_next (&l1, &n1))
    if (l1 != NILLINSTANCES)
      l = l1;
  l->next = vlist;
  return lx;
}


/** @return the list defined by the intersection of the operand lists (l1 and l2). */
extern ptlinstances
ptlinstances_intersection (ptlinstances l1, ptlinstances l2)
{
  ptexpcal n1;
  ptlinstances list, curlist;

  list = NILLINSTANCES;
  while (ptlinstances_get_next (&l1, &n1))
    if (ptlinstances_isin_exp (l2, n1))
      ptlinstances_isin_insert_exp (&list, &curlist, n1);
  return list;
}


/** @return a reverse copy of the operand list (l).
 */
extern ptlinstances
ptlinstances_reverse (ptlinstances l)
{
  ptlinstances liste_int;
  ptexpcal vexpcal;

  liste_int = NILLINSTANCES;
  while (ptlinstances_get_next (&l, &vexpcal))
    liste_int = ptlinstances_push_front (liste_int, vexpcal);
  return liste_int;
}



/** @return "true" if an element (velem) is in a list (l), "false" otherwise. */
extern boolean
ptlinstances_isin_exp (ptlinstances l, ptexpcal velem)
{
  int numero;

  if (l == NILLINSTANCES)
    return pk_false;
  numero = ptexpcal_get_num (velem);
  while (l != NILLINSTANCES)
    if (ptexpcal_get_num (ptlinstances_gethead (&l)) == numero)
      return pk_true;
  return pk_false;
}



/** @return "true" if  an element (velem) is in a sorted list (l), "false" otherwise. */
extern boolean
ptlinstances_isInSortedList (ptlinstances l, ptexpcal velem)
{
  int m, n;

  if (l == NILLINSTANCES)
    return pk_false;
  m = ptexpcal_get_num (velem);
  while (l != NILLINSTANCES)
    {
      n = ptexpcal_get_num (ptlinstances_gethead (&l));
      if (n == m)
	return pk_true;
      if (n > m)
	return pk_false;
    }
  return pk_false;
}



/**
   Removing an element (velem) of a list of expressions (headList).
   @return "true" if  the element is removed, "false" otherwise.
*/
extern boolean
ptlinstances_isin_remove_exp (ptlinstances * headList, ptexpcal velem)
{
  int numero, cnum;
  ptlinstances curlist, predlist, succlist;
  ptexpcal vcurrent;

  if ((*headList) == NILLINSTANCES)
    return pk_false;
  numero = ptexpcal_get_num (velem);
  predlist = NILLINSTANCES;
  succlist = *headList;
  curlist = succlist;
  while (ptlinstances_get_next (&succlist, &vcurrent))
    {
      cnum = ptexpcal_get_num (vcurrent);
      if (numero < cnum)
	return pk_false;
      if (numero == cnum)
	{
	  if (predlist == NILLINSTANCES)
	    *headList = succlist;
	  else
	    predlist->next = succlist;
	  return pk_true;
	}
      predlist = curlist;
      curlist = succlist;
    }
  return pk_false;
}

/** It inserts (unicity is guaranted) an element(x) in a list (lx).
    @return the new list (side effect on operand list) .*/
extern ptlinstances
ptlinstances_insert (ptlinstances lx, ptexpcal x)
{
  ptlinstances curlist;
  ptlinstances_isin_insert_exp (&lx, &curlist, x);
  return lx;
}

/**
  Inserting an element (velem) in a list (listhead).
  @return "false" if the element is already in the list, "true" otherwise.
  In this case, the list with velem as head element is returned (curlist)
*/
extern boolean
ptlinstances_isin_insert_exp (ptlinstances * headList, ptlinstances * curlist, ptexpcal velem)
{
  boolean bres;
  int numero, cnum;
  ptlinstances succ;

  bres = pk_true;			// LBesnard : added for windows compiler (RTC: run time check)
  if ((*headList) == NILLINSTANCES)
    {
      *headList = ptlinstances_push_front (NILLINSTANCES, velem);
      *curlist = *headList;
    }
  else
    {
      numero = ptexpcal_get_num (velem);
      cnum = ptexpcal_get_num ((*headList)->elem);
      if (numero < cnum)
	{
	  *headList = ptlinstances_push_front (*headList, velem);
	  *curlist = *headList;
	}
      else
	{
	  *curlist = *headList;
	  while (cnum > 0)
	    if (cnum == numero)
	      {
		cnum = 0;
		bres = pk_false;
	      }
	    else
	      {
		succ = (*curlist)->next;
		if (succ == NILLINSTANCES)
		  {
		    succ = ptlinstances_push_front (NILLINSTANCES, velem);
		    (*curlist)->next = succ;
		    cnum = 0;
		  }
		else
		  {
		    cnum = ptexpcal_get_num (succ->elem);
		    if (cnum > numero)
		      {
			succ = ptlinstances_push_front (succ, velem);
			(*curlist)->next = succ;
			cnum = 0;
		      }
		  }
		*curlist = succ;
	      }
	}
    }
  return bres;
}



/**
  @return a list obtained by the merging of operand lists (listToInsert, mainList) with side effects on mainList
*/
extern ptlinstances
ptlinstances_sortedfusion (ptlinstances listToInsert, ptlinstances mainList)
{
  ptlinstances vlinstances;

  if (listToInsert == NILLINSTANCES)
    return mainList;
  ptlinstances_isin_insert_exp (&mainList, &vlinstances, ptlinstances_gethead (&listToInsert));
  while (listToInsert != NILLINSTANCES)
    ptlinstances_isin_insert_exp (&mainList, &vlinstances, ptlinstances_gethead (&listToInsert));
  return mainList;
}


/** @return a copy of a list (lx). */
extern ptlinstances
ptlinstances_copy (ptlinstances lx)
{
  ptlinstances vlist, nlx;
  ptexpcal vexpcal;

  if (lx == NILLINSTANCES)
    return NILLINSTANCES;

  vlist = ptlinstances_create ();
  ptlinstances_gethead_attributs (lx, &vexpcal, &nlx);
  vlist->elem = vexpcal;
  vlist->next = ptlinstances_copy (nlx);
  return vlist;
}


/** Getting the i-th element of a list of expressions.
    @return true if the i-th element exists.
    @param l  (in) a list of expressions.
    @param i  (in) wanted element.
    @param ie  (out) the i-th element if it exists.
*/
extern boolean
ptlinstances_get_ith_element (ptlinstances l, int i, ptexpcal * ie)
{
  int n;

  if (i > 0)
    {
      n = 1;
      while (ptlinstances_get_next (&l, ie))
	{
	  if (n == i)
	    return pk_true;
	  n++;
	}
    }
  return pk_false;
}

/**
   @return the head element of a operand list (l).
*/
extern ptexpcal
ptlinstances_element (ptlinstances l)
{
  return l->elem;
}

/**
   @return the tail of a operand list (l).
*/
extern ptlinstances
ptlinstances_next (ptlinstances l)
{
  return l->next;
}


/**
   @return a list (l) without its last element, and the list (nlast) with the removed element. 
*/
extern ptlinstances
ptlinstances_remove_last_elem (ptlinstances l, ptlinstances * nlast)
{
  ptlinstances aux;

  if (l == NILLINSTANCES)
    {
      *nlast = NILLINSTANCES;
      return NILLINSTANCES;
    }
  if (l->next == NILLINSTANCES)
    {
      *nlast = NILLINSTANCES;
      return NILLINSTANCES;
    }

  aux = l;
  while (aux->next != NILLINSTANCES)
    {
      *nlast = aux;
      aux = aux->next;
    }
  (*nlast)->next = NILLINSTANCES;
  return l;
}

/** @return list defined by l1-l2 (side effect on l1). */
extern ptlinstances
ptlinstances_minus (ptlinstances l1, ptlinstances l2)
{
  ptexpcal exp;

  while (ptlinstances_get_next (&l2, &exp))
    l1 = ptlinstances_remove_exp (l1, exp);
  return l1;
}

/** 
    @return the list l in which e1 is added before e2. If e2 is not in the list, e1 is added at the end of l.
*/
extern ptlinstances
ptlinstances_add_pre (ptlinstances lexp, ptexpcal e1, ptexpcal e2)
{
  ptlinstances linst, pred;

  if (lexp == NILLINSTANCES)
    return ptlinstances_push_front (lexp, e1);
  if (lexp->elem == e1)
    return lexp;
  if (lexp->elem == e2)
    {
      lexp = ptlinstances_remove_exp (lexp, e1);
      return ptlinstances_push_front (lexp, e1);
    }
  pred = lexp;
  linst = lexp;
  lexp = lexp->next;
  while (lexp != NILLINSTANCES)
    {
      if (lexp->elem == e1)
	return linst;
      if (lexp->elem == e2)
	{
	  lexp = ptlinstances_remove_exp (lexp, e1);
	  lexp = ptlinstances_push_front (lexp, e1);
	  pred->next = lexp;
	  return linst;
	}
      pred = lexp;
      lexp = lexp->next;
    }
  ptlinstances_push_end (pred, e1);
  return linst;
}

/** Garbage collector.
    Operand list (l) becomes free.
*/
extern void
ptlinstances_free (ptlinstances l)
{
  ptlinstances ngclinstances;

  if (l != NILLINSTANCES)
    {
      ngclinstances = l;
      while (l->next != NILLINSTANCES)
	l = l->next;
      l->next = gclinstances;
      gclinstances = ngclinstances;
    }
}

/**
   @retun "true" if exp is an element of lexp,"false" otherwise.
   It replaces exp by nexp in lexp. no effect if exp is not an element of lexp.
*/
extern boolean
ptlinstances_replace (ptlinstances lexp, ptexpcal exp, ptexpcal nexp)
{
  while (lexp != NILLINSTANCES)
    if (lexp->elem == exp)
      {
	lexp->elem = nexp;
	return pk_true;
      }
    else
      lexp = lexp->next;
  return pk_false;
}
