/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* Gestion de la representation des valuations */

#include <limits.h>
#include <string.h>

/* ========== bigInteger TO BE DONE ==========
 */

/* IF GMP is used compile with gcc -lgmp command */
// GMP typedef mpz_t bigInteger; to come
typedef long long int bigInteger;
#define NILBIGINTEGER ((bigInteger )0)
#define IS_NILBIGINTEGER(ABIGINTEGER) (ABIGINTEGER == NILBIGINTEGER)

/* to compare repstr use  
   boolean strInteger_leq(ptrepstr str1, ptrepstr str2, boolean * strict)
 */

/** 
    @return a newly created empty integer cell
*/
static bigInteger
bigInteger_newEmpty ()
{
  bigInteger aBigInteger;
  // GMP mpz_init(aBigInteger);
  aBigInteger = 0;
  return aBigInteger;
}

/** 
    if aBigInteger is NIL, assign to aBigInteger a  newly created integer cell
    assign to (the resulting) aBigInteger an integer represented by aString 
    @param aString represents an unbounded integer in base 10
    @return (the resulting) aBigIntege
*/
static bigInteger
bigInteger_assign_str (bigInteger aBigInteger, ptrepstr aString)
{
  if (IS_NILBIGINTEGER (aBigInteger))
    aBigInteger = bigInteger_newEmpty ();
  // GMP else
  // GMP   mpz_clear(aBigInteger);
  // GMP mpz_set_str(aBigInteger, repstr2int(aString), 10);
  return aBigInteger;
}

/** 
    if aBigInteger is not NIL, cancel the current value of  aBigInteger
*/
static void
bigInteger_reset (bigInteger aBigInteger)
{
  if (IS_NILBIGINTEGER (aBigInteger))
    return;
  // mpz_clear(aBigInteger);
  aBigInteger = 0;
}

/* ========== integerCell ==========
 */
typedef struct integerCell *ptIntegerCell;
#define NIL_IntegerCell (ptIntegerCell)0
#define IS_NILINTEGERCELL(PTINTEGERCELL) ((PTINTEGERCELL) == (NIL_IntegerCell))

#define FALSE_INTEGERVALUE 0
#define TRUE_INTEGERVALUE 1

typedef struct integerCell
{
  boolean isRepresentable;	/* pk_true: is a compiler integer; false: is a too large integer */
  ptrepstr stringValue;
  compilerInteger integerValue;
//  bigInteger bigValue; 
} tintegerCell;

#define PTINTEGERCELL_SETSTANDARD(PTINTEGERCELL) (PTINTEGERCELL)->isRepresentable = pk_true
#define PTINTEGERCELL_ISSTANDARD(PTINTEGERCELL) ((PTINTEGERCELL)->isRepresentable)
#define PTINTEGERCELL_SETTOOLARGEINTEGER(PTINTEGERCELL) (PTINTEGERCELL)->isRepresentable = pk_false

#define PTINTEGERCELL_GETSTRINGVALUE(PTINTEGERCELL) ((PTINTEGERCELL)->stringValue)
#define PTINTEGERCELL_SETSTRINGVALUE(PTINTEGERCELL, AVALUE) (PTINTEGERCELL)->stringValue = (AVALUE)
#define PTINTEGERCELL_HASSTRINGVALUE(PTINTEGERCELL) ( (PTINTEGERCELL)->stringValue != NIL_repstr )
#define PTINTEGERCELL_SETNILSTRINGVALUE(PTINTEGERCELL) PTINTEGERCELL_SETSTRINGVALUE(PTINTEGERCELL, NIL_repstr)

#define PTINTEGERCELL_GETINTEGERVALUE(PTINTEGERCELL) ((PTINTEGERCELL)->integerValue)
#define PTINTEGERCELL_SETINTEGERVALUE(PTINTEGERCELL, AVALUE) (PTINTEGERCELL)->integerValue = (AVALUE)

// #define PTINTEGERCELL_SETBIGVALUE(PTINTEGERCELL, AVALUE) ((PTINTEGERCELL)->bigValue) = (AVALUE)
// #define PTINTEGERCELL_GETBIGVALUE(PTINTEGERCELL) ((PTINTEGERCELL)->bigValue)

/** assign aString to a newly created integer cell
    @param aString is the string notation of an integer value in base 10 (not checked)
 * safe provided that enough space memory is available
 */
static ptIntegerCell
integerCell_newBigInteger (ptrepstr aString)
{
  ptIntegerCell anIntegerCell = new integerCell ();
  PTINTEGERCELL_SETTOOLARGEINTEGER (anIntegerCell);
  PTINTEGERCELL_SETSTRINGVALUE (anIntegerCell, aString);
  return anIntegerCell;
}

/** assign anIntValue and aString to newly created integer cell
    @param anIntValue is a representable integer
    @param aString is the string notation of anIntValue in base 10 (not checked)
 * safe provided that aString is a consistant representation of anIntValue and enough space memory is available
 */
static ptIntegerCell
integerCell_newInteger (compilerInteger anIntValue, ptrepstr aString)
{
  ptIntegerCell anIntegerCell = new integerCell ();
  PTINTEGERCELL_SETSTANDARD (anIntegerCell);
  PTINTEGERCELL_SETSTRINGVALUE (anIntegerCell, aString);
  PTINTEGERCELL_SETINTEGERVALUE (anIntegerCell, anIntValue);
  return anIntegerCell;
}

/* MAIN entry points */

/** @return true if anIntegerCell contains a represented integer, false otherwise
    @param result is assigned the integer value when integerCell_getIntegerValue returns pk_true
    @param result remains unchanged otherwise
 * safe 
 */
static boolean
integerCell_getIntegerValue (ptIntegerCell anIntegerCell, compilerInteger * result)
{
  if (IS_NILINTEGERCELL (anIntegerCell))
    return pk_false;
  if (PTINTEGERCELL_ISSTANDARD (anIntegerCell))
    {
      (*result) = PTINTEGERCELL_GETINTEGERVALUE (anIntegerCell);
      return pk_true;
    }
  return pk_false;
}

/* ========== valueHolder ==========
 */
typedef struct valueHolder *ptValueHolder;
#define NILVALUEHOLDER (ptValueHolder)0
#define IS_NILVALUEHOLDER(PTVALUEHOLDER) ((PTVALUEHOLDER) == ( NILVALUEHOLDER))

typedef enum
{ free_tag, defined_tag, bounded_tag, valued_tag } definitionSTATUS;
typedef enum
{ innerAny, innerBool, innerInteger, innerBigInt } valueTYPE;

typedef struct valueHolder
{
  valueTYPE actualType;
  definitionSTATUS valueStatus;	/* infimum /= supremum when status is bounded_tag tag */
  ptIntegerCell infimum;
  ptIntegerCell supremum;	/* infimum = supremum when status is valued tag */
} tvalueHolder;

#define PTVALUEHOLDER_GETTYPE(PTVALUEHOLDER) ((PTVALUEHOLDER)->actualType)
#define PTVALUEHOLDER_SETTYPE(PTVALUEHOLDER, ACTUALTYPE) PTVALUEHOLDER_GETTYPE(PTVALUEHOLDER) = (ACTUALTYPE)
#define PTVALUEHOLDER_SETBOOLTYPE(PTVALUEHOLDER)    PTVALUEHOLDER_SETTYPE(PTVALUEHOLDER, innerBool)
#define PTVALUEHOLDER_SETINTEGERTYPE(PTVALUEHOLDER)   PTVALUEHOLDER_SETTYPE(PTVALUEHOLDER, innerInteger)
#define PTVALUEHOLDER_SETBIGINTEGERTYPE(PTVALUEHOLDER) PTVALUEHOLDER_SETTYPE(PTVALUEHOLDER, innerBigInt)

#define PTVALUEHOLDER_HASTYPE(PTVALUEHOLDER, ACTUALTYPE) ( PTVALUEHOLDER_GETTYPE(PTVALUEHOLDER) == (ACTUALTYPE) )
#define PTVALUEHOLDER_ISBOOL(PTVALUEHOLDER)    (PTVALUEHOLDER_HASTYPE(PTVALUEHOLDER, innerBool))
#define PTVALUEHOLDER_ISINTEGER(PTVALUEHOLDER)   (PTVALUEHOLDER_HASTYPE(PTVALUEHOLDER, innerInteger))
#define PTVALUEHOLDER_ISBIGINTEGER(PTVALUEHOLDER)   (PTVALUEHOLDER_HASTYPE(PTVALUEHOLDER, innerBigInt))

#define PTVALUEHOLDER_GETSTATUS(PTVALUEHOLDER) ((PTVALUEHOLDER)->valueStatus)
#define PTVALUEHOLDER_SETSTATUS(PTVALUEHOLDER, STATUS) PTVALUEHOLDER_GETSTATUS(PTVALUEHOLDER) = (STATUS)
#define PTVALUEHOLDER_SETFREE(PTVALUEHOLDER)    PTVALUEHOLDER_SETSTATUS(PTVALUEHOLDER,free_tag)
#define PTVALUEHOLDER_SETDEFINED(PTVALUEHOLDER) PTVALUEHOLDER_SETSTATUS(PTVALUEHOLDER,defined_tag)
#define PTVALUEHOLDER_SETBOUNDED(PTVALUEHOLDER) PTVALUEHOLDER_SETSTATUS(PTVALUEHOLDER,bounded_tag)
#define PTVALUEHOLDER_SETVALUED(PTVALUEHOLDER)  PTVALUEHOLDER_SETSTATUS(PTVALUEHOLDER,valued_tag)

#define PTVALUEHOLDER_ISSTATUS(PTVALUEHOLDER, STATUS) (PTVALUEHOLDER_GETSTATUS(PTVALUEHOLDER) == (STATUS))
#define PTVALUEHOLDER_ISFREE(PTVALUEHOLDER)    (PTVALUEHOLDER_ISSTATUS(PTVALUEHOLDER, free_tag))
#define PTVALUEHOLDER_ISDEFINED(PTVALUEHOLDER) (PTVALUEHOLDER_ISSTATUS(PTVALUEHOLDER, defined_tag))
#define PTVALUEHOLDER_ISBOUNDED(PTVALUEHOLDER) (PTVALUEHOLDER_ISSTATUS(PTVALUEHOLDER, bounded_tag))
#define PTVALUEHOLDER_ISVALUED(PTVALUEHOLDER)  (PTVALUEHOLDER_ISSTATUS(PTVALUEHOLDER, valued_tag))

#define PTVALUEHOLDER_GETINFIMUM(PTVALUEHOLDER) ((PTVALUEHOLDER)->infimum)
#define PTVALUEHOLDER_SETINFIMUM(PTVALUEHOLDER, ANINTEGERCELL) PTVALUEHOLDER_GETINFIMUM(PTVALUEHOLDER) = (ANINTEGERCELL)
#define PTVALUEHOLDER_SETNILINFIMUM(PTVALUEHOLDER) PTVALUEHOLDER_SETINFIMUM(PTVALUEHOLDER, NIL_IntegerCell)
#define PTVALUEHOLDER_HASINFIMUM(PTVALUEHOLDER) (!IS_NILINTEGERCELL(PTVALUEHOLDER_GETINFIMUM(PTVALUEHOLDER)))

#define PTVALUEHOLDER_GETSUPREMUM(PTVALUEHOLDER) ((PTVALUEHOLDER)->supremum)
#define PTVALUEHOLDER_SETSUPREMUM(PTVALUEHOLDER, ANINTEGERCELL) PTVALUEHOLDER_GETSUPREMUM(PTVALUEHOLDER) = (ANINTEGERCELL)
#define PTVALUEHOLDER_SETNILSUPREMUM(PTVALUEHOLDER) PTVALUEHOLDER_SETSUPREMUM(PTVALUEHOLDER, NIL_IntegerCell)
#define PTVALUEHOLDER_HASSUPREMUM(PTVALUEHOLDER) (!IS_NILINTEGERCELL(PTVALUEHOLDER_GETSUPREMUM(PTVALUEHOLDER)))

#define PTVALUEHOLDER_GETACTUAL(PTVALUEHOLDER) PTVALUEHOLDER_GETINFIMUM(PTVALUEHOLDER)
#define PTVALUEHOLDER_SETACTUAL(PTVALUEHOLDER, ANINTEGERCELL) { PTVALUEHOLDER_SETINFIMUM(PTVALUEHOLDER, ANINTEGERCELL); PTVALUEHOLDER_SETSUPREMUM(PTVALUEHOLDER, ANINTEGERCELL); }
#define PTVALUEHOLDER_SETNILACTUAL(PTVALUEHOLDER) { PTVALUEHOLDER_SETNILINFIMUM(PTVALUEHOLDER); PTVALUEHOLDER_SETNILSUPREMUM(PTVALUEHOLDER); }
#define PTVALUEHOLDER_HASACTUAL(PTVALUEHOLDER) PTVALUEHOLDER_HASINFIMUM(PTVALUEHOLDER)

// ( ISNIL_PTVALUEHOLDER(PTVALUEHOLDER) ? NILVALUE: )

#define PTVALUEHOLDER_GETINTEGERVALUE(PTVALUEHOLDER) ((PTVALUEHOLDER)->integerValue)
#define PTVALUEHOLDER_GETINTEGERBSUP(PTVALUEHOLDER) ((PTVALUEHOLDER)->integerBsup)


static ptValueHolder valueHolder_ERROR = NILVALUEHOLDER, valueHolder_TRUE =
  NILVALUEHOLDER, valueHolder_FALSE = NILVALUEHOLDER, valueHolder_BOOL =
  NILVALUEHOLDER, valueHolder_0 = NILVALUEHOLDER, valueHolder_1 = NILVALUEHOLDER;

static compilerInteger BigInt_0 = (compilerInteger) 0;
static char *string_INTMAX_MIN;
static int INTMAX_MAX_length, INTMAX_MIN_length;
static ptrepstr repstr_LLONG_MAX;

/** @return true if the string is not larger than the greatest representable integer
 * safe
 */
static boolean
integerCell_isNotTooLarge (ptrepstr anIntegerString)
{
  boolean strict;
  return repstrInteger_leq (anIntegerString, repstr_LLONG_MAX, &strict);
}

extern void
DEBUGindent_valueHolder (const char *leftIndent, int number, ptValueHolder avalueHolder)
{
  if (IS_NILVALUEHOLDER (avalueHolder))
    {
      printf ("%s valueHolder of %i is NIL\n", leftIndent, number);
      fflush (stdout);
      return;
    }
  printf ("%s %i has a valueHolder\n", leftIndent, number);
  fflush (stdout);
  if (PTVALUEHOLDER_ISFREE (avalueHolder))
    {
      printf ("    %s valueHolder is free\n", leftIndent);
      fflush (stdout);
    }
  else if (PTVALUEHOLDER_ISDEFINED (avalueHolder))
    {
      printf ("    %s valueHolder is defined\n", leftIndent);
      fflush (stdout);
    }
  else if (PTVALUEHOLDER_ISBOUNDED (avalueHolder))
    {
      printf ("    %s valueHolder is bounded\n", leftIndent);
      fflush (stdout);
    }
  else if (PTVALUEHOLDER_ISVALUED (avalueHolder))
    {
      printf ("    %s valueHolder is valued\n", leftIndent);
      fflush (stdout);
    }
  else
    {
      printf ("    %s valueHolder status is unknown\n", leftIndent);
      fflush (stdout);
    }
}


/** -->true when ?VPTVALUEHOLDER is nil or has meaningless fields */
static boolean
valueHolder_isFree (ptValueHolder vptValueHolder)
{
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return pk_true;
  return PTVALUEHOLDER_ISFREE (vptValueHolder);
}

/** @return true when ?VPTVALUEHOLDER is the value holder of an existing expression
    numerical fields are meaningless*/
static boolean
valueHolder_isDefined (ptValueHolder vptValueHolder)
{
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return pk_false;
  return PTVALUEHOLDER_ISDEFINED (vptValueHolder);
}

/** @return true when ?VPTVALUEHOLDER is the bounded value or the valued holder of an existing expression 
    value gield may be meaningless
 */
static boolean
valueHolder_isBounded (ptValueHolder vptValueHolder)
{
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return pk_false;
  return ((PTVALUEHOLDER_ISVALUED (vptValueHolder)) || (PTVALUEHOLDER_ISBOUNDED (vptValueHolder)));
}

/**-->True si ?VPTVALUEHOLDER est value (entier positif)*/
static boolean
valueHolder_isValued (ptValueHolder vptValueHolder)
{
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return pk_false;
  return PTVALUEHOLDER_ISVALUED (vptValueHolder);
}

/** @return true if vptValueHolder is valued and denotes an integer or a boolean
 *  @param isBig is set to pk_true if this integer is too large to be represented
                           false otherwise
*/
static boolean
valueHolder_isIntegerValued (ptValueHolder vptValueHolder, boolean * isBig)
{
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return pk_false;
  if (PTVALUEHOLDER_ISVALUED (vptValueHolder))
    {
      (*isBig) = PTVALUEHOLDER_ISBIGINTEGER (vptValueHolder);
      return pk_true;
    }
  return pk_false;
}

/** @return a new free value holder */
static ptValueHolder
valueHolder_newValueHolder (definitionSTATUS itsStatus)
{
  ptValueHolder vptValueHolder;
  // printf("DEBUG entering valueHolder_newValueHolder\n"); fflush(stdout);

  vptValueHolder = new valueHolder ();
  PTVALUEHOLDER_SETSTATUS (vptValueHolder, itsStatus);
  PTVALUEHOLDER_SETTYPE (vptValueHolder, innerAny);
  PTVALUEHOLDER_SETNILACTUAL (vptValueHolder);
  return vptValueHolder;
}

/**
   set vptValueHolder as a boolean constant descriptor 
   vptValueHolder is not nil (not checked)
   anIntValue is a consistent integer value for pk_true or false
   aString is "true" or "false"
*/
static void
unsafeValueHolder_assignBooleanValue (ptValueHolder vptValueHolder, compilerInteger anIntValue, ptrepstr aString)
{
  PTVALUEHOLDER_SETVALUED (vptValueHolder);
  PTVALUEHOLDER_SETBOOLTYPE (vptValueHolder);
  PTVALUEHOLDER_SETACTUAL (vptValueHolder, integerCell_newInteger (anIntValue, aString));
}

/**
   set vptValueHolder as a representable integer constant descriptor 
   vptValueHolder is not nil (not checked)
   anIntValue is an integer value
   aString is the decimal representation of anIntValue (not checked)
*/
static void
unsafeValueHolder_assignIntegerValue (ptValueHolder vptValueHolder, compilerInteger anIntValue, ptrepstr aString)
{
  PTVALUEHOLDER_SETVALUED (vptValueHolder);
  PTVALUEHOLDER_SETINTEGERTYPE (vptValueHolder);
  PTVALUEHOLDER_SETACTUAL (vptValueHolder, integerCell_newInteger (anIntValue, aString));
}

/**
   set vptValueHolder as a non representable integer constant descriptor 
   vptValueHolder is not nil (not checked)
   anIntegerString is the decimal representation of the large integer (not checked)
*/
static void
unsafeValueHolder_assignBigIntegerValue (ptValueHolder vptValueHolder, ptrepstr anIntegerString)
{
  PTVALUEHOLDER_SETVALUED (vptValueHolder);
  PTVALUEHOLDER_SETBIGINTEGERTYPE (vptValueHolder);
  PTVALUEHOLDER_SETACTUAL (vptValueHolder, integerCell_newBigInteger (anIntegerString));
}

/**
   set vptValueHolder as an integer constant descriptor 
   vptValueHolder is not nil (not checked)
   anIntegerString is the decimal representation of an integer (not checked)
   return pk_true if the integer is representable
*/
static boolean
unsafeValueHolder_assignStringIntegerValue (ptValueHolder vptValueHolder, ptrepstr anIntegerString)
{
  if (integerCell_isNotTooLarge (anIntegerString))
    {
      unsafeValueHolder_assignIntegerValue (vptValueHolder, repstr2integer (anIntegerString), anIntegerString);
      return pk_true;
    }
  printf ("WARNING the value %s is not an integer type value in this compiler\n", repstr2string (anIntegerString));
  fflush (stdout);
  unsafeValueHolder_assignBigIntegerValue (vptValueHolder, anIntegerString);
  return pk_true;
}

/**
   set vptValueHolder as a bounded integer descriptor 
   vptValueHolder is not nil (not checked)
   bi is smaller than bs  (not checked)
*/
static void
unsafeValueHolder_assignIntegerBounds (ptValueHolder vptValueHolder, compilerInteger bi, compilerInteger bs)
{
  PTVALUEHOLDER_SETBOUNDED (vptValueHolder);
  PTVALUEHOLDER_SETINTEGERTYPE (vptValueHolder);
  PTVALUEHOLDER_SETINFIMUM (vptValueHolder, integerCell_newInteger (bi, integer2repstr (bi)));
  PTVALUEHOLDER_SETSUPREMUM (vptValueHolder, integerCell_newInteger (bs, integer2repstr (bs)));
}

/**
   @return the representable integer value associated with vptValueHolder
   @return 0 if vptValueHolder is nil or if its associated value is not representable
*/
static compilerInteger
unsafeValueHolder_get_integerValue (ptValueHolder vptValueHolder)
{
  ptIntegerCell anIntegerCell;
  compilerInteger result;
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return 0;
  if (PTVALUEHOLDER_ISVALUED (vptValueHolder))
    {
      anIntegerCell = PTVALUEHOLDER_GETACTUAL (vptValueHolder);
      if (integerCell_getIntegerValue (anIntegerCell, &result))
	return result;
    }
  return 0;
}

/**
   @return true if vptValueHolder denotes a representable integer interval
   if this is the case, the lower bound is assigned to bi and the upper bound to bs
*/
static boolean
valueHolder_get_integerBounds (ptValueHolder vptValueHolder, compilerInteger * bi, compilerInteger * bs)
{
  if (IS_NILVALUEHOLDER (vptValueHolder))
    return pk_false;
  if (PTVALUEHOLDER_ISVALUED (vptValueHolder))
    {
      if (!integerCell_getIntegerValue (PTVALUEHOLDER_GETINFIMUM (vptValueHolder), bi))
	return pk_false;
      (*bs) = (*bi);
      return pk_true;
    }
  if (PTVALUEHOLDER_ISBOUNDED (vptValueHolder))
    {
      if (!integerCell_getIntegerValue (PTVALUEHOLDER_GETINFIMUM (vptValueHolder), bi))
	return pk_false;
      return integerCell_getIntegerValue (PTVALUEHOLDER_GETSUPREMUM (vptValueHolder), bs);
    }
  return pk_false;
}

/**
  @return the uniquely defined representation of Error ptValueHolder
*/
ptValueHolder
valueHolder_getError ()
{
  if (valueHolder_ERROR == NILVALUEHOLDER)
    {
      valueHolder_ERROR = valueHolder_newValueHolder (valued_tag);
      unsafeValueHolder_assignIntegerValue (valueHolder_ERROR, 0, integer2repstr (0));
    }
  return valueHolder_ERROR;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_getBooleanFree ()
{

  // printf("DEBUG entering valueHolder_getBooleanFree\n"); fflush(stdout);
  if (valueHolder_BOOL == NILVALUEHOLDER)
    {
      valueHolder_BOOL = valueHolder_newValueHolder (free_tag);
      PTVALUEHOLDER_SETBOOLTYPE (valueHolder_BOOL);
    }
  return valueHolder_BOOL;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_getFalse ()
{

  // printf("DEBUG entering valueHolder_getFalse\n"); fflush(stdout);
  if (valueHolder_FALSE == NILVALUEHOLDER)
    {
      valueHolder_FALSE = valueHolder_newValueHolder (valued_tag);
      unsafeValueHolder_assignBooleanValue (valueHolder_FALSE, 0, string2repstr ("false"));
    }
  return valueHolder_FALSE;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_getTrue ()
{
  // printf("DEBUG entering valueHolder_getFalse\n"); fflush(stdout);
  if (valueHolder_TRUE == NILVALUEHOLDER)
    {
      valueHolder_TRUE = valueHolder_newValueHolder (valued_tag);
      unsafeValueHolder_assignBooleanValue (valueHolder_TRUE, 1, string2repstr ("true"));
    }
  return valueHolder_TRUE;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_createIntegerConstant (compilerInteger val)
{
  ptValueHolder vptValueHolder;
  vptValueHolder = valueHolder_newValueHolder (valued_tag);
  unsafeValueHolder_assignIntegerValue (vptValueHolder, val, integer2repstr (val));
  // printf("DEBUG  exiting valueHolder_createIntegerConstant, value is %li, created holder is %li\n", (long int)val, (long int)vptValueHolder); fflush(stdout);
  return vptValueHolder;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_createStringIntegerConstant (ptrepstr integerStringValue)
{
  ptValueHolder vptValueHolder;
  vptValueHolder = valueHolder_newValueHolder (valued_tag);
  unsafeValueHolder_assignStringIntegerValue (vptValueHolder, integerStringValue);
  // printf("DEBUG  exiting valueHolder_createStringIntegerConstant, value is %s, created holder is %li\n", repstr2string(integerStringValue), (long int)vptValueHolder); fflush(stdout);
  return vptValueHolder;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_createIntegerBounds (compilerInteger bi, compilerInteger bs)
{
  ptValueHolder vptValueHolder;
  vptValueHolder = valueHolder_newValueHolder (bounded_tag);
  unsafeValueHolder_assignIntegerBounds (vptValueHolder, bi, bs);
  return vptValueHolder;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_get0 ()
{
  // printf("DEBUG entering valueHolder_getFalse\n"); fflush(stdout);
  if (valueHolder_0 == NILVALUEHOLDER)
    valueHolder_0 = valueHolder_createIntegerConstant (0);
  return valueHolder_0;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_get1 ()
{
  // printf("DEBUG entering valueHolder_getFalse\n"); fflush(stdout);
  if (valueHolder_1 == NILVALUEHOLDER)
    valueHolder_1 = valueHolder_createIntegerConstant (1);
  return valueHolder_1;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_provideIntegerRepresentation (compilerInteger val)
{
  if (val == 0)
    return valueHolder_get0 ();
  if (val == 1)
    return valueHolder_get1 ();
  return valueHolder_createIntegerConstant (val);
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptValueHolder
valueHolder_provideIntegerBoundsRepresentation (compilerInteger bi, compilerInteger bs)
{
  if (bi == bs)
    return valueHolder_provideIntegerRepresentation (bi);
  return valueHolder_createIntegerBounds (bi, bs);
}


/** @return true when aValueHolder actual value is equal to FALSE_INTEGERVALUE
*/
extern boolean
unsafeValueHolder_equalBool (ptValueHolder aValueHolder, int aValue)
{
  compilerInteger result;
  if (PTVALUEHOLDER_ISBOOL (aValueHolder))
    {
      if (PTVALUEHOLDER_ISVALUED (aValueHolder))
	{
	  ptIntegerCell anIntegerCell = PTVALUEHOLDER_GETACTUAL (aValueHolder);
	  if (integerCell_getIntegerValue (anIntegerCell, &result))
	    return (result == aValue);
	}
    }
  return pk_false;
}

/** @return true when aValueHolder actual value is equal to FALSE_INTEGERVALUE
*/
extern boolean
valueHolder_equalFalse (ptValueHolder aValueHolder)
{
  compilerInteger result;
  if (IS_NILVALUEHOLDER (aValueHolder))
    return pk_false;
  if (aValueHolder == valueHolder_FALSE)
    return pk_true;
  return unsafeValueHolder_equalBool (aValueHolder, FALSE_INTEGERVALUE);
}

/** @return true when aValueHolder actual value is equal to FALSE_INTEGERVALUE
*/
extern boolean
valueHolder_equalTrue (ptValueHolder aValueHolder)
{
  compilerInteger result;
  if (IS_NILVALUEHOLDER (aValueHolder))
    return pk_false;
  if (aValueHolder == valueHolder_TRUE)
    return pk_true;
  return unsafeValueHolder_equalBool (aValueHolder, TRUE_INTEGERVALUE);
}

/** @return true when aValueHolder actual value is equal to FALSE_INTEGERVALUE
*/
extern boolean
unsafeValueHolder_equalInteger (ptValueHolder aValueHolder, compilerInteger aValue)
{
  compilerInteger result;
  if (PTVALUEHOLDER_ISINTEGER (aValueHolder))
    {
      if (PTVALUEHOLDER_ISVALUED (aValueHolder))
	{
	  ptIntegerCell anIntegerCell = PTVALUEHOLDER_GETACTUAL (aValueHolder);
	  if (integerCell_getIntegerValue (anIntegerCell, &result))
	    return (result == aValue);
	}
    }
  return pk_false;
}

/** @return true when aValueHolder actual value is equal to 0
*/
extern boolean
valueHolder_equal0 (ptValueHolder aValueHolder)
{
  if (IS_NILVALUEHOLDER (aValueHolder))
    return pk_false;
  if (aValueHolder == valueHolder_0)
    return pk_true;
  return unsafeValueHolder_equalInteger (aValueHolder, 0);
}

/** @return true when aValueHolder actual value is equal to 1
*/
extern boolean
valueHolder_equal1 (ptValueHolder aValueHolder)
{
  if (IS_NILVALUEHOLDER (aValueHolder))
    return pk_false;
  if (aValueHolder == valueHolder_1)
    return pk_true;
  return unsafeValueHolder_equalInteger (aValueHolder, 1);
}

static void
DEBUG_compar (ptrepstr intStr1, ptrepstr intStr2)
{
  boolean strict;
  if (repstrInteger_leq (intStr1, intStr2, &strict))
    if (strict)
      printf ("DEBUG %s is strictly smaller than %s\n", repstr2string (intStr1), repstr2string (intStr2));
    else
      printf ("DEBUG %s is equal to %s\n", repstr2string (intStr1), repstr2string (intStr2));
  else
    printf ("DEBUG %s is strictly greater than %s\n", repstr2string (intStr1), repstr2string (intStr2));
  fflush (stdout);
}

/** 
   Associe la qualite DEFINI a ?VPTVALUEHOLDER 
   (c.a.d attribut associe a un signal en cours de definition) 
*/
static void
valueHolder_init ()
{
  if (valueHolder_ERROR != NILVALUEHOLDER)	/* yet initialized */
    return;
  // printf("DEBUG SHRT_MAX is %i\n", SHRT_MAX);
  // printf("DEBUG SHRT_MIN is %i\n", SHRT_MIN);
  // printf("DEBUG INT_MAX is %i\n", INT_MAX);
  // printf("DEBUG INT_MIN is %i\n", INT_MIN);
  // printf("DEBUG LONG_MAX is %li\n", LONG_MAX);
  // printf("DEBUG LONG_MIN is %li\n", LONG_MIN);
  // printf("DEBUG LLONG_MAX is %lli\n", LLONG_MAX);
  // printf("DEBUG LLONG_MIN is %lli\n", LLONG_MIN);
  repstr_LLONG_MAX = integer2repstr (LLONG_MAX);
  // printf("DEBUG repstr_LLONG_MAX is %lli\n", repstr2integer(repstr_LLONG_MAX));

  valueHolder_ERROR = valueHolder_getError ();
  valueHolder_BOOL = valueHolder_getBooleanFree ();
  valueHolder_FALSE = valueHolder_getFalse ();
  valueHolder_TRUE = valueHolder_getTrue ();
  valueHolder_0 = valueHolder_get0 ();
  valueHolder_1 = valueHolder_get1 ();
  string_INTMAX_MIN = repstr2string (integer2repstr (-(9223372036854775807 - 1)));
  // string_INTMAX_MIN = repstr2string(integer2repstr(INTMAX_MIN));
  // printf("DEBUG string_INTMAX_MIN is %s, its length is %i\n",  string_INTMAX_MIN, INTMAX_MIN_length); fflush(stdout);
  // printf("DEBUG exiting  valueHolder_init\n"); fflush(stdout);

  // DEBUG_compar(repstr_LLONG_MAX, repstr_LLONG_MAX);
  // DEBUG_compar(string2repstr("0"), string2repstr("0"));
  // DEBUG_compar(string2repstr("0"), string2repstr("1"));
  // DEBUG_compar(string2repstr("1"), string2repstr("0"));
  // DEBUG_compar(string2repstr("1"), string2repstr("1"));
  // DEBUG_compar(string2repstr("1"), string2repstr("10"));
  // DEBUG_compar(string2repstr("10"), string2repstr("1"));
  // DEBUG_compar(string2repstr("10"), string2repstr("10"));
}
