/* 
   Authors: Ola¯c Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** DEBUG
*/
extern void
DEBUG_mess_ptexpcal (const char *mess, const char *name, ptexpcal vexpcal)
{
  if (vexpcal == NILEXPCAL)
    {
      printf ("DEBUG %s: ptexpcal %s is NIL\n", mess, name);
      fflush (stdout);
      return;
    }
  printf ("DEBUG %s: ptexpcal %s adress is %li; its number is %i \n", mess, name, (long int) vexpcal, ptexpcal_get_num (vexpcal));
  fflush (stdout);
  printf ("DEBUG      Type is %i, Operator is %i\n", (int) GET_ACTUALTYPE (vexpcal), (int) GET_OPERATOR (vexpcal));
  DEBUGindent_valueHolder ("          ", ptexpcal_get_num (vexpcal), GET_VALUEHOLDER (vexpcal));
}

extern void
DEBUG_ptexpcal_TRUE (char *mess)
{
  DEBUG_mess_ptexpcal (mess, "ptexpcal_TRUE", ptexpcal_TRUE);
}

extern void
DEBUG_ptexpcal_FALSE (char *mess)
{
  DEBUG_mess_ptexpcal (mess, "ptexpcal_FALSE", ptexpcal_FALSE);
}

/** signal declare de l'expression exp (elinime l'acces au champs) */
extern ptexpcal
ptexpcal_getDeclaredSignal (ptexpcal anExpCal)
{
  noperators op;
  if (anExpCal != NILEXPCAL)
    {
      op = ptexpcal_get_operator (anExpCal);
      while ((op == opcham) || (op == opelem))
	{
	  anExpCal = ptexpcal_get_binaryArgstAt (anExpCal, 1);
	  op = ptexpcal_get_operator (anExpCal);
	}
    }
  return anExpCal;
}

/**-->classe de l'expression ?PT*/
extern cloper
ptexpcal_get_operatorClass (ptexpcal exp)
{
  if (exp == NILEXPCAL)
    return undef_cloper;
  return GET_OPERATORCLASS (exp);
}

/** -->true si l'expression ?VEXPCAL represente un signal */
extern boolean
ptexpcal_IsSignal (ptexpcal vexpcal)
{
  if (vexpcal == NILEXPCAL)
    return pk_false;
  if (GET_OPERATORCLASS (vexpcal) == variabl)
    return (ptlinstid_is_signal (ptexpcal_get_instance (vexpcal)));
  return pk_false;
}

/** Le signal ?VEXPCAL denote un flot d'entree. L'identificateur ?VEXPCAL a du etre declare en tant que signal. */
extern void
ptexpcal_belongs_InFlow (ptexpcal vexpcal)
{
  if (ptexpcal_IsSignal (vexpcal))
    ptlinstid_setInput (ptexpcal_get_instance (vexpcal));
}

/** Le signal ?VEXPCAL denote un flot de sorties. L'identificateur ?VEXPCAL a du etre declare en tant que signal.*/
extern void
ptexpcal_belongs_OutFlow (ptexpcal vexpcal)
{
  if (ptexpcal_IsSignal (vexpcal))
    ptlinstid_setOutput (ptexpcal_get_instance (vexpcal));
}

/** Le signal ?VEXPCAL denote un flot local. L'identificateur ?VEXPCAL a du etre declare en tant que signal. */
extern void
ptexpcal_belongs_LocalFlow (ptexpcal vexpcal)
{
  if (ptexpcal_IsSignal (vexpcal))
    ptlinstid_setLocal (ptexpcal_get_instance (vexpcal));
}

/** Le signal ?VEXPCAL denote un flot d'assertion. L'identificateur ?VEXPCAL a du etre declare en tant que signal. */
extern void
ptexpcal_belongs_AssertFlow (ptexpcal vexpcal)
{
  if (vexpcal == NILEXPCAL)
    return;
  SETAS_ASSERTIONSIGNAL (vexpcal, pk_true);
}

/**-->true si ?VEXPCAL represente une assertion. */
extern boolean
ptexpcal_IsAssertion (ptexpcal vexpcal)
{
  if (vexpcal == NILEXPCAL)
    return pk_false;
  return IS_ASSERTIONSIGNAL (vexpcal);
}



/**-->true si ?VEXPCAL represente un signal declare en tant qu'entree*/
extern boolean
ptexpcal_IsInputFlow (ptexpcal vexpcal)
{
  if (ptexpcal_IsSignal (vexpcal))
    return (ptlinstid_isInput (ptexpcal_get_instance (vexpcal)));
  return pk_false;
}


/**-->true si ?VEXPCAL represente un signal declare en tant que sortie*/
extern boolean
ptexpcal_IsOutputFlow (ptexpcal vexpcal)
{
  if (ptexpcal_IsSignal (vexpcal))
    return (ptlinstid_isOutput (ptexpcal_get_instance (vexpcal)));
  return pk_false;
}


/**-->true si ?VEXPCAL represente un signal declare en tant que sortie*/
extern boolean
ptexpcal_IsLocalFlow (ptexpcal vexpcal)
{
  if (ptexpcal_IsSignal (vexpcal))
    return (ptlinstid_isLocal (ptexpcal_get_instance (vexpcal)));
  return pk_false;
}


/** @return the current type of the operand expression ?VEXPCAL */
extern TypeType
ptexpcal_get_type (ptexpcal vexpcal)
{
  if (vexpcal == NILEXPCAL) return TypeTypeGetUndefinedType ();
  return GET_ACTUALTYPE (vexpcal);
}

extern TypeType
ptexpcal_getUnamedType (ptexpcal vexpcal)
{
  TypeType t;
  if (vexpcal == NILEXPCAL) return TypeTypeGetUndefinedType ();
  t = GET_ACTUALTYPE (vexpcal);
  if (TypeTestNamedClass (t)) t = NamedTypeGetTerm (t);
  return t;
}

///
extern boolean
ptexpcal_get_aDim (ptexpcal vexpcal, int idim, ptexpcal * vidim)
{
  ptexpcal vdim;
  TypeType TT;

  //TT = ptexpcal_get_type(vexpcal);
  TT = ptexpcal_getUnamedType (vexpcal);
  if (TypeTestArrayClass (TT))
    {
      vdim = ArrayMatrixTypeGetDimLst (TT);
      while ((idim > 0) && (vdim != NILEXPCAL))
	{
	  (*vidim) = ptexpcal_getNaryArgs (&vdim);
	  idim--;
	  if (idim == 0)
	    return pk_true;
	}
    }
  return pk_false;
}

/**
  -->True si le signal ?VSIG a un nombre de dimension au moins egal a ?indexDim,
  dans ce cas !val designe l'expression (entier) de cette dimension.
*/
extern boolean
ptexpcal_get_ValueDim (ptexpcal exp, int indexDim, int *val)
{
  ptexpcal velem;

  if (!ptexpcal_get_aDim (exp, indexDim, &velem))
    return pk_false;
  *val = (int) ptexpcal_get_integerValue (velem);
  return pk_true;
}

/** -->true si ?PT represente un signal non defini */
extern boolean
ptexpcal_IsFreeSignal (ptexpcal anExpcal)
{
  return (valueHolder_isFree (ptexpcal_getValueHolder (anExpcal)));
}

/** ?anExpcal represente un signal non defini */
extern void
ptexpcal_belongs_freeSignal (ptexpcal anExpcal)
{
  if (IS_NILEXPCAL (anExpcal))
    return;
  CLEAR_VALUEHOLDER (anExpcal);
}

/** Le signal represente par ?VEXPCAL est marque `defini' */
// PLG CHANGE TO AVOID SIDE EFFECT
// used in ExpraccessSet_g.c ExprManager::belongsDefined
// used in TypeChecking_g.c ExprManager::TypeCheckingEquation: applied to an identifier
extern void
ptexpcal_belongsDefined (ptexpcal anExpcal)
{
  if (anExpcal == NILEXPCAL)
    {
      printf ("##INTERNAL ERROR## in ptexpcal_belongsDefined, anExpcal is nil\n");
      fflush (stdout);
      return;
    }
  (void) ptexpcal_getForce_valueHolder (anExpcal, defined_tag);
}

/** -->true si ?PT represente un signal defini mais non value */
extern boolean
ptexpcal_IsDefinedSignal (ptexpcal pt)
{
  return (valueHolder_isDefined (ptexpcal_getValueHolder (pt)));
}


/**-->true si ?PT est value ou borne */
extern boolean
ptexpcal_IsBounded (ptexpcal pt)
{
  return (valueHolder_isBounded (ptexpcal_getValueHolder (pt)));
}


/**
   @return true if anExpcal value is a representable integer interval
   if this is the case, the lower bound is assigned to bi and the upper bound to bs
*/
extern boolean
ptexpcal_get_integerBounds (ptexpcal anExpcal, compilerInteger * bi, compilerInteger * bs)
{
  if (IS_NILEXPCAL (anExpcal))
    {
      printf ("##INTERNAL ERROR## a ptexpcal is nil\n");
      fflush (stdout);
      return pk_false;
    }
  return valueHolder_get_integerBounds (ptexpcal_getValueHolder (anExpcal), bi, bs);
}

/**-->true si le signal ?PT est value */
extern boolean
ptexpcal_IsAssigned (ptexpcal pt)
{
  return valueHolder_isValued (ptexpcal_getValueHolder (pt));
}

/** @return true if vptValueHolder is valued and denotes a representable integer or a boolean
*/
extern boolean
ptexpcal_hasIntegerValue (ptexpcal pt)
{
  boolean isBig;		// PLG put isBig as a parameter */
  if (valueHolder_isIntegerValued (ptexpcal_getValueHolder (pt), &isBig))
    return !isBig;
  return pk_false;
}

/** -->valeur du signal value DT
   A utiliser apres un test (expcal_IsAssigned(pt)) 
   (sinon retourne 0) */
extern compilerInteger
ptexpcal_get_integerValue (ptexpcal pt)
{
  return (unsafeValueHolder_get_integerValue (ptexpcal_getValueHolder (pt)));
}

/** -->valeur du signal value DT
   A utiliser apres un test (expcal_IsAssigned(pt)) 
   (sinon retourne 0) */
extern boolean
ptexpcal_equalValues (ptexpcal expcal1, ptexpcal expcal2)
{
  return ptexpcal_get_integerValue (expcal1) == ptexpcal_get_integerValue (expcal2);
}

/** @return true when expcal value is equal to the value of ptexpcal_FALSE 
*/
extern boolean
ptexpcal_equalFalse (ptexpcal expcal)
{
  if (IS_NILEXPCAL (expcal))
    return pk_false;
  if (IS_NILEXPCAL (ptexpcal_FALSE))
    return pk_false;
  if (expcal == ptexpcal_FALSE)
    return pk_true;
  return valueHolder_equalFalse (GET_VALUEHOLDER (expcal));
}

/** @return true when expcal value is equal to the value of ptexpcal_TRUE
*/
extern boolean
ptexpcal_equalTrue (ptexpcal expcal)
{
  if (IS_NILEXPCAL (expcal))
    return pk_false;
  if (IS_NILEXPCAL (ptexpcal_TRUE))
    return pk_false;
  if (expcal == ptexpcal_TRUE)
    return pk_true;
  return valueHolder_equalTrue (GET_VALUEHOLDER (expcal));
}

/** assign the value holder that denotes valeurentiere to anExpcal.
    It does not change anything else.
 */
extern boolean
ptexpcal_set_integerValueHolder (ptexpcal anExpcal, compilerInteger valeurentiere)
{
  ptValueHolder val;

  if (anExpcal == NILEXPCAL)
    {
      printf ("##INTERNAL ERROR## in ptexpcal_set_integerValue, anExpcal is nil\n");
      fflush (stdout);
      return pk_false;
    }
  if (IS_EVENT (anExpcal))
    val = valueHolder_getTrue ();
  else if (IS_BOOLEAN (anExpcal))
    {
      if (valeurentiere == 1)
	val = valueHolder_getTrue ();
      else
	val = valueHolder_getFalse ();
      SET_VALUEHOLDER (anExpcal, val);
    }
  else
    val = valueHolder_provideIntegerRepresentation (valeurentiere);
  SET_VALUEHOLDER (anExpcal, val);
  return pk_true;
}

/** Affectation des valeurs d'un signal borne ?anExpcal :?BI borne inferieure, ?BS  borne superieure */
extern boolean
ptexpcal_set_integerBoundsHolder (ptexpcal anExpcal, compilerInteger bi, compilerInteger bs)
{
  if (anExpcal == NILEXPCAL)
    {
      printf ("##INTERNAL ERROR## in ptexpcal_set_integerBounds, vxpcal is nil\n");
      fflush (stdout);
      return pk_false;
    }
  if (bi == bs)
    return ptexpcal_set_integerValueHolder (anExpcal, bi);
  if (IS_BOOLEAN (anExpcal))
    return pk_false;
  SET_VALUEHOLDER (anExpcal, valueHolder_createIntegerBounds (bi, bs));
  return pk_true;
}

/** assign the value holder that denotes valeurentiere to anExpcal
 *  set opconst and type */
extern void
ptexpcal_set_computedIntegerValue (ptexpcal anExpcal, compilerInteger valeurentiere)
{
  if (ptexpcal_set_integerValueHolder (anExpcal, valeurentiere))
    {
      SET_ACTUALTYPE (anExpcal, get_IntegerType ());
    }
}

/** assign the value holder that denotes valeurentiere to anExpcal
 *  set opconst and type */
extern void
ptexpcal_set_integerValue (ptexpcal anExpcal, compilerInteger valeurentiere)
{
  if (ptexpcal_set_integerValueHolder (anExpcal, valeurentiere))
    {
      SET_ACTUALTYPE (anExpcal, get_IntegerType ());
      anExpcal->U.CONST.opconst = integer2repstr (valeurentiere);
    }
}

/** Affecte la valeur ?VALEURENTIERE a ?anExpcal */
extern void
ptexpcal_set_stringIntegerValue (ptexpcal anExpcal, ptrepstr integerStringValue)
{
  if (anExpcal != NILEXPCAL)
    {
      SET_VALUEHOLDER (anExpcal, valueHolder_createIntegerConstant (integerStringValue));
    }
}

/** Reinitialisation de l'information de valeur */
extern void
ptexpcal_reset_value (ptexpcal anExpcal)
{
  SET_VALUEHOLDER (anExpcal, NILVALUEHOLDER);
}

/** Fin acces valeur */


/**  -->dimension declaree du signal ?VSIG */
extern ptexpcal
ptexpcal_get_realDimension (ptexpcal vsig)
{
  TypeType TT;

  TT = ptexpcal_getUnamedType (vsig);
  if (!TypeTestArrayClass (TT))
    return NILEXPCAL;
  return ArrayMatrixTypeGetDimLst (TT);
}

/**
  -->Nombre de dimensions de l'expression ?EXP.
*/
extern int
ptexpcal_get_nbDim (ptexpcal exp)
{
  ptexpcal ldim;

  ldim = ptexpcal_get_realDimension (exp);
  ptexpcal_get_indir (ldim, &ldim);
  return ptexpcal_get_nbelem (ldim);
}

/** Associe a ?anExpcal la dimension representee par ?DIM */
extern void
ptexpcal_set_dimension (ptexpcal anExpcal, ptexpcal dim)
{
  /* ptct_set_dim(ptexpcal_get_type(anExpcal), dim); */
  /* SANS EFFET : doit etre fait au niveau du type de ?anExpcal */
}

/** -->representation de la dimension de ?PT */
extern ptexpcal
ptexpcal_get_dimension (ptexpcal pt)
{
  TypeType TT;
  TT = ptexpcal_getUnamedType (pt);
  if (!TypeTestArrayClass (TT))
    return NILEXPCAL;
  return ArrayMatrixTypeGetDimLst (TT);
}

///
extern boolean
ptexpcal_satisfies_predefinedType (ptexpcal vexp)
{
  return TypeTestPredef (ptexpcal_getUnamedType (vexp));
}

/** Assign type TT to vexpcal
    Assumes vexpcal not typed, not valued
 */
extern void
ptexpcal_set_type (ptexpcal vexpcal, TypeType TT)
{
  TypeType vexpcalType;
  if (vexpcal == NILEXPCAL)
    return;
  if (HAS_DEFINEDACTUALTYPE (vexpcal))
    {
      vexpcalType = GET_ACTUALTYPE (vexpcal);
      if (vexpcalType == TT)
	return;
      // if ( !TypeType_hasAnyType(vexpcalType) ) ptexpcal_set_type_warnings(vexpcal, vexpcalType, TT); 
    }
  SET_ACTUALTYPE (vexpcal, TT);
  if (TT == PureType)
    {
      SET_VALUEHOLDER (vexpcal, valueHolder_getTrue ());
    }
}

/** Pas d'effet dans le graphe SIGNAL (ajout Graph Sacres) */
extern void
ptexpcal_set_signal (ptexpcal vsig)
{
}

/** Pas d'effet dans le graphe SIGNAL (ajout Graph Sacres) */
extern boolean
ptexpcal_satisfies_known (ptexpcal vsig)
{
  return pk_true;
}

/** Fixe le numero de l'expression ?vexp a ?vnum  */
extern void
ptexpcal_copy_num (ptexpcal tgtExpcal, ptexpcal srcExpcal)
{
  tgtExpcal->setnum (ptexpcal_get_num (srcExpcal));
}


/** -->Numero du signal  */
extern int
ptexpcal_get_num (ptexpcal velem)
{
  return GETNUMEXP (velem);
}

/**-->operateur SIGNAL de l'expression ?PT*/
extern noperators
ptexpcal_get_operator (ptexpcal pt)
{
  return GET_OPERATOR (pt);
}

/**
  -->expression E decrite par ?VHORLEXPCAL = E when h 
  ?VHORLEXPCAL sinon
*/
extern ptexpcal
ptexpcal_get_expWhen (ptexpcal vhorlexpcal)
{
  cloper op;

  op = GET_OPERATORCLASS (vhorlexpcal);
  if (op == opereh)
    return vhorlexpcal->U.EXPR_WHEN.esync;
  // The operator can be an extraction and the class can be oper2tp (Dec 2008 accepted for SignalLustre)
  if ((GET_OPERATOR (vhorlexpcal) == opextr) && (op == oper2tp))
    return vhorlexpcal->U.PBINARY.op1tp;
  return vhorlexpcal;
}

///
extern void
ptexpcal_set_expWhen (ptexpcal vwhen, ptexpcal vexp)
{
  ptexpcal x;
  if (GET_OPERATORCLASS (vwhen) != opereh)
    return;
  x = vwhen->U.EXPR_WHEN.esync;
  if (x != vexp)
    {
      ptexpcal_remove_luse (x, vwhen);
      vwhen->U.EXPR_WHEN.esync = vexp;
      SET_TRAVELINDEXLIST (vwhen, GET_TRAVELINDEXLIST (vexp));
      ptexpcal_append_luse (vexp, vwhen);
    }
}

/** Affecte ?vinstr comme instruction de définition de ?vsig */
extern void
ptexpcal_set_node (ptexpcal vsig, ptINode vinstr)
{
  ptlinstid_set_leval (ptexpcal_get_instance (vsig), vinstr);
}

/**
   -->instruction definissant l'instance ?vsig
*/
extern ptINode
ptexpcal_get_node (ptexpcal vsig)
{
  ptexpcal_getSignalAssignedTo (vsig, &vsig);
  return ptlinstid_get_node (ptexpcal_get_instance (vsig));
}

/** -->Identificateur associe au signal ?PT */
extern ptident
ptexpcal_get_identifier (ptexpcal pt)
{
  return (ptlinstid_get_identvar (ptexpcal_get_instance (pt)));
}

/** Affectation de l'initalisation ?VINIT au signal ?SIG */
extern void
ptexpcal_set_initExp (ptexpcal sig, ptexpcal vinit)
{
  ptlinstid_set_vali (ptexpcal_get_instance (sig), vinit);
}

/** The signal sig is declared as a state variable */
extern void
ptexpcal_set_stateTaq (ptexpcal sig, boolean b)
{
  ptlinstid_set_stateTag (ptexpcal_get_instance (sig), b);
}

/** @return true if the signal ?sig is a state variable, false otherwise */
extern boolean
ptexpcal_get_stateTaq (ptexpcal sig)
{
  return ptlinstid_get_stateTag (ptexpcal_get_instance (sig));
}

/** The signal sig is declared as a shared signal (global signal) if b. */
extern void
ptexpcal_set_sharedTaq (ptexpcal sig, boolean b)
{
  linstid_set_sharedTag (ptexpcal_get_instance (sig), b);
}

/** @return true if the signal ?sig is a shared signal (global signal), false otherwise */
extern boolean
ptexpcal_get_sharedTaq (ptexpcal sig)
{
  return linstid_get_sharedTag (ptexpcal_get_instance (sig));
}

/**-->Remplace le ?POS ieme argument de ?anExpcal denotant une expression binaire par ?VAL */
extern void
ptexpcal_set_binaryArgstAt (ptexpcal anExpcal, int pos, ptexpcal val)
{
  ptexpcal x;
  if (pos == 1)
    {
      x = anExpcal->U.SBINARY.op1;
      if (x != val)
	{
	  ptexpcal_remove_luse (x, anExpcal);
	  anExpcal->U.SBINARY.op1 = val;
	  ptexpcal_append_luse (val, anExpcal);
	}
    }
  else
    {
      x = anExpcal->U.SBINARY.op2;
      if (x != val)
	{
	  ptexpcal_remove_luse (x, anExpcal);
	  anExpcal->U.SBINARY.op2 = val;
	  ptexpcal_append_luse (val, anExpcal);
	}
    }
}




/**-->argument de ?PT pour un unaire*/
extern ptexpcal
ptexpcal_get_unaryArg (ptexpcal pt)
{
  return (pt->U.SUNARY.op);
}


/** Modification par ?VAL de l'operande de ?anExpcal (unaire) */
extern void
ptexpcal_set_unaryArg (ptexpcal anExpcal, ptexpcal val)
{
  ptexpcal x;

  x = anExpcal->U.SUNARY.op;
  if (x != val)
    {
      ptexpcal_remove_luse (x, anExpcal);
      anExpcal->U.SUNARY.op = val;
      ptexpcal_append_luse (val, anExpcal);
    }
}



/**
  Pour un binaire : !REPONSE1 : premier argument de ?PT !REPONSE2 : deuxieme
  argument de ?PT
*/
extern void
ptexpcal_get_binaryArgs (ptexpcal pt, ptexpcal * reponse1, ptexpcal * reponse2)
{
  *reponse1 = pt->U.SBINARY.op1;
  *reponse2 = pt->U.SBINARY.op2;
}



/**-->argument de ?PT en position ?POS pour un binaire*/
extern ptexpcal
ptexpcal_get_binaryArgstAt (ptexpcal pt, int pos)
{
  if (pos == 1)
    return (pt->U.SBINARY.op1);
  return (pt->U.SBINARY.op2);
}




/**
  Pour un ternaire : !REPONSE1 : premier argument de ?PT !REPONSE2 :
  deuxieme argument de ?PT !REPONSE3 : troisieme argument de ?PT
*/
extern void
ptexpcal_get_ternaryArgs (ptexpcal pt, ptexpcal * reponse1, ptexpcal * reponse2, ptexpcal * reponse3)
{
  *reponse1 = pt->U.STERNARY.op31;
  *reponse2 = pt->U.STERNARY.op32;
  *reponse3 = pt->U.STERNARY.op33;
}



/**-->argument de ?PT en position ?POS pour un ternaire*/
extern ptexpcal
ptexpcal_get_ternaryArgstAt (ptexpcal pt, int pos)
{
  if (pos == 1)
    return pt->U.STERNARY.op31;
  if (pos == 2)
    return pt->U.STERNARY.op32;
  return pt->U.STERNARY.op33;
}

/**-->modification du ?POS ieme argument de ?anExpcal pour un ternaire*/
extern void
ptexpcal_set_ternaryArgstAt (ptexpcal anExpcal, int pos, ptexpcal val)
{
  ptexpcal x;

  if (pos == 1)
    {
      x = anExpcal->U.STERNARY.op31;
      if (x != val)
	{
	  ptexpcal_remove_luse (x, anExpcal);
	  anExpcal->U.STERNARY.op31 = val;
	  ptexpcal_append_luse (val, anExpcal);
	}
    }
  else if (pos == 2)
    {
      x = anExpcal->U.STERNARY.op32;
      if (x != val)
	{
	  ptexpcal_remove_luse (x, anExpcal);
	  anExpcal->U.STERNARY.op32 = val;
	  ptexpcal_append_luse (val, anExpcal);
	}
    }
  else
    {
      x = anExpcal->U.STERNARY.op33;
      if (x != val)
	{
	  ptexpcal_remove_luse (x, anExpcal);
	  anExpcal->U.STERNARY.op33 = val;
	  ptexpcal_append_luse (val, anExpcal);
	}
    }
}


/**
  Pour un binaire temporel : !REPONSE1 : premier argument de ?PT !REPONSE2 :
  deuxieme argument de ?PT
*/
extern void
ptexpcal_get_defaultArgs (ptexpcal pt, ptexpcal * reponse1, ptexpcal * reponse2)
{
  if (ptexpcal_satisfies_defaultExpression (pt))
    {
      *reponse1 = pt->U.PBINARY.op1tp;
      *reponse2 = pt->U.PBINARY.op2tp;
    }
  else
    {
      *reponse1 = NILEXPCAL;
      *reponse2 = NILEXPCAL;
    }
}



/**-->argument de ?PT en position ?POS
  pour un binaire temporel*/
extern ptexpcal
ptexpcal_get_defaultArgstAt (ptexpcal pt, int pos)
{
  if (pos == 1)
    return pt->U.PBINARY.op1tp;
  return pt->U.PBINARY.op2tp;
}

///
extern void
ptexpcal_set_defaultArgstAt (ptexpcal pt, int pos, ptexpcal exp)
{
  ptexpcal x;

  if (pos == 1)
    {
      x = pt->U.PBINARY.op1tp;
      if (x != exp)
	{
	  ptexpcal_remove_luse (x, pt);
	  pt->U.PBINARY.op1tp = exp;
	  ptexpcal_append_luse (exp, pt);
	}
    }
  else
    {
      x = pt->U.PBINARY.op2tp;
      if (x != exp)
	{
	  ptexpcal_remove_luse (x, pt);
	  pt->U.PBINARY.op2tp = exp;
	  ptexpcal_append_luse (exp, pt);
	}
    }
}


///
extern boolean
ptexpcal_get_Nary (ptexpcal * list, ptexpcal * arg)
{
  ptexpcal_get_indir (*list, list);
  if (*list == NILEXPCAL)
    return pk_false;
  (*arg) = ptexpcal_getNaryArgs (list);
  ptexpcal_get_indir (*arg, arg);
  return pk_true;
}

/**
  @return the contribution clock of an expression (defined for when/default/function call) (pt).
*/
extern ptinsth
ptexpcal_getContributionClock (ptexpcal pt)
{
  if (ptexpcal_satisfies_WhenCell (pt))
    return pt->U.EXPR_WHEN.hc;
  if (ptexpcal_satisfies_defaultExpression (pt))
    return pt->U.PBINARY.hc;
  if (GET_OPERATORCLASS (pt) == call)
    return pt->U.EFUNC.hc;
  return NILINSTH;
}



/**
  Fixe l'horloge de contribution de l'expression ?PT (denotant un When ou un default) a H
*/
extern void
ptexpcal_set_ContributionClock (ptexpcal pt, ptinsth h)
{
  if (ptexpcal_satisfies_WhenCell (pt))
    pt->U.EXPR_WHEN.hc = h;
  else if (ptexpcal_satisfies_defaultExpression (pt))
    pt->U.PBINARY.hc = h;
  else;				/* ?? */
}




/**  --attention au type de ?PT--
    -->1er element de la liste representee par ?PT 
    !PT : liste privee du premier element
*/
extern ptexpcal
ptexpcal_getNaryArgs (ptexpcal * pt)
{
  ptexpcal vexp;

  ptexpcal_get_indir (*pt, pt);
  vexp = (*pt)->U.SBINARY.op1;
  ptexpcal_get_indir (vexp, &vexp);
  *pt = (*pt)->U.SBINARY.op2;
  ptexpcal_get_indir (*pt, pt);
  return vexp;
}


/**->Nombre d'element d'un Expcal Liste */
extern int
ptexpcal_get_nbelem (ptexpcal lexpr)
{
  int n;

  if (lexpr == NILEXPCAL)
    return 0;
  ptexpcal_get_indir (lexpr, &lexpr);
  if (ptexpcal_get_operatorClass (lexpr) != oper2)
    return 0;
  n = 0;
  while (lexpr != NILEXPCAL)
    {
      lexpr = ptexpcal_get_binaryArgstAt (lexpr, 2);
      ptexpcal_get_indir (lexpr, &lexpr);
      n++;
    }
  return n;
}

/**
  -->Exception accociee au signal ?VSIG (tableau de processus)
*/
extern ptexpcal
ptexpcal_get_exception (ptexpcal vsig)
{
  return vsig->U.IDENT.opexception;
}




/**-->true si ?VEXPCAL represente un parametre
 */
extern boolean
ptexpcal_IsParameter (ptexpcal vexpcal)
{
  if (ptexpcal_get_operatorClass (vexpcal) != variabl) return pk_false;
  return (ptlinstid_is_parameter (ptexpcal_get_instance (vexpcal)));
}



/**-->true si ?VEXPCAL represente une constante declaree */
extern boolean
ptexpcal_IsConstant (ptexpcal vexpcal)
{
  if (ptexpcal_get_operatorClass (vexpcal) != variabl)
    return pk_false;
  return (ptlinstid_is_constant (ptexpcal_get_instance (vexpcal)));
}



/** -->true si l'expression ?VEXPCAL represente un signal donnant accès à un champ d'une structure/bundle. */
extern boolean
ptexpcal_IsFieldInstance (ptexpcal vexpcal)
{
  if (ptexpcal_get_operatorClass (vexpcal) != variabl)
    return pk_false;
  return (ptlinstid_is_SignalField (ptexpcal_get_instance (vexpcal)));
}


/**-->true si l'identificateur ?VEXPCAL represente une etiquette  */
extern boolean
ptexpcal_IsLabel (ptexpcal vexpcal)
{
  if (ptexpcal_get_operatorClass (vexpcal) != variabl)
    return pk_false;
  return (ptlinstid_is_label (ptexpcal_get_instance (vexpcal)));
}

/**-->true si l'expression ?VEXPCAL represente un indice de motif  */
extern boolean
ptexpcal_IsIndex (ptexpcal vexpcal)
{
  if (ptexpcal_get_operatorClass (vexpcal) != variabl)
    return pk_false;
  return (ptlinstid_is_index (ptexpcal_get_instance (vexpcal)));
}

/** -->Identificateur (arbre) associe au signal ?PT */
extern ptident
ptexpcal_get_treeIdentifier (ptexpcal pt)
{
  ptexpcal x;
  if (!ptexpcal_get_DeclaredSignal (pt, &x))
    return NILIDENT;
  return (ptlinstid_get_originalident (ptexpcal_get_instance (x)));
}

/** Sets the original name (out of instanciation) of pt to id */
extern void
ptexpcal_set_treeIdentifier (ptexpcal pt, ptident id)
{
  ptexpcal x;

  if (ptexpcal_get_DeclaredSignal (pt, &x))
    ptlinstid_set_originalident (ptexpcal_get_instance (x), id);
}

/**-->representation de la constante ?PT*/
extern ptrepstr
ptexpcal_get_constant (ptexpcal pt)
{
  return pt->U.CONST.opconst;
}

/**
  -->True si ?VEXP represente un signal de type event
*/
extern boolean
ptexpcal_IsEvent (ptexpcal vexp)
{
  TypeType TT;
  TT = ptexpcal_getUnamedType (vexp);
  if (TypeTestPredef (TT))
    return (TT == PureType);
  return pk_false;
}



/** -->True si ?VEXP represente un signal de type booleen */
extern boolean
ptexpcal_IsBoolean (ptexpcal vexp)
{
  TypeType TT;
  TT = ptexpcal_getUnamedType (vexp);
  return (boolean) TypeTestLogical (TT);
}

/** -->True si le type de ?VEXPCALCUL est libre */
extern boolean
ptexpcal_IsFreeType (ptexpcal vexp)
{
  TypeType t;
  t = ptexpcal_get_type (vexp);
  if (TypeTestPredef (t))
    return pk_false;
  return TypeTestAnyClass (t);
}


///
extern boolean
ptexpcal_hasIntegerType (ptexpcal vexp)
{
  TypeType TT;
  TT = ptexpcal_getUnamedType (vexp);
  if (TypeTestPredef (TT))
    return TypeTestIntNumber (TT);
  return pk_false;
}

/** -->true si le type declare de ?VEXP est event */
extern boolean
ptexpcal_satisfies_ofEventType (ptexpcal vexp)
{
  /*  return ((boolean) (ptexpcal_IsEvent(vexp) && (ptlinstid_get_t(ptexpcal_get_instance(vexp)) == opevent))); */
  return ((boolean) (ptexpcal_IsEvent (vexp)));
}

/* -----FIN REVOIR (types) */


/** @return the declared type of a signal(vexpcal).
 */
extern TypeType
ptexpcal_get_declared_type (ptexpcal vexpcal)
{
  if (ptexpcal_get_operatorClass (vexpcal) != variabl) return UndefinedType;
  return (ptlinstid_get_declared_type(ptexpcal_get_instance (vexpcal)));
}


/** (correspond a sa declaration) */
extern ntrees
ptexpcal_get_treeOfSignalDeclaration (ptexpcal vexpcal)
{
  ptexpcal x;
  if (!ptexpcal_get_DeclaredSignal (vexpcal, &x))
    return NILTREE;
  return (ptlinstid_get_decl (ptexpcal_get_instance (x)));
}

/** Affectation du noeud de l'arbre de syntaxe abstraite où le signal ?VEXPCAL est créé */
extern void
ptexpcal_set_treeOfSignalDeclaration (ptexpcal vexpcal, ntrees vdecl)
{
  if (vexpcal != NILEXPCAL)
    ptlinstid_set_decl (ptexpcal_get_instance (vexpcal), vdecl);
}


/** Affecte l'identificateur ?VNAME au signal reference par ?PT */
extern void
ptexpcal_set_identifier (ptexpcal pt, ptident vname)
{
  ptlinstid_set_identvar (ptexpcal_get_instance (pt), vname);
}


/** type (indice, param, signal,..) du signal ?VEXPCAL  */
extern typinst
ptexpcal_get_typeinst (ptexpcal vexpcal)
{
  return ptlinstid_get_tinst (ptexpcal_get_instance (vexpcal));
}

/** type (indice, param, signal,..) du signal ?VEXPCAL  */
extern boolean
ptexpcal_has_typeinst (ptexpcal vexpcal, typinst tyInst)
{
  return ptlinstid_has_tinst (ptexpcal_get_instance (vexpcal), tyInst);
}

/** L'identificateur ?VEXPCAL denote un parametre */
extern void
ptexpcal_belongs_parameter (ptexpcal vexpcal)
{
  ptlinstid_set_tinst (ptexpcal_get_instance (vexpcal), tparam);
}

/** L'identificateur ?VEXPCAL denote un signal  */
extern void
ptexpcal_belongs_signal (ptexpcal vexpcal)
{
  ptlinstid_set_tinst (ptexpcal_get_instance (vexpcal), tsignal);
}

/**  Associe un traitement d'exception ?VEXCEPTION au signal indice ?VSIG. */
extern void
ptexpcal_set_exception (ptexpcal vsig, ptexpcal vexception)
{
  vsig->U.IDENT.opexception = vexception;
}

/**
  @return the uniquely defined representation of Error Value ptexpcal
*/
extern ptexpcal
ptexpcal_getError ()
{
  if (ptexpcal_ERROR == NILEXPCAL)
    {
      ptexpcal_ERROR = ptexpcal_newExpcal (opcstent, constant, NILVALUEHOLDER);
      SETNEWNUMEXP (ptexpcal_ERROR);
      SET_ACTUALTYPE (ptexpcal_ERROR, get_ErrorType ());
      ptexpcal_ERROR->U.CONST.opconst = char2repstr ('z');
    }
  return ptexpcal_ERROR;
}

/** -->Expression denotant la valeur "vrai" */
extern ptexpcal
ptexpcal_getTrue (void)
{
  if (ptexpcal_TRUE == NILEXPCAL)
    {
      ptexpcal_TRUE = ptexpcal_newExpcal (optrue, constant, valueHolder_getTrue ());
      SETNEWNUMEXP (ptexpcal_TRUE);
      SET_ACTUALTYPE (ptexpcal_TRUE, get_PureType ());
      ptexpcal_TRUE->U.CONST.opconst = string2repstr ("true");
    }
  return ptexpcal_TRUE;
}

/** -->Expression denotant la valeur "faux" */
extern ptexpcal
ptexpcal_getFalse (void)
{
  if (ptexpcal_FALSE == NILEXPCAL)
    {
      ptexpcal_FALSE = ptexpcal_newExpcal (opfalse, constant, valueHolder_getFalse ());
      SETNEWNUMEXP (ptexpcal_FALSE);
      SET_ACTUALTYPE (ptexpcal_FALSE, get_BooleanType ());
      ptexpcal_FALSE->U.CONST.opconst = string2repstr ("false");
    }
  return ptexpcal_FALSE;
}

/** -->Expression denotant la valeur 0 */
extern ptexpcal
ptexpcal_get0 (void)
{
  if (ptexpcal_0 == NILEXPCAL)
    {
      ptexpcal_0 = ptexpcal_newExpcal (opcstent, constant, valueHolder_get0 ());
      SETNEWNUMEXP (ptexpcal_0);
    }
  SET_ACTUALTYPE (ptexpcal_0, get_IntegerType ());
  ptexpcal_0->U.CONST.opconst = string2repstr ("0");
  return ptexpcal_0;
}

/** -->Expression denotant la valeur 1 */
extern ptexpcal
ptexpcal_get1 (void)
{
  if (ptexpcal_1 == NILEXPCAL)
    {
      ptexpcal_1 = ptexpcal_newExpcal (opcstent, constant, valueHolder_get1 ());
      SETNEWNUMEXP (ptexpcal_1);
      SET_ACTUALTYPE (ptexpcal_1, get_IntegerType ());
      ptexpcal_1->U.CONST.opconst = string2repstr ("1");
    }
  return ptexpcal_1;
}

/**
  -->nouvelle instance d'expression constante d'operateur ?OPE (classe
  ?PCL) pour la constante representee par la chaine ?X
*/
extern ptexpcal
ptexpcal_RIEN_create_constant (noperators ope, ptrepstr x)
{
  ptexpcal vexp;
  vexp = ptexpcal_newExpcal (ope, constant, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.CONST.opconst = x;
  return vexp;
}

/*
  -->nouvelle representation d'expression, de type OPIDENT, et ayant
  l'identificateur ?VIDENT comme arbre associe
*/
/** --code mort */
extern ptexpcal
ptexpcal_RIEN_create_identifier (ptident vident)
{
  ptexpcal vexp;

  vexp = ptexpcal_newExpcal (opident, variabl, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  return vexp;
}

/**
  -->nouvelle instance d'expression  d'operateur ?OPE (classe ?PCL) et
  d'argument ?X1
*/
extern ptexpcal
ptexpcal_RIEN_create_unaryExp (noperators ope, cloper pcl, ptexpcal x1)
{
  ptexpcal vexp;

  vexp = ptexpcal_newExpcal (ope, pcl, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.SUNARY.op = x1;
  ptexpcal_append_luse (x1, vexp);
  return vexp;
}

/**
  -->nouvelle instance d'expression binaire  d'operateur ?OPE (classe ?PCL)
  et d'arguments ?X1 et ?X2
*/
extern ptexpcal
ptexpcal_RIEN_create_binaryExp (noperators ope, cloper pcl, ptexpcal x1, ptexpcal x2)
{
  ptexpcal vexp;

  vexp = ptexpcal_newExpcal (ope, pcl, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.SBINARY.op1 = x1;
  ptexpcal_append_luse (x1, vexp);
  vexp->U.SBINARY.op2 = x2;
  ptexpcal_append_luse (x2, vexp);
  return vexp;
}

/**
  -->Nouvelle expression sur signaux (?VEXPDEDEF when VH) 
  ?INSTENT : ensemble des signaux de ?VEXPDEDEF, 
  OPE : operateur d'extraction
*/
extern ptexpcal
ptexpcal_RIEN_create_whenExp (noperators ope, ptexpcal vexpdedef, ptinsth vh)
{
  ptexpcal vexp;

  vexp = ptexpcal_newExpcal (ope, opereh, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.EXPR_WHEN.hc = vh;
  vexp->U.EXPR_WHEN.esync = vexpdedef;
  ptexpcal_append_luse (vexpdedef, vexp);
  return vexp;
}

/**
  -->nouvelle instance d'expression binaire temporelle d'operateur ?OPE et
  d'arguments ?X1 et ?X2
*/
extern ptexpcal
ptexpcal_RIEN_create_defaultExp (noperators ope, ptexpcal x1, ptexpcal x2)
{
  ptexpcal vexp;

  vexp = ptexpcal_newExpcal (ope, oper2tp, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.PBINARY.op1tp = x1;
  ptexpcal_append_luse (x1, vexp);
  vexp->U.PBINARY.op2tp = x2;
  ptexpcal_append_luse (x2, vexp);
  vexp->U.PBINARY.hc = NILINSTH;
  return vexp;
}

/**
  -->nouvelle instance d'expression ternaire d'operateur ?OPE (classe ?PCL)
  et d'arguments ?X1, ?X2 et ?X3
*/
extern ptexpcal
ptexpcal_RIEN_create_ternaryExp (noperators ope, cloper pcl, ptexpcal x1, ptexpcal x2, ptexpcal x3)
{
  ptexpcal vexp;

  vexp = ptexpcal_newExpcal (ope, pcl, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.STERNARY.op31 = x1;
  ptexpcal_append_luse (x1, vexp);
  vexp->U.STERNARY.op32 = x2;
  ptexpcal_append_luse (x2, vexp);
  vexp->U.STERNARY.op33 = x3;
  ptexpcal_append_luse (x3, vexp);
  return vexp;
}

///
extern ptexpcal
ptexpcal_RIEN_create_type (TypeType t)
{
  ptexpcal vexp;
  vexp = ptexpcal_newExpcal (optype, opertyp, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.ETYPE.opvtype = t;
  return vexp;
}

///
extern ptexpcal
ptexpcal_RIEN_create_funcCall (ptident id, ptexpcal lparams, ptexpcal inputs)
{
  ptexpcal vexp;
  vexp = ptexpcal_newExpcal (opappe, call, NILVALUEHOLDER);
  SETNEWNUMEXP (vexp);
  vexp->U.EFUNC.idFunction = id;
  vexp->U.EFUNC.Inputs = inputs;
  ptexpcal_append_luse (inputs, vexp);
  vexp->U.EFUNC.Params = lparams;
  ptexpcal_append_luse (lparams, vexp);
  return vexp;
}

/**
  @return a new ptexpcal representing the constant val
*/
extern ptexpcal
ptexpcal_provideIntegerConstant (compilerInteger val)
{
  if (val == 0)
    return ptexpcal_get0 ();
  if (val == 1)
    return ptexpcal_get1 ();
  return ptexpcal_newIntegerConstant (val);
}

/**
  @return a new ptexpcal representing the constant val
*/
extern ptexpcal
ptexpcal_provideStringIntegerConstant (ptrepstr val)
{
  char *valString = repstr2string (val);
  if (strcmp (valString, "0") == 0)
    return ptexpcal_get0 ();
  if (strcmp (valString, "1") == 0)
    return ptexpcal_get1 ();
  return ptexpcal_newStringIntegerConstant (val);
}

///
extern boolean
ptexpcal_get_funcCall (ptexpcal e, ptident * id, ptexpcal * lp, ptexpcal * inputs)
{
  ptexpcal x;
  if (GET_OPERATORCLASS (e) != call)
    return pk_false;
  *id = ptexpcal_GetIdentFunctionCall (e);
  x = ptexpcal_GetInputsFunctionCall (e);
  ptexpcal_get_indir (x, inputs);
  x = ptexpcal_GetParametersFunctionCall (e);
  ptexpcal_get_indir (x, lp);
  return pk_true;
}

///
extern ptexpcal
ptexpcal_GetInputsFunctionCall (ptexpcal e)
{
  return e->U.EFUNC.Inputs;
}

///
extern ptident
ptexpcal_GetIdentFunctionCall (ptexpcal e)
{
  return e->U.EFUNC.idFunction;
}

///
extern ptexpcal
ptexpcal_GetParametersFunctionCall (ptexpcal e)
{
  return e->U.EFUNC.Params;
}

extern void
ptexpcal_updateFunctionCall (ptexpcal fc, ptexpcal oexp, ptexpcal nexp)
{
  if (fc->U.EFUNC.Inputs == oexp)
    fc->U.EFUNC.Inputs = nexp;
  else
    {
      if (fc->U.EFUNC.Params != oexp)
	return;
      fc->U.EFUNC.Params = nexp;
    }
  ptexpcal_remove_luse (oexp, fc);
  ptexpcal_append_luse (nexp, fc);
}

///
extern TypeType
ptexpcal_get_exprType (ptexpcal vexp)
{
  return vexp->U.ETYPE.opvtype;
}

///
extern void
ptexpcal_set_exprType (ptexpcal vexp, TypeType t)
{
  vexp->U.ETYPE.opvtype = t;
}


/** @return the list of the nodes connected to an expression (exp). */
extern ptlINode
ptexpcal_get_NodesFeedBy (ptexpcal vexp)
{
  if (vexp == NILEXPCAL)
    return NILLINODE;
  return GET_LNODEINPUT (vexp);
}


/** The list of the nodes connected to an expression (exp) is defined by l. */
extern void
ptexpcal_set_NodesFeedBy (ptexpcal vexp, ptlINode l)
{
  SET_LNODEINPUT (vexp, l);
}


/** -->Liste des utilisations de ?VEXP */
extern ptlexpcal
ptexpcal_get_luse (ptexpcal vexp)
{
  if (vexp == NILEXPCAL)
    return NILLEXPCAL;
  return GET_LEXPOUTPUT (vexp);
}


/** Suppression de ?VEXPTO dans la liste d'utilisation de ?VEXPFROM*/
extern void
ptexpcal_remove_luse (ptexpcal vexpfrom, ptexpcal vexpto)
{
  if (vexpfrom != NILEXPCAL)
    SET_LEXPOUTPUT (vexpfrom, ptlinstances_remove_exp (GET_LEXPOUTPUT (vexpfrom), vexpto));
}

/** Ajout de ?VEXPTO dans la liste d'utilisation de ?VEXPFROM */
extern void
ptexpcal_append_luse (ptexpcal vexpfrom, ptexpcal vexpto)
{
  if ((vexpfrom != NILEXPCAL) && (vexpto != NILEXPCAL))
    SET_LEXPOUTPUT (vexpfrom, ptlinstances_insert (GET_LEXPOUTPUT (vexpfrom), vexpto));
}






/**
  -->nouvelle instance d'expression de liste (binaire) d'operateur ?OPE
  (classe ?PCL); ?X1 le premier element et  ?X2 l'elements suivants
*/
extern ptexpcal
ptexpcal_RIEN_create_binaryListExp (noperators ope, cloper pcl, ptexpcal x1, ptexpcal x2)
{
  return ptexpcal_RIEN_create_binaryExp (ope, pcl, x1, x2);
}



/** Retourne la valeur "vrai" 
   si ?EXP1 et ?EXP2 representent la meme expression ou 
   si ?EXP1 et ?EXP2 ont meme valeur, lorsqu'elles sont evaluees et ?LOOKVALUE.
*/
extern boolean
ptexpcal_isEquivalent (ptexpcal exp1, ptexpcal exp2, boolean lookvalue)
{
  if (exp1 == exp2)
    return pk_true;
  if (exp1 == NILEXPCAL)
    return (boolean) (exp2 == NILEXPCAL);
  if (exp2 == NILEXPCAL)
    return pk_false;

  if (lookvalue && ptexpcal_IsAssigned (exp1) && ptexpcal_IsAssigned (exp2))
    return ptexpcal_equalValues (exp1, exp2);
  if (ptexpcal_get_operator (exp1) != ptexpcal_get_operator (exp2))
    return pk_false;

  switch ((int) (ptexpcal_get_operatorClass (exp1)))
    {
    case variabl:
      if (ptexpcal_get_signal (exp1) == ptexpcal_get_signal (exp2))
	return pk_true;
      if (ptexpcal_IsIndex (exp1) && ptexpcal_IsIndex (exp2))
	return (ptexpcal_get_num (exp1) == ptexpcal_get_num (exp2));	// clone
      return pk_false;
    case constant:
      return streq (ptexpcal_get_constant (exp1), ptexpcal_get_constant (exp2));

    case oper1:
    case oper1tp:
      return ptexpcal_isEquivalent (ptexpcal_get_unaryArg (exp1), ptexpcal_get_unaryArg (exp2), lookvalue);

    case oper2:
    case oper2tp:
      if (!ptexpcal_isEquivalent (ptexpcal_get_binaryArgstAt (exp1, 1), ptexpcal_get_binaryArgstAt (exp2, 1), lookvalue))
	return pk_false;
      return ptexpcal_isEquivalent (ptexpcal_get_binaryArgstAt (exp1, 2), ptexpcal_get_binaryArgstAt (exp2, 2), lookvalue);

    case oper3:
    case oper3tp:
      if (!ptexpcal_isEquivalent (ptexpcal_get_ternaryArgstAt (exp1, 1), ptexpcal_get_ternaryArgstAt (exp2, 1), lookvalue))
	return pk_false;
      if (!ptexpcal_isEquivalent (ptexpcal_get_ternaryArgstAt (exp1, 2), ptexpcal_get_ternaryArgstAt (exp2, 2), lookvalue))
	return pk_false;
      return ptexpcal_isEquivalent (ptexpcal_get_ternaryArgstAt (exp1, 3), ptexpcal_get_ternaryArgstAt (exp2, 3), lookvalue);

    case opereh:
      if (!ptexpcal_isEquivalent (ptexpcal_get_expWhen (exp1), ptexpcal_get_expWhen (exp2), lookvalue))
	return pk_false;
      return (ptexpcal_getContributionClock (exp1) == ptexpcal_getContributionClock (exp2));

    case opertyp:
      // return (ptexpcal_get_exprType(exp1) == ptexpcal_get_exprType(exp2));
      return pk_true;		// ????? revoir controle de type

    case call:
      if (!eq_idents (ptexpcal_GetIdentFunctionCall (exp1), ptexpcal_GetIdentFunctionCall (exp2)))
	return pk_false;
      return ptexpcal_isEquivalent (ptexpcal_GetInputsFunctionCall (exp1), ptexpcal_GetInputsFunctionCall (exp2), lookvalue);

    default:
      return pk_false;
    }
}


/**
  Creation d'un nouvel exemplaire d'identificateur designe par sa pile ?R,
  l'arbre syntaxique de sa declaration ?AR, son statut ?ti son type ?TT
  Il est affecte de l'horloge ?HORL et de l'instruction ?VINSTR
*/
extern ptexpcal
ptexpcal_newIdentifier (ptident r, ntrees ar, typinst ti, TypeType TT, ptINode vinstr, ptinsth horl)
{
  ptexpcal vexp;
  ptlinstid vinstid;

  vinstid = ptlinstid_new2 (r, ar, ti, vinstr, horl, TT);
  vexp = ptexpcal_newExpcal (opident, variabl, NILVALUEHOLDER);	/* PLG WARNING SIDE EFFECT */
  SETNEWNUMEXP (vexp);
  ptexpcal_set_type (vexp, TT);
  vexp->U.IDENT.opvaria = vinstid;
  vexp->U.IDENT.opexception = NILEXPCAL;
  ptexpcal_belongs_LocalFlow (vexp);	/* doit etre eventuellement modifie par l'appelant */
  ptlinstid_set_expcalass (vinstid, vexp);
  return vexp;
}

/**
  Creation d'un nouvel exemplaire d'identificateur copie de ?VEXPCAL.
  Il est affecte de l'horloge ?VH et de l'instruction ?VINSTR.
  L'expression sur signaux retourne est la representation de ce nouvel exemplaire.
*/
extern ptexpcal
atcpinstidexpcal (ptexpcal vexpcal, ptINode vinstr, ptinsth vh)
{
  ptexpcal vexpcalcul;
  ptlinstid vinst, vinstid;

  vinst = ptexpcal_get_instance (vexpcal);
  vinstid = ptlinstid_copy2 (vinst, vinstr, vh);
  vexpcalcul = ptexpcal_duplicate (ptlinstid_get_expcalass (vinst), vinstid, ptexpcal_get_type (vexpcal));
  SETNEWNUMEXP (vexpcalcul);
  ptexpcal_belongs_LocalFlow (vexpcalcul); /* Must be updated by the caller if necessary */
  return vexpcalcul;
}

/**
  Creation d'un nouvel exemplaire d'identificateur !VNEXP obtenu par la copie de l'identificateur represente par ?VEXP. 
  !VNEXP est affecte de l'horloge ?VH et de l'expression de definition ?VINSTR
  ---> Expression sur signaux representant !VNEXP.
*/
extern ptexpcal
ptexpcal_cloneIdent (ptexpcal vexpcal, ptINode vinstr)
{
  ptexpcal vexpcalcul;
  ptlinstid vinst, vinstid;

  vinst = ptexpcal_get_instance (vexpcal);
  vinstid = ptlinstid_copy (vinst, vinstr /*, vh */ );
  vexpcalcul = ptexpcal_duplicate (vexpcal, vinstid, ptexpcal_get_type (vexpcal));
  return vexpcalcul;
}


/** -->expression d'initialisation de l'instance ?VEXPCAL */
extern ptexpcal
ptexpcal_get_initExp (ptexpcal vexpcal)
{
  ptexpcal vinit;

  if (vexpcal == NILEXPCAL)
    return NILEXPCAL;
  if (ptexpcal_get_indir (ptlinstid_get_vali (ptexpcal_get_instance (vexpcal)), &vinit))
    ptexpcal_set_initExp (vexpcal, vinit);
  return vinit;
}

/** 
   -->true si l'expression ?VEXP denote un default 
*/
extern boolean
ptexpcal_satisfies_defaultExpression (ptexpcal vexp)
{
  return (ptexpcal_get_operator (vexp) == opmele);
}


/** -->true si l'expression ?VEXP denote un when/cell */
extern boolean
ptexpcal_satisfies_WhenCell (ptexpcal vexp)
{
  return (ptexpcal_get_operatorClass (vexp) == opereh);
}

///
extern ptlinstances
ptexpcal_getFieldsAndSignalsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_true, pk_true, pk_false, tsignal);
}

///
extern ptlinstances
ptexpcal_getSignalsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_true, pk_false, pk_false, tsignal);
}


///
extern ptlinstances
ptexpcal_getParametersReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_true, pk_false, pk_false, tparam);
}


///
extern ptlinstances
ptexpcal_getConstantsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_true, pk_false, pk_false, tconstant);
}

///
extern ptlinstances
ptexpcal_getIndexesReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_true, pk_false, pk_false, tindex);
}

///
extern ptlinstances
ptexpcal_getLabelsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_true, pk_false, pk_false, tlabel);
}

/// @return all the identifiers of the operand expression (vexp).
extern ptlinstances
ptexpcal_getIdsReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_false, pk_false, pk_false, tsignal);
}

///
extern ptlinstances
ptexpcal_getPossiblyIndexesReferencedIn (ptexpcal vexp)
{
  return ptexpcal_getIdReferencedIn (vexp, pk_false, pk_false, pk_true, tsignal);
}


/** It assigns the clock (vh) to the expression (e), assumed to be a signal or a synchronous call. */
extern void
ptexpcal_set_clock (ptexpcal e, ptinsth vh)
{
  if (GET_OPERATORCLASS (e) == variabl)
    ptlinstid_set_vhorl (ptexpcal_get_instance (e), vh);
  else
    e->U.EFUNC.hc = vh;		// call
}

/** pk_true if exp is a signal instance (!s=exp) or a field access with the !s instance assigned to */
extern boolean
ptexpcal_getSignalAssignedTo (ptexpcal exp, ptexpcal * s)
{
  ptexpcal e;

  ptexpcal_get_indir (exp, &exp);
  if (ptexpcal_get_operatorClass (exp) == variabl)
    {
      *s = exp;
      return pk_true;
    }
  if (ptexpcal_get_operator (exp) == opelem)
    return ptexpcal_getSignalAssignedTo (ptexpcal_get_binaryArgstAt (exp, 1), s);
  if (!ptexpcal_getAttribute (exp, FIELD_SIGNAL, (attributeValue *) & e))
    return pk_false;
  *s = e;
  return pk_true;
}


/*@return tke clock of the signal e 
  @param e  (in) assumed to be a signal identifier, an access to an element, 
       a function call or an access to a structure element (field).
*/
extern ptinsth
ptexpcal_get_clock (ptexpcal e)
{
  ptlinstid vinstid;
  ptinsth clockRes;

  if (e == NILEXPCAL)
    return NILINSTH;
  if (GET_OPERATORCLASS (e) == call)
    return e->U.EFUNC.hc;
  ptexpcal_getSignalAssignedTo (e, &e);
  if (e == NILEXPCAL)
    return NILINSTH;
  vinstid = ptexpcal_get_instance (e);
  clockRes = ptlinstid_get_vhorl (vinstid);
  return clockRes;
}



/** Reinitialisation de l'expression sur signaux ?exp */
extern void
ptexpcal_reset_luse (ptexpcal exp)
{
  switch ((int) (ptexpcal_get_operatorClass (exp)))
    {
    case oper1:
      ptexpcal_remove_luse (exp->U.SUNARY.op, exp);
      exp->U.SUNARY.op = NILEXPCAL;
      break;

    case oper2:
      ptexpcal_remove_luse (exp->U.SBINARY.op1, exp);
      exp->U.SBINARY.op1 = NILEXPCAL;
      ptexpcal_remove_luse (exp->U.SBINARY.op2, exp);
      exp->U.SBINARY.op2 = NILEXPCAL;
      break;
    case oper2tp:
      ptexpcal_remove_luse (exp->U.PBINARY.op1tp, exp);
      exp->U.PBINARY.op1tp = NILEXPCAL;
      ptexpcal_remove_luse (exp->U.PBINARY.op2tp, exp);
      exp->U.PBINARY.op2tp = NILEXPCAL;

      break;
    case oper3:
      ptexpcal_remove_luse (exp->U.STERNARY.op31, exp);
      exp->U.STERNARY.op31 = NILEXPCAL;
      ptexpcal_remove_luse (exp->U.STERNARY.op32, exp);
      exp->U.STERNARY.op32 = NILEXPCAL;
      ptexpcal_remove_luse (exp->U.STERNARY.op32, exp);
      exp->U.STERNARY.op33 = NILEXPCAL;
      break;


      break;
    case opereh:
      ptexpcal_remove_luse (exp->U.EXPR_WHEN.esync, exp);
      exp->U.EXPR_WHEN.esync = NILEXPCAL;
      break;

      // pas d'occurrences
    case variabl:
    case constant:
    case oper1tp:
    case oper3tp:
      break;
    default:
      //      pK_messenger->Caseerror(currentLine, currentFile);
      ;
    }
}



/** Lien expression/horloge 
   retourne la valeur "vrai" si l'horloge !vhsub = when ?vexp existe dans le systeme */
extern boolean
ptexpcal_get_extraction (ptexpcal vexp, ptinsth * vhsub)
{
  if (HAS_CLOCKSAMPLER (vexp))
    {
      *vhsub = GET_CLOCKSAMPLER (vexp);
      return pk_true;
    }
  return pk_false;
}

/** Signale que l'horloge ?vhsub = when ?vexp existe dans le systeme */
extern void
ptexpcal_set_extraction (ptexpcal vexp, ptinsth vhsub)
{
  SET_CLOCKSAMPLER (vexp, vhsub);
}

/**  Getting travel indexes assigned to an expression.

 */
extern ptlinstances
ptexpcal_get_travelIndexes (ptexpcal vexp)
{
  return GET_TRAVELINDEXLIST (vexp);
}

/** Setting travel indexes to an expression. */
extern void
ptexpcal_set_travelIndexes (ptexpcal vexp, ptlinstances lindexes)
{
  SET_TRAVELINDEXLIST (vexp, lindexes);
}

/** Installation d'une indirection (equivalence) de ?vfrom vers ?vers */
extern void
ptexpcal_set_indir (ptexpcal vfrom, ptexpcal vers)
{
  if (vfrom != vers)
    {
      SET_OPERATOR (vfrom, opindirection);
      vfrom->U.SUNARY.op = vers;
      GET_OPERATORCLASS (vfrom) = oper1;
    }
}


/** !nexp = ?vexp avec ses indirections, True si !nexp est different de vexp */
extern boolean
ptexpcal_get_indir (ptexpcal vexp, ptexpcal * nexp)
{
  *nexp = vexp;
  if (vexp == NILEXPCAL)
    return pk_false;
  while ((GET_OPERATOR (*nexp)) == opindirection)
    *nexp = (*nexp)->U.SUNARY.op;
  return (*nexp != vexp);
}


///
extern boolean
ptexpcal_satisfies_cell (ptexpcal exp)
{
  return (ptexpcal_get_operator (exp) == opmemo);
}

///
extern ptexpcal
ptexpcal_get_basicExpr (ptexpcal vexp)
{
  noperators op;
  boolean b;

  if (vexp == NILEXPCAL)
    return vexp;
  b = pk_true;
  while (b)
    {
      op = ptexpcal_get_operator (vexp);
      if (op == opmemo || op == opextr)
	vexp = ptexpcal_get_expWhen (vexp);
      else
	b = pk_false;
    }
  return vexp;
}



///
extern boolean
ptexpcal_isNil (ptexpcal vexp)
{
  return (vexp == NILEXPCAL);
}



/** -->l'attribut de numéro ?NUM affecte a l'expression ?VEXP */
extern boolean
ptexpcal_getAttribute (ptexpcal vexp, int num, attributeValue * valAttribute)
{
  return vexp->getAttribute (num, valAttribute);
}

/** Affecte l'attribut ?VATTR de numéro ?NUM à l'expression ?VEXP */
extern void
ptexpcal_addAttribute (ptexpcal vexp, int num, attributeValue vattr)
{
  vexp->addAttribute (num, vattr);
}

/** Supprime l'attribut de  numéro ?NUM s'il est asscocié à l'expression ?VEXP. */
extern void
ptexpcal_removeAttribute (ptexpcal vexp, int num)
{
  vexp->removeAttribute (num);
}


///
extern boolean
ptexpcal_satisfies_marked (ptexpcal vexp, int num)
{
  return vexp->satisfiesMarked (num);
}

///
extern void
ptexpcal_addMark (ptexpcal vexp, int num)
{
  vexp->addMark (num);
}

///
extern void
ptexpcal_removeMark (ptexpcal vexp, int num)
{
  vexp->removeMark (num);
}

/** PRAGMAS */
extern ptlpragmas
ptexpcal_get_pragmas (ptexpcal vexp)
{
  return vexp->getPragmas ();
}

///Adding
extern void
ptexpcal_addPragma (ptexpcal vexp, ptpragma aPragma)
{
  vexp->addPragma (aPragma);
}

///Removing.
extern boolean
ptexpcal_removePragma (ptexpcal vexp, ptrepstr name)
{
  return vexp->removePragma (name);
}


/** @return "true" si VEXP est une expression d'initialisation constante ;
   @param  (in) vexp expression sur signal;
   @param  (out) VEXP est alors cette expression de definition.
*/
extern boolean
ptexpcal_get_ConstantInitExpr (ptexpcal exp, ptexpcal * vexp)
{
  ptexpcal vexp1, vexp2;
  noperators op;

  if (exp == NILEXPCAL)
    {
      *vexp = exp;
      return pk_true;
    }
  switch ((int) (ptexpcal_get_operatorClass (exp)))
    {
    case variabl:
      *vexp = exp;
      return (ptexpcal_IsParameter (exp) || ptexpcal_IsConstant (exp));

    case constant:
      *vexp = exp;
      return pk_true;

    case oper2:
      op = ptexpcal_get_operator (exp);
      if (op == oplelevec)
	{
	  do
	    if (!ptexpcal_get_ConstantInitExpr (ptexpcal_getNaryArgs (&exp), vexp))
	      return pk_false;
	  while (exp != NILEXPCAL);
	  return pk_true;
	}
      ptexpcal_get_binaryArgs (exp, &vexp1, &vexp2);
      if (op == opensvec)
	return ptexpcal_get_ConstantInitExpr (vexp2, vexp);
      if (op == opelevec)
	return ptexpcal_get_ConstantInitExpr (vexp2, vexp);
      if (!ptexpcal_get_ConstantInitExpr (vexp1, vexp))
	return pk_false;
      if (!ptexpcal_get_ConstantInitExpr (vexp2, vexp))
	return pk_false;
      *vexp = exp;
      return pk_true;

    default:
      return pk_false;
    }
}


///
extern boolean
ptexpcal_isReferenceToStateVar (ptexpcal vexp)
{
  noperators op;
  op = ptexpcal_get_operator (vexp);
  if (op == opvvarin)
    return pk_true;
  if ((op == opcham) || (op == opelem))
    return ptexpcal_isReferenceToStateVar (ptexpcal_get_unaryArg (vexp));
  return pk_false;
}


///
extern ptlinstances
ptexpcal_getPredecessors (ptexpcal vexp)
{
  ptlinstances l1, l2;

  if (vexp == NILEXPCAL)
    return NILLINSTANCES;
  switch ((int) (ptexpcal_get_operatorClass (vexp)))
    {
    case variabl:
      if (ptexpcal_IsSignal (vexp))
	if (ptexpcal_get_stateTaq (vexp))
	  return NILLINSTANCES;
      return ptlinstances_push_front (NILLINSTANCES, vexp);

    case constant:
      return NILLINSTANCES;

    case oper1:
      if (ptexpcal_isReferenceToStateVar (vexp))
	return NILLINSTANCES;
      if (ptexpcal_get_operator (vexp) == opsdec)
	return ptlinstances_push_front (NILLINSTANCES, vexp);
      return ptexpcal_getPredecessors (ptexpcal_get_unaryArg (vexp));

    case oper1tp:
      return ptexpcal_getPredecessors (ptexpcal_get_unaryArg (vexp));

    case oper2:
      if (ptexpcal_isReferenceToStateVar (vexp))
	return NILLINSTANCES;

    case oper2tp:
      l1 = ptexpcal_getPredecessors (ptexpcal_get_binaryArgstAt (vexp, 1));
      l2 = ptexpcal_getPredecessors (ptexpcal_get_binaryArgstAt (vexp, 2));
      return ptlinstances_sortedfusion (l1, l2);

    case oper3:
    case oper3tp:
      l1 = ptexpcal_getPredecessors (ptexpcal_get_ternaryArgstAt (vexp, 1));
      l2 = ptexpcal_getPredecessors (ptexpcal_get_ternaryArgstAt (vexp, 2));
      l2 = ptlinstances_sortedfusion (l1, l2);
      l1 = ptexpcal_getPredecessors (ptexpcal_get_ternaryArgstAt (vexp, 3));
      return ptlinstances_sortedfusion (l1, l2);

    case opereh:
      return ptexpcal_getPredecessors (ptexpcal_get_expWhen (vexp));

    case opertyp:
      return NILLINSTANCES;

    case call:
      return ptexpcal_getPredecessors (ptexpcal_GetInputsFunctionCall (vexp));
    default:
      return NILLINSTANCES;
    }
}


/** Garbage collector */
extern void
ptexpcal_free (ptexpcal vexp)
{
  if (vexp == NILEXPCAL)
    return;
  // Excluding of the predefined ones
  if (vexp == ptexpcal_TRUE)
    return;
  if (vexp == ptexpcal_FALSE)
    return;
  if (vexp == ptexpcal_0)
    return;
  if (vexp == ptexpcal_1)
    return;
  if (vexp == ptexpcal_ERROR)
    return;

  // ptexpcal_remove_luse (vexp->U.SBINARY.op1, vexp);
  // ptexpcal_remove_luse (vexp->U.SBINARY.op2, vexp);
  ptexpcal_clear (vexp);
  SET_VALUEHOLDER (vexp, NILVALUEHOLDER);	/* Access to the value holder (value, bounds) */
  SET_OPERATORCLASS (vexp, oper2);
  if (GET_OPERATORCLASS (vexp) == variabl)
    ptlinstid_free (ptexpcal_get_instance (vexp));
  vexp->U.SBINARY.op2 = gcExp;
  vexp->clear ();		// attributes, pragmas ...
  gcExp = vexp;
}

/**
   It links an expression (vexp) with the tree, produced by the parser, it comes from (vtree)
*/
extern void
ptexpcal_set_backAnnotation (ptexpcal vexp, BackAnnotationType vtree)
{
  vexp->backAnnotation = vtree;
}

/**
   
*/
extern BackAnnotationType
ptexpcal_get_backAnnotation (ptexpcal vexp)
{
  return vexp->backAnnotation;
}


/**
  Initialisations generales relatives aux Expression sur signaux.
  ?VOP operateur "event" de SIGNAL.
  Initilisations en vue du garbage
*/
extern void
ptexpcal_init (ptexpcal * ptexpcalFALSE, ptexpcal * ptexpcalTRUE, ptexpcal * ptexpcal0, ptexpcal * ptexpcal1)
{
  if (hasBeenInitialized)
    return;
  hasBeenInitialized = pk_true;
  numcourexp = 0;
  gcExp = NILEXPCAL;
  UndefinedType = TypeTypeGetUndefinedType ();
  ptexpcal_ERROR = ptexpcal_getError ();
  /* Predefined constant value */
  valueHolder_init ();
  (*ptexpcalTRUE) = ptexpcal_getTrue ();
  (*ptexpcalFALSE) = ptexpcal_getFalse ();
  (*ptexpcal0) = ptexpcal_get0 ();
  (*ptexpcal1) = ptexpcal_get1 ();
}

///
extern int
getExprCounter ()
{
  return numcourexp;
}
