/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   Module that manages the representation of the expressions (expcal type). 
   The valuation of the expression is integrated in this module.
*/

/** Undefined predefined type */
TypeType UndefinedType;


struct expcal:public ObjectCommonInfos
{
  boolean isAssertionSignal;
  ptlINode sourceNodeList;
  ptlexpcal targetExpcalList;
  ptlinstances travelIndexList;
  ptinsth clockSampler;
  noperators this_operator;	///< operator of the expression
  TypeType actualType;		///< type of the expression.
  ptValueHolder this_valueHolder;	/* NIL or pointer to the actual value holder */
  BackAnnotationType backAnnotation;	///< back annotation to the tree. (abstract tree, call tree)

  cloper opera;			///< operator class (selector of the union)
  union
  {
    struct
    {				/* Identifier */
      ptlinstid opvaria;
      ptexpcal opexception;
    } IDENT;
    struct
    {				/* Constant expression */
      ptrepstr opconst;
    } CONST;
    struct
    {				/* unary expression */
      ptexpcal op;
    } SUNARY;
    struct
    {				/* Synchronous binary Expression */
      ptexpcal op1, op2;
    } SBINARY;
    struct
    {				/* Binary Expression op1tp default op2tp */
      ptexpcal op1tp, op2tp;
      ptinsth hc;		/* horloge du noeud : borne sup des horloges des operandes (op1tp, op2tp) */
    } PBINARY;
    struct
    {				/* Ternary expression */
      ptexpcal op31, op32, op33;
    } STERNARY;
    struct
    {				/* Binary expression ( esynC WHEN hc, esync cell hc) */
      ptexpcal esync;
      ptinsth hc;
    } EXPR_WHEN;
    struct
    {				/* TYPE (appel de noeud) */
      TypeType opvtype;
    } ETYPE;
    struct
    {				/* external function call/conversion */
      ptident idFunction;
      ptexpcal Params, Inputs;
      ptinsth hc;
    } EFUNC;
  } U;
};

#define SETAS_ASSERTIONSIGNAL(AVEXPCAL, ATYPE) ((AVEXPCAL->isAssertionSignal)=ATYPE)
#define IS_ASSERTIONSIGNAL(AVEXPCAL) (AVEXPCAL->isAssertionSignal)

#define SET_LNODEINPUT(AVEXPCAL, ALNODE) ((AVEXPCAL->sourceNodeList)=ALNODE)
#define GET_LNODEINPUT(AVEXPCAL) (AVEXPCAL->sourceNodeList)

#define SET_LEXPOUTPUT(AVEXPCAL, ALEXPCAL) ((AVEXPCAL->targetExpcalList)=ALEXPCAL)
#define GET_LEXPOUTPUT(AVEXPCAL) (AVEXPCAL->targetExpcalList)

#define SET_TRAVELINDEXLIST(AVEXPCAL, ANINDEXLIST) ((AVEXPCAL->travelIndexList)=ANINDEXLIST)
#define GET_TRAVELINDEXLIST(AVEXPCAL) (AVEXPCAL->travelIndexList)

#define SET_CLOCKSAMPLER(AVEXPCAL, ACLOCK) (((AVEXPCAL)->clockSampler)=(ACLOCK))
#define GET_CLOCKSAMPLER(AVEXPCAL) ((AVEXPCAL)->clockSampler)
#define HAS_CLOCKSAMPLER(AVEXPCAL) (GET_CLOCKSAMPLER(AVEXPCAL) != NILINSTH)

#define SET_OPERATOR(AVEXPCAL, ANOPERATOR) (((AVEXPCAL)->this_operator)=(ANOPERATOR))
#define GET_OPERATOR(AVEXPCAL) ((AVEXPCAL)->this_operator)

#define GET_ACTUALTYPE(AVEXPCAL) ((AVEXPCAL)->actualType)
#define SET_ACTUALTYPE(AVEXPCAL, ATYPE) GET_ACTUALTYPE(AVEXPCAL)=(ATYPE)
#define HAS_EQACTUALTYPE(AVEXPCAL, ATYPE) (GET_ACTUALTYPE(AVEXPCAL) == ATYPE )
#define HAS_DEFINEDACTUALTYPE(AVEXPCAL) ( !HAS_EQACTUALTYPE(AVEXPCAL, UndefinedType) )

#define GET_VALUEHOLDER(AVEXPCAL) ((AVEXPCAL)->this_valueHolder)
#define SET_VALUEHOLDER(AVEXPCAL, AVALUE) GET_VALUEHOLDER(AVEXPCAL)=(AVALUE)
#define CLEAR_VALUEHOLDER(AVEXPCAL) SET_VALUEHOLDER(AVEXPCAL, NILVALUEHOLDER)
#define HAS_VALUEHOLDER(AVEXPCAL) (!(IS_NILVALUEHOLDER(GET_VALUEHOLDER(AVEXPCAL))))

#define GET_OPERATORCLASS(AVEXPCAL) ((AVEXPCAL)->opera)
#define SET_OPERATORCLASS(AVEXPCAL, OPERA) GET_OPERATORCLASS(AVEXPCAL)=(OPERA)

// #define SET_LOWERBOUND(AVEXPCAL, AVALUE) (((AVEXPCAL)->lowerBound)=(AVALUE))
// #define GET_LOWERBOUND(AVEXPCAL) ((AVEXPCAL)->lowerBound)
// #define HAS_LOWERBOUND(AVEXPCAL) ( !(ISNIL_PTVALUEHOLDER(GET_LOWERBOUND(AVEXPCAL))))

// #define SET_UPPERBOUND(AVEXPCAL, AVALUE) (((AVEXPCAL)->upperBound)=(AVALUE))
// #define GET_UPPERBOUND(AVEXPCAL) ((AVEXPCAL)->upperBound)
// #define HAS_UPPERBOUND(AVEXPCAL) ( !(ISNIL_PTVALUEHOLDER(GET_UPPERBOUND(AVEXPCAL))))

#define SETNEWNUMEXP(VEXP) VEXP->setnum(getNewNumber())
#define GETNUMEXP(VEXP) VEXP->getnum()


#define IS_BOOLEAN(AVEXPCAL)  (ptexpcal_getUnamedType(AVEXPCAL) == BooleanType)
#define IS_EVENT(AVEXPCAL)  (ptexpcal_getUnamedType(AVEXPCAL) == PureType)

/**
  Predefined constants : error, pk_true, pk_false.
*/
static ptexpcal ptexpcal_ERROR = NILEXPCAL, ptexpcal_TRUE = NILEXPCAL, ptexpcal_FALSE = NILEXPCAL, ptexpcal_0 = NILEXPCAL, ptexpcal_1 = NILEXPCAL;

/** Counters of expressions.*/
static int numcourexp = 0;

/** Internal tag for initialisation */
static boolean hasBeenInitialized = pk_false;

/** Local garbage collector */
static ptexpcal gcExp = NILEXPCAL;


static int
getNewNumber ()
{
  numcourexp = numcourexp + 1;
  return numcourexp;
}

/**
   Reset an expression (vexp).
*/
static void
ptexpcal_clear (ptexpcal vexp)
{
  SETAS_ASSERTIONSIGNAL (vexp, pk_false);
  SET_LNODEINPUT (vexp, NILLINODE);
  SET_LEXPOUTPUT (vexp, NILLINSTANCES);
  SET_TRAVELINDEXLIST (vexp, NILLINSTANCES);
  SET_CLOCKSAMPLER (vexp, NILINSTH);
  SET_ACTUALTYPE (vexp, UndefinedType);	/* type of the expression. */
  vexp->backAnnotation = NILBACKANNOTATION;	/* back annotation to the tree. */
}

/**
   Constructor.
   @return a new expression with ope as operator, of class operatorClass and 
   an access to the value holder(itsValue).   
*/
static ptexpcal
ptexpcal_newExpcal (noperators ope, cloper operatorClass, ptValueHolder itsValue)
{
  ptexpcal vexp;

  if (gcExp == NILEXPCAL)
    {
      vexp = new expcal ();
      ptexpcal_clear (vexp);
    }
  else
    {
      vexp = gcExp;
      gcExp = gcExp->U.SBINARY.op2;
    }
  vexp->setwhoami (exprObject);
  SET_OPERATOR (vexp, ope);	/* operator of the expression */
  SET_VALUEHOLDER (vexp, itsValue);	/* Access to the value holder (value, bounds) */
  SET_OPERATORCLASS (vexp, operatorClass);	///< operator class (selector of the union)
  return vexp;
}

/**
  @return a new ptexpcal representing the constant val
*/
static ptexpcal
ptexpcal_newIntegerConstant (compilerInteger val)
{
  ptValueHolder vptValueHolder;
  ptexpcal anExpcal;
  vptValueHolder = valueHolder_createIntegerConstant (val);
  anExpcal = ptexpcal_newExpcal (opcstent, constant, vptValueHolder);
  SETNEWNUMEXP (anExpcal);
  SET_ACTUALTYPE (anExpcal, get_IntegerType ());
  anExpcal->U.CONST.opconst = integer2repstr (val);
  return anExpcal;
}

/**
  @return a new ptexpcal representing the constant val
*/
static ptexpcal
ptexpcal_newStringIntegerConstant (ptrepstr val)
{
  ptValueHolder vptValueHolder;
  ptexpcal anExpcal;
  vptValueHolder = valueHolder_createStringIntegerConstant (val);
  anExpcal = ptexpcal_newExpcal (opcstent, constant, vptValueHolder);
  SETNEWNUMEXP (anExpcal);
  SET_ACTUALTYPE (anExpcal, get_IntegerType ());
  anExpcal->U.CONST.opconst = val;
  return anExpcal;
}

/** 
   @return a copy of an identifier (identifier).
*/
static ptexpcal
ptexpcal_duplicate (ptexpcal srcExpcal, ptlinstid vinstid, TypeType TT)
{
  ptexpcal newExpcal;

  if (srcExpcal == NILEXPCAL)
    return NILEXPCAL;
  newExpcal = ptexpcal_newExpcal (GET_OPERATOR (srcExpcal), GET_OPERATORCLASS (srcExpcal), GET_VALUEHOLDER (srcExpcal));	/* PLG WARNING SIDE EFFECT */
  newExpcal->setnum (srcExpcal->getnum ());
  ptexpcal_set_type (newExpcal, TT);
  newExpcal->U.IDENT.opvaria = vinstid;
  newExpcal->U.IDENT.opexception = NILEXPCAL;
  ptlinstid_set_expcalass (vinstid, newExpcal);
  return newExpcal;
}

/** -->representation de l'identificateur ?anExpcal*/
static ptlinstid
ptexpcal_get_signal (ptexpcal anExpcal)
{
  ptlinstid vinstid;
  if (anExpcal == NILEXPCAL) return NILLINSTID;
  vinstid = anExpcal->U.IDENT.opvaria;
  return vinstid;
}

///
static ptlinstances
ptexpcal_getIdReferencedIn (ptexpcal anExpcal, boolean select, boolean bfield, boolean possiblyIndexSelection, typinst tyId)
{
  ptlinstances l1, l2;
  ptexpcal e;

  if (anExpcal == NILEXPCAL)
    return NILLINSTANCES;
  if (ptexpcal_getAttribute (anExpcal, OPTIMIZED_ATTRIBUTE, (attributeValue *) & e))
    return ptexpcal_getIdReferencedIn (e, select, bfield, possiblyIndexSelection, tyId);

  switch ((int) (GET_OPERATORCLASS (anExpcal)))
    {
    case variabl:
      if (possiblyIndexSelection)
	{
	  if (ptexpcal_satisfies_marked (anExpcal, POSSIBLY_INDEX)) 
	    return ptlinstances_push_front (NILLINSTANCES, anExpcal);
	  return NILLINSTANCES;
	}

      if (!select)
	return ptlinstances_push_front (NILLINSTANCES, anExpcal);
      if (ptexpcal_has_typeinst (anExpcal, tyId)) 
	return ptlinstances_push_front (NILLINSTANCES, anExpcal);
      return NILLINSTANCES;

    case constant:
      return NILLINSTANCES;

    case oper1:
      if ((tyId == tindex) && (ptexpcal_get_operator (anExpcal) == opsdec))
	if (ptexpcal_getAttribute (anExpcal, LinkToArrayIndexAttribute, (attributeValue *) & e))
	  return ptlinstances_push_front (NILLINSTANCES, e);
    case oper1tp:
      return ptexpcal_getIdReferencedIn (ptexpcal_get_unaryArg (anExpcal), select, bfield, possiblyIndexSelection, tyId);

    case oper2:
    case oper2tp:
      if (bfield && ptexpcal_getSignalAssignedTo (anExpcal, &e))
	return ptlinstances_push_front (NILLINSTANCES, e);
      if (possiblyIndexSelection && (ptexpcal_get_operator (anExpcal) == opelem))
	l1 = NILLINSTANCES;
      else
	l1 = ptexpcal_getIdReferencedIn (ptexpcal_get_binaryArgstAt (anExpcal, 1), select, bfield, possiblyIndexSelection, tyId);
      l2 = ptexpcal_getIdReferencedIn (ptexpcal_get_binaryArgstAt (anExpcal, 2), select, bfield, possiblyIndexSelection, tyId);
      return ptlinstances_sortedfusion (l1, l2);

    case oper3:
    case oper3tp:
      l1 = ptexpcal_getIdReferencedIn (ptexpcal_get_ternaryArgstAt (anExpcal, 1), select, bfield, possiblyIndexSelection, tyId);
      l2 = ptexpcal_getIdReferencedIn (ptexpcal_get_ternaryArgstAt (anExpcal, 2), select, bfield, possiblyIndexSelection, tyId);
      l2 = ptlinstances_sortedfusion (l1, l2);
      l1 = ptexpcal_getIdReferencedIn (ptexpcal_get_ternaryArgstAt (anExpcal, 3), select, bfield, possiblyIndexSelection, tyId);
      return ptlinstances_sortedfusion (l1, l2);

    case opereh:
      return ptexpcal_getIdReferencedIn (ptexpcal_get_expWhen (anExpcal), select, bfield, possiblyIndexSelection, tyId);

    case opertyp:
      return NILLINSTANCES;
    case call:
      return ptexpcal_getIdReferencedIn (ptexpcal_GetInputsFunctionCall (anExpcal), select, bfield, possiblyIndexSelection, tyId);
    default:
      // pK_messenger->Caseerror(currentLine, currentFile);
      return NILLINSTANCES;
    }
}




/** @return true if a declared signal is referenced in the input parameter exp, false otherwise.
   @param anExpcal  (out) the found signal or the original expression.
   @param exp  (in) an signal expression
*/
static boolean
ptexpcal_get_DeclaredSignal (ptexpcal exp, ptexpcal * anExpcal)
{
  noperators op;

  *anExpcal = exp;
  if (exp == NILEXPCAL)
    return pk_false;
  op = ptexpcal_get_operator (exp);
  if (op == opident)
    return pk_true;
  if (op == opvvarin)
    return ptexpcal_get_DeclaredSignal (ptexpcal_get_unaryArg (exp), anExpcal);
  if (op == opcham)
    return ptexpcal_get_DeclaredSignal (ptexpcal_get_binaryArgstAt (exp, 1), anExpcal);
  return pk_false;
}

/** @return NIL if anExpcal is NIL
    @return the value holder associated with anExpcal if anExpcal is not NIL
 */
static ptValueHolder
ptexpcal_getValueHolder (ptexpcal anExpcal)
{
  if (anExpcal == NILEXPCAL)
    return NILVALUEHOLDER;
  return GET_VALUEHOLDER (anExpcal);
}


// static boolean
// ptexpcal_hasLowerBoundHolder(ptexpcal anExpcal)
// {
//   return HAS_LOWERBOUND(anExpcal);
// }

// static boolean
// ptexpcal_hasUpperBoundHolder(ptexpcal anExpcal)
// {
//   return HAS_UPPERBOUND(anExpcal);
// }


///
// PLG CHECK TO AVOID SIDE EFFECT
static ptValueHolder
ptexpcal_getForce_valueHolder (ptexpcal anExpcal, definitionSTATUS itsStatus)
{
  ptValueHolder aValueHolder;

  if (HAS_VALUEHOLDER (anExpcal))
    {
      return GET_VALUEHOLDER (anExpcal);
    }
  aValueHolder = valueHolder_newValueHolder (itsStatus);
  SET_VALUEHOLDER (anExpcal, aValueHolder);
  return aValueHolder;
}

static void
ptexpcal_set_type_warnings (ptexpcal vexpcal, TypeType vexpcalType, TypeType TT)
{
  if (!ptexpcal_IsFreeSignal (vexpcal))
    {
      printf ("INTERNAL WARNING in ptexpcal_set_type: vexpcal is yet bounded or valued \n");
      fflush (stdout);
    }
  if (TypeTestPredef (TT))
    {
      if (TypeTestPredef (vexpcalType))
	printf
	  ("INTERNAL WARNING in ptexpcal_set_type: vexpcal type is changed from %s predefined type to predefined type %s\n",
	   TypeTypePredefTypeStringValue (vexpcalType), TypeTypePredefTypeStringValue (TT));
      else
	printf
	  ("INTERNAL WARNING in ptexpcal_set_type: vexpcal type is changed from %s category to predefined type %s\n",
	   TypeTypeStringCategory (vexpcalType), TypeTypePredefTypeStringValue (TT));
    }
  else
    {
      if (TypeTestPredef (vexpcalType))
	printf
	  ("INTERNAL WARNING in ptexpcal_set_type: vexpcal type is changed from %s predefined type to category %s\n",
	   TypeTypePredefTypeStringValue (vexpcalType), TypeTypeStringCategory (TT));
      else
	printf
	  ("INTERNAL WARNING in ptexpcal_set_type: vexpcal type is changed from %s category to category %s\n",
	   TypeTypeStringCategory (vexpcalType), TypeTypeStringCategory (TT));
    }
  fflush (stdout);
}
