/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


typedef noperators typeNode; ///< opnil, opdefsig, opret, opdefpart, opldefpart, opmoti, opnodecall, opappe, opbund, opsubgraph.

struct instr_common : public ObjectCommonInfos
{
  ptinsth horl;        ///< Clock of the node
  ptINode ptINodebc;   ///< Block info (array, sub-graph)
  ptexpcal label;      ///< label of the node.

  ptldependence lsucc; ///< dependences: node conditionned successors
  ptldependence lpred; ///< dependences: node conditionned predecessors 
  ptlinsth lsucchorl;  ///< dependences: clock successors
  ptlinsth lpredhorl;  ///< dependences: clock predecessors

  ptldependence iterationPrecedenceLst;  ///< iteration precedence induced by the [?] operator (array management).
  ptinterface interface;    ///< Interface (?,!, locals,...) of the node
  safetyPropertyType safetyInfo; ///< Safety property of the node.
  BackAnnotationType backAnnotation; ///< back annotation to the parser tree.
};

struct INode: public instr_common
{
  ptexpcal inNode, outNode; ///< left part and right part of the node

  typeNode t;   ///< Type of the node 
  union
  {
    // ; nothing for t=opdefsig and t=opdefpart

    ptlinstid instidref; ///< Nil Node: the associated signal (t = opnil)
    ptlINode ldefpart; ///< Set of Part definitions (t = opldefpart)
    ptlinstances lfields; ///< list of signals (bundle , struct, tuple) node (t = opbund)

    /// Array of process (t = opmoti)
    struct
    { 
      ptINode indexNode;  ///< node of the indice
      ptINode internalNodeArray; ///< internal node: each node of the array depends on it.
      ptgfd initArray; ///< initialization part graph of the array
      ptgfd bodyArray; ///< body part graph of the array
    } ARRAY;
    
    /// Procedure call ; i-o call (t = opappe)
    struct
    { 
      boolean isIntrinsic;      /// Tag at true for an intrinsic
      ptident name;                /// Name 
      ptINode ProduceByInode;      /// Specific to Inputs produced by a node call
      ptexpcal Params;             /// Ordered parameters
    } PCALL;

    /// NodeCall : reference to a model (t = opnodecall).
    struct
    { 
      ptexpcal Params;  ///< Ordered parameters
      ptDCG self_Graph; ///<	Reference to the model 
      ptlINode lInternalsReads; ///< representation of the outputs (as inputs)
    } NCALL;

    /// sub-graph (t = opsubgraph)
    struct
    {
      ptgfd vgfd; ///< set of the nodes of the sub-graph
      ptINode InternalNode; ///< fictitious node for "input" of the sub-graph: each node of the sub-graph depends on it.
      boolean lignee; ///< false by default, set to true for a cluster.
    } SUBGRAPH;

  } U;
};
