/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/** It defines a node (node) as an undefined node. */
    void
 NodeManager::becomes_undefined(ptINode node)
{
  removeInSetCase(node);
  ptINode_set_rightOperand(node, NILEXPCAL);
  ptINode_set_oper(node, opnil);
}

/** It defines a node (aNodeaff) as an equation node "vexpadef := vexpdedef ".
    \see{get_rightOperand, get_leftOperand} */
void NodeManager::becomes_assign(ptINode aNodeaff, ptexpcal vexpadef, ptexpcal vexpdedef)
{
  ptINode_becomes_assign(aNodeaff, vexpadef, vexpdedef);
}

/**  It defines a node (n) as a part definition one ( left ::= right )
    \see{get_rightOperand, get_leftOperand, is_partialDefinition}
 */
void NodeManager::becomes_partialDefinition(ptINode n, ptexpcal left, ptexpcal right)
{
  ptINode_becomes_partialDefinition(n, left, right);
}

/** It defines a node (aNode) as set of partial definitions : a signal (s) is defined
    by a set of nodes n1,...,nm elements of the list ln.
*/
void NodeManager::becomes_setOfPartialDefinitions(ptINode aNode, ptexpcal s, ptlINode ln)
{
  ptINode_becomes_setOfPartialDefinitions(aNode, s, ln);
}

/** It sets a clock (h) of a node (aNode) . */
void NodeManager::set_nodeClock(ptINode aNode, ptinsth h)
{
  ptINode_set_horl(aNode, h);
}

/** It defines the node (aNode) as a call to a intrinsic function. 
    Used to tag the node built on functions defined in the SIGNAL module SIGNAL_PREDEF.LIB of the exe.dir directory. 
*/
void NodeManager::set_intrinsicfunc(ptINode aNode)
{
  ptINode_set_intrinsicfunc(aNode);
}

/** It defines the name of an assumed sub-graph (aSubGraph) by id. */
void NodeManager::set_clusterName(ptINode aSubGraph, ptident id)
{
  addAttribute(aSubGraph, attrClusterName, (attributeValue) id);
}

/** A sub-graph node (aNode) is marked as a train cluster (all the nodes of the sub-graph depend on the same inputs.) */
void NodeManager::set_trainClusterTag(ptINode aNode)
{
  ptINode_set_trainClusterTag(aNode);
}

/**
   It defines the inputs/outputs of an assumed sub-graph (aNode) by  le /ls.
*/
void NodeManager::set_inOutGraphRef(ptINode aNode, ptexpcal le, ptexpcal ls)
{
  // can be replace by setInOut().
  ptINode_set_inOutGraphRef(aNode, le, ls);
}

/**
   It defines the clock successors of a node (aNode) by  vldephorl.
*/
void NodeManager::set_clocksSuccessors(ptINode aNode, ptlinsth vldephorl)
{
  ptINode_set_clocksSuccessors(aNode, vldephorl);
}

/**
   It defines the node successors of a node (aNode) by vldep
*/
void NodeManager::set_successors(ptINode aNode, ptldependence vldep)
{
  ptINode_set_successors(aNode, vldep);
}

/**
   It defines the node predecessors of a node (aNode) by vldep
*/
void NodeManager::set_predecessors(ptINode aNode, ptldependence vldep)
{
  ptINode_set_predecessors(aNode, vldep);
}

/**
   It defines the clock predecessors of a node (aNode) by vlpred.
*/
void NodeManager::set_clockPredecessors(ptINode aNode, ptlinsth vlpred)
{
  ptINode_set_clockPredecessors(aNode, vlpred);
}

/**
   It defines the iteration predecessors of a node (aNode) by vldep.
   (a node n that reference x[?] induces the iteration dependence node(x) -> n)
*/
void NodeManager::set_iterationPrecedence(ptINode aNode, ptldependence vldep)
{
  ptINode_set_iterationPrecedence(aNode, vldep);
}

/**
   It defines the left part (outputs) of a node (node) by e.
*/
void NodeManager::set_leftOperand(ptINode node, ptexpcal e)
{
  ptINode_set_leftOperand(node, e);
}

/**
   It defines the right part (inputs or definition) of a node (node) by e.
*/
void NodeManager::set_rightOperand(ptINode node, ptexpcal e)
{
  ptINode_set_rightOperand(node, e);
}

/**
   A node (aNode) is declared/belongs a "partial definition defaultvalue" node.
 */
void NodeManager::set_partialDefinitionDefaultValue(ptINode aNode)
{
  addMark(aNode, PART_DEFAULT_DEFINITION);
}

/**
   It defines the safety property of a node (n) by vproperty.
   A node is by default opsafe.
*/
void NodeManager::set_safetyProperty(ptINode n, safetyPropertyType vproperty)
{
  ptINode_set_safetyProperty(n, vproperty);
}

/**
   It defines the back annotation (abtract tree produced by the parser + call tree) of a node (n) by ba.
*/
void NodeManager::setBackAnnotation(ptINode n, BackAnnotationType ba)
{
  ptINode_setBackAnnotation(n, ba);
}

/**
   It adds a link from a node (aNode) to a composed node (cNode) to specify that 
   it is an element of this composed node (used for array, sub-graph management).
*/
void NodeManager::set_inComposedNode(ptINode aNode, ptINode cNode)
{
  ptINode_set_inComposedNode(aNode, cNode);
}

/** It replaces the current definition (right part) of a node (aNode) by  vdef.
    The references are updated.
 */
void NodeManager::set_def(ptINode aNode, ptexpcal vdef)
{
  ptINode_set_def(aNode, vdef);
}

/**
   Sub-graph node.
   It sets for a sub-graph (node) its projection informations:
    - vhemb: must be the branching clocks of the nodes/clocks of the sub-graph
    - projectedSystemClocks: the list of the clocks used to have an endochronous hierarchy for 
      the sub-graph (parametrization until vhemb, for each node/clock of the sub-graph).
*/
void NodeManager::set_subGraphProjectedSystemInfos(ptINode node, ptinsth vhemb, ptlinsth projectedSystemClocks)
{
  ptgfd vgfd;
  if (ptINode_satisfies_subGraph(node))
    {
      vgfd = ptINode_get_subGraph(node);
      ptgfd_set_projectionInfos(vgfd, vhemb, projectedSystemClocks);
    }
}

/**
   It associates a label (vlab) to a node (n)
 */
void NodeManager::set_label(ptINode n, ptexpcal vlab)
{
  ptINode_set_label(n, vlab);
}

/**
   It defines the utility clock of a signal (vsig) in a subgraph (sg) by hu.
*/
void NodeManager::setUtilityClockInSubGraph(ptINode sg, ptexpcal vsig, ptinsth hu)
{
  ptTablePointerPointer vtable;

  if (!getAttribute(sg, UTILITYCLOCKINSUBGRAPH, (attributeValue *) & vtable))
    {
      vtable = new TablePointerPointerType();
      (*vtable)[vsig] = hu;
      addAttribute(sg, UTILITYCLOCKINSUBGRAPH, (attributeValue) vtable);
    }
  else
    (*vtable)[vsig] = hu;
}

/**
   It defines the assignment clock of a node (n) by hu.
*/
void NodeManager::setAssignmentClock(ptINode n, ptinsth hu)
{
  addAttribute(n, ASSIGNMENTCLOCK, (attributeValue) hu);
}

/**
   It defines the utility clock of a node (n) by hu.
*/
void NodeManager::setUtilityClock(ptINode n, ptinsth hu)
{
  addAttribute(n, UTILITYCLOCK, (attributeValue) hu);
}

/**
   It assigns to a node (n) the clock (vh) at which the signal (produced by n)
   is not defined by  the (assumed) operand constant value (vExpr).
*/
void NodeManager::setDefaultConstantValueInfos(ptINode n, ptinsth vh, ptexpcal vExpr)
{
  addAttribute(n, ASSIGNMENTCLOCKPART, vh);
  addAttribute(n, DEFAULCONSTANTVALUE, vExpr);
}

/* void NodeManager::freeNode(ptINode aNode)
{
  ptINode_delete(aNode);
  } */
