/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** @return  "true" if a node (node) is marked by val , "false" otherwise.*/
    boolean NodeManager::isMarked(ptINode node, typeMark val)
{
  return ptINode_satisfies_marked(node, val);
}

/** It adds a mark (val) to a node (node).*/
void NodeManager::addMark(ptINode node, typeMark val)
{
  ptINode_addMark(node, val);
}

/** A node (node) becomes unmarked by val.*/
void NodeManager::removeMark(ptINode node, typeMark val)
{
  ptINode_removeMark(node, val);
}

/** @return  
    - "true" if the num-th attribute is assigned to a node (node), then its value is returned (valAttribute)
    - "false" otherwise, valAttribute is not modified.
 */
boolean NodeManager::getAttribute(ptINode node, typeMark num, attributeValue * valAttribute)
{
  return ptINode_getAttribute(node, num, valAttribute);
}

/**
   It assigns the num-th attribute, with vattr as value, to a node (node).
*/
void NodeManager::addAttribute(ptINode node, typeMark num, attributeValue vattr)
{
  ptINode_addAttribute(node, num, vattr);
}

/**
   It removes the num-th attribute from the list of attributes of a node (node).
*/
void NodeManager::removeAttribute(ptINode node, typeMark num)
{
  ptINode_removeAttribute(node, num);
}

/** @return  the list of pragmas assigned to a node.*/
ptlpragmas NodeManager::getPragmas(ptINode aNode)
{
  return ptINode_get_pragmas(aNode);
}

/** It assigns a list of pragmas to a node.*/
void NodeManager::setPragmas(ptINode aNode, ptlpragmas pragmas)
{
  ptINode_set_pragmas(aNode, pragmas);
}

/** It adds a pragma (apragma) to the list of pragmas assigned to a node (aNode).
    @return "true" if the pragma is added, "false" otherwise (the node is already associated with).
*/
boolean NodeManager::addPragma(ptINode aNode, ptpragma apragma)
{
  return ptINode_addPragma(aNode, apragma);
}

/**
   It removes the pragma called name from the list of pragmas assigned to a node (aNode).
   @return "true" if the node is annotated with this pragma, "false" otherwise.
*/
boolean NodeManager::removePragma(ptINode aNode, ptrepstr name)
{
  return ptINode_remove_pragma(aNode, name);
}

/** It adds a list of  pragmas (pragmas) to a node (aNode).*/
void NodeManager::addPragmas(ptINode aNode, ptlpragmas pragmas)
{
  ptINode_addPragmas(aNode, pragmas);
}

/** It adds a mark (val) to all the nodes of a list of nodes (list). */
void NodeManager::marklINodes(ptlINode list, int val)
{
  ptINode aNode;

  while (ptlINode_get_elem(&list, &aNode))
    {
      ptINode_addMark(aNode, val);
      if (ptINode_satisfies_writeNode(aNode))
	ptINode_addMark(EM->getNode(ptINode_get_io_arg(aNode)), val);
    }
}

/** It removes a mark (val) for all the nodes of a list of nodes (list). */
void NodeManager::resetMarklINodes(ptlINode list, int val)
{
  ptINode aNode;

  while (ptlINode_get_elem(&list, &aNode))
    {
      ptINode_removeMark(aNode, val);
      ptINode_removeAttribute(aNode, val);
    }
}

/** It marks all the nodes of the graph manager by a mark (val). */
void NodeManager::markAll(int val)
{
  marklINodes(lread, val);
  marklINodes(lwrite, val);
  marklINodes(ldelay, val);
  marklINodes(ldefsig, val);
  marklINodes(lArray, val);
  marklINodes(listrefextern, val);
  marklINodes(lNodeCall, val);
  marklINodes(lSubGraph, val);
  marklINodes(lBundStruct, val);
  marklINodes(ldefpart, val);
}

/** It removes a mark (val) from all the nodes of the graph manager. */
void NodeManager::resetMarkAll(int val)
{
  resetMarklINodes(lread, val);
  resetMarklINodes(lwrite, val);
  resetMarklINodes(ldelay, val);
  resetMarklINodes(ldefsig, val);
  resetMarklINodes(lArray, val);
  resetMarklINodes(listrefextern, val);
  resetMarklINodes(lNodeCall, val);
  resetMarklINodes(lSubGraph, val);
  resetMarklINodes(lBundStruct, val);
  resetMarklINodes(ldefpart, val);
}
