/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

///
extern void ptINode_update(ptINode vinstr)
{
  ptexpcal vexp;
  if (ptexpcal_get_indir(vinstr->inNode, &vexp))
    {
      vinstr->inNode = vexp;
      vinstr->interface = NILINTERFACE;	// it must be recomputed.
    }
  if (ptexpcal_get_indir(vinstr->outNode, &vexp))
    {
      vinstr->outNode = vexp;
      vinstr->interface = NILINTERFACE;	// it must be recomputed.
    }
}

///
extern ptINode ptINode_create()
{
  ptINode vinstr;

  vinstr = new INode();
  ptINode_init_gen(vinstr);
  return vinstr;
}

extern void ptINode_delete(ptINode aNode)
{
  delete aNode;
}

/** -->nouveau descripteur d'instruction de type ?VT */
extern ptINode ptINode_init(ptINode vinstr, ptinsth h)
{
  ptINode_init_gen(vinstr);
  vinstr->horl = h;
  vinstr->t = opnil;
  ptINode_set_expcalref(vinstr, NILEXPCAL);
  return vinstr;
}

/** Caracterisation du noeud ?VINSTR en tant que lecture ou ecriture */
extern void atmknodelececr(ptINode vinstr, ptident id)
{
  ptINode_set_oper(vinstr, opappe);
  vinstr->U.PCALL.name = id;
  vinstr->U.PCALL.isIntrinsic = pk_false;
  vinstr->U.PCALL.Params = NILEXPCAL;
  vinstr->inNode = NILEXPCAL;
  vinstr->outNode = NILEXPCAL;
}

/********************************************************
 ** LECTURES ********************************************/

/** Caracterisation du noeud ?VINSTR en tant que lecture */
extern void atmknodelec(ptINode vinstr, ptexpcal exp, ptident id)
{
  atmknodelececr(vinstr, id);
  vinstr->outNode = exp;
}

/********************************************************
 ** ECRITURES ********************************************/

/** caracterisation d'un noeud de lecture */
extern void atmknodeecr(ptINode vinstr, ptexpcal exp, ptident id)
{
  atmknodelececr(vinstr, id);
  vinstr->inNode = exp;
}

/** ?exp alimente le noeud ?n */
extern void ptINode_append_luse(ptINode n, ptexpcal exp)
{
  ptlINode l;
  ptINode x;

  if (exp == NILEXPCAL)
    return;
  l = ptexpcal_get_NodesFeedBy(exp);
  x = (ptINode) n;
  ptlINode_insert_elem(&l, x);
  ptexpcal_set_NodesFeedBy(exp, l);
}

/** ?exp n'alimente plus le noeud ?n */
extern void ptINode_remove_luse(ptINode n, ptexpcal exp)
{
  ptlINode l;

  if (exp == NILEXPCAL)
    return;
  l = ptexpcal_get_NodesFeedBy(exp);
  ptINode x = (ptINode) n;
  ptlINode_remove_elem(&l, x);
  ptexpcal_set_NodesFeedBy(exp, l);
}

/** -->type du noeud designe par ?VINSTR */
extern noperators ptINode_get_oper(ptINode vinstr)
{
  return (vinstr->t);
}

/** -->numero du noeud ?VINSTR */
extern int ptINode_get_num(ptINode vinstr)
{
  return vinstr->getnum();
}

/** -->Interface du noeud ?VINSTR */
extern ptinterface ptINode_get_interface(ptINode vinstr)
{
  return (vinstr->interface);
}

/**
   -->Horloge du noeud ?VINSTR.
*/
extern ptinsth ptINode_get_nodeClock(ptINode vinstr)
{
  return (vinstr->horl);
}

/** Affectation de l'interface ?VINTER au noeud ?VINSTR */
extern void ptINode_set_interface(ptINode vinstr, ptinterface vinter)
{
  vinstr->interface = vinter;
}

/** ?H devient l'horloge de l'instruction "X */
extern void ptINode_set_horl(ptINode x, ptinsth h)
{
  x->horl = h;
}

/** Affecte ?vt comme opérateur de l'instruction ?vinstr */
extern void ptINode_set_oper(ptINode vinstr, noperators vt)
{
  vinstr->t = vt;
}

/** Affecte ?n comme numéro de l'instruction ?vinstr */
extern void ptINode_set_num(ptINode vinstr, int n)
{
  vinstr->setnum(n);
}

///
extern boolean ptINode_is_undefined(ptINode vinstr)
{
  return (ptINode_get_oper(vinstr) == opnil);
}

///
extern boolean ptINode_satisfies_lfields(ptINode vinstr)
{
  return (ptINode_get_oper(vinstr) == opbund);
}

///
extern void ptINode_resetAsEquation(ptINode vinstr)
{
  noperators op1;

  op1 = vinstr->t;
  ptINode_set_def(vinstr, NILEXPCAL);
  if ((op1 != opdefsig) && (op1 != opdefpart))
    ptINode_set_oper(vinstr, opdefsig);
}

/** -->True si ?VINSTR designe un appel de noeud */
extern boolean ptINode_is_nodeCall(ptINode vinstr)
{
  return (vinstr->t == opnodecall);
}

/** -------------------------- Safety property ---------------------------------*/
extern void ptINode_set_safetyProperty(ptINode n, safetyPropertyType code)
{
  n->safetyInfo = code;
}

///
extern safetyPropertyType ptINode_get_safetyProperty(ptINode n)
{
  return n->safetyInfo;
}

/* ---------------------------------------------------------------------------- */
/* --------------------------- Dependances managing---------------------------- */
/* ---------------------------------------------------------------------------- */

/**-->Liste des noeuds horloges successeurs de ?VINSTR. */
extern ptlinsth ptINode_get_clockSuccessors(ptINode vinstr)
{
  return vinstr->lsucchorl;
}

/** Les signaux definis dans le noeuds ?VINSTR sont utilises dans les horloges de ?VLDEPHORL */
extern void ptINode_set_clocksSuccessors(ptINode vinstr, ptlinsth vldephorl)
{
  vinstr->lsucchorl = vldephorl;
}

/** -->Liste des noeuds (signaux) successeurs de ?VINSTR. */
extern ptldependence ptINode_get_successors(ptINode vinstr)
{
  return vinstr->lsucc;
}

/** Les signaux definis dans le noeuds ?VINSTR sont utilises dans les instructions ?VLSUCC. */
extern void ptINode_set_successors(ptINode vinstr, ptldependence vlsucc)
{
  vinstr->lsucc = vlsucc;
}

/** Retourne les predecesseurs du noeud ?vinstr */
extern ptldependence ptINode_get_predecessors(ptINode vinstr)
{
  return vinstr->lpred;
}

/** vlpred = predecesseurs de ?vinstr */
extern void ptINode_set_predecessors(ptINode vinstr, ptldependence vlpred)
{
  vinstr->lpred = vlpred;
}

/** Retourne les horloges predecesseurs du noeud ?vinstr */
extern ptlinsth ptINode_get_clockPredecessors(ptINode vinstr)
{
  return vinstr->lpredhorl;
}

/** vlpred = predecesseurs de ?vinstr */
extern void ptINode_set_clockPredecessors(ptINode vinstr, ptlinsth vlpred)
{
  vinstr->lpredhorl = vlpred;
}

/** -->True si ?VINSTR designe une expression de retard */
extern boolean ptINode_is_delay(ptINode vinstr)
{
  return (vinstr->t == opret);
}

/********************************************************
 ** LECTURES / ECRITURES *********************************/

/**
  !OP1, !OP2 : expressions de partie gauche et de partie droite de
  l'instruction d'affectation ?VINSTR
*/
extern void ptINode_get_equation_arg(ptINode vinstr, ptexpcal * op1, ptexpcal * op2)
{
  ptexpcal vexp;

  *op1 = ptINode_get_leftOperand(vinstr);
  if (!ptINode_get_rightOperand(vinstr, &vexp))
    *op2 = NILEXPCAL;
  else
    {
      if (ptexpcal_satisfies_cell(vexp))
	vexp = ptexpcal_get_expWhen(vexp);
      *op2 = ptexpcal_get_expWhen(vexp);
    }
}

/** Rend pour l'instruction de decalage ?VINSTR : !RESRET := !SIGRET $ !VRET window !VFEN init vinit */
extern void ptINode_get_delay_arg(ptINode vinstr, ptexpcal * resret, ptexpcal * sigret, ptexpcal * vret, ptexpcal * vfen, ptexpcal * vinit)
{
  ptexpcal vexp;

  *resret = ptINode_get_leftOperand(vinstr);
  vexp = ptINode_get_exprret(vinstr);
  *vret = ptexpcal_get_ternaryArgstAt(vexp, 2);
  *vfen = ptexpcal_get_ternaryArgstAt(vexp, 3);

  vexp = ptexpcal_get_ternaryArgstAt(vexp, 1);
  *sigret = ptexpcal_get_binaryArgstAt(vexp, 1);
  *vinit = ptexpcal_get_binaryArgstAt(vexp, 2);
}

///
extern void ptINode_set_delayExpression(ptINode vinstr, ptexpcal ve)
{
  ptINode_remove_luse(vinstr, vinstr->inNode);
  vinstr->inNode = ve;
  ptINode_append_luse(vinstr, ve);
}

/** Caracterisation du noeud ?VINSTR en tant que retard */
extern void ptINode_becomes_delay(ptINode vinstr, ptexpcal resret, ptexpcal vexpret)
{
  ptINode_set_oper(vinstr, opret);
  vinstr->outNode = resret;
  vinstr->inNode = vexpret;
  ptINode_append_luse(vinstr, vexpret);
  ptINode_set_safetyProperty(vinstr, opunsafe);
}

/** It characterizes a node (n) as the definition of a structured signal (structure/bundle) composed of a list of fields (lf) */
extern void ptINode_becomes_lfields(ptINode n, ptlinstances lf)
{
  ptINode_set_oper(n, opbund);
  n->U.lfields = lf;
}

/** For an assumed node defining a structured signal (structure/bundle), it returns the list of the fields. */
extern ptlinstances ptINode_get_lfields(ptINode n)
{
  return n->U.lfields;
}

/** @return the read or write signal if the node(vinstr) is an i/o node, NILEXPCAL otherwise */
extern ptexpcal ptINode_get_io_arg(ptINode vinstr)
{
  ptexpcal l;

  if (ptINode_satisfies_readNode(vinstr))
    l = vinstr->outNode;
  else if (ptINode_satisfies_writeNode(vinstr))
    l = vinstr->inNode;
  else
    return NILEXPCAL;
  return ptexpcal_getNaryArgs(&l);
}

/** @return "true" if a node (node) is an i/o node, "false" otherwise.*/
extern boolean ptINode_satisfies_ioNode(ptINode node)
{
  if (ptINode_satisfies_readNode(node))
    return pk_true;
  if (ptINode_satisfies_writeNode(node))
    return pk_true;
  return pk_false;
}

/** It characterizes a node (n) as the definition of an equation (:= , ::=) but not a delay.*/
extern void ptINode_becomes_assign(ptINode vinstr, ptexpcal vexpadef, ptexpcal vexpdedef)
{
  ptINode_set_oper(vinstr, opdefsig);
  vinstr->outNode = vexpadef;
  vinstr->inNode = vexpdedef;
  ptINode_append_luse(vinstr, vexpdedef);
}

/** @return the delay expression (E $ N window M) of a delay node (vinstr)  */
extern ptexpcal ptINode_get_exprret(ptINode vinstr)
{
  return (vinstr->inNode);
}

/** @return the initial value of a delay node(visntr). */
extern ptexpcal ptINode_get_initDelay(ptINode vinstr)
{
  ptexpcal vexp;

  vexp = ptINode_get_exprret(vinstr);
  vexp = ptexpcal_get_binaryArgstAt(ptexpcal_get_ternaryArgstAt(vexp, 1), 2);
  if (vexp != NILEXPCAL)
    return vexp;
  vexp = ptINode_get_leftOperand(vinstr);
  return ptexpcal_get_initExp(vexp);
}

/** L'expression ?VHORLEXPCAL est le second operande de l'expression de definition de signal denote par ?VINSTR */
extern void ptINode_set_def(ptINode vinstr, ptexpcal vhorlexpcal)
{
  ptINode_remove_luse(vinstr, vinstr->inNode);
  if (vhorlexpcal != NILEXPCAL)
    ptexpcal_belongs_freeSignal(vinstr->outNode);
  vinstr->inNode = vhorlexpcal;
  if (vhorlexpcal != NILEXPCAL)
    ptINode_append_luse(vinstr, vhorlexpcal);
}

/** 
    Recuperation des infos d'un noeud 'appel' (VINSTR)
    !NOM : identificateur de la procedure 
    !OrderedInputs : liste des signaux d'entrees 
    !OrderedOutputs : liste des signaux de sorties
*/
extern void ptINode_get_procCall_arg(ptINode vinstr, ptident * nom, ptexpcal * param, ptexpcal * OrderedInputs, ptexpcal * OrderedOutputs)
{
  (*nom) = vinstr->U.PCALL.name;
  ptINode_update(vinstr);
  (*OrderedInputs) = vinstr->inNode;
  (*OrderedOutputs) = vinstr->outNode;
  (*param) = vinstr->U.PCALL.Params;
}

/**
   -->True si ?VINSTR designe une expression de definition de signal (pas un retard) 
*/
extern boolean ptINode_satisfies_equation(ptINode vinstr)
{
  return (vinstr->t == opdefsig);
}

/**
  -->Signal defini par le noeud ?VINSTR lorsque ce dernier designe une
  expression de definition (OPDEFSIG) ou un retard.
*/
extern ptexpcal ptINode_get_leftOperand(ptINode vinstr)
{
  ptINode_update(vinstr);
  return vinstr->outNode;
}

///
extern void ptINode_set_leftOperand(ptINode vinstr, ptexpcal exp)
{
  vinstr->outNode = exp;
}

///
extern void ptINode_set_rightOperand(ptINode vinstr, ptexpcal exp)
{
  vinstr->inNode = exp;
}

/** -->True si l'expression !vexp du  noeud ?VINSTR  X:= !vexp est definie */
extern boolean ptINode_get_rightOperand(ptINode vinstr, ptexpcal * vexp)
{
  *vexp = vinstr->inNode;
  return (*vexp != NILEXPCAL);
}

/** Caracterisation du noeud ?VINSTR en tant qu'appel externe */
extern void ptINode_becomes_extern(ptINode vinstr, ptident n, ptexpcal lp, ptexpcal vexplentren, ptexpcal vexplsortren)
{
  ptINode_set_oper(vinstr, opappe);
  vinstr->U.PCALL.name = n;
  vinstr->U.PCALL.isIntrinsic = pk_false;
  vinstr->inNode = vexplentren;
  vinstr->outNode = vexplsortren;
  vinstr->U.PCALL.Params = lp;
  ptINode_append_luse(vinstr, vexplentren);
}

///
extern void ptINode_set_inputExpressions(ptINode node, ptexpcal inputs)
{
  ptINode_remove_luse(node, node->inNode);
  node->inNode = inputs;
  ptINode_append_luse(node, inputs);
}

/** -->True si ?VINSTR designe une lecture */
extern boolean ptINode_satisfies_readNode(ptINode vinstr)
{
  if (vinstr->t != opappe)
    return pk_false;
  return (vinstr->U.PCALL.name == getReadId());
}

/** -->True si ?VINSTR designe une écriture */
extern boolean ptINode_satisfies_writeNode(ptINode vinstr)
{
  if (vinstr->t != opappe)
    return pk_false;
  return (vinstr->U.PCALL.name == getWriteId());
}

/** True si ?vinstr est un appel, dans ce cas !id denote le nom du noeud appele (procedure) */
extern boolean ptINode_get_nameCall(ptINode vinstr, ptident * id)
{
  if (vinstr->t != opappe)
    return pk_false;
  *id = vinstr->U.PCALL.name;
  return pk_true;
}

/** Set the name id for a call. */
extern void ptINode_set_NameCall(ptINode vinstr, ptident id)
{
  if (vinstr->t == opappe)
    vinstr->U.PCALL.name = id;
}

/** -->True si le noeud ?VINSTR denote une fonction intrinseque */
extern boolean ptINode_satisfies_intrinsicfunc(ptINode vinstr)
{
  if (vinstr->t == opappe)
    return (vinstr->U.PCALL.isIntrinsic);
  return pk_false;
}

/** Caracterisation du noeud ?VINSTR en tant que reference a une fonction intrinseque */
extern void ptINode_set_intrinsicfunc(ptINode vinstr)
{
  vinstr->U.PCALL.isIntrinsic = pk_true;
}

/** Creation du noeud ?vinstr defini en tant qu'appel du modele represente par son graphe ?vgraph et les valeurs effectives 
   (parametres, entrees, sorties) */
extern void ptINode_becomes_nodecall(ptINode vinstr, ptDCG vgraph, ptexpcal nodeCallParam, ptexpcal nodeCallInputs, ptexpcal nodeCallOutputs)
{
  ptINode_set_oper(vinstr, opnodecall);
  vinstr->U.NCALL.self_Graph = vgraph;
  vinstr->U.NCALL.Params = nodeCallParam;
  vinstr->inNode = nodeCallInputs;
  vinstr->outNode = nodeCallOutputs;
  vinstr->U.NCALL.lInternalsReads = NILLINODE;
  ptINode_append_luse(vinstr, nodeCallInputs);
}

///
extern void ptINode_set_nodecallProducedNodes(ptINode vinstr, ptlINode ln)
{
  vinstr->U.NCALL.lInternalsReads = ln;
}

///
extern ptlINode ptINode_get_nodecallProducedNodes(ptINode vinstr)
{
  return vinstr->U.NCALL.lInternalsReads;
}

///
extern void ptINode_update_NodeCall(ptINode vinstr, ptexpcal e1, ptexpcal e2)
{
  if (vinstr->U.NCALL.Params == e1)
    vinstr->U.NCALL.Params = e2;
  if (vinstr->inNode == e1)
    vinstr->inNode = e2;
  if (vinstr->outNode == e1)
    vinstr->outNode = e2;
}

/** Retourne les entrees/sorties (!nodeCallInputs/!nodeCallOutputs) d'un appel de noeud ?vinstr.*/
extern void ptINode_get_args_nodecall(ptINode vinstr, ptDCG * vg, ptexpcal * nodeCallParams, ptexpcal * nodeCallInputs, ptexpcal * nodeCallOutputs)
{
  ptexpcal vexp;

  (*vg) = vinstr->U.NCALL.self_Graph;
  if (ptexpcal_get_indir(vinstr->U.NCALL.Params, &vexp))
    vinstr->U.NCALL.Params = vexp;
  ptINode_update(vinstr);
  (*nodeCallParams) = vinstr->U.NCALL.Params;
  (*nodeCallInputs) = vinstr->inNode;
  (*nodeCallOutputs) = vinstr->outNode;
}

/**
 * Gestion des Sous-hierarchies 
 *	-->Sous-graphe défini par le neoud ?VINSTR
 */
extern ptDCG ptINode_get_nodeCall_graph(ptINode vinstr)
{
  return vinstr->U.NCALL.self_Graph;
}

/**
  --gestion des signaux indices--
  Specifie que l'instruction (bidon OPNIL) ?VINSTR se rapporte a l'instance ?VINSTREF.
*/
extern void ptINode_set_expcalref(ptINode vinstr, ptexpcal vinstref)
{
  ptlinstid vinst;
  /*   --Ref INSTID -- */
  if (ptINode_get_oper(vinstr) != opnil)
    return;
  if (vinstref != NILEXPCAL)
    vinst = ptexpcal_get_instance(vinstref);
  else
    vinst = NILLINSTID;
  vinstr->U.instidref = vinst;
}

/** -->True si le noeud ?vinstr est une entree produite par un NodeCall !nodecall */
extern boolean ptINode_is_outputNodeCall(ptINode vinstr, ptINode * nodecall)
{
  if (!ptINode_satisfies_readNode(vinstr))
    return pk_false;
  (*nodecall) = vinstr->U.PCALL.ProduceByInode;
  return ((*nodecall) != NILINODE);
}

/** Caracterisation du noeud ?VINSTR en tant que lecture environnement (nodeCall a NILINode)
   en tant que sortie de ?nodeCall sinon */
extern void ptINode_becomes_lecenv(ptINode vinstr, ptexpcal exp, ptINode nodecall)
{
  atmknodelec(vinstr, exp, getReadId());
  vinstr->U.PCALL.ProduceByInode = nodecall;
}

/** Caracterisation du noeud ?VINSTR en tant qu'écriture environnement */
extern void ptINode_becomes_ecrenv(ptINode vinstr, ptexpcal exp)
{
  atmknodeecr(vinstr, exp, getWriteId());
}

/* SUB-GRAPH  */
/* -------------------------------------------------------------------------------- */
/** -->True si ?VINSTR est un appel externe */
extern boolean ptINode_satisfies_subGraph(ptINode vinstr)
{
  return (vinstr->t == opsubgraph);
}

/** Sous-graphe designe par le noeud  ?VINSTR */
extern ptgfd ptINode_get_subGraph(ptINode vinstr)
{
  if (!ptINode_satisfies_subGraph(vinstr))
    return NILGFD;
  return vinstr->U.SUBGRAPH.vgfd;
}

/** Caracterisation du noeud ?VINSTR en tant que sous-graphe etiquette par ?LABEL.
   Les noeuds du sous-graphe sont geres dans la structure ?GFD */
extern void ptINode_becomes_subGraph(ptINode vinstr, ptexpcal label, ptgfd gfd)
{
  ptINode_set_oper(vinstr, opsubgraph);
  vinstr->U.SUBGRAPH.vgfd = gfd;
  ptINode_set_label(vinstr, label);
  vinstr->U.SUBGRAPH.InternalNode = ptINode_create();
  ptINode_set_horl(vinstr->U.SUBGRAPH.InternalNode, vinstr->horl);
  vinstr->inNode = NILEXPCAL;
  vinstr->outNode = NILEXPCAL;
  vinstr->U.SUBGRAPH.lignee = pk_false;
}

///
extern void ptINode_set_trainClusterTag(ptINode vinstr)
{
  vinstr->U.SUBGRAPH.lignee = pk_true;
}

///
extern boolean ptINode_is_trainCluster(ptINode vinstr)
{
  if (ptINode_satisfies_subGraph(vinstr))
    return vinstr->U.SUBGRAPH.lignee;
  return pk_false;
}

///
extern void ptINode_set_inOutGraphRef(ptINode vinstr, ptexpcal le, ptexpcal ls)
{
  if (ptINode_satisfies_subGraph(vinstr))
    {
      vinstr->inNode = le;
      vinstr->outNode = ls;
    }
}

///
extern boolean ptINode_get_inOutGraphRef(ptINode vinstr, ptexpcal * le, ptexpcal * ls)
{
  if (!ptINode_satisfies_subGraph(vinstr))
    return pk_false;
  *le = vinstr->inNode;
  *ls = vinstr->outNode;
  return pk_true;
}

/*
  It associates a label (vlabel) to a node (n)
 */
extern void ptINode_set_label(ptINode n, ptexpcal vlabel)
{
  n->label = vlabel;
}

extern boolean ptINode_get_label(ptINode vinstr, ptexpcal * vlab)
{
  if (vinstr->label == NILEXPCAL)
    return pk_false;
  (*vlab) = vinstr->label;
  return pk_true;
}

///
extern boolean ptINode_get_subGraphInputNode(ptINode subgraph, ptINode * intNode)
{
  if (!ptINode_satisfies_subGraph(subgraph))
    return pk_false;
  *intNode = subgraph->U.SUBGRAPH.InternalNode;
  return pk_true;
}

/* ProcCall  */
/* ----------------------------------------------------------------------- */

/** --> True if ?VINSTR is an external node, but not an I/O node */
extern boolean ptINode_satisfies_procCall(ptINode vinstr)
{
  if (vinstr->t != opappe)
    return pk_false;
  return !ptINode_satisfies_writeNode(vinstr) && !ptINode_satisfies_readNode(vinstr);
}

/** -->True if the node ?n is a list of part definition one. */
extern boolean ptINode_satisfies_setOfPartialDefinitions(ptINode n)
{
  return (n->t == opldefpart);
}

/** @return the set of expressions of a part definition */
extern ptlINode ptINode_get_SetPartDefinition(ptINode n)
{
  return n->U.ldefpart;
}

/** Set the expressions (ln)  of a part definition (n) */
extern void ptINode_set_SetPartDefinition(ptINode n, ptlINode ln)
{
  n->U.ldefpart = ln;
}

/** Update the set of expressions of a part definition */
extern void ptINode_becomes_setOfPartialDefinitions(ptINode n, ptexpcal s, ptlINode ln)
{
  ptINode_set_oper(n, opldefpart);
  ptINode_set_leftOperand(n, s);
  ptINode_set_SetPartDefinition(n, ln);
}

/** @return "true" if the node ?n is a part definition one  ( ::= ), "false" otherwise.*/
extern boolean ptINode_satisfies_partialDefinition(ptINode n)
{
  return (n->t == opdefpart);
}

/** It defines a node (n) by l ::= r  */
extern void ptINode_becomes_partialDefinition(ptINode n, ptexpcal l, ptexpcal r)
{
  ptINode_set_oper(n, opdefpart);
  ptINode_set_leftOperand(n, l);
  ptINode_set_rightOperand(n, r);
  ptINode_append_luse(n, r);
}

// ------------------ Array 

/** @return "true" if node is an array node.*/
extern boolean ptINode_satisfies_array(ptINode node)
{
  if (node == NILINODE)
    return pk_false;		// strange!!
  return (node->t == opmoti);
}

///
extern boolean ptINode_get_array_inputNode(ptINode n, ptINode * intNode)
{
  if (!ptINode_satisfies_array(n))
    return pk_false;
  *intNode = n->U.ARRAY.internalNodeArray;
  return pk_true;
}

/** Defining an array node by modifying a node descriptor.
  @return none
  @node (in) : a node descriptor.
  @param (in) i:index of the array;
  @vWith (in) : Graph of "with" part of the array.
  @vBody (in): Graph of the array.
   
 */
extern void ptINode_becomes_array(ptINode node, ptINode indexNode, ptgfd vWith, ptgfd vBody)
{
  ptINode_set_oper(node, opmoti);
  node->U.ARRAY.indexNode = indexNode;
  node->U.ARRAY.initArray = vWith;
  node->U.ARRAY.bodyArray = vBody;
  node->U.ARRAY.internalNodeArray = ptINode_create();
  ptINode_set_horl(node->U.ARRAY.internalNodeArray, node->horl);
}

/**
  @return the informations for a array node.
  @node (in) : an array node.
  @param (out) vIndexNode: the node assigned to the index of the array;
  @vWith (out) : Graph of "with" part of the array.
  @vBody (out): Graph of the array.
*/
extern void ptINode_get_array_arg(ptINode node, ptINode * vIndexNode, ptgfd * vWith, ptgfd * vBody)
{
  (*vIndexNode) = node->U.ARRAY.indexNode;
  (*vWith) = node->U.ARRAY.initArray;
  (*vBody) = node->U.ARRAY.bodyArray;
}

///
extern ptgfd ptINode_get_array_bodyGraph(ptINode node)
{
  return node->U.ARRAY.bodyArray;
}

///
extern ptgfd ptINode_get_array_initGraph(ptINode node)
{
  return node->U.ARRAY.initArray;
}

/** @return the node of the index of an array of processes. */
extern ptINode ptINode_get_array_index(ptINode node)
{
  return node->U.ARRAY.indexNode;
}

/**
  !LE, !LS  ensembles des signaux d'entree (resp. de sortie) du noeud NODE
  designant un ARRAY
*/
extern void ptINode_get_array_io(ptINode node, ptexpcal * inputs, ptexpcal * outputs)
{
  ptINode_update(node);
  (*inputs) = node->inNode;
  (*outputs) = node->outNode;
}

///
extern ptgfd ptINode_get_bodySubGraph(ptINode node)
{
  return node->U.ARRAY.bodyArray;
}

///
extern ptgfd ptINode_get_initSubGraph(ptINode node)
{
  return node->U.ARRAY.initArray;
}

// ---- A revoir 

/** --> True si ?NODE est interne a un noued composé. */
extern boolean ptINode_satisfies_InComposedNode(ptINode node)
{
  return (node->ptINodebc != NILINODE);
}

/** ?NODE est un noeud interne au noeud composé ?cNode */
extern void ptINode_set_inComposedNode(ptINode node, ptINode cNode)
{
  node->ptINodebc = cNode;
}

/** --> vrai pour ?node element du noeud composé !cnode. */
extern boolean ptINode_get_inComposedNode(ptINode node, ptINode * cnode)
{
  *cnode = node->ptINodebc;
  return ((*cnode) != NILINODE);
}

/** -->Liste des noeuds (signaux) successeurs de ?VINSTR. */
extern ptldependence ptINode_get_iterationPrecedence(ptINode vinstr)
{
  return vinstr->iterationPrecedenceLst;
}

/** Les signaux definis dans le noeuds ?VINSTR sont utilises dans les instructions ?VLSUCC. */
extern void ptINode_set_iterationPrecedence(ptINode vinstr, ptldependence vlsucc)
{
  vinstr->iterationPrecedenceLst = vlsucc;
}

/// ATTRIBUTES/PRAGMAS
extern boolean ptINode_satisfies_marked(ptINode node, typeMark val)
{
  return node->satisfiesMarked(val);
}

///
extern void ptINode_addMark(ptINode node, typeMark val)
{
  node->addMark(val);
}

///
extern void ptINode_removeMark(ptINode node, typeMark val)
{
  node->removeMark(val);
}

///
extern boolean ptINode_getAttribute(ptINode node, typeMark num, attributeValue * valAttribute)
{
  return node->getAttribute(num, valAttribute);
}

///
extern void ptINode_addAttribute(ptINode node, typeMark num, attributeValue vattr)
{
  node->addAttribute(num, vattr);
}

///
extern void ptINode_removeAttribute(ptINode node, typeMark num)
{
  node->removeAttribute(num);
}

///
extern ptlpragmas ptINode_get_pragmas(ptINode node)
{
  return node->getPragmas();
}

///
extern boolean ptINode_addPragma(ptINode node, ptpragma aPragma)
{
  return node->addPragma(aPragma);
}

///
extern void ptINode_set_pragmas(ptINode n, ptlpragmas l)
{
  n->setPragmas(l);
}

///
extern void ptINode_addPragmas(ptINode aNode, ptlpragmas pragmas)
{
  aNode->addPragmas(pragmas);
}

///
extern boolean ptINode_remove_pragma(ptINode n, ptrepstr name)
{
  return n->removePragma(name);
}

extern BackAnnotationType ptINode_getBackAnnotation(ptINode vinstr)
{
  return vinstr->backAnnotation;
}

extern void ptINode_setBackAnnotation(ptINode vinstr, BackAnnotationType vt)
{
  vinstr->backAnnotation = vt;
}

extern void ptINode_free(ptINode node)
{
  delete node;
}
