/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return "true" if the list is empty, "false" otherwise.
*/
    boolean ldependence::isEmptyList()
{
  return (this == NILLDEPENDENCE);
}

/**
   Constructor of an element of a list of dependence with el as value.
 */
ldependence::ldependence(ptdependence el)
{
  elem = el;
  next = NILLDEPENDENCE;
}

/**
   Destructor.
 */
ldependence::~ldependence()
{
}

/**
   @return the current list from which the element elem is removed.
*/
ptldependence ldependence::remove(ptdependence elem)
{
  ptldependence list, nlist, first, pred;
  ptdependence el;
  ptINode n1, n2;
  ptinsth h;

  if (isEmptyList())
    return this;
  h = elem->getDependence(&n1, &n2);
  first = this;
  list = first;
  pred = list;
  while (list->get_next(&el, &nlist))
    {
      if (el->is_dependence(n1, n2))
	{
	  if (list == first)
	    return nlist;
	  pred->next = nlist;
	  return first;
	}
      pred = list;
      list = nlist;
    }
  return first;
}

/**
   @return 
   - "true" if the dependence (vfrom -> vto) is an element of the list, aDependence is then thiselement
   - "false" otherwise, aDependence is not modified.
*/
boolean ldependence::exist_dependence(ptINode vfrom, ptINode vto, ptdependence * aDependence)
{
  ptldependence list;
  ptdependence el;

  list = this;
  while (list->get_next(&el, &list))
    if (el->is_dependence(vfrom, vto))
      {
	*aDependence = el;
	return pk_true;
      }
  return pk_false;
}

/**
   @return 
   - "true" if the list is not empty: el is then the current element, and nList is te next list.
   - "false" otherwise (empty list), el and nList are not modified.
*/
boolean ldependence::get_next(ptdependence * el, ptldependence * nList)
{
  if (isEmptyList())
    return pk_false;
  *el = elem;
  *nList = next;
  return pk_true;
}

/**
   @return the currnet list in which an element (elem) is head-inserted, "false" otherwise.
*/
ptldependence ldependence::insert(ptdependence elem)
{
  ptldependence l1;
  if (gcldependence == NILLDEPENDENCE)
    l1 = new ldependence(elem);
  else
    {
      l1 = gcldependence;
      l1->elem = elem;
      gcldependence = gcldependence->next;
    }
  l1->next = this;
  return l1;
}

/**
   @return the size of the current list.
*/
int ldependence::getSize()
{
  ptldependence liste;
  ptdependence el;
  int n;

  n = 0;
  liste = this;
  while (liste->get_next(&el, &liste))
    n++;
  return n;
}

/**
   Free a list of dependences.
*/
void ldependence::ptldependence_free()
{
  ptldependence lprev, l;
  ptdependence el;

  if (gcldependence == NILLDEPENDENCE)
    gcldependence = this;
  else
    {
      l = gcldependence;
      while (l->get_next(&el, &l))
	lprev = l;
      lprev->next = this;
    }
}
