/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor of the dependence "vFrom -> vTo when vWeight"
*/
    Idependence::Idependence(ptINode vFrom, ptINode vTo, ptinsth vWeight)
{
  from = vFrom;
  to = vTo;
  weight = vWeight;
  dependenceKind = dataFlowDep;
  //  setwhoami(dependenceObject);
}

/** Destructor. */
Idependence::~Idependence()
{
}

/**
   @return the weight associated with the dependence (vFrom -> vFrom)
*/
ptinsth Idependence::getDependence(ptINode * vFrom, ptINode * vTo)
{
  *vFrom = from;
  *vTo = to;
  return weight;
}

/**
   It fixes the weight of the dependence to vWeight.
*/
void Idependence::setClockDependence(ptinsth vWeight)
{
  weight = vWeight;
}

/**
   @return the weight associated with the dependence
*/
ptinsth Idependence::getClockDependence()
{
  return weight;
}

/**
   @return "true" if the dependence is an explicit dependence, "false" otherwise (\see dependenceKind)
*/
boolean Idependence::satisfiesExplicitDependence()
{
  return (dependenceKind == explicitDep);
}

/**
   It specifies that the dependence is an explicit dependence.
*/
void Idependence::setExplicitDependence()
{
  dependenceKind = explicitDep;
}

/**
   @return "true" if the dependence comes from a reinforcing, "false" otherwise (\see dependenceKind)
*/
boolean Idependence::satisfiesReinforcementDependence()
{
  return (dependenceKind == reinforcementDep);
}

/**
   It specifies that the dependence is a reinforcing.
*/
void Idependence::setReinforcementDependence()
{
  dependenceKind = reinforcementDep;
}

/**
   @return "true" if the dependence is a data flow dependence, "false" otherwise (\see dependenceKind)
*/
boolean Idependence::satisfiesDataFlowDependence()
{
  return (dependenceKind == dataFlowDep);
}

/**
   It specifies that the dependence is a data flow dependence.
*/
void Idependence::setDataFlowDependence()
{
  dependenceKind = dataFlowDep;
}

/**
   @return "true" is the current dependence is vfrom -> vto, "false" otherwise.
*/
boolean Idependence::is_dependence(ptINode vfrom, ptINode vto)
{
  ptinsth h;
  ptINode n1, n2;

  h = getDependence(&n1, &n2);
  if ((n1 == vfrom) && (n2 == vto))
    return pk_true;
  return pk_false;
}

/**
   It specifies that the dependence is a data flow dependence.
*/
TypeDependenceKind Idependence::getDependenceKind()
{
  return dependenceKind;
}

/**
   It specifies the kind of dependence to tdep.
*/
void Idependence::setDependenceKind(TypeDependenceKind tdep)
{
  dependenceKind = tdep;
}
