/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <map>
#include <vector>
#include "BackAnnotation.h"

#define maxDepthCall 20
typedef integer typeInstanciationPath;
typedef typeInstanciationPath *pttypeInstanciationPath;
typedef std::map<pttypeInstanciationPath, ptObjectCommonInfos> instanciationPathTableType;
typedef instanciationPathTableType *ptinstanciationPathTableType;
typedef std::map<ptinsth, ptident> ClockNamesTable;
typedef ClockNamesTable *ptClockNamesTable;

/* @name Traceability Manager Class.
   A Tracebility Manager is assigned to each Graph.
*/
class TraceabilityManager
{ 

#include "TraceabilityManager_m.h"
#include "TraceabilityManager_g.h"
#include "TraceabilityManagerRew_g.h"

private:
  ptClockManager CKM; ///< Connected Clock manager
  ptExprManager EM;   ///< Connected Expression manager
  ptNodeManager NM;   ///< Connected Node manager

  boolean mode; ///< Operations are applied when mode is "true", (false by default).
  ntrees callTreeRoot, ///< call tree associated with the graph (macro expansion).
    currentCallTree; ///< Current call tree (a node of the callTree member).
  ptinstanciationPathTableType instanciationPathTable; ///< association between an object (signal, node,...) and the call tree from which it is produced.

  boolean modeKeepSubstitutions; ///< When true, clocks and signals are kept for each functionality.
  ptStack ClockSystemsStack, ClockSystemsStackRef;
  ptStack SignalsStack, SignalsStackRef;
  ptStack ClockNamesTableStack, ClockNamesTableStackRef;
  ptinsth hundef;
};

