/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_

void
TraceabilityManager::setModeKeepSubstitutions (boolean b)
{
  modeKeepSubstitutions = b;
}

/**
   When modeKeepSubstitutions, it pushes the current infos (Clocks and Signals) of the graph.
*/
void TraceabilityManager::beginFunctionality()
{
  ptlinsth lh;
  ptlinstances lsig;
  ptClockNamesTable clocknametable;

  if (modeKeepSubstitutions)
    {
      clocknametable = new ClockNamesTable();
      lh = mkClockFormulaToKeep(clocknametable);
      // push the clock system and the set of signals of the graph.
      ClockSystemsStack = stackPush (ClockSystemsStack, (ptElemStack) lh);
      ClockNamesTableStack = stackPush (ClockNamesTableStack, (ptElemStack) clocknametable);

      if (!EM->getExprList(opident, &lsig)) lsig= NILLINSTANCES;
      lsig = ptlinstances_copy(lsig);
      SignalsStack = stackPush (SignalsStack, (ptElemStack) lsig);
    }
}


/**
   @return 
   - the previous clock rewritten in h since the last marked function,
   - or the universal undefined clock when such clock does not exist,
   - or the operand clock(h) when the last marked function is undefined.
*/
ptinsth TraceabilityManager::getPreviousObject(ptinsth h)
{
  pointer x;
  ptlinsth lh;

  if (ClockSystemsStack == ClockSystemsStackRef) return h;
  if (!stackTop(ClockSystemsStackRef, &x)) return h; // Occurs the first print.
  lh = (ptlinsth) x;
  return getClassOf(h, lh);
}

/**
   @return 
   - the previous signal rewritten in s since the last marked function,
   - or NILEXPCAL when such object does not exist,
   - or the operand signal(s) when the last marked function is undefined.
 */
ptexpcal TraceabilityManager::getPreviousObject(ptexpcal s)
{
  pointer x;
  ptlinstances lexp;
  ptStack currentStack;

  if (SignalsStack == SignalsStackRef) return s;
  if (!stackTop(SignalsStackRef, &x)) return NILEXPCAL; // Occurs the first print.
  lexp = (ptlinstances) x;
  return getClassOf(s, lexp);
}


/**
   The current "stack" of (clocks, Signals) is marked as the referenced one.
 */
void TraceabilityManager::setCurrentAsReference()
{
  ClockSystemsStackRef = ClockSystemsStack;
  SignalsStackRef =  SignalsStack;
  ClockNamesTableStackRef = ClockNamesTableStack;
}


void TraceabilityManager::printmodifiedSinceLastMarkedFunction()
{
  ptinsth vhnew, vhprev, vh;
  ptlinsth lh;
  pointer x;
  
  return;

  if (modeKeepSubstitutions)
    {
      if (ClockSystemsStack == NILSTACK)
	cout << " >>>>>> Empty stack !!!" << endl;
      else
	{
	  if (ClockSystemsStackRef == NILSTACK)
	    cout << " >>>>>> No ClockSystem Reference " << endl;
	  else
	    if (ClockSystemsStack == ClockSystemsStackRef)
	      cout << " >>>>>>  No modifications" << endl;
	    else
	      {
		/* removed 15 March 2016 LBesnard.
		if (stackTop(ClockSystemsStack, &x))
		  {
		    lh = (ptlinsth) x;
		    while (ptlinsth_get_next (&lh, &vh))
		      {
			vhprev = getPreviousObject(vh);
			printInfoClockRew(vh, vhprev); 
		      }
		  }
		*/
		if (stackTop(ClockSystemsStackRef, &x))
		  {
		    lh = (ptlinsth) x;
		    ptlinsth lh_save = lh;
		    cout << "; Mapping of variables" << endl;
		    while (ptlinsth_get_next (&lh, &vh))
		      {
			vhnew = CKM->getRepresentativeFormula(vh);
			// vh has been rewritten in vhnew ?
			if (( vhnew != vh ) /* && ptlinsth_isin_elem(lh_save, vhnew ) */)
			  printInfoClockRew2(vh, vhnew); 
		      }
		  }
	      }
	}
    }
}

ptlinsth TraceabilityManager::getClocksPreviousSystem()
{
  pointer x;
  ptlinsth lhprev;

  if (!modeKeepSubstitutions) return NILLINSTH;
  if (ClockSystemsStack == NILSTACK) return NILLINSTH; // Empty stack
  if (ClockSystemsStackRef == NILSTACK) return NILLINSTH; // No ClockSystem Reference
  if (ClockSystemsStack == ClockSystemsStackRef)
    {
      ptStack S;
      if (!stackPrevious(ClockSystemsStack, &S))  return NILLINSTH;
      if (!stackTop(S, &x))  return NILLINSTH;
      lhprev = (ptlinsth) x; 
    }
  else
    {
      if (!stackTop(ClockSystemsStackRef, &x)) return NILLINSTH;
      lhprev = (ptlinsth) x;
    }
  return lhprev;
}


/**
   
 */
void TraceabilityManager::getClocksWithNewName( ptlinsth lhpreviousSystem, ptlinsth *lhWithNewName, ptlPointer *lprevNames)
{
  pointer x;
  ptClockNamesTable idtablePrevious;
  ptident curId, prevId;
  ptStack S;
  ptinsth vh;

  *lhWithNewName = NILLINSTH;
  *lprevNames = NILLPOINTER;
  if (!modeKeepSubstitutions) return;
  if (ClockNamesTableStackRef == NILSTACK) return; // Empty stack
  if (ClockNamesTableStackRef == NILSTACK) return; // No ClockSystem Reference
  if (ClockNamesTableStack == ClockNamesTableStackRef)
    {
      if (!stackPrevious(ClockNamesTableStack, &S)) return ;
      if (!stackTop(S, &x)) return;
      idtablePrevious = (ptClockNamesTable) x;
      while (ptlinsth_get_next (&lhpreviousSystem, &vh))      
	if ( (*idtablePrevious).find (vh) != (*idtablePrevious).end ())
	  {
	    prevId = (*idtablePrevious)[vh];
	    curId = CKM->getName(vh); 
	    if (! eq_idents(curId, prevId))
	      {
		*lhWithNewName = ptlinsth_insert_elem ( *lhWithNewName, vh);
		*lprevNames = ptlPointer_add_front(*lprevNames, prevId);
	      }
	  }
    }
}

