/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   It initializes
   - the traceability mode to the value "false" (no operations)
   - the call tree to the nil tree value  (NILTREE).
   - the current call tree  to the nil tree value (NILTREE).
*/
TraceabilityManager::TraceabilityManager (ptNodeManager aNodeManager, ptClockManager aClockManager, ptExprManager aExprManager)
{
  modeKeepSubstitutions = pk_true;
  mode = pk_false;
  SET_NILTREE (callTreeRoot);
  SET_NILTREE (currentCallTree);
  instanciationPathTable = new instanciationPathTableType ();
  CKM = aClockManager;
  EM = aExprManager;
  NM = aNodeManager;
  // --  TraceabilityManager rewritings
  ClockSystemsStack = NILSTACK;
  ClockSystemsStackRef = NILSTACK;
  SignalsStack = NILSTACK;
  SignalsStackRef  = NILSTACK;
  ClockNamesTableStack  = NILSTACK;
  ClockNamesTableStackRef = NILSTACK;

  hundef = CKM->getUndefClock ();
}

/**
   Destructor.
*/
TraceabilityManager::~TraceabilityManager ()
{
}

boolean
TraceabilityManager::setMode (boolean b)
{
  boolean x;

  x = mode;
  mode = b;
  return x;
}


/**
   @return the mode of the traceability mode: "true" (on) the instanciation path table is built, "false" (off) otherwise.
*/
boolean
TraceabilityManager::getMode ()
{
  return mode;
}

void
TraceabilityManager::pushCallTree (ntrees vcall)
{
  ntrees x;

  if (mode)
    {
      if (IS_NILTREE (callTreeRoot))
	{
	  callTreeRoot = LL_make_lfeed_emptyList ();
	  currentCallTree = callTreeRoot;
	}
      x = LL_make_lfeed_emptyList ();
      currentCallTree = ntreeList_appendLast (currentCallTree, x);
      nmakeattrptr (x, Traceability_ATTRIBUTENUMBER, (char *) vcall);
      currentCallTree = x;
    }
}


void
TraceabilityManager::popCallTree (ntrees vcall)
{
  ntrees x;

  if (mode)
    if (ISNOT_NILTREE (currentCallTree))
      {
	x = nfather (currentCallTree);
	if (x == currentCallTree) SET_NILTREE (currentCallTree); else currentCallTree = x;
      }
}


/* 
   void TraceabilityManager::add (ptObjectCommonInfos vobject){
   // if (mode) (*instanciationPathTable)[vobject] = currentCallTree; } */


void
TraceabilityManager::add (ptObjectCommonInfos vobject, integer pos)
{
  pttypeInstanciationPath anInstanciationPath;
  integer i, n;
  ntrees t;

  if (mode)
    {
      t = currentCallTree;
      anInstanciationPath = newInstanciationPath ();
      n = depth (t);
      anInstanciationPath[n + 1] = 0;	// end mark.
      anInstanciationPath[n] = pos;	// port number.
      for (i = 1; i <= n; i++)
	{
	  anInstanciationPath[n - i] = nposit (t);
	  t = nfather (t);
	}
      (*instanciationPathTable)[anInstanciationPath] = vobject;
    }
}

/*
  void TraceabilityManager::remove(ptObjectCommonInfos vobject)
  { if (mode) instanciationPathTable->erase(vobject);}
*/

void TraceabilityManager::print(ptident graphname)
{
  instanciationPathTableType::iterator it; 
  ptObjectCommonInfos x;
  ptexpcal id;
  objectType objtype;
  boolean stop;
  pttypeInstanciationPath refPort;
  integer i;
  ofstream ofile;
  ptrepstr vfilename;
  if (mode)
    {
      if (!instanciationPathTable ->empty())
      {
	ptident vid = makeIdent_ident_string (graphname, "Traceability_IO");
	if ( sys_suffixedFilePathInUserOutputDir(ident2repstr(vid), "txt", &vfilename))
	  {
	    ofile.open (repstr2string(vfilename));
	    for (it = instanciationPathTable ->begin(); it != instanciationPathTable->end(); it++)
	      {
		refPort = (*it).first;
		i = 0;
		while (refPort[i] != 0)
		  {
		    ofile << refPort[i] << " | ";
		    i++;
		  }
		ofile << " 0 | ";            
		x = (*it).second; // identifier only currently.
		objtype = x->getwhoami();
		if (objtype == exprObject)
		  {
		    id = (ptexpcal) x;
		    ofile << "identifier : " << ident2string(EM->getName(id));
		  }
		else
		  {
		    ofile << "An expression for this line";
		  }
		ofile << endl;
	      }
	    ofile.close ();
	  }
      }
    }
}

boolean
TraceabilityManager::getinstanciationPathTable (ptinstanciationPathTableType * t)
{
  *t = instanciationPathTable;
  return mode;
}


ntrees TraceabilityManager::getCurrentCallTree()
{
  return currentCallTree;
}

