/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <iostream>
#include <fstream>
using namespace std;

_PROTECTED_ integer TraceabilityManager::depth (ntrees t)
{
  integer n;
  n = 0;
  if (ISNOT_NILTREE (t))
    while (pk_true)
      {
	if (t == callTreeRoot) return n;
	n++;
	t = nfather (t);
      }
  return n;
}


pttypeInstanciationPath
TraceabilityManager::newInstanciationPath ()
{
  pttypeInstanciationPath vline;
  integer i;

  //  nbEntry ++;
  vline = new typeInstanciationPath[maxDepthCall];
  for (i = 0; i < maxDepthCall; i++)
    vline[i] = 0;
  //  TraceablitiyMatrix.push_back(vline);
  return vline;
}



ptinsth TraceabilityManager::getClassOf(ptinsth h, ptlinsth lh)
{
  ptinsth vh;
  while (ptlinsth_get_next (&lh, &vh))
    if ( CKM->getRepresentativeFormula(vh) == CKM->getRepresentativeFormula(h))
      return vh;
  
  return hundef;
}


ptexpcal TraceabilityManager::getClassOf(ptexpcal s, ptlinstances lexp)
{
  ptexpcal newcurExp, curExp;
  while (ptlinstances_get_next (&lexp, &curExp))
    if (EM->getIndirection(curExp, &newcurExp))
      if (newcurExp == s) return curExp;
  return NILEXPCAL;
}



void TraceabilityManager::printInfoClockRew(ptinsth vh, ptinsth vhprev)
{
  ptident id_vhprev, id_vh;

  id_vh = CKM->getName(vh);
  if (vhprev == hundef) 
    cout << " Previous clock of " << vh << " ( " << ident2string(id_vh) << " )" << " is undefined !" << endl;
  else
    {
      id_vhprev = CKM->getName(vhprev);
      if ( ! eq_idents(id_vh, id_vhprev))
	cout << " " << ident2string(id_vh) << " was " << ident2string(id_vhprev) << endl;
      /**
	 if (vhprev == vh)
	 cout << " No modification for " << vh << " ( " << ident2string(id_vh) << " )" << endl;
	 else
	 cout << vh << "( " << ident2string(id_vh) << " )" << " was " << vhprev << "( " << ident2string(id_vhprev) << " )"<< endl;
      */
    }
}

void TraceabilityManager::printInfoClockRew2(ptinsth vh, ptinsth vhnew)
{
  ptident id_vhnew, id_vh;

  id_vh = CKM->getName(vh);
  id_vhnew = CKM->getName(vhnew);
  if ( ! eq_idents(id_vh, id_vhnew))
    cout << "(= " << ident2string(id_vh) << " " << ident2string(id_vhnew) << ")"  << endl;
}


ptlinsth TraceabilityManager::mkClockFormulaToKeep(ptlinsth lhTovisit, ptlinsth lh, ptClockNamesTable clocknametable)
{
  ptinsth nextFormula, vh;

  while (ptlinsth_get_next(&lhTovisit, &vh))
    {
      lh = ptlinsth_insert_elem (lh, vh);
      (*clocknametable)[vh] = CKM->getName(vh);
      nextFormula = CKM->getNextFormula(vh, pk_true);
      while (nextFormula != vh)
	{
	  lh = ptlinsth_insert_elem (lh, nextFormula);
	  nextFormula = CKM->getNextFormula(nextFormula, pk_true);
	}
    }
  return lh;
}


/**

 */
ptlinsth TraceabilityManager::mkClockFormulaToKeep( ptClockNamesTable clocknametable)
{
  ptlinsth lh;

  lh = mkClockFormulaToKeep( CKM->getlhroots(), NILLINSTH, clocknametable);
  lh = mkClockFormulaToKeep(  CKM->getlhplaced(), lh, clocknametable);
  return lh;
}
