/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         Named Types                      **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/


extern TypeType
NamedTypeCreate (NameClass Name, TypeType Term)
{
  // if a named type with Term as value exists, it is returned. Otherwise a new named type is created
  TypeType TT;
  if (NamedFind2 (Term, &TT))
    return TT;
  return NamedCreate (Name, Term);
}


extern TypeType
NamedTypeCreate2 (NameClass Name, TypeType Term)
{
  // if a named type with Term as value exists and Name as name, it is returned. Otherwise a new named type is created
  TypeType TT;
  if (!NamedFind (Name, Term, &TT))
    TT = NamedCreate (Name, Term);
  return TT;
}

extern BoolType
TypeTestNamedClass (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  if (TypeTestPredef (TT))
    return pk_false;
  return (macroCltestNamedClass (macroTygetclass (TT)));
}

extern NameClass
NamedTypeGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (MetaNamedGetName (TT));
}

extern TypeType
NamedTypeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (NamedGetTerm (TT));
}

extern TypeType
NamedTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (MetaNamedGetCurrent (TT));
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         Lambda Types                     **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
LambdaTypeCreate (TypeType Term)
{
  TypeType TT;
  if (LambdaFind (Term, &TT));
  else
    TT = LambdaCreate (Term);
  return TT;
}

extern TypeType
LambdaTypeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (LambdaGetTerm (TT));
}

extern BoolType
LambdaTypeTestDefined (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (LambdaTestDefined (TT));
}

extern BoolType
TypeTestLambdaClass (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroCltestLambdaClass (macroTygetclass (TT)));
}

extern TypeType
LambdaTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (LambdaGetCurrent (TT));
}

/*
  TT is assumed to be reduced.
  TT is assumed to have Subrange Class Term.
*/
extern TypeType
LambdaTypeSetBounds (TypeType TT, ExpcalCst Bounds, BoolType Generic)
  /*---------------***********---------------------------------------*/
{
  LambdaTypeRepr *Descr;
  TypeType RTT;
  Descr = (macroTylambdaSTree_getDescriptionr (TT));
  RTT = macroLambdagetterm (Descr);
  if (TypeTestSubrangeClass (RTT))
    {
      RTT = SubrangeTypeCreate (Bounds, SubrangeTypeGetTerm (RTT));
      if (Generic)
	RTT = LambdaTypeCreate (RTT);
      macroDesetcurrent (Descr, RTT);
      return RTT;
    }
  else if (TypeTestBitsClass (RTT))
    {
      RTT = BitsTypeCreate (Bounds);
      if (Generic)
	RTT = LambdaTypeCreate (RTT);
      macroDesetcurrent (Descr, RTT);
      return RTT;
    }
  else
    return TT;
}

/*
  TT is assumed to be reduced
  TT is assumed to have Subrange Class Term
*/
extern TypeType
LambdaTypeSetDimLst (TypeType TT, ExpcalCstList DimLst, BoolType Generic)
  /*---------------***********---------------------------------------*/
{
  LambdaTypeRepr *Descr;
  TypeType RTT;
  Descr = (macroTylambdaSTree_getDescriptionr (TT));
  RTT = macroLambdagetterm (Descr);
  if (TypeTestArrayClass (RTT))
    {
      RTT = ArrayTypeCreate (DimLst, ArrayMatrixTypeGetTerm (RTT));
      if (Generic)
	RTT = LambdaTypeCreate (RTT);
      macroDesetcurrent (Descr, RTT);
      return RTT;
    }
  else if (TypeTestMatrixClass (RTT))
    {
      RTT = MatrixTypeCreate (DimLst, ArrayMatrixTypeGetTerm (RTT));
      if (Generic)
	RTT = LambdaTypeCreate (RTT);
      macroDesetcurrent (Descr, RTT);
      return RTT;
    }
  else
    return TT;
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		            Any Types                     **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

/* --------------------------------------------------------
   AnyTypeCreate()
   ========================================================
   returns:
   a new AnyType
*/
extern TypeType
AnyTypeCreate ()
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  AnyTypeRepr *Descr;
  Descr = (AnyTypeRepr *) (malloc (sizeof (AnyTypeRepr)));
  TT = NewTypeItem (AnyClass, Descr);
  AppendClassEntry (AnyClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroDesetcurrent (Descr, TT) return TT;
}

extern BoolType
AnyTypeTestDefined (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (AnyTestDefined (TT));
}

extern BoolType
TypeTestAnyClass (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;		// Modif LBesnard Jan 2001
  return (macroCltestAnyClass (macroTygetclass (TT)));
}

extern TypeType
AnyTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (AnyGetDefinition (TT));
}

extern void
AnyTypeSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  AnySetCurrent (TT, TE);
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		          Meta Types                      **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

/* --------------------------------------------------------
   MetaTypeCreate(NameClass Name)
   ========================================================
   returns:
   a new MetaType named Name
*/
extern TypeType
MetaTypeCreate (NameClass Name)
  /*---------------***********---------------------------------------*/
{
  return (MetaCreate (Name));
}

extern BoolType
TypeTestMetaClass (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroCltestMetaClass (macroTygetclass (TT)));
}

extern NameClass
MetaTypeGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (MetaNamedGetName (TT));
}

extern TypeType
MetaTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (MetaNamedGetCurrent (TT));
}

extern BoolType
MetaTypeTestDefined (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (MetaTestDefined (TT));
}

/* --------------------------------------------------------
   MetaTypeNamedTypeSet(TypeType TT,TypeType Term)
   ========================================================
   returns:
   creates a NamedType RTT, giving the name of meta type TT to the
   type Term;
   RTT is set as the current type to TT and is returned as the result
   of the function.
*/
extern TypeType
MetaTypeNamedTypeSet (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  TypeType RTT;
  RTT = NamedTypeCreate2 (MetaTypeGetName (TT), TE);
  macroDesetcurrent ((macroTymetanamedSTree_getDescriptionr (TT)), RTT);
  return RTT;
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		          Abstract Types                  **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern BoolType
TypeTestAbstract (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestAbstract (macroTygetclass (TT)));
}
