/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |***                                                 File ClassAbstract_i.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**		           for Lambda types                         **|
	  |**		               Any types                            **|
	  |**		               Meta types                           **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         Lambda Types                     **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  TypeType Term;
} LambdaTypeRepr;
#define macroLambdagetterm(tyde) (tyde->Term)
#define macroLambdagetcurrent(tyde) (tyde->CurrentType)
#define macroLambdaset(tyde,term,current) {\
       (tyde->Term)=term;\
       (tyde->CurrentType)=current;\
    }
#define macroLambdasetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }
#define macroLambdasetterm(tyde,term) {\
       (tyde->Term)=term;\
    }
#define macroTylambdaSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(LambdaTypeRepr,tyty))

/* --------------------------------------------------------
   LambdaCreate(TypeType Term)
   ========================================================
   returns:
   a new LambdaType defined on term Term
*/
static TypeType
LambdaCreate (TypeType Term)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  LambdaTypeRepr *Descr;
  Descr = (LambdaTypeRepr *) (malloc (sizeof (LambdaTypeRepr)));
  TT = NewTypeItem (LambdaClass, Descr);
  AppendClassEntry (LambdaClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroLambdaset (Descr, Term, TT) return TT;
}

/* --------------------------------------------------------
   LambdaFind(TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no LambdaType built term Term
   exists
   pk_true if some existing LambdaType T is builtt on Term
   T is assigned to TT
*/
static BoolType
LambdaFind (TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  LambdaTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[LambdaClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (LambdaTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (TypeEqAdr (macroLambdagetterm (Descr), Term))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	}
    }
  return result;
}

static TypeType
LambdaGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroLambdagetterm (macroTylambdaSTree_getDescriptionr (TT)));
}

static TypeType
LambdaGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroDegetcurrent (macroTylambdaSTree_getDescriptionr (TT)));
}

static BoolType
LambdaTestDefined (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  LambdaTypeRepr *Descr;
  Descr = (macroTylambdaSTree_getDescriptionr (TT));
  return (!((macroDegetcurrent (Descr)) == TT));
}

static BoolType
LambdaRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = LambdaGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  else
    {
      switch (macroTygetclass (*RTT))
	{
	case SubrangeClass:
	case BitsClass:
	  break;
	case ArrayClass:
	case MatrixClass:
	  ArrayMatrixRecTestandGetCurrent (*RTT, RTT);
	  break;
	case LambdaClass:
	  LambdaRecTestandGetCurrent (*RTT, RTT);
	}
      return pk_true;
    }
}

static void
LambdaSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  LambdaTypeRepr *Descr;
  Descr = (macroTylambdaSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
LambdaReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType Term;
  if (LambdaRecTestandGetCurrent (TT, RTT))
    {
      if (TypeReduction (*RTT, RTT))
	LambdaSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      /* No current value, TT remains thus a lambda type */
      if (TypeReduction (LambdaGetTerm (TT), &Term))
	{
	  if (LambdaFind (Term, RTT));
	  else
	    *RTT = LambdaCreate (Term);
	  LambdaSetCurrent (TT, *RTT);
	  return pk_true;
	}
      else
	return pk_false;
    }
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		            Any Types                     **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
} AnyTypeRepr;
#define macroAnygetcurrent(tyde) (tyde->CurrentType)
#define macroAnysetcurrent(tyde,term) {\
       (tyde->CurrentType)=term;\
    }
#define macroTyanySTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(AnyTypeRepr,tyty))

static BoolType
AnyTestDefined (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  AnyTypeRepr *Descr;
  Descr = (macroTyanySTree_getDescriptionr (TT));
  return (!((macroDegetcurrent (Descr)) == (macroDegetmyself (Descr))));
}

static TypeType
AnyGetDefinition (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroDegetcurrent (macroTyanySTree_getDescriptionr (TT)));
}

static BoolType
AnyRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*----------***********************--------------------------*/
{
  *RTT = AnyGetDefinition (TT);
  if (*RTT == TT)
    return pk_false;
  TypeRecTestandGetCurrent (TT, RTT);
  return pk_true;
}

static void
AnySetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  AnyTypeRepr *Descr;
  Descr = (macroTyanySTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
AnyReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  if (AnyRecTestandGetCurrent (TT, RTT))
    {
      if (TypeReduction (*RTT, RTT))
	AnySetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    return pk_false;
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		      Meta Named Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  NameClass Name;
  TypeType Term;
} MetaNamedTypeRepr;

#define macroMetaNamedgetname(tyde) (tyde->Name)
#define macroMetaNamedgetcurrent(tyde) (tyde->CurrentType)
#define macroMetaNamedsetcurrent(tyde,term) {\
       (tyde->CurrentType)=term;\
    }
#define macroTymetanamedSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(MetaNamedTypeRepr,tyty))

static NameClass
MetaNamedGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroMetaNamedgetname (macroTymetanamedSTree_getDescriptionr (TT)));
}

static TypeType
MetaNamedGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroDegetcurrent (macroTymetanamedSTree_getDescriptionr (TT)));
}

static BoolType
MetaNamedRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = MetaNamedGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  MetaNamedRecTestandGetCurrent (*RTT, RTT);
  return pk_true;
}

static void
MetaNamedSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  MetaNamedTypeRepr *Descr;
  Descr = (macroTymetanamedSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

/*********|*********|*********|*********|*********|
	  |**					      	**|
	  |**		      Named                     **|
	  |**						**|
	  |*********|*********|*********|*********|*********/

#define macroNamedgetterm(tyde) (tyde->Term)
#define macroNamedset(tyde,name,term,current) {\
       (tyde->Name)=name;\
       (tyde->Term)=term;\
       (tyde->CurrentType)=current;\
    }
#define macroNamedsetterm(tyde,term) {\
       (tyde->Term)=term;\
    }

static TypeType
NamedGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroNamedgetterm (macroTymetanamedSTree_getDescriptionr (TT)));
}

/* --------------------------------------------------------
   NamedCreate(NameClass Name,TypeType Term)
   ========================================================
   returns:
   a new MetaType named Name
*/
static TypeType
NamedCreate (NameClass Name, TypeType Term)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  MetaNamedTypeRepr *Descr;
  Descr = (MetaNamedTypeRepr *) (malloc (sizeof (MetaNamedTypeRepr)));
  TT = NewTypeItem (NamedClass, Descr);
  AppendClassEntry (NamedClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroNamedset (Descr, Name, Term, TT);
  return TT;
}

/* --------------------------------------------------------
   NamedFind(NameClass Name, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no NamedType with name Name and type value TT
   exists
   pk_true if some existing NamedType T has 
   name Name and type value TT
   T is assigned to TT
*/
static BoolType
NamedFind (NameClass Name, TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  MetaNamedTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[NamedClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (MetaNamedTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (NameEq (macroMetaNamedgetname (Descr), Name) && TypeEqAdr (macroNamedgetterm (Descr), Term))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	}
    }
  return result;
}

/* --------------------------------------------------------
   NamedFind(NameClass Name, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no NamedType with name Name and type value TT
   exists
   pk_true if some existing NamedType T has 
   name Name and type value TT
   T is assigned to TT
*/
static BoolType
NamedFind2 (TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  TypeReprSubList ReprLst;
  MetaNamedTypeRepr *Descr;
  ReprLst = TypeClassTable[NamedClass].First;
  while (!(NullPointerDef (ReprLst)))
    {
      Descr = (MetaNamedTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (TypeEqAdr (macroNamedgetterm (Descr), Term))
	{
	  *TT = macroDegetmyself (Descr);
	  return pk_true;
	}
    }
  return pk_false;
}

static BoolType
NamedReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType Term;
  NameClass Name;
  if (MetaNamedRecTestandGetCurrent (TT, RTT))
    {
      if (TypeReduction (*RTT, RTT))
	MetaNamedSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      if (TypeReduction (NamedGetTerm (TT), &Term))
	{
	  Name = MetaNamedGetName (TT);
	  if (NamedFind (Name, Term, RTT));
	  else
	    *RTT = NamedCreate (Name, Term);
	  MetaNamedSetCurrent (TT, *RTT);
	  return pk_true;
	}
      else
	return pk_false;
    }
}

/*********|*********|*********|*********|*********|
	  |**					      	**|
	  |**		      Meta                      **|
	  |**						**|
	  |*********|*********|*********|*********|*********/

#define macroMetaset(tyde,name,current) {\
       (tyde->Name)=name;\
       (tyde->CurrentType)=current;\
    }

/* --------------------------------------------------------
   MetaCreate(NameClass Name)
   ========================================================
   returns:
   a new MetaType named Name
*/
TypeType
MetaCreate (NameClass Name)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  MetaNamedTypeRepr *Descr;
  Descr = (MetaNamedTypeRepr *) (malloc (sizeof (MetaNamedTypeRepr)));
  TT = NewTypeItem (MetaClass, Descr);
  AppendClassEntry (MetaClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroMetaset (Descr, Name, TT);
  return TT;
}

static BoolType
MetaTestDefined (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (MetaNamedGetCurrent (TT) != TT);
}

static BoolType
MetaReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  if (MetaNamedRecTestandGetCurrent (TT, RTT))
    {
      if (TypeReduction (*RTT, RTT))
	MetaNamedSetCurrent (TT, *RTT);
      return pk_true;
    }
  return pk_false;
}
