/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |***                                              File ClassApplication_g.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           GLOBAL FUNCTIONS                         **|
	  |**		           for Node types                           **|
	  |**		               GenericNode types                    **|
	  |**		               MetaNodes types                      **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		            Node types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
NodeTypeCreate (TypeType Domain, TypeType Image, ClkSystClass Clocks, GraphClass Graph)
{
  /* Nothing is currently done with Clocks and Graph */

  TypeType TT;
  if (NodeFind (Domain, Image, Clocks, Graph, &TT));
  else
    TT = NodeCreate (Domain, Image, Clocks, Graph);
  return TT;
}

/* --------------------------------------------------------
   TypeTestNodeClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a node class
   ------------------------------------------------------------*/
extern BoolType
TypeTestNodeClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestNodeClass (macroTygetclass (TT)));
}

extern TypeType
NodeTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (NodeGetCurrent (TT));
}

extern TypeType
NodeTypeGetDomain (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  return (NodeGetDomain (TT));
}

extern TypeList
NodeTypeGetInputTypeLst (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  return (NodeGetInputTypeLst (TT));
}

extern TypeType
NodeTypeGetImage (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  return (NodeGetImage (TT));
}

extern TypeList
NodeTypeGetOutputTypeLst (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  return (NodeGetOutputTypeLst (TT));
}

extern ClkSystClass
NodeTypeGetClocks (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (NodeGetClocks (TT));
}

extern GraphClass
NodeTypeGetGraph (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (NodeGetGraph (TT));
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     GenericNode types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
GenericNodeTypeCreate (TypeType Argument, TypeType Term)
{
  TypeType TT;
  if (GenericNodeFind (Argument, Term, &TT));
  else
    TT = GenericNodeCreate (Argument, Term);
  return TT;
}

/* --------------------------------------------------------
   TypeTestGenericNodeClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a generic node class
   ------------------------------------------------------------*/
extern BoolType
TypeTestGenericNodeClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestGenericNodeClass (macroTygetclass (TT)));
}


extern TypeType
GenericNodeTypeGetArgument (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (GenericNodeGetArgument (TT));
}

extern TypeType
GenericNodeTypeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (GenericNodeGetTerm (TT));
}

extern TypeType
GenericNodeTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (GenericNodeGetCurrent (TT));
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		        MetaNode types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

/* --------------------------------------------------------
   TypeTestMetaNodeClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a meta node class
   ------------------------------------------------------------*/
extern BoolType
TypeTestMetaNodeClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestMetaNodeClass (macroTygetclass (TT)));
}

extern TypeType
MetaNodeTypeCreate (TypeType TypeNameLst, TypeType Term)
{
  TypeType TT;
  if (MetaNodeFind (TypeNameLst, Term, &TT));
  else
    TT = MetaNodeCreate (TypeNameLst, Term);
  return TT;
}

extern TypeType
MetaNodeTypeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return MetaNodeGetTerm (TT);
}

extern TypeType
MetaNodeTypeGetArgument (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return MetaNodeGetArgument (TT);
}

extern TypeType
MetaNodeTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return MetaNodeGetCurrent (TT);
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     Application types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern BoolType
TypeTestApplication (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestApplication (macroTygetclass (TT)));
}
