/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |***                                              File ClassApplication_i.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**		           for Node types                           **|
	  |**		               GenericNode types                    **|
	  |**		               MetaNodes types                      **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		            Node types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  TypeType Domain;
  TypeType Image;
  ClkSystClass Clocks;
  GraphClass Graph;
} NodeTypeRepr;
#define macroNodegetdomain(tyde) (tyde->Domain)
#define macroNodegetimage(tyde) (tyde->Image)
#define macroNodegetclocks(tyde) (tyde->Clocks)
#define macroNodegetgraph(tyde) (tyde->Graph)
#define macroNodegetcurrent(tyde) (tyde->CurrentType)
#define macroNodeset(tyde,domain,image,clocks,graph,current) {\
       (tyde->Domain)=domain;\
       (tyde->Image)=image;\
       (tyde->Clocks)=clocks;\
       (tyde->Graph)=graph;\
       (tyde->CurrentType)=current;\
    }
#define macroNodesetdomain(tyde,domain) {\
       (tyde->Domain)=domain;\
    }
#define macroNodesetimage(tyde,image) {\
       (tyde->Image)=image;\
    }
#define macroNodesetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }
#define macroTynodeSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(NodeTypeRepr,tyty))

static TypeType
NodeGetDomain (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroNodegetdomain (macroTynodeSTree_getDescriptionr (TT)));
}

static TypeType
NodeGetImage (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroNodegetimage (macroTynodeSTree_getDescriptionr (TT)));
}

static ClkSystClass
NodeGetClocks (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroNodegetclocks (macroTynodeSTree_getDescriptionr (TT)));
}

static GraphClass
NodeGetGraph (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroNodegetgraph (macroTynodeSTree_getDescriptionr (TT)));
}

static TypeType
NodeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroNodegetcurrent (macroTynodeSTree_getDescriptionr (TT)));
}

static BoolType
NodeRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = NodeGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  NodeRecTestandGetCurrent (*RTT, RTT);
  return pk_true;
}

/* --------------------------------------------------------
   NodeCreate(TypeType Domain,TypeType Image,ClkSystClass Clocks, GraphClass Graph)
   ========================================================
   returns:
   a new NodeType 
   ------------------------------------------------------------*/
static TypeType
NodeCreate (TypeType Domain, TypeType Image, ClkSystClass Clocks, GraphClass Graph)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  NodeTypeRepr *Descr;
  Descr = (NodeTypeRepr *) (malloc (sizeof (NodeTypeRepr)));
  TT = NewTypeItem (NodeClass, Descr);
  AppendClassEntry (NodeClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroNodeset (Descr, Domain, Image, Clocks, Graph, TT) return TT;
}

/* --------------------------------------------------------
   NodeFind(TypeType Domain,TypeType Image,ClkSystClass Clocks, GraphClass Graph, TypeType *TT)
   ========================================================
   returns:
   false if no identical Node Type 
   exists
   pk_true if such a NodeType T exist
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
NodeFind (TypeType Domain, TypeType Image, ClkSystClass Clocks, GraphClass Graph, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  NodeTypeRepr *Descr;
  TypeType RefDomain;
  TypeType RefImage;
  result = pk_false;
  ReprLst = TypeClassTable[NodeClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (NodeTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      RefDomain = macroNodegetdomain (Descr);
      RefImage = macroNodegetimage (Descr);
      if (TypeEqAdr (RefDomain, Domain) &&
	  TypeEqAdr (RefImage, Image) && ClocksEq (macroNodegetclocks (Descr), Clocks) && GraphEq (macroNodegetgraph (Descr), Graph))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	};
    };
  return result;
}

static TypeList
NodeGetInputTypeLst (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  return (StructureGetTypeLst (NodeGetDomain (TT)));
}

static TypeList
NodeGetOutputTypeLst (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  return (StructureGetTypeLst (NodeGetImage (TT)));
}

static void
NodeSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  NodeTypeRepr *Descr;
  Descr = (macroTynodeSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
NodeReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT;
  TypeType DTT;
  TypeType ITT;
  BoolType ResD, ResI;
  ClkSystClass Clocks;
  GraphClass Graph;
  if (NodeRecTestandGetCurrent (TT, &ATT))
    {
      if (TypeReduction (ATT, RTT))
	NodeSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      ResD = StructureReduction (NodeGetDomain (TT), &DTT);
      ResI = StructureReduction (NodeGetImage (TT), &ITT);
      if (ResD || ResI)
	{
	  Clocks = NodeGetClocks (TT);
	  Graph = NodeGetGraph (TT);
	  if (NodeFind (DTT, ITT, Clocks, Graph, RTT));
	  else
	    *RTT = NodeCreate (DTT, ITT, Clocks, Graph);
	  NodeSetCurrent (TT, *RTT);
	  return pk_true;
	}
      else
	return pk_false;
    }
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     GenericNode types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  TypeType Argument;
  TypeType Term;
} GenericNodeTypeRepr;
#define macroGenericNodegetargument(tyde) (tyde->Argument)
#define macroGenericNodegetterm(tyde) (tyde->Term)
#define macroGenericNodegetcurrent(tyde) (tyde->CurrentType)
#define macroGenericNodeset(tyde,argument,term,current) {\
       (tyde->Argument)=argument;\
       (tyde->Term)=term;\
       (tyde->CurrentType)=current;\
    }
#define macroGenericNodesetargument(tyde,argument) {\
       (tyde->Argument)=argument;\
    }
#define macroGenericNodesetterm(tyde,term) {\
       (tyde->Term)=term;\
    }
#define macroGenericNodesetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }
#define macroGenericNodesetterm(tyde,term) {\
       (tyde->Term)=term;\
    }
#define macroGenericNodesetargument(tyde,argument) {\
       (tyde->Argument)=argument;\
    }
#define macroTygenericnodeSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(GenericNodeTypeRepr,tyty))

static TypeType
GenericNodeGetArgument (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroGenericNodegetargument (macroTygenericnodeSTree_getDescriptionr (TT)));
}

static TypeType
GenericNodeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroGenericNodegetterm (macroTygenericnodeSTree_getDescriptionr (TT)));
}

static TypeType
GenericNodeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroDegetcurrent (macroTygenericnodeSTree_getDescriptionr (TT)));
}

static BoolType
GenericNodeRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = GenericNodeGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  else
    {
      if (GenericNodeRecTestandGetCurrent (*RTT, RTT));
      return pk_true;
    }
}

/* --------------------------------------------------------
   GenericNodeCreate(TypeType Argument, TypeType Term)
   ========================================================
   returns:
   a new GenericNodeType 
*/
static TypeType
GenericNodeCreate (TypeType Argument, TypeType Term)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  GenericNodeTypeRepr *Descr;
  Descr = (GenericNodeTypeRepr *) (malloc (sizeof (GenericNodeTypeRepr)));
  TT = NewTypeItem (GenericNodeClass, Descr);
  AppendClassEntry (GenericNodeClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroGenericNodeset (Descr, Argument, Term, TT);
  return TT;
}

/* --------------------------------------------------------
   GenericNodeFind(TypeType Argument, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no identical GenericNodeType 
   exists
   pk_true if some existing GenericNodeType T has 
   the same Argument and Term
   T is assigned to TT
*/
static BoolType
GenericNodeFind (TypeType Argument, TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  GenericNodeTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[GenericNodeClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (GenericNodeTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (TypeEqAdr (macroGenericNodegetargument (Descr), Argument) && TypeEqAdr (macroGenericNodegetterm (Descr), Term))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	}
    }
  return result;
}

static void
GenericNodeSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  GenericNodeTypeRepr *Descr;
  Descr = (macroTygenericnodeSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
GenericNodeReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT, ATT1;
  TypeType TTT;
  BoolType ResA, ResT;
  if (GenericNodeRecTestandGetCurrent (TT, &ATT))
    {
      if (TypeReduction (ATT, RTT))
	GenericNodeSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      ResA = TypeReduction (GenericNodeGetArgument (TT), &ATT1);
      ResT = TypeReduction (GenericNodeGetTerm (TT), &TTT);
      if (ResA || ResT)
	{
	  if (GenericNodeFind (ATT1, TTT, RTT));
	  else
	    *RTT = GenericNodeCreate (ATT1, TTT);
	  GenericNodeSetCurrent (TT, *RTT);
	  return pk_true;
	}
      else
	return pk_false;
    }
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		        MetaNode types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  TypeType Argument;
  TypeType Term;
} MetaNodeTypeRepr;
/* Term is assumed to be a generic node or a node */
#define macroMetaNodegetargument(tyde) (tyde->Argument)
#define macroMetaNodegetterm(tyde) (tyde->Term)
#define macroMetaNodegetcurrent(tyde) (tyde->CurrentType)
#define macroMetaNodeset(tyde,argument,term,current) {\
       (tyde->Argument)=argument;\
       (tyde->Term)=term;\
       (tyde->CurrentType)=current;\
    }
#define macroMetaNodesetargument(tyde,argument) {\
       (tyde->Argument)=argument;\
    }
#define macroMetaNodesetterm(tyde,term) {\
       (tyde->Term)=term;\
    }
#define macroMetaNodesetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }
#define macroTymetanodeSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(MetaNodeTypeRepr,tyty))


static TypeType
MetaNodeGetArgument (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroMetaNodegetargument (macroTymetanodeSTree_getDescriptionr (TT)));
}

static TypeType
MetaNodeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroMetaNodegetterm (macroTymetanodeSTree_getDescriptionr (TT)));
}

static TypeType
MetaNodeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroMetaNodegetcurrent (macroTymetanodeSTree_getDescriptionr (TT)));
}

static BoolType
MetaNodeRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = MetaNodeGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  MetaNodeRecTestandGetCurrent (*RTT, RTT);
  return pk_true;
}

/* --------------------------------------------------------
   MetaNodeCreate(TypeType TypeNameLst, TypeType Term)
   ========================================================
   assumes:TypeNameLst is a tuple of metatype list
   Term is  node or a generic node
   returns:
   a new MetaNodeType having NameLst and Term as arguments
   ------------------------------------------------------------*/
static TypeType
MetaNodeCreate (TypeType TypeNameLst, TypeType Term)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  MetaNodeTypeRepr *Descr;
  Descr = (MetaNodeTypeRepr *) (malloc (sizeof (MetaNodeTypeRepr)));
  TT = NewTypeItem (MetaNodeClass, Descr);
  AppendClassEntry (MetaNodeClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroMetaNodeset (Descr, TypeNameLst, Term, TT) return TT;
}

/* --------------------------------------------------------
   MetaNodeFind(TypeType TypeNameLst, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no identical MetaNodeType 
   exists
   pk_true if such a MetaNodeType T exist
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
MetaNodeFind (TypeType TypeNameLst, TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  MetaNodeTypeRepr *Descr;
  TypeType RefTuple;
  result = pk_false;
  ReprLst = TypeClassTable[MetaNodeClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (MetaNodeTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      RefTuple = macroMetaNodegetargument (Descr);
      if (TypeEqAdr (RefTuple, TypeNameLst) && TypeEqAdr (macroMetaNodegetterm (Descr), Term))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	};
    };
  return result;
}

static void
MetaNodeSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  MetaNodeTypeRepr *Descr;
  Descr = (macroTymetanodeSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
MetaNodeReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT, ATT1;
  TypeType TTT;
  BoolType ResA, ResT;

  ATT = MetaNodeGetCurrent (TT);
  if (MetaNodeRecTestandGetCurrent (TT, &ATT))
    {
      if (TypeReduction (ATT, RTT))
	MetaNodeSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      ResA = TypeReduction (MetaNodeGetArgument (TT), &ATT1);
      ResT = TypeReduction (MetaNodeGetTerm (TT), &TTT);
      if (ResA || ResT)
	{
	  if (MetaNodeFind (ATT1, TTT, RTT));
	  else
	    *RTT = MetaNodeCreate (ATT1, TTT);
	  MetaNodeSetCurrent (TT, *RTT);
	  return pk_true;
	}
      else
	return pk_false;
    }
}
