/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |***                                              File ClassArrayMatrix_i.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**		           for Array types                          **|
	  |**		               Matrix types                         **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/
typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  ExpcalCstList DimLst;
  TypeType Term;
} ArrayMatrixTypeRepr;
#define macroArraygetdimlst(tyde) (tyde->DimLst)
#define macroArraygetterm(tyde) (tyde->Term)
#define macroArraygetcurrent(tyde) (tyde->CurrentType)
#define macroArrayset(tyde,dimlst,term,current) {\
       (tyde->DimLst)=dimlst;\
       (tyde->Term)=term;\
       (tyde->CurrentType)=current;\
    }
#define macroArraysetterm(tyde,term) {\
       (tyde->Term)=term;\
    }
#define macroArraysetcurrent(tyde,term) {\
       (tyde->CurrentType)=term;\
    }
#define macroTyarraymatrixSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(ArrayMatrixTypeRepr,tyty))

static ExpcalCstList
ArrayMatrixGetDimLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroArraygetdimlst (macroTyarraymatrixSTree_getDescriptionr (TT)));
}

static TypeType
ArrayMatrixGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroArraygetterm (macroTyarraymatrixSTree_getDescriptionr (TT)));
}

static TypeType
ArrayMatrixGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroDegetcurrent ((macroTyarraymatrixSTree_getDescriptionr (TT))));
}

static BoolType
ArrayMatrixRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = ArrayMatrixGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  ArrayMatrixRecTestandGetCurrent (*RTT, RTT);
  return pk_true;
}

/* --------------------------------------------------------
   ClassTestArrayMatrix(TypeType *TT)
   ========================================================
   returns:
   True if TT is a pointer to an array or a matrix and then TT is this array
   ------------------------------------------------------------*/
static BoolType
ClassTestArrayMatrix (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestArrayMatrix (macroTygetclass (TT)));
}

static BoolType
TestArrayClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestArrayClass (macroTygetclass (TT)));
}

/* --------------------------------------------------------
   ArrayTypeNormalize(ExpcalCstList *DimLst, TypeType Term)
   ========================================================
   assumes: Term is an instantiated type (ie it can not be a generic
   type provided with a current distinct effective type)
   returns:
   In normal form, Term is not an array
   DimLst shall contain all dimensions, result is the element type
   ------------------------------------------------------------*/
static TypeType
ArrayNormalize (ExpcalCstList * DimLst, TypeType Term)
  /*---------------***********---------------------------------------*/
{
  if (TypeTestArrayClass (Term))
    {
      ExpcalCstList DimLst1;
      ArrayMatrixTypeRepr *Descr;
      Descr = macroTyarraymatrixSTree_getDescriptionr (Term);
      Term = macroArraygetterm (Descr);
      DimLst1 = macroArraygetdimlst (Descr);
      DimLst1 = ExpcalListConcat (*DimLst, DimLst1);
      *DimLst = DimLst1;
    }
  return Term;
}

/* --------------------------------------------------------
   ArrayMatrixCreate(TypeClass Cl,ExpcalCstList DimLst, TypeType Term)
   ========================================================
   assumes: T is not an Array if Cl is arrayclass
   returns:
   a new Array (resp. Matrix) Type is created: DimLst is its Dimension
   Term is the type of its elements
   ------------------------------------------------------------*/
static TypeType
ArrayMatrixCreate (TypeClass Cl, ExpcalCstList DimLst, TypeType Term)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  ArrayMatrixTypeRepr *Descr;
  Descr = (ArrayMatrixTypeRepr *) (malloc (sizeof (ArrayMatrixTypeRepr)));
  TT = NewTypeItem (Cl, Descr);
  AppendClassEntry (Cl, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroArrayset (Descr, DimLst, Term, TT) return TT;
}

/* --------------------------------------------------------
   ArrayFind(ExpcalCstList *RefDimLst, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no Array Type with normalized *RefDimLst and Term
   as dimensions and Type exists
   pk_true if some identical Array Type T exists 
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
ArrayFind (ExpcalCstList * RefDimLst, TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  ArrayMatrixTypeRepr *Descr;
  Term = ArrayNormalize (RefDimLst, Term);
  result = pk_false;
  ReprLst = TypeClassTable[ArrayClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (ArrayMatrixTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (ExpcalCstListEq (macroArraygetdimlst (Descr), *RefDimLst) && TypeEqAdr (macroArraygetterm (Descr), Term))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	};
    };
  return result;
}

/* --------------------------------------------------------
   MatrixFind(ExpcalCstList DimLst, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no Matrix Type with DimLst and Term as dimensions and Type
   exists
   pk_true if some identical Matrix Type T exists
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
MatrixFind (ExpcalCstList DimLst, TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  ArrayMatrixTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[MatrixClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (ArrayMatrixTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (ExpcalCstListEq (macroArraygetdimlst (Descr), DimLst) && TypeEqAdr (macroArraygetterm (Descr), Term))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	};
    };
  return result;
}

static void
ArrayMatrixSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  ArrayMatrixTypeRepr *Descr;
  Descr = (macroTyarraymatrixSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
ArrayReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT;
  ExpcalCstList DimLst;
  if (ArrayMatrixRecTestandGetCurrent (TT, &ATT))
    {
      if (ArrayReduction (ATT, RTT))
	ArrayMatrixSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      if (TypeReduction (ArrayMatrixGetTerm (TT), RTT))
	{
	  DimLst = ArrayMatrixGetDimLst (TT);
	  if (ArrayFind (&DimLst, *RTT, &ATT));
	  else
	    ATT = ArrayMatrixCreate (ArrayClass, DimLst, *RTT);
	  ArrayMatrixSetCurrent (TT, ATT);
	  *RTT = ATT;
	  return pk_true;
	}
      else
	return pk_false;
    }
}

static BoolType
MatrixReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT;
  ExpcalCstList DimLst;
  ATT = ArrayMatrixGetCurrent (TT);
  if (ArrayMatrixRecTestandGetCurrent (TT, &ATT))
    {
      if (TypeReduction (ATT, RTT))
	ArrayMatrixSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      if (TypeReduction (ArrayMatrixGetTerm (TT), RTT))
	{
	  DimLst = ArrayMatrixGetDimLst (TT);
	  if (MatrixFind (DimLst, *RTT, &TT));
	  else
	    ATT = ArrayMatrixCreate (MatrixClass, DimLst, *RTT);
	  ArrayMatrixSetCurrent (TT, ATT);
	  *RTT = ATT;
	  return pk_true;
	}
      else
	return pk_false;
    }
}
