/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |**						           ClassProduct_g.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**		           for Tuple Types                          **|
	  |**		               Structure Types                      **|
	  |**		               Bundle Types                         **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		           Tuple Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
TupleTypeCreate (TypeList TypeLst)
{
  TypeType TT;
  if (TupleFind (TypeLst, &TT));
  else
    TT = TupleCreate (TypeLst);
  return TT;
}

/* --------------------------------------------------------
   TypeTestTupleClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a tuple class
   ------------------------------------------------------------*/
extern BoolType
TypeTestTupleClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestTupleClass (macroTygetclass (TT)));
}

extern TypeList
TupleTypeGetTypeLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (TupleGetTypeLst (TT));
}

extern TypeType
TupleTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (TupleGetCurrent (TT));
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		       Structure Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
StructureTypeCreate (NameList NameLst, TypeType Typelst)
{
  TypeType TT;
  if (!StructureFind (NameLst, Typelst, &TT))
    TT = StructureCreate (NameLst, Typelst);
  return TT;
}

/* --------------------------------------------------------
   TypeTestStructureClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a structure class
   ------------------------------------------------------------*/
extern BoolType
TypeTestStructureClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestStructureClass (macroTygetclass (TT)));
}

extern TypeType
StructureTypeGetArgument (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a structure type */
{
  return (StructureGetArgument (TT));
}

extern TypeList
StructureTypeGetTypeLst (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a structure type */
{
  return (StructureGetTypeLst (TT));
}

extern TypeType
StructureTypeGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (StructureGetCurrent (TT));
}

extern NameList
StructureTypeGetNameLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (StructureGetNameLst (TT));
}




/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         Bundle Types                     **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
BundleTypeCreate (TypeType Term, ClkSystClass Clocks)
{
  TypeType TT;
  if (BundleFind (Term, Clocks, &TT));
  else
    TT = BundleCreate (Term, Clocks);
  return TT;
}

/* --------------------------------------------------------
   TypeTestBundleClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a bundle class
   ------------------------------------------------------------*/
extern BoolType
TypeTestBundleClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  return (macroCltestBundleClass (macroTygetclass (TT)));
}


extern TypeType
BundleTypeGetTerm (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a bundle type */
{
  return macroBundlegetfields (macroTybundleSTree_getDescriptionr (TT));
}

extern ClkSystClass
BundleTypeGetClkSyst (TypeType TT)
  /*---------------***********---------------------------------------
    TT is assumed to be a bundle type */
{
  return macroBundlegetclocks (macroTybundleSTree_getDescriptionr (TT));
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		          Product Types                   **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern BoolType
TypeTestProduct (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestProduct (macroTygetclass (TT)));
}
