/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |**						           ClassProduct_i.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**		           for Tuple Types                          **|
	  |**		               Structure Types                      **|
	  |**		               Bundle Types                         **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		           Tuple Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  TypeList TypeLst;
} TupleTypeRepr;
#define macroTuplegettypelst(tyde) (tyde->TypeLst)
#define macroTuplegetcurrent(tyde) (tyde->CurrentType)
#define macroTupleset(tyde,typelst,current) {\
       (tyde->TypeLst)=typelst;\
       (tyde->CurrentType)=current;\
    }
#define macroTuplesetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }

#define macroTytupleSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(TupleTypeRepr,tyty))

static TypeList
TupleGetTypeLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroTuplegettypelst (macroTytupleSTree_getDescriptionr (TT)));
}

static TypeType
TupleGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroTuplegetcurrent (macroTytupleSTree_getDescriptionr (TT)));
}

static BoolType
TupleRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = TupleGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  TupleRecTestandGetCurrent (*RTT, RTT);
  return pk_true;
}

/* --------------------------------------------------------
   TupleCreate(TypeList TypeLst)
   ========================================================
   returns:
   a new TupleType with elements TypeLst
   ------------------------------------------------------------*/
static TypeType
TupleCreate (TypeList TypeLst)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  TupleTypeRepr *Descr;
  Descr = (TupleTypeRepr *) (malloc (sizeof (TupleTypeRepr)));
  TT = NewTypeItem (TupleClass, Descr);
  AppendClassEntry (TupleClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroTupleset (Descr, TypeLst, TT) return TT;
}

/* --------------------------------------------------------
   TupleFind(TypeClass Type, TypeList TypeLst, TypeType *TT)
   ========================================================
   returns:
   false if no TupleType with elements TypeLst
   exists
   pk_true if some existing TupleType T has 
   elements TypeLst
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
TupleFind (TypeList TypeLst, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  TupleTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[TupleClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (TupleTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (TypeLstEq (macroTuplegettypelst (Descr), TypeLst))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	}
    }
  return result;
}

static void
TupleSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  TupleTypeRepr *Descr;
  Descr = (macroTytupleSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
TupleReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT;
  TypeList LTT;
  if (TupleRecTestandGetCurrent (TT, &ATT))
    {
      if (TypeReduction (ATT, RTT))
	TupleSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      if (TypeListReduction (TupleGetTypeLst (TT), &LTT))
	{
	  if (TupleFind (LTT, &ATT));
	  else
	    ATT = TupleCreate (LTT);
	  TupleSetCurrent (TT, ATT);
	  *RTT = ATT;
	  return pk_true;
	}
      else
	return pk_false;
    }
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		       Structure Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
  NameList NameLst;
  TypeType Argument;
} StructureTypeRepr;
#define macroStructuregetnamelst(tyde) (tyde->NameLst)
#define macroStructuregetargument(tyde) (tyde->Argument)
#define macroStructuregetcurrent(tyde) (tyde->CurrentType)
#define macroStructureset(tyde,namelst,argument,current) {\
       (tyde->NameLst)=namelst;\
       (tyde->Argument)=argument;\
       (tyde->CurrentType)=current;\
    }

#define macroStructuresetargument(tyde,argument) {\
       (tyde->Argument)=argument;\
    }

#define macroStructuresetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }

#define macroTystructureSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(StructureTypeRepr,tyty))

static NameList
StructureGetNameLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroStructuregetnamelst (macroTystructureSTree_getDescriptionr (TT)));
}

static TypeType
StructureGetArgument (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroStructuregetargument (macroTystructureSTree_getDescriptionr (TT)));
}

static TypeType
StructureGetCurrent (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroDegetcurrent (macroTystructureSTree_getDescriptionr (TT)));
}

static BoolType
StructureRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = StructureGetCurrent (TT);
  if (*RTT == TT)
    return pk_false;
  StructureRecTestandGetCurrent (*RTT, RTT);
  return pk_true;
}

/* --------------------------------------------------------
   StructureCreate(NameList NameLst, TypeType Typelst)
   ========================================================
   returns:
   a new StructureType which field names and types
   are one to one associated in NameLst and list of types 
   in the TupleType Typelst
   ------------------------------------------------------------*/
static TypeType
StructureCreate (NameList NameLst, TypeType Typelst)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  StructureTypeRepr *Descr;
  Descr = (StructureTypeRepr *) (malloc (sizeof (StructureTypeRepr)));
  TT = NewTypeItem (StructureClass, Descr);
  AppendClassEntry (StructureClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroStructureset (Descr, NameLst, Typelst, TT) return TT;
}

/* --------------------------------------------------------
   StructureFind((NameList NameLst, TypeType Typelst, TypeType *TT)
   ========================================================
   returns:
   false if no StructureType which field names and types
   are one to one associated as they are in NameLst and list of types 
   (in the TupleType Typelst)
   exists
   pk_true if such a StructureType T exist
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
StructureFind (NameList NameLst, TypeType Typelst, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  StructureTypeRepr *Descr;
  TypeType RefTuple;
  result = pk_false;
  ReprLst = TypeClassTable[StructureClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (StructureTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      RefTuple = macroStructuregetargument (Descr);
      if (NameLstEq (macroStructuregetnamelst (Descr), NameLst) && TypeEqAdr (RefTuple, Typelst))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	};
    };
  return result;
}

static TypeList
StructureGetTypeLst (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a structure type */
{
  return (TupleGetTypeLst (StructureGetArgument (TT)));
}

static void
StructureSetCurrent (TypeType TT, TypeType TE)
  /*---------------***********---------------------------------------*/
{
  StructureTypeRepr *Descr;
  Descr = (macroTystructureSTree_getDescriptionr (TT));
  macroDesetcurrent (Descr, TE);
}

static BoolType
StructureReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  TypeType ATT;
  NameList NameLst;
  if (StructureRecTestandGetCurrent (TT, &ATT))
    {
      if (StructureReduction (ATT, RTT))
	StructureSetCurrent (TT, *RTT);
      return pk_true;
    }
  else
    {
      if (TupleReduction (StructureGetArgument (TT), RTT))
	{
	  NameLst = StructureGetNameLst (TT);
	  if (StructureFind (NameLst, *RTT, &ATT));
	  else
	    ATT = StructureCreate (NameLst, *RTT);
	  StructureSetCurrent (TT, ATT);
	  *RTT = ATT;
	  return pk_true;
	}
      else
	return pk_false;
    }
}





/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         Bundle Types                     **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  TypeType Argument;
  ClkSystClass Clocks;
  TypeType CurrentType;
} BundleTypeRepr;
#define macroBundlegetfields(tyde) (tyde->Argument)
#define macroBundlegetcurrent(tyde) (tyde->CurrentType)
#define macroBundlegetclocks(tyde) (tyde->Clocks)
#define macroBundleset(tyde,argument,clocks,current) {\
       (tyde->Argument)=argument;\
       (tyde->Clocks)=clocks;\
       (tyde->CurrentType)=current;\
    }
#define macroBundlesetfields(tyde,argument) {\
       (tyde->Argument)=argument;\
    }
#define macroBundlesetcurrent(tyde,current) {\
       (tyde->CurrentType)=current;\
    }

#define macroTybundleSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(BundleTypeRepr,tyty))


/* --------------------------------------------------------
   BundleCreate(TypeType Term, ClkSystClass Clocks)
   ========================================================
   returns:
   a new BundleType which Term type and Clocks as operands.
   ------------------------------------------------------------*/
static TypeType
BundleCreate (TypeType Term, ClkSystClass Clocks)
  /*---------------***********---------------------------------------*/
{
  /* Implemented by Lbesnard (incomplete) */
  /* Common Part */
  TypeType TT;
  BundleTypeRepr *Descr;
  Descr = (BundleTypeRepr *) (malloc (sizeof (BundleTypeRepr)));
  TT = NewTypeItem (BundleClass, Descr);
  AppendClassEntry (BundleClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroBundleset (Descr, Term, Clocks, TT) return TT;
}


/* --------------------------------------------------------
   BundleFind(TypeType Term, ClkSystClass Clocks, TypeType *TT))
   ========================================================
   returns:
   false if no BundleType with Term and Clocks as operands.
   pk_true if such a BundleType T exists
   ------------------------------------------------------------*/
static BoolType
BundleFind (TypeType Term, ClkSystClass Clocks, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  TypeReprSubList ReprLst;
  BundleTypeRepr *Descr;

  /* Implemented by Lbesnard (incomplete) */
  ReprLst = TypeClassTable[BundleClass].First;
  while (!NullPointerDef (ReprLst))
    {
      Descr = (BundleTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if ((macroBundlegetfields (Descr) == Term) && (macroBundlegetclocks (Descr) == Clocks))
	{
	  *TT = macroDegetmyself (Descr);
	  return pk_true;
	}
    }
  return pk_false;
}
