/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		          Extern Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
ExternTypeCreate (NameClass Name)
{
  TypeType TT;
  if (ExternFind (Name, &TT));
  else
    TT = ExternCreate (Name);
  return TT;
}

extern BoolType
TypeTestExternClass (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroCltestExternClass (macroTygetclass (TT)));
}

extern NameClass
ExternTypeGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (ExternGetName (TT));
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     Enumeration types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
EnumerationTypeCreate (NameClass Name, NameList NameLst)
{
  TypeType TT;
  if (EnumerationFind (Name, NameLst, &TT));
  else
    TT = EnumerationCreate (Name, NameLst);
  return TT;
}

extern BoolType
TypeTestEnumerationClass (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  return (macroCltestEnumerationClass (macroTygetclass (TT)));
}

extern NameClass
EnumerationTypeGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (EnumerationGetName (TT));
}

extern NameList
EnumerationTypeGetNameLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (EnumerationGetNameLst (TT));
}

extern BoolType
EnumTypeOrdinal (TypeType TT, NameClass Name, int *I)
{
  EnumerationTypeRepr *Descr;
  NameList NameLst;
  Descr = macroTyEnumerationSTree_getDescriptionr (TT);
  NameLst = macroEnumerationgetnamelst (Descr);
  return (NameListNameOrdinal (NameLst, Name, I));
}

/* return pk_true if the enumeration type (undefined name, NameLst) is in the type manager */
extern BoolType
EnumerationExist (NameList NameLst, TypeType * TT)
{
  return EnumerationFind2 (NameLst, TT);
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     Subrange types                       **|
	  |**		                       		          **|
	  |*********|*********|*********|*********|*********|*********/

extern TypeType
SubrangeTypeCreate (ExpcalCst Bounds, TypeType Term)
{
  TypeType TT;
  if (SubrangeFind (Bounds, Term, &TT));
  else
    TT = SubrangeCreate (Bounds, Term);
  return TT;
}

/* --------------------------------------------------------
   TypeTestSubrangeClass(TypeType TT)
   ========================================================
   returns:
   True if TT is a Subrange class
   ------------------------------------------------------------*/
extern BoolType
TypeTestSubrangeClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestSubrangeClass (macroTygetclass (TT)));
}

extern TypeType
SubrangeTypeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (SubrangeGetTerm (TT));
}

extern ExpcalCst
SubrangeTypeGetBounds (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (SubrangeGetBounds (TT));
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     Simple Type                          **|
	  |**		                       		          **|
	  |*********|*********|*********|*********|*********|*********/

extern BoolType
TypeTestSimple (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_true;
  else
    return (macroCltestSimple (macroTygetclass (TT)));
}
