/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |**                                                    File ClassSimple_i.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**		           for Extern types                         **|
	  |**		               Enumeration types                    **|
	  |**		               Subrange types                       **||
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		          Extern Types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  NameClass Name;
} ExternTypeRepr;
#define macroExterngetname(tyde) (tyde->Name)
#define macroExternset(tyde,name) {\
       (tyde->Name)=name;\
    }
#define  macroTyExternSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(ExternTypeRepr,tyty))

static NameClass
ExternGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroExterngetname (macroTyExternSTree_getDescriptionr (TT)));
}

/* --------------------------------------------------------
   ExternCreate(NameClass Name)
   ========================================================
   returns:
   a new ExternType named Name
*/
static TypeType
ExternCreate (NameClass Name)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  ExternTypeRepr *Descr;
  Descr = (ExternTypeRepr *) (malloc (sizeof (ExternTypeRepr)));
  TT = NewTypeItem (ExternClass, Descr);
  AppendClassEntry (ExternClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroExternset (Descr, Name) return TT;
}

/* --------------------------------------------------------
   ExternFind(NameClass Name, TypeType *TT)
   ========================================================
   returns:
   false if no ExternType with name Name exist
   pk_true if some existing ExternType T is named Name 
   T is assigned to TT
*/
static BoolType
ExternFind (NameClass Name, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  ExternTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[ExternClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (ExternTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (NameEq (macroExterngetname (Descr), Name))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	}
    };
  return result;
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     Enumeration types                    **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  NameClass Name;
  NameList NameLst;
} EnumerationTypeRepr;
#define macroEnumerationgetname(tyde) (tyde->Name)
#define macroEnumerationgetnamelst(tyde) (tyde->NameLst)
#define macroEnumerationset(tyde,name,namelst) {\
       (tyde->Name)=name;\
       (tyde->NameLst)=namelst;\
    }

#define  macroTyEnumerationSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(EnumerationTypeRepr,tyty))

static NameClass
EnumerationGetName (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroEnumerationgetname (macroTyEnumerationSTree_getDescriptionr (TT)));
}

static NameList
EnumerationGetNameLst (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroEnumerationgetnamelst (macroTyEnumerationSTree_getDescriptionr (TT)));
}

/* --------------------------------------------------------
   EnumerationCreate(NameClass Name, NameList NameLst)
   ========================================================
   returns:
   a new EnumerationType named Name with elements NameLst
   ------------------------------------------------------------*/
static TypeType
EnumerationCreate (NameClass Name, NameList NameLst)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  EnumerationTypeRepr *Descr;
  Descr = (EnumerationTypeRepr *) (malloc (sizeof (EnumerationTypeRepr)));
  TT = NewTypeItem (EnumerationClass, Descr);
  AppendClassEntry (EnumerationClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroEnumerationset (Descr, Name, NameLst) return TT;
}

/* --------------------------------------------------------
   EnumerationFind(NameClass Name, NameList NameLst, TypeType *TT)
   ========================================================
   returns:
   false if no EnumerationType with (?? name Name and ??) elements NameLst
   exists
   pk_true if some existing EnumerationType T has 
   (?? name Name and ??) elements NameLst
   T is assigned to TT
   ------------------------------------------------------------*/
static BoolType
EnumerationFind (NameClass Name, NameList NameLst, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  TypeReprSubList ReprLst;
  EnumerationTypeRepr *Descr;

  ReprLst = TypeClassTable[EnumerationClass].First;
  while (!NullPointerDef (ReprLst))
    {
      Descr = (EnumerationTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (NameEq (macroEnumerationgetname (Descr), Name) && NameLstEq (macroEnumerationgetnamelst (Descr), NameLst))
	{
	  *TT = macroDegetmyself (Descr);
	  return pk_true;
	}
    }
  return pk_false;
}


static BoolType
EnumerationFind2 (NameList NameLst, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  TypeReprSubList ReprLst;
  EnumerationTypeRepr *Descr;

  ReprLst = TypeClassTable[EnumerationClass].First;
  while (!NullPointerDef (ReprLst))
    {
      Descr = (EnumerationTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (NameLstEq (macroEnumerationgetnamelst (Descr), NameLst))
	{
	  *TT = macroDegetmyself (Descr);
	  return pk_true;
	}
    }
  return pk_false;
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		     Subrange types                       **|
	  |**		                       		          **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct
{
  TypeType WhoamI;
  ExpcalCst Bounds;
  TypeType Term;
} SubrangeTypeRepr;
#define macroSubrangegetbounds(tyde) (tyde->Bounds)
#define macroSubrangegetterm(tyde) (tyde->Term)
#define macroSubrangeset(tyde,bounds,term) {\
       (tyde->Bounds)=bounds;\
       (tyde->Term)=term;\
    }
#define macroSubrangesetterm(tyde,term) (tyde->Term)=term;

#define  macroTySubrangeSTree_getDescriptionr(tyty) (macroTySTree_getDescriptionr(SubrangeTypeRepr,tyty))

static TypeType
SubrangeGetTerm (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroSubrangegetterm (macroTySubrangeSTree_getDescriptionr (TT)));
}

static ExpcalCst
SubrangeGetBounds (TypeType TT)
  /*---------------***********---------------------------------------*/
{
  return (macroSubrangegetbounds (macroTySubrangeSTree_getDescriptionr (TT)));
}

/* --------------------------------------------------------
   SubrangeCreate(ExpcalCst Bounds , TypeType Term)
   ========================================================
   returns:
   a new SubrangeType with parameters VariableLst and term Term
*/
static TypeType
SubrangeCreate (ExpcalCst Bounds, TypeType Term)
  /*---------------***********---------------------------------------*/
{
  /* Common Part */
  TypeType TT;
  SubrangeTypeRepr *Descr;
  Descr = (SubrangeTypeRepr *) (malloc (sizeof (SubrangeTypeRepr)));
  TT = NewTypeItem (SubrangeClass, Descr);
  AppendClassEntry (SubrangeClass, Descr);
  macroDesetmyself (Descr, TT);
  /* Specific Part */
  macroSubrangeset (Descr, Bounds, Term) return TT;
}

/* --------------------------------------------------------
   SubrangeFind(ExpcalCst Bounds, TypeType Term, TypeType *TT)
   ========================================================
   returns:
   false if no SubrangeType with parameters VariableLst and term Term
   exists
   pk_true if some existing SubrangeType T has 
   parameters VariableLst and term Term
   T is assigned to TT
*/
static BoolType
SubrangeFind (ExpcalCst Bounds, TypeType Term, TypeType * TT)
  /*---------------***********---------------------------------------*/
{
  BoolType result;
  TypeReprSubList ReprLst;
  SubrangeTypeRepr *Descr;
  result = pk_false;
  ReprLst = TypeClassTable[SubrangeClass].First;
  while (!((result == pk_true) || (NullPointerDef (ReprLst))))
    {
      Descr = (SubrangeTypeRepr *) (GetCar (ReprLst));
      ReprLst = GetCdr (ReprLst);
      if (ExpcalCstEq (macroSubrangegetbounds (Descr), Bounds) && (TypeEqAdr (macroSubrangegetterm (Descr), Term)))
	{
	  result = pk_true;
	  *TT = macroDegetmyself (Descr);
	}
    }
  return result;
}
