/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/* --------------------------------------------------------
   NameList NameListCreate()
   ========================================================
   returns:
   a new empty NameList
*/
extern NameList
NameListCreate ()
{
  NameList Result;
  Result = macroNameListnew;
  macroNameListsetempty (Result) return Result;

}

/* --------------------------------------------------------
   NameListTestEmpty(NameList NameLst)
   ========================================================
   returns:
   pk_true if NameLst is empty
   false otherwise 
*/
extern BoolType
NameListTestEmpty (NameList NameLst)
{
  return (macroNameListtestempty (NameLst));
}

/* --------------------------------------------------------
   NameListCast2SubList(NameList NameLst)
   ========================================================
   returns:
   The sublist contained in NameLst
*/
extern NameSubList
NameListCast2SubList (NameList NameLst)
{
  return (macroNameListgetfirst (NameLst));
}

/* --------------------------------------------------------
   NameSubListGetHead(NameSubList NameLst)
   ========================================================
   returns:
   The first element of NameLst
*/
extern NameClass
NameSubListGetHead (NameSubList NameLst)
{
  return (macroNameSubListgethead (NameLst));
}


/* --------------------------------------------------------
   NameSubListGetTail(NameSubList NameLst)
   ========================================================
   returns:
   The sublist resulting from removing the first element in NameLst
*/
extern NameSubList
NameSubListGetTail (NameSubList NameLst)
{
  return (macroNameSubListgettail (NameLst));
}

/* --------------------------------------------------------
   NameSubListTestLast(NameSubList NameLst)
   ========================================================
   returns:
   pk_true if tail of NameLst is empty
*/
extern BoolType
NameSubListTestLast (NameSubList NameLst)
{
  return (macroNameSubListtestlast (NameLst));
}


/* --------------------------------------------------------
   NameListTestandGetOrdinal(NameList NameLst, int Ordinal, NameClass *ResName)
   ========================================================
   Assumes: Ordinal >= 0
   returns:
   pk_true if NameLst has more than Ordinal Elements
   and the corresponding name in ResName
   (first name has ordinal 0)
*/
extern BoolType
NameListTestandGetOrdinal (NameList NameLst, int Ordinal, NameClass * ResName)
{
  if (macroNameListtestempty (NameLst))
    return pk_false;
  else
    {
      NameSubList SubLst;
      SubLst = macroNameListgetfirst (NameLst);
      do
	{
	  if (Ordinal == 0)
	    {
	      *ResName = (macroNameSubListgethead (SubLst));
	      return pk_true;
	    }
	  else if (macroNameSubListtestlast (SubLst))
	    return pk_false;
	  SubLst = macroNameSubListgettail (SubLst);
	  Ordinal = Ordinal - 1;
	}
      while (pk_true);
    }
}


/* --------------------------------------------------------
   NameLstCompare(NameList NameLst1, NameList NameLst2)
   ========================================================
   returns:
   -1 if NameLst1 is smaller than NameLst2 for lexical comparison
   0 if NameLst1 and NameLst2 are identical
   1 if NameLst2 is smaller than NameLst1 for lexical comparison
*/
extern int
NameLstCompare (NameList NameLst1, NameList NameLst2)
{
  if (NameLst1 == NameLst2)
    return 0;
  else
    {
      if (macroNameListtestempty (NameLst1))
	if (macroNameListtestempty (NameLst2))
	  return 0;
	else
	  return -1;
      else if (macroNameListtestempty (NameLst2))
	return 1;
      else
	{
	  NameSubList SubLst1, SubLst2;
	  int result, goon;
	  SubLst1 = macroNameListgetfirst (NameLst1);
	  SubLst2 = macroNameListgetfirst (NameLst2);
	  result = 0;
	  goon = 2;
	  do
	    {
	      result = (NameCompare (macroNameSubListgethead (SubLst1), macroNameSubListgethead (SubLst2)));
	      if (result == 0)
		{
		  goon = NameSubLstFirstEnding (SubLst1, SubLst2);
		  if (goon == 2)
		    {
		      SubLst1 = macroNameSubListgettail (SubLst1);
		      SubLst2 = macroNameSubListgettail (SubLst2);
		    }
		  else
		    result = goon;
		}
	    }
	  while (result == 0 && goon == 2);
	  return (result);
	}
    }
}


/* --------------------------------------------------------
   NameLstEq(NameList NameLst1, NameList NameLst2)
   ========================================================
   returns:
   false if NameLst1 and NameLst2 are different
   pk_true if NameLst1 and NameLst2 are identical
*/
extern BoolType
NameLstEq (NameList NameLst1, NameList NameLst2)
{
  if (NameLstCompare (NameLst1, NameLst2) == 0)
    return pk_true;
  else
    return pk_false;
}

/* --------------------------------------------------------
   NameListNameOrdinal(NameList NameLst, NameClass Elem, int *Result)
   ========================================================
   returns:
   pk_true if Elem is a Name of NameLst
   then its ordinal is assigned to Result
   false if Elem is not a Name of NameLst
*/
extern BoolType
NameListNameOrdinal (NameList NameLst, NameClass Elem, int *Result)
{
  if (macroNameListtestempty (NameLst))
    return pk_false;
  else
    {
      int goon;
      NameSubList SubLst;
      *Result = 0;
      SubLst = macroNameListgetfirst (NameLst);
      do
	{
	  goon = (NameCompare (macroNameSubListgethead (SubLst), Elem));
	  if (goon == 0)
	    return pk_true;
	  else
	    {
	      if (macroNameSubListtestlast (SubLst))
		return pk_false;
	      else
		{
		  *Result = (*Result) + 1;
		  SubLst = macroNameSubListgettail (SubLst);
		}
	    }
	}
      while (pk_true);
    }
}


/* --------------------------------------------------------
   NameListAppendName(NameList NameLst, NameClass Elem)
   ========================================================
   side effect:
   Elem is append at the end of NameLst
*/
extern void
NameListAppendName (NameList NameLst, NameClass Elem)
{
  NameSubList RefCadR;
  RefCadR = macroNameSubListnew;
  macroNameSubListsethead (RefCadR, Elem) macroNameSubListsetlast (RefCadR);
  if (macroNameListtestempty (NameLst))
    {
      macroNameListsetfirst (NameLst, RefCadR);
    }
  else
    macroNameSubListsettail ((macroNameListgetlast (NameLst)), RefCadR);
  macroNameListsetlast (NameLst, RefCadR);
}

/* --------------------------------------------------------
   NameListInsertheadingName(NameList NameLst, NameClass Elem)
   ========================================================
   side effect:
   Elem is append at the beginning of NameLst
*/
extern void
NameListInsertheadingName (NameList NameLst, NameClass Elem)
{
  NameSubList RefCadR;
  RefCadR = macroNameSubListnew;
  macroNameSubListsethead (RefCadR, Elem) if (macroNameListtestempty (NameLst))
    {
      macroNameSubListsetlast (RefCadR);
    }
  else
    macroNameSubListsettail (RefCadR, (macroNameListgetfirst (NameLst)));
  macroNameListsetfirst (NameLst, RefCadR);
}

/* --------------------------------------------------------
   NameListRegister(NameList NameLst)
   ========================================================
   side effect:
   The ordered list of NameList is created
*/

extern NameList
NameListRegister (NameList NameLst)
{
  if (macroNameLstIndextestempty)
    {
      macroNameLstIndexsethead (NameLst);
      return NameLst;
    }
  else
    {
      int rescomp;
      NameDictEntry *NameLstNext;
      NameDictEntry *NameLstCurrent;
      NameLstCurrent = NameLstIndex;
      rescomp = NameLstCompare (NameLst, macroNameLstDictEntrygethead (NameLstCurrent));
      if (rescomp == -1)
	{
	  NameLstIndex = NameLstDictEntryNew (NameLst, NameLstCurrent);
	  return NameLst;
	};
      if (rescomp == 0)
	return (macroNameLstDictEntrygethead (NameLstCurrent));
      while (!(macroNameLstDictEntrytestlast (NameLstCurrent)))
	{
	  /* rescomp == 1: NameLst >= head(NameLstCurrent) */
	  NameLstNext = macroNameLstDictEntrygettail (NameLstCurrent);
	  rescomp = NameLstCompare (NameLst, macroNameLstDictEntrygethead (NameLstNext));
	  if (rescomp == -1)
	    {
	      NameLstNext = NameLstDictEntryNew (NameLst, macroNameLstDictEntrygettail (NameLstCurrent));
	      macroNameLstDictEntrysettail (NameLstCurrent, NameLstNext);
	      return NameLst;
	    };
	  if (rescomp == 0)
	    return (macroNameLstDictEntrygethead (NameLstNext));
	  NameLstCurrent = NameLstNext;
	};
      NameLstNext = NameLstDictEntryNew (NameLst, macroNameLstDictEntrygettail (NameLstCurrent));
      macroNameLstDictEntrysettail (NameLstCurrent, NameLstNext);
      return NameLst;
    }
}

/* --------------------------------------------------------
   NameListUnsortedCompare(NameList NameLst1, NameList NameLst2)
   ========================================================
   returns:
   -2 if NameLst1 and NameLst2 has no common element otherwise
   -1 if NameLst1 is smaller than NameLst2 for inclusion
   0 if NameLst1 and NameLst2 are identical
   1 if NameLst1 and NameLst2 has some common element
*/
extern int
NameListUnsortedCompare (NameList NameLst1, NameList NameLst2)
{
  NameSubList SubLst1;
  int result;
  int minres;
  if (macroNameListtestempty (NameLst1))
    return -2;
  if (macroNameListtestempty (NameLst2))
    return -2;
  if (NameLst1 == NameLst2)
    return 0;
  SubLst1 = macroNameListgetfirst (NameLst1);
  if (NameListNameOrdinal (NameLst2, macroNameSubListgethead (SubLst1), &result))
    minres = -1;
  else
    minres = -2;
  while ((!(macroNameSubListtestlast (SubLst1))) && (!(minres == 1)))
    {
      SubLst1 = macroNameSubListgettail (SubLst1);
      if (NameListNameOrdinal (NameLst2, macroNameSubListgethead (SubLst1), &result))
	{
	  if (minres == -2)
	    minres = 1;
	}
      else if (minres == -1)
	minres = 1;
    };
  return minres;
}



extern BoolType
NameListChangeName (NameList NameLst, NameClass Elem, NameClass NewElem)
{
  int goon;
  NameSubList SubLst;

  if (macroNameListtestempty (NameLst))
    return pk_false;
  if (!NameListNameOrdinal (NameLst, Elem, &goon))
    return pk_false;
  SubLst = macroNameListgetfirst (NameLst);
  while (pk_true)
    {
      goon = (NameCompare (macroNameSubListgethead (SubLst), Elem));
      if (goon == 0)
	{
	  macroNameSubListsethead (SubLst, NewElem);
	  return pk_true;
	}
      if (macroNameSubListtestlast (SubLst))
	return pk_false;
      SubLst = macroNameSubListgettail (SubLst);
    }
}



/* --------------------------------------------------------
   NameListInitiate()
   ========================================================
   side effect:
   The ordered list of NameList is created
*/

extern void
NameListInitiate ()
{
  NameLstIndex = NameLstDictEntryNew (EmptyNameList, LastNameDictEntry);
}
