/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/***
    |**
    |**                                                  File NameListManager_i.c
    |**						       	            **|
    |**						       	            **|
    |**						       	            **|
    |**	      	THIS FILE CONTAINS Name List management             **|
    |**		           LOCAL FUNCTIONS                          **|
    |**								    **|
    |*********|*********|*********|*********|*********|*********|*********/

/*===========================================================
  Name List Structure Types and macros
  ======================================================== */

struct LNameCell
{
  struct LNameCell *CdR;
  NameClass CaR;
};
struct LNameDescr
{
  NameSubList First;
  NameSubList Last;
};

#define macroNameSubListnew ((NameSubList)(malloc(sizeof(NameListCell))))
#define macroNameSubListtestlast(sublstarg) (sublstarg->CdR == ((NameSubList)0))
#define macroNameSubListgethead(sublstarg) (sublstarg->CaR)
#define macroNameSubListgettail(sublstarg) (sublstarg->CdR)
#define macroNameSubListsethead(sublstarg, Elem) (sublstarg->CaR)=Elem;
#define macroNameSubListsettail(sublstarg, Tail) (sublstarg->CdR)=Tail;
#define macroNameSubListsetlast(sublstarg) (sublstarg->CdR)=((NameSubList)0);

#define macroNameListnew ((NameList)(malloc(sizeof(NameListDescr))))
#define macroNameListtestempty(ListArg)  (ListArg->First == ((NameSubList)0))
#define macroNameListgetfirst(ListArg)  (ListArg->First)
#define macroNameListgetlast(ListArg)  (ListArg->Last)
#define macroNameListsetempty(ListArg) {\
  (ListArg->First = ((NameSubList)0));\
  (ListArg->Last = ((NameSubList)0));\
}
#define macroNameListsetfirst(ListArg,SubList)  (ListArg->First) = SubList;
#define macroNameListsetlast(ListArg,SubList)  (ListArg->Last) = SubList;

/*===========================================================
  Name List Local Functions
  ======================================================== */

/* --------------------------------------------------------
   NameSubLstFirstEnding(NameSubList SubLst1, NameSubList SubLst2)
   =====================
   returns:
   -1 if SubLst1 terminates a list and SubLst2 does not
   0 if SubLst1 and SubLst2 both terminate a list
   1 if SubLst2 terminates a list and SubLst1 does not
   2 if neither SubLst1 nor SubLst2 terminate a list
*/
static int
NameSubLstFirstEnding (NameSubList SubLst1, NameSubList SubLst2)
{
  if (macroNameSubListtestlast (SubLst1))
    if (macroNameSubListtestlast (SubLst2))
      return 0;
    else
      return -1;
  else if (macroNameSubListtestlast (SubLst2))
    return 1;
  else
    return 2;
}

/*===========================================================
  Name List Dictionary and macros
  ======================================================== */

typedef struct LNameDictEntry
{
  struct LNameDictEntry *CdR;
  NameList CaR;
} NameDictEntry;
#define LastNameDictEntry  ((NameDictEntry *) 0)

#define macroNameDictEntrynew (NameDictEntry *) malloc (sizeof (NameDictEntry));
#define macroNameLstDictEntrytestlast(RefDictEntry)\
     (RefDictEntry->CdR == LastNameDictEntry)
#define macroNameLstDictEntrygethead(RefDictEntry) (RefDictEntry->CaR)
#define macroNameLstDictEntrygettail(RefDictEntry) (RefDictEntry->CdR)
#define macroNameLstDictEntrysettail(RefDictEntry,Tail) (RefDictEntry->CdR)=Tail;
#define macroNameLstDictEntryset(RefDictEntry,Elem,Tail) {\
(RefDictEntry->CaR)=Elem;\
(RefDictEntry->CdR)=Tail;\
  };

NameDictEntry *NameLstIndex;

#define macroNameLstIndextestempty (NameLstIndex->CaR == (EmptyNameList))
#define macroNameLstIndexsethead(Elem) (NameLstIndex->CaR)=Elem;

/* --------------------------------------------------------
   NameList NameListDictEntryNew(NameList Elem, NameDictEntry *Tail)
   ========================================================
   returns:
   a new name list dictionary entry, followed by Tail, with Elem as content
*/
static NameDictEntry *
NameLstDictEntryNew (NameList Elem, NameDictEntry * Tail)
{
  NameDictEntry *Result;
  Result = macroNameDictEntrynew;
  macroNameLstDictEntryset (Result, Elem, Tail);
  return Result;
}

/*===========================================================
  Name Functions
  ======================================================== */

/* --------------------------------------------------------
   NameCompare(NameClass Name1, NameClass Name2)
   ========================================================
   returns:
   -1 if Name1 is smaller than Name2 for lexical comparison
   0 if Name1 and Name2 are identical
   1 if Name2 is smaller than Name1 for lexical comparison
*/
int
NameCompare (NameClass Name1, NameClass Name2)
{
  /*  return strcmp(Name1,Name2); */
  return strcompar (Name1, Name2);
}

/* --------------------------------------------------------
   NameEq(NameClass Name1, NameClass Name2)
   ========================================================
   returns:
   false if Name1 and Name2 are different
   pk_true if Name1 and Name2 are identical
*/
BoolType
NameEq (NameClass Name1, NameClass Name2)
{

  return streq (Name1, Name2);
  /* if (strcmp(Name1,Name2) == pk_true) return pk_true;  else return pk_false; */
}
