/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/***
    |**
    |**                                                    File TypeGarbage_g.c
    |**						       	            **|
    |**						       	            **|
    |**						       	            **|
    |**	            THIS FILE CONTAINS Type Garbage                 **|
    |**		           GLOBAL FUNCTIONS                         **|
    |**								    **|
    |*********|*********|*********|*********|*********|*********|*********/

/* Il y a un problème avec les valeurs courantes et les get
   qui vont les chercher automatiquement. C'est peut-être ce qu'il faut
   faire mais dans ce cas les autres composants et le type lui-même
   ne doivent pas nécessairement être marqués.
   Ou bien on marque tout sans se poser de questions.
   Ce n'est sans doute pas le seul pb du garbage
*/
extern void
RecursiveMarkTypeTerm (TypeType TT)
  /*--************------------------------
    Mark the entry corresponding to the user defined type TT and recursively
    its components.
    This procedure has important side effect and cannot be used without a
    following CleanTypeTable
  */
{
  if (!(macroTytestpredef (TT)))
    if (!(macroTytestmarked (TT)))
      {
	macroTysetmarked (TT);	/* set the type as marked */
	switch (macroTygetunmarkedclass (TT))
	  {
	  case ExternClass:
	  case EnumerationClass:
	  case SubrangeClass:
	  case BitsClass:
	    break;
	  case ArrayClass:
	  case MatrixClass:
	    RecursiveMarkTypeTerm (ArrayMatrixGetTerm (TT));
	    RecursiveMarkTypeTerm (ArrayMatrixGetCurrent (TT));
	    break;
	  case NamedClass:
	    RecursiveMarkTypeTerm (NamedGetTerm (TT));
	    RecursiveMarkTypeTerm (MetaNamedGetCurrent (TT));
	    break;
	  case LambdaClass:
	    RecursiveMarkTypeTerm (LambdaGetTerm (TT));
	    RecursiveMarkTypeTerm (LambdaGetCurrent (TT));
	    break;
	  case AnyClass:
	    if (AnyTypeTestDefined (TT))
	      RecursiveMarkTypeTerm (AnyGetDefinition (TT));
	    break;
	  case MetaClass:
	    if (MetaTypeTestDefined (TT))
	      RecursiveMarkTypeTerm (MetaNamedGetCurrent (TT));
	    break;
	  case TupleClass:
	    MarkTypeList (TupleGetTypeLst (TT));
	    RecursiveMarkTypeTerm (TupleGetCurrent (TT));
	    break;
	  case StructureClass:
	    RecursiveMarkTypeTerm (StructureGetArgument (TT));
	    RecursiveMarkTypeTerm (StructureGetCurrent (TT));
	    break;
	  case BundleClass:
	    NotImplemented ("Structure marking not implemented. Returns -1");
	    break;
	  case NodeClass:
	    RecursiveMarkTypeTerm (NodeGetDomain (TT));
	    RecursiveMarkTypeTerm (NodeGetImage (TT));
	    RecursiveMarkTypeTerm (NodeGetCurrent (TT));
	    break;
	  case GenericNodeClass:
	    RecursiveMarkTypeTerm (GenericNodeGetArgument (TT));
	    RecursiveMarkTypeTerm (GenericNodeGetTerm (TT));
	    RecursiveMarkTypeTerm (GenericNodeGetCurrent (TT));
	    break;
	  case MetaNodeClass:
	    RecursiveMarkTypeTerm (MetaNodeGetArgument (TT));
	    RecursiveMarkTypeTerm (MetaNodeGetTerm (TT));
	    RecursiveMarkTypeTerm (MetaNodeGetCurrent (TT));
	    break;
	  }
      }
}

extern void
TypeGTSetFutureIndex (int *Max, int *AvailableIndex)
{
  *Max = macroTTDgetnext;
  *AvailableIndex = 0;
  SetFutureWhoamIType (AvailableIndex, *Max);
}

extern TypeType
TypeGetFutureIndex (TypeType TT)
{
  TypeType Future;
  Future = macroDegetmyself (macroTySTree_getDescriptionr (AnonymousTypeRepr, TT));
  if (macrotestmarkeduserType (Future))
    return (macrogetunmarkType (Future));
  else
    return (Future);
}

extern void
TypeGarbageClean (int Max, int AvailableIndex)
{
  TypeGarbageLocal (Max);
  TypeListCleanRegistered (AvailableIndex);
}

extern void
TypeGarbage ()
{
  int max;
  int AvailableIndex;
  TypeGTSetFutureIndex (&max, &AvailableIndex);
  TypeGarbageClean (max, AvailableIndex);
  TypeManagerReset ();
}
