/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* --------------------------------------------------------
   TypeList TypeListCreate()
   ========================================================
   returns:
   a new empty TypeList
*/
extern TypeList
TypeListCreate ()
{
  return (ListofTypeCreate ());
}

/* --------------------------------------------------------
   TypeListTestEmpty(TypeList TypeLst)
   ========================================================
   returns:
   pk_true if TypeLst is empty
   false otherwise 
*/
extern BoolType
TypeListTestEmpty (TypeList TypeLst)
{
  return (macroTypeListtestempty (TypeLst));
}

/* --------------------------------------------------------
   TypeListCast2SubList(TypeList TypeLst)
   ========================================================
   returns:
   The sublist contained in TypeLst
*/
extern TypeSubList
TypeListCast2SubList (TypeList TypeLst)
{
  return (macroTypeListgetfirst (TypeLst));
}

/* --------------------------------------------------------
   TypeSubListGetHead(TypeSubList TypeLst)
   ========================================================
   returns:
   The first element of TypeLst
*/
extern TypeType /* TypeClass */
TypeSubListGetHead (TypeSubList TypeLst)
{
  return (macroTypeSubListgethead (TypeLst));
}


/* --------------------------------------------------------
   TypeSubListGetTail(TypeSubList TypeLst)
   ========================================================
   returns:
   The sublist resulting from removing the first element in TypeLst
*/
extern TypeSubList
TypeSubListGetTail (TypeSubList TypeLst)
{
  return (macroTypeSubListgettail (TypeLst));
}

/* --------------------------------------------------------
   TypeSubListTestLast(TypeSubList TypeLst)
   ========================================================
   returns:
   pk_true if tail of TypeLst is empty
*/
extern BoolType
TypeSubListTestLast (TypeSubList TypeLst)
{
  return (macroTypeSubListtestlast (TypeLst));
}


/* --------------------------------------------------------
   TypeListTestandGetOrdinal(TypeList TypeLst, int Ordinal, TypeType *ResTyp)
   ========================================================
   Assumes: Ordinal >= 0
   returns:
   pk_true if TypeLst has more than Ordinal Elements
   and the corresponding type in ResType
   (first type has ordinal 0)
*/
extern BoolType
TypeListTestandGetOrdinal (TypeList TypeLst, int Ordinal, TypeType * ResType)
{
  if (macroTypeListtestempty (TypeLst))
    return pk_false;
  else
    {
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TypeLst);
      do
	{
	  if (Ordinal == 0)
	    {
	      *ResType = (macroTypeSubListgethead (SubLst));
	      return pk_true;
	    }
	  else if (macroTypeSubListtestlast (SubLst))
	    return pk_false;
	  SubLst = macroTypeSubListgettail (SubLst);
	  Ordinal = Ordinal - 1;
	}
      while (pk_true);
    }
}


/* --------------------------------------------------------
   TypeLstCompare(TypeList TypeLst1, TypeList TypeLst2)
   ========================================================
   returns:
   -1 if TypeLst1 is smaller than TypeLst2 for lexical comparison
   0 if TypeLst1 and TypeLst2 are identical
   1 if TypeLst2 is smaller than TypeLst1 for lexical comparison
*/
extern int
TypeLstCompare (TypeList TypeLst1, TypeList TypeLst2)
{
  if (TypeLst1 == TypeLst2)
    return 0;
  else
    {
      if (macroTypeListtestempty (TypeLst1))
	if (macroTypeListtestempty (TypeLst2))
	  return 0;
	else
	  return -1;
      else if (macroTypeListtestempty (TypeLst2))
	return 1;
      else
	{
	  TypeSubList SubLst1, SubLst2;
	  int result, goon;
	  SubLst1 = macroTypeListgetfirst (TypeLst1);
	  SubLst2 = macroTypeListgetfirst (TypeLst2);
	  result = 0;
	  goon = 2;
	  do
	    {
	      result = (TypeLexCompare (macroTypeSubListgethead (SubLst1), macroTypeSubListgethead (SubLst2)));
	      if (result == 0)
		{
		  goon = SubListofTypeFirstEnding (SubLst1, SubLst2);
		  if (goon == 2)
		    {
		      SubLst1 = macroTypeSubListgettail (SubLst1);
		      SubLst2 = macroTypeSubListgettail (SubLst2);
		    }
		  else
		    result = goon;
		}
	    }
	  while (result == 0 && goon == 2);
	  return (result);
	}
    }
}


/* --------------------------------------------------------
   TypeLstEq(TypeList TypeLst1, TypeList TypeLst2)
   ========================================================
   returns:
   false if TypeLst1 and TypeLst2 are different
   pk_true if TypeLst1 and TypeLst2 are identical
*/
extern BoolType
TypeLstEq (TypeList TypeLst1, TypeList TypeLst2)
{
  if (TypeLstCompare (TypeLst1, TypeLst2) == 0)
    return pk_true;
  else
    return pk_false;
}

/* --------------------------------------------------------
   TypeListTypeOrdinal(TypeList TypeLst, TypeType Elem, int *Result)
   ========================================================
   returns:
   pk_true if Elem is a type of TypeLst
   then its ordinal is assigned to Result
   false if Elem is not a type of TypeLst
*/
extern BoolType
TypeListTypeOrdinal (TypeList TypeLst, TypeType Elem, int *Result)
{
  if (macroTypeListtestempty (TypeLst))
    return pk_false;
  else
    {
      int goon;
      TypeSubList SubLst;
      *Result = 0;
      SubLst = macroTypeListgetfirst (TypeLst);
      do
	{
	  goon = (TypeLexCompare (macroTypeSubListgethead (SubLst), Elem));
	  if (goon == 0)
	    return pk_true;
	  else
	    {
	      if (macroTypeSubListtestlast (SubLst))
		return pk_false;
	      else
		{
		  *Result = (*Result) + 1;
		  SubLst = macroTypeSubListgettail (SubLst);
		}
	    }
	}
      while (pk_true);
    }
}

/* --------------------------------------------------------
   TypeListAppendType(TypeList TypeLst, TypeType Elem)
   ========================================================
   side effect:
   Elem is append at the end of TypeLst
*/
extern void
TypeListAppendType (TypeList TypeLst, TypeType Elem)
{
  TypeSubList RefCadR;
  RefCadR = macroTypeSubListnew;
  macroTypeSubListsethead (RefCadR, Elem) macroTypeSubListsetlast (RefCadR);
  if (macroTypeListtestempty (TypeLst))
    {
      macroTypeListsetfirst (TypeLst, RefCadR);
    }
  else
    macroTypeSubListsettail ((macroTypeListgetlast (TypeLst)), RefCadR);
  macroTypeListsetlast (TypeLst, RefCadR);
}

/* --------------------------------------------------------
   TypeListInsertheadingType(TypeList TypeLst, TypeType Elem)
   ========================================================
   side effect:
   Elem is append at the end of TypeLst
*/
extern void
TypeListInsertheadingType (TypeList TypeLst, TypeType Elem)
{
  ListInsertheadingtype (TypeLst, Elem);
}

/* --------------------------------------------------------
   TypeListRegister(TypeList TypeLst)
   ========================================================
   side effect:
   The ordered list of TypeList is created
*/

extern TypeList
TypeListRegister (TypeList TypeLst)
{
  return (ListRegister (TypeLst));
}

/* --------------------------------------------------------
   TypeListInitiate()
   ========================================================
   side effect:
   The ordered list of TypeList is created
*/

extern void
TypeListInitiate ()
{
  TypeLstIndex = TypeLstDictEntryNew (EmptyTypeList, LastTypeDictEntry);
}
