/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********|*********|*********|*********|*********|*********|*********|
	  |**						       	            **|
	  |**                                                File TypeListManager_i.c
	  |**						       	            **|
	  |**						       	            **|
	  |**						       	            **|
	  |**	      	THIS FILE CONTAINS Type List management             **|
	  |**		           LOCAL FUNCTIONS                          **|
	  |**								    **|
	  |*********|*********|*********|*********|*********|*********|*********/

/*
  Ces traitements, en particulier pour la comparaison de type sont à changer
*/

/* --------------------------------------------------------
   TypeLexCompare(TypeType type1, TypeType type2)
   ========================================================
   returns:
   -1 if type1 is smaller than type2 for lexical comparison
   0 if type1 and type2 are identical
   1 if type2 is smaller than type1 for lexical comparison
*/
static int
TypeLexCompare (TypeType Type1, TypeType Type2)
{
  if (Type1 == Type2)
    return 0;
  else if (macroTytestpredef (Type1))
    if (macroTytestpredef (Type2))
      if (Type1 < Type2)
	return -1;
      else
	return 1;
    else
      return -1;
  else if (macroTytestpredef (Type2))
    return 1;
  else
    {
      TypeClass ClCl1, ClCl2;
      ClCl1 = macroTygetclass (Type1);
      ClCl2 = macroTygetclass (Type2);
      if (ClCl1 < ClCl2)
	return -1;
      else if (ClCl1 > ClCl2)
	return 1;
      else if (Type1 < Type2)
	return -1;
      else
	return 1;
    };
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         TypeSubList                      **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

struct LTypeCell
{
  struct LTypeCell *CdR;
  TypeType CaR;
};
struct LTypeDescr
{
  TypeSubList First;
  TypeSubList Last;
};

#define macroTypeSubListnew ((TypeSubList)(malloc(sizeof(TypeListCell))))
#define macroTypeSubListtestlast(sublstarg) (sublstarg->CdR == ((TypeSubList)0))
#define macroTypeSubListgethead(sublstarg) (sublstarg->CaR)
#define macroTypeSubListgettail(sublstarg) (sublstarg->CdR)
#define macroTypeSubListsethead(sublstarg, Elem) (sublstarg->CaR)=Elem;
#define macroTypeSubListsettail(sublstarg, Tail) (sublstarg->CdR)=Tail;
#define macroTypeSubListsetlast(sublstarg) (sublstarg->CdR)=((TypeSubList)0);

static BoolType
SubListofTypeTestLast (TypeSubList STLst)
  /*---------------************------------------------*/
{
  return (macroTypeSubListtestlast (STLst));
}

static TypeType
SubListofTypeGetHead (TypeSubList STLst)
  /*---------------************------------------------*/
{
  return (macroTypeSubListgethead (STLst));
}

static TypeSubList
SubListofTypeGetTail (TypeSubList STLst)
  /*---------------************------------------------*/
{
  return (macroTypeSubListgettail (STLst));
}

/* --------------------------------------------------------
   SubListofTypeFirstEnding(TypeSubList SubLst1, TypeSubList SubLst2)
   ========================================================
   returns:
   -1 if SubLst1 terminates a list and SubLst2 does not
   0 if SubLst1 and SubLst2 both terminate a list
   1 if SubLst2 terminates a list and SubLst1 does not
   2 if neither SubLst1 nor SubLst2 terminate a list
*/
static int
SubListofTypeFirstEnding (TypeSubList SubLst1, TypeSubList SubLst2)
{
  if (macroTypeSubListtestlast (SubLst1))
    if (macroTypeSubListtestlast (SubLst2))
      return 0;
    else
      return -1;
  else if (macroTypeSubListtestlast (SubLst2))
    return 1;
  else
    return 2;
}

/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         TypeList                         **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

#define macroTypeListnew ((TypeList)(malloc(sizeof(TypeListDescr))))
#define macroTypeListtestempty(ListArg)  (ListArg->First == ((TypeSubList)0))
#define macroTypeListgetfirst(ListArg)  (ListArg->First)
#define macroTypeListgetlast(ListArg)  (ListArg->Last)
#define macroTypeListsetempty(ListArg) {\
  (ListArg->First = ((TypeSubList)0));\
  (ListArg->Last = ((TypeSubList)0));\
}
#define macroTypeListsetfirst(ListArg,SubList)  (ListArg->First) = SubList;
#define macroTypeListsetlast(ListArg,SubList)  (ListArg->Last) = SubList;

static BoolType
ListofTypeTestEmpty (TypeList TLst)
  /*---------------************------------------------*/
{
  return (macroTypeListtestempty (TLst));
}

static TypeSubList
ListofTypeGetFirst (TypeList TLst)
  /*---------------************------------------------*/
{
  return (macroTypeListgetfirst (TLst));
}

static TypeList
ListofTypeCreate ()
{
  TypeList Result;
  Result = macroTypeListnew;
  macroTypeListsetempty (Result) return Result;
}

static void
ListInsertheadingtype (TypeList TypeLst, TypeType Elem)
{
  TypeSubList RefCadR;
  RefCadR = macroTypeSubListnew;
  macroTypeSubListsethead (RefCadR, Elem) if (macroTypeListtestempty (TypeLst))
    {
      macroTypeSubListsetlast (RefCadR);
    }
  else
    macroTypeSubListsettail (RefCadR, (macroTypeListgetfirst (TypeLst)));
  macroTypeListsetfirst (TypeLst, RefCadR);
}


/*********|*********|*********|*********|*********|*********|
	  |**						       	  **|
	  |**		         Dictionary                       **|
	  |**							  **|
	  |*********|*********|*********|*********|*********|*********/

typedef struct LTypeDictEntry
{
  struct LTypeDictEntry *CdR;
  TypeList CaR;
} TypeDictEntry;
#define LastTypeDictEntry  ((TypeDictEntry *) 0)

#define macroTypeDictEntrynew (TypeDictEntry *) malloc (sizeof (TypeDictEntry));
#define macroTypeLstDictEntrytestlast(RefDictEntry)\
     (RefDictEntry->CdR == LastTypeDictEntry)
#define macroTypeLstDictEntrygethead(RefDictEntry) (RefDictEntry->CaR)
#define macroTypeLstDictEntrygettail(RefDictEntry) (RefDictEntry->CdR)
#define macroTypeLstDictEntrysettail(RefDictEntry,Tail) (RefDictEntry->CdR)=Tail;
#define macroTypeLstDictEntryset(RefDictEntry,Elem,Tail) {\
(RefDictEntry->CaR)=Elem;\
(RefDictEntry->CdR)=Tail;\
  };

TypeDictEntry *TypeLstIndex;

#define macroTypeLstIndextestempty (TypeLstIndex->CaR == (EmptyTypeList))
#define macroTypeLstIndexsethead(Elem) (TypeLstIndex->CaR)=Elem;

/* --------------------------------------------------------
   TypeList TypeListDictEntryNew(TypeList Elem, TypeDictEntry *Tail)
   ========================================================
   returns:
   a new name list dictionary entry, followed by Tail, with Elem as content
*/
static TypeDictEntry *
TypeLstDictEntryNew (TypeList Elem, TypeDictEntry * Tail)
{
  TypeDictEntry *Result;
  Result = macroTypeDictEntrynew;
  macroTypeLstDictEntryset (Result, Elem, Tail);
  return Result;

}

static TypeList
ListRegister (TypeList TypeLst)
{
  if (macroTypeLstIndextestempty)
    {
      macroTypeLstIndexsethead (TypeLst);
      return TypeLst;
    }
  else
    {
      int rescomp;
      TypeDictEntry *TypeLstNext;
      TypeDictEntry *TypeLstCurrent;
      TypeLstCurrent = TypeLstIndex;
      rescomp = TypeLstCompare (TypeLst, macroTypeLstDictEntrygethead (TypeLstCurrent));
      if (rescomp == -1)
	{
	  TypeLstIndex = TypeLstDictEntryNew (TypeLst, TypeLstCurrent);
	  return TypeLst;
	};
      if (rescomp == 0)
	return (macroTypeLstDictEntrygethead (TypeLstCurrent));
      while (!(macroTypeLstDictEntrytestlast (TypeLstCurrent)))
	{
	  /* rescomp == 1: TypeLst >= head(TypeLstCurrent) */
	  TypeLstNext = macroTypeLstDictEntrygettail (TypeLstCurrent);
	  rescomp = TypeLstCompare (TypeLst, macroTypeLstDictEntrygethead (TypeLstNext));
	  if (rescomp == -1)
	    {
	      TypeLstNext = TypeLstDictEntryNew (TypeLst, macroTypeLstDictEntrygettail (TypeLstCurrent));
	      macroTypeLstDictEntrysettail (TypeLstCurrent, TypeLstNext);
	      return TypeLst;
	    };
	  if (rescomp == 0)
	    return (macroTypeLstDictEntrygethead (TypeLstNext));
	  TypeLstCurrent = TypeLstNext;
	};
      TypeLstNext = TypeLstDictEntryNew (TypeLst, macroTypeLstDictEntrygettail (TypeLstCurrent));
      macroTypeLstDictEntrysettail (TypeLstCurrent, TypeLstNext);
      return TypeLst;
    }
}

static BoolType
TypeSubListReduction (BoolType CreateNew, TypeSubList SubLst, TypeList * RLst)
  /*---------------************------------------------*/
{
  TypeType ATT;
  TypeType RTT;
  ATT = macroTypeSubListgethead (SubLst);
  if (TypeReduction (ATT, &RTT))
    {
      if ((macroTypeSubListtestlast (SubLst)))
	{
	  *RLst = ListofTypeCreate ();
	}
      else
	TypeSubListReduction  (pk_true, macroTypeSubListgettail (SubLst), RLst);
      ListInsertheadingtype (*RLst, RTT);
      return pk_true;
    }
  else if ((macroTypeSubListtestlast (SubLst)))
    if (CreateNew)
      {
	*RLst = ListofTypeCreate ();
	ListInsertheadingtype (*RLst, ATT);
	return pk_true;
      }
    else
      return pk_false;
  else if (TypeSubListReduction (CreateNew, macroTypeSubListgettail (SubLst), RLst))
    {
      ListInsertheadingtype (*RLst, ATT);
      return pk_true;
    }
  else
    return pk_false;
}

static BoolType
TypeListReduction (TypeList TLst, TypeList * RLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      if (TypeSubListReduction (pk_false, SubLst, RLst))
	{
	  *RLst = ListRegister (*RLst);
	  return pk_true;
	}
      else
	return pk_false;
    };
}

static BoolType
TypeListTestGeneric (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestGeneric (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}

static BoolType
TypeListTestLogical (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestLogical (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}

static BoolType
TypeListTestIntNumber (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestIntNumber (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}

static BoolType
TypeListTestRealNumber (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestRealNumber (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}

static BoolType
TypeListTestComplexNumber (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestComplexNumber (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}

static BoolType
TypeListTestNumerical (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestNumerical (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}

static BoolType
TypeListTestPartOrder (TypeList TLst)
  /*---------------************------------------------*/
{
  if (macroTypeListtestempty (TLst))
    return pk_false;
  else
    {
      BoolType Goon, Res;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  Res = TypeTestPartOrder (macroTypeSubListgethead (SubLst));
	  if ((macroTypeSubListtestlast (SubLst)) || (!Res))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
      return Res;
    };
}
