/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/***
    |**
    |**
    |** File TypeMemoryManager_i.c
    |** ==========================
    |**    						            **|
    |**	      THIS FILE CONTAINS  Memory allocation for             **|
    |**                    TYPE REPRESENTATION tables               **|
    |**		                                                    **|
    |**	       						            **|
    |*********|*********|*********|*********|*********|*********|*********/

#define NullPointerDef(Pointer) ((void *)(Pointer) == NullPointer)
typedef struct
{
  TypeType WhoamI;
  TypeType CurrentType;
} AnonymousTypeRepr;
typedef Pointer RefTypeRepr;
typedef struct
{
  TypeClass Category;
  RefTypeRepr RefDescr;
} TypeItemType;
typedef TypeItemType *TypeArray;
typedef struct
{
  int MaxSize;
  int Factor;
  TypeArray Base;
  int Next;
  int Length;
} TypeDescriptorType;

typedef TypeDescriptorType *RefDescriptorType;
typedef RefDescriptorType *RefRefDescriptorType;

TypeDescriptorType *RefTypeTableDescriptor;

#define macroTTDgetnext (RefTypeTableDescriptor->Next)
#define macroTTDgetlength (RefTypeTableDescriptor->Length)
#define macroTTDsetnext(v) RefTypeTableDescriptor->Next = v;

#define macroTygetclass(tyty) (((RefTypeTableDescriptor->Base)[tyty]).Category)
#define macroTySTree_getDescriptionr(cast,tyty) ((cast *)(((RefTypeTableDescriptor->Base)[tyty]).RefDescr))
#define macroTysetclass(tyty,Cl) (((RefTypeTableDescriptor->Base)[tyty]).Category) = Cl;
#define macroTysetdescr(tyty,De) (((RefTypeTableDescriptor->Base)[tyty]).RefDescr) = De;

#define macroTytestNodeType(tyty) ((((RefTypeTableDescriptor->Base)[tyty]).Category) == NodeClass)

static void
TypeTableInitiate (RefRefDescriptorType Descriptor, int InitialSize, int MaxSize, int ResizingFactor, int ElemByteSize, const char *TableName)
{
  *Descriptor = (RefDescriptorType) malloc (sizeof (TypeDescriptorType));
  (*Descriptor)->MaxSize = MaxSize;
  (*Descriptor)->Factor = ResizingFactor;
  (*Descriptor)->Next = 0;
  (*Descriptor)->Length = InitialSize;
  (*Descriptor)->Base = (TypeArray) malloc (InitialSize * ElemByteSize);
  if (NullPointerDef ((*Descriptor)->Base))
    {
      ERRORDescriptor ("BINGO !!!", "initial", *Descriptor);
      //exit(1);
      return;
    };
  TRACEDescriptor ("initial", *Descriptor);
}

static void
TypeTableResizing (RefDescriptorType Descriptor, const char *TableName)
{
  TypeArray result;
  int NewLength;
  NewLength = (Descriptor->Length);
  NewLength = NewLength * (Descriptor->Factor);
  if (NewLength >= (Descriptor->MaxSize))
    {
      ERRORDescriptor ("Bad Parameters", "new", Descriptor);
      return;
    };
  result = (TypeArray) malloc ((NewLength * sizeof (TypeItemType)));
  if (NullPointerDef (result))
    {
      ERRORDescriptor ("BINGO !!!", "new", Descriptor);
      return;
    };
  memmove ((void *) result, (void *) (Descriptor->Base), (Descriptor->Next) * sizeof (TypeItemType));
  Descriptor->Length = NewLength;
  Descriptor->Base = (TypeArray) result;
  TRACEDescriptor ("new", Descriptor);
}

/* --------------------------------------------------------
   TypeEqAdr(TypeType TT1, TypeType TT2)
   ========================================================
   returns:
   pk_true if types TT1 and TT2 have the same adress
   false if types TT1 and TT2 have different adresses
*/
static BoolType
TypeEqAdr (TypeType TT1, TypeType TT2)
  /*---------------************------------------------
   */
{
  return (TT1 == TT2);
}


/* --------------------------------------------------------
   NewTypeItem(TypeClass Category, Pointer RefDescr)
   ========================================================
   returns:
   a new TypeType inserted in the TypeTable
*/
static TypeType
NewTypeItem (TypeClass Category, Pointer RefDescr)
  /*---------------***********---------------------------------------*/
{
  TypeType result;
  result = macroTTDgetnext;
  if (result >= (macroTTDgetlength))
    TypeTableResizing (RefTypeTableDescriptor, TypeTableName);
  macroTTDsetnext ((result + 1));
  macroTysetclass (result, Category) macroTysetdescr (result, RefDescr) return result;
}
