/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/**
   |**                                                        TypeSigprint_i.c
   |**                                                        ================
   |**						       	            **|
   |**	      	THIS FILE CONTAINS T Y P E Signal Printing          **|
   |**		           LOCAL FUNCTIONS                          **|
   |**								    **|
   |*********|*********|*********|*********|*********|*********|*********/

/************ 
	      |** Required: the following type declaration is required
	      |************/
/*
  typedef int TypeType;
*/

static void
SigPrintExpcal (RefFILE stderrorout, ExpcalCst Exp)
  /*---------------************------------------------
   */
{
  macrowString (stderrorout, "NotImplemented: Print Expcal");
}

static void
SigNamePrint (RefFILE stderrorout, NameClass Name)
  /*---------------************------------------------ */
{
  macrowString (stderrorout, Name);
}

static const char *SigTypeStr[16] = {
  "undefined",
  "label",
  "event",
  "boolean",
  "short",
  "integer",
  "long",
  "real",
  "dpreal",
  "complex",
  "dcomplex",
  "char",
  "string",
  "byte",
  "word",
  "ERROR"
};

static void
SigPredefTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a predef type */
{
  macrowString (stderrorout, SigTypeStr[-TT]);
}

static const char *SigClassStr[] = {
  "-enum-",
  "extern",
  "-lambda-",
  "-subrange-",
  "-any-",
  "-error-",
  "-named-",
  "-meta-",
  "-bits-",
  "-array-",
  "-matrix-",
  "-tuple-",
  "structure",
  "bundle",
  "-node-",
  "-generic node",
  "-meta node"
};


/*
  Forward reference
*/
void SigTypePrint (RefFILE stderrorout, TypeType TT);

static void
SigNameListPrint (RefFILE stderrorout, NameList NLst)
  /*---------------************------------------------ */
{
  if (NameListTestEmpty (NLst));
  else
    {
      BoolType Goon;
      NameSubList NSubLst;
      NSubLst = NameListCast2SubList (NLst);
      Goon = pk_true;
      do
	{
	  SigNamePrint (stderrorout, NameSubListGetHead (NSubLst));
	  if (NameSubListTestLast (NSubLst))
	    Goon = pk_false;
	  else
	    {
	      macrowvirg (stderrorout);
	      NSubLst = NameSubListGetTail (NSubLst);
	    }
	}
      while (Goon);
    }
}

static void
SigEnumerationTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be an enumeration type */
{
  SigNamePrint (stderrorout, EnumerationGetName (TT));
  macroweq (stderrorout);
  macrowlpar (stderrorout);
  SigNameListPrint (stderrorout, EnumerationGetNameLst (TT));
  macrowrpar (stderrorout);
}

static void
SigExternTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be an extern type */
{
  macrowextern (stderrorout);
  SigNamePrint (stderrorout, ExternGetName (TT));
}

static void
SigSubrangeTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Subrange type */
{
  macrowsubrange (stderrorout);
  SigPrintExpcal (stderrorout, SubrangeGetBounds (TT));
}

static void
SigAnyTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Any type */
{
  if (AnyTestDefined (TT))
    SigTypePrint (stderrorout, AnyGetDefinition (TT));
  else
    macrowany (stderrorout);

}

static void
SigLambdaTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Lambda type */
{
  macrowlambda (stderrorout);
  macrowddot (stderrorout);
  SigTypePrint (stderrorout, LambdaGetTerm (TT));
}

static void
SigMetaTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Meta type */
{
  macrowmeta (stderrorout);
  macrowddot (stderrorout);
  if (MetaTestDefined (TT))
    SigTypePrint (stderrorout, MetaNamedGetCurrent (TT));
  else
    SigNamePrint (stderrorout, MetaNamedGetName (TT));
}

static void
SigMetaTypeDeclPrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Meta type */
{
  macrowtype (stderrorout);
  SigNamePrint (stderrorout, MetaNamedGetName (TT));
}

static void
SigNamedTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Named type */
{
  SigTypePrint (stderrorout, NamedGetTerm (TT));
}

static void
SigNamedTypeDeclPrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Named type */
{
  macrowtype (stderrorout);
  SigNamePrint (stderrorout, MetaNamedGetName (TT));
  macroweq (stderrorout);
  SigTypePrint (stderrorout, NamedGetTerm (TT));
}

static void
SigBitsTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a Bits type */
{
  macrowbits (stderrorout);
  SigPrintExpcal (stderrorout, BitsGetDim (TT));
}

static void
SigArrayMatrixTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be an array (resp. matrix) type */
{
  macrowarray (stderrorout);
  SigPrintExpcal (stderrorout, ArrayMatrixGetDimLst (TT));
  SigTypePrint (stderrorout, ArrayMatrixGetTerm (TT));
}


static void
SigTypeListPrint (RefFILE stderrorout, TypeList TLst)
  /*---------------************------------------------
    TT is assumed to be a tuple of types */
{
  if (ListofTypeTestEmpty (TLst));
  else
    {
      BoolType Goon;
      TypeSubList SubLst;
      SubLst = ListofTypeGetFirst (TLst);
      Goon = pk_true;
      do
	{
	  SigTypePrint (stderrorout, SubListofTypeGetHead (SubLst));
	  if (SubListofTypeTestLast (SubLst))
	    Goon = pk_false;
	  else
	    {
	      macrowvirg (stderrorout);
	      SubLst = SubListofTypeGetTail (SubLst);
	    }
	}
      while (Goon);
    }
}

static void
SigTupleTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a tuple of types */
{
  macrowlpar (stderrorout);
  SigTypeListPrint (stderrorout, TupleGetTypeLst (TT));
  macrowrpar (stderrorout);
}

static void
SigTypeNameListCoPrint (RefFILE stderrorout, TypeList TLst, NameList NLst)
  /*---------------************------------------------
   */
{
  if (ListofTypeTestEmpty (TLst));
  else
    {
      BoolType Goon;
      TypeSubList TSubLst;
      NameSubList NSubLst;
      TSubLst = ListofTypeGetFirst (TLst);
      NSubLst = NameListCast2SubList (NLst);
      Goon = pk_true;
      do
	{
	  SigTypePrint (stderrorout, SubListofTypeGetHead (TSubLst));
	  macrowspace (stderrorout);
	  SigNamePrint (stderrorout, NameSubListGetHead (NSubLst));
	  if (SubListofTypeTestLast (TSubLst))
	    Goon = pk_false;
	  else
	    {
	      macrowvirg (stderrorout);
	      TSubLst = SubListofTypeGetTail (TSubLst);
	      NSubLst = NameSubListGetTail (NSubLst);
	    }
	}
      while (Goon);
    }
}

static void
SigStructureClassPrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a structure type representation */
{
  SigTypeNameListCoPrint (stderrorout, StructureGetTypeLst (TT), StructureGetNameLst (TT));
}

static void
SigStructureTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a structure type */
{
  macrowstruct (stderrorout);
  macrowlpar (stderrorout);
  SigStructureClassPrint (stderrorout, TT);
  macrowrpar (stderrorout);
}

static void
SigNodeClassPrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  macrowsin (stderrorout);
  SigStructureClassPrint (stderrorout, NodeGetDomain (TT));
  macrowsout (stderrorout);
  SigStructureClassPrint (stderrorout, NodeGetImage (TT));
}

static void
SigNodeTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  macrowlpar (stderrorout);
  SigNodeClassPrint (stderrorout, TT);
  macrowrpar (stderrorout);
}

static void
SigGenericNodeClassPrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  TypeType Term;
  SigTypePrint (stderrorout, GenericNodeGetArgument (TT));
  Term = GenericNodeGetTerm (TT);
  if (macroTytestNodeType (Term))
    {
      macrowracc (stderrorout);
      SigNodeTypePrint (stderrorout, Term);
    }
  else
    {
      macrowvirg (stderrorout);
      SigGenericNodeClassPrint (stderrorout, Term);
    }
}

static void
SigGenericNodeTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a node type */
{
  macrowlacc (stderrorout);
  SigGenericNodeClassPrint (stderrorout, TT);
}

static void
SigMetaNodeTypePrint (RefFILE stderrorout, TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a meta node type */
{
  TypeType Term;
  macrowlacc (stderrorout);
  Term = MetaNodeGetArgument (TT);
  SigTypeListPrint (stderrorout, TupleGetTypeLst (Term));
  Term = MetaNodeGetTerm (TT);
  if (macroTytestNodeType (Term))
    {
      macrowracc (stderrorout);
      SigNodeTypePrint (stderrorout, Term);
    }
  else
    {
      macrowvirg (stderrorout);
      SigGenericNodeClassPrint (stderrorout, Term);
    }
}
