/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Construtor of a TravelClockAttributes object.
*/
  TravelClockAttributes::TravelClockAttributes ()
{
  lispredgen = NILLINODE;
  lhspredgen = NILLINSTH;
  nbspredtree = NILINFOGENCODESUBTREE;
  reached = pk_false;
}

/**
   @return the set of eligible nodes of the current object.
*/
ptlINode
TravelClockAttributes::getFreeNodesInSubTree ()
{
  return lispredgen;
}

/**
   The list (li) becomes the list of eligible nodes of the current object.
*/
void
TravelClockAttributes::setFreeNodesInSubTree (ptlINode li)
{
  lispredgen = li;
}

/**
   The list (lh) becomes the list of eligible clocks of the current object.
*/
void
TravelClockAttributes::setFreeClocksInSubTree (ptlinsth lh)
{
  lhspredgen = lh;
}

/**
   @return the set of eligible clocks of the current object.
*/
ptlinsth
TravelClockAttributes::getFreeClocksInSubTree ()
{
  return lhspredgen;
}


/**
   v becomes the specific attributes (infoGenCodeSubTree), associated with a node of a hierarchy,  of the current object.
*/
void
TravelClockAttributes::setInfoGenCodeSubTree (ptinfoGenCodeSubTree v)
{
  nbspredtree = v;
}

/**
   @return the specific attributes (infoGenCodeSubTree) associated with the current object.
*/
ptinfoGenCodeSubTree
TravelClockAttributes::getInfoGenCodeSubTree ()
{
  return nbspredtree;
}


/**
   The current object is marked reached/unreached when b/not b.
*/
void
TravelClockAttributes::setReached (boolean b)
{
  reached = b;
}

/**
   @return "true" if the current object is marked reached, "false" otherwise.
*/
boolean TravelClockAttributes::isReached ()
{
  return reached;
}
