/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor of a TravelINodeAttributes object.
 */
  TravelINodeAttributes::TravelINodeAttributes ()
{
  initialize ();		// reached = pk_false; 
}

/**
   It initializes the current object :
   - number of selected predecessors: 0
   - number of (waited) predecessors: 0
   - not selected 
   - not reached by one of its predecessor.
 */
void
TravelINodeAttributes::initialize ()
{
  nbpredparc = 0;
  nbpred = 0;
  parc = pk_false;
  reached = pk_false;
}

/** @return the number of (waited) predecessors of the current object. */
integer TravelINodeAttributes::getnbpred ()
{
  return nbpred;
}

/** @return "true" if the number of (waited) predecessors of the current object is equal to 0, "false" otherwise.
 */
boolean TravelINodeAttributes::isnullnbpred ()
{
  return (nbpred == 0);
}

/** It decreases the  number of (waited) predecessors of the current object.
 */
void
TravelINodeAttributes::decrnbpred ()
{
  nbpred--;
}


/** It increases the  number of (waited) predecessors of the current object.
 */
void
TravelINodeAttributes::incrnbpred ()
{
  nbpred++;
}


/** @return the value of the selected attribute of the current node.
 */
boolean TravelINodeAttributes::isAlreadySelected ()
{
  return parc;
}


/** @return "true" if the internal counter (number of selected predecessors) is equal to \e vpred, "false" otherwise.
 */
boolean TravelINodeAttributes::numberOfSelectedPredEqualTo (integer vpred)
{
  return (boolean) (nbpredparc == vpred);
}

/** @return the result of the test of the equality of the internal counters
    (number of selected predecessors, number of (waited) predecessors) of the current object.
 */
boolean TravelINodeAttributes::allPredSelected ()
{
  return (boolean) (nbpredparc == nbpred);
}

/** @return the difference between vpred and the number of selected predecessors of the current object.
 */
integer TravelINodeAttributes::get_nbWaitedPredecessors (integer vpred)
{
  return vpred - nbpredparc;
}

/** It increases the number of selected predecessors of the current object.
*/
void
TravelINodeAttributes::incrNbOfSelectedPred ()
{
  nbpredparc++;
}


/**
   The current object is marked selected.
*/
void
TravelINodeAttributes::setSelected ()
{
  parc = pk_true;
}


/**  The current object is marked "unselected" and 
     counter of waited predecessors is set to 0.
*/
void
TravelINodeAttributes::setUnselected ()
{
  parc = pk_false;
  nbpred = 0;
}


/**
   It resets the  number of selected predecessors  of the current object.
*/
void
TravelINodeAttributes::resetCounterSelectedPred ()
{
  nbpredparc = 0;
}


/**
   The current object is marked reached/unreached when b/not b. 
 */
void
TravelINodeAttributes::setReached (boolean b)
{
  reached = b;
}


/**
   @return "true" if the current object is marked reached, "false" otherwise.
*/
boolean TravelINodeAttributes::isReached ()
{
  return reached;
}

/**
   It resets the fields of the current object.
*/
void
TravelINodeAttributes::reset ()
{
  setUnselected ();
  nbpredparc = 0;
}


/**
   @return the number of visited predecessors.
*/
integer TravelINodeAttributes::getNbVisitedPred ()
{
  return nbpredparc;
}
