/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    It is a specialisation of InterFormatTranslator::formulaCoding().

    It produces the boolean representation of a clock formula (clk) at tick.
    For a non child of the tick, 
    - the boolean is x:= left_tt(clk, k) where k is the boolean clock of the child of tick on the branch (tick,clk)
    - the clock is removed from the children of its father clock
    - the clock is added as a new child of tick.
 */
  _VIRTUAL_ ptinsth DCGBool2DCGFlatTranslator::formulaCoding (ptinsth clk)
{
  ptexpcal
    vhExp,
    c;
  ptinsth
    clkFather;

  clkFather = CKM->getFather (clk);
  if (!CKM->satisfiesTickClock (clkFather))
    {
      // Removing from the children of its fater clock.
      CKM->removeChild (clkFather, clk);
      // making the definition; 
      c = CKM->getBooleanSignal (clk);
      vhExp = NM->createLeftt (c, btick, htick);
      clk = CKM->getRepresentativeFormula (endDefinition (clk, vhExp, htick));
      // Adding as a child of the tick.
      clk = CKM->addPlaced (clk, htick);
    }
  return CKM->getRepresentativeFormula (clk);
}

/**
   Redefinition. Here, it is without effects.

   Called by InterFormatTranslator::clocksCoding().
*/
_VIRTUAL_ void
DCGBool2DCGFlatTranslator::propagateClock (ptinsth clk)
{
}

/**
   It defines the attributes member of the class
   - htick: the tick clock of the graph.
   - btick: the boolean that represents the tick clock.
*/
_VIRTUAL_ void
DCGBool2DCGFlatTranslator::tickCoding ()
{
  htick = tickClock;		// CKM->getTickClock(); 
  btick = CKM->getBooleanSignal (htick);
  CKM->addMark (htick, CST_MYMARK);
}
