/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define xxinfos(sg) {if (sg != NILINODE) { pK_messenger->printString(" of Subgraph Number:"); fprintf(stdout,"%d", NM->get_num(sg)); pK_messenger->printString(" \n"); } \
  else pK_messenger->printString(" of <Scheduler>  \n");}
/**
   In debugging mode (DEBUG attribute), it prints informations about a node (inst) of a sub-graph (sg) at initial clock if bhinit.
*/
  void
DCGBool2DCGSchTranslator::DEBUGGING (ptINode inst, ptINode sg, boolean bhinit)
{
  if (DEBUG)
    {
      pK_messenger->printString (" generate node number :");
      printNode (inst);
      if (!bhinit)
	xxinfos (sg)
	else
	pK_messenger->printString (" of <initialize block> \n ");
      printOutputs (inst);
    }
}

/**
   In debugging mode (DEBUG attribute), it prints informations about a clock (vh) of a sub-graph (sg).
*/
void
DCGBool2DCGSchTranslator::DEBUGGING (ptinsth h, ptINode sg)
{
  if (DEBUG)
    {
      pK_messenger->printString (" generate clock:");
      printClock (h);
      xxinfos (sg);
    }
}

/**
   In debugging mode (DEBUG attribute), it prints the number of a node (n) following by the  string \e "used)".
*/
void
DCGBool2DCGSchTranslator::printNode (ptINode n)
{
  if (DEBUG)
    {
      fprintf (stdout, " ( %d, ", NM->get_num (n));
      fprintf (stdout, " used ) ");
    }
}

/**
   In debugging mode (DEBUG attribute), it prints the name of a clock (h)
*/
void
DCGBool2DCGSchTranslator::printClock (ptinsth h)
{
  if (DEBUG)
    {
      if (CKM->satisfiesSub (h))
	fprintf (stdout, "[ %s ] ", repstr2string (ident2repstr (EM->getName (CKM->getBooleanSignal (h)))));
      else
	fprintf (stdout, repstr2string (ident2repstr (CKM->getName (h))));
    }
}

/**
   In debugging mode (DEBUG attribute), it calls printClock() for each element of a list (lh).
*/
void
DCGBool2DCGSchTranslator::printlClocks (ptlinsth lh)
{
  ptinsth h;

  if (DEBUG)
    while (ptlinsth_get_next (&lh, &h))
      printClock (h);
}

/**
   In debugging mode (DEBUG attribute), it calls printNode() for each element of a list (ln).
*/
void
DCGBool2DCGSchTranslator::printlNodes (ptlINode ln)
{
  ptINode node;
  if (DEBUG)
    while (ptlINode_get_elem (&ln, &node))
      printNode (node);
}



/**
   In debugging mode (DEBUG attribute), it calls printNode() for each node N such that (M->N) is an element of a list of dependences (ldep).
*/
void
DCGBool2DCGSchTranslator::printlNodes (ptldependence ldep)
{
  ptINode node, n1;
  ptdependence adep;

  if (DEBUG)
    while (ldep->get_next (&adep, &ldep))
      {
	adep->getDependence (&n1, &node);
	printNode (node);
      }
}


/**
   In debugging mode (DEBUG attribute), it prints the name of the signals defined in a node (node).
*/
void
DCGBool2DCGSchTranslator::printOutputs (ptINode node)
{
  ptlinstances lsig;
  ptexpcal sig;

  if (DEBUG)
    {
      lsig = NM->getInodelOutputFlows (node);
      fprintf (stdout, "       Outputs: {");
      while (ptlinstances_get_next (&lsig, &sig))
	fprintf (stdout, "%s ", repstr2string (ident2repstr (EM->getName (sig))));
      fprintf (stdout, "}\n");
    }
}


/**
   In debugging mode (DEBUG attribute), it prints informations for a graph of clusters.
   For each cluster, it gives
   - the "generated" nodes and the "locked" ones (with the number of not reached predecessors). The successors are also printed.
   - the "generated" clocks and the "locked" ones. The successors are also printed.
*/
void
DCGBool2DCGSchTranslator::DEBUGGING ()
{
  integer i;
  ptlINode ln;
  ptlinsth lh;
  ptinsth h;
  ptINode node, sg;

  if (DEBUG)
    {
      if (nbSets == 0)
	return;			/* pas de lignees */
      fprintf (stdout, "\n ********************** \n Nb SUBGRAPHS (Scheduller not included) :\n", "%d", nbSets);
      /* le scheduller n'est pas decompile; */
      for (i = 1; i < nbSets; i++)
	{
	  sg = SubGraphTable[i]->aSubGraph;
	  fprintf (stdout, ">>>> nodes/Clocks of subgraph:  %d \n", NM->get_num (sg));
	  if (NM->get_subGraphDefinitions (sg, &ln, &lh))
	    {
	      while (ptlINode_get_elem (&ln, &node))
		{
		  fprintf (stdout, "   node : ");
		  printNode (node);
		  printOutputs (node);
		  int nbWait, n;
		  n = PARC->ptINodeAttr_getNumberOfPred (node);
		  nbWait = PARC->ptINodeAttr_getNumberOfWaitedPred (node, n);
		  fprintf (stdout, "       <nbpred:> %d ", n);
		  if (!PARC->isNodeAlreadySelected (node))
		    fprintf (stdout, "       <wait:> %d ^^^^^", nbWait);
		  else
		    fprintf (stdout, "       generated");
		  fprintf (stdout, "\n       Successors: {");

		  printlClocks (NM->get_clockSuccessors (node));
		  printlNodes (NM->get_successors (node));
		  fprintf (stdout, "}\n");
		}
	      while (ptlinsth_get_next (&lh, &h))
		{
		  fprintf (stdout, "      clock :");
		  printClock (h);
		  if (!PARC->ptinsthAttr_isAlreadySelected (h))
		    fprintf (stdout, "not generated");
		  fprintf (stdout, "\n       Successors: {");
		  printlClocks (CKM->getClockReferencesTo (h));
		  printlNodes (CKM->getNodeReferencesTo (h));
		  fprintf (stdout, "}\n");
		}
	      fprintf (stdout, "---------------end-----\n");
	    }

	}
    }
}
