/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isEventClockedFormat(t) ((t == DCG_BASIC) || (t == DCG_POLY) || (t == DCG_ENDO))
#define isIdentExp(vexp) (EM->getOperatorClass(vexp) == variabl)

#define isSignalOfInterface(vexp) (!EM->isSharedSignal(vexp) && !EM->isStateVar(vexp) && !EM->isSharedSignalBySeveralGraphs(vexp))

_PRIVATE_
/** @return the list of the output clocks  associated with a node of the hierarchy (vh). */
  ptlinsth ProfileBuilder::stInterface_get_lhoout (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_outputClocks (vstru);
}

/** @return the list of the local clocks associated with a node of the hierarchy (vh). */
ptlinsth ProfileBuilder::stInterface_get_lholoc (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_localClocks (vstru);
}



/** @return the list of the input clocks associated with a node of the hierarchy (vh). */
ptlinsth ProfileBuilder::stInterface_get_lhoin (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_inputClocks (vstru);
}



/** @return the list of the parameters associated with a node of the hierarchy (vh). */
ptlinstances ProfileBuilder::stInterface_get_lpar (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_parameters (vstru);
}


/** @return the list of the constants associated with a node of the hierarchy (vh). */
ptlinstances ProfileBuilder::stInterface_get_constants (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_constants (vstru);
}

/** @return the list of the input signals associated with a node of the hierarchy (vh). */
ptlinstances ProfileBuilder::stInterface_get_lent (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_inputSignals (vstru);
}


/** @return the list of the output signals associated with a node of the hierarchy (vh). */
ptlinstances ProfileBuilder::stInterface_get_lsort (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_outputSignals (vstru);
}

/** @return the list of the local signals associated with a node of the hierarchy (vh). */
ptlinstances ProfileBuilder::stInterface_get_lloc (ptinsth vh)
{
  ptinterface
    vstru;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  return ptinterface_get_localSignals (vstru);
}

/** It adds a signal (vexp) as an input signal to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_ent (ptinsth vh, ptexpcal vexp)
{
  ptinterface vstru;
  ptlinstances lsig;
  ptexpcal vsig;

  if (!isIdentExp (vexp))
    {
      lsig = EM->getSignalsReferencedIn (vexp);
      while (ptlinstances_get_next (&lsig, &vsig))
	stInterface_insert_ent (vh, vsig);
    }
  else
    {
      vexp = EM->getDeclaredSignal (vexp);
      if (isSignalOfInterface (vexp))
	{
	  vstru = CKM->getInterfaceSubHierarchy (vh);
	  vstru = ptinterface_insert_inputSignal (vstru, vexp);
	  CKM->setInterfaceSubHierarchy (vh, vstru);
	}
    }
}

/** It adds a signal (vexp) as an output signal to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_sort (ptinsth vh, ptexpcal vexp)
{
  ptinterface vstru;
  ptlinstances lsig;
  ptexpcal vsig;

  if (!isIdentExp (vexp))
    {
      lsig = EM->getSignalsReferencedIn (vexp);
      while (ptlinstances_get_next (&lsig, &vsig))
	stInterface_insert_sort (vh, vsig);
    }
  else
    {
      vexp = EM->getDeclaredSignal (vexp);
      if (!ptlinstances_isin_exp (inputs, vexp))
	{
	  if (isSignalOfInterface (vexp))
	    {
	      vstru = CKM->getInterfaceSubHierarchy (vh);
	      vstru = ptinterface_insert_outputSignal (vstru, vexp);
	      CKM->setInterfaceSubHierarchy (vh, vstru);
	    }
	}
    }
}

/** It adds a signal (vexp) as a local signal to a node of the hierarchy (vh). */
void ProfileBuilder::addLocal(ptinsth vh, ptexpcal vexp)
{
  ptinterface vstru;
  vstru = CKM->getInterfaceSubHierarchy (vh);
  vstru = ptinterface_insert_localSignal (vstru, vexp);
  CKM->setInterfaceSubHierarchy (vh, vstru);
}

/** It adds a signal (vexp) as a local signal to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_loc (ptinsth vh, ptexpcal vexp)
{
  ptlinstances lsig;
  ptexpcal vsig;

  if (!isIdentExp (vexp))
    {
      lsig = EM->getSignalsReferencedIn (vexp);
      while (ptlinstances_get_next (&lsig, &vsig))
	stInterface_insert_loc (vh, vsig);
    }
  else
    {
      vexp = EM->getDeclaredSignal (vexp);
      if (!EM->IsOutputFlow (vexp))
	if (isSignalOfInterface (vexp))
	  addLocal(vh, vexp);
    }
}


/** It adds a parameter (vexp) to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_lpar (ptinsth vh, ptexpcal vexp)
{
  ptinterface vstru;

  vexp = EM->getDeclaredSignal (vexp);
  vstru = CKM->getInterfaceSubHierarchy (vh);
  vstru = ptinterface_insert_parameter (vstru, vexp);
  CKM->setInterfaceSubHierarchy (vh, vstru);
}


/** It adds a clock (vhin) as an input clock to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_hoin (ptinsth vh, ptinsth vhin)
{
  ptinterface vstru;
  ptexpcal vsig;
  ptlinstances lsig;

  if (!CKM->satisfiesPredefinedClock (vh))
    {
      vstru = CKM->getInterfaceSubHierarchy (vh);
      if (CKM->isSubstitutable (vhin))
	// restricted to extraction.
	{
	  lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vhin));
	  while (ptlinstances_get_next (&lsig, &vsig))
	    stInterface_insert_ent (vh, vsig);
	}
      else if (!CKM->isClockRepresentedBySignal (vhin, &vsig))
	{
	  vstru = ptinterface_insert_inputClock (vstru, vhin);
	  CKM->setInterfaceSubHierarchy (vh, vstru);
	}
    }
}



/** It adds a clock (vhout) as an output clock to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_hoout (ptinsth vh, ptinsth vhout)
{
  ptinterface vstru;
  ptexpcal vsig;
  ptlinstances lsig;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  if (CKM->isSubstitutable (vhout))
    // restricted to extraction.
    {
      lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vhout));
      while (ptlinstances_get_next (&lsig, &vsig))
	stInterface_insert_sort (vh, vsig);
    }
  else if (CKM->isClockRepresentedBySignal (vhout, &vsig))
    stInterface_insert_sort (vh, vsig);
  else
    vstru = ptinterface_insert_outputClock (vstru, vhout);
  CKM->setInterfaceSubHierarchy (vh, vstru);
}

/** It adds a clock (vhloc) as a local clock, and if it is not an output, to a node of the hierarchy (vh). */
void
ProfileBuilder::stInterface_insert_holoc (ptinsth vh, ptinsth vhloc)
{
  ptinterface vstru;
  ptexpcal vsig;
  ptlinstances lsig;

  vstru = CKM->getInterfaceSubHierarchy (vh);
  if (CKM->isSubstitutable (vhloc))
    // restricted to extraction.
    {
      lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vhloc));
      while (ptlinstances_get_next (&lsig, &vsig))
	stInterface_insert_loc (vh, vsig);
    }
  else if (CKM->isClockRepresentedBySignal (vhloc, &vsig))
    stInterface_insert_loc (vh, vsig);
  else if (!ptinterface_is_outputClock (vstru, vhloc))
    vstru = ptinterface_insert_localClock (vstru, vhloc);
  CKM->setInterfaceSubHierarchy (vh, vstru);
}
