/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" is the node (n) is selected for the translation, "false" otherwie.
   @param n (in) object of the original graph.
*/
  _VIRTUAL_ boolean AbstractHomomorphism::isSelectedNode (ptINode n)
{
  return pk_true;
}

/**
   @return "true" is the node of a signal (x) is selected, "false" otherwie.
   @param x (in) object of the original graph.
*/
_VIRTUAL_ boolean
AbstractHomomorphism::isSelectedSignal (ptexpcal x)
{
  return isSelectedNode (fEM->getNode (x));
}

/**
   Called for a declaration, in the generated graph, of a signal (vsig).
   @param vsig (in) a clock of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDclIdentifier (ptexpcal vsig)
{
}

/**
   Called for a undefined signal declaration (vsig).
   @param vsig (in) a clock of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismUndefinedIdentifier (ptexpcal vsig)
{
}


/**
   Called for a declaration of a clock (vh).
   @param vh (in) a clock of the original graph.
 
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDclClock (ptinsth vh)
{
}

/**
   Called at the beginning of a translation of a node(n).
   @param n (in) object of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::beginHomomorphismNode (ptINode n)
{
}

/**
   Called after the translation of a node (no) in a new one (ng).
   @param no (in) object of the original graph.
   @param ng (in) object of the generated graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::endHomomorphismNode (ptINode no, ptINode ng)
{
}

/**
   Called for a undefined signal declaration (vsig): no effect.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismHierarchy (ptlinsth lh)
{
}

/**
   Called for the translation of the dependences associated with a node (aNode).
   @param aNode (in) object of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDependence (ptINode aNode)
{
}

/**
   Called for the translation of dependences associated with a clock (h).
   @param h (in) object of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDependence (ptinsth h)
{
}

/**
   Called for the translation of the definition of a clock (h).
   @param h (in) object of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismClock (ptinsth h)
{
}

/**
   Called for the translation of a clock constraint on placed formula (vh1 = vh2).
   @param vh1 (in) object of the original graph.
   @param vh2 (in) object of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismConstraint (ptinsth vh1, ptinsth vh2)
{
}

/**
   Called at the end of the translation.
*/
_VIRTUAL_ void
AbstractHomomorphism::finalize ()
{
}

/**
   @return the list of the current instances (images) of the element of a list (ls).
*/
_VIRTUAL_ ptlinstances
AbstractHomomorphism::getCurrentInstances (ptlinstances ls)
{
  return NILLINSTANCES;
}

_VIRTUAL_ ptexpcal
AbstractHomomorphism::getCombedExprAttributes (ptexpcal e)
{
  return NILEXPCAL;
}

/**
   Called for defining a image of a signal (vsig) in the produced graph.
   The links between the two graphs can be done by attribute.
   @param vsig (in) object of the original graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismAttribute (ptexpcal vsig)
{
}

/**
   @return the morphism of a node (aNode).
   Called for the translation of an equation node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismEqu (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: Equ not implemented");
  return sNM->mk_node_undefined ();
}


/**
   @return the morphism of a node (aNode).
   Called for the translation of a delay node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismDelay (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: Delay not implemented");
  return sNM->mk_node_undefined ();
}


/**
   @return the morphism of a node (aNode).
   Called for the translation of a node call node
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismNodecall (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: Node Call not implemented");
  return sNM->mk_node_undefined ();
}



/**
   @return the morphism of a node (aNode).
   Called for the translation of a procedure call node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismProccall (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: Proc Call not implemented");
  return sNM->mk_node_undefined ();
}

/**
   @return the morphism of a node (aNode).
   Called for the translation of a write call node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismWrite (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: Write Proc Call not implemented");
  return sNM->mk_node_undefined ();
}


/**
   @return the morphism of a node (aNode).
   Called for the translation of a read call node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismRead (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: Read Proc Call not implemented");
  return sNM->mk_node_undefined ();
}


/**
   @return the morphism of a node (aNode).
   Called for the translation of an array node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismArray (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: array not implemented");
  return sNM->mk_node_undefined ();
}

/**
   @return the morphism of a node (aNode).
   Called for the translation of a sub-graph node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismSubGraph (ptINode vinstr)
{
  pK_messenger->printString ("Homomorphism: SubGraph not implemented");
  return sNM->mk_node_undefined ();
}


/**
   @return the morphism of a node (aNode).
   Called for the translation of a bundle/structure (list of fields) node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismBundle (ptINode aNode)
{
  pK_messenger->printString ("Homomorphism: Structure/Bundle not implemented");
  return sNM->mk_node_undefined ();
}


/**
   @return the morphism of a node (aNode).
   Called for the translation of a set of partial definitions node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismPartialDefs (ptINode aNode)
{
  pK_messenger->printString ("Homomorphism: Partial Definition List not implemented");
  return sNM->mk_node_undefined ();
}

/**
   @return the morphism of a node (aNode).
   Called for the translation of a partial definition node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismPartialDef (ptINode aNode)
{
  pK_messenger->printString ("Homomorphism: a Partial Definition not implemented");
  return sNM->mk_node_undefined ();
}

/**
   @return the morphism of a node (aNode).
   Called for the translation of an undefined node.
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::homomorphismUndefined (ptINode aNode)
{
  // pK_messenger->printString("Homomorphism: a Undefined Definition -- not implemented");
  return sNM->mk_node_undefined ();
}

/**
   @return the morphism of a node (aNode).

   It calls to specific method for each kind of node (they must be redefined):
   - homomorphismUndefined()
   - homomorphismEqu()
   - homomorphismDelay()
   - homomorphismProccall() with some particular cases: homomorphismRead(), homomorphismWrite()
   - homomorphismNodecall()
   - homomorphismSubGraph()
   - homomorphismBundle()
   - homomorphismPartialDefs()
   - homomorphismPartialDef()
*/
_VIRTUAL_ ptINode
AbstractHomomorphism::nodeHomomorphism (ptINode aNode)
{
  ptINode cloneNode;

  cloneNode = NILINODE;
  switch ((int) (fNM->get_oper (aNode)))
    {
    case opnil:
      cloneNode = homomorphismUndefined (aNode);
      break;

    case opdefsig:
      cloneNode = homomorphismEqu (aNode);
      break;

    case opret:
      cloneNode = homomorphismDelay (aNode);
      break;

    case opappe:
      if (fNM->is_readNode (aNode))
	cloneNode = homomorphismRead (aNode);
      else if (fNM->is_writeNode (aNode))
	cloneNode = homomorphismWrite (aNode);
      else
	cloneNode = homomorphismProccall (aNode);
      break;

    case opnodecall:
      cloneNode = homomorphismNodecall (aNode);
      break;

    case opsubgraph:
      cloneNode = homomorphismSubGraph (aNode);
      break;

    case opmoti:
      cloneNode = homomorphismArray (aNode);
      break;

    case opbund:
      cloneNode = homomorphismBundle (aNode);
      break;

    case opldefpart:
      cloneNode = homomorphismPartialDefs (aNode);
      break;

    case opdefpart:
      cloneNode = homomorphismPartialDef (aNode);
      break;

    default:
      pK_messenger->errorMethodFile ("nodeHomomorphism", "AbstractHomomorrphism");
    }
  return cloneNode;
}

/**
   It applies homomorphismDependence() for each node of ln.
   @param ln (in) is a list of elements of the first original.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDependences (ptlINode ln)
{
  ptINode aNode;

  while (ptlINode_get_elem (&ln, &aNode))
    homomorphismDependence (aNode);
}

/**
   It applies homomorphismDependence(ptinsth ) for each node of ln.
   @param lh (in) is a list of elements of the first original.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDependences (ptlinsth lh)
{
  ptinsth h;

  while (ptlinsth_get_next (&lh, &h))
    homomorphismDependence (h);
  homomorphismDependence (fCKM->getInitClock ());
}

/**
   It applies homomorphismDependence(ptINode) for each node of a signal of lsig.
   @param lsig is a list of elements of the first original.
  
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismdependences (ptlinstances lsig)
{
  ptexpcal vsig;

  while (ptlinstances_get_next (&lsig, &vsig))
    homomorphismDependence (fEM->getNode (vsig));
}

/**
   It sets the pragmas of the original graph to the generated graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismPragmas ()
{
  ptlpragmas lp;
  ptpragma vp, np;
  ptrepstr name;
  ptObjPragma obj;
  ntrees stmt;

  lp = fGraph->getPragmas ();
  while (ptlpragmas_get_next (&lp, &vp))
    {
      getPragma (vp, &name, &obj, &stmt);
      np = newpragma (name, getCurrentInstances ((ptlinstances) obj), copyEnonce (stmt));
      sGraph->addPragma (np);
    }
}

/**
   It defines the name of the generated graph by the name of the original one catenated with a suffix (suff).
*/
_VIRTUAL_ void
AbstractHomomorphism::setNameNewGraph (ptident suff)
{
  ptident id;

  id = fGraph->getName ();
  id = makeIdent_ident_ident (id, suff);
  sGraph->setName (id);
}


/**
   It initializes the translation of the objets of the g1 graph (original) into the generated one (g2).
   - the graphs are visibles in the class by fGraph(g1) and sGraph (g2).
   - the managers of g1 (resp. g2) becomes accessible in the class.
   - attribute numbers are initialized.
   - a \ref GraphBuilder "graph builder" is created for g2.
*/
_VIRTUAL_ void
AbstractHomomorphism::initialize (ptDCG g1, ptDCG g2)
{
  g2->setKind (g1->getKind ());

  fgraph_type = g1->getFormatLevel ();
  fGraph = g1;
  fEM = g1->getExpressionManager ();
  fCKM = g1->getClockManager ();
  fNM = g1->getNodeManager ();
  fNameManager = g1->getNameManager ();

  sGraph = g2;
  sEM = g2->getExpressionManager ();
  sCKM = g2->getClockManager ();
  sNM = g2->getNodeManager ();
  sNameManager = g2->getNameManager ();

  sCKM->setModeSynchroSolving (pk_true);

  attributeNumber = getMark ();
  unNodes = g1->getUnused (&unClocks);
  BDEBUG = pk_false;
  error = pk_false;
  hundef = sCKM->getUndefClock ();	// universal
  igc = new intemkIGraph (g2);	// target graph builder.
  lConstraintsNullClock=NILLINSTH;
  fCKM->getNullClockConstraints (&lConstraintsNullClock);
  lAssertionsNullClock=NILLINSTH;
  fCKM->getNullClockAssertions (&lAssertionsNullClock);

}

/**
   It applies homomorphismDclIdentifier(ptexpcal) for each element of the
   list of expression with op as operator (assumed to be identifiers)
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDclIdentifiers (noperators op)
{
  ptexpcal vsig;
  ptlinstances ls;

  if (fEM->getExprList (op, &ls))
    while (ptlinstances_get_next (&ls, &vsig))
      homomorphismDclIdentifier (vsig);
}

/**
   It applies homomorphismDclIdentifier(noperators) for the "opident" operator.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDclIdentifiers ()
{
  homomorphismDclIdentifiers (opident);
}

/**
   It applies homomorphismUndefinedIdentifier() for each element of ls (assumed to be
   some undefined signals).
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismUndefinedIdentifiers (ptlinstances ls)
{
  ptexpcal vsig;

  while (ptlinstances_get_next (&ls, &vsig))
    homomorphismUndefinedIdentifier (vsig);
}

/**
   It applies homomorphismDclClock(ptinsth) for each clock of the system.
   It deals of the declaration part.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismDclClocks ()
{
  ptlinsth lh;
  ptinsth h;

  lh = ptlinsth_duplicate (fCKM->getlhroots ());
  lh = ptlinsth_sorted_fusion (lh, fCKM->getlhplaced ());
  while (ptlinsth_get_next (&lh, &h))
    {
      h = fCKM->getRepresentativeFormula (h);
      homomorphismDclClock (h);
    }
}


/**
   It applies homomorphismDclClock(ptinsth) for each clock of the system.
   It deals of the definition part.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismClocks (ptlinsth lh)
{
  ptinsth h;

  while (ptlinsth_get_next (&lh, &h))
    {
      h = fCKM->getRepresentativeFormula (h);
      homomorphismClock (h);
    }

}


/** 
    It translates a node (aNode) of the original graph (assumed), into another one (n) of the generated one.
    @return "true" if the operation succeeds (n is available), "false" otherwise.
    It calls
    - isSelectedNode(), 
    - beginHomomorphismNode(),
    - nodeHomomorphism();
    - then endHomomorphismNode 
    to achieve this tranlsation.
    \note We suppose that a node is translated to another node.
*/
_VIRTUAL_ boolean
AbstractHomomorphism::homomorphismNode (ptINode aNode, ptINode * n)
{
  if (!isSelectedNode (aNode))
    return pk_false;

  if (!getNodeAttribute (aNode, n))
    {
      beginHomomorphismNode (aNode);
      *n = nodeHomomorphism (aNode);
      endHomomorphismNode (aNode, *n);
    }
  return pk_true;
}

/**
   It applies homomorphismNode() for each node of a list of nodes (ln).
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismNodes (ptlINode ln)
{
  ptINode aNode, n;
  ptlINode ln1;

  ln1 = ln;
  while (ptlINode_get_elem (&ln, &aNode))
    if (!fNM->is_setOfPartialDefinitions (aNode))
      homomorphismNode (aNode, &n);

  ln = ln1;
  while (ptlINode_get_elem (&ln, &aNode))
    if (fNM->is_setOfPartialDefinitions (aNode))
      homomorphismNode (aNode, &n);

}

/**
   It updates the interface of the generated graph: the image of the inputs/outputs signals becomes
   the inputs/outputs of the generated graph.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismInterface ()
{
  ptlinstances lin, lout, lin1, lout1;

  lin1 = fGraph->getInputs ();
  lin = getCurrentInstances (lin1);
  sGraph->setInputs (lin);
  lout1 = fGraph->getOutputs ();
  lout = getCurrentInstances (lout1);
  sGraph->setOutputs (lout);
}

/**
   It applies homomorphismAttribute() for each identifiers of a list of signals (ls).
   It is to fix links between the signals of the original graph and the signals of the new one (using attribute
   for example).
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismAttributes (ptlinstances ls)
{
  ptexpcal vsig;

  while (ptlinstances_get_next (&ls, &vsig))
    if (isSelectedSignal (vsig))
      homomorphismAttribute (vsig);
}

/**
   It applies homomorphismAttribute() for each identifiers of the original graph.
   It is to fix links between the signals of the original graph and the signals of the new one (using attribute
   for example).
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismAttributes ()
{
  ptlinstances ls;

  if (fEM->getExprList (opident, &ls))
    homomorphismAttributes (ls);
}

/**
   It applies homomorphismConstraint(ptinsth, ptinsth) for constraints of placed formulas.
*/
_VIRTUAL_ void
AbstractHomomorphism::homomorphismConstraints ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  lh1 = fCKM->getPlacedNullClocks ();
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      if (getClockAttribute (vh1, &vh2))
	sCKM->addPlacedNullClocks (vh2);
    }
  fCKM->getConstraintsOnPlacedClocks (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      homomorphismConstraint (vh1, vh2);
    }
}


/**
   It adds all undefined identifiers, of the new graph, as inputs.
*/
_VIRTUAL_ void
AbstractHomomorphism::addInputs ()
{
  ptlinstances ls, ladd;
  ptexpcal vsig;

  if (!sEM->getExprList (opident, &ls))
    return;
  ladd = NILLINSTANCES;
  while (ptlinstances_get_next (&ls, &vsig))
    if (sEM->IsSignal (vsig))
      if (!isIOSignal (vsig))
	if (sNM->is_undefined (sEM->getNode (vsig)))
	  {
	    sGraph->addInput (vsig);
	    ladd = ptlinstances_insert (ladd, vsig);
	  }
  igc->addInNodes (ladd);
}

/**
   It sets to b the internal tag error.
*/
_VIRTUAL_ void
AbstractHomomorphism::setTagError (boolean b)
{
  error = b;
}

/**
   @return the value of the internal tag error.
*/
_VIRTUAL_ boolean
AbstractHomomorphism::isOnError ()
{
  return error;
}

/**
   @return "true" if aSig is a I/O signal of the original graph, "false" otherwise.
*/
boolean AbstractHomomorphism::isIOSignal (ptexpcal aSig)
{
  if (ptlinstances_isin_exp (sGraph->getInputs (), aSig))
    return pk_true;
  if (ptlinstances_isin_exp (sGraph->getOutputs (), aSig))
    return pk_true;
  return pk_false;
}

/**
   @return "true" if the image (n) in the generated graph of a node (aNode) of the original graph exists, "false" otherwise.
*/
boolean AbstractHomomorphism::getNodeAttribute (ptINode aNode, ptINode * n)
{
  return fNM->getAttribute (aNode, attributeNumber, (attributeValue *) n);
}

/**
   It assigns to a node (aNode), assumed to be of the original graph, its clone (n) in the produced graph.
*/
void
AbstractHomomorphism::setNodeAttribute (ptINode aNode, ptINode n)
{
  fNM->addAttribute (aNode, attributeNumber, (attributeValue) n);
}

/**
   @return "true" if the image (nsig) in the generated graph of a signal (sig) of the original graph exists, "false" otherwise.
*/
boolean AbstractHomomorphism::getSigAttribute (ptexpcal sig, ptexpcal * nsig)
{
  return fEM->getAttribute (sig, attributeNumber, (attributeValue *) nsig);
}

/**
   It assigns to a signal (sig), assumed to be of the original graph, its clone (nsig) in the produced graph.
*/
void
AbstractHomomorphism::setSigAttribute (ptexpcal sig, ptexpcal nsig)
{
  fEM->addAttribute (sig, attributeNumber, (attributeValue) nsig);
  //  fEM->setLinkObject2Graphs(sig, nsig, sGraph);
}

/**
   @return "true" if the image (nh) in the generated graph of a clock (h) of the original graph exists, "false" otherwise.
*/
boolean AbstractHomomorphism::getClockAttribute (ptinsth h, ptinsth * nh)
{
  return fCKM->getAttribute (h, attributeNumber, (attributeValue *) nh);
}

/**
   It assigns to a clock(h), assumed to be of the original graph, its clone (nh) in the produced graph.
*/
void
AbstractHomomorphism::setClockAttribute (ptinsth h, ptinsth nh)
{
  fCKM->addAttribute (h, attributeNumber, (attributeValue) nh);
  //  fCKM->setLinkObject2Graphs(h, nh, sGraph);
  sCKM->setName (nh, fCKM->getName (h)); // Copy of the original name March 2016 LBesnard
}


/**
   It returns the Signal string of the type t.
*/
ptrepstr AbstractHomomorphism::__getTypeStr (TypeType t)
{
  if (TypeTestPredef (t))
    {
      switch ((int) t)
	{
	case PureType:
	  return string2repstr ("event");
	case BooleanType:
	  return string2repstr ("logical");
	case ShortType:
	  return string2repstr ("short");
	case IntegerType:
	  return string2repstr ("integer");
	case LongType:
	  return string2repstr ("long");
	case RealType:
	  return string2repstr ("real");
	case DRealType:
	  return string2repstr ("dpreal");
	case ComplexType:
	  return string2repstr ("complex");
	case DComplexType:
	  return string2repstr ("dcomplex");
	case CharType:
	  return string2repstr ("char");
	case StringType:
	  return string2repstr ("string");
	case ByteType:
	case WordType:
	case ErrorType:
	case LabelType:
	  return nilstr;
	}
    }
  if (TypeTestExternClass (t))
    return ExternTypeGetName (t);
  if (TypeTestMetaClass (t))
    return MetaTypeGetName (t);
  if (TypeTestEnumerationClass (t))
    return EnumerationTypeGetName (t);
  if (TypeTestNamedClass (t))
    return NamedTypeGetName (t);
  return string2repstr ("XXXTYPE");

  /*  TypeTestStructureClass --> nameTyped for a structure type declaration.
     TypeTestBundleClass --> nameTyped for a structure type declaration.
     if(TypeTestSubrangeClass(t)) return cNOTIMPLEMENTED();
     if(TypeTestBitsClass(t)) return cNOTIMPLEMENTED();
     if(TypeTestLambdaClass(t))  return cNOTIMPLEMENTED();
     if(TypeTestNodeClass(t))  return cNOTIMPLEMENTED();
     if(TypeTestGenericNodeClass(t))  return cNOTIMPLEMENTED();
     if(TypeTestMetaNodeClass(t)) return cNOTIMPLEMENTED();
     if(TypeTestMatrixClass(t))  return cNOTIMPLEMENTED();;
     if(TypeTestTupleClass(t))  return cNOTIMPLEMENTED();
   */
}

/* 
   It returns the Signal string of the type t prefixed by _.
   Used to build a enumerated type.
*/
ptrepstr AbstractHomomorphism::getTypeStrAsEnumVal (TypeType t)
{
  ptrepstr
    vstr;
  vstr = __getTypeStr (t);
  strconc2str (string2repstr ("_"), vstr, pk_false, &vstr);
  return vstr;
}

/**
   @return "true" if the graph is a graph with events (DCG_BASIC, DCG_POLY, or DCG_ENDO), "false" otherwise.
*/
boolean AbstractHomomorphism::isEventClockedFormat ()
{
  return ((fgraph_type == DCG_BASIC) || (fgraph_type == DCG_POLY) || (fgraph_type == DCG_ENDO));
}

/**
   Debuging.
*/
void
AbstractHomomorphism::TRACE (const char *mess)
{
  //  printf("DEBUG ---DEBUG --- %s\n", mess);
}
