/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor. */
  Duplicator::Duplicator ()
{
}


/** Destructor.*/
_VIRTUAL_ Duplicator::~Duplicator ()
{
  bExportHierarchy = pk_true;
}

/**
   @return the mark used to keep the link between objects of the original graph with the objects of the new one.
   (This attribute is not cleaned at the end of the translation: it becomes a permanent attribute.
*/
typeMark
Duplicator::getUsedMark ()
{
  return attributeNumber;
}


/**
   @return a copy of the operand graph (g).
   It calls copySelected() for a global selection.
*/
ptDCG Duplicator::copy (ptDCG g)
{
  return copySelected (g, DUPLICATEALL, pk_true);
}

/**
   @return a copy of the operand graph (g), but the hierarchy is not imported.
   It calls copySelected() for a global selection.
*/
ptDCG Duplicator::copyWithoutHierarchy (ptDCG g)
{
  return copySelected (g, DUPLICATEALL, pk_false);
}


/**
   It copies a graph (g1) in another one (g2). The hierarchy is exported in the copy if bExportHierarchy is "true".
   It calls 
   - the initializing step (initialize()), 
   - the copy (copyGraph())
   - then, the finalizing (finalize()).

   Note that a link between the objects of the graph g1 and the objects of the graph g2 is kept. It is set by the ExprManager::setLinkObject2Graphs().
   The user may use the method ExprManager::getLinkObject2Graphs() to get the link value.
*/
void
Duplicator::copyIn (ptDCG g1, ptDCG g2, boolean bExportHierarchy)
{
  initialize (g1, g2, bExportHierarchy);
  copyGraph ();
  finalize ();
}

/**
   @return a copy of the operand graph (g) using a mark (n) for selecting the objects 
   The objects are assumed marked before this call except if n=0 (n=0 stands for all objects are selected). 
   The operation of copy is defined in mkCopy(). The hierarchy is exported in the copy if bExportHierarchy is "true".
*/
ptDCG Duplicator::copySelected (ptDCG g, typeMark n, boolean bExportHierarchy)
{
  selectMark = n;
  return mkCopy (g, bExportHierarchy);
}


/**
   @return a copy of the operand graph (g) using a mark (n) for selecting the objects.
   The objects are assumed marked before this call.
   The copy is defined in mkCopy() method. The copy is assumed to be a copy of the sub-hierarchy with vhroot as the clock root.
   (All the selected nodes are assumed to be in the sub-hierarchy, the interface of the root clock is assigned to the returned graph).
*/
ptDCG Duplicator::copyHierarchy(ptDCG g, typeMark n, ptinsth vhroot)
{
  ptDCG ng;
  ng = copySelected (g, n, pk_false);
  replaceInterface(ng, fCKM->getInterfaceSubHierarchy (vhroot));
  return ng;
}
