/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return the homomorphism of the type of a clock (vsig) in the new graph.
*/
  _VIRTUAL_ TypeType Homomorphism::homomorphismType (ptinsth clk)
{
  return homomorphismType (get_PureType ());
}

/**
   @return a clock (h) as a signal (ie. true when h).
*/
_VIRTUAL_
  ptexpcal
Homomorphism::clock2expr (ptinsth h)
{
  return sEM->mkWhenOrCellExp (opextr, sEM->getTrue (), h);
}


/** 
    Without effect:  we suppose that the system (and its hierachy) has been 
    imported in the produced graph. 
*/
_VIRTUAL_ void
Homomorphism::homomorphismHierarchy (ptlinsth lh)
{
}


/**
   @return "true" if the clock is selected for the translation, "false" otherwise.
   It is selected iff it is not in the unused clocks of the original graph.
*/
_VIRTUAL_ boolean
Homomorphism::isSelectedClock (ptinsth h)
{
  return !ptlinsth_isin_elem (unClocks, h);
}

/** 
    @return the copy of the clock clk in the current builting graph.
    This clock have been built before the "cost homomorphism".
*/
_VIRTUAL_ ptinsth
Homomorphism::getClockCopy (ptinsth clk)
{
  ptinsth nclk;
  if (fCKM->getAttribute (clk, usedMarkBefore, (attributeValue *) & nclk))
    return nclk;
  return sCKM->getUndefClock ();	// else pb;
}

/**
   @return the created signal associated with a clock (clk). This signal is available via an attribute of the clock.
*/
_VIRTUAL_ ptexpcal
Homomorphism::assignAttribute (ptinsth clk)
{
  ptexpcal e;
  ptident name;

  name = homomorphismIdent (clk);
  e = sEM->mkIdentExp (name, tsignal, homomorphismType (clk), sNM->mk_node_undefined (), sCKM->getUndefClock (), MK_ANNOT_NIL());
  sEM->setClock (e, getClockCopy (clk));
  fCKM->addAttribute (clk, attributeNumber, (attributeValue) e);
  return e;
}

/**
   It assigns the homomorphism signal (attribute) to a clock (clk, clock of the original graph)
   and also to its copy in the new graph.
*/
_VIRTUAL_ void
Homomorphism::homomorphismDclClock (ptinsth clk)
{
  ptexpcal e;
  ptinsth nclk;

  clk = fCKM->getRepresentativeFormula (clk);
  if (isSelectedClock (clk))
    if (fCKM->getAttribute (clk, usedMarkBefore, (attributeValue *) & nclk))
      // Existe dans le graphe du morphisme
      if (!fCKM->getAttribute (clk, attributeNumber, (attributeValue *) & e))
	{
	  e = assignAttribute (clk);
	  imageSetAttributeClock (nclk, e);
	}
}


/**
   new graph objects: a clock (nclk) of the new graph is linked with its 
   boolean representation (vattr) in this new graph.
*/
_VIRTUAL_ void
Homomorphism::imageSetAttributeClock (ptinsth nclk, ptexpcal vattr)
{
  sCKM->addAttribute (nclk, attributeNumber, (attributeValue) vattr);
}


/**
   @return the boolean representation in this new graph of a clock (nclk) of this new graph.
*/
_VIRTUAL_ ptexpcal
Homomorphism::imageGetAttributeClock (ptinsth nclk)
{
  ptexpcal e;
  if (sCKM->getAttribute (nclk, attributeNumber, (attributeValue *) & e))
    return e;
  return NILEXPCAL;		// ne devrait pas produire
}

/** @return "true" if the homomorphism-signal of a clock (clk) is available, "false" otherwise (the information is not 
    available via the attribute and clk is not a variable).
 */
_VIRTUAL_ boolean
Homomorphism::getAttribute (ptinsth clk, ptexpcal * hs)
{
  ptinsth nclk;
  // Attention au "tick"
  if (fCKM->satisfiesNullClock (clk))
    {
      *hs = NILEXPCAL;
      return pk_true;
    }				// pas de reference (blindage)
  if (fCKM->getAttribute (clk, attributeNumber, (attributeValue *) hs))
    return pk_true;
  if (!fCKM->satisfiesVari (clk))
    return pk_false;
  *hs = assignAttribute (clk);
  if (fCKM->getAttribute (clk, usedMarkBefore, (attributeValue *) & nclk))
    imageSetAttributeClock (nclk, *hs);
  return pk_true;
}

/**
   It defines the morphism of a formula (h) that represents a free variable.
   The associated boolean signal (x) is defined by x:= pk_true when M(h).
*/
_VIRTUAL_ void
Homomorphism::homoVari (ptinsth h)
{
  ptexpcal homoSignal;

  h = fCKM->getRepresentativeFormula (h);
  getAttribute (h, &homoSignal);
  homomorphismExp (homoSignal, fEM->getTrue (), getClockCopy (h));
}

/**
   It defines the morphism of a formula (h) that represents an extration by a positive condition (C).
   The associated boolean signal (x) is defined by x:= C when M(^C).
*/
_VIRTUAL_ void
Homomorphism::homoExtract (ptinsth h)
{
  ptexpcal Cond, homoSignal;
  ptinsth clkCond;

  getAttribute (h, &homoSignal);
  Cond = fCKM->getConditionSub (h);
  Cond = fEM->mkUnaryExpTp (opwhen1, Cond, fEM->getBackAnnotation (Cond));
  clkCond = fCKM->getClockSub (h);
  clkCond = fCKM->getRepresentativeFormula (clkCond);
  clkCond = getClockCopy (clkCond);
  homomorphismExp (homoSignal, Cond, clkCond);
}


/**
   It defines the morphism of a formula (h) that represents an extration by a negative condition (not C).
   The associated boolean signal (x) is defined by x:= not C when M(^C).
*/
_VIRTUAL_ void
Homomorphism::homoExtractNega (ptinsth h)
{
  ptexpcal homoSignal, Cond;
  ptinsth clkCond;

  getAttribute (h, &homoSignal);
  Cond = fCKM->getConditionSub (h);
  Cond = fEM->mkUnaryExp (opnon, Cond, fEM->getBackAnnotation (Cond));
  Cond = fEM->mkUnaryExpTp (opwhen1, Cond, fEM->getBackAnnotation (Cond));
  clkCond = fCKM->getClockSub (h);
  clkCond = fCKM->getRepresentativeFormula (clkCond);
  clkCond = getClockCopy (clkCond);
  homomorphismExp (homoSignal, Cond, clkCond);
}

/**
   @return the morphism of a binary clock (original graph) by scanning the associated translation defined 
   via a scanner object restricted to the outputs. 
   @param ScanPragObj (in) a scanner ready to analyze the translation of the outputs
   @param h (in) binary clock; current translated formula (original node)
   @param h1 (in)  the first operand of h
   @param h2 (in)  the second operand of h

   It takes into account the following tokens (see morphismPredefinedNames.h): 
   - tk_input, tk_output, tk_homo (&h). &h must be followed by &{i,o,p....}
   - end scanning: tk_2pe
   - separator: tk_virg
*/
_VIRTUAL_ ptexpcal
Homomorphism::homoClockBinaryOutputs (ptStrScanner ScanPragObj, ptinsth h, ptinsth h1, ptinsth h2)
{
  boolean bHomo, b, stop;
  int iOp;
  ptexpcal homoSignal, CurrentOutputs;
  ScanPragmaToken tk;

  CurrentOutputs = NILEXPCAL;
  getAttribute (h, &homoSignal);
  CurrentOutputs = addExp (homoSignal, CurrentOutputs);
  return CurrentOutputs;

  // Outputs def
  bHomo = pk_false;
  do
    {
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_homo:		/* &h homomorphism, must be followed by &{i,o,p....} */
	  bHomo = pk_true;
	  break;

	case tk_input:		/* &i */
	  if (bHomo)
	    {
	      b = ScanPragObj->getInt (&iOp);
	      if (b)
		{
		  if (iOp == 1)
		    getAttribute (h1, &homoSignal);
		  else if (iOp == 2)
		    getAttribute (h2, &homoSignal);
		  else
		    b = pk_false;	// Error
		  if (b)
		    CurrentOutputs = addExp (homoSignal, CurrentOutputs);
		}
	      bHomo = pk_false;
	    }
	  else;			// Non traite

	  break;
	case tk_output:	/* &o */
	  if (bHomo)
	    {
	      b = ScanPragObj->getInt (&iOp);
	      if (b)
		{
		  if (iOp == 1)
		    getAttribute (h, &homoSignal);
		  else
		    b = pk_false;
		  if (b)
		    CurrentOutputs = addExp (homoSignal, CurrentOutputs);
		}
	      bHomo = pk_false;
	    }
	  else;			// Non traite
	  break;
	case tk_clock:		// A completer

	  break;

	case tk_type:
	case tk_param:
	case tk_name:
	case tk_dim:
	case tk_end:
	case tk_error:
	case tk_ident:
	  stop = pk_true;
	  setTagError (pk_true);
	  break;

	case tk_2pe:		//   := 
	  stop = pk_true;
	  break;
	default:		// tk_virg, tk_po, tk_pf, tk_ao, tk_af
	  ;
	}
    }
  while (!stop);
  return CurrentOutputs;
}

/**
   @return the morphism of a binary clock (original graph) by scanning the associated translation defined 
   via a scanner object restricted to the parameters. 
   @param ScanPragObj (in) a scanner ready to analyze the translation of the parameters.
   @param h (in) binary clock; current translated formula (original node)
   @param h1 (in)  the first operand of h
   @param h2 (in)  the second operand of h

   It takes into account the following tokens (see morphismPredefinedNames.h): 
   - tk_type, tk_dim
   - end scanning: tk_af
   - separator: tk_virg
*/
_VIRTUAL_ ptexpcal
Homomorphism::homoClockBinaryParameters (ptStrScanner ScanPragObj, ptinsth h, ptinsth h1, ptinsth h2)
{
  ptexpcal nexp, hexp;
  TypeType TT;
  ScanPragmaToken tk;

  tk = ScanPragObj->getToken ();
  switch (tk)
    {
    case tk_af:		// }   : End of parameters
      return NILEXPCAL;

    case tk_virg:
      return homoClockBinaryParameters (ScanPragObj, h, h1, h2);

    case tk_type:
      nexp = homoClockBinaryParameters (ScanPragObj, h, h1, h2);
      TT = get_PureType ();
      hexp = sEM->mkExpType (TT);
      return addExp (hexp, nexp);

    case tk_dim:		// Pas de sens (event)
      nexp = homoClockBinaryParameters (ScanPragObj, h, h1, h2);
      hexp = NILEXPCAL;
      return addExp (hexp, nexp);

    default:			/*
				   case tk_homo: case tk_input: case tk_output: case tk_clock:
				   case tk_param: case tk_name: case tk_end:
				   case tk_error:  case tk_ident: case tk_2pe:
				   case tk_po: case tk_pf: case tk_ao:  */
      setTagError (pk_true);
      return NILEXPCAL;
      break;
    }
}


/**
   @return the morphism of a binary clock (original graph) by scanning the associated translation defined 
   via a scanner object restricted to the inputs. 
   @param ScanPragObj (in) a scanner ready to analyze the translation of the inputs
   @param h (in) binary clock; current translated formula (original node)
   @param h1 (in)  the first operand of h
   @param h2 (in)  the second operand of h

   It takes into account the following tokens (see morphismPredefinedNames.h): 
   - tk_homo (&h). &h must be followed by &{i,o,p....}
   - end scanning: tk_pf
   - separator: tk_virg
*/
_VIRTUAL_ ptexpcal
Homomorphism::homoClockBinaryInputs (ptStrScanner ScanPragObj, ptinsth h, ptinsth h1, ptinsth h2)
{
  boolean b;
  int iOp;
  ptexpcal homoSignal, nexp;
  ScanPragmaToken tk;

  tk = ScanPragObj->getToken ();
  switch (tk)
    {
    case tk_pf:
      return NILEXPCAL;
    case tk_virg:
      return homoClockBinaryInputs (ScanPragObj, h, h1, h2);

    case tk_homo:		/* &h homomorphism, must be followed by &{i,o,p....} */
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  b = ScanPragObj->getInt (&iOp);
	  nexp = homoClockBinaryInputs (ScanPragObj, h, h1, h2);
	  if (b)
	    {
	      if (iOp == 1)
		getAttribute (h1, &homoSignal);
	      else if (iOp == 2)
		getAttribute (h2, &homoSignal);
	      else
		b = pk_false;	// Error
	      if (b)
		nexp = addExp (homoSignal, nexp);
	      else		/* error */
		;
	    }
	  return nexp;
	  break;
	case tk_output:	/* &o */
	  b = ScanPragObj->getInt (&iOp);
	  nexp = homoClockBinaryInputs (ScanPragObj, h, h1, h2);
	  if (b)
	    {
	      if (iOp == 1)
		getAttribute (h, &homoSignal);
	      else
		b = pk_false;
	      if (b)
		nexp = addExp (homoSignal, nexp);
	      else		/* error */
		;
	    }
	  return nexp;
	  break;
	case tk_clock:		/* &c  --Not implemented --- SKIP-- */
	  nexp = homoClockBinaryInputs (ScanPragObj, h, h1, h2);
	  return nexp;
	  break;
	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}
    default:			/* tk_end, tk_po, tk_ao, tk_af, tk_2pe, tk_type, tk_param, tk_name, tk_dim, tk_error, tk_ident */
      setTagError (pk_true);
      return NILEXPCAL;
      break;
    }
}

/**
   It builds the definition of the signal-homomorphism for the formula defined by <em> h = h1 vop h2 </em>.
   The translation is given by a definition in the table of morphism (^+, ^*, ^- operators). The translation 
   is guided by a scanner on the definition. It calls:
   - homoClockBinaryOutputs() for the translation of the outputs
   - homoClockBinaryParameters() for the translation of the parameters
   - homoClockBinaryInputs() for the translation of the inputs
   .
   The results are used to build in the new graph the definition of the signal-homomorphism.
   \code 
   Example: 
   operator ^+ = (?x,y; !z; ) pragmas MORPHISM "(&h&o, &done):=CostClkPlus{&t}(&h&i1, &h&i2, &hck&o, &wait)" end pragmas;
   - homoClockBinaryOutputs() analyzes (&h&o, &done)
   - homoClockBinaryParameters() analyzes {&t}
   - homoClockBinaryInputs() analyzes (&h&i1, &h&i2, &hck&o, &wait)
   \endcode
*/
_VIRTUAL_ void
Homomorphism::homoClockBinary (noperators vop, ptinsth h, ptinsth h1, ptinsth h2)
{
  ptStrScanner ScanPragObj;
  ptINode n;
  ptrepstr vstr;
  ptident CurrentIdCall;
  ptexpcal CurrentParameters, CurrentInputs, CurrentOutputs;
  ScanPragmaToken tk;

  if (scannerOnOp (vop, &ScanPragObj))
    {
      // Outputs
      CurrentOutputs = homoClockBinaryOutputs (ScanPragObj, h, h1, h2);
      // Process name
      if (ScanPragObj->getIdent (&vstr))
	CurrentIdCall = repstr2ident (vstr);
      else
	CurrentIdCall = NILIDENT;
      // Parameters
      tk = ScanPragObj->getToken ();
      if (tk == tk_ao)		// { 
	CurrentParameters = homoClockBinaryParameters (ScanPragObj, h, h1, h2);
      else
	CurrentParameters = NILEXPCAL;
      // Inputs
      tk = ScanPragObj->getToken ();	// tk_po
      if (tk == tk_po)
	CurrentInputs = homoClockBinaryInputs (ScanPragObj, h, h1, h2);
      else
	CurrentInputs = NILEXPCAL;
      n = sNM->mk_procCall (CurrentIdCall, CurrentParameters, CurrentInputs, CurrentOutputs, NILBACKANNOTATION);
      insertNode (n, h);
      delete ScanPragObj;
    }
}

/* It builds the definition of the signal-homomorphism for the formula (formula). It calls
   - homoVari() for free variable,
   - homoExtract() for a positive extraction,
   - homoExtractNega() for a negative extraction,
   - homoClockBinary() for upper bound, lower bound and complemtary operator
 */
_VIRTUAL_ void
Homomorphism::formulaHomomorphism (ptinsth formula)
{
  ptinsth h1, h2;

  switch (fCKM->getOperator (formula))
    {
    case sub:
      homoExtract (formula);
      break;

    case nosub:
      homoExtractNega (formula);
      break;

    case infer:
      if (fCKM->getMinusArgs (formula, &h1, &h2))
	homoClockBinary (opcomp, formula, h1, h2);
      else
	{
	  fCKM->getBinaryArgs (formula, &h1, &h2);
	  homoClockBinary (opbinf, formula, h1, h2);
	}
      break;

    case super:
      fCKM->getBinaryArgs (formula, &h1, &h2);
      homoClockBinary (opbsup, formula, h1, h2);
      break;

    case hcomp:
      fCKM->getUnaryArg (formula, &h1);
      h2 = fCKM->getBranching (formula, h1);
      homoClockBinary (opcomp, formula, h1, h2);
      break;

    case vari:
      homoVari (formula);
      break;

    case hnot:			// pas d'occurrences
    case predef:
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}

/**
   It produces for a "selected" clock (h) of the original graph, the definition of the its signal-homomorphism in the new graph.
   It is selected if the clone of h exists in the new graph.
*/
_VIRTUAL_ void
Homomorphism::homomorphismClock (ptinsth h)
{
  ptinsth hCostGraph;
  boolean G_internalNode_save;

  h = fCKM->getRepresentativeFormula (h);
  if (!fCKM->getAttribute (h, usedMarkBefore, (attributeValue *) & hCostGraph))
    return;
  if (!isSelectedClock (h))
    return;
  hCostGraph = h;
  G_internalNode_save = G_internalNode;
  G_internalNode = pk_true;
  formulaHomomorphism (hCostGraph);
  G_internalNode = G_internalNode_save;
  /* --- Class info ---
     ptinsth h1;
     h1 = fCKM->getNextFormula(h, pk_true);
     while (h1 != h)
     {
     formulaHomomorphism(h1);
     h1 = fCKM->getNextFormula(h1, pk_true);
     }
   */
}
