/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @ingroup initHomoGroup
   It initializes
   - the name (idDependence, attribute member of the Homomorphism class) of the process associated with the dependence operator.
   - the value of the tag keepDependenceOp (attribute member of the Homomorphism class): "true" when "-->" is the name of t he operator
   associated with the dependence operator, "false" otherwise.
*/
  void
Homomorphism::getDependenceCallId ()
{
  ptrepstr vstr;
  boolean stop;
  ScanPragmaToken tk;
  ptStrScanner ScanPragObj;

  if (scannerOnOp (opdepd, &ScanPragObj))
    {
      vstr = nilstr;
      stop = pk_false;
      do
	{
	  tk = ScanPragObj->getToken ();
	  switch (tk)
	    {
	    case tk_ident:
	      strconc2str (vstr, ScanPragObj->getIdent (), pk_false, &vstr);
	      break;
	    case tk_end:
	    case tk_error:
	      stop = pk_true;
	      break;
	    default:
	      ;
	    }
	}
      while (!stop);
      idDependence = repstr2ident (vstr);
      keepDependenceOp = eq_idents (idDependence, string2ident ("-->"));
    }
}


/**
   @return an output of a node (node): 
   - for a subgraph, it is the label (convention)
   - when there are many outputs, it takes the first element of the set,
   - otherwise, the output of the node

   @param node (in/out) a node, when it is a subgraph node, node (out) is the node of the label, otherwise it is not modify.
*/
_VIRTUAL_ ptexpcal
Homomorphism::getOutOf (ptINode * node)
{
  ptexpcal e, out;

  // Restriction
  if (sNM->is_subgraph (*node))
    {
      sNM->get_label (*node, &e);
      getAttribute (e, &out);
      *node = sEM->getNode (out);
    }
  else
    {
      out = sNM->get_leftOperand (*node);
      if (sEM->satisfiesList (out))
	{
	  sEM->getNaryArgs (&out, &e);
	  out = e;
	}
    }
  return out;
}


/**
   @param vfrom (in) node of the new graph
   @param vto (in) node of the new graph
   @param h (in) dependence clock, in the new graph, of the dependence vfrom --> vto.

   It adds a node associated with the dependence vfrom --> vto when h in the new graph.
    x := idDependence (outputOf(vfrom), outputOf(vTo), h)
*/
_VIRTUAL_ void
Homomorphism::insertDependenceInDef (ptINode vfrom, ptINode vto, ptinsth h)
{
  ptident id;
  ptexpcal CurrentParameters, CurrentInputs, CurrentOutputs, outVto, outVfrom, out;
  ptINode n;
  ptlinstances l;

  outVfrom = getOutOf (&vfrom);
  outVto = getOutOf (&vto);

  //  out = newSignalAt(nilId, sEM->getClock(outVto));
  out = newSignalAt (nilId, h);

  if (sNM->is_readNode (vto))
    {
      // Changement de l'entrée avec conservation du nom original
      id = sEM->getName (outVto);
      sEM->setName (outVto, sEM->getName (out));
      sEM->setName (out, id);
      l = sGraph->getInputs ();
      l = ptlinstances_remove_exp (l, outVto);
      l = ptlinstances_insert (l, out);
      sGraph->setInputs (l);
    }

  if (sCKM->satisfiesUndefClock (h))
    h = sNM->get_nodeClock (vto);
  /*
     sNM->get_rightOperand(vto, &inputs);
     sEM->getNaryArgs(&inputs, &finput); // !inputs =< E2,...,En>
     sNM->set_rightOperand(vto, sEM->mkNaryExp(oplesign, out, inputs, NILBACKANNOTATION));
   */
  // Creation du noeud tenant compte de la dependence
  // Outputs
  CurrentOutputs = addExp (out, NILEXPCAL);
  // Parameters
  CurrentParameters = NILEXPCAL;
  // Inputs
  CurrentInputs = addExp (clock2expr (h), NILEXPCAL);
  CurrentInputs = addExp (outVto, CurrentInputs);	//  CurrentInputs=addExp(finput, CurrentInputs);
  CurrentInputs = addExp (outVfrom, CurrentInputs);
  // La suite sera a adpater pour accepter autre chose qu'un appel
  n = sNM->mk_procCall (idDependence, CurrentParameters, CurrentInputs, CurrentOutputs, NILBACKANNOTATION);
  sNM->set_safetyProperty (n, opunsafe); // otherwise it will be removed 

  insertNode (n, h);
}

/**
   @param vfrom (in) node of the new graph
   @param vto (in) node of the new graph
   @param h (in) dependence clock, in the new graph, of the dependence vfrom --> vto.
   if \e keepDependenceOp (attribute member)
   - it adds the dependence vfrom --> vto when h in the new graph (the dependence is imported)
   - otherwise, the dependence is "inserted" in the definiton of the output of vfrom (insertDependenceInDef()).
*/
_VIRTUAL_ void
Homomorphism::importDependence (ptINode vfrom, ptINode vto, ptinsth h)
{
  if (keepDependenceOp)
    sNM->addExplicitDependence (vfrom, vto, h);
  else
    insertDependenceInDef (vfrom, vto, h);
}

/**
   Each \e explicite dependence aNode -> vto is imported in the new graph (importDependence()).
*/
_VIRTUAL_ void
Homomorphism::homomorphismDependence (ptINode aNode)
{
  ptINode vto, vfrom, nvto, nNode, nodecall;
  ptldependence larcs;
  ptdependence arc;
  ptinsth h, nh;

  if (fNM->getAttribute (aNode, attributeNumber, (attributeValue *) & nNode))
    {
      /* ---- A terminer -------
         lh = fNM->get_clockSuccessors(aNode);
         while (ptlinsth_get_next(&lh, &h))
         if (fCKM->getAttribute(h, attributeNumber, (attributeValue*) &)) 
         sNM->addDependence(nNode, nh);
       */
      larcs = fNM->get_successors (aNode);
      while (larcs->get_next (&arc, &larcs))
	if (arc->satisfiesExplicitDependence ())
	  {
	    h = arc->getDependence (&vfrom, &vto);	// vfrom :: aNode
	    // vform, vto not produced by a node call.
	    if (!fNM->is_outputNodeCall (vfrom, &nodecall))
	      if (isSelectedNode (vto))
		if (!fNM->is_outputNodeCall (vto, &nodecall))
		  {
		    if (fNM->getAttribute (vto, attributeNumber, (attributeValue *) & nvto))
		      {
			h = fCKM->getRepresentativeFormula (h);
			nh = getClockCopy (h);
			if (sCKM->satisfiesUndefClock (nh))
			  nh = sCKM->getTickClock ();
			importDependence (nNode, nvto, nh);
		      }
		    else
		      setTagError (pk_true);
		  }
	  }
    }
}


/**
   It calls homomorphismDependence(ptINode) to each node of a list (ln).
*/
_VIRTUAL_ void
Homomorphism::homomorphismDependences (ptlINode ln)
{
  ptINode aNode;

  if (fgraph_type != DCG_SCH)
    while (ptlINode_get_elem (&ln, &aNode))
      homomorphismDependence (aNode);
}

/**
   Without effects: not implemented.
*/
_VIRTUAL_ void
Homomorphism::homomorphismDependence (ptinsth h)
{
}


/**
   It calls homomorphismDependence(ptinsth) to each clocks of a list (lh).
*/
_VIRTUAL_ void
Homomorphism::homomorphismDependences (ptlinsth lh)
{
  ptinsth h;
  if (fgraph_type != DCG_SCH)
    {
      while (ptlinsth_get_next (&lh, &h))
	homomorphismDependence (h);
      homomorphismDependence (fCKM->getInitClock ());
    }
}
