/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return a copy of a signal (vsig).
*/
  _VIRTUAL_ ptexpcal Homomorphism::getIdCopy (ptexpcal vsig)
{
  ptexpcal
    nsig;
  ptinsth
    clkCopy,
    clk;

  nsig = sEM->mkIdentCopy (vsig, sNM->mk_node_undefined (), sCKM->getUndefClock ());
  fEM->setTreeOfSignalDeclaration (nsig, fEM->getTreeOfSignalDeclaration (vsig));
  fEM->addAttribute (vsig, usedMarkBefore, (attributeValue) nsig);
  clk = fEM->getClock (vsig);
  clkCopy = getClockCopy (clk);
  sEM->setClock (nsig, clkCopy);
  return nsig;
}

/**
   @returns the combedExpr expression (original graph) in which the leafs 
   are replaced by their new values (new graph).
   It uses the attributes.
*/
_VIRTUAL_
  ptexpcal
Homomorphism::getCombedExprAttributes (ptexpcal combedExpr)
{
  ptexpcal e, ne, l;
  if (combedExpr == NILEXPCAL)
    return NILEXPCAL;
  fEM->getNaryArgs (&combedExpr, &e);
  l = getCombedExprAttributes (combedExpr);
  getAttribute (e, &ne);
  return sEM->mkNaryExp (oplesign, ne, l, MK_ANNOT_NIL() );
}

/**
   @returns the ls list (original graph) in which the leafs are replaced by their new values (new graph).
   It uses the attributes.
*/
_VIRTUAL_ ptlinstances
Homomorphism::getCurrentInstances (ptlinstances ls)
{
  ptexpcal vsig, e;
  ptlinstances lsn;

  lsn = NILLINSTANCES;
  while (ptlinstances_get_next (&ls, &vsig))
    if (getAttribute (vsig, &e))
      lsn = ptlinstances_insert (lsn, e);
  return lsn;
}

/**
   Unused. Replaced by getCurrentInstances().
*/
_VIRTUAL_ ptlinstances
Homomorphism::getAttributesOf (ptlinstances ls)
{
  ptlinstances outputs;
  ptexpcal vsig, nsig;

  outputs = NILLINSTANCES;
  while (ptlinstances_get_next (&ls, &vsig))
    {
      getAttribute (vsig, &nsig);
      outputs = ptlinstances_insert (outputs, nsig);
    }
  return outputs;
}

/**
   @return the ls list (original graph) in which the leafs are replaced by their new values (new graph).
   It uses the attributes.
*/
_VIRTUAL_ ptlinstances
Homomorphism::getCurrentInstances (ptlinsth ls)
{
  ptinsth h;
  ptlinstances lsn;
  ptexpcal e;

  lsn = NILLINSTANCES;
  while (ptlinsth_get_next (&ls, &h))
    if (getAttribute (h, &e))
      lsn = ptlinstances_insert (lsn, e);
  return lsn;
}

/** It associated as attribute, if it does not exist, a new signal with the operand (vsig).
    The new signal can have a type different from the type of the operand signal.
    \see setAttribute, mkHomoId, homomorphismIdent(ptexpcal)
*/
_VIRTUAL_ void
Homomorphism::homomorphismDclIdentifier (ptexpcal vsig)
{
  ptexpcal nsig;
  ptinsth h;

  if ((fEM->getLuse (vsig) != NILLEXPCAL) || fEM->getExtractionLink (vsig, &h))	// Added July 2010
    if (!fEM->getAttribute (vsig, attributeNumber, (attributeValue *) & nsig))
      {
	nsig = mkHomoId (vsig, homomorphismIdent (vsig));
	setAttribute (vsig, nsig);
      }
}

/**
   It attributes a signal (vsig) of the original graph by a signal (nsig) of the new graph.
*/
_VIRTUAL_ void
Homomorphism::setAttribute (ptexpcal vsig, ptexpcal nsig)
{
  ptexpcal ns;

  fEM->addAttribute (vsig, attributeNumber, (attributeValue) nsig);
  if (fEM->getAttribute (vsig, usedMarkBefore, (attributeValue *) & ns))
    fEM->addAttribute (ns, attributeNumber, (attributeValue) nsig);
}

/**
   @return "true" if the homomorphism-signal (hs) of the operand signal (vsig) exists, "false" otherwise.
*/
_VIRTUAL_ boolean
Homomorphism::getAttribute (ptexpcal vsig, ptexpcal * hs)
{
  fEM->getSignalAssignedTo (vsig, &vsig);
  if (fEM->getAttribute (vsig, attributeNumber, (attributeValue *) hs))
    return pk_true;
  return pk_false;
}

/** It removes the homomorphism attribute associated with vsig.
 */
_VIRTUAL_ void
Homomorphism::removeAttribute (ptexpcal vsig)
{
  fEM->removeAttribute (vsig, attributeNumber);
}

/**
   @return the clock copy (new graph object) for a signal (original graph).
*/
_VIRTUAL_ ptinsth
Homomorphism::getSignalClockCopy (ptexpcal vsig)
{
  return getClockCopy (fEM->getClock (vsig));
}


/**
   @return a new signal in the new graph with preId as basename and clk as clock.
*/
_VIRTUAL_ ptexpcal
Homomorphism::newSignalAt (ptident preId, ptinsth clk)
{
  ptexpcal nsig;
  nsig = sEM->mkIdentExp (preId, tsignal, homomorphismType (clk), sNM->mk_node_undefined (), clk, MK_ANNOT_NIL());
  sEM->setClock (nsig, clk);
  return nsig;
}

/**
   It applies specific actions according to the class identifier of vsig.
   nsig is the homomorphism signal of vsig.
   \note local to mkHomoId() method.
*/
_VIRTUAL_ void
Homomorphism::specificActionHomoId (ptexpcal vsig, ptexpcal nsig)
{
  typinst t;

  t = fEM->getClassIdentifier (vsig);
  switch (t)
    {
    case tlabel:
    case tsignal:
    case tfield:
      sEM->setClock (nsig, getSignalClockCopy (vsig));
      break;
    case tparam:
      sGraph->addConstant (nsig);
      break;
    case tconstant:
      sGraph->addConstant (nsig);
      break;
    default:;			// tindice, ttyp:
    }
}

/**
   @return a new identifier (declared in the new graph) with id as name, built from vsig (original graph).
   Specific actions according to the class identifier are isolated in specificActionHomoId().
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkHomoId (ptexpcal vsig, ptident id)
{
  ptexpcal nsig;
  typinst t;
  ntrees vsigDecl;
  TypeType TT;

  // The kind of identifier (signal, constant, label,...) is modified !!!
  t = fEM->getClassIdentifier (vsig);
  if (t == tlabel)
    t = tsignal;
  if (t == tfield)
    t = tsignal;
  if (t == tparam)
    t = tconstant;
  vsigDecl = fEM->getTreeOfSignalDeclaration (vsig);
  TT = homomorphismType (vsig);
  nsig = sEM->mkIdentExp (id, t, TT, sNM->mk_node_undefined (), sCKM->getUndefClock (), new BackAnnotation(NILTREE, vsigDecl));
  fEM->setTreeOfSignalDeclaration (nsig, vsigDecl);
  specificActionHomoId (vsig, nsig);
  return nsig;
}

/**
   @return the homomorphism of the type of a signal (vsig) in the new graph.
*/
_VIRTUAL_ TypeType
Homomorphism::homomorphismType (ptexpcal vsig)
{
  return homomorphismType (fEM->getType (vsig));
}


/**
   @return a new list (object of the new graph):  {homoExp} + expList (if homoExp is defined).
*/
_VIRTUAL_ ptexpcal
Homomorphism::addExp (ptexpcal homoExp, ptexpcal expList)
{
  if (homoExp != NILEXPCAL) return sEM->mkNaryExp (oplesign, homoExp, expList, MK_ANNOT_NIL());
  return NILEXPCAL;
}

/**
   @return a list built by the concatenating of le and expList.
   Side effect on expList.
*/
_VIRTUAL_ ptexpcal
Homomorphism::addExpList (ptexpcal le, ptexpcal expList)
{
  ptexpcal e;
  if (sEM->isNILEXPCAL (le))
    return expList;
  if (!sEM->satisfiesList (le))
    return addExp (le, expList);
  if (sEM->getNaryArgs (&le, &e))
    return addExp (e, addExpList (le, expList));
  return expList;
}

/**
   @return the signal "done" defined by the node, inserted in the graph at the clock clk: 
   <em> done := idNode params (s1, s2) </em>.
   Special cases:
   - s1 undefined, it returns s2
   - s2 undefined, it returns s1
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkWaitNodeAt (ptexpcal s1, ptexpcal s2, ptinsth clk, ptident idNode, ptexpcal params)
{
  ptexpcal ns, ins, outs;
  ptINode n;

  if (sEM->isNILEXPCAL (s1))
    return s2;
  if (sEM->isNILEXPCAL (s2))
    return s1;
  ns = newSignalAt (doneId, clk);
  ins = NILEXPCAL;
  ins = sEM->mkNaryExp (oplesign, s1, ins, MK_ANNOT_NIL());
  ins = sEM->mkNaryExp (oplesign, s2, ins, MK_ANNOT_NIL());
  outs = sEM->mkNaryExp (oplesign, ns, NILEXPCAL, MK_ANNOT_NIL());
  n = sNM->mk_procCall (idNode, params, ins, outs, NILBACKANNOTATION);
  //  sNM->set_safetyProperty(n, opsafe); 
  insertNode (n, clk);
  return ns;
}

/**
   @return the "done" signal associated with a node (p).
*/
_VIRTUAL_ ptexpcal
Homomorphism::getSetSignalDone (ptINode p)
{
  ptexpcal signalDoneProcucedGraph;
  ptINode y;

  if (fNM->is_outputNodeCall (p, &y))
    p = y;
  if (!fNM->getAttribute (p, doneAttribute, (attributeValue *) & signalDoneProcucedGraph))
    {
      signalDoneProcucedGraph = newSignalAt (doneId, currentClkDone);
      fNM->addAttribute (p, doneAttribute, (attributeValue) signalDoneProcucedGraph);
    }
  return signalDoneProcucedGraph;
}

/**
   @return the expression "undefinedvalue when clk".
   undefinedvalue is a predefined value (member of the class).
*/
_VIRTUAL_ ptexpcal
Homomorphism::undefinedValueAt (ptinsth clk)
{
  return sEM->mkWhenOrCellExp (opextr, G_UndefinedValue, clk);
}



/**
   @return the signal defined by the "wait" node (built) associated with 
   the current node taking into account the kind of graph (sbDCplusWaitSignal())
   and the explicit dependences (mkWaitNodeAt()). The result can also be the undefinedValueAt() for an "internal added node".

   The current node is given by the context information (currentNodeOriginalGraph context).

   Example: for the following v->a, u-> b, x:= a+b
   The "wait" signal (wait(x)) associated with x is defined by:\n
   wait(x) := waiting(done(v), done(u)) defined at h, where 
    - h is the value of definedAtClk
    - waiting is an intrinsic process name.
   
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkWaitSignalCurrentNode (ptinsth clk, ptinsth definedAtClk)
{
  ptldependence lpred;
  ptdependence aDep;
  ptexpcal waitSignal;
  ptINode node, n, p;
  ptinsth h;
  boolean b;

  if (G_internalNode)
    return undefinedValueAt (clk);

  node = currentNodeOriginalGraph;
  if (sbDCplusWaitSignal (node, clk, definedAtClk, &waitSignal))
    return waitSignal;
  lpred = fNM->get_predecessors (node);
  waitSignal = NILEXPCAL;
  b = pk_false;
  while (lpred->get_next (&aDep, &lpred))
    if (aDep->satisfiesExplicitDependence ())
      {
	h = aDep->getDependence (&p, &n);
	waitSignal = mkWaitNodeAt (waitSignal, getSetSignalDone (p), definedAtClk, waitId, NILEXPCAL);
	b = pk_true;
      }
  if (b)
    return waitSignal;
  return undefinedValueAt (clk);
}

/**
   @return the signal "done" associated with the current node. It is created if it does not exist.
   
   The current node is given by a context information (currentNodeOriginalGraph context).
   For a internal node (G_internalNode context information), it is always created,
   otherwise an attribute gives it (getSetSignalDone()).
   
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkDoneSignalCurrentNode ()
{
  if (G_internalNode)
    return newSignalAt (doneId, currentClkDone);
  return getSetSignalDone (currentNodeOriginalGraph);
}

/**
   It specifies that an error occured during the translation.
   @return the null expression (NILEXPCAL) useful to continue the translation.
*/
_VIRTUAL_ ptexpcal
Homomorphism::error ()
{
  setTagError (pk_true);
  return NILEXPCAL;
}

/**
   @return a new signal defined by the translation associated with the "opnil" operator in the morphism table.
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkNilAt (ptinsth clk)
{
  ptexpcal eh, exp;
  eh = newSignalAt (nilId, clk);
  exp = fEM->mkConstantExp (opnil, nilstr);
  homomorphismExp (eh, exp, clk);
  return eh;
}

/**
   @return "true" if the operation succeeds, "false" otherwise (restriction: element of the list are signals).

   It packs a list of expressions in a "pack node" defined at a clock (clk). For example, for a list such that (e1,e2,...),
   it produces the nodes pe := pack(H(e1), pack (H(e2), pack(...)) where 
   - "pack" is an intrinsic function,
   - H(x) is the morphism of x in the new graph.

   This method is local to the isExprListOperand() method.
   
*/
_VIRTUAL_ boolean
Homomorphism::mkPackExpressionAt (ptexpcal exp, ptinsth clk, ptexpcal * pe)
{
  ptexpcal e, e1;
  boolean b;

  *pe = NILEXPCAL;
  b = pk_false;
  while (fEM->getNaryArgs (&exp, &e))
    if (fEM->IsSignal (e))	// Restriction
      if (isSelectedNode (fEM->getNode (e)))
	{
	  e1 = homomorphismOperand (e, clk);
	  (*pe) = mkWaitNodeAt (*pe, e1, clk, packId, NILEXPCAL);
	  b = pk_true;
	}
  return b;
}

/**
   @return true if exp is the null expression or a list of expressions (well-formed) , "false" otherwise.
   @param exp (in) an expression of the original graph
   @param clk (in) a context clock for node inserting (new graph)
   @param e (out) available when it returns "false" value, equal to e1 when exp = {null, e1}, otherwise equal to exp.
   @param hexp (out) the homomorphism of the packed list (see mkPackExpressionAt())
*/
_VIRTUAL_ boolean
Homomorphism::isExprListOperand (ptexpcal exp, ptinsth clk, ptexpcal * e, ptexpcal * hexp)
{
  ptexpcal e1, se;

  if (fEM->isNILEXPCAL (exp))
    {
      *hexp = undefinedValueAt (clk);
      return pk_true;
    }
  if (!fEM->satisfiesList (exp))
    {
      *e = exp;
      return pk_false;
    }
  se = exp;
  if (fEM->getNaryArgs (&exp, &e1))
    if (fEM->isNILEXPCAL (exp))
      {
	*e = e1;
	return pk_false;
      }
  return mkPackExpressionAt (se, clk, hexp);
}


/** @return the image in the new graph of a clock (h) if it is not the predefined "initial" clock, otherwise clk.
    It assumes that clk is not the predefined "initial" clock and it is a clock of the new graph.
 */
_VIRTUAL_ ptinsth
Homomorphism::getContextClock (ptinsth h, ptinsth clk)
{
  h = getClockCopy (fCKM->getRepresentativeFormula (h));
  if (sCKM->satisfiesInitClock (h))
    return clk;
  return h;
}


/**
   @return always "true".
   It produces eh = morphism(exp), with clk as context clock
   (It uses the opcstent operator for the entry point of the morphism table).
*/
_VIRTUAL_ boolean
Homomorphism::homomorphismNamedConstant (ptexpcal exp, ptinsth clk, ptexpcal * eh)
{
  ptINode n;
  ptrepstr vstr;
  noperators op;

  *eh = newSignalAt (nilId, clk);
  op = opcstent;		// ???? choice
  getStrOperatorName (op, &vstr);
  n = homomorphismExp (vstr, op, (*eh), NILEXPCAL, exp, clk);
  return pk_true;
}

/**
   @return "true" if the image (eh) of an identifier (exp) is available, "false" otherwise.
   For a named Constant, it calls homomorphismNamedConstant() method.
*/
_VIRTUAL_ boolean
Homomorphism::homomorphismVariabl (ptexpcal exp, ptinsth clk, ptexpcal * eh)
{
  // the next 2 lines have been exchanged in July 2010 (LBesnard).
  if (fEM->IsConstant (exp))
    if (homomorphismNamedConstant (exp, clk, eh))
      return pk_true;
  if (getAttribute (exp, eh))
    return pk_true;
  // IsLabel(exp) IsIndice(exp)
  return pk_false;
}


/**
   @return "true" if the image (eh) of an x-expression (exp) is available, "false" otherwise.
   x-expression is an identifier or field access ((homomorphismVariabl()).
*/
_VIRTUAL_ boolean
Homomorphism::homomorphismSignalAccess (ptexpcal exp, ptinsth clk, ptexpcal * eh)
{
  cloper clop;
  noperators op;

  clop = fEM->getOperatorClass (exp);
  if (clop == variabl)
    {
      if (!homomorphismVariabl (exp, clk, eh))
	*eh = mkNilAt (clk);
      return pk_true;
    }
  if (clop != oper2)
    return pk_false;
  op = fEM->getOperator (exp);
  if (op != opcham)
    return pk_false;		// opelem x[..]
  if (!homomorphismVariabl (exp, clk, eh))
    *eh = mkNilAt (clk);
  return pk_true;
}

/**
   @return the image of an leaf-expression (exp) defined 
   - isExprListOperand() when exp is a list of identifiers
   - homomorphismSignalAccess() when exp is an identifier access (x, x.y)
   - otherwise the definition of the opnil operator in the morphism table.
   
   Assumed : the expression exp is variabl class expression.
*/
_VIRTUAL_ ptexpcal
Homomorphism::homomorphismOperand (ptexpcal exp, ptinsth clk)
{
  ptexpcal eh;
  boolean G_internalNode_save;

  if (isExprListOperand (exp, clk, &exp, &eh))
    return eh;
  if (exp == NILEXPCAL)
    return NILEXPCAL;
  if (!homomorphismSignalAccess (exp, clk, &eh))
    {
      G_internalNode_save = G_internalNode;
      G_internalNode = pk_true;
      eh = newSignalAt (nilId, clk);
      homomorphismExp (eh, exp, clk);
      G_internalNode = G_internalNode_save;
    }
  return eh;
}

/**
   @return the image of the dimensions (array type) of a signal (exp) or null expression otherwise.
   @param exp (in) an expression of the original graph.
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkExprDimType (ptexpcal exp)
{
  TypeType TT;
  ptexpcal dim;

  TT = fEM->getType (exp);
  if (!TypeTestArrayClass (TT))
    dim = NILEXPCAL;
  else
    dim = ArrayMatrixTypeGetDimLst (TT);
  return dim;			// copie dans sEM
}

/** 
    @return the iOp-th operand of o or the null expression when an error occurs.

    @param o (in) an expression of the produced graph (be careful), assumed to be a list
    @param iOp (in) an positive int.
 */
_VIRTUAL_ ptexpcal
Homomorphism::getOutputOperandAtIndex (ptexpcal o, int iOp)
{
  ptexpcal e;
  int i;

  if (o == NILEXPCAL)
    return NILEXPCAL;
  if (iOp <= 0)
    return NILEXPCAL;
  if (!sEM->satisfiesList (o))
    {
      if (iOp == 1)
	return o;
      return NILEXPCAL;		// error
    }
  i = 0;
  while (i <= iOp)
    {
      if (!sEM->getNaryArgs (&o, &e))
	return NILEXPCAL;
      i++;
    }
  return e;
}

/**
   @return the opIndex operand of an expression (exp).
   The null expression is returned when an error is detected.
   When exp is an identifier expression, it is returned.
*/
_VIRTUAL_ ptexpcal
Homomorphism::getOperandAtIndex (ptexpcal exp, int opIndex)
{
  cloper t;

  t = fEM->getOperatorClass (exp);
  if (!fEM->existChild (exp, opIndex))
    {
      if (opIndex != 1)
	return error ();
      if (t != variabl)
	return error ();
    }

  switch (t)
    {
    case variabl:
    case constant:
      return exp;

    case oper1:
    case oper1tp:
      return fEM->getUnaryArg (exp);

    case oper2:
    case oper2tp:		// default
      return fEM->getBinaryArgstAt (exp, opIndex);

    case oper3tp:		// delay, win
    case oper3:
      return fEM->getBinaryArgstAt (exp, opIndex);

    case opereh:		// E when h or E cell h
      if (opIndex != 1)
	return NILEXPCAL;
      return fEM->getWhenOrCellArg (exp);

    default:			// case opertyp:
      return NILEXPCAL;
    }
}

/**
   @return the dimension of the opIndex-th operand of an expression (exp).
   The null expression is returned when an error is detected.
   It references mkExprDimType(), getOperandAtIndex().
*/
_VIRTUAL_ ptexpcal
Homomorphism::getDimTypeOperandAtIndex (ptexpcal exp, int opIndex)
{
  ptexpcal e;
  e = getOperandAtIndex (exp, opIndex);
  if (fEM->isNILEXPCAL (e))
    return NILEXPCAL;
  return mkExprDimType (e);
}



/**
   @return the type of the opIndex-th operand of an expression (exp) represented as an expression.
   The null expression is returned when an error is detected.
   It references mkExprElemType(), getOperandAtIndex().
*/
_VIRTUAL_ ptexpcal
Homomorphism::getTypeOperandAtIndex (ptexpcal exp, int opIndex)
{
  ptexpcal e;
  e = getOperandAtIndex (exp, opIndex);
  if (fEM->isNILEXPCAL (e))
    return NILEXPCAL;
  return mkExprElemType (e);
}

/**
   @return the image of the opIndex-th operand of an expression (exp) of the original graph.
   The clock (clk) is a context clock (new graph) for fixing the clock of the constants.
*/
_VIRTUAL_ ptexpcal
Homomorphism::homomorphismOperandAtIndex (ptexpcal exp, int opIndex, ptinsth clk)
{
  ptexpcal e1, eh;
  cloper t;
  ptinsth h;

  t = fEM->getOperatorClass (exp);
  if (!fEM->existChild (exp, opIndex))
    {
      if (opIndex != 1)
	return error ();
      if (t != variabl)
	return error ();
    }
  switch (t)
    {
    case variabl:
      if (homomorphismVariabl (exp, clk, &eh))
	return eh;
      return error ();

    case constant:
      return error ();

    case oper1:
    case oper1tp:
      e1 = fEM->getUnaryArg (exp);
      return homomorphismOperand (e1, clk);

    case oper2:
      e1 = fEM->getBinaryArgstAt (exp, opIndex);
      return homomorphismOperand (e1, clk);

    case oper2tp:		// default
      e1 = fEM->getBinaryArgstAt (exp, opIndex);
      return homomorphismOperand (e1, clk);

    case oper3tp:		// delay, win
    case oper3:
      e1 = fEM->getTernaryArgstAt (exp, opIndex);
      return homomorphismOperand (e1, clk);

    case opereh:		// E when h or (E1 when h1) cell h
      if (opIndex == 1)
	{
	  e1 = fEM->getWhenOrCellArg (exp);
	  if (fEM->IsCellExpr (exp))
	    h = fEM->getContributionClock (e1);
	  else if (!fEM->getExprClock (e1, &h))
	    h = fEM->getContributionClock (exp);

	  h = getContextClock (h, clk);
	  return homomorphismOperand (e1, h);
	}
      //(opIndex == 2)
      h = fEM->getContributionClock (exp);
      h = getContextClock (h, clk);
      getAttribute (h, &eh);
      return eh;

      // case opertyp:  ??????

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return NILEXPCAL;
}

/**
   @return the clock of an expression (e) if it is not an expression on constants, or 
   the context clock (contextClock).
*/
_VIRTUAL_ ptinsth
Homomorphism::getHomoExprClock (ptexpcal e, ptinsth contextClock)
{
  ptinsth h;
  if (!fEM->getExprClock (e, &h))
    return contextClock;
  return getContextClock (h, contextClock);
}

/** 
    @return the image of the contribution clock of the opIndex-th operand of an expression(exp).
    It assumes that exp is an expression of the original graph.
    The clock (clk) is a context clock (of the new graph) for fixing the clock of the constants.
 */
_VIRTUAL_ ptinsth
Homomorphism::getClockOperandAtIndex (ptexpcal exp, int opIndex, ptinsth clk)
{
  ptexpcal e;
  ptinsth h;
  cloper cl;

  cl = fEM->getOperatorClass (exp);
  if (!fEM->existChild (exp, opIndex))
    {
      if (opIndex != 1)
	return hundef;
      if (cl != variabl)
	return hundef;
    }

  switch (cl)
    {
    case oper1tp:
      return getHomoExprClock (fEM->getUnaryArg (exp), clk);

    case oper2tp:		// default
      e = fEM->getBinaryArgstAt (exp, opIndex);
      h = fEM->getContributionClock (exp);
      h = getContextClock (h, clk);
      return getHomoExprClock (e, h);

    case oper3tp:		// delay, win
      e = fEM->getTernaryArgstAt (exp, opIndex);
      return getHomoExprClock (e, clk);

    case opereh:		// E when h or E cell h
      if (opIndex == 1)
	{
	  e = fEM->getWhenOrCellArg (exp);
	  if (fEM->IsCellExpr (exp))
	    h = fEM->getContributionClock (e);
	  else
	    h = fEM->getContributionClock (exp);
	  h = getContextClock (h, clk);
	  return getHomoExprClock (e, h);
	}
      //(opIndex == 2)
      h = fEM->getContributionClock (exp);
      return getContextClock (h, clk);


    default:
      // variabl, constant, oper1, oper2, oper3
      return getHomoExprClock (exp, clk);
    }
}


/**
   @return the type of an expression as a reference to enumerated value of the defined type in homomorphismTypeCode() method.

   Example: if the type of exp is real, it returns the expression opvenum(__TC__, "_real").
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkExprElemType (ptexpcal exp)
{
  TypeType TT;
  ptexpcal e1;
  ptrepstr vstr;

  TT = fEM->getType (exp);
  if (TypeTestArrayClass (TT))
    TT = ArrayMatrixTypeGetTerm (TT);
  vstr = getTypeStrAsEnumVal (TT);
  e1 = sEM->mkExpType (TypeCodeEnum);
  return sEM->mkBinaryExp (opvenum, e1, sEM->mkConstantExp (opcstenum, vstr), MK_ANNOT_NIL());
}

/**
   @return the morphism of an expression (exp) (original graph) by scanning the associated translation defined 
   via a scanner object.
   @param exp (in) expression definition of the current translated node
   @param ScanPragObj (in) a scanner ready to analyze the translation of the parameters
   @param o (in) the outputs (new graph) of the node currently translated (it may reference the type of the i-th ouputs)

   It takes into account the following tokens (see morphismPredefinedNames.h):
   - tk_type, tk_cardwait, tk_dim 
   - end scanning: tk_af
   - separator: tk_virg
   
   \code
   For the diff operator, with the following morphism definition (table) \n 
   operator /= = (?x,y; !z; ) pragmas MORPHISM "(&h&o, &done) :=CostNEQ{&t&i1,&t&i2}(&h&i1, &h&i2, &hck&o, &wait)" end pragmas;\n
   This method translates only the text { &t&i1, &t&i2 }. It specifies that the process "CostNEQ" is parametrized by 
   two expressions: the type (&t) of the first input (&i1) and the  the type (&t) of the second input (&i2).
   \endcode
*/
_VIRTUAL_ ptexpcal
Homomorphism::homoParameters (ptStrScanner ScanPragObj, ptexpcal o, ptexpcal exp)
{
  ptexpcal nexp, hexp, eTT;
  ScanPragmaToken tk;
  boolean b;
  int iOp, nbWait;

  if (ScanPragObj->getCurrentToken () == tk_af)
    return NILEXPCAL;		// End of parameters
  tk = ScanPragObj->getToken ();
  switch (tk)
    {
    case tk_af:		// }   : End of parameters
      return NILEXPCAL;

    case tk_virg:
      return homoParameters (ScanPragObj, o, exp);

    case tk_type:
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 1;
	  nexp = homoParameters (ScanPragObj, o, exp);
	  eTT = getTypeOperandAtIndex (exp, iOp);
	  nexp = addExp (eTT, nexp);
	  return nexp;
	  break;
	case tk_output:	/* &o --> Homotype (???) */
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 1;
	  nexp = homoParameters (ScanPragObj, o, exp);
	  eTT = getTypeOperandAtIndex (o, iOp);
	  nexp = addExp (eTT, nexp);
	  return nexp;
	  break;

	default:
	  nexp = homoParameters (ScanPragObj, o, exp);
	  hexp = mkExprElemType (exp);
	  return addExp (hexp, nexp);
	  break;
	}

      //      nexp = homoParameters( ScanPragObj, o, exp);
      //      hexp = mkExprElemType(exp);
      //      return addExp(hexp, nexp);

    case tk_cardwait:
      nexp = homoParameters (ScanPragObj, o, exp);
      nbWait = fNM->getCardExplicitPredecessors (currentNodeOriginalGraph);
      hexp = sEM->mkIntegerConstant (nbWait);
      return addExp (hexp, nexp);


    case tk_dim:		//&d  " dimension of the type ", must be followed by &p &i &o -- Not Implemented --
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  b = ScanPragObj->getInt (&iOp);
	  nexp = homoParameters (ScanPragObj, o, exp);
	  if (b)
	    {
	      eTT = getDimTypeOperandAtIndex (exp, iOp);
	      nexp = addExp (eTT, nexp);
	    }
	  else
	    setTagError (pk_true);
	  return nexp;
	  break;
	case tk_output:	/* &o */
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 0;
	  nexp = homoParameters (ScanPragObj, o, exp);
	  eTT = getDimTypeOperandAtIndex (o, iOp);
	  nexp = addExp (eTT, nexp);
	  return nexp;
	  break;

	default:
	  //      setTagError(pk_true);
	  return NILEXPCAL;
	  break;
	}
    default:			/*
				   case tk_homo: case tk_input: case tk_output: case tk_clock:
				   case tk_param: case tk_name: case tk_end:
				   case tk_error:  case tk_ident: case tk_2pe:
				   case tk_po: case tk_pf: case tk_ao:  */
      setTagError (pk_true);
      return NILEXPCAL;
      break;
    }
}


/**
   @return mkDoneSignalCurrentNode().

   Currently out is unused (to be completed).
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkDoneSignalOutputs (ptexpcal out)
{
  return mkDoneSignalCurrentNode ();	// To be completed
}


/**
   @return the morpshism defined by the analyze of the text associated to a scanner (ScanPragObj) for the outputs.

   @param ScanPragObj (in) a scanner ready to analyze the translation of the outputs. Moreover, the current token is a '(' character.
   @param out (in) the output of the current translated node (new graph).
   @param clk (in) clock context for fixing the clocks of the constants (new graph).
   @param bsignaldone (out) "true" if a signal "done" has been produced (&done is referenced in the pragmas MORPHISM), "false" otherwise.
   
   It takes into account the following tokens (see morphismPredefinedNames.h):
   - tk_homo --followed by tk_output (&o)--, tk_list --&lexp followed by tk_output (&o)--, tk_done, 
   - end scanning: tk_pf
   - separator: tk_virg
   
   \code
   For the diff operator, with the following morphism definition (table) \n 
   operator /= = (?x,y; !z; ) pragmas MORPHISM "(&h&o, &done) :=CostNEQ{&t&i1,&t&i2}(&h&i1, &h&i2, &hck&o, &wait)" end pragmas;\n
   This method translates only the text (&h&o, &done). It specifies that the process "CostNEQ" has two outputs
   - the morphsim of the output of the orignal node
   - a "done" node for keeping ordering (sequentializing).
   \endcode
*/
_VIRTUAL_ ptexpcal
Homomorphism::severalOutputs (ptStrScanner ScanPragObj, ptexpcal out, ptinsth clk, boolean * bsignaldone)
{
  ptexpcal nexp;
  ScanPragmaToken tk;
  int iOp;

  tk = ScanPragObj->getToken ();
  switch (tk)
    {
      // End scanning
    case tk_pf:
      return NILEXPCAL;
      // Separator.
    case tk_virg:
      return severalOutputs (ScanPragObj, out, clk, bsignaldone);

      // &lexp, must be followed by tk_output (&o)
    case tk_list:
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_output:
	  nexp = severalOutputs (ScanPragObj, out, clk, bsignaldone);
	  return addExpList (out, nexp);
	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}

      // &h homomorphism, must be followed by &o
    case tk_homo:
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_output:
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 1;
	  nexp = severalOutputs (ScanPragObj, out, clk, bsignaldone);
	  return addExp (getOutputOperandAtIndex (out, iOp), nexp);
	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}
      // &done 
    case tk_done:
      nexp = severalOutputs (ScanPragObj, out, clk, bsignaldone);
      *bsignaldone = pk_true;	// => noeud au tick du programe
      return addExp (mkDoneSignalOutputs (out), nexp);

      // default
    default:
      setTagError (pk_true);
      return NILEXPCAL;
    }
}

/**
   @return the morpshism defined by the analyze of the text associated to a scanner (ScanPragObj) for the outputs.
   It verifies also when the end of outputs is detected by the scanner, that the next token is e : character. If is is 
   found the next token is read. The result is produced by severalOutputs() method.

   @param ScanPragObj (in) a scanner ready to analyze the translation of the outputs. Moreover, the current token is a '(' character.
   @param out (in) the output of the current translated node (new graph).
   @param clk (in) clock context for fixing the clocks of the constants (new graph).
   @param bsignaldone (out) "true" if a signal "done" has been produced (&done is referenced in the pragmas MORPHISM), "false" otherwise.
*/
_VIRTUAL_ ptexpcal
Homomorphism::severalOutputs2pe (ptStrScanner ScanPragObj, ptexpcal out, ptinsth clk, boolean * bsignaldone)
{
  ScanPragmaToken tk;
  ptexpcal nexp;

  // currentToken is tk_po char at first call
  nexp = severalOutputs (ScanPragObj, out, clk, bsignaldone);
  tk = ScanPragObj->getToken ();
  // := waited
  if (tk != tk_2pe)
    setTagError (pk_true);
  else
    ScanPragObj->getToken ();
  return nexp;
}


/**
   @return out as a list-expression, or if out is undefined a new signal, with clk as clock, defined by newSignalAt() method.
*/
_VIRTUAL_ ptexpcal
Homomorphism::defaultOutput (ptexpcal out, ptinsth clk)
{
  if (out == NILEXPCAL)
    out = newSignalAt (nilId, clk);	// ?? a revoir
  else if (sEM->satisfiesList (out))
    return out;
  return addExp (out, NILEXPCAL);
}

/**
   @return the morpshism defined by the analyze of the text associated to a scanner (ScanPragObj) for the outputs.
   @param ScanPragObj (in) a scanner ready to analyze the translation of the outputs 
   @param out (in) the output of the current translated node (new graph).
   @param clk (in) clock context for fixing the clocks of the constants (new graph).
   @param bsignaldone (out) "true" if a signal "done" has been produced (&done is referenced in the pragmas MORPHISM), "false" otherwise

   It calls 
   - defaultOutput() on tk_ident token (in this case bsignaldone is always "false")
   - severalOutputs2pe on tk_ident token
   
   It may produce a null expression when an error occurs.
*/
_VIRTUAL_ ptexpcal
Homomorphism::mkOutputs (ptStrScanner ScanPragObj, ptexpcal out, ptinsth clk, boolean * bsignaldone)
{
  ScanPragmaToken tk;

  *bsignaldone = pk_false;
  tk = ScanPragObj->getCurrentToken ();
  switch (tk)
    {
    case tk_ident:		/* <id> (i1, ...,in)  => one output */
      return defaultOutput (out, clk);
    case tk_po:		/* ( o1, ..., on) := <id>(....) => several outputs */
      return severalOutputs2pe (ScanPragObj, out, clk, bsignaldone);
    default:
      setTagError (pk_true);
      return NILEXPCAL;
    }
}

/**
   @return the morpshim of an expression (e).
   e can be 
   - a function call, the result is given by the translation associated with the opappe operator.
   - a leaf, the result is produced by the HomomorphismOperand() method.
*/
_VIRTUAL_ ptexpcal
Homomorphism::homomorphismLeafOrFunctionCall (ptexpcal e, ptinsth clk)
{
  boolean G_internalNode_save;
  ptexpcal eh, hlout, lin, lp;
  ptident id;

  if (sEM->getOperatorClass (e) != call)
    return homomorphismOperand (e, clk);
  G_internalNode_save = G_internalNode;
  G_internalNode = pk_true;
  eh = newSignalAt (nilId, clk);
  hlout = sEM->mkNaryExp (oplesign, eh, NILEXPCAL, MK_ANNOT_NIL());
  sEM->getFunctionCall (e, &id, &lp, &lin);
  homomorphismExp (ident2repstr (id), opappe, hlout, lp, lin, clk);
  G_internalNode = G_internalNode_save;
  return eh;
}

/**
   Called when &lexp&i (number of inputs is variable) is detected by the scanner.
   @return the morphism of each operand (order is kept) of exp, followed by nexp (new graph object) expression.
   For each element it calls homomorphismLeafOrFunctionCall() method. This method needs a clock context (clk from the new graph) 
   for fixing the clocks of the constants.
*/
_VIRTUAL_ ptexpcal
Homomorphism::addHomoList (ptexpcal exp, ptinsth clk, ptexpcal nexp)
{
  ptexpcal e;
  noperators op;

  if (fEM->isNILEXPCAL (exp))
    return nexp;
  // urgence : opensvec, oplevec
  op = fEM->getOperator (exp);
  if ((op == opensvec) || (op == oplelevec))
    {
      exp = fEM->convertLinstances2ListExpcal (fEM->getSignalsReferencedIn (exp));
      if (fEM->isNILEXPCAL (exp))
	return nexp;
    }
  if (!fEM->satisfiesList (exp))
    return addExp (homomorphismLeafOrFunctionCall (exp, clk), nexp);
  if (fEM->getNaryArgs (&exp, &e))
    return addExp (homomorphismLeafOrFunctionCall (e, clk), addHomoList (exp, clk, nexp));
  return nexp;
}

/**
   @return the morphism of an expression (exp) (original graph) by scanning the associated translation defined 
   via a scanner object restricted to the inputs.
   @param ScanPragObj (in) a scanner ready to analyze the translation of the inputs
   @param exp (in) expression definition of the current translated node (original node)
   @param o (in) the outputs of the node currently translated (new graph)
   @param params (in) the parameters of the node currently translated
   @param clk (in) clock context for fixing the clocks of the constants (new graph).

   It takes into account the following tokens (see morphismPredefinedNames.h):
   - tk_list, tk_homo, tk_homoClock, tk_clock, tk_wait
   - end scanning: tk_pf
   - separator: tk_virg

   \code
   For the diff operator, with the following morphism definition (table) \n 
   operator /= = (?x,y; !z; ) pragmas MORPHISM "(&h&o, &done) :=CostNEQ{&t&i1,&t&i2}(&h&i1, &h&i2, &hck&o, &wait)" end pragmas;\n
   This method translates only the text (&h&i1, &h&i2, &hck&o, &wait). It specifies that the process "CostNEQ" has 4 inputs: 
   - the image (&h) of the first input (&i1), 
   - the image (&h) of the second input (&i2), 
   - the image of the clock (&hck) of the output (&o)
   - and the signal defined by the "waiting" of its predecessors (explicit dependence or reinforcing).   
   \endcode
*/
_VIRTUAL_ ptexpcal
Homomorphism::homoInputs (ptStrScanner ScanPragObj, ptexpcal o, ptexpcal exp, ptinsth clk, ptexpcal params)
{
  ptexpcal nexp, e;
  ScanPragmaToken tk;
  int iOp;
  boolean b;
  ptinsth h;

  tk = ScanPragObj->getToken ();
  switch (tk)
    {
    case tk_pf:
      return NILEXPCAL;
    case tk_virg:
      return homoInputs (ScanPragObj, o, exp, clk, params);

    case tk_list:		/* &lexp, must be followed by &i (restriction) */
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  return addHomoList (exp, clk, nexp);
	  break;

	case tk_output:	/* &o */
	  setTagError (pk_true);
	  return NILEXPCAL;
	  break;

	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}

    case tk_homo:		/* &h homomorphism, must be followed by &{i,o,p....} */
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  b = ScanPragObj->getInt (&iOp);
	  nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  if (b)
	    nexp = addExp (homomorphismOperandAtIndex (exp, iOp, clk), nexp);
	  else
	    // new new
	  if (homomorphismVariabl (exp, clk, &e))
	    nexp = addExp (e, nexp);
	  else
	    setTagError (pk_true);
	  return nexp;
	  break;

	case tk_output:	/* &o */
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 1;
	  if (ScanPragObj->getCurrentToken () == tk_pf)
	    nexp = NILEXPCAL;
	  else
	    nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  nexp = addExp (getOutputOperandAtIndex (o, iOp), nexp);
	  return nexp;
	  break;

	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}

    case tk_homoClock:		/* &hck  can be followed by &i &o */
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  b = ScanPragObj->getInt (&iOp);
	  nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  if (b)
	    nexp = addExp (imageGetAttributeClock (getClockOperandAtIndex (exp, iOp, clk)), nexp);
	  else
	    setTagError (pk_true);
	  return nexp;
	  break;

	case tk_output:	/* &o */
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 1;
	  if (ScanPragObj->getCurrentToken () == tk_pf)
	    nexp = NILEXPCAL;
	  else
	    nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  e = getOutputOperandAtIndex (o, iOp);
	  if (e != NILEXPCAL)
	    if (sEM->getExprClock (e, &h))
	      nexp = addExp (imageGetAttributeClock (h), nexp);
	    else
	      setTagError (pk_true);
	  else
	    nexp = addExp (imageGetAttributeClock (clk), nexp);
	  return nexp;
	  break;

	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}
      break;


    case tk_clock:		/* &ck */
      tk = ScanPragObj->getToken ();
      switch (tk)
	{
	case tk_input:		/* &i */
	  b = ScanPragObj->getInt (&iOp);
	  nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  if (b)
	    {
	      h = getClockOperandAtIndex (exp, iOp, clk);
	      nexp = addExp (clock2expr (h), nexp);
	    }
	  else
	    setTagError (pk_true);
	  return nexp;
	  break;

	case tk_output:	/* &o */
	  if (!ScanPragObj->getInt (&iOp))
	    iOp = 1;
	  nexp = homoInputs (ScanPragObj, o, exp, clk, params);
	  e = getOutputOperandAtIndex (o, iOp);
	  if (sEM->getExprClock (e, &h))
	    nexp = addExp (clock2expr (h), nexp);
	  else
	    setTagError (pk_true);
	  return nexp;
	  break;

	default:
	  setTagError (pk_true);
	  return NILEXPCAL;
	}
      break;

    case tk_wait:		// taking into account dependencies
      nexp = homoInputs (ScanPragObj, o, exp, clk, params);
      e = mkWaitSignalCurrentNode (clk, currentClkDone /* tickClockProducedGraph */ );
      return addExp (e, nexp);

      /* 
         ptINode n;   
         case tk_ident:
         ptexpcal x = newSignalAt(nilId, hundef);
         n = homomorphismExp(ScanPragObj->getIdent(), x, params, exp, clk);
         insertNode(n, h);
         nexp = homoInputs(ScanPragObj, o, exp, clk, params);
         return addExp(x, nexp);
       */

    default:			/* tk_end, tk_po, tk_ao, tk_af, tk_2pe, tk_type, tk_param, tk_name, tk_dim, tk_error, tk_ident */
      setTagError (pk_true);
      return NILEXPCAL;
      break;
    }
}


/**
   @return the node produced by the morphism of a node (original graph) by scanning the associated translation defined 
   via a scanner object.
   @param ScanPragObj (in) a scanner ready to analyze the translation of an operator
   @param strName (in) the name associated with the current operator (it can be referenced by &n in the morphism table)
   @param out (in) the outputs of the node currently translated (new graph)
   @param params (in) the parameters of the node currently translated
   @param exp (in) expression definition of the node currently translated (original node)
   @param clk (in) clock context for fixing the clocks of the constants (new graph).

   Using the tokens provided by the scanner, it calls:
   - mkOutputs() to defined the ouputs in the new graph
   - homomorphismIdent(ptStrScanner, ptrepstr) to produce the name of the node in the new graph
   - homoParameters() for the translation of the parameters
   - homoInputs() for  the translation of the inputs
   then it builds the new node and inserts it in the new graph.

   \code
   For the diff operator, with the following morphism definition (table) \n 
   operator /= = (?x,y; !z; ) pragmas MORPHISM "(&h&o, &done) :=CostNEQ{&t&i1,&t&i2}(&h&i1, &h&i2, &hck&o, &wait)" end pragmas;\n
   this method produces a new node in the new graph  using the pragmas MORPHISM.
   \endcode
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismExp (ptStrScanner ScanPragObj, ptrepstr strName, ptexpcal out, ptexpcal params, ptexpcal exp, ptinsth clk)
{
  ptINode n;
  ptident CurrentIdCall;
  ptexpcal CurrentParameters, CurrentInputs, CurrentOutputs;
  ScanPragmaToken tk;
  boolean bsignaldone;

  tk = ScanPragObj->getToken ();
  CurrentOutputs = mkOutputs (ScanPragObj, out, clk, &bsignaldone);
  // Process name
  CurrentIdCall = homomorphismIdent (ScanPragObj, strName);
  // Parameters
  tk = ScanPragObj->getCurrentToken ();
  if (tk == tk_ao)		/* { */
    {
      CurrentParameters = homoParameters (ScanPragObj, out, exp);
      tk = ScanPragObj->getToken ();
    }
  else
    {
      CurrentParameters = NILEXPCAL;
      tk = ScanPragObj->getCurrentToken ();
    }
  // Inputs
  if (tk == tk_po)		/* ( */
    CurrentInputs = homoInputs (ScanPragObj, out, exp, clk, params);
  else
    CurrentInputs = NILEXPCAL;

  // La suite sera a adpater pour accepter autre chose qu'un appel
  n = sNM->mk_procCall (CurrentIdCall, CurrentParameters, CurrentInputs, CurrentOutputs, NILBACKANNOTATION);
  sNM->set_safetyProperty (n, opsafe);	// FOR TEST
  if (bsignaldone)
    clk = currentClkDone;
  insertNode (n, clk);		// A la bsup des e/s
  return n;
}

/**
   @return the image in the new graph of a node from the original graph.
   The image is created by the following actions:
   - it creates a scanner for an operator (op), if the operator (op) is not defined in the table, the result will be the undefined node.
   - it calls homomorphismExp(ptStrScanner, ptrepstr , ptexpcal , ptexpcal , ptexpcal , ptinsth )

   @param vstr (in) the name associated with the operator (it can be referenced by &n in the morphism table)
   @param op (in) the operator of the current translated node.
   @param out (in) the outputs of the node currently translated (new graph)
   @param params (in) the parameters of the node currently translated (original graph)
   @param def (in) expression definition of the node currently translated (original node)
   @param clk (in) clock context for fixing the clocks of the constants (new graph).

   Example: for x:= a+b, vstr = "+", op=+, out= morphism(x), params=nil, def=a+b, clk= morphism(^x)
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismExp (ptrepstr vstr, noperators op, ptexpcal out, ptexpcal params, ptexpcal def, ptinsth clk)
{
  ptStrScanner ScanPragObj;
  if (!scannerOn (op, vstr, out, params, def, clk, &ScanPragObj))
    return sNM->mk_node_undefined ();
  return homomorphismExp (ScanPragObj, vstr, out, params, def, clk);
}


/**
   @return the image in the new graph of a node from the original graph.
   It is teh result of the call to
   homomorphismExp(ptrepstr , noperators, ptexpcal , ptexpcal , ptexpcal , ptinsth )

   @param out (in) the outputs of the node currently translated (new graph)
   @param exp (in) expression definition of the node currently translated (original node)
   @param clk (in) clock context for fixing the clocks of the constants (new graph).

   Example: for x:= a+b, out= morphism(x), exp=a+b, clk= morphism(^x).
   It calls homomorphismExp("+", opplus, out, NILEXPCAL, exp, clk)
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismExp (ptexpcal out, ptexpcal exp, ptinsth clk)
{
  noperators op;
  ptrepstr vstr;

  op = fEM->getOperator (exp);
  getStrOperatorName (op, &vstr);
  return homomorphismExp (vstr, op, out, NILEXPCAL, exp, clk);
}
