/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" if a node (n) is selected for the translation, "false" otherwise.
*/
  _VIRTUAL_ boolean Homomorphism::isSelectedNode (ptINode n)
{
  if (ptlINode_isIn (unNodes, n))
    return pk_false;
  if (sNM->is_partialDefinition (n))
    return pk_false;		// done via ldefpart operator
  return pk_true;
}

/**
   @return a undefined node and signals an error.
*/
_VIRTUAL_
  ptINode
Homomorphism::errorOnNode (ptINode n)
{
  pK_messenger->printString ("Homomorphism : ERROR on a node");
  return sNM->mk_node_undefined ();
}


/**
   Finalizing the translation of a node (n1) of the orignal graph in a node (n) of the new graph.
   It adds a link from n1 to n using the attributeNumber attribut.
*/
_VIRTUAL_ void
Homomorphism::endHomomorphismNode (ptINode n1, ptINode n)
{
  //  sNM ->addPragmas(n, fNM->getPragmas(n1));  // Il faudra dupliquer ( contexte)
  fNM->addAttribute (n1, attributeNumber, (attributeValue) n);
}

/**
   Initializing the translation of a node (n).
*/
_VIRTUAL_ void
Homomorphism::beginHomomorphismNode (ptINode n)
{
  currentNodeOriginalGraph = n;
  // e= mkWaitSignalCurrentNode(clk, currentClkDone/* tickClockProducedGraph*/);
  G_internalNode = pk_false;
}



/**
   @return for an expression (expression) assumed to be of the form "E when/cell h"
   - E is ^E is h,
   - expression otherwise.

   Local to homomorphismEqu() method.
*/
ptexpcal Homomorphism::getSubTermFilteredClock (ptexpcal expression)
{
  ptinsth
    hd,
    vh;
  ptexpcal
    basicExpr;
  noperators
    vop;

  if (fEM->IsCellExpr (expression))
    return expression;
  hd = fEM->getContributionClock (expression);
  if (fCKM->satisfiesInitClock (hd))
    return fEM->getBasicExpr (expression);
  basicExpr = fEM->getBasicExpr (expression);
  vop = fEM->getOperator (basicExpr);
  if ((vop == opmele) || (vop == opleft_tt))
    return basicExpr;
  if (vop == opnombm)
    return basicExpr;
  if (fEM->getSynchronousExprHorl (basicExpr, &vh))
    {
      vh = fCKM->getRepresentativeFormula (vh);
      if (vh == hd)
	return basicExpr;
    }
  return expression;
}

/**
   @return the morphism of a node (node) when it is an equation node (X:=E)
   The translation is given in the morphism table by the operator(E).
*/
_VIRTUAL_
  ptINode
Homomorphism::homomorphismEqu (ptINode node)
{
  ptexpcal vopg, vopd, hvopg;
  ptinsth h;

  vopg = fNM->get_leftOperand (node);
  fNM->get_rightOperand (node, &vopd);
  h = fEM->getContributionClock (vopd);
  vopd = getSubTermFilteredClock (vopd);
  h = getContextClock (h, getClockCopy (fNM->get_nodeClock (node)));
  if (getAttribute (vopg, &hvopg))
    return homomorphismExp (hvopg, vopd, h);
  return errorOnNode (node);
}

/**
   @return the morphism of a node (node) when it is a delay node.
   The translation is given in the morphism table by the operators opret($) and opfen (window)
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismDelay (ptINode node)
{
  ptexpcal vopg, vopd, vDelay, vWin, hvopg, vinit;
  ptinsth h;
  boolean b;

  fNM->get_delay_arg (node, &vopg, &vopd, &vDelay, &vWin, &vinit);
  //  vinit = fEM->getInitialValue(vopg);
  // Construction d'une expression conforme a la syntaxe Signal (non typee)
  b = fEM->setModeTypeChecking (pk_false);	// <<<<<--
  if (vWin == NILEXPCAL)
    /* cas d'un retard simple */
    vopd = fEM->mkTernaryExp (opret, vopd, vDelay, vinit, MK_ANNOT_NIL());
  else if (vDelay == NILEXPCAL)
    /* cas d'une fenetre simple */
    vopd = fEM->mkTernaryExp (opfen, vopd, vWin, vinit, MK_ANNOT_NIL());
  else
    /* cas d'un retard et d'une fenetre */
    {
      vopd = fEM->mkTernaryExp (opret, vopd, vDelay, NILEXPCAL, MK_ANNOT_NIL());
      vopd = fEM->mkTernaryExp (opfen, vopd, vWin, vinit, MK_ANNOT_NIL());
    }
  fEM->setModeTypeChecking (b);	// <<<<<--
  fEM->setType (vopd, fEM->getType (vopg));
  h = getClockCopy (fNM->get_nodeClock (node));
  if (getAttribute (vopg, &hvopg))
    return homomorphismExp (hvopg, vopd, h);
  return errorOnNode (node);
}


/**
   @return a node produced by the definition associated with the opappe operator in the morphism table.
   It calls homomorphismExp(ptrepstr, noperators , ptexpcal, ptexpcal , ptexpcal , ptinsth)
   with (name, opappe, out, params, def, clk) as parameters.

   It manages the attribute member of the class \e markExternalCall:
   - it is set to "true". It will be referenced in insertNode() to tag the node by a mark ("DONTDECLARE") to specify 
   that the declaration of the model is not necessary.
   - it is popped at the end of the method.

   The name can be referenced (by &n) in the declaration of the definition associated with the opappe operator 
   in the translation table. For example:\n
   operator opappe = (? ! ) pragmas MORPHISM "(&h&o, &done) :=Cost&n{}(&h&i1, &h&i2 , &hck&o, &wait)" end pragmas; %
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismExternal (ptrepstr name, ptexpcal out, ptexpcal params, ptexpcal def, ptinsth clk)
{
  ptINode n;
  boolean s;

  s = markExternalCall;
  markExternalCall = pk_true;
  n = homomorphismExp (name, opappe, out, params, def, clk);
  markExternalCall = s;
  return n;
}

/**
   @return the morphism of a node (node) when it is a external procedure call node.
   homomorphismExternal() is called with the name of the external procedure call.
   This name can be referenced (by &n) in the declaration of the definition associated with the opappe operator 
   in the translation table. For example:\n
   operator opappe = (? ! ) pragmas MORPHISM "(&h&o, &done) :=Cost&n{}(&h&i1, &h&i2 , &hck&o, &wait)" end pragmas; %
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismProccall (ptINode node)
{
  ptexpcal lin, lout, par, hlout;
  ptident name;
  ptinsth clk;

  fNM->get_procCall_arg (node, &name, &par, &lin, &lout);
  hlout = getCombedExprAttributes (lout);
  clk = getClockCopy (fNM->get_nodeClock (node));
  if (existPredefinedPragma (fNM->getPragmas (node), PRAGMA_CONVERSION))
    {
      fEM->getNaryArgs (&hlout, &hlout);
      return homomorphismExp (ident2repstr (name), opconv, hlout, par, lin, clk);
    }
  return homomorphismExternal (ident2repstr (name), hlout, par, lin, clk);
}

/** 
    @return the morphism of a node (node) when it is a read node (associated with an input of the program).
    The translation is given by the opread operator in the table of morphism.
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismRead (ptINode node)
{
  ptexpcal hvsig, vsig, exp, ns;
  ptINode m, hn;
  ptinsth clk;
  boolean b;

  vsig = fNM->get_io_arg (node);
  if (!getAttribute (vsig, &hvsig))
    return errorOnNode (node);
  if (!externalGraph)
    {
      removeAttribute (vsig);
      /*
         id = homomorphismDcl(fEM->getName(vsig));
         ns = mkHomoId(vsig,id);
       */
      fEM->getAttribute (vsig, iomark, (attributeValue *) & ns);
      setAttribute (vsig, ns);
      sGraph->addInput (ns);
      if (!igc->addInNode (ns, NILINODE, &m))
	m = sNM->mk_node_undefined ();
      // voir si utile-----------
      b = fEM->setModeTypeChecking (pk_false);	// <<<<<--
      exp = fEM->mkUnaryExp (opread, vsig, fEM->getBackAnnotation (vsig));
      fEM->setModeTypeChecking (b);	// <<<<<--
      clk = sEM->getClock (hvsig);
      hn = homomorphismExp (hvsig, exp, clk);
      //-----------
    }
  else
    {
      if (!igc->addInNode (hvsig, NILINODE, &m))
	m = sNM->mk_node_undefined ();
      sGraph->addInput (hvsig);
    }
  return m;
}

/** 
    @return the morphism of a node (node) when it is a write node (associated with an output of the program).
    The translation is given by the opwrite operator in the table of morphism.
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismWrite (ptINode node)
{
  ptexpcal hvsig, vsig, exp, ns;
  ptINode hn;
  ptinsth clk;
  boolean b;

  vsig = fNM->get_io_arg (node);
  if (!getAttribute (vsig, &hvsig))
    return errorOnNode (node);
  /*
     id = homomorphismDcl(fEM->getName(vsig)); 
     ns = mkHomoId(vsig,id); 
   */
  fEM->getAttribute (vsig, iomark, (attributeValue *) & ns);
  sGraph->addOutput (ns);
  b = fEM->setModeTypeChecking (pk_false);	// <<<<<--why ?
  exp = fEM->mkUnaryExp (opwrite, vsig, fEM->getBackAnnotation (vsig));
  fEM->setModeTypeChecking (b);	// <<<<<-- why?
  clk = sEM->getClock (hvsig);
  hn = homomorphismExp (ns, exp, clk);
  removeAttribute (vsig);
  setAttribute (vsig, ns);
  return hn;
}


/**
   @return the morphism of a node (node) when it is an undefined node.
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismUndefined (ptINode node)
{
  ptexpcal s, hs;

  s = fNM->get_leftOperand (node);
  if (!getAttribute (s, &hs))
    return errorOnNode (node);
  return sEM->getNode (hs);
}

/**
   @return the morphism of a node (node) when it is a set of partial definitions.
   Each partial equation is translated homomorphismExp(ptexpcal, ptexpcal, ptinsth ).
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismPartialDefs (ptINode node)
{
  ptexpcal s, hs, out, vdef;
  ptlINode ln;
  ptINode n, hn;
  ptinsth h;

  s = fNM->get_leftOperand (node);
  if (!getAttribute (s, &hs))
    return errorOnNode (node);

  sEM->addMark (hs, PART_DEFINITION);
  ln = fNM->get_setOfPartialDefinitions_arg (node);	// set of X ::= E
  while (ptlINode_get_elem (&ln, &n))
    {				// n is a part definition node
      h = fNM->get_contributionClockEquation (n);
      h = getContextClock (h, getClockCopy (fNM->get_nodeClock (n)));
      // if (fCKM->satisfiesInitClock(h)) h = fNM->get_nodeClock(n); h = getClockCopy(h);
      out = newSignalAt (nilId, h);
      fNM->get_rightOperand (n, &vdef);
      hn = homomorphismExp (out, vdef, h);
      sNM->mk_node_equation (hs, out, h, ptlinstances_insert (NILLINSTANCES, out), NILBACKANNOTATION);
    }
  sEM->removeMark (hs, PART_DEFINITION);
  return sEM->getNode (hs);
}

/**
   @return aNode.
   Without effect: nothing to do, see homomorphismPartialDefs() method
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismPartialDef (ptINode aNode)
{
  return aNode;
}

/**
   @return the morphism of a node (node) when it is a node call.
   homomorphismExternal() is called with the name of the referenced graph.
   This name can be referenced (by &n) in the declaration of the definition associated with the opappe operator 
   in the translation table. For example:\n
   operator opappe = (? ! ) pragmas MORPHISM "(&h&o, &done) :=Cost&n{}(&h&i1, &h&i2 , &hck&o, &wait)" end pragmas; %
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismNodecall (ptINode node)
{
  ptexpcal lin, lout, par, hlout;
  ptident name;
  ptinsth clk;
  ptDCG vg;

  fNM->get_nodeCall_arg (node, &vg, &par, &lin, &lout);
  name = vg->getName ();
  hlout = getCombedExprAttributes (lout);
  clk = getClockCopy (fNM->get_nodeClock (node));
  return homomorphismExternal (ident2repstr (name), hlout, par, lin, clk);
}

/** 
    @return the node associated with a "label" (hlabel) of a subgraph (node).
    This node is created by the translation associated with the operator opentsor
    given in the morphism table.
    \see homomorphismExp(ptexpcal, ptexpcal, ptinsth ).
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismLabelSubGraph (ptexpcal hlabel, ptINode node)
{
  ptinsth clk;
  ptexpcal insOuts, lins, louts;

  beginHomomorphismNode (node);
  clk = sEM->getClock (hlabel);
  lins = fEM->convertLinstances2ListExpcal (fNM->getInodelInputFlows (node));
  louts = fEM->convertLinstances2ListExpcal (fNM->getInodelOutputFlows (node));
  // op = opentsor;  getStrOperatorName(opentsor, &vstr);
  //return homomorphismExp(vstr, op, hlabel, NILEXPCAL, exp, clk);
  insOuts = fEM->mkBinaryExp (opentsor, lins, louts, MK_ANNOT_NIL());	// construire une interface?
  return homomorphismExp (hlabel, insOuts, clk);
}


/**
   @return the morphism of a node (node) when it is a subgraph.
   See homomorphismLabelSubGraph() for the label translation.
   \note  doc must be completed.
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismSubGraph (ptINode node)
{
  ptinsth h, nh, saveClk, hnsg;
  ptlinsth lh, lh1;
  ptlINode ln;
  ptINode n, nsg, aNode, save;
  ptexpcal vlabel, vlabelCopy, e, hvlabel;
  ptexpcal lins, louts, hlins, hlouts;
  ptident name;

  if (sNM->getAttribute (node, attributeNumber, (attributeValue *) & n))
    return n;

  fNM->get_label (node, &vlabel);
  vlabelCopy = getIdCopy (vlabel);
  if (!getAttribute (vlabel, &hvlabel))
    return errorOnNode (node);
  // hvlabel: signal attribut (costLabel par exemple)
  setAttribute (vlabelCopy, hvlabel);
  nsg = sNM->mk_node_subgraph (vlabelCopy, fNM->getBackAnnotation (node));
  if (fNM->get_clusterName (node, &name))
    {
      name = makeIdent_ident_ident (name, idMorphismSuffix);
      sNM->set_clusterName (nsg, name);
    }

  h = fNM->get_nodeClock (node);
  // push  
  save = currentBlock;
  currentBlock = nsg;
  hnsg = getClockCopy (h);
  saveClk = currentClkDone;
  currentClkDone = hnsg;
  // -------

  fNM->get_inOutGraphRef (node, &lins, &louts);
  hlins = getCombedExprAttributes (lins);
  hlouts = getCombedExprAttributes (louts);
  sNM->set_inOutGraphRef (nsg, hlins, hlouts);
  if (fNM->is_trainCluster (node))
    sNM->set_trainClusterTag (nsg);

  fNM->get_subGraphDefinitions (node, &ln, &lh);
  while (ptlINode_get_elem (&ln, &aNode))
    homomorphismNode (aNode, &n);

  lh1 = lh;
  while (ptlinsth_get_next (&lh, &h))
    if (isSelectedClock (h))
      if (!fCKM->getAttribute (h, attributeNumber, (attributeValue *) & e))
	homomorphismClock (h);

  lh = lh1;
  while (ptlinsth_get_next (&lh, &h))
    if (isSelectedClock (h))
      {
	nh = getClockCopy (h);
	sNM->subGraphInsertClock (nsg, nh);
	// et le calcul aussi ?
      }
  sNM->unlinkNodesFromClock (nsg);
  // Pop
  currentBlock = save;
  currentClkDone = saveClk;
  // -----
  insertNode (nsg, hnsg);
  homomorphismLabelSubGraph (hvlabel, node);
  return nsg;
}


/**
   @return the morphism of a node (node) when it is a bundle (list of fields).
   It calls homomorphismExternal() with the type of the bundle as name.
   This name can be referenced (by &n) in the declaration of the definition associated with the opappe operator 
   in the translation table. For example:\n
   operator opappe = (? ! ) pragmas MORPHISM "(&h&o, &done) :=Cost&n{}(&h&i1, &h&i2 , &hck&o, &wait)" end pragmas; %
*/
_VIRTUAL_ ptINode
Homomorphism::homomorphismBundle (ptINode node)
{
  ptlinstances ls;
  ptexpcal lins, vout, hvout;
  ptinsth clk;
  ptrepstr name;

  ls = fNM->get_lfields_arg (node);
  lins = sEM->convertLinstances2ListExpcal (ls);
  vout = fNM->get_leftOperand (node);
  if (!getAttribute (vout, &hvout))
    return errorOnNode (node);
  name = getTypeStrAsEnumVal (fEM->getType (vout));
  clk = getClockCopy (fNM->get_nodeClock (node));
  return homomorphismExternal (name, hvout, NILEXPCAL, lins, clk);
}


/**
   It applies the homomorphismNodes() for the nodes of each sub-graph of the original graph.
*/
_VIRTUAL_ void
Homomorphism::homomorphismSubGraphs ()
{
  ptlINode l;
  if (fNM->getSubGraphs (&l))
    homomorphismNodes (l);
}


/**
   It inserts a node (n) in the new graph, at the clock h.
   It takes into account some context informations: the node is also inserted in a sub-graph, if the current block is set.
   Moreover, the node is marked "DONTDECLARE" (global mark of the compiler) to specify that the declaration of the model is not necessary,
   if markExternalCall is "false" (markExternalCall is an attribute of the class).
*/
_VIRTUAL_ void
Homomorphism::insertNode (ptINode n, ptinsth h)
{
  sNM->insertNode (n, h);
  if (!markExternalCall)
    sNM->addMark (n, DONTDECLARE);
  if (sNM->is_subgraph (currentBlock))
    sNM->subGraphInsertNode (currentBlock, n);
}


/**
   It achieves the morphism for the write nodes of the program.
   It applies homomorphismNodes() on the list of the write nodes.
*/
_VIRTUAL_ void
Homomorphism::homomorphismInterfaceOutputs ()
{
  ptlINode ln;
  ptlinstances ls, outputs;
  ptexpcal nsig;

  if (!externalGraph)
    if (fNM->getOutputs (&ln))
      homomorphismNodes (ln);

  ls = fGraph->getOutputs ();
  outputs = getCurrentInstances (ls);	// getAttributesOf(ls);
  igc->addOutNodes (outputs);

  if (externalGraph)
    while (ptlinstances_get_next (&outputs, &nsig))
      sGraph->addOutput (nsig);
}

/**
   It achieves the morphism for the read nodes of the program.
   It applies homomorphismNodes() on the list of the read nodes
*/
_VIRTUAL_ void
Homomorphism::homomorphismInterfaceInputs ()
{
  ptlINode ln;
  if (fNM->getInputs (&ln))
    homomorphismNodes (ln);
}

/**
   It achieves the morphism for the read/write nodes of the program.
   It calls 
   - homomorphismInterfaceInputs() 
   - homomorphismInterfaceOutputs()
*/
_VIRTUAL_ void
Homomorphism::homomorphismInterface ()
{
  homomorphismInterfaceInputs ();
  homomorphismInterfaceOutputs ();
}


/**
   For each node of ln (assumed to be a read/write node of x of the original graph), 
   it creates the clone of x in the new graph.
   An attribute (iomark, attrribute of the class) is used to link (x and its clone).
   This mark will be used in homomorphismRead(), homomorphismWrite().
*/
_VIRTUAL_ void
Homomorphism::mkAttributeInterface (ptlINode ln)
{
  ptexpcal vsig, ns;
  ptINode node;

  while (ptlINode_get_elem (&ln, &node))
    {
      vsig = fNM->get_io_arg (node);
      ns = mkHomoId (vsig, homomorphismDcl (fEM->getName (vsig)));
      fEM->addAttribute (vsig, iomark, (attributeValue) ns);
    }
}

/**
   It applies mkAttributeInterface(ptlINode) for each read-node and write-node of the graph.
*/
_VIRTUAL_ void
Homomorphism::mkAttributeInterface ()
{
  ptlINode ln;

  if (!externalGraph)
    {
      if (fNM->getInputs (&ln))
	mkAttributeInterface (ln);
      if (fNM->getOutputs (&ln))
	mkAttributeInterface (ln);
    }
}
