/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   Constructor of a table useful for the morphisms of Signal programs.
   - the table is filled by mkOperatorTable().
   - IdentLocate() gives  access to an element of the table.
*/
  HomomorphismTable::HomomorphismTable ()
{
  IdentTable = new IdentifierTable ();
}

/**
   Destructor.
*/
_VIRTUAL_ HomomorphismTable::~HomomorphismTable ()
{
}

/**
   It translates a graph of a module (aModule) into a homomorphism table.
   \see module2Table()
 */
_VIRTUAL_
  boolean
HomomorphismTable::mkOperatorTable (ptDCG aModule)
{
  return module2Table (aModule);
}


/**
   @return "true" if an identifier (id) is an entry point of the homomorphism table, "false" otherwise.
   @param id (in) an operator identifier
   @param lg (out) the set of the graph with id as name (polymorphism: it is possible to have several definitions for the same operators).
   <em> The caller must choose the good one </em>. This parameter is available if the identifier is defined in the table.
*/
_VIRTUAL_ boolean
HomomorphismTable::IdentLocate (ptident id, ptlDCG * lg)
{
  ptelemHashTable elem;

  if (eq_idents (id, NILIDENT))
    return pk_false;
  if (!IdentTable->findOrInsert (id, &elem))
    return pk_false;
  (*lg) = (ptlDCG) (elem->getExternalRef ());
  return ((*lg) != NILLDCG);
}
