/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return "true" if the operation succeeds, "false" otherwise (the parameter is not a graph of a module)
   For each graph of the module (aModule) if inserts the element in the table (SymbolTableAddGraph()).
*/
  _VIRTUAL_ boolean HomomorphismTable::module2Table (ptDCG aModule)
{
  ptlDCG
    lg;
  ptDCG
    aGraph;
  pointer
    elem;
  boolean
    b;

  if (!aModule->getModuleGraphs (&lg))
    return pk_false;
  b = pk_true;
  while (ptlPointer_get_elemNext (&lg, &elem))
    {
      aGraph = (ptDCG) elem;
      SymbolTableAddGraph (aGraph);
    }
  return b;
}


/**
   It adds in the entry of the current table, associated with the name of the operand graph (anOperator).
   The same entry can have several graph (polymorphism).
*/
_VIRTUAL_ void
HomomorphismTable::SymbolTableAddGraph (ptDCG anOperator)
{
  ptelemHashTable elem;
  ptident id;
  ptlDCG lg;

  id = anOperator->getName ();
  if (IdentTable->findOrInsert (id, &elem))
    lg = (ptlPointer) (elem->getExternalRef ());
  else
    lg = NILLDCG;
  if (ptlPointer_insert_elem (&lg, (pointer) anOperator))
    elem->setExternalRef ((pointer) lg);
}
