/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define STRING_MORPHISM_SUFFIX "_M"
#define UNDEFINED_IDENT NILIDENT
#define isEventClockedFormat(t) ((t == DCG_BASIC) ||(t == DCG_POLY) || (t == DCG_ENDO))
#define MK_ANNOT_NIL()  NILBACKANNOTATION


/** @defgroup initHomoGroup Homomorphism: initializing step. */
/**
   @ingroup initHomoGroup
   It initializes the associative table tableIdentOperator.
   To each operator(int) is associated the identifier used in the homomorphism table
   to represent it. Some operators are associated with a predefined undefined name (UNDEFINED_IDENT), they
   are assumed not defined in the morphism table.

   \code
   tableIdentOperator[opplus] = string2ident("+");
   and in the homomorphism table, we have
   operator + = (?x,y; !z; ) pragmas Morphism "(&h&o, &done) :=CostPlus{&t&i1,&t&i2}( &h&i1, &h&i2, &hck&o, &wait)" end pragmas;
   \endcode
*/
  _VIRTUAL_ void
Homomorphism::InitOp2stringOpTable ()
{
  tableIdentOperator[opnom] = string2ident (H_RENAME);
  tableIdentOperator[opchar] = string2ident ("char");
  tableIdentOperator[opcomplex] = string2ident ("complex");
  tableIdentOperator[opcstcar] = string2ident (H_CONSTANT);
  tableIdentOperator[opcstcha] = string2ident (H_CONSTANT);
  tableIdentOperator[opcstdp] = string2ident ("double");
  tableIdentOperator[opcstent] = string2ident (H_CONSTANT);
  tableIdentOperator[opcstreel] = string2ident (H_CONSTANT);
  tableIdentOperator[opdcomplex] = string2ident ("dcomplex");
  tableIdentOperator[opdpreal] = string2ident ("dpreal");
  tableIdentOperator[opevent] = string2ident ("event");
  tableIdentOperator[opfalse] = string2ident (H_CONSTANT);
  tableIdentOperator[ophnull] = string2ident ("^0");

  tableIdentOperator[opimplicit] = UNDEFINED_IDENT;
  tableIdentOperator[opinteger] = string2ident ("integer");
  tableIdentOperator[oplogical] = string2ident ("logical");
  tableIdentOperator[oplong] = string2ident ("long");
  tableIdentOperator[opnil] = string2ident ("opnil");

  tableIdentOperator[opnullproc] = UNDEFINED_IDENT;

  tableIdentOperator[opreal] = string2ident ("real");
  tableIdentOperator[opretun] = UNDEFINED_IDENT;
  tableIdentOperator[opshort] = string2ident ("short");
  tableIdentOperator[opstring] = string2ident ("string");
  tableIdentOperator[optabvide] = UNDEFINED_IDENT;
  //  tableIdentOperator[optrue] = string2ident("true");
  tableIdentOperator[optrue] = string2ident (H_CONSTANT);
  tableIdentOperator[opunsafe] = string2ident ("unsafe");
  tableIdentOperator[opcomment] = UNDEFINED_IDENT;

  // tableIdentOperator[opcorpp] = UNDEFINED_IDENT;
  tableIdentOperator[opevnt] = string2ident ("^");
  tableIdentOperator[opexte] = UNDEFINED_IDENT;
  tableIdentOperator[opnega] = string2ident ("-");
  tableIdentOperator[opnon] = string2ident ("not");
  tableIdentOperator[oppcar] = UNDEFINED_IDENT;
  tableIdentOperator[opsdec] = UNDEFINED_IDENT;
  tableIdentOperator[opstru] = UNDEFINED_IDENT;
  tableIdentOperator[optran] = string2ident ("tr");
  tableIdentOperator[opvvarin] = UNDEFINED_IDENT;
  //  tableIdentOperator[opvconp] = UNDEFINED_IDENT;
  //  tableIdentOperator[opvtypp] = UNDEFINED_IDENT;
  tableIdentOperator[opwhen1] = string2ident ("when");

  tableIdentOperator[opcstenum] = string2ident ("enum");
  tableIdentOperator[opindirection] = UNDEFINED_IDENT;

  tableIdentOperator[opread] = string2ident (H_READ);
  tableIdentOperator[opwrite] = string2ident (H_WRITE);

  // tableIdentOperator[opbinary] = string2ident("opbinary");
  tableIdentOperator[opappe] = string2ident (H_EXTERNALCALL);
  tableIdentOperator[opbinf] = string2ident ("^*");
  tableIdentOperator[opbsup] = string2ident ("^+");
  tableIdentOperator[opbund] = UNDEFINED_IDENT;

  tableIdentOperator[opcas] = UNDEFINED_IDENT;

  tableIdentOperator[opcham] = string2ident (H_FIELDACCESS);
  tableIdentOperator[opcomp] = string2ident ("^-");
  tableIdentOperator[opconc] = string2ident ("||");
  tableIdentOperator[opcons] = string2ident ("=");
  tableIdentOperator[opconv] = string2ident ("Conversion");	//UNDEFINED_IDENT;
  tableIdentOperator[opcplx] = string2ident ("@");
  tableIdentOperator[opdeccon] = string2ident ("constant");
  tableIdentOperator[opdecconp] = UNDEFINED_IDENT;
  tableIdentOperator[opdecsig] = string2ident (H_IDENT);

  tableIdentOperator[opdecvar] = string2ident ("statevar");

  tableIdentOperator[opdefegal] = string2ident (":=:");

  tableIdentOperator[opdefpart] = string2ident ("::=");
  tableIdentOperator[opdefpartdef] = string2ident ("::= defaultvalue");

  tableIdentOperator[opdefsig] = string2ident (":=");
  tableIdentOperator[opdesc] = UNDEFINED_IDENT;

  tableIdentOperator[opdescmodu] = UNDEFINED_IDENT;

  tableIdentOperator[opdiff] = string2ident ("/=");
  tableIdentOperator[opdivi] = string2ident ("/");
  tableIdentOperator[opegal] = string2ident ("=");
  tableIdentOperator[opegeg] = string2ident ("==");
  tableIdentOperator[opelem] = string2ident (H_ELEMACCESS);
  tableIdentOperator[opelevec] = string2ident ("[]");
  tableIdentOperator[opensvec] = string2ident (H_ENSVEC);
  tableIdentOperator[opentsor] = string2ident (H_LABEL);
  tableIdentOperator[opet] = string2ident ("and");
  tableIdentOperator[opextr] = string2ident ("when");

  tableIdentOperator[opformact] = string2ident ("action");

  tableIdentOperator[opformfunc] = string2ident ("function");
  tableIdentOperator[opformnode] = string2ident ("node");
  tableIdentOperator[opformproc] = string2ident ("process");

  tableIdentOperator[opiegal] = string2ident ("<=");
  tableIdentOperator[opiieg] = string2ident ("<<=");
  tableIdentOperator[opindmot] = UNDEFINED_IDENT;
  tableIdentOperator[opinfe] = string2ident ("<");
  tableIdentOperator[opinit] = string2ident ("init");
  tableIdentOperator[opinst] = string2ident ("opinst");

  tableIdentOperator[opintvff] = UNDEFINED_IDENT;
  tableIdentOperator[opintvfo] = UNDEFINED_IDENT;
  tableIdentOperator[opintvof] = UNDEFINED_IDENT;
  tableIdentOperator[opintvoo] = UNDEFINED_IDENT;

  tableIdentOperator[opmele] = string2ident ("default");
  tableIdentOperator[opmmul] = string2ident ("*.");
  tableIdentOperator[opmod] = string2ident ("modulo");
  tableIdentOperator[opmodu] = string2ident ("module");
  tableIdentOperator[opmoin] = string2ident ("-");
  tableIdentOperator[opmult] = string2ident ("*");
  tableIdentOperator[opnomba] = string2ident ("after");
  tableIdentOperator[opnombd] = string2ident ("from");
  tableIdentOperator[opnombm] = string2ident ("count");
  tableIdentOperator[opobli] = UNDEFINED_IDENT;
  tableIdentOperator[opou] = string2ident ("or");

  tableIdentOperator[opouex] = string2ident ("xor");

  tableIdentOperator[oppconf] = UNDEFINED_IDENT;
  tableIdentOperator[oppetiq] = string2ident (H_LABEL);	// string2ident("::");
  tableIdentOperator[opparc] = UNDEFINED_IDENT;
  tableIdentOperator[opplus] = string2ident ("+");
  tableIdentOperator[oppuis] = string2ident ("**");
  tableIdentOperator[oprecu] = string2ident ("\\");
  tableIdentOperator[oprepe] = string2ident ("|*");
  tableIdentOperator[oprest] = UNDEFINED_IDENT;
  tableIdentOperator[opsegal] = string2ident (">=");
  tableIdentOperator[opsequ] = string2ident ("next");
  tableIdentOperator[opspec] = string2ident ("spec");
  tableIdentOperator[opsupe] = string2ident (">");
  tableIdentOperator[optvect] = UNDEFINED_IDENT;

  tableIdentOperator[optypefunc] = UNDEFINED_IDENT;
  tableIdentOperator[optypenode] = UNDEFINED_IDENT;
  tableIdentOperator[optypeproc] = UNDEFINED_IDENT;

  tableIdentOperator[opvenum] = string2ident (H_ENUMVALUEACCESS);

  tableIdentOperator[opvar] = string2ident ("var");
  tableIdentOperator[opident] = string2ident (H_RENAME);

  tableIdentOperator[opleft_tt] = string2ident ("leftTT");

  //tableIdentOperator[oplists] = UNDEFINED_IDENT;
  tableIdentOperator[opchai] = UNDEFINED_IDENT;
  tableIdentOperator[oplcas] = UNDEFINED_IDENT;

  tableIdentOperator[oplcons] = UNDEFINED_IDENT;
  tableIdentOperator[oplcrohor] = UNDEFINED_IDENT;
  tableIdentOperator[opldechor] = UNDEFINED_IDENT;
  tableIdentOperator[opldecs] = UNDEFINED_IDENT;
  tableIdentOperator[opldecl] = UNDEFINED_IDENT;

  tableIdentOperator[opldep] = UNDEFINED_IDENT;

  tableIdentOperator[opleflot] = UNDEFINED_IDENT;
  tableIdentOperator[oplepara] = UNDEFINED_IDENT;
  tableIdentOperator[oplesign] = UNDEFINED_IDENT;
  tableIdentOperator[oplevect] = UNDEFINED_IDENT;
  tableIdentOperator[oplelevec] = string2ident (H_LELEVEC);
  tableIdentOperator[oplequhor] = UNDEFINED_IDENT;
  tableIdentOperator[oplexchor] = UNDEFINED_IDENT;
  tableIdentOperator[oplfeed] = UNDEFINED_IDENT;
  tableIdentOperator[oplident] = UNDEFINED_IDENT;

  tableIdentOperator[oplimpo] = UNDEFINED_IDENT;
  tableIdentOperator[oplmodu] = UNDEFINED_IDENT;

  tableIdentOperator[oplopra] = UNDEFINED_IDENT;
  tableIdentOperator[oplparc] = UNDEFINED_IDENT;
  tableIdentOperator[oplprag] = UNDEFINED_IDENT;
  tableIdentOperator[oplsign] = UNDEFINED_IDENT;
  //  tableIdentOperator[oplsync] = UNDEFINED_IDENT;
  tableIdentOperator[opltype] = UNDEFINED_IDENT;
  tableIdentOperator[opltypep] = UNDEFINED_IDENT;

  tableIdentOperator[optenum] = UNDEFINED_IDENT;

  tableIdentOperator[opcomments] = UNDEFINED_IDENT;
  tableIdentOperator[oplsuff] = UNDEFINED_IDENT;
  tableIdentOperator[oplread] = UNDEFINED_IDENT;
  tableIdentOperator[oplwrite] = UNDEFINED_IDENT;
  tableIdentOperator[opldefpart] = UNDEFINED_IDENT;

  // tableIdentOperator[opternary] = string2ident("opternary");

  tableIdentOperator[opact] = string2ident ("action");
  tableIdentOperator[opactp] = string2ident ("action");	// private

  tableIdentOperator[opborn] = UNDEFINED_IDENT;
  tableIdentOperator[opcond] = string2ident ("if");
  tableIdentOperator[opdepd] = string2ident (H_DEPENDENCE);
  tableIdentOperator[openum] = UNDEFINED_IDENT;
  tableIdentOperator[opfen] = string2ident ("window");
  tableIdentOperator[opfunc] = string2ident ("function");
  tableIdentOperator[opfuncp] = UNDEFINED_IDENT;
  tableIdentOperator[opinte] = UNDEFINED_IDENT;

  tableIdentOperator[opmemo] = string2ident ("cell");
  tableIdentOperator[opmoti] = string2ident ("array");

  tableIdentOperator[opnode] = string2ident ("node");
  tableIdentOperator[opnodep] = UNDEFINED_IDENT;

  tableIdentOperator[oppcas] = UNDEFINED_IDENT;
  tableIdentOperator[opprag] = UNDEFINED_IDENT;
  tableIdentOperator[opprocess] = string2ident ("process");
  tableIdentOperator[opprocessp] = UNDEFINED_IDENT;
  tableIdentOperator[opret] = string2ident ("$");
  tableIdentOperator[optype] = string2ident (H_TYPE);

  tableIdentOperator[opnodecall] = UNDEFINED_IDENT;
  tableIdentOperator[opsubgraph] = UNDEFINED_IDENT;

  tableIdentOperator[opleflot] = string2ident (H_TUPLE_ASS);	// x:= (e1,e2,...) 
}

/**
   @ingroup initHomoGroup
   It defines the G_HomoType attribute member.
   If the optype operator is defined in the homomorphism table via the name H_TYPE (see morphismPredefinedNames.h)
   the definition is taken otherwise _HomoType_.

   \code
   operator _TYPE_         = (? ! ) pragmas Morphism "HomoType" end pragmas;
   \endcode
*/
_VIRTUAL_ void
Homomorphism::mkhomomorphismType ()
{
  ptrepstr idstr;
  ptStrScanner ScanPragObj;

  if (!scannerOnOp (optype, &ScanPragObj))
    idstr = string2repstr ("_HomoType_");
  else if (!ScanPragObj->getIdent (&idstr))
    idstr = string2repstr ("_HomoType_");
  G_HomoType = ExternTypeCreate (idstr);
  sEM->mkExpType (G_HomoType);	// necessaire pour avoir la dcl du type avant la constante G_xxxx
}

/**
   @ingroup initHomoGroup
   It initializes the homomorphism translation from an original graph (g1) to a new one (g2).
   In particular it initializes the member of the class: 
   - str_pragma_morphism, 
   - the associative table (operator, name of the operator) (InitOp2stringOpTable()), 
   - currentClkDone, doneAttribute, 
   - some identifiers :nilId, idMorphismSuffix, doneId, waitId, packId, idInternal, waitIdsbdcplus
   - G_UndefinedValue, G_HomoType (mkhomomorphismType())
   - markExternalCall: default value is false
   - aCounter is set to 0 (used to enchure the unicity of some calls (referenced in the morphism table by &cpt)
   - externalGraph: attribute of g1 graph.
*/
_VIRTUAL_ void
Homomorphism::initialize (ptDCG g1, ptDCG g2)
{
  AbstractHomomorphism::initialize (g1, g2);
  str_pragma_morphism = getPredefinedPragma (PRAGMA_MORPHISM);
  InitOp2stringOpTable ();
  currentClkDone = getClockCopy (fCKM->getTickClock ());
  doneAttribute = getMark ();
  nilId = getIdentNil ();
  idMorphismSuffix = string2ident (STRING_MORPHISM_SUFFIX);
  doneId = string2ident ("done");
  waitId = string2ident ("waiting");
  packId = string2ident ("pack");
  idInternal = string2ident ("_internal");
  waitIdsbdcplus = string2ident ("waiting2");
  mkhomomorphismType ();
  G_UndefinedValue = sEM->mkIdentExp (string2ident ("NilHomoTypeValue"), tconstant, G_HomoType, NILEXPCAL, sNM->mk_node_undefined (), sCKM->getUndefClock (),  MK_ANNOT_NIL());
  getDependenceCallId ();
  //---Urgency
  if (fgraph_type == DCG_SCH)
    fGraph->getDCGSchInfos (&nbSets, &SubGraphTable, &lNodesHinit, &lClocksNotToBeInitiliazed);

  markExternalCall = pk_false;
  aCounter = 0;			// used for unicity of some calls (referenced in morphism table by &cpt for ensure unicity of some calls)
  externalGraph = g1->isGraphOfExternalProcess ();
}


/**
   @ingroup initHomoGroup
   It splits a list of nodes (ln) in two lists.
   @return the subset of the nodes of ln wihtout the node of lio.
   @param ln (in) a liust of node.
   @param lio (out) subset of ln that are io node (read, write nodes).

*/
_VIRTUAL_ ptlINode
Homomorphism::splitIONode (ptlINode ln, ptlINode * lio)
{
  ptINode n;
  ptlINode l1;

  l1 = NILLINODE;
  *lio = NILLINODE;
  while (ptlINode_get_elem (&ln, &n))
    if (fNM->is_ioNode (n))
      ptlINode_insert_elem (lio, n);
    else
      ptlINode_insert_elem (&l1, n);
  return l1;
}

/** @defgroup tableAccessGroup Homomorphism: homomorphism table access.*/
/**
   @ingroup tableAccessGroup
   @return "true" if an operator (op) is defined in the tableIdentOperator, "false" otherwise.
   @param op (in) an operator.
   @param vid (out) the identifier assocaited with the operator. Available when it returns "true".
*/
_VIRTUAL_ boolean
Homomorphism::getIdentOperatorName (noperators op, ptident * vid)
{
  int i;
  i = (int) op;
  if ((i <= 0) || (i > SIG_NB_OPERATORS))
    return pk_false;
  (*vid) = tableIdentOperator[i];
  return pk_true;
}

/**
   @ingroup tableAccessGroup
   @return 
   - "true" if an operator (op) is defined with a real name (different from UNDEFINED_IDENT) in the tableIdentOperator. The 
   name is returned in vstr.
   - "false" otherwise, the name (vstr) is nilstr value.
   
*/
_VIRTUAL_ boolean
Homomorphism::getStrOperatorName (noperators op, ptrepstr * vstr)
{
  ptident id;

  if (!getIdentOperatorName (op, &id))
    return pk_false;
  if (id == UNDEFINED_IDENT)
    {
      *vstr = nilstr;
      return pk_false;
    }
  *vstr = ident2repstr (id);
  return pk_true;
}

/**
   @ingroup tableAccessGroup
   It returns the value of the morphism pragma associated with an identifier (id).
   \code
   For id referencing the "+" operator, it returns the tree representing
      "(&h&o, &done) :=CostPlus{&t&i1,&t&i2}( &h&i1, &h&i2, &hck&o, &wait)"
   It is extracted from the definition given in the table (pragmaOperatorTable):
   operator + = (?x,y; !z; ) pragmas Morphism "(&h&o, &done) :=CostPlus{&t&i1,&t&i2}( &h&i1, &h&i2, &hck&o, &wait)" end pragmas;
   \endcode
*/
_VIRTUAL_ ntrees
Homomorphism::getPragmaIdent (ptident id)
{
  ptlDCG lg;
  pointer e;
  ptDCG g;
  ptlpragmas lp;

  if (!pragmaOperatorTable->IdentLocate (id, &lg))
    return NILTREE;
  // It takes the first element of the list (to be completed)
  if (!ptlPointer_get_elemNext (&lg, &e))
    return NILTREE;
  g = (ptDCG) e;
  lp = g->getPragmas ();
  return getEntryCodeOp (lp, str_pragma_morphism);
}

/**
   @ingroup tableAccessGroup
   @return "true" if the tree vt is defined, "false" ohterwise.
   @param op (in) an operator.
   @param vt (in) the pragma MORPHISM associated with op if it exists.
   @param ScanPragObj (out) a scanner object for analyzing vt (if it is defined).
*/
_VIRTUAL_ boolean
Homomorphism::mkScanPragObj (noperators op, ntrees vt, ptStrScanner * ScanPragObj)
{
  ptrepstr vstr;

  if (ISNOT_NILTREE (vt))
    {
      *ScanPragObj = new StrScanner (vt, pk_true);
      return pk_true;
    }
  if (!getStrOperatorName (op, &vstr))
    strconc2str (string2repstr ("Operator code : "), integer2repstr (op), pk_false, &vstr);
  pK_messenger->globalMessageStr (ErrEmptyCode, vstr);
  return pk_false;
}

/**
   @ingroup tableAccessGroup
   @return "true" if the morphism of an operator (op) is defined in the associated morphism table, "false" otherwise.
   @param op (in) an operator.
   @param ScanPragObj (out) a scanner object for analyzing the pragma MORPHISM associated to op (if its the morphism is defined),
*/
_VIRTUAL_ boolean
Homomorphism::scannerOnOp (noperators op, ptStrScanner * ScanPragObj)
{
  ptident id;
  ntrees vt;

  if (!getIdentOperatorName (op, &id))
    SET_NILTREE (vt);
  else
    vt = getPragmaIdent (id);
  return mkScanPragObj (op, vt, ScanPragObj);
}

/**
   @ingroup tableAccessGroup
   @return true if an operator (op) is defined in the homomorphism table, "false" otherwise.
   Currently the parameters vstr, out, params def are unused.
   When it returns "true" the scanner object ScanPragObj is set: it is ready to analyze the pragma
   MORPHISM assocaited with the operator.
   \see getIdentOperatorName, IdentLocate, mkScanPragObj
*/
_VIRTUAL_ boolean
Homomorphism::scannerOn (noperators op, ptrepstr vstr, ptexpcal out, ptexpcal params, ptexpcal def, ptinsth clk, ptStrScanner * ScanPragObj)
{
  ptident id;
  ntrees vt;
  ptlDCG lg;
  pointer e;
  ptDCG g;
  ptlpragmas lp;
  int n, nb;
  boolean found;

  found = pk_false;
  if (getIdentOperatorName (op, &id))
    if (pragmaOperatorTable->IdentLocate (id, &lg))
      {
	n = ptlPointer_get_size (lg);
	if (n == 1)
	  {
	    ptlPointer_get_elemNext (&lg, &e);
	    g = (ptDCG) e;
	    found = pk_true;
	  }
	else if (n != 0)
	  {
	    // several definitions: solving by the arity of the operator (should be completed by the types)
	    nb = STree_getOperatorArity (op);
	    while (!found && ptlPointer_get_elemNext (&lg, &e))
	      {
		g = (ptDCG) e;
		found = (ptlinstances_get_size (g->getInputs ()) == nb);
	      }
	  }
      }
  if (!found)
    SET_NILTREE (vt);
  else
    {
      lp = g->getPragmas ();
      vt = getEntryCodeOp (lp, str_pragma_morphism);
    }
  return mkScanPragObj (op, vt, ScanPragObj);
}

/**
   @ingroup tableAccessGroup
   Unused method: Not documented. 
 */
_VIRTUAL_ boolean
Homomorphism::scannerOnIdent (ptident id, ptStrScanner * ScanPragObj)
{
  ntrees vt;
  vt = getPragmaIdent (id);
  if (ISNOT_NILTREE (vt))
    {
      *ScanPragObj = new StrScanner (vt, pk_true);
      return pk_true;
    }
  pK_messenger->globalMessageIdent (ErrEmptyCode, id);
  return pk_false;
}


/**
   @return a new identifier name using the scanner object (ScanPragObj) 
   assumed to be initialized for the translation of such object.
   if the current token of the scanner object is
   - "ident" (tk_ident), the name is given by the identifier following the token
   - "name" (tk_name) the name is strId
   - "counter" (tk_cpt)  the internal counter (aCounter) is increased, the name is then the value of this counter.
*/
_VIRTUAL_ ptident
Homomorphism::homomorphismIdent (ptStrScanner ScanPragObj, ptrepstr strId)
{
  ptrepstr vstr;
  boolean stop;
  ScanPragmaToken tk;

  vstr = nilstr;
  stop = pk_false;
  tk = ScanPragObj->getCurrentToken ();
  do
    {
      switch (tk)
	{
	case tk_ident:
	  strconc2str (vstr, ScanPragObj->getIdent (), pk_false, &vstr);
	  break;
	case tk_name:
	  strconc2str (vstr, strId, pk_false, &vstr);
	  break;

	case tk_cpt:
	  aCounter++;
	  strconc2str (vstr, integer2repstr (aCounter), pk_false, &vstr);
	  break;

	default:
	  stop = pk_true;
	  break;
	}
      if (!stop)
	tk = ScanPragObj->getToken ();
    }
  while (!stop);
  return repstr2ident (vstr);
}

/**
   @return a new identifier name associated with id (from original graph) if the declaration of identifier is 
   defined in the homomorphism table, undefined name otherwise.
   When the declaration of identifier is defined in the homomorphism table, a scanner object is created and the 
   identifier name is produced from it (see homomorphismIdent(ptStrScanner, ptrepstr)).
*/
_VIRTUAL_ ptident
Homomorphism::homomorphismDcl (ptident id)
{
  ptStrScanner ScanPragObj;

  if (!scannerOnOp (opdecsig, &ScanPragObj))
    return NILIDENT;
  ScanPragObj->getToken ();
  return homomorphismIdent (ScanPragObj, ident2repstr (id));
}

/**
   @return the identifier name associated with a signal (vsig) in the new graph.
   It calls homomorphismDcl(ptident) for this.
*/
_VIRTUAL_ ptident
Homomorphism::homomorphismIdent (ptexpcal vsig)
{
  ptident id;

  id = fEM->getName (vsig);
  if (!externalGraph)
    {
      if (fEM->IsInputFlow (vsig))
	id = makeIdent_ident_ident (id, idInternal);
      if (fEM->IsOutputFlow (vsig))
	id = makeIdent_ident_ident (id, idInternal);
    }
  return homomorphismDcl (id);
}

/**
   @return the identifier name associated with a clock (clk) in the new graph.
   It calls homomorphismDcl(ptident) for this.
*/
_VIRTUAL_ ptident
Homomorphism::homomorphismIdent (ptinsth clk)
{
  ptident id;

  id = fCKM->getName (clk);
  if (id == NILIDENT)
    if (!isEventClockedFormat (fgraph_type))
      id = fEM->getOriginalName (fCKM->getBooleanSignal (clk));
  return homomorphismDcl (id);
}


/**
   @return the homomorphism type of a type (TT) in the new graph.
*/
_VIRTUAL_ TypeType
Homomorphism::homomorphismType (TypeType TT)
{
  ptStrScanner ScanPragObj;
  if (!scannerOnOp (optype, &ScanPragObj))
    return TT;
  return G_HomoType;
}

/** 
    It codifies the types of the program by a enumerated type (referenced by &t in MORPHISM pragma).
    For example, type __TC__ = enum (_event, _logical, _short, _integer, _long, _real, _dpreal, _complex, _dcomplex, _char, _string);
 */
_VIRTUAL_ void
Homomorphism::homomorphismTypeCode ()
{
  ptlinstances ltypes;
  TypeType TT;
  ptexpcal elem;
  NameList NL;
  ptident id;
  NameClass typeName;
  int i;

  NL = NameListCreate ();
  NL = NameListRegister (NL);
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (PureType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (BooleanType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (ShortType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (IntegerType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (LongType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (RealType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (DRealType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (ComplexType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (DComplexType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (CharType));
  NameListAppendName (NL, (NameClass) getTypeStrAsEnumVal (StringType));
  if (fGraph->getUserTypesOneLevel (&ltypes))
    while (ptlinstances_get_next (&ltypes, &elem))
      {
	TT = fEM->getExpType (elem);
	typeName = (NameClass) getTypeStrAsEnumVal (TT);
	if (!NameListNameOrdinal (NL, typeName, &i))
	  NameListAppendName (NL, typeName);
      }
  NL = NameListRegister (NL);
  id = fNameManager->mkNameType (string2ident (H_TYPECODING_CST));
  TypeCodeEnum = EnumerationTypeCreate (ident2repstr (id), NL);
  sEM->mkExpType (TypeCodeEnum);
}

/** @defgroup finalHomoGroup Homomorphism: finalizing step. */
/**
   @ingroup finalHomoGroup
   Redefintion of the AbstractHomomorphism::finalize() method.
*/
_VIRTUAL_ void
Homomorphism::finalize ()
{
}

/**
   @ingroup finalHomoGroup
   It finalizes the translation.
*/
_VIRTUAL_ void
Homomorphism::homomorphismEnd ()
{
  fGraph->unmarkObjects (doneAttribute);
  if (externalGraph)
    sGraph->setGraphOfExternalProcess (fGraph->getSIGNALTreeDeclaration (), fGraph->getImplementationLanguage ());
  sGraph->addConstant (G_UndefinedValue);
}
