/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" if a node (node) is found in the sbDC+ graph, "false" otherwise.
   When the node (node) is found, prevNode (out) is the predecessor of node in the graph.   
*/
  _VIRTUAL_ boolean Homomorphism::getInfosSchedulling (ptINode node, ptINode * prevNode)
{
  boolean
    bDelay;
  int
    i;
  ptlINode
    ln;

  bDelay = fNM->is_delayNodeOrDelayClusterNode (node);
  i = 0;
  while (i < nbSets)
    {
      if (bDelay)
	ln = SubGraphTable[i]->ldelays;
      else
	ln = SubGraphTable[i]->ln;
      if (ptlINode_isIn_getPrevious (ln, node, prevNode))
	return pk_true;
      i++;
    }
  return pk_false;
}

/**
   @return the number of closed blocks (not implemented, return 1).
   It will be interesting to take into account (for a sbDC+ graph) the number of blocks to close/open (nbOpen) for a
   node (n1) and the previous one (n2) in the sBDC+ graph.
   Currently, it returns 1 and nbOpen=0.
*/
_VIRTUAL_ int
Homomorphism::getNbOpenCloseBlocks (ptINode n1, ptINode n2, int *nbOpen)
{
  *nbOpen = 0;
  return 1;
}

/**
   @return null expression (not implemented).
   It will be interesting to take into account (for a sbDC+ graph) the number of blocks to close/open for the 
   current node between its predecessor in the sequentialization.
*/
_VIRTUAL_ ptexpcal
Homomorphism::buildParams (int nbClose, int nbOpen)
{
  return NILEXPCAL;
}


/**
   When the graph is a sbDC+ one (sequentialized) the reinforcing must be taken into account.
   @return "true" if the graph is sbDC+ one, "false" otherwise.
   @param node (in) a node 
   @param clk (in) a clock context used to produced the undefined expression node at clk.
   @param definedAtClk (in)  a clock context used to produced the "wait" node.
   @param wa (out) the signal defining the precedence.
   \see mkWaitNodeAt()
*/

_VIRTUAL_ boolean
Homomorphism::sbDCplusWaitSignal (ptINode node, ptinsth clk, ptinsth definedAtClk, ptexpcal * wa)
{
  ptINode prevNode;
  int nbClose, nbOpen;
  ptexpcal params;

  if (fgraph_type != DCG_SCH)
    return pk_false;
  if (getInfosSchedulling (node, &prevNode))
    {
      nbClose = getNbOpenCloseBlocks (node, prevNode, &nbOpen);
      if (prevNode != NILINODE)
	{
	  params = buildParams (nbClose, nbOpen);
	  *wa = mkWaitNodeAt (NILEXPCAL, getSetSignalDone (prevNode), definedAtClk, waitIdsbdcplus, params);
	}
      else
	*wa = undefinedValueAt (clk);
    }
  else
    *wa = undefinedValueAt (clk);
  return pk_true;
}
