/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
/**
   Redefinition .
   The name of the clock will be set later. It is set to undefined name.
*/
  _VIRTUAL_ void
Inclusion::setClockName (ptinsth aClock, ptinsth aClockCopy)
{
  sCKM->setName (aClockCopy, NILIDENT);
  // The  name will be set later.
}

/**
   @return "true" if the node is seltected for the translation, "false" otherwise.
*/
_VIRTUAL_ boolean
Inclusion::isSelectedNode (ptINode n)
{
  if (fNM->is_readNode (n))
    return pk_false;
  if (fNM->is_writeNode (n))
    return pk_false;
  return Duplicator::isSelectedNode (n);
}

/**
   It assigns the i-th effective value (effective) to the i-th formal expression (formal) for each element of the lists.
   Assumed: the lists have the same number of elements.
   The link between expressions is etablished by an attribute (attributeNumber).
*/
_VIRTUAL_ void
Inclusion::ConnectInstances (ptexpcal formal, ptexpcal effective)
{
  ptexpcal cur_formal, cur_eff;

  while (formal != NILEXPCAL)
    {
      cur_formal = fEM->getNaryArgs (&formal);
      cur_eff = fEM->getNaryArgs (&effective);
      fEM->addAttribute (cur_formal, attributeNumber, (attributeValue) cur_eff);
    }
}

/**
   Similar to ConnectInstances() for the nodes.
   It links the i-th node of the effective value (effective) to the i-th node of the formal expression (formal) for each element of the lists.
   The link between nodes is etablished by an attribute (attributeNumber).
*/
_VIRTUAL_ void
Inclusion::ConnectInputNodes (ptexpcal formal, ptexpcal effective)
{
  ptexpcal cur_formal, cur_eff;
  ptINode read, effNode;

  while (formal != NILEXPCAL)
    {
      cur_formal = fEM->getNaryArgs (&formal);
      cur_eff = fEM->getNaryArgs (&effective);
      if (fEM->IsSignal (cur_eff))	// a modifier
	{
	  read = fEM->getNode (cur_formal);
	  effNode = fEM->getNode (cur_eff);
	  fNM->addAttribute (read, attributeNumber, (attributeValue) effNode);
	}
    }
}

/**
   Similar to ConnectInstances() for the clocks.
   It links the i-th clock of the effective value (effective) to the i-th clock of the formal expression (formal) for each element of the lists.
   The link between clock etablished by an attribute (attributeNumber).
*/
_VIRTUAL_ void
Inclusion::ConnectClocks (ptexpcal formal, ptexpcal effective)
{
  ptexpcal cur_formal, cur_eff;
  ptinsth hformal, heff;

  while (formal != NILEXPCAL)
    {
      cur_formal = fEM->getNaryArgs (&formal);
      cur_eff = fEM->getNaryArgs (&effective);
      if (fEM->getSynchronousExprHorl (cur_eff, &heff))	// a modifier
	{
	  heff = fCKM->getRepresentativeFormula (heff);
	  hformal = fCKM->getRepresentativeFormula (fEM->getClock (cur_formal));
	  setAttribute (hformal, heff);
	}
    }
}


/**
   The pragmas of the "included" model, but not the PRAGMA_UNEXPANDED one, are inserted in the 
   graph in which the objects are inserted.
*/
_VIRTUAL_ void
Inclusion::homomorphismPragmas ()
{
  ptlpragmas lprag;
  ptpragma vpragma;

  lprag = fGraph->getPragmas ();
  while (ptlpragmas_get_next (&lprag, &vpragma))
    if (!isPragmaKeyword (vpragma, PRAGMA_UNEXPANDED))
      sGraph->addPragma (vpragma);
}

/**
   The nodes of  the signals of lout are removed from the graph.
   lout are assumed to be the outputs of the current node call expansion.
*/
void
Inclusion::resetOutputsNodeCall (ptexpcal lout)
{
  ptINode inNode;

  while (lout != NILEXPCAL)
    {
      inNode = fEM->getNode (fEM->getNaryArgs (&lout));
      fNM->removegfd (inNode);
    }
}


/**
   This method expands the node call (nodeCall) in the graph g by inserting a copy of 
   the referenced graph in the context of the call. 

   When includeAsSubGraph, the created nodes are elements of a new subgraph (characterized by a label).
*/
_VIRTUAL_ void
Inclusion::include (ptDCG g, ptINode nodeCall, boolean includeAsSubGraph)
{
  ptlinsth lh;
  ptlINode ln;
  ptDCG g1;
  ptexpcal nodeCallParams, nodeCallInputs, nodeCallOutputs, linForm, loutForm, lparForm;

  fNM->get_nodeCall_arg (nodeCall, &g1, &nodeCallParams, &nodeCallInputs, &nodeCallOutputs);
  initialize (g1, g, pk_true);	// Objects of g1 will be copied into g.
  sCKM->initSystemForModification ();
  resetOutputsNodeCall (nodeCallOutputs);
  if (includeAsSubGraph)
    pushLabel ();

  lparForm = fEM->convertLinstances2ListExpcal (g1->getParameters ());
  linForm = fEM->convertLinstances2ListExpcal (g1->getInputs ());
  loutForm = fEM->convertLinstances2ListExpcal (g1->getOutputs ());

  ConnectInstances (lparForm, nodeCallParams);
  ConnectInstances (linForm, nodeCallInputs);
  ConnectInputNodes (linForm, nodeCallInputs);
  ConnectInstances (loutForm, nodeCallOutputs);
  ConnectClocks (linForm, nodeCallInputs);
  ConnectClocks (loutForm, nodeCallOutputs);

  // Getting nodes/clocks to be copied.
  fNM->getAllNodes (&ln);
  ln = ptlINode_Copy (ln);
  lh = ptlinsth_duplicate (fCKM->getlhroots ());
  lh = ptlinsth_sorted_fusion (lh, fCKM->getlhplaced ());
  // -------------end

  homomorphismNodes (ln);
  homomorphismClocks (lh);
  homomorphismHierarchy (lh);
  homomorphismDependences (lh);
  homomorphismDependences (ln);
  homomorphismPragmas ();

  homomorphismdependences (g1->getInputs ());
  g->setParameters (ptlinstances_sortedfusion (lparameters, g->getParameters ()));
  g->setConstants (ptlinstances_sortedfusion (lconstants, g->getConstants ()));

  if (includeAsSubGraph)
    popLabel (getCurrentInstances (g1->getInputs ()), getCurrentInstances (g1->getOutputs ()));

  finalize ();
}

/**
   Useful when a sub graph is built (from includeGraphAsSubGraph()).
   It creates a sub graph in which  all the created nodes/clocks during the expansion will be inserted.
   This sub graph is characterized by a label.
*/
void
Inclusion::pushLabel ()
{
  ptexpcal vexpcal;
  ptINode node;
  ptinsth clkLabel;
  ptident id;

  vexpcal = sEM->mkLabelIdent (getLabelId (), sNM->mk_node_undefined (), hundef, NILBACKANNOTATION, pk_true);
  node = sNM->mk_node_subgraph (vexpcal, NILBACKANNOTATION);
  id = makeIdent_ident_int (fGraph->getName (), sNM->get_num (node));
  sNM->set_clusterName (node, id);

  clkLabel = igc->setClockVariable (vexpcal);
  igc->pushLabel (node, clkLabel);
  CreatedSubGraph = node;
}


/**
   It finalizes the creating of a sub graph, previously initilaizes by pushLabel().
   It fixes the clock of the sub graph, inserts it in the graph.
   The pragma PRAGMA_UNEXPANDED is transfered from the expanded model to the sub graph.
*/
void
Inclusion::popLabel (ptlinstances lin, ptlinstances lout)
{
  ptexpcal vlabel;
  ptElemStack node;
  ptinsth vh1, vh;
  ptINode aSubGraph;
  ptlpragmas lprag;
  ptpragma vpragma;

  if (igc->topLabel (&node))
    {
      aSubGraph = (ptINode) node;
      sNM->get_label (aSubGraph, &vlabel);
      vh1 = sCKM->makeSup (igc->upperBoundClockOfIdentifiers (lin), igc->upperBoundClockOfIdentifiers (lout));
      vh = vh1;
      vh = sCKM->makeSynchro (vh, igc->getSignalClockSynchronizedWith (vlabel, vh));

      sNM->insertNode (aSubGraph, vh);
      sNM->set_subGraphProjectedSystemInfos (aSubGraph, vh, NILLINSTH);	/* revoir */
      sNM->unlinkNodesFromClock (aSubGraph);
      igc->popLabel ();
      igc->takeIntoAccountContexts (aSubGraph);
      // Assignment of the  "unexpanded" pragma to the subgraph if it is assigned to the included graph.
      lprag = fGraph->getPragmas ();
      if (getPragmaKeyword (lprag, PRAGMA_UNEXPANDED, &vpragma))
	sNM->addPragma (aSubGraph, vpragma);
    }
}
