/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return the signals of the selected sub-graph (inputs + outputs + locals).
*/
  _VIRTUAL_ ptlinstances SubGraphExtract::selectingSignals ()
{
  ptlinstances lsig;

  lsig = ptlinstances_sortedfusion (inputsSubGraph, NILLINSTANCES);
  lsig = ptlinstances_sortedfusion (outputsSubGraph, lsig);
  lsig = ptlinstances_sortedfusion (fNM->getInodelLocalFlows (subGraph), lsig);
  return lsig;
}

/**
   @return the list of the nodes of the current subgraph (subGraph).
*/
_VIRTUAL_
  ptlINode
SubGraphExtract::selectingNodes ()
{
  ptlINode ln1;
  ptlinsth lh1;

  fNM->get_subGraphDefinitions (subGraph, &ln1, &lh1);
  return ptlINode_Copy (ln1);
}


/**
   @return 
    - the clocks of the original graph if exportClockSystem,
    - the clocks of the current subgraph + the clocks of the signals of the sub-graph (see selectingSignals()), otherwise.
*/
_VIRTUAL_ ptlinsth
SubGraphExtract::selectingClocks (boolean bExportClocksOfSubGraph, boolean exportClockSystem)
{
  ptlINode ln1;
  ptlinsth lh, lh1;
  boolean b;
  ptinsth vh, vhemb;
  ptinterface vinterface;
  
  if (exportClockSystem)
    {
      lh = ptlinsth_duplicate (fCKM->getlhroots ());
      lh = ptlinsth_sorted_fusion (lh, fCKM->getlhplaced ());
      if ((lConstraintsNullClock != NILLINSTH) || (lAssertionsNullClock != NILLINSTH)) lh = ptlinsth_sorted_insert_elem (lh, fCKM->getNullClock (), &b);
    }
  else if (bExportClocksOfSubGraph)
    {
      fNM->mkSubGraphProjectedSystemInfos (subGraph);
      fNM->get_subGraphProjectedSystemInfos (subGraph, &vhemb, &lh);
    }
  else
    {
      fNM->get_subGraphDefinitions (subGraph, &ln1, &lh1);
      // --- ok for bDC+ format, to be completed for DC+.
      lh = ptlinsth_duplicate (lh1);
      // Getting the clock of the subgraph. rootClock: branching clock of the nodes of the subgraph.
      fNM->get_subGraphProjectedSystemInfos (subGraph, &rootClock, &lh1);
      getAttribute (rootClock);	// defined as a "variable"
      lh = ptlinsth_sorted_fusion (lh, lh1);
      lh = fEM->addClockSignals (selectingSignals (), lh);
      lh = ptlinsth_remove_elem (lh, fCKM->getNullClock (), &b);
    }
  return lh;
}

/**
   @return lh + the clocks of the dependencies of the selected nodes.
*/
_VIRTUAL_ ptlinsth
SubGraphExtract::selectingDependenceClocks (ptlINode ln, ptlinsth lh)
{
  ptINode aNode, vfrom, vto;
  ptlINode lnsave;
  ptldependence larcs;
  ptinsth h;
  boolean b;
  ptdependence arc;

  lnsave = ln;
  while (ptlINode_get_elem (&ln, &aNode))
    {
      larcs = fNM->get_successors (aNode);
      while (larcs->get_next (&arc, &larcs))
	{
	  h = arc->getDependence (&vfrom, &vto);
	  if (ptlINode_isIn (lnsave, vto))
	    {
	      h = fCKM->getRepresentativeFormula (h);
	      lh = ptlinsth_sorted_insert_elem (lh, h, &b);
	    }
	}
    }

  return lh; // removed Aug 2016
  
  // Added July 2016 LBesnard.
  ptexpcal vsig;
  ptINode n;
  ptlinstances l;

  l = inputsSubGraph;
  while (ptlinstances_get_next (&l, &vsig))
    {
      n = fEM->getNode (vsig);
      larcs = fNM->get_successors (n);
      while (larcs->get_next (&arc, &larcs))
	{
	  h = arc->getDependence (&vfrom, &vto);
	  if (ptlINode_isIn (lnsave, vto))
	    {
	      h = fCKM->getRepresentativeFormula (h);
	      lh = ptlinsth_sorted_insert_elem (lh, h, &b);
	    }
	}
    }
  return lh;
}


/** Selecting objetcs of the selected subgraph (subGraph global member), as global attributes:
    - inputsSubGraph, outputsSubGraph: the inputs (resp. outpust) of the sub-graph.
    - lsig: the signals of the subgraph.
    - ln: nodes of the subgraph.
    - lh: clocks of the subgraph.
*/
_VIRTUAL_ void
SubGraphExtract::selectingObjects (ptlINode * ln, ptlinsth * lh, boolean bExportClocksOfSubGraph, boolean exportClockSystem)
{
  ptlinsth lh1;

  inputsSubGraph = fNM->getInodelInputFlows (subGraph);
  outputsSubGraph = fNM->getInodelOutputFlows (subGraph);
  *ln = selectingNodes ();
  *lh = selectingClocks (bExportClocksOfSubGraph, exportClockSystem);
  *lh = selectingDependenceClocks (*ln, *lh);
}

/**
   It sets the pragmas associated with the sub-graph to the generated graph.
   @param lsig (in) list of signals that are selected for defining the objects associated with a pragma.
*/
_VIRTUAL_ void
SubGraphExtract::homomorphismPragmasSubGraph ()
{
  // The pragmas of the original graph are not propagated.
  ptlpragmas lp1;
  ptpragma vp;
  ntrees stmt;
  ptrepstr name;
  ptObjPragma obj;

  lp1 = fNM->getPragmas (subGraph);
  while (ptlpragmas_get_next (&lp1, &vp))
    {
      getPragma (vp, &name, &obj, &stmt);
      sGraph->removePragma (name);
      sGraph->addPragma (vp);	// Caution: ok if there are no "objects" associated with.
    }
}

/**
   The dependences of the inputs of the subgraph are translated.
*/
_VIRTUAL_ void
SubGraphExtract::setDependencesInputsSubGraph ()
{
  ptexpcal vsig;
  ptINode n, nNode;
  ptldependence larcs;
  ptlinsth lh;
  ptlinstances l;

  l = inputsSubGraph;
  while (ptlinstances_get_next (&l, &vsig))
    {
      n = fEM->getNode (vsig);
      nNode = sEM->getNode (getAttribute (vsig));
      lh = fNM->get_clockSuccessors (n);
      larcs = fNM->get_successors (n);
      homomorphismDependence (nNode, lh, larcs);
    }
  l = outputsSubGraph;
  while (ptlinstances_get_next (&l, &vsig))
    {
      n = fEM->getNode (vsig);
      nNode = sEM->getNode (getAttribute (vsig));
      lh = fNM->get_clockSuccessors (n);
      larcs = fNM->get_successors (n);
      homomorphismDependence (nNode, lh, larcs);
    }
  
}

/**
   It defines the inputs/outputs of the produced graph.
*/
_VIRTUAL_ void
SubGraphExtract::set_io (ptlinsth limportedClocks)
{
  ptlinstances lin, lout;

  sGraph->setParameters (lparameters);
  sGraph->setConstants (lconstants);


  lout = getCurrentInstances (outputsSubGraph);
  lin = addUndefinedSignals (NILLINSTANCES, limportedClocks);
  igc->addInNodes (lin);
  igc->addOutNodes (lout);

  lin = ptlinstances_sortedfusion (getCurrentInstances (inputsSubGraph), lin);
  sGraph->setInputs (lin);
  sGraph->setOutputs (lout);
}


/**
   It defines the inputs/outputs of the produced graph.
*/
_VIRTUAL_ void
SubGraphExtract::setInputNodes()
{
  ptexpcal nsig, vsig;
  ptlinstances lsn;
  ptINode n, nodeNewGraph;

  lsn = inputsSubGraph;
  while (ptlinstances_get_next (&lsn, &vsig))
    if (isSelectedSignal (vsig))
      {
	nsig = getAttribute (vsig);
	igc->addInNode(nsig, NILINODE, &nodeNewGraph);
	n = fEM->getNode (vsig);
	setNodeAttribute (n, nodeNewGraph);
      }
}

/**
   It creates the objects of the produced graph by importing the selected objects (nodes (ln) and clocks (lh)) of the original graph.
*/
_VIRTUAL_ void
SubGraphExtract::setControlClocks()
{
  ptinsth h;

  h = fGraph->getStep();
  sGraph->setStep(getAttribute (h));
  h = fGraph->getIOstep();
  sGraph->setIOstep(getAttribute (h));
}


/**
   It creates the objects of the produced graph by importing the selected objects (nodes (ln) and clocks (lh)) of the original graph.
*/
_VIRTUAL_ void
SubGraphExtract::extracting (ptlINode ln, ptlinsth lh, boolean exportClockSystem)
{
  ptlinstances ls;

  homomorphismNodes (ln);
  homomorphismClocks (lh);
  if (getExportHierarchy ())
    homomorphismHierarchy (lh);	
  homomorphismDependences (lh);
  homomorphismDependences (ln);
  homomorphismPragmasSubGraph ();
  ls = selectingSignals ();
  homomorphismAttributes (ls);
  if (exportClockSystem)
    {
      homomorphismConstraints ();
      setControlClocks();
    }
  homomorphismRDVConstraints(lh);
  ptlinstances_free (ls);
}

/**
   @return the operand list (l) + the signals, referenced in the produced graph, 
   that are not defined (they will be declared as inputs in the produced graph).
*/
_VIRTUAL_ ptlinstances
SubGraphExtract::addUndefinedSignals (ptlinstances l, ptlinsth limportedClocks)
{
  ptlinstances ls;
  ptexpcal vsig;
  ptINode n;
  ptinsth vh1, h;
  boolean bimport;

  if (sEM->getExprList (opident, &ls))
    while (ptlinstances_get_next (&ls, &vsig))
      if (!sEM->IsEvent (vsig))	// the "tick" for example.
	{
	  n = sEM->getNode (vsig);
	  if (sNM->is_undefined (n))
	    {
	      bimport = !sEM->isSignalUnifiedWithClock (vsig, &vh1);
	      if (bimport)
		{
		  if (!ptlinstances_isin_exp (l, vsig))
		    l = ptlinstances_insert (l, vsig);
		}
	    }
	}
  return l;
}


/**
   Specific part for a subgraph extraction.

   For a referenced clock H not cloned (element of the global list ClocksNotCloned of the Duplicator class),
   the cloned clock is added if H is an undersampling (when C) and C is an input of the subgraph.

   Then the Duplicator::finalize() method is called.
*/
_VIRTUAL_ void
SubGraphExtract::finalize ()
{
  ptlinsth lh;
  ptinsth vh;
  ptexpcal vCondClone, vcond;

  if (ClocksNotCloned != NILLINSTH)
    {
      lh = ClocksNotCloned;
      while (ptlinsth_get_next (&lh, &vh))
	if (fCKM->satisfiesSub (vh))
	  {
	    vcond = fCKM->getConditionSub (vh);	// condition in the original graph
	    if (getSigAttribute (vcond, &vCondClone))
	      homomorphismClock (vh);	// vh is then removed from the list ClocksNotCloned.
	  }
    }
  Duplicator::finalize ();
}


_VIRTUAL_ boolean
SubGraphExtract::isExportedSignal (ptexpcal vsig)
{
  return ptlINode_isIn (lSelectedNodes, fEM->getNode (vsig));
}



_VIRTUAL_ void
SubGraphExtract::homomorphismRDVConstraints (ptlinsth lh)
{
  ptlinsth lh1, lh2;
  ptinsth nvh1, vh1, vh2, nvh2;

  while (ptlinsth_get_next (&lh, &vh1))
    {
      if (fCKM->isWaitingClock(vh1, &vh2))
	{
	  if (getClockAttribute (vh1, &nvh1))
	    if (getClockAttribute (vh2, &nvh2))
	      {
		sCKM->addRDVClocks(nvh1, nvh2);
	      }
	}
    }
}

