/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/// PREDEFINED OBJECT OF THE VIRTUAL MACHINE.
AssignmentClock *pK_AssignmentClock = new AssignmentClock();

_PRIVATE_

/**
   @return the undefined clock if vexpr is "vDefinedSig $1", the contribution clock of vexpr otherwise.
*/
boolean AssignmentClock::assignmentClockExpr(ptexpcal vDefinedSig, ptexpcal vexpr, ptexpcal * vinit, ptinsth * vhass)
{
  ptexpcal vcurinit;
  ptinsth h1, h2;

  // cell
  if (EM->IsCellExpr(vexpr))
    {
      *vhass = EM->getContributionClock(EM->getWhenOrCellArg(vexpr));
      return pk_true;
    }
  // default
  if (EM->getOperator(vexpr) == opmele)
    {
      if (!assignmentClockExpr(vDefinedSig, EM->getBinaryArgstAt(vexpr, 1), vinit, &h1))
	return pk_false;
      if (!assignmentClockExpr(vDefinedSig, EM->getBinaryArgstAt(vexpr, 2), vinit, &h2))
	return pk_false;
      *vhass = CKM->makeSup(h1, h2);
      return pk_true;
    }
  // E when hc
  *vhass = EM->getContributionClock(vexpr);
  vexpr = EM->getBasicExpr(vexpr);
  if (EM->isDelayOneOf(vexpr, vDefinedSig, &vcurinit))
    {
      if (*vinit == NILEXPCAL)
	(*vinit) = vcurinit;
      else if (!EM->IsCompatibleWith(*vinit, vcurinit))
	return pk_false;
      *vhass = hundef;
      return pk_true;
    }
  return pk_true;
}

/**
   @return "true" if the assignment clock for the operand node (node) is defined in vhAssignment, "false" otherwise.
   vhAssignment is equal to the undefined clock if the node is x ::= x$1, or  x ::= defaultvalue x$1.
   The assigment clock is assigned to the signal defined in node.
   A node is assumed to be an equation (not a delay).
   \See :setAssignmentClock(), getAssignmentClock().
*/
boolean AssignmentClock::assignmentClockEquation(ptINode node, ptexpcal * vinit, ptinsth * vhAssignment)
{
  ptexpcal vDefinedSig, vexpdedef, vdedef;

  if (!NM->get_rightOperand(node, &vdedef))
    return pk_false;
  vDefinedSig = NM->get_leftOperand(node);
  vexpdedef = EM->getBasicExpr(vdedef);
  return assignmentClockExpr(vDefinedSig, vexpdedef, vinit, vhAssignment);
}

/**
   It computes the assignment clock for each node of the operand list (lnodes).
   A node is assumed to be an equation (not a delay).
*/
void AssignmentClock::assignmentClockEquations(ptlINode lnodes)
{
  ptINode node;
  ptinsth vhAssignment;
  ptexpcal vDefinedSig, vinit;

  while (ptlINode_get_elem(&lnodes, &node))
    {
      vinit = NILEXPCAL;
      if (assignmentClockEquation(node, &vinit, &vhAssignment))
	{
	  vDefinedSig = NM->get_leftOperand(node);
	  EM->setInitialValue(vDefinedSig, vinit);
	  EM->setAssignmentClock(vDefinedSig, vhAssignment);
	}
    }
}

void AssignmentClock::assignmentClockPartialEquation(ptINode node)
{
  ptlINode ln;
  ptINode n;
  ptinsth vhAssignment, vh;
  ptexpcal vinit, vDefinedSig;
  boolean b;

  vDefinedSig = NM->get_leftOperand(node);
  ln = NM->get_setOfPartialDefinitions_arg(node);
  vhAssignment = hundef;
  b = pk_true;
  vinit = NILEXPCAL;
  while (b && ptlINode_get_elem(&ln, &n))
    if (assignmentClockEquation(n, &vinit, &vh))
      vhAssignment = CKM->makeSup(vhAssignment, vh);
    else
      b = pk_false;
  if (b)
    {
      EM->setInitialValue(vDefinedSig, vinit);
      EM->setAssignmentClock(vDefinedSig, vhAssignment);
    }
}

/**
   It computes the assignment clock for each node of the operand list (lnodes).
   A node is assumed to be an equation (not a delay).
*/
void AssignmentClock::assignmentClockPartialEquations(ptlINode lnodes)
{
  ptINode node;

  while (ptlINode_get_elem(&lnodes, &node))
    assignmentClockPartialEquation(node);
}
