/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   The functionality is applied to a graph (g).
*/
  Pruner::Pruner (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();

  unusedNodes1 = AGRAPH->getUnused (&unusedClocks1);	// Unused objects currently. Some of them may be used now!

  unusedNodes = NILLINODE;
  unusedClocks = NILLINSTH;
}

/**
   Destructor.
*/
Pruner::~Pruner ()
{
}


/**
   It provides for the selected graph (see constructor) the set of nodes (ln) and clocks (lh) that are not useful for
   the production of the ouptuts and the consumption of the inputs of t he graph. The algorithm is composed of two parts:
   - a colouring step (See Colouring::apply(), and \ref Colouring): it selects the nodes/clocks useful  for 
   the production of the ouptuts and the consumptions of the inputs
   - a garbaging step (garbaging()): it produces the output lists (unmarked objects of the first step).

   The parameter \e bLastClockDefaultUsed is used for the colouring, when it is "true" it specifies that the last constribution clock of a default must be
   marked.

   The mark, used for the colouring, is a local one: the graph is cleaned at the end of the functionality.
*/
void
Pruner::apply (boolean bLastClockDefaultUsed, ptlINode * ln, ptlinsth * lh)
{
  ptDCGColouring x;

  NM->removeDependenceNullClock ();

  aMark = getMark ();
  x = new Colouring (AGRAPH);
  x->apply (aMark, bLastClockDefaultUsed);

  garbaging ();
  *ln = unusedNodes;
  *lh = unusedClocks;
  CKM->resetMarkAll (aMark);
  NM->resetMarkAll (aMark);
  delete x;
  AGRAPH->setUnused (unusedNodes1, unusedClocks1);
}
