/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   The clock (clk) is added to the list (unusedClocks) if it is not useful marked.
*/
  void
Pruner::mkUnusedClock (ptinsth clk)
{
  boolean b;
  clk = CKM->getRepresentativeFormula (clk);
  if (CKM->satisfiesPredefinedClock (clk))
    return;
  if (!CKM->satisfiesMarked (clk, aMark))
    unusedClocks = ptlinsth_sorted_insert_elem (unusedClocks, clk, &b);
}

/**
   For each clock, it calls mkUnusedClock().
*/
void
Pruner::mkUnusedClocks ()
{
  ptlinsth lh;
  ptinsth vh;

  lh = CKM->getlhroots ();
  while (ptlinsth_get_next (&lh, &vh))
    mkUnusedClock (vh);

  lh = CKM->getlhplaced ();
  while (ptlinsth_get_next (&lh, &vh))
    mkUnusedClock (vh);

  lh = CKM->getFormulaList (sub);
  while (ptlinsth_get_next (&lh, &vh))
    {
      mkUnusedClock (vh);
      if (CKM->getComplInFather (vh, &vh))
	mkUnusedClock (vh);
    }
}


/**
   It adds a node (node) to the set of unuseful nodes (unusedNodes) if it is unmarked.   
*/
void
Pruner::mkUnusedNode (ptINode node)
{
  ptlINode ln;
  ptlinstances lsig;
  ptexpcal s;

  if (!NM->isMarked (node, aMark))
    {
      ptlINode_insert_elem (&unusedNodes, node);
      if (NM->is_setOfPartialDefinitions (node))	// set of partial definitions
	{
	  ln = NM->get_setOfPartialDefinitions_arg (node);
	  mkUnusedNodes (ln);
	}
      else if (NM->is_lfields (node))	// struct-bundle
	{
	  lsig = NM->get_lfields_arg (node);
	  while (ptlinstances_get_next (&lsig, &s))
	    ptlINode_insert_elem (&unusedNodes, EM->getNode (s));
	}
    }
  else
    ptlINode_remove_elem (&unusedNodes1, node);

}


/**
   It applies mkUnusedNode() to each element of a list (ls)
*/
void
Pruner::mkUnusedNodes (ptlINode ls)
{
  ptINode node;
  while (ptlINode_get_elem (&ls, &node))
    mkUnusedNode (node);
}

/**
   @return the unuseful nodes of a list of nodes (ls)
*/
void
Pruner::mkUnusedNodes ()
{
  ptlINode ls;

  if (NM->getEquations (&ls))
    mkUnusedNodes (ls);
  if (NM->getDelays (&ls))
    mkUnusedNodes (ls);
  if (NM->getProcCalls (&ls))
    mkUnusedNodes (ls);
  if (NM->getIterates (&ls))
    mkUnusedNodes (ls);
  if (NM->getGroupsOfPartialEquations (&ls))
    mkUnusedNodes (ls);
  if (NM->getBundles (&ls))
    mkUnusedNodes (ls);
  if (NM->getSubGraphs (&ls))
    mkUnusedNodes (ls);
}


/**
   It builds the list of the unuseful nodes/clocks of the graph.
*/
void
Pruner::garbaging ()
{
  mkUnusedNodes ();
  mkUnusedClocks ();
}
