/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   Use init() to set the graph.
*/
  UtilityClock::UtilityClock ()
{
}

/**
   Constructor.
   The functionality is applied to a graph (g).
   A predefined object of this class, called pK_UtilityClock, is available.
*/
UtilityClock::UtilityClock (ptDCG g)
{
  init (g);
}


/** Destructor. */
UtilityClock::~UtilityClock ()
{
}


/**
   It initializes the utility clock functionality for the operand graph g.
   To compute the utility clocks calls UtilityClock::apply().
*/
void
UtilityClock::init (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  the_nullClock_UC_copy = CKM->getNullClock ();
  aMark = getMark ();
}


/**
   It computes the utility clock of the signals of the selected graph.

   To select a graph, see the constructors : UtilityClock(ptDCG), UtilityClock().
*/
void
UtilityClock::apply ()
{
  SelectedObjectMark = -1;	// -1: all nodes of the graph are selected
  go ();
}


/**
   Computes the clock utility of the inputs of a subgraph (aSubGraph) of a graph (g).
   XXX.
*/
void
UtilityClock::SubGraphClockUtility (ptDCG g, ptINode aSubGraph)
{
  ptlINode ln;
  ptlinsth lh;

  init (g);
  SelectedObjectMark = getMark ();
  SubGraph = aSubGraph;
  go ();
}
