/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   It removes a delay tree (vmemtree) from a set of such trees (lmemtrees).
 */
    ptlmemnode mkaggregationsAndMultiplexings::memtreeRemoveScalarDelayOne(ptmemnode vmemtree, ptlmemnode lmemtrees)
{
  ptlmemnode vlchild;
  ptINode nodeDelay;
  ptexpcal /* res, sig, vret, vfen, */ vresinst;
  ptmemnode rootMemtree;

  lmemtrees = ptlmemnode_remove_elem(lmemtrees, vmemtree);
  rootMemtree = ptmemnode_get_real_root(vmemtree);
  if (rootMemtree == vmemtree)
    ptexpcal_set_node_tree_delay(ptmemnode_get_signal(vmemtree), NILMEMNODE);
  else if (rootMemtree != NILMEMNODE)
    ptmemnode_set_equiv(rootMemtree, ptlmemnode_remove_elem(ptmemnode_get_equiv(rootMemtree), vmemtree));

  vlchild = ptmemnode_get_lchild(vmemtree);
  while (ptlmemnode_getNaryArgs(&vlchild, &vmemtree))
    {
      vresinst = ptmemnode_get_signal(vmemtree);
      nodeDelay = EM->getNode(vresinst);
      ptexpcal_set_node_tree_delay(vresinst, NILMEMNODE);
      if (ptmemnode_get_lchild(vmemtree) != NILLMEMNODE)
	lmemtrees = ptlmemnode_insert_elem(vmemtree, lmemtrees);
    }
  return lmemtrees;
}

/** @return "true" if the delay tree node (vmemtree) is a root of such a structure 
    reduced to one child and this node is a delay of 1, "false" otherwise.
*/
boolean mkaggregationsAndMultiplexings::isSimpleDelayTree(ptmemnode vmemtree)
{
  ptlmemnode vlchild;
  ptmemnode vchild;
  boolean bDelayOne;
  ptexpcal vexp;

  vexp = ptmemnode_get_signal(vmemtree);
  // if ( EM->getNumberOfDimensions(vexp) != 0) return pk_false; tested before
  /* When there is a cycle in the delay definitions, the root is a delay.
     Example : x := y $1 init x0 | y := x $1 init y0
     Fixing bug #7003
   */
  if (NM->is_delay(EM->getNode(vexp)))
    return pk_false;
  if (ptmemnode_get_equiv(vmemtree) != NILLMEMNODE)
    return pk_false;

  vlchild = ptmemnode_get_lchild(vmemtree);
  bDelayOne = pk_true;
  while (bDelayOne && ptlmemnode_getNaryArgs(&vlchild, &vchild))
    {
      if (ptmemnode_get_lchild(vchild) != NILLMEMNODE)
	return pk_false;
      if (ptmemnode_get_equiv(vchild) != NILLMEMNODE)
	return pk_false;
      bDelayOne = (ptmemnode_get_delay_value(vchild) == 1) && (ptmemnode_get_window_value(vchild) == 1);
      // \note { added Dec 2006: the type of the signals must be the same (no extension).}
      // if (bDelayOne) bDelayOne = EM->getType( ptmemnode_get_signal(vchild)) == EM->getType(vexp);
    }
  return bDelayOne;
}

/** For a tree of delays verifying the isSimpleDelayTree() condition (ie zx := x $1, and it is the only delay on x)
    If the signals are scalars, they are never multiplexed in the same area even the MULTIPLEXINGDELAYS is pk_true.
    When it is not multiplexed, the memtreeRemoveScalarDelayOne() method is applied.
 */
void mkaggregationsAndMultiplexings::setlmemtreesDelayOneScalarFirst()
{
  ptlmemnode vlmemt, lmemtrees;
  ptmemnode vmemtree;
  boolean bContinue, bapply;
  ptexpcal vexp;
  TypeType vtype;

  lmemtrees = AGRAPH->getDelaysTree();
  bContinue = pk_true;
  while (bContinue)
    {
      bContinue = pk_false;
      vlmemt = lmemtrees;
      while (ptlmemnode_getNaryArgs(&vlmemt, &vmemtree))
	{
	  if (MULTIPLEXINGDELAYS)
	    {
	      vexp = ptmemnode_get_signal(vmemtree);
	      vtype = EM->getUnamedType(vexp);
	      bapply = TypeTestPredef(vtype) || TypeTestEnumerationClass(vtype);
	    }
	  else
	    bapply = pk_true;
	  if (bapply)
	    if (isSimpleDelayTree(vmemtree))
	      {
		lmemtrees = memtreeRemoveScalarDelayOne(vmemtree, lmemtrees);
		bContinue = pk_true;
	      }
	}
    }
  AGRAPH->setDelaysTree(lmemtrees);
}

/**
   It updates the delay trees of the graph:
   - it cleans the delay trees of the graph (removeMemtreesNullClock()),
   - it applies particular method for "$ 1" (setlmemtreesDelayOneScalarFirst())
   - it computes then the size of the arrays for a circular managing (each delay can be represented by a index on this array).
*/
void mkaggregationsAndMultiplexings::setlmemtrees()
{
  ptlmemnode vlmemt, lmemtrees;
  ptmemnode vmemtree, mem;
  ptexpcal vexp;

  removeMemtreesNullClock();
  setlmemtreesDelayOneScalarFirst();

  lmemtrees = AGRAPH->getDelaysTree();
  /* Assignment of the sizes for a circular management. */
  vlmemt = lmemtrees;
  while (ptlmemnode_getNaryArgs(&vlmemt, &vmemtree))
    {
      vexp = ptmemnode_get_signal(vmemtree);
      if (ptexpcal_satisfies_node_tree_delay(vexp, &mem))
	if (EM->getNumberOfDimensions(vexp) != 0)
	  ptmemnode_belongs_typuseinec(mem, useout);
      ptmemnode_set_size_all(vmemtree, ptmemnode_size_compute(vmemtree, 0, 0, 0) + 1);
    }
}

/**
   Local to setlmemtrees() method.
   It removes each delay tree of the graph when the signal associated with the root is a signal synchonized with the null clock.
*/
void mkaggregationsAndMultiplexings::removeMemtreesNullClock()
{
  ptlmemnode vlmemt, lmemtrees;
  ptmemnode vmemtree;
  ptexpcal vexp;

  lmemtrees = AGRAPH->getDelaysTree();
  vlmemt = lmemtrees;
  while (ptlmemnode_getNaryArgs(&vlmemt, &vmemtree))
    {
      vexp = ptmemnode_get_signal(vmemtree);
      if (CKM->satisfiesNullClock(EM->getClock(vexp)))
	lmemtrees = ptlmemnode_remove_elem(lmemtrees, vmemtree);
    }
  AGRAPH->setDelaysTree(lmemtrees);
}

/**
   @return "true" if the operation succeeds (optimization is applied), "false" otherwise.
   For a root of a delay tree (vmemtree) verifying the condition of isSimpleDelayTree(),
   it is possible to merge  x (associate with the root) and zx (associated with the child) n the same memory,
   - if x is not a memory (cell operator).
   - if the sucessors of x can be produced before the update of the memory (zx). To achieve this, a reinforcing of the graph is
   tested (reinforcingFromTo()). It there is no circuit after the reinforcing, the merging is possible. In this case, attributes
   are added (global compiler attributes defined in the CompilerAttributes.h file)
      - zx is attributed with the OPTIMIZED_MEMORY_ATTRIBUTE attribute, with x as value.
      - the node of zx is marked with the OPTIMIZED_MEMORY_ATTRIBUTE attribute,
      - the node of x is marked with the BELONG_MEMORY attribute.

      It is not applied for constant nodes (element of lNodesInitClk).
*/
boolean mkaggregationsAndMultiplexings::delayMemoryOptimisation(ptmemnode vmemtree, ptlINode lNodesInitClk, boolean bclu)
{
  ptlmemnode vlchild;
  ptmemnode vmemtree1, x;
  ptexpcal signalX, signalZX;
  boolean b;
  ptINode n1, n2, blockn1, blockn2;

  if (!isSimpleDelayTree(vmemtree))
    return pk_false;
  vlchild = ptmemnode_get_lchild(vmemtree);
  if (!ptlmemnode_getNaryArgs(&vlchild, &vmemtree1))
    return pk_false;
  if (ptlmemnode_getNaryArgs(&vlchild, &x))
    return pk_false;
  signalX = ptmemnode_get_signal(vmemtree);
  signalZX = ptmemnode_get_signal(vmemtree1);
  GET_OPTIMIZED_MEMORY_ATTRIBUTE(signalX, &signalX);
  GET_OPTIMIZED_MEMORY_ATTRIBUTE(signalZX, &signalZX);
  signalX = EM->getAggregation(signalX);
  signalZX = EM->getAggregation(signalZX);
  if (signalX == signalZX) return pk_false;
  n1 = EM->getNode(signalZX);
  // Restrictions
  if (bclu  && EM->IsOutputFlow(signalZX)) return pk_false;
  n2 = EM->getNode(signalX);
  if (NM->is_cellNode(n2)) return pk_false;		// cell
  if (ptlINode_isIn(lNodesInitClk, n2)) return pk_false; // constant nodes.

  NM->get_inComposedNode(n1, &blockn1);	// array or subgraph. ok for aggregration for clusters.
  NM->get_inComposedNode(n2, &blockn2);
  if ((blockn1 != NILINODE) && (blockn2 != NILINODE))
    if (!NM->is_trainCluster(blockn1) || !NM->is_trainCluster(blockn2))
      if (blockn1 != blockn2)
	return pk_false;
  // end Restrictions

  b = reinforcingFromTo(signalZX, signalX, htop);
  if (b)
    {
      // The reinforcing does not introduce a circuit -> merging x and zx ; Adding attributes
      EM->addAttribute(signalZX, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) signalX);
      // NM->addMark(n1, OPTIMIZED_MEMORY_ATTRIBUTE);  // no references.
      NM->addMark(n2, BELONG_MEMORY);
      NM->addMark(n2, DontOptimize);
    }
  return b;
}

/** 
    @return "true" if the signal assigned to vmemtree, is not also assigned to another element of vlmemt, "false" otherwise.
    ( Example : For  zx := x $ init I1 and zx1 := x $ init I2.
      If I1, I2 are not compatible initialisations, two trees are built with x as root.)
*/
boolean mkaggregationsAndMultiplexings::isSingleMemtree(ptmemnode vmemtree, ptlmemnode vlmemt)
{
  ptmemnode vmemtree1;
  ptexpcal S1, S2;
  S1 = ptmemnode_get_signal(vmemtree);
  GET_OPTIMIZED_MEMORY_ATTRIBUTE(S1, &S1);
  while (ptlmemnode_getNaryArgs(&vlmemt, &vmemtree1))
    if (vmemtree1 != vmemtree)
      {
	S2 = ptmemnode_get_signal(vmemtree1);
	GET_OPTIMIZED_MEMORY_ATTRIBUTE(S2, &S2);
	if (S1 == S2)
	  return pk_false;
      }
  return pk_true;
}
