/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define GET_OPTIMIZED_MEMORY_ATTRIBUTE(S1, ptE)   EM->getAttribute(S1, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *)ptE)

_PRIVATE_
/**
   It adds a signal (vsig) to the locals (global list called lssigloc)
   if it is not an input nor an output.
*/
    void
 mkaggregationsAndMultiplexings::addlocal(ptexpcal vsig)
{
  if (!EM->IsOutputFlow(vsig) && !EM->IsInputFlow(vsig))
    lssigloc = ptlinstances_insert(lssigloc, vsig);
}

/**
   It calls  mkaggregationsAndMultiplexings::addlocal() for each element of a list of signals (l).
*/
void mkaggregationsAndMultiplexings::addlocals(ptlinstances l)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&l, &vsig))
    addlocal(vsig);
}

/**
   Local circuit detector used after a reinforcing of the graph (reinforcingFromTo() method).
   @return "true" if the graph contains some circuits, "false" otherwise.
*/
boolean mkaggregationsAndMultiplexings::graphWithCycle()
{
  if (CircuitDetector == (ptCircuitCalculus) NULL)
    CircuitDetector = new CircuitCalculus(AGRAPH);
  return CircuitDetector->circuit_detection();
}

/**
   @return "true" if it exists a cell in the succesors of the operand node , "false" otherwise.
 */
boolean mkaggregationsAndMultiplexings::isSuccMemory(ptINode node)
{
  ptdependence aDep;
  ptldependence ldep;
  ptINode vfrom, vto;

  ldep = NM->get_successors(node);
  while (ldep->get_next(&aDep, &ldep))
    {
      aDep->getDependence(&vfrom, &vto);
      if (NM->is_cellNode(vto)) return pk_true;
    }
  return pk_false;
}


/**
   It reinforces the graph by adding a dependence from the sucessors of n1 (node of s1) to the node n2 (node of s2).
   For each added dependence, a detection of cycle is done.
   When a circuit is added, the added dependencies are removed.
   @return "true" if the reinforcing succeeds, "false" otherwise.
*/
boolean mkaggregationsAndMultiplexings::reinforcingFromTo(ptexpcal s1, ptexpcal s2, ptinsth h)
{
  ptINode node, n1, n2, vfrom, vto;
  ptdependence aDep, nDep;
  ptldependence ldep, laddedDep;
  boolean b;
  ptinsth k;
  ptlINode ln;

  if (EM->getExtractionLink(s1, &k)) return pk_false;
  n1 = EM->getNode(s1);
  n2 = EM->getNode(s2);

  if (isSuccMemory(n1)) return pk_false;
  ldep = NM->get_successors(n1);
  laddedDep = NILLDEPENDENCE;
  while (ldep->get_next(&aDep, &ldep))
    {
      aDep->getDependence(&vfrom, &vto);
      if (! NM->isInternalNodeOf(n2, vto))
	if (NM->addDependence(vto, n2, htop))
	  {
	    NM->getDependence(vto, n2, &nDep);
	    laddedDep = laddedDep->insert(nDep);
	    // Adding the depenencies for partial definitions.
	    if (NM->is_setOfPartialDefinitions(n2))
	      {
		ln = NM->get_setOfPartialDefinitions_arg(n2);
		while (ptlINode_get_elem(&ln, &node))
		  if (NM->addDependence(vto, node, htop))
		    {
		      NM->getDependence(vto, node, &nDep);
		      laddedDep = laddedDep->insert(nDep);
		    }
	      }
	  }
    }
  if (laddedDep != NILLDEPENDENCE) b = !graphWithCycle(); else b = pk_true;
  if (!b)
    {
      // A cycle is inserted, removing added dependencies.
      ldep = laddedDep;
      while (ldep->get_next(&aDep, &ldep))
	NM->removeDependence(aDep);
    }
  if (laddedDep != NILLDEPENDENCE)
    delete laddedDep;
  return b;
}

/**
   @return the associated aggregation attribute of a signal (exp), noag by default.
*/
typeag mkaggregationsAndMultiplexings::signal_get_typeag(ptexpcal exp)
{
  pointer pt;
  if (!EM->getAttribute(exp, TYPEAGGREG, &pt))
    return noag;
  return *((typeag *) pt);
}

/**
   It associates the attribute (ta) with a signal (exp), the associated value is expWith (ta not equal to noag).
   When 
   - ta=noag, the AGGREGATTRIBUTE, and TYPEAGGREG attributes are removed from the signal.
   - otherwise, 
      - the TYPEAGGREG attribute is associated with exp with ta as value,
      - the AGGREGATTRIBUTE attribute is associated with exp with expWith as value.
*/
void mkaggregationsAndMultiplexings::signal_set_typeag(ptexpcal exp, typeag ta, ptexpcal expWith)
{
  switch ((int)ta)
    {
    case noag:
      EM->removeAttribute(exp, AGGREGATTRIBUTE);
      EM->removeAttribute(exp, TYPEAGGREG);
      return;

    case agid:
      EM->addAttribute(exp, TYPEAGGREG, &vagid);
      break;
    case agwhen:
      EM->addAttribute(exp, TYPEAGGREG, &vagwhen);
      break;
    case agcell:
      EM->addAttribute(exp, TYPEAGGREG, &vagcell);
      break;
    }

  markedList = ptlinstances_insert(markedList, exp);
  EM->addAttribute(exp, AGGREGATTRIBUTE, (attributeValue) expWith);
}

/**
   It returns "true" if at most one of the signals expg and vinstexpd (the aggregating of expd) are initialized, "false" otherwise.
*/
boolean mkaggregationsAndMultiplexings::atmostOneInitialized(ptexpcal expg, ptexpcal expd, ptexpcal * vinstexpd)
{
  ptexpcal vinitG;
  ptmemnode mem;

  *vinstexpd = EM->getAggregation(expd);
  vinitG = EM->getInitialValue(expg);

  if (vinitG == NILEXPCAL)
    return pk_true;
  if (EM->getInitialValue(*vinstexpd) != NILEXPCAL)
    return pk_false;
  EM->setInitialValue(*vinstexpd, vinitG);
  if (ptexpcal_satisfies_node_tree_delay(*vinstexpd, &mem))
    ptmemnode_set_cellinit(mem, vinitG);
  return pk_true;
}

/** It tries to aggregate the output (x) of an equation (i), x:=E, in E.
    @return "true" if the operation succeeds, "false" otherwise.
    \note If the node (i) is x:=zx with zx:=x$1, and zx and x are merged in the same memory, the aggregating of x and zx is suspended: 
    the node (i) is kept in the member list Glist. It will be visited later (endfunctionality()).
*/
boolean mkaggregationsAndMultiplexings::aggregateEqu(ptINode i)
{
  ptexpcal vinstexpd, expg, expd;
  typeag agexpg;
  ptinsth hexpd, h1;
  boolean bRefStateVar;
  ptmemnode mem;
  ptINode vblock;

  if (NM->isMarked(i, DontOptimize))
    return pk_false;
  //  if (NM->is_inComposedNode(i)) return pk_false; changed to (Sep 2014, LBesnard)
  if (NM->get_inComposedNode(i, &vblock))
    if (NM->is_array(vblock))
      return pk_false;

  NM->get_equation_arg(i, &expg, &expd);
  if (EM->isNILEXPCAL(expd))
    return pk_false;
  if (EM->getOperatorClass(expg) != variabl)
    return pk_false;
  // Same type , we can have x := y with  [1]Type x and y scalar;
  if (EM->getType(expg) != EM->getType(expd))
    return pk_false;
  if (ptlinstances_isin_exp(AGRAPH->getOutputs(), expg))
    return pk_false;
  if (ptexpcal_satisfies_node_tree_delay(expg, &mem))
    return pk_false;
  if (!EM->IsSignal(expd))
    {
      //      return pk_false; // added Feb 2014 (LBesnard, similar to x:= y $1)
      if (EM->getOperator(expd) != opvvarin)
	return pk_false;
      bRefStateVar = pk_true;
    }
  else
    bRefStateVar = pk_false;
  // No aggregating for fields of structures.

  if (bRefStateVar)
    {
      signal_set_typeag(expg, agid, expd);
      return pk_true;
    }

  hexpd = NM->get_contributionClockEquation(i);
  h1 = EM->getClock(expd);
  if (hexpd == NM->get_nodeClock(i))
    if (h1 == hexpd)
      agexpg = agid;
    else
      agexpg = agwhen;
  else
    {
      if (h1 != hexpd)
	return pk_false;
      if (signal_get_typeag(expd) == agwhen)
	return pk_false;
      if (atmostOneInitialized(expg, expd, &vinstexpd))
	agexpg = agcell;
      else
	{
	  if (!EM->IsCompatibleWith(EM->getInitialValue(vinstexpd), EM->getInitialValue(expg)))
	    return pk_false;
	  agexpg = agcell;
	}
    }
  if (agexpg == noag)
    return pk_false;
  if (GET_OPTIMIZED_MEMORY_ATTRIBUTE(expd, &vinstexpd))	//EM->getAttribute(expd, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *)&vinstexpd))
    {
      if (agexpg != agcell)
	ptlINode_insert_elem(&Glist, i);
      return pk_false;
    }
  signal_set_typeag(expg, agexpg, expd);
  return pk_true;
}

/**
  The input signals of the node, referenced in a delay tree, are marked as used in an external call.
*/
void mkaggregationsAndMultiplexings::aggregateProcCall(ptINode i)
{
  ptlinstances linst1;
  ptexpcal exp1;
  ptmemnode mem;
  boolean b;
  ptident name;

  if (NM->is_readNode(i))
    return;
  linst1 = NM->getInodelInputFlows(i);
  if (linst1 == NILLINSTANCES)
    return;
  b = pk_false;
  if (NM->is_intrinsicfunc(i))
    if (NM->get_nameCall(i, &name))
      b = isExceptionId(name);

  while (ptlinstances_get_next(&linst1, &exp1))
    {
      exp1 = EM->getAggregation(exp1);
      if (b)			// if (!EM->IsOutputFlow(exp1) && !EM->IsInputFlow(exp1))
	addlocal(exp1);

      if (ptexpcal_satisfies_node_tree_delay(exp1, &mem))
	if (EM->getNumberOfDimensions(exp1) != 0)
	  ptmemnode_belongs_typuseinec(mem, usein);
    }
}

/**
   It initializes the functionality:
   - the suspended aggregating list is set to nil value.
   - the current local signals are get (lssigloc attribute member),
   - the traveller object is initialized.
*/
void mkaggregationsAndMultiplexings::initfunctionality()
{
  Glist = NILLINODE;
  lssigloc = AGRAPH->getLocalSignals();
  PARC->initialize();
}

/**
   End of the functionality.
   During the travel, the nodes ( u:=zx such that zx:=x$1, such that zx is merged with x (OPTIMIZED_MEMORY_ATTRIBUTE)) 
   are kept in a global list (Glist). It tries here to aggregate u in x by a reinforcing of the graph (reinforcingFromTo()).
*/
void mkaggregationsAndMultiplexings::aggregation_with_optimized_memory_attribute()
{
  ptlINode l;
  ptINode n;
  ptexpcal expg, expd, other;

  l = Glist;
  while (ptlINode_get_elem(&l, &n))
    {
      NM->get_equation_arg(n, &expg, &expd);
      if (GET_OPTIMIZED_MEMORY_ATTRIBUTE(expd, &other))
	// if the reinforcing does not introduce a circuit -> merging nodes
	if (reinforcingFromTo(expg, other, htop))
	  EM->addAttribute(expg, AGGREGATTRIBUTE, (attributeValue *) other);
    }
}

void mkaggregationsAndMultiplexings::stateVariables()
{
  ptlinstances lexpr;
  ptexpcal exp, stateVar;

  if (EM->getExprList(opvvarin, &lexpr))	// lexpr = Sets of ?x expressions
    while (ptlinstances_get_next(&lexpr, &exp))
      {
	stateVar = EM->getUnaryArg(exp);
	addlocal(stateVar);
      }
}

/** It finalizes the functionnality:
    - the list of local signals are associated with the graph.
    - the traveller object is finalized (finalize()),
    - the suspended aggregating are visited (aggregation_with_optimized_memory_attribute()).
*/
void mkaggregationsAndMultiplexings::endfunctionality(ptinsth ghdc)
{
  stateVariables();
  AGRAPH->setLocalSignals(lssigloc);
  PARC->finalize();
  aggregation_with_optimized_memory_attribute();
}

/** It travels the nodes associated with the predefined initial clock. */
void mkaggregationsAndMultiplexings::parcoursinstrhinit(void)
{
  ptINode i;
  ptlsucc linst;

  PARC->initTravelForInitialClock();
  graphTravelling(NILINODE);
  // linst = CKM-> getNodeReferencesTo(CKM->getInitClock());
  AGRAPH->getNodesAtInitClock(&linst);
  while (ptlINode_get_elem(&linst, &i))
    if (!PARC->isNodeAlreadySelected(i))
      {
	PARC->ptINodeAttr_setSelected(i);
	PARC->signalSelectedNode(i);
      }
}

/**
   It applies some operations for a selected clock (h), selected in graphTravelling():
   - it updates the local signals (lssigloc). The locals signals associated to the clock are added to the list.
   - it signals to its successors that it is visited (traveller object).
*/
void mkaggregationsAndMultiplexings::clockProcessing(ptinsth h)
{
  PARC->signalSelectedClock(h);
  if (!CKM->satisfiesInitClock(h))
    {
      addlocals(ptinterface_get_localSignals(CKM->getInterfaceSubHierarchy(h)));
      /* Added March 2014 (LB) */
      if (CKM->isSubstitutable(h))
	if (CKM->satisfiesSub(h))
	  addlocals(EM->getSignalsReferencedIn(CKM->getConditionSub(h)));
    }
}

/**
  It travels the nodes/clocks of the graph for a block context (blockNode) using the \ref GraphTravel API.
  When a node is selected, it calls nodeProcessing(), for a clock it calls clockProcessing().
*/
void mkaggregationsAndMultiplexings::graphTravelling(ptINode blockNode)
{
  ptINode inst;
  ptinsth horl;
  boolean bIsNode;

  //  array context block opening.
  if (blockNode != NILINODE)
    PARC->openBlock(blockNode);

  // getting the nodes/clocks selected by the greedy algorithm.
  while (PARC->getNextNode(&bIsNode, &inst, &horl))
    if (bIsNode)
      nodeProcessing(inst);
    else
      clockProcessing(horl);
  //  closing array context block closing.
  if (blockNode != NILINODE)
    PARC->closeBlock();
}

/**
   It applies some operations for a selected node (node), selected in graphTravelling().
    - it calls specific aggregrating operations (only for):
        - aggregateEqu() for equation node,
	- aggregateProcCall() for procedure/node call node.
    - it updates the local signals (lssigloc) for composed nodes.
    - it signals to its successors that it is visited (traveller object).
*/
void mkaggregationsAndMultiplexings::nodeProcessing(ptINode node)
{
  ptINode n;
  ptexpcal x;

  switch (NM->get_oper(node))
    {
    case opldefpart:
    case opdefpart:
    case opret:
      break;

    case opdefsig:
      aggregateEqu(node);
      break;

    case opmoti:
      PARC->signalArraySelectedNode(node);
      graphTravelling(node);
      addlocals(NM->getInodelLocalFlows(node));
      break;

    case opappe:
    case opnodecall:
      aggregateProcCall(node);
      break;

    case opsubgraph:
      PARC->signalSubGraphSelectedNode(node);
      graphTravelling(node);
      addlocals(NM->getInodelLocalFlows(node));
      break;

    case opbund:
      x = NM->get_leftOperand(node);
      if (EM->getSignalAssignedTo(x, &x))
	addlocal(x);
      break;
    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
  PARC->signalSelectedNode(node);
}
