/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/* Constructor.
   The functionality is applied to a graph (g), <em> a bDC+ graph </em>.
   Here we suppose that  the nodes of this graph are marked by the predefined pragma PRAGMA_RUN_ON.
   The pragma is of the following form: \n
   PRAGMA_RUN_ON "i1" "i2"... where ij is a string ("ij" is the name of the partition).
   \note restriction: the multi-values are accept in the value of the pragma, but there is no code duplication, currently.
*/
 RunOnMark::RunOnMark(ptDCG g):Clusters(g)
{
  LNodesWithoutPragma = NILLINODE;	// booleans clocks.
  nbsets = -1;
  referencedInMark = getMark();
  selectedClockList = NILLINSTH;
}

/**
   Destructor.
*/
_VIRTUAL_ RunOnMark::~RunOnMark()
{
}

/**
   It colours the objects of the graph.
   The nodes are marked by colouringNodes() method, the clocks in colouringCLocks() method.
   @return "true" if the operation succeeds, "false" otherwise (no pragma runon, objects are not all localized).
*/
_VIRTUAL_ boolean RunOnMark::colouring()
{
  selectedClockList = selectingClocks();
  colouringNodes();
  colouringCLocks();
  return !error;
}

/**
   It redefines the partitionning operation by calling the following operations:
   - partitionningInitialize()
   - assignAllNodes()
   - assignClocks() on selected clocks (selectingClocks() method)
   - finalizeUseInternalRepresentation()
   - subGraphsBuildProfile()
   - partitionningFinalize()

   Each resulting subgraph is annotated (pragma) by the RunOn pragma, and the "Processor" pragma if it exists in the original graph.
*/
_VIRTUAL_ ptlINode RunOnMark::partitionning()
{
  partitionningInitialize();
  assignAllNodes();
  assignClocks(selectedClockList);
  finalizeUseInternalRepresentation();
  insertSubGraphs();
  subGraphsBuildProfile();
  partitionningFinalize();
  return lSubGraph;
}

_VIRTUAL_ boolean RunOnMark::partitionningNodes(ptlINode * ln)
{
  colouringNodes();
  if (error)
    return pk_false;
  partitionningInitialize();
  assignAllNodes();
  finalizeUseInternalRepresentation();
  insertSubGraphs();
  subGraphsBuildProfile();
  partitionningFinalize();
  *ln = lSubGraph;
  return pk_true;
}
