/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isEventClockedFormat(t) ((t == DCG_BASIC) ||(t == DCG_POLY) || (t == DCG_ENDO))

_PROTECTED_
/**
   @return true if a clock (vh), is an extraction (when C, when not C) 
   and C is an input of the clustering, false otherwise.
*/
    boolean TrainClusters::isAnExtractOnInput(ptinsth vh, ptINode * aNode)
{
  ptexpcal exp;

  (*aNode) = NILINODE;
  if (!CKM->satisfiesSub(vh))
    return pk_false;
  exp = CKM->getConditionSub(vh);
  if (EM->getOperator(exp) == opnon)
    exp = EM->getUnaryArg(exp);
  (*aNode) = EM->getNode(exp);
  if (ptlinstances_isin_exp(InputsGraph, exp))
    return pk_true;		// added Oct 2014 LB
  return (ptlINode_isIn(SelectedNodes, *aNode));
}

boolean TrainClusters::mustBeAppliedTo(ptINode aNode)
{
  ptINode cn;

  if (ptlINode_isIn(SelectedNodes, aNode))
    return pk_false;
  if (ptlINode_isIn(unusedNodes, aNode))
    return pk_false;
  if (NM->getInArrayNode(aNode, &cn))
    return pk_false;
  if (aNode == TickNode)
    return pk_false;
  return pk_true;
}

/**
   It applies assignNode(ptINode) for each node of a list of  nodes (Lnodes), excluding the element of the selected ones (SelectedNodes)
   and the unuseful ones (unusedNodes).
*/
void TrainClusters::assignNodes(ptlINode lNodes)
{
  ptINode aNode;

  while (ptlINode_get_elem(&lNodes, &aNode))
    if (mustBeAppliedTo(aNode))
      assignNode(aNode);
}

/**
   It assigns a clock (vh) to a sub-graph. 
   The set of marks (lm) associated with clock gives the associated sub-graph (see getSetSubGraphOf() method).
*/
void TrainClusters::assignClock(ptinsth vh)
{
  ptlMark lm;
  ptINode aSubGraph, aNode;

  if (!CKM->getInComposedNode(vh, &aNode))
    {
      if (isEventClockedFormat(graph_type))
	lm = ptinsthAttr_get_marks(vh);
      else
	{
	  if (isAnExtractOnInput(vh, &aNode))
	    return;
	  lm = ptINodeAttr_get_marks(aNode);	// aNode is always defined in BOOLEAN clock format.
	}
      if (getSetSubGraphOf(lm, &aSubGraph))
	NM->subGraphInsertClock(aSubGraph, vh);
    }
}

/**
   It applies assignClock(ptinsth) for each clock of lh, excluding the element of the selected ones (SelectedClocks)
   and the unuseful ones (unusedClocks).
*/
void TrainClusters::assignClocks(ptlinsth lh)
{
  ptinsth vh;

  if (!isEventClockedFormat(graph_type))
    while (ptlinsth_get_next(&lh, &vh))
      if (CKM->getRepresentativeFormula(vh) == vh)
	if (!ptlinsth_sorted_isin_elem(SelectedClocks, vh))
	  if (!ptlinsth_sorted_isin_elem(unusedClocks, vh))
	    assignClock(vh);
}

/**
   It synthesyzes of the i/o signals of the sub-graphs.
*/
void TrainClusters::mkGraphIOReferences()
{
  ptlINode lsg;
  ptINode aSubGraph;
  ptlinstances lsig, entrees, sorties;
  ptexpcal lentrees, lsorties;

  lsg = lSubGraph;
  while (ptlINode_get_elem(&lsg, &aSubGraph))
    {
      lsig = NM->getInodelInputFlows(aSubGraph);
      entrees = ptlinstances_intersection(lsig, InputsGraph);
      // entrees = ptlinstances_sortedfusion(ptlinstances_intersection(lsig, lOutGraph), entrees);
      lentrees = EM->convertLinstances2ListExpcal(entrees);
      lsig = NM->getInodelOutputFlows(aSubGraph);
      sorties = ptlinstances_intersection(lsig, OutputsGraph);
      lsorties = EM->convertLinstances2ListExpcal(sorties);
      NM->set_inOutGraphRef(aSubGraph, lentrees, lsorties);
    }
}

void TrainClusters::setDependenceBooleanFormat(ptINode anode)
{
  ptlinsth lh, clocks;
  ptlINode lnodes;
  ptINode sg, n;
  ptinsth hAt, vh_n, vh;

  if (!isEventClockedFormat(graph_type))
    if (NM->is_subgraph(anode))
      {
	NM->get_subGraphDefinitions(anode, &lnodes, &clocks);
	while (ptlINode_get_elem(&lnodes, &n))
	  {
	    vh_n = NM->get_nodeClock(n);
	    lh = NM->getInodelInputClocks(n);
	    while (ptlinsth_get_next(&lh, &vh))
	      {
		if (CKM->getBooleanClockAt(vh, CKM->getBranching(vh, vh_n), &hAt))
		  {
		    // fprintf(stderr, " setDependenceBooleanFormat TODO \n");
		    // if (clockGetSubGraph(vh, &sg) && (sg != anode)) NM->addDependence(sg, anode, CKM->getUndefClock());
		  }
	      }
	  }
      }
}

/** It updates the dependences between the generated sub-graphs and also the kept nodes (lkeptNodes).
*/
void TrainClusters::setDependencesSubGraphs(ptlINode lkeptNodes)
{
  ptINode aSubGraph, anode;
  ptlINode lsg, ln, lnsave;
  ptlinsth lh;
  ptinsth vh;

  /* selected nodes -> clusters dependencies */
  ln = ptlINode_fusion(lkeptNodes, lSubGraph);
  lnsave = ln;
  while (ptlINode_get_elem(&ln, &anode))
    {
      setDependenceBooleanFormat(anode);
      // lsg = lSubGraph; replaced by: ( modified Aug 2016 --dependences on inputs!)
      lsg = lnsave;
      while (ptlINode_get_elem(&lsg, &aSubGraph))
	{
	  if (NM->is_subgraph(aSubGraph) || NM->is_subgraph(anode))
	    NM->setDependencesToSubGraph(anode, aSubGraph);
	}
    }

  /* Adding external clock of the cluster -> cluster Dependencies */
  lsg = lSubGraph;
  while (ptlINode_get_elem(&lsg, &aSubGraph))
    {
      lh = NM->getInodelInputClocks(aSubGraph);
      while (ptlinsth_get_next(&lh, &vh))
	{
	  vh = CKM->getRepresentativeFormula(vh);
	  NM->addDependence(vh, aSubGraph);
	}
    }
}

/**
   It applies
   - InterfaceCalculusSubGraphs() method to build the interface of the produced partitions (sub-graphs)
   - Clusters::bdcplusInterfaceCalculusSubGraphs() method to for bDC+ graph interface 
   - Clusters::InterfaceCalculusSubGraphsUpdate() method to update the interfaces of the sub-graphs..
*/
_VIRTUAL_ void TrainClusters::subGraphsBuildProfile()
{
  // Specialisation of Clusters::subGraphsBuildProfile() (why?)
  InterfaceCalculusSubGraphs();
  bdcplusInterfaceCalculusSubGraphs();
  InterfaceCalculusSubGraphsUpdate();
}

boolean TrainClusters::isOptimizedDelay(ptINode node)
{
  //  return pk_false;
  ptexpcal vce;
  if (!NM->is_delay(node))
    return pk_false;
  return EM->getAttribute(NM->get_leftOperand(node), OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) & vce);
}

/**
    It updates the dependences induced by the partitionnning.
    - dependences from Clusters to the kept nodes,
    - dependences from the kept nodes to Clusters (See setDependencesSubGraphs()),
    - dependences from the Clusters to the "output nodes" of the graph.
*/
_VIRTUAL_ void TrainClusters::setDependences()
{
  ptINode aSubGraph, anode, n;
  ptlINode ln;
  ptinsth vh;
  ptlMark lm;

  ln = SelectedNodes;
  while (ptlINode_get_elem(&ln, &anode))
    {
      vh = NM->get_nodeClock(anode);	// io node
      lm = ptinsthAttr_get_marks(vh);
      if (getSubGraphOf(lm, &aSubGraph) || isAnExtractOnInput(vh, &aSubGraph))
	NM->addDependence(aSubGraph, anode, vh);
    }

  setDependencesSubGraphs(ptlINode_Copy(SelectedNodes));

  if (NM->getOutputs(&ln))
    while (ptlINode_get_elem(&ln, &anode))
      {
	n = EM->getNode(NM->get_io_arg(anode));
	if (nodeGetSubGraph(n, &aSubGraph))
	  {
	    NM->addDependence(aSubGraph, anode, NM->get_nodeClock(anode));
	  }
      }
}

/**
   @return the list of sub-graphs produced by a partitionning defined by the following operations:
   - Cluster::partitionningInitialize()
   - Cluster::assignAllNodes()
   - assignClocks() 
   - Cluster::insertSubGraphs()
   - subGraphsBuildProfile()
   - setDependences()
   - mkGraphIOReferences()
   - Cluster::partitionningFinalize()
*/
ptlINode TrainClusters::partitionning()
{
  partitionningInitialize();
  assignAllNodes();
  assignClocks(CKM->getClocks());
  insertSubGraphs();
  subGraphsBuildProfile();
  setDependences();
  mkGraphIOReferences();
  partitionningFinalize();
  return lSubGraph;
}
