/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor of a partitionning operation applied to a graph (g).
    It initializes the internal marks: SelectedMark = 0 and OtherMark = 1.
    It is a intermediate class used for binary partitionnings such that (boolean, others), (states, others),...
*/
 separate_selected_others::separate_selected_others(ptDCG g):RunOnMark(g)
{
  SelectedMark = 0;
  OtherMark = 1;
  pragmaValueSelectedPart = nilstr;
  pragmaValueOtherPart = nilstr;
}

void separate_selected_others::initcolouring()
{
  setMinMaxMark(SelectedMark);
  setMinMaxMark(OtherMark);
}

/**
   It colours the objects of the graph using two marks: SelectedMark and OtherMark.
   The nodes are marked by colouringNodes() method, the clocks in colouringCLocks() method.
   The number of sub sets is managed in the method setMinMaxMark().
*/
_VIRTUAL_ boolean separate_selected_others::colouring()
{
  initcolouring();
  colouringNodes();
  colouringCLocks();
  return pk_true;
}
