/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
/**
   It marks a clock (h) by the internal SelectedMark mark.
*/
    void
 separate_selected_others::setSelectedMark(ptinsth h)
{
  ptinsthAttr_append_mark(h, SelectedMark);
}

/**
   This virtual method must be redefined in the class that inherits from it;
   @return false by default.
*/
_VIRTUAL_ boolean separate_selected_others::isSelectedNode(ptINode aNode)
{
  return pk_false;
}

/**
   No effect.
*/
_VIRTUAL_ void separate_selected_others::colouringClock(ptinsth h)
{
}

/**
   It marks a node (aNode) by SelectedMark if isSelectedNode(), or OtherMark otherwise
*/
_VIRTUAL_ void separate_selected_others::colouringNode(ptINode aNode)
{
  integer aMark;

  if (!NM->is_ioNode(aNode))
    {
      if (isSelectedNode(aNode))
	aMark = SelectedMark;
      else
	aMark = OtherMark;
      ptINodeAttr_append_mark(aNode, aMark);
    }
}

/**
   It marks the clocks of the graph, calling colouringClock(ptinsth) for each clock of the system.
*/
_VIRTUAL_ void separate_selected_others::colouringCLocks()
{
  ptlinsth lh;
  ptinsth vh;

  lh = getSelectedClocks();
  while (ptlinsth_get_next(&lh, &vh))
    colouringClock(CKM->getRepresentativeFormula(vh));
}

/**
   The subgraph (aSubGraph) is added in the set of sub-graphs (lSubGraph, attribute memeber of the class).
*/
_VIRTUAL_ void separate_selected_others::selectSubGraph(ptINode aSubGraph)
{
  ptlINode_insert_elem(&lSubGraph, aSubGraph);
}

_VIRTUAL_ ptlinsth separate_selected_others::selectingClocks()
{
  ptlinsth lh, lh1;
  ptinsth vh;
  boolean bdummy;

  lh1 = CKM->getClocks();
  lh = NILLINSTH;
  while (ptlinsth_get_next(&lh1, &vh))
    if (!ptlinsth_sorted_isin_elem(unusedClocks, vh))
      lh = ptlinsth_sorted_insert_elem(lh, vh, &bdummy);

  return lh;
}

/**
   It associates the name of the pragma associated with the a part (i) of the binary partition.
   Assumed : i = 1 (selected part) or i = 2 (other part)
*/
_VIRTUAL_ void separate_selected_others::setPragmaValuePart(integer ipart, ptrepstr vstr)
{
  if (ipart == 1)
    pragmaValueSelectedPart = vstr;
  else if (ipart == 2)
    pragmaValueOtherPart = vstr;
  else;				// error ?
}

/** 
    It assigns the predefined pragma call "Partition" to a subgraph ( aSubGraph ) with the value associated with i
    It is assumed that i = 1 stands for the selected part and i = 2 the other part.
 */
_VIRTUAL_ void separate_selected_others::setpragmapartition(ptINode aSubGraph, integer i)
{
  ptrepstr vstr;
  if (i == 1)
    vstr = pragmaValueSelectedPart;
  else if (i == 2)
    vstr = pragmaValueOtherPart;
  else
    vstr = string2repstr("UndefinedPart");

  NM->addPragma(aSubGraph, newPredefinedPragma(PRAGMA_SELECTED_OTHERS_PARTITION, NILOBJPRAGMA, repstr2PragmaStatement(vstr)));
}

/** 
    The vector of Subgraphs is stored in the general list (lSubGraph of the class Cluster) to apply methods of Cluster class.
    This management is previously opened in partitionningInitialize() method.
*/
_VIRTUAL_ void separate_selected_others::finalizeUseInternalRepresentation()
{
  integer i;
  ptINode aSubGraph;
  ptrepstr vstr;

  lSubGraph = NILLINODE;
  for (i = 0; i < nbsets; i++)
    {
      aSubGraph = SubGraphTable[i];
      selectSubGraph(aSubGraph);
      vstr = integer2repstr(i + 1);	/* PLG useless ? */
      setpragmapartition(aSubGraph, i + 1);
    }
}
