/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define TypeType2int(TT) (integer)TT
    void
 typeSelection::initSelectedTypes()
{
  integer i, n;
  ptlinstances lTypes;

  for (i = 0; i <= nbPredefinedType; i++)
    selectedPredefinedTypes[i] = pk_false;

  if (EM->getUserTypes(&lTypes))
    {
      n = ptlinstances_get_size(lTypes);
      TypesTab.reserve(n + 1);
      for (i = 0; i <= n; i++)
	TypesTab[i] = pk_false;
      ptlinstances_free(lTypes);
    }
}

/**
   It defines the operand type (TT) to be selected for the colouring step of  the partitionning.
*/
_VIRTUAL_ void typeSelection::selectType(TypeType TT)
{
  if (TypeTestPredef(TT))
    selectedPredefinedTypes[-TypeType2int(TT)] = pk_true;
  else
    TypesTab[TypeType2int(TT)] = pk_true;
}

/**
   @return "true" if the operand type (TT) is a selected one, "false" otherwise.
*/
_VIRTUAL_ boolean typeSelection::isSelectedType(TypeType TT)
{
  if (TypeTestPredef(TT))
    return selectedPredefinedTypes[-TypeType2int(TT)];
  return TypesTab[TypeType2int(TT)];
}

/**
   @return "true" if the operand expression (vexp) references only selected types, "false" otherwise.
   The selection of the types is done by the method a specialtion of the typeSelection:: selectTypes() method.
*/
_VIRTUAL_ boolean typeSelection::isExprBuiltOnSelecedTypes(ptexpcal vexp)
{
  TypeType TT;
  noperators vop, op;
  ptexpcal vexpcal1, vexpcal2, nexp;
  ptrepstr vstr;
  integer val;

  if (vexp == NILEXPCAL)
    return pk_true;
  switch ((int)(EM->getOperatorClass(vexp)))
    {
    case variabl:
    case constant:
      TT = EM->getType(vexp);
      return isSelectedType(TT);

    case oper1:
    case oper1tp:
      if (!isSelectedType(EM->getType(vexp)))
	return pk_false;
      vexpcal1 = EM->getUnaryArg(vexp);
      return isExprBuiltOnSelecedTypes(vexpcal1);

    case oper2:
    case oper2tp:
      vop = EM->getOperator(vexp);
      if (LL_is_listOp(vop))
	return pk_false;	// list.
      if (!isSelectedType(EM->getType(vexp)))
	return pk_false;
      if (vop == opvenum)
	return pk_true;

      vexpcal1 = EM->getBinaryArgstAt(vexp, 1);
      if (!isExprBuiltOnSelecedTypes(vexpcal1))
	return pk_false;
      vexpcal2 = EM->getBinaryArgstAt(vexp, 2);
      return isExprBuiltOnSelecedTypes(vexpcal2);

    case opereh:
      if (!isSelectedType(EM->getType(vexp)))
	return pk_false;
      vexpcal1 = EM->getWhenOrCellArg(vexp);
      return isExprBuiltOnSelecedTypes(vexpcal1);

    case opertyp:
    case oper3:
    case oper3tp:
      return pk_false;
    case call:
      return pk_false;

    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
  return pk_false;
}

/**
   Specialisation of the "abstract" method declared in the separate_selected_others::isSelectedNode().
   @return "true" if a node (aNode) produces a boolean result? "false" otherwise.
   \note restriction: only equation (one ouput) and delays are tested.
   It must be completed for procedure call, node call.
*/
_VIRTUAL_ boolean typeSelection::isSelectedNode(ptINode aNode)
{
  ptexpcal re, expg, vret, vfen, vinit;

  if (ptlINode_isIn(lSelectedNodes, aNode))
    return pk_true;
  switch ((int)(NM->get_oper(aNode)))
    {
    case opdefsig:
      // case opdefpart:
      expg = NM->get_leftOperand(aNode);
      if (!isExprBuiltOnSelecedTypes(expg))
	return pk_false;
      if (!NM->get_rightOperand(aNode, &re))
	return pk_false;
      return isExprBuiltOnSelecedTypes(EM->getBasicExpr(re));

    case opret:
    case opfen:
      NM->get_delay_arg(aNode, &expg, &re, &vret, &vfen, &vinit);
      if (!isExprBuiltOnSelecedTypes(expg))
	return pk_false;
      return isExprBuiltOnSelecedTypes(re);

    case opmoti:
      return pk_false;
    case opappe:
      return pk_false;
    case opnodecall:
      return pk_false;
    case opsubgraph:
      return pk_false;
      // case opldefpart: return pk_false;
    default:
      pK_messenger->Caseerror(currentLine, currentFile);
      return pk_false;
    }
}

_VIRTUAL_ void typeSelection::finalyzeInterface(ptlINode ln)
{
  ptexpcal vsig;
  ptlinstances lsig;
  ptINode selectedpart, otherpart;
  ptinterface selectedpartInterface, otherpartInterface;

  ptlINode_get_elem(&ln, &selectedpart);
  ptlINode_get_elem(&ln, &otherpart);
  selectedpartInterface = NM->getSet_interface(selectedpart);
  otherpartInterface = NM->getSet_interface(otherpart);
  lsig = AGRAPH->getInputs();
  while (ptlinstances_get_next(&lsig, &vsig))
    if (isExprBuiltOnSelecedTypes(vsig))
      ptinterface_insert_inputSignal(selectedpartInterface, vsig);
    else
      ptinterface_insert_inputSignal(otherpartInterface, vsig);

  lsig = AGRAPH->getOutputs();
  while (ptlinstances_get_next(&lsig, &vsig))
    if (isExprBuiltOnSelecedTypes(vsig))
      ptinterface_insert_outputSignal(selectedpartInterface, vsig);
    else
      ptinterface_insert_outputSignal(otherpartInterface, vsig);
}
