/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return vt + the exclusive clocks for a list of clocks (lh).
   For each clock H of lh, a composition of list of exclusive clocks.
   (| H ^# H1 ^# ... Hn
    | .... 
    |)   
 */
ntrees DCG2SignalTree::addExclusiveClocks(ptlinsth lh, ntrees vt)
{
  ptinsth vh, vhe;
  ptlinsth lhe;
  ntrees vt1;
  ptident id;
  int n;

  while (ptlinsth_get_next(&lh, &vh))
    if (CKM->getAttribute(vh, EXCLUSIVE_CLOCKS, (attributeValue *) & lhe))
      {
	n = CKM->getNum(vh);
	id = mkhonametree1(vh);
	while (ptlinsth_get_next(&lhe, &vhe))
	  if (CKM->getNum(vhe) > n)
	    {
	      vt1 = LL_make_lexchor_list(STree_scopy(id));
	      vt1 = ntreeList_appendLast(vt1, mkhonametree1(vhe));
	      STree_addExprToComposition_Post(&vt, vt1);
	    }
      }
  return vt;
}

/**
   @return the exclusive clocks of a graph (g).
   It is a composition of pair of exclusive clocks.
   (| H ^# H1 |
    | .... 
    |)
 */
ntrees DCG2SignalTree::mkExclusiveClocks(ptDCG g)
{
  ntrees vt, process;

  DCG2SignalTreeini(g);
  SET_NILTREE(vt);
  vt = addExclusiveClocks(CKM->getlhroots(), vt);
  vt = addExclusiveClocks(CKM->getlhplaced(), vt);
  process = STree_mkSkeletonModel(createSuffIdent(g->getName(), ecSuffix), opprocess);
  if (ISNOT_NILTREE(vt))
    STree_addExpressionOfProcessToaModel(process, vt);
  return process;
}
