/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_ ptINode DCG2SignalTree::setCurrentBlock(ptINode vCurrentBlock)
{
  ptINode x;
  x = CurrentBlock;
  CurrentBlock = vCurrentBlock;
  return x;
}

/**
   @return "true" if a node (node) is unified with a clock (vh), vh is set, "false" otherwise and vh is not modified.
*/
boolean DCG2SignalTree::isNodeUnifiedWithClock(ptINode node, ptinsth * vh)
{
  ptexpcal vLeftPart;

  vLeftPart = NM->get_leftOperand(node);
  if (vLeftPart != NILEXPCAL)
    if (EM->IsSignal(vLeftPart))
      return EM->isSignalUnifiedWithClock(vLeftPart, vh);
  return pk_false;
}

/** 
    @return "true" if a node (inode) is a selected node for the operation, "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::isSelectedNode(ptINode inode)
{
  return !ptlINode_isIn(unNodes, inode);
}

/** 
    @return "true" if a node (vnode) is assigned to the current block, "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::isNodeCurrentBlockFilteringIO(ptINode vnode)
{
  if (NM->is_readNode(vnode)  && !DEBUG_MODE) return pk_false;
  if (NM->is_writeNode(vnode) && !DEBUG_MODE) return pk_false;
  return isNodeCurrentBlock(vnode);
}

/** 
    @return "true" if a node (vnode) is assigned to the current block, "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::isNodeCurrentBlock(ptINode vnode)
{
  ptinsth h;
  ptINode cn;

  if (!isSelectedNode(vnode)) return pk_false;
  if (isEventClockedFormat(graph_type))
    if (isNodeUnifiedWithClock(vnode, &h))
      return pk_false;

  NM->get_inComposedNode(vnode, &cn);
  return (cn == CurrentBlock);
}

/**  @return the Signal abstract tree representation assigned to a read node (vnode). */
_VIRTUAL_ ntrees DCG2SignalTree::mkReadDefinition(ptINode vnode)
{
  ptINode x;
  if (!DEBUG_MODE && NM->is_outputNodeCall(vnode, &x))
    return NILTREE;
  return mkdefsigtree(NM->get_io_arg(vnode), LL_make_appe_tree(STree_scopy(getReadId()), NILTREE));
}

/**  @return the Signal abstract tree representation assigned to a write node (vnode). */
_VIRTUAL_ ntrees DCG2SignalTree::mkWriteDefinition(ptINode vnode)
{
  ntrees vt;
  // vt = ptexpcal2ntrees(EM->mkNaryExp(oplident, NM->get_io_arg(vnode), NILEXPCAL, NILTREE));
  vt = ntreeList_appendLast(LL_make_lident_emptyList(), ptexpcal2ntrees(NM->get_io_arg(vnode)));
  return LL_make_appe_tree(STree_scopy(getWriteId()), vt);
}

/**  @return the Signal abstract tree representation assigned to a external call (vnode), i.e. the graph is unknown.. */
_VIRTUAL_ ntrees DCG2SignalTree::mkProcCallDefinition(ptINode vnode)
{
  ptident id;
  ptexpcal e, lentren, lsortren, par;
  ntrees vt;

  // I/O nodes are filtered (before).
  if (NM->is_readNode(vnode))
    return mkReadDefinition(vnode);
  if (NM->is_writeNode(vnode))
    return mkWriteDefinition(vnode);

  NM->get_procCall_arg(vnode, &id, &par, &lentren, &lsortren);
  if (existPredefinedPragma(NM->getPragmas(vnode), PRAGMA_CONVERSION))
    {
      vt = STree_mkNullaryTree(identType2SignalType(ident2repstr(id)));	// predefined type.
      vt = LL_make_conv_tree(vt, mkListIdentCall(lentren, 2));

      return LL_make_defsig_tree(mkListIdentCall(lsortren, 3), vt);
    }
  if (isCartesianProductId(id))
    {
      vt = ptexpcal2ntrees(lentren);	// lentren is oppcar( e1,..., en)
      return LL_make_defsig_tree(mkListIdentCall(lsortren, 3), vt);
    }
  if (isAssertionId(id))
    {
      if (EM->getNaryArgs(&lentren, &e))
	{
	  vt = ptexpcal2ntrees(e);
	  return LL_make_appeassert_tree(vt);
	}
      else;			// internal error on assert.
    }
  return mkCallTree(id, par, lentren, lsortren);
}

/**
   @return  vt + the signals of a list (lsig) with null clock as clock.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkTreeNullClockSignals(ptexpcal lsig, ntrees vt)
{
  ptexpcal aSig;
  ptinsth hn;

  hn = CKM->getNullClock();
  while (EM->getNaryArgs(&lsig, &aSig))
    if (EM->IsSignal(aSig))
      if (CKM->satisfiesNullClock(EM->getClock(aSig)))
	{
	  if (IS_NILTREE(vt))
	    vt = mac_STree_make_initSynchro(mkhonametree(hn));
	  vt = ntreeList_appendLast(vt, ptexpcal2ntrees(aSig));
	}
  return vt;
}

/**  @return the Signal abstract tree representation of a node call (vnode), 
     i.e.the graph is known and it must be transalted.
     It it exists some signals referenced in the node call with null clock as clock, the 
     synchronisation is added to the Signal abstract tree representation.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkNodeCallDefinition(ptINode vnode)
{
  ptident nom;
  ptexpcal lent, lsort, lparam;
  ntrees vt, vt1, vtreehnull;
  ptDCG vg;

  NM->get_nodeCall_arg(vnode, &vg, &lparam, &lent, &lsort);
  nom = vg->getName();	/** vg doit etre declare */
  mustBePrinted(vg);
  vt = mkCallTree(nom, lparam, lent, lsort);
  vtreehnull = mkTreeNullClockSignals(lent, NILTREE);
  vtreehnull = mkTreeNullClockSignals(lsort, vtreehnull);
  if (IS_NILTREE(vtreehnull))
    return vt;
  vt1 = vt;
  vt = STree_getNilTree();
  STree_addExprToComposition_Post(&vt, vt1);
  STree_addExprToComposition_Post(&vt, vtreehnull);
  return vt;
}

/** 
    @return the Signal abstract tree of the expression : <em> h \b count exp </em>. 
*/
_VIRTUAL_ ntrees DCG2SignalTree::counter2ntrees(ptinsth h, ptexpcal exp)
{
  ntrees t1;

  t1 = mkhonametree(h);
  if (!isEventClockedFormat(graph_type))
    t1 = LL_make_when1_tree(t1);
  return LL_make_nombm_tree(t1, ptexpcal2ntrees(exp));
}

/**
   Assumed: expr is the basic expression of upExpr (i.e. e when hUpExpr)
   @return expr if ^e = hUpExpr, upExpr otherwise.
*/
_VIRTUAL_ ptexpcal DCG2SignalTree::SpecificTranslation(ptexpcal expr, ptexpcal upExpr, ptinsth hUpExpr)
{
  ptinsth vh;

  if (EM->getSynchronousExprHorl(expr, &vh))
    {
      vh = CKM->getRepresentativeFormula(vh);
      if (vh == hUpExpr)
	return expr;
    }
  return upExpr;
}

/** 
    @return the Signal abstract tree representation assigned to the equation (vnode).
    It can be an equation (:=), or a partial definition (::= ) or a default partial definition ( ::= defaultvalue).

    \note the delays are translated in an another method (\see{mkTreeDelay()}).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkEquationDefinition(ptINode vnode)
{
  ptexpcal vLeftPart, vRightPart, basicExpr;
  ptinsth hg, hd;
  ntrees vt;
  noperators vop;
  boolean done;

  if (!NM->get_rightOperand(vnode, &vRightPart))
    return NILTREE;
  vLeftPart = NM->get_leftOperand(vnode);
  if (!EM->IsCellExpr(vRightPart))
    {
      hg = CKM->getRepresentativeFormula(EM->getClock(vLeftPart));
      hd = CKM->getRepresentativeFormula(EM->getContributionClock(vRightPart));
      if (CKM->satisfiesInitClock(hd))
	{
	  vRightPart = EM->getBasicExpr(vRightPart);
	  vt = LL_make_extr_tree(ptexpcal2ntrees(vRightPart), mkhonametree(hg));
	}
      else
	{
	  done = pk_false;
	  if (hg == hd)
	    {
	      basicExpr = EM->getBasicExpr(vRightPart);
	      vop = EM->getOperator(basicExpr);
	      if ((vop == opmele) || (vop == opleft_tt))
		vRightPart = basicExpr;
	      else if (vop == opnombm)
		{
		  done = pk_true;
		  vt = counter2ntrees(hg, EM->getBinaryArgstAt(basicExpr, 2));
		}
	      else
		vRightPart = SpecificTranslation(basicExpr, vRightPart, hg);
	    }
	  if (!done)
	    vt = ptexpcal2ntrees(vRightPart);
	}
    }
  else
    {
      vt = ptexpcal2ntrees(vRightPart);
      nchangchild(vt, notKernptexpcal2ntrees(EM->getInitialValue(vLeftPart)), 3);
    }

  return finalyze_equation(getSignalOperator(vnode), vLeftPart, vt);
}

_VIRTUAL_ ntrees DCG2SignalTree::finalyze_equation(noperators vop, ptexpcal vLeftPart, ntrees vt)
{
  if (vop == opdefsig)
    return mkdefsigtree(vLeftPart, vt);
  return STree_mkEquation(vop, ptexpcal2ntrees(vLeftPart), vt);
}

noperators DCG2SignalTree::getSignalOperator(ptINode vnode)
{
  ptexpcal vout, vsig;
  noperators vop;

  if (NM->isDefaultValueNode(vnode))
    return opdefpartdef;

  vop = NM->get_oper(vnode);
  if ((vop == opdefsig) || (vop == opret))
    {
      vout = NM->get_leftOperand(vnode);
      if (vout == NILEXPCAL)
	return vop;
      if (!EM->satisfiesList(vout))
	{
	  if (EM->isSharedSignalBySeveralGraphs(vout))
	    return opdefpart;
	}
      else
	while (EM->getNaryArgs(&vout, &vsig))
	  {
	    if (EM->isSharedSignalBySeveralGraphs(vsig))
	      return opdefpart;
	  }
      return opdefsig;		// vop;
    }
  return opdefpart;
}

/** 
    @return the Signal abstract tree representation of an array definition (anArrayNode). 
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkArrayDefinition(ptINode anArrayNode)
{
  ptexpcal vIndex, uBound, indexDef;
  ntrees vtreeBody, vtreeInit;
  ptlINode ln;
  ptlinsth lh;
  ptINode indexNode, x;

  // push array
  x = setCurrentBlock(anArrayNode);

  indexNode = NM->get_indexNode(anArrayNode);
  vIndex = NM->get_index(indexNode);
  indexDef = NM->get_index_def(indexNode);
  uBound = EM->getTernaryArgstAt(indexDef, 2);
  // Body
  NM->get_array_bodyDefinitions(anArrayNode, &ln, &lh);
  vtreeBody = inodes2ntrees(ln, NILTREE);
  if (isEventClockedFormat(graph_type))
    vtreeBody = clocks2ntrees(lh, vtreeBody);

  // init
  NM->get_array_initDefinitions(anArrayNode, &ln, &lh);
  vtreeInit = inodes2ntrees(ln, NILTREE);
  if (isEventClockedFormat(graph_type))
    vtreeInit = clocks2ntrees(lh, vtreeInit);

  //   pop array
  setCurrentBlock(x);

  return LL_make_moti_tree(LL_make_indmot_tree(notKernptexpcal2ntrees(vIndex), notKernptexpcal2ntrees(uBound)), vtreeBody, vtreeInit);
}

/**  
     @return the Signal abstract tree representation assigned to a delay node (vnode).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkDelayDefinition(ptINode vnode)
{
  ptexpcal vexpsigres, vexpsigret, vexpret, vexpfen, vinit;

  NM->get_delay_arg(vnode, &vexpsigres, &vexpsigret, &vexpret, &vexpfen, &vinit);
  return finalyze_equation(getSignalOperator(vnode), vexpsigres, mkTreeDelay(vexpsigret, vexpret, vexpfen, vinit, vexpsigres));
}

/** 
    @return the nil tree for a bundle node (n). Not (yet) implemented.
*/
ntrees DCG2SignalTree::mkBundleDefinition(ptINode n)
{
  return NILTREE;
}

/** 
    @return the Signal tree (vtNode) of a node (node) with its assigned label if it exists, the Signal tree (vtNode) otherwise.
    If th node is a subgraph, the clock of the subgraph is added to the resulted tree.
*/
ntrees DCG2SignalTree::addLabelTo(ptINode node, ntrees vtNode)
{
  ntrees vLabel, vtclk, vt;
  ptinsth h;
  ptexpcal vlab;

  vLabel = getLabelTree(node);
  if (IS_NILTREE(vLabel)) return vtNode;
  if (getLinkNodeToTree(node, &vt)) return vtNode; // already declared.

  SET_NILTREE(vt);
  STree_addExprToComposition_Post(&vt, LL_make_petiq_tree(vLabel, vtNode));
  STree_addLabelInModel(STree_scopy(vLabel), ptinsthAttr_get_sigtree(vRootClock));	// Declaring the label at the top level.

  // Adding lab ^= clock, for a subgraph.
  if (NM->get_label(node, &vlab))
    {
      h = EM->getClock(vlab);
      if (!CKM->satisfiesFictitiousRoot(h) && !CKM->satisfiesUndefClock(h))
	{
	  vtclk = STree_mkSynchroIni(STree_scopy(vLabel));
	  vtclk = ntreeList_appendLast(vtclk, mkWhenClock1(CKM->getRepresentativeFormula(h)));
	  STree_addExprToComposition_Post(&vt, vtclk);
	}
    }
  return vt;
}

/**
   @return the abstract tree label: reference_to(aSubGraph) for the sub graph node aSubGraph, 
   where label is the label associated with the sub graph.
*/
ntrees DCG2SignalTree::mkReferenceToSubGraphDefinition(ptINode aSubGraph, boolean babstract)
{
  return addLabelTo(aSubGraph, mkSubGraphCallTree(aSubGraph, babstract));
}

/**  
     @return the SIGNAL Abstract Tree representation of a node (vnode).
*/
ntrees DCG2SignalTree::mkNodeDefinition(ptINode vnode)
{
  ntrees t;
  noperators vop;

  vop = (int)(NM->get_oper(vnode));
  switch (vop)
    {
    case opldefpart:
      return NILTREE;

    case opdefsig:
    case opdefpart:
      t = mkEquationDefinition(vnode);
      break;
    case opmoti:
      t = mkArrayDefinition(vnode);
      break;
    case opret:
      t = mkDelayDefinition(vnode);
      break;
    case opappe:
      t = mkProcCallDefinition(vnode);
      break;
    case opnodecall:
      t = mkNodeCallDefinition(vnode);
      break;
    case opsubgraph:
      t = mkReferenceToSubGraphDefinition(vnode, pk_false);
      break;
    case opbund:
      t = mkBundleDefinition(vnode);
      break;

    case opnil:
      SET_NILTREE(t);
      break;
    default:
      pK_messenger->Caseerror(currentLine, currentFile);
      return NILTREE;
    }
  if (vop != opsubgraph)
    t = addLabelTo(vnode, t);
  mkNodeDebugInfos(vnode, t);
  mkMemoryOptimizedAttributreTree(vnode);
  return t;
}

/** 
    @return the SIGNAL Abstract Tree representation (a composition) of a list of nodes (li), 
    added to the operand list (composition).
*/
ntrees DCG2SignalTree::inodes2ntrees(ptlINode li, ntrees composition)
{
  ptINode i;
  ptinsth h;

  while (ptlINode_get_elem(&li, &i))
    if (!isNodeUnifiedWithClock(i, &h))
      STree_addExprToComposition_Post(&composition, mkNodeDefinition(i));
  return composition;
}

/** 
    @return "true" if the trees (model) is modified, "false" otherwise.
    It adds to a model (model) the Signal abstract tree representation of a list of nodes (lnodes).

    It calls mkNodeDefinition(). Moreover, for a sub-graph, the definition is also inserted (SubGraph2sigTree()).
*/
_VIRTUAL_ boolean DCG2SignalTree::buildtree(ptlINode lnodes, ntrees model)
{
  boolean use;
  ptINode blockNode, vnode;
  ntrees SubProcess;
  ntrees vlabel, t, x, y;

  // lbesnard:  x, G_TreeGraph have been added for solving some problems of the visibility of labels.
  use = pk_false;
  while (ptlINode_get_elem(&lnodes, &vnode))
    if (isNodeCurrentBlockFilteringIO(vnode))
      {
	use = pk_true;
	if (!NM->isMarked(vnode, subgraphAlreadyInserted))
	  {
	    t = mkNodeDefinition(vnode);
	    if (!NM->is_subgraph(vnode))
	      {
		STree_addExpressionOfProcessToaModel(model, t);
		setLinkNodeToTree(vnode, t);
	      }
	    else
	      {
		NM->addMark(vnode, subgraphAlreadyInserted);
		vlabel = getLabelTree(vnode);

		SubProcess = SubGraph2sigTreeInitialize(vnode, pk_false);
		SubGraph2sigTree(vnode, pk_false, SubProcess);
		if (NM->get_inComposedNode(vnode, &blockNode))
		  getLinkNodeToTree(blockNode, &x);
		else
		  x = G_TreeGraph;
		STree_addProcessDeclaration(x, SubProcess);
		// If no dependence on vnode (i.e. restructuration, it is declared at its clock (i.e. in the model abstact tree, otherwise at the top level).
		if ((NM->get_successors(vnode) == NILLDEPENDENCE) && (NM->get_clockSuccessors(vnode) == NILLINSTH)
		    && (NM->get_predecessors(vnode) == NILLDEPENDENCE) && (NM->get_clockPredecessors(vnode) == NILLINSTH))
		  y = model;
		else
		  y = x;
		// Si pas de dependences (sbo, sso,....)
		STree_addExpressionOfProcessToaModel(y, t);
		STree_addLabelInModel(vlabel, y);
	      }
	  }
      }
  return use;
}

/**
   For debugging:
*/
_VIRTUAL_ boolean DCG2SignalTree::addNodeInfos(ptINode node, ntrees t)
{
  ptINode n;
  ptrepstr vstr;
  ntrees vattr;

  if (!DEBUG_MODE)
    return pk_false;
  if (IS_NILTREE(t))
    return pk_false;

  vattr = mknumtree(NM->get_num(node));
  if (NM->is_array(node))
    {
      vattr = ntreeList_appendLast(LL_make_lident_emptyList(), vattr);
      n = NM->get_indexNode(node);
      makeRepstr_string_int(" index node ", NM->get_num(n), &vstr);
      vattr = ntreeList_appendLast(vattr, LL_make_nom_atom(vstr));
      NM->get_array_inputNode(node, &n);
      makeRepstr_string_int(" internal array node: ", NM->get_num(n), &vstr);
      vattr = ntreeList_appendLast(vattr, LL_make_nom_atom(vstr));
    }
  else				// SubGraph
  if (NM->get_subGraphInputNode(node, &n))
    {
      vattr = ntreeList_appendLast(LL_make_lident_emptyList(), vattr);
      makeRepstr_string_int(" internal subGraph node: ", NM->get_num(n), &vstr);
      vattr = ntreeList_appendLast(vattr, LL_make_nom_atom(vstr));
    }
  STree_setVMAnnotations(t, vattr);
  return pk_true;
}

/**
   For debugging:
   The tree (t) is annotated with the number associated with a node (node).
*/
_VIRTUAL_ void DCG2SignalTree::mkNodeDebugInfos(ptINode node, ntrees t)
{
  ntrees vattr;
  ptinsth h;
  ptlinsth lh;
  ptldependence lsucc;
  ptdependence aDep;
  ptinsth hdep;
  ptINode n11, n12;
  ptrepstr vstr;

  if (addNodeInfos(node, t))
    {
      // Sucessors
      vattr = LL_make_lident_emptyList();
      lsucc = NM->get_successors(node);

      while (lsucc->get_next(&aDep, &lsucc))
	{
	  hdep = aDep->getDependence(&n11, &n12);
	  if (NM->is_writeNode(n12))
	    makeRepstr_string_int("! ", NM->get_num(n12), &vstr);
	  else
	    {
	      if (NM->is_readNode(n12)) 
		makeRepstr_string_repstr("? ", ident2repstr(EM->getName(NM->get_io_arg(n12))), &vstr);
	      else
	      vstr = integer2repstr(NM->get_num(n12));
	    }
	  vattr = ntreeList_appendLast(vattr, LL_make_nom_atom(vstr));

	}
      if (ntreearity(vattr) > 0)
	{
	  vattr = mac_STree_make_setOf("Node Succ", vattr);
	  STree_setVMAnnotations(t, vattr);
	}

      vattr = LL_make_lident_emptyList();
      lh = NM->get_clockSuccessors(node);
      while (ptlinsth_get_next(&lh, &h))
	{
	  h = CKM->getRepresentativeFormula(h);
	  vattr = ntreeList_appendLast(vattr, mkhonametree(h));
	}

      if (ntreearity(vattr) > 0)
	{
	  vattr = mac_STree_make_setOf("Clock Succ", vattr);
	  STree_setVMAnnotations(t, vattr);
	}
      // Predecessors
      vattr = LL_make_lident_emptyList();
      lsucc = NM->get_predecessors(node);

      while (lsucc->get_next(&aDep, &lsucc))
	{
	  hdep = aDep->getDependence(&n11, &n12);
	  if (NM->is_readNode(n11))
	    {
	      // makeRepstr_string_int("? ", NM->get_num(n11), &vstr);
	      makeRepstr_string_repstr("? ", ident2repstr(EM->getName(NM->get_io_arg(n11))), &vstr);
	    }
	  else
	    {
	      vstr = integer2repstr(NM->get_num(n11));
	    }
	  vattr = ntreeList_appendLast(vattr,  LL_make_nom_atom(vstr));
	}
      if (ntreearity(vattr) > 0)
	{
	  vattr = mac_STree_make_setOf("Node Pred", vattr);
	  STree_setVMAnnotations(t, vattr);
	}

      vattr = LL_make_lident_emptyList();
      lh = NM->get_clockPredecessors(node);
      while (ptlinsth_get_next(&lh, &h))
	{
	  h = CKM->getRepresentativeFormula(h);
	  vattr = ntreeList_appendLast(vattr, mkhonametree(h));
	}

      if (ntreearity(vattr) > 0)
	{
	  vattr = mac_STree_make_setOf("Clock Pred", vattr);
	  STree_setVMAnnotations(t, vattr);
	}
    }
}

/**
   Not yet implemented.
   It should produce a trace of the memory optimizations.
*/
void DCG2SignalTree::mkMemoryOptimizedAttributreTree(ptINode vnode)
{
  return;

// To be tested
//  ptlinstances lo;  ptexpcal s1, s2;  ntrees l;
/*     lo = NM->getInodelOutputFlows(vnode); */
/*     while (ptlinstances_get_next(&lo, &s1)) */
/*       if (EM->getAttribute(s1, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *)s2)) */
/* 	{ */
/* 	  l =  LL_make_lident_emptyList(); */
/* 	  l = ntreeList_appendLast(l, STree_scopy(EM->getName(s1))); */
/* 	  l = ntreeList_appendLast(l, STree_scopy(EM->getName(s2))); */
/* 	  STree_addExprToComposition_Post(&lMemOptimizedAttribute, l); */
/* 	} */
}

/**
   @return the Signal abstract tree representation of the label of a sub-graph (vnode).
*/
_VIRTUAL_ ntrees DCG2SignalTree::getLabelTree(ptINode vnode)
{
  ptexpcal lab;
  if (NM->get_label(vnode, &lab))
    return ptexpcal2ntrees(lab);
  return NILTREE;
}

/**
   @return the Signal abstract tree representation of the name of a node (vnode).
   This name is 
   - the label for a sub-graph
   - the name of one output for the others when it exists such an element, NILTREE otherwise.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkNameNode(ptINode vnode)
{
  ptexpcal sig;
  if (!NM->getDependenceNodeLabel(vnode, &sig))
    return NILTREE;
  return ptexpcal2ntrees(sig);
}

_VIRTUAL_ ntrees DCG2SignalTree::lnodes2trees(ptlINode ln)
{
  ptINode node;
  ntrees vt, vt1;

  SET_NILTREE(vt);
  while (ptlINode_get_elem(&ln, &node))
    {
      vt1 = mkNodeDefinition(node);
      if (ISNOT_NILTREE(vt1))
	{
	  if (IS_NILTREE(vt))
	    vt = LL_make_lfeed_list(vt1);
	  else
	    vt = ntreeList_appendLast(vt, vt1);
	}
    }
  return vt;
}
