/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return the boolean type when t is the pure type and the graph is not a DC+ one, t otherwise.
*/
    _VIRTUAL_ TypeType DCG2SignalTree::__event2booleanConversion(TypeType t)
{
  // unused now. the transformation boolean/event is done by the Interformat transformation.
  return t;

  if (isEventClockedFormat(graph_type))
    return t;
  if (!TypeTestPredef(t))
    return t;
  if (t == PureType)
    return BooleanType;
  return t;
}

/**
   @return the SIGNAL Abstract Tree representation of a type (t) assumed to be a structure type or a bundle type.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkStructBundleDescription(TypeType t)
{
  ntrees description_type, t1;
  NameList NL;
  TypeList fieldsType;
  TypeType TT;
  TypeSubList NSubLstType;
  boolean Goon;
  NameSubList NSubLst;

  if (!(TypeTestStructureClass(t) || TypeTestBundleClass(t)))
    return NILTREE;
  description_type = LL_make_ldecs_emptyList();
  NL = StructureTypeGetNameLst(t);	// field Name
  NSubLst = NameListCast2SubList(NL);
  fieldsType = StructureTypeGetTypeLst(t);
  NSubLstType = TypeListCast2SubList(fieldsType);

  Goon = pk_true;
  while (Goon)
    {
      t1 = LL_make_lsign_list(repstr2ident(NameSubListGetHead(NSubLst)));
      TT = TypeSubListGetHead(NSubLstType);
      t1 = LL_make_decsig_tree(TypeType2ntrees(TT), t1);
      description_type = ntreeList_appendLast(description_type, t1);
      if (NameSubListTestLast(NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail(NSubLst);
	  NSubLstType = TypeSubListGetTail(NSubLstType);
	}
    }
  return description_type;
}

/**
   @return the SIGNAL Abstract Tree representation of the clock part of a type (t) assumed to be a bundle type.
   \note Not implemented.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkBundleClockDescription(TypeType t)
{
  // incomplete
  return STree_mkAtomNilTree();
}

/**
   @return the SIGNAL Abstract Tree representation of a bundle type (t).
   It returns nil tree when t is not a bundle type.
*/
_VIRTUAL_ ntrees DCG2SignalTree::TypeTypeBundle2ntrees(TypeType t)
{
  if (!TypeTestBundleClass(t))
    return NILTREE;
  return LL_make_bund_tree(mkStructBundleDescription(BundleTypeGetTerm(t)), mkBundleClockDescription(t));
}

/**
   @return the SIGNAL Abstract Tree representation of a structure type (t).
   It returns nil tree when t is not a structure type.
*/
_VIRTUAL_ ntrees DCG2SignalTree::TypeTypeStructure2ntrees(TypeType t)
{
  if (!TypeTestStructureClass(t))
    return NILTREE;
  return LL_make_stru_tree(mkStructBundleDescription(t));
}

/**
   @return "true" if a type (TT) is an element of l marked by m, "false" otherwise.
   TT is in l iff it has already been genrated.
*/
_VIRTUAL_ boolean DCG2SignalTree::TestTypeHasBeenGenerated(TypeType TT, ptlinstances l, typeMark m)
{
  ptexpcal elem;

  if (TypeTestPredef(TT))
    return pk_true;
  while (ptlinstances_get_next(&l, &elem))
    if (EM->getOperator(elem) == optype)
      if (EM->getExpType(elem) == TT)
	if (EM->isMarked(elem, m))
	  return pk_true;
  return pk_false;
}

/**
   @return "true" if all the element of l are marked by m, "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::ConstantsHaveBeenGenerated(ptlinstances l, typeMark m)
{
  ptexpcal elem;

  while (ptlinstances_get_next(&l, &elem))
    if (!EM->isMarked(elem, m))
      return pk_false;
  return pk_true;
}

/**
   @return "true" if the types referenced by a type (tt) are elements of l marked by m, "false" otherwise.
   The marked elements are already generated.
*/
_VIRTUAL_ boolean DCG2SignalTree::TestSubTypesHaveBeenGenerated(TypeType t, ptlinstances l, typeMark m)
{
  TypeList l1;
  TypeType t1;
  TypeSubList SubLst1;
  ptlinstances lcstes;
  boolean b;

  if (TypeTestPredef(t))
    return pk_true;
  if (TypeTestExternClass(t))
    return TestTypeHasBeenGenerated(t, l, m);

  if (TypeTestArrayClass(t))
    {
      // Verifier que les cosntantes des ArrayMatrixTypeGetDimLst(t) sont generees
      lcstes = EM->getConstantsReferencedIn(ArrayMatrixTypeGetDimLst(t));
      b = ConstantsHaveBeenGenerated(lcstes, m);
      ptlinstances_free(lcstes);
      if (!b)
	return pk_false;
      return TestSubTypesHaveBeenGenerated(ArrayMatrixTypeGetTerm(t), l, m);
    }
  if (TypeTestEnumerationClass(t))
    return pk_true;
  if (TypeTestMetaClass(t))
    return TestTypeHasBeenGenerated(t, l, m);
  if (TypeTestNamedClass(t))
    return TestSubTypesHaveBeenGenerated(NamedTypeGetTerm(t), l, m);
  if (TypeTestStructureClass(t))
    return TestSubTypesHaveBeenGenerated(StructureTypeGetArgument(t), l, m);
  if (TypeTestBundleClass(t))
    return TestSubTypesHaveBeenGenerated(BundleTypeGetTerm(t), l, m);

  if (TypeTestTupleClass(t))
    {
      l1 = TupleTypeGetTypeLst(t);
      SubLst1 = TypeListCast2SubList(l1);
      while (pk_true)
	{
	  t1 = TypeSubListGetHead(SubLst1);
	  if (!TestTypeHasBeenGenerated(t1, l, m))
	    return pk_false;
	  if (TypeSubListTestLast(SubLst1))
	    return pk_true;
	  SubLst1 = TypeSubListGetTail(SubLst1);
	}
    }
  return pk_false;
  /*
     if (TypeTestBundleClass ( t ))  return TypeTypeBundle2ntrees(t);
     if (TypeTestSubrangeClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestBitsClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestLambdaClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestGenericNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
     if (TypeTestMetaNodeClass ( t )) return STree_mkNOTIMPLEMENTED();
     if (TypeTestMatrixClass ( t ))  return STree_mkNOTIMPLEMENTED();;
     if (TypeTestTupleClass ( t ))  return STree_mkNOTIMPLEMENTED();
   */

}

/**
   @param vtype (out) the Abstract tree Signal representation of a the operand type t.
   @param t (in) a type.
   @return "true" if vtype is defined, "false" otherwise.
*/
boolean DCG2SignalTree::TypeTypeUser2ntrees(TypeType t, ntrees * vtype)
{
  ntrees vinit, name, description_type;
  NameList NL;
  boolean Goon;
  NameSubList NSubLst;

  vinit = STree_mkAtomNilTree();
  if (TypeTestExternClass(t))
    {
      name = repstr2ident(ExternTypeGetName(t));
      description_type = STree_mkUndefinedExternalValue();
      (*vtype) = STree_mkDclType(name, description_type, vinit);
      return pk_true;
    }
  if (TypeTestEnumerationClass(t))
    {
      name = repstr2ident(EnumerationTypeGetName(t));
      description_type = LL_make_tenum_emptyList();
      NL = EnumerationTypeGetNameLst(t);
      NSubLst = NameListCast2SubList(NL);
      Goon = pk_true;
      while (Goon)
	{
	  description_type = ntreeList_appendLast(description_type, repstr2ident(NameSubListGetHead(NSubLst)));
	  if (NameSubListTestLast(NSubLst))
	    Goon = pk_false;
	  else
	    NSubLst = NameSubListGetTail(NSubLst);
	}
      (*vtype) = STree_mkDclType(name, description_type, vinit);
      return pk_true;
    }

  if (TypeTestMetaClass(t))
    {
      (*vtype) = STree_mkDclType(repstr2ident(MetaTypeGetName(t)), NILTREE, NILTREE);
      return pk_true;
    }

  if (TypeTestNamedClass(t))
    {
      (*vtype) = STree_mkDclType(repstr2ident(NamedTypeGetName(t)), TypeType2ntrees(NamedTypeGetTerm(t)), NILTREE);
      return pk_true;
    }
  return pk_false;

  /** 
    if (TypeTestSubrangeClass ( t )) return STree_mkNOTIMPLEMENTED();
    if (TypeTestBitsClass ( t )) return STree_mkNOTIMPLEMENTED();
    if (TypeTestLambdaClass ( t ))  return STree_mkNOTIMPLEMENTED();
    if (TypeTestNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
    if (TypeTestGenericNodeClass ( t ))  return STree_mkNOTIMPLEMENTED();
    if (TypeTestMetaNodeClass ( t )) return STree_mkNOTIMPLEMENTED();
    if (TypeTestMatrixClass ( t ))  return STree_mkNOTIMPLEMENTED();;
    if (TypeTestTupleClass ( t ))  return STree_mkNOTIMPLEMENTED();

  */
}

/**  
     @return the Signal Abstract tree representation of a type (t).
*/
ntrees DCG2SignalTree::TypeType2ntrees(TypeType t)
{
  ntrees treeTypeBase;

  // t = event2booleanConversion(t);
  if (TypeTestPredef(t))
    return STree_mkNullaryTree(predefTypeType2SignalType(t));
  if (TypeTestArrayClass(t))
    {
      treeTypeBase = TypeType2ntrees(ArrayMatrixTypeGetTerm(t));
      return LL_make_tvect_tree(notKernptexpcal2ntrees(ArrayMatrixTypeGetDimLst(t)), treeTypeBase);
    }
  if (TypeTestExternClass(t))
    return repstr2ident(ExternTypeGetName(t));
  if (TypeTestEnumerationClass(t))
    return repstr2ident(EnumerationTypeGetName(t));
  if (TypeTestMetaClass(t))
    return repstr2ident(MetaTypeGetName(t));
  if (TypeTestStructureClass(t))
    return TypeTypeStructure2ntrees(t);
  if (TypeTestBundleClass(t))
    return TypeTypeBundle2ntrees(t);
  if (TypeTestNamedClass(t))
    return repstr2ident(NamedTypeGetName(t));

  if (TypeTestSubrangeClass(t))
    return STree_mkNOTIMPLEMENTED();
  if (TypeTestBitsClass(t))
    return STree_mkNOTIMPLEMENTED();
  if (TypeTestLambdaClass(t))
    return STree_mkNOTIMPLEMENTED();
  if (TypeTestNodeClass(t))
    return STree_mkNOTIMPLEMENTED();
  if (TypeTestGenericNodeClass(t))
    return STree_mkNOTIMPLEMENTED();
  if (TypeTestMetaNodeClass(t))
    return STree_mkNOTIMPLEMENTED();
  if (TypeTestMatrixClass(t))
    return STree_mkNOTIMPLEMENTED();;
  if (TypeTestTupleClass(t))
    return STree_mkNOTIMPLEMENTED();

  return STree_mkNullaryTree(opimplicit);
}

/**
   @return "true" if the type or the constant referenced in an expression (elem) can be generated, "false" otherwise.
   It can be generated if the types/constants it references are already generated. 
   The already generated types/constants (l) are marked by m.
   - for a constant, it calls TestTypeHasBeenGenerated()
   - for a type, it calls TestSubTypesHaveBeenGenerated().
*/
_VIRTUAL_ boolean DCG2SignalTree::testCanBeGenerated(ptexpcal elem, typeMark m, ptlinstances l)
{
  TypeType TT;

  // Case : elem is a type.
  if (EM->getOperator(elem) == optype)
    {
      TT = EM->getExpType(elem);
      if (TypeTestExternClass(TT))
	return pk_true;
      return TestSubTypesHaveBeenGenerated(TT, l, m);
    }

  // Case: elem is a constant
  TT = EM->getType(elem);
  return TestTypeHasBeenGenerated(TT, l, m);
  // Il faudra verifier que les expressions d'initialisation sont declarees
  /** 
     if (!EM->IsExternalConstant(elem))
     {
     vnode = EM->getNode(aConst);
     NM->get_equation_arg(vnode, &vLeftPart, &vRightPart);
     }
  */
}
