/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isEventClockedFormat(t) ((t == DCG_BASIC) ||(t == DCG_POLY) || (t == DCG_ENDO))

#define mac_STRee_get_argIfEvent(vt)  ((noper(vt) == opevnt) ? nchild(1, vt) : vt)
#define mac_STree_make_comment(numComment) LL_make_comment_atom(pK_messenger->recupmessStr(numComment))
#define mac_STree_make_setOf(vNameSet, vSet) LL_make_egal_tree( LL_make_nom_atom( string2repstr(vNameSet)), vSet)
#define mac_STree_make_initSynchro(vt)  LL_make_lequhor_list( mac_STRee_get_argIfEvent(vt))
#define isIOSignal(aSig) (EM->IsInputFlow(aSig) || EM->IsOutputFlow(aSig))

_PROTECTED_
/// Number of elements in a list of trees (there is a limit in the SIGNAL syntactic analyzer)
#define CST_NB_MAX_LIST_ELEM 1000
/// PREDEFINED OBJECT OF THE VIRTUAL MACHINE.
    ptDCG2SignalTree pK_Graph2SignalTreeBuilder = new DCG2SignalTree();

/**
   @return a new ident built on the opernad prefix (identPrefix).
   The global attribute bAddedSignal is set to pk_true.
*/
ptident DCG2SignalTree::mkNewIdentKernelMode()
{
  identPrefix = identPrefixKernel;
  return mkNewIdent();
}

/**
   @return the union of the two operand lists (lextref, l1) with side effects on lextref. 
   The equality is tested on the node call name.

   It assumes that le lists are list of node calls.
*/
ptlINode DCG2SignalTree::fusionExternalProcs(ptlINode lextref, ptlINode l1)
{
  ptINode n1, n2;
  ptident id1, id2;
  boolean f;
  ptlINode l2;

  while (ptlINode_get_elem(&l1, &n1))
    {
      NM->get_nameCall(n1, &id1);
      l2 = lextref;
      f = pk_false;
      while (!f && ptlINode_get_elem(&l2, &n2))
	{
	  NM->get_nameCall(n2, &id2);
	  f = eq_idents(id1, id2);
	}
      if (!f)
	ptlINode_insert_elem(&lextref, n1);
    }
  return lextref;
}

/** It adds a graph (g) in the list of the graphs to be translated. */
_VIRTUAL_ void DCG2SignalTree::mustBePrinted(ptDCG g)
{
  ptlPointer_insert_elem(&lIGraph, (pointer) g);
}

/** 
    It associates the Signal Abstract Tree representation (vhtree) with a clock (vh).
*/
_VIRTUAL_ void DCG2SignalTree::ptinsthAttr_set_sigtree(ptinsth vh, ntrees vhtree)
{
  if (flatMode)
    if (!CKM->satisfiesFictitiousRoot(vh))
      return;
  CKM->addAttribute(vh, SignalTreeAttribute, (attributeValue) vhtree);
}

_VIRTUAL_ void DCG2SignalTree::setLinkNodeToTree(ptINode vnode, ntrees vtree)
{
  // if (flatMode) if (!CKM->satisfiesFictitiousRoot(vh)) return;
  NM->addAttribute(vnode, SignalTreeAttribute, (attributeValue) vtree);
}

_VIRTUAL_ boolean DCG2SignalTree::getLinkNodeToTree(ptINode vnode, ntrees * vtree)
{
  ntrees attribut;
  if (!NM->getAttribute(vnode, SignalTreeAttribute, (attributeValue *) & attribut))
    return pk_false;
  *vtree = attribut;
  return pk_true;
}

/** 
    @return the Signal Abstract Tree representation assigned to a clock (vh).
*/
_VIRTUAL_ ntrees DCG2SignalTree::ptinsthAttr_get_sigtree(ptinsth vh)
{
  ntrees attribut;

  if (flatMode)
    vh = AGRAPH->getRootClock();
  else
    vh = CKM->getRepresentativeFormula(vh);
  if (!CKM->getAttribute(vh, SignalTreeAttribute, (attributeValue *) & attribut))
    return NILTREE;
  return attribut;
}

/** 
    It removes the Signal Abstract Tree representation associated to a clock (vh).
*/
_VIRTUAL_ void DCG2SignalTree::ptinsthAttr_remove_sigtree(ptinsth vh)
{
  CKM->removeAttribute(vh, SignalTreeAttribute);
  CKM->removeMark(vh, VisitedMark);
}

/**
   It removes the attribute associated with the clocks of a list (lh) for the translation.
   (ptinsthAttr_remove_sigtree()).
*/
_VIRTUAL_ void DCG2SignalTree::rmAttributes(ptlinsth lh)
{
  ptinsth vh;
  while (ptlinsth_get_next(&lh, &vh))
    ptinsthAttr_remove_sigtree(CKM->getRepresentativeFormula(vh));
}

/**
   It removes the attribute (SignalTreeAttribute) associated with all the clocks of the hierarchy for the translation.
   It removes the attribute (subgraphAlreadyInserted) associated with the subgraph nodes.
   It resets the list of the printed dependences (global list: printedDependences).
   \see resetSigTreelClock(ptlinsth)
*/
_VIRTUAL_ void DCG2SignalTree::cleanGraph()
{
  ptlINode ln;

  rmAttributes(CKM->getlhroots());
  rmAttributes(CKM->getlhplaced());
  if (NM->getSubGraphs(&ln))
    NM->resetMarklINodes(ln, subgraphAlreadyInserted);
  NM->resetMarkAll(SignalTreeAttribute);
  printedDependences.clear();
}

/** 
    @return the name of the current graph taking into account 
    - the attribute called ATTRIBUTE_SUFFIX if it exists, 
    - or the subformat: the name of the graph is catenated with the DCG type name (BASIC, POLY, ENDO, BOOL, FLAT, SCH).
*/
_VIRTUAL_ ptident DCG2SignalTree::mkGraphNameWithSubFormat(ptDCG g)
{
  ptident suff;

  if (g->getAttribute(ATTRIBUTE_SUFFIX, (attributeValue *) & suff))
    return newIdent_concIdentIdent(g->getName(), suff);
  return g->getLevelName();
}

/**
   It defines the attribute member of the class for the managing of the flat mode:
   - flatMode is set to bflat
   - G_LocSig in flat mode will contain all the local signals of the hierarchy.
   - G_Clocks in flat mode will contain all the clocks of the hierarchy.
   - G_TreeGraph is set to vtProcess (upper model)

   This mode is used when the graph contains labels (the visibility of labels is local to the declaration of a model).
*/
_VIRTUAL_ void DCG2SignalTree::flatModeInit(ntrees vtProcess, boolean bflat)
{
  setFlatMode(bflat);
  genInterface = !flatMode;
  G_LocSig = NILLINSTANCES;
  G_Clocks = NILLINSTH;
  G_TreeGraph = vtProcess;
}

/**
   Added to allow specialisation of such a translator.
   For this translation (Graph to Signal Tree), do nothing, but must be obviously kept.
*/
_VIRTUAL_ void DCG2SignalTree::specificTranslatorInit()
{
}

/** 
    Initializing the translator for a graph (g) with the characteristics of the graph.
*/
void DCG2SignalTree::initTranslatorFor(ptDCG g)
{
  AGRAPH = g;
  CKM = AGRAPH->getClockManager();
  EM = AGRAPH->getExpressionManager();
  NM = AGRAPH->getNodeManager();
  NameM = AGRAPH->getNameManager();
  lpragmas = AGRAPH->getPragmas();
  inputs = AGRAPH->getInputs();
  outputs = AGRAPH->getOutputs();
  graph_type = AGRAPH->getFormatLevel();
  unNodes = AGRAPH->getUnused(&unClocks);
  isEndochronous = AGRAPH->isEndochronous();
}

/** Initializing the translator for the graph g. */
void DCG2SignalTree::DCG2SignalTreeini(ptDCG g)
{

  initTranslatorFor(g);
  SignalTreeAttribute = getMark();
  VisitedMark = getMark();
  subgraphAlreadyInserted = getMark();
  RunOnSetOnSubgraphs = pk_false;
  if (notInitialized)
    {
      traSuffix = string2repstr("TRA");
      abstractSuffix = string2repstr("ABSTRACT");
      ecSuffix = string2repstr("EXCLUSIVE_CLOCKS");
      identPrefixKernel = string2ident("__KID_");
      identPrefix = getIdentNil();
      notInitialized = pk_false;
    }
  mkNewIdentTag = pk_false;
  noDoubleSynchroSameId = pk_false;
  cpt_names = 0;

  specificTranslatorInit();
  G_TickIsReferenced = pk_false;
  bDependenceWithClock = pk_false;
}

void DCG2SignalTree::setInfosTracabilty(ptexpcal vexp, ntrees vt)
{
  ntrees vattr;

  return;
  // if (TraceabilityMode) ;
  if (DEBUG_MODE)
    if (ISNOT_NILTREE(vt))
      {
	vattr = mknumtree(EM->getNum(vexp));
	STree_setVMAnnotations(vt, vattr);
      }
}

/**
   @return 
   - "true" if 
      - the current mode (flatMode) is the "flat" mode 
      - or when the graph is a graph of clusters,
      - or when the hierarchy must be printed in flat mode (induced by the labels)
   - "false" otherwise.
   \see setFlatMode()
*/
_VIRTUAL_ boolean DCG2SignalTree::getHierarchyModePrinting()
{
  ptINode aNode;
  ptlINode lsg;

  if (flatMode)
    return pk_true;
  // return pk_false; to see the clock hierarchy for a program with labels.

  // When the graph is a  graph of clusters, it return pk_true
  if (NM->getSubGraphs(&lsg))
    while (ptlINode_get_elem(&lsg, &aNode))
      if (NM->is_trainCluster(aNode))
	return pk_true;

  if (flatmodeRequiredForDependencies(AGRAPH->getExplicitDependences()))
    return pk_true;
  return flatmodeRequiredForDependencies(AGRAPH->getReinforcedDependences());
}

/**
   @return the tree of the identifier (id).
*/
ntrees DCG2SignalTree::mknumtreewithprefix(int n, const char *aString)
{
  ptrepstr vstr;

  makeRepstr_string_int(aString, n, &vstr);
  return LL_make_nom_atom(vstr);
}

/**
   @return the tree of the identifier Num= n.
*/
ntrees DCG2SignalTree::mknumtree(int n)
{
  return mknumtreewithprefix(n, " Num= ");
}

/**
   It finalizes the translation of the IGraph2signal functionality by modifying the resulting tree.
   The tree is modified when the "flat mode" is set. In this case, the compositions of the model are flatenned.
   @return the possibly modified operand tree (vt).
*/
ntrees DCG2SignalTree::DCG2SignalTreeFinal(ntrees vt)
{
  cleanTree(vt);
  return vt;
}

/**
   It adds, as assertions, to the model (process) the bounds informations of the output signals of the graph.
*/
_VIRTUAL_ void DCG2SignalTree::AssertionValue(ntrees process)
{
  ptlinstances ls;
  ptexpcal vexp;
  ntrees t1;

  ls = AGRAPH->getOutputs();
  while (ptlinstances_get_next(&ls, &vexp))
    if (getInfoValue(vexp, &t1))
      STree_addSpecification(process, LL_make_appeassert_tree(t1));
}

/** 
    @return "true" if the graph contains train clusters (trainclusters), "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::getTrainList(ptlINode * trainclusters)
{
  ptlINode l;
  ptINode n;

  (*trainclusters) = NILLINODE;
  if (!NM->getSubGraphs(&l))
    return pk_false;
  while (ptlINode_get_elem(&l, &n))
    if (NM->is_trainCluster(n))
      ptlINode_insert_elem(trainclusters, n);
  return ((*trainclusters) != NILLINODE);
}
