/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
*/
  DCG2Automaton::DCG2Automaton ()
{
  BDEBUG = pk_false;
  BTRACE = pk_true;
  UserSelectedSignals = NILLINSTANCES;
  aDuplicator = new Duplicator ();
  aProfileBuilder = new ProfileBuilder ();
  cpt = 0;
  ls_invisible_states = NILLINSTANCES;
  vautomaton = new Automaton ();	// result.
  def_managers (vautomaton);
  ATTR_LINKGRAPHS = getMark ();
}

/**
   Destructor.
*/
DCG2Automaton::~DCG2Automaton ()
{
}

/**
   The resulting automaton is assigned to the graph (setAutomaton()).
   Use getAutomaton() method to get it.
*/
ptAutomaton
DCG2Automaton::build (ptDCG_poly og, ptlinstances lSelectedStateVar)
{
  ptlStates lStates;
  ptlTransitions lTransitions;
  int nbstates, vtime, beginAt;

  beginAt = sysgetcputime ();

  init (og);
  UserSelectedSignals = lSelectedStateVar;
  nbvar = ptlinstances_get_size (lSelectedStateVar);
  buildAutomaton (lSelectedStateVar);
  og->setAutomaton (vautomaton);

  BTRACE_LEVEL0 (printf ("\n\n--------------------------------------------------------------------\n");
		 fflush (stdout));
  printf ("     o Nb State variables = %d\n", ptlinstances_get_size (UserSelectedSignals));
  lStates = vautomaton->getStates ();
  nbstates = ptlPointer_get_size (lStates);
  BTRACE_LEVEL0 (printf ("     o Automaton, Nb States = %d\n", nbstates);
		 fflush (stdout));
  lTransitions = vautomaton->getTransitions ();
  DEBUG_print_transitions (lTransitions);
  BTRACE_LEVEL0 (printf ("     o Automaton, Nb Transitions = %d  (max = %d)\n", ptlPointer_get_size (lTransitions), (int) pow ((float) nbstates, 2));
		 fflush (stdout));
  vtime = ((int) sysgetcputime () - beginAt);
  BTRACE_LEVEL0 (pK_messenger->printTime (vtime);
		 fflush (stdout));
  BTRACE_LEVEL0 (printf ("\n--------------------------------------------------------------------\n");
		 fflush (stdout));


  //  ntrees_printTreeMemoryState();

  return vautomaton;
}
