/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PRIVATE_
#include "math.h"
#define BTRACE_LEVEL0(text) text
#define BTRACE_LEVEL1(text) text
// #define BTRACE_LEVEL1(text) 
#define VSUFF_AUTO "AUTO"
#define updateClocks(g) g->ClockCalculus();
#define TestSameObjects(obj1, obj2) obj1==obj2
#define decl_managers(g) ptExprManager EM_##g; ptClockManager CKM_##g; ptNodeManager NM_##g; ptNameManager NameManager_##g;
#define def_managers(g)  EM_##g = g->getExpressionManager();CKM_##g = g->getClockManager();NM_##g = g->getNodeManager();NameManager_##g = g->getNameManager();
#define require_managers(g) decl_managers(g) def_managers(g)
#define getStateName(g) g->getName()
#define getLabelStateName(g) g->getLabelStateName()
#define mkRefToEnumValue(EM, g) EM->mkBinaryExp(opvenum, vautomaton->getExpressionManager() ->mkExpType(TypeStateEnum), EM -> mk_enumPtexpcal(ident2repstr(getLabelStateName(g))), NILBACKANNOTATION)
#define mkRefToEnumValueId(EM, id) EM->mkBinaryExp(opvenum, vautomaton->getExpressionManager()->mkExpType(TypeStateEnum), EM->mk_enumPtexpcal(ident2repstr(id)), NILBACKANNOTATION);
#define GET_ATTR_LINKGRAPHS(MANAGER, obj1, vtable) (MANAGER->getAttribute(obj1, ATTR_LINKGRAPHS, (attributeValue *) &vtable))
#define SET_ATTR_LINKGRAPHS(MANAGER, obj1, vtable) (MANAGER->addAttribute (obj1, ATTR_LINKGRAPHS, (attributeValue) vtable))
#define GET_TRANSITIONS_FROM(vTransitionsTable, S1, lts, it)	\
  it = vTransitionsTable->find( S1 );\
  if (it == vTransitionsTable->end()) lts = NILLPOINTER; else lts = (ptlTransitions) (*it).second;
/**
   Initialize the translator for a graph (g).
*/
  void
DCG2Automaton::init (ptDCG_poly g)
{
  G_ORI = g;
  def_managers (G_ORI);
  NL = NameListCreate ();
  NL = NameListRegister (NL);
  TypeStateEnum = EnumerationTypeCreate (string2repstr ("StateType"), NL);
  NameListAppendName (NL, (NameClass) ident2repstr (G_ORI->getName ()));
  outputs_ORI = G_ORI->getOutputs ();
  inputs_ORI = G_ORI->getOutputs ();
  TheNullClock = CKM_G_ORI->getNullClock ();
}

/**
   Garbaging a map(graph, pointer list).
*/
void
DCG2Automaton::free (ptTablePointerPointer vtable)
{
  TablePointerPointerType::iterator it;
  for (it = vtable->begin (); it != vtable->end (); it++)
    {
      ptlPointer_free ((ptlTransitions) (*it).second);
    }
  delete vtable;
}

/**
   @return the number of elements of second for all element of a (first : graph,  second: pointer list).
*/
int
DCG2Automaton::get_size (ptTablePointerPointer vtable)
{
  int n;
  TablePointerPointerType::iterator it;

  n = 0;
  for (it = vtable->begin (); it != vtable->end (); it++)
    {
      n = n + ptlPointer_get_size ((ptlTransitions) (*it).second);
    }
  return n;
}

/**
   @return "true" if a signal (vsig), of the expression Manager EM,  is a clone of an another one (secondsig) of an another graph, "false" otherwise.
*/
boolean
DCG2Automaton::getLinkObject2Graphs (ptExprManager EM, ptexpcal obj1, ptDCG g, ptexpcal * obj2)
{
  ptTablePointerPointer vtable;
  TablePointerPointerType::iterator it;
  boolean b;

  if (!GET_ATTR_LINKGRAPHS (EM, obj1, vtable))
    return pk_false;
  it = vtable->find (g);
  if (it != vtable->end ())
    {
      (*obj2) = (ptexpcal) (*it).second;
      return pk_true;
    }
  b = pk_false;
  for (it = vtable->begin (); !b && it != vtable->end (); it++)
    b = getLinkObject2Graphs (EM, (ptexpcal) (*it).second, g, obj2);

  if (!b)
    return pk_false;
  (*vtable)[g] = *obj2;
  return pk_true;
}

/**
   @return "true" if a clock (o1) has a clone (o2) in an another graph (g), "false" otherwise.
*/
boolean DCG2Automaton::getLinkObject2Graphs (ptClockManager CKM, ptinsth o1, ptDCG g, ptinsth * o2)
{
  ptTablePointerPointer
    vtable;
  TablePointerPointerType::iterator it;

  if (!GET_ATTR_LINKGRAPHS (CKM, o1, vtable))
    return pk_false;
  it = vtable->find (g);
  if (it == vtable->end ())
    return pk_false;
  (*o2) = (ptinsth) (*it).second;
  return pk_true;
}

/**
   It keeps the link between two clocks of two graphs (by morphisms).
   o1 is a clock of this clockmanager, o2 is a clock of an another graph (g).
   If o2 is undefined, the link is broken.
*/
void
DCG2Automaton::setLinkObject2Graphs (ptClockManager CKM, ptinsth o1, ptinsth o2, pointer g)
{
  ptTablePointerPointer vtable;

  if (!GET_ATTR_LINKGRAPHS (CKM, o1, vtable))
    {
      vtable = new TablePointerPointerType ();
      (*vtable)[g] = o2;
      SET_ATTR_LINKGRAPHS (CKM, o1, vtable);
    }
  else
    (*vtable)[g] = o2;
}

/**
   It sets the link between two signals of two different graphs (by morphisms): 
   firstsig is an object of the Expression Manager EM, secondsig a signal of g.
   If secondsig is undefined, the link (firstsig->g) is broken.
*/
void
DCG2Automaton::setLinkObject2Graphs (ptExprManager EM, ptexpcal firstsig, ptexpcal secondsig, ptDCG g)
{
  ptTablePointerPointer vtable;

  if (!GET_ATTR_LINKGRAPHS (EM, firstsig, vtable))
    {
      vtable = new TablePointerPointerType ();
      (*vtable)[g] = secondsig;
      SET_ATTR_LINKGRAPHS (EM, firstsig, vtable);
    }
  else
    (*vtable)[g] = secondsig;
}

/**
   Keeping the links between objects of two graphs: new_DCG is assumed to be a copy of ori_DCG: signals/clocks of ori_DCG 
   have been attributed by the duplicator usign the mark vmark.
   This method adds for x a clock/signal of ori_DCG the link (x --> (new_DCG, x')).
   This link is useful to obtain the clone of an object in a graph (get
*/
void
DCG2Automaton::setGraphLinks (ptDCG ori_DCG, ptDCG new_DCG, typeMark vmark)
{
  ptExprManager EM;
  ptClockManager CKM;
  ptlinstances lsig;
  ptexpcal vsig, vsig_clone;
  ptinsth h, h_clone;
  ptlinsth lh, lh_aux;

  // The signals 
  EM = ori_DCG->getExpressionManager ();
  if (EM->getExprList (opident, &lsig))
    while (ptlinstances_get_next (&lsig, &vsig))
      if (EM->getAttribute (vsig, vmark, (attributeValue *) & vsig_clone))
	setLinkObject2Graphs (EM, vsig, vsig_clone, new_DCG);
  // The clocks 
  CKM = ori_DCG->getClockManager ();
  lh_aux = CKM->getClocks ();
  lh = lh_aux;
  while (ptlinsth_get_next (&lh, &h))
    if (CKM->getAttribute (h, vmark, (attributeValue *) & h_clone))
      setLinkObject2Graphs (CKM, h, h_clone, new_DCG);
  ptlinsth_free (lh_aux);
}

void
DCG2Automaton::updatelinks (ptDCG g1, ptDCG g2)
{
  ptlinstances ls;
  ptexpcal vsig, vclone;

  require_managers (g1);
  if (EM_g1->getExprList (opident, &ls))
    while (ptlinstances_get_next (&ls, &vsig))
      getLinkObject2Graphs (EM_g1, vsig, g2, &vclone);

}


/**
   The variables (EM, CKM, NM) of the translator are set to the managers of a graph (g).
*/
void
DCG2Automaton::setCurrentGraph (ptDCG_poly g)
{
  EM = g->getExpressionManager ();
  CKM = g->getClockManager ();
  NM = g->getNodeManager ();
}

/**
   @return "true" if a signal (vsig) has a clone (returned in vclone) in a graph (g), "false" otherwise.
   if g is the orignal graph, clone(vsig) = vsig.
*/
boolean DCG2Automaton::getCloneSignal (ptexpcal vsig, ptDCG_poly g, ptexpcal * vclone)
{
  if (g == G_ORI)
    {
      *vclone = vsig;
      return pk_true;
    }
  if (getLinkObject2Graphs (g->getExpressionManager (), vsig, g->getDCG (), vclone))
    return pk_true;
  printf ("Error: Signal not found in a state !!! \n");
  return pk_false;
}

/**
   @return "true" if a clock (vclk) has a clone (returned in vclone) in a graph (g), "false" otherwise.
   if g is the orignal graph, clone(vclk) = vclk
*/
boolean
DCG2Automaton::getCloneClock (ptinsth vclk, ptDCG_poly g, ptinsth * vclone)
{
  ptClockManager CKM = g->getClockManager ();
  if (g == G_ORI)
    {
      *vclone = vclk;
      return pk_true;
    }
  if (getLinkObject2Graphs (CKM, vclk, g->getDCG (), vclone))
    return pk_true;
  printf ("  *** Warning: clock not found in a state !!! \n");
  *vclone = vclk;
  return pk_true;
}


/**
   @return the coding of the initial value of a signal defined by a delay (vDelay).
   If the initial value of vDelay is pk_true(resp. pk_false), it returns the value of the enumerated type associated with the gT (resp. gF) graph,
*/
ptexpcal
DCG2Automaton::mkInitValue (ptDCG_poly gT, ptDCG_poly gF, ptexpcal vDelay)
{
  ptexpcal vinit;
  ptident id;

  vinit = EM->getInitialValue (vDelay);
  if (EM->eq_trueExpcal (vinit))
    id = getStateName (gT);
  else
    id = getStateName (gF);
  return mkRefToEnumValueId (EM, id);
}

/**
  It adds the Valuation pragma to a graph(g) applied to an object (id) and with vstrvalue as statement.  
*/
void
DCG2Automaton::addValuationPragma (ptDCG_poly g, ptident id, ptrepstr vstrvalue)
{
  ptObjPragma obj = NILOBJPRAGMA;
  ntrees stmt = NILENONCEPRAGMA;
  ptlinstances lsig = NILLINSTANCES;
  ptpragma vpragma;
  ptrepstr vstrq;

  getDoubleQuotedString (ident2repstr (id), &vstrq);
  stmt = addEnonceValue (stmt, vstrq);
  getDoubleQuotedString (vstrvalue, &vstrq);
  stmt = addEnonceValue (stmt, vstrq);
  vpragma = newpragma (string2repstr ("Valuation"), (ptObjPragma) NILLINSTANCES, stmt);
  g->addPragma (vpragma);
}

/**
   In debug mode, it unparses the graph(gbasic).
*/
void
DCG2Automaton::printBasic (ptDCG_basic gbasic)
{
  ptident id;

  if (BDEBUG)
    {
      id = gbasic->getName ();
      gbasic->setName (makeIdent_ident_repstr (id, integer2repstr (cpt)));
      ptDCG_poly gp = DCG_poly::new_prePoly (gbasic->getDCG ());
      gp->setIntrinsicClocks ();
      gp->unparse ("beforeValuation");
      gp->initGraphForModifications ();
      gbasic->setName (id);
    }
}


void
DCG2Automaton::changeName (ptDCG g, ptlinstances ls)
{
  ptexpcal vsig;
  ptinsth vh;
  ptident id;
  ptExprManager EM = g->getExpressionManager ();
  ptClockManager CKM = g->getClockManager ();
  ptNameManager aNameManager = vautomaton->getNameManager ();

  while (ptlinstances_get_next (&ls, &vsig))
    if (!EM->IsInputFlow (vsig))
      {
	id = aNameManager->mkIdentName (EM->getOriginalName (vsig));
	EM->setName (vsig, id);
	if (EM->isSignalUnifiedWithClock (vsig, &vh))
	  CKM->setName (vh, id);
      }
}

/**
   @return a copy of a graph(g), and the clone (nvsig) of vsig (a signal of g), in the copy.
*/
ptDCG_basic
DCG2Automaton::mkCopy (ptDCG_poly g, ptexpcal vsig, ptexpcal * nvsig)
{
  ptDCG_basic vg_basic;
  ptDCG ori_DCG, new_DCG;
  typeMark vmark;

  ori_DCG = g->getDCG ();
  new_DCG = aDuplicator->copyWithoutHierarchy (ori_DCG);
  changeName (new_DCG, new_DCG->getOutputs ());
  vmark = aDuplicator->getUsedMark ();
  g->getExpressionManager ()->getAttribute (vsig, vmark, (attributeValue *) nvsig);
  vg_basic = new DCG_basic (new_DCG, pk_false, pk_true);
  printBasic (vg_basic);
  // vg_basic-> setName(makeIdent_ident_int(g->getName(), cpt )); // MODIF 11 Mars
  vg_basic->setName (makeIdent_ident_int (G_ORI->getName (), cpt));	// MODIF 11 Mars
  setGraphLinks (ori_DCG, new_DCG, vmark);
  return vg_basic;
}


void
DCG2Automaton::removeSig (ptDCG_basic vg_basic, ptexpcal nvsig)
{
  ptExprManager EM;

  //  A complÃšter : vÃ©rifier que le noeud n'a pas de sucesseurs....
  EM = vg_basic->getExpressionManager ();
  if (EM->mayBeRemoved (nvsig))
    vg_basic->getDCG ()->removeNode (EM->getNode (nvsig));
}


/**
   @return a copy of a graph(g) in which a boolean signal (vsig) is valuated to bsigValue.
   vsig is a signal of g.
*/
ptDCG_poly
DCG2Automaton::mkDCGPolyBooleanValuation (ptDCG_poly g, ptexpcal vsig, boolean bsigValue)
{
  ptExprManager EM_COPY;
  ptNodeManager NM_COPY;
  ptClockManager CKM_COPY;
  ptexpcal value;
  ptrepstr vstrvalue;
  ptexpcal nvsig;
  ptDCG_basic vg_basic;
  ptDCG_poly vg_poly;
  boolean bm;

  cpt++;
  vg_basic = mkCopy (g, vsig, &nvsig);

  EM_COPY = vg_basic->getExpressionManager ();
  NM_COPY = vg_basic->getNodeManager ();
  CKM_COPY = vg_basic->getClockManager ();

  if (bsigValue)
    value = EM_COPY->getTrue ();
  else
    value = EM_COPY->getFalse ();
  NM_COPY->modifiedDefinition (nvsig, value, CKM_COPY->getInitClock (), EM_COPY->getClock (nvsig));
  EM_COPY->replace (nvsig, value);
  EM_COPY->rewritingExprEH ();
  EM_COPY->substituteEvaluatedExpression ();
  removeSig (vg_basic, nvsig);
  // vg_basic-> eliminateWhenEvent();  It must not be applied
  // vg_basic-> transformBooleanClocks2Events(); It must not be applied

  bm = pK_messenger->setSilentModeMessagesTo (pk_true);
  vg_poly = DCG_poly::DCG_basic2poly (vg_basic);
  vg_poly->setIntrinsicClocks ();
  vg_poly->delayTreeBuild ();

  vstrvalue = EM_COPY->getConstant (value);
  addValuationPragma (vg_poly, EM_COPY->getName (nvsig), vstrvalue);
  bm = pK_messenger->setSilentModeMessagesTo (bm);
  if (g != G_ORI)
    updatelinks (G_ORI->getDCG (), vg_poly->getDCG ());
  return vg_poly;
}

/**
   @return the undersampling [NVSIG] and hnosub=[not NVSIG], where C is the clone of vsig in g.
*/
ptinsth
DCG2Automaton::mkExtractions (ptDCG_poly g, ptexpcal vsig, ptinsth * hnosub)
{
  ptExprManager EM = g->getExpressionManager ();
  ptClockManager CKM = g->getClockManager ();
  ptexpcal vcloneSig;
  ptinsth hCloneSig, hsub;

  getCloneSignal (vsig, g, &vcloneSig);
  hCloneSig = EM->getClock (vcloneSig);
  /* if (EM->eq_trueExpcal(vcloneSig))  
     {
     *hnosub = TheNullClock;
     return hCloneSig;
     }
     if (EM->eq_falseExpcal(vcloneSig))
     {
     *hnosub = hCloneSig;
     return  TheNullClock;
     } */
  hsub = CKM->makeSub (hCloneSig, vcloneSig, ptlinstances_insert (NILLINSTANCES, vcloneSig));
  *hnosub = CKM->makeNoSub (vcloneSig);
  return hsub;
}

/**
   @return "true" if it exists a transition (vtransition) such that g-[]->g in a list of transitions (lTransitions), "false" otherwise.
*/
boolean
DCG2Automaton::getNoEvolutionTransition (ptTransitionsTable vTransitionsTable, State g, ptTransition * vtransition)
{
  pointer velem;
  State s1, s2;
  ptinsth vh;
  ptlTransitions lTransitions;
  TablePointerPointerType::iterator it;

  GET_TRANSITIONS_FROM (vTransitionsTable, g, lTransitions, it);
  while (ptlPointer_get_elemNext (&lTransitions, &velem))
    {
      *vtransition = (ptTransition) velem;
      (*vtransition)->get (&s1, &s2, &vh);
      if ((s1 == s2) && (s1 == g))
	return pk_true;
    }
  return pk_false;
}

/**
  Getting a transition from a state (g).

  @return true if it exists a transition g-[vclkTransition]->gTo in a list of transitions (lTransitions), "false" otherwise.
  vclkTransition, gTo, lTransitions are modified.
*/
boolean
DCG2Automaton::getTransitionFrom (ptlTransitions * lTransitions, State g, State * gTo, ptinsth * vclkTransition, ptTransition * vtransition)
{
  pointer velem;
  State s1;

  while (ptlPointer_get_elemNext (lTransitions, &velem))
    {
      *vtransition = (ptTransition) velem;
      (*vtransition)->get (&s1, gTo, vclkTransition);
      if ((s1 != *gTo) /* && (s1 == g) */ )
	return pk_true;
    }
  return pk_false;
}


/**
   Getting a transition from a state (g) to an another state (gTo).

   @return "true" if it exists a transition g-[vclkTransition]->gTo in a list of transitions (lTransitions), "false" otherwise.
   vclkTransition, gTo, lTransitions are modified.
*/
boolean
DCG2Automaton::getTransitionFromTo (ptlTransitions lTransitions, State g, State gTo, ptTransition * vtransition)
{
  pointer velem;
  State s2, s1;
  ptinsth vclkTransition;

  while (ptlPointer_get_elemNext (&lTransitions, &velem))
    {
      *vtransition = (ptTransition) velem;
      (*vtransition)->get (&s1, &s2, &vclkTransition);
      if ((s1 == g) && (s2 == gTo))
	return pk_true;
    }
  return pk_false;
}

/**
   @return true if it exists a transition s1-[ ]-> aState and (s1!=aState), false otherwise.
*/
boolean
DCG2Automaton::existTransitionTo (State aState, ptTransitionsTable vTransitionsTable)
{
  pointer velem;
  State s2, s1;
  ptinsth vclkTransition;
  ptTransition vtransition;
  ptlTransitions ltransitions;
  TablePointerPointerType::iterator it;

  for (it = vTransitionsTable->begin (); it != vTransitionsTable->end (); it++)
    {
      ltransitions = (ptlTransitions) (*it).second;
      while (ptlPointer_get_elemNext (&ltransitions, &velem))
	{
	  vtransition = (ptTransition) velem;
	  vtransition->get (&s1, &s2, &vclkTransition);
	  if ((s1 != aState) && (s2 == aState))
	    return pk_true;
	}
    }
  return pk_false;
}


/**
   It removes the unreachable states of lStates and the transitions of lTransitions with null clock as trigger.
*/
void
DCG2Automaton::rmUnreachableStates (ptlStates * lStates, ptTransitionsTable vTransitionsTable)
{
  pointer velem;
  State s1, s2;
  ptTransition vtransition;
  ptinsth vclkTransition;
  ptlTransitions lTransitions, lts;
  ptDCG_poly g;
  ptClockManager CKM;

  TablePointerPointerType::iterator it;
  TablePointerInteger::iterator it1;

  BTRACE_LEVEL1 (printf ("         rmUnreachableStates::Step1, |Transitions| = %d\n", get_size (vTransitionsTable));
		 fflush (stdout);
    );

  for (it = vTransitionsTable->begin (); it != vTransitionsTable->end (); it++)
    {
      lts = (ptlTransitions) (*it).second;
      lTransitions = lts;

      while (ptlPointer_get_elemNext (&lts, &velem))
	{
	  vtransition = (ptTransition) velem;
	  vtransition->get (&s1, &s2, &vclkTransition);
	  g = (ptDCG_poly) s1;
	  CKM = g->getClockManager ();
	  vclkTransition = CKM->getRepresentativeFormula (vclkTransition);
	  vtransition->setClock (vclkTransition);
	  if (CKM->eqClock (vclkTransition, TheNullClock))
	    {
	      BTRACE_LEVEL1 (if (BTRACE) printf (" >>>>>>>>>>>>>>> Transition with null clock!! <<<<<<<<<<<<<< \n"));
	      ptlPointer_remove_elem (&lTransitions, velem);
	      (*it).second = lTransitions;
	    }
	}
    }
  BTRACE_LEVEL1 (printf ("         rmUnreachableStates::Step2\n");
		 fflush (stdout);
    );

  for (it1 = nbpreds.begin (); it1 != nbpreds.end (); it1++)
    {
      if ((*it1).second == 0)
	{
	  BTRACE_LEVEL1 (printf (" >>>>>>>>>>>>>>> Removing an Unreachable State: %s\n", ident2string (getStateName (g)));
	    );
	  ptlPointer_remove_elem (lStates, velem);
	  (*it1).second = -1;
	}
    }

  BTRACE_LEVEL1 (printf ("         rmUnreachableStates::END\n");
		 fflush (stdout);
    );
}

void
DCG2Automaton::addTransitionFrom (ptTransitionsTable vTransitionsTable, State S, ptTransition vTransition)
{
  TablePointerPointerType::iterator it;
  ptlTransitions lts;

  GET_TRANSITIONS_FROM (vTransitionsTable, S, lts, it);
  lts = ptlPointer_add (lts, vTransition);
  (*vTransitionsTable)[S] = lts;
}


void
DCG2Automaton::addTransition (ptTransitionsTable vTransitionsTable, State S1, State S2, ptinsth vClkTransition)
{
  ptDCG_poly G1, G2;

  if (!CKM->eqClock (vClkTransition, TheNullClock))
    {
      G1 = (ptDCG_poly) S1;
      G2 = (ptDCG_poly) S2;
      addTransitionFrom (vTransitionsTable, S1, new Transition (S1, S2, vClkTransition));
      if (S1 != S2)
	nbpreds[S2] = nbpreds[S2] + 1;
    }
}

/**
   @return the delayed signal of vsig.
   Assumed: vsig is defined by a delay node.
*/
ptexpcal
DCG2Automaton::getDelayedSignal (ptexpcal vsig)
{
  ptexpcal x, vDelayedSig, vret, vfen, vinit;

  NM->get_delay_arg (EM->getNode (vsig), &x, &vDelayedSig, &vret, &vfen, &vinit);
  return vDelayedSig;
}

/**
   @return the list of transition  NextTransitions in which two added transitions:
   - S1->[ h1 ]->S1 with h1 =  (step(S1) ^- vClkS1S2) ^*  vclkNoEvol
   - S1->[ h2 ]->S2 with h2 =  vClkS1S2 ^*  vclkNoEvol
*/
void
DCG2Automaton::mkTransitionsSubStateFrom (ptTransitionsTable NextTransitions, ptDCG_poly S1, ptDCG_poly S2, ptinsth vClkS1S2, ptinsth vclkNoEvol,
					  ptinsth * h_StayIn_S1, ptinsth * h_NoStayIn_S1)
{
  ptinsth h1_step, h1;
  ptClockManager CKM_S1;

  CKM_S1 = S1->getClockManager ();
  getCloneClock (vclkNoEvol, S1, &h1);
  addTransition (NextTransitions, S1, S1, CKM_S1->makeMinus (h1, vClkS1S2));
  addTransition (NextTransitions, S1, S2, CKM_S1->makeInf (h1, vClkS1S2));
  h1_step = S1->getStep ();
  *h_StayIn_S1 = CKM_S1->makeMinus (h1_step, vClkS1S2);
  *h_NoStayIn_S1 = vClkS1S2;
  return;

  /* ptinsth h1_step;
     CKM_S1 = S1->getClockManager( );
     getCloneClock(vclkNoEvol, S1, &h1);
     h1_step = S1->getStep( );
     *h_StayIn_S1 = CKM_S1->makeMinus(h1_step, vClkS1S2);
     *h_NoStayIn_S1 = CKM_S1->makeInf(h1, vClkS1S2);
     addTransition(NextTransitions, S1, S1, CKM_S1->makeInf(*h_StayIn_S1, h1));
     addTransition(NextTransitions, S1, S2, *h_NoStayIn_S1 );
   */
}

/**
   Invisible state management: it declares in g a signal (vInvState_in_g) used to manage a invisible state (when a boolean delay is not present in a graph), 
   and a clone in g11 (vInvState_in_g11), and in g22 (vInvState_in_g12). Th signals are initialized to vinit.
*/
void
DCG2Automaton::declareInvisibleState (ptDCG_poly g, ptDCG_poly g11, ptDCG_poly g12, ptexpcal vinit, ptexpcal * vInvState_in_g,
				      ptexpcal * vInvState_in_g11, ptexpcal * vInvState_in_g12)
{
  ptident id;
  require_managers (g);
  require_managers (g11);
  require_managers (g12);

  BTRACE_LEVEL1 (printf (" >>>>>>>>>>>>>>> INVISIBLE STATE...\n");
    );
  id = string2ident ("HIST");	// History of a set of states.
  *vInvState_in_g = EM_g->mkIdentExp (id, tsignal, TypeStateEnum, vinit, NM_g->mk_node_undefined (), CKM_g->getUndefClock (), NILBACKANNOTATION);
  EM_g->setStateTag (*vInvState_in_g, pk_true);
  EM_g->setSharedSignalBySeveralGraphs (*vInvState_in_g, pk_true);
  ls_invisible_states = ptlinstances_insert (ls_invisible_states, *vInvState_in_g);

  // a clone of vInvState_in_g in g11 (out)
  *vInvState_in_g11 = EM_g11->mkIdentCopy (*vInvState_in_g, NM_g11->mk_node_undefined (), CKM_g11->getUndefClock ());
  EM_g11->setStateTag (*vInvState_in_g11, pk_true);
  EM_g11->setSharedSignalBySeveralGraphs (*vInvState_in_g11, pk_true);

  aProfileBuilder->becomesInputOutput (g11->getDCG (), *vInvState_in_g11);

  // a clone of vInvState_in_g in g12 (out)
  *vInvState_in_g12 = EM_g12->mkIdentCopy (*vInvState_in_g, NM_g12->mk_node_undefined (), CKM_g12->getUndefClock ());
  EM_g12->setStateTag (*vInvState_in_g12, pk_true);
  EM_g12->setSharedSignalBySeveralGraphs (*vInvState_in_g12, pk_true);

  aProfileBuilder->becomesInputOutput (g12->getDCG (), *vInvState_in_g12);
}

/**
   Invisible state management: it defines a signel (invState) by the equation invState := enumstate(g) when vclk, where enumstate(g) is the enum value (See TypeStateEnum type) that represents the state g.
*/
void
DCG2Automaton::mkInvisibleStateDef (ptexpcal invState, ptDCG_poly g, ptinsth vclk)
{
  ptident id;
  require_managers (g);
  ptexpcal e1;
  ptINode n;

  id = getStateName (g);
  e1 = mkRefToEnumValueId (EM_g, id);
  n = NM_g->insertEquation (invState, e1, vclk, vclk);
}

/*
  Invisible state management.
  @return the clock when[? vsig == TypeStateEnum.id], where
  - id is a value of the enumerated type that represents a state of the automaton 
  - vsig is a state variable.
*/
ptinsth
DCG2Automaton::makeWhenClockwhen (ptDCG_poly g, ptexpcal vsig, ptident id)
{
  ptexpcal e1, e2, e;
  require_managers (g);

  e1 = mkRefToEnumValueId (EM_g, id);
  e2 = EM_g->mkUnaryExp (opvvarin, vsig, NILBACKANNOTATION);
  e = EM_g->mkBinaryExp (opegal, e2, e1, NILBACKANNOTATION);
  return CKM_g->makeSub (g->getStep (), e, NILLINSTANCES);
}


/**
   Invisible state management.
   @return the clock (k) that must be used to modify the clock of the transition g -[h ^* k]->gTo to solve the non determinism induced by a delayed signal not defined in a graph(g).
*/
ptinsth
DCG2Automaton::mkClockReturnIn (ptDCG_poly g, ptexpcal invState_g, ptDCG_poly gTo, ptexpcal invState_gTo)
{
  ptinsth was_in_gTo;

  was_in_gTo = gTo->getStep ();
  mkInvisibleStateDef (invState_gTo, gTo, was_in_gTo);
  return makeWhenClockwhen (g, invState_g, getStateName (gTo));
}

/**
   @return the transitions of the automaton with NextStates as set of states, derived from the current automaton (lcurrentStates, lcurrentTransitions) using a boolan signal (vDelayedSig).

   Nextstates are the states derived from lcurrentStates by valuations (true/false) of a boolean delay zc := vDelayedSig $1.
   For S in lcurrentStates, it has been produced two states (St = S[zc=true], Sf = S[zc=false]) in NextStates.

   - the transition S-[HSS]->S is replaced by 4 transitions (STEP specifies the fastest clock of the program)
          - St-[h]->St, with h = HSS ^* (STEP ^-[not vDelayedSig])
	  - Sf-[h]->Sf, with h = HSS ^* (STEP ^- [vDelayedSig],)
	  - St-[h]->Sf  with h = HSS ^* [not vDelayedSig]
	  - Sf-[h]->St  with h = HSS ^* [vDelayedSig]
   - a transition S-[HSS']->S' (S!=S') is replaced by 4 transitions
          - St-[h]->St', with h = (STEP ^-[not vDelayedSig]) ^* HSS'
	  - St-[h]->Sf', with h = [not vDelayedSig] ^* HSS'
	  - Sf-[h]->St', with h = [vDelayedSig] ^* HSS'
	  - Sf-[h]->Sf', with h = (STEP ^-[vDelayedSig]) ^* HSS'
   
*/
ptTransitionsTable
DCG2Automaton::mkTransitions (ptTransitionsTable lcurrentTransitions, ptexpcal vsig, ptlStates lcurrentStates, ptlStates NextStates)
{
  ptlTransitions lts;
  ptTransitionsTable NextTransitions;
  pointer elem;
  ptClockManager CKM_T, CKM_F;
  ptDCG_poly g, vg_polyT, vg_polyF, g1, g11, g12;
  ptinsth h_NoStayIn_FF, h_NoStayIn_TT, h_StayIn_FF, h_StayIn_TT, h1, h_TT, h_FF, h_TF, h_FT, vclkNoEvol, vclkTransition;
  ptexpcal vinit, invState_g, invState_g11, invState_g12, vDelayedSig;
  ptTransition vTransition, vNoEvolTransition;
  boolean b;
  ptlStates lst;
  TablePointerPointerType::iterator it;
  int vtime, beginAt;

  BTRACE_LEVEL1 (printf ("          ::mkTransitions--(Step1) = |Transitions| = %d", get_size (lcurrentTransitions));
		 fflush (stdout));

  beginAt = (int) sysgetcputime ();
  NextTransitions = new TablePointerPointerType ();
  // First step, consider only the graphs, in which the signal vsig is present. 
  // (ie g is rewritten in two graphs vg_polyT, &vg_polyF by the partitionning vsig=TRUE, vsig=FALSE)
  lst = lcurrentStates;
  while (ptlPointer_get_elemNext (&lst, &elem))
    {
      g = (ptDCG_poly) elem;
      setCurrentGraph (g);
      getNoEvolutionTransition (lcurrentTransitions, g, &vNoEvolTransition);

      if (getAttribute (g, &vg_polyT, &vg_polyF))
	{
	  // the signals, vsig and vDelayedSig, are present in g.
	  vclkNoEvol = vNoEvolTransition->getClock ();	// g -[vclkNoEvol]->g 
	  vDelayedSig = getDelayedSignal (vsig);
	  CKM_T = vg_polyT->getClockManager ();
	  CKM_F = vg_polyF->getClockManager ();

	  h_TT = mkExtractions (vg_polyT, vDelayedSig, &h_TF);
	  mkTransitionsSubStateFrom (NextTransitions, vg_polyT, vg_polyF, h_TF, vclkNoEvol, &h_StayIn_TT, &h_NoStayIn_TT);

	  h_FT = mkExtractions (vg_polyF, vDelayedSig, &h_FF);
	  mkTransitionsSubStateFrom (NextTransitions, vg_polyF, vg_polyT, h_FT, vclkNoEvol, &h_StayIn_FF, &h_NoStayIn_FF);

	  // lts = lcurrentTransitions;
	  GET_TRANSITIONS_FROM (lcurrentTransitions, g, lts, it)
	    while (getTransitionFrom (&lts, (State) g, (State *) & g1, &vclkTransition, &vTransition))
	    {
	      // g -[vclkTransition]->g1 and (g1 != g)
	      if (getAttribute (g1, &g11, &g12))
		{
		  // the signals, vsig and vDelayedSig, are present in g1.
		  // Building the induced transitions: vg_polyT -[]->g11, vg_polyT-[]->g12, vg_polyF -[]->g11, vg_polyF-[]->g12.
		  getCloneClock (vclkTransition, vg_polyT, &h1);
		  addTransition (NextTransitions, vg_polyT, g11, CKM_T->makeInf (h1, h_StayIn_TT));
		  addTransition (NextTransitions, vg_polyT, g12, CKM_T->makeInf (h1, h_NoStayIn_TT /*h_TF */ ));

		  getCloneClock (vclkTransition, vg_polyF, &h1);
		  addTransition (NextTransitions, vg_polyF, g11, CKM_F->makeInf (h1, h_NoStayIn_FF /*h_FT */ ));
		  addTransition (NextTransitions, vg_polyF, g12, CKM_F->makeInf (h1, h_StayIn_FF));
		}
	      else
		{
		  // the signals, vsig and vDelayedSig, are absent in g1.
		  // induced transitions: vg_polyT -[]->g1, vg_polyF -[]->g1
		  getCloneClock (vclkTransition, vg_polyT, &h1);
		  addTransition (NextTransitions, vg_polyT, g1, h1);
		  getCloneClock (vclkTransition, vg_polyF, &h1);
		  addTransition (NextTransitions, vg_polyF, g1, h1);
		}
	    }
	  updateClocks (vg_polyT);
	  updateClocks (vg_polyF);
	}
    }
  BTRACE_LEVEL1 (printf ("          :: mkTransitions--(Step2) ");
		 fflush (stdout));
  // Second step, consider the graphs in which the signal vsig is absent.
  lst = lcurrentStates;
  while (ptlPointer_get_elemNext (&lst, &elem))
    {
      g = (ptDCG_poly) elem;
      setCurrentGraph (g);
      getNoEvolutionTransition (lcurrentTransitions, g, &vNoEvolTransition);

      if (!getAttribute (g, &vg_polyT, &vg_polyF))
	{
	  // The signals, vsig and vDelayedSig, are absent in g.
	  addTransitionFrom (NextTransitions, g, vNoEvolTransition);

	  b = pk_false;
	  GET_TRANSITIONS_FROM (lcurrentTransitions, g, lts, it);
	  while (getTransitionFrom (&lts, (State) g, (State *) & g1, &vclkTransition, &vTransition))
	    {
	      if (getAttribute (g1, &g11, &g12))
		{
		  // The signals, vsig and vDelayedSig, are present in g1, but not in g.
		  /* Building the induced transitions: g -[]->g11, g-[]->g12.
		     The automaton becomes non deterministic. Adding a invisible_state to resolve it. */
		  vinit = mkInitValue (g11, g12, vsig);

		  declareInvisibleState (g, g11, g12, vinit, &invState_g, &invState_g11, &invState_g12);
		  h1 = mkClockReturnIn (g, invState_g, g11, invState_g11);
		  addTransition (NextTransitions, g, g11, CKM->makeInf (h1, vclkTransition));

		  h1 = mkClockReturnIn (g, invState_g, g12, invState_g12);
		  addTransition (NextTransitions, g, g12, CKM->makeInf (h1, vclkTransition));
		  b = pk_true;
		}
	      else
		// The signals, vsig and vDelayedSig, are absent in g and g1.
		addTransitionFrom (NextTransitions, g, vTransition);	// NextTransitions = ptlPointer_add(NextTransitions, vTransition);
	    }
	  if (b)
	    updateClocks (g);
	}
    }

  BTRACE_LEVEL1 (printf ("          ::end mkTransitions        ");
		 vtime = ((int) sysgetcputime () - beginAt);
		 pK_messenger->printTime (vtime);
    );
  BTRACE_LEVEL1 (printf ("\n");
		 fflush (stdout));
  return NextTransitions;
}

/**
   It adds to the global table vtableAttr the fact that vg_polyT (resp. vg_polyF) is derived from g + a valuation to True(Resp. False) for a delay, ie vtableAttr[g]={ vg_polyT, vg_polyF}.
   Moreover, it declares a enumerate value for each graph in the TypeStateEnum type.
*/
void
DCG2Automaton::setAttribute (ptDCG_poly g, ptDCG_poly vg_polyT, ptDCG_poly vg_polyF)
{
  ptlPointer ls;

  ls = ptlPointer_add (NILLPOINTER, vg_polyT);
  ls = ptlPointer_add (ls, vg_polyF);
  vtableAttr[g] = ls;
  NameListAppendName (NL, (NameClass) ident2repstr (getStateName (vg_polyT)));
  NameListAppendName (NL, (NameClass) ident2repstr (getStateName (vg_polyF)));
}

/**
   @return true if it exists to states (gt, gf) derived from one (g), "false" otherwise.
*/
boolean
DCG2Automaton::getAttribute (ptDCG_poly g, ptDCG_poly * gt, ptDCG_poly * gf)
{
  ptlPointer ls;
  pointer vaux;

  if (vtableAttr.find (g) == vtableAttr.end ())
    return pk_false;
  ls = (ptlPointer) vtableAttr[g];
  ptlPointer_get_elemNext (&ls, &vaux);
  *gt = (ptDCG_poly) vaux;
  ptlPointer_get_elemNext (&ls, &vaux);
  *gf = (ptDCG_poly) vaux;
  return pk_true;
}

/*
  @return 
  the states of the automaton produced by the valuations (true/false) of a boolean signal (vsig) for each state of lcurrentStates. 
  For lcurrentStates={ S1,...,Sn}, it produces { S11, S12, ...Sn1, Sn2} where Si1 = Si with vsig is True, Si2 = Si with vsig is False.

  It also returns the new initial state (NextInitialState) of the automata derived from the current one (currentInitialState).

*/
ptlStates
DCG2Automaton::mkStates (ptlStates lcurrentStates, State currentInitialState, ptexpcal vsig, State * NextInitialState)
{
  ptlStates NextStates;
  pointer elem;
  ptDCG_poly g, vg_polyT, vg_polyF;
  ptexpcal cloneSig;
  boolean b;
  int vtime, beginAt;

  BTRACE_LEVEL1 (printf ("          ");
		 fflush (stdout);
    );
  beginAt = (int) sysgetcputime ();

  NextStates = NILLSTATES;
  while (ptlPointer_get_elemNext (&lcurrentStates, &elem))
    {
      g = (ptDCG_poly) elem;
      setCurrentGraph (g);
      BTRACE_LEVEL1 (printf ("#");
		     fflush (stdout));

      b = getCloneSignal (vsig, g, &cloneSig);
      b = b && !CKM->eqClock (EM->getClock (cloneSig), TheNullClock);
      if (b)
	{
	  vg_polyF = mkDCGPolyBooleanValuation (g, cloneSig, pk_false);
	  NextStates = ptlPointer_add (NextStates, vg_polyF);
	  vg_polyT = mkDCGPolyBooleanValuation (g, cloneSig, pk_true);
	  NextStates = ptlPointer_add (NextStates, vg_polyT);
	  // Next Initial State.
	  if (TestSameObjects (currentInitialState, g))
	    if (EM->eq_trueExpcal (EM->getInitialValue (vsig)))
	      *NextInitialState = vg_polyT;
	    else
	      *NextInitialState = vg_polyF;
	  // updateClocks(vg_polyT); updateClocks(vg_polyF);
	  setAttribute (g, vg_polyT, vg_polyF);
	}
      else
	{			// vsig is absent in state g.
	  NextStates = ptlPointer_add (NextStates, g);
	  if (TestSameObjects (currentInitialState, g))
	    *NextInitialState = g;
	  addValuationPragma (g, EM->getName (vsig), string2repstr ("absent"));
	}
    }

  BTRACE_LEVEL1 (vtime = ((int) sysgetcputime () - beginAt);
		 pK_messenger->printTime (vtime);
    );
  BTRACE_LEVEL1 (printf ("\n");
		 fflush (stdout));
  return NextStates;
}



void
DCG2Automaton::removeRefToIn (ptDCG g1, ptDCG g2)
{
  ptExprManager EM;
  ptClockManager CKM;
  ptlinsth lh, lh_aux;
  ptinsth h;
  ptlinstances lsig;
  ptexpcal vsig;
  ptTablePointerPointer vtable;

  EM = g2->getExpressionManager ();
  if (EM->getExprList (opident, &lsig))
    while (ptlinstances_get_next (&lsig, &vsig))
      if (GET_ATTR_LINKGRAPHS (EM, vsig, vtable))
	vtable->erase ((pointer) g1);

  CKM = g2->getClockManager ();
  lh_aux = CKM->getClocks ();
  lh = lh_aux;
  while (ptlinsth_get_next (&lh, &h))
    if (GET_ATTR_LINKGRAPHS (CKM, h, vtable))
      vtable->erase ((pointer) g1);
  ptlinsth_free (lh_aux);
}

/**
   Garbaging: removing the elements of lcurrentStates, not in NextStates.
*/
void
DCG2Automaton::deleteGraphs (ptlStates lcurrentStates, ptlStates NextStates)
{
  pointer elem;
  ptDCG_poly g;
  ptDCG aDCG;

  BTRACE_LEVEL1 (printf ("     Removing Graphs BEGIN\n          ");
		 fflush (stdout));
  while (ptlPointer_get_elemNext (&lcurrentStates, &elem))
    {
      if (!ptlPointer_IsIn (NextStates, elem))
	{
	  g = (ptDCG_poly) elem;
	  if (g != G_ORI)
	    {
	      BTRACE_LEVEL1 (printf ("!");
			     fflush (stdout));
	      aDCG = g->getDCG ();
	      removeRefToIn (aDCG, G_ORI->getDCG ());
	      delete aDCG;
	    }
	}
    }
  BTRACE_LEVEL1 (printf ("\n     Removing Graphs END \n");
		 fflush (stdout));
}

void
DCG2Automaton::prelude (ptlinstances lStatesVar)
{
  ptexpcal vsig;
  ptINode aNode;
  ptexpcal resret, sigret, vret, vfen, vinit;

  while (ptlinstances_get_next (&lStatesVar, &vsig))
    {
      aNode = EM_G_ORI->getNode (vsig);
      NM_G_ORI->get_delay_arg (aNode, &resret, &sigret, &vret, &vfen, &vinit);
      EM_G_ORI->addMark (sigret, DONT_CHANGE_TYPE);	// no boolean2event conversion!!
    }
}

/**
   It builds the automaton (vautomaton) associated with a Signal program (G_ORI) for a list of state variables (lStatesVar).
   See mkListOfStatesVariables() for the ordering of the elements of the lists.
*/
void
DCG2Automaton::buildAutomaton (ptlinstances lStatesVar)
{
  ptexpcal vsig;
  ptlStates lcurrentStates, NextStates;
  ptTransitionsTable NextTransitions, lcurrentTransitions;
  State currentInitialState, NextInitialState;
  boolean bm, bmess;
  int vtime, beginAt;
  int nbVars, cptLoc;

  beginAt = (int) sysgetcputime ();
  bm = G_ORI->getDCG ()->getSetSuspendClockNameAction (pk_true);
  prelude (lStatesVar);
  bmess = pK_messenger->setSilentModeMessagesTo (pk_true);
  cpt = 1;
  lcurrentTransitions = new TablePointerPointerType ();
  // Initial Automaton.
  lcurrentStates = ptlPointer_add (NILLPOINTER, G_ORI);
  addTransition (lcurrentTransitions, G_ORI, G_ORI, G_ORI->getStep ());
  currentInitialState = G_ORI;
  // ---
  if (lStatesVar == NILLINSTANCES)
    {
      NextStates = lcurrentStates;
      NextTransitions = lcurrentTransitions;
      NextInitialState = currentInitialState;
    }
  else
    {
      nbVars = ptlinstances_get_size (lStatesVar);
      BTRACE_LEVEL1 (printf ("     o Nb State variable = %d\n", nbVars);
		     fflush (stdout));
      cptLoc = 0;
      while (ptlinstances_get_next (&lStatesVar, &vsig) /* && (cptLoc < 4) */ )	// cptLoc : added for testing (Lbesnard)
	{
	  cptLoc++;
	  BTRACE_LEVEL1 (printf ("     o State variable = %s\n", ident2string (EM_G_ORI->getName (vsig)));
			 printf ("       - nb States = %d\n", ptlPointer_get_size (lcurrentStates)));
	  NextStates = mkStates (lcurrentStates, currentInitialState, vsig, &NextInitialState);
	  NextTransitions = mkTransitions (lcurrentTransitions, vsig, lcurrentStates, NextStates);
	  rmUnreachableStates (&NextStates, NextTransitions);

	  deleteGraphs (lcurrentStates, NextStates);
	  // Next step
	  BTRACE_LEVEL1 (printf (" ptlPointer_free(lcurrentStates)\n"));
	  ptlPointer_free (lcurrentStates);
	  BTRACE_LEVEL1 (printf ("free(lcurrentTransitions)\n"));
	  free (lcurrentTransitions);
	  BTRACE_LEVEL1 (printf ("free(lcurrentTransitions) --after- \n"));
	  lcurrentStates = NextStates;
	  lcurrentTransitions = NextTransitions;
	  currentInitialState = NextInitialState;
	}
    }
  BTRACE_LEVEL1 (printf ("  buildAutomaton:: end of iteration."); fflush (stdout));
  BTRACE_LEVEL1 (vtime = ((int) sysgetcputime () - beginAt); pK_messenger->printTime (vtime); fflush (stdout));
  bm = G_ORI->getDCG ()->getSetSuspendClockNameAction (bm);
  finalize (NextStates, NextTransitions, NextInitialState);
  bmess = pK_messenger->setSilentModeMessagesTo (bmess);
}

/**
   TO BE COMPLETED.
*/
void
DCG2Automaton::mkOutputDefinitions ()
{
  ptlinstances lsig;
  ptexpcal vexp, vsig_in_g, vsig;
  ptlStates lStates;
  pointer velem;
  ptDCG g;
  ptINode node;
  require_managers (vautomaton);
  ptinsth h, h_in_g, hundef_auto;

  hundef_auto = CKM_vautomaton->getUndefClock ();
  lsig = outputs_ORI;
  while (ptlinstances_get_next (&lsig, &vsig))
    {
      vexp = NILEXPCAL;
      lStates = vautomaton->getStates ();
      if (EM->isSignalUnifiedWithClock (vsig, &h))
	{
	  while (ptlPointer_get_elemNext (&lStates, &velem))
	    {
	      g = (ptDCG) velem;
	      if (getLinkObject2Graphs (g->getClockManager (), h, g, &h_in_g))
		{
		  vsig_in_g = EM_vautomaton->mkWhenOrCellExp (opextr, EM_vautomaton->getTrue (), h_in_g);
		  if (vexp == NILEXPCAL)
		    vexp = vsig_in_g;
		  else
		    vexp = EM_vautomaton->mkBinaryExpTp (opmele, vexp, vsig_in_g, NILBACKANNOTATION);
		}
	    }
	}
      else
       {
	 while (ptlPointer_get_elemNext (&lStates, &velem))
	   {
	     g = (ptDCG) velem;
	     if (getLinkObject2Graphs (g->getExpressionManager (), vsig, g, &vsig_in_g))
	       {
		 if (vexp == NILEXPCAL)
		   vexp = vsig_in_g;
		 else
		   vexp = EM_vautomaton->mkBinaryExpTp (opmele, vexp, vsig_in_g, NILBACKANNOTATION);
	       }
	   }
       }
      if (vexp != NILEXPCAL)	// vsig := vexp
	{
	  EM_vautomaton->setSignalUnifiedWithClock (vsig, hundef_auto);
	  node = EM_vautomaton->getNode (vsig);
	  NM_vautomaton->becomes_undefined (node);
	  node = NM_vautomaton->mk_node_equation (vsig, vexp, hundef_auto, NILLINSTANCES, NILBACKANNOTATION);
	  NM_vautomaton->insertinSet (node);
	}
    }
}

/**
   The clock (vh) is exported of a graph (g). 
   If the clock is substituable (undersampling), the signals referenced in the conditions are exported.
*/
void
DCG2Automaton::becomesOutput (ptDCG g, ptinsth vh)
{
  ptlinstances louts, lsig, lins;
  ptexpcal exp;
  ptClockManager CKM = g->getClockManager ();
  ptExprManager EM = g->getExpressionManager ();

  if (CKM->isSubstitutable (vh))
    {
      lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vh));
      // changeName(g, lsig); // Done in shareDeclarations();
      louts = g->getOutputs ();
      lins = g->getInputs ();
      while (ptlinstances_get_next (&lsig, &exp))
	{
	  if (!ptlinstances_isin_exp (lins, exp))
	    if (!ptlinstances_isin_exp (louts, exp))
	      louts = ptlinstances_push_end (louts, exp);
	}
      g->setOutputs (louts);
    }

  aProfileBuilder->becomesOutput (g, vh);
}

/**
   Sharing the signals of the states, except the outputs.
 */
void
DCG2Automaton::shareDeclarations (ptlStates lStates)
{
  ptlinstances ls;
  ptexpcal vsig_in_g, vsig;
  ptlStates lst;
  boolean b;
  ptinterface vinterface_automaton;
  pointer velem;
  ptDCG g;
  ptExprManager EM;

  if (EM_G_ORI->getExprList (opident, &ls))
    {
      vinterface_automaton = vautomaton->getInterface ();
      while (ptlinstances_get_next (&ls, &vsig))
	if (!EM_G_ORI->IsInputFlow (vsig) && !EM_G_ORI->IsOutputFlow (vsig))
	  if (!ptlinstances_isin_exp (UserSelectedSignals, vsig))
	    {
	      lst = lStates;
	      b = pk_false;
	      while (ptlPointer_get_elemNext (&lst, &velem))
		{
		  g = ((ptDCG_poly) velem)->getDCG ();
		  if (isUsedIn (vsig, g, &vsig_in_g))
		    {
		      EM->setName (vsig_in_g, EM_G_ORI->getName (vsig));
		      EM = g->getExpressionManager ();
		      // EM->setSharedSignal(vsig_in_g, pk_true);
		      EM->setSharedSignalBySeveralGraphs (vsig_in_g, pk_true);
		      b = pk_true;
		    }
		}
	      // Attention au type du sÃ©lectionnÃ©. (booleen->event)...
	      if (b)
		{
		  vsig_in_g = mkStateVariableAutomaton (vsig);
		  EM_vautomaton->setStateTag (vsig_in_g, pk_false);
		  EM_vautomaton->setSharedSignal (vsig_in_g, pk_true);
		  ptinterface_insert_localSignal (vinterface_automaton, vsig_in_g);
		}
	    }
    }
}

/**
   Finalize step: Defining the transitions of the the automaton.
*/
void
DCG2Automaton::finalize_setAutomatonTransitions_old (ptTransitionsTable vTransitionsTable)
{
  TablePointerPointerType::iterator it;
  ptDCG aDCG;
  ptlTransitions lTransitions;
  ptTransition vtransition;
  State s1, s2;
  pointer velem;
  ptinsth vh;

  for (it = vTransitionsTable->begin (); it != vTransitionsTable->end (); it++)
    {
      aDCG = ((ptDCG_poly) (*it).first)->getDCG ();
      lTransitions = (ptlTransitions) (*it).second;
      while (ptlPointer_get_elemNext (&lTransitions, &velem))
	{
	  vtransition = (ptTransition) velem;
	  vtransition->get (&s1, &s2, &vh);
	  vautomaton->addTransition (aDCG /* ((ptDCG_poly) s1) ->getDCG() */ , ((ptDCG_poly) s2)->getDCG (), vh);
	  delete vtransition;	// gc
	}
      ptlPointer_free ((ptlTransitions) (*it).second);	// gc
    }
}

/**
   Finalize step: Defining the transitions of the the automaton.
*/
void
DCG2Automaton::finalize_setAutomatonTransitions (ptTransitionsTable vTransitionsTable)
{
  TablePointerPointerType::iterator it;
  ptDCG aDCG1, aDCG2;
  ptlTransitions lts, last, lTransitions;
  ptTransition vtransition;
  State s1, s2;
  pointer velem;
  ptinsth vh;

  last = NILLPOINTER;
  lts = NILLPOINTER;
  for (it = vTransitionsTable->begin (); it != vTransitionsTable->end (); it++)
    {
      aDCG1 = ((ptDCG_poly) (*it).first)->getDCG ();
      lTransitions = (ptlTransitions) (*it).second;
      if (lts == NILLPOINTER)
	lts = lTransitions;
      if (last != NILLPOINTER)
	ptlPointer_conc (last, lTransitions);
      while (ptlPointer_get_elemNext (&lTransitions, &velem))
	{
	  vtransition = (ptTransition) velem;
	  vtransition->get (&s1, &s2, &vh);
	  aDCG2 = ((ptDCG_poly) s2)->getDCG ();
	  vtransition->set (aDCG1, aDCG2, vh);
	  printf (" s1 = %s, s2 = %s \n", ident2string (aDCG1->getName ()), ident2string (aDCG2->getName ()));
	  if (lTransitions != NILLPOINTER)
	    last = lTransitions;
	}
    }
  vautomaton->setTransitions (lts);
}


/**
   It creates the automaton. The states becomes DCG graphs (to be independant of the level).
*/
void
DCG2Automaton::finalize (ptlStates lStates, ptTransitionsTable vTransitionsTable, State InitialState)
{
  ptlTransitions lTransitions;
  pointer velem;
  ptDCG_poly g;
  State s1, s2;
  ptinsth vh;
  ptTransition vtransition;
  ptDCG aDCG;
  ptlStates lstates;
  TablePointerPointerType::iterator it;
  int vtime, beginAt, n;

  beginAt = (int) sysgetcputime ();
  BTRACE_LEVEL1 (printf ("\nFinalize: 0\n");
		 fflush (stdout));
  shareDeclarations (lStates);
  BTRACE_LEVEL1 (printf ("Finalize: 1\n");
		 fflush (stdout));
  lstates = lStates;
  n = 0;
  // Fixing the identifier of the clocks.
  // The operation was suspended during the construction to limit the number of identifiers in the tree memory.
  while (ptlPointer_get_elemNext (&lstates, &velem))
    {
      g = (ptDCG_poly) velem;
      aDCG = g->getDCG ();
      BTRACE_LEVEL1 (printf ("/%d", n++);
		     fflush (stdout));
      aDCG->setClocksName (true);
      changeName (aDCG, g->getOutputs ());
    }

  BTRACE_LEVEL1 (printf ("\nFinalize: 2\n");
		 fflush (stdout));
  for (it = vTransitionsTable->begin (); it != vTransitionsTable->end (); it++)
    {
      aDCG = ((ptDCG_poly) (*it).first)->getDCG ();
      lTransitions = (ptlTransitions) (*it).second;
      while (ptlPointer_get_elemNext (&lTransitions, &velem))
	{
	  vtransition = (ptTransition) velem;
	  vtransition->get (&s1, &s2, &vh);
	  becomesOutput (aDCG, vh);
	  // The interface is (re)computed later in nonProductiveSignalElimination() (See below)
	}
    }
  BTRACE_LEVEL1 (printf ("Finalize: 3\n");
		 fflush (stdout);
    );
  // ----- Defining the states of the the automaton.
  vautomaton->setName (makeIdent_ident_string (G_ORI->getName (), VSUFF_AUTO));
  lstates = lStates;
  while (ptlPointer_get_elemNext (&lstates, &velem))
    {
      g = (ptDCG_poly) velem;
      g->nonProductiveSignalElimination ();
      vautomaton->addState (g->getDCG ());
    }

  BTRACE_LEVEL1 (printf ("Finalize: 4\n");
		 fflush (stdout));
  finalize_setAutomatonTransitions (vTransitionsTable);

  BTRACE_LEVEL1 (printf ("Finalize: 5\n");
		 fflush (stdout));
  // --- Initial state, state variable associated with the automaton.
  mkDeclarations (((ptDCG_poly) InitialState)->getDCG ());

  BTRACE_LEVEL1 (printf ("Finalize: 6\n");
		 fflush (stdout);
    );
  // --- finalizing the definitions of the outputs (not shared signals).
  mkOutputDefinitions ();

  BTRACE_LEVEL1 (printf ("End Finalize (total time)\n");
		 vtime = ((int) sysgetcputime () - beginAt);
		 pK_messenger->printTime (vtime);
    );
  BTRACE_LEVEL1 (printf ("\n");
		 fflush (stdout));

}

/**
   @return a state variable of name id, of type TT, and initialized to vinit.
   This signal is declared in the automaton.
*/
ptexpcal
DCG2Automaton::mkStateVariableAutomaton (ptident id, TypeType TT, ptexpcal vinit)
{
  ptexpcal vState;
  vState = EM_vautomaton->mkIdentExp (id, tsignal, TT, vinit, NM_vautomaton->mk_node_undefined (), CKM_vautomaton->getUndefClock (), NILBACKANNOTATION);
  EM_vautomaton->setStateTag (vState, pk_true);
  return vState;
}

/**
   @return a copy of a signal. This copy is declared as a state variable in the automaton.
*/
ptexpcal
DCG2Automaton::mkStateVariableAutomaton (ptexpcal vsig)
{
  ptexpcal vsigCopy;
  vsigCopy = EM_vautomaton->mkIdentCopy (vsig, NM_vautomaton->mk_node_undefined (), CKM_vautomaton->getUndefClock ());
  EM_vautomaton->setStateTag (vsigCopy, pk_true);
  return vsigCopy;
}

/**
   @return a copy of a signal. This copy is declared as a shared signal in the automaton.
ptexpcal DCG2Automaton::mkSharedSignalAutomaton(ptexpcal vsig)
{
  ptexpcal  vsigCopy;
  vsigCopy = EM_vautomaton -> mkIdentCopy(vsig, NM_vautomaton->mk_node_undefined(), CKM_vautomaton->getUndefClock());
  EM_vautomaton ->setSharedSignal(vsigCopy, pk_true);
  return vsigCopy;
}
*/

/**
   @return true if the clone of vsig in g is useful, false otherwise.
*/
boolean
DCG2Automaton::isUsedIn (ptexpcal vsig, ptDCG g, ptexpcal * vsig_in_g)
{
  ptExprManager EM;
  ptlinstances ls_g;
  ptlINode unusedNodes;
  ptlinsth unusedClocks;
  ptinsth h;

  EM = g->getExpressionManager ();
  if (!getLinkObject2Graphs (EM, vsig, g, vsig_in_g))
    return pk_false;
  if (EM->isSignalUnifiedWithClock (*vsig_in_g, &h))
    return pk_true;
  if (!EM->getExprList (opident, &ls_g))
    return pk_false;
  if (!ptlinstances_isin_exp (ls_g, *vsig_in_g))
    return pk_false;
  unusedNodes = g->getUnused (&unusedClocks);
  return !ptlINode_isIn (unusedNodes, EM->getNode (*vsig_in_g));
}


/**
   It declares the signals that are shared by several graphs.
*/
void
DCG2Automaton::DeclSharedSignalsAutomaton ()
{
  ptlinstances ls;
  ptexpcal vsig;
  ptinterface vinterface;

  vinterface = vautomaton->getInterface ();
  // Declaring the invisible state variables of the automaton.
  ls = ls_invisible_states;
  while (ptlinstances_get_next (&ls, &vsig))
    ptinterface_insert_localSignal (vinterface, mkStateVariableAutomaton (vsig));
}

/**
   Add the declaation of the clocks of the dependences in the automaton.
*/
void
DCG2Automaton::ExportClocksTransitions ()
{
  ptlTransitions lTransitions;
  ptTransition vtransition;
  pointer velem;
  State s1, s2;
  ptinsth vh;
  ptinterface vinterface_automaton;
  ptClockManager CKM;
  int vtime, beginAt;
  ptlinsth lh;

  beginAt = (int) sysgetcputime ();
  vinterface_automaton = vautomaton->getInterface ();
  lh = ptinterface_get_localClocks (vinterface_automaton);

  lTransitions = vautomaton->getTransitions ();
  while (ptlPointer_get_elemNext (&lTransitions, &velem))
    {
      vtransition = (ptTransition) velem;
      vtransition->get (&s1, &s2, &vh);
      CKM = ((ptDCG) s1)->getClockManager ();
      // unicity is guaranteed for the clock.
      if (!CKM->isSubstitutable (vh))
	{
	  lh = ptlinsth_insert_elem (lh, vh);
	  //ptinterface_insert_localClock( vinterface_automaton, vh);
	}
    }
  ptinterface_set_localClocks (vinterface_automaton, lh);
  BTRACE_LEVEL1 (vtime = ((int) sysgetcputime () - beginAt);
		 pK_messenger->printTime (vtime);
    );
  BTRACE_LEVEL1 (printf ("\n");
		 fflush (stdout));
}

/**
   It imports in the automaton, the outputs of a state (g).
*/
void
DCG2Automaton::DeclareLocalOutputs (ptDCG g)
{
  ptinterface vinterface_automaton;
  ptlinstances lsig;
  ptexpcal vsig_copy, vsig;
  def_managers (vautomaton);

  vinterface_automaton = vautomaton->getInterface ();
  lsig = g->getOutputs ();

  while (ptlinstances_get_next (&lsig, &vsig))
    {
      vsig_copy = EM_vautomaton->mkIdentCopy (vsig, NM_vautomaton->mk_node_undefined (), CKM_vautomaton->getUndefClock ());
      EM_vautomaton->setName (vsig_copy, EM_vautomaton->getName (vsig));
      ptinterface_insert_localSignal (vinterface_automaton, vsig_copy);
    }
}

/**
   It fixes the interface of the automaton by the interface of the original graph + the clocks of the transitions.
   It adds 
   - a state variable to manage the current state of the automaton.
   - a labelled node, for each state, that references the state (a state is a process).
*/
void
DCG2Automaton::mkDeclarations (ptDCG InitialState)
{
  ptexpcal vlab;
  ptinterface vinterface;
  ptDCG g;
  ptlStates lStates;
  pointer velem;
  ptident initialState;
  ptINode node;

  vinterface = vautomaton->getInterface ();
  ptinterface_set_inputSignals (vinterface, G_ORI->getInputs ());
  ptinterface_set_outputSignals (vinterface, G_ORI->getOutputs ());

  BTRACE_LEVEL1 (printf ("\n");
		 fflush (stdout));
  BTRACE_LEVEL1 (printf ("   * mkDeclarations 1\n");
		 fflush (stdout));
  ExportClocksTransitions ();
  /* Declaration of the current state of the automaton if "invisible states" are used. */
  initialState = getLabelStateName (InitialState);

  vautomaton->setInitialState (initialState);

  // Declaration of the labelled nodes.
  BTRACE_LEVEL1 (printf ("   * mkDeclarations 2\n");
		 fflush (stdout));

  lStates = vautomaton->getStates ();
  while (ptlPointer_get_elemNext (&lStates, &velem))
    {
      g = (ptDCG) velem;
      node = NM_vautomaton->mk_node_call (g->getDCG (), NILEXPCAL, NILEXPCAL, NILEXPCAL, NILBACKANNOTATION);

      NM_vautomaton->insertinSet (node);
      vlab = EM_vautomaton->mkLabelIdent (getLabelStateName (g), NM_vautomaton->mk_node_undefined (), CKM_vautomaton->getUndefClock (), NILBACKANNOTATION, pk_false);
      NM_vautomaton->set_label (node, vlab);
      DeclareLocalOutputs (g);
    }
  BTRACE_LEVEL1 (printf ("   * mkDeclarations 3\n"); fflush (stdout));

  DeclSharedSignalsAutomaton ();
}

void
DCG2Automaton::DEBUG_print_transitions (ptlTransitions lTransitions)
{
  ptTransition vtransition;
  pointer velem;
  State s2, s1;
  ptinsth vh;
  ptDCG aDCG1, aDCG2;

  return;
  while (ptlPointer_get_elemNext (&lTransitions, &velem))
    {
      vtransition = (ptTransition) velem;
      vtransition->get (&s1, &s2, &vh);
      aDCG2 = (ptDCG) s2;
      aDCG1 = (ptDCG) s1;
      printf (" s1 = %s, s2 = %s \n", ident2string (aDCG1->getName ()), ident2string (aDCG2->getName ()));
    }
}
