/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/**
   @defgroup lmemnodeGrp atDelayTree:list of delay trees.   
*/
/**
   @ingroup lmemnodeGrp
   @return the list composed of the element mn followed by lmn.
*/
extern ptlmemnode
ptlmemnode_insert_elem (ptmemnode mn, ptlmemnode lmn)
{
  ptlmemnode vlmem;

  if (gclmemnode == NILLMEMNODE)
    vlmem = new lmemnode ();
  else
    {
      vlmem = gclmemnode;
      gclmemnode = gclmemnode->next;
    }
  vlmem->elem = mn;
  vlmem->next = lmn;
  return vlmem;
}


/**
   @ingroup lmemnodeGrp
   @return "true" if the list vlmem is not empty, "false" otherwise.
   @param elem (out) current element of the list, available when the list is not empty.
   @param vlmem (in/out) is the next element of the in-list.
*/
extern boolean
ptlmemnode_getNaryArgs (ptlmemnode * vlmem, ptmemnode * elem)
{
  if ((*vlmem) == NILLMEMNODE)
    return pk_false;
  (*elem) = (*vlmem)->elem;
  *vlmem = (*vlmem)->next;
  return pk_true;
}


/** @ingroup lmemnodeGrp
    @return the current element of  vlmem.
 */
extern ptmemnode
ptlmemnode_get_elem (ptlmemnode vlmem)
{
  return vlmem->elem;
}

/** 
    @ingroup lmemnodeGrp
    It removes an element (ele) from a list (lmem).
 */
extern ptlmemnode
ptlmemnode_remove_elem (ptlmemnode lmem, ptmemnode ele)
{
  ptlmemnode vlmem, vlpred;
  boolean trouve;

  vlmem = lmem;
  vlpred = lmem;
  trouve = pk_false;
  while ((vlmem != NILLMEMNODE) && !trouve)
    {
      trouve = (boolean) (ptlmemnode_get_elem (vlmem) == ele);
      if (!trouve)
	{
	  vlpred = vlmem;
	  vlmem = ptlmemnode_get_next (vlmem);
	}
    }
  if (!trouve)
    return lmem;
  if (vlmem == lmem)
    return ptlmemnode_get_next (lmem);
  setlmemnext (vlpred, ptlmemnode_get_next (vlmem));
  return lmem;
}

/** 
    @ingroup lmemnodeGrp
    @return the tail of the list (vlmem). 
*/
extern ptlmemnode
ptlmemnode_get_next (ptlmemnode vlmem)
{
  return vlmem->next;
}


/** Garbage collector.
    Operand list (l) becomes free.
*/
extern void
ptlmemnode_free (ptlmemnode l)
{
  ptlmemnode ngclmemnode;

  if (l != NILLMEMNODE)
    {
      ngclmemnode = l;
      while (l->next != NILLMEMNODE)
	l = l->next;
      l->next = gclmemnode;
      gclmemnode = ngclmemnode;
    }
}

/** @return a new node of a delay tree. It is associated with a signal (vexp). */
extern ptmemnode
makeMemnode (ptexpcal vexp, typuseinec useInExternalCall)
{
  ptmemnode vmem;

  vexp = ptexpcal_getUnindexedSignal (vexp);
  vmem = ptmemnode_RIEN_create (vexp, ptexpcal_get_initExp (vexp), useInExternalCall);
  ptexpcal_set_node_tree_delay (vexp, vmem);
  return vmem;
}


/** It sets to ?VRET (resp. ?VFEN) the value of the delay (resp. window) for the delay-tree node ?vmem. */
extern void
ptmemnode_set_delayWindow (ptmemnode vmem, ptexpcal vret, ptexpcal vfen)
{
  vmem->ret = vret;
  vmem->fen = vfen;
}

/**
  Fixe a ?VTAILLE le modulo de l'indice associe a ?VMEM
*/
extern void
ptmemnode_set_size (ptmemnode vmem, integer vtaille)
{
  vmem->taille = vtaille;
}


/** Definit ?EXP comme expression d'init des cell agreges a ?VMEM. */
extern void
ptmemnode_set_cellinit (ptmemnode vmem, ptexpcal exp)
{
  vmem->cellini = exp;
}


/** -->Signal affecte au noeud ?VMEN d'un arbre de retard */
extern ptexpcal
ptmemnode_get_signal (ptmemnode vmem)
{
  return vmem->vinst;
}



/** -->Pere de ?VMEM dans l'arborescence des retards */
extern ptmemnode
ptmemnode_get_father (ptmemnode vmem)
{
  return vmem->father;
}



/** It returns pk_true if ?VMEM is a root of a delay tree. */
extern boolean
ptmemnode_satisfies_root (ptmemnode vmem)
{
  if (vmem == NILMEMNODE)
    return pk_false;
  if (vmem->father == NILMEMNODE)
    return pk_true;
  return (vmem->father == vmem);
}




/** It returns the signal assigned to the root of the delay tree node ?MN. */
extern ptexpcal
ptmemnode_get_signal_root (ptmemnode mn)
{
  mn = ptmemnode_get_root (mn);
  return mn->vinst;
}



/** -->les fils du noeud ?VMEM d'un arbre de retards */
extern ptlmemnode
ptmemnode_get_lchild (ptmemnode vmem)
{
  return vmem->lchild;
}

/** It returns the set of tree delay node equivalent to !vmem */
extern ptlmemnode
ptmemnode_get_equiv (ptmemnode vmem)
{
  return vmem->equiv;
}



/** -->Expression du retard du noeud ?VMEM d'une arborescence de retard */
extern ptexpcal
ptmemnode_get_delay_exp (ptmemnode vmem)
{
  return vmem->ret;
}



/** -->Expression de la fenetre glissante du noeud ?VEM element d'une arboresecence de retard. */
extern ptexpcal
ptmemnode_get_window_exp (ptmemnode vmem)
{
  return vmem->fen;
}



/** It returns the initial value of the index assigned to the delay tree node ?VMEM. */
extern integer
ptmemnode_get_initial_index (ptmemnode vmem)
{
  return vmem->init;
}



/** -->nombre d'indices compactes dans ?VMEM */
extern integer
ptmemnode_get_nb_indexes (ptmemnode vmem)
{
  return vmem->nbic;
}



/** It returns the arrea size assigned to the node ?VMEM. */
extern integer
ptmemnode_get_size (ptmemnode vmem)
{
  return vmem->taille;
}



/** -->Expression d'initialisation associee a ?VMEM */
extern ptexpcal
ptmemnode_get_init_exp (ptmemnode vmem)
{
  return vmem->valini;
}



/** It returns the initializing value of the delay-tree node ?vmem when it is a cell. */
extern ptexpcal
ptmemnode_get_cell_init_exp (ptmemnode vmem)
{
  return vmem->cellini;
}



/**
  It returns the type of the use of the node ?MN for external call.
*/
extern typuseinec
ptmemnode_get_use_in_external_call (ptmemnode mn)
{
  return mn->useinec;
}

/** It returns the value of the window for the delay tree node ?mem. */
extern integer
ptmemnode_get_window_value (ptmemnode mem)
{
  ptexpcal vexp;

  vexp = ptmemnode_get_window_exp (mem);
  if (vexp == NILEXPCAL)
    return 1;
  return getMaxValue (vexp);
}



/** Element d'une arboresecence de retard */
extern integer
ptmemnode_get_delay_value (ptmemnode mem)
{
  ptexpcal vexp;

  vexp = ptmemnode_get_delay_exp (mem);
  if (vexp == NILEXPCAL)
    return 0;
  return getMaxValue (vexp);
}


/** It returns the name of the identifier for the tree delay node ?VMEM. */
extern ptident
ptmemnode_get_signal_ident (ptmemnode vmem)
{
  return ptexpcal_get_identifier (ptmemnode_get_signal (vmem));
}



/**-->le representant de ?MN attache a l'instance
 */
extern ptmemnode
ptmemnode_get_real_root (ptmemnode mn)
{
  return ptexpcal_get_node_tree_delay (ptmemnode_get_signal (mn));
}



/**
  It returns the number assigned to the root of the delay tree of ?MN.
*/
extern integer
ptmemnode_get_root_num (ptmemnode mn)
{
  ptmemnode mnroot;

  mn = ptmemnode_get_root (mn);
  mnroot = ptmemnode_get_real_root (mn);
  return atgetmemnumroot (mn, mnroot);
}



/** It returns the number of the index assigned to the delay tree node ?MN. */
extern integer
ptmemnode_get_index_num (ptmemnode mn)
{
  if (ptmemnode_get_equiv (mn) == NILLMEMNODE)
    return ptmemnode_get_root_num (mn);
  return 0;
}

/** It removes the child ?VCHILD of ?VFATHER. */
extern void
ptmemnode_remove_child (ptmemnode vfather, ptmemnode vchild)
{
  setmemlchild (vfather, ptlmemnode_remove_elem (ptmemnode_get_lchild (vfather), vchild));
  setmemfather (vchild, NILMEMNODE);
}



/** It appends ?VCHILD  as child of ?VFATHER. */
extern void
ptmemnode_insert_child (ptmemnode vfather, ptmemnode vchild)
{
  setmemlchild (vfather, ptlmemnode_insert_elem (vchild, ptmemnode_get_lchild (vfather)));
  setmemfather (vchild, vfather);
}



/**
  It sets ?VINIT as initial value for the index assigned to ?VMEM. 
*/
extern void
ptmemnode_set_initial_index (ptmemnode vmem, integer vinit)
{
  vmem->init = vinit;
}



/**
  Fixe a ?VLEQUIV le champs de equivalents de ?VMEM
*/
extern void
ptmemnode_set_equiv (ptmemnode vmem, ptlmemnode vlequiv)
{
  vmem->equiv = vlequiv;
}



/**
  It sets to ?VNBIC the number of compacted indexes of the delay tree node ?VMEM.
*/
extern void
ptmemnode_set_nb_indexes (ptmemnode vmem, integer vnbic)
{
  vmem->nbic = vnbic;
}



/** -->representant de ?VMEMREF par duplication */
extern ptmemnode
ptmemnode_equiv_copy (ptmemnode vmemref)
{
  ptmemnode vmem;
  ptlmemnode vequiv;

  vequiv = ptmemnode_get_equiv (vmemref);
  vmem = ptmemnode_RIEN_create (ptmemnode_get_signal (vmemref), NILEXPCAL, nouse);
  ptmemnode_set_equiv (vmemref, ptlmemnode_insert_elem (vmem, vequiv));
  return vmem;
}


/**
  Subsitue ?ELEASUB dans ?LMEM par ?ELEDESUB
*/
extern void
sublmemnode (ptlmemnode lmem, ptmemnode eleasub, ptmemnode eledesub)
{
  ptlmemnode vlmem, vlpred;
  boolean trouve;
  ptmemnode elem;

  vlmem = lmem;
  trouve = pk_false;
  while ((vlmem != NULL) && !trouve)
    {
      vlpred = vlmem;
      if (ptlmemnode_getNaryArgs (&vlmem, &elem))
	trouve = (elem == eleasub);
    }
  if (!trouve)
    setlmemnext (vlpred, ptlmemnode_insert_elem (eledesub, NILLMEMNODE));
  else
    setlmemele (vlpred, eledesub);
}


/**
  Fixe ?USE comme type d'utilisation du signal retarde ?VINST dans les
  appels externes. si inutilisé devient use si utilisé pas du même type
  devient utilisé in et out
*/
extern void
ptmemnode_belongs_typuseinec (ptmemnode mn, typuseinec use)
{
  typuseinec t;
  t = mn->useinec;
  if (t == nouse)
    mn->useinec = use;
  else if (t != use)
    mn->useinec = useinout;
}



/** Fixe a ?T la taille des descripteurs de l'arborescence de retard de racine ?VMEM */
extern void
ptmemnode_set_size_all (ptmemnode vmem, integer t)
{
  ptlmemnode vlchild;

  ptmemnode_set_size (vmem, t);
  vlchild = ptmemnode_get_lchild (vmem);
  while (vlchild != NILLMEMNODE)
    {
      ptmemnode_set_size_all (ptlmemnode_get_elem (vlchild), t);
      vlchild = ptlmemnode_get_next (vlchild);
    }
}

///
extern integer
ptmemnode_size_compute (ptmemnode vmem, integer t, integer vinitpere, integer vnbicpere)
{
  integer vfen, vinit, vnbic;
  ptlmemnode vlchild;
  ptmemnode elem;

  vfen = ptmemnode_get_window_value (vmem);
  vinit = vinitpere + ptmemnode_get_delay_value (vmem) + (vfen - 1);
  ptmemnode_set_initial_index (vmem, vinit);
  if (vfen == 1)
    vnbic = vnbicpere;
  else
    vnbic = vnbicpere + 1;
  ptmemnode_set_nb_indexes (vmem, vnbic);
  if (t < vinit)
    t = vinit;
  vlchild = ptmemnode_get_lchild (vmem);
  while (ptlmemnode_getNaryArgs (&vlchild, &elem))
    t = ptmemnode_size_compute (elem, t, vinit, vnbic);
  return t;
}
