/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
///
typedef struct memnode
/* Trees of Delays */
{
  ptexpcal vinst;
  ptmemnode father;		/* noeud pere dans l'arbre */
  ptlmemnode lchild;		/* liste des fils */
  ptlmemnode equiv;
  ptexpcal ret,			/* décalage */
    fen;			/* fenêtre */
  integer init,			/* */
    nbic,			/* */
    taille;			/* taille du modulo */
  ptexpcal valini;		/* valeurs initiales */
  ptexpcal cellini;		/* idem pour les cells */
  typuseinec useinec;		/* Utilisation dans un appel externe */
}				///
memnode;

///
typedef struct lmemnode
/* Liste d'elements de type memnode */
{
  ptmemnode elem;
  ptlmemnode next;
}				///
lmemnode;


static ptlmemnode gclmemnode = NILLMEMNODE;	/* Garbage */

/**-->Nouveau noeud d'une arborescence de retard,
  associe a ?VINST, d'initialisation ?VALINI et decore de l'attribut
  ?USEINEC qui specifie son type d'utilisation dans un appel externe
*/
static ptmemnode
ptmemnode_RIEN_create (ptexpcal vinst, ptexpcal valini, typuseinec useinec)
{
  ptmemnode vmem;

  vmem = new memnode ();
  vmem->vinst = vinst;
  vmem->father = NILMEMNODE;
  vmem->lchild = NILLMEMNODE;
  vmem->equiv = NILLMEMNODE;
  vmem->ret = NILEXPCAL;
  vmem->fen = NILEXPCAL;
  vmem->init = 0;
  vmem->nbic = 0;
  vmem->taille = 0;
  vmem->valini = valini;
  vmem->cellini = NILEXPCAL;
  vmem->useinec = useinec;
  return vmem;
}





/**-->numero de la racine de l'arbre contenant ?MN      */
static integer
atgetmemnumroot (ptmemnode mn, ptmemnode mnroot)
{
  ptlmemnode lmn;
  integer num;
  ptmemnode elem;
  boolean found;

  if (mn == mnroot)
    return 0;
  num = 1;
  lmn = ptmemnode_get_equiv (mnroot);
  found = pk_false;
  while (ptlmemnode_getNaryArgs (&lmn, &elem) && !found)
    if (mn == elem)
      found = pk_true;
    else
      num = num + 1;
  return num;
}




/** -->racine de ?MN dans l'arborescence des retards. */
static ptmemnode
ptmemnode_get_root (ptmemnode mn)
{
  while ((mn->father != NILMEMNODE) && (mn->father != mn))
    mn = mn->father;
  return mn;
}


/** Fixe ?ELE comme element du descripteur ?VLMEM */
static void
setlmemele (ptlmemnode vlmem, ptmemnode ele)
{
  vlmem->elem = ele;
}


/** Fixe a ?VLNEXT descripteur suivant de ?VLMEM */
static void
setlmemnext (ptlmemnode vlmem, ptlmemnode vlnext)
{
  vlmem->next = vlnext;
}


/**
  Attache ?VFATHER comme pere de ?VCHILD
*/
static void
setmemfather (ptmemnode vchild, ptmemnode vfather)
{
  vchild->father = vfather;
}


/** Attache ?VLCHILD comme fils de ?VMEM */
static void
setmemlchild (ptmemnode vmem, ptlmemnode vlchild)
{
  vmem->lchild = vlchild;
}

/** @return the max-value of vexp (assumed bounded) */
static compilerInteger
getMaxValue (ptexpcal vexp)
{
  compilerInteger bs, bi;

  if (ptexpcal_IsAssigned (vexp))
    return ptexpcal_get_integerValue (vexp);
  if (ptexpcal_IsBounded (vexp))
    {
      ptexpcal_get_integerBounds (vexp, &bi, &bs);
      return bs;
    }
  return 0;
}
