/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
/**
   Structure conversion.
   @return a combed expression (list) from a list of expressions (liste_exp).
*/
  ptexpcal downstreamRetiming::mkexpcalfromptlinstances (ptlinstances * liste_exp)
{
  ptexpcal vexpcal, expd, expres;

  ptlinstances_get_next (liste_exp, &vexpcal);
  if (*liste_exp != NILLINSTANCES)
    {
      expd = mkexpcalfromptlinstances (liste_exp);
      expres = EM->mkNaryExp (oplelevec, vexpcal, expd, NILBACKANNOTATION);
    }
  else
    expres = EM->mkNaryExp (oplelevec, vexpcal, NILEXPCAL, NILBACKANNOTATION);
  return expres;
}

/**
    It extracts a list (l_oper) from another one (l_liste_init).
    @return "true" if the extracted list is not reduced to null expression, "false" otherwise.
*/
boolean
downstreamRetiming::extractlistfromptexpcal (ptexpcal * l_liste_init, ptlinstances * l_oper)
{
  ptexpcal liste_init, liste_init_int, vexpcal;
  ptlinstances l_oper_int;
  boolean b;

  if (*l_liste_init == NILEXPCAL)
    {
      *l_oper = NILLINSTANCES;
      return pk_true;
    }
  liste_init = EM->getNaryArgs (l_liste_init);
  if (liste_init == NILEXPCAL)
    return pk_false;
  vexpcal = EM->getNaryArgs (&liste_init);
  liste_init_int = *l_liste_init;
  b = extractlistfromptexpcal (&liste_init_int, &l_oper_int);
  if (liste_init != NILEXPCAL)
    {
      *l_liste_init = EM->mkNaryExp (oplelevec, liste_init, liste_init_int, NILBACKANNOTATION);
      *l_oper = ptlinstances_push_front (l_oper_int, vexpcal);
      return pk_true;
    }
  else
    {
      *l_oper = ptlinstances_push_front (l_oper_int, vexpcal);
      return pk_false;
    }
}


/**
   @param liters (in/out) list of iterators. It is set to the tail of the list (out).
   @param bi (out) lower bound of the first element of liters(in)
   @param bs (out) upper bound of the first element of liters(in)
   @param pas (out) step of the first element of liters(in).
   \note Assumed: the elements are valued.
*/
void
downstreamRetiming::dec_parcours_init (ptexpcal * liters, long int *bi, long int *bs, long int *pas)
{
  ptexpcal parcours, expg, expd, exp1, exp2, exp3, expcalres;

  parcours = EM->getNaryArgs (liters);
  EM->getBinaryArgs (parcours, &expg, &expd);
  expcalres = expg;
  EM->getTernaryArgs (expd, &exp1, &exp2, &exp3);
  if (exp1 != NULL)
    *bi = EM->getValue (exp1);
  else
    *bi = 1;
  if (exp2 != NULL)
    *bs = EM->getValue (exp2);
  if (exp3 != NULL)
    *pas = EM->getValue (exp3);
  else
    *pas = 1;
}


/**
   @return an initialisation expression built on travels taking as model another expression (vexpcal).
   The signals referenced in vexpcal are substituted by vexpi in the retruned exrpession.
*/
ptexpcal
downstreamRetiming::mk_exp_init (ptexpcal vexpcal, ptexpcal vexpi)
{
  ptexpcal expres, expres1, expres2;
  ptexpcal exp1, exp2;
  noperators op;
  cloper cl;

  op = EM->getOperator (vexpcal);
  cl = EM->getOperatorClass (vexpcal);
  switch (cl)
    {
    case constant:
      return vexpcal;
    case variabl:
      return vexpi;
    case oper1:
      exp1 = EM->getUnaryArg (vexpcal);
      expres1 = mk_exp_init (exp1, vexpi);
      expres = EM->mkUnaryExp (op, expres1, EM->getBackAnnotation (vexpcal));
      return expres;

    case oper2:
      exp1 = EM->getBinaryArgstAt (vexpcal, 1);
      expres1 = mk_exp_init (exp1, vexpi);
      exp2 = EM->getBinaryArgstAt (vexpcal, 2);
      expres2 = mk_exp_init (exp2, vexpi);
      expres = EM->mkBinaryExp (op, expres1, expres2, EM->getBackAnnotation (vexpcal));
      return expres;
    default:
      ;
    }
  return NILEXPCAL;
}


/**
   @return the ind-th element of vexpcal if it exists, null expression otherwise.
   More precisely, 
   - for a vectorial element [i]:E, if (i=ind) it returns E, null expression otherwise.
   - for an iterator on expression {i \e from bi \e to bs \e step pas}:E,
   it returns [ind]:E', where E' is E in which the references to i are replaced by ind.
   if ind is not an element of the "enumeration" (ie ind != bi+ n*pas), it returns null expression.
   - otherwise dec_expr_simple_init() is called
*/
ptexpcal
downstreamRetiming::dec_elem_vect_init (int ind, ptexpcal * vexpcal)
{
  noperators ope;
  long int val, i;
  long int bi, bs, pas;
  ptexpcal vexpg, vexpd, expind, expcalres, expcalint, expcalind;

  expcalres = NILEXPCAL;
  ope = EM->getOperator (*vexpcal);
  switch (ope)
    {
    case opelevec:
      EM->getBinaryArgs (*vexpcal, &vexpg, &vexpd);
      expind = EM->getNaryArgs (&vexpg);
      val = (int) (EM->getValue (expind));
      if (val == ind)
	expcalres = vexpd;
      return expcalres;

    case opensvec:
      EM->getBinaryArgs (*vexpcal, &vexpg, &vexpd);
      dec_parcours_init (&vexpg, &bi, &bs, &pas);
      expcalint = dec_elem_vect_init (ind, &vexpd);
      for (i = bi; i <= bs; i += pas)
	{
	  if (i == ind)
	    {
	      expcalind = EM->mkIntegerConstant (ind);
	      expcalres = mk_exp_init (expcalint, expcalind);
	    }
	}
      return expcalres;

    default:
      return dec_expr_simple_init (ind, vexpcal);
    };
}

/**
   @return the ind-th element of vexpcal (dec_elem_vect_init())
*/
ptexpcal
downstreamRetiming::dec_enum_vect_init (int ind, ptexpcal * vexpcal)
{
  ptexpcal velem, vexpcalint, vexpcalres;

  while (*vexpcal != NULL)
    {
      velem = EM->getNaryArgs (vexpcal);
      vexpcalint = dec_elem_vect_init (ind, &velem);
      if (vexpcalint != NILEXPCAL)
	vexpcalres = vexpcalint;
    };
  return vexpcalres;
}



/**
   @return 
    - vexpcal if it is not a list of vectorial element (oplelevec operator),
    - the ind-th element of vexpcal (dec_enum_vect_init())
*/
ptexpcal
downstreamRetiming::dec_expr_simple_init (int ind, ptexpcal * vexpcal)
{
  noperators ope;

  ope = EM->getOperator (*vexpcal);
  if (ope == oplelevec)
    return dec_enum_vect_init (ind, vexpcal);
  return *vexpcal;
}




/**
   @param vexpcal (in) a initialized signal (delay),
   @param val (in) the value of the delay of vexpcal.
   @param l_init (out) list of val last expressions of the initialisations of vexpcal.

   It calls dec_expr_simple_init() method  for each selected element.
*/
void
downstreamRetiming::mklistinit (ptexpcal vexpcal, int val, ptlinstances * l_init)
{
  ptexpcal explinit, expinit;
  int i;

  for (i = val; i > 0; i--)
    {
      explinit = EM->getInitialValue (vexpcal);
      expinit = dec_expr_simple_init (i - 1, &explinit);
      *l_init = ptlinstances_push_front (*l_init, expinit);
    };

}


/**
   @return a list of vectorial elements \n
   <em> [cptinit]:liste_exp_1, ..., [cptinit+(n-1)]:liste_exp_n </em>
   where liste_exp_i denotes the i-th element of the operand list (liste_exp)
*/
ptexpcal
downstreamRetiming::mkexpcalinit (int cptinit, ptlinstances * liste_exp)
{
  ptexpcal vexpcal, expd, expval, expval1, expinit, expinitres;

  ptlinstances_get_next (liste_exp, &vexpcal);
  EM->getEquivConstant (vexpcal, &vexpcal, pk_true);
  expval = EM->mkIntegerConstant (cptinit);
  expval1 = EM->mkNaryExp (oplesign, expval, NILEXPCAL, NILBACKANNOTATION);
  expinit = EM->mkBinaryExp (opelevec, expval1, vexpcal, NILBACKANNOTATION);
  if ((*liste_exp) != NILLINSTANCES)
    {
      expd = mkexpcalinit ((cptinit + 1), liste_exp);
      expinitres = EM->mkNaryExp (oplelevec, expinit, expd, NILBACKANNOTATION);
    }
  else
    expinitres = EM->mkNaryExp (oplelevec, expinit, NILEXPCAL, NILBACKANNOTATION);
  return expinitres;
}

/**
   @return an expression of initialisation using l_liste_init 
   and takes into account as model vexpcal (for the vectorial elements).

   - For each element of the operand list (l_liste_init):
      - it extracts an element (mkinitialisation()),
      - it builds the initilaisation expression (mkexpcaldinit())
      - the result of the previous operation is added to a list.
   
   - the result of the previous operation is used to finalize the initialisation expression (mkexpcalinit())
   - the result of the previous operation is simplified (simplifyExpr()).
   
*/
ptexpcal
downstreamRetiming::mkinitialisation (ptexpcal vexpcal, ptexpcal * l_liste_init)
{
  ptlinstances liste_oper, liste_oper_res;
  ptexpcal exp, expres;
  boolean b;

  b = pk_true;
  liste_oper_res = NULL;

  while (b)
    {
      b = extractlistfromptexpcal (l_liste_init, &liste_oper);
      exp = mkexpcaldinit (vexpcal, &liste_oper);
      liste_oper_res = ptlinstances_push_front (liste_oper_res, exp);
    };
  liste_oper_res = ptlinstances_reverse (liste_oper_res);
  expres = mkexpcalinit (0, &liste_oper_res);
  return simplifyExpr (expres);
}

/**
   @return "true" if all the elements of the list (vexpDim) are valuated to 0, "false" otherwise.
 */
boolean
downstreamRetiming::listOfZero (ptexpcal vexpDim)
{
  ptexpcal anElem;
  while (EM->getNaryArgs (&vexpDim, &anElem))
    {
      if (EM->getOperatorClass (anElem) != constant)
	return pk_false;
      if (!EM->IsAssigned (anElem))
	return pk_false;
      if (EM->getValue (anElem) != 0)
	return pk_false;
    }
  return pk_true;
}

/**
   It returns a simplified expression of the operand one (vexpinit).
   The simplification is applied for a list of vectorial elements: for [0,...0]:valexp, it is simplified in valexp.
 */
ptexpcal
downstreamRetiming::simplifyExpr (ptexpcal vexpinit)
{
  noperators op;
  ptexpcal listExp, anElem, vexpDim, vexpValue;

  if (EM->isNILEXPCAL (vexpinit))
    return vexpinit;
  op = EM->getOperator (vexpinit);
  if (op != oplelevec)
    return vexpinit;
  listExp = vexpinit;
  if (!EM->getNaryArgs (&listExp, &anElem))
    return vexpinit;
  if (!EM->isNILEXPCAL (listExp))
    return vexpinit;
  if (EM->getOperator (anElem) != opelevec)
    return vexpinit;
  /* anElem ::=  ele_vec (<expr_dimension>, <expr_signal>) */
  EM->getBinaryArgs (anElem, &vexpDim, &vexpValue);
  if (listOfZero (vexpDim))
    return vexpValue;
  return vexpinit;
  //yyy;
}

/**
   @return the expression \e EXPCAL $ \e VAL_DELAY window \e 0 init \e vexpinit.
*/
ptexpcal
downstreamRetiming::mkexpcaldelay (ptexpcal vexpcal, ptexpcal vexpinit, int val_delay)
{
  ptexpcal expdelay, expval, expwind;

  expval = EM->mkIntegerConstant (val_delay);
  expwind = EM->mkIntegerConstant (0);
  expdelay = EM->mkTernaryExp (opret, EM->mkBinaryExp (opinit, vexpcal, vexpinit, NILBACKANNOTATION), expval, NILEXPCAL, NILBACKANNOTATION);

  return expdelay;
}


/**
   @return a clone of a synchronous expression (vexpcal) in which 
   the leafs are replaced by the elements of loperres (keeping the ordrer).
   The vexpcal expression is travelled from left to right.
*/
ptexpcal
downstreamRetiming::mkexpcaldinit (ptexpcal vexpcal, ptlinstances * loperres)
{
  ptexpcal expres, expres1, expres2;
  ptexpcal exp1, exp2;
  noperators op;
  cloper cl;

  op = EM->getOperator (vexpcal);
  cl = EM->getOperatorClass (vexpcal);

  switch (cl)
    {
    case constant:
    case variabl:
    case call:
      ptlinstances_get_next (loperres, &expres);
      break;
    case oper1:
      exp1 = EM->getUnaryArg (vexpcal);
      expres1 = mkexpcaldinit (exp1, loperres);
      expres = EM->mkUnaryExp (op, expres1, EM->getBackAnnotation (vexpcal));
      break;
    case oper2:
      exp1 = EM->getBinaryArgstAt (vexpcal, 1);
      expres1 = mkexpcaldinit (exp1, loperres);
      exp2 = EM->getBinaryArgstAt (vexpcal, 2);
      expres2 = mkexpcaldinit (exp2, loperres);
      expres = EM->mkBinaryExp (op, expres1, expres2, EM->getBackAnnotation (vexpcal));
      break;
    default:;
    };
  return expres;
}
