/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Constructor.
   The functionality is applied to a graph (g).
*/
  downstreamRetiming::downstreamRetiming (ptDCG g)
{
  AGRAPH = g;
  EM = g->getExpressionManager ();
  CKM = g->getClockManager ();
  NM = g->getNodeManager ();
}

/**
   Destructor.
*/
downstreamRetiming::~downstreamRetiming ()
{
}


/**
   It rewrites the expressions such that y:=f(zx1,....zxn) with zxk = xk $ ik init (...)
   into  y := y' $ j init (...)  with 
   y' := f(zx1',..., zxn') and zxk' := Xk $ ( ik -j) init (...)
   
   Algorithm : for each clock of he clock system, it visits it associated sub graph (retimingClock())..
*/
void
downstreamRetiming::applyRetiming ()
{
  ptinsth vh;
  ptlinsth lh;

  retimMark = getMark ();

  lh = CKM->getlhroots ();
  while (ptlinsth_get_next (&lh, &vh))
    if (vh == CKM->getRepresentativeFormula (vh))
      retimingClock (vh);

  lh = CKM->getlhplaced ();
  while (ptlinsth_get_next (&lh, &vh))
    if (vh == CKM->getRepresentativeFormula (vh))
      retimingClock (vh);

  applyRetimingFunctionCall ();
}
