/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return "true" if the operation succeeds, "false" otherwise (cycle on the references).
   It builds the delay trees of the selected graph in the constructor. 
   
   The delays build from the same signal are represented by a tree with the signal as root.
   For each delay node, it calls the addDelay(). Then, verifications on the compatibility of the
   initial values are made (verifyInitCompatibility() method).

   It can be used by the code generators for example (or unifications) for optimizings: a tree can be 
   implemented as an array, the signals (nodes of the tree) can be managed circularly using indexes on this area.

   \code
   Example: zx:=x$1, zx1:=x$3, zzx:= zx $1 can be represented by a tree with :
   - x as root
   - zx and zx1 are children of x
   - zzx a child of zx.

   If the initialisations are compatibles, the signals (x, zx,zzx, zx1) can be represented by an array of []
   
   \endcode   

*/
  boolean mkDelayTree::delayTreeBuilder ()
{
  boolean
    pb;
  ptlINode
    lNodes;
  ptINode
    n;

  pb = pk_false;
  if (NM->getDelays (&lNodes))
    {
      lmemtrees = AGRAPH->getDelaysTree ();
      while (ptlINode_get_elem (&lNodes, &n))
	if (!addDelay (n))
	  pb = pk_true;

      if (!pb)
	verifyInitCompatibility ();
      AGRAPH->setDelaysTree (lmemtrees);
    }
  return !pb;
}


/**
   Constructor.
   The functionality is applied to a graph (vGraph)
*/
mkDelayTree::mkDelayTree (ptDCG vGraph)
{
  EM = vGraph->getExpressionManager ();
  CKM = vGraph->getClockManager ();
  NM = vGraph->getNodeManager ();
  AGRAPH = vGraph;
}

/**
   TODO
*/
void
mkDelayTree::printDelayTree ()
{
  ptlINode lNodes;
  ptINode n;


  lmemtrees = AGRAPH->getDelaysTree ();
  while (ptlINode_get_elem (&lNodes, &n));
}
