/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return
   - nouse if the signal vexp is not used in an external call.
   - useout if it is defined by an external call.
   - usein if it is an input of an external call.

   The interface of externall call  must be respected in code generation step.
*/
  typuseinec mkDelayTree::isuseinec (ptexpcal vinst)
{
  ptINode
    n1,
    vinstr;
  boolean
    find;
  ptexpcal
    lent,
    lsort,
    par;
  ptident
    nom;
  ptldependence
    ldep;
  ptdependence
    adep;

  vinstr = EM->getNode (vinst);
  if (EM->getNumberOfDimensions (vinst) == 0)
    return nouse;
  if (NM->isProcessCall (vinstr))
    find = pk_true;
  else
    {
      find = pk_false;
      ldep = NM->get_successors (vinstr);
      while (!find && (ldep->get_next (&adep, &ldep)))
	{
	  adep->getDependence (&n1, &vinstr);
	  find = NM->isProcessCall (vinstr);
	}
    }

  if (!find)
    return nouse;
  NM->get_procCall_arg (vinstr, &nom, &par, &lent, &lsort);
  find = pk_false;
  while (!find && (lsort != NILEXPCAL))
    find = (vinst == EM->getNaryArgs (&lsort));
  if (find)
    return useout;
  return usein;
}

/** @return "true" if it exists a cycle in the references in a delay tree, "false" otherwise.
    It searches if the signal (vsigdelay) is a signal associated with a delay tree node of the branch starting at the 
    node vmemsig until the root of its delay tree.
 */
boolean mkDelayTree::memnodeCycle (ptmemnode vmemsig, ptexpcal vsigdelay)
{
  ptmemnode
    vmemnode;

  vmemnode = ptmemnode_get_father (vmemsig);
  while (vmemnode != NILMEMNODE)
    {
      if (ptmemnode_get_signal (vmemnode) == vsigdelay)
	return pk_true;
      vmemnode = ptmemnode_get_father (vmemnode);
    }
  return pk_false;
}


/**
   It makes or merges the delay node <em> vresdelay := vsigdelay $ vret window vfen </em> in the delay tree forest.
   It takes into account the different cases (vresdelay already or not in a delay tree, vsigdelay already or not in a delay tree)
   and verifies that there is not a cycle on the references.
   @return "true" if the operation succeeds, "false" otherwise (cycle).
*/
boolean mkDelayTree::ptlmemnode_add_delay (ptlmemnode * lmemtrees, ptexpcal vresdelay, ptexpcal vsigdelay, ptexpcal vret, ptexpcal vfen)
{
  // It returns always pk_true.

  ptmemnode
    vmemsigdelay,
    vmemres;

  if (!EM->getSignalAssignedTo (vsigdelay, &vsigdelay))
    return pk_true;
  EM->getSignalAssignedTo (vresdelay, &vresdelay);

  vmemres = ptexpcal_get_node_tree_delay (vresdelay);
  vmemsigdelay = ptexpcal_get_node_tree_delay (vsigdelay);

  if ((vmemres == NILMEMNODE) && (vmemsigdelay == NILMEMNODE))
    {
      /* pas de noeud ni pour RES ni SIG : => création des noeuds */
      vmemres = makeMemnode (vresdelay, isuseinec (vresdelay));
      vmemsigdelay = makeMemnode (vsigdelay, isuseinec (vsigdelay));
      /* => assignation du retard et du fenetrage */
      ptmemnode_set_delayWindow (vmemres, vret, vfen);
      /* accrochage de RES sous SIG */
      ptmemnode_insert_child (vmemsigdelay, vmemres);
      /* => nouvelle racine de l'arbre des retards */
      *lmemtrees = ptlmemnode_insert_elem (vmemsigdelay, *lmemtrees);
      return pk_true;
    }

  if (vmemres == NILMEMNODE)
    {
      /* pas de noeud pour RES : => création */
      vmemres = makeMemnode (vresdelay, isuseinec (vresdelay));
      /* => assignation du retard et du fenetrage */
      ptmemnode_set_delayWindow (vmemres, vret, vfen);
      /* accrochage de RES sous SIG */
      ptmemnode_insert_child (vmemsigdelay, vmemres);
      return pk_true;
    }

  if (vmemsigdelay == NILMEMNODE)
    {
      /* => assignation du retard et du fenetrage */
      ptmemnode_set_delayWindow (vmemres, vret, vfen);
      /* pas de noeud pour SIG : => création */
      vmemsigdelay = makeMemnode (vsigdelay, isuseinec (vsigdelay));
      /* accrochage de RES sous SIG */
      ptmemnode_insert_child (vmemsigdelay, vmemres);
      /* accrochage de SIG àl'ancienne place de RES */
      sublmemnode (*lmemtrees, vmemres, vmemsigdelay);
      return pk_true;
    }

  /* Cycle : for example x := y $1 init V0 | y := x $ 1 init V1. */
  if (!memnodeCycle (vmemsigdelay, vresdelay))
    {
      /* => assignation du retard et du fenetrage */
      ptmemnode_set_delayWindow (vmemres, vret, vfen);
      /* accrochage de RES sous SIG */
      ptmemnode_insert_child (vmemsigdelay, vmemres);
      /* suppression de RES de la liste des racines */
      *lmemtrees = ptlmemnode_remove_elem (*lmemtrees, vmemres);
    }
  return pk_true;
}

/** @return "true" if the operation succeeds, "false" otherwise.
    It adds ( ptlmemnode_add_delay() method) the node (i), assumed to be a delay node, in the delay tree if the
    clock of the delay is not the null clock.
*/
boolean mkDelayTree::addDelay (ptINode i)
{
  ptexpcal
    vret,
    vfen,
    exp1,
    exp2,
    vinit;

  NM->get_delay_arg (i, &exp1, &exp2, &vret, &vfen, &vinit);
  if (CKM->satisfiesNullClock (EM->getClock (exp1)))
    return pk_true;
  return ptlmemnode_add_delay (&lmemtrees, exp1, exp2, vret, vfen);
}



/** It removes a sub tree with mem as root from it father (mempere). The removed sub tree becomes a
    child of an "equivalent" of its father.
    The equivalent 
    - is created if it does not exists, it is added to the lmemtrees list;
    - must have a initial value compatible with initssarbre.
*/
void
mkDelayTree::dettachchild (ptmemnode mempere, ptmemnode mem, ptexpcal initssarbre, ptlmemnode * lmemtrees)
{
  ptlmemnode lmn;
  ptmemnode memsecroot;
  boolean trouve;

  lmn = ptmemnode_get_equiv (mempere);
  trouve = pk_false;
  while (ptlmemnode_getNaryArgs (&lmn, &memsecroot) && !trouve)
    {
      if (EM->IsCompatibleWith (ptmemnode_get_init_exp (memsecroot), initssarbre))
	if (EM->IsCompatibleWith (ptmemnode_get_init_exp (ptlmemnode_get_elem (ptmemnode_get_lchild (memsecroot))), initssarbre))
	  trouve = pk_true;
    }
  if (!trouve)
    {
      memsecroot = ptmemnode_equiv_copy (mempere);
      *lmemtrees = ptlmemnode_insert_elem (memsecroot, *lmemtrees);
    }
  ptmemnode_remove_child (mempere, mem);
  ptmemnode_insert_child (memsecroot, mem);
}

/** 
    It verifies the compatibility (ExprManager::IsCompatibleWith()) of the initial values for a sub tree with mempere as root.
    This operation can induce (see dettachchild()) some splitting of the delay tree (when the initialisations are incompatible)
*/
void
mkDelayTree::verifinit (ptmemnode mempere, integer nbfen, ptexpcal * initssarbre, ptlmemnode * lmemtrees)
{
  ptlmemnode lmem;
  ptmemnode mem;
  ptexpcal initfrere;
  boolean okbrothers, okfather, ok;
  integer nbfen1, delaypere, delaycurrent;
  ptexpcal vCellIni;

  lmem = ptmemnode_get_lchild (mempere);
  initfrere = NILEXPCAL;
  while (ptlmemnode_getNaryArgs (&lmem, &mem))
    {
      if (ptmemnode_get_window_value (mem) > 1)
	nbfen1 = nbfen + 1;
      else
	nbfen1 = nbfen;
      verifinit (mem, nbfen1, initssarbre, lmemtrees);
      delaypere = ptmemnode_get_delay_value (mempere);
      delaycurrent = ptmemnode_get_delay_value (mem);
      if ((delaypere == 0) && (delaycurrent == 0))
	okfather = EM->IsCompatibleWith (ptmemnode_get_init_exp (mempere), *initssarbre);
      else
	okfather = pk_true;
      if (okfather && (nbfen1 > 1))
	okfather = EM->IsConstantExpr (ptmemnode_get_init_exp (mem));
      okbrothers = EM->IsCompatibleWith (*initssarbre, initfrere);
      ok = okbrothers && okfather;
      if (ok)
	{
	  if (initfrere == NILEXPCAL)
	    initfrere = *initssarbre;
	  vCellIni = ptmemnode_get_cell_init_exp (mem);
	  ok = EM->IsCompatibleWith (vCellIni, initfrere);
	  if (ok && (nbfen1 > 1))
	    okfather = EM->IsConstantExpr (vCellIni);
	  // Deja fait (2 lignes au dessus ) : if (ok) ok = EM->IsCompatibleWith(vCellIni, initfrere);
	  //ATTENTION : if (!(okbrothers && okfather)) dettachcellag(mem);
	}
      else
	dettachchild (mempere, mem, *initssarbre, lmemtrees);
    }
  if (initfrere == NILEXPCAL)
    *initssarbre = ptmemnode_get_init_exp (mempere);
  else
    *initssarbre = initfrere;
}


/** It verifies the compatibility of the initial values (see verifinit() method) for each tree of the list
    of delay trees (lmemtrees member).
*/
void
mkDelayTree::verifyInitCompatibility ()
{
  ptlmemnode vlmemt;
  ptmemnode vmemtree;
  ptexpcal expbid;

  vlmemt = lmemtrees;
  expbid = NILEXPCAL;
  while (ptlmemnode_getNaryArgs (&vlmemt, &vmemtree))
    verifinit (vmemtree, 0, &expbid, &lmemtrees);
}

/** For a delay (node n), x := y $ M init [...], a cycle has been detected in the tree of delays:
    The signal y is an element of the delay tree with x as root.
\verbatim
    Example, x := y $1 init x0 | y := x $1 init y0.
         It is equivalent to x := (x $2) init x0, y0
         x is then defined by 
            (| x := U[i] 
             | U := [x0, y0] 
             | i := 0 when (zi = 2) default zi +1 | zi := i $1 init 2
             | i^= U ^=x
             |)
\endverbatim
  */
boolean mkDelayTree::substitute (ptINode n)
{
  return pk_false;
}
