/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   @return a new signal, with vAt as clock, defined by a cell expression vx when vhx cell vAt.
*/
  ptexpcal statesVariables::makeCellAt (ptexpcal model, ptexpcal vx, ptexpcal vinit, ptinsth vhx, ptinsth vAt)
{
  ptexpcal mx;
  ptINode node;

  mx = mkCopyIdentExp (model, vAt);
  EM->setClock (mx, vAt);
  EM->setInitialValue (mx, vinit);
  node = NM->mk_node_equation (mx, vx, vhx, EM->getSignalsReferencedIn (vx), NILBACKANNOTATION);
  NM->insertNode (node, vAt);
  return mx;
}

/** Extension for a cell.
    @return "true" if the operation succeeds, "false" otherwise.
    It extends the cell definition (node) to the clock vhto.
    In  the cell are rewritten in Basic Signal (cellModeTranslation attribute),
    it applies makeMemoryofAt() method, otherwise makeCellAt() method.
*/
boolean
statesVariables::extendACellToClock (ptINode node, ptinsth vhto)
{
  ptexpcal vx, vy, mx, zmx;
  ptinsth vhy, vhx;

  vy = NM->get_leftOperand (node);
  NM->get_rightOperand (node, &vx);
  vx = EM->getWhenOrCellArg (vx);
  vhx = EM->getContributionClock (vx);
  vx = EM->getBasicExpr (vx);
  if (cellModeTranslation)
    zmx = makeMemoryofAt (vx, vhx, EM->getInitialValue (vy), vhto, &mx);
  else
    mx = makeCellAt (vy, vx, EM->getInitialValue (vy), vhx, vhto);

  NM->resetNodeForModification (node);	// removeNode(node);
  vhy = EM->getClock (vy);
  NM->modifiedDefinition (vy, mx, vhy, vhy);
  return pk_true;
}

/** It applies extendACellToClock() method to each cell of a list of nodes (linstr).
    @return "true" if the operation succeeds, "false" otherwise.
    It extends the cell definitions to the clock vhto.
*/
boolean
statesVariables::extendCellsToClock_loc (ptlINode linstr, ptinsth vhto)
{
  ptINode vinstr;
  boolean b;

  b = pk_true;
  while (ptlINode_get_elem (&linstr, &vinstr) && b)
    if (NM->is_cellNode (vinstr))
      b = extendACellToClock (vinstr, vhto);
  return b;
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It applies extendCellsToClock_loc() method for the nodes associated with a clock (vhfrom)
   It extends the cell definitions to the clock vhto.
*/
boolean
statesVariables::extendCellsToClock (ptinsth vhfrom, ptinsth vhto)
{
  return extendCellsToClock_loc (getNodesAssignedTo (vhfrom), vhto);
}


/**
   @return "true" if the operation succeeds, "false" otherwise.
   It applies extendCellsToClock() method for each clock of the hierarchy dominated by vhfrom.
   For each cell of the hierarchy, it extends the cell definition to the clock vhto.
   Assumed: vhfrom is under vhto in the hierarchy of clocks.
*/
boolean
statesVariables::extendCellFromClockToClock (ptinsth vhfrom, ptinsth vhto)
{
  ptinsth vh;
  ptlinsth lchild;
  boolean b;

  /* for a clock (vhfrom) */
  if (vhfrom != vhto)
    b = extendCellsToClock (vhfrom, vhto);
  else
    b = pk_true;
  if (b)
    {
      /* For all children of clock of vhform */
      lchild = CKM->getChildren (vhfrom);
      while (ptlinsth_get_next (&lchild, &vh) && b)
	b = extendCellFromClockToClock (vh, vhto);
    }
  return b;
}
