/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

static ptClockManager CKM;

/** It travels the hierarchy with h as root and computes the number of node by level in vinfo. */
static void
traverse (ptinsth h, list_type vinfo)
{
  ptlinsth children;
  ptinsth e;
  long int counter;

  if (list_is_at_end (vinfo))
    {
      list_insert_last (vinfo, (itemValue_type) 0);
      list_set_to_last (vinfo);
    }
  counter = (long int) list_get_current (vinfo);
  counter++;
  list_set_current (vinfo, (itemValue_type) counter);
  children = CKM->getChildren (h);
  while (ptlinsth_get_next (&children, &e))
    {
      list_forward (vinfo);
      traverse (e, vinfo);
      list_backward (vinfo);
    }
}

/**
   It travels the hierarchy with sr as root (traverse()) and produces
   some statistics about the hierarchy: number of clocks, depth, breadth
*/
static void
hierstats (ptinsth sr)
{
  long int curr;
  int breadth, depth, size;
  list_type vinfo;

  vinfo = list_create ();
  traverse (sr, vinfo);
  list_set_to_first (vinfo);
  breadth = 0;
  depth = 0;
  size = 0;
  while (!list_is_at_end (vinfo))
    {
      curr = (long int) list_get_current (vinfo);
      size = size + curr;
      if (breadth < curr)
	breadth = curr;
      depth++;
      list_forward (vinfo);
    }
  pK_messenger->globalMessage (infostatistics);
  pK_messenger->globalMessageInt (infonbclocks, size);
  pK_messenger->globalMessageInt (infodepthStat, depth);
  pK_messenger->globalMessageInt (infobreadth, breadth);

}
