
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __DCG_H__
#define __DCG_H__

enum GraphLevelType {DCG_BASIC, DCG_POLY, DCG_ENDO, DCG_BOOL, DCG_FLAT, DCG_SCH};
typedef enum {FiacreTypes} SelectedClassType; 

#include "Common_DCG_Automaton.h"
#include "Automaton.h"

class DCG: public ObjectCommonInfos, public Common_DCG_Automaton
{
#include "DCG_LL_m.h"
#include "DCG_m.h"
#include "DCG_LL_g.h"
#include "DCG_g.h"

 protected:
  GraphLevelType vgraph_level;  ///< Sub-format of the graph.

  noperators kind;   ///< Type of the graph : process, function, node, module.
  ptlINode list_extern_proc; ///< List of external procedure calls.
  ptlinstances list_extern_funcs;  ///< list of external referenced in expressions (functions)
  ptlINode lNodesAtInitClock; ///< the list of the nodes at the initial clock (constant signals).

  ptlinstances list_indices;///< List of the indexes.
  ptlINode unusedNodes ; ///< List of the unused nodes.
  ptlinsth unusedClocks;  ///< List of the unused clocks.
  ptlmemnode delays_tree;  ///< List of the delay trees.
  
  ptlinstances lSigNullClock;  ///< list of signals at null clock.
  ptlinsth lConstraintsNullClock;   ///< List of clock formula to be proved equivalent with the null clock.
  
  boolean externalNode; ///< set to "true" for an external graph.
  ntrees SIGNAL_Tree; ///< Acces to the representation of the graph as a SIGNAL abstract tree.

  ptlPointer cycles; /// list of the detected cycles in the DCG.
  ptlPointer PseudoCycles; /// list of the null clock cycles in the DCG.
  boolean modeVerbose; ///< verbose mode when it is true (trace of the functionality).

  ptinsth /* idle,*/ step, iostep, complete;
  ptAutomaton vautomaton; ///< associated automaton.
  ptlinstances lssiglocgc;   ///< list of all the local signals of the graph.

  // DCG_SCH infos.(It should be removed from here)
  ptelemTableGenCode *SubGraphTable;   ///< Access to the list of ordered subgraphs.
  int nbSets;  ///< Number of elements in the SubGraphTable.
  ptlINode lNodesHinit;  ///< Access to the list of the node which defines constant signals.
  ptlinstances lClocksNotToBeInitiliazed;   ///< set of boolean clocks that must not be initialized to "bottom"
  static boolean bSuspendClockNameAction;  ///< global name when true the assigment of an identifier to the clocks is suspended.
  boolean checkExplicitTypingApplied;
};


#endif // __DCG_H__
