
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ Common_DCG_Automaton::Common_DCG_Automaton ()
{
}

Common_DCG_Automaton::~Common_DCG_Automaton ()
{
  delete aExprManager;
  delete aClockManager;
  delete aNodeManager;
  //  delete aTraceabilityManager;
  // a voir pour lstatesvar, locGraphs;
}

void
Common_DCG_Automaton::initialyze (ptNameManager nm)
{
  aNameManager = nm;
  aExprManager = new ExprManager ();
  aClockManager = new ClockManager ();
  aNodeManager = new NodeManager ();
  aTraceabilityManager = new TraceabilityManager (aNodeManager, aClockManager, aExprManager);
  aNodeManager->initialize(aClockManager, aExprManager, aTraceabilityManager);
  aExprManager->initialize (aNodeManager, aClockManager, aNameManager, aTraceabilityManager);
  aClockManager->initialize (aExprManager);
  locGraphs = NILLDCG;
  lstatesvar = NILLINSTANCES;
}

/** id becomes the name of the object. */
void
Common_DCG_Automaton::setName (ptident id)
{
  vname = id;
}

/** @return the name of the graph, previously set by setName(). */
ptident Common_DCG_Automaton::getName ()
{
  return vname;
}

/** 
    @return the clock system manager of the graph.
*/
ptClockManager Common_DCG_Automaton::getClockManager ()
{
  return aClockManager;
}

/** @return the node manager of the graph.*/
ptNodeManager Common_DCG_Automaton::getNodeManager ()
{
  return aNodeManager;
}

/** @return the expresion manager of the graph.*/
ptExprManager Common_DCG_Automaton::getExpressionManager ()
{
  return aExprManager;
}

/** @return the name manager of the graph.*/
ptNameManager Common_DCG_Automaton::getNameManager ()
{
  return aNameManager;
}


/** @return the traceability manager of the graph.*/
ptTraceabilityManager Common_DCG_Automaton::getTraceabilityManager ()
{
  return aTraceabilityManager;
}


/** @return the set of state variables of the program.*/
ptlinstances Common_DCG_Automaton::getStatesVar ()
{
  return lstatesvar;
}

/** l (in): set of state variables. */
void
Common_DCG_Automaton::setStatesVar (ptlinstances l)
{
  lstatesvar = l;
}

/** It adds vsig in the set of state variables.*/
void
Common_DCG_Automaton::addStatesVar (ptexpcal vsig)
{
  ptlinstances bidon, liste;
  liste = lstatesvar;
  if (ptlinstances_isin_insert_exp (&liste, &bidon, vsig))
    lstatesvar = liste;
}

/**
   Management of the graphs of a module or the states of an automaton.\n
   It adds a graph (vgraph) as a graph of a model defined in the module/automaton.
*/
void
Common_DCG_Automaton::addLocalGraph (ptDCG vgraph)
{
  ptlPointer_insert_elem (&locGraphs, (attributeValue) vgraph);
}


/**
   @return the local graphs.
*/
ptlDCG
Common_DCG_Automaton::getLocalGraphs ()
{
  return locGraphs;
}


/** 
    @return true if it exists non predefined types (returned in lTypes) for the current object(DCG, Automaton), 
    false otherwise.
*/
boolean
Common_DCG_Automaton::getUserTypes (ptlinstances * lTypes)
{
  return getExpressionManager ()->getUserTypes (lTypes);
}
