

/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define GET_ROOT_INTERFACE() getClockManager()-> getInterfaceSubHierarchy(getRootClock())


_PRIVATE_

boolean DCG::bSuspendClockNameAction = pk_false;	// static member.


/**
   Initializing an empty graph. It creates new managers (clock manager, node manager, 
   expression manager, name manager) and it declares it as a DCGBasic one.
 */
void
DCG::initialyze (ptNameManager nm)
{
  setwhoami (graphObject);
  Common_DCG_Automaton::initialyze (nm);
  setFormatLevel (DCG_BASIC);

  vname = getIdentNil ();
  unusedNodes = NILLINODE;
  unusedClocks = NILLINSTH;
  delays_tree = NILLMEMNODE;

  list_extern_proc = NILLINODE;	/* external procedure call referenced as nodes */
  list_extern_funcs = NILLINSTANCES;	/* list of external referenced in expressions (functions) */
  list_indices = NILLINSTANCES;	/* Indexes List declarations */

  lSigNullClock = NILLINSTANCES;	/* Signals synchronous with the null clock */
  lNodesAtInitClock = NILLINODE;	// constant signals.
  lConstraintsNullClock = NILLINSTH;	/* formula to be proved equivalent with the null clock. */
  externalNode = pk_false;
  SET_NILTREE (SIGNAL_Tree);

  cycles = NILLPOINTER;
  PseudoCycles = NILLPOINTER;

  modeVerbose = pk_false;
  // Predefined clocks are set to the undefined clock.
  iostep = aClockManager->getUndefClock ();
  step = iostep;
  complete = iostep;
  // idle = step;
  vautomaton = NILAUTOMATON;

  lssiglocgc = NILLINSTANCES;	/* used for code generators (only?) */
  lClocksNotToBeInitiliazed = NILLINSTANCES;
  lNodesHinit = NILLINODE;
  nbSets = -1;
  // *SubGraphTable= (ptelemTableGenCode) NILPOINTER;
  checkExplicitTypingApplied = false;
}

/**
   @return the list (ldep) + the typed (tdep) dependences assigned to a node (aNode).
*/
ptldependence DCG::getTypedDependencesNode (TypeDependenceKind tdep, ptINode aNode, ptldependence ldep)
{
  ptldependence ld;
  ptdependence aDep;
  ptINode vFrom, vTo;
  ptinsth vh1;
  ptexpcal lsig, aSig;

  require_managers (EM, NM, CKM);
  ld = NM->get_successors (aNode);
  while (ld->get_next (&aDep, &ld))
    if (aDep->getDependenceKind () == tdep)
      {
	vh1 = aDep->getDependence (&vFrom, &vTo);
	vh1 = CKM->getRepresentativeFormula (vh1);
	aDep->setClockDependence (vh1);
	ldep = ldep->insert (aDep);
      }

  if (NM->isNodeCall (aNode))
    {
      lsig = NM->get_leftOperand (aNode);
      while (EM->getNaryArgs (&lsig, &aSig))
	ldep = getTypedDependencesNode (tdep, EM->getNode (aSig), ldep);
    }
  return ldep;
}

/**
   @return the list (ldep) + the typed (tdep) dependences assigned to a set of nodes (lNodes).
*/
ptldependence DCG::getTypedDependencesNodes (TypeDependenceKind tdep, ptlINode lNodes, ptldependence ldep)
{
  ptINode
    aNode;

  while (ptlINode_get_elem (&lNodes, &aNode))
    ldep = getTypedDependencesNode (tdep, aNode, ldep);
  return ldep;
}

ptldependence DCG::getTypedDependences (TypeDependenceKind tdep, boolean b)
{
  ptldependence ldep;
  ptlINode lNodes;

  ldep = NILLDEPENDENCE;
  if (aNodeManager->getInputs (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  if (aNodeManager->getEquations (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  if (aNodeManager->getDelays (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  if (aNodeManager->getProcCalls (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  if (aNodeManager->getNodeCalls (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  if (aNodeManager->getSubGraphs (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  if (aNodeManager->getGroupsOfPartialEquations (&lNodes))
    ldep = getTypedDependencesNodes (tdep, lNodes, ldep);
  // if (b) if (aNodeManager->getOutputs( &lNodes)) ldep = getTypedDependencesNodes(tdep, lNodes, ldep);
  return ldep;
}


void
DCG::setClocksName (ptlinsth lh, boolean breprBySignal)
{
  ptinsth h;

  while (ptlinsth_get_next (&lh, &h))
    setClockName (h, breprBySignal);
}


/**
   For a clock (clk), it sets names to its equivalent formulae, and of its associated assertions.
*/
void
DCG::setNamesConstraintsAssertions (ptinsth clk)
{
  ptident id, id1;
  ptinsth clk1;

  require_clock_manager (CKM);
  id = CKM->getName (clk);
  // set a name to each formula of the class
  clk1 = CKM->getNextFormula (clk, pk_true);
  while (clk1 != clk)
    {
      id1 = CKM->getName (clk1);
      if (eq_idents (id1, NILIDENT) || eq_idents (id1, id))
	CKM->setName (clk1, getNameManager ()->new_pureClockIdent (CKM->getNum (clk1)));
      clk1 = CKM->getNextFormula (clk1, pk_true);
    }

  // set a name to each assertion of the class
  clk1 = clk;
  while (CKM->getNextAssertion (clk1, &clk1))
    {
      id1 = CKM->getName (clk1);
      if (eq_idents (id1, NILIDENT) || eq_idents (id1, id))
	CKM->setName (clk1, getNameManager ()->new_pureClockIdent (CKM->getNum (clk1)));
    }
}
