/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   @return "true" if  the translation of a DCGEndo (gendo) to a DCGBool (gb) is perfomed, "false" otherwise.
*/
    _STATIC_ boolean DCG_bool::transformToBooleanControl(ptDCG_endo gendo, ptDCG_bool * gb)
{
  // printf("DEBUG entering DCG_bool::transformToBooleanControl\n"); fflush(stdout);
  if (!gendo->DCG_endo::transformToBooleanControl())
    return pk_false;
  *gb = new DCG_bool(gendo);
  if (!(*gb)->setArrayClocks())
    return pk_false;		// Added March 18 2014
  (*gb)->setComposedNodesDependences();
  (*gb)->buildClockProcessInterface();
  return pk_true;
}

 DCG_bool::DCG_bool(ptDCG_endo gendo):DCG_endo(gendo)
{
  gendo->setFormatLevel(DCG_BOOL);
}

DCG_bool::DCG_bool()
{
}

/**
   \copydoc DCG::mkBooleanClockScopeExtand()
*/
ptinsth DCG_bool::mkBooleanClockScopeExtand(ptinsth h1, ptinsth h2)
{
  return getDCG()->mkBooleanClockScopeExtand(h1, h2);
}

/** 
   This  method can be applied to a graph of a module (set of processes) or a process.
   @param bstrong : at pk_true, the clock of the dependences are taken into account.
   \copydoc buildIOClockSubHierarchyOneLevel
 */
boolean DCG_bool::buildIOClockSubHierarchy(boolean bstrong)
{
  for_all_graphs_do(buildIOClockSubHierarchyOneLevel(bstrong));
  return pk_true;
}

/** 
   This  method can be applied to a graph of a module (set of processes) or a process.

   \copydoc totallyOrderOneLevel
   The assert nodes are visible for code generation when bCheck (code will be generated for verifying assumptions at run time).
   @return the value "true" if the translation is executed, false when the graph is a DC+ one.
*/
boolean DCG_bool::Ordering(boolean btotally, boolean bCheck)
{
  ptrepstr codeOp;

  codeOp = string2repstr(csttotallyOrder);
  if (isVerboseMode())
    pK_messenger->messagestr(mbeginOperName, codeOp);
  for_all_graphs_do(OrderingOneLevel(btotally, bCheck));
  setFormatLevel(DCG_SCH);
  if (isVerboseMode())
    pK_messenger->messagestr(mendOperName, codeOp);
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc flattenCHOneLevel
*/
boolean DCG_bool::flattenCH()
{
  ptDCG g;
  pointer elem;
  boolean bModule;
  ptlDCG lg, lg1;
  int vmyMark;
  ptrepstr codeOp;

  codeOp = string2repstr(cstflattenCH);
  if (isVerboseMode())
    pK_messenger->messagestr(mbeginOperName, codeOp);
  vmyMark = getMark();
  bModule = getModuleGraphs(&lg);
  if (!bModule)
    ptlPointer_insert_elem(&lg, (pointer) getDCG());
  while (ptlPointer_get_elemNext(&lg, &elem))
    {
      g = (ptDCG) elem;
      if (!g->satisfiesMarked(vmyMark))
	{
	  g->addMark(vmyMark);
	  if (!g->isGraphOfExternalProcess())
	    {
	      init(g);
	      gw->flattenCHOneLevel();
	      if (g->getReferencedGraphs(&lg1))
		lg = ptlPointer_fusion(lg, lg1);
	    }
	  else
	    setFormatLevel(DCG_FLAT);
	}
      ptlPointer_remove_elem(&lg, elem);
    }
  resetMark(vmyMark);
  if (bModule)
    setFormatLevel(DCG_FLAT);
  if (isVerboseMode())
    pK_messenger->messagestr(mendOperName, codeOp);
  buildClockProcessInterface();
  return pk_true;
}

/** 
   This  method can be applied to a graph of a module (set of processes) or a process.

   \copydoc addExceptionsForConstraintsOneLevel

   @return true if the operation is done, false when the graph is a DC+ one.
 */
boolean DCG_bool::addExceptionsForConstraints()
{
  // printf("DEBUG entering DCG_bool::addExceptionsForConstraints \n"); fflush(stdout);
  for_all_graphs_do(addExceptionsForConstraintsOneLevel());
  // printf("DEBUG  exiting DCG_bool::addExceptionsForConstraints \n"); fflush(stdout);
  return pk_true;
}

/** 
    It performs the user-partitionning by analyzing the pragma "RunOn i " where i denotes a string. 
    Nodes annotated by the same string value are in the same class or sub-graph. 
    @param ln (out) is a list of "sub-graph" nodes produced by the partitionning.

    @return "true" if the operation succeeds (all objects are localized), "false" otherwise.

    \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module (set of processes).
 
*/
boolean DCG_bool::partitionningRunOn(ptlINode * ln)
{
  ptRunOnMark aRunOnMark;
  ptrepstr codeOp;
  boolean b;
 
  // Take care of shared signal (no reference to ?)
  require_node_manager(NM);
  codeOp = string2repstr(cstRunOnPartition);
  pK_messenger->messagestr(messOperName, codeOp);
  makeNewInputs();
  aRunOnMark = new RunOnMark(getDCG());
  if (aRunOnMark->colouring())
    {
      NodesInterfaceCalculus();
      *ln = aRunOnMark->partitionning();
      b = pk_true;
    }
  else
    {
      pK_messenger->message(infoRunOnNotFound);
      b = pk_false;
    }

  delete aRunOnMark;
  if (isVerboseMode())
    pK_messenger->messagestr(mendOperName, codeOp);

  return b;
}
