

/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
*/
  Service::Service ()
{
  berror = pk_false;
  btranslation = pk_false;
  bproduce = pk_false;
}


/**
   It translates the operand graph (g) in a DCG_XXX graph (defined by the user).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
Service::translation (ptDCG g)
{
  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("Service::translation"));
  return pk_false;
}

/**
   It prints the text associated to the DCG_XXX.
*/
void
Service::print ()
{
  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("Service::print"));
}

/**
   @return "true" if the AST (vast) associated to the DCG_XXX is built, "false" otherwise.
*/
boolean
Service::produce_tree ()
{
  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("Service::produce_tree"));
  return pk_false;
}

/**
   It generates the text associated with the DCG_XXX graph. It calls produce_tree() followed by the AST::XXXUnparse.
*/
void
Service::unparse ()
{
  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("Service::unparse"));
}


/**
   It sets the error tag to pk_true.
*/
void
Service::setOnError ()
{
  berror = pk_true;
}

/**
   @return the status of the current service (false when an error is detected).
*/
boolean
Service::isOnError ()
{
  return berror;
}


/**
   It specifies that the translation method has been applied for the current service.
*/
void
Service::setTranslationApplied ()
{
  btranslation = pk_true;
}

/**
   @return "true" if the translation method is applied, "false" otherwise.
*/
boolean
Service::isTranslationApplied ()
{
  return btranslation;
}



/**
   It specifies that the produce_tree method has been applied for the current service.
*/
void
Service::setProduceApplied ()
{
  bproduce = pk_true;
}

/**
   @return "true" if the produce_tree method has been already applied, "false" otherwise.
*/
boolean
Service::isProduceApplied ()
{
  return bproduce;
}
