/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

typedef class kernel2graph *ptkernel2graph;
#include <map>
typedef std::map<TypeType, ptreptyp> BundleTableType;

/**
   Kernel to Graph translation.
*/
class kernel2graph
{
#include "kernel2graph_m.h"
#include "kernel2graphTypes_i.h"
#include "kernel2graphModule_i.h"
#include "kernel2graphEsig_i.h"
#include "kernel2graphEproc_i.h"
#include "kernel2graph_g.h"

private:
  static ptStack callsStack ; /**< Class variable, stack of th calls.*/

  ptDCG AGRAPH; /**< graph assigned to the compiled model.*/
  ptExprManager EM; /**< Expression manager assigned to the AGRAPH graph.*/
  ptClockManager CKM; /**< Clock manager assigned to the AGRAPH graph.*/
  ptNodeManager NM; /**< Node manager assigned to the AGRAPH graph.*/
  ptNameManager nameManager;/**< Name manager assigned to the AGRAPH graph.*/
  ptTraceabilityManager TM; /**< Traceability manager assigned to the AGRAPH graph.*/

  ptintemkIGraph igc; /**< graph constructor interface */
  ptGraphManager GRAPHS_MANAGER; /**< graphs manager. */

  /* kernel-tree model infos. */
  ptrepfi vfi; /**< model in kernel-tree representation. */
  ptklExprSig parglobal; /**< Expression of the parameters. */

  /* Graph infos */
  ptinsth hinit, the_nullClock_K2G_copy, hundef;	/**< predefined clocks. */
  ptexpcal vexptrue, vexpfalse, ptexp1, ptexp0; /**< predefined expressions: "true", "false", "1", "0".*/

  ptlinstances lparameters, linputs, loutputs, lstateVar;
  ntrees currentTree; /**< the current back annotation to the abtract tree. TO BE REMOVED and replaced by an access to back_annot_obj->get_Abstree() */
  BackAnnotationType back_annot_obj; /**< the current back annotation to the abtract tree(currentTree) and the call tree. */

  TypeType UndefinedType;
  typeMark INPUT_ADDED_FOR_UNEXPANDED, OUTPUT_ADDED_FOR_UNEXPANDED, XMARK, REF_STATEVAR_ATTR;
  boolean ACCESS_MODE_INPUT, 
    bWithAssertions, 
    inAssertBlock; /**< set to true for the composition of assert(|  |) */
  boolean isSignalThreadModel,///< true when a model is declared SIGNAL thread, false otherwise.
    isAutomatonModel;  ///< true when a model is an automaton, false otherwise.
  boolean btm; ///< traceability management (managed when btm is "true")
  integer numappel; // temporary
  static BundleTableType  BundleTable;

  ptinsth vhOversamplingSignalThread, vhDispatchSignalThread;

};
