/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Local to \ref sigProduceModule method.
   Compiling the models of the operand list (lsModel).
   @return "true" if the operation succeeds, "false" otherwise.
*/
  boolean kernel2graph::compilingModels (ptlrepid lsModel)
{
  int numero;
  boolean notEmpty, bres;
  ptrepfi aModel;
  ptlrepid vrepid;
  ptDCG vgraph;
  ptklExprSig param;
  ptkernel2graph translator;

  bres = pk_true;
  numero = 1;
  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = ptlrepid_getRepidAt (lsModel, numero, &vrepid);
      if (notEmpty)
	{
	  aModel = getModelRepid (vrepid);
	  param = NILEXPRSIG;
	  translator = new kernel2graph (GRAPHS_MANAGER, igc->getLabelStack (), btm);
	  if (!translator->kernelTree2DCG (aModel, param, pk_false))
	    bres = pk_false;
	  vgraph = translator->getDCG ();
	  GRAPHS_MANAGER->setLinkGraph (vgraph, (pointer) aModel);
	  delete translator;
	  AGRAPH->addLocalGraph (vgraph);

	  numero = numero + 1;
	}
    }
  return bres;
}

/**
   Local to sigProcess2Module()  method.
   Compiling a Signal module assumed set in the attribute member called vfi.
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
kernel2graph::sigProduceModule ()
{
  ptlrepid locales, constantsTypesloc, modeleList, modules;
  ptklExprProc corps;

  produceGraphInit (kl_getNameModel (vfi));
  NEW_BACK_ANNOT( kl_getBodyModel (vfi, &corps));

  /*
     param = kl_getParametersModel(vfi);
     if (!instdclparameters(param, parglobal))
     {
     sigTreeInterface2DCG(param, NILLREPID, NILLREPID);
     lparameters = getCurrentInstances(param, NILLINSTANCES, pk_true);
     } */

  lparameters = NILLINSTANCES;
  modules = kl_getModules (vfi);

  pushModules (modules);
  constantsTypesloc = kl_getConstantsTypesModel (vfi);
  locales = kl_getLocalFlowsModel (vfi);
  compiling_locals (constantsTypesloc, locales, NILLREPID);

  linputs = NILLINSTANCES;
  loutputs = NILLINSTANCES;
  modeleList = kl_getLocModels (vfi);
  compilingModels (modeleList);

  AGRAPH->setParameters (lparameters);
  AGRAPH->setInputs (linputs);
  AGRAPH->setOutputs (loutputs);

  AGRAPH->setLocalSignals (getCurrentInstances (locales, NILLINSTANCES, pk_false));

  compiling_locals_finalize (constantsTypesloc, locales, NILLREPID);
  popModules (modules);

  return pK_messenger->getErrorCode ();
}
