/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
*/
  kernel2graph::kernel2graph (ptGraphManager aGraphManager, boolean traceabilityMode)
{
  pK_messenger->setErrorCode (pk_true);
  pK_messenger->gmessload (mkfgmess);
  AGRAPH = new DCG ();
  AGRAPH->addPragma (newPredefinedPragma (PRAGMA_MAIN, NILOBJPRAGMA, NILTREE));
  kernel2graphInit (aGraphManager, NILSTACK, traceabilityMode);
}

/** 
    Constructor used for models annotated with the pragma "unexpanded" or Signal thread..
*/
kernel2graph::kernel2graph (ptGraphManager aGraphManager, ptStack LabelsStack, boolean traceabilityMode)
{
  AGRAPH = new DCG ();
  kernel2graphInit (aGraphManager, LabelsStack, traceabilityMode);
}

/** 
    Constructor used for models annotated with the pragma "unexpanded" or Signal thread..
*/
kernel2graph::kernel2graph (ptGraphManager aGraphManager, ptNameManager nm, ptStack LabelsStack, boolean traceabilityMode)
{
  AGRAPH = new DCG (nm);
  kernel2graphInit (aGraphManager, LabelsStack, traceabilityMode);
}

/**
   Destructor.
*/
kernel2graph::~kernel2graph ()
{
}

/**
   Compiling a module (kernelTreeModel) represented as a kernel tree.
   @return "true" if the graph has been built, "false" otherwise. 
   Use getDCG() method to get the built graph.   
*/
boolean kernel2graph::sigProcess2Module (ptrepfi kernelTreeModel)
{
  parglobal = NILEXPRSIG;
  vfi = kernelTreeModel;
  return sigProduceModule ();
}


/**
   Producing graph(s) for a model (process or module) represented as a kernel Tree.
   @return true if the graph has been built, "false" otherwise. Use getDCG() method to get the built graph.   
   @param kernelTreeModel (in) : a Signal model represented as a kernel tree.
   @param param (in) : the optional parameters.
   @param isaModule (in) : boolean to specific, when pk_true, if the kernelTreeModel is a module.
*/
boolean kernel2graph::kernelTree2DCG (ptrepfi kernelTreeModel, ptklExprSig param, boolean isaModule)
{
  ptklExprProc corps;
  noperators op;

  getManagers ();
  NEW_BACK_ANNOT(kl_getBodyModel (kernelTreeModel, &corps));
  if (isaModule) op = opmodu; else op = noper (currentTree);
  AGRAPH->setKind (op);
  if (isaModule) return sigProcess2Module (kernelTreeModel);
  parglobal = param;
  vfi = kernelTreeModel;
  sigProduceGraph ();
  printTraceabilityInfos();
  return igc->graphFinalyze ();
}


/**
   @return the graph of the current translation (when the operation succeeds), produced by one of the methods:
   sigProcess2Module() or kernelTree2DCG().
*/
ptDCG kernel2graph::getDCG ()
{
  return AGRAPH;
}
