/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return a new representation of the reference to a prototype of a model (val).
*/
extern ptIdentProp
mkIdentPropProcess (ptrepfi val)
{
  ptIdentProp v;
  v = mkIdentPropTag (protoModel);
  v->U.valf = val;
  return v;
}


/**
   @return a new representation of the reference to a prototype of a type (val).
*/
extern ptIdentProp
mkIdentPropTyp (ptreptyp val)
{
  ptIdentProp v;
  v = mkIdentPropTag (protoType);
  v->U.valty = val;
  return v;
}


/**
   @return a new representation of the reference to a prototype of a flow/constant/index/parameter (val).
*/
extern ptIdentProp
mkIdentProp (ptrepproto val)
{
  ptIdentProp v;
  v = mkIdentPropTag (protoIdent);
  v->U.valpr = val;
  return v;
}

/**
   @return a new representation of the reference to a prototype of a model type (val).
*/
extern ptIdentProp
mkIdentModelType (ptrepinterface val)
{
  ptIdentProp v;
  v = mkIdentPropTag (protoModelType);
  v->U.modelType = val;
  return v;
}


/**
  @return the model associated with an assumed reference to such an object (v). 
*/
extern ptrepfi
ptIdentProp_getProcess (ptIdentProp v)
{
  return GETPROCESSVAL (v);
}


/**
   @return the identifier prototype associated with an assumed reference to such an object (v). 
*/
extern ptrepproto
ptIdentProp_getIdent (ptIdentProp v)
{
  return GETIDENTVAL (v);
}

/**
   @return the type prototype associated with an assumed reference to such an object (v).
*/
extern ptreptyp
ptIdentProp_getType (ptIdentProp v)
{
  return GETTYPEVAL (v);
}

/**
   @return "true" if v is a reference to an object with ty class, "false" otherwise.
*/
extern boolean
ptIdentProp_equal (ptIdentProp v, typid ty)
{
  return (GETIDENDTYPID (v) == ty);
}

/**
   @return "true" if v is a reference to an x identifier, "false" otherwise.
*/
extern boolean
ptIdentProp_IdClassEqual (ptIdentProp v, typval x)
{
  if (GETIDENDTYPID (v) != protoIdent)
    return pk_false;
  return (ptrepproto_getIdClass (GETIDENTVAL (v)) == x);
}

/**
   @return the type of identifier of an element (v) assumed to be a reference to a identifier model.
*/
extern typval
ptIdentProp_getIdClass (ptIdentProp v)
{
  return ptrepproto_getIdClass (GETIDENTVAL (v));
}

/**
   It defines by vt the class of the identifier referenced in v.
*/
extern void
ptIdentProp_setIdClass (ptIdentProp v, typval vt)
{
  ptrepproto_setClassId (GETIDENTVAL (v), vt);
}

/**
   @return the type of the identifier referenced by v.
*/
extern typid
ptIdentProp_getIdClassType (ptIdentProp v)
{
  return GETIDENDTYPID (v);
}

/**
   It links v to its assocatied stack.
 */
extern void
ptIdentProp_setBackStack (ptIdentProp v, ptIdStack vpile)
{
  v->backIdStack = vpile;
}

/**
   It specifies that an object (v) is defined in a module (vmodule)
*/
extern void
ptIdentProp_setModule (ptIdentProp v, ptrepfi vmodule)
{
  v->backModule = vmodule;
}

/**
   @return "true" if an element (v) is defined in a module-vmodule), "false" otehrwise.
*/
extern boolean
ptIdentProp_getModule (ptIdentProp v, ptrepfi * vmodule)
{
  (*vmodule) = v->backModule;
  return ((*vmodule) != NILREPFI);
}

/**
   @return the type prototype associated with an assumed reference to such an object (v).
*/
extern ptrepinterface
ptIdentProp_getModelType (ptIdentProp v)
{
  return GETMODELTYPE (v);
}
