/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/** Identifier prototype representation.
    It gives access to internal representation associated with an idenfifier when it is the name of
        - a model of process, 
        - a flow (signal, parameters, index, constant), 
        - or type.
*/
typedef struct IdentProp
{
  ptIdStack backIdStack;	///< Back annotation to the associated stack (utility?)
  ptrepfi backModule;		///< For an object defined in a module, back information available.
  typid t;			///< Operator selecting the union case.
  union
  {
    ptrepfi valf;		///< Access to a model
    ptrepproto valpr;		///< Access to a signal, parameter, index, named constant
    ptreptyp valty;		///< Access to a type
    ptrepinterface modelType;	///< Access to a model type
  } U;
}
tIdentProp;

#define GETPROCESSVAL(v) (v->U.valf)
#define GETTYPEVAL(v) (v->U.valty)
#define GETIDENTVAL(v) (v->U.valpr)
#define GETMODELTYPE(v) (v->U.modelType)
#define GETIDENDTYPID(v) (v->t)

static ptIdentProp
mkIdentPropTag (typid v)
{
  ptIdentProp vpt;
  vpt = (ptIdentProp) malloc ((unsigned) (sizeof *vpt));
  vpt->backModule = NILREPFI;
  vpt->t = v;
  return vpt;
}
