/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/



/** 
    Second step of the compiling (kernel2graph translation).\n
    It pushes a type (t) to the associated stack with v.
    (case : type A = \<definition de type\>)
*/
extern void
kl_pushTypeInstance (ptIdStack v, TypeType t)
{
  kl_push_elem (v, (ptElemStack) t);
}


/**
   Second step of the compiling (kernel2graph translation).\n
   It pops the top of the associated stack with v.	
*/
extern void
kl_popTypeInstance (ptIdStack v)
{
  kl_pop_elem (v);
}

/**
   Second step of the compiling (kernel2graph translation).\n
   @return 
   - "true" if stack associated with v is not empty, then TT is the type at the top of the stack
   - "false" otherwise.
*/
extern boolean
kl_topTypeInstance (ptIdStack v, TypeType * TT)
{
  ptElemStack el;

  if (!kl_top_elem (v, &el))
    return pk_false;
  *TT = (TypeType) el;
  return pk_true;
}

/** 
    Second step of the compiling (kernel2graph translation).\n
    It pushes to the associated stack with v, the signal vsig.
*/
extern void
kl_pushInstance (ptIdStack v, ptexpcal vsig)
{
  kl_push_elem (v, (ptElemStack) vsig);
}

/**
   Second step of the compiling (kernel2graph translation).\n
   It pops the top of the associated stack with v.
*/
extern void
kl_popInstance (ptIdStack v)
{
  kl_pop_elem (v);
}

/** Second step of the compiling (kernel2graph translation).\n
    @return the top value (an signal instance) of the associated stack with v.
*/
extern ptexpcal
kl_topInstance (ptIdStack v)
{
  ptElemStack el;
  if (kl_top_elem (v, &el))
    return (ptexpcal) el;
  return NILEXPCAL;

}

/**
   Second step of the compiling (kernel2graph translation).\n
   @return the n-th element, from the top,  of the associated stack with v.
*/
extern ptexpcal
kl_topInstanceMinus (ptIdStack v, integer n)
{
  ptElemStack el;
  if (kl_top_elem_minus (v, n, &el))
    return (ptexpcal) el;
  return NILEXPCAL;
}
