/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return a new undefined expression of processes, linked with an abstract tree (p), used for back annotations.
*/
extern ptklExprProc
kl_mkNilExprProc (ntrees p)
{
  return mkKernelExprProc (NILLREPID, NILLREPID, NILLREPID, p, opnil);
}


/**
   @return a new composition expression of vptcf1 and vptcf2, characterized by 
   - the signals, not defined in the expression, that it references (inputs),
   - the signals it produces (outputs).
   - the abstract tree (vtree) from which it is produced 
*/
extern ptklExprProc
kl_mkComposition (ptlrepid inputs, ptlrepid outputs, ntrees vtree, ptklExprProc vptcf1, ptklExprProc vptcf2)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, NILLREPID, vtree, oplfeed);
  vklExprProc->U.composition.opcol1 = vptcf1;
  vklExprProc->U.composition.opcol2 = vptcf2;
  return vklExprProc;
}

/** 
    If the operand expression (eproc) is a compostion, it specifies that it is a set of assertions.
    
*/
extern void
kl_setCompositionAssertion (ptklExprProc eproc)
{
  eproc->assertion = pk_true;
}

/**
   @copydoc kl_mkIterate()
   Moreover, ub defines the upper bound of the index.
*/
extern ptklExprProc
kl_mkArray (ptlrepid inputs, ptlrepid outputs, ptlrepid lp, ntrees vtree, ptklExprSig vind, ptklExprSig ub, ptklExprProc pw, ptklExprProc body)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, lp, vtree, opmoti);
  vklExprProc->U.array.indice = vind;
  vklExprProc->U.array.lvarmodi = pw;
  vklExprProc->U.array.corpsdo = body;
  vklExprProc->U.array.bsup = ub;
  return vklExprProc;
}

/**
   @return a new iterate expression of processes, characterized by 
   - the signals, not defined in the expression, that it references (inputs),
   - the signals that it produces (outputs),
   - the parameters (lp) that it references,
   - the index (vind),
   - the initial part (pw),
   - the body part (body),
   - the abstract tree (vtree) from which it is produced .
*/
extern ptklExprProc
kl_mkIterate (ptlrepid inputs, ptlrepid outputs, ptlrepid lp, ntrees vtree, ptklExprSig vind, ptklExprProc pw, ptklExprProc body)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, lp, vtree, opiter);
  vklExprProc->U.array.indice = vind;
  vklExprProc->U.array.lvarmodi = pw;
  vklExprProc->U.array.corpsdo = body;
  vklExprProc->U.array.bsup = NILEXPRSIG;
  return vklExprProc;
}


/**
   @return a new equation, partial or not, expression on processes, according to the value ope (X := E, X ::= E ou X ::= default E).
   It is characterized by 
   - the signals, not defined in the expression, that it references (inputs),
   - the signals that it produces (outputs),
   - the parameters (lpar) that it references,
   - the left part expression (lf) of the equation,
   - the rigth part expression (vdef) of the equation,
   - the abstract tree (vtree) from which it is produced .
*/
extern ptklExprProc
kl_mkEquation (noperators ope, ptlrepid inputs, ptlrepid outputs, ptlrepid lpar, ntrees vtree, ptklExprSig lp, ptklExprSig vdef)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, lpar, vtree, ope);
  vklExprProc->U.equation.opaff1 = lp;
  vklExprProc->U.equation.opaff2 = vdef;
  return vklExprProc;
}


/**
   @return a new expression of synchronization, characterized by 
   - the signals, not defined in the expression, that it references (inputs),
   - the synchronisation operator (vop); i.e. ^=, ^<, ^#, ^>
   - the list of expressions (arg) of the expression
   - the abstract tree (vtree) from which it is produced .
*/
extern ptklExprProc
kl_mkClockOp (ptlrepid inputs, ntrees vtree, ptklExprSig arg, noperators vop)
{
  ptklExprProc vklExprProc;
  vklExprProc = mkKernelExprProc (inputs, NILLREPID, NILLREPID, vtree, vop);
  vklExprProc->U.opltl = arg;
  return vklExprProc;
}


/**
   It defines the reference to the model xxxxxxxxxxxxxxxxx.
*/
extern void
kl_setmodel (ptklExprProc vklExprProc, ptIdStack varid, ptrepfi vmodel)
{
  if (vklExprProc != NILEXPFIL)
    {
      vklExprProc->U.modelRef.nom = NILLREPID;
      appendProcess (&vklExprProc->U.modelRef.nom, varid, vmodel);
    }
}

/**
  @return a new invocation of a model of process characterized by
  - the signals, not defined in the expression, that it references (inputs),
  - the signals that it produces (outputs),
  - the model of the process (vmodel),
  - the stack associated with the name of the model (varid)
  - the effective paramaters of the invocation.
  - the abstract tree (vtree) from which it is produced .
*/
extern ptklExprProc
kl_mkMacro (ptlrepid inputs, ptlrepid outputs, ntrees vtree, ptIdStack varid, ptrepfi vmodel, ptklExprSig pareff)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, NILLREPID, vtree, opinst);
  vklExprProc->U.modelRef.lparam = pareff;
  kl_setmodel (vklExprProc, varid, vmodel);
  return vklExprProc;
}


/**
   @return a new "case " expression de processes, characterized by
  - the signals, not defined in the expression, that it references (inputs),
  - the signals that it produces (outputs),
  - the parameters that it references (lp),
  - the selector identifier (id)
  - the list of cases (l_case
  - the "otherwise" part (p_else),
  - the abstract tree (vtree) from which it is produced .
*/
extern ptklExprProc
kl_mkCase (ptlrepid inputs, ptlrepid outputs, ptlrepid lp, ntrees vtree, ptklExprSig id, ptlcase l_case, ptklExprProc p_else)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, lp, vtree, oppcas);
  vklExprProc->U.acase.ident = id;
  vklExprProc->U.acase.l_case = l_case;
  vklExprProc->U.acase.p_else = p_else;
  return vklExprProc;
}


/**
   @return a new "dependence" expression de processes defined by vfrom -> vto when vat.
   Moreover, it is characterized 
   - by the signals, not defined in the expression, that it references (inputs),
   - and the back annotation (vtree) to the abstract tree.
*/
extern ptklExprProc
kl_mkDependence (ptlrepid inputs, ntrees vtree, ptklExprSig vfrom, ptklExprSig vto, ptklExprSig vat)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, NILLREPID, NILLREPID, vtree, opdepd);
  vklExprProc->U.dependence.vfrom = vfrom;
  vklExprProc->U.dependence.vto = vto;
  vklExprProc->U.dependence.vat = vat;
  return vklExprProc;
}




/**
  (Used for confined process), 
  It associates with an expresson of processes (vklExprProc) the following declarations
  - the constants and types (const_typ_loc)
  - the local signals  (sig_loc)
  - the lables (labs)
*/
extern void
kl_setLocals (ptklExprProc vklExprProc, ptlrepid const_typ_loc, ptlrepid sig_loc, ptlrepid labs)
{
  if (vklExprProc != NILEXPFIL)
    {
      vklExprProc->cst_typ_loc = insertend (const_typ_loc, vklExprProc->cst_typ_loc);
      vklExprProc->sig_loc = insertend (sig_loc, vklExprProc->sig_loc);
      vklExprProc->llabels = labs;
    }
}


/**
  It specifies that an expression on processes (vklExprProc) is labelled by vlab.
  */
extern void
kl_setLabelExpr (ptklExprProc vklExprProc, ptklExprSig vlab)
{
  if (vklExprProc != NILEXPFIL) vklExprProc->etiq = vlab;
}


/**
   It specifies that an expression on processes (vklExprProc) is controlled by vOn.
*/
extern void
kl_setControlExpr (ptklExprProc vklExprProc, ptklExprSig vOn)
{
  if (vklExprProc != NILEXPFIL) vklExprProc->On = vOn;
}


/**
  @return a new invocation of a model of process (positional macro-expansion of a model) characterized by
  - the signals, not defined in the expression, that it references (inputs),
  - the signals that it produces (outputs),
  - the parameters that it references (lparams),
  - a reference to the model (vrefModel) (see kl_mkMacro()),
  - the effective parameters of the instance(pareff),
  - the abstract tree (vtree) from which it is produced.
*/
extern ptklExprProc
kl_mkProcCall (ptlrepid inputs, ptlrepid outputs, ptlrepid lparams, ntrees vtree, ptklExprProc vrefModel, ptklExprSig pareff)
{
  ptklExprProc vklExprProc;

  vklExprProc = mkKernelExprProc (inputs, outputs, lparams, vtree, opappe);
  vklExprProc->U.call.opmodel = vrefModel;
  vklExprProc->U.call.largs = pareff;
  return vklExprProc;
}

/**
   @return a new expression of processes that represents the body of an external model (external "language"), 
   characterized by
   - the abstract tree (vtree) of the model.
   - the "langage" of the implementation (vl).
*/
extern ptklExprProc
kl_mkExternal (ntrees vtree, ntrees vl)
{
  ptklExprProc vklExprProc;
  vklExprProc = mkKernelExprProc (NILLREPID, NILLREPID, NILLREPID, vtree, opexte);
  vklExprProc->U.vlanguage = vl;
  return vklExprProc;
}


/**
   @return a new expression of processes that represents an empty sytem of equations.
   It is characterized by the abstract tree (vtree) of the model.
*/
extern ptklExprProc
kl_mkEmptySystem (ntrees vtree)
{
  ptklExprProc vklExprProc;
  vklExprProc = mkKernelExprProc (NILLREPID, NILLREPID, NILLREPID, vtree, opnullproc);
  return vklExprProc;
}


/**
   Expression on signals.\n
  @return the expression on signals for encapsulating a reference to a model (vposi).
*/
extern ptklExprSig
kl_mkProcCallExpr (ptklExprProc vposi, ntrees vt)
{
  ptklExprSig pt;
  pt = mkKernelExprSig (opappe, call, vt);
  pt->U.opposi = vposi;
  return pt;
}



/**
  Expression on signals.\n
  @return a new expression for a reference to a identifier (nom). It is a signal/constant/index or a parameter.
*/
extern ptklExprSig
kl_mkIdentifierExpr (ptlrepid nom, ntrees vt)
{
  ptklExprSig pt;

  pt = mkKernelExprSig (opident, variabl, vt);
  pt->U.opvari = nom;
  return pt;
}


/**
   Expression on signals.\n
   @return a new representation of a constant (vconst) of type given by an operator (ope).
*/
extern ptklExprSig
kl_mkConstExpr (noperators ope, ptrepstr vconst, ntrees vt)
{
  ptklExprSig pt;

  pt = mkKernelExprSig (ope, constant, vt);
  pt->U.opconst = vconst;
  return pt;
}

/**
   Expression on signals.\n
   @return a new  unary expression on signals characterized by 
   - its class (cl),
   - a Signal operator (ope),
   - its argument (op1).
*/
extern ptklExprSig
kl_mkUnaryExpr (cloper cl, noperators ope, ptklExprSig op1, ntrees vt)
{
  ptklExprSig pt;

  pt = mkKernelExprSig (ope, cl, vt);
  pt->U.op = op1;
  return pt;
}

/**
  Expression on signals.\n
  @return a new  binary expression on signals characterized by 
   - its class (cl),
   - a Signal operator (ope),
   - two arguments (op1, op2).
*/
extern ptklExprSig
kl_mkBinaryExpr (cloper cl, noperators ope, ptklExprSig op1, ptklExprSig op2, ntrees vt)
{
  ptklExprSig pt;

  pt = mkKernelExprSig (ope, cl, vt);
  pt->U.BinExp.op21 = op1;
  pt->U.BinExp.op22 = op2;
  return pt;
}

/**
   Expression on signals.\n
   It defines the second operand of an assumed binary expression (v) by op2.
*/
extern void
kl_setBinaryExprChild2 (ptklExprSig v, ptklExprSig op2)
{
  v->U.BinExp.op22 = op2;
}


/**
   Expression on signals.\n
   @return a new expression that represents a list of expressions with ope as operator, initialized with op1. 
*/
extern ptklExprSig
kl_mkListExpr (noperators ope, ptklExprSig op1, ntrees vt)
{
  return kl_mkBinaryExpr (oper2, ope, op1, NILEXPRSIG, vt);
}

/**
   Expression on signals.\n
   @return a new list expression with ope as operator, by adding op1 before op2.
*/
extern ptklExprSig
kl_mkInsertHeadList (noperators ope, ptklExprSig op1, ptklExprSig op2, ntrees vt)
{
  return kl_mkBinaryExpr (oper2, ope, op1, op2, vt);
}

/**
   Expression on signals.\n
   @return a new  binary expression on signals characterized by 
   - its class (cl),
   - a Signal operator (ope),
   - three arguments (op1, op2, op3).
*/
extern ptklExprSig
kl_mkTernaryExpr (cloper cl, noperators ope, ptklExprSig op1, ptklExprSig op2, ptklExprSig op3, ntrees vt)
{
  ptklExprSig pt;

  pt = mkKernelExprSig (ope, cl, vt);
  pt->U.TerExp.op31 = op1;
  pt->U.TerExp.op32 = op2;
  pt->U.TerExp.op33 = op3;
  return pt;
}

/**
   Expression on signals.\n
   @return a new expression that references a type (vtyp) with ope as SIGNAL operator.
   
   Used for the representation of the types in parameter declarations.
   */
extern ptklExprSig
kl_mkTypeExpr (noperators ope, ptreptyp vtyp, ntrees vt)
{
  ptklExprSig pt;

  pt = mkKernelExprSig (ope, opertyp, vt);
  pt->U.optyp = vtyp;
  return pt;
}

/**
   It assigns the interface (inputs,outputs) of an expression of process (vklExprProc).
*/
extern void
klset_inout (ptklExprProc vklExprProc, ptlrepid inputs, ptlrepid outputs)
{
  vklExprProc->inList = insertlist (inputs, NILLREPID);
  vklExprProc->outList = insertlist (outputs, NILLREPID);
}
