/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** Constructor.
    @return a new representation of a model, typed by v (model, ident, type) 
    with nom as associated stack.
    The next element of the created representation is the nil element.
*/
extern ptlrepid
mkRepidsigRef (ptIdStack nom, ptIdentProp v)
{
  return mkrepid (nom, NILLREPID, v);
}

/**
   @return the stack associated with vlrepid.
*/
extern ptIdStack
nameId (ptlrepid vlrepid)
{
  return vlrepid->rep;
}

/**
   @return the type of model associated with vlrepid.
*/
extern ptIdentProp
ptlrepid_getIdentProp (ptlrepid vlrepid)
{
  return GETIDENPROP (vlrepid);
}


/**
   @return the next element of vlrepid.
*/
extern ptlrepid
succid (ptlrepid vlrepid)
{
  return vlrepid->next;
}

/**
   The next element of vlrepid becomes v.
*/
extern void
setsuccid (ptlrepid vlrepid, ptlrepid v)
{
  vlrepid->next = v;
}

/** @return the identifier associated with the stack of vlrepid. */
extern ptident
getextidentrepid (ptlrepid vlrepid)
{
  return identpile (nameId (vlrepid));
}

/**
  It inserts a new model characterized by its stack (elem) in a sorted list (deblist).
  @return "true"if the element is not already in the list, "false" otherwise.
*/
extern boolean
ptlrepid_insert (ptlrepid * deblist, ptIdStack elem)
{
  boolean b;
  integer n;
  ptlrepid succ;

  if ((*deblist) == NILLREPID)
    {
      *deblist = mkrepid (elem, NILLREPID, kl_top (elem));
      return pk_true;
    }

  n = atpilegetnum (elem);
  if (n < atpilegetnum ((*deblist)->rep))
    {
      *deblist = mkrepid (elem, *deblist, kl_top (elem));
      return pk_true;
    }
  succ = *deblist;
  do
    {
      if (atpilegetnum (succ->rep) == n)
	{
	  succ = NILLREPID;
	  b = pk_false;
	}
      else if (succ->next == NILLREPID)
	{
	  succ->next = mkrepid (elem, NILLREPID, kl_top (elem));
	  succ = NILLREPID;
	  b = pk_true;
	}
      else if (atpilegetnum (succ->next->rep) > n)
	{
	  succ->next = mkrepid (elem, succ->next, kl_top (elem));
	  succ = NILLREPID;
	  b = pk_true;
	}
      else
	succ = succ->next;
    }
  while (succ != NILLREPID);
  return b;
}


/**
   @return for an assumed  model of identifier (vrepid), the representation of its prototype. 
   model of identifier: flow identifier, constant identifdier, index identifier, parameter identifier,...
*/
extern ptrepproto
getIdRepid (ptlrepid vrepid)
{
  return ptIdentProp_getIdent (GETIDENPROP (vrepid));
}

/**
   @return for an assumed  model of type (vrepid), the representation of its prototype. 
*/
extern ptreptyp
getTypeRepid (ptlrepid vrepid)
{
  return ptIdentProp_getType (GETIDENPROP (vrepid));
}

/**
   @return for an assumed model of a model type (vrepid), the representation of its prototype. 
*/
extern ptrepinterface
getModelype (ptlrepid vrepid)
{
  return ptIdentProp_getModelType (GETIDENPROP (vrepid));
}

/**
   It pushes (kl_pushModel()) each element of a list of declarations of models (ptliste) in its associated stack.
*/
extern void
pushModels (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      kl_pushModel (GETSTACKVAL (ptliste), ptlrepid_getIdentProp (ptliste));
      ptliste = ptliste->next;
    }
}

/**
   It pops (kl_pop) each element of a list of declarations of model (ptliste)
*/
extern void
popModels (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      kl_pop (GETSTACKVAL (ptliste));
      ptliste = ptliste->next;
    }
}


/**
  It pushes (kl_push()) each element of a list of declaration of parameters (ptliste) in its associated stack
*/
extern void
pushParameters (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      if (isTypeId (ptliste) || isModelId (ptliste))
	kl_pushType (GETSTACKVAL (ptliste), GETIDENPROP (ptliste));
      else
	kl_pushParameter (GETSTACKVAL (ptliste), GETIDENPROP (ptliste));
      ptliste = ptliste->next;
    }
}


/**
   It pushes (kl_pushConstant()) each element of a list of declarations of constants (ptliste) in its associated stack
 */
extern void
pushConstants (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      kl_pushConstant (GETSTACKVAL (ptliste), GETIDENPROP (ptliste));
      ptliste = ptliste->next;
    }
}

/**
   It pushes (kl_pushLabel()) each element of a list of declarations of constants (ptliste) in its associated stack.
*/
extern void
pushLabels (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      kl_pushLabel (GETSTACKVAL (ptliste), GETIDENPROP (ptliste));
      ptliste = ptliste->next;
    }
}


/**
   It pushes (kl_pushType()) each element of a list of declarations of types (ptliste) in its associated stack.
*/
extern void
pushTypes (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      kl_pushType (GETSTACKVAL (ptliste), GETIDENPROP (ptliste));
      ptliste = ptliste->next;
    }
}

/**
   It pushes (kl_pushFlow()) each element of a list of declarations of flows (ptliste) in its associated stack.
*/
extern void
pushFlows (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      kl_pushFlow (GETSTACKVAL (ptliste), GETIDENPROP (ptliste));
      ptliste = ptliste->next;
    }
}


/**
   It pops each element of a liste of declaration of flows (ptliste) from associated stack.   
*/
extern void
popFlows (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      ptlrepid_pop (ptliste);
      ptliste = ptliste->next;
    }
}

/** 
    It pops each element of a liste of declaration of types (ptliste) from associated stack.   
*/
extern void
popTypes (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      if (ptIdentProp_equal (GETIDENPROP (ptliste), protoType))
	popTypeRepid (ptliste);
      ptliste = ptliste->next;
    }
}


/** @return the sicz of a liste of declaration (l). */
extern integer
ptlrepid_size (ptlrepid l)
{
  integer nb;

  nb = 0;
  while (l != NILLREPID)
    {
      nb = nb + 1;
      l = l->next;
    }
  return nb;
}


/**
   @return "true" if elem is the stack of an element of a list (l), "false" otherwise.
*/
extern boolean
ptlrepid_isIn_Unsorted (ptlrepid l, ptIdStack elem)
{
  ptlrepid aux;
  integer cnum, lnum;

  if (l == NILLREPID)
    return pk_false;
  cnum = atpilegetnum (elem);
  aux = l;
  do
    {
      lnum = atpilegetnum (aux->rep);
      if (cnum == lnum)
	return pk_true;
      aux = aux->next;
    }
  while (aux != NILLREPID);
  return pk_false;
}

/**
   @return "true" if an element (elem) is an element of a list (list), "false" otherwise.
*/
extern boolean
ptlrepid_isIn_list (ptlrepid elem, ptlrepid list)
{
  return ptlrepid_isIn_Unsorted (list, elem->rep);
}


/**
  @return
  - "true" if elem is the stack of an element of liste, it is removed form the list.
  - "false" otherwise.
*/
extern boolean
ptlrepid_remove_Unsorted (ptlrepid * liste, ptIdStack elem)
{
  integer cnum, lnum;
  ptlrepid aux, prec;

  if ((*liste) == NILLREPID)
    return pk_false;
  cnum = atpilegetnum (elem);
  aux = *liste;
  do
    {
      lnum = atpilegetnum (aux->rep);
      if (cnum == lnum)
	{
	  if ((*liste) == aux)
	    *liste = aux->next;
	  else
	    prec->next = aux->next;
	  return pk_true;
	}
      prec = aux;
      aux = aux->next;
    }
  while (aux != NILLREPID);
  return pk_false;
}



/**
   @return the rank of vpile in delist, 0 if no present.
*/
extern integer
getRankOfRepid (ptlrepid deblist, ptIdStack vpile)
{
  integer i;
  boolean result;

  if (deblist == NILLREPID)
    return 0;
  result = pk_false;
  i = 1;
  do
    {
      if (deblist->rep == vpile)
	result = pk_true;
      else if (deblist->next == NILLREPID)
	{
	  i = 0;
	  result = pk_true;
	}
      else
	{
	  i = i + 1;
	  deblist = deblist->next;
	}
    }
  while (!result);
  return i;
}



/**
   @return the stack of the poselem-th element of a list of declaration (deblist), NILPILE when it does not exist.
 */
extern ptIdStack
ptlrepid_getStackRepidAt (ptlrepid deblist, integer poselem)
{
  ptIdStack v;

  if (deblist == NILLREPID)
    return NILPILE;
  while (poselem > 0)
    {
      poselem = poselem - 1;
      if (poselem == 0)
	v = deblist->rep;
      else if (deblist->next == NILLREPID)
	v = NILPILE;
      else
	deblist = deblist->next;
    }
  return v;
}


/**
  It inserts in a sorted list (mainlist) a copy of each element of a unsorted list (lToInsert).
  @return the result of the insertions (modified mainList).
*/
extern ptlrepid
insertunsort (ptlrepid lToInsert, ptlrepid mainList)
{
  ptlrepid vlrepid;

  while (lToInsert != NILLREPID)
    {
      ptlrepid_insertrepid (&mainList, &vlrepid, lToInsert);
      lToInsert = lToInsert->next;
    }
  return mainList;
}

/**
  It inserts a list(lToInsert) at  the end of an another one (mainlist).
  @return the result of the merge.
*/
extern ptlrepid
insertend (ptlrepid lToInsert, ptlrepid mainList)
{
  ptlrepid elc, smainList;

  if (lToInsert == NILLREPID)
    return mainList;
  if (mainList == NILLREPID)
    {
      mainList = duprepid (lToInsert, NILLREPID);
      lToInsert = lToInsert->next;
    }
  smainList = mainList;
  while (mainList->next != NILLREPID)
    mainList = mainList->next;
  while (lToInsert != NILLREPID)
    {
      elc = duprepid (lToInsert, NILLREPID);
      mainList->next = elc;
      mainList = mainList->next;
      lToInsert = lToInsert->next;
    }
  return smainList;
}



/**
   It produces  llresult = LRESULT - the elements of another list (lToRemove).
*/
extern void
diffLists (ptlrepid lToRemove, ptlrepid * lresult)
{
  while (lToRemove != NILLREPID)
    {
      ptlrepid_remove_Unsorted (lresult, lToRemove->rep);
      lToRemove = lToRemove->next;
    }
}

/**
   It pushes the element (ptliste) assumed to be a declaration of an identifier (kl_pushConstant()) or a type (kl_pushType()).
*/
extern void
pushConstantOrType (ptlrepid ptliste)
{
  ptIdentProp v;
  v = GETIDENPROP (ptliste);
  switch ((int) ptIdentProp_getIdClassType (v))
    {
    case protoType:
      kl_pushType (GETSTACKVAL (ptliste), v);
      break;
    case protoIdent:
      kl_pushConstant (GETSTACKVAL (ptliste), v);
      break;
    }
}

/**
   It pushes (pushConstantOrType()) each element of a list of declarations (constants or types).
*/
extern void
pushConstantsTypes (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      pushConstantOrType (ptliste);
      ptliste = ptliste->next;
    }
}

/**
   It pops (popConstantOrType()) each element of a list (ptliste) of declarations (constants or types).
*/
extern void
popConstantOrType (ptlrepid ptliste)
{
  ptlrepid_pop (ptliste);
}

/**
   It pops (popConstantOrType()) each element of a list (ptliste) of declarations (constants or types).
*/
extern void
popConstantsTypes (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      popConstantOrType (ptliste);
      ptliste = ptliste->next;
    }
}


/**
   @return the model associated with a declaration (l) assumed to be a declaration of a model.
 */
extern ptrepfi
getModelRepid (ptlrepid l)
{
  return ptIdentProp_getProcess (GETIDENPROP (l));
}

/**
   @return "true" if vlrepid is a declaration of a type, "false" otherwise.
 */
extern boolean
isTypeId (ptlrepid vlrepid)
{
  return ptIdentProp_equal (GETIDENPROP (vlrepid), protoType);
}

/**
   @return "true" if vlrepid is a declaration of a model or a formal process, "false" otherwise.
*/
extern boolean
isModelId (ptlrepid vlrepid)
{
  return ptIdentProp_equal (GETIDENPROP (vlrepid), protoModel);
}


/**
   @return "true" if vlrepid is a declaration of a constant/flow/index/parameter, "false" otherwise.
 */
extern boolean
isFlowId (ptlrepid vlrepid)
{
  return ptIdentProp_equal (GETIDENPROP (vlrepid), protoIdent);
}

/**
   vlrepid is assumed to be a declaration of a constant/flow/index/parameter identifier.
   It returns:
   - init the initial value, nil by default
   - vtyp the type of the identifier
   - bstate pk_true for a state variable declaration, false otherwise
   - vtree the back information (tracability) to the abstract tree produced by the parser.
*/
extern void
getInfoProto (ptlrepid vrepid, ptklExprSig * init, ptreptyp * vtyp, categoryType * vcat, ntrees * vtree)
{
  ptrepproto vproto;
  vproto = getIdRepid (vrepid);
  ptrepproto_getInfos (vproto, init, vtyp, vcat, vtree);
}

/**
   vlrepid is assumed to be a declaration of a constant/flow/index/parameter.
   @return the back information (tracability) to the abstract tree produced by the parser.
*/
extern ntrees
flowIdGetTree (ptlrepid vrepid)
{
  ptklExprSig init;
  ptreptyp vtyp;
  categoryType vcat;;
  ntrees vtree;

  getInfoProto (vrepid, &init, &vtyp, &vcat, &vtree);
  return vtree;
}

/**
   vlrepid is assumed to be a declaration of a constant/flow/index/parameter.
   @return "true" is it denotes a state variable, "false" otherwise.
*/
extern boolean
isStateVariable (ptlrepid vrepid)
{
  ptklExprSig init;
  ptreptyp vtyp;
  categoryType vcat;
  ntrees vtree;

  getInfoProto (vrepid, &init, &vtyp, &vcat, &vtree);
  return (vcat == statevarSignal);
}


/**
   vlrepid is assumed to be a declaration of a constant/flow/index/parameter.
   @return "true" is it denotes a state variable, "false" otherwise.
*/
extern boolean
isSharedVariable (ptlrepid vrepid)
{
  ptklExprSig init;
  ptreptyp vtyp;
  categoryType vcat;
  ntrees vtree;

  getInfoProto (vrepid, &init, &vtyp, &vcat, &vtree);
  return (vcat == sharedSignal);
}


/**
   vlrepid is assumed to be a declaration of a constant/flow/index/parameter.
   @return "true" is it denotes an event, "false" otherwise.
*/
extern boolean
isEventFlow (ptlrepid vrepid)
{
  ptklExprSig init;
  ptreptyp vtyp;
  categoryType vcat;
  ntrees vtree;

  getInfoProto (vrepid, &init, &vtyp, &vcat, &vtree);
  if (vtyp == NILREPTYP)
    return pk_false;
  return (opevent == kl_getBasicType (vtyp));
}

/**
   It removes the poselem-th element of a list of declarations (deblist).
   @return the modified list.
*/
extern ptlrepid
removerepid (ptlrepid deblist, integer poselem)
{
  ptlrepid result;

  if (poselem == 1)
    return deblist->next;
  result = deblist;
  while (poselem > 2)
    {
      deblist = deblist->next;
      poselem = poselem - 1;
    }
  deblist->next = deblist->next->next;
  return result;
}



/**
   It adds at the end of a list (deblist), a new representation of the model of process (vrepfi).
   @return "true" if the element was not already in the list, "false" otherwise.
*/
extern boolean
appendrepfi (ptlrepid * deblist, ptIdStack elem, ptrepfi vrepfi, ptlrepid * vlrepid)
{
  ptlrepid succ;
  boolean result;

  *vlrepid = mkrepid (elem /*NILPILE*/, NILLREPID, mkIdentPropProcess (vrepfi));
  if ((*deblist) == NILLREPID)
    {
      *deblist = *vlrepid;
      return pk_true;
    }
  succ = *deblist;
  result = pk_true;
  do
    {
      if (ptIdentProp_getProcess (GETIDENPROP (succ)) == vrepfi)
	result = pk_false;
      if (succ->next == NILLREPID)
	{
	  succ->next = *vlrepid;
	  succ = NILLREPID;
	}
      else
	succ = succ->next;
    }
  while (succ != NILLREPID);
  return result;
}

/**
   It inserts in a sorted list (mainlist) a copy of each element of a sorted list (lToInsert).
   @return the result of the insertions (modified mainList).
*/
extern ptlrepid
insertlist (ptlrepid lToInsert, ptlrepid mainList)
{
  ptlrepid vlrepid;

  if (lToInsert != NILLREPID)
    {
      ptlrepid_insertrepid (&mainList, &vlrepid, lToInsert);
      lToInsert = lToInsert->next;
      while (lToInsert != NILLREPID)
	{
	  ptlrepid_insertrepid (&vlrepid, &vlrepid, lToInsert);
	  lToInsert = lToInsert->next;
	}
    }
  return mainList;
}

/**
   It pops a declaration (vrepid)
*/
extern void
ptlrepid_pop (ptlrepid vrepid)
{
  ptIdStack vpile;
  vpile = vrepid->rep;
  kl_pop (vpile);
}


/**
   It pops a type declaration (vrepid) 
*/
extern void
popTypeRepid (ptlrepid vrepid)
{
  ptlrepid_pop (vrepid);
}

/** 
    It pops an index declaration (vrepid) 
*/
extern void
popIndexRepid (ptlrepid vrepid)
{
  ptlrepid_pop (vrepid);
}


/** @return
    - "true" if the n-th element of a list (vrepid) exists, then velem is this element
    - "false" otherwise.
*/
extern boolean
ptlrepid_getRepidAt (ptlrepid vrepid, integer n, ptlrepid * velem)
{
  ptlrepid lrepidcour;

  lrepidcour = vrepid;
  while ((lrepidcour != NILLREPID) && (n != 1))
    {
      lrepidcour = lrepidcour->next;
      n--;
    }
  *velem = lrepidcour;
  return (boolean) (lrepidcour != NILLREPID);
}

/**
   Assumed: vlrepid is a list of declarations of types.
   @return
    - "true" if the n-th element of a list (vlrepid) exists, then vtypid is the representation of the selected type
    - "false" otherwise.
*/
extern boolean
ptlrepid_getIdTypeOf (ptlrepid vlrepid, integer n, typid * vtypid)
{
  integer i;
  i = 1;
  while ((i < n) && (vlrepid != NILLREPID))
    {
      vlrepid = vlrepid->next;
      i++;
    }
  if (vlrepid == NILLREPID)
    return pk_false;
  *vtypid = ptIdentProp_getIdClassType (GETIDENPROP (vlrepid));
  return pk_true;
}


/**
  It adds a new declaration of a model of identifier (constant/flow/index/parameter) to a list of declarations (deblist).
  The identifer is characterized by its associated stack (elem) and its specific attributes (vrepproto:init, type,...)
  @return "true" if the element is not already in the list, it is then added at the end "false" otherwise.
  vlrepid is the declaration of the element.
*/
extern boolean
appendProtoId (ptlrepid * deblist, ptIdStack elem, ptrepproto vrepproto, ptlrepid * vlrepid)
{
  *vlrepid = mkrepid (elem, NILLREPID, mkIdentProp (vrepproto));
  return appendrepid (deblist, *vlrepid);
}


/**
  It adds a new declaration of a model of type to a list of declarations (deblist).
  The identifer is characterized by its associated stack (elem) and its specific attributes (vreptyp)
  @return "true" if the element is not already in the list, it is then added at the end "false" otherwise. 
  vlrepid is the declaration of the element.
*/
extern boolean
appendProtoTyp (ptlrepid * deblist, ptIdStack elem, ptreptyp vreptyp, ptlrepid * vlrepid)
{
  *vlrepid = mkrepid (elem, NILLREPID, mkIdentPropTyp (vreptyp));
  return appendrepid (deblist, *vlrepid);
}

/**
  It adds a new declaration of a model of a model type (ie an interface) to a list of declarations (deblist).
  The identifer is characterized by its associated stack (elem) and its specific attributes (vreptyp)
  @return "true" if the element is not already in the list, it is then added at the end "false" otherwise. 
  vlrepid is the declaration of the element.
*/
extern boolean
appendProtoModelType (ptlrepid * deblist, ptIdStack elem, ptrepinterface vreptyp, ptlrepid * vlrepid)
{
  *vlrepid = mkrepid (elem, NILLREPID, mkIdentModelType (vreptyp));
  return appendrepid (deblist, *vlrepid);
}

/**
  It adds a new declaration of a model of process to a list of declarations (deblist).
  The identifer is characterized by its associated stack (elem) and its specific attributes (vrepfi)
  @return "true" if the element is not already in the list, it is then added at the end "false" otherwise. 
*/
extern boolean
appendProcess (ptlrepid * deblist, ptIdStack elem, ptrepfi vrepfi)
{
  ptlrepid vlrepid;

  vlrepid = mkrepid (elem, NILLREPID, mkIdentPropProcess (vrepfi));
  return appendrepid (deblist, vlrepid);
}

/**
   Assumed: vlrepid is a list of declarations of  model of identifiers (constant/flow/index/parameter)
   @return 
   - "true" if the n-th element of a list (vrepid) exists, "false" otherwise.
   It returns (when it exist):
   - init : the expression of its initial value,
   - vtyp: its type,
   - bstate: equal to pk_true for a state variable, false otherwise
   - vtree the back information (tracability) to the abstract tree produced by the parser.
      
   See also getInfoProto().
*/
extern boolean
ptlrepid_getIdInfosAt (ptlrepid vrepid, integer n, ptklExprSig * init, ptreptyp * vtyp, categoryType * vcat, ntrees * vtree)
{
  ptlrepid lrepidcour;
  if (!ptlrepid_getRepidAt (vrepid, n, &lrepidcour))
    return pk_false;
  getInfoProto (lrepidcour, init, vtyp, vcat, vtree);
  return pk_true;
}

/**
   Assumed: vlrepid is a list of declarations of  model of identifiers (constant/flow/index/parameter)
   @return 
   - "true" if the n-th element of a list (vrepid) exists, then vtyp is the representation of the selected type
   - "false" otherwise.   
*/
extern boolean
ptlrepid_getIdInfosAttype (ptlrepid vrepid, integer n, ptreptyp * vtyp)
{
  ptlrepid lrepidcour;

  if (!ptlrepid_getRepidAt (vrepid, n, &lrepidcour))
    return pk_false;
  (*vtyp) = getTypeRepid (lrepidcour);
  return pk_true;
}


/** 
   It pushes a module (vfi):
   - its constants/types (pushConstantsTypes()) 
   - its local models (pushModels()).
*/
extern void
pushModule (ptrepfi vfi)
{
  pushConstantsTypes (kl_getConstantsTypesModel (vfi));
  pushModels (kl_getLocModels (vfi));
}

/**
   It pushes a list of declarations of imports (ptliste). It applies pushModule() for each element.
*/
extern void
pushModules (ptlrepid ptliste)
{
  ptrepfi vfi;

  while (ptliste != NILLREPID)
    {
      vfi = getModelRepid (ptliste);
      pushModule (vfi);
      ptliste = ptliste->next;
    }
}

/** 
   It pops a module (vfi):
   - its constants/types (popConstantsTypes()) 
   - its local models (popModels()).
*/
extern void
popModule (ptrepfi vfi)
{
  popConstantsTypes (kl_getConstantsTypesModel (vfi));
  popModels (kl_getLocModels (vfi));
}

/**
   It pops a list of declarations of imports (ptliste). It applies popModule() for each element (from last to first).
*/
extern void
popModules (ptlrepid ptliste)
{
  if (ptliste == NILLREPID)
    return;
  popModules (ptliste->next);
  popModule (getModelRepid (ptliste));
}



/**
   Assumed: vlrepid is a list of declarations of process models (ptliste).
   Each process is declared as intrinsic model.
*/
extern void
setInstrinsicAttribute (ptlrepid ptliste)
{
  ptrepfi vfi;
  while (ptliste != NILLREPID)
    {
      vfi = getModelRepid (ptliste);
      kl_belongsIntrinsicModel (vfi);
      ptliste = ptliste->next;
    }
}



/**
   The objects of the list of declarations (l) are tagged to be elements defined in a library (vmodule)
*/
extern void
setDclInModule (ptlrepid l, ptrepfi vmodule)
{
  while (l != NILLREPID)
    {
      ptIdentProp_setModule (GETIDENPROP (l), vmodule);
      l = l->next;
    }
}


/**
   @return l + a copy of el.
*/
extern ptlrepid
addRepid (ptlrepid l, ptlrepid el)
{
  ptlrepid elc;

  elc = duprepid (el, NILLREPID);
  appendrepid (&l, elc);
  return l;
}

/**
   @return "true" if it exists a model (model) with id as name in a list of local models, taking into account the imports, of a model (vfi),
   "false" otherwise.
*/
extern boolean
klGetModelOf (ptrepfi vfi, ptident id, ptrepfi * model)
{
  ptlrepid linstrinsics;
  integer n, i;

  if (vfi == NILREPFI)
    return pk_false;
  if (kllrepidGetModelOf (kl_getLocModels (vfi), id, model))
    return pk_true;
  linstrinsics = kl_getModules (vfi);

  n = ptlrepid_size (linstrinsics);
  for (i = 1; i <= n; i++)
    {
      if (kllrepidGetModelOf (kl_getLocModels (getModelRepid (linstrinsics)), id, model))
	return pk_true;
      linstrinsics = succid (linstrinsics);
    }
  return pk_false;
}

/**
   @return the list l in which an element (elem) is inserted if it is not already in.
*/
extern ptlrepid
ptlrepid_add_elem (ptlrepid l, ptlrepid elem)
{
  integer n, cnum;
  ptlrepid succ, curlist;
  boolean b;

  if (l == NILLREPID)
    return elem;
  n = atpilegetnum (elem->rep);
  cnum = atpilegetnum (l->rep);
  if (n < cnum)
    {
      elem->next = l;
      return elem;
    }
  curlist = l;
  b = pk_true;
  while (b)
    {
      b = pk_false;
      if (cnum != n)
	{
	  succ = curlist->next;
	  if (succ == NILLREPID)
	    curlist->next = elem;
	  else
	    {
	      cnum = atpilegetnum (succ->rep);
	      if (cnum > n)
		{
		  elem->next = succ;
		  curlist->next = elem;
		}
	      else
		{
		  b = pk_true;
		  curlist = succ;
		}
	    }
	}
    }
  return l;
}

/**
   It pops each element of a list (ptliste) of declarations of model types.
*/
extern void
popModelTypes (ptlrepid ptliste)
{
  while (ptliste != NILLREPID)
    {
      popConstantOrType (ptliste);
      ptliste = ptliste->next;
    }
}

/**
   (test = same stack)
   @return 
   - "true" if an element (vlrepid) is not in a list (deblist), then it is added at the end of deblist.
   - "false" otherwise, deblist is not modified.
*/
extern boolean
appendrepid (ptlrepid * deblist, ptlrepid vlrepid)
{
  ptIdStack elem;
  ptlrepid succ;
  boolean result;

  if ((*deblist) == NILLREPID)
    {
      *deblist = vlrepid;
      result = pk_true;
    }
  else
    {
      elem = vlrepid->rep;
      succ = *deblist;
      result = pk_true;
      do
	{
	  if (succ->rep == elem)
	    result = pk_false;
	  if (succ->next == NILLREPID)
	    {
	      succ->next = vlrepid;
	      succ = NILLREPID;
	    }
	  else
	    succ = succ->next;
	}
      while (succ != NILLREPID);
    }
  vlrepid->next = NILLREPID;
  return result;
}


boolean searchId(ptlrepid l, ptident id, ptlrepid *vrep)
{
  int i,n;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      if (eq_idents(id, getextidentrepid (l)))
	{
	  *vrep = l;
	  return pk_true; 
	}
      l = succid (l);
    }
  return pk_false;
}
