/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** representation of a list of declarations (kernel representation). */
typedef struct lrepid
{
  ptIdStack rep;		///< stack of the identifier
  ptlrepid next;		///< next element of the list
  ptIdentProp vidprop; /**<
			  Identifier prototype representation (model, ident, type).
			  ident: flow identifier, constant identifdier, index identifier, parameter identifier,...
		       */
} tlrepid;

#define GETIDENPROP(x) (x->vidprop)
#define GETSTACKVAL(x) (x->rep)

/**
   @return a new representation of a model, typed by vid (model, ident, type) 
   with nom as associated stack, and with suiv as next element.
*/
static ptlrepid
mkrepid (ptIdStack nom, ptlrepid suiv, ptIdentProp vid)
{
  ptlrepid vlrepid;

  vlrepid = (ptlrepid) malloc ((unsigned) (sizeof *vlrepid));
  vlrepid->rep = nom;
  vlrepid->next = suiv;
  vlrepid->vidprop = vid;
  return vlrepid;
}


/**
   @return a copy of a model(model), with suiv as next element.
*/
static ptlrepid
duprepid (ptlrepid model, ptlrepid suiv)
{
  return mkrepid (model->rep, suiv, model->vidprop);
}

/**
   It inserts a copy of an element (vlrepid) in a sorted list (deblist) and curlist is the element in the list.
   @return "true"if the element is not already in the list, "false" otherwise.
*/
static boolean
ptlrepid_insertrepid (ptlrepid * deblist, ptlrepid * curlist, ptlrepid vlrepid)
{
  integer numero, cnum;
  ptlrepid succ;
  boolean b;

  if ((*deblist) == NILLREPID)
    {
      *deblist = duprepid (vlrepid, NILLREPID);
      *curlist = *deblist;
      return pk_true;
    }
  numero = atpilegetnum (vlrepid->rep);
  cnum = atpilegetnum ((*deblist)->rep);
  if (numero < cnum)
    {
      *deblist = duprepid (vlrepid, *deblist);
      *curlist = *deblist;
      return pk_true;
    }
  *curlist = *deblist;
  while (cnum > 0)
    if (cnum == numero)
      {
	cnum = 0;
	b = pk_false;
      }
    else
      {
	succ = (*curlist)->next;
	if (succ == NILLREPID)
	  {
	    succ = duprepid (vlrepid, NILLREPID);
	    (*curlist)->next = succ;
	    cnum = 0;
	    b = pk_true;
	  }
	else
	  {
	    cnum = atpilegetnum (succ->rep);
	    if (cnum > numero)
	      {
		succ = duprepid (vlrepid, succ);
		(*curlist)->next = succ;
		cnum = 0;
		b = pk_true;
	      }
	  }
	*curlist = succ;
      }
  return b;
}


/**
   @return "true" if it exists a model (vfi) with id as name in a list of process declaration (list), "false" otherwise.
*/
static boolean
kllrepidGetModelOf (ptlrepid list, ptident id, ptrepfi * vfi)
{
  while (list != NILLREPID)
    {
      *vfi = getModelRepid (list);
      if (eq_idents (id, kl_getNameModel (*vfi)))
	return pk_true;
      list = list->next;
    }
  return pk_false;
}
