/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** 
    @return the ordinal associated with a stack (vs).
*/
extern integer
atpilegetnum (ptIdStack vs)
{
  return vs->num;
}

/**
   @return the identifier associated with a stack (vs).
 */
extern ptident
identpile (ptIdStack vs)
{
  return vs->repident;
}

/**
   @return the stack associated with an identifier (id).
 */
extern ptIdStack
get_stackOf (ptident id)
{
  ptelemHashTable elem;

  if (!IdentTable->findOrInsert (id, &elem))
    elem->setExternalRef ((pointer) kl_mkIdStack (id));
  return ((ptIdStack) (elem->getExternalRef ()));
}


/** 
    @return the current top  of a stack (vs) if it is a model
*/
extern boolean
kl_getTopIfModel (ptIdStack vs, ptrepfi * vfi)
{
  ptIdentProp v;

  v = kl_top (vs);
  if (v == NILIDENTPROP)
    return pk_false;
  if (!ptIdentProp_equal (v, protoModel))
    return pk_false;
  *vfi = ptIdentProp_getProcess (v);
  return pk_true;
}


/** @return "true" if a stack is empty, "false" otherwise. */
extern boolean
kl_isEmptyIdStack (ptIdStack vs)
{
  return stackIsEmpty (getStack (vs));
}


/** 
    @return the stack associated with the "undefined" identifier.
*/
extern ptIdStack
getidnil ()
{
  return (get_stackOf (vidnil));
}

/** @return the top of a stack (vs).*/
extern boolean
kl_top_elem (ptIdStack vs, ptElemStack * v)
{
  return stackTop (getStack (vs), v);
}

extern boolean
kl_top_elem_minus (ptIdStack vs, integer n, ptElemStack * v)
{
  ptStack x;
  integer i;

  if (n <= 0)
    return pk_false;
  i = 0;
  x = getStack (vs);
  while (i < n)
    {
      stackPrevious (x, &x);
      i++;
    }
  return stackTop (x, v);
}
