/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return a new stack associated with an identifier (pid).
 */
extern ptIdStack
kl_mkIdStack (ptident pid)
{
  ptIdStack val;

  numpile = numpile + 1;
  val = (ptIdStack) malloc ((unsigned) (sizeof *val));
  val->repident = pid;
  val->num = numpile;
  val->vStack = NILSTACK;
  return val;
}



/**
   It pushes a parameter (v) in a stack (vs)
*/
extern void
kl_pushParameter (ptIdStack vs, ptIdentProp v)
{
  ptIdentProp_setIdClass (v, par);
  kl_push (vs, v);
}


/**
   It pushes a contant (v) in a stack (vs)
*/
extern void
kl_pushConstant (ptIdStack vs, ptIdentProp v)
{
  ptIdentProp_setIdClass (v, cste);
  kl_push (vs, v);
}

/**
   It pushes a flow (v) in a stack (vs)
*/
extern void
kl_pushFlow (ptIdStack vs, ptIdentProp v)
{
  ptIdentProp_setIdClass (v, sig);
  kl_push (vs, v);
}

/**
   It pushes a label (v) in a stack (vs)
*/
extern void
kl_pushLabel (ptIdStack vs, ptIdentProp v)
{
  ptIdentProp_setIdClass (v, eti);
  kl_push (vs, v);
}


/**
   It pushes a type (v) in a stack (vs)
*/
extern void
kl_pushType (ptIdStack vs, ptIdentProp v)
{
  kl_push (vs, v);
}

/**
   It pushes a type (v) in a stack (vs)
*/
extern void
kl_pushModelType (ptIdStack vs, ptIdentProp v)
{
  kl_push (vs, v);
}


/** It pushes a index (v) in a stack (vs).*/
extern void
kl_pushIndex (ptIdStack vs, ptIdentProp v)
{
  ptIdentProp_setIdClass (v, ind);
  kl_push (vs, v);
}


/** @return "true" if the top of a stack is a type, "false" otherwise */
extern boolean
kl_isTopaType (ptIdStack vs)
{
  ptIdentProp v;

  v = kl_top (vs);
  if (v == NILIDENTPROP)
    return pk_false;
  return ptIdentProp_equal (v, protoType);
}


/** It pushes a model (v) in in a stack (vs). */
extern void
kl_pushModel (ptIdStack vs, ptIdentProp v)
{
  kl_push_elem (vs, (ptElemStack) v);
}


/** It pushes v in a stack vs.*/
extern void
kl_push (ptIdStack vs, ptIdentProp v)
{
  kl_push_elem (vs, (ptElemStack) v);
}


/** @return the top of a stack (vs).*/
extern ptIdentProp
kl_top (ptIdStack vs)
{
  ptElemStack vp;
  if (!kl_top_elem (vs, &vp))
    return NILIDENTPROP;
  return (ptIdentProp) vp;
}

/** It pops the current top of a stack (vs).
*/
extern void
kl_pop (ptIdStack vs)
{
  kl_pop_elem (vs);
}

/** It pushes v in a stack vs.*/
extern void
kl_push_elem (ptIdStack vs, ptElemStack v)
{
  setStack (vs, stackPush (getStack (vs), v));
}


/** It pops the current top of a stack (vs).
*/
extern void
kl_pop_elem (ptIdStack vs)
{
  setStack (vs, stackPop (getStack (vs)));
}
