/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return "true" if a model (vfi) is a type (vrepid) conversion , "false" otherwise.
*/
extern boolean
kl_isTypeConversionModel (ptrepfi vfi, ptlrepid * vrepid)
{
  *vrepid = vfi->ConversionModel;
  return ((*vrepid) != NILLREPID);
}

/** 
    @return the list of the local models of a model (vfi).
*/
extern ptlrepid
kl_getLocModels (ptrepfi vfi)
{
  return vfi->locfi;
}

/**
   @return the list of the local models of a model (vfi).
*/
extern ptlrepid
kl_getModelTypes (ptrepfi vfi)
{
  return vfi->modelTypes;
}

/**
   @return "true" if a model (vfi) is a predefined one, "false" otherwise.
*/
extern boolean
kl_isIntrinsicModel (ptrepfi vfi)
{
  return (vfi->modelType == 1);
}

/**
   @return "true" if a model (vfi) is a formal one (used as parameter), "false" otherwise.
*/
extern boolean
kl_isFormalModel (ptrepfi vfi)
{
  return (vfi->modelType == 2);
}

/**
   @return the local signals of a model (vfi).
*/
extern ptlrepid
kl_getLocalFlowsModel (ptrepfi vfi)
{
  return vfi->sig_loc;
}


/** @return the used modules  (use M1,....,Mn) of a model (vfi).*/
extern ptlrepid
kl_getModules (ptrepfi vfi)
{
  return vfi->lmodules;
}


/** @return the declared labels of a model (vfi). */
extern ptlrepid
kl_getLabelsModel (ptrepfi vfi)
{
  return vfi->labels;
}

/** @return the parameters of a model (vfi). */
extern ptlrepid
kl_getParametersModel (ptrepfi vfi)
{
  return vfi->interface->param;
}


/** @return the local types/constants of a model (vfi). */
extern ptlrepid
kl_getConstantsTypesModel (ptrepfi vfi)
{
  return (vfi->cst_typ_loc);
}

/** @return the name of a model (vfi). */
extern ptident
kl_getNameModel (ptrepfi vfi)
{
  return vfi->nom;
}

/**
   @return the back annotation to the abstract tree of a model (vfi) and 
   the expression definition (expfi) of the model. (See kl_getSpecModel() for the "spec" part).
*/
extern ntrees
kl_getBodyModel (ptrepfi vfi, ptklExprProc * expfi)
{
  *expfi = vfi->corps;
  return vfi->abstractTree;
}

/** @return the back annotation to the abstract tree of a model (vfi). */
extern ntrees
kl_getTreeReference (ptrepfi vfi)
{
  return vfi->abstractTree;
}

/** @return the safety property (safe, unsafe,...) of a model (vfi). */
extern safetyPropertyType
kl_getSafetyProperty (ptrepfi vfi)
{
  return vfi->interface->safetyProperty;
}

/** @return the "spec" part associated with a model (vfi). 
    (See kl_getBodyModel() for the body part).
*/
extern boolean
kl_getSpecModel (ptrepfi vfi, ptklExprProc * vexp)
{
  (*vexp) = vfi->interface->graph_ext;
  return (*vexp != NILEXPFIL);
}


/**
   @return the declared outputs of a model (vfi) and its declared inputs (inputs).
*/
extern ptlrepid
kl_getInsOutsModel (ptrepfi vfi, ptlrepid * inputs)
{
  *inputs = vfi->interface->inList;
  return vfi->interface->outList;
}


/** @return the pragmas associated with a model (vfi).*/
extern ptlPointer
kl_getPragmaModel (ptrepfi vfi)
{
  return vfi->interface->l_pragmas;
}

/** It returns the constants/types (lcst_typ), local signals (lsig), and labels of a model (vfi). */
extern void
kl_getLocalDeclarations (ptrepfi vfi, ptlrepid * lcst_typ, ptlrepid * lsig, ptlrepid * llabels)
{
  (*lcst_typ) = vfi->cst_typ_loc;
  (*lsig) = vfi->sig_loc;
  (*llabels) = vfi->labels;
}

/** @return the error code associaed with a model (vfi) */
extern boolean
kl_getErrorCode (ptrepfi vfi)
{
  return vfi->errorCode;
}

/**
   @return the intrinsics (back annotation) of the compiler, set by kl_setRefToIntrinsics().
 */
extern ptrepfi
kl_getRefToIntrinsics (ptrepfi vfi)
{
  return vfi->lintrinsics;
}

/**
   @return the signals of a model (vfi) referenced outside of it and not declared as outputs.
*/
extern ptlrepid
kl_getExternalReferencesOut (ptrepfi vfi)
{
  return vfi->externRefOut;
}

/**
   @return the signals of a model (vfi) referenced in it and not declared as inputs.
*/
extern ptlrepid
kl_getExternalReferencesIn (ptrepfi vfi)
{
  return vfi->externRefIn;
}

/**
   @return the assertion tag value previously set by kl_setAssertionsFlag().
   When it is "true", the model references the assert() predefined model.
*/
extern boolean
kl_getAssertionsFlag (ptrepfi vfi)
{
  return vfi->bcontainsAssertions;
}


/**
   @return true if a model vfi is a formal one (used as parameter), false otherwise.
   When it is a formal one, the effective model (veff) is returned (it can be null).
*/
extern boolean
kl_getEffectiveModel (ptrepfi vfi, ptrepfi * veff)
{
  if (!kl_isFormalModel (vfi))
    return pk_false;
  *veff = vfi->EffectiveModel;
  return pk_true;
}


/**
   Interface.
   @return the interface of a model (vfi).
*/
extern ptrepinterface
kl_getInterfaceModel (ptrepfi vfi)
{
  return vfi->interface;
}

extern boolean
kl_getCurrentStateVarAutomaton (ptrepfi vfi, ptlrepid * varCurrentStateAutomaton)
{
  *varCurrentStateAutomaton = vfi->varCurrentStateAutomaton;
  return ((*varCurrentStateAutomaton) != NILLREPID);
}


/**
   Interface.
   @return the name of an interface (vinterface).
*/
extern ptident
kl_getInterfaceName (ptrepinterface vinterface)
{
  return vinterface->nom;
}

/**
   Interface.
   @return the inputs assigned to an interface (vinterface).
*/
extern ptlrepid
kl_getInterfaceInputs (ptrepinterface vinterface)
{
  return vinterface->inList;
}

/**
   Interface.
   @return the outputs assigned to an interface (vinterface).
*/
extern ptlrepid
kl_getInterfaceOutputs (ptrepinterface vinterface)
{
  return vinterface->outList;
}

/**
   Interface.
   @return the parameters assigned to an interface (vinterface).
*/
extern ptlrepid
kl_getInterfaceParameters (ptrepinterface vinterface)
{
  return vinterface->param;
}

/**
   Interface.
   @return the specification assigned to an interface (vinterface).
*/
extern ptklExprProc
kl_getInterfaceSpec (ptrepinterface vinterface)
{
  return vinterface->graph_ext;
}

/**
   Interface.
   @return the property (safe, unsafe, deterministic,...) assigned to an interface (vinterface).
*/
extern safetyPropertyType
kl_getInterfaceProperty (ptrepinterface vinterface)
{
  return vinterface->safetyProperty;
}

/**
   Interface.
   @return the pragmas assigned to an interface (vinterface).
*/
extern ptlPointer
kl_getInterfacePragmas (ptrepinterface vinterface)
{
  return vinterface->l_pragmas;
}

/**
   Interface.
   @return the back annotation (reference to the abstract tree produced by the parser) assigned to an interface (vinterface).
*/
extern ntrees
kl_getInterfaceBackAnnotation (ptrepinterface vinterface)
{
  return vinterface->t;
}

/**
   Interface.
   @return the operator assigned to an interface (vinterface).
   This operator specifies that the interface is an interface of a process, a node, a function or an action.
*/
extern noperators
kl_getInterfaceOperator (ptrepinterface vinterface)
{
  return vinterface->op;
}
