/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   Constructor.\n
   @return a new prototype of a process model characterized by its name (nomfi), with p as abstract tree representation.
   Defaults:
  - it is not a predefined model
  - all the lists (inputs, outputs,..) are empty.
  - declared as opsafe
  - without errors
  - without assertions
*/
extern ptrepfi
kl_mkModel (ntrees p, ptident nomfi, noperators op)
{
  ptrepfi vfi;

  vfi = (struct repfi *) malloc ((unsigned) (sizeof *vfi));
  vfi->nom = nomfi;
  vfi->modelType = 0;
  vfi->sig_loc = NILLREPID;
  vfi->cst_typ_loc = NILLREPID;
  vfi->labels = NILLREPID;
  vfi->modelTypes = NILLREPID;
  vfi->locfi = NILLREPID;
  vfi->lmodules = NILLREPID;
  vfi->corps = NILEXPFIL;
  vfi->abstractTree = p;
  vfi->errorCode = pk_false;
  vfi->ConversionModel = NILLREPID;
  vfi->lintrinsics = NILREPFI;
  vfi->externRefOut = NILLREPID;
  vfi->externRefIn = NILLREPID;
  vfi->bcontainsAssertions = pk_false;
  vfi->interface = kl_mkInterface (p, NILIDENT, op);
  vfi->EffectiveModel = NILREPFI;
  vfi->varCurrentStateAutomaton = NILLREPID;
  return vfi;
}

/**
   Constructor.\n
   @return a copy of a model (vp).
*/
extern ptrepfi
kl_DuplicateModel (ptrepfi vp)
{
  ptrepfi vfi;
  ptrepinterface vinter, vinter2;

  vinter2 = vp->interface;
  vfi = kl_mkModel (vp->abstractTree, vp->nom, vinter2->op);
  vfi->modelType = vp->modelType;
  vfi->sig_loc = vp->sig_loc;
  vfi->cst_typ_loc = vp->cst_typ_loc;
  vfi->labels = vp->labels;
  vfi->locfi = vp->locfi;
  vfi->modelTypes = vp->modelTypes;
  vfi->lmodules = vp->lmodules;
  vfi->corps = vp->corps;
  vfi->abstractTree = vp->abstractTree;
  vfi->errorCode = vp->errorCode;
  vfi->ConversionModel = vp->ConversionModel;
  vfi->lintrinsics = NILREPFI;
  vfi->externRefOut = vp->externRefOut;
  vfi->externRefIn = vp->externRefIn;
  vfi->bcontainsAssertions = vp->bcontainsAssertions;
  vfi->varCurrentStateAutomaton = vp->varCurrentStateAutomaton;
  vinter = vfi->interface;
  vfi->EffectiveModel = vp->EffectiveModel;
  // Interface.
  vinter->nom = vinter2->nom;
  vinter->l_pragmas = vinter2->l_pragmas;
  vinter->inList = vinter2->inList;
  vinter->outList = vinter2->outList;
  vinter->param = vinter2->param;
  vinter->graph_ext = vinter2->graph_ext;
  vinter->safetyProperty = vinter2->safetyProperty;
  return vfi;
}



/**
   Constructor.\n
   @return a model for a type (typId) conversion, with id as name.
*/
extern ptrepfi
kl_mkTypeConversionModel (ptident id, ptlrepid typId)
{
  ptrepfi vfi;

  vfi = kl_mkModel (NILTREE, id, opprocess);
  vfi->ConversionModel = typId;
  return vfi;
}

/**
   A model (vfi) is annotated with a list of intrinsics (li).
*/
extern void
kl_setRefToIntrinsics (ptrepfi vfi, ptrepfi li)
{
  vfi->lintrinsics = li;
}

/**
   A model (vfi) is characterized by its expression of definition (body).
*/
extern void
kl_setBodyModel (ptrepfi vfi, ptklExprProc body)
{
  vfi->corps = body;
}


/**
   A model (vfi) is characterized by its "spec" part (v).
*/
extern void
kl_setSpecModel (ptrepfi vfi, ptklExprProc v)
{
  vfi->interface->graph_ext = v;
}

/**
   A model (vfi) is characterized by its safety value(code).
*/
extern void
kl_setSafetyProperty (ptrepfi vfi, safetyPropertyType code)
{
  vfi->interface->safetyProperty = code;
}

/**
  A model (vfi) is characterized by its parameters(lp), its inputs(inputs) and its outputs(outputs).
*/
extern void
kl_setParamsInputsOutputs (ptrepfi vfi, ptlrepid lp, ptlrepid inputs, ptlrepid outputs)
{
  vfi->interface->param = lp;
  vfi->interface->inList = inputs;
  vfi->interface->outList = outputs;
}

/**
  A model (vfi) is characterized by  its local signals.
*/
extern void
kl_setLocalFlowModel (ptrepfi vfi, ptlrepid l)
{
  vfi->sig_loc = l;
}


/**
   A model (vfi) is characterized by the modules (l) it references.
 */
extern void
kl_setModules (ptrepfi vfi, ptlrepid l)
{
  vfi->lmodules = l;
}

/**
   A model (vfi) is characterized by its local models (l).
*/
extern void
kl_setLocModels (ptrepfi vfi, ptlrepid l)
{
  vfi->locfi = l;
}

/**
   A model (vfi) is characterized by its expression of definition (body).
*/
extern void
kl_setModelTypes (ptrepfi vfi, ptlrepid lmodeltypes)
{
  vfi->modelTypes = lmodeltypes;
}


/**
   A model (vfi) is characterized by its label declarations (l).
*/
extern void
kl_setLabelsModel (ptrepfi vfi, ptlrepid l)
{
  vfi->labels = l;
}

/**
  A model (vfi) is characterized by its constants/types declarations (l).
*/
extern void
kl_setConstantsTypesModel (ptrepfi vfi, ptlrepid l)
{
  vfi->cst_typ_loc = l;
}

/**
    A model (vfi) is characterized by its pragmas(l).
*/
extern void
kl_setPragmasModel (ptrepfi vfi, ptlPointer l)
{
  vfi->interface->l_pragmas = l;
}

/**
  A model (vfi) belongs an intrinsic model.
*/
extern void
kl_belongsIntrinsicModel (ptrepfi vfi)
{
  vfi->modelType = 1;
}

/**
  A model (vfi) belongs a formal model (parameter of a model).
*/
extern void
kl_belongsFormalModel (ptrepfi vfi)
{
  vfi->modelType = 2;
}

/**
   A model (vfi) references the assert() model.
*/
extern void
kl_setAssertionsFlag (ptrepfi vfi, boolean b)
{
  vfi->bcontainsAssertions = b;
}


/**
   A model (vfi) is linked with its abstract tree representation (vt).
   It is a back annotation used to stick annotations (messages, errors)
*/
extern void
kl_setTreeReference (ptrepfi vfi, ntrees vt)
{
  vfi->abstractTree = vt;
}

/**
   It specifies that a model (vfi) is erroneous (when cc), correct otherwise.
*/
extern void
kl_setError (ptrepfi vfi, boolean cc)
{
  vfi->errorCode = cc;
}


/**
   It assigns to a model (vfi) the list of the signals (not declared as outputs in vfi) by referenced outside.
*/
extern void
kl_setExternalReferencesOut (ptrepfi vfi, ptlrepid l)
{
  vfi->externRefOut = l;
}

/**
   It assigns to a model (vfi) the list of the signals (not declared as inputs in vfi) by referenced in the model.
*/
extern void
kl_setExternalReferencesIn (ptrepfi vfi, ptlrepid l)
{
  vfi->externRefIn = l;
}

/**
   It assigns to a formal model (vfi) the effective model (veff)
*/
extern void
kl_setEffectiveModel (ptrepfi vfi, ptrepfi veff)
{
  vfi->EffectiveModel = veff;
}

/**
   It assigns to a model (vfi) the operand interface (vinter)
*/
extern void
kl_setInterfaceModel (ptrepfi vfi, ptrepinterface vinter)
{
  vfi->interface = vinter;
}

extern void
kl_setCurrentStateVarAutomaton (ptrepfi vfi, ptlrepid vStateVarCurrentState)
{
  vfi->varCurrentStateAutomaton = vStateVarCurrentState;
}

/**
   Constructor.\n
   @return a new prototype of an interface of a model characterized by its name (id), with p as abstract tree representation.
   This interface is characterized by its operator (ope) (process, function, action, node).
*/
extern ptrepinterface
kl_mkInterface (ntrees p, ptident id, noperators ope)
{
  ptrepinterface vinter;

  vinter = (struct repinterface *) malloc ((unsigned) (sizeof *vinter));
  vinter->nom = id;
  vinter->inList = NILLREPID;
  vinter->outList = NILLREPID;
  vinter->param = NILLREPID;
  vinter->graph_ext = NILEXPFIL;
  vinter->safetyProperty = opsafe;
  vinter->l_pragmas = NILLPOINTER;
  vinter->t = p;
  vinter->op = ope;
  return vinter;
}

/**
   Interface.
   It assigns a name (id) to an interface.
*/
extern void
kl_setInterfaceName (ptrepinterface vinterface, ptident id)
{
  vinterface->nom = id;
}

/**
   Interface.
   It assigns the inputs (l) to an interface (vinterface).
*/
extern void
kl_setInterfaceInputs (ptrepinterface vinterface, ptlrepid l)
{
  vinterface->inList = l;
}

/**
   Interface.
   It assigns the outputs (l) to an interface (vinterface).
*/
extern void
kl_setInterfaceOutputs (ptrepinterface vinterface, ptlrepid l)
{
  vinterface->outList = l;
}

/**
   Interface.
   It assigns the parameters (l) to an interface (vinterface).
*/
extern void
kl_setInterfaceParameters (ptrepinterface vinterface, ptlrepid l)
{
  vinterface->param = l;
}


/**
   Interface.
   It assigns the specifications (exp) to an interface (vinterface).
*/
extern void
kl_setInterfaceSpec (ptrepinterface vinterface, ptklExprProc exp)
{
  vinterface->graph_ext = exp;
}

/**
   Interface.
   It assigns a property(prop) to an interface (vinterface).
   A property is an element of {safe, unsafe, deterministic}.
*/
extern void
kl_setInterfaceProperty (ptrepinterface vinterface, safetyPropertyType prop)
{
  vinterface->safetyProperty = prop;
}

/**
   Interface.
   It assigns the pragmas (prags) to an interface (vinterface).
*/
extern void
kl_setInterfacePragmas (ptrepinterface vinterface, ptlPointer prags)
{
  vinterface->l_pragmas = prags;
}

/**
   Interface.
   It assigns the reference to the abstract tree produced by the parser (vt) to an interface (vinterface).
   The interface is built from the referenced tree.
*/
extern void
kl_setInterfaceBackAnnotation (ptrepinterface vinterface, ntrees vt)
{
  vinterface->t = vt;
}

/**
   Interface.
   It assigns an operator (ope) to an interface (vinterface).
   This operator specifies that the interface is an interface of a process, a node, a function or an action.
*/
extern void
kl_setInterfaceOperator (ptrepinterface vinterface, noperators ope)
{
  vinterface->op = ope;
}
