/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   Kernel tree representation of an interface of a model (process, function, node, action).
*/
typedef struct repinterface
{
  noperators op;		///< process, action, function, automaton,
  ptident nom;			///< name of the interface.
  ptlrepid param,		///< parameters of the model
    inList,			///< inputs of the model
    outList;			///< outputs of the model
  ptlPointer l_pragmas;		///< The  pragmas of the model.
  ptklExprProc graph_ext;	///< Specification  of the process (external graph)
  safetyPropertyType safetyProperty;	///< safety property.
  ntrees t;			///< back annotation on abstract tree.
}
trepinterface;

/**
   Kernel tree representation of a SIGNAL model (process, node, modul,...)
*/
typedef struct repfi
{
  ptident nom;			///< name of the model.
  int modelType;		///< 0  by default, 1 for a predefined process, 2 for a formal process.
  boolean errorCode,		///< error found on the production of the model when it is pk_true
    bcontainsAssertions;	///< with assertions if it its pk_true.
  ntrees abstractTree;		///< back annotation on abstract tree.
  ptlrepid sig_loc,		///< local signals of the model
    cst_typ_loc,		///< Constants and Types of the model
    locfi,			///< Local processes  of the model
    modelTypes,			///< Local model types
    labels,			///< Labels of the model
    lmodules,			///< modules used in the model (use M1,....)
    externRefOut,		///< Signals referenced outsite the model but not declared as outputs.
    externRefIn;		///< Signals referenced in the models but not declared in it.
  ptklExprProc corps;		///< Body of the process (definition)
  ptlrepid ConversionModel;	///< Name of the type for a conversion model (integer replaced by Integer, Integer defined in lintrinsics)
  ptrepfi lintrinsics;		///< List of the intrinsics, used for conversion models.
  ptrepinterface interface;	///< interface of the model.
  ptrepfi EffectiveModel;	///< For a formal process (ie modelType=2), when it is set the current model is associated with.
  // Added for automaton.
  ptlrepid varCurrentStateAutomaton;	///< name given used to reference the current state of an automaton. (using an intrinsic model).
}
trepfi;
