/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return a new identifier (signal, parameter, index, constant) prototype representation
   characterized by its type (typ) an initial value (val), and a back annotation (arbre) to the abstract tree (parser).
   vCategory specifies the category (statevar, shared, or standard signal).
   By default, the class (signal, parameter, index, constant) is no set.
*/
extern ptrepproto
mkrepproto (ptreptyp typ, ptklExprSig val, categoryType vCategory, ntrees arbre, ntrees vtypetree)
{
  ptrepproto reppro;

  reppro = (struct repproto *) malloc ((unsigned) (sizeof *reppro));
  reppro->OTHER.deftyp = typ;
  reppro->OTHER.vali = val;
  reppro->OTHER.vcategory = vCategory;
  reppro->OTHER.typedecl = vtypetree;
  reppro->arbreid = arbre;
  return reppro;
}


extern ptrepproto
mkrepprotoLabel (ntrees arbre)
{
  ptrepproto reppro;

  reppro = (struct repproto *) malloc ((unsigned) (sizeof *reppro));
  reppro->arbreid = arbre;
  reppro->typeId = eti;
  reppro->LABEL.LabelledExpr = NILEXPFIL;
  reppro->LABEL.bStateAutomaton = pk_false;
  return reppro;
}

extern void
setRepprotoType (ptrepproto reppro, ptreptyp typ)
{
  reppro->OTHER.deftyp = typ;
}

extern void
setRepprotoValue (ptrepproto reppro, ptklExprSig val)
{
  reppro->OTHER.vali = val;
}



/**
   It defines the class (vt) of the identifier prototype (signal, parameter, index, constant).
*/
extern void
ptrepproto_setClassId (ptrepproto vproto, typval vt)
{
  vproto->typeId = vt;
}

/**
   It returns the characteristics of an identifier prototype (vproto).
   - its type (vtyp) 
   - its initial value (init), 
   - the state tag status (bstate: pk_true for a signal defined by partial equations).
   - the back annotation (arbre) to the abstract tree.
*/
extern void
ptrepproto_getInfos (ptrepproto vproto, ptklExprSig * init, ptreptyp * vtyp, categoryType * vCategory, ntrees * arbre)
{
  *init = vproto->OTHER.vali;
  *vtyp = vproto->OTHER.deftyp;
  *arbre = vproto->arbreid;
  *vCategory = vproto->OTHER.vcategory;
}

extern ptklExprSig
ptrepproto_getInit (ptrepproto vproto)
{
  return vproto->OTHER.vali;
}

/**
   @return the class of the identifier prototype.
*/
extern typval
ptrepproto_getIdClass (ptrepproto vproto)
{
  return (vproto->typeId);
}

/**
   @return the back annotation to the abstract tree associated with an identifier prototype (vproto).
*/
extern ntrees
ptrepproto_getTree (ptrepproto vproto)
{
  return vproto->arbreid;
}


/**
   @return the back annotation to the type of the abstract tree associated with an identifier prototype (vproto).
*/
extern ntrees
ptrepproto_getTypeTree (ptrepproto vproto)
{
  return vproto->OTHER.typedecl;
}

/**
   When the identifier (vproto) is a label, it returns the expression it labels, NILEXPFIL otherwise.
*/
extern ptklExprProc
ptrepproto_getLabelledExpr (ptrepproto vproto)
{
  if (vproto->typeId == eti)
    return vproto->LABEL.LabelledExpr;
  return NILEXPFIL;
}

/*
  Associate a reference to the expression of process to a label(vproto)
*/
extern void
ptrepproto_setLabelledExpr (ptrepproto vproto, ptklExprProc vexpr)
{
  if (vproto->typeId == eti)
    vproto->LABEL.LabelledExpr = vexpr;
}

extern void
ptrepproto_setLabelStateAutomaton (ptrepproto vproto, boolean b)
{
  if (vproto->typeId == eti)
    vproto->LABEL.bStateAutomaton = b;
}


extern boolean
ptrepproto_isLabelStateAutomaton (ptrepproto vproto)
{
  if (vproto->typeId == eti)
    return vproto->LABEL.bStateAutomaton;
  return pk_false;
}
