/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*! \file IdentifierTable_g.c */

_PUBLIC_
/** Constructor.

   It Creates a hash table for identifiers.
*/
    IdentifierTable::IdentifierTable()
{
  initialize();
}

/** Destructor.
*/
IdentifierTable::~IdentifierTable()
{
}

/**
   It (re)itinializes a table of identifiers.
*/
void
 IdentifierTable::initialize()
{
  int i;
  for (i = 0; i <= lasthash; i++)
    namhashtab[i] = NILELEMHASHTABLE;
}

/** 
  @param vid (in) an identifier.
  @param p (out) : reference to the entry assigned to the identifier in the table.
  @return "True" if the identifier vid is in the table, "False" otherwise : in this case, a new entry
  is added in the table for the identifier. The user must update the value (externalRef) of the elemHashTable object.
*/
boolean IdentifierTable::findOrInsert(ptident vid, ptelemHashTable * p)
{
  short indhash;
  boolean isabsent;
  ptelemHashTable elt;

  indhash = get_short_identTableEntry(vid);
  elt = namhashtab[indhash];
  isabsent = pk_true;
  while (isabsent && (elt != NILELEMHASHTABLE))
    if (eq_idents(elt->getId(), vid))
      isabsent = pk_false;
    else
      elt = elt->getnext();

  if (isabsent) {
    elt = new elemHashTable(vid);
    elt->setNext(namhashtab[indhash]);
    namhashtab[indhash] = elt;
  }
  *p = elt;
  return !isabsent;
}

/**
   Removing the identifier vid from the current table.
   @return true if the operation succeeds, false otherwise (i.e. the identifier is not in the current table).
*/
boolean IdentifierTable::remove(ptident vid)
{
  int indhash;
  ptelemHashTable elt, prev;
  boolean first;

  indhash = get_short_identTableEntry(vid);
  elt = namhashtab[indhash];
  first = pk_true;
  while (elt != NILELEMHASHTABLE)
    if (eq_idents(elt->getId(), vid)) {
      if (first)
	namhashtab[indhash] = elt->getnext();
      else
	prev->setNext(elt->getnext());
      return pk_true;
    }
    else {
      first = pk_false;
      prev = elt;
      elt = elt->getnext();
    }
  return pk_false;
}
