/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* ===> The following ident independent repstr constructors should be moved to repstr related functions */

/* ====> repstr name building with 2 components */

/** XString : Catenate the 2 XString aRepstr1, aRepstr2,
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param repstrResult (out) result of the catenating operation.
*/
extern boolean makeRepstr_repstr_repstr(ptrepstr aRepstr1, ptrepstr aRepstr2, ptrepstr * repstrResult)
{
  return strconc2str(aRepstr1, aRepstr2, pk_false, repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param aRepstr, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_repstr_string(ptrepstr aRepstr, const char *aString, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr(aRepstr, string2repstr(aString), repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param aRepstr, anInt, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_repstr_int(ptrepstr aRepstr, int anInt, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr(aRepstr, integer2repstr(anInt), repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param aRepstr, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_string_repstr(const char *aString, ptrepstr aRepstr, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr(string2repstr(aString), aRepstr, repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param anIdent, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_string_int(const char *aString, int anIndex, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr(string2repstr(aString), integer2repstr(anIndex), repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param anIdent, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_char_int(char aChar, int anIndex, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr(char2repstr(aChar), integer2repstr(anIndex), repstrResult);
}

/* ====> repstr name building with 3 components */

/** XString : Catenate the 3 XString aRepstr1, aRepstr2, aRepstr3,
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param repstrResult (out) result of the catenating operation.
*/
extern boolean makeRepstr_repstr_repstr_repstr(ptrepstr aRepstr1, ptrepstr aRepstr2, ptrepstr aRepstr3, ptrepstr * repstrResult)
{
  if (makeRepstr_repstr_repstr(aRepstr1, aRepstr2, repstrResult))
    return makeRepstr_repstr_repstr(*repstrResult, aRepstr3, repstrResult);
  return pk_false;
}

/** XString : Catenate the 3 XString aRepstr1, aRepstr2, aRepstr3,
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param repstrResult (out) result of the catenating operation.
*/
extern boolean makeRepstr_repstr_string_repstr(ptrepstr aRepstr1, const char *aString, ptrepstr aRepstr2, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr_repstr(aRepstr1, string2repstr(aString), aRepstr2, repstrResult);
}

/** XString : Catenate the 3 XString aRepstr1, aRepstr2, aRepstr3,
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param repstrResult (out) result of the catenating operation.
*/
extern boolean makeRepstr_repstr_string_int(ptrepstr aRepstr, const char *aString, int anInt, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr_repstr(aRepstr, string2repstr(aString), integer2repstr(anInt), repstrResult);
}

/** XString : Catenate the 3 XString aRepstr1, aRepstr2, aRepstr3,
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param repstrResult (out) result of the catenating operation.
*/
extern boolean makeRepstr_repstr_string_char(ptrepstr aRepstr1, const char *aString, char aChar, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr_repstr(aRepstr1, string2repstr(aString), char2repstr(aChar), repstrResult);
}

/** XString : Catenate the 3 XString aRepstr1, aRepstr2, aRepstr3,
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param repstrResult (out) result of the catenating operation.
*/
extern boolean makeRepstr_repstr_char_int(ptrepstr aRepstr1, char aChar, int anIndex, ptrepstr * repstrResult)
{
  return makeRepstr_repstr_repstr_repstr(aRepstr1, char2repstr(aChar), integer2repstr(anIndex), repstrResult);
}

/* ====> repstr name building with 4 components */

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param aRepstr, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean
makeRepstr_repstr_repstr_repstr_repstr(ptrepstr aRepstr1, ptrepstr aRepstr2, ptrepstr aRepstr3, ptrepstr aRepstr4, ptrepstr * repstrResult)
{
  if (makeRepstr_repstr_repstr_repstr(aRepstr1, aRepstr2, aRepstr3, repstrResult))
    return makeRepstr_repstr_repstr(*repstrResult, aRepstr4, repstrResult);
  return pk_false;
}

/* ===> end of ident independent repstr constructors */

/** Constructor.

  @param str : a repstr-string.
  @return a new ptident identifier built from str.
*/
extern ptident repstr2nameIdent(ptrepstr str)
{
  return LL_make_ident_tree(LL_make_nom_atom(str), LL_make_lsuff_emptyList());
}

/** Constructor.

   @param aString : a C string.
   @return a new ptident identifier built from aString.
*/
extern ptident string2ident(const char *aString)
{
  int stringLength = strlen(aString);
  char *stringEnd;
  ptrepstr prefix;
  ntrees identSuffix, identRes;

  stringEnd = (char *)memchr(aString, '_', stringLength);	// pointer to the first occurrence of '_' or null pointer
  if (stringEnd == (char *)NULL)
    identRes = repstr2nameIdent(string2repstr(aString));
  else {
    if (stringEnd == aString)
      prefix = string2repstr("");
    else
      prefix = substring2repstr(aString, stringEnd - aString);
    identSuffix = CStringSplit_makelist(stringEnd + 1, oplsuff, opnom, '_');
    identRes = LL_make_ident_tree(LL_make_nom_atom(prefix), identSuffix);
  }
  return identRes;
}

/**
  @param str : a repstr-string.
  @return a new ptident identifier built from str.
*/
extern ptident repstr2ident(ptrepstr str)
{
  return string2ident(repstr2string(str));
}

/** Constructor.

   @param s : a C string.
   @return a new ptident identifier built from s.
*/
// extern ptident string2ident(const char *s)
// {
//   return repstr2ident(string2repstr(s));
// }

/**
  Conversion.

  @return a string built as the concatenation of all elements of id separated by sep
  @return nilstr if memeory allocation fails
*/
extern ptrepstr ident2repstr_sep(ptident id, char sep)
{
  register ptrepstr vrepstr;
  ptreg vreg;
  noperators op;

  if (id == NILIDENT)
    return nilstr;
  op = noper(id);
  if (op == opnom || op == opop)
    return nstringof(id);	// added
  if (strregnew(&vreg)) {
    vrepstr = appendidtoreg(vreg, id, sep);
    strregfree(vreg);
    return vrepstr;
  }
  /* Memory allocation error */
  return nilstr;
}

/**
  Conversion.

  @return a string built as the concatenation of all elements of id using the suffSeparator as separator.
*/
extern ptrepstr ident2repstr(ptident id)
{
  return ident2repstr_sep(id, suffSeparator);
}

/** 
   Conversion.

   @return the conversion of a ptident to a C-string.
 */
extern char *ident2string(ptident id)
{
  ptrepstr r;
  r = ident2repstr(id);
  return repstr2string(r);
}

/** Constructor.
   Inserts STR as a string element before POS in id.
   @return (modified) id.
*/
extern ptident insert_repstrInIdent(ptident id, ptrepstr str, int pos)
{
  return identInsert_suffix(id, LL_make_nom_atom(str), pos);
}

/** 
   Copying.

   @return a copy of id. 
*/
extern ptident copy_ident(ptident id)
{
  return (ptident) ncopytree(id, pk_false);
}

/** 
   Testing equality.

   @return true if the identifiers id1 and id2 are identical. 
*/
extern boolean eq_idents(ptident id1, ptident id2)
{
  return ntreesEq((ntrees) id1, (ntrees) id2);
}

/**
  Consersion to lower cases.

  @return the string in lower case given by the development of ID ; the suffixes are separated by a 
  suffSeparator ('_') character.
*/
extern ptrepstr ident2lowerRepstr(ptident id)
{
  ptrepstr vrepstr;
  ptreg vreg;

  if (strregnew(&vreg)) {
    vrepstr = regConcStore_lowerCaseId(vreg, id, suffSeparator);
    strregfree(vreg);
    return vrepstr;
  }
  return nilstr;
}

/** Size.
   @param  nbempty [out]: number of nil suffixes.
   @param  p [in]: an identifier.
   @param incrsep [in] : an integer >= 0.
   @return the size of p with incrsep character used to separate suffixes.

*/
extern int identsize_getnbempty(ptident p, int incrsep, int *nbempty)
{
  int lng, i;
  ntrees filsc, fils;
  noperators opera;

  if (IS_NILTREE(p)) {
    *nbempty = 1;
    return 0;
  }
  *nbempty = 0;
  opera = noper(p);
  lng = 0;
  if ((opera == opident) || (opera == opidentfield)) {
    fils = nchild(1, p);
    if (IS_NILTREE(fils))
      *nbempty = (*nbempty) + 1;
    else
      lng = lng + inclng(fils, nbempty);
    fils = nchild(2, p);
    if (ISNOT_NILTREE(p)) {
      opera = noper(fils);
      if (opera == oplsuff)
	if (nisemptyList(fils)) {
	  *nbempty = (*nbempty) + 1;
	  lng = lng + incrsep;
	}
	else
	  for (i = 1; i <= ntreearity(fils); i++) {
	    filsc = nchild(i, fils);
	    lng = lng + incrsep + inclng(filsc, nbempty);
	  }
      else
	lng = lng + incrsep + inclng(fils, nbempty);
    }
  }
  else
    lng = lng + inclng(p, nbempty);
  return lng;
}

/** Size.

   @return the size of p with incrsep character used to separate suffixes.
*/
extern int identsize(ptident p, int incrsep)
{
  int nbempty;

  return identsize_getnbempty(p, incrsep, &nbempty);
}

/**   hash.

     @return the hash code assigned to id ; this value is Tree Machine dependent
 */
extern short hash_ident(ptident id)
{
  return strhash(ident2repstr(id));
}

/**
  catenating.

  @return the concatenating of id1 and id2.
*/
extern ptident makeIdent_ident_ident(ptident id1, ptident id2)
{
  ptrepstr vstr;
  strconc2str(ident2repstr(id1), ident2repstr(id2), pk_false, &vstr);
  return repstr2ident(vstr);
}

/**
  catenating.

  @return the concatenating of aRepstr, aString .
*/
extern ptident makeIdent_ident_repstr(ptident anIdent, ptrepstr aRepstr)
{
  return makeIdent_ident_ident(anIdent, repstr2ident(aRepstr));
}

/**
  catenating.

  @return the concatenating of id1 and id2.
*/
extern ptident makeIdent_ident_string(ptident id1, const char *aString)
{
  ntrees newId = ncopytree(id1, pk_false);
  identAppend_stringSuffix(newId, aString);
  return newId;
}

/**
     ptident operation.

     @return the number of components of the identifier id. */
extern int identGet_componentNumber(ptident id)
{
  noperators op;
  if (id == NILIDENT)
    return 0;
  op = noper(id);
  if (op == opnom || op == opop)
    return 1;
  id = nchild(2, id);
  if (nisemptyList(id))
    return 1;
  return 1 + ntreearity(id);
}

/**
  ptident operation.

  @return the string of ID but the POS first suffixes are forgoten.
  Suffixes are separated by the suffSeparator ('_') character.
*/
extern ptrepstr ideltostr(ptident id, int pos)
{
  int i;
  if (pos == 1)
    return identGet_strPrefix(id);
  i = identGet_componentNumber(id);
  if (i >= pos)
    return identGet_strAtIndexInsuffix(pos, id);
  return nilstr;
}

/** Constructor.
   
   @return a new identifier composed of vident as prefix and num as suffix.
 */
extern ptident makeIdent_ident_int(ptident vident, int num)
{
  int vpos;

  if (ideltostr(vident, 1) == nilstr) {
    vident = copy_ident(vident /*vidempty */ );
    vpos = 2;
  }
  else {
    vident = copy_ident(vident);
    vpos = 1 + identGet_componentNumber(vident);
  }
  identInsert_intSuffix(vident, num, vpos);
  return vident;
}

/** Constructor.
   
   @return a new identifier composed of vident as prefix and num as suffix.
 */
extern ptident makeIdent_ident_int_1(ptident vident, int num)
{
  vident = copy_ident(vident);
  identInsert_intSuffix(vident, num, 2);
  return vident;
}

/**
  catenating.

  @return the concatenating of aRepstr, aString .
*/
extern ptident makeIdent_repstr_string(ptrepstr aRepstr, const char *aString)
{
  return makeIdent_ident_ident(repstr2ident(aRepstr), string2ident(aString));
}

/**
  catenating.

  @return the concatenating of aRepstr, aString .
*/
extern ptident makeIdent_repstr_ident(ptrepstr aRepstr, ptident anIdent)
{
  return makeIdent_ident_ident(repstr2ident(aRepstr), anIdent);
}

/**
  catenating.

  @return the concatenating of aRepstr, aString .
*/
extern ptident makeIdent_repstr_int(ptrepstr aRepstr, int anInt)
{
  return makeIdent_ident_int(repstr2ident(aRepstr), anInt);
}

/**
  catenating.

  @return the concatenating of aString, anIdent .
*/
extern ptident makeIdent_string_ident(const char *aString, ptident anIdent)
{
  return makeIdent_ident_ident(string2ident(aString), anIdent);
}

/**
  catenating.

  @return the concatenating of aString, aRepstr .
*/
extern ptident makeIdent_string_repstr(const char *aString, ptrepstr aRepstr)
{
  return makeIdent_ident_ident(string2ident(aString), repstr2ident(aRepstr));
}

/**
  catenating.

  @return the concatenating of aRepstr, aString .
*/
extern ptident makeIdent_ident_string_int(ptident anIdent, const char *aString, int anInt)
{
  return makeIdent_ident_int(makeIdent_ident_string(anIdent, aString), anInt);
}

/**
  catenating.

  @return the concatenating of aString1, aRepstr, aString2, anIdent.
*/
extern ptident makeIdent_string_repstr_string_ident(const char *aString1, ptrepstr aRepstr, const char *aString2, ptident anIdent)
{
  ptident identres;
  identres = makeIdent_string_repstr(aString1, aRepstr);
  identres = makeIdent_ident_string(identres, aString2);
  return makeIdent_ident_ident(identres, anIdent);
}

/** 
   Constuctor.

   @return vident if vident is an atom, or if vident is a identifier with non-nil value for first child.
   A copy of the "empty" identifier is returned otherwise.
*/
extern ptident getSignalIdentifier(ptident vident)
{
  if (noper(vident) != opident) return vident;
  if (ideltostr(vident, 1) != nilstr) return vident;
  // return copy_ident (vidempty);
  return copy_ident(vident);
}

/** Copying.

  @return a copy of vident, in which the suffix is replaced by num.
*/
extern ptident identChangeSuffix(ptident vident, int vnum)
{
  ptrepstr vsuff;

  vsuff = integer2repstr(vnum);
  vident = copy_ident(vident);
  identChange_inSuffix(vident, vsuff, identGet_componentNumber(vident));
  return vident;
}

/** Copying.
    @return a copy of vident, in which the prefix is replaced by Lnum.  
*/
extern ptident identChangePrefix(ptident vident, int num)
{
  noperators op;
  ptrepstr newPrefix;

  newPrefix = make_localRepstr(num);
  op = noper(vident);
  if (op == opnom || op == opop)
    return repstr2ident(newPrefix);
  return identMake_identSuffix(newPrefix, vident);
}

/** Constructor.
   
   @return a new identifier composed of vident as prefix and num as suffix.
 */
extern ptident newIdent_intSuffixed2(ptident vident, int num)
{
  ptident id;
  int i, maxi;

  id = make_localIdent(num);
  maxi = identGet_componentNumber(vident);
  for (i = 1; i <= maxi; i++)
    insert_repstrInIdent(id, ideltostr(vident, i), i + 1);
  return id;
}

/** 
  Constructor.
  @return a new identifier copy of id and vstr added as last suffix.
 */
extern ptident createSuffIdent(ptident id, ptrepstr vstr)
{
  int vpos;
  if (vstr != nilstr) {
    id = copy_ident(id);
    vpos = identGet_componentNumber(id) + 1;
    insert_repstrInIdent(id, vstr, vpos);
  }
  return id;
}

/** 
  Constructor.
  @return a new identifier copy of id and vstr added as last suffix.
 */
extern ptident newIdent_concIdentIdent(ptident id1, ptident id2)
{
  return createSuffIdent(id1, ident2repstr(id2));
}

/** Constructor.
   
   @return a new identifier composed of vident as prefix and num as suffix.
 */
extern ptident newIdent_stringSuffixed(ptident vident, const char *suf)
{
  return createSuffIdent(vident, string2repstr(suf));
}

/**
   Constructor.
   @return a new identifier copy of id and vstr added as last suffix, name of field structure/bundle.
*/
extern ptident createSuffIdentField(ptident id, ptrepstr vstr)
{
  int vpos;
  if (vstr != nilstr) {
    id = copy_ident(id);
    nsetop(id, opidentfield);
    vpos = identGet_componentNumber(id) + 1;
    insert_repstrInIdent(id, vstr, vpos);
  }
  return id;
}

/**
   It reducecs the size of the identifier(id) ...
   Currently, return id (TO BE COMPLETED).
*/
extern ptident reduceIdent(ptident id, int lgmaxId)
{
  int l, n;

  return id;
  // to be completed. Not Tested
  l = identsize(id, 1);
  n = identGet_componentNumber(id);	// nb components of id.
  while ((n > 2) && (l > lgmaxId)) {
    ndestroyListElem(nchild(2, id), 1);
    l = identsize(id, 1);
    n--;
  }
  return id;
}

/** 
   List of identifiers managing.
   It inserts the identifier id as pos-th element in the list lid.
*/
extern void namlisput(ptidentl lid, ptident id, int pos)
{
  ninsertleft(pos, id, lid);
}

/**
  List of identifiers managing.
  
  @return list of empty list identifier.
 */
extern ptidentl namlfridempty(void)
{
  return LL_make_lident_emptyList();
}

/** 
   List of identifiers managing.
   
   @return true if id is a list of identifier.
 */
extern boolean namislident(ptident id)
{
  return (noper(id) == oplident);
}

/** 
   List of identifiers managing.
   
   @return a list-tree de type OPLIDENT with s as element.
 */
extern ptidentl namlfrid(ptident s)
{
  return LL_make_lident_list(s);
}

/** Predefined identifier.
    @return  identifier assigned to the empty string. */
extern ptident getIdentNil(void)
{
  return vidnil;
}

/**
   List of identifiers managing.
   @return the pos-th element of the list lid.
   \pre size of(lid) >= pos.
 */
extern ptident namidfrl(ptidentl lid, int pos)
{
  return nchild(pos, lid);
}

/* ====> ptident related makeRepstr*/
/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param anIdent, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_string_ident(const char *aString, ptident anIdent, ptrepstr * repstrResult)
{
  return strconc2str(string2repstr(aString), ident2repstr(anIdent), pk_false, repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param anIdent, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_ident_repstr(ptident anIdent, ptrepstr aRepstr, ptrepstr * repstrResult)
{
  return strconc2str(ident2repstr(anIdent), aRepstr, pk_false, repstrResult);
}

/** XString : Catenating an XString followed by a char * string
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param anIdent, aString, (in) two string.
  @param repstrResult (out) result of the catenating operation of str2 to str1.
*/
extern boolean makeRepstr_ident_string(ptident anIdent, const char *aString, ptrepstr * repstrResult)
{
  return strconc2str(ident2repstr(anIdent), string2repstr(aString), pk_false, repstrResult);
}

/**
  Initializing.

  @return none.
*/
extern void namini()
{
  vidempty = string2ident(cstrempty);
  vidnil = repstr2ident(nilstr);
}

/**
  Finalizing.

  @return none.
*/
extern void namend()
{

}
