/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return the identifier used for coding an input node. */
extern ptident getReadId()
{
  return idinputenv;
}

/** @return the identifier used for coding an input node (present). */
extern ptident getPresentId()
{
  return idPresent;
}

/** 
@return true if id is the "present" instrinsic function.
*/
extern boolean isPresentId(ptident id)
{
  return (eq_idents(id, idPresent));
}

/** 
@return the identifier used for coding an output node. */
extern ptident getWriteId()
{
  return idoutputenv;
}

/** 
@return the prefix identifier used for the name of the boolean clocks. */
extern ptident getPrefixBooleanClock(void)
{
  return copy_ident(idPrefixBooleanClock);
}

/** 
@return the identifier id prefixed by the boolean clock prefix. */
extern ptident namBooleanClockfrom(ptident id)
{
  ptident idC;
  int n, i;
  ntrees lsuff;

  idC = getPrefixBooleanClock();
  if (noper(id) != opident)
    return insert_repstrInIdent(idC, nstringof(id), 2);
  idC = insert_repstrInIdent(idC, ideltostr(id, 1), 2);
  lsuff = nchild(2, id);
  n = ntreearity(lsuff);
  if (n > 0)
    for (i = 1; i <= n; i++)
      idC = insert_repstrInIdent(idC, ideltostr(lsuff, i), i + 2);
  return idC;
}

/** 
    @return the name of the predefined init clock.
*/
extern ptident getInitClockId(void)
{
  return vstralpha;
}

/** 
@return the name of the predefined null clock. */
extern ptident getNullClockId(void)
{
  return vstrnull;
}

/**
   @return the name used for the IOStep clock
*/
extern ptident getIOStepClockId(void)
{
  return vstrIOStep;
}

/** 
@return the name of the predefined upper bound clock (fictitious root clock). */
extern ptident getUpbClockId(void)
{
  return vstrhupb;
}

/** 
@return the name for label. */
extern ptident getLabelId()
{
  return vidLabel;
}

/** 
@return the prefix identifier used for builting the name of the clocks. */
extern ptident getClockId()
{
  return vstrhorl;
}

/** 
@return the prefix identifier used for builting the name of the types. */
extern ptident getTypeId()
{
  return vidType;
}

/** 
@return the prefix identifier used for builting the name of the indexes. */
extern ptident getIndexId()
{
  return vidIndex;
}

/** 
@return the prefix identifier used for builting the name of the constants. */
extern ptident getConstId()
{
  return vidConst;
}

/** 
@return true if id is the "STATUS" instrinsic function.
*/
extern boolean isStatusId(ptident id)
{
  return (eq_idents(id, vstrstatus));
}

/** 
@return the identifier of the internal predefined function "STATUS". */
extern ptident getStatusId()
{
  return vstrstatus;
}

/** 
@return the identifier used to built an exception node.
   It is a internal intrinsic function. */
extern ptident getExceptionId()
{
  return vidException;
}

/** 
@return true if id is the identifier used to specify an exception. */
extern boolean isExceptionId(ptident id)
{
  return eq_idents(id, vidException);
}

/** 
@return the identifier of the ASSERTION intrinsic function. */
extern ptident getAssertionId()
{
  return vstrassert;
}

/** 
@return true if id is the identifier of the "assert" intrinsic function. */
extern boolean isAssertionId(ptident id)
{
  return (eq_idents(id, vstrassert));
}

/** 
@return the prefix identifier used for builting the name of the delays. */
extern ptident getDelayId()
{
  return NameAssociatedWithOperatorTable[opret];
}

/** 
@return the prefix identifier used for builting the name of the windows. */
extern ptident getWindowId()
{
  return NameAssociatedWithOperatorTable[opfen];
}

/** 
@return the prefix identifier used for builting the name of the sub expressions. */
extern ptident getSubexprId()
{
  return vSubExpr;
}

/** 
    @return the prefix identifier used for builting the name of the sub expressions in "when E". */
extern ptident getConditionId()
{
  return NameAssociatedWithOperatorTable[opwhen1];
}

/** 
    @return the prefix identifier used for builting the name of the sub expressions in x[E]. */
extern ptident getElemAccessId()
{
  return NameAssociatedWithOperatorTable[opelem];
}

/** 
    @return true if id is the identifier of the "dim" intrinsic function.
*/
extern boolean isDimId(ptident id)
{
  return (eq_idents(id, vstrdim));
}

/** 
@return true if id is the identifier of the "size" intrinsic function. */
extern boolean isSizeId(ptident id)
{
  return (eq_idents(id, vstrsize));
}

/** 
@return true if id is the identifier of the "bounds" intrinsic function. */
extern boolean isBoundsId(ptident id)
{
  return (eq_idents(id, vstrbounds));
}

/** 
@return true if id is the identifier of the "Nil_Process" intrinsic function. */
extern boolean isNilProcId(ptident id)
{
  return (eq_idents(id, vnilproc));
}

/** 
@return true if id is the identifier of the "indices" intrinsic function. */
extern boolean isIndicesId(ptident id)
{
  return (eq_idents(id, vstrindices));
}

/** 
@return the identifier of the "indices" intrinsic function.
*/
extern ptident getIndicesId()
{
  return vstrindices;
}


/** 
@return true if id is the identifier of the cartesian product (internal) intrinsic function. */
extern boolean isCartesianProductId(ptident id)
{
  return (eq_idents(id, vstrCartesianProduct));
}

/** 
@return the identifier of the cartesian product (internal) intrinsic function. */
extern ptident getCartesianProductId()
{
  return vstrCartesianProduct;
}

/** @return True if ?ID is the identifier sample. */
extern boolean isSampleId(ptident id)
{
  return eq_idents(id, vstrsample);
}

/** @return the identifier of the sample (internal) intrinsic function. */
extern ptident getSampleId()
{
  return vstrsample;
}

/** @return True if ?ID is the identifier clk_affine. */
extern boolean isAffineClkId(ptident id)
{
  return eq_idents(id, vstraffineclk);
}

/** @return the identifier of the affine clock intrinsic function. */
extern ptident getAffineClkId()
{
  return vstraffineclk;
}

/** @return True if ?ID is the identifier unsample */
extern boolean isUnsampleId(ptident id)
{
  return eq_idents(id, vstrunsample);
}

/** @return the identifier of the unsample (internal) intrinsic function. */
extern ptident getUnsampleId()
{
  return vstrunsample;
}

/** 
    @return the identifier used to export (Signal program) the clock hierarchy.
*/
extern ptident getSubTreeId()
{
  return idSubTreeOf;
}

extern ptident getConstraintsId()
{
  return idConstraints;
}

extern ptident getMicroStepName()
{
  return idMicroStep;
}

extern ptident getIdUndersampling(boolean bpos)
{
  if (bpos)
    return TT_id;
  else
    return FF_id;
}

/**
   @return the operator for an type identifier (strnom).
   For an unknown predefined type name, it returns the opnil operator.
*/
extern noperators identType2SignalType(ptrepstr strnom)
{
  if (streq(strnom, string2repstr(cvBoolean)))
    return oplogical;
  if (streq(strnom, string2repstr(cvInteger)))
    return opinteger;
  if (streq(strnom, string2repstr(cvShort)))
    return opshort;
  if (streq(strnom, string2repstr(cvLong)))
    return oplong;
  if (streq(strnom, string2repstr(cvReal)))
    return opreal;
  if (streq(strnom, string2repstr(cvDreal)))
    return opdpreal;
  if (streq(strnom, string2repstr(cvEvent)))
    return opevent;
  if (streq(strnom, string2repstr(cvComplex)))
    return opcomplex;
  if (streq(strnom, string2repstr(cvDcomplex)))
    return opdcomplex;
  if (streq(strnom, string2repstr(cvChar)))
    return opchar;
  if (streq(strnom, string2repstr(cvString)))
    return opstring;
  return opnil;
}

/**
   @return the string assigned to an operator of a predefined type (ope).
   For other operators, it returns the nil string.
*/
extern ptrepstr predefinedType2ident(noperators ope)
{
  switch ((int)ope) {
  case opevent:
    return string2repstr(cvEvent);
  case oplogical:
    return string2repstr(cvBoolean);
  case opshort:
    return string2repstr(cvShort);
  case opinteger:
    return string2repstr(cvInteger);
  case oplong:
    return string2repstr(cvLong);
  case opreal:
    return string2repstr(cvReal);
  case opdpreal:
    return string2repstr(cvDreal);
  case opcomplex:
    return string2repstr(cvComplex);
  case opdcomplex:
    return string2repstr(cvDcomplex);
  case opchar:
    return string2repstr(cvChar);
  case opstring:
    return string2repstr(cvString);
  }
  return nilstr;
}

/**
 */
extern ptident getPredicateId()
{
  return vPredicate;
}

extern boolean getNameAssociatedWithOperator(noperators nop, ptident * id)
{
  if (NameAssociatedWithOperatorTable.find(nop) == NameAssociatedWithOperatorTable.end())
    return pk_false;
  (*id) = NameAssociatedWithOperatorTable[nop];
  return pk_true;
}

extern ptident getPreviousId()
{
  return previousId;
}

extern ptident getStrVar()
{
  return idVar;
}

extern ptident getAnonymousId()
{
  return AnonymousId;
}

extern ptident getTransitionId()
{
  return transitionId;
}

extern ptident getAutomatonClockId()
{
  return AutomatonClockId;
}

extern ptident getAutomatonSetClockId()
{
  return AutomatonSetClockId;
}


extern ptident getInitialStateId()
{
  return initialStateId;
}

extern ptident getAutomatonIsIn()
{
  return AutomatonIsIn;
}

/** 
    @return true if id is the identifier of the "Current State" intrinsic function.
 */
extern boolean isAutomatonIsIn(ptident id)
{
  return (eq_idents(id, AutomatonIsIn));
}

/**
   @return true if id is the identifier of the "SetCurrentStateId" intrinsic function.
 */
extern ptident getSetCurrentStateId()
{
  return SetCurrentStateId;
}

/** 
    @return true if id is the identifier of the "Current State" intrinsic function. */
extern boolean isSetCurrentStateId(ptident id)
{
  return (eq_idents(id, SetCurrentStateId));
}

/** 
    @return true if id is the identifier of the "transition" intrinsic function. */
extern boolean isTransitionId(ptident id)
{
  return (eq_idents(id, transitionId));
}

/** 
    @return true if id is the identifier of the "automaton clock" intrinsic function. */
extern boolean isAutomatonClockId(ptident id)
{
  return (eq_idents(id, AutomatonClockId));
}


/** 
    @return true if id is the identifier of the "automaton clock" intrinsic function. */
extern boolean isAutomatonSetClockId(ptident id)
{
  return (eq_idents(id, AutomatonSetClockId));
}


/** 
    @return true if id is the identifier of the "Initial state" intrinsic function. */
extern boolean isInitialStateId(ptident id)
{
  return (eq_idents(id, initialStateId));
}

extern boolean isSendId(ptident id)
{
  return (eq_idents(id, idSend));
}

extern boolean isReceiveId(ptident id)
{
  return (eq_idents(id, idReceive));
}

extern ptident getSendId()
{
  return idSend;
}

extern ptident getReceiveId()
{
  return idReceive;
}

extern ptident getDispatchId()
{
  return DispatchId;
}

extern ptident getCompleteId()
{
  return  CompleteId;
}


extern ptident getThreadClockId()
{
  return ThreadClockId;
}

/** 
    @return true if id is the identifier of the "Thread_Clock" intrinsic function. */
extern boolean isThreadClockId(ptident id)
{
  return (eq_idents(id, ThreadClockId));
}

extern ptident getThreadBooleanClockId()
{
  return ThreadBooleanClockId;
}

/** 
    @return true if id is the identifier of the "Thread_Clock" intrinsic function. */
extern boolean isThreadBooleanClockId(ptident id)
{
  return (eq_idents(id, ThreadBooleanClockId));
}


/** 
   Initializing.

   predefined identifier table (intrinsics,prefixes, suffixes) initializing.
*/
extern void predefSignalNameIni()
{
  AnonymousId = string2ident("XZX");

  vstrassert = string2ident(cstrassert);
  vstrstatus = string2ident(cstrstatus);

  vstrhorl = string2ident(cstrhorl);
  vstralpha = createSuffIdent(vstrhorl, ideltostr(string2ident(cstralpha), 1));
  vstrhupb = createSuffIdent(vstrhorl, ideltostr(string2ident(cstrhupb), 1));
  vstrnull = createSuffIdent(vstrhorl, ideltostr(string2ident(cstnull), 1));

  idinputenv = string2ident(cstinenv);
  idoutputenv = string2ident(cstoutenv);
  idPresent = string2ident("present");

  vidLabel = string2ident(cst_nlabel);
  vidType = string2ident(cst_Type);
  vidConst = string2ident(cst_cst);
  vidIndex = string2ident(cst_I);
  vidException = string2ident(cst_exception);

  vstrbounds = string2ident(cst_bounds);
  vstrindices = string2ident(cst_indices);
  vstrsize = string2ident(cst_size);
  vstrdim = string2ident(cst_dim);
  vstrCartesianProduct = string2ident(cst_CartesianProduct);

  vstrsample = string2ident(cst_sample);
  vstraffineclk = string2ident(cst_affineclk);
  vstrunsample = string2ident(cst_unsample);

  vSubExpr = string2ident(cst_subexpr);
  previousId = string2ident("previous");
  vnilproc = string2ident(StrNilProc);
  idSubTreeOf = string2ident(cst_idSubTreeOf);
  idConstraints = string2ident(cst_constraints);
  idMicroStep = string2ident(cst_MicroStep);
  idPrefixBooleanClock = string2ident(cst_PrefixBooleanClock);
  TT_id = string2ident("TT");
  FF_id = string2ident("FF");
  idVar = string2ident("var");
  vPredicate = string2ident("PREDICATE");
  vstrIOStep = string2ident("IOStep");

  idSend = string2ident("pK_Send");
  idReceive = string2ident("pK_Receive");

  // Predefined names for automata: they must be the same than those declared in $SignalLib_Std/AUTOMATON_INTRINSICS.SIG.
  transitionId = string2ident("Automaton_Transition");
  AutomatonClockId = string2ident("Automaton_Clock");
  AutomatonSetClockId = string2ident("Automaton_SetClock");
  initialStateId = string2ident("Automaton_Initial_State");
  AutomatonIsIn = string2ident("Automaton_IsIn");
  SetCurrentStateId = string2ident("Automaton_SetCurrentStateId");

  regExpStarId = string2ident("Regular_Star");
  regExpConcId = string2ident("Regular_Conc");
  regExpSumId = string2ident("Regular_Sum");
  regExpPowId = string2ident("Regular_Pow");

  // The identifiers associated with the some operators
  NameAssociatedWithOperatorTable[opleft_tt] = string2ident("extended_");
  NameAssociatedWithOperatorTable[opmemo] = string2ident("cell");
  NameAssociatedWithOperatorTable[opvar] = string2ident("cell");
  NameAssociatedWithOperatorTable[opnombm] = string2ident("count");
  NameAssociatedWithOperatorTable[opnomba] = string2ident("expanded_counterAfter");
  NameAssociatedWithOperatorTable[opnombd] = string2ident("expanded_counterFrom");
  NameAssociatedWithOperatorTable[opret] = string2ident("pre");
  NameAssociatedWithOperatorTable[opfen] = string2ident("win");
  NameAssociatedWithOperatorTable[opelem] = string2ident("elem");
  NameAssociatedWithOperatorTable[opwhen1] = string2ident("cond");

  // DispatchId,CompleteId, ThreadClockId, ThreadBooleanClockId are used in the AADL2SIGNAL translation
  DispatchId = string2ident("Dispatch_Thread"); 
  CompleteId = string2ident("Complete_Thread"); 
  ThreadClockId = string2ident("Thread_Clock"); 
  ThreadBooleanClockId  = string2ident("Thread_BooleanClock");
}
