/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   The parser is never called in interactive mode.
   This Unix function is here for portability.
*/
extern int isatty(int x)
{ return 0;}

#if xKHOST == xKwinNT
/**
   The parser is never called in interactive mode.
   This Unix function is here for portability.
*/
static int fileno(FILE *x) { return 0;}
#endif

// Bison dependent.
// xKHOST == xKwinNT, xKsolaris, xKlinux xKmacOs
  #if xKHOST == xKmacOs
  #define INC_DECL char *text
  #else
  #define INC_DECL 
  #endif

#define yy_current_buffer YY_CURRENT_BUFFER_LVALUE
#define YY_NO_UNISTD_H 1 /* to avoid  <unistd.h> */


/* The following structure must be the same than the generated one in signal_bison.c  */
/* typedef struct YYLTYPE YYLTYPE;
struct YYLTYPE
{
  int first_line;
  int first_column;
  int last_line;
  int last_column;
};
#define yyltype YYLTYPE 
#define YYLTYPE_IS_DECLARED 1
#define YYLTYPE_IS_TRIVIAL 1
*/

typedef YYLTYPE *ptyyltype;
#define yyltype YYLTYPE
#define ptYYLTYPE ptyyltype

#define YY_DECL int siglex(void*, ptYYLTYPE llocp)


#define YY_NEVER_INTERACTIVE 1

#define NBPHYLA 54
static const char* phylaTable[NBPHYLA] =
{
"[ATTR]",
"[BORN]",
"[CAS]",
"[CHAI]",
"[CST_CHA]",
"[DEC_S]",
"[DESC]",
"[DESC_MODU]",
"[ELE_VEC]",
"[ENT_SOR]",
"[I_TYPE]",
"[IDENT]",
"[IND_MOT]",
"[L_CAS]",
"[L_CONS]",
"[L_E_SIGN]",
"[L_IDENT]",
"[L_PARC]",
"[L_SIGN]",
"[MODU]",
"[O_SPEC]",
"[OL_ATTR]",
"[OL_DECL]",
"[OL_DEC_S]",
"[OL_E_PARA]",
"[OL_OPRA]",
"[OL_PRAG]",
"[PARC]",
"[PRAG]",
"[DECL]",
"[DEF_CONS]",
"[DEF_TYPE]",
"[E_CST]",
"[ENU_VAL]",
"[IDENT_OP]",
"[INTE]",
"[O_CHAI]",
"[O_IDENT]",
"[OL_E_SIGN]",
"[REF_PRO]",
"[SIGN]",
"[E_PROC]",
"[TYPE]",
"[O_E_PROC]",
"[TYPE_S]",
"[V_PROC]",
"[V_TYPE]",
"[O_CORP]",
"[E_SIGN]",
"[LSIG]",
"[D_E_SIGN]",
"[E_PARA]",
"[O_E_SIGN]",
"[V_CONS]"
};

#define errSyntax 1 
#define errunrecchar 2

static boolean ADDEDPHYLUM=0;
static integer VPHYLUM=without_phylum;
static boolean PHYLUMHASBEENADDED=0;
static ptrepstr __siganafilename__;

#define LINENUMBER(x) (ADDEDPHYLUM ? (x-1) : x)


static boolean ongoing_comments = false;
static ntrees commentsTree;
static ntrees commentLine;



