%option 8bit
%s entrypoint sentences comment
%{
#define free(x) free((char *)x)

#define YY_USER_INIT { BEGIN(entrypoint); }

#include <string.h>

static int sigline   = 0;
static int sigcolumn = 0;
static int sigi;


static void siglexposition(ptYYLTYPE loc)
{
  loc->first_line   = sigline;
  loc->first_column = sigcolumn;

  for (sigi=0 ; sigi<yyleng ; sigi++) 
     if (yytext[sigi]=='\n') { sigcolumn = 0; sigline++; } else sigcolumn++;

  loc->last_line   = sigline;
  loc->last_column = sigcolumn;
}


static void saveComments(ptYYLTYPE loc)
{
  ptrepstr r;
  if (!string2Srepstr(yytext, &r))
     pK_messenger -> parsing_errorLineColumn(__siganafilename__, LINENUMBER(loc->first_line), LINENUMBER(loc->first_column), erroutofboundstring);
  // printf("DEBUG saveComments=%s\n",yytext);
  commentLine = LL_make_comment_atom(r);
  if (ongoing_comments)
     commentsTree = ntreeList_appendLast(commentsTree,commentLine);
  else
     { commentsTree = LL_make_comments_list(commentLine);
       ongoing_comments = true;
     }   
}

static void Unput(int c);

#define siglexreturn(tk) { siglexposition(llocp); return tk; }

extern "C" int yywrap(void) { return 1;}

/* PHYLUM management */
static int insertPHYLUM(char *buf, int max_size)
{
  int i, c= '*', n, lg;
  const char * strphylum;

  strphylum = phylaTable[VPHYLUM];
  lg = strlen(strphylum);
  for ( n = 0; n < lg ; n++ ) buf[n] = strphylum[n];
  buf[n] = '\n';
  n++;
  for (; n < max_size && (c = getc( yyin )) != EOF ; n++ ) buf[n] = (char) c;
  //  if ( c == EOF && ferror( yyin ) ) YY_FATAL_ERROR( "input in flex scanner failed" ); 
  PHYLUMHASBEENADDED = 1; 
  return n;
}

/* ------------IMPORTED HERE TO TAKE INTO ACCOUNT PHYLUM management --- */
/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
    if (ADDEDPHYLUM && !PHYLUMHASBEENADDED) result = insertPHYLUM(buf, max_size); \
    else if ( ((result = fread( buf, 1, max_size, yyin )) == 0) && ferror( yyin ) ) \
	    /* YY_FATAL_ERROR( "input in flex scanner failed" ) */ ;
#endif

%}

ss \\a
ea \\b
sp \\f
sl \\n
rc \\r
th \\t
tv \\v
as \\\\
dq \\\"
sq \\'
pi \\\?
pc \\\%
echap {ss}|{ea}|{sp}|{sl}|{rc}|{th}|{tv}|{as}|{dq}|{sq}|{pi}|{pc}
octal \\[0-7]([0-7][0-7]?)?
car_hexa [a-fA-F0-9]
hexa \\x{car_hexa}{car_hexa}?
code_car {echap}|{octal}|{hexa}
format [\t\14\15]
car_accent []
%%

<comment>\n                          { siglexposition(llocp); break; }
<comment>([^%\\\n]*|{code_car})*$    { siglexposition(llocp); saveComments(llocp); break; }
<comment>([^%\\\n]*|{code_car})*\%   { siglexposition(llocp);
                                       yytext[yyleng-1]=0;
                                       saveComments(llocp);
                                       BEGIN sentences;
                                       break; };
<sentences>\n                        { siglexposition(llocp); break; }
<sentences>" "|{format}              { siglexposition(llocp); break; }

<sentences>";" siglexreturn(KW_PVIR);
<sentences>"," siglexreturn(KW_VIRG);
<sentences>"module" siglexreturn(KW_module);
<sentences>"MODULE" siglexreturn(KW_module);
<sentences>"=" siglexreturn(KW_EQUL);
<sentences>"end" siglexreturn(KW_end);
<sentences>"END" siglexreturn(KW_end);
<sentences>"use" siglexreturn(KW_use);
<sentences>"USE" siglexreturn(KW_use);
<sentences>"external" siglexreturn(KW_external);
<sentences>"EXTERNAL" siglexreturn(KW_external);
<sentences>"process" siglexreturn(KW_process);
<sentences>"PROCESS" siglexreturn(KW_process);
<sentences>"private" siglexreturn(KW_private);
<sentences>"PRIVATE" siglexreturn(KW_private);
<sentences>"node" siglexreturn(KW_node);
<sentences>"NODE" siglexreturn(KW_node);
<sentences>"function" siglexreturn(KW_function);
<sentences>"FUNCTION" siglexreturn(KW_function);
<sentences>"action" siglexreturn(KW_action);
<sentences>"ACTION" siglexreturn(KW_action);
<sentences>"procedure" siglexreturn(KW_procedure);
<sentences>"PROCEDURE" siglexreturn(KW_procedure);
<sentences>"automaton" siglexreturn(KW_automaton);
<sentences>"AUTOMATON" siglexreturn(KW_automaton);
<sentences>"operator" siglexreturn(KW_operator);
<sentences>"OPERATOR" siglexreturn(KW_operator);
<sentences>"{" siglexreturn(KW_BOUV);
<sentences>"}" siglexreturn(KW_BFER);
<sentences>"(" siglexreturn(KW_POUV);
<sentences>")" siglexreturn(KW_PFER);
<sentences>"?" siglexreturn(KW_INTE);
<sentences>"!" siglexreturn(KW_XCLA);
<sentences>"safe" siglexreturn(KW_safe);
<sentences>"SAFE" siglexreturn(KW_safe);
<sentences>"deterministic" siglexreturn(KW_deterministic);
<sentences>"DETERMINISTIC" siglexreturn(KW_deterministic);
<sentences>"unsafe" siglexreturn(KW_unsafe);
<sentences>"UNSAFE" siglexreturn(KW_unsafe);
<sentences>"spec" siglexreturn(KW_spec);
<sentences>"SPEC" siglexreturn(KW_spec);
<sentences>"pragmas" siglexreturn(KW_pragmas);
<sentences>"PRAGMAS" siglexreturn(KW_pragmas);
<sentences>"type" siglexreturn(KW_type);
<sentences>"TYPE" siglexreturn(KW_type);
<sentences>"init" siglexreturn(KW_init);
<sentences>"INIT" siglexreturn(KW_init);
<sentences>"event" siglexreturn(KW_event);
<sentences>"EVENT" siglexreturn(KW_event);
<sentences>"boolean" siglexreturn(KW_boolean);
<sentences>"BOOLEAN" siglexreturn(KW_boolean);
<sentences>"short" siglexreturn(KW_short);
<sentences>"SHORT" siglexreturn(KW_short);
<sentences>"integer" siglexreturn(KW_integer);
<sentences>"INTEGER" siglexreturn(KW_integer);
<sentences>"long" siglexreturn(KW_long);
<sentences>"LONG" siglexreturn(KW_long);
<sentences>"real" siglexreturn(KW_real);
<sentences>"REAL" siglexreturn(KW_real);
<sentences>"dreal" siglexreturn(KW_dreal);
<sentences>"DREAL" siglexreturn(KW_dreal);
<sentences>"complex" siglexreturn(KW_complex);
<sentences>"COMPLEX" siglexreturn(KW_complex);
<sentences>"dcomplex" siglexreturn(KW_dcomplex);
<sentences>"DCOMPLEX" siglexreturn(KW_dcomplex);
<sentences>"char" siglexreturn(KW_char);
<sentences>"CHAR" siglexreturn(KW_char);
<sentences>"string" siglexreturn(KW_string);
<sentences>"STRING" siglexreturn(KW_string);
<sentences>"enum" siglexreturn(KW_enum);
<sentences>"ENUM" siglexreturn(KW_enum);
<sentences>"[" siglexreturn(KW_COUV);
<sentences>"]" siglexreturn(KW_CFER);
<sentences>"struct" siglexreturn(KW_struct);
<sentences>"STRUCT" siglexreturn(KW_struct);
<sentences>"bundle" siglexreturn(KW_bundle);
<sentences>"BUNDLE" siglexreturn(KW_bundle);
<sentences>"constant" siglexreturn(KW_constant);
<sentences>"CONSTANT" siglexreturn(KW_constant);
<sentences>"shared" siglexreturn(KW_shared);
<sentences>"SHARED" siglexreturn(KW_shared);
<sentences>"statevar" siglexreturn(KW_statevar);
<sentences>"STATEVAR" siglexreturn(KW_statevar);
<sentences>"label" siglexreturn(KW_label);
<sentences>"LABEL" siglexreturn(KW_label);
<sentences>"ref" siglexreturn(KW_ref);
<sentences>"REF" siglexreturn(KW_ref);
<sentences>"assert" siglexreturn(KW_assert);
<sentences>"ASSERT" siglexreturn(KW_assert);
<sentences>"|" siglexreturn(KW_VERT);
<sentences>"::" siglexreturn(KW_DPINDPIN);
<sentences>"on" siglexreturn(KW_on);
<sentences>"ON" siglexreturn(KW_on);
<sentences>"where" siglexreturn(KW_where);
<sentences>"WHERE" siglexreturn(KW_where);
<sentences>"/" siglexreturn(KW_SLAH);
<sentences>"array" siglexreturn(KW_array);
<sentences>"ARRAY" siglexreturn(KW_array);
<sentences>"of" siglexreturn(KW_of);
<sentences>"OF" siglexreturn(KW_of);
<sentences>"iterate" siglexreturn(KW_iterate);
<sentences>"ITERATE" siglexreturn(KW_iterate);
<sentences>"to" siglexreturn(KW_to);
<sentences>"TO" siglexreturn(KW_to);
<sentences>"with" siglexreturn(KW_with);
<sentences>"WITH" siglexreturn(KW_with);
<sentences>"case" siglexreturn(KW_case);
<sentences>"CASE" siglexreturn(KW_case);
<sentences>"in" siglexreturn(KW_in);
<sentences>"IN" siglexreturn(KW_in);
<sentences>":" siglexreturn(KW_DPIN);
<sentences>"else" siglexreturn(KW_else);
<sentences>"ELSE" siglexreturn(KW_else);
<sentences>"[." siglexreturn(KW_COUVPOIN);
<sentences>".]" siglexreturn(KW_POINCFER);
<sentences>":=" siglexreturn(KW_DPINEQUL);
<sentences>"::=" siglexreturn(KW_DPINDPINEQUL);
<sentences>"defaultvalue" siglexreturn(KW_defaultvalue);
<sentences>"DEFAULTVALUE" siglexreturn(KW_defaultvalue);
<sentences>":=:" siglexreturn(KW_DPINEQULDPIN);
<sentences>"^=" siglexreturn(KW_CHAPEQUL);
<sentences>"^<" siglexreturn(KW_CHAPINFE);
<sentences>"^>" siglexreturn(KW_CHAPSUPE);
<sentences>"^#" siglexreturn(KW_CHAPDIES);
<sentences>"-->" siglexreturn(KW_MINUMINUSUPE);
<sentences>"when" siglexreturn(KW_when);
<sentences>"WHEN" siglexreturn(KW_when);
<sentences>"^0" siglexreturn(KW_CHAP0);
<sentences>"next" siglexreturn(KW_next);
<sentences>"NEXT" siglexreturn(KW_next);
<sentences>"default" siglexreturn(KW_default);
<sentences>"DEFAULT" siglexreturn(KW_default);
<sentences>"after" siglexreturn(KW_after);
<sentences>"AFTER" siglexreturn(KW_after);
<sentences>"from" siglexreturn(KW_from);
<sentences>"FROM" siglexreturn(KW_from);
<sentences>"count" siglexreturn(KW_count);
<sentences>"COUNT" siglexreturn(KW_count);
<sentences>"^+" siglexreturn(KW_CHAPPLUS);
<sentences>"^-" siglexreturn(KW_CHAPMINU);
<sentences>"^*" siglexreturn(KW_CHAPETOI);
<sentences>"[:" siglexreturn(KW_COUVDPIN);
<sentences>"[/:" siglexreturn(KW_COUVSLAHDPIN);
<sentences>"if" siglexreturn(KW_if);
<sentences>"IF" siglexreturn(KW_if);
<sentences>"then" siglexreturn(KW_then);
<sentences>"THEN" siglexreturn(KW_then);
<sentences>".." siglexreturn(KW_POINPOIN);
<sentences>"step" siglexreturn(KW_step);
<sentences>"STEP" siglexreturn(KW_step);
<sentences>"xor" siglexreturn(KW_xor);
<sentences>"XOR" siglexreturn(KW_xor);
<sentences>"or" siglexreturn(KW_or);
<sentences>"OR" siglexreturn(KW_or);
<sentences>"and" siglexreturn(KW_and);
<sentences>"AND" siglexreturn(KW_and);
<sentences>"not" siglexreturn(KW_not);
<sentences>"NOT" siglexreturn(KW_not);
<sentences>"==" siglexreturn(KW_EQULEQUL);
<sentences>"<<=" siglexreturn(KW_INFEINFEEQUL);
<sentences>"/=" siglexreturn(KW_SLAHEQUL);
<sentences>">" siglexreturn(KW_SUPE);
<sentences>">=" siglexreturn(KW_SUPEEQUL);
<sentences>"<" siglexreturn(KW_INFE);
<sentences>"<=" siglexreturn(KW_INFEEQUL);
<sentences>"+" siglexreturn(KW_PLUS);
<sentences>"-" siglexreturn(KW_MINU);
<sentences>"|+" siglexreturn(KW_VERTPLUS);
<sentences>"*" siglexreturn(KW_ETOI);
<sentences>"*." siglexreturn(KW_ETOIPOIN);
<sentences>"modulo" siglexreturn(KW_modulo);
<sentences>"MODULO" siglexreturn(KW_modulo);
<sentences>"|*" siglexreturn(KW_VERTETOI);
<sentences>"**" siglexreturn(KW_ETOIETOI);
<sentences>"@" siglexreturn(KW_ARRO);
<sentences>"cell" siglexreturn(KW_cell);
<sentences>"CELL" siglexreturn(KW_cell);
<sentences>"var" siglexreturn(KW_var);
<sentences>"VAR" siglexreturn(KW_var);
<sentences>"$" siglexreturn(KW_DOLL);
<sentences>"window" siglexreturn(KW_window);
<sentences>"WINDOW" siglexreturn(KW_window);
<sentences>"tr" siglexreturn(KW_tr);
<sentences>"TR" siglexreturn(KW_tr);
<sentences>"\\\\" siglexreturn(KW_BSLABSLA);
<sentences>"^" siglexreturn(KW_CHAP);
<sentences>"." siglexreturn(KW_POIN);
<sentences>"#" siglexreturn(KW_DIES);
<sentences>"<<" siglexreturn(KW_INFEINFE);
<sentences>">>" siglexreturn(KW_SUPESUPE);
<sentences>"false" siglexreturn(KW_false);
<sentences>"FALSE" siglexreturn(KW_false);
<sentences>"true" siglexreturn(KW_true);
<sentences>"TRUE" siglexreturn(KW_true);
<entrypoint>"[ATTR]"  {BEGIN sentences; siglexreturn(EN_ATTR);};
<entrypoint>"[BORN]"  {BEGIN sentences; siglexreturn(EN_BORN);};
<entrypoint>"[CAS]"  {BEGIN sentences; siglexreturn(EN_CAS);};
<entrypoint>"[CHAI]"  {BEGIN sentences; siglexreturn(EN_CHAI);};
<entrypoint>"[CST_CHA]"  {BEGIN sentences; siglexreturn(EN_CST_CHA);};
<entrypoint>"[DEC_S]"  {BEGIN sentences; siglexreturn(EN_DEC_S);};
<entrypoint>"[DESC]"  {BEGIN sentences; siglexreturn(EN_DESC);};
<entrypoint>"[DESC_MODU]"  {BEGIN sentences; siglexreturn(EN_DESC_MODU);};
<entrypoint>"[ELE_VEC]"  {BEGIN sentences; siglexreturn(EN_ELE_VEC);};
<entrypoint>"[ENT_SOR]"  {BEGIN sentences; siglexreturn(EN_ENT_SOR);};
<entrypoint>"[I_TYPE]"  {BEGIN sentences; siglexreturn(EN_I_TYPE);};
<entrypoint>"[IDENT]"  {BEGIN sentences; siglexreturn(EN_IDENT);};
<entrypoint>"[IND_MOT]"  {BEGIN sentences; siglexreturn(EN_IND_MOT);};
<entrypoint>"[L_CAS]"  {BEGIN sentences; siglexreturn(EN_L_CAS);};
<entrypoint>"[L_CONS]"  {BEGIN sentences; siglexreturn(EN_L_CONS);};
<entrypoint>"[L_E_SIGN]"  {BEGIN sentences; siglexreturn(EN_L_E_SIGN);};
<entrypoint>"[L_IDENT]"  {BEGIN sentences; siglexreturn(EN_L_IDENT);};
<entrypoint>"[L_PARC]"  {BEGIN sentences; siglexreturn(EN_L_PARC);};
<entrypoint>"[L_SIGN]"  {BEGIN sentences; siglexreturn(EN_L_SIGN);};
<entrypoint>"[MODU]"  {BEGIN sentences; siglexreturn(EN_MODU);};
<entrypoint>"[O_SPEC]"  {BEGIN sentences; siglexreturn(EN_O_SPEC);};
<entrypoint>"[OL_ATTR]"  {BEGIN sentences; siglexreturn(EN_OL_ATTR);};
<entrypoint>"[OL_DECL]"  {BEGIN sentences; siglexreturn(EN_OL_DECL);};
<entrypoint>"[OL_DEC_S]"  {BEGIN sentences; siglexreturn(EN_OL_DEC_S);};
<entrypoint>"[OL_E_PARA]"  {BEGIN sentences; siglexreturn(EN_OL_E_PARA);};
<entrypoint>"[OL_OPRA]"  {BEGIN sentences; siglexreturn(EN_OL_OPRA);};
<entrypoint>"[OL_PRAG]"  {BEGIN sentences; siglexreturn(EN_OL_PRAG);};
<entrypoint>"[PARC]"  {BEGIN sentences; siglexreturn(EN_PARC);};
<entrypoint>"[PRAG]"  {BEGIN sentences; siglexreturn(EN_PRAG);};
<entrypoint>"[DECL]"  {BEGIN sentences; siglexreturn(EN_DECL);};
<entrypoint>"[DEF_CONS]"  {BEGIN sentences; siglexreturn(EN_DEF_CONS);};
<entrypoint>"[DEF_TYPE]"  {BEGIN sentences; siglexreturn(EN_DEF_TYPE);};
<entrypoint>"[E_CST]"  {BEGIN sentences; siglexreturn(EN_E_CST);};
<entrypoint>"[ENU_VAL]"  {BEGIN sentences; siglexreturn(EN_ENU_VAL);};
<entrypoint>"[IDENT_OP]"  {BEGIN sentences; siglexreturn(EN_IDENT_OP);};
<entrypoint>"[INTE]"  {BEGIN sentences; siglexreturn(EN_INTE);};
<entrypoint>"[O_CHAI]"  {BEGIN sentences; siglexreturn(EN_O_CHAI);};
<entrypoint>"[O_IDENT]"  {BEGIN sentences; siglexreturn(EN_O_IDENT);};
<entrypoint>"[OL_E_SIGN]"  {BEGIN sentences; siglexreturn(EN_OL_E_SIGN);};
<entrypoint>"[REF_PRO]"  {BEGIN sentences; siglexreturn(EN_REF_PRO);};
<entrypoint>"[SIGN]"  {BEGIN sentences; siglexreturn(EN_SIGN);};
<entrypoint>"[E_PROC]"  {BEGIN sentences; siglexreturn(EN_E_PROC);};
<entrypoint>"[TYPE]"  {BEGIN sentences; siglexreturn(EN_TYPE);};
<entrypoint>"[O_E_PROC]"  {BEGIN sentences; siglexreturn(EN_O_E_PROC);};
<entrypoint>"[TYPE_S]"  {BEGIN sentences; siglexreturn(EN_TYPE_S);};
<entrypoint>"[V_PROC]"  {BEGIN sentences; siglexreturn(EN_V_PROC);};
<entrypoint>"[V_TYPE]"  {BEGIN sentences; siglexreturn(EN_V_TYPE);};
<entrypoint>"[O_CORP]"  {BEGIN sentences; siglexreturn(EN_O_CORP);};
<entrypoint>"[E_SIGN]"  {BEGIN sentences; siglexreturn(EN_E_SIGN);};
<entrypoint>"[LSIG]"  {BEGIN sentences; siglexreturn(EN_LSIG);};
<entrypoint>"[D_E_SIGN]"  {BEGIN sentences; siglexreturn(EN_D_E_SIGN);};
<entrypoint>"[E_PARA]"  {BEGIN sentences; siglexreturn(EN_E_PARA);};
<entrypoint>"[O_E_SIGN]"  {BEGIN sentences; siglexreturn(EN_O_E_SIGN);};
<entrypoint>"[V_CONS]"  {BEGIN sentences; siglexreturn(EN_V_CONS);};
<entrypoint>. {BEGIN sentences; Unput(yytext[0]);};
<entrypoint>\n {BEGIN sentences; Unput(yytext[0]);};
<sentences>([_a-zA-Z]|{car_accent})([_a-zA-Z0-9]|{car_accent})*                   siglexreturn(IDENT);
<sentences>[0-9]+                                  siglexreturn(ENTIER);
<sentences>[0-9]+\.[0-9]+([eE][-+]?[0-9]+)?        siglexreturn(REEL);
<sentences>[0-9]+[eE][-+]?[0-9]+                   siglexreturn(REEL);
<sentences>[0-9]+\.[0-9]+[dD][-+]?[0-9]+           siglexreturn(DREEL);
<sentences>[0-9]+[dD][-+]?[0-9]+                   siglexreturn(DREEL);
<sentences>'([^'\\\n\t\14\15]|{code_car})'         { /* shiftleft();
                                                     suplast(); */
                                                     siglexreturn(CAR); }
<sentences>\"([^"\\\n\t\14\15]*|{code_car})*\"     { /* shiftleft();
                                                     suplast(); */
                                                     siglexreturn(CHAINE);}
<sentences>\%                                      { siglexposition(llocp);
                                                     BEGIN comment;
                                                     break; }
<sentences>.	                                   { siglexposition(llocp);
/*                                                     pK_messenger -> erreurstr(errSyntax, strmakestrwithnum(LINENUMBER(llocp->first_line)));	*/
						     pK_messenger -> parsing_errorLineColumn(__siganafilename__, LINENUMBER(llocp->first_line), LINENUMBER(llocp->first_column), errunrecchar);
                                                     break; }
%%
// input in C++ is a predefined stream.
// So, this function do the comparison with the compiler (c or C++).
// Read the next character in the input string.
static int Input()
{
  char c;
#ifdef __cplusplus
  c =  yyinput();
#else
  c =  input();
#endif
  switch(c) 
  {
    case '\n' :
      sigline++;
      sigcolumn = 0;
      break;
    default :
      sigcolumn++;
  }
  return c;
}


// Sets a character in the input stream. It will be the next character scanned.
static void Unput(int c)
{
  unput(c);
  switch(c) {
    case '\n' :
      sigline--;
      sigcolumn = 0;
      break;
    default :
      sigcolumn--;
  }
}

static void init_lex(integer phylumCode)
{
  /* PHYLA  Management */
  ADDEDPHYLUM = (phylumCode != without_phylum);
  VPHYLUM = phylumCode;
  PHYLUMHASBEENADDED = 0;
  /* end-PHYLA  Management */
  root_signal=LL_make_nil_atom();

  sigline = 1;
  sigcolumn = 1;
  YY_USER_INIT;
}
