/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __StrScanner_h__
#define __StrScanner_h__

typedef enum {tk_end,       /**<  end of analyze */
	      tk_error,     /**<  error */
	      tk_ident,     /**<  identifier */
	      tk_progName, /**<   \b &m name of the program. */
	      tk_name,      /**<  \b &n name of a node (external call)*/
	      tk_param,     /**<  \b &p<integer> parameter (&p5 stands for the 5th parameter). */
	      tk_input,     /**<  \b &i<integer> input (&i2 stands for the 2th input).*/
	      tk_output,    /**<  \b &o<integer> output (&o3 stands for the 3th output).*/
	      tk_type,      /**<  \b &t type */
	      tk_baseType,  /**<  \b &b base type */
	      tk_clock,     /**<  \b &ck clock */
	      tk_dim,       /**<  \b &d  dimension */
	      tk_2pe,       /**<  \b := */
	      tk_virg,      /**<  \b ,  */
	      tk_po,        /**<  \b (  */
	      tk_pf,        /**<  \b )  */
	      tk_ao,        /**<  \b {  */
	      tk_af,        /**<  \b }  */
	      tk_homo,      /**<  \b &h  homomorphism */
	      tk_homoClock, /**<  \b &hck  clock homomorphism */
	      tk_cardwait,  /**<  \b &|wait| number of predecessors */
	      tk_wait,      /**<  \b &wait */
	      tk_done,      /**<  \b &done done output */
              tk_list       /**<  \b &lexp followedy &i,&o or &p stands for list of inputs/outputs/parameters. */,
              tk_adr,       /**<  \b && stand for adress of */
              tk_cpt,       /**<  \b &cpt internal counter */
              tk_symbol      /**<  \b &symb reference to a symbol */
} ScanPragmaToken;


typedef class StrScanner *ptStrScanner;
#define NILSTRSCANNER   (ptStrScanner)pK_NIL

const integer StrScanner_NBKW = 18;

class StrScanner
{
#include "StrScanner_m.h"
#include "StrScanner_g.h"
protected:
  ptrepstr this_currentIdent;  ///< current identifier.
  ScanPragmaToken   currentToken;  ///< current  token read.

  ptrepstr curStr;  ///< Current line of the Current parametrized Text.
  integer indexCurStr;///< index on curStr.
  integer lgCurStr;   ///< size of curStr.

  ntrees CurrentParametrizedText;   ///< Current parametrized Text.
  integer iCurrentParametrizedText; ///< Index on the Current parametrized Text.
  integer nbLines;     ///< nbLines of the Current parametrized Text.
  boolean modeSkipSpace; ///< spacing mode. When it is "true", the space character is skipped.
  struct { ptrepstr kw; ScanPragmaToken kwcode; } kwTable[StrScanner_NBKW]; ///< associative table (name, token).
};
		 
#endif
