/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/**
   @return the string of the i-th predefined pragma.
   The set of indexes is available in the file $Signal_incl/SignalEnvPragmas.h
*/
extern ptrepstr
getPredefinedPragma (int i)
{
  if ((i < 0) || (i > LAST_PRAGMA_NUMBER))
    return nilstr;
  //  if (!SignalPredefinedPragmasTableLoaded) SignalPredefinedPragmasTableIni();
  return SignalPredefinedPragmasTable[i];
}


extern boolean
isPredefinedPragmaName (ptrepstr vstr, int *i)
{
  for ((*i) = 0; (*i) <= LAST_PRAGMA_NUMBER; (*i)++)
    if (streq (getPredefinedPragma (*i), vstr))
      return pk_true;

  return pk_false;
}


/**
   @return the first string of a list (enonce).
*/
extern char *
VMvaluePragma2String (ntrees enonce)
{
  if (IS_NILTREE (enonce))
    return (char *) NULL;
  return repstr2string (nstringof (nchild (1, enonce)));
}

/** @return  the i-th term of the text (enonce).
    A pragma text is a set of strings.
 */
extern ptrepstr
getPragmaTerm (ntrees enonce, int i)
{
  ptrepstr EnonceValue;
  EnonceValue = nstringof (nchild (i, enonce));
  return ExtractString (EnonceValue);
}


/** It adds a string (v) to a value of pragma (enonce).
 */
extern ntrees
addEnonceValue (ntrees enonce, ptrepstr v)
{
  ntrees y;

  y = LL_make_cha_atom (v);
  if (IS_NILTREE (enonce))
    return LL_make_chai_list (y);
  return ntreeList_appendLast (enonce, y);
}

/**
   @return a copy of value of  pragma (e).
*/
extern ntrees
copyEnonce (ntrees e)
{
  return STree_scopy (e);
}


/**
   @return a value of pragma composed of a string (repId).
*/
extern ntrees
repstr2PragmaStatement (ptrepstr repId)
{
  ptrepstr nrepId;

  if (repId == nilstr)
    return STree_mkEmptyString ();
  if (!getDoubleQuotedString (repId, &nrepId))
    repId = nrepId;
  return LL_make_chai_list (LL_make_cha_atom (repId));
}


/**
   @return the i-th argument, without double quotes, of a pragma statement (stmt).
   assumed : stmt is a pragma statement (list of strings).
*/
extern ptrepstr
getIthPramaStatementWithoutQuotes (ntrees stmt, int i)
{
  return ExtractString (nstringof (nchild (i, stmt)));
}

extern void
SignalPredefinedPragmasTableIni ()
{
  SignalPredefinedPragmasTableIniLoc ();
}


/**
   @return the name of the pragma, taking into account some aliases.
*/
extern ptrepstr
getTreePragmaName (ntrees aPragma)
{
  ptrepstr v;
  v = STree_getIdentStr (nchild (1, aPragma));
  if (v == getPredefinedPragma (PRAGMA_BINDING))
    return getPredefinedPragma (PRAGMA_RUN_ON);
  return v;
}
