/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


#include <string.h>

/** Pragma representation. */
typedef struct pragma
{
  ptrepstr name;
  pointer objects;
  ntrees stmts;
}
tpragma;

static ptrepstr SignalPredefinedPragmasTable[LAST_PRAGMA_NUMBER + 1];

static void
SignalPredefinedPragmasTableIniLoc ()
{
  SignalPredefinedPragmasTable[PRAGMA_COMMENT] = string2repstr ("Comment");
  SignalPredefinedPragmasTable[PRAGMA_NOTE] = string2repstr ("Note");

  SignalPredefinedPragmasTable[PRAGMA_MAIN] = string2repstr ("Main");
  SignalPredefinedPragmasTable[PRAGMA_UNEXPANDED] = string2repstr ("Unexpanded");
  SignalPredefinedPragmasTable[PRAGMA_SEPARATE] = string2repstr ("Separate");

  // Used for separate compiling.
  SignalPredefinedPragmasTable[PRAGMA_CLUSTER] = string2repstr ("Cluster");
  SignalPredefinedPragmasTable[PRAGMA_BLACK_BOX] = string2repstr ("BlackBox");
  SignalPredefinedPragmasTable[PRAGMA_GREY_BOX] = string2repstr ("GreyBox");
  SignalPredefinedPragmasTable[PRAGMA_DELAYCLUSTER] = string2repstr ("State");


  // It allows to associate code for a target language.
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_C] = string2repstr ("C_Code");
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_CPP] = string2repstr ("CPP_Code");
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_JAVA] = string2repstr ("Java_Code");


  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_C_IMPORT] = string2repstr ("C_Code_Import");
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_CPP_IMPORT] = string2repstr ("CPP_Code_Import");
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_JAVA_IMPORT] = string2repstr ("Java_Code_Import");

  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_C_DECLARE] = string2repstr ("C_Code_Declare");
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_CPP_DECLARE] = string2repstr ("CPP_Code_Declare");
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_JAVA_DECLARE] = string2repstr ("Java_Code_Declare");

  // Used for morphism (profiling)
  SignalPredefinedPragmasTable[PRAGMA_PROCESSOR_TYPE] = string2repstr ("ProcessorType");
  // Link with Sigali tools
  SignalPredefinedPragmasTable[PRAGMA_SIGALI] = string2repstr ("Sigali");

  SignalPredefinedPragmasTable[PRAGMA_DEFINED_CLOCK_HIERARCHY] = string2repstr ("DefinedClockHierarchy");

  // Partitionning information RunOn <string> equivalent to Binding
  SignalPredefinedPragmasTable[PRAGMA_RUN_ON] = string2repstr ("RunOn");
  SignalPredefinedPragmasTable[PRAGMA_BINDING] = string2repstr ("Binding");
  SignalPredefinedPragmasTable[PRAGMA_BINDINGIO] = string2repstr ("BindingIO");
  // Description of the architecture 
  SignalPredefinedPragmasTable[PRAGMA_PROCESSOR_NAME] = string2repstr ("Processor");
  SignalPredefinedPragmasTable[PRAGMA_BUS] = string2repstr ("Bus");
  SignalPredefinedPragmasTable[PRAGMA_CONNECTION] = string2repstr ("Connection");

  // Distribution.
  SignalPredefinedPragmasTable[PRAGMA_MASTER] = string2repstr ("Master");
  SignalPredefinedPragmasTable[PRAGMA_SLAVES] = string2repstr ("Slaves");
  SignalPredefinedPragmasTable[PRAGMA_COMM_ENV] = string2repstr ("Environment");
  SignalPredefinedPragmasTable[PRAGMA_COMM_RECEIVE] = string2repstr ("Receiving");
  SignalPredefinedPragmasTable[PRAGMA_COMM_SEND] = string2repstr ("Sending");
  SignalPredefinedPragmasTable[PRAGMA_TOPOLOGY] = string2repstr ("Topology");
  SignalPredefinedPragmasTable[PRAGMA_TARGET] = string2repstr ("Target");
  SignalPredefinedPragmasTable[PRAGMA_SELECTED_OTHERS_PARTITION] = string2repstr ("Partition");
  SignalPredefinedPragmasTable[PRAGMA_MORPHISM] = string2repstr ("Morphism");

  // Misc : for conversion, integer(x) is represented as a call in the graph.
  SignalPredefinedPragmasTable[PRAGMA_CONVERSION] = string2repstr ("Conversion");
  SignalPredefinedPragmasTable[PRAGMA_RESTRICTED_MODES] = string2repstr ("Restricted modes");

  // unused now ?
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_THREADSJAVA] = string2repstr ("ThJava_CODE");	// ??
  SignalPredefinedPragmasTable[PRAGMA_CODEGEN_VHDL] = string2repstr ("VHDL_CODE");	// ??

  SignalPredefinedPragmasTable[PRAGMA_SIGNAL_THREAD] = string2repstr ("Signal_Thread");

}
