/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright(C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox(http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/// PREDEFINED OBJECT OF THE VIRTUAL MACHINE.
ptSignalPrinter pK_SignalPrinter = new SignalPrinter ();

#define X1_indent 5
#define smallindent 3
#define actindent 2

#define comcom "%"
#define comincom "*"
#define comincomend "%------<"
#define errSymbol "~"
#define cstSymbolHolo "..."

_PROTECTED_
/**
   @return true if a tree(t) is the null object or an atom with opnil as operator.
*/
  boolean SignalPrinter::isSignalNilTree (ntrees t)
{
  if (IS_NILTREE (t))
    return pk_true;
  return (noper (t) == opnil);
}

/**
   @return true if a tree(t) is the null object or an atom with opnil as operator.
*/
boolean SignalPrinter::isChildNot (ntrees t, ntrees * vt)
{
  ntrees
    t1;
  *vt = t;
  if (IS_NILTREE (t))
    return pk_false;
  t1 = nchild (1, t);
  if (noper (t1) != opnon)
    return pk_false;
  *vt = t1;
  return pk_true;
}

/** 
    Printing the first child of a declaration of signals/constants(operators: opdecsig, opdeccon, opdecconp).
*/
_VIRTUAL_ void
SignalPrinter::printtype ()
{
  ntrees vtype;
  boolean bspace;

  vtype = nchild (1, operande);
  if (ISNOT_NILTREE (vtype))
    bspace = (noper (vtype) == opident);
  else
    bspace = pk_false;
  recsondecunparse (1);
  if (bspace)
    snspace ();
}

_VIRTUAL_ void
SignalPrinter::printSecondChildDeclaration ()
{

  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
  else
    {
      this_intePrinter->printerror ();
      //      this_intePrinter->printkey(kwpvir);
    }
}

/**
   @return "true" is op is a comment operator.
*/
_VIRTUAL_ boolean
SignalPrinter::iscomment (noperators op)
{
  return (boolean) (op == opcomment) || (op == opcomments);
}


///
_VIRTUAL_ boolean
SignalPrinter::isparproc (noperators op)
{
  return (boolean) (op == oppconf) || (op == oplfeed) || (op == oplseq) || (op == opmoti) || (op == opiter) || (op == oppcas);
}


/**
   @return "true" if op is the operator of a "primary" process expression,
   that may need to be parenthesized in some syntactic contexts,
   "false" otherwise.
*/
_VIRTUAL_ boolean
SignalPrinter::isprimproc (noperators op)
{
  return (boolean) (op == opappe) || (op == opdefpart) || (op == opdefpartdef) || (op == opdefsig) ||
    (op == opinst) || (op == oppcontrol) || (op == oppetiq) || (op == oplcrohor) ||
    (op == opldechor) || (op == oplequhor) || (op == oplexchor) || (op == opdepd) || (op == opnullproc);
}


///
_VIRTUAL_ boolean
SignalPrinter::isfactsig (noperators op)
{
  return (boolean) (op == opnom) || (op == opident) || (op == opidentfield) || (op == opcstcar) || (op == opcstdp) ||
    (op == opcstent) || (op == opcstreel) || (op == opfalse) ||
    (op == optrue) || (op == opsdec) || (op == opcham) || (op == opvenum) || (op == opelem) || (op == opchai) || (op == opleflot);
}


///
_VIRTUAL_ boolean
SignalPrinter::isprimsig (noperators op)
{
  return (boolean) (isfactsig (op)) || (op == optabvide) || (op == oppcar) ||
    (op == opappe) || (op == opconv) || (op == opinst) || (op == oplevect) || (op == oplelevec) || (shorten_clocks && (op == opwhen1));
}


///
_VIRTUAL_ boolean
SignalPrinter::isEventOperator (noperators op)
{
  return (boolean) (op == opevnt);
}


///
_VIRTUAL_ boolean
SignalPrinter::isMaskingOperator (noperators op)
{
  return (boolean) (op == opobli);
}


/**
   pos = pk_true postlude, pk_false for prelude.
*/
_VIRTUAL_ boolean
SignalPrinter::unparse_annotation (ntrees varb, boolean pos, int holoph)
{
  boolean b, result;
  ptrepstr b0, b1, bEnd;

  result = pk_false;
  if (!isSignalNilTree (varb))
    {
      if (this_intePrinter->open_unparsePrintableTreeAttribute (&b0, &bEnd, &b1))
	{
	  //      sigpristr(b0);
	  unparse_tree (varb, holoph);
	  //      sigpristr(bEnd);
	}
      else
	{
	  if (iscomment (noper (varb)))
	    {
	      if (pos)
		snspace ();
	      this_intePrinter->reset_unparsePrintableTreeAttribute (b0, pk_false);
	      b = stdMode;
	      stdMode = pk_false;
	      unparse_tree (varb, holoph);
	      stdMode = b;
	      this_intePrinter->reset_unparsePrintableTreeAttribute (bEnd, !pos);
	    }
	  else
	    {
	      result = pk_true;
	      this_intePrinter->reset_unparsePrintableTreeAttribute (b1, pk_true);
	      unparse_tree (varb, holoph);
	      this_intePrinter->reset_unparsePrintableTreeAttribute (b1, pk_true);
	    };
	}
      this_intePrinter->close_unparsePrintableTreeAttribute ();
    }
  return result;
}

/**
   Printing the prefix annotations(if pos) associated with a tree(p).
   The annotations are trees, they are holoph-printed(holoph).
   pos = false for pre-comment, pk_true for post-comment.
*/
_VIRTUAL_ boolean
SignalPrinter::unparse_printableTreeAttribute (ntrees p, boolean pos, int holoph)
{
  boolean result;
  ntrees varb;

  result = pk_false;
  if (STree_getPrintableTreeAttribute (p, !pos, &varb))
    {
      result = unparse_annotation (varb, pos, holoph);
      return pk_true;
    }
  if (!pos)
    return pk_false;
  varb = STree_getVMAnnotation (p);
  result = result || unparse_annotation (varb, pos, holoph);
  return result;
}

/**
   Printing of the i-th child of a tree(p). This child must be a leaf(nullary tree).
*/
_VIRTUAL_ void
SignalPrinter::sonconswrite (ntrees p, int i)
{
  ntrees operand;

  operand = nchild (i, p);
  if (IS_NILTREE (operand))
    this_intePrinter->printerror ();
  else if (noper (operand) == opnil)
    this_intePrinter->printerror ();
  else
    inteconsprint (operand);
}



/** Printing a list of suffixes(p). */
_VIRTUAL_ void
SignalPrinter::lsuffIdent (ntrees p, int tkw)
{
  int nbfils;
  int i;

  nbfils = ntreearity (p);
  for (i = 1; i <= nbfils; i++)
    {
      this_intePrinter->printkey (tkw);
      sonconswrite (p, i);
    }
}



/** Printing an identifier(p) represented as a binary tree: first child is a leaf, second a list of suffixes.
 */
_VIRTUAL_ void
SignalPrinter::identprint (ntrees p)
{
  ntrees operand;

  snfit (this_intePrinter->sigidentsize (p) + 1);
  sonconswrite (p, 1);
  operand = nchild (2, p);
  if (IS_NILTREE (operand))
    this_intePrinter->printerror ();
  else if (noper (operand) == opnil)
    this_intePrinter->printerror ();
  else if (noper (p) == opident)
    lsuffIdent (operand, kw_);
  else
    lsuffIdent (operand, kwpoin);	// ident of a signal associoated with a field of a structure/bundle.
  snglue (pk_true);
}



/** 
    Printing an identifier(p) represented as a binary tree(\ref identprint function)
    or a nullary tree.
*/
_VIRTUAL_ void
SignalPrinter::identornameprint (ntrees p)
{
  noperators opera;

  opera = noper (p);
  if ((opera == opnom) || (opera == opop))
    inteconsalphaprint (p);
  else if ((opera == opident) || (opera == opidentfield))
    identprint (p);
}



/**
   @return "true" if the nofils-th child of the current tree( \b treeunparseTree, class attribute) exists, "false" otherwise.
   When it exists the class attribute member 
   operande and voperator are set:
   - \b operande: nofils-th child of treeunparseTree,
   - \b operator: the operator of operande.   
*/
_VIRTUAL_ boolean
SignalPrinter::ischildprint (int nofils)
{
  operande = nchild (nofils, treeunparseTree);
  if (IS_NILTREE (operande))
    return pk_false;
  voperator = noper (operande);
  return (boolean) (voperator != opnil);
}


/**
   It prints the \b operande tree, with brackets if embed.
   Moreover, a specific string is printed when the xxholophrastexx is reached.
*/
_VIRTUAL_ void
SignalPrinter::embedarg (boolean embed)
{
  if (vholoph < actindent)
    this_intePrinter->printholo ();
  else
    {
      if (vholoph > actindent)
	vholoph = vholoph - 1;
      if (embed)
	{
	  this_intePrinter->printkey (kwpouv);
	  unparse_subtree (operande, vholoph);
	  this_intePrinter->printkey (kwpfer);
	}
      else
	unparse_subtree (operande, vholoph);
    }
}


///
_VIRTUAL_ void
SignalPrinter::sigargunp ()
{
  embedarg (!isprimsig (voperator));
}


///
_VIRTUAL_ void
SignalPrinter::sigfactargunp ()
{
  embedarg (!isfactsig (voperator));
}


///
_VIRTUAL_ void
SignalPrinter::unalphasigexp (int tkw, int fils)
{
  this_intePrinter->printalphakeysp (tkw);
  if (ischildprint (fils))
    sigargunp ();
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::unsigexp (int tkw, int fils)
{
  this_intePrinter->printkey (tkw);
  if (ischildprint (fils))
    sigargunp ();
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::unsigfact (int fils)
{
  if (ischildprint (fils))
    sigfactargunp ();
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::sigarglesign (int fils)
{
  boolean b;

  if (ischildprint (fils))
    {
      if (vholoph < actindent)
	this_intePrinter->printholo ();
      else
	{
	  b = vholoph > actindent;
	  if (b)
	    vholoph = vholoph - 1;
	  if (voperator == oplesign)
	    {
	      this_intePrinter->printkey (kwbouv);
	      unparse_subtree (operande, vholoph);
	      this_intePrinter->printkey (kwbfer);
	    }
	  else
	    unparse_subtree (operande, vholoph);
	  if (b)
	    vholoph = vholoph + 1;
	}
    }
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::markunparse (int holoph)
{
  snmark (holoph - actindent);
  unparse_subtree (operande, holoph);
  snrelease (holoph - actindent);
}


///
_VIRTUAL_ void
SignalPrinter::markunparsegluekey (int rkey, int holoph)
{
  snmark (holoph - actindent);
  unparse_subtree (operande, holoph);
  this_intePrinter->printreleasegluekey (rkey, holoph - actindent);
}


///
_VIRTUAL_ void
SignalPrinter::markkeyunparsekey (int lkey, int rkey, int holoph)
{
  this_intePrinter->printkeymark (lkey, holoph - actindent);
  unparse_subtree (operande, holoph);
  this_intePrinter->printkeyrelease (rkey, holoph - actindent);
}


///
_VIRTUAL_ void
SignalPrinter::markgluekeyunparsekey (int lkey, int rkey, int holoph)
{
  this_intePrinter->printgluekeymark (lkey, holoph - actindent);
  unparse_subtree (operande, holoph);
  this_intePrinter->printgluekeyrelease (rkey, holoph - actindent);
}


///
_VIRTUAL_ void
SignalPrinter::markkeyunparse (int lkey, int holoph)
{
  this_intePrinter->printkeymark (lkey, holoph - actindent);
  unparse_subtree (operande, holoph);
  snrelease (holoph - actindent);
}


///
_VIRTUAL_ void
SignalPrinter::markspkeyunparse (int lkey, int holoph)
{
  this_intePrinter->printspkeymark (lkey, holoph - actindent);
  unparse_subtree (operande, holoph);
  snrelease (holoph - actindent);
}


///
_VIRTUAL_ void
SignalPrinter::markspdkeyunparse (int lkey1, int lkey2, int holoph)
{
  this_intePrinter->printspkeymark (lkey1, holoph - actindent);
  this_intePrinter->printalphakeysp (lkey2);
  unparse_subtree (operande, holoph);
  snrelease (holoph - actindent);
}


///
_VIRTUAL_ boolean
SignalPrinter::marktkeyunparse (int lkey, int holoph, int son)
{
  if (ischildprint (son))
    {
      markkeyunparse (lkey, holoph - 1);
      return pk_true;
    }
  return pk_false;
}


///
boolean _VIRTUAL_
SignalPrinter::marktalphakeyunparse (int lkey, int holoph, int son)
{
  if (ischildprint (son))
    {
      markspkeyunparse (lkey, holoph - 1);
      return pk_true;
    }
  this_intePrinter->printspalphakey (lkey);
  return pk_false;
}


///
boolean _VIRTUAL_
SignalPrinter::marktalphadkeyunparse (int lkey1, int lkey2, int holoph, int son)
{
  if (ischildprint (son))
    {
      markspdkeyunparse (lkey1, lkey2, holoph - 1);
      return pk_true;
    }
  this_intePrinter->printspalphakey (lkey1);
  this_intePrinter->printalphakeysp (lkey2);
  return pk_false;
}


///
void _VIRTUAL_
SignalPrinter::tabunparse (int holoph, int indent)
{
  if (holoph > actindent)
    {
      sntab (indent);
      unparse_subtree (operande, holoph - 1);
      snbacktab ();
    }
  else
    {
      separate ();
      unparse_subtree (operande, holoph - 1);
    }
}


///
_VIRTUAL_ void
SignalPrinter::recsonunparse (int pos)
{
  if (ischildprint (pos))
    unparse_subtree (operande, vholoph);
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::recsondecunparse (int pos)
{
  if (ischildprint (pos))
    unparse_subtree (operande, vholoph - 1);
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::recosonunparse (int pos)
{
  if (ischildprint (pos))
    unparse_subtree (operande, vholoph);
}


///
_VIRTUAL_ void
SignalPrinter::recopinst ()
{
  if (ischildprint (1))
    unparse_subtree (operande, vholoph);
  else
    this_intePrinter->printerror ();
  if (ischildprint (2))
    markgluekeyunparsekey (kwbouv, kwbfer, vholoph - 1);
  else
    {
      this_intePrinter->printkey (kwbouv);
      this_intePrinter->printkey (kwbfer);
    }
}


///
_VIRTUAL_ void
SignalPrinter::recaoper (int pos, boolean nilall)
{
  if (ischildprint (pos))
    markgluekeyunparsekey (kwbouv, kwbfer, vholoph - 1);
  else
    {
      this_intePrinter->printkey (kwbouv);
      if (!nilall)
	this_intePrinter->printerror ();
      this_intePrinter->printkey (kwbfer);
    }
}


///
_VIRTUAL_ void
SignalPrinter::reccoper (int pos, boolean nilall)
{
  if (ischildprint (pos))
    markgluekeyunparsekey (kwcouv, kwcfer, vholoph - 1);
  else
    {
      this_intePrinter->printkey (kwcouv);
      if (!nilall)
	this_intePrinter->printerror ();
      this_intePrinter->printkey (kwcfer);
    }
}


///
_VIRTUAL_ void
SignalPrinter::recroper (int pos, boolean nilall)
{
  if (ischildprint (pos))
    markgluekeyunparsekey (kwpouv, kwpfer, vholoph - 1);
  else
    {
      this_intePrinter->printkey (kwpouv);
      if (!nilall)
	this_intePrinter->printerror ();
      this_intePrinter->printkey (kwpfer);
    }
}


///
_VIRTUAL_ void
SignalPrinter::embedproc (boolean embed)
{
  if (vholoph < actindent)
    this_intePrinter->printholo ();
  else
    {
      if (vholoph > actindent)
	vholoph = vholoph - 1;
      if (embed)
	{
	  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
	  markkeyunparse (kwvert, vholoph);
	  snspace ();
	  this_intePrinter->printkey (kwvert);
	  this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
	}
      else
	markunparse (vholoph);
    }
}


///
_VIRTUAL_ void
SignalPrinter::embedincproc (boolean embed)
{
  if (vholoph < actindent)
    this_intePrinter->printholo ();
  else
    {
      if (embed)
	{
	  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
	  markkeyunparse (kwvert, vholoph);
	  snspace ();
	  this_intePrinter->printkey (kwvert);
	  this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
	}
      else
	markunparse (vholoph);
    }
}


///
_VIRTUAL_ void
SignalPrinter::sigargprocunp ()
{
  embedproc (!(isparproc (voperator)) && (voperator != opexte));
}

///
_VIRTUAL_ void
SignalPrinter::sigargincprocunp ()
{
  embedincproc (!(isparproc (voperator)));
}


///
_VIRTUAL_ void
SignalPrinter::sigargprimprocunp ()
{
  embedproc (isprimproc (voperator) && (voperator != opexte));
}


///
_VIRTUAL_ void
SignalPrinter::sigargincprimprocunp ()
{
  embedincproc (isprimproc (voperator) && (voperator != opexte));
}


///
_VIRTUAL_ void
SignalPrinter::binsigexp (int tkw, int op, boolean balpha)
{
  if (ischildprint (1))
    if (op == voperator)
      if (vholoph < actindent)
	this_intePrinter->printholo ();
      else if (vholoph > actindent)
	unparse_subtree (operande, vholoph - 1);
      else
	unparse_subtree (operande, vholoph);
    else
      embedarg (!(isprimsig (voperator)));
  else
    this_intePrinter->printerror ();

  if (balpha)
    this_intePrinter->printspalphakey (tkw);
  else
    this_intePrinter->printkey (tkw);

  if (ischildprint (2))
    embedarg (!(isprimsig (voperator)));
  else
    this_intePrinter->printerror ();
}


///
_VIRTUAL_ void
SignalPrinter::binsigexp (int tkw, int op)
{
  binsigexp (tkw, op, pk_false);
}


///
_VIRTUAL_ void
SignalPrinter::binalphasigexp (int tkw, int op)
{
  binsigexp (tkw, op, pk_true);
}


///
_VIRTUAL_ void
SignalPrinter::binsighexp (int tkw, int op)
{
  if (ischildprint (1))
    if (op == voperator)
      if (vholoph < actindent)
	this_intePrinter->printholo ();
      else if (vholoph > actindent)
	unparse_subtree (operande, vholoph - 1);
      else
	unparse_subtree (operande, vholoph);
    else
      embedarg (!(isEventOperator (voperator) || isprimsig (voperator)));
  else
    this_intePrinter->printerror ();

  this_intePrinter->printspalphakey (tkw);

  if (ischildprint (2))
    embedarg (!(isprimsig (voperator)));
  else
    this_intePrinter->printerror ();
}


/**
   Printing a restructuration expression.
   \bsg <expr_restruct> ::= <expr_temporel> ":" <expr_temporel> ; \esg
   
*/
_VIRTUAL_ void
SignalPrinter::binsigrestrexp ()
{
  if (ischildprint (1))
    embedarg ((voperator == ophnull) || (voperator == oprest) || (voperator == opsequ));
  else
    this_intePrinter->printerror ();
  this_intePrinter->printspalphakey (kwdpin);
  if (ischildprint (2))
    embedarg ((voperator == ophnull) || (voperator == oprest) || (voperator == opsequ));
  else
    this_intePrinter->printerror ();
}


/**
   Printing a sequence  expression.
   \bsg <expr_sequence> ::= <expr_sequence> "next" <expr_restruct>  \esg
*/
_VIRTUAL_ void
SignalPrinter::binsigseqexp ()
{
  if (ischildprint (1))
    embedarg (voperator == ophnull);
  else
    this_intePrinter->printerror ();
  this_intePrinter->printspalphakey (kwnext);
  if (ischildprint (2))
    embedarg (voperator == ophnull);
  else
    this_intePrinter->printerror ();
}

/**
   Printing a binary process expression with tkw as separator(masking).
*/
_VIRTUAL_ void
SignalPrinter::recunproexp (int tkw)
{
  int loclevel;

  if (ischildprint (1))
    {
      loclevel = unarylevel;
      if (isMaskingOperator (voperator))
	{
	  unarylevel = unarylevel + 1;
	  unparse_subtree (operande, vholoph);
	  this_intePrinter->printlkeysp (tkw, vholoph - actindent);
	}
      else
	{
	  unarylevel = 0;
	  sigargprimprocunp ();
	  snmark (vholoph - actindent);
	  this_intePrinter->printalphakeysp (tkw);
	}
      unarylevel = loclevel;
    }
  else
    {
      this_intePrinter->printerror ();
      this_intePrinter->printalphakeysp (tkw);
      loclevel = -1;
    }
  if (ischildprint (2))
    markunparse (vholoph - 1);
  else
    this_intePrinter->printerror ();
  if (loclevel == 0)
    snrelease (vholoph - actindent);
}


/** Printing a list tree using a separator (seplst) to separate each element and a glue parameter (coller) to shape of the text.
 */
_VIRTUAL_ void
SignalPrinter::decl0 (int seplst, int coller)
{
  int nbfils;
  int i;

  nbfils = ntreearity (treeunparseTree);
  if (nbfils == 0)
    this_intePrinter->printerror ();
  else
    for (i = 1; i <= nbfils; i++)
      {
	operande = nchild (i, treeunparseTree);
	if (IS_NILTREE (operande))
	  this_intePrinter->printerror ();
	else
	  unparse_subtree (operande, vholoph - 1);
	if (i != nbfils)
	  if (coller == 0)
	    this_intePrinter->printkey (seplst);
	  else if (coller == 1)
	    this_intePrinter->printkeysp (seplst);
	  else if (coller > 1)
	    this_intePrinter->printspalphakey (seplst);
	  else if (coller == -2)
	    snspace ();
	  else if (seplst > 0)
	    {
	      this_intePrinter->printkeysp (seplst);
	      wline (vholoph - actindent);
	    }
	  else
	    {
	      wline (vholoph - actindent);
	      if (coller < -1)
		wline (vholoph - actindent);
	    }
      }
}


/** Printing a list tree using a separator (seplst) to separate each element and a glue parameter (coller) to shape of the text.
    A final separator is added at the end of the list.
 */
_VIRTUAL_ void
SignalPrinter::decl1 (int seplst, int coller)
{
  decl0 (seplst, coller);
  this_intePrinter->printkey (seplst);
}

/**
   Printing a list of dependences.
*/
_VIRTUAL_ void
SignalPrinter::decDependences ()
{
  int nbfils;
  int i;

  nbfils = ntreearity (treeunparseTree);
  if (nbfils == 0)
    this_intePrinter->printerror ();
  else
    for (i = 1; i <= nbfils; i++)
      {
	sigarglesign (i);
	if (i != nbfils)
	  this_intePrinter->printspalphakey (kwminuminusupe);
      }
}

/**
   Printing a list of pragmas(operande) with  an holophrase if holoph=0.
*/
_VIRTUAL_ void
SignalPrinter::decpragmas (int holoph)
{
  this_intePrinter->printalphakeysp (kwpragmas);
  wline (holoph - actindent);
  vholoph = holoph - 1;
  if (holoph > actindent)
    {
      sntab (smallindent);
      decl0 (0, -1);
      snbacktab ();
    }
  else
    {
      separate ();
      decl0 (0, -1);
    }
  vholoph = holoph + 1;
  this_intePrinter->printlkeysp (kwend, holoph - actindent);
  this_intePrinter->printkey (kwpragmas);
  wline (holoph - actindent);
}

/**
   Printing a list of composition with seplst as separator.
*/
_VIRTUAL_ void
SignalPrinter::declfeed (int seplst)
{
  int nbfils;
  int i;
  boolean b;

  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
  nbfils = ntreearity (treeunparseTree);
  if (nbfils == 0)
    {
      this_intePrinter->printkey (seplst);
      this_intePrinter->printerror ();
    }
  else
    {
      b = (nbfils == 1);
      for (i = 1; i <= nbfils; i++)
	{
	  operande = nchild (i, treeunparseTree);
	  markkeyunparse (seplst, vholoph - 1);
	  if (b)
	    snspace ();
	  else
	    wline (vholoph - actindent);
	}
    }
  this_intePrinter->printkey (seplst);
  this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
}



/**
   model type printing without first keyword.
*/
_VIRTUAL_ void
SignalPrinter::processtypeunp ()
{
  recsonunparse (1);
  if (!marktalphakeyunparse (kwequl, vholoph - 1, 2))
    this_intePrinter->printerror ();
  if (ischildprint (3))
    unparse_subtree (operande, vholoph);
}


/**
   Printing children of a module(operande tree)
*/
_VIRTUAL_ void
SignalPrinter::moduleunparse ()
{
  ntrees nomproc, vt;
  boolean b;

  vt = operande;
  if (ischildprint (1))
    {
      nomproc = operande;
      unparse_subtree (operande, vholoph);
    }
  else
    {
      this_intePrinter->printerror ();
      SET_NILTREE (nomproc);
    }
  this_intePrinter->printspkey (kwequl);
  wline (vholoph - actindent);
  if (ischildprint (2))
    tabunparse (vholoph - 1, smallindent);
  else
    this_intePrinter->printerror ();
  this_intePrinter->printlkeysp (kwend, vholoph - actindent);
  if (ISNOT_NILTREE (nomproc))
    {
      this_intePrinter->intedebcomprint ();
      identornameprint (nomproc);
      this_intePrinter->intefincomprint ();
    }

  postlude_root (vt);
}


/**
   Printing a nullary tree: opnom operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opnom ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opchar operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opchar ()
{
  this_intePrinter->printalphakey (kwchar);
}

/**
   Printing a nullary tree: opcomplex operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcomplex ()
{
  this_intePrinter->printalphakey (kwcomplex);
}

/**
   Printing a nullary tree: opcstcar operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcstcar ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opcstcha operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcstcha ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opcstdp operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcstdp ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opcstent operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcstent ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opcstreel operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcstreel ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opdcomplex operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opdcomplex ()
{
  this_intePrinter->printalphakey (kwdcomplex);
}

/**
   Printing a nullary tree: opdeterm operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opdeterm ()
{
  this_intePrinter->printalphakey (kwdeterministic);
}

/**
   Printing a nullary tree: opdpreal operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opdpreal ()
{
  this_intePrinter->printalphakey (kwdreal);
}

/**
   Printing a nullary tree: opevent operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opevent ()
{
  this_intePrinter->printalphakey (kwevent);
}

/**
   Printing a nullary tree: opfalse operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opfalse ()
{
  this_intePrinter->printalphakey (kwfalse);
}

/**
   Printing a nullary tree: ophnull operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_ophnull ()
{
  this_intePrinter->printalphakey (kwchap0);
}

/**
   Printing a nullary tree: opimplicit operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opimplicit ()
{
}

/**
   Printing a nullary tree: opinteger operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opinteger ()
{
  this_intePrinter->printalphakey (kwinteger);
}

/**
   Printing a nullary tree: oplogical operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_oplogical ()
{
  this_intePrinter->printalphakey (kwboolean);
}

/**
   Printing a nullary tree: oplong operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_oplong ()
{
  this_intePrinter->printalphakey (kwlong);
}

/**
   Printing a nullary tree: opnil operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opnil ()
{
}

/**
   Printing a nullary tree: opnullproc operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opnullproc ()
{
  this_intePrinter->printkey (kwpouv);
  this_intePrinter->printkey (kwvert);
  snspace ();
  this_intePrinter->printkey (kwvert);
  this_intePrinter->printkey (kwpfer);
}

/**
   Printing a nullary tree: opop  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opop ()
{
  inteconsalphaprint (treeunparseTree);
}

/**
   Printing a nullary tree: opreal  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opreal ()
{
  this_intePrinter->printalphakey (kwreal);
}

/**
   Printing a nullary tree: opretun  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opretun ()
{
}

/**
   Printing a nullary tree: opsafe  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opsafe ()
{
  this_intePrinter->printalphakey (kwsafe);
}


/**
   Printing a nullary tree: opshort  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opshort ()
{
  this_intePrinter->printalphakey (kwshort);
}

/**
   Printing a nullary tree: opstring  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opstring ()
{
  this_intePrinter->printalphakey (kwstring);
}

/**
   Printing a nullary tree: optabvide  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_optabvide ()
{
  this_intePrinter->printcspkey (kwcouv);
  this_intePrinter->printalphakey (kwcfer);
}

/**
   Printing a nullary tree: optrue  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_optrue ()
{
  this_intePrinter->printalphakey (kwtrue);
}

/**
   Printing a nullary tree: opunsafe  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opunsafe ()
{
  this_intePrinter->printalphakey (kwunsafe);
}

/**
   Printing a nullary tree: opcomment  operator. 
*/
_VIRTUAL_ void
SignalPrinter::decta0_opcomment ()
{
  inteconsprint (treeunparseTree);
}

/** 
    Printing a nullary tree(leaf) using globals(operande, voperator)
*/
_VIRTUAL_ void
SignalPrinter::decta0 ()
{
  switch ((int) voperator)
    {
    case opnom:
      decta0_opnom ();
      break;
    case opchar:
      decta0_opchar ();
      break;
    case opcomplex:
      decta0_opcomplex ();
      break;
    case opcstcar:
      decta0_opcstcar ();
      break;
    case opcstcha:
      decta0_opcstcha ();
      break;
    case opcstdp:
      decta0_opcstdp ();
      break;
    case opcstent:
      decta0_opcstent ();
      break;
    case opcstreel:
      decta0_opcstreel ();
      break;
    case opdcomplex:
      decta0_opdcomplex ();
      break;
    case opdeterm:
      decta0_opdeterm ();
      break;
    case opdpreal:
      decta0_opdpreal ();
      break;
    case opevent:
      decta0_opevent ();
      break;
    case opfalse:
      decta0_opfalse ();
      break;
    case ophnull:
      decta0_ophnull ();
      break;
    case opimplicit:
      decta0_opimplicit ();
      break;
    case opinteger:
      decta0_opinteger ();
      break;
    case oplogical:
      decta0_oplogical ();
      break;
    case oplong:
      decta0_oplong ();
      break;
    case opnil:
      decta0_opnil ();
      break;
    case opnullproc:
      decta0_opnullproc ();
      break;
    case opop:
      decta0_opop ();
      break;
    case opreal:
      decta0_opreal ();
      break;
    case opretun:
      decta0_opretun ();
      break;
    case opsafe:
      decta0_opsafe ();
      break;
    case opshort:
      decta0_opshort ();
      break;
    case opstring:
      decta0_opstring ();
      break;
    case optabvide:
      decta0_optabvide ();
      break;
    case optrue:
      decta0_optrue ();
      break;
    case opunsafe:
      decta0_opunsafe ();
      break;
    case opcomment:
      decta0_opcomment ();
      break;
    default:
      this_intePrinter->printerror ();
    }
}

/**
	Printing an unary tree: opevnt operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opevnt ()
{
  unsigexp (kwchap, 1);
}

/**
	Printing an unary tree: opexte operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opexte ()
{
  this_intePrinter->printalphakey (kwexternal);
  recosonunparse (1);
}

/**
	Printing an unary tree: oplab operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_oplab ()
{
  if (ischildprint (1))
    {
      markkeyunparse (kwlabel, vholoph);
      //      this_intePrinter->printkeysp(kwpvir);
    }
}

/**
	Printing an unary tree: opnega operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opnega ()
{
  unsigexp (kwminu, 1);
}

/**
	Printing an unary tree: opnon operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opnon ()
{
  unalphasigexp (kwnot, 1);
}

/**
	Printing an unary tree: oppcar operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_oppcar ()
{
  this_intePrinter->printkey (kwinfeinfe);
  recsonunparse (1);
  this_intePrinter->printkey (kwsupesupe);
}

/**
	Printing an unary tree: opref operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opref ()
{
  if (ischildprint (1))
    {
      markkeyunparse (kwref, vholoph);
      //      this_intePrinter->printkeysp(kwpvir);
    }
}

/**
	Printing an unary tree: opsdec operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opsdec ()
{
  unsigfact (1);
  this_intePrinter->printkey (kwcouv);
  this_intePrinter->printkey (kwinte);
  this_intePrinter->printkey (kwcfer);
}

/**
	Printing an unary tree: opstru operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opstru ()
{
  this_intePrinter->printalphakeysp (kwstruct);
  if (ischildprint (1))
    markkeyunparsekey (kwpouv, kwpfer, vholoph);
}

/**
	Printing an unary tree: optran operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_optran ()
{
  unalphasigexp (kwtr, 1);
}

/**
	Printing an unary tree: opvvarin operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opvvarin ()
{
  recsonunparse (1);
  this_intePrinter->printkey (kwinte);
}

/**
   Printing an unary tree: opwhen1 operator.
*/
_VIRTUAL_ void
SignalPrinter::decta1_opwhen1 ()
{
  int tk;
  ntrees t;
  if (shorten_clocks)
    {
      if (isChildNot (operande, &t))
	tk = kwcouvslahdpin;
      else
	tk = kwcouvdpin;
      this_intePrinter->printkey (tk);
      treeunparseTree = t;
      recsondecunparse (1);
      this_intePrinter->printkey (kwcfer);
    }
  else
    unalphasigexp (kwwhen, 1);
}

/**
   Printing an unary tree: oppassert operator: assert(| ... |).
*/
_VIRTUAL_ void
SignalPrinter::decta1_oppassert ()
{
  this_intePrinter->printalphakeysp (kwassert);
  if (ischildprint (1))
    unparse_subtree (operande, vholoph - 1);
}

/**
   Printing an unary tree: opappeassert operator: assert(C).
*/
_VIRTUAL_ void
SignalPrinter::decta1_opappeassert ()
{
  this_intePrinter->printalphakeysp (kwassert);
  if (ischildprint (1))
    markkeyunparsekey (kwpouv, kwpfer, vholoph);
  //  unalphasigexp(kwassert, 1);
}


/** 
    Printing a unary tree using globals(operande, voperator).
    It switches, to decta1_XX method, according to the value(XX) of  the current operator(voperator).
*/
_VIRTUAL_ void
SignalPrinter::decta1 ()
{
  switch ((int) voperator)
    {
    case opevnt:
      decta1_opevnt ();
      break;
    case opexte:
      decta1_opexte ();
      break;
    case oplab:
      decta1_oplab ();
      break;
    case opnega:
      decta1_opnega ();
      break;
    case opnon:
      decta1_opnon ();
      break;
    case oppcar:
      decta1_oppcar ();
      break;
    case opref:
      decta1_opref ();
      break;
    case opsdec:
      decta1_opsdec ();
      break;
    case opstru:
      decta1_opstru ();
      break;
    case optran:
      decta1_optran ();
      break;
    case opvvarin:
      decta1_opvvarin ();
      break;
    case opwhen1:
      decta1_opwhen1 ();
      break;
    case oppassert:
      decta1_oppassert ();
      break;			/* assert(| ... |) */
    case opappeassert:
      decta1_opappeassert ();
      break;			/* assert(c) */
    default:
      this_intePrinter->printerror ();
    }
}



/**
   Printing a binary tree: opappe operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opappe ()
{
  recsonunparse (1);
  recroper (2, pk_true);
}

/**
   Printing a binary tree: opbinf operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opbinf ()
{
  binsighexp (kwchapetoi, opbinf);
}

/**
   Printing a binary tree: opbsup operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opbsup ()
{
  binsighexp (kwchapplus, opbsup);
}

/**
   Printing a binary tree: opbund operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opbund ()
{
  this_intePrinter->printalphakeysp (kwbundle);
  if (ischildprint (1))
    markkeyunparsekey (kwpouv, kwpfer, vholoph);
  if (ischildprint (2))
    {
      wline (vholoph - actindent);
      this_intePrinter->printalphakeysp (kwspec);
      sigargprocunp ();
    };
}

/**
   Printing a binary tree: opcas operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opcas ()
{
  sigarglesign (1);
  if (ischildprint (2))
    {
      this_intePrinter->printkeysp (kwdpin);
      sigargprimprocunp ();
    }
  else
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opcham operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opcham ()
{
  unsigfact (1);
  this_intePrinter->printkey (kwpoin);
  recsondecunparse (2);
}

/**
   Printing a binary tree: opcomp operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opcomp ()
{
  binsighexp (kwchapminu, 0);
}

/**
   Printing a binary tree: opconc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opconc ()
{
  binalphasigexp (kwvertplus, opconc);
}

/**
   Printing a binary tree: opcons operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opcons ()
{
  recsonunparse (1);
  if (!marktalphakeyunparse (kwequl, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opconv operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opconv ()
{
  recsonunparse (1);
  recroper (2, pk_false);
}

/**
   Printing a binary tree: opcplx operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opcplx ()
{
  binsigexp (kwarro, 0);
}

/**
   Printing a binary tree: opdeccon operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdeccon ()
{
  this_intePrinter->printalphakeysp (kwconstant);
  printtype ();
  printSecondChildDeclaration ();
}

/**
   Printing a binary tree: opdecpar operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdecpar ()
{
  this_intePrinter->printalphakeysp (kwshared);
  printtype ();
  printSecondChildDeclaration ();
}

/**
   Printing a binary tree: opdecconp operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdecconp ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwconstant);
  printtype ();
  printSecondChildDeclaration ();
}

/**
   Printing a binary tree: opdecsig operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdecsig ()
{
  printtype ();
  printSecondChildDeclaration ();
}

/**
   Printing a binary tree: opdecvar operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdecvar ()
{
  this_intePrinter->printalphakeysp (kwstatevar);
  printtype ();
  printSecondChildDeclaration ();
}

/**
   Printing a binary tree: opdefegal operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdefegal ()
{
  recsonunparse (1);
  if (!marktalphakeyunparse (kwdpinequldpin, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opdefpart operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdefpart ()
{
  recsonunparse (1);
  if (!marktalphakeyunparse (kwdpindpinequl, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opdefpartdef operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdefpartdef ()
{
  recsonunparse (1);
  if (!marktalphadkeyunparse (kwdpindpinequl, kwdefaultvalue, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opdefsig operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdefsig ()
{
  recsonunparse (1);
  if (!marktalphakeyunparse (kwdpinequl, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opdesc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdesc ()
{
  if (ischildprint (1))
    unparse_subtree (operande, vholoph);
  if (ischildprint (2))
    sigargincprimprocunp ();
}

/**
   Printing a binary tree: opdescmodu operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdescmodu ()
{
  if (ischildprint (1))
    unparse_subtree (operande, vholoph);
  if (ischildprint (2))
    unparse_subtree (operande, vholoph);
}

/**
   Printing a binary tree: opdiff operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdiff ()
{
  binsigexp (kwslahequl, 0);
}

/**
   Printing a binary tree: opdivi operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opdivi ()
{
  binsigexp (kwslah, 0);
}

/**
   Printing a binary tree: opegal operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opegal ()
{
  binsigexp (kwequl, 0);
}

/**
   Printing a binary tree: opegeg operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opegeg ()
{
  binsigexp (kwequlequl, 0);
}

/**
   Printing a binary tree: opelem operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opelem ()
{
  unsigfact (1);
  reccoper (2, pk_false);
}

/**
   Printing a binary tree: opelevec operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opelevec ()
{
  reccoper (1, pk_true);
  this_intePrinter->printkey (kwdpin);
  recsondecunparse (2);
}

/**
   Printing a binary tree: opensvec operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opensvec ()
{
  recaoper (1, pk_false);
  this_intePrinter->printkey (kwdpin);
  recsondecunparse (2);
}

/**
   Printing a binary tree: opentsor operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opentsor ()
{
  this_intePrinter->printkeymark (kwpouv, vholoph - actindent);
  if (marktkeyunparse (kwinte, vholoph, 1))
    wline (vholoph - actindent - 1);
  if (marktkeyunparse (kwxcla, vholoph, 2))
    wline (vholoph - actindent - 1);
  this_intePrinter->printkeyrelease (kwpfer, vholoph - actindent);
}

/**
   Printing a binary tree: opet operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opet ()
{
  binalphasigexp (kwand, opet);
}

/**
   Printing a binary tree: opextr operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opextr ()
{
  binalphasigexp (kwwhen, opextr);
}

/**
   Printing a binary tree: opformact operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opformact ()
{
  this_intePrinter->printalphakeysp (kwaction);
  recsonunparse (1);
  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
}

/**
   Printing a binary tree: opformact operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opformautomaton ()
{
  this_intePrinter->printalphakeysp (kwautomaton);
  recsonunparse (1);
  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
}

/**
   Printing a binary tree: opformfunc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opformfunc ()
{
  this_intePrinter->printalphakeysp (kwfunction);
  recsonunparse (1);
  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
}

/**
   Printing a binary tree: opformnode operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opformnode ()
{
  this_intePrinter->printalphakeysp (kwnode);
  recsonunparse (1);

  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
}

/**
   Printing a binary tree: opformproc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opformproc ()
{
  this_intePrinter->printalphakeysp (kwprocess);
  recsonunparse (1);

  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
}

/**
   Printing a binary tree: opformprocedure operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opformprocedure ()
{
  this_intePrinter->printalphakeysp (kwprocedure);
  recsonunparse (1);

  if (ischildprint (2))
    markunparsegluekey ( /*kwpvir */ 0, vholoph + 1);
}

/**
   Printing a binary tree: opiegal operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opiegal ()
{
  binsigexp (kwinfeequl, 0);
}

/**
   Printing a binary tree: opiieg operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opiieg ()
{
  binsigexp (kwinfeinfeequl, 0);
}

/**
   Printing a binary tree: opindmot operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opindmot ()
{
  recsondecunparse (1);
  this_intePrinter->printspalphakey (kwto);
  recsondecunparse (2);
}

/**
   Printing a binary tree: opinfe operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opinfe ()
{
  binsigexp (kwinfe, 0);
}

/**
   Printing a binary tree: opinit operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opinit ()
{
  recsonunparse (1);
  this_intePrinter->printspalphakey (kwinit);
  recsondecunparse (2);
}

/**
   Printing a binary tree: opinst operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opinst ()
{
  recopinst ();
}

/**
   Printing a binary tree: opintvff operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opintvff ()
{
  this_intePrinter->printkey (kwcouvpoin);
  recsondecunparse (1);
  this_intePrinter->printkey (kwvirg);
  recsondecunparse (2);
  this_intePrinter->printkey (kwpoincfer);
}

/**
   Printing a binary tree: opintvfo operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opintvfo ()
{
  this_intePrinter->printkey (kwcouvpoin);
  recsondecunparse (1);
  this_intePrinter->printkey (kwvirg);
  recsondecunparse (2);
  this_intePrinter->printkey (kwcouvpoin);
}

/**
   Printing a binary tree: opintvof operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opintvof ()
{
  this_intePrinter->printkey (kwpoincfer);
  recsondecunparse (1);
  this_intePrinter->printkey (kwvirg);
  recsondecunparse (2);
  this_intePrinter->printkey (kwpoincfer);
}

/**
   Printing a binary tree: opintvoo operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opintvoo ()
{
  this_intePrinter->printkey (kwpoincfer);
  recsondecunparse (1);
  this_intePrinter->printkey (kwvirg);
  recsondecunparse (2);
  this_intePrinter->printkey (kwcouvpoin);
}

/**
   Printing a binary tree: opmele operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opmele ()
{
  binalphasigexp (kwdefault, opmele);
}

/**
   Printing a binary tree: opmmul operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opmmul ()
{
  binsigexp (kwetoipoin, 0);
}

/**
   Printing a binary tree: opmod operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opmod ()
{
  binalphasigexp (kwmodulo, 0);
}

/**
   Printing a binary tree: opmodu operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opmodu ()
{
  this_intePrinter->printalphakeysp (kwmodule);
  moduleunparse ();
}

/**
   Printing a binary tree: opmoin operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opmoin ()
{
  binsigexp (kwminu, 0);
}

/**
   Printing a binary tree: opmult operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opmult ()
{
  binsigexp (kwetoi, opmult);
}

/**
   Printing a binary tree: opnomba operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opnomba ()
{
  binalphasigexp (kwafter, 0);
}

/**
   Printing a binary tree: opnombd operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opnombd ()
{
  binalphasigexp (kwfrom, 0);
}

/**
   Printing a binary tree: opnombm operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opnombm ()
{
  binalphasigexp (kwcount, 0);
}

/**
   Printing a binary tree: opobli operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opobli ()
{
  recunproexp (kwslah);
}

/**
   Printing a binary tree: opou operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opou ()
{
  binalphasigexp (kwor, opou);
}

/**
   Printing a binary tree: opouex operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opouex ()
{
  binalphasigexp (kwxor, opouex);
}

/**
   Printing a binary tree: oppconf operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oppconf ()
{
  boolean b;

  b = (noper (nfather (operande)) == opdesc);
  if (ischildprint (1))
    sigargprocunp ();
  else
    {
      this_intePrinter->printkey (kwpouv);
      this_intePrinter->printkey (kwvert);
      snspace ();
      this_intePrinter->printkey (kwvert);
      this_intePrinter->printkey (kwpfer);
    }
  if (ischildprint (2))
    if (ntreearity (operande) != 0)	/*la liste des declarations peut etre vide */
      if (b)
	{
	  this_intePrinter->printlkeysp (kwwhere, vholoph - actindent);
	  wline (vholoph - actindent);
	  unparse_subtree (operande, vholoph - 1);
	  this_intePrinter->printlkey (kwend, vholoph - actindent);
	}
      else
	{			/* Pour where dans une expression de processus */
	  this_intePrinter->printspkeymark (kwwhere, vholoph - actindent);
	  unparse_subtree (operande, vholoph - 1);
	  this_intePrinter->printkeyrelease (kwend, vholoph - actindent);
	}

}

/**
   Printing a binary tree: oppcontrol operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oppcontrol ()
{
  this_intePrinter->printalphakeysp (kwon);
  recsondecunparse (1);
  if (!marktalphakeyunparse (kwdpindpin, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: oppetiq operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oppetiq ()
{
  recsondecunparse (1);
  if (!marktalphakeyunparse (kwdpindpin, vholoph - 1, 2))
    this_intePrinter->printerror ();
}

/**
   Printing a binary tree: opparc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opparc ()
{
  if (ischildprint (1))
    unparse_subtree (operande, vholoph - 1);
  recsondecunparse (2);
}


/**
   Printing a binary tree: opplus operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opplus ()
{
  binsigexp (kwplus, opplus);
}

/**
   Printing a binary tree: oppuis operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oppuis ()
{
  binsigexp (kwetoietoi, 0);
}

/**
   Printing a binary tree: oprecu operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oprecu ()
{
  binsigexp (kwbslabsla, 0);
}

/**
   Printing a binary tree: oprepe operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oprepe ()
{
  binalphasigexp (kwvertetoi, 0);
}

/**
   Printing a binary tree: oprest operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_oprest ()
{
  binsigrestrexp ();
}

/**
   Printing a binary tree: opsegal operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opsegal ()
{
  binsigexp (kwsupeequl, 0);
}

/**
   Printing a binary tree: opsequ operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opsequ ()
{
  binsigseqexp ();
}

/**
   Printing a binary tree: opspec operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opspec ()
{
  if (ischildprint (1))
    {
      unparse_subtree (operande, vholoph - 1);
      if (ischildprint (2))
	{
	  wline (vholoph - actindent);
	  this_intePrinter->printalphakeysp (kwspec);
	  sigargincprocunp ();
	}
    }
  else if (ischildprint (2))
    {
      this_intePrinter->printalphakeysp (kwspec);
      sigargincprocunp ();
    }
}

/**
   Printing a binary tree: opsupe operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opsupe ()
{
  binsigexp (kwsupe, 0);
}

/**
   Printing a binary tree: optvect operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_optvect ()
{
  reccoper (1, pk_true);
  recsonunparse (2);
}

/**
   Printing a binary tree: opvenum operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opvenum ()
{
  if (ischildprint (1))
    sigfactargunp ();
  this_intePrinter->printkey (kwdies);
  recsondecunparse (2);
}

_VIRTUAL_ void
SignalPrinter::spalphakeSigargunp (int kw, int ison)
{
  if (ischildprint (ison))
    {
      this_intePrinter->printspalphakey (kw);
      sigargunp ();
    }
}

/**
   Printing a binary tree: opvar operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opvar ()
{
  unalphasigexp (kwvar, 1);
  spalphakeSigargunp (kwinit, 2);
}

/**
   Printing a binary tree: opident operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opident ()
{
  separate ();
  identprint (treeunparseTree);
}

/**
   Printing a binary tree: opident operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opidentfield ()
{
  separate ();
  identprint (treeunparseTree);
}

/**
   Printing a binary tree: opleft_tt operator.
*/
_VIRTUAL_ void
SignalPrinter::decta2_opleft_tt ()
{
  this_intePrinter->printalphakey (kwlefttt);
  this_intePrinter->printalphakey (kwpouv);
  binsigexp (kwvirg, 0);
  this_intePrinter->printalphakey (kwpfer);
}

/** 
    Printing a binary-tree using globals(operande, voperator)
    It switches, to decta2_XX method, according to the value(XX) of  the current operator(voperator).
*/
_VIRTUAL_ void
SignalPrinter::decta2 ()
{
  switch ((int) voperator)
    {
    case opappe:
      decta2_opappe ();
      break;
    case opbinf:
      decta2_opbinf ();
      break;
    case opbsup:
      decta2_opbsup ();
      break;
    case opbund:
      decta2_opbund ();
      break;
    case opcas:
      decta2_opcas ();
      break;
    case opcham:
      decta2_opcham ();
      break;
    case opcomp:
      decta2_opcomp ();
      break;
    case opconc:
      decta2_opconc ();
      break;
    case opcons:
      decta2_opcons ();
      break;
    case opconv:
      decta2_opconv ();
      break;
    case opcplx:
      decta2_opcplx ();
      break;
    case opdeccon:
      decta2_opdeccon ();
      break;
    case opdecconp:
      decta2_opdecconp ();
      break;
    case opdecpar:
      decta2_opdecpar ();
      break;
    case opdecsig:
      decta2_opdecsig ();
      break;
    case opdecvar:
      decta2_opdecvar ();
      break;
    case opdefegal:
      decta2_opdefegal ();
      break;
    case opdefpart:
      decta2_opdefpart ();
      break;
    case opdefpartdef:
      decta2_opdefpartdef ();
      break;
    case opdefsig:
      decta2_opdefsig ();
      break;
    case opdesc:
      decta2_opdesc ();
      break;
    case opdescmodu:
      decta2_opdescmodu ();
      break;
    case opdiff:
      decta2_opdiff ();
      break;
    case opdivi:
      decta2_opdivi ();
      break;
    case opegal:
      decta2_opegal ();
      break;
    case opegeg:
      decta2_opegeg ();
      break;
    case opelem:
      decta2_opelem ();
      break;
    case opelevec:
      decta2_opelevec ();
      break;
    case opensvec:
      decta2_opensvec ();
      break;
    case opentsor:
      decta2_opentsor ();
      break;
    case opet:
      decta2_opet ();
      break;
    case opextr:
      decta2_opextr ();
      break;
    case opformact:
      decta2_opformact ();
      break;
    case opformautomaton:
      decta2_opformautomaton ();
      break;
    case opformfunc:
      decta2_opformfunc ();
      break;
    case opformnode:
      decta2_opformnode ();
      break;
    case opformproc:
      decta2_opformproc ();
      break;
    case opformprocedure:
      decta2_opformprocedure ();
      break;
    case opiegal:
      decta2_opiegal ();
      break;
    case opiieg:
      decta2_opiieg ();
      break;
    case opindmot:
      decta2_opindmot ();
      break;
    case opinfe:
      decta2_opinfe ();
      break;
    case opinit:
      decta2_opinit ();
      break;
    case opinst:
      decta2_opinst ();
      break;
    case opintvff:
      decta2_opintvff ();
      break;
    case opintvfo:
      decta2_opintvfo ();
      break;
    case opintvof:
      decta2_opintvof ();
      break;
    case opintvoo:
      decta2_opintvoo ();
      break;
    case opmele:
      decta2_opmele ();
      break;
    case opmmul:
      decta2_opmmul ();
      break;
    case opmod:
      decta2_opmod ();
      break;
    case opmodu:
      decta2_opmodu ();
      break;
    case opmoin:
      decta2_opmoin ();
      break;
    case opmult:
      decta2_opmult ();
      break;
    case opnomba:
      decta2_opnomba ();
      break;
    case opnombd:
      decta2_opnombd ();
      break;
    case opnombm:
      decta2_opnombm ();
      break;
    case opobli:
      decta2_opobli ();
      break;
    case opou:
      decta2_opou ();
      break;
    case opouex:
      decta2_opouex ();
      break;
    case oppconf:
      decta2_oppconf ();
      break;
    case oppcontrol:
      decta2_oppcontrol ();
      break;
    case oppetiq:
      decta2_oppetiq ();
      break;
    case opparc:
      decta2_opparc ();
      break;
    case opplus:
      decta2_opplus ();
      break;
    case oppuis:
      decta2_oppuis ();
      break;
    case oprecu:
      decta2_oprecu ();
      break;
    case oprepe:
      decta2_oprepe ();
      break;
    case oprest:
      decta2_oprest ();
      break;
    case opsegal:
      decta2_opsegal ();
      break;
    case opsequ:
      decta2_opsequ ();
      break;
    case opspec:
      decta2_opspec ();
      break;
    case opsupe:
      decta2_opsupe ();
      break;
    case optvect:
      decta2_optvect ();
      break;
    case opvenum:
      decta2_opvenum ();
      break;
    case opvar:
      decta2_opvar ();
      break;
    case opident:
      decta2_opident ();
      break;
    case opidentfield:
      decta2_opidentfield ();
      break;
    case opleft_tt:
      decta2_opleft_tt ();
      break;

    default:
      this_intePrinter->printerror ();
    }
}


/**
   Printing a list tree: opcomments operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opcomments ()
{
  decl0 (0, -1);
}

/**
   Printing a list tree: opchai operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opchai ()
{
  decl0 (0, -1);
}

/**
   Printing a list tree: oplattr operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplattr ()
{
  decl0 (0, -2);
}


/**
   Printing a list tree: oplcas operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplcas ()
{
  decl0 (0, -1);
}

/**
   Printing a list tree: oplcons operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplcons ()
{
  decl0 (kwvirg, 1);
}

/**
   Printing a list tree: oplcrohor operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplcrohor ()
{
  decl0 (kwchapinfe, 2);
}

/**
   Printing a list tree: opldechor operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opldechor ()
{
  decl0 (kwchapsupe, 2);
}

/**
   Printing a list tree: opldecs operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opldecs ()
{
  //  decl0(0, -1);
  decl1 (kwpvir, -1);
}

/**
   Printing a list tree: opldecl operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opldecl ()
{
  // decl0(0, -1);
  decl1 (kwpvir, -1);
}

/**
   Printing a list tree: opldep operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opldep ()
{
  decDependences ();
}

/**
   Printing a list tree: opleflot operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opleflot ()
{
  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
  decl0 (kwvirg, 1);
  this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
}

/**
   Printing a list tree: oplepara operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplepara ()
{
  decl0 (kwvirg, 1);
}

/**
   Printing a list tree: oplesign operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplesign ()
{
  decl0 (kwvirg, 1);
}

/**
   Printing a list tree: oplevect operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplevect ()
{
  this_intePrinter->printgluekeymark (kwcouv, vholoph - actindent);
  decl0 (kwvirg, 0);
  this_intePrinter->printgluekeyrelease (kwcfer, vholoph - actindent);
}

/**
   Printing a list tree: oplelevec operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplelevec ()
{
  this_intePrinter->printgluekeymark (kwcouv, vholoph - actindent);
  decl0 (kwvirg, 0);
  this_intePrinter->printgluekeyrelease (kwcfer, vholoph - actindent);
}

/**
   Printing a list tree: oplequhor operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplequhor ()
{
  decl0 (kwchapequl, 2);
}


/**
   Printing a list tree: oplexchor operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplexchor ()
{
  decl0 (kwchapdies, 2);
}

/**
   Printing a list tree: oplfeed operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplfeed ()
{
  declfeed (kwvert);
}


/**
   Printing a list tree: oplident operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplident ()
{
  decl0 (kwvirg, 1);
}


/**
   Printing a list tree: oplimpo operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplimpo ()
{
  this_intePrinter->printspalphakey (kwuse);
  decl0 (kwvirg, 1);
  //  this_intePrinter->printkeysp(kwpvir);
}

/**
   Printing a list tree: oplmodu operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplmodu ()
{
  // decl0(0, -1);
  decl1 (kwpvir, -1);
}

/**
   Printing a list tree: oplopra operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplopra ()
{
  this_intePrinter->printgluekeymark (kwbouv, vholoph - actindent);
  decl0 (kwvirg, 0);
  this_intePrinter->printgluekeyrelease (kwbfer, vholoph - actindent);
}

/**
   Printing a list tree: oplparc operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplparc ()
{
  decl0 (kwvirg, 1);
}

/**
   Printing a list tree: oplprag operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplprag ()
{
  decpragmas (vholoph);
}

/**
   Printing a list tree: oplseq operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplseq ()
{
  declfeed (kwpvir);
}


/**
   Printing a list tree: oplsign operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_oplsign ()
{
  decl0 (kwvirg, 1);
}

/**
   Printing a list tree: opltype operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opltype ()
{
  this_intePrinter->printalphakeysp (kwtype);
  snmark (vholoph - actindent);
  decl0 (kwvirg, 1);
  this_intePrinter->printreleasegluekey ( /*kwpvir */ 0, vholoph - actindent);
}

/**
   Printing a list tree: opltypep operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_opltypep ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwtype);
  snmark (vholoph - actindent);
  decl0 (kwvirg, 1);
  this_intePrinter->printreleasegluekey ( /*kwpvir */ 0, vholoph - actindent);
}


/**
   Printing a list tree: optenum operator. 
*/
_VIRTUAL_ void
SignalPrinter::declist_optenum ()
{
  this_intePrinter->printalphakeysp (kwenum);
  this_intePrinter->printgluekeymark (kwpouv, vholoph - actindent);
  decl0 (kwvirg, 1);
  this_intePrinter->printgluekeyrelease (kwpfer, vholoph - actindent);
}


/** 
    Printing a list tree using class attribute members: operande, voperator.
    It switches, to declist_XX method, according to the value(XX) of  the current operator(voperator).
  */
_VIRTUAL_ void
SignalPrinter::declist ()
{
  switch ((int) voperator)
    {
    case opcomments:
      declist_opcomments ();
      break;
    case opchai:
      declist_opchai ();
      break;
    case oplattr:
      declist_oplattr ();
      break;
    case oplcas:
      declist_oplcas ();
      break;
    case oplcons:
      declist_oplcons ();
      break;
    case oplcrohor:
      declist_oplcrohor ();
      break;
    case opldechor:
      declist_opldechor ();
      break;
    case opldecs:
      declist_opldecs ();
      break;
    case opldecl:
      declist_opldecl ();
      break;
    case opldep:
      declist_opldep ();
      break;
    case opleflot:
      declist_opleflot ();
      break;
    case oplepara:
      declist_oplepara ();
      break;
    case oplesign:
      declist_oplesign ();
      break;
    case oplevect:
      declist_oplevect ();
      break;
    case oplelevec:
      declist_oplelevec ();
      break;
    case oplequhor:
      declist_oplequhor ();
      break;
    case oplexchor:
      declist_oplexchor ();
      break;
    case oplfeed:
      declist_oplfeed ();
      break;
    case oplident:
      declist_oplident ();
      break;
    case oplimpo:
      declist_oplimpo ();
      break;
    case oplmodu:
      declist_oplmodu ();
      break;
    case oplopra:
      declist_oplopra ();
      break;
    case oplparc:
      declist_oplparc ();
      break;
    case oplprag:
      declist_oplprag ();
      break;
    case oplseq:
      declist_oplseq ();
      break;
    case oplsign:
      declist_oplsign ();
      break;
    case opltype:
      declist_opltype ();
      break;
    case opltypep:
      declist_opltypep ();
      break;
    case optenum:
      declist_optenum ();
      break;

    default:
      this_intePrinter->printerror ();
    }
}


/**
   Model declaration printing without first keyword.
  (action, process, node, private node,...)
*/
_VIRTUAL_ void
SignalPrinter::processUnparsing ()
{
  ntrees nomproc;
  //  boolean b;

  //  b = IS_NILTREE(nfather(operande)) || (nfather(operande)==operande);
  if (ischildprint (1))
    {
      nomproc = operande;
      unparse_subtree (operande, vholoph);
    }
  else
    {
      this_intePrinter->printerror ();
      SET_NILTREE (nomproc);
    }
  this_intePrinter->printspkey (kwequl);
  wline (vholoph - actindent);
  if (ischildprint (2))
    tabunparse (vholoph + 1, X1_indent);
  else
    this_intePrinter->printerror ();
  wline (vholoph - actindent);
  if (ischildprint (3))
    tabunparse (vholoph - 1, smallindent);
  else
    this_intePrinter->printerror ();
  wline (vholoph - actindent);
  // if((ISNOT_NILTREE(nomproc)) && viscomm)
  //   {
  //     this_intePrinter->intedebcomprint();
  //     identornameprint(nomproc);
  //     this_intePrinter->intefincomprint();
  //  }
  //  if (b) this_intePrinter->printkeysp(kwpvir);
}

/**
  Printing a ternary tree :  opact operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opact ()
{
  this_intePrinter->printalphakeysp (kwaction);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opactp operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opactp ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwaction);
  processUnparsing ();
}

/**
  Printing a ternary tree :  opautomaton operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opautomaton ()
{
  this_intePrinter->printalphakeysp (kwautomaton);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opautomatonp operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opautomatonp ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwautomaton);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opborn operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opborn ()
{
  spalphakeSigargunp (kwin, 1);	// if(ischildprint(1)) { this_intePrinter->printspalphakey(kwin); sigargunp(); }
  spalphakeSigargunp (kwto, 2);	// if(ischildprint(2)) { this_intePrinter->printspalphakey(kwto); sigargunp(); }
  spalphakeSigargunp (kwstep, 3);	// if(ischildprint(3)) { this_intePrinter->printspalphakey(kwstep); sigargunp(); }
}

/**
   Printing a ternary tree :  opcond operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opcond ()
{
  this_intePrinter->printalphakeysp (kwif);
  recsondecunparse (1);
  this_intePrinter->printspalphakey (kwthen);
  recsondecunparse (2);
  unalphasigexp (kwelse, 3);
}


/**
   Printing a ternary tree :  opdepd operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opdepd ()
{
  this_intePrinter->printgluekeymark (kwbouv, vholoph - actindent);
  sigarglesign (1);
  this_intePrinter->printspalphakey (kwminuminusupe);
  sigarglesign (2);
  this_intePrinter->printgluekeyrelease (kwbfer, vholoph - actindent);
  /* marktalphakeyunparse(kwwhen, vholoph - 1, 3); */
  if (!marktalphakeyunparse (kwwhen, vholoph, 3))
    this_intePrinter->printerror ();
}

/**
   Printing a ternary tree :  openum operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_openum ()
{
  binsigexp (kwpoinpoin, 0);
  spalphakeSigargunp (kwstep, 3);	// if(ischildprint(3)) { this_intePrinter->printspalphakey(kwstep); sigargunp(); }
}

/**
   Printing a ternary tree :  opfen operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opfen ()
{
  binalphasigexp (kwwindow, 0);
  spalphakeSigargunp (kwinit, 3);	// if(ischildprint(3)) { this_intePrinter->printspalphakey(kwinit); sigargunp(); }
}

/**
   Printing a ternary tree :  opfunc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opfunc ()
{
  this_intePrinter->printalphakeysp (kwfunction);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opfuncp operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opfuncp ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwfunction);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opnode operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opnode ()
{
  this_intePrinter->printalphakeysp (kwnode);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opnodep operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opnodep ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwnode);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opoper operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opoper ()
{
  this_intePrinter->printalphakeysp (kwoperator);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opprocedure operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opprocedure ()
{
  this_intePrinter->printalphakeysp (kwprocedure);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opprocedurep operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opprocedurep ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwprocedure);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opprocess operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opprocess ()
{
  this_intePrinter->printalphakeysp (kwprocess);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opprocessp operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opprocessp ()
{
  this_intePrinter->printalphakeysp (kwprivate);
  this_intePrinter->printkeysp (kwprocess);
  processUnparsing ();
}

/**
   Printing a ternary tree :  opinte operator.(interface)
*/
_VIRTUAL_ void
SignalPrinter::decta3_opinte ()
{
  if (ischildprint (1))		// parameters
    {
      markkeyunparsekey (kwbouv, kwbfer, vholoph);
      wline (vholoph - actindent);
    }
  recsonunparse (2);		// inputs/output
  if (ischildprint (3))		// external graph
    {
      wline (vholoph - actindent);
      unparse_subtree (operande, vholoph);
    }
}

/**
   Printing a ternary tree :  opiter operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opiter ()
{
  this_intePrinter->printalphakeysp (kwiterate);
  recsondecunparse (1);

  wline (vholoph - actindent);
  this_intePrinter->printalphakeysp (kwof);
  if (ischildprint (2))
    markunparse (vholoph - 1);
  else
    this_intePrinter->printerror ();
  if (ischildprint (3))
    {
      wline (vholoph - actindent);
      markkeyunparse (kwwith, vholoph - 1);
    }
  this_intePrinter->printlkeysp (kwend, vholoph - actindent);
}

/**
   Printing a ternary tree :  opmemo operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opmemo ()
{
  binalphasigexp (kwcell, 0);
  spalphakeSigargunp (kwinit, 3);	// if(ischildprint(3)) { this_intePrinter->printspalphakey(kwinit); sigargunp(); }
}

/**
   Printing a ternary tree :  opmoti operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opmoti ()
{
  this_intePrinter->printalphakeysp (kwarray);
  recsondecunparse (1);

  wline (vholoph - actindent);
  this_intePrinter->printalphakeysp (kwof);
  if (ischildprint (2))
    markunparse (vholoph - 1);
  else
    this_intePrinter->printerror ();
  if (ischildprint (3))
    {
      wline (vholoph - actindent);
      markkeyunparse (kwwith, vholoph - 1);
    }
  this_intePrinter->printlkeysp (kwend, vholoph - actindent);
}



/**
   Printing a ternary tree :  oppcas operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_oppcas ()
{
  this_intePrinter->printalphakeysp (kwcase);
  recsonunparse (1);

  this_intePrinter->printspkey (kwin);
  wline (vholoph - actindent);
  if (ischildprint (2))
    markunparse (vholoph - 1);

  else
    this_intePrinter->printerror ();
  if (ischildprint (3))
    {
      wline (vholoph - actindent);
      snmark (vholoph - actindent);
      this_intePrinter->printkeysp (kwelse);
      sigargprimprocunp ();
      snrelease (vholoph - actindent);
    }
  this_intePrinter->printlkeysp (kwend, vholoph - actindent);
}


/**
   Printing a ternary tree :  opprag operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opprag ()
{
  recsondecunparse (1);
  if (ischildprint (2))
    {
      snspace ();
      unparse_subtree (operande, vholoph - 1);
    }
  if (ischildprint (3))
    {
      snspace ();
      markunparse (vholoph - 1);
    }
}



/**
   Printing a ternary tree :  opret operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_opret ()
{
  if (ischildprint (1))
    sigargunp ();
  else
    this_intePrinter->printerror ();
  this_intePrinter->printkey (kwdoll);
  if (ischildprint (2))
    if (voperator != opretun)
      sigargunp ();
    else;
  else
    this_intePrinter->printerror ();
  spalphakeSigargunp (kwinit, 3);	// if(ischildprint(3)) { this_intePrinter->printspalphakey(kwinit); sigargunp(); }
}

/**
   Printing a ternary tree :  optype operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optype ()
{
  recsonunparse (1);
  if (!marktalphakeyunparse (kwequl, vholoph - 1, 2))
    this_intePrinter->printerror ();
  if (ischildprint (3))
    {
      this_intePrinter->printspalphakey (kwinit);
      unparse_subtree (operande, vholoph - 1);
    }
}

/**
   Printing a ternary tree :  optypeact operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optypeact ()
{
  this_intePrinter->printalphakeysp (kwaction);
  processtypeunp ();
}

/**
   Printing a ternary tree :  optypeautomaton operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optypeautomaton ()
{
  this_intePrinter->printalphakeysp (kwautomaton);
  processtypeunp ();
}

/**
   Printing a ternary tree :  optypefunc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optypefunc ()
{
  this_intePrinter->printalphakeysp (kwfunction);
  processtypeunp ();
}

/**
   Printing a ternary tree :  optypenode operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optypenode ()
{
  this_intePrinter->printalphakeysp (kwnode);
  processtypeunp ();
}

/**
   Printing a ternary tree :  optypeproc operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optypeproc ()
{
  this_intePrinter->printalphakeysp (kwprocess);
  processtypeunp ();
}

/**
   Printing a ternary tree :  optypeprocedure operator.
*/
_VIRTUAL_ void
SignalPrinter::decta3_optypeprocedure ()
{
  this_intePrinter->printalphakeysp (kwprocedure);
  processtypeunp ();
}


/** 
    Printing a ternary tree using globals(operande, voperator)
    It switches, to decta3_XX method, according to the value(XX) of  the current operator(voperator).
*/
_VIRTUAL_ void
SignalPrinter::decta3 ()
{
  switch ((int) voperator)
    {
    case opact:
      decta3_opact ();
      break;
    case opactp:
      decta3_opactp ();
      break;
    case opautomaton:
      decta3_opautomaton ();
      break;
    case opautomatonp:
      decta3_opautomatonp ();
      break;
    case opborn:
      decta3_opborn ();
      break;
    case opcond:
      decta3_opcond ();
      break;
    case opdepd:
      decta3_opdepd ();
      break;
    case openum:
      decta3_openum ();
      break;
    case opfen:
      decta3_opfen ();
      break;
    case opfunc:
      decta3_opfunc ();
      break;
    case opfuncp:
      decta3_opfuncp ();
      break;
    case opinte:
      decta3_opinte ();
      break;
    case opiter:
      decta3_opiter ();
      break;
    case opmemo:
      decta3_opmemo ();
      break;
    case opmoti:
      decta3_opmoti ();
      break;
    case opnode:
      decta3_opnode ();
      break;
    case opnodep:
      decta3_opnodep ();
      break;
    case opoper:
      decta3_opoper ();
      break;
    case oppcas:
      decta3_oppcas ();
      break;
    case opprag:
      decta3_opprag ();
      break;
    case opprocedure:
      decta3_opprocedure ();
      break;
    case opprocedurep:
      decta3_opprocedurep ();
      break;
    case opprocess:
      decta3_opprocess ();
      break;
    case opprocessp:
      decta3_opprocessp ();
      break;
    case opret:
      decta3_opret ();
      break;
    case optype:
      decta3_optype ();
      break;
    case optypeact:
      decta3_optypeact ();
      break;
    case optypeautomaton:
      decta3_optypeautomaton ();
      break;
    case optypefunc:
      decta3_optypefunc ();
      break;
    case optypenode:
      decta3_optypenode ();
      break;
    case optypeproc:
      decta3_optypeproc ();
      break;
    case optypeprocedure:
      decta3_optypeprocedure ();
      break;
    default:
      this_intePrinter->printerror ();
    }
}

_VIRTUAL_ void
SignalPrinter::prelude (ntrees p)
{
  if (viscomm)
    if (unparse_printableTreeAttribute (p, pk_false, vholoph - 1))
      separate ();
}


_VIRTUAL_ void
SignalPrinter::postlude_root (ntrees p)
{
  if (IS_NILTREE (nfather (p)) || (nfather (p) == p))
    this_intePrinter->printkeysp (kwpvir);
}


_VIRTUAL_ void
SignalPrinter::postlude (ntrees p)
{
  if (viscomm)
    if (unparse_printableTreeAttribute (p, pk_true, vholoph - 1))
      separate ();
  if (STree_isModel (p))
    postlude_root (p);
}


/**
   It prints a tree(p) if holoph is greater than 0, otherwise a specific string(\ref printholo() function)
   The following operations are 
   - the prefix comments are printed (\ref unparse_printableTreeAttribute function)
   - it recursively prints the tree(p) according to its arity:
   - \ref decta0() for a leaf
   - \ref decta1() for unary tree
   - \ref decta2() for binary tree
   - \ref decta3() for ternary tree
   - \ref declist() for list tree
   During this "unparsing", Signal syntatic sugar is added.
   - the postfix comments are printed(\ref unparse_printableTreeAttribute function).

   Some globals are managed in this function:
   - operande = p
   - voperator= noper(operande)

*/
_VIRTUAL_ void
SignalPrinter::unparse_subtree (ntrees p, int holoph)
{
  ntrees p_save;
  int holoph_save;

  // push
  holoph_save = vholoph;
  vholoph = holoph;
  p_save = treeunparseTree;
  treeunparseTree = p;

  if (vholoph > 0)
    {
      prelude (p);
      voperator = noper (treeunparseTree);
      operande = p;
      if (LL_is_listOp (voperator))
	declist ();
      else if (LL_is_ternaryOp (voperator))
	decta3 ();
      else if (LL_is_binaryOp (voperator))
	decta2 ();
      else if (LL_is_unaryOp (voperator))
	decta1 ();
      else if (LL_is_atomOp (voperator))
	decta0 ();
      else
	return;			/* PLG Internal error */
      postlude (p);

    }
  else
    {
      voperator = noper (treeunparseTree);
      this_intePrinter->printholo ();
    }
  // pop
  treeunparseTree = p_save;
  vholoph = holoph_save;
}



/**
   It prints a tree(p) with an holophrase if holoph=0.

   It pushes context information before calling \ref treeunparse function,
   then  context information is popped.
*/
_VIRTUAL_ void
SignalPrinter::unparse_tree (ntrees p, int holoph)
{
  boolean viscomm_save;
  int unarylevel_save;

  // Push
  unarylevel_save = unarylevel;
  unarylevel = 0;
  viscomm_save = viscomm;
  viscomm = this_intePrinter->snflagon ();

  if (IS_NILTREE (p))
    this_intePrinter->printerror ();
  else
    {
      if (holoph == -1)
	holoph = 10000;
      unparse_subtree (p, holoph);
    }
  holoph = actindent + 1;
  if (stdMode)
    wline (holoph - actindent);

  // Pop
  viscomm = viscomm_save;
  unarylevel = unarylevel_save;
}

/**
   It prints a tree(p) with an holophrase if holoph=0.
   The text is not terminated by a line feed.
*/
_VIRTUAL_ void
SignalPrinter::unparse_tree1 (ntrees p, int holoph)
{
  boolean s;
  s = stdMode;
  stdMode = pk_false;
  unparse_tree (p, holoph);
  stdMode = s;
}

/**
   Printer initialisations:
   - An \ref intePrinter object is associated with the printer, if it is not yet done.
   - A "printer mode" is defined in this interface. It allows to print or not the comments.
   
   For a graphical mode, the comments are not printed, and for a textual one, they are printed.
   By default, the "textual mode" is set. To change it, use SignalPrinter::setPrinterContext() method.
   @return "true" if the operation succeeds, "false" otherwise.
*/
_VIRTUAL_ boolean
SignalPrinter::initPrinter ()
{
  // printf("DEBUG entering SignalPrinter::initPrinter()\n"); fflush(stdout);
  if (notannounced)
    {
      notannounced = pk_false;
      // printf("DEBUG      in  SignalPrinter::initPrinter() 1-1\n"); fflush(stdout);
      this_intePrinter = new intePrinter (string2repstr (comcom), string2repstr (comcom),
					  string2repstr (comincom), string2repstr (comincomend),
					  string2repstr (errSymbol), string2repstr (cstSymbolHolo));
      // printf("DEBUG      in  SignalPrinter::initPrinter() 1-2\n"); fflush(stdout);
      if (!this_intePrinter->load_KWLangTable ("SIGNAL"))
	return pk_false;
    }
  // printf("DEBUG      in  SignalPrinter::initPrinter() 2\n"); fflush(stdout);
  inteprint_initialize ();
  // printf("DEBUG  exiting SignalPrinter::initPrinter()\n"); fflush(stdout);
  return pk_true;
}
