/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define SEPARATOR '&'
#define escapeSep '\\'
#define p_LIST  "lexp"
#define p_STRING "h"
#define p_TYPE "t"
#define p_BASETYPE "b"
#define p_PROGNAME  "m"
#define p_MODELNAME "n"
#define p_PARAM "p"
#define p_INPUT "i"
#define p_OUTPUT "o"
#define p_CLOCK "ck"
#define p_DIM "d"
#define p_NB_WAIT "|wait|"
#define p_WAIT "wait"
#define p_HOMO_CLOCK "hck"
#define p_DONE "done"
#define p_COUNTER "cpt"
#define p_SYMBOL "symbol"
/**
   Initializing associative table (string key word, token)
*/
  void
StrScanner::initKwTable ()
{
  kwTable[0].kw = string2repstr (p_STRING);
  kwTable[0].kwcode = tk_homo;
  kwTable[1].kw = string2repstr (p_HOMO_CLOCK);
  kwTable[1].kwcode = tk_homoClock;
  kwTable[2].kw = string2repstr (p_TYPE);
  kwTable[2].kwcode = tk_type;
  kwTable[3].kw = string2repstr (p_MODELNAME);
  kwTable[3].kwcode = tk_name;
  kwTable[4].kw = string2repstr (p_PARAM);
  kwTable[4].kwcode = tk_param;
  kwTable[5].kw = string2repstr (p_INPUT);
  kwTable[5].kwcode = tk_input;
  kwTable[6].kw = string2repstr (p_OUTPUT);
  kwTable[6].kwcode = tk_output;
  kwTable[7].kw = string2repstr (p_CLOCK);
  kwTable[7].kwcode = tk_clock;
  kwTable[8].kw = string2repstr (p_DIM);
  kwTable[8].kwcode = tk_dim;
  kwTable[9].kw = string2repstr (p_NB_WAIT);
  kwTable[9].kwcode = tk_cardwait;
  kwTable[10].kw = string2repstr (p_WAIT);
  kwTable[10].kwcode = tk_wait;
  kwTable[11].kw = string2repstr (p_DONE);
  kwTable[11].kwcode = tk_done;
  kwTable[12].kw = string2repstr (p_LIST);
  kwTable[12].kwcode = tk_list;
  kwTable[13].kw = string2repstr (p_BASETYPE);
  kwTable[13].kwcode = tk_baseType;
  kwTable[14].kw = string2repstr (p_COUNTER);
  kwTable[14].kwcode = tk_cpt;
  kwTable[15].kw = char2repstr (SEPARATOR);
  kwTable[15].kwcode = tk_adr;
  kwTable[16].kw = string2repstr (p_PROGNAME);
  kwTable[16].kwcode = tk_progName;
  kwTable[17].kw = string2repstr (p_SYMBOL);
  kwTable[17].kwcode = tk_symbol;
  // The last one must be equal to (StrScanner_NBKW -1)
}

/**
    @return false in case of memory allocation failure
   @return "true" if a token (vStrKw) is scanned, "false" otherwise.
*/
boolean
StrScanner::scankw (ptrepstr * vStrKw)
{
  char c;
  boolean stop, bunput, b;
  ptreg strRegister;
  int i;

  stop = pk_false;
  if (strregnew (&strRegister))
    {
      i = 0;
      bunput = pk_true;
      while ((!stop) && nextChar (&c))
	{
	  i++;
	  switch (c)
	    {
	    case escapeSep:
	      if (nextChar (&c))
		strregconccha (strRegister, c);
	      else
		stop = pk_true;
	      break;

	    case SEPARATOR:
	      // if (nextCharGlue(&c)) { if (c == SEPARATOR) currentToken = tk_adr; else unput(1);}
	      stop = pk_true;
	      break;

	    case ':':
	    case ',':
	    case '(':
	    case ')':
	    case '{':
	    case '}':
	      stop = pk_true;
	      break;
	    default:
	      strregconccha (strRegister, c);
	      if (i == 1)
		if ((c == 'i') || (c == 'o') || (c == 'p') || (c == 'm'))
		  {
		    stop = pk_true;
		    bunput = pk_false;
		  }
	    }
	}

      if (nregGet_usedSize (strRegister) != 0)
	{
	  if (bunput && stop)
	    unput (1);
	  strregremovelastchar (strRegister, ' ');
	  b = strregstore (strRegister, vStrKw);
	}
      else
	b = pk_false;
      strregfree (strRegister);
      return b;
    }
  /* Memory allocation error */
  return pk_false;
}

/**
   @return the token after '&' separator, tk_error if not found.
*/
ScanPragmaToken
StrScanner::getTokenAfterSEPARATOR ()
{
  ptrepstr vkw;
  int i;

  if (!scankw (&vkw))
    return tk_error;
  i = 0;
  while (i < StrScanner_NBKW)
    {
      if (streq (vkw, kwTable[i].kw))
	return kwTable[i].kwcode;
      i++;
    }
  return tk_error;
}

/**
   @return "true" if the end of the text to scanned is detected, "false" otherwise.
*/
boolean StrScanner::endOfText ()
{
  return (iCurrentParametrizedText > nbLines);
}

/**
   @return "true" if the end of the current line is detected, "false" otherwise.
*/
boolean StrScanner::endOfLine ()
{
  return (indexCurStr > lgCurStr);
}

/**
   @return 
   - "true" if it exists a new line to scan; the line is prepared for the scanner (\ref setStr()),
   - "false" othersie.
*/
boolean
StrScanner::nextLine ()
{
  iCurrentParametrizedText++;
  if (endOfText ())
    return pk_false;
  setStr (ExtractString (getPragmaTerm (CurrentParametrizedText, iCurrentParametrizedText)));
  return pk_true;
}


/**
   @return "true" if a next character (c) is found in the text to scan, "false" otherwise.
*/
boolean
StrScanner::nextCharGlue (char *c)
{
  if (endOfLine ())
    if (!nextLine ())
      return pk_false;
  *c = strgetchar (curStr, indexCurStr);
  indexCurStr++;
  if ((*c) != escapeSep)
    return pk_true;
  if (endOfLine ())
    if (!nextLine ())
      return pk_false;
  *c = strgetchar (curStr, indexCurStr);
  indexCurStr++;
  return pk_true;
}



/**
   str becomes the text to scan.
*/
void
StrScanner::setStr (ptrepstr str)
{
  curStr = str;
  indexCurStr = 1;
  lgCurStr = strsize (curStr);
}




/**
   local method of getInt().
   @return  
   - tk_end if theend of text is found,
   - tk_ident if the current word to scan is an int; the string representation of this int is stored.
   - tk_error otherwise.
*/
ScanPragmaToken
StrScanner::getIntToken ()
{
  char c;
  boolean stop, omode;
  ptreg strRegister;

  omode = setMode (pk_false);

  currentToken = tk_error;	// Default
  stop = pk_false;
  if (strregnew (&strRegister))
    {
      while ((!stop) && nextChar (&c))
	if (isDecimalDigit (c))
	  strregconccha (strRegister, c);
	else
	  stop = pk_true;

      if (nregGet_usedSize (strRegister) != 0)
	{
	  if (stop)
	    unput (1);
	  if (!strregstore (strRegister, &this_currentIdent))
	    {
	      strregfree (strRegister);
	      return tk_error;
	    }
	  currentToken = tk_ident;
	}
      else if (!stop)
	currentToken = tk_end;
      else
	unput (1);
      strregfree (strRegister);
      setMode (omode);
      return currentToken;
    }
  /* Memory allocation error */
  return tk_error;
}


/**
   @param vt (in) it is the representation of a pragma. The value of a pragma is a list (nbl parameter) of strings.
   the vstr parameter is the first value of this list.
   A null tree is allowed for scanning on the vstr string.
   @param nbl (in) number of children of vt. nbl must be equal to 0 or 1 when vt is the nil tree.
   @param vstr (in) the first value of the vt pragma, or the string to scan.
   @param b (in)  must be "true" to skip space during scanning.
*/
void
StrScanner::StrScannerIni (ntrees vt, int nbl, ptrepstr vstr, boolean b)
{
  CurrentParametrizedText = vt;
  nbLines = nbl;
  iCurrentParametrizedText = 1;
  setStr (vstr);
  currentToken = tk_error;	// Default
  modeSkipSpace = b;
}
