/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** Printing the external representation of a cycle (treecycle)
   in the file \<UserOutputDirectory\>/nomprogsig_cyc.SIG.
*/
extern boolean
SignalUnparseCycles (int graphi, ptident nomprogsig, ntrees treecycle)
{
  ntrees aTree;
  ptrepstr processName;
  ptrepstr filePathName;
  if (IS_NILTREE (treecycle))
    return pk_true;
  aTree = STree_mkSkeletonModel (createSuffIdent (nomprogsig, string2repstr (hocyclesuff)), opprocess);
  STree_setBody (aTree, treecycle);
  processName = STreeModelModule_getNameStr (aTree);
  if (pK_SignalPrinter->fileUnparse (aTree, xfilebuf, processName, 9999, ofile, &filePathName))
    {
      pK_messenger->globalMessageStr (infocycle, filePathName);
      return pk_true;
    }
  pK_messenger->globalMessageStr (ErrWritingFile, processName);
  return pk_false;
}


/** Printing Parser result  (file _LIS.SIG) */
extern boolean
SignalUnparseParserResult (ptident nomprogsig, ntrees processTree, ntrees paramTree, int infomess)
{
  ptident id;
  boolean b, cMode;
  ptrepstr filePathName;
  ptrepstr fileName;

  cMode = pK_SignalPrinter->setPrinterContext (pk_false);	// to see the errors in graphical context.
  b = pk_true;
  if (ISNOT_NILTREE (paramTree))
    {
      id = createSuffIdent (nomprogsig, string2repstr ("PARAM_LIS"));
      fileName = ident2repstr (id);
      if (pK_SignalPrinter->fileUnparse (paramTree, xfilebuf, fileName, 9999, ofile, &filePathName))
	pK_messenger->globalMessageStr (infomess, filePathName);
      else
	{
	  pK_messenger->globalMessageStr (ErrWritingFile, fileName);
	  b = pk_false;
	}
    }
  if (ISNOT_NILTREE (processTree))
    {
      id = createSuffIdent (nomprogsig, string2repstr ("LIS"));
      fileName = ident2repstr (id);
      if (pK_SignalPrinter->fileUnparse (processTree, xfilebuf, fileName, 9999, ofile, &filePathName))
	pK_messenger->globalMessageStr (infomess, filePathName);
      else
	{
	  pK_messenger->globalMessageStr (ErrWritingFile, fileName);
	  b = pk_false;
	}
    }
  pK_SignalPrinter->setPrinterContext (cMode);
  return b;
}

/** Printing Parser result (mpp) in pathName */
extern boolean
pathNameFormatUnparse (char *pathName, ntrees mpp)
{
  if (ISNOT_NILTREE (mpp))
    return ntrees_formatSave (pathName, mpp, (char *) "Signal", sysGetSignalVersion ());
  return pk_true;
}

/** Printing Parser result  (file .isf) */
extern boolean
fileFormatUnparse (ptrepstr nomprogsig, ntrees mpp)
{
  ptrepstr id;
  boolean b = pk_true;

  if (ISNOT_NILTREE (mpp))
    {
      if (sys_suffixedFilePathInUserOutputDir (nomprogsig, (char *) "isf", &id))
	{
	  char *pathName = repstr2string (id);
	  printf ("     - generate Intermediate Tree Format %s\n", pathName);	/* pkmess infolisisf */
	  return ntrees_formatSave (pathName, mpp, (char *) "Signal", sysGetSignalVersion ());
	}
      return pk_false;
    }
  printf ("     - generate SSME format WARNING %s is an empty Signal Tree\n", repstr2string (nomprogsig));	/* pkmess infolisisf */
  return pk_true;
}

extern boolean
SignalUnparseIsf (ptident nomprogsig, ntrees mpp)
{
  return fileFormatUnparse (ident2repstr (nomprogsig), mpp);
}

/** Printing a tree in the file \<UserOutputDirectory\>/FOO.SIG.
   FOO is given by vt tree.
*/
extern boolean
SignalUnparse (ntrees vt, int infomess)
{
  ptrepstr fileName;
  if (IS_NILTREE (vt)) return pk_true;
  fileName = STreeModelModule_getNameStr (vt);
  if (fileName == nilstr) fileName = string2repstr ("ANONYMOUS");
  return decsig (vt, fileName, infomess);
}

extern boolean
SignalUnparseFile (ntrees vt, ptrepstr fileName, int infomess)
{
  if (fileName == nilstr) return SignalUnparse (vt, infomess);
  if (IS_NILTREE (vt)) return pk_true;
  return decsig (vt, fileName, infomess);
}

/**
   @return the buffer used by the pretty printer.
*/
extern ptrepbuf
getPrinterBuffer ()
{
  return xfilebuf;
}

/**
   Entry point for the initialisation of the module.
   @param fb  (in) printing buffer.
   It will be done at the first call to the pretty printer.
*/
extern void
decsigini (ptrepbuf aBuffer)
{
  xfilebuf = aBuffer;
}
