/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return "true" if a tree (t) is annotated with a "COMMENT" (varb) in prefix (resp. postfix) position if bpre (resp. not bpre), "false" otherwise.
*/
extern boolean
STree_getPrintableTreeAttribute (ntrees aTree, boolean bpre, ntrees * varb)
{
  short i;
  if (bpre)
    i = TREE_COMMENTS_PRE_ATTRIBUTENUMBER;
  else
    i = TREE_COMMENTS_POST_ATTRIBUTENUMBER;
  return get_attr (i, aTree, varb);
}


/**
   It adds a "COMMENT" annotation (cstr) to  a tree (t) in prefix position if bpre, postfix position otherwise.
*/
extern void
STree_addStrComment (ntrees aTree, ptrepstr cstr, boolean bpre)
{
  STree_addPrintableTreeAttribute (aTree, LL_make_comment_atom (cstr), bpre);
}


/**
   It adds a "COMMENT" annotation (vattrtree) to  a tree (aTree) in prefix position if bpre, postfix position otherwise.
*/
extern void
STree_addPrintableTreeAttribute (ntrees aTree, ntrees vattrtree, boolean bpre)
{
  short numAttr;
  if (bpre)
    numAttr = TREE_COMMENTS_PRE_ATTRIBUTENUMBER;
  else
    numAttr = TREE_COMMENTS_POST_ATTRIBUTENUMBER;
  add_attr (numAttr, aTree, vattrtree);
}

/**
   It adds an annotation (cstr) to  a tree (t).
*/
extern void
STree_setVMAnnotationsStr (ntrees t, ptrepstr cstr)
{
  STree_setVMAnnotations (t, LL_make_comment_atom (cstr));
}


/**
   It adds an annotation (cstr) to  a tree (t).
*/
extern void
STree_setVMAnnotations (ntrees t, ntrees vattr)
{
  noperators op;
  ntrees t1;

  op = noper (t);
  if ((op == opdefsig) || (op == opdefpart) || (op == opdefpartdef))
    {
      t1 = nchild (2, t);
      if ((ISNOT_NILTREE (t1)) && (noper (t1) != opnil))
	t = t1;
    }

  // printf("DEBUG      in STree_setVMAnnotations, numattr is %i\n", TREE_VM_ATTRIBUTENUMBER);
  add_attr (TREE_VM_ATTRIBUTENUMBER, t, vattr);	// for traceability.
}


/**
   @return the list (l) in which the annotated sub-trees of a tree (t) are added.
   An element (e) of the list is a leaf with an int value. This int gives the address of an annotated node.
   The function nintof(e) gives the int value.
*/
extern ntrees
STree_getVMAnnotations (ntrees t, ntrees l)
{
  int i, n;

  if (ISNOT_NILTREE (t))
    {
      if (nattrexist (t, TREE_VM_ATTRIBUTENUMBER))
	{
	  if (IS_NILTREE (l))
	    l = LL_make_anyList_emptyList ();
	  l = ntreeList_appendLast (l, LL_make_intVal_atom ((long int) t));
	}
      if (!LL_is_atomOp (noper (t)))
	{
	  n = ntreearity (t);
	  for (i = 1; i <= n; i++)
	    l = STree_getVMAnnotations (nchild (i, t), l);
	}
    }
  return l;
}

/**
   @return "true" if a  tree(t) is annotated with an error/warning, "false" otherwise.
*/
extern boolean
STree_isVMAnnotated (ntrees t)
{
  return nattrexist (t, TREE_VM_ATTRIBUTENUMBER);
}

/**
   @return the annotation (error/warning) associated with a tree(t) or NILTREE when it does not exist.
   This annotation, when it exists, is a list of comments.
*/
extern ntrees
STree_getVMAnnotation (ntrees t)
{
  ntrees val;
  if (!get_attr (TREE_VM_ATTRIBUTENUMBER, t, &val))
    return NILTREE;
  return val;
}

/**
   It resets the annotations added by the VM for the trees with t as root.
*/
extern void
STree_resetVMAnnotations (ntrees t)
{
  int i, n;

  if (ISNOT_NILTREE (t))
    {
      // if (nattrexist(t, TREE_VM_ATTRIBUTENUMBER))
      ndestroyattr (t, TREE_VM_ATTRIBUTENUMBER);
      if (!LL_is_atomOp (noper (t)))
	{
	  n = ntreearity (t);
	  for (i = 1; i <= n; i++)
	    STree_resetVMAnnotations (nchild (i, t));
	}
    }
}
