/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/**
   @return the identifier of an expression (t).
   Applied rules, for t 
     - x[]  -> STree_getIdentOf(x)
     - x[?] -> STree_getIdentOf(x)
     - x.y  -> STree_getIdentOf(x)
     - x\\E -> STree_getIdentOf(x)
     - t otherwise.
*/
extern ntrees
STree_getIdentOf (ntrees t)
{
  noperators op;

  while (pk_true)
    {
      op = noper (t);
      if ((op == opcham) || (op == opelem) || (op == oprecu) || (op == opsdec))
	t = nchild (1, t);
      else
	return t;
    }
}

/**
   @return the name of a model (m), NILTREE when m is not a model.
   For m a list of modules, it returns the name of the first module.
 */
extern ntrees
STreeModelModule_getName (ntrees m)
{
  noperators op;
  ntrees s;

  op = noper (m);
  if (op == oplmodu)
    {
      m = nchild (1, m);
      op = noper (m);
    }
  switch ((int) op)
    {
    case opact:
    case opnode:
    case opprocess:
    case opfunc:
    case opmodu:
    case opoper:
    case opautomaton:
    case opprocedure:

      // privates
    case opprocessp:
    case opfuncp:
    case opnodep:
    case opactp:
    case opautomatonp:
    case opprocedurep:

      return nchild (1, m);
    }
  return NILTREE;
}


/**
   @return the name of a model (m).
*/
extern ptrepstr
STreeModelModule_getNameStr (ntrees m)
{
  ptident id;

  id = STreeModelModule_getName (m);
  if (IS_NILTREE (id))
    return nilstr;
  return ident2repstr (id);
}



/** Getting arity.
    @return the arity of the SIGNAL voperator operator.
 */
extern int
STree_getOperatorArity (noperators voperator)
{
  // out of bounds --> -1
  if (LL_is_listOp (voperator))
    return 4;
  if (LL_is_ternaryOp (voperator))
    return 3;
  if (LL_is_binaryOp (voperator))
    return 2;
  if (LL_is_unaryOp (voperator))
    return 1;
  if (LL_is_atomOp (voperator))
    return 0;
  return -1;			// liste
}


/** Getting root of a tree. 
    @return true when t is not a nil object, false otherwise.
    @param t  (in) a tree.
    @param r (out) the root of the tree.
*/
extern boolean
STree_getRootTree (ntrees t, ntrees * r)
{
  ntrees v;

  if (IS_NILTREE (t))
    return pk_false;
  while (pk_true)
    {
      v = nfather (t);
      if (IS_NILTREE (v))
	{
	  *r = t;
	  return pk_true;
	}
      t = v;
    }
}

/** 
   Getting name of a model for a call.
   @param p_appel (in) a call object (assumed).
*/
extern ntrees
STree_getNameCalledModel (ntrees p_appel)
{
  noperators op;

  op = noper (p_appel);
  if (op == opappe)
    {
      p_appel = nchild (1, p_appel);
      op = noper (p_appel);
    }
  if ((op == opnom) || (op == opident) || (op == opop))
    return p_appel;
  return nchild (1, p_appel);
}

/**
   Getting the definition of a signal (id) in the body of a process (p_proc)
   @return "true" if it is found, "false" otherwise.
   @param vDef (out) reference to the definition (the first one for a partial definition)
   @param name (in) an identifier (assumed).
   @param p_proc (in) a model (assumed).
*/
extern boolean
STreeModel_getDefinitionOf (ptident id, ntrees p_proc, ntrees * vDef)
{
  int ope;
  boolean fini;
  ntrees p, p_expr, p_sav;

  p_expr = STreeModel_getBody (p_proc);
  if (IS_NILTREE (p_expr))
    return pk_false;
  fini = pk_false;
  while (!fini)
    {
      while (ISNOT_NILTREE (p_expr))
	{
	  p_sav = p_expr;
	  ope = noper (p_expr);
	  if (LL_is_atomOp (ope))
	    IS_NILTREE (p_expr);
	  else
	    {
	      switch (ope)
		{
		case opdefsig:
		  p = nchild (1, p_expr);
		  if (eq_idents (id, p))
		    {
		      (*vDef) = p_sav;
		      return pk_true;
		    }
		  p_expr = ntrees_getNext (p_sav);
		  break;
		case opmoti:
		  p_expr = nchild (3, p_expr);
		  break;
		case opappe:
		case opinst:
		  p_expr = ntrees_getNext (p_sav);
		  break;
		case oplfeed:
		  p_expr = nchild (1, p_expr);
		  break;
		default:
		  p_expr = ntrees_getNext (p_sav);
		  break;
		}
	    }
	}
      p_expr = ntrees_getNext (p_sav);
      while ((IS_NILTREE (p_expr)) && (p_sav != p_proc))
	{
	  p_sav = nfather (p_sav);
	  p_expr = ntrees_getNext (p_sav);
	}
      if (p_sav == p_proc)
	fini = pk_true;
      else
	fini = (STree_isLocalDeclList (p_expr) || STree_isBodyTree (p_expr));
    }
  return pk_false;
}



/**
  Getting a call in a body of a process.
  @return true if a call to a model (name) is found in the body of p_proc model, false otherwise.
  @param aCall (out) reference to the (first?) calling tree.
  @param name (in) an identifier (assumed).
  @param p_proc (in) a model (assumed).
*/
extern boolean
STreeModel_getCallToIn (ptident name, ntrees p_proc, ntrees * aCall)
{
  int ope;
  boolean fini, found;
  ntrees p_expr, p_sav;
  ptident id;

  p_expr = STreeModel_getBody (p_proc);
  if (IS_NILTREE (p_expr))
    return pk_false;
  fini = pk_false;
  while (!fini)
    {
      found = pk_false;
      while (ISNOT_NILTREE (p_expr))
	{
	  p_sav = p_expr;
	  ope = noper (p_expr);
	  if (LL_is_atomOp (ope))
	    SET_NILTREE (p_expr);
	  else
	    {
	      switch (ope)
		{
		case opdefsig:
		  p_expr = nchild (2, p_expr);
		  break;
		case opmoti:
		  p_expr = nchild (3, p_expr);
		  break;
		case opappe:
		case opinst:
		  id = STree_getNameCalledModel (p_expr);
		  if (eq_idents (id, name))
		    {
		      *aCall = p_expr;
		      return pk_true;
		    }
		  else
		    p_expr = ntrees_getNext (p_sav);
		  break;
		default:
		  p_expr = nchild (1, p_expr);
		  break;
		}
	    }
	}
      if (!found)
	{
	  p_expr = ntrees_getNext (p_sav);
	  while (IS_NILTREE (p_expr))
	    {
	      p_sav = nfather (p_sav);
	      p_expr = ntrees_getNext (p_sav);
	    }
	  fini = (STree_isLocalDeclList (p_expr) || STree_isBodyTree (p_expr));
	}
      else
	fini = pk_true;
    }
  return found;
}



/**
   @return the model in which the operand tree (p) is, NILTREE if such object does not exist.
*/
extern ntrees
STree_getModelInWhich (ntrees p)
{
  p = nfather (p);
  while (ISNOT_NILTREE (p) && !STree_isModel (p))
    p = nfather (p);
  return (p);
}


/**
  @return the father trees of a process (p). 
  If the process is directly under a module, it returns the module.
  If such object does not exist, it returns NILTREE.
*/
extern ntrees
STree_getModelFather (ntrees p)
{
  ntrees father, res;

  SET_NILTREE (res);
  if (ISNOT_NILTREE ((father = nfather (p))))
    {
      if (ISNOT_NILTREE ((father = nfather (father))))
	{
	  if (ISNOT_NILTREE ((father = nfather (father))))
	    {
	      if (noper (father) == opmodu)
		res = father;
	      else
		res = nfather (father);
	    }
	}
    }
  return (res);
}

/**
  @return the local declarations of a model (p).
  Assumed: p is a model.
*/
extern ntrees
STreeModel_getLocalDeclarations (ntrees p)
{
  ntrees x;

  if (noper (p) == opmodu)
    {
      x = nchild (2, p);
      if (IS_NILTREE (x))
	return NILTREE;
      return nchild (2, x);
    }
  p = nchild (3, p);
  if (noper (p) == opdesc)
    p = nchild (2, p);
  if (IS_NILTREE (p))
    return NILTREE;
  if (noper (p) == oppconf)
    p = nchild (2, p);
  if (IS_NILTREE (p))
    return NILTREE;
  if (noper (p) == opldecl)
    return p;
  return NILTREE;
}

/**
   @return the index-th local model of a process (aModel) if it exists, niltree otherwise.
*/
extern ntrees
STree_getIthModel (ntrees aModel, int index)
{
  ntrees lDecl, child;
  int i, nbProc;
  boolean notFound;

  lDecl = STreeModel_getLocalDeclarations (aModel);
  notFound = pk_true;
  if (ISNOT_NILTREE (lDecl))
    if (index <= ntreearity (lDecl))
      {
	i = 1;
	nbProc = 0;
	while ((i <= ntreearity (lDecl)) && notFound)
	  {
	    if (STree_isModel (nchild (i, lDecl)))
	      nbProc++;
	    notFound = !(nbProc == index);
	    i++;
	  }
      }
  if (notFound)
    SET_NILTREE (child);
  else
    child = nchild (i - 1, lDecl);
  return (child);
}

/**
   @return the parameters of a model (p).
   assumed: p is a model.
*/
extern ntrees
STree_getParameters (ntrees p)
{
  ntrees p_inte;

  if (!STree_isModel (p))
    return NILTREE;
  p_inte = nchild (2, p);
  if (ISNOT_NILTREE (p_inte))
    return (nchild (1, p_inte));
  return (NILTREE);
}


/**
   @return the numbers of parameters of a model (p).
   assumed: p is a model.
*/
extern int
STree_getNbParameters (ntrees p)
{
  ntrees x;

  x = STree_getParameters (p);
  return (get_nb_elem_decl (x));
}

/**
   @return the i-th parameter of a model (p).
   assumed: p is a model.
*/
extern boolean
STree_getIthParameter (ntrees p, int rank, ntrees * s)
{
  return getIthDecl (STree_getParameters (p), rank, s);
}

/**
   @return the interface of a model (p).
   assumed: p is a model.
*/
extern ntrees
STree_getInterface (ntrees p)
{
  return (nchild (2, p));
}

/**
   @return the tree composed of the inputs/outputs of a model (p).
   assumed: p is a model.
*/
extern ntrees
STree_getInputsOutputs (ntrees p)
{
  ntrees p_inte;

  p_inte = STree_getInterface (p);
  if (ISNOT_NILTREE (p_inte))
    return (nchild (2, p_inte));
  return (NILTREE);
}


/**
   @return the list of the outputs of a model (p).
   assumed: p is a model.
*/
extern ntrees
STree_getOutputs (ntrees p)
{
  ntrees p_inte;

  p_inte = STree_getInterface (p);
  if (ISNOT_NILTREE (p_inte))
    return (nchild (2, nchild (2, p_inte)));
  return (NILTREE);
}


/**
   @return the number of outputs of a model (p).
   assumed: p is a model.
*/
extern int
STree_getNbOutputs (ntrees p)
{
  ntrees x;

  x = STree_getOutputs (p);
  return (get_nb_elem_decl (x));
}


/**
   @return the i-th output of a model (p).
   assumed: p is a model.
*/
extern boolean
STree_getIthOutput (ntrees p, int rank, ntrees * s)
{
  return getIthDecl (STree_getOutputs (p), rank, s);
}


/**
   @return the list of the inputs of a model (p).
   assumed: p is a model.
*/
extern ntrees
STree_getInputs (ntrees p)
{
  ntrees p_inte;

  p_inte = STree_getInterface (p);
  if (IS_NILTREE (p_inte))
    return (NILTREE);
  return nchild (1, nchild (2, p_inte));

}


/**
   @return the i-th input of a model (p).
   assumed: p is a model.
*/
extern boolean
STree_getIthInput (ntrees p, int rank, ntrees * s)
{
  return getIthDecl (STree_getInputs (p), rank, s);
}

/**
   @return the number of inputs of a model (p).
   assumed: p is a model.
*/
extern int
STree_getNbInputs (ntrees p)
{
  ntrees x;

  x = STree_getInputs (p);
  return get_nb_elem_decl (x);
}


/**
   @return the external graph of a model (p).
   assumed: p is a model.
*/
extern ntrees
STree_getSpec (ntrees p)
{
  ntrees p_inte;

  p_inte = nchild (2, p);
  if (IS_NILTREE (p_inte))
    return (NILTREE);
  return nchild (3, p_inte);
}

/**
   @return 
   - "true" if a model (p) is characterized by a set of attributes (lattr) (safe, unsafe, deterministic).
   - "false" otherwise, lattr is not modified.
*/
extern boolean
STree_getAttributeSpec (ntrees p, ntrees * lattr)
{
  ntrees vt;

  vt = STree_getSpec (p);
  if (IS_NILTREE (vt))
    return pk_false;
  return STree_getSpecAttributes (vt, lattr);
}

/**
   @return 
   - "true" if a model (p) is characterized by a set of attributes (leq) (safe, unsafe, deterministic).
   - "false" otherwise, lattr is not modified.
*/
extern boolean
STree_getEquationsSpec (ntrees p, ntrees * leq)
{
  ntrees vt;

  vt = STree_getSpec (p);
  if (IS_NILTREE (vt))
    return pk_false;
  return STree_getSpecEquations (vt, leq);
}

/**
   @return "true" if a specification (vt) has a not empty list of attributes (lattr), "false otherwise.
*/
boolean
STree_getSpecAttributes (ntrees vt, ntrees * lattr)
{
  if (noper (vt) != opspec)
    return pk_false;
  vt = nchild (1, vt);
  if (IS_NILTREE (vt))
    return pk_false;
  if (noper (vt) == opnil)
    return pk_false;
  *lattr = vt;
  return pk_true;
}

/**
   @return "true" if a specification (vt) has a not empty list of equations (leq), "false otherwise.
*/
boolean
STree_getSpecEquations (ntrees vt, ntrees * leq)
{
  if (noper (vt) != opspec)
    return pk_false;
  vt = nchild (2, vt);
  if (IS_NILTREE (vt))
    return pk_false;
  if (noper (vt) == opnil)
    return pk_false;
  *leq = vt;
  return pk_true;
}

/**
   @return "true" if a model (p) is annotated with a pragma named vname; the first one is returned , "false" otherwise.
*/
extern boolean
STree_getPragmaNamed (ntrees p, ptrepstr vname, ntrees * vp)
{
  ntrees lp, ap, obj, stmt;
  ptrepstr id;
  int i, n;

  lp = STree_getPragmas (p);
  n = ntreearity (lp);
  for (i = 1; i <= n; i++)
    {
      ap = nchild (i, lp);
      STree_getPragmaValues (ap, &id, &obj, &stmt);
      if (streq (id, vname))
	{
	  *vp = ap;
	  return pk_true;
	}
    }
  return pk_false;
}

/**
   Getting the definition of a pragma (aPragma).
   \bsg   <pragma> ::=<identificateur> <o_l_objets_pragma> <enonce_pragma> \esg
   @param aPragma  (in) assumed to be a pragma.
   @param name  (out) an identifier
   @param p_obj  (out) the optional list of objects associated wit hthe pragma.
   @param enonce  (out) the value of the pragma (enonce_pragma)
*/
extern void
STree_getPragmaValues (ntrees aPragma, ptrepstr * name, ntrees * p_obj, ntrees * enonce)
{
  (*name) = STree_getIdentStr (nchild (1, aPragma));
  (*p_obj) = nchild (2, aPragma);	// not process
  (*enonce) = nchild (3, aPragma);
}

/**
   @return the name of the signal/constant from a declaration (decl).
   For a list of declarations, it takes the first element of the list.
*/
extern ntrees
STree_getNameOfDeclaration (ntrees decl)
{
  if (noper (decl) == opcons)
    return nchild (1, decl);
  // For a list: the first element.
  if (noper (decl) == opldecs)
    decl = nchild (1, decl);
  if (noper (decl) == opdecsig)
    decl = nchild (1, nchild (2, decl));
  if (noper (decl) == opinit)
    return nchild (1, decl);
  return decl;
}

/**
   @return the declaration of the signal/constant without its type from a declaration (decl).
   For a list of declarations, it takes the first element of the list.
*/
extern ntrees
STree_getDeclarationWithoutType (ntrees decl)
{
  if (noper (decl) == opldecs)
    decl = nchild (1, decl);
  if (noper (decl) == opdecsig)
    decl = nchild (1, nchild (2, decl));
  return decl;
}


/**
   @return the expression of the initial value associated with a signal declaration  (decl) 
   or niltree otherwise. 
*/
extern ntrees
STree_getInitExpr (ntrees decl)
{
  if (noper (decl) == opinit)
    return nchild (2, decl);
  return NILTREE;
}


/**
   @return the string associated with a signal (p) when it is a identifier, or an access to an element of an array
   or an access to a field of a structure, or nilstr otherwise.

   - For X,  it returns X.
   - For X.Y, it returns X.
   - For X[ ], it returns X;
*/
extern ptrepstr
STree_getIdentStr (ntrees p)
{
  switch (noper (p))
    {
    case opnom:
    case opident:
      return ident2repstr (p);
    case opelem:
    case opcham:
      return STree_getIdentStr (nchild (1, p));
    }
  return nilstr;
}


/**
  @return the operator of type for a declaration (p_decl) of a signal/constant or a
  list declarations (in this case, it takes the first element of the list).
  
  @param p_decl  (in) an abstract tree ; checked to be a declaration or a list of declarations of signals/constants.
  @param arb_typ (out) the type represented as an abstract tree.
  
*/
extern noperators
STreeDecl_getType (ntrees p_decl, ntrees * arb_typ)
{
  noperators ope, ope_typ;
  ntrees p;

  ope = noper (p_decl);
  if ((ope == opdecsig) || (ope == opdeccon) || (ope == opdecconp) || (ope == opdecvar) || (ope == opdecpar))
    {
      *arb_typ = nchild (1, p_decl);
      ope_typ = noper (*arb_typ);
    }
  else if (ope == opldecs)
    {
      p = nchild (1, p_decl);
      *arb_typ = nchild (1, p);
      ope_typ = noper (*arb_typ);
    }
  else
    {
      SET_NILTREE (*arb_typ);
      ope_typ = opimplicit;
    }
  return (ope_typ);
}

/** @return the tree that represents the implicit type.
 */
extern ntrees
STree_getImplicitTypeTree (void)
{
  return LL_make_implicit_atom ();
}

/** @return the NIL tree representation.
 */
extern ntrees
STree_getNilTree (void)
{
  return NILTREE;
}

/** @return the description of a model (p) (process, action, node,....). It is given by the rules:
   \bsg
   <decl_processus> ::= "process" <identificateur> "=" <def_interface> <description> ;
   <description> ::= <o_l_pragmas> <o_corps> ;
   \esg
*/
extern ntrees
STree_getDescription (ntrees p)
{
  return nchild (3, p);
}


/**
   @return the pragmas of a model (p).
   assumed: p is a model or a module.
*/
extern ntrees
STree_getPragmas (ntrees p)
{
  if (noper (p) == opmodu)
    return (nchild (1, nchild (2, p)));
  return (nchild (1, nchild (3, p)));
}



/**
   @return the expression of process (body) of a model (m)
   Assumed: m is a model.
*/
extern ntrees
STreeModel_getBody (ntrees m)
{
  ntrees p;

  if (IS_NILTREE (m))
    return NILTREE;

  p = nchild (3, m);
  if (IS_NILTREE (p))
    return NILTREE;

  if (noper (p) == opdesc)
    p = nchild (2, p);
  if (IS_NILTREE (p))
    return NILTREE;

  if (noper (p) == oppconf)
    p = nchild (1, p);
  return (p);
}




/** 
    @return the father of the tree of the declaration of signals according to \b where parameter.
    @param vtreeproc (in) a Signal model (assumed),
    @param where (in) 0 for local, 1 for parameter, 2 for inputs, 3  for outputs.
    @param iDecl (out) the index of the son of the returned tree to select to have the where-declaration area.
*/
extern ntrees
STree_getDeclaration (ntrees vtreeproc, int where, int *iDecl)
{
  ntrees vt, p_desc;
  boolean loc;

  loc = (where == 0);
  if (loc)
    {
      p_desc = nchild (3, vtreeproc);
      vt = nchild (2, p_desc);
      if ((IS_NILTREE (vt)) || (noper (vt) != oppconf))
	vt = STree_addLocalBlock (p_desc);
      *iDecl = 2;
    }
  else
    {
      vt = nchild (2, vtreeproc);
      if (IS_NILTREE (vt))
	vt = STree_addEmptyInterface (vtreeproc);
      *iDecl = where;
      if (where != 1)
	{			/* inputs/outputs */
	  vt = nchild (2, vt);
	  (*iDecl) = (*iDecl) - 1;
	}
    }
  return vt;
}


/** 
    @return the (first) string of "external" tree.
    \bsg <indication_externe> ::= "external" | "external" <chaine> \esg
 */
extern ptrepstr
STree_getExternalValue (ntrees t)
{
  if (IS_NILTREE (t))
    return nilstr;
  if (noper (t) != opexte)
    return nilstr;
  t = nchild (1, t);
  if (noper (t) == opnil)
    return nilstr;
  t = nchild (1, t);		// opchai: liste de chaines.
  return nstringof (t);
}


/**
   @return  "true" if the identifier (id) is found in the local declarations of a model (p) and the 
   declaration is returned (vdecl), "false" otherwise.
   The parameter op can be set for the selection:
   - opdecsig: id is a a signal
   - opdeccon:  id is a constant
   - optype: id is a type.
   - oplab: id is a label
   - opdecvar: id is a  state 
   - opdecpar: id is a shared
   - opnil for a search in all the local declarations of the model.
*/
extern boolean
STree_getDeclarationId (ptident id, ntrees p, noperators op, ntrees * vdecl)
{
  ntrees ldecl;

  // searching in inputs, outputs, then in locals.
  ldecl = STree_getInputs (p);
  if (STreeListDecl_getId (id, ldecl, op, vdecl))
    return pk_true;
  ldecl = STree_getOutputs (p);
  if (STreeListDecl_getId (id, ldecl, op, vdecl))
    return pk_true;
  ldecl = STreeModel_getLocalDeclarations (p);
  return STreeListDecl_getId (id, ldecl, op, vdecl);
}

/**
   @return the i-th module of a list of module (lm), or lm if it is not one.
*/
extern ntrees
STree_getModule (ntrees lm, int i)
{
  if (noper (lm) == oplmodu)
    return nchild (i, lm);
  return lm;
}


/**
   @return the child of a composition (vt) when it is reduced to one element, otherwise vt.
*/
extern ntrees
STree_getSimplifiedComposition (ntrees vt)
{
  if (IS_NILTREE (vt))
    return vt;
  if (noper (vt) != oplfeed)
    return vt;
  if (ntreearity (vt) == 1)
    return nchild (1, vt);
  return vt;
}
