/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.
   
   Contact: polychronycontact@listes.irisa.fr 
   
   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).
   
   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.
   
   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.
   
   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define OPIS_SIGNALIDENTIFIER(voperator) ( (voperator == opnom) || (voperator == opident) )
/**
   @return true if an operator (ope) is a Signal unary operator.
*/
extern boolean
STree_isUnaryExprProc (int ope)
{
  return (ope == opobli || ope == oppconf);
}

/**
   @return true if  an operator (ope) is a Signal process expression operator.
*/
extern boolean
STree_isExprProc (int ope)
{
  return (ope == opappe || ope == oplequhor || ope == opdefsig
	  || ope == opinst || ope == opobli || ope == oplfeed
	  || ope == opmoti || ope == opcond || ope == opdefpart || ope == oplequhor || ope == oplexchor || ope == oppconf || ope == opconv);
}

/**
   @return true if a tree (t) is the null object or an atom with opnil as operator.
*/
extern boolean
STree_isSignalNilTree (ntrees aTree)
{
  if (IS_NILTREE (aTree))
    return pk_true;
  return (noper (aTree) == opnil);	/* noper(p) returns OPUNDEF if p is NIL */
}

/**
   @return true if a tree (p) is a list object.
*/
extern boolean
STree_isList (ntrees aTree)
{
  if (IS_NILTREE (aTree))
    return pk_false;
  return nislist (aTree);
}

/**
   @return true if a tree (p) is a leaf object.
*/
extern boolean
STree_isLeaf (ntrees p)
{
  return (LL_is_atomOp (noper (p)));	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the Signal tree (p) is a function, false otherwise.
*/
extern boolean
STree_isFunction (ntrees p)
{
  return (noper (p) == opfunc);	/* noper(p) returns OPUNDEF if p is NIL */
}

/**
   @return true if the Signal tree (p) is a node, false otherwise.
*/
extern boolean
STree_isNode (ntrees p)
{
  return (noper (p) == opnode);	/* noper(p) returns OPUNDEF if p is NIL */
}

/**
   @return true if the Signal tree (p) is an action, false otherwise.
*/
extern boolean
STree_isAction (ntrees p)
{
  return (noper (p) == opact);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the Signal tree (p) is a module, false otherwise.
*/
extern boolean
STree_isModule (ntrees p)
{
  return (noper (p) == opmodu);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the Signal tree (p) is a list of modules, false otherwise.
*/
extern boolean
STree_isModulesList (ntrees p)
{
  return (noper (p) == oplmodu);	/* noper(p) returns OPUNDEF if p is NIL */
}

/** 
    @return true if an object (p) denotes a list of local declarations. 
*/
extern boolean
STree_isLocalDeclList (ntrees p)
{
  return (noper (p) == opldecl);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the operand tree (p) is an interface object.
*/
extern boolean
STree_isInterface (ntrees p)
{
  return (noper (p) == opinte);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the operand tree (p) is an external graph specification object.
*/
extern boolean
STree_isSpec (ntrees p)
{
  int ope;
  ntrees pere;

  ope = noper (p);		/* noper(p) returns OPUNDEF if p is NIL */
  if (ope == opspec)
    return pk_true;
  if (ope == opnil)
    {
      pere = nfather (p);	/* NIL TESTED */
      if (ISNOT_NILTREE (pere))
	return ((noper (pere) == opinte) && (nposit (p) == 3));	/* noper(p) returns OPUNDEF if p is NIL */
    }
  return pk_false;
}


/**
   @return true if the operand tree p represents an input/output specification object.
*/
extern boolean
STree_isInputOutput (ntrees p)
{
  return (noper (p) == opentsor);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the operand tree p represents a process model object.
*/
extern boolean
STree_isProcess (ntrees p)
{
  return ((noper (p) == opprocess));	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the operand tree p represents a model(process, node, function, action) object,
   false otherwise (including module, NILTREE)
*/
extern boolean
STree_isModel (ntrees p)
{
  switch ((int) noper (p))	/* noper(p) returns OPUNDEF if p is NIL */
    {
    case opprocess:
    case opfunc:
    case opnode:
    case opprocessp:
    case opfuncp:
    case opnodep:
    case opact:
    case opactp:
    case opprocedure:
    case opprocedurep:
    case opautomaton:
    case opautomatonp:
    case opoper:
      return pk_true;

    default:
      return pk_false;
    }
}

/**
   @return true if the operand tree p represents a private model(process, node, function, action) object,
   false otherwise.
*/
extern boolean
STree_isPrivateDeclaration (ntrees p)
{
  switch ((int) noper (p))	/* noper(p) returns OPUNDEF if p is NIL */
    {
    case opprocessp:
    case opfuncp:
    case opnodep:
    case opactp:
    case opprocedurep:
    case opautomatonp:
      return pk_true;

    default:
      return pk_false;
    }
}

/**
   @return true if the Signal tree (p) is an extern model, false otherwise including NILTREE
   An extern model has no expression and no local declarations.
*/
extern boolean
STree_isExternModel (ntrees p)
{
  ntrees p_exp, p_decl_loc;

  if (!STree_isModel (p))
    return (pk_false);		/* STree_isModel(p) returns false if p is NIL */
  p_exp = STreeModel_getBody (p);
  p_decl_loc = STreeModel_getLocalDeclarations (p);
  return (((IS_NILTREE (p_exp)) || (noper (p_exp) == opnil)) && (IS_NILTREE (p_decl_loc)));	/* noper(p) returns OPUNDEF if p is NIL */
}

/**
  @return true if p is the body of a model, false otherwise.
*/
extern boolean
STree_isBodyTree (ntrees p)
{
  int ope;
  ntrees pere;

  pere = nfather (p);		/* NIL TESTED */
  ope = noper (pere);		/* noper(p) returns OPUNDEF if p is NIL */
  if (ope == opdesc) return (pk_true);
  if (ope != oppconf) return (pk_false);		/* on peut avoir un oppconf à l'intérieur d'une expression */
  pere = nfather (pere);	/* NIL TESTED */
  return (noper (pere) == opdesc);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
  @return true if p is the name of a model, false otherwise.
*/
extern boolean
STree_isNameOfModel (ntrees p)
{
  int ope;
  ope = noper (p); /* noper(p) returns OPUNDEF if p is NIL */
  return (STree_isModel (nfather (p)) && OPIS_SIGNALIDENTIFIER (ope));	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
   @return true if the tree p is the parameter list declaration of a process, false otherwise.
*/
extern boolean
STree_isParametersOfModel (ntrees p)
{
  int ope;
  ope = noper (p);		/* noper(p) returns OPUNDEF if p is NIL */
  return ((STree_isInterface (nfather (p)) && ((ope == opldecs) || ope == opnil)));
}


/**
  @return true if the tree p is a list of effective parameters, false otherwise.
*/
extern boolean
STree_isEffectiveParameter (ntrees p)
{
  return (noper (p) == oplepara);	/* noper(p) returns OPUNDEF if p is NIL */
}


/**
  @return true if the tree p is a list of declaration of signals, false otherwise.
*/
extern boolean
STree_isSignalDecl (ntrees p)
{
  int ope;

  ope = noper (p);		/* noper(p) returns OPUNDEF if p is NIL */
  return ((ope == opdecsig) || (ope == opldecs));
}

/**
  @return true if the tree p is a list of declaration of signals, false otherwise.
*/
extern boolean
STree_isSharedSignalDecl (ntrees p)
{
  return (noper (p) == opdecpar);
}

/**
  @return true if the tree p is a list of declaration of constants, false otherwise. 
*/
extern boolean
STree_isConstantDecl (ntrees p)
{
  int ope;

  ope = noper (p);		/* noper(p) returns OPUNDEF if p is NIL */
  return ((ope == opdeccon) || (ope == opdecconp));
}

/**
  @return true if the tree p is a list of declaration of types, false otherwise. 
*/
extern boolean
STree_isTypeDecl (ntrees p)
{
  int ope;

  ope = noper (p);		/* noper(p) returns OPUNDEF if p is NIL */
  return ((ope == opltype) || (ope == opltypep));
}

/**
  @return true if there is no parameters associated to a model (proc).
  proc may be a model or a module; in this last case, returns always pk_true.
 */
extern boolean
STree_isModelWithoutParameters (ntrees proc)
{
  ntrees inte, lParam;

  if (!STree_isModel (proc))
    return (pk_true);
  inte = nchild (2, proc);
  if (ISNOT_NILTREE (inte))
    lParam = nchild (1, inte);
  else
    SET_NILTREE (lParam);
  return ((IS_NILTREE (lParam)) || (ISNOT_NILTREE (lParam) && noper (lParam) == opnil));	/* noper(p) returns OPUNDEF if p is NIL */
}



/**
  @return true if p is a list of pragmas or nil tree, false otherwise.
*/
extern boolean
STree_isPragmas (ntrees p)
{
  int ope;
  ope = noper (p);		/* noper(p) returns OPUNDEF if p is NIL */
  return ((ope == oplprag) || (ope == opnil));
}



/**
  @return true if the two trees p_nom1 and p_nom2 are same identifiers, false otherwise.
*/
extern boolean
STree_isIdentEq (ntrees p_nom1, ntrees p_nom2)
{
  ptrepstr str1, str2;

  str1 = ident2repstr (p_nom1);
  str2 = ident2repstr (p_nom2);
  return (str1 == str2);
}


 /**
   @return "true" if an identifier (declid) is in a list of declarations(l_liste)
 */
extern boolean
STree_isInDeclList (ntrees l_liste, ntrees declid)
{
  int i, j, nblp, nbp;
  boolean trouver;
  ntrees nom_to_ins, liste, idc, liste_princ;

  trouver = pk_false;
  if (ISNOT_NILTREE (l_liste) && ISNOT_NILTREE (declid))
    {
      nom_to_ins = STree_getNameOfDeclaration (declid);
      nblp = ntreearity (l_liste);
      if (noper (l_liste) != opnil)	/* noper(p) returns OPUNDEF if p is NIL */
	{
	  i = 1;
	  while ((i <= nblp) && !trouver)
	    {
	      liste_princ = nchild (i, l_liste);
	      if (STree_isSignalDecl (liste_princ))
		{
		  liste = nchild (2, liste_princ);
		  nbp = ntreearity (liste);
		  j = 1;
		  while ((j <= nbp) && !trouver)
		    {
		      idc = nchild (j, liste);
		      if (noper (idc) == opinit)	/* noper(p) returns OPUNDEF if p is NIL */
			trouver = STree_isIdentEq (nom_to_ins, nchild (1, idc));
		      else
			trouver = STree_isIdentEq (nom_to_ins, idc);
		      j++;
		    }
		}
	      i++;
	    }
	}
    }
  return (trouver);
}


/**
   @return true if the no-th element of the list of parameters l_param (assumed) is a type, false otherwise.
*/
extern boolean
STree_isParamAtype (ntrees l_param, int no)
{
  int i, nb, nb_elem, no_cherche;
  boolean b_res, pas_trouve;
  ntrees p_liste;

  nb = ntreearity (l_param);
  no_cherche = no;
  pas_trouve = pk_true;
  b_res = pk_false;
  i = 1;
  while (pas_trouve && i <= nb)
    {
      p_liste = nchild (i, l_param);
      nb_elem = ntreearity (p_liste);
      if (no_cherche <= nb_elem)
	{
	  b_res = (noper (nchild (no_cherche, p_liste)) == optype);	/* noper(p) returns OPUNDEF if p is NIL */
	  pas_trouve = pk_false;
	}
      else
	{
	  i++;
	  no_cherche = no_cherche - nb_elem;
	}
    }
  return (b_res);
}

/** 
    @return true if vt is undefined object (nil tree, tree with opnil operator or opnullproc operator), false otherwise.
*/
extern boolean
STree_isUndefinedTree (ntrees vt)
{
  if (IS_NILTREE (vt))
    return pk_true;
  return (noper (vt) == opnil);	/* noper(p) returns OPUNDEF if p is NIL */

}

/**
  @return true if the two trees pt1 (in) and pt2 (in) are equals. 
*/
extern boolean
STree_isTreeEq (ntrees pt1, ntrees pt2)
{
  return ntreesEq (pt1, pt2);
}


/**
  Searching a model in a list of models. 
  @return true if the model with p_nom as name exists in the list of models.
  @param p_nom (in) an identifier (assumed).
  @param list_proc (in) a list of declarations (models, signals,...) (assumed).
  @param p_decl  (out) the model of p_nom when it exists.
 */
extern boolean
STree_isModelInList (ntrees p_nom, ntrees list_proc, ntrees * p_decl)
{
  int i, nbproc;
  boolean trouver;
  ntrees proc;

  trouver = pk_false;
  if (ISNOT_NILTREE (list_proc))
    {
      nbproc = ntreearity (list_proc);
      i = 1;
      while (!trouver && (i <= nbproc))
	{
	  proc = nchild (i, list_proc);
	  if (STree_isModel (proc))
	    {
	      trouver = eq_idents (p_nom, STreeModelModule_getName (proc));
	      *p_decl = proc;
	    }
	  i++;
	}
    }
  return (trouver);
}


/**
   @return "true" if p is an operator model, "false" otherwise.
*/
extern boolean
STree_isOperatorModel (ntrees p)
{
  return (noper (p) == opoper);	/* noper(p) returns OPUNDEF if p is NIL */
}


/** 
   @return true if the tree (t) represents the implicit type object. 
*/
extern boolean
STree_isImplicitTypeTree (ntrees t)
{
  return (noper (t) == opimplicit);	/* noper(p) returns OPUNDEF if p is NIL */
}

/**
   @return "true" is elem is an element of a list (list), "false" otherwise.
   Assumed : list is a list.
*/
extern boolean
STree_isInList (ntrees list, ntrees elem)
{
  int j, n;

  n = ntreearity (list);
  for (j = 1; j <= n; j++)
    if (STree_isTreeEq (nchild (j, list), elem))
      return pk_true;
  return pk_false;
}

/**
   @return "true" if vt is an empty list (or nil tree), "false" otherwise.
*/
extern boolean
STree_isEmptyList (ntrees vt)
{
  if (IS_NILTREE (vt))
    return pk_true;
  if (!STree_isList (vt))
    return pk_false;
  return (ntreearity (vt) == 0);
}

/**
   @return "true" if the operand tree (vt) is a composition.
*/
boolean
STree_isComposition (ntrees vt)
{
  return (ISNOT_NILTREE (vt) && (noper (vt) == oplfeed));	/* noper(p) returns OPUNDEF if p is NIL */
}

/**
   @return "true" if the operand tree (vt) is a confined expression of processus (i.e (| ...|) where ...end).
*/
boolean
STree_isConfinedProcess (ntrees vt)
{
  return (ISNOT_NILTREE (vt) && (noper (vt) == oppconf));	/* noper(p) returns OPUNDEF if p is NIL */
}
