/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define OPTION_HEADER        '-'
#define SUBOPTION_HEADER     '--'
//#define char_colon    ':'
#define char_eq    '='

#define SIGNAL               "signal"

/* batch SIGNAL compilers Options*/
/*  ====> input file options */
#define INISF                "inisf" /* in is Internal Signal Format" */
#define COMMANDFILE          "com" /* setCommandFile */
/*      SOURCEFILE           is not tagged: setSourceFile */
#define PARAMREAD            "par" /* setParameterFile */

/* ====> output file options */
#define OUTISF               "outisf"  /* generate Internal Signal Format" */
#define OUTPUTROOT           "d"   /* setOutputDirectory */
#define OUTPUTROOTUC         "D"   /* setOutputDirectory */
#define WARNINGS             "war"   /* setopt_warn */
#define VERBOSE              "v"   /* setopt_verbose */
#define COSIMULATIONMODE     "tm"   /* setopt_traceability_mode */
#define KEEPTRACEREW         "kt"   /* setopt_traceability_mode */
#define deprecated_PRINTOPTIONS         "print" /* setopt_gen_print_when */

/* ====> graph class transformation options */
#define MAKEBASIC            "basic"    /* setopt_gen_basic; -basic is implicit option with all other options; without any option -basic --tra is implicit */
#define MAKEGENBASIC         "BASIC"    /* setopt_gen_basic; -basic is implicit option with all other options; without any option -basic --tra is implicit */
#define MAKEPOLY             "poly"     /* setopt_gen_poly; induces implicit -basic; without other options -poly implicitly means -poly --tra */
#define MAKEGENPOLY          "POLY"     /* setopt_gen_poly; induces implicit -basic; without other options -poly implicitly means -poly --tra */
#define deprecated_PROF                 "depth"    /* setopt_solvingDepth */
#define MAKEENDO             "endo"     /* setopt_gen_endo; induces implicit -poly; without other options -endo implicitly means -endo --tra  */
#define MAKEGENENDO          "ENDO"     /* setopt_gen_endo; induces implicit -poly; without other options -endo implicitly means -endo --tra  */
#define deprecated_MAKEBOOL  "bc"       /* DEPRECATED setopt_gen_bool; induces implicit -endo; without other options -bc implicitly means -bc --tra  */
#define MAKEBOOL             "bool"       /* setopt_gen_bool; induces implicit -endo; without other options -bc implicitly means -bc --tra  */
#define MAKEGENBOOL          "BOOL"       /* setopt_gen_bool; induces implicit -endo; without other options -bc implicitly means -bc --tra  */
#define deprecated_MAKESEQUENTIAL       "sd"       /* DEPRECATED setopt_gen_sequential; induces implicit -bc; without other options -sd implicitly means -sd --tra  */
#define MAKESEQUENTIAL       "seq"       /* setopt_gen_sequential; induces implicit -bc; without other options -sd implicitly means -sd --tra  */
#define MAKEGENSEQUENTIAL    "SEQ"       /* setopt_gen_sequential; induces implicit -bc; without other options -sd implicitly means -sd --tra  */
#define MAKEFLAT             "flat"       /* setopt_gen_flat; induces implicit -bc; without other options -fl implicitly means -fl --tra  */
#define MAKEGENFLAT          "FLAT"       /* setopt_gen_flat; induces implicit -bc; without other options -fl implicitly means -fl --tra  */
#define MAKEMONO             "mono"     /* setopt_gen_flat; induces implicit -bc; without other options -fl implicitly means -fl --tra  */
#define MAKEGENMONO          "MONO"     /* setopt_gen_flat; induces implicit -bc; without other options -fl implicitly means -fl --tra  */

/* ====> kind of output */
#define LIST                 "lis"   /* setopt_gen_list */
#define NOTRACE              "notra"   /* setopt_gen_trace */
#define TRACE                "tra"   /* setopt_gen_trace */
#define TRACEUC              "TRA"   /* setopt_gen_trace */
#define SME                  "sme"   /* setopt_sme */
#define SSME                 "ssme"   /* setopt_ssme */
#define GENEAUTO             "geneauto"   /* setopt_geneauto */
#define ABSTRACTION          "spec"   /* setopt_abstraction */
#define EXCLUSIVECLOCK       "ec"     /* setopt_ExclusiveClock */
#define MKMODULEHEADER       "header" /* setopt_header useless with tra alone*/

/* ====> graph options */
#define DRETIMING            "dr"   /* setopt_retiming */
/* #define LRETIMING            "-lr"  NOT IMPLEMENTED */
#define REWCOND              "crew"   /* setopt_crew_bool_ctrl, setopt_crew_definitions */
#define US                   "su"   /* setopt_us */
#define DEFAULT_SPLITTING    "ds"   /* setopt_default_splitting */
#define PARTIALDEF_MERGING   "pdg"   /* setopt_partialdef_merging */
#define SPLITDATA            "sdata"   /* */
#define ENUM2INT             "e2i"   /* */
#define SUBSTITUTE           "s"   /* setopt_substitute */
#define KERNEL               "kernel"   /* setopt_kernel*/
#define UTILITY_CLOCK_OPT    "uc"   /* setopt_utility_clock */
#define PROFILING            "profiling"   /* setopt_profiling */
#define POLYENDO             "polyendo"

/* ====> partition options */
#define SEP_STATES_OTHER     "sso"   /* setopt_separate_states_others */
#define SEP_BOOL_OTHER       "sbo"   /* setopt_separate_booleans_others */
#define SEP_FINITE_TYPES_OTHER "sfto"   /* setopt_separate_finiteTypes_others */
#define CLUSTERING             "clu"   /* setopt_clustering */
#define TAGBASEDPARTITION      "dist"     /* setopt_tagBased_partition -cluster + -threads */

/* ====> kind of exported file */
#define Z3Z                  "z3z"   /* setopt_gen_z3z */
#define Z3ZUC                "Z3Z"   /* setopt_gen_z3z */
#define SMT                 "smt"
#define SMTUC               "SMT"
#define CODELUSTRE           "lustre"   /* setopt_gen_lustre; not fully tested/implemented */
#define CODELUSTREUC         "LUSTRE"   /* setopt_gen_lustre; not fully tested/implemented */
#define CODEC                "c"   /* setopt_gen_c_modules */
#define CODECUC              "C"   /* setopt_gen_c_modules */
#define CODETHREADS          "threads"   /* setopt_clustering, setopt_gen_threads*/
#define CODECPP              "c++"   /* setopt_gen_cpp_modules */
#define CODECPPUC            "C++"   /* setopt_gen_cpp_modules */
#define CODEJAVA             "java"   /* setopt_gen_java_modules */
#define CODEJAVAUC           "JAVA"   /* setopt_gen_java_modules */
#define CODESYNDEX           "syndex"   /* setopt_gen_syndex */
#define CODESYNDEXUC         "SYNDEX"   /* setopt_gen_syndex */
#define DOTCODE              "dot"   /* setopt_gen_dot */


#define OPTFORCE             "force"   /* setopt_alwaysCode */
#define CHECK                "check"   /* setopt_check useless with tra alone */
#define UNDELAYMULTIPLEXING  "umd"   /* setopt_delayMultiplexing */
#define UNDELAYOPTIMIZING    "udo"   /* setopt_delayOptimizing */


/* ====> listing options */
#define STATISTICS           "stat"   /* setopt_stat */
#define DEBUG                "dbg"   /* setopt_debug */

/* ====> wizzard option */
#define TESTFUNC             "test"   /* setopt_TestFunctionality */
