/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* ===> Host interactions */

/**
   @return "true" if the full path name str exists, "false" otherwise.
*/
extern boolean
sysExistFile (ptrepstr str)
{
  char *A;
  int l, i;
  FILE *fg;

  l = strsize (str);
  if (l == 0)
    return pk_false;
  A = allocate_string (l + 1);
  for (i = 1; i <= l; i++)
    A[i - 1] = strgetchar (str, i);
  A[l] = nullChar;
  fg = fopen (A, "r");
  if (fg == NULL)
    return pk_false;
  fclose (fg);
  free (A);
  return pk_true;
}

/**
   It removes a file or a directory (fname).
*/
extern void
sysdelete (ptrepstr fname)
{
  char *s;
  s = repstr2string (fname);
  remove (s);			/* This is the ISO C function to remove a file, declared in stdio.h */
  free (s);
}

/**
   Making a directory with ud as name.
*/
extern boolean
sysmkdirectory (ptrepstr ud)
{
  char *command;
  int l, i, lt, lmkdir, retcode;

  if (strsize (ud) == 0)
    return pk_false;
  GETDOUBLEQUOTEDSTRING (ud);	// OS dependant.
  l = strsize (ud);
  lmkdir = strlen (CMD_MKDIR);
  lt = l + lmkdir;
  command = allocate_string (lt + 1);
  for (i = 1; i <= lt; i++)
    command[i - 1] = ' ';
  strcpy (command, CMD_MKDIR);
  for (i = 1; i <= l; i++)
    command[i + lmkdir - 1] = strgetchar (ud, i);
  command[lt] = nullChar;
  retcode = system (command);
  // printf(" code retour system : %d \n", retcode);
  return pk_true;
}

/* ===> Host path */
/**
   @return true if the operation succeeds.
   @param fname (in): a pK string 
   @param suff (in): a suffix.
   @param res (out):  the pK-string obtained by the catenating of fname and suff 
   using the system separator(s)
*/
extern boolean
sys_add_typeSuffix2repstr (ptrepstr fname, const char *suff, ptrepstr * res)
{
  if (strconcstrcha (fname, fileSuffixSeparator, res))
    return strconcstrString (*res, (char *) suff, pk_false, res);
  return pk_false;
}

/** Extracts the directory name from the name filename.
 */
extern char *
sys_extract_dirPath (char *filename, boolean keepSeparator)
{
  char *dirname;
  if (filename == (char *) 0)
    return filename;
  dirname = copyString (filename);
  if (strKillSuffix (dirname, fsystemSeparator, keepSeparator))
    return dirname;
  return (char *) 0;
}

/* ===> Polychrony tool set path */
/**
 * fileName is a file that should be in Polychrony binary directory
 * @returns the full path name of fileName suffixed by suff
 * in a new mallocated string
 * This function may fail
 */
extern char *
sys_suffixedFileInPolyBin (const char *fileName, const char *suff)
{
  char *res = _sys_suffixedFileInPolyBin (fileName, strlen (suff));
  res = strcat (res, suff);
  return res;
}

/**
 * requires signaldir correctly defined as the full name of Polychrony "bin" directory.
 * @returns the full path name of fileName in signaldir, suffixed by step
 * in a new mallocated string
 */
extern char *
sys_digitSuffixedFileInPolyBin (const char *fileName, int step)
{
  char *res = _sys_suffixedFileInPolyBin (fileName, 1);
  int prefixLength = strlen (res);
  res[prefixLength] = step + 48;
  res[prefixLength + 1] = NULL_char;
  return res;
}

/**
   @returns the full path name of CodeGenerator.defs file in the directory given by the 
   environment variable Signal_bin
   \note CodeGenerator.defs is a library of predefined code, assigned with some Signal operators (Matrices, vectors), 
   for C, C++ and Java code generators.
 */
extern char *
sysGetFileCodeGenPredef ()
{
  return (getFilePathName (signaldir, CodeGeneratorPredef));
}

/**
   It defines an environment variable (eVar) by an associated data (vVar).
   It can be used to avoid the access to the execution environment (shell variables).
*/
extern void
sysSetEnvDef (char *eVar, const char *vVar)
{
  char *vstr;

  vstr = copyString ((char *) vVar);
  setEnvDef (eVar, vstr);
  if (strcmp (eVar, SIGNAL_BIN_NAME) == 0)
    signaldir = vstr;
}

/* ===> User defined output path */

/** 
    Setting current output directory "userdirname" to ud.
    @return "false" on internal error, true otherwise
*/
extern boolean
sys_set_userOutputDir (ptrepstr ud)
{
  int ud_lg = strsize (ud);
  if (ud_lg == 0)
    ud = nilstr;
  else if (strgetchar (ud, strsize (ud)) != fsystemSeparator)
    if (!strconcstrcha (ud, fsystemSeparator, &ud))
      return pk_false;
  userdirname = ud;
  return pk_true;
}

/** 
    build the path of filename in pathString directory; return it in fullPath
 
    @return "false" on internal error, or if  fileName is empty
*/
extern boolean
sysMake_strPath_repstr_repstr (ptrepstr pathString, ptrepstr fileName, ptrepstr * fullPath)
{
  int string_lg = strsize (fileName);
  if (string_lg == 0) return pk_false;
  string_lg = strsize (pathString);
  if (string_lg == 0)
    {
      (*fullPath) = fileName;
      return pk_true;
    }
  if (strconcstrcha (pathString, fsystemSeparator, fullPath))
    return strconc2str (*fullPath, fileName, pk_false, fullPath);
  return pk_false;
}

/**
   It sets the current output directory to od.
   od is a correct and complete directory path
   \see sys_set_userOutputDir().
*/
extern void
sys_restore_userOutputDir (ptrepstr od)
{
  userdirname = od;
}

/**
   @return the value of the current output directory.
*/
extern ptrepstr
sys_get_userOutputDir ()
{
  return userdirname;
}

/** 
    Setting current output directory "userdirname" to ud.
    @return "false" on internal error, true otherwise
*/
extern boolean
sys_setInner_userOutputDir (ptrepstr ud)
{
  if (strsize (userdirname) == 0)
    return sys_set_userOutputDir (ud);
  if (strconc2str (userdirname, ud, pk_false, &ud))
    return sys_set_userOutputDir (ud);
  return pk_false;
}

/**
   generate path file by concatenating fname to current output directory (static variable userdirname)
   if it does not exist, creates the current output directory (static variable userdirname)
   @return "true" if the operation succeeds, "false" otherwise.
   @param fname (in) the name of a file.
   @param res (out) the generated path file
   
   See sys_set_userOutputDir()
*/
extern boolean
sys_suffixedFilePathInUserOutputDir (ptrepstr fname, const char *suffix, ptrepstr * res)
{
  ptrepstr suffixedName = string2repstr (getSuffixedFileName (repstr2string (fname), suffix));
  if (strsize (userdirname) == 0)
    {
      *res = suffixedName;
      return pk_true;
    }
  if (!sysExistFile (userdirname))
    if (!sysmkdirectory (userdirname))
      return pk_false;
  return strconc2str (userdirname, suffixedName, pk_false, res);
}

/**
   generate path file by concatenating fname to current output directory (static variable userdirname)
   if it does not exist, creates the current output directory (static variable userdirname)
   @return "true" if the operation succeeds, "false" otherwise.
   @param fname (in) the name of a file.
   @param res (out) the generated path file
   
   See sys_set_userOutputDir()
*/
extern boolean
sys_filePathInUserOutputDir (ptrepstr fname, ptrepstr * res)
{
  if (strsize (userdirname) == 0)
    {
      *res = fname;
      return pk_true;
    }
  if (!sysExistFile (userdirname))
    if (!sysmkdirectory (userdirname))
      return pk_false;
  return strconc2str (userdirname, fname, pk_false, res);
}


/**
   requires SIGNAL_LIBRARY_PATH correctly defined as the set of directories that contains Signal libraries.
   @return true if the extended fullPathName dirName/filename exists
   where dirName is a path extracted of the SIGNAL_LIBRARY_PATH string.
*/
extern boolean
sysGetDirnameLibrary (ptrepstr filename, ptrepstr * dirName, ptrepstr * fullPathName)
{
  char *path;
  int lg, pos;

  path = sysGetEnvDef (CSTLIB_ENV_NAME);
  if (path == NULL)
    return pk_false;
  lg = strlen (path);
  pos = 0;
  while (pos < lg)
    if (extractStrDirname (path, pathSeparator, fsystemSeparator, dirName, &pos))
      if (strconc2str (*dirName, filename, pk_false, fullPathName))
	if (sysExistFile (*fullPathName))
	  return pk_true;
  return pk_false;
}

/** 
    @return the value of the pK_VERSION environment shell variable.
    (this variable is managed in sigConfig file of Polychrony "bin" directory)
*/
extern char *
sysGetSignalVersion ()
{
  return (sysGetEnvDef (isdtversion));
}

/** 
    Assumes the variable exists in the environment.
    @return the INTEGER value of the environment shell variable called name.
 */
extern int
sysGetenvIntegerValue (const char *name)
{
  return (atoi (sysGetEnvDef (name)));	// TODO get_int_HostValue(name) ??
}

/**
   @return  Elapsed real time (in 100th of seconds). 
*/
extern double
sysgetcputime (void)
{
  return (double) (clock () / ((CLOCKS_PER_SEC) / 100));
}



/**
   Misc. Used for V4.13.10 to V4.14 translation.
   It deletes the phylum [OL_E_PARA] at the begining of the fileName file.
*/
extern boolean
sysUpdateParametersFile (char *fileName)
{

  FILE *pf;
  pf = (FILE *) fopen ((char *) fileName, "r+");
  if (pf == NULL)
    return pk_false;
  removeStringBeginOfFile (pf, "[OL_E_PARA]");
  fclose (pf);
  return pk_true;
}

/**
   @return the data associated with a key (vstr). This value is
   - the data previously set by a call to the method sysSetEnvDef(),
   - the data given by the execution environment: the key (vstr) is seen as a shell variable,
   - the null string otherwise.
*/
extern char *
sysGetEnvDef (const char *vstr)
{
  return getEnvDef ((char *) vstr);
}

/**
   It defines an environment variable (eVar) by an associated data (vVar) if it is not yet assigned.
*/
extern void
sysSetEnvDefDefault (const char *eVar, const char *vVar)
{
  char *val;

  if (!isEnvDef ((char *) eVar))
    {
      val = getEnvDef ((char *) eVar);
      if (val == NIL_String)
	sysSetEnvDef ((char *) eVar, vVar);
    }
}

/**
   It sets the "new line" mode to b.
   When b is pk_true a new line is sent in locwriteln(), otherwise it is not.
   it is useful for the code generators: for example, a string cannot be on several lines.
   @return the mode used before the call.
 */
extern boolean
setnlmode (boolean c)
{
  boolean x;
  x = nlmode;
  nlmode = c;
  return x;
}

/**
   Module initiliazing:
    - file registers;
    - getting environment informations: binary path of the compiler.
*/
extern boolean
sysini ()
{
  initFileSystemEnv ();
  signaldir = sysGetEnvDef (SIGNAL_BIN_NAME);
  if (signaldir == NIL_String)
    {
      printf (" Undefined Shell variable: Signal_bin");
      return pk_false;
    }
  filreg3.name = nilstr;
  filreg4.name = nilstr;
  filreg5.name = nilstr;
  filreg6.name = nilstr;
  userdirname = nilstr;
  nlmode = pk_true;
  return pk_true;
}

/**
   @return the name of a temporary file
*/
extern char *
sysGetnameTempfile ()
{
  char *name;
  OS_DEPENDANT_tmpnam (name);
  return name;
}


/**
   @return 
   - "true" if a file (fname) is found in the Library environment, cfname is the full pathname of fname,
   - "false" otherwise, cfname is not modified.
*/
extern boolean
existFile (char *fname, char * *cfname)
{
  ptrepstr fileName, dirname;

  if (!sysGetDirnameLibrary (string2repstr (fname), &dirname, &fileName))
    return pk_false;
  (*cfname) = repstr2string (fileName);
  return pk_true;
}


/**
   @return the name of the extension of a file (filename), nilstr by default.
*/
extern ptrepstr
sysGetFileExtension (char *filename)
{
  char *vextension;
  int i, j, lg, pos;
  char c;

  lg = strlen (filename);
  if (lg == 0)
    return nilstr;
  pos = lg;
  while (pos >= 0)
    {
      c = filename[pos];
      if (c == fsystemSeparator)
	return nilstr;
      if (c == fileSuffixSeparator)
	{
	  if (pos == lg)
	    return nilstr;
	  vextension = allocate_string (lg - pos + 1);
	  j = 0;
	  for (i = pos + 1; i <= lg; i++)
	    {
	      vextension[j] = filename[i];
	      j++;
	    }
	  vextension[lg - pos + 1] = nullChar;
	  return string2repstr (vextension);
	}
      pos--;
    }
  return nilstr;
}
