/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/**
   @return true if an attribute (num) exists in a list of attributes (list), "false" otherwise.
   The value (val) associated with the num-attribute is returned.
*/
extern boolean
ptlAttributes_get (ptlAttributes list, int num, attributeValue * val)
{
  ptAttribute curAttr;

  if (ptlAttributes_satisfies_nil (list))
    return pk_false;
  while (ptlAttributes_get_next (&list, &curAttr))
    if (ptAttribute_get_num (curAttr) == num)
      {
	*val = ptAttribute_get_val (curAttr);
	return pk_true;
      }
  return pk_false;
}

/**
   It inserts an attribute (num, valattr) in a list of attributes (lattr).
   Unicity is guaranteed by the ordinal (num). When it exists in the list the value is replaced by the new one (valattr).
   @return the new list.
*/
extern ptlAttributes
ptlAttributes_insert (ptlAttributes lattr, int num, attributeValue valattr)
{
  ptAttribute attr, curAttr;
  ptlAttributes curlist, succ;
  int curNum, cnum;

  attr = ptAttribute_RIEN_create (num, valattr);
  if (ptlAttributes_satisfies_nil (lattr))
    return ptlAttributes_RIEN_create (attr, NILLATTRIBUTES);

  curAttr = ptlAttributes_get_elem (lattr);
  curNum = ptAttribute_get_num (curAttr);
  if (num < curNum)
    return ptlAttributes_RIEN_create (attr, lattr);
  if (num == curNum)
    {
      curAttr->valAttr = valattr;
      return lattr;
    }

  curlist = lattr;
  while (pk_true)
    {
      curAttr = ptlAttributes_get_elem (curlist);
      curNum = ptAttribute_get_num (curAttr);
      if (num == curNum)
	{
	  curAttr->valAttr = valattr;
	  return lattr;
	}
      succ = ptlAttributes_get_tail (curlist);
      if (ptlAttributes_satisfies_nil (succ))
	{
	  succ = ptlAttributes_RIEN_create (attr, NILLATTRIBUTES);
	  curlist->next = succ;
	  return lattr;
	}
      cnum = ptAttribute_get_num (ptlAttributes_get_elem (succ));
      if (cnum > num)
	{
	  succ = ptlAttributes_RIEN_create (attr, succ);
	  curlist->next = succ;
	  return lattr;
	}
      curlist = succ;
    }
}

/** It removes an attribute (num) from a list (li).
    @return "true" if the element has been removed from the list, "false" otherwise.
*/
extern ptlAttributes
ptlAttributes_remove (ptlAttributes list, int num)
{
  ptlAttributes curr, pred;
  int cnum;

  if (ptlAttributes_satisfies_nil (list))
    return list;
  cnum = ptAttribute_get_num (ptlAttributes_get_elem (list));
  if (num < cnum)
    return list;
  if (num == cnum)
    return list->next;

  pred = list;
  curr = list->next;
  while (pk_true)
    {
      if (ptlAttributes_satisfies_nil (curr))
	return list;
      cnum = ptAttribute_get_num (ptlAttributes_get_elem (curr));
      if (cnum == num)
	{
	  pred->next = curr->next;
	  return list;
	}
      pred = curr;
      curr = curr->next;
    }
}
