/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** An attribute is a couple (num, value) */
struct Attribute
{
  int num; /**< ordinal of the atribute */
  attributeValue valAttr; /**< value of the attribute */
};


/** List of attribute (element, next) */
struct lAttributes
{
  ptlAttributes next;
  ptAttribute elem;
};

/** Constructor.
    @return a new list of attributes composed of an attribute (val) followed by l.
*/
static ptlAttributes
ptlAttributes_RIEN_create (ptAttribute val, ptlAttributes l)
{
  ptlAttributes vmark;

  vmark = new lAttributes ();
  vmark->elem = val;
  vmark->next = l;
  return vmark;
}

/**
   @return "true" if the list is empty, "false" otherwise.
*/
static boolean
ptlAttributes_satisfies_nil (ptlAttributes list)
{
  return (list == NILLATTRIBUTES);
}

/**
   @return "true" if a list of attributes (list) is not empty, "false" otherwise.
   The header attribute (vattr) is returned, and the list of attributes (list) is set to the next.
*/
static boolean
ptlAttributes_get_next (ptlAttributes * list, ptAttribute * vattr)
{
  if ((*list) == NILLATTRIBUTES)
    return pk_false;
  *vattr = (*list)->elem;
  *list = (*list)->next;
  return pk_true;
}


/** @return the attribute associated with the head of a list (list). */
static ptAttribute
ptlAttributes_get_elem (ptlAttributes list)
{
  if (list == NILLATTRIBUTES)
    return NILATTRIBUTE;
  return list->elem;
}

/** @return the successor of a list (list). */
static ptlAttributes
ptlAttributes_get_tail (ptlAttributes list)
{
  if (list == NILLATTRIBUTES)
    return NILLATTRIBUTES;
  return list->next;
}


/** Attribute Constructor.
    @return a new attribute (num, val).
*/
static ptAttribute
ptAttribute_RIEN_create (int num, attributeValue val)
{
  ptAttribute attr;

  attr = new Attribute ();
  attr->num = num;
  attr->valAttr = val;
  return attr;
}

/**
   @return the ordinal of an attribute (attr).
*/
static int
ptAttribute_get_num (ptAttribute attr)
{
  return attr->num;
}

/**
   @return the value of an attribute (attr).
*/
static attributeValue
ptAttribute_get_val (ptAttribute attr)
{
  return attr->valAttr;
}
