/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** 
    Switch on the concatenation mode when not(b), otherwise the "separated" mode.
*/
extern void
snglue (boolean b)
{
  separer = b;
}


/* It adds a space charater.*/
extern void
snspace (void)
{
  buf->space ();
}



/** It adds a space charater and set the concatenation mode on. */
extern void
printspace (void)
{
  buf->space ();
  separer = pk_false;
}



/** It sets the concatenation mode on and increments of n the tabulation position.
*/
extern void
sntab (integer n)
{
  separer = pk_false;
  buf->tabulation (n);
}



/** It sets the tabulation position to its previous value.*/
extern void
snbacktab (void)
{
  buf->backTabulation ();
}



/**
   It sets a mark on the current  position of the buffer if (level>0).
*/
extern void
snmark (integer level)
{
  if (level > 0)
    buf->mark (0);
}



/** It removes the current mark, previously set by snmark(), if (level>0).
 */
extern void
snrelease (integer level)
{
  if (level > 0)
    buf->backMark ();
}



/** It asks lb characters in the buffer. */
extern void
snfit (integer lg)
{
  if (buf->isOver (lg))
    buf->empty ();
}



/** It adds a character (caract) to the buffer. This operation is not guaranteed.
 */
extern void
sncharwrite (char caract)
{
  buf->writeChar (caract);
}



/** It writes the buffer.
 */
extern void
snline (void)
{
  buf->empty ();
}

/** It writes a new line.
 */
extern void
newline (void)
{
  buf->newline ();
}


/** It adds a character (caract) to the buffer. This operation is guaranteed.
 */
extern void
sncharpri (char caract)
{
  buf->printChar (caract);
}



/**  It adds a string (str) to the buffer. This operation is not guaranteed. */
extern void
sigwristr (ptrepstr str)
{
  buf->writeRepstr (str);
}



/**  It adds a string (str) to the buffer. This operation is guaranteed. */
extern void
sigpristr (ptrepstr str)
{
  buf->printString (str);
}


/**  It adds a string (str) to the buffer. This operation is guaranteed. */
extern void
sigpristring (char *str)
{
  buf->printString (str);
}

/**  It adds the string, associated with a nullary tree (t), to the buffer. This operation is guaranteed.
 */
extern void
inteconsprint (ntrees p)
{
  sigpristr (nstringof (p));
}

/**
   It inserts, in "separated" mode, a space to the buffer.
*/
extern void
separate ()
{
  if (separer)
    snspace ();
}

/**
   It adds in "separated" mode a space or switch the mode (concatenation to separated) then
   it inserts the string (str) in the buffer (operation is guaranteed).
*/
extern void
inteconsstrprint (ptrepstr str)
{
  if (separer)
    snspace ();
  else
    separer = pk_true;
  sigpristr (str);
}

/**
   It adds in "separated" mode a space or switch the mode (concatenation to separated) then
   it inserts the string, associated with a nullary tree (t), in the buffer (operation is guaranteed).
*/
extern void
inteconsalphaprint (ntrees t)
{
  if (separer)
    snspace ();
  else
    separer = pk_true;
  sigpristr (nstringof (t));
}




/**
   if level>0, it writes the buffer, then set the insertion mode to concatenation; otherwise it adds a space in "separated mode".
*/
extern void
wline (integer level)
{
  if (level > 0)
    {
      snline ();
      separer = pk_false;
    }
  else if (separer)
    printspace ();
}



/**
   It adds in "separated" mode a space or switch the mode (concatenation to separated) then
   it adds the string, associated with a nullary tree(t), to the buffer (operation not guaranteed).
*/
extern void
inteconsalphawrite (ntrees p)
{
  if (separer)
    snspace ();
  else
    separer = pk_true;
  sigwristr (nstringof (p));
}

/**
  Initialising of the module.
  Insertion mode is set to concatenation.
*/
extern void
inteprint_initialize ()
{
  separer = pk_false;
}

/**
   Opening a printing session.
   @param oMode (in) output mode: 0 for screen output, 1 or 2 graphical output.
   @param cbuf (in) a buffer, the operations will produce text in this text buffer.
*/
extern boolean
inteprint_open_stdOut (ptrepbuf cbuf, integer oMode)
{
  savebuf = buf;
  buf = cbuf;
  return cbuf->buffer_open (nilstr, oMode);
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It opens a printing session on a file with filePathName as name (it fails if filePathName is undefined).
   @param cbuf (in) a buffer, the operations will produce text in this text buffer.
   @param filePathName (in) filename.
*/
extern boolean
inteprint_open_buffer (ptrepbuf cbuf, ptrepstr filePathName)
{
  savebuf = buf;
  buf = cbuf;
  return cbuf->buffer_open (filePathName, 0);
}

/**
  @return "true" if the operation succeeds, "false" otherwise.
  It opens the session for writing on file with unsuffixedFilePathName.suff as name (it fails if unsuffixedFilePathName is undefined)
  @param cbuf (in) a buffer, the operations will produce text in this text buffer.
  @param unsuffixedFilePathName (in) file path name prefix.
  @param suff (in) file name suffix.
  @param fname (out) the effective opened file name.
*/
extern boolean
inteprint_open_bufferNewName (ptrepbuf cbuf, ptrepstr unsuffixedFilePathName, const char *suff, ptrepstr * filePathName)
{
  if (unsuffixedFilePathName == nilstr) return pk_false;
  if (!sys_add_typeSuffix2repstr (unsuffixedFilePathName, suff, filePathName)) return pk_false;
  if (!sys_filePathInUserOutputDir (*filePathName, filePathName)) return pk_false;
  inteprint_open_buffer (cbuf, *filePathName);
  return pk_true;
}

/** Closing a printing session.
 */
extern void
intpriclose ()
{
  buf->buffer_close ();
  buf = savebuf;
}

/**
   It sets to b the char encoding mode:
   b must be "false" for XML, "true" otherwise.
*/
extern void
setCharEncodingMode (boolean b)
{
  buf->setCharEncodingMode (b);
}
