
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   @return the string foo_body.h, for processName=foo
*/
  ptrepstr inteprintgen::makeStr_fileName (ptrepstr prefix, const char *typeSuffix)
{
  ptrepstr name;
  if (makeRepstr_repstr_string (prefix, typeSuffix, &name))
    return name;
  return nilstr;
}

/**
   @return the string foo_body.h, for processName=foo
*/
ptrepstr
inteprintgen::makeStr_fileName (ptrepstr prefix, const char *innerSuffix, const char *typeSuffix)
{
  ptrepstr name;
  if (makeRepstr_repstr_string (prefix, innerSuffix, &name))
    if (makeRepstr_repstr_string (name, typeSuffix, &name))
      return name;
  return nilstr;
}

/**
   @return the string foo_body.h, for processName=foo
*/
ptrepstr
inteprintgen::makeStr_fileName (ptident id, const char *typeSuffix)
{
  return makeStr_fileName (ident2repstr (id), typeSuffix);
}

/**
   @return the string foo_body.h, for processName=foo
*/
ptrepstr
inteprintgen::makeStr_fileName (ptident id, const char *innerSuffix, const char *typeSuffix)
{
  return makeStr_fileName (ident2repstr (id), innerSuffix, typeSuffix);
}

/**
   It associates a file (filePathName) to the printing buffer.
*/
boolean
inteprintgen::open_newBuffer (ptrepstr filePathName, int v1, int v2)
{
  savebuf = buf;
  buf = new buffer (v1, v2);
  return buf->buffer_open (filePathName, 0);
}

/**
   It associates a file (filePathName) to the printing buffer.
*/
boolean
inteprintgen::open_newBuffer (ptrepstr fileName, int v1, int v2, ptrepstr * filePathName)
{
  if (sys_filePathInUserOutputDir (fileName, filePathName))
    return open_newBuffer (*filePathName, v1, v2);
  return pk_false;
}

/**
   It associates a file unsuffixedFileName.suffix  to the printing buffer.
*/
boolean
inteprintgen::open_newBuffer (ptrepstr unsuffixedFileName, const char *suffix, int v1, int v2, ptrepstr * filePathName)
{
  ptrepstr fileName;
  if (sys_add_typeSuffix2repstr (unsuffixedFileName, suffix, &fileName))
    return open_newBuffer (fileName, v1, v2, filePathName);
  return pk_false;
}

/** It empties the associated buffer. (It writes a line) */
_VIRTUAL_ void
inteprintgen::printLine ()
{
  buf->empty ();
}

/**
   It writes a space character.
*/
_VIRTUAL_ void
inteprintgen::printSpace ()
{
  buf->space ();
}

/**  Syntax independant.\n
     It writes a character (c).
*/
_VIRTUAL_ void
inteprintgen::printchar (char c)
{
  buf->printChar (c);
}


/**  Syntax independant.\n
     It adds a left margin tablation.
*/
_VIRTUAL_ void
inteprintgen::indent ()
{
  buf->tabulation (lgindent);
}

/** Syntax independant.\n
    It deletes a left margin tabulation.
*/
_VIRTUAL_ void
inteprintgen::backindent (int n)
{
  int i;
  if (n > 0)
    for (i = 1; i <= n; i++)
      buf->backTabulation ();
}

/**
   Syntax independant.\n
   It sets a mark to the current position. 
*/
_VIRTUAL_ void
inteprintgen::mark ()
{
  buf->mark (0);
}

/** Deprecated.Use printString(ptrepstr). */
_VIRTUAL_ void
inteprintgen::printstr (ptrepstr str)
{
  printString (str);
}

/** It writes the operand string. */
_VIRTUAL_ void
inteprintgen::printString (ptrepstr str)
{
  if (!isNilString (str))
    {
      buf->fit (strsize (str));
      cwristr (str);
    }
}

/** It writes an ASCII string (vstr) following by a space. */
_VIRTUAL_ void
inteprintgen::printstrspace (ptrepstr vstr)
{
  printstr (vstr);
  printSpace ();
}

/** It writes ':' character. */
_VIRTUAL_ void
inteprintgen::printColon ()
{
  printchar (':');
}

/** It writes '.' character. */
_VIRTUAL_ void
inteprintgen::printDot ()
{
  printchar ('.');
}

/** It prints the '_' character. */
_VIRTUAL_ void
inteprintgen::printunderscore ()
{
  printchar ('_');
}


/** It writes '?' character. */
_VIRTUAL_ void
inteprintgen::printQuestionMark ()
{
  printchar ('?');
}

/** It writes ',' character. */
_VIRTUAL_ void
inteprintgen::printComma ()
{
  printchar (',');
}


/** It writes ';' character. */
_VIRTUAL_ void
inteprintgen::printSemiColon ()
{
  printchar (';');
}

/** It writes '(' character. */
_VIRTUAL_ void
inteprintgen::printLP ()
{
  printchar ('(');
}

/** It writes ')' character. */
_VIRTUAL_ void
inteprintgen::printRP ()
{
  printchar (')');
}

/** It writes '[' character. */
_VIRTUAL_ void
inteprintgen::printLeftSquareBracket ()
{
  printchar ('[');
}

/** It writes '{' character. */
_VIRTUAL_ void
inteprintgen::printLBracket ()
{
  printchar ('{');
}

/** It writes '}' character. */
_VIRTUAL_ void
inteprintgen::printRBracket ()
{
  printchar ('}');
}

/** It writes ']' character. */
_VIRTUAL_ void
inteprintgen::printRightSquareBracket ()
{
  printchar (']');
}



/** It writes "Polychrony" key word.*/
_VIRTUAL_ void
inteprintgen::printkwPolychrony (boolean suffUnderscore)
{
  printstr (string2repstr (kwPolychrony));
  if (suffUnderscore)
    printunderscore ();
}

/** It writes the "not yet implemented" key word. */
_VIRTUAL_ void
inteprintgen::printkwnotimpl ()
{
  buf->printAlfa (kwnotimpl, lgkwnotimpl);
}

/** It writes an identifier (id). */
_VIRTUAL_ void
inteprintgen::printident (ptident id)
{
  printstr (ident2repstr (id));
}

/** It writes an identifier (id) following by a space. */
_VIRTUAL_ void
inteprintgen::printidentspace (ptident id)
{
  printident (id);
  printSpace ();
}


/** It closes the current file. */
_VIRTUAL_ void
inteprintgen::closefile ()
{
  buf->buffer_close ();
  buf = savebuf;
}

/**
   It prints ';' character.
*/
_VIRTUAL_ void
inteprintgen::printSemiColonLine ()
{
  buf->printChar (';');
  printLine ();
}


/**
   It prints ';' character and signals n back indentations.
*/
_VIRTUAL_ void
inteprintgen::printSemiColonLineBackIndent (int n)
{
  printSemiColonLine ();
  backindent (n);
}

/**
   It prints ");" string.
*/
_VIRTUAL_ void
inteprintgen::printRPSemiColonLine ()
{
  buf->print2Chars (')', ';');
  printLine ();
}


/**
   It prints ")" string.
*/
_VIRTUAL_ void
inteprintgen::printRPline ()
{
  buf->printChar (')');
  printLine ();
}


/**
   It prints ")" string and prints the line.
*/
_VIRTUAL_ void
inteprintgen::printRPlineindent ()
{
  printRPline ();
  indent ();
}

/**
   It prints an int (n).
*/
_VIRTUAL_ void
inteprintgen::printint (long int n)
{
  ptrepstr str;

  if (n < 0)
    {
      str = integer2repstr (-n);
      buf->fit (strsize (str) + 1);
      buf->printChar ('-');
    }
  else
    {
      str = integer2repstr (n);
      buf->fit (strsize (str));
    }
  buf->writeRepstr (str);
}


/**
   It prints a reference kew word.
*/
_VIRTUAL_ void
inteprintgen::printkwref ()
{
  buf->printChar ('&');
}


/** It prints the "+" operator. */
_VIRTUAL_ void
inteprintgen::printkwplus ()
{
  buf->printChar ('+');
}

/** It prints the "-" operator. */
_VIRTUAL_ void
inteprintgen::printkwminus ()
{
  buf->printChar ('-');
}

/** It prints the "*" operator. */
_VIRTUAL_ void
inteprintgen::printstar ()
{
  buf->printChar ('*');
}

/** It prints the "/" operator. */
_VIRTUAL_ void
inteprintgen::printkwdivi ()
{
  buf->printChar ('/');
}


/** It prints "+1" string. */
_VIRTUAL_ void
inteprintgen::printplusOne ()
{
  buf->print2Chars ('+', '1');
}

/** It prints "+v" string. */
_VIRTUAL_ void
inteprintgen::printplus (int v)
{
  if (v == 0)
    return;
  printkwplus ();
  printint (v);
}

/**
   It prints the "if" key-word.
*/
_VIRTUAL_ void
inteprintgen::printkwif ()
{
  buf->printAlfa (kwif, lgkwif);
  buf->space ();
}

/**
   It prints the "else" key-word.
*/
_VIRTUAL_ void
inteprintgen::printkwelse ()
{
  buf->printAlfa (kwelse, lgkwelse);
  buf->space ();
}

/**
   It prints the "end" key-word.
*/
_VIRTUAL_ void
inteprintgen::printkwend ()
{
  buf->printAlfa (kwend, lgkwend);
  buf->space ();
}

/**
   It prints the "then" key-word.
*/
_VIRTUAL_ void
inteprintgen::printkwthen ()
{
  buf->printAlfa (kwthen, lgkwthen);
  buf->space ();
}

/**
   It prints the opening of a block statement.
*/
_VIRTUAL_ void
inteprintgen::printblockopening ()
{
  buf->printChar ('{');
  printLine ();
}

/**
   It opens  a block statement (printblockopening()) and signals an indentation (indent()).
*/
_VIRTUAL_ void
inteprintgen::blockIndentOpening ()
{
  printblockopening ();
  indent ();
}

_VIRTUAL_ void
inteprintgen::blockIndentClosing ()
{
  backindent (1);
  printBlockClosingBackIndent ();
  printLine ();
}


/** It closes a block statement. */
_VIRTUAL_ void
inteprintgen::printblockclosing ()
{
  buf->printChar ('}');
  printLine ();
}

/** It closes a block statement previously open by blockIndentOpening() and signals a back-indentation (backindent()). */
_VIRTUAL_ void
inteprintgen::printBlockClosingBackIndent ()
{
  printblockclosing ();
  backindent (1);
}

/** It signals nbck back-indentations. */
_VIRTUAL_ void
inteprintgen::printBlockClosingBackIndent (int nbck)
{
  if (nbck > 0)
    {
      while (nbck > 0)
	{
	  backindent (1);	// printBlockClosingBackIndent();      
	  nbck--;
	}
      printLine ();
    }
}



/**
   It prints a string (com) as a comment.
   It calls printBeginComment(), writeRepstr(), printEndComment().
*/
_VIRTUAL_ void
inteprintgen::printcomment (ptrepstr com)
{
  printBeginComment ();
  buf->writeRepstr (com);
  printEndComment ();
  printLine ();
}


/**
   It generates the "modulo v" expression.
*/
_VIRTUAL_ void
inteprintgen::printmodulo (int v)
{
  buf->printChar ('%');
  printint (v);
}

/** 
    It generates the pow operator.
 */
_VIRTUAL_ void
inteprintgen::printPow ()
{
  printstr (string2repstr ("**"));
}


/**
   It generates "();"
*/
_VIRTUAL_ void
inteprintgen::genLPRPSemiColonLine ()
{
  printLPRP ();
  printSemiColonLine ();
}

/** It prints "()" string. */
_VIRTUAL_ void
inteprintgen::printLPRP ()
{
  printchar ('(');
  printchar (')');
}

/**
   It print an identifier (id) followed by an undersocre character.
*/
_VIRTUAL_ void
inteprintgen::printidentUnderscore (ptident id)
{
  printident (id);
  printunderscore ();
}


/**
   It generates a Polychrony string (str).
*/
_VIRTUAL_ void
inteprintgen::cwristr (ptrepstr str)
{
  buf->writeRepstr (str);
}


/** It writes the opening of a comment. (c LIKE) */
_VIRTUAL_ void
inteprintgen::printBeginComment ()
{
  buf->print2Chars ('/', '*');
}

/** It writes the closing of a comment.(c LIKE) */
_VIRTUAL_ void
inteprintgen::printEndComment ()
{
  buf->print2Chars ('*', '/');
}



/**
   It writes the key word "while".
*/
_VIRTUAL_ void
inteprintgen::printkwWhile ()
{
  buf->writeAlfa (kwwhile, lgkwwhile);
}

/**
   It writes the key word "do".
*/
_VIRTUAL_ void
inteprintgen::printkwDo ()
{
  buf->writeAlfa (kwdo, lgkwdo);
}

/**
   It generates the boolean constant "TRUE" when cb, "FALSE" otherwise.
*/
_VIRTUAL_ void
inteprintgen::printkwlogicons (boolean cb)
{
  if (cb)
    buf->printAlfa (kwTrue, lgkwTrue);
  else
    buf->printAlfa (kwFalse, lgkwFalse);
}

/**
   It generates the boolean constant "true" when cb, "false" otherwise.
*/
_VIRTUAL_ void
inteprintgen::printkwlogicons_m (boolean cb)
{
  if (cb)
    buf->printAlfa (kwTrueL, lgkwTrue);
  else
    buf->printAlfa (kwFalseL, lgkwFalse);
}

/**
   It generates the boolean constant "true" when cb, "false" otherwise.
*/
_VIRTUAL_ void
inteprintgen::printkwfor ()
{
  buf->fit (lgkwfor + 2);
  buf->writeAlfa (kwfor, lgkwfor);
}


/**
   It generates the c-string vstr.
*/
_VIRTUAL_ void
inteprintgen::printString (const char *vstr)
{
  buf->printString (vstr);
}


/** It writes '|' character. */
_VIRTUAL_ void
inteprintgen::printBar ()
{
  printchar ('|');
}

/**
   It writes 'in' key word..
*/
_VIRTUAL_ void
inteprintgen::printkwin ()
{
  buf->fit (3);
  buf->writeAlfa ("in", 2);
}
