/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   It closes a log-session.
*/
  void
messenger::parsing_logFileClose ()
{
  if (isParserLogFileOpen)
    {
      aBuffer->buffer_close ();
      isParserLogFileOpen = pk_false;
    }
}

/**
   It resets the log-session (log-file is reset).
*/
boolean
messenger::parsing_logFileReset ()
{
  if (isParserLogFileOpen) parsing_logFileClose ();
  if (aBuffer->buffer_open (vParserLogFile, 0 /* 0->stdout */ ))
    {
      isParserLogFileOpen = pk_true;
      return pk_true;
    }
  return pk_false;
}

void
messenger::parsing_logFileInitialize ()
{
  parsing_logFileReset ();
  aBuffer->printString (prefixBeginCodeop);
  aBuffer->printString ("Parser");
  aBuffer->empty ();
}


/**
   It prints the log file when berr and signals to the log-session the end of parsing operation.
*/
void
messenger::parsing_logFileFinalize (boolean berr)
{
  if (berr) parsing_logFilePrint (nilstr);
  parsing_logFileInitialize (); 
  /* parsing_logFileReset ();
     aBuffer->printString (prefixEndCodeop);
     aBuffer->printString ("Parser");
     aBuffer->empty (); */
}



/**
   Assumed: a log-session is opened.
   It inserts in the log-file 
   the verrno-th error message followed by the l-th line of the file (vcurrentFile) and
   a mark under the col-th of this copied line.
*/
void
messenger::parsing_errorLineColumn (ptrepstr vcurrentFile, int l, int col, int verrno)
{
  int j;
  ptrepfil f;
  ptrepstr vstr;
  char c;
  boolean stop;

  if (l < 0)
    l = 0;
  if (col < 0)
    col = 0;
  aBuffer->printString (polychronyLineParser);
  aBuffer->printInteger (l);
  aBuffer->printChar (' ');
  aBuffer->printString (polychronyColumnParser);
  aBuffer->printInteger (col);
  aBuffer->printChar (' ');
  aBuffer->printString (polychronyErrorMessParser);
  vstr = getfullmess (err, verrno);
  aBuffer->printString (vstr);
  aBuffer->empty ();

  f = sysopen (vcurrentFile, pk_false);
  // Go to the line of the error.
  for (j = 1; j < l;)
    if (sysnewline (f))
      j++;
  // Output of the error line.
  stop = pk_false;
  while (!stop)
    if (syseoline (f))
      stop = pk_true;
    else if (!sysgetcha (f, &c))
      stop = pk_true;
    else if (c == '\t')
      aBuffer->printChar (' ');
    else
      aBuffer->printChar (c);
  sysclose (f);
  aBuffer->printChar ('\n');
  aBuffer->printNChars (' ', col);
  aBuffer->printChar (polychronyErrorMark);
  aBuffer->printNChars (' ', col);
  aBuffer->empty ();
  setErrorCode (pk_false);
}

/**
   Printing (filePathName==nilstr) or saving (in filePathName) the log-file.
*/
void
messenger::parsing_logFilePrint (ptrepstr filePathName)
{
  if (isParserLogFileOpen) parsing_logFileClose ();
  if (aBuffer->buffer_open (filePathName, 0))
    LogPutAndClose ();
  else
    PKExit ();
}

/**
   Saving the log-file to the Polychrony_syntactic_errors.log of the current directory.
*/
void
messenger::parsing_logFileSave (char *vfile)
{
  ptrepstr LogFile;

  LogFile = string2repstr (vfile);
  makeRepstr_repstr_string (LogFile, "_SYNTAX_ERRORS_LOG", &LogFile);
  parsing_logFilePrint (LogFile);
}

/** 
   Getting in tab, nmax chars of the log-file.
   @return the index of the substring of tab to be displayed.
*/
char *
messenger::getInfoParser (char *tab, int nmax)
{
  ptrepfil f;
  char c;
  int n;
  boolean stop;

  if (isParserLogFileOpen) parsing_logFileClose ();
  f = sysopen (vParserLogFile, pk_false);
  stop = pk_false;
  n = 0;
  while ((n < nmax) && !stop)
    {
      if (syseoline (f))
	{
	  tab[n] = '\n';
	  n++;
	}
      if (sysgetcha (f, &c))
	tab[n] = c;
      else
	stop = pk_true;
      n++;
    }
  sysclose (f);
  tab[n] = '\0';
  char *x = strstr (tab, polychronyErrorMessParser);
  if (x == NULL)
    x = tab;
  else
    x = x + strlen (polychronyErrorMessParser);
  char *y = strstr (x, polychronyEndCodeOp);
  if (y != NULL)
    *y = '\0';
  return x;
}
