/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @defgroup tabkw table: Managing table of keywords.
    @defgroup tabmess table: Managing table of messages.
 */

/**
   @ingroup tabkw

   @param tabname (in) the name of table without suffix.
   @param res (out) the table of key-words reads in the $Signal_bin/\<tabname\>.kw, if the operatioon succeeds.
   @return "true" is the operatioon succeeds, "false" otherwise.
*/

extern boolean
tabkwload (const char *tabname, ptreptab * res)
{
  char *filname;

  // printf("DEBUG entering tabkwload, tabname is %s\n", tabname); fflush(stdout); 
  filname = sys_suffixedFileInPolyBin (tabname, KWSUFF);
  return strTableLoad (res, filname);
}


/**
   @ingroup tabmess
   Pushing the current message table number.
 */
extern void
savesetindpass ()
{
  saveindpass = indpasse;
  indpasse = 0;
}


/**
   @ingroup tabmess
   Poping the message table number previously pushed by savesetindpass() function.
*/
extern void
restoreindpass ()
{
  indpasse = saveindpass;
}

/** 
    @ingroup tabmess
    Initiliasing of the message tables.
*/
extern void
setnulltab ()
{
  int i;

  indpasse = 0;
  for (i = 0; i <= maxpasse; i++)
    taberrinterne[i] = NIL_reptab;
  for (i = 0; i <= maxpasse; i++)
    taberrfatale[i] = NIL_reptab;
  for (i = 0; i <= maxpasse; i++)
    tabinfo[i] = NIL_reptab;
  for (i = 0; i <= maxpasse; i++)
    taberr[i] = NIL_reptab;
  for (i = 0; i <= maxpasse; i++)
    tabwarning[i] = NIL_reptab;
  for (i = 0; i <= maxpasse; i++)
    tabcomment[i] = NIL_reptab;
}


/**
    @ingroup tabmess
    Initialising for printing a message of a table (message, index on this message is initialized.
    @param mtyp (in) the name of the selected table (error, warning, info, comment, fatal error, internal error)
    @param mnum (in) the ordinal of the message in the selected table.
*/
extern void
initmess (diagtype mtyp, int mnum)
{
  switch ((int) mtyp)
    {
    case texte:
      iemmess = strTableGet (tabcomment[indpasse], mnum);
      break;
    case info:
      iemmess = strTableGet (tabinfo[indpasse], mnum);
      break;
    case warn0:
    case warn1:
    case warn2:
      iemmess = strTableGet (tabwarning[indpasse], mnum);
      break;
    case err:
      iemmess = strTableGet (taberr[indpasse], mnum);
      break;
    case fatale:
      iemmess = strTableGet (taberrfatale[indpasse], mnum);
      break;
    case interne:
      iemmess = strTableGet (taberrinterne[indpasse], mnum);
      break;
    }
}

/**
    @ingroup tabmess
    @return 
    - 0 when the end of message is found.
    - the current char of the selected message otherwise. The index on the selected message is increased.
    The use of these function must be preceeded by a call to \ref startmess() function.
*/
extern char
getmesschar ()
{
  char c;

  if (pmsch > strsize (iemmess))
    return 0;
  c = strgetchar (iemmess, pmsch);
  pmsch = pmsch + 1;
  return c;
}


/**
   @ingroup tabmess
   @param mtyp (in) the name of the selected table (error, warning, info, comment, fatal error, internal error)
   @param mnum (in) the ordinal of the message in the selected table.
   @return the mnum message of the selected table.
*/
extern ptrepstr
getfullmess (diagtype mtyp, int mnum)
{
  ptrepstr vstring;

  switch ((int) mtyp)
    {
    case texte:
      vstring = strTableGet (tabcomment[indpasse], mnum);
      break;
    case info:
      vstring = strTableGet (tabinfo[indpasse], mnum);
      break;
    case warn0:
    case warn1:
    case warn2:
      vstring = strTableGet (tabwarning[indpasse], mnum);
      break;
    case err:
      vstring = strTableGet (taberr[indpasse], mnum);
      break;
    case fatale:
      vstring = strTableGet (taberrfatale[indpasse], mnum);
      break;
    case interne:
      vstring = strTableGet (taberrinterne[indpasse], mnum);
      break;
    }
  return vstring;
}


/**
    @ingroup tabmess
    It sets the index on the selected message at the begining.
    @return the first character of the selected message.
    Use getmesschar() to get the next characters.
*/
extern char
startmess ()
{
  pmsch = 1;
  return getmesschar ();
}



/**
   @ingroup tabmess
   Loading table of messages from the file messname.indpas of the directory $Signal_bin.
   @param messname (in) : the name of the table.
   @param indpas (in) : step of the compiler.
   @return "true" if the table is loaded, "false" otherwise (file cannot be opened, string memory is full).
*/
extern boolean
tabdiaload (const char *messname, int indpas)
{
  char *filname;
  ptreptab res;

  indpasse = indpas;
  if (loadedTab[indpasse])
    return pk_true;
  filname = sys_digitSuffixedFileInPolyBin (messname, indpas);
  return tabloaddia (&res, filname);
}


extern void
tabtoolsIni ()
{
  int i;
  for (i = 0; i <= maxpasse; i++)
    loadedTab[i] = pk_false;
}
