/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
// #define isIdentifierOp(op) ((op == opnom) || (op == opident) || (op == opcham) || (op == opelem))
// #define isClockOperator(op) ((op == opcomp) || (op == opbinf) || (op == opbsup) /* || (op == ophnull) */ )
// #define isClockRelationOperator(op) ((op == oplcrohor) || (op == opldechor) || (op ==  oplequhor) || (op == oplexchor))
// #define isRelationalOp(op)  ((op == opegeg) || (op == opiieg) || (op == opsupe) || (op == opinfe) || (op ==  opsegal) || (op == opiegal) || (op == opegal) || (op == opdiff))
// #define isLogicalOp(op)  ((op == opouex) || (op == opet) || (op == opou) || (op ==  opnon))
#define isXmiBlock(op) ((op == oppetiq) || (op == opobli) || (op == oplfeed)  || (op == oppcas) || (op == oppconf) || (op == opmoti) || (op == opiter))
/**
   When an operator is not yet implemented, the translator will return the "false" value.
   Currently, only some "manipulations of arrays" are not implemented. They can be rewritten handmade.
 */
  void
SME_SyntaxRestriction::notOK (ntrees e)
{
  Ok = pk_false;
  Trees2Kernel::notYetImplemented (e);
}

/**
   Assumed: vdecl is an signal identifier in the declaration area.
   @return "true" if the type (vtype) is set, "false" otherwise.
*/
boolean
SME_SyntaxRestriction::getTypeOf (ntrees vdecl, ntrees * vtype)
{

  while (noper (vdecl) != opdecsig)
    vdecl = nfather (vdecl);
  *vtype = nchild (1, vdecl);
  return pk_true;
}

/** the construction (p) is not yet implemented.
    It is accepted.
*/
_VIRTUAL_ void
SME_SyntaxRestriction::notYetImplemented (ntrees p)
{
  //  mkklErrorAnnotTree(errnotimple, p);
}

/**
   @return the model of the referenced model in an expression (vklExprProc).
*/
ptrepfi SME_SyntaxRestriction::getDeclarationModel (ptklExprProc vklExprProc)
{
  ptrepfi
    vfi;
  ptlrepid
    varid;
  noperators
    code;
  ntrees
    vt;
  ptklExprSig
    klExprSig1,
    klExprSig2,
    lparam;

  code = kl_getExprProcType (vklExprProc);
  if (code == opappe)
    {
      vklExprProc = kl_getModelProcCall (vklExprProc, &klExprSig1, &vt);
      code = kl_getExprProcType (vklExprProc);
    }

  if (code == opdefsig)
    {
      kl_getEquation (vklExprProc, &klExprSig1, &klExprSig2, &vt);
      varid = kl_getFlowExprModelProcCallname (klExprSig2);
      return getModelRepid (varid);
    }
  if (code == opinst)
    {
      kl_getMacro (vklExprProc, &varid, &lparam, &vt);
      return getModelRepid (varid);
    }
  else;				//?
  return NILREPFI;
}

/**
   @return a list of identifiers from a list of expressions (outputs) to be conform with the SIGNAL syntax.
*/
ntrees SME_SyntaxRestriction::mkldecl (ntrees outputs)
{
  ntrees
    ldecl;
  int
    i;

  if (!STree_isList (outputs))
    ldecl = LL_make_lsign_list (STree_scopy (outputs));
  else
    {
      ldecl = LL_make_lsign_emptyList ();
      for (i = 1; i <= ntreearity (outputs); i++)
	ldecl = ntreeList_appendLast (ldecl, STree_scopy (nchild (i, outputs)));
    }
  return ldecl;
}

void
SME_SyntaxRestriction::addDeclarationToConfinedExpr (ntrees p_conf, ntrees vtype, ntrees lsig)
{
  STree_addDeclarationToConfinedExpr (p_conf, LL_make_decsig_tree (vtype, lsig));
}

/**
   @return a confined expression "<np> where <vtype> <lsig> end."
*/
ntrees
SME_SyntaxRestriction::addConfinedExpression (ntrees np, ntrees vtype, ntrees lsig)
{
  ntrees p_conf;
  p_conf = STree_mkConfinedExpression (np, NILTREE);
  addDeclarationToConfinedExpr (p_conf, vtype, lsig);
  return p_conf;
}

/**
   Searching for a type tree (op) in the ancestors.
   @return "true" if it is found, "false" otherwise.
*/
boolean
SME_SyntaxRestriction::upUntil (ntrees f, noperators op, ntrees * res)
{
  boolean found;
  ntrees f1;

  found = pk_false;
  while (!found)
    {
      f1 = nfather (f);
      if (IS_NILTREE (f1))
	return pk_false;
      if (noper (f1) == op)
	found = pk_true;
      else
	f = f1;
    }
  *res = f1;
  return pk_true;
}

/**
   If f is in the context "f where ...end;", the elements of nDefs are added to f and the declarations (lids) to the where part.
   Otherwise, it replaces F in a composition by a confined expression "(| F | refModel |) where lsig; end;" where F is an ancestor of f.
*/
void
SME_SyntaxRestriction::insertBlock (ntrees f, ntrees nDefs, ntrees lsig)
{

  ntrees decl, p_conf, np, f1;
  int rank, i;

  if (noper (f) != oplfeed)
    {
      if (!upUntil (f, oplfeed, &f1))
	return;
      f = f1;
    }

  // Adding the new definitions to the f-composition.
  if (noper (nDefs) == oplfeed)
    for (i = 1; i <= ntreearity (nDefs); i++)
      f = ntreeList_appendLast (f, nchild (i, nDefs));
  else
    f = ntreeList_appendLast (f, nDefs);

  // Adding the new declarations (lsig)
  f1 = nfather (f);
  if (noper (f1) != oppconf)
    {
      rank = nposit (f);
      p_conf = addConfinedExpression (f, STree_getImplicitTypeTree (), lsig);
      nchangchild (f1, p_conf, rank);
    }
  else
    addDeclarationToConfinedExpr (f1, STree_getImplicitTypeTree (), lsig);
}


/**
   @return the tree that represents the list, with the operator opl, of the identifiers of list.
*/
ntrees SME_SyntaxRestriction::mkListTree (ptlrepid list, noperators opl)
{
  ptident
    id;
  ntrees
    treeList;

  if (list == NILLREPID)
    treeList = STree_mkEmptyTree ();
  else
    {
      treeList = nmakemptylist (opl);
      while (list != NILLREPID)
	{
	  id = STree_scopy (getextidentrepid (list));
	  treeList = ntreeList_appendLast (treeList, id);
	  list = succid (list);
	}
    }
  return treeList;
}

/**
   @return a list tree of new identifiers built from the list operand (one by element).
*/
ntrees SME_SyntaxRestriction::mkListTreeNewIdentifiers (ptlrepid list)
{
  ntrees
    treeList,
    id;

  treeList = LL_make_leflot_emptyList ();
  while (list != NILLREPID)
    {
      id = getNewUndeclaredIdent (STree_scopy (getextidentrepid (list)));
      treeList = ntreeList_appendLast (treeList, id);
      list = succid (list);
    }
  return treeList;
}

/**
   @return a new name with id as prefix. An internal counter (cptIds) 
   of the class is used to produce this identifier to ensure the unicity.
*/
ntrees SME_SyntaxRestriction::getNewUndeclaredIdent (ntrees id)
{
  ptIdStack
    name;
  int
    i;
  boolean
    b;

  if (eq_idents (id, idNil))
    {
      cptIds++;
      id = makeIdent_ident_int (id, cptIds);
      i = cptIds;
      b = pk_true;
    }
  else
    {
      i = 0;
      b = pk_false;
    }
  while (pk_true)
    {
      if (locate (id, &name))
	if (kl_isEmptyIdStack (name))
	  {
	    if (b)
	      cptIds = i;
	    return id;
	  }
      i++;
      id = makeIdent_ident_int (id, i);
    }
}

/**
   @return "true" if the tree is modified, "false" otherwise.
   Rules:
   - x + P{}(x1, ..., xm) is rewritten into (| x + x' | x':= P{}(x1, ..., xm) |)
   - P{}() is rewritten into (o1,...,on):= p{} (i1,...,in) where the outputs/inputs identifiers are those of the declaration of the model P.
*/
boolean SME_SyntaxRestriction::mkXmiCallTree (ntrees p, ptklExprProc vklExprProc)
{
  boolean
    b;
  ntrees
    f,
    vouts,
    np,
    wh;
  noperators
    op,
    opf;
  ptlrepid
    inputs,
    outputs;
  ptrepfi
    vfi;
  int
    nb,
    i;

  op = noper (p);
  f = nfather (p);
  opf = noper (f);
  if ((opf == opdefsig) || (opf == opdefpart) || (opf == opdefpartdef))
    return pk_false;
  b = (opf == oplfeed) || (opf == oppetiq);
  if (op == opappe)
    {
      vfi = getDeclarationModel (vklExprProc);
      outputs = kl_getInsOutsModel (vfi, &inputs);
      if (b)
	vouts = mkListTree (outputs, opleflot);
      else
	vouts = mkListTreeNewIdentifiers (outputs);
      if (outputs == NILLREPID)
	return pk_false;		// Added July 2009.
      nb = ntreearity (vouts);
      if (nb == 1)
	vouts = nchild (1, vouts);
      i = nposit (p);
      np = p;
      if (b)
	{
	  if (nb > 0)
	    {
	      np = LL_make_defsig_tree (vouts, np);
	      nchangchild (f, np, i);
	      return pk_true;
	    }
	  return pk_false;		// well formed call
	}
      // example: x + P{}(x1, ..., xm) ; rewritten into (| x + x' | x':= P{}(x1, ..., xm) |)


      np = LL_make_defsig_tree (vouts, np);
      vouts = STree_scopy (vouts);
      nchangchild (f, vouts, i);
      insertBlock (f, np, mkldecl (vouts));
      return pk_true;
    }

  if (op != opinst)
    return pk_false;

  if (b || (opf == opappe))	// opf = opappe: cas P{}() no inputs 
    {
      // example (| P{} ....
      vfi = getDeclarationModel (vklExprProc);
      outputs = kl_getInsOutsModel (vfi, &inputs);
      if (b)
	{
	  i = nposit (p);
	  np = LL_make_appe_tree (p, mkListTree (inputs, oplesign));
	  wh = f;
	}
      else
	{
	  i = nposit (f);
	  wh = nfather (f);
	  np = f;		// STree_scopy(f);

	}
      if (outputs != NILLREPID)
	{
	  vouts = mkListTree (outputs, opleflot);
	  if (ntreearity (vouts) == 1)
	    vouts = nchild (1, vouts);
	  np = LL_make_defsig_tree (vouts, np);
	}
      nchangchild (wh, np, i);
    }
  else
    return pk_false;		// error?
  return pk_true;
}

/**
   @return "true" if p is an event expression, elem is the argument of p (the rule is recursively applied), "false" otherwise.
*/
boolean
SME_SyntaxRestriction::simplifyEvent (ntrees p, ntrees * elem)
{
  noperators op;

  op = noper (*elem);
  if (op != opevnt)
    return pk_false;
  do
    {
      *elem = nchild (1, *elem);
      op = noper (*elem);
    }
  while (op == opevnt);
  return pk_true;
}

/**
   @return the tree "true when <t>"; used to eliminate the unary extraction.
*/
ntrees
SME_SyntaxRestriction::mkWhen1 (ntrees t)
{
  return LL_make_extr_tree (STree_mkNullaryTree (optrue), t);
}

/**
   @return the tree "^t".
*/
ntrees
SME_SyntaxRestriction::mkEvent (ntrees t)
{
  return LL_make_evnt_tree (t);
}

/**
   It rewrites an unary expression (p).
   Rules:
   - the unary when is replaced by a binary one 
   - for "E when (E1 default E2)", a signal is added for the second operand (default operator only, the meta model is not homogeneous)
   - "not pk_true (resp. not pk_false)" is rewritten in "true (resp. pk_false)".
   - "not (true when C)" is rewritten in "false when C", "not (false when C)" in "true when C"
*/
void
SME_SyntaxRestriction::xmiSyntaxExpression1 (ntrees p)
{
  noperators op, vop;
  ntrees elem, vout, s;
  int i;

  vop = noper (p);
  if (LL_is_atomOp (vop))
    return;
  if (vop == opvvarin)		// X? is replaced by X
    {
      elem = nfather (p);
      i = nposit (p);
      nchangchild (elem, nchild (1, p), i);
    }
  else if (vop == opwhen1)
    {
      elem = nfather (p);
      i = nposit (p);
      vout = mkWhen1 (nchild (1, p));
      nchangchild (elem, vout, i);
    }
  else if (vop == opextr)	// E when ( E1 default E2)
    {
      elem = nchild (2, p);
      op = noper (elem);
      if (isIdentifierOp (op) || isRelationalOp (op) || isLogicalOp (op))
	return;
      vout = getNewUndeclaredIdent (STree_scopy (idNil));
      nchangchild (p, STree_scopy (vout), 2);
      addEquation (vout, elem);
    }
  else if (vop == opnon)
    {
      elem = nchild (1, p);
      op = noper (elem);
      if (op == optrue)
	nsetop (elem, opfalse);
      else if (op == opfalse)
	nsetop (elem, optrue);
      else if (op == opextr)
	{
	  s = nchild (1, elem);
	  op = noper (s);
	  if (op == optrue)
	    {
	      nsetop (s, opfalse);
	      nchangchild (nfather (p), elem, nposit (p));
	    }
	  else if (op == opfalse)
	    {
	      nsetop (s, optrue);
	      nchangchild (nfather (p), elem, nposit (p));
	    }
	}
    }
}

/**
   It rewrites an clock expression (p) used in a synchronisation expression ( ^=, ^>,....).
   The meta model accepts only the clock operators (^+, ^*, ^-, ^) in such a context.
*/
_VIRTUAL_ void
SME_SyntaxRestriction::xmiSyntaxClockExpression (ntrees p)
{
  ntrees elem, vout, s;
  noperators op, vop;
  boolean bce, b;
  int i;

  if (ISNOT_NILTREE (p))
    {
      xmiSyntaxExpression1 (p);
      vop = noper (p);
      if (LL_is_atomOp (vop))
	return;
      bce = isClockOperator (vop) || isClockRelationOperator (vop);
      for (i = 1; i <= ntreearity (p); i++)
	{
	  elem = nchild (i, p);
	  op = noper (elem);
	  b = pk_false;
	  if (vop == opevnt)
	    {
	      if (simplifyEvent (p, &elem))
		{
		  nchangchild (p, elem, i);
		  op = noper (elem);
		}
	      b = !isIdentifierOp (op);
	    }
	  else if (bce)
	    {
	      if (simplifyEvent (p, &elem))
		{
		  nchangchild (p, elem, i);
		  op = noper (elem);
		}
	      if (op != ophnull)
		if (!isIdentifierOp (op))
		  /* if (op == opmele) nsetop(elem, opbsup); else */ b = !isClockOperator (op);
	    }
	  else if (op != ophnull)
	    if (!isIdentifierOp (op))
	      b = isClockOperator (op) /*|| (op == opwhen1) */ ;
	  if (b)
	    {
	      vout = getNewUndeclaredIdent (STree_scopy (idNil));
	      nchangchild (p, STree_scopy (vout), i);
	      addEquation (vout, elem);
	    }
	}
    }
}

/**
   It rewrites the clock expressions (p) not used in a synchronisation expression ( ^=, ^>,....).
   The meta model accepts only the clock operators (^+, ^*, ^-, ^) in such a context.
*/
_VIRTUAL_ void
SME_SyntaxRestriction::xmiSyntaxExpression (ntrees p)
{
  ntrees x;
  noperators op, vop;
  boolean bce, b;
  int i;

  if (ISNOT_NILTREE (p))
    {
      xmiSyntaxExpression1 (p);
      vop = noper (p);
      if (LL_is_atomOp (vop))
	return;
      for (i = 1; i <= ntreearity (p); i++)
	xmiSyntaxExpression (nchild (i, p));

      switch ((int) (vop))
	{
	case opcomp:
	  i = nposit (p);
	  x = mkWhen1 (LL_make_mele_tree (LL_make_non_tree (mkEvent (nchild (2, p))), mkEvent (nchild (1, p))));
	  nchangchild (nfather (p), x, i);
	  break;

	case opbinf:
	  i = nposit (p);
	  x = LL_make_extr_tree (mkEvent (nchild (1, p)), mkEvent (nchild (2, p)));
	  nchangchild (nfather (p), x, i);
	  break;

	case opbsup:
	  i = nposit (p);
	  x = LL_make_mele_tree (mkEvent (nchild (1, p)), mkEvent (nchild (2, p)));
	  nchangchild (nfather (p), x, i);
	  break;

	case ophnull:		// pk_true when false.
	  break;
	}
    }
}

/**
   It adds 
   - the equation id:= rp to a composition (lfeed), created if it is undefined.
   - id to the list of identifiers (lids), created if it is undefined.
*/
void
SME_SyntaxRestriction::addEquation (ntrees id, ntrees rp)
{
  ntrees aDef;

  if (IS_NILTREE (lids))
    lids = LL_make_lsign_emptyList ();
  lids = ntreeList_appendLast (lids, STree_scopy (id));
  aDef = LL_make_defsig_tree (STree_scopy (id), rp);
  if (IS_NILTREE (lfeed))
    lfeed = LL_make_lfeed_emptyList ();
  lfeed = ntreeList_appendLast (lfeed, aDef);
}

/**
   Local to SME_SyntaxRestriction::cExpType().
 */
void
SME_SyntaxRestriction::addDecTypeVect (ntrees p, noperators voptype, ntrees ldeclTypes)
{
  if (voptype == optvect)
    {
      do
	{
	  p = nchild (2, p);
	  voptype = noper (p);
	}
      while (voptype == optvect);
      if ((voptype == optenum) || (voptype == opstru) || (voptype == opbund))
	addDecType (p, nfather (p), voptype, ldeclTypes);
    }
}

/**
   Local to SME_SyntaxRestriction::cExpType().
*/
void
SME_SyntaxRestriction::addDecType (ntrees p, ntrees f, noperators voptype, ntrees ldeclTypes)
{
  ntrees f1, name, typeDecl;
  int i;
  noperators op;
  boolean found;

  if (voptype != optvect)
    {
      i = nposit (p);
      name = getNewUndeclaredIdent (STree_scopy (idNil));
      nchangchild (f, STree_scopy (name), i);

      typeDecl = STree_mkDclType (name, p, NILTREE);

      if (IS_NILTREE (ldeclTypes))
	{
	  found = pk_false;
	  while (!found)
	    {
	      f1 = nfather (f);
	      op = noper (f1);
	      if (STree_isModel (f1))
		found = pk_true;
	      else if ((op == oppconf) || (op == opdesc) || (op == opltype))
		found = pk_true;
	      else
		f = f1;
	    }
	  if (op == opltype)
	    {
	      ldeclTypes = f1;
	      ldeclTypes = ntreeList_insertFirst (typeDecl, ldeclTypes);
	    }
	  else
	    {
	      ldeclTypes = STree_addTypeToTypes (NILTREE, typeDecl);
	      if (op == oppconf)
		STree_addDeclarationToConfinedExpr (f1, ldeclTypes);
	      else if (op == opdesc)
		{
		  f1 = LL_make_pconf_tree (nchild (2, f1), LL_make_ldecl_emptyList ());;
		  STree_addDeclarationToConfinedExpr (f1, ldeclTypes);
		}
	      else
		{		// a unnamed type in the interface.
		  f = nfather (f1);
		  if (IS_NILTREE (f))
		    mkklErrorAnnotTree (errTypeRestriction, f1);	// it will be necessary to make a module.
		  else
		    f = ntreeList_insertFirst (ldeclTypes, f);
		}
	    }
	}
      else
	ldeclTypes = ntreeList_insertFirst (typeDecl, ldeclTypes);

      if ((voptype == opstru) || (voptype == opbund))
	{
	  ntrees l = p;
	  if (firstChild (&l))
	    do
	      if ((voperator == optenum) || (voperator == opstru) || (voperator == opbund))
		addDecType (operande, p, voperator, ldeclTypes);
	      else
		addDecTypeVect (p, voptype, ldeclTypes);
	    while (nextChild (&l));
	}
    }
  else
    addDecTypeVect (p, voptype, NILTREE);
}


/**
   Compiling a type, specialisation of Trees2Kernel::cExpType().

   Specialisation: 
   Because the meta model accepts only the named types, the unnamed types referenced in a type (p) must be declared.
   These declarations are added in addDecType() and addDecTypeVect() methods.
   
   example:
   \verbatim
   type titi= struct
   ( struct
      ( int x;
	real z; 
	) U;
      int x;
      real z; 
     );
   \endverbatim
   is rewritten in
   \verbatim
   type _XMI_Id_1 = struct ( int x;
                             real z; ),
             titi = struct ( _XMI_Id_1 U;
	                     int x;
                             real z; );
   \endverbatim
*/
_VIRTUAL_ ptreptyp SME_SyntaxRestriction::cExpType (ntrees p)
{
  ptreptyp
    deftyp;
  ntrees
    f;
  noperators
    voperator;

  deftyp = Trees2Kernel::cExpType (p);
  f = nfather (p);
  voperator = noper (p);
  if (noper (f) != optype)
    {				// Adding a declaration of type for struct, bundle, enum types.
      if ((voperator == optenum) || (voperator == opstru) || (voperator == opbund) || (voperator == optvect))
	addDecType (p, f, voperator, NILTREE);
    }
  else if (voperator == optvect)
    {
      do
	{
	  p = nchild (2, p);
	  voperator = noper (p);
	}
      while (voperator == optvect);
      if ((voperator == optenum) || (voperator == opstru) || (voperator == opbund))
	addDecType (p, nfather (p), voperator, NILTREE);
    }
  return (deftyp);
}

/**
   Compiling a labelled expression (p), specialisation of Trees2Kernel::cExpLabelledExp().

   Specialisation:
   - the classical compiling is applied: see Trees2Kernel::cExpLabelledExp().
   - because in the meta model, the label is associated with a subprocess, 
   an expression "e :: Expr", and Expr is not a composition, is rewritten in e ::(| Expr |)
*/
_VIRTUAL_ boolean SME_SyntaxRestriction::cExpLabelledExp (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean
    b;
  ntrees
    s2;

  b = Trees2Kernel::cExpLabelledExp (p, inputs, outputs, vklExprProc);
  s2 = nchild (2, p);
  if (noper (s2) != oplfeed)
    nchangchild (p, LL_make_lfeed_list (s2), 2);
  return b;
}

/**
   Compiling a labelled expression (p), specialisation of Trees2Kernel::cExpCase().
   
*/
_VIRTUAL_ boolean SME_SyntaxRestriction::cExpCase (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ntrees
    p_case,
    p_elem;
  int
    i,
    nb;
  boolean
    b;

  b = Trees2Kernel::cExpCase (p, inputs, outputs, vklExprProc);
  p_case = nchild (2, p);
  nb = ntreearity (p_case);
  for (i = 1; i <= nb; i++)
    {				/* liste d'enumerations du case:  "valeur"": "expression de processus" */
      p_elem = nchild (i, p_case);
      if (isChild (2, p_elem))
	if (noper (operande) != oplfeed)
	  nchangchild (p_elem, LL_make_lfeed_list (operande), 2);
    }
  if (isoChild (3, p))		/* partie else du case */
    if (noper (operande) != oplfeed)
      nchangchild (p, LL_make_lfeed_list (operande), 3);

  return b;
}

/**
   Local to rewritListEleVec().

   @return a confined expression composed of the iterate structure built from the operand tree (t) and the definion of the associated indexes.
   The operand tree (t) is (assumed) defined by iterations built from an vectorial definition
   - { i in bi1 to bs1 step s1, .....}: []: expression
   or 
   - { i in bi1 to bs1 step s1, .....}: expression

   Example:
   \verbatim
   I:= [ { i to 14 step 1 } : [i+3]:i+i/2 ] 
   \endverbatim
   is rewritten in
   \verbatim
   (| i := 0..14 step 1
    | iterate i
      of I[i+3] := i+(i/2)
      end
    |) where i; end 
   \endverbatim
*/
ntrees SME_SyntaxRestriction::makeIterate (ntrees t, ntrees ensvec)
{
  ntrees
    x,
    lindexes,
    p,
    vdef,
    last,
    iterateTree,
    bounds,
    le,
    lf,
    lp,
    v;
  ntrees
    vindex,
    bi,
    bs,
    step;
  int
    nb,
    i;
  noperators
    op;

  lindexes = LL_make_lsign_emptyList ();
  lf = LL_make_lfeed_emptyList ();
  lp = nchild (1, ensvec);
  nb = ntreearity (lp);
  for (i = 1; i <= nb; i++)
    {
      p = nchild (i, lp);
      if (isoChild (1, p))
	vindex = operande;
      else
	vindex = getNewUndeclaredIdent (STree_scopy (idNil));
      bounds = nchild (2, p);
      if (isoChild (1, bounds))
	bi = operande;
      else
	bi = LL_make_ent_atom (integer2repstr (0));
      if (isoChild (2, bounds))
	bs = operande;
      else
	notOK (bounds);
      if (isoChild (3, bounds))
	step = operande;
      else
	step = LL_make_ent_atom (integer2repstr (1));

      lf = ntreeList_appendLast (lf, LL_make_defsig_tree (STree_scopy (vindex), nmaketernary (openum, bi, bs, step)));
      lindexes = ntreeList_appendLast (lindexes, vindex);
    }
  nb = ntreearity (lindexes);
  SET_NILTREE (iterateTree);
  for (i = nb; i > 0; i--)
    {
      iterateTree = nmaketernary (opiter, STree_scopy (nchild (i, lindexes)), iterateTree, STree_mkAtomNilTree ());
      if (i == nb)
	last = iterateTree;
    }
  lf = ntreeList_appendLast (lf, iterateTree);
  v = nchild (2, ensvec);
  op = noper (v);

  if (op == oprest)		// [I,J]:E
    {
      le = LL_make_lesign_emptyList ();
      x = nchild (1, v);	// oplevect --> oplesign conversion.
      for (i = 1; i <= ntreearity (x); i++)
	le = ntreeList_appendLast (le, nchild (i, x));
      vdef = LL_make_elem_tree (t, le);	// t[I,J]
      x = nchild (2, v);
    }
  else
    {
      vdef = LL_make_elem_tree (t, lindexes);	// lindexes devrait etre une "l_e_sign"
      x = v;
    }

  vdef = LL_make_defsig_tree (vdef, x);	// t[I,J] := x;
  nchangchild (last, vdef, 2);
  return addConfinedExpression (lf, STree_getImplicitTypeTree (), STree_scopy (lindexes));
}

/**
   @return "true" if an tree (p) is a sub-tree of a declaration, "false" otherwise.
*/
boolean SME_SyntaxRestriction::isInADeclaraction (ntrees p)
{
  noperators
    op;
  ntrees
    f;

  f = nfather (p);
  op = noper (f);
  if (op == opinit)
    return pk_true;
  if (op == opdeccon)
    return pk_true;
  if ((op == opret) || (op == opfen))
    {
      if (nchild (2, f) == p)
	return pk_true;
      if (nchild (3, f) == p)
	return pk_true;
    }
  if (op == oplfeed)
    return pk_false;
  if (op == opdefsig)
    return pk_false;
  return isInADeclaraction (f);
}

/**
   It rewrites a definition of an array such that [ {i to n}: E1, [E2]:E3 ] using the 'iterate' and 'next' operators.

   See makeIterate().
*/
void
SME_SyntaxRestriction::rewritListEleVec (ntrees p)
{
  ntrees xmodule, sdef, vdef, newId, vnext, lf, v, s, f, vtype, vdecl, lident;
  noperators opc, op;
  int i, nb, rank;
  typid proto;
  typval idClass;

  if (!isInADeclaraction (p))
    {
      // By construction (syntactic rule): [E]:E is always after an "ensvec": [ {i to n}: E1, [E2]:E3 ]

      f = nfather (p);
      op = noper (f);
      if ((op != opdefsig) && (op != opdefpart))
	notOK (p);		// Restriction.
      else
	{
	  rank = nposit (p);
	  nb = ntreearity (p);
	  if (nb == 1)
	    {
	      s = nchild (1, p);
	      opc = noper (s);
	      // if (opc == opelevec) nchangchild(f, s, 2);  It cannot occur in this case, it is always after a "opensvec".
	      if (opc == opensvec)
		{
		  v = makeIterate (nchild (1, f), s);
		  nchangchild (nfather (f), v, nposit (f));
		}
	      else
		notOK (operande);
	    }
	  else
	    {
	      lf = LL_make_lfeed_emptyList ();
	      lident = LL_make_lsign_emptyList ();
	      SET_NILTREE (vnext);
	      for (i = 1; i <= nb; i++)
		{
		  s = nchild (i, p);
		  opc = noper (s);
		  if (opc == opelevec)	// [E]:E after a opensvec (always): [ {i to n}: E1, [E2]:E3 ]
		    {
		      vdef = STree_scopy (s);
		      s = nchild (1, vdef);
		      if (noper (s) == oplesign)
			nsetop (s, opleflot);
		      nsetop (vdef, oprest);
		      vnext = LL_make_sequ_tree (vnext, vdef);
		    }
		  else if (opc == opensvec)
		    {
		      newId = getNewUndeclaredIdent (STree_scopy (idNil));
		      v = makeIterate (newId, s);
		      lf = ntreeList_appendLast (lf, v);
		      newId = STree_scopy (newId);
		      if (IS_NILTREE (vnext))
			vnext = newId;
		      else
			vnext = LL_make_sequ_tree (vnext, newId);
		      lident = ntreeList_appendLast (lident, STree_scopy (newId));
		    }
		  else
		    notOK (operande);
		}
	      sdef = nchild (1, f);
	      vdef = LL_make_defsig_tree (sdef, vnext);
	      lf = ntreeList_appendLast (lf, vdef);
	      vdef = STree_getIdentOf (vdef);
	      if (getTreeDecl (sdef, &vdecl, &xmodule, &proto, &idClass))
		if (getTypeOf (vdecl, &vtype))
		  vtype = STree_scopy (vtype);
		else
		  vtype = STree_getImplicitTypeTree ();
	      else
		vtype = STree_getImplicitTypeTree ();
	      lf = addConfinedExpression (lf, vtype, lident);
	      nchangchild (nfather (f), lf, nposit (f));
	    }
	  // notOK(p);
	}
    }
}

/**
   Compiling a list of expressions (p), specialisation of Trees2Kernel::cExpList().
   Specialisation:
   - the classical compiling is applied: see Trees2Kernel::cExpList().
   - it calls rewritListEleVec() for a list of vectorial elements (oplelevec operator).
*/
_VIRTUAL_ ptklExprSig SME_SyntaxRestriction::cExpList (ntrees * p, noperators ope, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig
    vexp_res,
    vklExprSig1,
    vklExprSig2;
  ntrees
    x;

  x = *p;
  vexp_res = Trees2Kernel::cExpList (p, ope, les, lparam, indexList);
  if (noper (x) == oplelevec)
    rewritListEleVec (x);
  return (vexp_res);
}

/**
   Compiling a list of dependences (p), specialisation of Trees2Kernel::cExplDependences().

   Specialisation:
   - the classical compiling is applied: see Trees2Kernel::cExplDependences().
   - It rewrites the dependence 
   \f$ L_SIG_1 --> L_SIG_2 --> ...--> L_SIG_n \f$
   in
   \f$ (| L_SIG_1 --> L_SIG_2 | L_SIG_2 --> L_SIG_3 ... |  L_SIG_n-1 --> L_SIG_n |) \f$
   because the meta model does not accept the first one.
*/
_VIRTUAL_ boolean SME_SyntaxRestriction::cExplDependences (ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  boolean
    b;
  int
    i,
    nb,
    rank;
  ntrees
    dep,
    f;

  b = Trees2Kernel::cExplDependences (p, inputs, vklExprProc);
  nb = ntreearity (p);
  if (nb > 2)
    {
      rank = nposit (p);
      f = nfather (p);
      for (i = 1; i < nb; i++)
	{
	  dep = LL_make_ldep_list (nchild (i, p));
	  dep = ntreeList_appendLast (dep, STree_scopy (nchild (i + 1, p)));
	  ninsertright (rank, dep, f);
	}
      ntreesDestroy (p);
    }
  return b;
}

/**
   Compiling a masking (p), specialisation of Trees2Kernel::cExpMasking().

   Specialisation:
   - the classical compiling is applied: see Trees2Kernel::cExpMasking(). 
   - The masking is replaced by a confined expression: \f$ Q/x1,...xn \f$ is replaced by \f$ (| Q |) where x1,...xn end\f$
*/
_VIRTUAL_
  boolean
SME_SyntaxRestriction::cExpMasking (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;
  ntrees lident, t;

  b = Trees2Kernel::cExpMasking (p, inputs, outputs, vklExprProc);
  lident = nchild (2, p);
  nsetop (p, oppconf);
  t = LL_make_ldecl_list (LL_make_decsig_tree (STree_getImplicitTypeTree (), mkldecl (lident)));
  nchangchild (p, t, 2);
  return b;
}

/**
   Specialisation of Trees2Kernel::initklparcours().

   It reinits the global objects(lids, lfeed) used to define the new equations when auxillary signals (lids) are created.
   The created equations/identifiers are added in the current compiled model in SME_SyntaxRestriction::finalklparcours().
 */
_VIRTUAL_ void
SME_SyntaxRestriction::initklparcours (ntrees p, noperators op)
{
  ntrees sl, slf;

  //  if (isXmiBlock(op)) 
  {
    // push
    sl = lids;
    slf = lfeed;

    // reinit
    SET_NILTREE (lids);
    SET_NILTREE (lfeed);
    // pop
    lids = sl;
    lfeed = slf;
  }
}


/**
   Specialisation of Trees2Kernel::finalklparcours().

   It adds the list of added equations (lfeed) and the created signals (lids) to the current compiled tree (p).
   The initialisation of the used objects (lfeed, lids) are initialized in SME_SyntaxRestriction::initklparcours().
 */
_VIRTUAL_ void
SME_SyntaxRestriction::finalklparcours (ntrees p, noperators op)
{
  if (ISNOT_NILTREE (lids))
    //   if (isXmiBlock(op))
    {
      insertBlock (p, lfeed, lids);
      SET_NILTREE (lids);
      SET_NILTREE (lfeed);
    }
}


/**
  Compiling the import objects (p), specialisation of Trees2Kernel::cDecImports().
  In the meta model, only one "use" construction is allowed for a model.

  Specialisation:
  - the classical compiling is applied: see Trees2Kernel::cDecImports().
  - it replaces a set of \f$ use M1; ... use Mn; \f$ by one \f$ use M1,...Mn; \f$ keeping the order.
*/
_VIRTUAL_ ptlrepid
SME_SyntaxRestriction::cDecImports (ntrees p)
{
  ntrees list, limports, firstUse;
  int i;
  ptlrepid lvfiModules;

  lvfiModules = Trees2Kernel::cDecImports (p);
  list = p;
  if (!isUndefinedBodyTree (list))
    {
      SET_NILTREE (firstUse);
      if (firstChild2 (&list))
	do
	  {
	    if (voperator == oplimpo)	// use "M1, M2, ....Mn"
	      if (IS_NILTREE (firstUse))
		firstUse = operande;
	      else
		{
		  limports = operande;
		  for (i = 1; i <= ntreearity (limports); i++)
		    firstUse = ntreeList_appendLast (firstUse, nchild (i, limports));
		  ntreesDestroy (limports);
		}
	  }
	while (nextChild (&list));
    }
  return lvfiModules;
}

/**
   It searches the occurrences of a label (labeldecl), in tree (inTree).
   @param vdef  (out) is the definition of the label, (the labelled process by labeldecl)
   @param nb (out) the number of references of the label (vdef excluded)

   Remark : the algorithm stops when vdef is defined and nb >0.
 */
void
SME_SyntaxRestriction::labelReferenced (ntrees labeldecl, ntrees inTree, ntrees * vdef, int *nb)
{
  noperators voperator;
  int i, n;
  ntrees p;

  if (ISNOT_NILTREE (inTree))
    {
      voperator = noper (inTree);
      if ((voperator == opnom) || (voperator == opident))
	{
	  if (eq_idents (labeldecl, inTree))
	    {
	      p = nfather (inTree);
	      if ((noper (p) == oppetiq) && (nposit (inTree) == 1))	// p is a tree such that "labeldecl :: Eproc"
		{
		  (*vdef) = p;
		}
	      else
		(*nb) = (*nb) + 1;
	    }
	}
      else
	{
	  if (LL_is_atomOp (voperator))
	    return;
	  n = ntreearity (inTree);
	  for (i = 1; i <= n; i++)
	    if (((*nb) == 0) || (IS_NILTREE (*vdef)))
	      labelReferenced (labeldecl, nchild (i, inTree), vdef, nb);
	}
    }
}

/**
   @return "true" if the label is a used as a key word in the Signal program, false otherwise.
   List of the kept label : _constraints_.
*/
_VIRTUAL_ boolean
SME_SyntaxRestriction::iskeptLabel (ntrees labeldecl)
{
  if (eq_idents (labeldecl, string2ident ("_constraints_")))
    return pk_true;
  return pk_false;
}

/**
   It compiles the declaration of a label (labeldecl).
   @return letiq + the reached label.
*/
_VIRTUAL_ ptlrepid
SME_SyntaxRestriction::cDecLabel (ntrees labeldecl, ptlrepid letiq)
{
  ntrees pconf, vdef;
  int n;

  letiq = Trees2Kernel::cDecLabel (labeldecl, letiq);
  // it "definition" (E::P) is removed (replaced by P )if it is not referenced.
  // The label _constraints_ is kept.
  if (!iskeptLabel (labeldecl))
    if (upUntil (labeldecl, oppconf, &pconf))
      {
	SET_NILTREE (vdef);
	n = 0;
	labelReferenced (labeldecl, nchild (1, pconf), &vdef, &n);
	if ((n == 0) && (ISNOT_NILTREE (vdef)))
	  // e ::Eproc  replaced by Eproc 
	  nchangchild (nfather (vdef), nchild (2, vdef), nposit (vdef));
      }
  return letiq;
}
