/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define XmlVersion  "1.0"
#define SSMEVersion "2.0"
#define xmlns_xmi   "http://www.omg.org/XMI"
#define xmlns_xsi   "http://www.w3.org/2001/XMLSchema-instance"
#define xmlns_ssme  "http://www.irisa.fr/espresso/ssme.ecore"
#define UTF_str     "UTF-8"

#define comBegin "<!--"
#define comEnd "-->"
#define comincom " "
#define comincomend " "
#define errSymbol "ERROR"
#define cstSymbolHolo "..."

#define pushmodes(tk, babs) tksave = currentnamelist; abssave = abstractMode; currentnamelist = (char*) tk; abstractMode = babs;
#define popmodes( )  currentnamelist = tksave; abstractMode = abssave;


ptSSME_Printer pK_SSME_Printer = new SSME_Printer ();

_PROTECTED_ boolean SSME_Printer::isfathera (noperators vop)
{
  ntrees
    t;
  if (IS_NILTREE (treeunparseTree))
    return pk_false;
  t = nfather (treeunparseTree);
  return (boolean) (noper (t) == vop);
}

void
SSME_Printer::printError ()
{
  printXSI (errSymbol);		//does not exist in SSME ...
}

void
SSME_Printer::printString (const char *s)
{
  if (s != NULL)
    sigpristring ((char *) s);
}

/**
   Global initialisation of the printer.
   Some of the initialisations are done in reset().
*/
void
SSME_Printer::GeneralInit ()
{
  reset ();
  currentnamelist = NULL;
  abstractMode = pk_false;
}


/**
   Printer initialisations:
   - An \ref intePrinter object is associated with the printer, if it is not yet done.
   - A "printer mode" is defined in this interface. It allows to print or not the comments.
   
   For a graphical mode, the comments are not printed, and for a textual one, they are printed.
   By default, the "textual mode" is set. To change it, use intePrinter::setPrinterContext() method.

   It loads:
   - the Signal-SME table of keywords, called SIGNALSSME.kw, defined in $Signal_bin directory.
   This table contains the redefined key words of SIGNAL.
   - a specific table (SMEKEYWORDS.kw defined in $Signal_bin directory)

   @return "true" if the operations succeed, "false" otherwise.
*/
_VIRTUAL_ boolean
SSME_Printer::initPrinter ()
{
  if (notannounced)
    {
      notannounced = pk_false;
      this_intePrinter = new intePrinter (string2repstr (comBegin), string2repstr (comEnd),
					  string2repstr (comincom), string2repstr (comincomend),
					  string2repstr (errSymbol), string2repstr (cstSymbolHolo));
      /* Loading tables (SIGNALSSME.kw, defined in the $Signal_bin (data) directory).  THERE IS NO TABLE FOR SSME. */
      // if (!this_intePrinter->load_KWLangTable("SIGNALSSME")) return pk_false;
      this_intePrinter->setPrinterContext (pk_true);
    }
  return pk_true;
}


_VIRTUAL_ void
SSME_Printer::closeblock ()
{
  sncharpri ('>');
  snline ();
}

_VIRTUAL_ void
SSME_Printer::openblock ()
{
  sncharpri ('<');
}

/**
   It initialises the printer for printing a XML string (some characters needs codification).
   \see setCharEncodingMode().
*/
void
SSME_Printer::beginString ()
{
  sncharpri ('"');
  setCharEncodingMode (pk_false);	// XML characters.
}

/**
   It finalyzes the printer for printing a XML string, the XML mode is previously opened in beginString().
   \see setCharEncodingMode().
*/
void
SSME_Printer::endString ()
{
  sncharpri ('"');
  setCharEncodingMode (pk_true);	// std characters.
}


void
SSME_Printer::printQuotedStr (const char *vstr)
{
  beginString ();
  sigpristring ((char *) vstr);
  endString ();
}


void
SSME_Printer::printkwB (const char *kw)
{
  openblock ();
  printString (kw);
  closeblock ();
}

void
SSME_Printer::printkwE (boolean bsimple, const char *kw)
{
  if (bsimple)
    {
      sncharpri ('/');
      closeblock ();
    }
  else
    {
      openblock ();
      sncharpri ('/');
      printString (kw);
      closeblock ();
    }
}

_VIRTUAL_ void
SSME_Printer::printXSI (const char *str)
{
  if (str != NULL)
    {
      snspace ();
      printString ("xsi:type");
      sncharpri ('=');
      beginString ();
      printString ("ssme");
      sncharpri (':');
      printString (str);
      endString ();
      snspace ();
    }
}

_VIRTUAL_ void
SSME_Printer::printXSI (ntrees t)
{
  if (STree_isOperatorModel (t))
    {
      printXSI (tkOperatorModel);
      closeblock ();
    }
  else if (STree_isModel (t))
    {
      printXSI (tkProcessModel);
      printModelAttributes (noper (t));
      closeblock ();
    }
}

boolean
SSME_Printer::isclassattribue (ntrees t)
{
  if (STree_isLeaf (t)) return pk_true;
  if (noper (t) == opident) return pk_true;
  return pk_false;
}

boolean
SSME_Printer::isFilteredOperator ()
{
  // incoherence ssme - signal : pas de tpe implicit en ssme ("undefined" n'est pas acceptÃ© en valeur de type-voir le mÃ©ta-modÃšle--).
  if (voperator == opimplicit)
    return pk_true;
  if (voperator == opretun)
    return pk_true;
  return pk_false;
}

/** Printing a list of suffixes(p). */
_VIRTUAL_ void
SSME_Printer::lsuffIdent (ntrees p, const char *tkw)
{
  int nbfils;
  int i;

  nbfils = ntreearity (p);
  for (i = 1; i <= nbfils; i++)
    {
      printString (tkw);
      inteconsprint (nchild (i, p));
    }
}

/** Printing an identifier(p) represented as a binary tree: first child is a leaf, second a list of suffixes.
 */
_VIRTUAL_ void
SSME_Printer::identprint (ntrees p)
{
  ntrees operand;

  printXSI (tkIdentifier);
  snspace ();
  printString ("atomString");
  sncharpri ('=');
  beginString ();
  snfit (this_intePrinter->sigidentsize (p) + 1);
  operand = nchild (1, p);
  inteconsprint (operand);
  operand = nchild (2, p);
  if (IS_NILTREE (operand))
    this_intePrinter->printerror ();
  else if (noper (operand) == opnil)
    this_intePrinter->printerror ();
  else if (noper (p) == opident)
    lsuffIdent (operand, "_");
  else
    lsuffIdent (operand, ".");	// ident of a signal associated with a field of a structure/bundle.
  endString ();
  snspace ();
}


/**
   model type printing without first keyword.
*/
_VIRTUAL_ void
SSME_Printer::printChild0 (int ichild, const char *tk, boolean babs)
{
  boolean b;
  char *tksave;
  boolean abssave;

  if (ischildprint (ichild))
    {
      pushmodes (tk, babs);
      if (STree_isList (operande))
	unparse_subtree (operande, vholoph);
      else if (!isFilteredOperator ())
	{
	  b = isclassattribue (operande);
	  openblock ();
	  printString (tk);
	  if (!b) if (babs) printXSI (operande); else closeblock ();
	  unparse_subtree (operande, vholoph);
	  printkwE (b, tk);
	}
      popmodes ();
    }
}

/**
   model type printing without first keyword.
*/
_VIRTUAL_ void
SSME_Printer::printChild (int ichild, const char *tk)
{
  printChild0 (ichild, tk, pk_false);
}

/**
   model type printing without first keyword.
*/
_VIRTUAL_ void
SSME_Printer::printChildAbs (int ichild, const char *tk)
{
  printChild0 (ichild, tk, pk_true);
}


void
SSME_Printer::printModelAttributes (noperators op)
{
  switch ((int) (op))
    {
    case opprocessp:
      printKind ("visibilityKind", "private");
      break;
    case opprocess:
      break;

    case opfunc:
      printKind ("modelKind", "function");
      break;
    case opfuncp:
      printKind ("visibilityKind", "private");
      printKind ("modelKind", "function");
      break;

    case opnode:
      printKind ("modelKind", "node");
      break;
    case opnodep:
      printKind ("visibilityKind", "private");
      printKind ("modelKind", "node");
      break;

    case opact:
      printKind ("modelKind", "action");
      break;
    case opactp:
      printKind ("visibilityKind", "private");
      printKind ("modelKind", "action");
      break;

    case opprocedure:
      printKind ("modelKind", "procedure");
      break;
    case opprocedurep:
      printKind ("visibilityKind", "private");
      printKind ("modelKind", "procedure");
      break;

    case opautomaton:
      printKind ("modelKind", "automaton");
      break;
    case opautomatonp:
      printKind ("visibilityKind", "private");
      printKind ("modelKind", "automaton");
      break;
    }
}

void
SSME_Printer::printOperator (noperators op)
{
  printString ("ssme");
  sncharpri (':');
  if (op == oplmodu)
    printString (tkListModule);
  else if (op == opmodu)
    printString (tkModule);
  else
    printString (tkProcessModel);
}

_VIRTUAL_ void
SSME_Printer::printKind (const char *kind, const char *valueKind)
{
  snspace ();
  printString (kind);
  sncharpri ('=');
  beginString ();
  printString (valueKind);
  endString ();
}



// ssmename="KEWORDTABLE(kw)"
_VIRTUAL_ void
SSME_Printer::printValue (const char *ssmename, const char *kw)
{
  snspace ();
  printString (ssmename);
  sncharpri ('=');
  beginString ();
  //  snglue(pk_false);this_intePrinter->printalphakey(kw);
  printString (kw);
  endString ();
  snspace ();
}


_VIRTUAL_ void
SSME_Printer::printObj0 (const char *ssmeclass, const char *ssmename, const char *kw)
{
  printXSI (ssmeclass);
  if (kw != NULL)
    printValue (ssmename, kw);
}

// Nullary tree.
_VIRTUAL_ void
SSME_Printer::printAtom (const char *xsi)
{
  printXSI (xsi);
  snspace ();
  printString ("atomString");
  sncharpri ('=');
  if ((voperator == opcstcha) || (voperator == opcstcar))
    inteconsprint (treeunparseTree);
  else
    {
      beginString ();
      inteconsprint (treeunparseTree);
      endString ();
    }
  snspace ();
}



/**
   Printing a nullary tree: opnom operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opnom ()
{
  printAtom (tkIdentifier);
}

_VIRTUAL_ void
SSME_Printer::printConstant (const char *ssmeclass)
{
  printAtom (ssmeclass);
}

/**
   Printing a nullary tree: opchar operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opchar ()
{
  printObj0 (tkPredefinedType, "type", kwchar);
}

/**
   Printing a nullary tree: opcomplex operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcomplex ()
{
  printObj0 (tkPredefinedType, "type", kwcomplex);
}

/**
   Printing a nullary tree: opcstcar operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcstcar ()
{
  printConstant (tkConstantCharacter);
}

/**
   Printing a nullary tree: opcstcha operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcstcha ()
{
  printConstant (tkConstantString);
}

/**
   Printing a nullary tree: opcstdp operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcstdp ()
{
  printConstant (tkConstantDoubleReal);
}

/**
   Printing a nullary tree: opcstent operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcstent ()
{
  printConstant (tkConstantInteger);
}

/**
   Printing a nullary tree: opcstreel operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcstreel ()
{
  printConstant (tkConstantReal);
}

/**
   Printing a nullary tree: opdcomplex operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opdcomplex ()
{
  printObj0 (tkPredefinedType, "type", kwdcomplex);
}


/**
   Printing a nullary tree: opdpreal operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opdpreal ()
{
  printObj0 (tkPredefinedType, "type", kwdreal);
}

/**
   Printing a nullary tree: opevent operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opevent ()
{
  printObj0 (tkPredefinedType, "type", kwevent);
}

/**
   Printing a nullary tree: opfalse operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opfalse ()
{
  printObj0 (tkConstantLogical, "constantLogicalKind", kwfalse);
}

/**
   Printing a nullary tree: ophnull operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_ophnull ()
{
  printXSI (tkNullClock);
}

/**
   Printing a nullary tree: opimplicit operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opimplicit ()
{
  printObj0 (tkPredefinedType, "type", NULL);
}

/**
   Printing a nullary tree: opinteger operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opinteger ()
{
  printObj0 (tkPredefinedType, "type", kwinteger);
}

/**
   Printing a nullary tree: oplogical operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_oplogical ()
{
  printObj0 (tkPredefinedType, "type", kwboolean);
}

/**
   Printing a nullary tree: oplong operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_oplong ()
{
  printObj0 (tkPredefinedType, "type", kwlong);
}

/**
   Printing a nullary tree: opnil operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opnil ()
{
}

/**
   Printing a nullary tree: opnullproc operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opnullproc ()
{
  printXSI (tkNullProc);
}

/**
   Printing a nullary tree: opop  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opop ()
{
  printXSI (tkOperator);
  snspace ();
  printString ("operatorKind");
  sncharpri ('=');
  beginString ();
  inteconsprint (treeunparseTree);
  endString ();
}

/**
   Printing a nullary tree: opreal  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opreal ()
{
  printObj0 (tkPredefinedType, "type", kwreal);
}

/**
   Printing a nullary tree: opretun  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opretun ()
{
}

/**
   Printing a nullary tree: opsafe  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opsafe ()
{
  printValue ("processAttributeKind", kwsafe);
}

/**
   Printing a nullary tree: opunsafe  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opunsafe ()
{
  printValue ("processAttributeKind", kwunsafe);
}

/**
   Printing a nullary tree: opdeterm operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opdeterm ()
{
  printValue ("processAttributeKind", kwdeterministic);
}

/**
   Printing a nullary tree: opshort  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opshort ()
{
  printObj0 (tkPredefinedType, "type", kwshort);
}

/**
   Printing a nullary tree: opstring  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opstring ()
{
  printObj0 (tkPredefinedType, "type", kwstring);
}

/**
   Printing a nullary tree: optabvide  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_optabvide ()
{
  sncharpri ('[');
  sncharpri (']');
}

/**
   Printing a nullary tree: optrue  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_optrue ()
{
  printObj0 (tkConstantLogical, "constantLogicalKind", kwtrue);
}


/**
   Printing a nullary tree: opcomment  operator. 
*/
_VIRTUAL_ void
SSME_Printer::decta0_opcomment ()
{
  // inteconsprint(treeunparseTree);
}


// Unary expressions

_VIRTUAL_ void
SSME_Printer::unsigexp (const char *vclass, const char *vstrchild)
{
  printXSI (vclass);
  closeblock ();
  printChildAbs (1, vstrchild);
}

_VIRTUAL_ void
SSME_Printer::unsigexp (const char *str)
{
  unsigexp (str, "exprSignal");
}

/**
	Printing an unary tree: opevnt operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opevnt ()
{
  unsigexp (tkClock);
}

/**
	Printing an unary tree: opexte operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opexte ()
{
  unsigexp (tkExternalAttribute, "name");
}

/**
	Printing an unary tree: oplab operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_oplab ()
{
  unsigexp (tkLabels, "listIdentifier");
}

/**
	Printing an unary tree: opnega operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opnega ()
{
  unsigexp (tkUnaryMinus);
}

/**
	Printing an unary tree: opnon operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opnon ()
{
  unsigexp (tkNot);
}

/**
	Printing an unary tree: oppcar operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_oppcar ()
{
  unsigexp (tkCartesianProduct, "listExprSignal");
}

/**
	Printing an unary tree: opref operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opref ()
{
  unsigexp (tkExternalReferences, "listIdentifier");
}

/**
	Printing an unary tree: opsdec operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opsdec ()
{
  unsigexp (tkPreviousIterationRef);
}

/**
	Printing an unary tree: opstru operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opstru ()
{
  printTupleType (pk_false);
}

/**
	Printing an unary tree: optran operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_optran ()
{
  unsigexp (tkTransposition);
}

/**
	Printing an unary tree: opvvarin operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opvvarin ()
{
  // x ?
  unsigexp (tkStateVariableValue, "identifier");
}

/**
   Printing an unary tree: opwhen1 operator.
*/
_VIRTUAL_ void
SSME_Printer::decta1_opwhen1 ()
{
  unsigexp (tkUnaryWhen);
}

/**
   Printing an unary tree: oppassert operator: assert(| ... |).
*/
_VIRTUAL_ void
SSME_Printer::decta1_oppassert ()
{
  unsigexp (tkAssertProcess, "processExpression");
}

/**
   Printing an unary tree: opappeassert operator: assert(C).
*/
_VIRTUAL_ void
SSME_Printer::decta1_opappeassert ()
{
  unsigexp (tkAssertSignal, "exprSignal");
}

// end Unary



_VIRTUAL_ void
SSME_Printer::binsigexpOperands (const char *ch1, const char *ch2)
{
  printChild (1, ch1);
  printChild (2, ch2);
}

_VIRTUAL_ void
SSME_Printer::binsigexpOperands (const char *ch1, const char *ch2, boolean b1, boolean b2)
{
  printChild0 (1, ch1, b1);
  printChild0 (2, ch2, b2);
}

_VIRTUAL_ void
SSME_Printer::printVisibility (noperators voperator)
{
  if ((voperator == opdecconp) || (voperator == opltypep))
    printKind ("visibility", "private");
}


_VIRTUAL_ void
SSME_Printer::printModelKind (noperators voperator)
{
  if (voperator == opformact)
    printKind ("modelKind", "action");
  else if (voperator == opformfunc)
    printKind ("modelKind", "function");
  else if (voperator == opformnode)
    printKind ("modelKind", "node");
  else if (voperator == opformproc)
    printKind ("modelKind", "process");
  else if (voperator == opformprocedure)
    printKind ("modelKind", "procedure");
  else if (voperator == opformautomaton)
    printKind ("modelKind", "automaton");
}

void
SSME_Printer::printXsiVisibility (const char *str)
{
  printXSI (str);
  printVisibility (voperator);
  printModelKind (voperator);
  closeblock ();
}


_VIRTUAL_ void
SSME_Printer::binsigexp (const char *str, const char *ch1, const char *ch2)
{
  printXsiVisibility (str);
  binsigexpOperands (ch1, ch2);
}


_VIRTUAL_ void
SSME_Printer::binsigexp (const char *str, const char *ch1, const char *ch2, boolean b1abs, boolean b2abs)
{
  printXsiVisibility (str);
  binsigexpOperands (ch1, ch2, b1abs, b2abs);
}

_VIRTUAL_ void
SSME_Printer::binsigexp2 (const char *str, const char *ch1, const char *ch2)
{
  binsigexpOperands (ch1, ch2);
}

_VIRTUAL_ void
SSME_Printer::binsigexp (const char *str)
{
  printXSI (str);
  closeblock ();
  binsigexpOperands ("exprSignal1", "exprSignal2", pk_true, pk_true);
}

/**
   Printing a binary tree: opappe operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opappe ()
{
  binsigexp (tkProcessCall, "referenceProcess", "listExprSignal", pk_true, pk_false);
}

/**
   Printing a binary tree: opbinf operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opbinf ()
{
  binsigexp (tkClockProduct);
}

/**
   Printing a binary tree: opbsup operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opbsup ()
{
  binsigexp (tkClockUnion);
}

_VIRTUAL_ void
SSME_Printer::printTupleType (boolean bbundle)
{
  printXSI (tkTupleType);
  if (bbundle)
    printKind ("typeKind", "bundle");
  else
    printKind ("typeKind", "struct");
  closeblock ();
  printChild0 (1, "listSignalsDeclaration", pk_true);
  if (bbundle)
    {
      printkwB ("graphSpecification");
      printChild0 (2, "exprProcess", pk_true);
      printkwE (pk_false, "graphSpecification");
    }
}

/**
   Printing a binary tree: opbund operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opbund ()
{
  printTupleType (pk_true);
}

/**
   Printing a binary tree: opcas operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opcas ()
{
  binsigexpOperands ("enumerationValues", "exprProcess", pk_true, pk_true);
}

/**
   Printing a binary tree: opcham operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opcham ()
{
  binsigexp (tkFieldAccess, "exprSignal", "identifier", pk_true, pk_false);
}

/**
   Printing a binary tree: opcomp operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opcomp ()
{
  binsigexp (tkClockDifference);
}

/**
   Printing a binary tree: opconc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opconc ()
{
  binsigexp (tkConcatenation);
}

/**
   Printing a binary tree: opcons operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opcons ()
{
  binsigexp (tkConstantDescription, "identifier", "constantDescriptionValue",  pk_false, pk_true);
}

/**
   Printing a binary tree: opconv operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opconv ()
{
  binsigexp (tkConversion, "predefinedType", "exprSignal", pk_false, pk_true);
}

/**
   Printing a binary tree: opcplx operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opcplx ()
{
  binsigexp (tkComplexConstructor);
}

/**
   Printing a binary tree: opdeccon operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdeccon ()
{
  binsigexp (tkConstantsDeclaration, "type", "listConstantDefinition", pk_true, pk_true);
}

/**
   Printing a binary tree: opdecconp operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdecconp ()
{
  binsigexp (tkConstantsDeclaration, "type", "listConstantDefinition", pk_true, pk_true);
}

_VIRTUAL_ void
SSME_Printer::declsigs (const char *kw)
{
  binsigexp (kw, "type", "listSignal", pk_true, pk_false);
}

/**
   Printing a binary tree: opdecpar operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdecpar ()
{
  declsigs (tkSharedVariablesDeclaration);
}

/**
   Printing a binary tree: opdecsig operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdecsig ()
{
  declsigs (tkSignalsDeclaration);
}

/**
   Printing a binary tree: opdecvar operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdecvar ()
{
  declsigs (tkStateVariablesDeclaration);
}

/**
   Printing a binary tree: opdefegal operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdefegal ()
{
  binsigexp (tkIdentityEquation, "exprSignalLeft", "exprSignalRight", pk_true, pk_true);
}

/**
   Printing a binary tree: opdefpart operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdefpart ()
{
  binsigexp (tkSignalsPartialDefinition, "exprSignalLeft", "exprSignalRight", pk_true, pk_true);
}

/**
   Printing a binary tree: opdefpartdef operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdefpartdef ()
{
  binsigexp (tkSignalsPartialDefaultDefinition, "exprSignalLeft", "exprSignalRight", pk_true, pk_true);
}

/**
   Printing a binary tree: opdefsig operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdefsig ()
{
  binsigexp (tkSignalsDefinition, "exprSignalLeft", "exprSignalRight", pk_true, pk_true);
}

/**
   Printing a binary tree: opdesc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdesc ()
{
  ntrees vt, tbody;
  boolean b;
  noperators op;

  tbody = nchild (2, treeunparseTree);
  op = noper (tbody);
  // to be conform with the ssme meta-model.
  b = pk_false;
  if ((op != opexte) && (op != opnil))
    if (op != oppconf)
      {
	vt = treeunparseTree;
	nchangchild (treeunparseTree, LL_make_pconf_tree (tbody, NILTREE), 2);
	b = pk_true;
      }
  printChild0 (1, "listPragma", pk_false);
  if (op != opnil)
    printChild0 (2, "body", pk_true);
  if (b)
    nchangchild (vt, tbody, 2);
}

/**
   Printing a binary tree: opdescmodu operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdescmodu ()
{
  binsigexpOperands ("listPragma", "listDeclaration", pk_true, pk_true);
}

/**
   Printing a binary tree: opdiff operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdiff ()
{
  binsigexp (tkNotEqual);
}

/**
   Printing a binary tree: opdivi operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opdivi ()
{
  binsigexp (tkDiv);
}

/**
   Printing a binary tree: opegal operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opegal ()
{
  binsigexp (tkEqual);
}

/**
   Printing a binary tree: opegeg operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opegeg ()
{
  binsigexp (tkEqualEqual);
}

/**
   Printing a binary tree: opelem operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opelem ()
{
  binsigexp (tkArrayElem, "exprSignal", "listExprSignal", pk_true, pk_false);
}

/**
   Printing a binary tree: opelevec operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opelevec ()
{
  binsigexp (tkElemDefinition, "exprDimension", "exprSignal");
}

/**
   Printing a binary tree: opensvec operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opensvec ()
{
  binsigexp (tkIteration, "partialIteration", "elemDefinition", pk_true, pk_true);
}

/**
   Printing a binary tree: opentsor operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opentsor ()
{
  binsigexpOperands ("inputs", "outputs", pk_true, pk_true);
}

/**
   Printing a binary tree: opet operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opet ()
{
  binsigexp (tkAnd);
}

/**
   Printing a binary tree: opextr operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opextr ()
{
  binsigexp (tkWhen);
}

/**
   Printing a binary tree: opformact operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opformact ()
{
  binsigexp (tkFormalProcess, "modelType", "modelName");
}

/**
   Printing a binary tree: opformfunc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opformfunc ()
{
  binsigexp (tkFormalProcess, "modelType", "modelName");
}

/**
   Printing a binary tree: opformnode operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opformnode ()
{
  binsigexp (tkFormalProcess, "modelType", "modelName");
}

/**
   Printing a binary tree: opformproc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opformproc ()
{
  binsigexp (tkFormalProcess, "modelType", "modelName");
}

/**
   Printing a binary tree: opformact operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opformprocedure ()
{
  binsigexp (tkFormalProcess, "modelType", "modelName");
}

/**
   Printing a binary tree: opformact operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opformautomaton ()
{
  binsigexp (tkFormalProcess, "modelType", "modelName");
}


/**
   Printing a binary tree: opiegal operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opiegal ()
{
  binsigexp (tkLessOrEqual);
}

/**
   Printing a binary tree: opiieg operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opiieg ()
{
  binsigexp (tkLessLessEqual);
}

/**
   Printing a binary tree: opindmot operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opindmot ()
{
  binsigexp (tkArrayIndex, "identifier", "upperBound", pk_false, pk_true);
}

/**
   Printing a binary tree: opinfe operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opinfe ()
{
  binsigexp (tkLess);
}

/**
   Printing a binary tree: opinit operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opinit ()
{
  //  binsigexp(tkSignalNaming, "identifier", "initValue", pk_true, pk_true );
  binsigexpOperands ("identifier", "initValue", pk_true, pk_true);
}

/**
   Printing a binary tree: opinst operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opinst ()
{
  binsigexp (tkProcessExpansion, "identifier", "listExprParam", pk_false, pk_true);
}

_VIRTUAL_ void
SSME_Printer::printCaseInterval (const char *kindInterval)
{
  printXSI (tkCaseInterval);
  printKind ("intervalsKind", kindInterval);
  closeblock ();
  binsigexpOperands ("lowerBound", "upperBound");
}

/**
   Printing a binary tree: opintvff operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opintvff ()
{
  printCaseInterval (tkLeftClosedRightClosed);

}

/**
   Printing a binary tree: opintvfo operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opintvfo ()
{
  printCaseInterval (tkLeftClosedRightOpened);
}

/**
   Printing a binary tree: opintvof operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opintvof ()
{
  printCaseInterval (tkLeftOpenedRightClosed);
}

/**
   Printing a binary tree: opintvoo operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opintvoo ()
{
  printCaseInterval (tkLeftOpenedRightOpened);
}

/**
   Printing a binary tree: opmele operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opmele ()
{
  binsigexp (tkDefault);
}

/**
   Printing a binary tree: opmmul operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opmmul ()
{
  binsigexp (tkArrayProduct);
}

/**
   Printing a binary tree: opmod operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opmod ()
{
  binsigexp (tkModulo);
}

/**
   Printing a binary tree: opmodu operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opmodu ()
{
  //  binsigexp(tkModule, "identifier", "moduleDescription", pk_false, pk_true);    
  printChild (1, "identifier");
  printChild (2, "moduleDescription");
}

/**
   Printing a binary tree: opmoin operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opmoin ()
{
  binsigexp (tkMinus);
}

/**
   Printing a binary tree: opmult operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opmult ()
{
  binsigexp (tkProduct);
}

/**
   Printing a binary tree: opnomba operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opnomba ()
{
  binsigexp (tkAfter);
}

/**
   Printing a binary tree: opnombd operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opnombd ()
{
  binsigexp (tkFrom);
}

/**
   Printing a binary tree: opnombm operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opnombm ()
{
  binsigexp (tkCount);
}

/**
   Printing a binary tree: opobli operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opobli ()
{
  binsigexp (tkMasking, "exprProcess", "listIdentifier", pk_true, pk_false);
}

/**
   Printing a binary tree: opou operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opou ()
{
  binsigexp (tkOr);
}

/**
   Printing a binary tree: opouex operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opouex ()
{
  binsigexp (tkXor);
}

/**
   Printing a binary tree: oppconf operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oppconf ()
{
  binsigexp (tkRestrictionProcess, "exprProcess", "listDeclarations", pk_true, pk_true);
}

/**
   Printing a binary tree: oppcontrol operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oppcontrol ()
{
  binsigexp (tkControlledProcess, "exprSignal", "exprProcess", pk_true, pk_true);
}


/**
   Printing a binary tree: oppetiq operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oppetiq ()
{
  binsigexp (tkLabelledProcess, "label", "exprProcess", pk_false, pk_true);
}

/**
   Printing a binary tree: opparc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opparc ()
{
  binsigexp (tkPartialIteration, "index", "bounds");
}


/**
   Printing a binary tree: opplus operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opplus ()
{
  binsigexp (tkPlus);
}

/**
   Printing a binary tree: oppuis operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oppuis ()
{
  binsigexp (tkPower);
}

/**
   Printing a binary tree: oprecu operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oprecu ()
{
  binsigexp (tkRecover);
}

/**
   Printing a binary tree: oprepe operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oprepe ()
{
  binsigexp (tkRepetition);
}

/**
   Printing a binary tree: oprest operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_oprest ()
{
  binsigexp (tkArrayRestructuration);
}

/**
   Printing a binary tree: opsegal operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opsegal ()
{
  binsigexp (tkGreaterOrEqual);
}

/**
   Printing a binary tree: opsequ operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opsequ ()
{
  binsigexp (tkSequentialDefinition);
}

/**
   Printing a binary tree: opspec operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opspec ()
{
  printChild0 (1, "processAttributes", pk_true);
  printkwB ("graphSpecification");
  printChild0 (2, "exprProcess", pk_true);
  printkwE (pk_false, "graphSpecification");
}

/**
   Printing a binary tree: opsupe operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opsupe ()
{
  binsigexp (tkGreater);
}

/**
   Printing a binary tree: optvect operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_optvect ()
{
  binsigexp (tkArrayType, "exprDimension", "type", pk_true, pk_true);
}

/**
   Printing a binary tree: opvenum operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opvenum ()
{
  binsigexp (tkConstantEnum, "typeIdentifier", "constantEnumValue");
}

/**
   Printing a binary tree: opvar operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opvar ()
{
  binsigexp (tkVar);
}

/**
   Printing a binary tree: opident operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opident ()
{
  identprint (treeunparseTree);
}

/**
   Printing a binary tree: opident operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opidentfield ()
{
  identprint (treeunparseTree);
}

/**
   Printing a binary tree: opleft_tt operator.
*/
_VIRTUAL_ void
SSME_Printer::decta2_opleft_tt ()
{
  binsigexp ("left_tt");
}

/**
   Printing a ternary tree :  opborn operator.
*/
_VIRTUAL_ void
SSME_Printer::tersigexp (const char *vclass, char *child1, char *child2, char *child3)
{
  printChild (1, child1);
  printChild (2, child2);
  printChild (3, child3);
}


_VIRTUAL_ void
SSME_Printer::tersigexp (const char *vclass, const char *child1, const char *child2, const char *child3, boolean b1, boolean b2, boolean b3)
{
  printXSI (vclass);
  closeblock ();
  printChild0 (1, child1, b1);
  printChild0 (2, child2, b2);
  printChild0 (3, child3, b3);
}

_VIRTUAL_ void
SSME_Printer::tersigexp (const char *vclass)
{
  printXSI (vclass);
  closeblock ();
  tersigexp (vclass, "exprSignal1", "exprSignal2", "exprSignal3", pk_true, pk_true, pk_true);
}


/**
   Printing a ternary tree :  opborn operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opborn ()
{
  tersigexp (tkBounds, "lowerBound", "upperBound", "step", pk_true, pk_true, pk_true);
}

/**
   Printing a ternary tree :  opcond operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opcond ()
{
  tersigexp (tkIfThenElse);
}


/**
   Printing a ternary tree :  opdepd operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opdepd ()
{
  tersigexp (tkConditionedDependence, "signalsFrom", "signalsTo", "conditionExpr", pk_false, pk_false, pk_true);
}

/**
   Printing a ternary tree :  openum operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_openum ()
{
  tersigexp (tkIndex);
}

/**
   Printing a ternary tree :  opfen operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opfen ()
{
  tersigexp (tkWindow);
}


/**
   Model declaration printing without first keyword.
  (action, process, node, private node,...)
*/
_VIRTUAL_ void
SSME_Printer::modelUnparsing ()
{
  if (voperator == opoper)
    printChild (1, "operatorIdentifier");
  else
    printChild (1, "identifier");
  printChildAbs (2, "interfaceDefinition");
  printChild (3, "description");

}

/**
  Printing a ternary tree :  opact operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opact ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opactp operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opactp ()
{
  modelUnparsing ();
}


/**
   Printing a ternary tree :  opfunc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opfunc ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opfuncp operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opfuncp ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opnode operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opnode ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opnodep operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opnodep ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opoper operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opoper ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opprocess operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opprocess ()
{
  modelUnparsing ();
}


/**
   Printing a ternary tree :  opprocessp operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opprocessp ()
{
  modelUnparsing ();
}


/**
   Printing a ternary tree :  opautomaton operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opautomaton ()
{
  modelUnparsing ();
}


/**
   Printing a ternary tree :  opautomatonp operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opautomatonp ()
{
  modelUnparsing ();
}


/**
   Printing a ternary tree :  opprocedure operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opprocedure ()
{
  modelUnparsing ();
}


/**
   Printing a ternary tree :  opprocedurep operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opprocedurep ()
{
  modelUnparsing ();
}

/**
   Printing a ternary tree :  opinte operator.(interface)
*/
_VIRTUAL_ void
SSME_Printer::decta3_opinte ()
{
  tersigexp (tkProcessModelInterface, "listFormalParameter", "ioProfile", "externalGraph", pk_true, pk_false, pk_false);
}

/**
   Printing a ternary tree :  opiter operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opiter ()
{
  tersigexp (tkIterationProcess, "iterationIndex", "exprProcess", "iterationInit", pk_false, pk_true, pk_true);
}

/**
   Printing a ternary tree :  opmemo operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opmemo ()
{
  tersigexp (tkCell);
}

/**
   Printing a ternary tree :  opmoti operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opmoti ()
{
  tersigexp (tkArrayProcess, "arrayIndex", "exprProcess", "iterationInit", pk_false, pk_true, pk_true);
}

/**
   Printing a ternary tree :  oppcas operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_oppcas ()
{
  tersigexp (tkCaseProcess, "identifier", "listCaseBranch", "elseCase", pk_false, pk_false, pk_true);
}


/**
   Printing a ternary tree :  opprag operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opprag ()
{
  // tersigexp(tkPragma, "identifier","listPragmaObject", "pragmaStatement", pk_false, pk_true, pk_false);

  printXSI (tkPragma);
  closeblock ();
  printChild0 (1, "identifier", pk_false);
  printChild0 (2, "listPragmaObject", pk_true);
  printkwB ("pragmaStatement");
  printChild0 (3, "bidon", pk_false);
  printkwE (pk_false, "pragmaStatement");
}

/**
   Printing a ternary tree :  opret operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_opret ()
{
  tersigexp (tkDelay);
}

/**
   Printing a ternary tree :  optype operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optype ()
{
  if (!isSignalNilTree (nchild (3, treeunparseTree)))
    tersigexp (tkTypeMorphism, "identifier", "typeDescriptionValue", "initValue", pk_false, pk_true, pk_true);
  else
    binsigexp (tkTypeDescription, "identifier", "typeDescriptionValue", pk_true, pk_true);
}

/**
   model type printing without first keyword.
*/
_VIRTUAL_ void
SSME_Printer::processtypeunp (const char *kind)
{
  printKind ("modelKind", kind);
  tersigexp (tkProcessType, "identifier", "interfaceDefinition", "listPragma", pk_false, pk_true, pk_false);
  //  printChild(1, "identifier"); printChildAbs(2, "interfaceDefinition");  printChild(3, "listPragma" );

}



/**
   Printing a ternary tree :  optypeact operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optypeact ()
{
  processtypeunp ("action");
}

/**
   Printing a ternary tree :  optypeautomaton operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optypeautomaton ()
{
  processtypeunp ("automaton");
}

/**
   Printing a ternary tree :  optypeprocedure operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optypeprocedure ()
{
  processtypeunp ("procedure");
}


/**
   Printing a ternary tree :  optypefunc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optypefunc ()
{
  processtypeunp ("function");
}

/**
   Printing a ternary tree :  optypenode operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optypenode ()
{
  processtypeunp ("node");
}

/**
   Printing a ternary tree :  optypeproc operator.
*/
_VIRTUAL_ void
SSME_Printer::decta3_optypeproc ()
{
  processtypeunp ("process");
}

/**
   it generates <? xml version="1.0" encoding="UTF-8" ?>
*/
_VIRTUAL_ void
SSME_Printer::SSMEBegin ()
{
  setCharEncodingMode (pk_true);	// STD char encoding mode.
  openblock ();
  printString ("?xml version");
  snspace ();
  sncharpri ('=');
  snspace ();
  printQuotedStr (XmlVersion);
  snspace ();
  printString ("encoding");
  snspace ();
  sncharpri ('=');
  snspace ();
  printQuotedStr (UTF_str);
  sncharpri ('?');
  closeblock ();
}

_VIRTUAL_ void
SSME_Printer::SSMEEnd ()
{
  setCharEncodingMode (pk_true);	// reset XML mode (set in SSME_Printer::SMEBegin())
}

_VIRTUAL_ void
SSME_Printer::prelude (ntrees p)
{
  noperators ope;

  //  if(viscomm) if (unparse_printableTreeAttribute(treeunparseTree, pk_false, vholoph - 1)) separate();
  if (p == ssmeTreeRoot)
    {
      openblock ();
      ope = noper (p);
      printOperator (ope);	// process, process_p, node, node_p, func, func_p, act, act_p, ...
      snspace ();
      printString ("xmi:version");
      sncharpri ('=');
      printQuotedStr (SSMEVersion);
      snspace ();
      printString ("xmlns:xmi");
      sncharpri ('=');
      printQuotedStr (xmlns_xmi);
      snspace ();
      printString ("xmlns:xsi");
      sncharpri ('=');
      printQuotedStr (xmlns_xsi);
      snspace ();
      printString ("xmlns:ssme");
      sncharpri ('=');
      printQuotedStr (xmlns_ssme);
      printModelAttributes (ope);
      closeblock ();
    }
}

_VIRTUAL_ void
SSME_Printer::postlude (ntrees p)
{
  //  if(viscomm) if(unparse_printableTreeAttribute(treeunparseTree, pk_true, vholoph - 1)) separate();
  if (p == ssmeTreeRoot)
    {
      openblock ();
      sncharpri ('/');
      printOperator (noper (p));
      closeblock ();
    }

}

_VIRTUAL_ void
SSME_Printer::decl0 (int k, int j)
{
  int nbfils;
  int i;
  noperators vopl;
  ntrees t, ct;
  boolean b;

  nbfils = ntreearity (treeunparseTree);
  if (nbfils == 0)
    printError ();
  else
    {
      vopl = voperator;
      for (i = 1; i <= nbfils; i++)
	{
	  b = pk_false;
	  if (vopl == oplsign)
	    {
	      t = nchild (i, treeunparseTree);
	      if (ISNOT_NILTREE (t))
		if (noper (t) != opinit)
		  {
		    ct = treeunparseTree;
		    b = pk_true;
		    nchangchild (treeunparseTree, LL_make_init_tree (t, NILTREE), i);
		  }
	    }
	  printChild0 (i, currentnamelist, abstractMode);
	  if (b)
	    nchangchild (ct, t, i);
	}
    }
}

/** Printing a list tree using a separator (seplst) to separate each element and a glue parameter (coller) to shape of the text.
    A final separator is added at the end of the list.
 */
_VIRTUAL_ void
SSME_Printer::decl1 (int seplst, int coller)
{
  decl0 (seplst, coller);
  // this_intePrinter->printkey(seplst);  
}

/**
   Printing a list tree:xxxx
*/
_VIRTUAL_ void
SSME_Printer::printList (const char *vxsi, const char *elemlistname, boolean elemabstract)
{
  char *tksave;
  boolean abssave;

  openblock ();
  printString (currentnamelist);
  printXSI (vxsi);
  printVisibility (voperator);
  closeblock ();
  pushmodes (elemlistname, elemabstract);
  decl0 (0, 1);
  popmodes ();
  printkwE (pk_false, currentnamelist);
}

_VIRTUAL_ void
SSME_Printer::declist_oplfeed ()
{
  printList (tkComposition, "listExprProcess", pk_true);
}

/**
   Printing a list tree: oplsign operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplsign ()
{
  if (abstractMode)
    printList ("UNDEF", "exprSignal", pk_false);
  else
    decl0 (0, 1);
}

/**
   Printing a list tree: oplsign operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplesign ()
{
  char *tk;
  // context....
  if (isfathera (opcas))
    {
      openblock ();
      printString (currentnamelist);
      printXSI (tkCaseEnumeration);
      closeblock ();
      printkwB ("listExprSignal");
      printList2 ("exprSignal", pk_true);
      printkwE (pk_false, "listExprSignal");
      printkwE (pk_false, currentnamelist);
    }
  else if (isfathera (opdepd) || isfathera (opldep))
    {
      openblock ();
      printString (currentnamelist);
      closeblock ();
      printkwB ("listExprSignal");
      printList2 ("exprSignal", pk_true);
      printkwE (pk_false, "listExprSignal");
      printkwE (pk_false, currentnamelist);
    }
  else
    printList (NULL, "exprSignal", pk_true);
}


/**
   Printing a list tree: opleflot operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_opleflot ()
{
  if (abstractMode)
    printList (tkListFlows, "exprSignal", pk_true);
  else
    decl0 (0, 1);
}

/**
   Printing a list tree: oplequhor operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplequhor ()
{
  printList (tkClockEquality, "exprSignal", pk_true);
}


/**
   Printing a list tree: oplequhor operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplexchor ()
{
  printList (tkClockExclusiveRelation, "exprSignal", pk_true);
}


/**
   Printing a list tree: oplequhor operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_opldechor ()
{
  printList (tkClockInfRelation, "exprSignal", pk_true);
}

/**
   Printing a list tree: oplequhor operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplcrohor ()
{
  printList (tkClockSupRelation, "exprSignal", pk_true);
}


_VIRTUAL_ void
SSME_Printer::declist_opltype ()
{
  printList (tkTypesDeclaration, "typeDefinition", pk_true);
}

/**
   Printing a list tree: opltypep operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_opltypep ()
{
  printList (tkTypesDeclaration, "typeDefinition", pk_true);
}


_VIRTUAL_ void
SSME_Printer::printList2 (const char *tk, boolean mode)
{
  char *tksave;
  boolean abssave;

  pushmodes (tk, mode);
  decl0 (0, 1);
  popmodes ();
}

/**
   Printing a list tree: oplmodu operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplmodu ()
{
  printList2 ("listModule", pk_false);
}


/**
   Printing a list tree: oplprag operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplprag ()
{
  printList2 ("listPragma", pk_false);
}

/**
   Printing a list tree: oplopra operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplopra ()
{
  printList2 ("listPragmaObject", pk_false);
}


/**
   Printing a list tree: opchai operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_opchai ()
{
  if (abstractMode)
    {
      printXSI (tkStringValue);
      closeblock ();
    }
  printList2 ("listConstantString", pk_false);
}

/**
   Printing a list tree: opldep operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_opldep ()
{
  int nbfils, i;
  char *tk;
  ntrees t, vt;
  boolean b;

  openblock ();
  printString (currentnamelist);
  printXSI (tkUnconditionedDependences);
  closeblock ();

  nbfils = ntreearity (treeunparseTree);
  if (nbfils == 0)
    printError ();
  else
    {
      vt = treeunparseTree;
      tk = (char *) "signals";
      for (i = 1; i <= nbfils; i++)
	{
	  b = pk_false;
	  t = nchild (i, treeunparseTree);
	  if (noper (t) != oplesign)
	    {
	      nchangchild (treeunparseTree, LL_make_lesign_list (t), i);
	      b = pk_true;
	    }

	  printChild0 (i, tk, pk_false);
	  if (b)
	    nchangchild (vt, t, i);
	  tk = (char *) "dependenceElementsList";
	}
    }
  printkwE (pk_false, currentnamelist);
}



/**
   Printing a list tree: oplelevec operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplelevec ()
{
  openblock ();
  printString (currentnamelist);
  printXSI (tkArrayElementsEnumeration);
  closeblock ();
  printList2 ("listPartialArrayDefinition", pk_true);
  printkwE (pk_false, currentnamelist);
}

/**
   Printing a list tree: oplelevec operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_oplevect ()
{
  openblock ();
  printString (currentnamelist);
  printXSI (tkElemDefinitionEnumeration);
  closeblock ();
  printkwB ("listExprSignal");
  printList2 ("exprSignal", pk_true);
  printkwE (pk_false, "listExprSignal");
  printkwE (pk_false, currentnamelist);
}


_VIRTUAL_ void
SSME_Printer::declist_oplimpo ()
{
  printList (tkModulesImport, "listModule", pk_false);
}

/**
   Printing a list tree: optenum operator. 
*/
_VIRTUAL_ void
SSME_Printer::declist_optenum ()
{
  printList (tkEnumerationType, "listIdentifier", pk_false);
}


/*
op liste non traitÃ© -- a revoir 
   declist_oplattr();
   declist_oplcas();
   declist_oplcons();
   declist_opldecs();
   declist_opldecl();
   declist_oplepara();
   declist_oplident();
   declist_oplparc();
*/
