/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   It applies SignalUnparse() on t.
   @return always "true".
*/
  boolean pKServicesAPImkGraph::pipo (ntrees t)
{
  SignalUnparse (t, undefinedMessage);
  return pk_true;
}


/** 
    @return "true" if the operation succeeds, "false" otherwise.
    It prints in the  file FOO_LIS.SIG a program (prog) compiled with some parameters (param).
    FOO is given by the name of the model/module of prog.
    errored is pk_true iff prog or param is errored 
    Assumed : prog is a model or a module.
 */
boolean pKServicesAPImkGraph::unparse_SignalIsf (ntrees prog)
{
  ptident
    id;
  if (STree_isSignalNilTree (prog))
    return pk_false;
  id = STreeModelModule_getName (prog);
  if (id == NILIDENT)
    id = string2ident ("__FOO__");
  return SignalUnparseIsf (id, prog);
}

/** 
    @return "true" if the operation succeeds, "false" otherwise.
    It prints in the  file FOO_LIS.SIG a program (prog) compiled with some parameters (param).
    FOO is given by the name of the model/module of prog.
    errored is pk_true iff prog or param is errored 
    Assumed : prog is a model or a module.
 */
boolean pKServicesAPImkGraph::unparse_SignalLis (ntrees prog, ntrees param, boolean errored)
{
  ptident
    id;
  if (STree_isSignalNilTree (prog))
    return pk_false;
  id = STreeModelModule_getName (prog);
  if (id == NILIDENT)
    id = string2ident ("__FOO__");
  if (errored)
    return SignalUnparseParserResult (id, prog, param, errinfolis);
  return SignalUnparseParserResult (id, prog, param, infolis);
}

/**
   It translates the representation of a Signal program, represented by its abtract tree (abstractSignalTree) into
   its representation in XMI, according to the meta-model of the language Signal.
   The tree provided by the SIGNAL parser, is:
   - first, modified in the SME_SyntacticRestriction functionality, because the meta-model 
   does not allow to represent all the syntax of the SIGNAL language.
   This functionality is a specialisation of the \ref Trees2Kernel "reduction to Basic Signal".
   
   - then, the modified tree is annotated with attributes to have the good informations for the printer. 
   This is also a specialisation of the \ref Trees2Kernel "reduction to Basic Signal"

   - last, the "sme" (the XMI of the SIGNAL language) file is produced by the XmiPrinter fuctionality. 
   This one is a specialisation of the Signal textual printer.

   @param abstractSignalTree (in) the abtract tree of the Signal program to translate.
   @param dirName (in) the directory in which the xmi file is generated. 
   The name of the file is given by the name of the Signal model, suffixed by ".sme".
   @return the value 0 if the operation succeeds, >0 otherwise.
*/
integer pKServicesAPImkGraph::unparse_SME (ntrees abstractSignalTree)
{
  printf ("##INFO## Signal to SME is no more implemented in signal version V4.20 and later, since Jul 21, 2014");
  return polychronyOptionError;
}

/**
   It translates the representation of a Signal program, represented by its abtract tree (termAbstractTree) into
   its representation in XMI, according to the meta-model of the language Signal.
   The tree provided by the SIGNAL parser, is:
   - first, modified in the SME_SyntacticRestriction functionality, because the meta-model 
   does not allow to represent all the syntax of the SIGNAL language.
   This functionality is a specialisation of the \ref Trees2Kernel "reduction to Basic Signal".
   
   - then, the modified tree is annotated with attributes to have the good informations for the printer. 
   This is also a specialisation of the \ref Trees2Kernel "reduction to Basic Signal"

   - last, the "sme" (the XMI of the SIGNAL language) file is produced by the XmiPrinter fuctionality. 
   This one is a specialisation of the Signal textual printer.

   @param termAbstractTree (in) the abtract tree of the Signal program to translate.
   @param dirName (in) the directory in which the xmi file is generated. 
   The name of the file is given by the name of the Signal model, suffixed by ".sme".
   @return the value 0 if the operation succeeds, >0 otherwise.
*/
integer pKServicesAPImkGraph::unparse_SME (ntrees termAbstractTree, char *dirName)
{
  // SMESyntaxRestrictionTranslator = new SME_SyntaxRestriction(); 
  // SMEtrans = new SME_Translator(tloadedLib); 
  printf ("##INFO## Signal to SME is no more implemented in signal version V4.20 and later, since Jul 21, 2014");
  return polychronyOptionError;
}

integer pKServicesAPImkGraph::unparse_parse2SME (char *srcfile, char *dirName)
{
  // SMESyntaxRestrictionTranslator = new SME_SyntaxRestriction(); PLG SME_SyntaxRestriction to be removed
  // SMEtrans = new SME_Translator(tloadedLib);                    PLG SME_Translator to be removed
  printf ("##INFO## Signal to SME is no more implemented in signal version V4.20 and later, since Jul 21, 2014");
  return polychronyOptionError;
}


/**
   It translates the representation of a textual Signal program  into
   its representation in SSME (xmi), according to the meta-model of the language Signal.
   @param srcfile (in) the file that contains the Signal program to translate.
   @param dirName (in) the directory in which the xmi file is generated. 
   the name of the file is given by the name of the Signal model, suffixed by ".ssme".
   @return the value 0 if the operation succeeds, >0 otherwise.
   
   It calls:
   - the Signal Parser (method SIGNALParseSrcfile()).
   - then the Signal/SSME translator applied on the abtract tree returned by the parser. 
   See signalTree2SSME(ntrees).
*/
integer pKServicesAPImkGraph::signal2SSME (char *srcfile, char *dirName)
{
  ntrees vSignalTree;
  ptrepstr enteringOutputDir;

  enteringOutputDir = sys_get_userOutputDir ();

  if (!SignalParseFile (srcfile, &vSignalTree))
    return polychronySyntaxErrorProg;
  sys_set_userOutputDir (string2repstr (dirName));
  signalTree2SSME (vSignalTree);
  sys_restore_userOutputDir (enteringOutputDir);
  return polychronyNoError;
}

/**
   It translates the representation of a Signal program, represented by its abtract tree (termAbstractTree) into
   its representation in SSME (Xmi), according to the meta-model of the language Signal.
   The tree provided by the SIGNAL parser, is:
   - first, modified in the SSME_SyntacticRestriction functionality, because the meta-model 
   does not allow to represent all the syntax of the SIGNAL language.
   This functionality is a specialisation of the \ref Trees2Kernel "reduction to Basic Signal".
   
   - then, the modified tree is annotated with attributes to have the good informations for the printer. 
   This is also a specialisation of the \ref Trees2Kernel "reduction to Basic Signal"

   - last, the "ssme" (XMI form of the SIGNAL language) file is produced by the SSME_Printer fuctionality. 
   This one is a specialisation of the Signal textual printer.

   @param termAbstractTree (in) the abtract tree of the Signal program to translate.
   The name of the file is given by the name of the Signal model, suffixed by ".sme".
   @return the value 0 if the operation succeeds, >0 otherwise.
*/
void
pKServicesAPImkGraph::signalTree2SSME (ntrees termAbstractTree)
{
  SSME_unparse (termAbstractTree);
}

/**
   It exports the interfaces of the definitions of a module as a header file.
   @param srcfile (in) the file that contains the Signal program to translate.
   @param dirName (in) the directory in which the xmi file is generated. 
   the name of the file is given by the name of the Signal model, suffixed by ".SIG".
   @return the value 0 if the operation succeeds, >0 otherwise.
   
   Actions :
   
   - It calls the Signal Parser (method SIGNALParseSrcfile()).
   - It cleans the "body" part of the model 
   - It prints then the header file.
*/
integer pKServicesAPImkGraph::exportHeader (ntrees termAbstractTree)
{
  ntrees srcAbstractTree, trgtAbstractTree;
  ptrepstr fileId;

  if (STree_isModulesList (termAbstractTree))
    {
      srcAbstractTree = STree_getModule (termAbstractTree, 1);
      if (!STree_extractModuleHeaders (srcAbstractTree, &trgtAbstractTree))
	return polychronyUnexpected;
    }
  else if (!(STree_extractModelHeader (termAbstractTree, &trgtAbstractTree)))
    return polychronyUnexpected;
  if (sys_add_typeSuffix2repstr (ident2repstr (STreeModelModule_getName (trgtAbstractTree)), "h", &fileId))
    {
      STree_addGeneralComments (trgtAbstractTree);
      SignalUnparseFile (trgtAbstractTree, fileId, InfoHeaderPrinter);
      return polychronyNoError;
    }
  return polychronyInternalError;
}
