/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define signalname "SIGNAL"
#define predefname "PREDEF  "

#define TRACE(s)

_PROTECTED_
/** 
    @return true if one element of the list lm is in "error" mode.
*/
    boolean pKServicesAPImkGraph::printReferencedLibraryModuleList(ptlrepid lm, boolean bforce)
{
  integer n, i;
  ntrees t;
  ptrepfi m;
  boolean b;

  b = pk_false;
  n = ptlrepid_size(lm);
  for (i = 1; i <= n; i++)
    {
      m = getModelRepid(lm);
      if (printReferencedLibrary(m, bforce))
	b = pk_true;
      if (bforce || !kl_getErrorCode(m))
	{
	  b = pk_true;
	  // unicity ?
	  t = kl_getTreeReference(m);
	  if (!nattrexist(t, TreeCompilerPrint_ATTRIBUTENUMBER))
	    {
	      nmakeattrint(t, TreeCompilerPrint_ATTRIBUTENUMBER, 0);
	      unparse_SignalLis(t, NILTREE, pk_false);
	    }
	}
      lm = succid(lm);
    }
  return b;
}

/** 
    @return true if a module is in mode error, false otherwise.
 */
boolean pKServicesAPImkGraph::printReferencedLibraryLoc(ptrepfi model, boolean bforce)
{
  ptlrepid l;
  boolean b;
  integer n, i;
  ptrepfi m;

  if (model == NILREPFI)
    return pk_false;
  l = kl_getModules(model);
  b = printReferencedLibraryModuleList(l, bforce);
  l = kl_getLocModels(model);
  n = ptlrepid_size(l);
  for (i = 1; i <= n; i++)
    {
      m = getModelRepid(l);
      if (printReferencedLibraryLoc(m, bforce))
	b = pk_true;
      l = succid(l);
    }
  return b;
}

void pKServicesAPImkGraph::resetMark(ptrepfi model)
{
  ptlrepid l;
  integer n, i;
  ptrepfi m;
  ntrees t;

  if (model != NILREPFI)
    {
      t = kl_getTreeReference(model);
      ndestroyattr(t, TreeCompilerPrint_ATTRIBUTENUMBER);

      l = kl_getModules(model);
      n = ptlrepid_size(l);
      for (i = 1; i <= n; i++)
	{
	  m = getModelRepid(l);
	  resetMark(m);
	  l = succid(l);
	}
      l = kl_getLocModels(model);
      n = ptlrepid_size(l);
      for (i = 1; i <= n; i++)
	{
	  m = getModelRepid(l);
	  resetMark(m);
	  l = succid(l);
	}
    }
}

/** 
    @return true if a module is in mode error, false otherwise.
    It prints (printReferencedLibraryLoc()) the Signal text of the referenced libraries of a model (model)
    only for the errored one (bforce is "false") or always (bforce is "true").
    Then, the used marks are cleaned resetMark(model).
*/
boolean pKServicesAPImkGraph::printReferencedLibrary(ptrepfi model, boolean bforce)
{
  boolean b;
  b = printReferencedLibraryLoc(model, bforce);
  resetMark(model);
  return b;

}

/**
  Compiling a Signal program from an abstract tree.
  - step 1 : production of a kernel-tree representation (global variable ptrepfi pKServices_vfi)  of the model.
  - step2: production of a parametrized instance of such a model (\see makeInstanceModel function).
  
  @param x (in) : the abstract-tree of the compiled program/module.
  @param idprogram (out) : name of the compiled program (module/process).
  @return a non null value when errors are detected, 0 otherwise. (see polychronyExitStatus.h)
*/
integer pKServicesAPImkGraph::makeModel(ntrees x, ptident * idprogram)
{
  initcallTrees2Kernel();
  pK_messenger->setErrorCode(pk_true);
  pKServices_vfi = pK_Tree2Kernel->Trees2KernelStep1(x, &cc, idprogram, pk_true);
  if (cc)
    return polychronyNoError;
  optGenLis = pk_true;
  return polychronySolvingRefError;
}

/**
   Initializing for kernel reduction compiling step.
*/
void pKServicesAPImkGraph::initcallTrees2Kernel()
{
  pK_messenger->setcputime();
  pKServices_vfi = NILREPFI;
}

/**
   Global initializing of the virtual machine.
   - Tree memory
   - string memory
   - name manager
   - stack manager (why here ?)
   - message manager
*/
void pKServicesAPImkGraph::userbegin_init()
{
  ntreesMemoryInitialize();
  stringMemoryInitialize();
  //  pK_Graph2SignalTreeBuilder->setInterfaceMode (ToolContext == eclipsePlugin);      // useful/necessary for signal2xmi.
  namini();
  predefSignalNameIni();
  SignalPredefinedPragmasTableIni();
  atpileini(getIdentNil());
  pK_messenger->gmessini();
}

/**
   Global  reset (when not ginit) of the virtual machine.
   - Tree memory
   - string memory
   - name manager
   - stack manager (why here ?)
   - message manager
*/
void pKServicesAPImkGraph::userbegin_reset()
{
  boolean bGUI;
  bGUI = ToolContext == pKGUI;
  if (!bGUI)
    {
      pK_messenger->parsing_logFileClose();
      ntrees_memoryReset();
      pK_SignalPrinter->reset();
      TypeGarbage();
    }

  namini();
  predefSignalNameIni();
  SignalPredefinedPragmasTableIni();
  atpileini(getIdentNil());
  if (pK_messenger->gmessini()) ;	/* success is assumed */

  if (!bGUI)
    {
      pK_SignalPrinter->setPrinterContext(bGUI);
      pK_messenger->gmessload(systmess);
    }
}

/**
   Global initializing of the virtual machine:
   - system interface (\ref SystemInterfaceAPI)
   - buffer (\ref buffer)
   - Signal pretty printer (decsigini()),
   - loading global messages,(gmessload())
   - type manager (TypeManagerInit())
*/
void pKServicesAPImkGraph::commonini()
{
  integer i;
  integer TableMaxSize;
  ptrepbuf this_buffer;

  sysini();
  if (pK_messenger->parsing_logFileReset()) ;
  pK_SignalPrinter->setPrinterContext(ToolContext == pKGUI);
  this_buffer = new buffer(45, 256);
  decsigini(this_buffer);
  i = pK_messenger->gmessload(systmess);
  sysSetEnvDefDefault(SHELL_TableMaxSize, "4096");
  TableMaxSize = sysGetenvIntegerValue(SHELL_TableMaxSize);
  TypeManagerInit(TableInitialSize, TableMaxSize, TableResizingFactor);
}

/**
   unparse_SME translator: it prints the SME textual form of a Signal tree (vt).
   The tree must be annotated by attributes, before calling the printer.
   \see unparse_SME(ntrees, char*) method.
*/
void pKServicesAPImkGraph::printSMEtree(ntrees vt, ptident idprogram)
{
  ptSME_Printer SMEPrinter;
  ptrepstr effname;
  ptlPointer l;
  ptrepbuf buf;
  ntrees aLib;
  pointer velem;

  // if (noper(vt) == oplmodu) vt = nchild(1, vt);
  vt = STree_getModule(vt, 1);

  SMEPrinter = new SME_Printer();
  buf = getPrinterBuffer();
  if (SMEPrinter->fileUnparse(vt, buf, ident2repstr(idprogram), &effname))
    {
      pK_messenger->globalMessageStr(InfoSMEPrinter, effname);
      if (nattrexist(vt, SME_ReferencedModuleATTRIBUTENUMBER))
	{
	  l = (ptlPointer) nattrptr(vt, SME_ReferencedModuleATTRIBUTENUMBER);
	  while (ptlPointer_get_elemNext(&l, &velem))
	    {
	      SMEPrinter->reset();
	      aLib = (ntrees) velem;
	      SMEPrinter->fileUnparse(aLib, buf, STreeModelModule_getNameStr(aLib), &effname);
	      pK_messenger->globalMessageStr(InfoSMEPrinter, effname);
	    }
	}
    }
  else
    pK_messenger->globalMessageStr(InfoFileNotGenerated, string2repstr("sme"));
  delete SMEPrinter;
}

void pKServicesAPImkGraph::printSignal(ntrees vt)
{
  boolean b;
  b = pK_SignalPrinter->setPrinterContext(pk_false);	// to print the annotations of the trees.
  SignalUnparse(vt, infotra);
  pK_SignalPrinter->setPrinterContext(b);
}

/** 
    @return "true" if the parsing of a file (srcfile) succeeds, "false" otherwise.
    @param vSignalTree: the resulting tree of the parsing.
    @param srcfile: the complete path of the file to parse.
*/
boolean pKServicesAPImkGraph::SignalParseFile(char *srcfile, ntrees * vSignalTree)
{
  boolean bc, cc;
  integer messTableOld;

  messTableOld = pK_messenger->gmessload(parsermess);
  cc = SIGNALParseSrcfile(srcfile, vSignalTree, &bc);
  pK_messenger->gmessrestore(messTableOld);
  return cc;
}

/**
   SSME_unparse translator: it prints the SSME textual form of a Signal tree (vt).
   The tree must be annotated by attributes, before calling the printer.
*/
void pKServicesAPImkGraph::SSME_unparse(ntrees termAbstractTree)
{
  ptrepstr effname;
  ptident idprogram = STreeModelModule_getName(termAbstractTree);

  if (pK_SSME_Printer->fileUnparse(termAbstractTree, getPrinterBuffer(), ident2repstr(idprogram), &effname))
    pK_messenger->globalMessageStr(InfoSSMEPrinter, effname);
}

void pKServicesAPImkGraph::session_logFileClose(integer vstatus)
{
  pK_messenger->gmessload(systmess);
  if (vstatus == polychronyNoError)
    pK_messenger->message(announceSignalCompletion);
  else
    pK_messenger->message(announceSignalFailure);
  pK_messenger->gmessload(systmess);
  pK_messenger->session_logFileClose(vstatus);
}

/**
  Compiling a Signal program = production of an instance of a model.
  Assume : \ref makeModel function has been applied before this function; the representation of the model is built (global variable pKServices_vfi).
  @param mpp : the abstract-tree of the compiled program/module.
  @param mparam: the abstract-tree of the user parameters.
  @param vgraph (out) : graph of the compiled instance.
  @return a non null value when errors are detected, 0 otherwise.(see polychronyExitStatus.h)
*/
integer pKServicesAPImkGraph::makeInstanceModel(ntrees mpp, ntrees mparam, ptDCG * vgraph)
{
  boolean bres, b;
  integer status;
  ptkernel2graph translator;
  ntrees instModel;
  noperators op;

  instModel = STree_mkReferenceToModelWithParameter(mparam, mpp);
  status = polychronyNoError;
  pK_messenger->setErrorCode(pk_true);

  b = pK_Tree2Kernel->Trees2KernelStep2(instModel, pKServices_vfi, &parglobal, &cc);
  if (b)
    pK_messenger->setcputime();
  if (b && cc)
    {
      ptGraphManager aGraphManager = pK_GraphManager;
      translator = new kernel2graph(aGraphManager, traceabilityMode);
      op = noper(this_termAbstractTree);
      bres = translator->kernelTree2DCG(pKServices_vfi, parglobal, (op == oplmodu) || (op == opmodu));
      if (!bres)
	{
	  optGenLis = pk_true;
	  status = polychronyTypeError;	/* the application is not well typed */
	}
      else
	*vgraph = translator->getDCG();
      delete translator;
    }
  else if (!cc)
    {
      optGenLis = pk_true;
      status = polychronySolvingRefError;	/* Pb forme du processus */
    }

  return status;
}
