/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __SEMAPHOREMANAGER_H__
#define __SEMAPHOREMANAGER_H__



/**********************************************************************************/
/****************************   SEMAPHORE_OBJECT : BEGIN **************************/
/**********************************************************************************/

typedef class Semaphore * ptSemaphore;
class Semaphore
{
   public :
   APEX_Semaphore_type sema;
   Semaphore(APEX_Semaphore_type s);
};

/**********************************************************************************/
/*****************************   SEMAPHORE_OBJECT : END   *************************/
/**********************************************************************************/




/**********************************************************************************/
/*****************************   MANAGER_NODE  : BEGIN   **************************/
/**********************************************************************************/


/* Type of manager node contents */
struct SemAssoc_type{
  Comm_ComponentID_type SemaphoreID; 
  ptSemaphore ptrSemaphore;
};


/* Manager node type definition */
typedef class SemaphoreManagerNode *ptSemaphoreManagerNode;
class SemaphoreManagerNode{

 private :
  SemAssoc_type contents;                                  /* Contents of a node */
  ptSemaphoreManagerNode next;                                      /* Next node */

 public :
  SemaphoreManagerNode();
  SemaphoreManagerNode(SemAssoc_type componentRecord);
  ~SemaphoreManagerNode();
  void setValue(SemAssoc_type value);
  void setNext(ptSemaphoreManagerNode ptvalue);
  SemAssoc_type getValue();
  ptSemaphoreManagerNode getNext();
};


/**********************************************************************************/
/******************************   MANAGER_NODE  : END   ***************************/
/**********************************************************************************/




/**********************************************************************************/
/********************************   MANAGER : BEGIN  ******************************/
/**********************************************************************************/


typedef class SemaphoreManager *ptSemaphoreManager;
class SemaphoreManager{

 private :
  ptSemaphoreManagerNode first;                      /* First node in the manager */
  ptSemaphoreManagerNode last;                        /* Last node in the manager */
  int max_size;                                      /* Size limit of the manager */
  int current_size;                     /* Current number of nodes in the manager */

  void setFirst(ptSemaphoreManagerNode ptvalue);
  void setLast(ptSemaphoreManagerNode ptvalue);
  ptSemaphoreManagerNode getFirst();
  ptSemaphoreManagerNode getLast();

 public :
  SemaphoreManager();
  SemaphoreManager(int size);
  ~SemaphoreManager();
 
  bool SemaphoreRecord(Comm_ComponentName_type semaphore_Name, 
                       int current_Value, 
                       int maximum_Value, 
                       QueuingDiscipline_type queuing_Discipline, 
                       Comm_ComponentID_type& semaphoreID);
   /* If the manager was not full, a new semaphore node is created, initialized
     with input parameters and it is returned true. Else, it is returned false. */

  bool SemaphoreCheckName(Comm_ComponentName_type semaphore_Name, Comm_ComponentID_type& semaphoreID); 
   /* If a semaphore named semaphore_Name exists in the manager, semaphoreID is 
      its identifier and it is returned true. Else, it is returned false */

  bool SemaphoreCheckID(Comm_ComponentID_type semaphoreID, APEX_Semaphore_type& sema);
   /* If there is an identifier semaphoreID in the manager, sema is the 
     semaphore associated and it is returned true. Else, it is returned false */

  void SemaphoreUpdate(APEX_Semaphore_type sema);
  /* The semaphore from the manager, which has the same name as sema, is updated */

  bool Full();
  /* Returns true if the manager size limit is reached. Else, it is returned 
    false */

  void printFile(Comm_ComponentID_type semaphoreID, char *result_file);
  /* Prints semaphoreID status in result_file */

};

/*********************************************************************************/
/*********************************   MANAGER : END  ******************************/
/*********************************************************************************/


//static ptSemaphoreManager SEMAPHORE_MANAGER;

#endif // __SEMAPHOREMANAGER_H__










