/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __PRIORITYRECEIVEQUEUE_H__
#define __PRIORITYRECEIVEQUEUE_H__



/***********************************************************************************/
/***************************   PRIORITY_RECEIVEQUEUE_NODE   ************************/
/***********************************************************************************/


typedef class PRIORITY_ReceiveQueueNode *ptPRIORITY_ReceiveQueueNode;
class PRIORITY_ReceiveQueueNode
{
 private:
  ProcessID_type contents;                                   /* Contents of a node */
  ptPRIORITY_ReceiveQueueNode next;                                   /* Next node */

 public:
  PRIORITY_ReceiveQueueNode(ProcessID_type Process_ID);
  ~PRIORITY_ReceiveQueueNode();
  void setContents(ProcessID_type Process_ID);
  void setNext(ptPRIORITY_ReceiveQueueNode nextNode);
  ProcessID_type getContents();
  ptPRIORITY_ReceiveQueueNode getNext();
};



/***********************************************************************************/
/*****************************   PRIORITY_RECEIVEQUEUE   ***************************/
/***********************************************************************************/


typedef class PRIORITY_ReceiveQueue *ptPRIORITY_ReceiveQueue;
class PRIORITY_ReceiveQueue
{
  private:
   ptPRIORITY_ReceiveQueueNode first;          /* First node of the PRIORITY_QUEUE */
   ptPRIORITY_ReceiveQueueNode last;            /* Last node of the PRIORITY_QUEUE */
   int max_size;                               /* Size limit of the PRIORITY_QUEUE */
   int current_size;              /* Current number of nodes in the PRIORITY_QUEUE */
  
   void setFirst(ptPRIORITY_ReceiveQueueNode node);
   void setLast(ptPRIORITY_ReceiveQueueNode node);
 
  public:
   PRIORITY_ReceiveQueue();
   PRIORITY_ReceiveQueue(int nbproc);
   ~PRIORITY_ReceiveQueue();
   ptPRIORITY_ReceiveQueueNode getFirst();
   ptPRIORITY_ReceiveQueueNode getLast();

   int getCurrentSize();

   void insert(ProcessID_type Process_ID);
   /* Insert Process_ID in the QUEUE */

   void remove(ProcessID_type& Process_ID);
   /* Remove the node with the highest value from the QUEUE and Process_ID is the
      value of this node */

   void clean();
   /* delete all nodes in the QUEUE */

   bool empty();
   /* The QUEUE is empty */

   bool full();
   /* The QUEUE size limit is reached */

   void printFile(char *result_file);
   /* prints the QUEUE in result_file */

};


#endif // __PRIORITYRECEIVEQUEUE_H__










