module AADL_EVENTDATAPORT =

use FIFOS_LIB;

process InEventDataPort_Behavior =
  { type msg_type;
    msg_type def_value; 
    integer size, dequeue_number; 
    string port_name;}
  (? msg_type write_flow;
     event Frozen_time_event;
     event clk_output;
   ! msg_type read_flow;)
  pragmas
    Comment 
      "It models the behavior of an input event data port of a thread."
      "At Frozen_time_event, frozen current elements of in_fifo to frozen_fifo."
      "The thread can access the frozen_fifo each time when a dequeue action is performed." 
      "Note that the dequeue_number is constrainted in the associated property."
  end pragmas
  (| (out1, cnt1) := fifo{msg_type, size, def_value}(write_flow)
   | cnt1 ^= write_flow ^+ out1
   | out1 ^= Frozen_time_event
   | clk_output ^= read_flow   
   | read_flow := AT (out1 when ((cnt1 cell ^clk_output) >0), clk_output)
   |)
  where
    integer cnt1;
    msg_type out1;
  end;
    
process OutEventDataPort_Behavior =
  { type msg_type;
    msg_type def_value;
    integer size; 
    string port_name;
    }
  (? msg_type write_flow;  
     event Output_time_event;
   ! msg_type sent_flow;)
  pragmas
    Comment 
     "It models the behavior of an output event data port of a thread."
     "The output is queued, and sent out at Output_time_event."
  end pragmas

  (| (out1, cnt1) := fifo{msg_type, size, def_value}(write_flow)
   | out1 ^= Output_time_event
   | cnt1 ^= write_flow ^+ out1
   | sent_flow :=  AT (out1 when ((cnt1 cell Output_time_event) >0), Output_time_event)
   |)  
  where
    msg_type out1;
    integer cnt1;  
  end;    
end;
