/*
 * Revision Control Information
 * split bdd.h: inner declarations
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd.h,v $
 * $Author: PLG $
 * $Revision: 1.2 $
 * $Date: 2011/03/02 $
 * see bdd.h
 */
/* requires
 * #include <stddef.h>
 * #include <stdio.h>
 * #include "bdd_export.h"
 * ...
 */
#ifndef bdd_inner_h /* { */
#define bdd_inner_h

#undef BDD_AUTOMATED_STATISTICS_GATHERING
#define BDD_INLINE_ITE
#define BDD_INLINE_ITE_CONSTANT
#undef BDD_NO_GC
#undef BDD_STATS
#undef BDD_MEMORY_USAGE

// #define boolean int


/* 
 * Initialization data structure.
 */
typedef struct __st_bdd_mgr_init {
    struct {
        boolean on;                   /* TRUE/FALSE: is the cache on */
        UnsignedInteger resize_at;       /* percentage at which to resize (e.g. 85% is 85); doesn't apply to adhoc */
        UnsignedInteger max_size;        /* max allowable number of buckets; for adhoc, max allowable number of entries */
    } ITE_cache,
      ITE_const_cache,
      adhoc_cache;
    struct {
      boolean on;                     /* TRUE/FALSE: is the garbage collector on */
    } garbage_collector;
    struct {
      void (*daemon)();               /* used for callback when memory limit exceeded */
      UnsignedInteger limit;             /* upper bound on memory allocated by the manager; in megabytes */
    } memory;
    struct {
      float ratio;                    /* allocate new bdd_nodes to achieve ratio of used to unused nodes */
      UnsignedInteger init_blocks;       /* number of bdd_nodeBlocks initially allocated */
    } nodes;
} bdd_mgr_init;

/*
 * BDD Manager Allocation And Destruction
 */

/*
 * BDD Variable Allocation
 */

/*
 * BDD Formula Management
 */

/* 
 * Operations on BDD Formulas
 */

/*
 * Queries about BDD Formulas
 */

/*
 * Statistics and Other Queries
 */
typedef struct bdd_cache_stats {
    UnsignedInteger hits;
    UnsignedInteger misses;
    UnsignedInteger collisions;
    UnsignedInteger inserts;
} bdd_cache_stats;

typedef struct bdd_stats {
    struct {
	bdd_cache_stats hashtable;   /* the unique table; collisions and inserts fields not used */ 
	bdd_cache_stats itetable;
	bdd_cache_stats consttable;
	bdd_cache_stats adhoc;
    } cache;		/* various cache statistics */
    struct {
	UnsignedInteger calls;
	struct {
	    UnsignedInteger trivial;
	    UnsignedInteger cached;
	    UnsignedInteger full;
	} returns;
    } ITE_ops,
      ITE_constant_ops,
      adhoc_ops;
    struct {
	UnsignedInteger total;
    } blocks;		/* bdd_nodeBlock count */
    struct {
	UnsignedInteger used;
	UnsignedInteger unused;
	UnsignedInteger total;
        UnsignedInteger peak;
    } nodes;		/* bdd_node count */
    struct {
	UnsignedInteger used;
	UnsignedInteger unused;
	UnsignedInteger total;
        UnsignedInteger blocks; 
    } extptrs;		/* bdd_t count */
    struct {
	UnsignedInteger times;     /* the number of times the garbage-collector has run */
        UnsignedInteger nodes_collected; /* cumulative number of nodes collected over life of manager */
        long runtime;           /* cumulative CPU time spent garbage collecting */
    } gc;
    struct {
        Integer first_sbrk;         /* value of sbrk at start of manager; used to analyze memory usage */
        Integer last_sbrk;          /* value of last sbrk (see "man sbrk") fetched; used to analyze memory usage */
        UnsignedInteger manager;
	UnsignedInteger nodes;
	UnsignedInteger hashtable;
	UnsignedInteger ext_ptrs;
	UnsignedInteger ITE_cache;
	UnsignedInteger ITE_const_cache;
	UnsignedInteger adhoc_cache;
	UnsignedInteger total;
    } memory;           /* memory usage */
} bdd_stats;

typedef Integer bdd_literal;	/* integers in the set { 0, 1, 2 } */

/*
 *    foreach macro in the most misesque tradition
 *    bdd_gen_free always returns 0
 */

/*
 *    foreach_bdd_cube(fn, gen, cube)
 *    bdd_t_pointer fn;
 *    bdd_gen_pointer gen;
 *    array_t_pointercube;	- return
 *
 *    foreach_bdd_cube(fn, gen, cube) {
 *        ...
 *    }
 */
#define foreach_bdd_cube(fn, gen, cube)\
  for((gen) = bdd_first_cube(fn, &cube);\
      ((gen)->status != bdd_EMPTY) ? TRUE: bdd_gen_free(gen);\
      (void) bdd_next_cube(gen, &cube))

/*
 * These are NOT to be used directly; only indirectly in the macros.
 */
/* 
 * Miscellaneous
 */

/*
 *    These are the hooks for stuff that uses bdd's
 *
 *    There are three hooks, and users may use them in whatever
 *    way they wish; these hooks are guaranteed to never be used
 *    by the bdd package.
 */
typedef struct bdd_external_hooks {
    char *network;
    char *mdd;
} bdd_external_hooks;

/*
 * Default settings.
 */
#define BDD_NO_LIMIT                      ((1<<30)-2)
#define BDD_DFLT_ITE_RESIZE_AT            75
#define BDD_DFLT_ITE_MAX_SIZE             100000
#define BDD_DFLT_ITE_CONST_RESIZE_AT      75
#define BDD_DFLT_ITE_CONST_MAX_SIZE       1000003
#define BDD_DFLT_ADHOC_RESIZE_AT          0
#define BDD_DFLT_ADHOC_MAX_SIZE           100000
#define BDD_DFLT_DAEMON                   (void *)NULL
#define BDD_DFLT_MEMORY_LIMIT             BDD_NO_LIMIT   
#define BDD_DFLT_NODE_RATIO               2.0
#define BDD_DFLT_INIT_BLOCKS              10

#endif /* } */

