/* PLG 2011/05/09
 * extracted from bdd_int
 * requires BDD_StdTypes.h bdd_int.h bdd_int_hashcache.h bdd_int_safenodeframe.h bdd_int_bddBlock.h bdd_inner.h
 */

/*
 *    The bdd_manager contains all of the information necessary to manage
 *    a garbage-collecting heap of bdd_nodes.   Garbage-collection is to 
 *    happen transparently to the user, so no pointers can be exported.
 *
 *    Note that this is a BIG structure.   That is because it must hold EVERY
 *    single pointer and variable which is to be used by the bdd package.
 *    Each bdd_manager contains the _complete_ set of state variables required
 *    for the bdd package to operate on a heap of bdd's.
 */
/* typedef struct bdd_manager bdd_manager; already typedef'd in bdd.h */
struct __bdd_manager {
  char *undef1;                       /* used for debugging */
  struct {
    struct {
      UnsignedInteger nkeys;		/* the number of keys in the hash table */
      UnsignedInteger nbuckets;	/* the size of the buckets array */
      bdd_node_pointer *buckets;		/* the buckets array */
      Integer rehash_at_nkeys;	/* the number of keys at which we rehash */
    } hashtable;		/* all live bdd_nodes are in this hash table */
    struct {
      struct {
	UnsignedInteger index;		/* the roving index in the block */
	bdd_bddBlock *block;		/* the block in which we look */
      } pointer;			/* the roving pointer in the map */
      UnsignedInteger free;		/* the number which are known to be free */
      UnsignedInteger nmap;		/* the total number of bdd_t's in map */
      bdd_bddBlock *map;		/* mapping bdd_t*'s to bdd_node*'s */
    } external_refs;	/* external references from the package user */
    struct {
      bdd_safeframe *frames;	/* a list of safe frames on the stack */
    } internal_refs;	/* internal references from the bdd package */
    struct {
      struct {
	boolean on;			/* is caching even turned on? */
	boolean invalidate_on_gc;	/* always invalidate the cache on a gc */
	UnsignedInteger resize_at;         /* percentage at which to resize (e.g. 85% is 85) */
	UnsignedInteger max_size;          /* don't resize if nbuckets exceeds this */
	UnsignedInteger nbuckets;		/* buckets in the cache */
	UnsignedInteger nentries;		/* entries in the cache */
	bdd_hashcache_entry **buckets;	/* the cache */
      } itetable;		        /* cache of the hashtable entries */
      struct {
	boolean on;			/* is caching even turned on? */
	boolean invalidate_on_gc;	/* always invalidate the cache on a gc */
	UnsignedInteger resize_at;         /* percentage at which to resize (e.g. 85% is 85) */
	UnsignedInteger max_size;          /* don't resize if nbuckets exceeds this */
	UnsignedInteger nbuckets;		/* buckets in the cache */
	UnsignedInteger nentries;		/* entries in the cache */
	bdd_constcache_entry **buckets;	/* the cache */
      } consttable;		/* for the exclusive use of ITE_constant */
      struct {
	boolean on;			/* is caching even turned on? */
	UnsignedInteger max_size;          /* don't resize if nbuckets exceeds this */
	st_table_pointer table;		/* may be NIL - of {bdd_adhoccache_key, bdd_node_pointer } */
      } adhoc;			/* cache used for various nefarious purposes (transient) */
    } cache;		/* a cache for past computations (partially voided on gc) */
    struct {
      struct {
	bdd_nodeBlock_pointer top;	/* the top of the in-use list (append only) */
	bdd_nodeBlock_pointer *tail;	/* either &top or &p->next of some member of the list */
      } inuse;			/* an in-use list of nodeBlock */
      bdd_nodeBlock_pointer free;	/* a free list of unused blocks in this space */
    } half[2];
    UnsignedInteger init_node_blocks;  /* number of bdd_nodeBlocks initially allocated */
    struct {
      bdd_nodeBlock_pointer block;
      UnsignedInteger index;
    } pointer;		/* a pointer to the next block/object at which to allocate */
    struct {
      boolean on;			/* is garbage-collection even turned on? */
      Integer halfspace;		/* either 0 or 1, indicating which halfspace is in use */
      float node_ratio;           /* ratio of used to unused nodes */
      struct {
	Integer open_generators;	/* must be zero to perform a garbage-collect */
      } status;
      struct {
	struct {
	  bdd_nodeBlock_pointer block;
	  UnsignedInteger index;
	} start;		/* to communicate where to start gc passes */
      } during;			/* used _during_ gc only */
    } gc;			/* garbage-collection relevancies */
    bdd_stats stats;
  } heap;
  struct {
    bdd_node_pointer one;			/* the constant 1 (there is no constant 0) */
    UnsignedInteger nvariables;	/* variables in the bdd (same as array_n(order) */
  } bdd;			/* stuff relevant to bdd's and algorithms on bdd's */
  bdd_external_hooks hooks;
  struct {
    void (*daemon)();               /* used for callback when memory limit exceeded */
    UnsignedInteger limit;             /* upper bound on memory allocated by the manager; in megabytes */
  } memory;
}bdd_manager;

/*
 *    boolean
 *    BDD_IS_CONSTANT(manager, node)
 *    bdd_manager_pointer manager;
 *    bdd_node_pointer node;
 *
 */
#define BDD_IS_CONSTANT(manager, node)	(BDD_REGULAR(node) == (manager)->bdd.one)
/*
 *    The constants available in the system
 *
 *    bdd_node_pointer 
 *    BDD_ONE(manager)
 *    bdd_manager_pointer manager;
 *
 *        the constant one
 *
 *    bdd_node_pointer 
 *    BDD_ZERO(manager)
 *    bdd_manager_pointer manager;
 *
 *        the constant zero (derived from one by complementation)
 */
#define BDD_ONE(manager)		(manager)->bdd.one
#define BDD_ZERO(manager)		BDD_NOT(BDD_ONE(manager))

#define ALLOC_BDD_MANAGER ((bdd_manager_pointer) malloc(sizeof(bdd_manager)))

/*
 *    int bdd_ITE_hash(manager, f, g, h) is used in bdd_hashcache_entry_hash, garb_collect.c, hash_cache.c
 *    bdd_manager_pointer manager;
 *    bdd_node_pointer f;		- may not be nil
 *    bdd_node_pointer g;		- may not be nil
 *    bdd_node_pointer h;		- may not be nil
 */
#define bdd_ITE_hash(manager, f, g, h)	\
	bdd_generic_hash(f, g, h, (manager)->heap.cache.itetable.nbuckets)

/*
 *    int bdd_const_hash(manager, f, g, h) is used in bdd_constcache_entry_hash, garb_collect.c, const_cache.c
 *    bdd_manager_pointer manager;
 *    bdd_node_pointer f;		- may not be nil
 *    bdd_node_pointer g;		- may not be nil
 *    bdd_node_pointer h;		- may not be nil
 */
#define bdd_const_hash(manager, f, g, h)	\
	bdd_generic_hash(f, g, h, (manager)->heap.cache.consttable.nbuckets)

/*
 *    int bdd_raw_node_hash(manager, variableId, T, E) is used in bdd_node_hash, find_or_add.c
 *    bdd_manager_pointer manager;
 *    bdd_variableId variableId;
 *    bdd_node_pointer T;		- may be nil
 *    bdd_node_pointer E;		- may be nil
 */
#define bdd_raw_node_hash(manager, variableId, T, E)			\
	bdd_generic_hash(variableId, T, E, (manager)->heap.hashtable.nbuckets)

/*
 *    int bdd_node_hash(manager, node) is used in garb_collect.c, resize_table.c
 *    bdd_manager_pointer manager;
 *    bdd_node_pointer node;		- may not be nil
 */
#define bdd_node_hash(manager, node)	\
	bdd_raw_node_hash(manager, (node)->id, BDD_REGULAR(node)->T, BDD_REGULAR(node)->E)

/*
 *    int
 *    bdd_hashcache_entry_hash(manager, entry, nbuckets)
 *    bdd_manager_pointer manager;
 *    bdd_hashcache_entry *entry;	- may not be nil
 */
#define bdd_hashcache_entry_hash(manager, entry)	\
	bdd_ITE_hash(manager, (entry)->ITE.f, (entry)->ITE.g, (entry)->ITE.h)

/*
 *    int
 *    bdd_constcache_entry_hash(manager, entry)
 *    bdd_manager_pointer manager;
 *    bdd_constcache_entry *entry;	- may not be nil
 */
#define bdd_constcache_entry_hash(manager, entry)	\
	bdd_const_hash(manager, (entry)->ITE.f, (entry)->ITE.g, (entry)->ITE.h)

/*
 *    void
 *    bdd_safeframe__start(manager, sf)
 *    bdd_manager_pointer manager;	- by value
 *    bdd_safeframe sf;		- by value (sortof - by reference b/c its a macro)
 *
 *    void
 *    bdd_safeframe__end(manager)
 *    bdd_manager_pointer manager;	- by value
 *
 *    Usage: DO NOT USE THESE (they are only useful in bdd_new_node)
 *           use the versions below which do the assertion checks.
 */
#define bdd_safeframe__start(manager, sf)	\
		    { (sf).nodes = (bdd_safenode *)NULL; \
		      (sf).prev = (manager)->heap.internal_refs.frames; \
		      (manager)->heap.internal_refs.frames = &(sf); }
#define bdd_safeframe__end(manager)			\
		    { (manager)->heap.internal_refs.frames = \
		              (manager)->heap.internal_refs.frames->prev; }

/*
 *    void
 *    bdd_safenode__link(manager, sn, n)
 *    bdd_manager_pointer manager;	- by value
 *    bdd_safenode sn;		- by value (sortof - by reference b/c its a macro)
 *    bdd_node_pointer n;		- by value (sortof - by reference b/c its a macro)
 *
 *    void
 *    bdd_safenode__declare(manager, sn, n)
 *    bdd_manager_pointer manager;	- by value
 *    bdd_safenode sn;		- by value (sortof - by reference b/c its a macro)
 *
 *    Usage: DO NOT USE THESE (they are only useful in bdd_new_node)
 *           use the versions below which do the assertion checks.
 */
#define bdd_safenode__link(manager, sn, n)	\
  { (sn).node = (NIL_bdd_node_pointer);		\
		      (sn).arg = &(n); \
		      (sn).next = (manager)->heap.internal_refs.frames->nodes; \
		      (manager)->heap.internal_refs.frames->nodes = &(sn); }

#define bdd_safenode__declare(manager, sn)	\
  { (sn).node = (NIL_bdd_node_pointer);			 \
    (sn).arg = (NIL_bdd_node_pointer_star );					\
		      (sn).next = (manager)->heap.internal_refs.frames->nodes; \
		      (manager)->heap.internal_refs.frames->nodes = &(sn); }

/*
 *    void
 *    bdd_safenode_link(manager, sn, n)
 *    bdd_manager_pointer manager;	- by value
 *    bdd_safenode sn;		- by value (sortof - by reference b/c its a macro)
 *    bdd_node_pointer n;		- by value (sortof - by reference b/c its a macro)
 *
 *    void
 *    bdd_safenode_declare(manager, sn, n)
 *    bdd_manager_pointer manager;	- by value
 *    bdd_safenode sn;		- by value (sortof - by reference b/c its a macro)
 *
 *    Usage: (see below)
 */
#define bdd_safenode_link(manager, sn, n)	\
		    { bdd_safenode__link(manager, sn, n); }

#define bdd_safenode_declare(manager, sn)	\
		    { bdd_safenode__declare(manager, sn); }

/*
 *    void
 *    bdd_safeframe_start(manager, sf)
 *    bdd_manager_pointer manager;	- by value
 *    bdd_safeframe sf;		- by value (sortof - by reference b/c its a macro)
 *
 *    void
 *    bdd_safeframe_end(manager)
 *    bdd_manager_pointer manager;	- by value
 *
 *    Usage: (see below)
 */
#define bdd_safeframe_start(manager, sf)	\
		    { bdd_safeframe__start(manager, sf); }
#define bdd_safeframe_end(manager)		\
		    { bdd_safeframe__end(manager); }

