/* PLG 2011/05/09
 * extracted from bdd_int
 */
/*
 *    The bdd_node heap is made up of a chain of nodeBlocks, each
 *    of which contain some number of bdd_nodes.
 *
 *    Note now magical the incantiation is for BDD_NODES_PER_NODEBLOCK.
 *    This is done to ensure that the memory allocator which underlys
 *    this package is not too smart for our own good.  Typical behaviors
 *    are that memory allocators add their internal overhead, round
 *    up to the nextclosest power of 2 and allocate that amount of space
 *    to satisfy the user's (large) space request.   We want that internal
 *    number to be exactly a power of two (or real close to one but less-than)
 *
 *    Want: BDD_NODES_PER_NODEBLOCK == 2^k - BDD_MALLOC_OVERHEAD_ESTIMATE
 *    ... for appropriate value of k (k is called BDD_NODEBLOCK_PO2 here)
 */
#define BDD_MALLOC_OVERHEAD_ESTIMATE		8	/* assume malloc adds 4 bytes of header info */

#define BDD_NODEBLOCK_PO2			15	/* 2^15 = 32K - the magic Power 'O 2 (PO2) */

#define BDD_NODES_PER_NODEBLOCK			\
	(((1<<(BDD_NODEBLOCK_PO2-1)) -		\
	  BDD_MALLOC_OVERHEAD_ESTIMATE -	\
	  /* used */ sizeof (Integer) -		\
	  /* next */ sizeof (long)) / sizeof (bdd_node))

typedef struct bdd_nodeBlock {
    Integer used;
    bdd_nodeBlock_pointer next;
    struct __bdd_node subheap[BDD_NODES_PER_NODEBLOCK];
} bdd_nodeBlock;	/* total size plus malloc overhead should be an exact power of 2 */
