#include <stdlib.h>
#include "BDD_StdTypes.h"
#include "ansi.h"
#include "Alloc.h"

#include <stddef.h>
#include "boolean.h"
#include "bdd_export_type.h"
#include "bdd_export.h"

#include "st.h"

#include "bdd_int.h"
#include "bdd_int_hashcache.h"
#include "bdd_int_safenodeframe.h"
#include "bdd_int_bddBlock.h"
#include <stdio.h>
#include "bdd_inner.h"
#include "bdd_int_manager.h"

#include "st_g.h"
#include "bdd_util_g.h"
#include "safe_mem_g.h"

/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/adhoc_cache.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:01:27 $
 * $Log: adhoc_cache.c,v $
 * Revision 1.2  1992/09/19  02:01:27  shiple
 * Version 2.4
 * Prefaced compile time debug switches with BDD_. Added code to limit size of adhoc cache.
 * In bdd_adhoccache_lookup, just return if caching is off. Added usage of
 * bdd_will_exceed_mem_limit in bdd_adhoccache_insert.
 *
 * Revision 1.1  1992/07/29  00:26:36  shiple
 * Initial revision
 *
 * Revision 1.3  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.2  1992/04/24  22:51:34  sis
 * Core leak fixed.
 *
 * Revision 1.1  91/03/27  14:31:23  shiple
 * Initial revision
 * 
 * Revision 1.1  91/03/27  14:29:37  shiple
 * Initial revision
 * 
 *
 */

/*
 *    cmp - a st_table comparison function
 *
 *    return {0, 1} on {same, different}.
 */
static Integer
cmp (bdd_adhoccache_key * n1, bdd_adhoccache_key * n2)
{
  boolean samep;

  samep = n1->f == n2->f && n1->g == n2->g && n1->v == n2->v;
  return (samep ? /* equal */ 0 : /* different */ 1);
}

/* 
 *    hash - a st_table hash function
 *
 *    return an integer in the range 0 .. modulus-1
 */
static Integer
hash (bdd_adhoccache_key * key, Integer modulus)
{
  Integer hval;

  hval = bdd_generic_hash (key->f, key->g, key->v, modulus);
  return (hval);
}

/*
 *    bdd_adhoccache_init - initialize the adhoc cache for the manager
 *
 *    return nothing, just do it.
 */
void
bdd_adhoccache_init (bdd_manager_pointer manager)
{
  manager->heap.cache.adhoc.table = st_init_table (cmp, hash);
}

/*
 *    bdd_adhoccache_insert - insert an entry into the adhoc cache
 *
 *    return nothing, just do it.
 */
void
bdd_adhoccache_insert (bdd_manager_pointer manager, bdd_node_pointer f, bdd_node_pointer g, bdd_int v, bdd_node_pointer value)
			/* may be (NIL_bdd_node_pointer) */
			/* may be (NIL_bdd_node_pointer) */
			/* may be any integer */
{
  bdd_adhoccache_key *key;
  Integer status;

  if (!manager->heap.cache.adhoc.on)
    {
      return;			/* has no effect if caching is off */
    }

  /* 
   * If we have already reached the max number of entries, then just
   * return.  Note that manager->heap.stats.cache.adhoc.inserts is not incremented.
   */
  if (st_count (manager->heap.cache.adhoc.table) > manager->heap.cache.adhoc.max_size)
    {
      return;
    }

  /* 
   * If allocating a bdd_adhoccache_key will cause the manager's memory limit to be
   * exceeded, then just return.   Note that manager->heap.stats.cache.adhoc.inserts 
   * is not incremented.
   */
  if (bdd_will_exceed_mem_limit (manager, sizeof (bdd_adhoccache_key), pk_false) == pk_true)
    {
      return;
    }

  /*
   * Allocate a new key.
   */
  key = ALLOC (bdd_adhoccache_key, 1);
  if (key == ((bdd_adhoccache_key *) NULL))
    {
      (void) bdd_memfail (manager, "bdd_adhoccache_insert");
    }

  /*
   * Set the values of the key.
   */
  key->f = f;
  key->g = g;
  key->v = v;

  /*
   * Insert the new key.  Since we previous lookup on this entry was not successful, there should
   * not be another entry already under the key.  st_insert returns 1 if there was an entry already under
   * the key, and 0 otherwise.  Since the table has collision chains, there should not be any collisions.
   */
  status = st_insert (manager->heap.cache.adhoc.table, (char *) key, (char *) value);
  manager->heap.stats.cache.adhoc.inserts++;
}

/*
 *    bdd_adhoccache_lookup - lookup an entry in the adhoc cache
 *
 *    return {pk_true, pk_false} on {found, not found}.
 */
boolean
bdd_adhoccache_lookup (bdd_manager_pointer manager, bdd_node_pointer f, bdd_node_pointer g, bdd_int v, bdd_node_pointer * value)
			/* may be (NIL_bdd_node_pointer) */
			/* may be (NIL_bdd_node_pointer) */
			/* may be any integer */
			/* return */
{
  bdd_adhoccache_key key;


  if (!manager->heap.cache.adhoc.on)
    {
      /*
       * Always fails if caching is off.
       */
      return (pk_false);
    }

  key.f = f;
  key.g = g;
  key.v = v;

  if (st_lookup (manager->heap.cache.adhoc.table, (char *) &key, (char **) value))
    {
      manager->heap.stats.cache.adhoc.hits++;
      return (pk_true);
    }

  manager->heap.stats.cache.adhoc.misses++;
  return (pk_false);
}

/*
 *    bdd_adhoccache_uninit - destroy a adhoc cache
 *
 *    delete all the keys
 *
 *    return nothing, just do it.
 */
void
bdd_adhoccache_uninit (bdd_manager_pointer manager)
{
  st_generator *gen;
  bdd_adhoccache_key *key;
  bdd_node_pointer value;

  st_foreach_item (manager->heap.cache.adhoc.table, gen, (char **) &key, (char **) &value)
  {
    FREE (key);
  }
  st_free_table (manager->heap.cache.adhoc.table);
  manager->heap.cache.adhoc.table = (NIL_st_table);
}
