/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd_cofactor.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:23:19 $
 *
 */

#include <stddef.h>
#include "boolean.h"
#include "BDD_StdTypes.h"
#include "bdd_export_type.h"
#include "ansi.h"
#include "bdd_export.h"

#include "bdd_int_safenodeframe.h"
#include "bdd_int_hashcache.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "util.h"

#include "id_ord.h"
#include "bdd_int.h"
#include "bdd_int_bddBlock.h"
#include "bdd_inner.h"
#include "bdd_int_manager.h"

#include "id_ord_g.h"
#include "bdd_util_g.h"
#include "adhoc_cache_g.h"
#include "external_ptr_g.h"
#include "find_or_add_g.h"
#include "ite_g.h"

/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd_cofactor.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:23:19 $
 * $Log: bdd_cofactor.c,v $
 * Revision 1.2  1992/09/19  02:23:19  shiple
 * Version 2.4
 * > Prefaced compile time debug switches with BDD_. Added typecast to void to some function calls.
 * Added adhoc_ops stats. Restructured _cofactor so that the cases where one branch of g
 * is ZERO are properly hashed and accounted for in the statistics.
 *
 * Revision 1.1  1992/07/29  00:26:44  shiple
 * Initial revision
 *
 * Revision 1.2  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.1  92/01/08  17:34:25  sis
 * Initial revision
 * 
 * Revision 1.1  91/03/27  14:35:28  shiple
 * Initial revision
 * 
 *
 */

/*
 *    ARGH!   The word ``cofactor'' is an exported function in mis,
 *    so we are not allowed to use it here!   ARGH!  These guys don't
 *    believe in package prefixes!  ARGH!
 */
static bdd_node_pointer _cofactor (bdd_manager_pointer manager, bdd_node_pointer f, bdd_node_pointer g);

/*
 *    bdd_cofactor - cofactor f by g
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_cofactor (bdd_t_pointer f, bdd_t_pointer g)
{
  bdd_safeframe frame;
  bdd_safenode ret;
  bdd_manager_pointer manager;
  bdd_t_pointer h;

  if (f == (NIL_bdd_t_pointer) || g == (NIL_bdd_t_pointer))
    fail ("bdd_cofactor: invalid BDD");

  if (g->node == BDD_ZERO (g->bdd))
    fail ("bdd_cofactor: cofactor wrt zero not defined");

  if (f->bdd != g->bdd)
    fail ("bdd_cofactor: different bdd managers");

  manager = f->bdd;		/* either this or g->bdd will do */

  /*
   *    Start the safe frame now that we know the input is correct.
   *    f and g are external pointers so they need not be protected
   */
  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, ret);

  (void) bdd_adhoccache_init (manager);

  ret.node = _cofactor (manager, f->node, g->node);

  /*
   *    Free the cache, end the safe frame and return the (safe) result
   */
  (void) bdd_adhoccache_uninit (manager);
  h = bdd_make_external_pointer (manager, ret.node);
  bdd_safeframe_end (manager);

  return (h);
}

/*
 *    _cofactor - recursively perform the cofactoring
 *
 *    ARGH!   The word ``cofactor'' is an exported function in mis,
 *    so we are not allowed to use it here!   ARGH!  These guys don't
 *    believe in package prefixes!  ARGH!
 *
 *    return the result of the reorganization
 */
static bdd_node_pointer
_cofactor (bdd_manager_pointer manager, bdd_node_pointer f, bdd_node_pointer g)
{
  bdd_safeframe frame;
  bdd_safenode safe_f, safe_g, ret, f_T, f_E, g_T, g_E, var, co_T, co_E;
  bdd_variableId fId, gId;

  bdd_safeframe_start (manager, frame);
  bdd_safenode_link (manager, safe_f, f);
  bdd_safenode_link (manager, safe_g, g);
  bdd_safenode_declare (manager, ret);
  bdd_safenode_declare (manager, f_T);
  bdd_safenode_declare (manager, f_E);
  bdd_safenode_declare (manager, g_T);
  bdd_safenode_declare (manager, g_E);
  bdd_safenode_declare (manager, var);
  bdd_safenode_declare (manager, co_T);
  bdd_safenode_declare (manager, co_E);

  manager->heap.stats.adhoc_ops.calls++;

  if (BDD_IS_CONSTANT (manager, f))
    {
      /*
       *    f is either zero or one
       */
      manager->heap.stats.adhoc_ops.returns.trivial++;
      bdd_safeframe_end (manager);
      return (f);
    }

  if (g == BDD_ONE (manager))
    {
      /*
       *    If the thing to cofactor by is a constant
       *    then the result is just the function itself
       */
      manager->heap.stats.adhoc_ops.returns.trivial++;
      bdd_safeframe_end (manager);
      return (f);
    }

  fId = BDD_REGULAR (f)->id;
  gId = BDD_REGULAR (g)->id;

  (void) bdd_get_branches (f, &f_T.node, &f_E.node);
  (void) bdd_get_branches (g, &g_T.node, &g_E.node);

  /*
   *    If there is some possibility that the function may have
   *    been computed before then look up the function in the cache ...
   */
  if (bdd_adhoccache_lookup (manager, f, g, /* v */ 0, &ret.node))
    {
      /*
       *    The answer was already in the cache, so just return it.
       */
      manager->heap.stats.adhoc_ops.returns.cached++;
      bdd_safeframe_end (manager);
      return (ret.node);
    }

  /*
   *    In the following code, we must worry about taking a _cofactor
   *    by the zero function.  So, in each case this is special-cased.
   *    For those cases where we call _cofactor recursively on both branches,
   *    we subsequently call bdd_find_or_add to get the node (var, 1, 0),
   *    where the id of var = (fId > gId) ? fId: gId.
   */
  if (bdd_get_ord (fId) > bdd_get_ord (gId))
    {
      if (g_E.node == BDD_ZERO (manager))
	{
	  /*
	   *    cofactor(f, g_T)
	   */
	  ret.node = _cofactor (manager, f, g_T.node);
	}
      else if (g_T.node == BDD_ZERO (manager))
	{
	  /*
	   *    cofactor(f, g_E)
	   */
	  ret.node = _cofactor (manager, f, g_E.node);
	}
      else
	{
	  /*
	   *    ITE(v, cofactor(f, g_T), cofactor(f, g_E))
	   */
	  co_T.node = _cofactor (manager, f, g_T.node);
	  co_E.node = _cofactor (manager, f, g_E.node);
	  var.node = bdd_find_or_add (manager, gId, BDD_ONE (manager), BDD_ZERO (manager));
	  ret.node = bdd__ITE_ (manager, var.node, co_T.node, co_E.node);
	}
    }
  else if (fId == gId)
    {
      if (g_E.node == BDD_ZERO (manager))
	{
	  /*
	   *    cofactor(f_T, g_T)
	   */
	  ret.node = _cofactor (manager, f_T.node, g_T.node);
	}
      else if (g_T.node == BDD_ZERO (manager))
	{
	  /*
	   *    cofactor(f_E, g_E)
	   */
	  ret.node = _cofactor (manager, f_E.node, g_E.node);
	}
      else
	{
	  /*
	   *    ITE(fId, cofactor(f_T, g_T), cofactor(f_E, g_E))
	   */
	  co_T.node = _cofactor (manager, f_T.node, g_T.node);
	  co_E.node = _cofactor (manager, f_E.node, g_E.node);
	  var.node = bdd_find_or_add (manager, fId, BDD_ONE (manager), BDD_ZERO (manager));
	  ret.node = bdd__ITE_ (manager, var.node, co_T.node, co_E.node);
	}
    }
  else
    {				/* fId < gId */
      /*
       *    ITE(fId, cofactor(f_T, g), cofactor(f_E, g))
       */
      co_T.node = _cofactor (manager, f_T.node, g);
      co_E.node = _cofactor (manager, f_E.node, g);
      var.node = bdd_find_or_add (manager, fId, BDD_ONE (manager), BDD_ZERO (manager));
      ret.node = bdd__ITE_ (manager, var.node, co_T.node, co_E.node);
    }

  (void) bdd_adhoccache_insert (manager, f, g, /* v */ 0, ret.node);
  manager->heap.stats.adhoc_ops.returns.full++;

  bdd_safeframe_end (manager);
  return (ret.node);
}
