/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd_start.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:46:41 $
 *
 */

#include <string.h>

#include <stdlib.h>
#include "BDD_StdTypes.h"
#include "ansi.h"
#include "Alloc.h"

#include <stddef.h>
#include "boolean.h"
#include "bdd_export_type.h"
#include "bdd_export.h"
#include "id_ord.h"
#include "bdd_int.h"
#include "bdd_int_hashcache.h"
#include "bdd_int_safenodeframe.h"
#include "bdd_int_bddBlock.h"
#include <stdio.h>
#include "bdd_inner.h"
#include "bdd_int_manager.h"
#include "bdd_util_g.h"
#include "find_or_add_g.h"
#include "safe_mem_g.h"

/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd_start.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:46:41 $
 * $Log: bdd_start.c,v $
 * Revision 1.2  1992/09/19  02:46:41  shiple
 * Version 2.4
 * Prefaced compile time debug switches with BDD_. Added typecast to void to some function calls.
 * Initialized new fields added to manager and stats data structures.  Added use of bdd_mgr_init.
 * Changed ITE and ITE_const caches to be arrays of pointers.  Added bdd_register_daemon.
 *
 * Revision 1.1  1992/07/29  00:26:53  shiple
 * Initial revision
 *
 * Revision 1.2  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.1  92/01/08  17:34:29  sis
 * Initial revision
 * 
 * Revision 1.1  91/03/27  14:35:34  shiple
 * Initial revision
 * 
 *
 */

/*
 * bdd_set_mgr_init_dflts - initialize the input bdd_mgr_init with the default settings
 */
static void
bdd_set_mgr_init_dflts (st_bdd_mgr_init_pointer mgr_init)
{
  mgr_init->ITE_cache.on = pk_true;
  mgr_init->ITE_cache.resize_at = BDD_DFLT_ITE_RESIZE_AT;
  mgr_init->ITE_cache.max_size = BDD_DFLT_ITE_MAX_SIZE;
  mgr_init->ITE_const_cache.on = pk_true;
  mgr_init->ITE_const_cache.resize_at = BDD_DFLT_ITE_CONST_RESIZE_AT;
  mgr_init->ITE_const_cache.max_size = BDD_DFLT_ITE_CONST_MAX_SIZE;
  mgr_init->adhoc_cache.on = pk_true;
  mgr_init->adhoc_cache.resize_at = BDD_DFLT_ADHOC_RESIZE_AT;
  mgr_init->adhoc_cache.max_size = BDD_DFLT_ADHOC_MAX_SIZE;
  mgr_init->garbage_collector.on = pk_true;
  mgr_init->memory.daemon = BDD_DFLT_DAEMON;
  mgr_init->memory.limit = BDD_DFLT_MEMORY_LIMIT;
  mgr_init->nodes.ratio = BDD_DFLT_NODE_RATIO;
  mgr_init->nodes.init_blocks = BDD_DFLT_INIT_BLOCKS;
}

/*
 *    bdd_start_with_params - initialize a bdd manager with the supplied parameters
 *
 *    return the new bdd manager
 */
static bdd_manager_pointer
bdd_start_with_params (Integer nvariables, st_bdd_mgr_init_pointer mgr_init)
			/* the maximum number of variables allowable in this bdd manager */
{
  bdd_manager_pointer manager;
  Integer i, h;

  //    manager = ALLOC(bdd_manager, 1);
  manager = ALLOC_BDD_MANAGER;
  if (manager == (NIL_bdd_manager_pointer))
    (void) bdd_memfail ((NIL_bdd_manager_pointer), "bdd_start");

  manager->undef1 = 0;		/* for debugging purposes */

  /*
   *    Initialize the heap part of the bdd_manager
   */

  manager->heap.hashtable.nkeys = 0;
  manager->heap.hashtable.nbuckets = BDD_HASHTABLE_INITIAL_SIZE;
  manager->heap.hashtable.buckets = ALLOC (bdd_node_pointer, manager->heap.hashtable.nbuckets);
  if (manager->heap.hashtable.buckets == (NIL_bdd_node_pointer_star))
    (void) bdd_memfail ((NIL_bdd_manager_pointer), "bdd_start");
  manager->heap.hashtable.rehash_at_nkeys = manager->heap.hashtable.nbuckets * BDD_HASHTABLE_MAXCHAINLEN;
  for (i = 0; i < manager->heap.hashtable.nbuckets; i++)
    {
      manager->heap.hashtable.buckets[i] = (NIL_bdd_node_pointer);
    }

  manager->heap.external_refs.pointer.block = (NIL_bdd_bddBlock);
  manager->heap.external_refs.pointer.index = 0;
  manager->heap.external_refs.nmap = 0;
  manager->heap.external_refs.free = 0;
  manager->heap.external_refs.map = (NIL_bdd_bddBlock);

  manager->heap.internal_refs.frames = (NIL_bdd_safeframe);

  manager->heap.cache.itetable.on = mgr_init->ITE_cache.on;
  manager->heap.cache.itetable.invalidate_on_gc = pk_false;
  manager->heap.cache.itetable.resize_at = mgr_init->ITE_cache.resize_at;
  manager->heap.cache.itetable.max_size = mgr_init->ITE_cache.max_size;
  manager->heap.cache.itetable.nbuckets = BDD_CACHE_INITIAL_SIZE;
  manager->heap.cache.itetable.nentries = 0;
  manager->heap.cache.itetable.buckets = ALLOC (bdd_hashcache_entry *, manager->heap.cache.itetable.nbuckets);
  if (manager->heap.cache.itetable.buckets == (NIL_bdd_hashcache_entry_star))
    {
      (void) bdd_memfail ((NIL_bdd_manager_pointer), "bdd_start");
    }
  for (i = 0; i < manager->heap.cache.itetable.nbuckets; i++)
    {
      manager->heap.cache.itetable.buckets[i] = (NIL_bdd_hashcache_entry);
    }

  manager->heap.cache.consttable.on = mgr_init->ITE_const_cache.on;
  manager->heap.cache.consttable.invalidate_on_gc = pk_false;
  manager->heap.cache.consttable.resize_at = mgr_init->ITE_const_cache.resize_at;
  manager->heap.cache.consttable.max_size = mgr_init->ITE_const_cache.max_size;
  manager->heap.cache.consttable.nbuckets = BDD_CACHE_INITIAL_SIZE;
  manager->heap.cache.consttable.nentries = 0;
  manager->heap.cache.consttable.buckets = ALLOC (bdd_constcache_entry *, manager->heap.cache.consttable.nbuckets);
  if (manager->heap.cache.consttable.buckets == (NIL_bdd_constcache_entry_star))
    {
      (void) bdd_memfail ((NIL_bdd_manager_pointer), "bdd_start");
    }
  for (i = 0; i < manager->heap.cache.consttable.nbuckets; i++)
    {
      manager->heap.cache.consttable.buckets[i] = (NIL_bdd_constcache_entry);
    }

  manager->heap.cache.adhoc.on = mgr_init->adhoc_cache.on;	/* if defined, caching is on */
  manager->heap.cache.adhoc.max_size = mgr_init->adhoc_cache.max_size;
  manager->heap.cache.adhoc.table = (NIL_st_table);	/* initially undefined */

  for (h = 0; h < sizeof_el (manager->heap.half); h++)
    {
      manager->heap.half[h].inuse.top = (NIL_bdd_nodeBlock);
      manager->heap.half[h].inuse.tail = &manager->heap.half[h].inuse.top;
      manager->heap.half[h].free = (NIL_bdd_nodeBlock);
    }

  manager->heap.init_node_blocks = mgr_init->nodes.init_blocks;
  manager->heap.pointer.block = (NIL_bdd_nodeBlock);
  manager->heap.pointer.index = 0;

  manager->heap.gc.on = mgr_init->garbage_collector.on;
  manager->heap.gc.halfspace = 0;	/* start at halfspace 0 */
  manager->heap.gc.node_ratio = mgr_init->nodes.ratio;
  manager->heap.gc.status.open_generators = 0;


  manager->heap.gc.during.start.index = 0;
  manager->heap.gc.during.start.block = (NIL_bdd_nodeBlock);
  manager->memory.daemon = mgr_init->memory.daemon;
  manager->memory.limit = mgr_init->memory.limit;

  /*
   *    Clear out the stats
   */
  manager->heap.stats.cache.hashtable.hits = 0;
  manager->heap.stats.cache.hashtable.misses = 0;
  manager->heap.stats.cache.hashtable.collisions = 0;
  manager->heap.stats.cache.hashtable.inserts = 0;

  manager->heap.stats.cache.itetable = manager->heap.stats.cache.hashtable;
  manager->heap.stats.cache.consttable = manager->heap.stats.cache.hashtable;
  manager->heap.stats.cache.adhoc = manager->heap.stats.cache.hashtable;

  manager->heap.stats.ITE_ops.calls = 0;
  manager->heap.stats.ITE_ops.returns.trivial = 0;
  manager->heap.stats.ITE_ops.returns.cached = 0;
  manager->heap.stats.ITE_ops.returns.full = 0;

  manager->heap.stats.ITE_constant_ops = manager->heap.stats.ITE_ops;
  manager->heap.stats.adhoc_ops = manager->heap.stats.ITE_ops;

  manager->heap.stats.blocks.total = 0;

  manager->heap.stats.nodes.used = 0;
  manager->heap.stats.nodes.unused = 0;
  manager->heap.stats.nodes.total = 0;
  manager->heap.stats.nodes.peak = 0;

  manager->heap.stats.extptrs.used = 0;
  manager->heap.stats.extptrs.total = manager->heap.external_refs.nmap;
  manager->heap.stats.extptrs.unused = manager->heap.stats.extptrs.total;
  manager->heap.stats.extptrs.blocks = 0;

  manager->heap.stats.gc.times = 0;
  manager->heap.stats.gc.nodes_collected = 0;
  manager->heap.stats.gc.runtime = 0;

  manager->heap.stats.memory.first_sbrk = manager->heap.stats.memory.last_sbrk = (Integer) malloc (0);
  manager->heap.stats.memory.manager = 0;
  manager->heap.stats.memory.nodes = 0;
  manager->heap.stats.memory.hashtable = 0;
  manager->heap.stats.memory.ext_ptrs = 0;
  manager->heap.stats.memory.ITE_cache = 0;
  manager->heap.stats.memory.ITE_const_cache = 0;
  manager->heap.stats.memory.adhoc_cache = 0;
  manager->heap.stats.memory.total = 0;


  /*
   *    Now initialize all of the bdd-algorithm stuff. This first call to find_or_add causes 
   *    manager->heap.init_node_blocks to be allocated. 
   */
  manager->bdd.one = bdd_find_or_add (manager, /* variableId */ BDD_ONE_ID, (NIL_bdd_node_pointer), (NIL_bdd_node_pointer));
  manager->bdd.nvariables = nvariables;

  /*
   *    With the understanding that zeroing out the structure will
   *    mean niling out all of the pointers inside the structure.
   */
  memset (&manager->hooks, 0, sizeof (manager->hooks));

  return (manager);
}

/*
 *    bdd_start - initialize a bdd manager
 *
 *    Given the number of variables, create and initialize a bdd_manager with
 *    the default settings.
 *
 *    return the new bdd manager
 */
bdd_manager_pointer
bdd_start (Integer nvariables)
			/* the maximum number of variables allowable in this bdd manager */
{
  bdd_mgr_init mgr_init;

  (void) bdd_set_mgr_init_dflts (&mgr_init);

  return (bdd_start_with_params (nvariables, &mgr_init));
}

/* Modified by Loic Besnard.
 * bdd_register_daemon
 *
 * Register the function to be called when the memory limit is exceeded.  
void
bdd_register_daemon(manager, daemon)
bdd_manager_pointer manager;
void (*daemon)();
{
    manager->memory.daemon = daemon; 
    return;
}
 */
