/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/find_or_add.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 03:05:06 $
 *
 */

#include <stddef.h>
#include "boolean.h"
#include "BDD_StdTypes.h"
#include "bdd_export_type.h"
#include "ansi.h"
#include "bdd_export.h"

#include "bdd_int.h"
#include "bdd_int_hashcache.h"
#include "bdd_int_safenodeframe.h"
#include "bdd_int_bddBlock.h"
#include <stdio.h>
#include "bdd_inner.h"
#include "bdd_int_manager.h"

#include "id_ord.h"
#include "id_ord_g.h"
#include "new_node_g.h"
#include "resize_table_g.h"

#define BDD_COMPUTE_LAST(node)  \
{ \
   (node)->last = (node)->id; \
   if (((node)->T != (NIL_bdd_node_pointer)) && (! BDD_IS_CONSTANT(manager,(node)->T))) \
     (node)->last = BDD_REGULAR((node)->T)->last; \
   if (((node)->E != (NIL_bdd_node_pointer)) && (! BDD_IS_CONSTANT(manager,(node)->E))) \
     if (bdd_get_ord((node)->last) < bdd_get_ord(BDD_REGULAR((node)->E)->last)) \
       (node)->last = BDD_REGULAR((node)->E)->last; \
}

/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/find_or_add.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 03:05:06 $
 * $Log: find_or_add.c,v $
 * Revision 1.2  1992/09/19  03:05:06  shiple
 * > Version 2.4
 * Prefaced compile time debug switches with BDD_.  Renamed resize_tables to resize_hashtable.
 * Added code to update hash hits/misses; moved/added comments.
 *
 * Revision 1.1  1992/07/29  00:27:02  shiple
 * Initial revision
 *
 * Revision 1.2  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.1  92/01/08  17:34:38  sis
 * Initial revision
 * 
 * Revision 1.1  91/03/27  14:35:41  shiple
 * Initial revision
 * 
 *
 */

/*
 *    bdd_find_or_add - find a node in the hashtable or add it
 *
 *    Find a node whose id is given and which has T, E for its data.
 *    If no such node is in the table, then make a new one.
 *
 *    return the (possibly new) node in the hashtable
 */
bdd_node_pointer
bdd_find_or_add (bdd_manager_pointer manager, bdd_variableId variableId, bdd_node_pointer T, bdd_node_pointer E)
{
  bdd_safeframe frame;
  bdd_safenode safe_T, safe_E, new, chain, ret;
  Integer pos;

  bdd_safeframe_start (manager, frame);
  bdd_safenode_link (manager, safe_T, T);
  bdd_safenode_link (manager, safe_E, E);
  bdd_safenode_declare (manager, new);
  bdd_safenode_declare (manager, chain);
  bdd_safenode_declare (manager, ret);

  /*
   *    Search the chain (short, its length is supposed
   *    to be less than BDD_HASHTABLE_MAXCHAINLEN).
   */

  pos = bdd_raw_node_hash (manager, variableId, T, E);

  for (chain.node = manager->heap.hashtable.buckets[pos]; chain.node != (NIL_bdd_node_pointer); chain.node = chain.node->next)
    {

      if (chain.node->id == variableId && chain.node->T == T && chain.node->E == E)
	{
	  /*
	   *    Found it, blow the frame and return the thing found.
	   *    Update the stats.
	   */
	  bdd_safeframe_end (manager);
	  manager->heap.stats.cache.hashtable.hits++;
	  return (chain.node);
	}
    }

  /*
   *    Its not there, so add it, but recall that adding a new node to
   *    the hashtable may cause a garbage-collection.   So, we must ensure
   *    that all values that we want preserved across the gc are tied down.
   *
   *    If one more entry will cause the table to explode,
   *    then make the table bigger before going on.
   */
  if (manager->heap.hashtable.nkeys + 1 >= manager->heap.hashtable.rehash_at_nkeys)
    {
      (void) bdd_resize_hashtable (manager);	/* definitely don't do THIS if gc is occurring */
    }

  ret.node = bdd_new_node (manager, variableId, T, E, (NIL_bdd_node_pointer));

  BDD_COMPUTE_LAST (ret.node);	/* moi */

  /*
   *    Link in the new node. Recompute pos just in case a gc happenned (which would screw up the pos).
   */
  pos = bdd_raw_node_hash (manager, variableId, T, E);

  ret.node->next = manager->heap.hashtable.buckets[pos];
  manager->heap.hashtable.buckets[pos] = ret.node;
  manager->heap.hashtable.nkeys++;
  manager->heap.stats.cache.hashtable.misses++;

  bdd_safeframe_end (manager);
  return (ret.node);
}
