/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/new_node.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 03:19:08 $
 *
 */
#include <stdlib.h>
#include "BDD_StdTypes.h"
#include "ansi.h"
#include "Alloc.h"

#include <stddef.h>
#include "boolean.h"
#include "bdd_export_type.h"
#include "bdd_export.h"

#include "bdd_int.h"
#include "bdd_int_hashcache.h"
#include "bdd_int_safenodeframe.h"
#include "bdd_int_bddBlock.h"
#include <stdio.h>
#include "bdd_inner.h"
#include "bdd_int_manager.h"

#include "bdd_int_nodeBlock.h"

#include "bdd_util_g.h"
#include "garb_collect_g.h"
#include "safe_mem_g.h"

/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/new_node.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 03:19:08 $
 * $Log: new_node.c,v $
 * Revision 1.2  1992/09/19  03:19:08  shiple
 * Version 2.4
 * Prefaced compile time debug switches with BDD_.  Added peak node usage.
 * Fixed bug in function new_block which caused an additional node block
 * to be added to the free list at the start of each garbage collection.
 * Changed name new_block to get_node_block, added function get_block_from_free_list.
 * Added use of manager->heap.init_node_blocks and manager->heap.gc.node_ratio.
 * Add usage of bdd_will_exceed_mem_limit in add_block_to_free_list.
 *
 * Revision 1.1  1992/07/29  00:27:10  shiple
 * Initial revision
 *
 * Revision 1.2  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.1  92/01/08  17:34:40  sis
 * Initial revision
 * 
 * Revision 1.1  91/03/27  14:35:45  shiple
 * Initial revision
 * 
 *
 */

static void get_node_block (bdd_manager_pointer manager);
static Integer add_block_to_free_list (bdd_manager_pointer manager, Integer optionalp);
static void get_block_from_free_list (bdd_manager_pointer manager);


/*
 *    bdd_new_node - get a new node and fill in its value
 *
 *    The new node is not put in any hash table or cache
 *
 *    return the new node
 */
bdd_node_pointer
bdd_new_node (bdd_manager_pointer manager, bdd_variableId variableId, bdd_node_pointer T, bdd_node_pointer E, bdd_node_pointer old)
/* NIL for a new node, non-nil during gc (only used to communicate debug info) */
{
  bdd_safeframe frame;
  bdd_safenode safe_T, safe_E, safe_old, new;
  Integer index;

  /*
   *    WATCHOUT - be sure not to call the version of ``safe frame end''
   *    which does frame assertion checks.   This routine is called inside
   *    of the garbage-collector, and so the frames are in an inconsistent
   *    state by definition.    
   */
  bdd_safeframe__start (manager, frame);
  bdd_safenode__link (manager, safe_T, T);
  bdd_safenode__link (manager, safe_E, E);
  bdd_safenode__link (manager, safe_old, old);
  bdd_safenode__declare (manager, new);

  /*
   *    If this is the first block, or, the block is full,
   *    then get a new block for the heap.  This means getting a block from
   *    the free list, if it is nonempty, or allocating a new block.  If
   *    this function is called from relocate within bdd_garbage_collect, then
   *    it will always be the case that the free list is nonempty.  That's because
   *    the number of objects collected must be less than or equal to the original
   *    number of objects.
   */
  if (manager->heap.pointer.block == (NIL_bdd_nodeBlock) || manager->heap.pointer.index >= sizeof_el (manager->heap.pointer.block->subheap))
    {

      /*
       * It should never occur that we are during gc and the free list is empty.
       */
      (void) get_node_block (manager);
    }

  index = manager->heap.pointer.index++;
  manager->heap.pointer.block->used++;
  new.node = &manager->heap.pointer.block->subheap[index];

  new.node->id = variableId;
  /* new.node->last = BDD_USER_0_ID; ajouter lbesnard. */
  new.node->T = T;
  new.node->E = E;
  new.node->next = (NIL_bdd_node_pointer);

  /*
   *    Read between the #ifdefs carefully.  What is going on here is
   *    rather simple, but the ifdefs get too much in the way to make
   *    it very readable.
   *
   *    if (old == nil) {
   *        - then not during gc so ...
   *        - assign the current gc generation
   *        - assign a new uniqueId
   *        - log the creation of the object
   *    } else {
   *        - else it is during a gc
   *        - copy over the previous uniqueId
   *        - copy over the previus age (of creation)
   *        - do not log any creation (b/c it really wasn't)
   *    }
   */
  if (old == (NIL_bdd_node_pointer))
    {				/* not during gc */
    }
  else
    {				/* during gc */
    }


  manager->heap.stats.nodes.used++;
  manager->heap.stats.nodes.peak = MAX (manager->heap.stats.nodes.peak, manager->heap.stats.nodes.used);
  manager->heap.stats.nodes.unused--;

  /*
   *    WATCHOUT - be sure not to call the version of ``safe frame end''
   *    which does frame assertion checks.   This routine is called inside
   *    of the garbage-collector, and so the frames are in an inconsistent
   *    state by definition.    
   */
  bdd_safeframe__end (manager);
  return (new.node);
}

static void
get_node_block (bdd_manager_pointer manager)
{
  boolean request_status, return_status;
  Integer i;

  if (manager->heap.half[manager->heap.gc.halfspace].free != (NIL_bdd_nodeBlock))
    {
      /*
       * If the free list is not empty, then just take a block off the free list.
       * If we are garbage collecting and relocating objects, then this condition should
       * always be satisfied.
       */
      (void) get_block_from_free_list (manager);
    }
  else
    {
      if (manager->heap.half[manager->heap.gc.halfspace].inuse.top == (NIL_bdd_nodeBlock))
	{
	  /*
	   * The free list is empty and the inuse list is empty.  Thus, add manager->heap.init_node_blocks to
	   * the free list, and then take one off the free list (a little bit redundant).
	   * Use the request_status flag to make sure that we get at least one block.
	   */
	  request_status = pk_false;
	  for (i = 1; i <= manager->heap.init_node_blocks; i++)
	    {
	      return_status = add_block_to_free_list (manager, request_status);

	      /*
	       * If no more blocks can be allocated, then stop asking for more.
	       */
	      if (return_status == pk_false)
		{
		  break;
		}
	      request_status = pk_true;	/* getting more becomes optional */
	    }
	  (void) get_block_from_free_list (manager);
	}
      else
	{
	  /*
	   * The free list is empty, and the inuse list is full.  Time to garbage collect.
	   */
	  (void) bdd_garbage_collect (manager);

	  /*
	   * We want to provide a buffer for the heap to grow.  Otherwise, the heap
	   * could be, say 98% full, and we will be garbage collecting very frequently.
	   */

	  /*
	   * If no garbage was collected, then the inuse list will be filled up and the 
	   * free list will be empty.  In this case, we must get at least one new block,
	   * so we set request_status to pk_false.  Otherwise, getting more blocks is optional.
	   */
	  if ((manager->heap.pointer.index >= sizeof_el (manager->heap.pointer.block->subheap))
	      && (manager->heap.half[manager->heap.gc.halfspace].free == (NIL_bdd_nodeBlock)))
	    {
	      request_status = pk_false;
	    }
	  else
	    {
	      request_status = pk_true;
	    }

	  /*
	   * Policy: While the ratio of used to unused nodes is greater than manager->heap.gc.node_ratio, 
	   * keep adding blocks to the free list as long as they are available.  This is to provide room for growth.
	   */
	  while (manager->heap.stats.nodes.used > (manager->heap.stats.nodes.unused * manager->heap.gc.node_ratio))
	    {
	      return_status = add_block_to_free_list (manager, request_status);

	      /*
	       * If no more blocks can be allocated, then stop asking for more.
	       */
	      if (return_status == pk_false)
		{
		  break;
		}
	      request_status = pk_true;	/* getting more becomes optional */
	    }

	  /*
	   * If the inuse list is filled up, then at this point, we are guaranteed to have at least 
	   * one block on the free list.  In this case, take a block off the free list.  If the inuse
	   * list is not filled up, then the next bdd_node will be allocated in the current inuse block.
	   */
	  if (manager->heap.pointer.index >= sizeof_el (manager->heap.pointer.block->subheap))
	    {
	      (void) get_block_from_free_list (manager);
	    }
	}
    }				/* close: if (manager->heap.half[manager->heap.gc.halfspace].free != (NIL_bdd_nodeBlock)) { */
  return;
}

/*
 * Assume there is at least one block on the free list.  Take a block off the free list
 * and put it onto the inuse list.  Set the manager to allocate next bdd_node in the
 * first slot of the block taken from the free list.
 */
static void
get_block_from_free_list (bdd_manager_pointer manager)
{
  bdd_nodeBlock_pointer f;

  /*
   * We assume that there is at least one block on the free list.
   */

  /*
   * Take the next block off of the head of the free list, and update the free list pointer.
   */
  f = manager->heap.half[manager->heap.gc.halfspace].free;
  manager->heap.half[manager->heap.gc.halfspace].free = f->next;

  /*
   * Initialize it.
   */
  f->used = 0;
  f->next = (NIL_bdd_nodeBlock);

  /*
   * Update the inuse pointers.
   */
  *manager->heap.half[manager->heap.gc.halfspace].inuse.tail = f;
  manager->heap.half[manager->heap.gc.halfspace].inuse.tail = &f->next;

  /*
   * Tell manager where to allocate next bdd_node.
   */
  manager->heap.pointer.block = f;
  manager->heap.pointer.index = 0;

  /*
   *    Do not modify any of stats.blocks.total, stats.nodes.unused or stats.nodes.total
   *    because neither changed when a block on the free list is used; these stats
   *    were changed in add_block_to_free_list().
   */
}

/*
 *    add_block_to_free_list
 *
 *    Make two new blocks and stick one in each halfspace's free list
 *
 */
static Integer
add_block_to_free_list (bdd_manager_pointer manager, Integer optionalp)
/* if memory allocation fails, then don't call fail() */
{
  bdd_nodeBlock_pointer b[2];
  Integer h;

  /*
   * Check to see if allocating 2 more blocks will cause the manager memory limit to be exceeded.
   * If so, and if optionalp is set, then just return pk_false, indicating that we were not
   * successful in allocating more blocks. If optionalp is pk_false, then call bdd_will_exceed_mem_limit
   * again, this time indicating that we want to quit this computation.
   */
  if (bdd_will_exceed_mem_limit (manager, (2 * sizeof (bdd_nodeBlock)), pk_false) == pk_true)
    {
      if (optionalp)
	{
	  return pk_false;
	}
      else
	{
	  (void) bdd_will_exceed_mem_limit (manager, (2 * sizeof (bdd_nodeBlock)), pk_true);
	}
    }

  b[0] = (bdd_nodeBlock_pointer) ALLOC (bdd_nodeBlock, 1);
  b[1] = (bdd_nodeBlock_pointer) ALLOC (bdd_nodeBlock, 1);

  if (b[0] == (NIL_bdd_nodeBlock) || b[1] == (NIL_bdd_nodeBlock))
    {
      /*
       *    We've blown the memory limit; back of the greedy approach
       *    Could be that we've got enough to go on already...
       */
      if (optionalp)
	{
	  if (b[0] != (NIL_bdd_nodeBlock))
	    FREE (b[0]);
	  if (b[1] != (NIL_bdd_nodeBlock))
	    FREE (b[1]);
	  /*
	   * We were NOT successful in allocating a new pair of blocks, but the optionalp 
	   * flag is set.
	   */
	  return pk_false;
	}
      else
	{
	  (void) bdd_memfail (manager, "add_block_to_free_list");
	}
    }

  for (h = 0; h < sizeof_el (manager->heap.half); h++)
    {
      b[h]->used = 0;
      b[h]->next = (NIL_bdd_nodeBlock);

      b[h]->next = manager->heap.half[h].free;
      manager->heap.half[h].free = b[h];
    }

  manager->heap.stats.nodes.unused += sizeof_el (manager->heap.pointer.block->subheap);
  manager->heap.stats.nodes.total += sizeof_el (manager->heap.pointer.block->subheap);
  manager->heap.stats.blocks.total++;

  /*
   * We were successful in allocating a new pair of blocks.
   */
  return pk_true;
}
