/*
 * Revision Control Information
 *
 * $Source: /vol/opua/opua2/sis/sis-1.1/common/src/sis/util/RCS/safe_mem.c,v $
 * $Author: sis $
 * $Revision: 1.2 $
 * $Date: 1992/05/06 19:03:25 $
 *
 */
/* LINTLIBRARY */

#include "BDD_StdTypes.h"
#include <stdio.h>
#include <stdlib.h>
#include "ansi.h"
#include "safe_mem_g.h"

/*
 *  These are interface routines to be placed between a program and the
 *  system memory allocator.  
 *
 *  It forces well-defined semantics for several 'borderline' cases:
 *
 *	malloc() of a 0 size object is guaranteed to return something
 *	    which is not 0, and can safely be freed (but not dereferenced)
 *	free() accepts (silently) an 0 pointer
 *	realloc of a 0 pointer is allowed, and is equiv. to malloc()
 *	For the IBM/PC it forces no object > 64K; note that the size argument
 *	    to malloc/realloc is a 'long' to catch this condition
 *
 *  The function pointer MMoutOfMemory() contains a vector to handle a
 *  'out-of-memory' error (which, by default, points at a simple wrap-up 
 *  and exit routine).
 */

/* Modified by Loic Besnard IRISA */


/* MMout_of_memory -- out of memory for lazy people, flush and exit */
static void
MMout_of_memory (Integer size)
{
  (void) fflush (stdout);
  (void) fprintf (stderr, "\nout of memory allocating %ld bytes\n", size);
  exit (1);
}


void (*MMoutOfMemory) () = MMout_of_memory;


char *
MMalloc (Integer size)
{
  char *p;

#ifdef IBMPC
  if (size > 65000L)
    {
      if (MMoutOfMemory != (void (*)()) 0)
	(*MMoutOfMemory) (size);
      return ((char *) NULL);
    }
#endif
  if (size <= 0)
    size = sizeof (long);
  if ((p = (char *) malloc ((unsigned) size)) == ((char *) NULL))
    {
      if (MMoutOfMemory != (void (*)()) 0)
	(*MMoutOfMemory) (size);
      return ((char *) NULL);
    }
  return p;
}


char *
MMrealloc (char *obj, Integer size)
{
  char *p;

#ifdef IBMPC
  if (size > 65000L)
    {
      if (MMoutOfMemory != (void (*)()) 0)
	(*MMoutOfMemory) (size);
      return ((char *) NULL);
    }
#endif
  if (obj == ((char *) NULL))
    return MMalloc (size);
  if (size <= 0)
    size = sizeof (long);
  if ((p = (char *) realloc (obj, (unsigned) size)) == ((char *) NULL))
    {
      if (MMoutOfMemory != (void (*)()) 0)
	(*MMoutOfMemory) (size);
      return ((char *) NULL);
    }
  return p;
}
