/*
 * Revision Control Information
 *
 * $Source: /vol/opua/opua2/sis/sis-1.1/common/src/sis/st/RCS/st.c,v $
 * $Author: sis $
 * $Revision: 1.2 $
 * $Date: 1992/05/06 19:02:54 $
 *
 */
#include <stdlib.h>
#include "BDD_StdTypes.h"
#include "ansi.h"
#include "Alloc.h"

#include <stddef.h>
#include "boolean.h"
#include "bdd_export_type.h"
#include "bdd_export.h"

#include "st.h"
#include "st_g.h"
#include "safe_mem_g.h"

static void rehash (register st_table_pointer table);

#ifndef ABS
#define ABS(a)			((a) < 0 ? -(a) : (a))
#endif

#define ST_NUMCMP(x,y) ((x) != (y))
#define ST_NUMHASH(x,size) (ABS((Integer)x)%(size))
#define ST_PTRHASH(x,size) ((Integer)((UnsignedInteger)(x)>>2)%size)
#define EQUAL(func, x, y) \
    ((((func) == st_numcmp) || ((func) == st_ptrcmp)) ?\
      (ST_NUMCMP((x),(y)) == 0) : ((*func)((x), (y)) == 0))

static Integer
st_numhash (char *x, Integer size)
{
  return ST_NUMHASH (x, size);
}

static Integer
st_numcmp (char *x, char *y)
{
  return ST_NUMCMP (x, y);
}

/*
 *  do_hash((void *)key, (void *)table)	
 */
#define do_hash(key, table)						\
    ((table->hash == st_ptrhash) ? ST_PTRHASH((key), (table)->num_bins) :\
     (table->hash == st_numhash) ? ST_NUMHASH((key), (table)->num_bins) :\
     (*table->hash)((key), (table)->num_bins))

/*Integer st_ptrhash(char *x, Integer size);
Integer st_numcmp(char *x, char *y);
Integer st_ptrcmp(char *x, char *y);
*/
static st_table_pointer
st_init_table_with_params (Integer (*compare) ( /* ??? */ ), Integer (*hash) ( /* ??? */ ), Integer size, Integer density, double grow_factor,
			   Integer reorder_flag)
{
  Integer i;
  st_table_pointer new;

  new = ALLOC_ST_TABLE;
  new->compare = compare;
  new->hash = hash;
  new->num_entries = 0;
  new->max_density = density;
  new->grow_factor = grow_factor;
  new->reorder_flag = reorder_flag;
  if (size <= 0)
    {
      size = 1;
    }
  new->num_bins = size;
  new->bins = ALLOC (st_table_entry *, size);
  for (i = 0; i < size; i++)
    {
      new->bins[i] = 0;
    }
  return new;
}

st_table_pointer
st_init_table (Integer (*compare) ( /* ??? */ ), Integer (*hash) ( /* ??? */ ))
{
  return st_init_table_with_params (compare, hash, ST_DEFAULT_INIT_TABLE_SIZE,
				    ST_DEFAULT_MAX_DENSITY, ST_DEFAULT_GROW_FACTOR, ST_DEFAULT_REORDER_FLAG);
}

void
st_free_table (st_table_pointer table)
{
  register st_table_entry *ptr, *next;
  Integer i;

  for (i = 0; i < table->num_bins; i++)
    {
      ptr = table->bins[i];
      while (ptr != (NIL_st_table_entry))
	{
	  next = ptr->next;
	  FREE (ptr);
	  ptr = next;
	}
    }
  FREE (table->bins);
  FREE (table);
}

#define PTR_NOT_EQUAL(table, ptr, user_key)\
  (ptr != (NIL_st_table_entry) && !EQUAL(table->compare, user_key, (ptr)->key))

#define FIND_ENTRY(table, hash_val, key, ptr, last) \
    (last) = &(table)->bins[hash_val];\
    (ptr) = *(last);\
    while (PTR_NOT_EQUAL((table), (ptr), (key))) {\
	(last) = &(ptr)->next; (ptr) = *(last);\
    }\
  if ((ptr) != (NIL_st_table_entry) && (table)->reorder_flag) {	\
	*(last) = (ptr)->next;\
	(ptr)->next = (table)->bins[hash_val];\
	(table)->bins[hash_val] = (ptr);\
    }

Integer
st_lookup (st_table_pointer table, register char *key, char **value)
{
  Integer hash_val;
  register st_table_entry *ptr, **last;

  hash_val = (Integer) do_hash (key, table);

  FIND_ENTRY (table, hash_val, key, ptr, last);

  if (ptr == (NIL_st_table_entry))
    {
      return 0;
    }
  else
    {
      if (value != ((char **) NULL))
	*value = ptr->record;
      return 1;
    }
}

#define ADD_DIRECT(table, key, value, hash_val, new)\
{\
    if (table->num_entries/table->num_bins >= table->max_density) {\
	rehash(table);\
	hash_val = do_hash(key,table);\
    }\
    \
    new = ALLOC(st_table_entry, 1);\
    \
    new->key = key;\
    new->record = value;\
    new->next = table->bins[hash_val];\
    table->bins[hash_val] = new;\
    table->num_entries++;\
}

Integer
st_insert (register st_table_pointer table, register char *key, char *value)
{
  Integer hash_val;
  st_table_entry *new;
  register st_table_entry *ptr, **last;

  hash_val = (Integer) do_hash (key, table);

  FIND_ENTRY (table, hash_val, key, ptr, last);

  if (ptr == (NIL_st_table_entry))
    {
      ADD_DIRECT (table, key, value, hash_val, new);
      return 0;
    }
  else
    {
      ptr->record = value;
      return 1;
    }
}

static void
rehash (register st_table_pointer table)
{
  register st_table_entry *ptr, *next, **old_bins = table->bins;
  Integer i, old_num_bins = table->num_bins, hash_val;

  table->num_bins = table->grow_factor * old_num_bins;

  if (table->num_bins % 2 == 0)
    {
      table->num_bins += 1;
    }

  table->num_entries = 0;
  table->bins = ALLOC (st_table_entry *, table->num_bins);
  for (i = 0; i < table->num_bins; i++)
    {
      table->bins[i] = 0;
    }

  for (i = 0; i < old_num_bins; i++)
    {
      ptr = old_bins[i];
      while (ptr != (NIL_st_table_entry))
	{
	  next = ptr->next;
	  hash_val = (Integer) do_hash (ptr->key, table);
	  ptr->next = table->bins[hash_val];
	  table->bins[hash_val] = ptr;
	  table->num_entries++;
	  ptr = next;
	}
    }
  FREE (old_bins);
}

Integer
st_ptrhash (char *x, Integer size)
{
  return ST_PTRHASH (x, size);
}

Integer
st_ptrcmp (char *x, char *y)
{
  return ST_NUMCMP (x, y);
}

st_generator *
st_init_gen (st_table_pointer table)
{
  st_generator *gen;

  gen = ALLOC (st_generator, 1);
  gen->table = table;
  gen->entry = (NIL_st_table_entry);
  gen->index = 0;
  return gen;
}


Integer
st_gen (st_generator * gen, char **key_p, char **value_p)
{
  register Integer i;

  if (gen->entry == (NIL_st_table_entry))
    {
      /* try to find next entry */
      for (i = gen->index; i < gen->table->num_bins; i++)
	{
	  if (gen->table->bins[i] != (NIL_st_table_entry))
	    {
	      gen->index = i + 1;
	      gen->entry = gen->table->bins[i];
	      break;
	    }
	}
      if (gen->entry == (NIL_st_table_entry))
	{
	  return 0;		/* that's all folks ! */
	}
    }
  *key_p = gen->entry->key;
  if (value_p != 0)
    *value_p = gen->entry->record;
  gen->entry = gen->entry->next;
  return 1;
}


void
st_free_gen (st_generator * gen)
{
  FREE (gen);
}
