/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** List of the variables of a bdd.

  @param  bdd (in) : a bdd descriptor.
  @return the list of the variables referenced in the bdd operand.
 */
extern supp_type
pK_bdd_support (bdd_type bdd)
{
  /*  --ACCESS to bdd list managing --- */
  bdd_gen_pointer gen;
  bdd_node_pointer node;
  supp_type result;
  boolean stop;
  itemValue_type aux;

  result = (supp_type) list_create ();
  list_insert_last ((list_type) result, (itemValue_type) BDD_ZERO_ID);
  foreach_bdd_node (bdd, gen, node)
  {
    list_set_to_first ((list_type) result);
    stop = pk_false;
    while (!stop)
      {
	if (node->id != BDD_ZERO_ID)
	  {
	    aux = list_get_current ((list_type) result);
	    if ((itemValue_type) (node->id) <= aux)	/* WARNING itemValue_type are node->id values ??? */
	      {
		stop = pk_true;
		if ((itemValue_type) (node->id) != aux)
		  list_insert_before ((list_type) result, (itemValue_type) (node->id));
	      }
	  }
	else
	  stop = pk_true;
	list_forward ((list_type) result);
	/* en effet le dernier elt de la liste est le plus grd */
	/* donc on sortira toujours par le if  <= aux */
      }
  }
  list_set_to_last ((list_type) result);
  list_remove_current ((list_type) result);
  /* on retire l'element inutile qu'est BDDD_ZERO_ID */
  return result;
}

/****************************************************************************/
extern supp_type
supp_create (bdd_type bdd)
{
  return (supp_type) pK_bdd_support (bdd);
}

/****************************************************************************/
extern void
supp_destroy (supp_type supp)
{
  list_destroy (supp);
}

/****************************************************************************/
extern boolean
supp_is_member (supp_type supp, bddSupportElement_type var_id)
{
  bddSupportElement_type current;
  list_set_to_first (supp);
  while (!list_is_at_end (supp))
    {
      current = (bddSupportElement_type) list_get_current (supp);
      if (current > var_id)
	return pk_false;
      if (current == var_id)
	return pk_true;
      list_forward (supp);
    }
  return pk_false;
}

/****************************************************************************/
extern supp_type
supp_clone (supp_type s)
{
  list_type result;
  listItem_type save_position;
  result = list_create ();
  save_position = list_get_position (s);
  list_set_to_first (s);
  while (!list_is_at_end (s))
    {
      list_insert_last (result, list_get_current (s));
      list_forward (s);
    }
  list_set_position (s, save_position, pk_true);
  return (supp_type) result;
}

/****************************************************************************/
extern void
supp_insert (supp_type s, bddSupportElement_type id)
{
  listItem_type save_position;

  save_position = list_get_position (s);
  list_set_to_first (s);
  while (!list_is_at_end (s) && (id > (bddSupportElement_type) list_get_current (s)))
    list_forward (s);
  if (list_is_at_end (s))
    list_insert_last (s, (itemValue_type) id);
  else
    {
      if (id != (bddSupportElement_type) list_get_current (s))
	list_insert_before (s, (itemValue_type) id);
    }
  list_set_position (s, save_position, pk_true);
}

/****************************************************************************/
extern void
supp_union (supp_type accu, supp_type added)
{
  unsigned long int id_add;

  list_set_to_first (accu);
  if (!list_is_at_end (added))
    list_set_to_first (added);

  while ((!list_is_at_end (added)))
    {
      id_add = (unsigned long int) list_get_current (added);
      supp_insert (accu, id_add);
      list_forward (added);
    }
}

/****************************************************************************/
extern void
supp_diff (supp_type accu, supp_type substracted)
{
  unsigned long int id, id_subst;

  list_set_to_first (accu);
  list_set_to_first (substracted);

  while ((!list_is_at_end (accu)) && (!list_is_at_end (substracted)))
    {
      id_subst = (unsigned long int) list_get_current (substracted);
      id = (unsigned long int) list_get_current (accu);
      if (id < id_subst)
	{
	  /* id_subst may be further */
	  list_forward (accu);
	}
      else
	{
	  if (id == id_subst)
	    {
	      list_remove_current (accu);
	      list_forward (substracted);
	    }
	  else
	    {
	      /* id_subst is definitely not in accu */
	      list_forward (substracted);
	    }
	}
    }
}

/****************************************************************************/
extern void
supp_remove_current (supp_type s)
{
  list_remove_current (s);
}

/****************************************************************************/
extern void
supp_set_to_max (supp_type s)
{
  list_set_to_last (s);
}

/****************************************************************************/
extern void
supp_set_to_min (supp_type s)
{
  list_set_to_first (s);
}

/****************************************************************************/
extern bddSupportElement_type
supp_get_current (supp_type s)
{
  return (bddSupportElement_type) list_get_current (s);
}

/****************************************************************************/
extern void
supp_forward (supp_type s)
{
  list_forward (s);
}

/****************************************************************************/
extern void
supp_backward (supp_type s)
{
  list_backward (s);
}

/****************************************************************************/
extern boolean
supp_is_at_end (supp_type s)
{
  return list_is_at_end (s);
}
