_PUBLIC_
/* ========== File attributes ==========
 */
/* PLG 2005-05-04
 * success: this_fileName is not NIL 
 *          returns this file name 
 * failure: this_fileName is NIL 
 *          returns ""
 */
const char *
NBuffer::getName ()
{
  if (IS_NULL_STRING (this_fileName))
    return "";
  return this_fileName;
}

/* PLG 2005-05-04
 * success: this is safe and open
 * failure: this is in error
 *            prints an error message
 *          this is not open
 */
boolean
NBuffer::isOpen ()
{
  if (IS_OPEN ())
    return pk_true;
  if (ISNOT_SAFE ())
    illegalUse ();
  return pk_false;
}

/* ========== this_container available space ==========
 */

/* PLG 2005-05-04
 * success: returns number of available spaces in this actual line
 * failure: this is not open or in error, 
 *          return -1;
 */
int
NBuffer::getAvailableSpace ()
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	return ACTUALMARGIN_AVAILABLE_SPACE ();
      return MARGINFREE_AVAILABLE_SPACE ();
    }
  illegalUse ();
  return -1;
}

/* PLG 2005-05-04
 * success: this is safe, open and this_container is empty
 * failure: this is not open or is in error, or this_container is not empty
 */
boolean
NBuffer::isEmpty ()
{
  if (IS_EMPTY ())
    return pk_true;
  if (ISNOT_OPEN ())
    illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * success: this is safe, open and this actual line has less than <required> available spaces
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          this actual line has at least <required> available spaces
 */
boolean
NBuffer::isOver (int required)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	return (ACTUALMARGIN_OVER (required));
      return (MARGINFREE_OVER (required));
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-04-19
 * success: this is safe, open and this actual line has less than <required> available spaces
 *             where required is <str> size
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          this actual line has at least <required> available spaces
 */
boolean
NBuffer::isOver_ (const char *str)
{
  if (IS_OPEN ())
    {
      if (IS_NULL_STRING (str))	/* strlen(str) > 0 when false */
	return pk_false;
      if (MAY_HAVE_MARGIN ())
	return (ACTUALMARGIN_OVER (strlen (str)));
      return (MARGINFREE_OVER (strlen (str)));
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * success: this is safe, open and this actual line has at least <required> available spaces
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          this actual line has less than <required> available spaces
 */
boolean
NBuffer::sfit (int required)
{
  if (IS_OPEN ())
    {
      if (required <= 0)
	return pk_true;
      return FIT (required);
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * equivalent to sfit(required) where required is <str> length
 */
boolean
NBuffer::sfit_ (const char *str)
{
  if (IS_OPEN ())
    {
      if (IS_NULL_STRING (str))	/* strlen(str) > 0 when false */
	return pk_true;
      return FIT (strlen (str));
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * equivalent to sfit(required) where required is the size of <intValue> decimal representation
 */
boolean
NBuffer::sfit_ (int intValue)
{
  if (IS_OPEN ())
    {
      if (intValue < 0)
	{
	  return FIT (1 + get_strPosIntSize (-intValue));
	}
      return FIT (get_strPosIntSize (intValue));
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * success: this is safe, open and this_container content begins with <str>
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          this_container content does not begin with <str>
 */
boolean
NBuffer::isLinePrefixed (const char *str)
{
  if (IS_OPEN ())
    {
      if (IS_NULL_STRING (str))	/* strlen(str) > 0 when false */
	return pk_true;
      return (strstr (this_container, str) == this_container);
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * success: this is safe, open 
 *          this_container is emptied, its content is lost
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
boolean
NBuffer::reset ()
{
  if (ISNOT_OPEN ())
    {
      illegalUse ();
      return pk_false;
    }
  if (IS_NOTEMPTY ())
    {
      EMPTY_CONTAINER ();
      SET_SPACED ();
    }
  if (MAY_HAVE_MARGIN ())
    {
      SET_ACTUALMARGIN ();
    }
  return pk_true;
}

/* ========== Margins ==========
 */

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *             a mark is set at current position of this virtual line
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
boolean
NBuffer::mark ()
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	pushIndentation (VIRTUAL_INDEX ());
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *             a mark is set <ll> positions after (before if <ll> < 0) current position of this virtual line
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
boolean
NBuffer::mark (int ll)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	pushIndentation (VIRTUAL_INDEX () + ll);
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *            the most recent mark is removed
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          there is not a current mark: no effect
 */
boolean
NBuffer::backMark ()
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	popIndentation ();
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *            <purgeLine> is pk_true, this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *            a new mark is set at first position of this virtual line
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
boolean
NBuffer::freeze_indent (boolean purgeLine)
{
  if (ISNOT_OPEN ())
    {
      illegalUse ();
      return pk_false;
    }
  if (MAY_HAVE_MARGIN ())
    {
      if (purgeLine)
	if (!PURGE_MARGINLINE ())
	  return pk_false;
      pushIndentation (0);
    }
  return pk_true;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *          <purgeLine> is pk_true and this_container is not empty 
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *          the most recent mark is removed
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
boolean
NBuffer::release_indent (boolean purgeLine)
{
  if (ISNOT_OPEN ())
    {
      illegalUse ();
      return pk_false;
    }
  if (MAY_HAVE_MARGIN ())
    {
      if (purgeLine)
	if (!PURGE_MARGINLINE ())
	  return pk_false;
      popIndentation ();
    }
  return pk_true;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *             a tabulation is set <ll> positions after (before if <ll> < 0) the most recent tabulation
 *             this tabulation value is set as a mark
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
boolean
NBuffer::tabulation (int ll)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	pushLeftIndentation (top (leftIndentationMarks) + ll);
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *            the most recent mark and tabulation are removed
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          there is not a current mark: no effect
 */
boolean
NBuffer::backTabulation ()
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	popLeftIndentation ();
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * assumes correct balancing of mark-backMark, tabulation-backTabulation and freeze_indent-release_indent
 * success: this is safe, open 
 *            <factor> <= 0 this has no effect
 *            <factor> > 0, the <factor> most recent marks and tabulations are removed
 * failure: this_file is in error or is not open 
 *               prints an error message
 *          there is not a current mark: no effect
 */
boolean
NBuffer::backTabulation (int factor)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	popLeftIndentation (factor);
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* ========== current prefix ==========
 */

/* PLG 2005-05-04
 * current prefix is a string printed after indentation spaces and before actual characters
 * success: this is safe, open 
*          prefix (NIL_String by default) is removed 
 *          returns this removed prefix value 
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
const char *
NBuffer::removeLeftMark ()
{
  if (IS_OPEN ())
    {
      if (HAS_PREFIX ())
	{
	  const char *prevString = this_prefix;
	  RESET_PREFIXFREE ();
	  return prevString;
	}
      return NIL_String;
    }
  illegalUse ();
  return NIL_String;
}

/* PLG 2005-05-04
 * current prefix is a string printed after indentation spaces and before actual characters
 * success: this is safe, open 
 *            this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
*          prefix (NIL_String by default) is removed 
 *          returns this removed prefix value 
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
const char *
NBuffer::line_removeLeftMark ()
{
  if (IS_OPEN ())
    {
      if (HAS_PREFIX ())
	{
	  const char *prevString = this_prefix;
	  if (!(PURGE_LINE ()))
	    return NIL_String;
	  RESET_PREFIXFREE ();
	  return prevString;
	}
      if (PURGE_PREFIXFREELINE () /*SET_SPACED() */ );
      return NIL_String;
    }
  illegalUse ();
  return NIL_String;
}

/* PLG 2005-05-04
 * current prefix is a string printed after indentation spaces and before actual characters
 * success: this is safe, open 
 *            this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *          current prefix (NIL_String by default) is removed 
 *          returns this removed prefix value 
 *          prefix becomes <lMargin>
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
const char *
NBuffer::line_setLeftMark (const char *lMargin)
{
  if (IS_NULL_STRING (lMargin))	/* strlen(str) > 0 when false */
    return line_removeLeftMark ();
  if (IS_OPEN ())
    {
      const char *prevString;
      if (HAS_PREFIX ())
	{
	  prevString = this_prefix;
	  if (!(PURGE_LINE ()))	/* SET_SPACED() */
	    return NIL_String;
	  RESET_PREFIXFREE ();
	}
      else
	{
	  prevString = NIL_String;
	  if (!PURGE_PREFIXFREELINE ())	/* SET_SPACED() */
	    return NIL_String;
	}
      resetLeftMark (lMargin, strlen (lMargin));
      return prevString;
    }
  illegalUse ();
  return NIL_String;
}

/* PLG 2005-05-04
 * current prefix is a string printed after indentation spaces and before actual characters
 * success: this is safe, open 
 *            <purgeLine> is pk_true, this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *          current prefix (NIL_String by default) is removed 
 *          returns this removed prefix value 
 *          prefix becomes <lMargin>
 * failure: this_file is in error or is not open 
 *               prints an error message
 */
const char *
NBuffer::setLeftMark (const char *lMargin)
{
  if (IS_NULL_STRING (lMargin))
    {				/* strlen(str) > 0 when false */
      return removeLeftMark ();
    }
  if (IS_OPEN ())
    {
      const char *prevString;
      if (HAS_PREFIX ())
	{
	  prevString = this_prefix;
	  RESET_PREFIXFREE ();
	}
      else
	prevString = NIL_String;
      resetLeftMark (lMargin, strlen (lMargin));
      return prevString;
    }
  illegalUse ();
  return NIL_String;
}

/* ========== separator printing ==========
 */

/* PLG 2005-05-04, PLG 2007-03-09
 * success: this is safe, open
 *              this_container is full wrt current margin rule and not space terminated
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *              this_container is not empty, not full and not space terminated
 *                 a space is appent
 *              this_container is space terminated
 * failure: this_file is in error or is not open, or put failed
 *               prints an error message
 */
boolean
NBuffer::space ()
{
  if (IS_EMPTY ())
    return pk_true;
  if (IS_OPEN ())
    {
      if (ISNE_NOTSPACE_TERMINATED ())
	return Uprint_sp ();
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * success: this is safe, open
 *              this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 * failure: this_file is in error or is not open, or put failed
 *               prints an error message
 */
boolean
NBuffer::line ()
{
  if (IS_EMPTY ())
    return pk_true;
  if (IS_OPEN ())
    {
      return PUT_LINE ();
    }				/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * success: this is safe, open
 *              this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 * failure: this_file is in error or is not open, or put failed
 *               prints an error message
 */
boolean
NBuffer::lineIfOver (int required)
{
  if (IS_OPEN ())
    return putLineIfNotFit_fit (required);
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * success: this is safe, open
 *              this_container is not empty
 *                 this_container content is previously put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *              <factor> empty strings are put and eol terminated with current margin configuration 
 * failure: this_file is in error or is not open, or put failed
 *               prints an error message
 */
boolean
NBuffer::line (int factor)
{
  if (IS_OPEN ())
    {
      if (!(PURGE_LINE ()))
	return pk_false;		/* SET_SPACED() */
      return put_emptyLines (factor);
    }				/* failure: error is set */
  illegalUse ();
  return pk_false;
}

/* ========== Gluing ==========
 */

/* ========== Char printing ==========
 */

/* PLG 2005-05-04, PLG 2007-03-16
 * success: this is safe, open
 *              this_container is full
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied
 *               <cc> is appent to this_container content
 * failure: this_file is in error or is not open
 *               prints an error message
 */
boolean
NBuffer::print (char cc)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	{
	  if (IS_ACTUALMARGIN_FULL ())	/* SET_SPACED() */
	    if (!(put_marginLine ()))
	      return pk_false;
	}
      else if (IS_MARGINFREE_FULL ())	/* SET_SPACED() */
	if (!(put_marginFreeLine ()))
	  return pk_false;
      ADD_CHAR_NS (cc);
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-11, PLG 2007-03-16
 * equivalent to 
 *       space(); print(cc);
 */
boolean
NBuffer::print_sp__ (char cc)
{
  if (IS_EMPTY ())
    {
      ADD_CHAR_NS (cc);
      return pk_true;
    }
  if (IS_OPEN ())
    {
      if (ISNE_SPACE_TERMINATED ())
	return Uprint (cc);	/* SET_NOTSPACED() */
      return Uprint_sp_ (cc);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       print(cc); space(); 
 */
boolean
NBuffer::space (char cc)
{
  if (IS_OPEN ())
    return Uprint__sp (cc);	/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-12, PLG 2007-03-09
 * equivalent to 
 *       space(); print(cc); space(); 
 */
boolean
NBuffer::print_sp__sp (char cc)
{
  if (IS_EMPTY ())
    return Uprint__sp (cc);	/* SET_SPACED() */
  if (IS_OPEN ())
    {
      if (ISNE_SPACE_TERMINATED ())
	return Uprint__sp (cc);	/* SET_SPACED() */
      return Uprint_sp__sp (cc);
    }				/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-12, PLG 2007-03-09
 * equivalent to 
 *       print(cc); mark(); 
 */
boolean
NBuffer::print_mark (char cc)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	{
	  if (appendMargin_ (cc))
	    {			/* SET_NOTSPACED() */
	      pushIndentation (VIRTUAL_INDEX ());
	      return pk_true;
	    }
	  return pk_false;
	}
      return appendMarginFree_ (cc);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-12, PLG 2007-03-09
 * equivalent to 
 *       print(cc); mark(incr); 
 */
boolean
NBuffer::print_mark (char cc, int incr)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	{
	  if (appendMargin_ (cc))
	    {			/* SET_NOTSPACED() */
	      pushIndentation (VIRTUAL_INDEX () + incr);
	      return pk_true;
	    }
	  return pk_false;
	}
      return appendMarginFree_ (cc);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       backMark(); print(cc); 
 */
boolean
NBuffer::backMark_print (char cc)
{
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	{
	  popIndentation ();
	  return appendMargin_ (cc);
	}			/* SET_NOTSPACED() */
      return appendMarginFree_ (cc);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       print(cc); line();
 */
boolean
NBuffer::line (char cc)
{
  if (IS_OPEN ())
    {
      if (Uprint (cc))
	{			/* SET_NOTSPACED() */
	  return (PUT_LINE ());
	}
      return pk_false;
    }				/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <cc1> and <cc2>
 */
boolean
NBuffer::print (char cc1, char cc2)
{
  if (IS_OPEN ())
    {
      return_Uprint_CHAR_CHAR_NS (cc1, cc2);
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       space(cc); space(); 
 */
boolean
NBuffer::space (char cc1, char cc2)
{
  if (IS_OPEN ())
    {
      if (Uprint__sp (cc1))	/* SET_SPACED() */
	return Uprint__sp (cc2);	/* SET_SPACED() */
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <cc1> and <cc2>
 */
boolean
NBuffer::print (char cc1, char cc2, char cc3)
{
  if (IS_OPEN ())
    {
      if (putLineIfNotFit_fit (3))
	{			/* SET_SPACED() if container is emptied */
	  ADD_CHAR (cc1);
	  ADD_CHAR_CHAR_NS (cc2, cc3);
	  return pk_true;
	}
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <factor> occurrences of <cc> 
 */
boolean
NBuffer::printNChars (char cc, long int factor)
{
  if (IS_OPEN ())
    {
      if (factor <= 0)
	return pk_true;
      return Uprint_eolIfSplit_nChar (cc, factor);
    }
  illegalUse ();
  return pk_false;
}

/* ========== String printing ==========
 */

/* PLG 2005-05-04, PLG 2007-03-09
 * success: this is safe, open
 *              this_container has less than <strValue> size spaces available
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied and then
 *              this_container has at least <strValue> size spaces available
 *                 <strValue> is appent to this_container content
 *              this_container has less than  <strValue> size spaces available
 *                 <strValue> size is greater than lineSize, the maximal number of characters in a put line
 *                    splitting is allowed
 *                       <strValue> is split into strings LLi of length equal to lineSize
 *                          all LLi are put in this_file and continuation terminated without indentation
 *                          last suffix is put in this_file and eol terminated without indentation
 *                    splitting is not allowed
 *                       <strValue> is put in this_file and eol terminated without indentation
 *                 <strValue> size is atmost lineSize 
 *                    <strValue> is put in this_file and eol terminated without indentation
 * failure: this_file is in error or is not open or put failed
 *               prints an error message
 */
boolean
NBuffer::print (const char *strValue)
{
  if (IS_OPEN ())
    {
      if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
	return pk_true;
      return Uprint_eolIfSplit_str (strValue);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-11, PLG 2007-03-09
 * equivalent to 
 *       space(); print(strValue);
 */
boolean
NBuffer::print_sp__ (const char *strValue)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return space ();
  if (IS_EMPTY ())
    {
      return Uprint_eolIfSplit_str (strValue);
    }				/* SET_NOTSPACED() */
  if (IS_OPEN ())
    {
      if (ISNE_NOTSPACE_TERMINATED () && !Uprint_sp ())	/* SET_SPACED() */
	return pk_false;
      return Uprint_eolIfSplit_str (strValue);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       print(strValue); space(); 
 */
boolean
NBuffer::space (const char *strValue)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return space ();		/* SET_SPACED() */
  if (IS_OPEN ())
    return Uprint_eolIfSplit_str_sp (strValue);	/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       space(); print(strValue); space(); 
 */
boolean
NBuffer::print_sp__sp (const char *strValue)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return space ();		/* SET_SPACED() */
  if (IS_EMPTY ())
    {
      return Uprint_eolIfSplit_str_sp (strValue);
    }				/* SET_SPACED() */
  if (IS_OPEN ())
    {
      if (ISNE_NOTSPACE_TERMINATED () && (!Uprint_sp ()))	/* SET_SPACED() */
	return pk_false;
      return Uprint_eolIfSplit_str_sp (strValue);
    }				/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       print(strValue); mark(); 
 */
boolean
NBuffer::print_mark (const char *strValue)
{
  if (IS_OPEN ())
    {
      if (IS_NULL_STRING (strValue) || Uprint_eolIfSplit_str (strValue))
	{			/* SET_NOTSPACED() */
	  if (MAY_HAVE_MARGIN ())
	    pushIndentation (VIRTUAL_INDEX ());
	  return pk_true;
	}
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       print(strValue); mark(incr); 
 */
boolean
NBuffer::print_mark (const char *strValue, int incr)
{
  if (IS_OPEN ())
    {
      if (IS_NULL_STRING (strValue) || Uprint_eolIfSplit_str (strValue))
	{			/* SET_NOTSPACED() */
	  if (MAY_HAVE_MARGIN ())
	    pushIndentation (VIRTUAL_INDEX () + incr);
	  return pk_true;
	}
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       backMark(); print(strValue); 
 */
boolean
NBuffer::backMark_print (const char *strValue)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return backMark ();
  if (IS_OPEN ())
    {
      if (MAY_HAVE_MARGIN ())
	popIndentation ();
      return Uprint_eolIfSplit_str (strValue);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *       print(strValue); line(); 
 */
boolean
NBuffer::line (const char *strValue)
{
  //  fprintf(stderr, "DEBUG NBuffer::line, strValue is \n++<%s>++\n", strValue);
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return line ();
  if (IS_OPEN ())
    {
      if (Uprint_eolIfSplit_str (strValue))
	return PURGE_LINE ();
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * success: this is safe, open
 *              this_container is not empty
 *                 this_container content is put and eol terminated with current margin configuration 
 *                 this_container is emptied and then
 *              <strValue> size is greater than lineSize, the maximal number of characters in a put line
 *                    splitting is allowed
 *                       <strValue> is split into strings LLi of length equal to lineSize
 *                          all LLi are put in this_file and continuation terminated without indentation
 *                          last suffix is put in this_file and eol terminated without indentation
 *                    splitting is not allowed
 *                       <strValue> is put in this_file and eol terminated without indentation
 *               <strValue> size is atmost lineSize 
 *                    <strValue> is put in this_file and eol terminated without indentation
 * failure: this_file is in error or is not open or put failed
 *               prints an error message
 */
boolean
NBuffer::noIndent_ln (const char *strValue)
{
  // fprintf(stderr,"DEBUG NBuffer::noIndent_ln\n");
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return line ();
  if (IS_OPEN ())
    {
      if (!(PURGE_LINE ()))
	return pk_false;		/* SET_SPACED() */
      if (ISNOT_EMPTYPREFIX ())
	{
	  return_FF_on_putPREFIX_err ();
	}
      return_FF_on_putSTRING_err (strValue);
      PUT_EOL ();
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <strValue1> and <strValue2>
 */
boolean
NBuffer::print (const char *strValue1, const char *strValue2)
{
  if (IS_NULL_STRING (strValue1))	/* strlen(strValue) > 0 when false */
    return print (strValue2);
  if (IS_NULL_STRING (strValue2))	/* strlen(strValue) > 0 when false */
    return print (strValue1);
  if (IS_OPEN ())
    {
      return Uprint_eolIfSplit_str_str (strValue1, strValue2);
    }				/* SET_NOTSPACED() */
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to 
 *      space(strValue1); space(strValue2);
 */
boolean
NBuffer::space (const char *strValue1, const char *strValue2)
{
  if (IS_NULL_STRING (strValue1))	/* strlen(strValue) > 0 when false */
    return space (strValue2);
  if (IS_NULL_STRING (strValue2))	/* strlen(strValue) > 0 when false */
    return space (strValue1);
  if (IS_OPEN ())
    {
      if (Uprint_eolIfSplit_str_sp (strValue1))	/* SET_SPACED() */
	return Uprint_eolIfSplit_str_sp (strValue2);	/* SET_SPACED() */
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-16
 * equivalent to 
 *      line(strValue1); line(strValue2); 
 */
boolean
NBuffer::line (const char *strValue1, const char *strValue2)
{
  if (IS_NULL_STRING (strValue1))	/* strlen(strValue) > 0 when false */
    return line (strValue2);
  if (IS_NULL_STRING (strValue2))	/* strlen(strValue) > 0 when false */
    return line (strValue1);
  if (IS_OPEN ())
    {
      return Uprint_eolIfSplit_str_str (strValue1, strValue2);
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-16
 * equivalent to 
 *       print(strValue1); line(); print(strValue2); line(); 
 */
boolean
NBuffer::print (int nbElem, const char *strArray[])
{
  if (IS_OPEN ())
    {
      int ii = 0;
      while (ii < nbElem)
	{
	  if (ISNOT_NULL_STRING (strArray[ii]))
	    return Uprint_eolIfSplit_strArray (nbElem - ii, (strArray + ii));
	  ii = ii + 1;
	}
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * iteratively spaces <nbElem> first elements of <strArray> 
 * <strArray> is assumed to contain at least <nbElem> elements
 */
boolean
NBuffer::space (int nbElem, const char *strArray[])
{
  if (IS_OPEN ())
    {
      int ii = 0;
      for (ii = 0; ii < nbElem; ii++)
	{
	  if (ISNOT_NULL_STRING (strArray[ii]))
	    return Uprint_eolIfSplit_strArray_sp (nbElem - ii, (strArray + ii));
	}
      if (IS_NOTSP_TERMINATED ())
	return Uprint_sp ();	/* SET_SPACED() */
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <cc> and <strValue>
 */
boolean
NBuffer::print (char cc, const char *strValue)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return print (cc);
  if (IS_OPEN ())
    return Uprint_eolIfSplit_chr_str (cc, strValue);
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <strValue> and <cc>
 */
boolean
NBuffer::print (const char *strValue, char cc)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return print (cc);		/* SET_NOTSPACED() */
  if (IS_OPEN ())
    return Uprint_eolIfSplit_str_chr (strValue, cc);
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-16
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <cc1> and <strValue> and <cc2>
 */
boolean
NBuffer::print (char cc1, const char *strValue, char cc2)
{
  if (IS_NULL_STRING (strValue))
    return print (cc1, cc2);
  /* strlen(strValue) > 0 when false */
  if (IS_OPEN ())
    {
      return Uprint_eolIfSplit_chr_str_chr (cc1, strValue, cc2);
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04, PLG 2007-03-09
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of '"', <strValue> and '"'
 */
boolean
NBuffer::print_CString (const char *strValue)
{
  if (IS_NULL_STRING (strValue))
    return print ('"', '"');
  /* strlen(strValue) > 0 when false */
  if (IS_OPEN ())
    {
      return Uprint_eolIfSplit_chr_str_chr ('"', strValue, '"');
    }
  illegalUse ();
  return pk_false;
}

/* ========== Integer printing ==========
 */

/* PLG 2005-05-04, PLG 2007-03-15
 * equivalent to print(decimalRepresentation) 
 *       where decimalRepresentation is the string decimal representation of argNumber
 */
boolean
NBuffer::print (long int argNumber)
{
  if (IS_OPEN ())
    {
      if (argNumber >= 0)
	{
	  if (argNumber < 10)
	    return Uprint (CHAR_DIGIT (argNumber));	/* SET_NOTSPACED() */
	  return Uprint_eolIfSplit_int (argNumber);
	}
      if (argNumber >= -9)
	{			/* strlen(strValue) > 0 when false */
	  return_Uprint_CHAR_CHAR_NS ('-', CHAR_DIGIT (-argNumber));
	}
      return Uprint_eolIfSplit_chr_int ('-', -argNumber);
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * equivalent to 
 *       print(argNumber); space();
 */
boolean
NBuffer::space (long int argNumber)
{
  if (IS_OPEN ())
    {
      if (argNumber >= 0)
	{
	  if (argNumber < 10)
	    return Uprint__sp (CHAR_DIGIT (argNumber));	/* SET_SPACED() */
	  if (!Uprint_eolIfSplit_int (argNumber))
	    return pk_false;
	}
      else if (argNumber >= -9)
	{			/* strlen(strValue) > 0 when false */
	  return Uprint__sp ('-', CHAR_DIGIT (-argNumber));
	}
      else
	{
	  if (!Uprint_eolIfSplit_chr_int ('-', -argNumber))
	    return pk_false;
	}
      return Uprint_sp ();
    }				/* SET_SPACED() */
  illegalUse ();
  return pk_false;
}

/* ========== Miscellaneous printing ==========
 */

/* PLG 2005-05-04
 * positiveNumber >= 0
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <cc> 
 *            and the string decimal representation of <positiveNumber>
 * failure: positiveNumber < 0
 */
boolean
NBuffer::print (char cc, long int positiveNumber)
{
  if (IS_OPEN ())
    {
      if (positiveNumber >= 0)
	{
	  if (positiveNumber >= 10)
	    return Uprint_eolIfSplit_chr_int (cc, positiveNumber);
	  return_Uprint_CHAR_CHAR_NS (cc, CHAR_DIGIT (positiveNumber));
	}
      stderr_error ("OUPS!? second argument of NBuffer::print(char cc, int positiveNumber) is negative");
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * positiveNumber >= 0
 * equivalent to print(concatenation) 
 *       where concatenation is the concatenation of <strValue> 
 *            and the string decimal representation of <positiveNumber>
 * failure: positiveNumber < 0
 */
boolean
NBuffer::print (const char *strValue, long int positiveNumber)
{
  if (IS_NULL_STRING (strValue))
    return print (positiveNumber);
  if (IS_OPEN ())
    {
      if (positiveNumber >= 0)
	{
	  if (positiveNumber >= 10)
	    return Uprint_eolIfSplit_str_int (strValue, positiveNumber);
	  return Uprint_eolIfSplit_str_chr (strValue, CHAR_DIGIT (positiveNumber));
	}
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* PLG 2005-05-04
 * positiveNumber >= 0
 *          equivalent to print(concatenation) 
 *                where concatenation is <cc> followed by
 *                     string decimal representation of <positiveNumber> followed by
 *                     <strValue>
 * failure: positiveNumber < 0
 */
/* WARNING PRAMETERS <positiveNumber> and <strValue> should be exchanged */
boolean
NBuffer::print (char cc, long int positiveNumber, const char *strValue)
{
  if (IS_NULL_STRING (strValue))	/* strlen(strValue) > 0 when false */
    return print (cc, positiveNumber);
  if (IS_OPEN ())
    {
      if (positiveNumber >= 0)
	{
	  if (positiveNumber >= 10)
	    return Uprint_eolIfSplit_chr_str_int (cc, strValue, positiveNumber);
	  return Uprint_eolIfSplit_chr_str_chr (cc, strValue, CHAR_DIGIT (positiveNumber));
	}
      return pk_false;
    }
  illegalUse ();
  return pk_false;
}

/* ========================================== */

/* PLG 2005-04-19
 * Ferme le Buffer. 
*/
boolean
NBuffer::close ()
{
  if (IS_OPEN ())
    {
      if (!PURGE_LINE ())
	return pk_false;
      if ((this_file != stdout) && (this_file != stderr))
	fclose (this_file);
      if (MAY_HAVE_MARGIN ())
	closeStacks ();
      SET_CLOSED ();
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/*
  Open file named name in writing mode
  Should return a boolean 
  and emit a mail when failure occurs
*/
boolean
NBuffer::open (const char *aFileName)
{
  const char *errMess;
  if (IS_OPEN ())
    {
      setInError ("file is yet open in open(char *)");
      return pk_false;
    }
  if (IS_CLOSED ())
    {
      if (IS_NULL_STRING (aFileName))
	{			/* strlen(strValue) > 0 when false */
	  setInError ("this_fileName is NIL in open(char *)");
	  return pk_false;
	}
      this_file = wFOpen (aFileName, &errMess);
      if (this_file == NIL_File)
	{
	  setInError (errMess);
	  return pk_false;
	}
      this_fileName = aFileName;
      activateContainer ();
      return pk_true;
    }
  illegalUse ();
  return pk_false;
}

/*
  Link an open file to this Buffer
*/
boolean
NBuffer::open (FILE * aFile, const char *aFileName)
{
  if (IS_OPEN ())
    setInError ("file is yet open in open(FILE *, char *)");
  else if (IS_CLOSED ())
    {
      if (aFile == NIL_File)
	setInError ("cannot open NIL file open(FILE * , char * )");
      else
	{
	  if (this_file == NIL_File)
	    {			/* PLG aFile = NIL_File a la creation */
	      this_file = aFile;
	      this_fileName = aFileName;
	      activateContainer ();
	      return pk_true;
	    }
	  if (this_file == aFile)
	    {
	      activateContainer ();
	      return pk_true;
	    }
	  setInError ("an other file holds this buffer in open(FILE * , char * )");
	}
    }
  else
    illegalUse ();
  return pk_false;
}

/* PLG 2007_03-02 correct added DONE
 * this creates a formatted buffer: printed strings can be split, there is no indent
 *     breakCharList is the lsit of char where splitting is preferred
 *     continuationMark is the string that is inserted fater splitting and before line print
 * used in PKMAKE FormattedFileNode
 */
NBuffer::NBuffer (int taille, const char *breakCharList, char continuationMark, boolean * correct)
{
  if (initSplitting (taille, breakCharList, continuationMark))
    {
      SET_MARGINFREE ();
      if (initSContainer ())
	return;
    }
  (*correct) = pk_false;
  setInError ("memory allocation");
}

/* PLG 2007_03-02 correct added DONE
 * this creates a formatted buffer: printed strings can be split, there is no indent
 *     breakCharList is the lsit of char where splitting is preferred
 *     continuationMark is the string that is inserted fater splitting and before line print
 * used in LogPrinter
 */
NBuffer::NBuffer (int maxindent, int taille, const char *breakCharList, char continuationMark, boolean * correct)
{
  if (initSplitting (taille, breakCharList, continuationMark))
    {
      initIndent (maxindent);
      if (initSContainer ())
	return;
    }
  (*correct) = pk_false;
  setInError ("memory allocation");
}

/* PLG 2007_03-02 correct added DONE
  CONSTRUCTEUR.
  Cree et retourne un Buffer avec ?MAXINDENT comme marge gauche maximale 
  et ?TAILLE comme longueur de ligne. 
  Pour etre utilise, il devra etre ouvert par OPEN.
  It will always be possible to write 2 characters in a line
 * used in LogPrinter
*/
NBuffer::NBuffer (int maxindent, int taille, boolean * correct)
{
  if (!init (maxindent, taille))
    {
      (*correct) = pk_false;
      setInError ("memory allocation");
    }
}

/* PLG 2007_03-02 correct added DONE
  CONSTRUCTEUR.
  Cree et retourne un Buffer sans marge gauche
  de taille ?TAILLE comme longueur de ligne. 
  Pour etre utilise, il devra etre ouvert par OPEN.
  It will always be possible to write 2 characters in a line
 * used in PKMAKE DirectFileNode
*/
NBuffer::NBuffer (int taille, boolean * correct)
{
  if (!init (stdLineIndent, taille))
    {
      (*correct) = pk_false;
      setInError ("memory allocation");
    }
}

/*
  CONSTRUCTEUR.
  Cree et retourne un Buffer avec ?MAXINDENT comme marge gauche maximale 
  et ?TAILLE comme longueur de ligne. 
  Pour etre utilise, il devra etre ouvert par OPEN.
  It will always be possible to write 2 characters in a line
*/
NBuffer::NBuffer (boolean * correct)
{
  if (!init (stdLineIndent, stdLineSize))
    {
      (*correct) = pk_false;
      setInError ("memory allocation");
    }
}

/* destructor */
NBuffer::~NBuffer ()
{
  if (IS_OPEN ())
    close ();
  this_stringRegister = deletePKString (this_stringRegister);
}
