/**
 * Cstring list manipulations
 */
_PUBLIC_
/* ========== Attributes
 */
/** sets aCStringList as successor of this
 */
  void
CStringList::setNext (ref2CStringList aCStringList)
{
  // fprintf(stderr,"DEBUG CStringList::setNext, this adress is %i, this cdr adress was %i\n", (int)this, (int)cdr);
  cdr = aCStringList;
}

/** returns next cell
 */
ref2CStringList
CStringList::next ()
{
  return cdr;
}

/** returns the cell reached after applying next index times
 * if index is negative sets found to false and returns this 
 * if index is too large sets found to false and returns last cell value 
 * if index is 0 sets found to pk_true returns this 
 * returns next(index-1, found) 
 */
ref2CStringList
CStringList::next (int index, boolean * found)
{
  ref2CStringList prevCStringList = this;
  if (index <= 0)
    {
      *found = (index == 0);
      return this;
    }
  while (index > 0)
    {
      if (prevCStringList->cdr == NIL_CStringList)
	{
	  *found = pk_false;
	  return prevCStringList;
	}
      prevCStringList = prevCStringList->cdr;
      index = index - 1;
    }
  *found = pk_true;
  return prevCStringList;
}

/** set this cell value
 */
void
CStringList::setContent (char *aString)
{
  car = aString;
}

/** returns value of this cell
 */
char *
CStringList::content ()
{
  return car;
}

/** set index-th cell to aString
 * failure index-th cell does not exist: list remain unchanged
 */
void
CStringList::setContent (int index, char *aString)
{
  boolean found;
  ref2CStringList elem = next (index, &found);
  if (found)
    elem->car = aString;
}

/** returns value of next(index)
 * failure index-th cell does not exist: returns NIL_String
 */
char *
CStringList::content (int index, boolean * found)
{
  ref2CStringList elem = next (index, found);
  if (*found)
    return elem->car;
  return NIL_String;
}

/** returns this size
 */
int
CStringList::getThisLength ()
{
  if (car == NIL_String)
    return 0;
  return strlen (car);
}

/** returns the number of elements in this
 */
int
CStringList::length ()
{
  int length = 1;
  ref2CStringList prevCStringList = this;
  while (prevCStringList->cdr != NIL_CStringList)
    {
      length = length + 1;
      prevCStringList = prevCStringList->cdr;
    }
  return length;
}

/** returns total addition of these element sizes
 */
int
CStringList::cumulatedSize ()
{
  ref2CStringList prevCStringList = this;
  int size = 0;

  while (prevCStringList != NIL_CStringList)
    {
      size = size + strlen (prevCStringList->car);
      prevCStringList = prevCStringList->cdr;
    }
  return size;
}

/** appends newLast at this end
 * has side effect on this list
 */
void
CStringList::append (ref2CStringList newLast)
{
  // fprintf(stderr,"DEBUG CStringList::append(ref2CStringList newLast), this adress is %i, this cdr adress was %i\n", (int)this, (int)cdr);
  ref2CStringList prevCStringList = this;
  while (prevCStringList->cdr != NIL_CStringList)
    {
      prevCStringList = prevCStringList->cdr;
    }
  prevCStringList->cdr = newLast;
}

/** appends a new cell containing str to the end of this list
 */
void
CStringList::append (char *str)
{
  // fprintf(stderr,"DEBUG CStringList::append(char *str), this adress is %i, this cdr adress was %i\n", (int)this, (int)cdr);
  append (new CStringList (str));
}

/** returns index of str in this list
 * returns -1 if str does not belong to this liste
 */
int
CStringList::find (char *str)
{
  ref2CStringList curCStringList = this;
  int i = 0;
  while (strcmp (curCStringList->car, str) != 0)
    {
      if (curCStringList->cdr == NIL_CStringList)
	return -1;
      i = i + 1;
      curCStringList = curCStringList->cdr;
    }
  return i;
}

/* ========== Builder
 */

/** creates and returns a new string list cell valued by <aString> followed by succ
 */
CStringList::CStringList (char *aString, ref2CStringList succ)
{
  cdr = succ;
  car = aString;
  // fprintf(stderr, "DEBUG CStringList::CStringList(char * aString, ref2CStringList succ) this adress is %i, this cdr adress is %i\n", (int)this, (int)cdr);
  // fprintf(stderr,"DEBUG car is <\n%s\n>\n", car);
}

/** creates and returns a new string list cell valued by <aString>
 */
CStringList::CStringList (char *aString)
{
  cdr = NIL_CStringList;
  car = aString;
  // fprintf(stderr,"DEBUG CStringList::CStringList(char * aString) this adress is %i\n", (int)this);
  // fprintf(stderr, "DEBUG car is <\n%s\n>\n", car);
}

/** creates an empty string list cell
 */
CStringList::CStringList ()
{
  cdr = NIL_CStringList;
  car = NIL_String;
  // fprintf(stderr, "DEBUG CStringList::CStringList(), this adress is %i\n", (int)this);
}

/** destroy this and its successors
   does not destroy strings
 */
_VIRTUAL_
CStringList::~
CStringList ()
{
  // fprintf(stderr,"DEBUG CStringList::~CStringList, this adress is %i, this cdr adress is %i\n", (int)this, (int)cdr);
  if (cdr != NIL_CStringList)
    delete cdr;
  // fprintf(stderr,"DEBUG CStringList::~CStringList END, this adress is %i\n", (int)this);
}
