/*  =============================================================================
 * standard basic functions
 */

#define IS_NIL_STRING(STR) ( STR == NIL_String )

/*  2004-12-27
 * success: characters in <str> belong to <atomAllowedChars>
 * 2004-12-27 */
extern boolean
are_leftChar_in_rightString (const char *str, const char *atomAllowedChars)
{
  return (strspn (str, atomAllowedChars) == strlen (str));
}

/**
  success: returns a unique empty string _emptyCString
           sets _emptyCString if required
  failure: memory allocation failed
           stderr print an error message
           returns NIL_String
*/

extern char *
emptyCString ()
{
  if (_emptyCString == NIL_String)
    {
      _emptyCString = (char *) calloc (1, sizeof (char));
      if (_emptyCString == NULL)
	{
	  stderr_alloc ("emptyCString");
	  return NIL_String;
	}
    }
  return _emptyCString;
}

/** COMMENT BIDON 1
 */
#define toto
/* 2005-01-04
 * success: returns a new malloc'd container large enough to hold max (<nb>, 0) not NULL characters 
 *             followed by a NULL_char string termination 
 *             this container is initiated with empty string
 * failure: memory allocation failed
 *          an error message is stderr printed
 *          returns NIL_String 
*/
extern char *
Cstrmalloc (int nb)
{
  char *container;
  if (nb < 0)
    nb = 0;
  container = (char *) malloc ((nb + 1) * sizeof (char));
  if (container == NULL)
    {
      stderr_alloc ("Cstrmalloc");
      return NIL_String;
    }
  container[0] = NULL_char;
  return container;
}

/* 2004-10-14
 * this remains unchanged
 * returns pk_true if atom contents is equal to aString
 */
extern boolean
areStringEq (const char *string1, const char *string2)
{
  return (strcmp (string1, string2) == 0);
}

/* =============================================================================
 * string representated integer functions
 */

/** compare non negative integers integer1 and integer2 represented by C strings
 *  @param(integer1, integer2) are assumed to be string representations of natural numbers
 *          each integer representation can be nil, "0" or a sequence of numerical characters that is not headed by '0'
 *          assumptions are not checked
 *  @return true iff integer1 <= integer2 as integer values
 *  @param(strict) is 
 *             is assigned pk_true when integer1 < integer2, 
 *             is assigned false when integer1 == integer2
 *             remains unchanged when  integer1 > integer2, 
 *  WARNING null string is arbitrarily taken for -1
 */
extern boolean
CstringInteger_leq (const char *integer1, const char *integer2, boolean * strict)
{
  int strlength1, strlength2;
  if (integer1 == integer2)
    {
      (*strict) = pk_false;
      return pk_true;
    }
  /* integer1 and integer2 are not both null */
  if (IS_NIL_STRING (integer1))
    {
      (*strict) = pk_true;
      return pk_true;
    }
  if (IS_NIL_STRING (integer2))
    return pk_false;
  /* none of integer1 and integer2 is null */
  strlength1 = strlen (integer1);
  strlength2 = strlen (integer2);
  if (strlength1 < strlength2)
    {
      (*strict) = pk_true;
      return pk_true;
    }
  if (strlength1 > strlength2)
    return pk_false;
  /* strlength1 == strlength2 */
  strlength1 = strcmp (integer1, integer2);
  if (strlength1 > 0)
    return pk_false;
  (*strict) = strlength1 < 0;
  return pk_true;
}

/* =============================================================================
 * length functions
 */
/** COMMENT BIDON 2
 */
/** <stringNumber> <= 0, returns 0
 * <stringNumber> > 0, returns cumulated size of <stringNumber> first elements of <stringArray[]>
 */
extern int
get_strArraySize (const char *stringArray[], int stringNumber)
{
  if (stringNumber > 0)
    {
      int arraySize = 0, ii;
      const char *curString;
      for (ii = 0; ii < stringNumber; ii++)
	{
	  curString = stringArray[ii];
	  if (curString != NIL_String)
	    arraySize = arraySize + strlen (curString);
	}
      return arraySize;
    }
  return 0;
}

/* <stringNumber> <= 0, returns 0
 * <stringNumber> > 0, <sepSize> == 0, returns cumulated size of <stringNumber> first elements of <stringArray[]>
 * <stringNumber> > 0, <sepSize> > 0,  returns (sepLgth + strLgth) where 
 *         sepLgth is the cumulated size of separators between not empty strings
 *         strLgth the cumulated size of the <stringNumber> first elements in stringArray
 */
extern int
get_strCondSepArraySize (const char *stringArray[], int stringNumber, int sepSize)
{
  if (stringNumber > 0)
    {
      if (sepSize > 0)
	{
	  int totalLength = 0, ii, curLg;
	  const char *curString;
	  for (ii = 0; ii < stringNumber; ii++)
	    {
	      curString = stringArray[ii];
	      if (curString != NIL_String)
		{
		  curLg = strlen (curString);
		  if (curLg > 0)
		    {
		      if (totalLength > 0)
			totalLength = totalLength + curLg + sepSize;
		      else
			totalLength = curLg;
		    }
		}
	    }
	  return totalLength;
	}
      return get_strArraySize (stringArray, stringNumber);
    }
  return 0;
}

/* val < 0, returns 0
 * val >= 0, returns the size of the string decimal representation of val
 */
extern short
get_strPosIntSize (long int val)
{
  short lgth = 1;
  if (val < 0)
    return 0;
  while (val >= 10)
    {
      val = val / 10;
      lgth = lgth + 1;
    }
  return lgth;
}

/* =============================================================================
 * standard copy functions
 */

/* copies in container: cc
 * followed by NULL_char terminator
 * assumes enough space available in container
 */
extern void
unsafe_cpyChar (char *container, char cc)
{
  container[0] = cc;
  container[1] = NULL_char;
}

/** copies in container: cc1, cc2 
 * followed by NULL_char terminator
 * assumes enough space available in container
 */
extern void
unsafe_cpyCharChar (char *container, char cc1, char cc2)
{
  container[0] = cc1;
  container[1] = cc2;
  container[2] = NULL_char;
}

/* copies in container: nn times cc 
 * followed by NULL_char terminator
 * if nn = 0 container becomes empty
 * unsafe: enough space is available in container
 */
extern void
unsafe_cpyNChar (char *container, char cc, int nn)
{
  if (nn > 0)
    {
      int ii;
      for (ii = 0; ii < nn; ii++)
	container[ii] = cc;
      container[nn] = NULL_char;
    }
  else
    container[0] = NULL_char;
}

/* copies <vstring> in <container>
 * terminated by NULL_char terminator
 * assumes enough space available in <container>
 * returns <vstring> length
 */
extern int
unsafe_cpyString (char *container, const char *vstring)
{
  if (vstring == NIL_String)
    {
      container[0] = NULL_char;
      return 0;
    }
  container = strcpy (container, vstring);
  return strlen (vstring);
}

/* copies <stringNumber> first elements of <stringArray[]> in <container>
 * terminated by NULL_char terminator
 * assumes enough space available in container
 * assumes at least <stringNumber> elements in <stringArray[]>
 */
extern void
unsafe_cpyStrArray (char *container, const char *stringArray[], int stringNumber)
{
  int ii;
  for (ii = 0; ii < stringNumber; ii++)
    if (stringArray[ii] != NIL_String)
      container = strcat (container, stringArray[ii]);
}


/* copies <stringNumber> first elements of <stringArray[]> in <container>
 * separated by sep and terminated by NULL_char terminator
 * NIL_String is separated by sep
 * assumes enough space available in container
 * assumes at least <stringNumber> elements in <stringArray[]>
 */
extern void
unsafe_cpySepStrArray (char *container, char sep, const char *stringArray[], int stringNumber)
{
  if (stringNumber > 0)
    {
      int ii;
      for (ii = 0; ii < stringNumber - 1; ii++)
	{
	  if (stringArray[ii] != NIL_String)
	    container = strcat (container, stringArray[ii]);
	  unsafe_catChar (container, sep);
	}
      container = strcat (container, stringArray[stringNumber]);
    }
}

/* copies in container concatenation of cc1, str, cc2
 * followed by NULL_char terminator
 * assumes enough space available in container
 * returns <vString> length +2
 */
extern int
unsafe_cpyCharStringChar (char *container, char cc1, const char *vString, char cc2)
{
  int nn;
  container[0] = cc1;
  if (vString == NIL_String)
    nn = 2;
  else
    {
      nn = strlen (vString) + 2;
      container[1] = NULL_char;
      container = strcat (container, vString);
    }
  container[nn - 1] = cc2;
  container[nn] = NULL_char;
  return nn;
}

/* copies in container: C string mark '"' followed by <vString> followed by  '"'
 * followed by NULL_char terminator
 * assumes enough space available in container
 * returns <vString> length +2
 */
extern int
unsafe_cpyCString (char *container, const char *vString)
{
  return unsafe_cpyCharStringChar (container, '"', vString, '"');
}

/* copies at most <strLg> first chars of <vstring> in <container>
 *          followed by NULL_char terminator
 * assumes not NIL <vstring> 
 * assumes <vstring> length  not smaller than <strLg>
 * assumes enough space available in container 
*/
extern void
unsafe_NcpyPrefix (char *container, const char *vstring, int strLg)
{
  if (strLg > 0)
    {
      container = strncpy (container, vstring, strLg);
      container[strLg] = NULL_char;
    }
}

/* copies at most <strLg> first chars of <vstring> in <container>
 *          followed by NULL_char terminator
 * if <vstring> length  is greater than <strLg>, vstring truncated prefix is copied in container 
 * if <vstring> length  is smaller than <strLg>, vstring is copied in container and completed by NUL_char
 * assumes enough space available in container 
*/
extern void
unsafe_NcpyString (char *container, const char *vstring, int strLg)
{
  if (strLg > 0)
    {
      if (vstring == NIL_String)
	container[0] = NULL_char;
      else
	{
	  container = strncpy (container, vstring, strLg);
	  container[strLg] = NULL_char;
	}
    }
}

/* copies at most <strLg> first chars of <vstring> in <container>
 *          followed by NULL_char terminator
 * if <vstring> length  is greater than <strLg>, vstring truncated prefix is copied in container 
 * if <vstring> length  is smaller than <strLg>, vstring is copied in container 
 * assumes enough space available in container 
*/
extern int
unsafe_atMostNcpyString (char *container, const char *vstring, int strLg)
{
  if (strLg > 0)
    {
      if (vstring == NIL_String)
	container[0] = NULL_char;
      else
	{
	  int strLgth = strlen (vstring);
	  if (strLgth < strLg)
	    strLg = strLgth;
	  container = strncpy (container, vstring, strLg);
	  container[strLg] = NULL_char;
	  return strLgth;
	}
    }
  return 0;
}

/* assumes enough space available in <container>
 * assumes <posInt> >= 0
 * copies the XX char * decimal representation of <posInt> in <container>
 * XX IS NOT FOLLOWED by NULL_char terminator
 * returns number of copied characters
 */
extern int
unsafe_recCastIntPosVal (char *container, int posInt)
{
  if (posInt >= 10)
    {
      int n = posInt / 10;
      int pos = unsafe_recCastIntPosVal (container, n);
      posInt = posInt - 10 * n;
      container[pos] = CHAR_DIGIT (posInt);
      return pos + 1;
    }
  container[0] = CHAR_DIGIT (posInt);
  return 1;
}

/* copies the XX char * decimal representation of <val> absolute value in <container>
 * followed by NULL_char terminator
 * assumes enough space available in container
 * returns number of copied decimal characters
 */
extern int
unsafe_castIntAbsVal (char *container, int posInt)
{
  int lastIndex;
  if (posInt < 0)
    posInt = -posInt;
  lastIndex = unsafe_recCastIntPosVal (container, posInt);
  container[lastIndex] = NULL_char;
  return lastIndex;
}

/* if lastIndex <= 0 <container> remains unchanged
 * copies the XX char * decimal representation of <val> absolute value in <container>
 * followed by NULL_char terminator
 *
 * begins with the low wheight digit at index <lastIndex>-1 in <container>
 * left fills XX with '0' if length of XX is smaller than available
 * left truncates XX if length of XX is greater than required
 * assumes enough space available in container
 */
extern void
unsafe_cpyIntAbsVal (char *container, int val, int lastIndex)
{
  if (lastIndex > 0)
    {
      div_t quot_rem;
      int j;
      if (val < 0)
	val = -val;
      container[lastIndex] = NULL_char;
      for (j = lastIndex - 1; j >= 0; j--)
	{
	  quot_rem = div (val, 10);
	  container[j] = CHAR_DIGIT ((int) (quot_rem.rem));
	  val = quot_rem.quot;
    }}
}

/* appends to the end of container cc
   followed by NULL_char terminator
 * assumes enough space available in container
 */
extern void
unsafe_catChar (char *container, char c)
{
  int n = strlen (container);
  container[n] = c;
  container[n + 1] = NULL_char;
}

/* appends to the end of container cc1, cc2
   followed by NULL_char terminator
 * assumes enough space available in container
 */
extern void
unsafe_catCharChar (char *container, char cc1, char cc2)
{
  int n = strlen (container);
  container[n] = cc1;
  container[n + 1] = cc2;
  container[n + 2] = NULL_char;
}

/* appends to the end of container nn times cc
   followed by NULL_char terminator
 * assumes enough space available in container
 */
extern void
unsafe_catNChar (char *container, char cc, int nn)
{
  unsafe_cpyNChar ((char *) (container + strlen (container)), cc, nn);
}

/* if lastIndex <= 0 does nothing
 * copies the XX char * decimal representation of <val> absolute value in <container>
 * begins with the low wheight digit at index lastIndex-1 after the end of current <container> content
 * left fills XX with '0' if length of XX is smaller than available
 * left truncates XX if length of XX is greater than required
 * NULL char is written after the low weight digit
 * assumes enough space available in container
 */
extern void
unsafe_catIntAbsVal (char *container, int val, int lastIndex)
{
  unsafe_cpyIntAbsVal ((char *) (container + strlen (container)), val, lastIndex);
}

/* appends to the end of container at most charNb first characters of <vString> 
 * followed by NULL_char terminator
 * success: cumulated size of <container> actual content and <vstring> is at most <containerSize>
 * failure: container remains unchanged
 * assumes enough space available in container
 */
extern void
unsafe_NcatString (char *container, const char *vstring, int charNb)
{
  unsafe_NcpyString ((char *) (container + strlen (container)), vstring, charNb);
}

/* concatenate at most charNb first chars of vstring in container
 * terminates container with NULL_char
   container size is assumed to be large enough (not checked) to contain conainerSize supplementary characters
   failure: strlen(container)+min(charNb,strlen(vstring)) is greater than containerSize
            container remains unchanged
*/
extern boolean
unsafe_NcatStringAtMost (char *container, const char *vstring, int charNb, int containerSize)
{
  int strLg;
  if ((vstring == NIL_String) || (charNb <= 0))
    return pk_true;
  strLg = strlen (vstring);
  if (charNb > strLg)
    charNb = strLg;
  if (charNb <= (containerSize - strlen (container)))
    {
      container = strncat (container, vstring, charNb);
      return pk_true;
    }
  return pk_false;
}

/* appends "<vString>" to container 
 * assumes enough space available in container
 * returns <vString> length + 2
 */
extern int
unsafe_catCString (char *container, const char *vString)
{
  return unsafe_cpyCString ((char *) (container + strlen (container)), vString);
}

/* 2004-08-12
 * returns -1 if cc does not belong to stringValue
 * returns length of the longest prefix before first cc occurence if cc belongs to stringValue
 */
extern int
prefixLengthUntilChar (const char *stringValue, char cc)
{
  int length;
  char *suffixStr;

  if (stringValue == NIL_String)
    return -1;
  suffixStr = strchr (stringValue, cc);
  if (suffixStr == NIL_String)
    return -1;
  length = suffixStr - stringValue;
  return length;
}

/* 2004-08-12
 * returns length of the longest prefix of stringValue that does not contain character occurring in refList
 */
extern int
prefixLengthUntilCharInStr (const char *stringValue, const char *refList)
{
  if (stringValue == NIL_String)
    return 0;
  return strcspn (stringValue, refList);
}

/*
 * <stringValue> is not NIL
 * <refList> does not contain NULL_char
 * <maxSize> is at least 0, is at most <stringValue> length
 * success: <stringValue> is the concatenation of prefix, charX, suffix1, suffix2
 *           where charX belongs to <refList>
 *           suffix1 has no occurrence of a character belonging to <refList>
 *           length of prefix + length of suffix1 + 1 is <maxSize>
 *          returns length of prefix 
 * failure: none of the <maxSize> first characters of <stringValue> belongs to <refList>
 *          returns -1
 */
extern int
largestPrefixLengthUntilCharInStr (const char *stringValue, const char *refList, int maxSize)
{
  int length, charIndex, nCharIndex = 0;
  do
    {
      charIndex = nCharIndex - 1;
      length = strcspn ((char *) (stringValue + nCharIndex), refList);
      nCharIndex = nCharIndex + length + 1;
    }
  while (nCharIndex <= maxSize);
  return charIndex;
}

/*
 * returns -1 if sequence cc1,cc2 does not belong to stringValue 
 * returns length of the longest prefix before first cc1, cc2 consecutive occurrences if cc1,cc2 belongs to stringValue
 */
extern int
prefixLengthUntilCharChar (const char *stringValue, char cc1, char cc2)
{
  int length, charIndex = 0;

  if (stringValue == NIL_String)
    return -1;
  while (stringValue[charIndex] != NULL_char)
    {
      length = prefixLengthUntilChar ((char *) (stringValue + charIndex), cc1);
      if (length < 0)
	return -1;
      charIndex = charIndex + length;
      if (stringValue[charIndex + 1] == cc2)
	return charIndex;
      charIndex = charIndex + 1;
    }
  return -1;
}

/* =============================================================================
 * standard creation functions
 */

/* success: return a new mallocat'd string containing <cc>
   failure: prints an error message to stderr
            returns NIL_String
*/
extern char *
newCstr_Char (char cc)
{
  char *container = (char *) malloc (2);
  if (container == NULL)
    {
      stderr_alloc ("unsafe_Cstrmalloc");
      return NIL_String;
    }
  container[0] = cc;
  container[1] = NULL_char;
  return container;
}

/* success: returns a new mallocat'd container large enough to hold max(0,<size>) not NULL characters 
 *             followed by a NULL_char string termination 
 *             container is initiated with <cc>
   failure: prints an error message to stderr
            returns NIL_String
  */
extern char *
newCstrN_Char (char cc, int size)
{
  char *container = Cstrmalloc (size);
  if (container != NIL_String)
    unsafe_cpyNChar (container, cc, size);
  return container;
}

/* success: vstring is NIL_String returns NIL_String
 *          vstring is not NIL_String 
 *             return a new malloc'd string containing a copy of NULL_char terminated <vstring>
 * failure: prints an error message to stderr
 *          returns NIL_String
*/
extern char *
newCstr_String (const char *vstring)
{
  if (vstring != NIL_String)
    {
      char *container;
      int strLength = strlen (vstring);
      if (strLength > 0)
	{
	  if (unsafe_Cstrmalloc (strLength, &container))
	    container = strcpy (container, vstring);
	  return container;
	}
      return emptyCString ();
    }
  return NIL_String;
}

/* 2005-01-04
 * success: returns a new mallocat'd container large enough to hold max(0,<bound>) not NULL characters
 *             followed by a NULL_char string termination 
 *          <vstring> is NIL_String container is emptied
 *          <vstring> is not NIL_String container is initiated with at most <bound> firt characters of <vstring> 
 * failure: prints an error message to stderr
 *          return NIL_String
 */
extern char *
newCstrN_String (const char *vstring, int bound)
{
  if (bound > 0)
    {
      char *container;
      if (unsafe_Cstrmalloc (bound, &container))
	{
	  unsafe_NcpyString (container, vstring, bound);
	  return container;
	}
    }
  return emptyCString ();
}

/* success: return a new malloc'd container large enough to hold strlen(<vstring>)+max(0,<overFlow>) characters
 *          initiated with a NULL_char terminated copy of <vstring>
 * failure: memory allocation failed
 *          prints an error message to stderr
 *          return NIL_String
 */
extern char *
newCstr_container (const char *vstring, int overFlow)
{
  char *container;
  if (vstring == NIL_String)
    return Cstrmalloc (overFlow);
  if (overFlow <= 0)
    return newCstr_String (vstring);
  if (unsafe_Cstrmalloc (strlen (vstring) + overFlow, &container))
    container = strcpy (container, vstring);
  return container;
}

/* @return a new mallocat'd string resulting of the concatenation of vstring and cc
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_catChar (const char *vstring, char cc)
{
  char *container = newCstr_container (vstring, 1);
  if (container != NIL_String)
    unsafe_catChar (container, cc);
  return container;
}

/* @return a new mallocat'd string resulting of the concatenation of vstring1 and vstring2
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_catStr (const char *vstring1, const char *vstring2)
{
  char *container;
  if (vstring2 == NIL_String)
    return newCstr_String (vstring1);
  container = newCstr_container (vstring1, strlen (vstring2));
  if (container != NIL_String)
    container = strcat (container, vstring2);
  return container;
}

/* @return a new mallocat'd string resulting of the concatenation of <stringNumber> first elements of <stringArray>
   assumes stringNumber<=stringArray size
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_StrArray (const char *stringArray[], int stringNumber)
{
  int ii;
  int tailLength = stringNumber;
  for (ii = 0; ii < stringNumber; ii++)
    {
      if (stringArray[ii] == NIL_String)
	tailLength = tailLength - 1;
      else
	{
	  stringArray = &(stringArray[ii]);
	  break;
	}
    }
  if (tailLength <= 2)
    {
      if (tailLength <= 0)
	return NIL_String;
      if (tailLength == 1)
	return newCstr_String (stringArray[0]);
      return newCstr_catStr (stringArray[0], stringArray[1]);
    }
  else
    {
      char *container = Cstrmalloc (get_strArraySize (stringArray, tailLength));
      if (container != NIL_String)
	unsafe_cpyStrArray (container, stringArray, tailLength);
      return container;
    }
}

/* @return a new mallocat'd string resulting of the concatenation of vstring1, sep, vstring2
   if suffix is NIL_String or empty, the string terminates with sep
   if vstring is NIL_String or empty, the string begins with sep
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_catSepStr (const char *vstring1, char sep, const char *vstring2)
{
  char *container;
  if (vstring2 == NIL_String)
    return newCstr_catChar (vstring1, sep);
  container = newCstr_container (vstring1, strlen (vstring2) + 1);
  if (container == NIL_String)
    return NIL_String;
  unsafe_catChar (container, sep);
  container = strcat (container, vstring2);
  return container;
}

/* @return a new mallocat'd string resulting of the concatenation of <stringNumber> first elements of <stringArray>
 * NIL_String is separated by sep
   assumes 0<=stringNumber<=stringArray size
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_sepStrArray (char sep, const char *stringArray[], int stringNumber)
{
  if (stringNumber <= 2)
    {
      if (stringNumber <= 0)
	return NIL_String;
      if (stringNumber == 1)
	return newCstr_String (stringArray[0]);
      return newCstr_catSepStr (stringArray[1], sep, stringArray[2]);
    }
  else
    {
      char *container = Cstrmalloc (get_strArraySize (stringArray, stringNumber) + stringNumber - 1);
      if (container != NIL_String)
	unsafe_cpySepStrArray (container, sep, stringArray, stringNumber);
      return container;
    }
}

/* @return a new mallocat'd string equal to the greatest substring of vstring beginning at index 
 * that doesn't contain occurrence of ' ' or sep
 * assumes 0<=index<=strlen(vstring)
 * sets the new value of index to its next value for which  vstring[index] is not ' '
 * index initial value remains unchanged if  vstring[index] is not ' '
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstrsubstringupto (const char *vstring, char sep, int *index)
{
  if (vstring != NIL_String)
    {
      char cc;
      int nb, n = *index;
      cc = vstring[n];
      while ((cc != sep) && (cc != ' ') && (cc != NULL_char))
	{
	  n++;
	  cc = vstring[n];
	}
      nb = n - (*index);
      if (nb > 0)
	{
	  char *res = newCstrN_String (&(vstring[*index]), nb);
	  *index = n + 1;
	  return res;
	}
    }
  return NIL_String;
}

/* if vstring is a C string removes its embedding C string mark in a new mallocat'd string 
   otherwise returns vstring
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_removeCStringMarks (char *vstring)
{
  if (vstring != NIL_String)
    {
      int lg = strlen (vstring);
      if (lg > 1)
	if ((vstring[0] == '"') && (vstring[lg] == '"'))
	  {
	    if (lg == 2)
	      return NIL_String;
	    return newCstrN_String (&(vstring[1]), lg - 2);
	  }
    }
  return vstring;
}

/* @return a new mallocat'd string containing the representation of the positive integer val
   failure: prints an error message to stderr
            return NIL_String
 */
extern char *
newCstr_posInt (long int val)
{
  short lgth = get_strPosIntSize (val);
  if (lgth > 0)
    {
      char *container;
      if (unsafe_Cstrmalloc (lgth, &container))
	unsafe_cpyIntAbsVal (container, val, lgth);
      return container;
    }
  return NIL_String;
}

/* sets the new value of index to its next value for which  vstring[index] is not ' '
   assumes 0<=index<=vstring length
 * index initial value remains unchanged if  vstring[index] is not ' ' or vstring is NIL_String
 */
extern int
setIndexOnNonSpace (const char *vstring, int index)
{
  if (vstring != NIL_String)
    while (vstring[index] == ' ')
      index++;
  return index;
}

/*
 * returns pk_true if <ref> is a prefix of <source>
 *         (thus  pk_true if <ref> is NIL_String)
 */
extern boolean
isstrprefix (const char *ref, const char *source)
{
  if (ref == NIL_String)
    return pk_true;
  if (source == NIL_String)
    return pk_false;
  return (strncmp (ref, source, strlen (ref)) == 0);
}

/*
 * returns pk_true if length of <ref> is <length> and ref is a prefix of <source>
 */
extern boolean
isstrnprefix (const char *ref, const char *source, int length)
{
  if (ref == NIL_String)
    return (length == 0);
  if (strlen (ref) != length)
    return pk_false;
  if (source == NIL_String)
    return pk_false;
  if (strncmp (ref, source, length) == 0)
    return pk_true;
  return pk_false;
}

/* prefix compares cc with prefix pstr2 of length lg2 in str2 
 * assumes 0<=lg2<= length str2
 * returns 0 if lg2 = 1 and first character of pstr2 is cc
 * returns -1 if pstr1 < pstr2
 * returns 1 if pstr2 < pstr1
 */
extern int
CstrHeaderCharCompare (char cc, const char *str2, int lg2)
{
  if ((str2 == NIL_String) || (lg2 == 0))
    return 1;
  if (cc == str2[0])
    {
      if (lg2 == 1)
	return 0;
      return -1;
    }
  if (cc < str2[0])
    return -1;
  return 1;
}

/* compares prefix pstr1 of length lg1 in str1 with prefix pstr2 of length lg2 in str2 
 * assumes 0<=lg1<= length str1, 0<=lg2<= length str2
 * returns 0 if prefixes are equal
 * returns -1 if pstr1 < pstr2
 * returns 1 if pstr2 < pstr1
 */
extern int
CstrPrefixCompare (const char *str1, int lg1, const char *str2, int lg2)
{
  int res, lg;

  if (str1 == NIL_String)
    {
      if (str2 == NIL_String)
	return 0;
      return -1;
    }
  if (str2 == NIL_String)
    return 1;
  if (lg1 <= lg2)
    lg = lg1;
  else
    lg = lg2;
  res = strncmp (str1, str2, lg);
  if (res == 0)
    {
      if (lg1 == lg2)
	return 0;
      if (lg1 < lg2)
	return (-1);
      return 1;
    }
  return res;
}

/* 2004-08-12
   success: ch is a letter or ch is  a digit or ch is '_' (allowed not first characters for C identifier)
   failure: ch is a not a letter and ch is not a digit and ch is not '_'
 */
extern boolean
isCIdChar (char ch)
{
  return (isalnum (ch) || ch == '_');
}

/* converts uppercase characters in str to lowercase 
 * proceeds on lg characters
 * assumes lg at most equal to the length of str
 * assumes ansi behaviour of tolower
 */
extern void
stringToLower (char *str, int lg)
{
  if (str != NIL_String)
    {
      int i;
      for (i = 0; i < lg; i++)
	str[i] = (char) (tolower (str[i]));
    }
}

/* converts lowercase characters in str to uppercase 
 * proceeds on g characters
 * assumes lg at most equal to the length of str
 * assumes ansi behaviour of toupper
 */
extern void
stringToUpper (char *str, int lg)
{
  if (str != NIL_String)
    {
      int i;
      for (i = 0; i < lg; i++)
	str[i] = (char) (toupper (str[i]));
    }
}
