/*
 * This file is strongly host system dependent
 */

ref2LstRefVarType knownVariableList = NIL_LstRefVarType;

int commandLineSize;
char *commandBuffer = NIL_String;
char *workingDirectoryPrefix = NIL_String;
int WDLength;
char *variableBuffer = NIL_String;
char **curErrMess;

/* ===================================================================== 
 * host variable syntax dependant functions
 *
 */

/* ANSI
 * prints an error <message> on stderr
 */
static void
sys_stderrLine (const char *message)
{
  fprintf (stderr, "%s\n", message);
}

/* ANSI
 * prints an error <message> on stderr
 */
static void
sys_lineErr (const char *message)
{
  fprintf (stderr, "## ERROR in fileSystemEnv ##: %s\n", message);
}

/* ANSI
 * prints on stderr an error message made of space separated concatenation of <mess1> and <mess2>
 */
static void
sys_lineErr2 (const char *mess1, const char *mess2)
{
  fprintf (stderr, "## ERROR in fileSystemEnv ##: %s %s\n", mess1, mess2);
}

/* ANSI
 */
static void
init_Command ()
{
  commandLineSize = FILENAME_MAX + 64;
  commandBuffer = (char *) calloc (commandLineSize + 1, 1);
}

/* ANSI
 */
static void
init_Variable ()
{
  variableBuffer = (char *) calloc (FILENAME_MAX + 1, 1);
}

/* ANSI
 * workingDirectoryPrefix is a string container
 *           large enough to contain current path string (not checked)
 * set working directory variable to host current directory
 */
static void
init_WDContent ()
{
  char *WDN = getenv (_ENV_WORKINGDIR);
  if (WDN == NULL)		// added LBesnard
    workingDirectoryPrefix = strcpy (workingDirectoryPrefix, ".");
  else
    workingDirectoryPrefix = strcpy (workingDirectoryPrefix, WDN);
  WDLength = strlen (workingDirectoryPrefix);
  _MAKEDIRECTORYPREFIX (workingDirectoryPrefix, WDLength);
}

/* NOT ANSI
 *     requires _MAXPATHLEN 
 * workingDirectoryPrefix is a new allocated string container
 *           large enough to contain current path string (not checked)
 * set working directory variable to host current directory
 */
static void
init_WD ()
{
  workingDirectoryPrefix = (char *) malloc (_MAXPATHLEN + 1);
  init_WDContent ();
}

/* ===================================================================== 
 * Memory level variable management
 *
 */

/*
 * <aVariable> is a not NIL_HostVariable (not checked)
 * creates and return a new aliasing entry in association table
 */
static void
register_Variable (ref2HostVariable aVariable)
{
  ref2LstRefVarType newatom = (ref2LstRefVarType) malloc ((unsigned) (sizeof (*newatom)));
  newatom->car = aVariable;
  newatom->cdr = knownVariableList;
  knownVariableList = newatom;
}

/*
 * <varName> is a not empty NULL_char terminated string (not checked)
 * if <varName> starts with a character with _NOTENVNAMECHARSET, variable name is the rest of <varName>
 *   else variable name is <varName>
 * success: variable name is declared in host
 *          creates and returns a new aliasing from variable name 
 *              to a new m'allocated copy of its system associated value
 * failure: returns NIL_HostVariable
 */
static ref2HostVariable
new_HostVariableMirror (const char *varName)
{
  ref2HostVariable refVariable;
  char *aStringValue = getenv (varName);
  if (aStringValue == NULL)
    {
      sys_lineErr2 ("Undefined host variable", varName);
      return NIL_HostVariable;
    }
  refVariable = (ref2HostVariable) malloc ((unsigned) (sizeof (*refVariable)));
  refVariable->envName = varName;
  refVariable->value = newCstr_String (aStringValue);
  register_Variable (refVariable);
  refVariable->inError = pk_false;
  return refVariable;
}

/*
 * set known variable list empty
 */
static void
reset_VariableList ()
{
  ref2LstRefVarType deletedCell;
  while (knownVariableList != NIL_LstRefVarType)
    {
      deletedCell = knownVariableList;
      knownVariableList = knownVariableList->cdr;
      free (deletedCell);
    }
}

/*
 * <varName> is a NUL_char terminated string (not checked !)
 * success: returns a pointer to the pair (var varName, var value)
 * failure: cannot be translated 
 *          prints an error message on stderr
 *          returns NIL_HostVariable
 */
static ref2HostVariable
find_HostVariable_ (const char *varName)
{
  ref2HostVariable refVariable;
  ref2LstRefVarType searchedList = knownVariableList;
  while (searchedList != NIL_LstRefVarType)
    {
      refVariable = searchedList->car;
      if (areStringEq ((refVariable)->envName, varName))
	return refVariable;
      searchedList = searchedList->cdr;
    }
  return new_HostVariableMirror (varName);
}

/*
 * <source> is a NUL_char terminated string (not checked) denoting a variable environnement XX
 * XX is the longest string in <source>, free of character in ":$ /!\n" 
 * success: returns a pointer to the pair (XX, copy of XX host value)
 * failure: XX is empty or cannot be translated 
 *          prints an error message on stderr
 *          returns NIL_HostVariable
 */
static ref2HostVariable
find_HostVariable_heading_ (const char *source)
{
  if (source != NIL_String)
    {
      size_t length = strcspn (source, _NOTENVNAMECHARSET);
      if (length > 0)
	{
	  ref2HostVariable refVariable;
	  char *varName;
	  ref2LstRefVarType searchedList = knownVariableList;
	  while (searchedList != NIL_LstRefVarType)
	    {
	      refVariable = searchedList->car;
	      if (isstrnprefix ((refVariable)->envName, source, length))
		return refVariable;
	      searchedList = searchedList->cdr;
	    }
	  varName = newCstrN_String (source, length);
	  return new_HostVariableMirror (varName);
	}
    }
  sys_lineErr ("host variable name is empty");
  return NIL_HostVariable;
}

/*
 * source is NIL_String or a NUL_char terminated string (not checked !)
 * source begins with a var name equal to the longest string, free of ' ', '!', '\n'
 * success: not empty var name is defined in the environnement
 *              returns its value; it must remain unchanged
 *          sets nameLength equal to var name length that is > 0 
 * failure: var name is empty
 *          prints an error message on stderr
 *          sets nameLength equal to 0
 *          returns NIL_String
 */
static char *
get_string_HostValue_heading_length_ (const char *source, int *nameLength)
{
  ref2HostVariable refVariable;
  refVariable = find_HostVariable_heading_ (source);
  if (refVariable == NIL_HostVariable)
    {
      *nameLength = 0;
      return NIL_String;
    }
  *nameLength = strlen (refVariable->envName);
  return refVariable->value;
}

/* <source> is not NIL_String (not checked>
 * uses variableBuffer 
 *    requires variableBuffer large enough to receive length characters (not checked)
 * copies <length> first characters of <source> in variableBuffer
 */
static void
resetPathBuffer (const char *source, size_t length)
{
  if (length > 0)
    {
      source = (const char *) memcpy (variableBuffer, source, length);
      variableBuffer[length] = '\0';
    }
  else
    variableBuffer[0] = '\0';
}

/* <source> is a NULL terminated string 
 * uses variableBuffer 
 * success: appends the string <source> to variableBuffer from pos targetIndex
 *          returns <source> length
 * failure: not enough space available in variableBuffer
 *          returns 0
 */
static int
append2Target (const char *source, int targetIndex)
{
  int length = strlen (source);
  if ((targetIndex + length) <= FILENAME_MAX)
    {
      source = (const char *) memcpy ((variableBuffer + targetIndex), source, length);
      variableBuffer[targetIndex + length] = NULL_char;
      return length;
    }
  return 0;
}

/**
  @return  a new, created, aliasing entry in association table.
 */
static void
registerVariable (ref2VariableType newVariable)
{
  ref2LstRefVarType newatom = (ref2LstRefVarType) malloc ((unsigned) (sizeof (*newatom)));
  newatom->car = newVariable;
  newatom->cdr = knownVariableList;
  knownVariableList = newatom;
}

/**
   @return a new, created, aliasing from varName to pathName.
 */
static ref2VariableType
newVariable (char *varName, char *pathName)
{
  ref2VariableType refVariable = (ref2VariableType) malloc ((unsigned) (sizeof (*refVariable)));
  refVariable->envName = varName;
  refVariable->value = pathName;
  refVariable->inError = (pathName == NULL);
  return refVariable;
}


/**
   source contains a symbolic name.
   @return a pointer to the pair (var name, var value).
   If var name cannot be translated set var value to NIL_String
 */
static ref2VariableType
findVariable (char *source)
{
  return find_HostVariable_heading_ (source);

  /*ref2VariableType refVariable;
     char *varName;
     char *pathName;
     ref2LstRefVarType searchedList = knownVariableList;
     size_t length = getVariableLength(source);

     while (searchedList != NIL_LstRefVarType) 
     {
     refVariable = searchedList->car;
     if (isstrnprefix((refVariable)->envName, source, length))  return refVariable;
     searchedList = searchedList->cdr;
     }

     varName = clonePrefix(source, length);
     pathName = getEnvStrDef(varName); 
     refVariable =  newVariable(varName, pathName);
     registerVariable(refVariable);
     return refVariable; */
}


/** 
    source contains a symbolic name.
    @return "true" if source is already defined and refVariable is this variable, "false" otherwise.
 */
static boolean
isDefinedVariable (const char *source, ref2VariableType * refVariable)
{
  ref2LstRefVarType searchedList = knownVariableList;
  size_t length = strcspn (source, _NOTENVNAMECHARSET);	//  size_t length = getVariableLength(source);

  while (searchedList != NIL_LstRefVarType)
    {
      (*refVariable) = searchedList->car;
      if (isstrnprefix ((*refVariable)->envName, source, length))
	return pk_true;
      searchedList = searchedList->cdr;
    }
  return pk_false;
}


static char *
strduplicate (const char *absoluteFilePathName)
{
  char *x;
  int n;

  n = (int) strlen (absoluteFilePathName);
  x = (char *) malloc (n + 1);
  memset (x, 0, n);
  x = strcat (x, absoluteFilePathName);
  return x;
}
