/* 2004-08-12
 * a PKMAKE variant of a GNU variant of lineget
 * Read from aFile into *lineptr (and null-terminate it). 
 *   *lineptr is a pointer to lSize+1 characters space
 *                 (including space for terminating null char). 
 * success: a newline is seen when at most lSize characters have been read
 *                 returns in <lineptr> the characters that have been read upto the first newline occurrence
 *              returns the number (>0,<=lSize) of read characters (including end of line)
 *          end-of-file is encountered when at least one character and at most lSize characters have been read
 *                 returns in <lineptr> the characters that have been read
 *              returns -1
 *          end-of-file is encountered before reading any character
 *                 returns empty string in <lineptr> (first char is NULL_char)
 *              returns 0
 * failure: lSize characters have been read without encountering a newline occurrence and <aFile> being not ended
 *                 returns in <lineptr> the <lSize> characters that have been read
 *              returns -2
 *           an error occurred during aFile reading
 *              returns -3
 *          
 */
extern int
getpKLine (char *lineptr, int lSize, FILE * aFile)
{
  char *res = fgets (lineptr, lSize + 1, aFile);
  if (res == lineptr)
    {
      if (lineptr[strlen (res) - 1] == '\n')
	return (strlen (res));
      /* at most lSize characters have been read without encountering a newline occurrence */
      if (feof (aFile))
	return -1;
      return -2;
    }
  /* res is null pointer:
     an error occurred or remaining file content is empty */
  lineptr[0] = NULL_char;
  if (feof (aFile))
    return 0;
  return -3;
}

/* close a file descriptor
 * host system direct dependency: 
 *       function fclose
 * input
 *       file to close
 */
extern void
rwFClose (FILE * file)
{
  fclose (file);
}

/* checks if fileName is a readable file name
 * success: returns pk_true
 * failure: returns false
*/

extern boolean
file_exist (const char *fileName)
{
  FILE *fg;

  fg = fopen (fileName, "rt");
  if (fg == NULL)
    return pk_false;
  fclose (fg);
  return pk_true;
}

/* open a file in writing mode and returns its new system file descriptor
 * host system direct dependency: 
 *         function fopen
 * PK direct dependency: 
 *         constant NIL_String, NIL_File
 * input
 *       filePathName: string terminated by a standard 0 refering to a file full identification
 *                 filePathName contains only authorised characters in the host system
 * success
 *    output 
 *       errMess: NIL_String
 *    returns open file system descriptor
 * failure 
 *    output 
 *       errMess: error message
 *    returns NIL_File
 */
extern FILE *
wFOpen (const char *filePathName, const char **errMess)
{
  FILE *writeFile;
  if (filePathName == NIL_String)
    {
      *errMess = "empty path name";
      return (NIL_File);
    }
  writeFile = fopen (filePathName, "w");
  if (writeFile == NULL)
    {
      *errMess = "opening failed in writing mode";
      return NIL_File;
    }
  *errMess = NIL_String;
  return writeFile;
}

/* open a file in reading mode and returns its new system file descriptor
 * host system direct dependency: 
 *         function fopen
 * PK direct dependency: 
 *         constant NIL_String, NIL_File
 * input
 *       filePathName: string terminated by a standard 0 refering to a file full identification
 *                 filePathName contains only authorised characters in the host system
 * success
 *    output 
 *       errMess: NIL_String
 *    returns open file system descriptor
 * failure 
 *    output 
 *       errMess: error message
 *    returns NIL_File
 */
extern FILE *
rFOpen (const char *filePathName, const char **errMess)
{
  FILE *readFile;
  if (filePathName == (NIL_String))
    {
      *errMess = "empty path name";
      return NIL_File;
    }
  readFile = fopen (filePathName, "r");
  if (readFile == NULL)
    {
      *errMess = "opening failed in reading mode";
      return NIL_File;
    }
  *errMess = NIL_String;
  return readFile;
}
