/*
 * This file is strongly host system dependent
 */

/* 
 * returns one hour duration
 * host system direct dependency  
 *             function time 
 */
extern time_t
one_hour ()
{
  return oneFHour;		/*ANSI */
}

/* 
 * returns a time_t value formated as YYYY-MM-JJ HH:MM:SS
 *
 */
extern char *
strcvtTime (time_t value)
{
  static char buffer[timeStrSize + 1];
  const char *format = "%Y-%m-%d %H:%M:%S";
  struct tm *tm = localtime (&value);	/*ANSI */
  (void) strftime (buffer, timeStrSize + 1, format, tm);	/*ANSI */
  return buffer;
}

/* 
 * start Time date is the date of the first call to getStartTime or getNow
 * set start Time date if it is not set
 * returns start Time
 */
extern time_t
getStartTime ()
{
  setStartTime ();
  return startTime;
}

/* 
 * set start Time date if it is not set
 * returns current date
 */
extern time_t
getNow ()
{
  setStartTime ();
  return get_now ();		/*ANSI */
}

/* 
 * Apocalypse is one year later than start Time 
 * set start Time date if it is not set
 * returns Apocalypse
 */
extern time_t
getApocalypse ()
{
  setStartTime ();
  return apocalypse;
}

/*@returns  cpu time in milliseconds*/
extern int
getcputime (void)
{
  long duration = clock ();

  setStartTime ();
  if (CLOCKS_PER_SEC >= 1000)
    return (int) (duration / ((CLOCKS_PER_SEC) / 1000));
  else
    {
      fprintf (stderr, "warning CLOCKS_PER_SEC=%ld\n", CLOCKS_PER_SEC);
      return (int) duration;
    }
}

/*@returns time in seconds since start time */
extern int
getTotalTime (void)
{
  time_t now = getNow ();
  return (int) (difftime (now, startTime));
}

/* 
 * Non ANSI <stdio> facility 
 * get filePath last modification date; file may be missing
 * host system direct dependency: 
 *         types struct stat
 *         variable errno
 *         function stat
 * input
 *       <filePath>: string terminated by a standard 0 referring to a file full identification
 *                 <filePath> contains only authorised characters in the host system
 * success: <errMess> becomes NIL_String
 *          file <filePath> exists returns its local last modification date in system units
 *          file <filePath> does not exist returns "apocalypse" time, greater than current time
 * failure: <errMess> contains an error message
 *          returns NIL_time (smaller than any time)
 */
extern time_t
getFDate (const char *filePath, const char **errMess)
{
#if (xKHOST != xKwinNT)
  _STATTYPE buf;
  time_t fileTime;
  setStartTime ();
  if (filePath == NIL_String)
    {
      *errMess = "empty file path name";
      return NIL_time;
    }
  errno = 0;			/*ANSI */
  _ASSIGNFTIME (fileTime, filePath, buf);
  if (errno == 0)
    return fileTime;
  if (_ERRFILEMISSING ())
    {
      errno = 0;
      *errMess = "file does not exist";
      return apocalypse;
    }
  if (_ERRFILEPERMISSION ())
    *errMess = "permission denied";
  else
    *errMess = "file access error";
  return NIL_time;
#endif
  *errMess = "file access error";
  return NIL_time;
}
