#define print_LINEREF_sp(_NUMBER_) space("/line"); space(_NUMBER_)
#define print_COLUMNREF_sp(_NUMBER_) space("/column"); space(_NUMBER_)
#define print_LINECOLUMNREF_sp(_LN_, _COL_) print_LINEREF_sp(_LN_); print_COLUMNREF_sp(_COL_)

_PUBLIC_
/* sets silent mode to b
   returns previous state of silent mode
 */
  boolean LogPrinter::setSilentModeMessagesTo (boolean b)
{
  boolean
    x = silentMode;
  silentMode = b;
  return x;
}

/* prints positive int <val> followed by space followed by car
 */
void
LogPrinter::putmesstimeunit (int val, char car)
{
  if (val < 10)
    print ('0');
  space ((long int) val);
  space (car);
}

/* converts vtime in hh, mn, sec, centiseconds
 */
void
LogPrinter::intTimeConversion (int vtime, int *hh, int *mn, int *sec, int *cs)
{
  int ts, tm;

  ts = vtime / 100;
  tm = ts % 3600;
  *hh = ts / 3600;
  *mn = tm / 60;
  *sec = tm % 60;
  *cs = vtime % 100;
}

/* prints "time" followed by <vtime> in hours, minuts, seconds, centiseconds units
 */
void
LogPrinter::printDuration (int vtime)
{
  int hh, mn, sec, cs;

  intTimeConversion (vtime, &hh, &mn, &sec, &cs);
  print ("        time   :");
  if (hh > 0)
    putmesstimeunit (hh, 'h');
  if (mn > 0)
    putmesstimeunit (mn, 'm');
  putmesstimeunit (sec, 's');
  putmesstimeunit (cs, ' ');
}

/* prints cpu-time since the previous call. */
void
LogPrinter::printCpuTimeDuration ()
{
  double x;

  x = getcputime ();
  printDuration ((int) (x - xtime));
  xtime = x;
}


/* returns cpu-time since program starting */
int
LogPrinter::getFullDuration ()
{
  return (int) (getcputime () - beginAt);
}

/* @return error code condition. */
boolean
LogPrinter::getErrorCode (void)
{
  return cc;
}

/* Sets  error code condition  to ?scc */
void
LogPrinter::setErrorCode (boolean scc)
{
  cc = scc;
}

/* appends "line" message followed by space
 *       followed by <lineNumber> followed by space
 */
void
LogPrinter::print_lineRef_sp (int lineNumber)
{
  print_LINEREF_sp ((long int) lineNumber);
}

/* appends "line" message followed by space followed by 
 *       <lineNumber> followed by space followed by
 *       "column" message followed by space followed by 
 *        <columnNumber> followed by space
 */
void
LogPrinter::print_lineColumnRef_sp (int lineNumber, int columnNumber)
{
  print_LINECOLUMNREF_sp ((long int) lineNumber, (long int) columnNumber);
}

/* appends "aware" message followed by <strMess> followed by space
   registers an error occurrence
 */
void
LogPrinter::error__sp (const char *strMess)
{
  setErrorCode (pk_false);
  space (awareMark);
  space (strMess);
}

/* appends "aware" message followed by space followed by column number
 * followed by <strMess> followed by space
 * registers an error occurrence
 */
void
LogPrinter::errorColumn_sp (int columnNumber, const char *strMess)
{
  setErrorCode (pk_false);
  space (awareMark);
  print_COLUMNREF_sp ((long int) columnNumber);
  space (strMess);
}

/* appends "aware" message followed by space followed by line number
 * followed by <strMess> followed by space
 * registers an error occurrence
 */
void
LogPrinter::errorLine_sp (int lineNumber, const char *strMess)
{
  setErrorCode (pk_false);
  space (awareMark);
  print_LINEREF_sp ((long int) lineNumber);
  space (strMess);
}

/* appends "aware" message followed by space followed by line number
 * followed by <strMess> and output this
 * registers an error occurrence
 */
void
LogPrinter::errorLine_ln (int lineNumber, const char *strMess)
{
  setErrorCode (pk_false);
  space (awareMark);
  print_LINEREF_sp ((long int) lineNumber);
  line (strMess);
}


/* appends "aware" message followed by space followed by line number
 * followed by column number followed by <strMess> and output this
 * registers an error occurrence
 */
void
LogPrinter::errorLineColumn_sp (int lineNumber, int columnNumber, const char *strMess)
{
  setErrorCode (pk_false);
  space (awareMark);
  print_LINECOLUMNREF_sp ((long int) lineNumber, (long int) columnNumber);
  space (strMess);
}

void
LogPrinter::init (boolean silentm, const char *aware_Mark)
{
  xtime = getcputime ();
  beginAt = xtime;
  silentMode = silentm;
  awareMark = aware_Mark;
}

void
LogPrinter::init_open (boolean silentm, const char *aware_Mark, FILE * logFile, const char *fileName, boolean * correct)
{
  if (*correct)
    {
      init (silentm, aware_Mark);
      (*correct) = open (logFile, fileName);
    }
}

/* PLG 2007_03_02 new <correct> DONE
 * init IntePrinter::IntePrinter(int verLevel, int lineLgth, const char * brkStr, char continuationChar, boolean *correct)
 */
LogPrinter::LogPrinter (boolean silentm, const char *aware_Mark, FILE * logFile, const char *fileName,
			int lineLgth, const char *brkStr, char continuationMark, boolean * correct):
NBuffer (lineLgth / 2, lineLgth, brkStr, continuationMark, correct)
{
  init_open (silentm, aware_Mark, logFile, fileName, correct);
}

/* PLG 2007_03_02 new <correct> DONE
 * init IntePrinter::IntePrinter(int verLevel, int lineLgth, boolean *correct)
 */
LogPrinter::LogPrinter (boolean silentm, const char *aware_Mark, FILE * logFile, const char *fileName, int lineLgth, boolean * correct):
NBuffer (lineLgth / 2, lineLgth, correct)
{
  init_open (silentm, aware_Mark, logFile, fileName, correct);
}

/* PLG 2007_03_02 new <correct> DONE
 * init logtools in Signal
 */
LogPrinter::LogPrinter (boolean silentm, const char *aware_Mark, const char *fileName, boolean * correct):
NBuffer (23, 127, correct)
{
  if (*correct)
    {
      init (silentm, aware_Mark);
      (*correct) = open (fileName);
    }
}
