/* a LGString that is not NIL_LGString has a not NIL_String buf
 * it can have _emptyString buf
 */
/* ============================
 * Container functions
 */
/* 2005-01-04
 * stringRegister is NIL returns 0
 * stringRegister is not NIL returns stringRegister size 
 */
extern int
strregGetSize (LGString stringRegister)
{
  if (stringRegister == NIL_LGString)
    return 0;
  return (stringRegister->size);
}

/* 2005-01-04
 * success: stringRegister size is <= 0
 * failure: stringRegister size is > 0
 */
extern boolean
strregIsNulSize (LGString stringRegister)
{
  if (stringRegister == NIL_LGString)
    return pk_true;
  return ((stringRegister->size) == 0);
}

/* 2005-01-04
 * stringRegister is not NIL returns stringRegister available free space 
 * stringRegister is NIL returns 0 
 */
extern int
strregGetAvailableSize (LGString stringRegister)
{
  if (stringRegister != NIL_LGString)
    if ((stringRegister->size) > 0)
      return (stringRegister->size) - strlen (stringRegister->buf);
  return 0;
}

/* 2005-01-04
 * success: stringRegister has at least <requiredSize> available free space 
 * failure: 
 */
extern boolean
strregIsAvailable (LGString stringRegister, int requiredSize)
{
  if (stringRegister == NIL_LGString)
    return pk_false;
  else if (requiredSize <= 0)
    return pk_true;
  else
    {
      int containerSize = stringRegister->size;
      if (containerSize <= 0)
	return pk_false;
      return (containerSize >= (strlen (stringRegister->buf) + requiredSize));
    }
}

/* 2005-01-04
 * success: returns a free register or a new m'allocated register in <stringRegister>
 *              with enough space to hold <requiredSize> not NULL character
 *              new register size is max (0, <requiredSize>)
 *              new register container is emptied
 * failure: descriptor memory allocation failed or container memory allocation failed
 *          an error message is stderr printed
 *          <stringRegister> remains unchanged
 */
extern boolean
new_sizedPKString (int requiredSize, LGString * stringRegister)
{
  if (!(findContainer (stringRegister, requiredSize)))
    {
      LGString wRegister = (LGString) malloc (sizeof (*wRegister));
      if (wRegister == NULL)
	{
	  fprintf (stderr, "## system ERROR ##: Memory allocation error in new_sizedPKString\n");
	  return pk_false;
	}
      if (requiredSize < 0)
	requiredSize = 0;
      wRegister->buf = Cstrmalloc (requiredSize);
      if ((wRegister->buf) == NIL_String)
	return pk_false;
      wRegister->size = requiredSize;
      (*stringRegister) = wRegister;
    }
  return pk_true;
}

/* 2005-01-04
 * success: returns a free register or a new m'allocated register in <stringRegister>
 *              with enough space to hold <requiredSize> not NULL character
 *              new register size is max (0, <requiredSize>)
 *              new register container is emptied
 *          returns its container in <stringContainer>
 * failure: descriptor memory allocation failed or container memory allocation failed
 *          an error message is stderr printed
 *          <stringRegister> remains unchanged
 */
extern boolean
new_sizedPKStringContainer (int requiredSize, LGString * stringRegister, char **stringContainer)
{
  if (new_sizedPKString (requiredSize, stringRegister))
    {
      (*stringContainer) = (*stringRegister)->buf;
      return pk_true;
    }
  return pk_false;
}

/* 2005-01-04
 * success: empties stringRegister container
 * failure: stringRegister is NIL or its string container is NIL
 */
extern void
strregEmpty (LGString stringRegister)
{
  if (stringRegister != NIL_LGString)
    if ((stringRegister->buf) != NIL_String)
      stringRegister->buf[0] = NULL_char;
}

/* 2005-01-04
 * success: <stringContainer> is set to not NIL <stringRegister> container
 * failure: <stringRegister> is NIL_LGString or <stringRegister> container is NIL_String
 *          <stringContainer> reamains unchanged
 */
extern boolean
getPKStringContainer (LGString stringRegister, char **stringContainer)
{
  if (stringRegister == NIL_LGString)
    return pk_false;
  if (stringRegister->size == 0)
    return pk_false;
  (*stringContainer) = stringRegister->buf;
  return pk_true;
}

/** string container : Frees the string container pointed to by stringRegister
    if its size is > 0
    future access to deleted PKString has unpredictable result
    @return the unique null container.
    _SEEALSO_{strregnew}.
*/
extern LGString
deletePKString (LGString stringRegister)
{
  if (stringRegister != NIL_LGString)
    if (stringRegister->size > 0)
      freeContainer (stringRegister);
  return NIL_LGString;
}

/* ============================
 * string functions
 */
/** 
    @return  0 if stringRegister is NIL
    @return  stringRegister actual string length if stringRegister is not NIL
*/
extern int
strregGetLength (LGString stringRegister)
{
  if (stringRegister != NIL_LGString)
    if (stringRegister->size > 0)
      return strlen (stringRegister->buf);
  return 0;
}

/** creates a new allocated copy of stringRegister char * content
    if stringRegister is not NIL_LGString and its length is >0  
    @return the new copy
    empties stringRegister
    if stringRegister is NIL_LGString or its length is >0  
    returns the unique null string 
    _SEEALSO_{strregnew}.
*/
extern char *
detachPKString (LGString stringRegister)
{
  if (stringRegister != NIL_LGString)
    if (stringRegister->size > 0)
      {
	char *wContainer = stringRegister->buf;
	char *stringResult = newCstr_String (wContainer);
	wContainer[0] = NULL_char;
	return stringResult;
      }
  return NIL_String;
}

/**
   @return a new allocated container loaded with c
*/
extern LGString
new_charPKString (char c)
{
  LGString wRegister;
  if (new_sizedPKString (1, &wRegister))
    {
      (wRegister->buf)[0] = c;
      (wRegister->buf)[1] = NULL_char;
      return wRegister;
    }
  return NIL_LGString;
}

/**
   Allocates a temporary string (a string container) loaded with the decimal representation of val
   if val is negative, the representation begins with '-'
   @return the new allocated temporary string.
*/
extern LGString
new_intPKString (int integerValue)
{
  LGString wRegister;
  int valueLength;

  if (integerValue >= 0)
    {
      valueLength = get_strPosIntSize (integerValue);
      if (!new_sizedPKString (valueLength + 1, &wRegister))
	return NIL_LGString;
    }
  else
    {
      integerValue = -integerValue;
      valueLength = get_strPosIntSize (integerValue);
      if (new_sizedPKString (valueLength + 2, &wRegister))
	{
	  (wRegister->buf)[0] = '-';
	  (wRegister->buf)[1] = NULL_char;
	}
      else
	return NIL_LGString;
    }
  unsafe_catIntAbsVal (wRegister->buf, integerValue, valueLength);
  return wRegister;
}

/**
   If ( requiredSize <= 0 ) returns nullContainer
   If ( requiredSize > 0 ) 
   Allocates requiredSize character spaces to a temporary string (a string container) 
   first characters of wRegister are filled with min(requiredSize, stringValue length) first characters of str
   @return the new allocated temporary string.
*/
extern LGString
new_nPKString (const char *stringValue, int requiredSize)
{
  LGString wRegister;
  if (requiredSize <= 0)
    return nullContainer ();
  if (new_sizedPKString (requiredSize, &wRegister))
    {
      unsafe_NcpyString (wRegister->buf, stringValue, requiredSize);
      return (wRegister);
    }
  return NIL_LGString;
}

/**
   If (requiredFreeSpace <= 0 ) returns new_nPKString(stringValue,usedSpace)
   If ( usedSpace <= 0 ) returns new_sizedPKString(requiredFreeSpace)
   If (requiredFreeSpace > 0 ) and ( usedSpace > 0 )
   allocates usedSpace+requiredFreeSpace characters to a temporary string (a string container) 
   first characters of wRegister are filled with min(usedSpace, stringValue length) first characters of stringValue
   @return the new allocated temporary string.
*/
extern LGString
new_stringPKString (const char *stringValue, int usedSpace, int requiredFreeSpace)
{
  LGString wRegister;
  if (requiredFreeSpace <= 0)
    return new_nPKString (stringValue, usedSpace);
  if (usedSpace <= 0)
    {
      if (new_sizedPKString (requiredFreeSpace, &wRegister))
	return wRegister;
      return NIL_LGString;
    }
  if (new_sizedPKString (usedSpace + requiredFreeSpace, &wRegister))
    {
      unsafe_NcpyString (wRegister->buf, stringValue, usedSpace);
      return wRegister;
    }
  return NIL_LGString;
}

/** string container:Reducing length of stringRegister string content to atmost maxLength characters  
    @return none
*/
extern void
strregStringCut (LGString stringRegister, int maxLength)
{
  if (stringRegister != NIL_LGString)
    if ((maxLength < (stringRegister->size)) && (maxLength >= 0))
      (stringRegister->buf)[maxLength] = NULL_char;
}

/** string container:appending a char.
    success: stringRegister has at least one free available slot
    c is concatenated to stringRegister content
    failure: stringRegister remains unchanged
*/
extern boolean
strregconccha (LGString stringRegister, char c)
{
  if (stringRegister != NIL_LGString)
    {
      char *wContainer = stringRegister->buf;
      int stringLength = strlen (wContainer);
      if (stringLength < stringRegister->size)
	{
	  wContainer[stringLength] = c;
	  wContainer[stringLength + 1] = NULL_char;
	  return pk_true;
	}
    }
  return pk_false;
}

/** success: stringRegister has at least factor free available slot or factor is at most 0
    if factor >0 factor occurrences of c are concatenated to stringRegister content
    failure: stringRegister remains unchanged
*/
extern boolean
strregconcmcha (LGString stringRegister, char c, int factor)
{
  if (factor <= 0)
    return pk_true;
  if (stringRegister != NIL_LGString)
    {
      char *wContainer = stringRegister->buf;
      int stringLength = strlen (wContainer);
      if ((stringLength + factor) <= stringRegister->size)
	{
	  int ii;
	  for (ii = 0; ii < factor; ii++)
	    wContainer[stringLength + ii] = c;
	  wContainer[stringLength + factor] = NULL_char;
	  return pk_true;
	}
    }
  return pk_false;
}

/** success: stringRegister has at least maxLength free available slots or  stringValue is NIL_String 
 *              or min(maxLength,  stringValue length) is at most 0
 if maxLength >0 min(maxLength,  stringValue length) first characters of  stringValue are concatenated to stringRegister content
 failure: stringRegister remains unchanged
*/
extern boolean
strregconcnString (LGString stringRegister, const char *stringValue, int maxLength)
{
  if (stringRegister == NIL_LGString)
    return pk_false;
  return unsafe_NcatStringAtMost (stringRegister->buf, stringValue, maxLength, stringRegister->size);
}

/** success:  stringValue is NIL_String or stringRegister has at least ( stringValue length) free available slots
    stringValue is concatenated to stringRegister content
    failure: stringRegister remains unchanged
*/
extern boolean
strregconcString (LGString stringRegister, const char *stringValue)
{
  if (stringRegister == NIL_LGString)
    return pk_false;
  if (stringValue == NIL_String)
    return pk_true;
  if (strregGetAvailableSize (stringRegister) < strlen (stringValue))
    return pk_false;
  stringRegister->buf = strcat (stringRegister->buf, stringValue);
  return pk_true;
}

/**
 * appends to stringRegister the string representation of i
 * or -i if i is negative
 * returns false if there is not enough space in stringRegister
 */

extern boolean
strregconcInt (LGString stringRegister, int integerValue)
{
  char *wContainer;
  int stringLength;
  int valueLength;

  if (stringRegister == NIL_LGString)
    return pk_false;
  wContainer = stringRegister->buf;
  stringLength = strlen (wContainer);
  if (integerValue >= 0)
    {
      valueLength = get_strPosIntSize (integerValue);
      if ((stringLength + valueLength + 1) > stringRegister->size)
	return pk_false;
    }
  else
    {
      valueLength = get_strPosIntSize (-integerValue);
      if ((stringLength + valueLength + 2) > stringRegister->size)
	return pk_false;
      wContainer[stringLength] = '-';
      wContainer[stringLength + 1] = NULL_char;
    }
  unsafe_catIntAbsVal (wContainer, integerValue, valueLength);
  return pk_true;
}
