/* string container */

#define stdRegLength  511

/* string container list */
typedef struct ContainerLst *ref2ContainerLst;
#define NIL_ContainerLst ((ref2ContainerLst)0)

typedef struct ContainerLst
{
  LGString car;
  ref2ContainerLst cdr;
} stContainerLst;

typedef struct stringContainer
{
  int size;			/* number of non nul characters that can be contained in buf */
  char *buf;			/* if buf is NIL_String then size is 0 */
} stStringContainer;

/* 
 * list of free containers containing at least one slot
 */
ref2ContainerLst freeContainerLst = NIL_ContainerLst;
ref2ContainerLst activeContainerLst = NIL_ContainerLst;
LGString nullLGString = NIL_LGString;
char *_emptyString = NIL_String;

/* 2005-01-04
 * success: <aLGString> is NIL_LGString
 *              a warning message is stderr printed
 *          <aLGString> is not NIL_LGString 
 *              Frees the string container contained in aLGString
 *              Empties it 
 *              links it in free container list
 * failure: memory allocation failed
 *          an error message is stderr printed
 */
void
freeContainer (LGString aLGString)
{
  if (aLGString == NIL_LGString)
    fprintf (stderr, "## system WARNING ##: cannot free NIL string in freeContainer\n");
  else
    {
      ref2ContainerLst regLst = (ref2ContainerLst) malloc (sizeof (*regLst));
      if (regLst == NULL)
	fprintf (stderr, "## system ERROR ##: Memory allocation error in freeContainer\n");
      else
	{
	  (aLGString->buf)[0] = NULL_char;
	  regLst->car = aLGString;
	  regLst->cdr = freeContainerLst;
	  freeContainerLst = regLst;
	}
    }
}

/*
  @returns null container named nullLGString
  sets nullLGString if required
*/
LGString
nullContainer ()
{
  if (nullLGString == NIL_LGString)
    {
      _emptyString = emptyCString ();
      nullLGString = (LGString) malloc (sizeof (*nullLGString));
      if (nullLGString == NULL)
	{
	  fprintf (stderr, "## system ERROR ##: Memory allocation error in newEmptyContainer\n");
	  return NIL_LGString;
	}
      nullLGString->size = 0;
      nullLGString->buf = _emptyString;
    }
  return nullLGString;
}

/* 2005-01-04
 * success: <aFreeContainer> is set to a new reallocated empty string container
 *          large enough to hold max(0, <size>) not NULL characters
 * failure: no suitable container is available
 */
boolean
findContainer (LGString * freeRegister, int size)
{
  LGString currentRegister;
  ref2ContainerLst previousCell;
  ref2ContainerLst currentCell;

  if (freeContainerLst == NIL_ContainerLst)
    return pk_false;
  currentCell = freeContainerLst;
  currentRegister = currentCell->car;
  if ((currentRegister->size) >= size)
    {
      freeContainerLst = freeContainerLst->cdr;
      *freeRegister = currentRegister;
      return pk_true;
    }
  while ((currentCell->cdr) != NIL_ContainerLst)
    {
      previousCell = currentCell;
      currentCell = currentCell->cdr;
      currentRegister = currentCell->car;
      if (currentRegister->size >= size)
	{
	  previousCell->cdr = currentCell->cdr;
	  *freeRegister = currentRegister;
	  return pk_true;
	}
    }
  return pk_false;
}
