_PUBLIC_
/* prevents further reading not preceded by lineOpen:
 *         line analyzer current atom becomes empty
 */
const char *
FileAnalyzer::get_fileName ()
{
  return this_fileName;
}

/* prevents further reading not preceded by lineOpen:
 *         line analyzer current atom becomes empty
 */
void
FileAnalyzer::setFileReadingCompleted ()
{
  setFileReadingEnd ();
}

/* prevents further reading not preceded by lineOpen:
 *         line analyzer current atom becomes empty
 */
void
FileAnalyzer::setFileReadingErrored ()
{
  setInError ();
}

/* 
 * retuns pk_true if FileAnalyzer is properly set for reading file
 */
boolean
FileAnalyzer::isFileReadingOK ()
{
  return fileReading_OK;
}

/* 2005-01-04
 * load this_fileBuffer with at most this_fileBufferSize new characters read in "this_file"
 *          if end of line is encountered it is the last not NULL_char in this_fileBuffer
 * success: at most this_fileBufferSize-1 characters followed by an end of line as been read
 *          increments line counter
 *          this_lineAnalyzer cursor is set at this_lineAnalyzer main buffer
 *          removes end of line mark from this_lineAnalyzer main buffer if it is the last read char
 *          returns this_lineAnalyzer main buffer new length
 *          this_lineAnalyzer atom remains unchanged
 * failure: eof has been reached before reading a character
 *               sets file completed
 *               this_lineAnalyzer atom becomes empty
 *               returns -1 
 *          read line is too long wrt container space
 *               this_lineAnalyzer atom remains unchanged
 *               returns -2
 *          other (errored) readings
 *               sets this errored
 *               this_lineAnalyzer atom becomes empty
 *               returns -3
 */
int
FileAnalyzer::nextFileLine ()
{
  if (fileReading_OK)
    {
      int length = getpKLine (this_fileBuffer, this_fileBufferSize - 1, this_file);
      /* <length> > 0: <length> characters have been read and copied in <this_fileBuffer> (including end of line)
       * length = 0: eof has been reach before reading any character
       * length = -1: eof has been reach after reading at least one character
       * length = -2: the number of characters read before meeting eol occurence is greater than this_fileBufferSize-1
       * length = -3: an error occurred
       */
      if (length >= -1)
	{
	  if (length == 0)
	    {
	      setFileReadingEnd ();
	      return -1;
	    }
	  this_lineCounter = this_lineCounter + 1;
	  return this_lineAnalyzer->lineOpenString (this_fileBuffer);
	}
      if (length == -2)
	{
	  this_lineCounter = this_lineCounter + 1;
	  return -2;
	}
      setInError ();
      return -3;
    }
  if (isFileReadingCompleted ())
    return -1;
  return -3;
}

/*
 * returns line counter
 */
int
FileAnalyzer::getLineNumber ()
{
  return this_lineCounter;
}

/*
 * print line counter and strMess in a new line of logBuffer
 * logBuffer is a write open file
 */
void
FileAnalyzer::trace_line_ln (ref2LogPrinter logBuffer)
{
  logBuffer->print_lineRef_sp (this_lineCounter);
  this_lineAnalyzer->ln_traceLine_ln (logBuffer);
}

/*
 * print line counter, index in line and strMess in a new line of logBuffer
 * logBuffer is a write open file
 */
void
FileAnalyzer::print_lineColumnRef_sp (ref2LogPrinter logBuffer)
{
  logBuffer->print_lineColumnRef_sp (this_lineCounter, (this_lineAnalyzer->get_cursor ()) + 1);
}

/*
 * print line counter, index in line and strMess in a new line of logBuffer
 * logBuffer is a write open file
 */
void
FileAnalyzer::trace_lineColumn_ln (ref2LogPrinter logBuffer)
{
  logBuffer->print_lineColumnRef_sp (this_lineCounter, (this_lineAnalyzer->get_cursor ()) + 1);
  this_lineAnalyzer->ln_traceLine_ln (logBuffer);
}

/*
 * print line counter and strMess in a new line of logBuffer
 */
void
FileAnalyzer::errorLine_ln (ref2LogPrinter logBuffer, const char *strMess)
{
  logBuffer->errorLine_ln (this_lineCounter, strMess);
}

/*
 * print line counter, index in line and strMess in a new line of logBuffer
 */
void
FileAnalyzer::errorLineColumn_sp (ref2LogPrinter logBuffer, const char *strMess)
{
  logBuffer->line ();
  logBuffer->errorLineColumn_sp (this_lineCounter, this_lineAnalyzer->get_cursor () + 1, strMess);
}

/*
 * print line counter, index in line and strMess in a new line of logBuffer
 */
void
FileAnalyzer::errorLineColumn_ln (ref2LogPrinter logBuffer, const char *strMess)
{
  logBuffer->line ();
  errorLineColumn_sp (logBuffer, strMess);
  logBuffer->line ();
}

/*
 * print line counter, index in line and message in a new line of logBuffer
 */
void
FileAnalyzer::error_reading (ref2LogPrinter logBuffer, const char *mess)
{
  logBuffer->line ("in file ", this_fileName);
  logBuffer->errorLine_ln (this_lineCounter, mess);
}

/*
 * print line counter, index in line and message in a new line of logBuffer
 */
void
FileAnalyzer::error_reading (ref2LogPrinter logBuffer, const char *mess1, const char *mess2)
{
  logBuffer->line ("in file ", this_fileName);
  logBuffer->errorLine_sp (this_lineCounter, mess1);
  logBuffer->line (mess2);
}

/*
 * print line counter, index in line and message in a new line of logBuffer
 */
void
FileAnalyzer::error_readingUnexpected_sp (ref2LogPrinter logBuffer)
{
  logBuffer->line ("in file ", this_fileName);
  if (isFileReadingCompleted ())
    logBuffer->error__sp ("unexpected end of file in");
  else if (this_lineAnalyzer->is_lineCompleted ())
    logBuffer->errorLine_sp (this_lineCounter, (char *) "unexpected end of line in");
  else
    errorLineColumn_sp (logBuffer, (char *) "unexpected char in");
}

/*
 * print line counter, index in line and message in a new line of logBuffer
 */
void
FileAnalyzer::error_readingUnexpected (ref2LogPrinter logBuffer, const char *mess)
{
  error_readingUnexpected_sp (logBuffer);
  logBuffer->line (mess);
}

/*
 * print line counter, index in line and message in a new line of logBuffer
 */
void
FileAnalyzer::error_readingUnexpected (ref2LogPrinter logBuffer, const char *mess1, const char *mess2)
{
  error_readingUnexpected_sp (logBuffer);
  logBuffer->line (mess1, mess2);
}

/* closes current open file
 * has no effect if no file is currently open
 */
void
FileAnalyzer::closeFileReading ()
{
  if (fileReading_OK)
    if (!fileReading_Close)
      {
	freezeFileReading ();
	rwFClose (this_file);
      }
}

/* associates a buffer to an open file and load a first line
 * success: returns pk_true
 * failure: assign an error message to mess
 *          this remains unchanged
 *          returns false
 */
boolean
FileAnalyzer::openFileReading (FILE * aFile, const char *name, const char **mess)
{
  if (fileReading_Close)
    return attachFile (name, aFile);
  *mess = "a file is already open in this buffer";
  return pk_false;
}

/* open file this_fileName 
 * associates a buffer to file
 * success: returns pk_true
 * failure: assign an error message to mess
 *          this remains unchanged
 *          returns false
 */
boolean
FileAnalyzer::openFileReading (const char *name, const char **mess)
{
  FILE *readFile;
  readFile = rFOpen (name, mess);
  if (readFile == NIL_File)
    return pk_false;
  return openFileReading (rFOpen (name, mess), name, mess);
}

/**
  Allocates a temporary string (a string register) filled with NULL_char
  sets this in frozen mode (open or openFileReading is required before reading)
  @param ?lSize required space for the string container (NULL_char excluded)
  @param ?aSize maximal atom size (NULL_char excluded)
  @param ?aLineAnalyzer is provided to analyze each line
*/
FileAnalyzer::FileAnalyzer (int lSize, ref2LineAnalyzer aLineAnalyzer)
{
  init (lSize, aLineAnalyzer);
}

FileAnalyzer::~FileAnalyzer ()
{
  if (this_lineAnalyzer != NIL_LineAnalyzer)
    this_lineAnalyzer->lineClose ();
  if (this_fileBufferSize > 0)
    this_fileRegister = deletePKString (this_fileRegister);
}
