/* SYNTAX used in LineAnalyzer_xx comments
 *
 * [Character] is the set of all characters but [NULL_char]
 * [Character*] ::= {[Character]}*
 * [Character+] ::= [Character] [Character*]
 * [String] ::= [Character*] [NULL_char]
 *
 * [White] ::= ' ' '\n' '\v' '\f' '\t',  the set of ANSI C white spaces 
 * [Character_but_White] is a [Character] that does not belong to [White]
 * [White*] ::=  {[White]}* 
 * [White+] ::=  [White] [White*]
 * [WhiteString] ::= [White*] [NULL_char]
 * [not_WhiteString] ::= [Character_but_White] [String]
 * 
 */

/*
 * a string is terminated by NULL_char
 * a line is a string terminated by end of line mark
 * a line tagged operations consider that end of line mark 
 *        in last position in a string does not belong to the line
 *
 */

#define WHITESPACE " \n\v\f\t"

/*
 * ==:: if thisLineActualLength >= 0 then 0 <= thisLineCursor <= thisLineActualLength
 * ==:: if thisLineActualLength < 0 then thisLineActualLength == thisLineCursor
 *
 * a CLOSE string must be open before using it safely
 * ==> if string is CLOSE:             thisLineActualLength < 0, thisLineActualLength <= thisLineCursor

 * a NOTEMPTY string is a not empty string 
 * ==> if string is NOTEMPTY:                     0 < thisLineActualLength, 0 <= thisLineCursor <= thisLineActualLength
 * a NOTCOMPLETED string is open and has not reached its NULL_char string termination
 * ==> if string is NOTCOMPLETED:                 0 <= thisLineCursor < thisLineActualLength
 * a NOTLASTPOSITION string is a not empty string which end is not reached
 * ==> if string is NOTLASTPOSITION:                  0 <= thisLineCursor < thisLineActualLength-1
 */
#define isNOTEMPTY() ( (thisLineActualLength > 0) )
#define isCOMPLETED() ( (thisLineCursor >= thisLineActualLength) )
#define isNOTCOMPLETED() ( (thisLineCursor < thisLineActualLength) )
#define isNOTLASTPOSITION() ( ( thisLineCursor < thisLineActualLength-1 ) )

#define isSTARTPOS() ( ( isNOTEMPTY() && ( thisLineCursor == 0 ) ) )
#define isSTART(c) ( ( isNOTEMPTY() && ( thisLineCursor == 0 ) && ( thisLineBuffer[0] == c ) ) )
#define isCURRENT(c) ( ( isNOTCOMPLETED() ) && ( thisLineBuffer[thisLineCursor] == c ) )
#define isFORWARD(c) ( ( isNOTLASTPOSITION() ) && ( thisLineBuffer[thisLineCursor+1] == c ) )
/* the following are unsafe 
 */
#define unsafe_getCURRENTCHAR() ( thisLineBuffer[thisLineCursor] )
#define unsafe_isCURRENTCHAR(c) ( ( unsafe_getCURRENTCHAR() == c ) )
#define unsafe_getFORWARDCHAR() ( thisLineBuffer[thisLineCursor+1] )

#define unsafe_getTAIL() ( ((char *)(thisLineBuffer+thisLineCursor)) )
#define IS_WHITE() ( ( strspn(unsafe_getTAIL(), WHITESPACE) > 0 ) )
#define SKIP_WHITEsTAR() ( thisLineCursor = thisLineCursor+strspn(unsafe_getTAIL(), WHITESPACE) )
#define STRCONTAINSCHR(str,chr) ( strchr(str, unsafe_getCURRENTCHAR()) != NIL_String )
