_PROTECTED_
/** =============================================================================
 * line functions
 */
/** PLG 2004-10-25
 * this atom remains unchanged
 * thisLineBuffer becomes NIL_String
 * thisLineCursor and thisLineActualLength are set to -1
 */
  void
LineAnalyzer::set_line ()
{
  thisLineBuffer = NIL_String;
  thisLineActualLength = -1;
  thisLineCursor = -1;
}

/* PLG 2005-01-03
 * this atom remains unchanged
 * this buffer value becomes <aStr> (WARNING NOT A COPY)
 * this length becomes <aStr> length
 * this cursor position becomes 0
 * attributes:
 *      thisLineBuffer is set equal to <aStr>
 *      thisLineActualLength becomes thisLineBuffer actual length
 *      thisLineCursor is set to 0
 */
void
LineAnalyzer::set_line (char *aStr)	/* cannot be const */
{
  thisLineBuffer = aStr;
  if (aStr == NIL_String)
    thisLineActualLength = 0;
  else
    thisLineActualLength = strlen (thisLineBuffer);
  thisLineCursor = 0;
}

/* PLG 2005-01-03
 * this atom remains unchanged provided that it does not share common space with this buffer
 * removes terminating [eol] character in this buffer
 * this buffer is [NULL_char] 
 *      this remains unchanged
 * this buffer is [Character+] [NULL_char] and  last character in [Character+] is not [eol]
 *      this remains unchanged
 * this buffer is [Character*] [eol] [NULL_char]
 *      this buffer becomes [Character*] [NULL_char]
 */
void
LineAnalyzer::remove_lineLastEol ()
{
  if ((thisLineActualLength > 0) && thisLineBuffer[thisLineActualLength - 1] == '\n')
    {
      thisLineActualLength = thisLineActualLength - 1;
      thisLineBuffer[thisLineActualLength] = NULL_char;
    }
}

/* =============================================================================
 * atom operations
 */

/* PLG 2005-01-04
 * this buffer remains unchanged provided that it does not share common space with this atom register container
 * reset atom attributes to fit this atom register
 * success: thisAtomBuffer becomes thisAtomRegister container
 *          thisAtomLength becomes 0
 * failure: thisAtomBuffer becomes NIL_String
 * thisAtomSize is set according to thisAtomBuffer
 */
boolean
LineAnalyzer::reset_atom ()
{
  if (getPKStringContainer (thisAtomRegister, &thisAtomBuffer))
    {
      thisAtomSize = strregGetSize (thisAtomRegister);
      empty_atom ();
      return pk_true;
    }
  thisAtomBuffer = NIL_String;
  thisAtomSize = 0;
  return pk_false;
}

/** PLG 2004-10-25
 * this buffer remains unchanged
 * reset atom attributes to fit this atom register
 * this atom buffer becomes the unique  null container
 * this atom content becomes [NULL_char]
 * this atom length becomes 0
*/
void
LineAnalyzer::set_atomNull ()
{
  if (!new_sizedPKString (0, &thisAtomRegister))
    thisAtomRegister = NIL_LGString;
  thisAtomSize = 0;
  thisAtomBuffer = NIL_String;
  thisAtomLength = 0;
}

/* PLG 2004-10-25
 * this buffer remains unchanged
 * <size> <= 0 
 *      this atom buffer becomes the unique null container
 * <size> > 0 
 *      this atom buffer becomes a new PK'allocated container which size is at least <size>
 * this atom content becomes [NULL_char]
 * this atom length becomes 0
 */
void
LineAnalyzer::set_atomNew (int size)
{
  if (new_sizedPKStringContainer (size, &thisAtomRegister, &thisAtomBuffer))
    thisAtomSize = strregGetSize (thisAtomRegister);
  else
    {
      thisAtomRegister = NIL_LGString;
      thisAtomBuffer = NIL_String;
      thisAtomSize = 0;
    }
  thisAtomLength = 0;
}

/* PLG 2005-01-04
 * success: length > 0
 *             if this atom size is smaller than <length>, deletes this atom register 
 *                creates a new one large enough to hold <length> not NULL characters
 *             a copy of min(<length>, <source> length) first not NULL characters of <source> is loaded in this atom 
 *             this atom length is set to the actual length of this atom buffer 
 *          length <=0
 *             this atom is emptied if its size is > 0
 *             this atom length is set to 0
 * failure: a memory error occurred
 *          thisAtomRegister becomes NIL_LGString
 *          thisAtomBuffer becomes NIL_String
 */
boolean
LineAnalyzer::reset_atom (char *source, int length)
{
  if (length > 0)
    {
      if (thisAtomSize < length)
	{
	  if (thisAtomSize > 0)
	    thisAtomRegister = deletePKString (thisAtomRegister);
	  if (new_sizedPKStringContainer (length, &thisAtomRegister, &thisAtomBuffer))
	    {
	      thisAtomSize = strregGetSize (thisAtomRegister);
	      thisAtomLength = 0;
	    }
	  else
	    {
	      thisAtomRegister = NIL_LGString;
	      thisAtomBuffer = NIL_String;
	      thisAtomSize = 0;
	      thisAtomLength = 0;
	      return pk_false;
	    }
	}
      else if (!reset_atom ())
	{
	  thisAtomLength = 0;
	  return pk_false;
	}
      if (source != NIL_String)
	{
	  thisAtomLength = strlen (source);
	  thisAtomBuffer = strncpy (thisAtomBuffer, source, length);
	  if (length <= thisAtomLength)
	    {
	      thisAtomBuffer[length] = NULL_char;
	      thisAtomLength = length;
	    }
	}
    }
  else
    empty_atom ();		/* empties atom container; thisAtomLength is set to 0 */
  return pk_true;
}
