/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*	m_base.h							     */
/*									     */
/*  Ce module contient les fonctions de DEBUG     			     */
/*									     */
/*****************************************************************************/

#ifndef __M_BASE__
#define __M_BASE__


/*  */
#include <stdio.h>
#include <string.h>
#include "varglobal.h"
#include "stringmemStr_type.h"
#include "stringmemRegister_type.h"

/* macros de creation des objets de la zone des arbres */

/* ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== */
/* ===== ===== ===== ===== ===== THE FOLLOWING MACRO MAY RETURN NIL ===== ===== ===== ===== ===== ===== */

/* macros d'acces et de modification des champs d'un node*/
/* e est un pointeur sur node*/

/* ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== WARNING ===== */
/* ===== ===== ===== ===== ===== MOST OF THE FOLLOWING MACROS ARE NOT SAFE ===== ===== ===== ===== ===== */


/* macros d'acces et de modification des champs d'un repere */
/* r est un pointeur sur repere */


static void DEBUG_INDENT(int indentVal)
{
  int i;
  for (i=0; i<indentVal; i++)
    printf(" ");
}

static void DEBUG_INDENTMESS_LN(int indentVal, const char * mess) 
{ 
  DEBUG_INDENT(indentVal);
  printf("%s\n",mess); fflush(stdout);
}

static void DEBUG_PRINTADRINTREE(untypedValueType aStruct)
{
  long int nodeMemoryPos;
  if ( aStruct == NIL_untypedValueType )
    printf("-");
  else if ( TREEMEMORY_ISTREECOMPONENTADR(aStruct) ) {
    nodeMemoryPos = TREEMEMORY_GETTREECOMPONENTPOS(aStruct);
    printf("%li", nodeMemoryPos); }
  else
    printf("#");
}

static boolean DEBUG_NODE(int indentVal, int depth, ptnode aNode);
static boolean DEBUG_ATTRIBUTE(int indentVal, int depth, ptattribute anAttribute);

static void DEBUG_OLINE(int indentVal)
{
  DEBUG_INDENT(indentVal);
  printf("======================================================================\n");
}

static void DEBUG_CLINE(int indentVal)
{
  DEBUG_INDENT(indentVal);
  printf("----------------------------------------------------------------------\n");
}

static void DEBUG_stringRegister(ptStringRegister aStringRegister)
{
  printf("DEBUG register buffer address is %li\n", (long)stringRegister_GET_BUF(aStringRegister)); fflush(stdout);
  printf("DEBUG register start index is %i\n", stringRegister_GET_STARTINDEX(aStringRegister)); fflush(stdout);
  printf("DEBUG register used size is %i\n", stringRegister_GET_USEDSIZE(aStringRegister)); fflush(stdout);
  printf("DEBUG register string value is <<%s>>\n", stringRegister_GET_CSTRINGSTART(aStringRegister)); fflush(stdout);
}


static void DEBUG_regIndex(const char* mess, ptreg rIndex)
{
  printf("\n------------------- \nDEBUG %s, register index is %i\n", mess, rIndex); fflush(stdout);
  if ( rIndex >= 0 ) {
    DEBUG_stringRegister(indexRegisterCAST_stringRegister(rIndex)); }
  printf("\n------------------- end DEBUG %s\n", mess); fflush(stdout);
}


/* debug operator and its value
   @return false if value is not a tre 
 */
static boolean DEBUG_OPER_VALUE(int oper, untypedValueType value)
{
  printf("OP:%i", USEROPER(oper));
  if ( oper <= TCOMMUNVALEUR ) {
    switch (oper) 
      {
      case OPUNDEF: printf("(UNDEF), value:-"); return false;
      case OPTROU: printf("(TROU), value:-"); return false;
      case OPNIL: printf("(NIL), value:-"); return false;
      case OPENTIER: printf("(ENTIER), value:%li", (long int)value); return false;
      case OPCAR: printf("(CAR), value:%c", (char)(long int)value); return false;
      case OPCHAINE: {
	int lg, i;
	char c;
	TMStringCharAdr chaine;
	char * adchaine;
	printf("(CHAINE), value:%c",'"');
	chaine = TMSTRING_CAST_TMSTRINGCHARADR((TMString)value);
	lg = TMSTRINGCHARADR_GET_LENGTH(chaine);
	adchaine = TMSTRINGCHARADR_GET_CSTRING(chaine) ;
	for (i = 1; i < lg + 1 ; i++) {
	  c = *adchaine;
	  printf("%c", c);
	  adchaine = adchaine + 1; }
	printf("%c",'"');
	return false; }
      case OPSTAR: printf("(STAR), value:-"); return false;
      case OPREPERE: printf("(REPERE), value:%li", (long int)value); return false;
      case OPARBRE: printf("(ARBRE), value:"); DEBUG_PRINTADRINTREE(value); return true;
      default: printf("## INTERNAL ERROR ##"); return false;
      } }
  else 
    printf(", value:");
  DEBUG_PRINTADRINTREE(value);
  return true;
}

/** print an any node descriptor common part
   @param aDescritor
   @param isTree is assigned true if the value is a tree, false otherwise, when the function returns true
   @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static boolean DEBUG_SONCOMMONCONTENT(ptfilsd aDescriptor, int * pos, boolean * isTree)
{
  long int memoryPos;
  untypedValueType nodeValue;
  int oper;
  if ( aDescriptor == NIL_filsd ) {
    printf("-]\n"); fflush(stdout);
    return false; }
  if ( ! TREEMEMORY_ISTREECOMPONENTADR (aDescriptor) ) {
    printf("#:%li]\n", (long int)aDescriptor); fflush(stdout);
    return false; }
  memoryPos = TREEMEMORY_GETTREECOMPONENTPOS(aDescriptor);
  (*pos) = (int)(ANYFILS_GET_POS(aDescriptor));
  nodeValue = ANYFILS_GET_UNTYPED_VALUE(aDescriptor);
  oper = (int)ANYFILS_GET_INNEROPER(aDescriptor);
  printf("%li, PHY:%i, POS:%i, ", memoryPos, (int)ANYFILS_GET_PHYL(aDescriptor), *pos);
  (*isTree) = DEBUG_OPER_VALUE(oper, nodeValue);
  printf(" UP:");
  return true;
}

/*
  @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static boolean DEBUG_SONINLISTNODE(int indentVal, int depth, ptfilslist aDescriptor)
{
  ptnode container;
  int pos;
  boolean isTree;
  DEBUG_INDENT(indentVal);
  printf("[LIST: ");
  if ( DEBUG_SONCOMMONCONTENT((ptfilsd)aDescriptor, &pos, &isTree)  ) {
    container = (FILSLIST_GET_NODE_CONTAINER(aDescriptor));
    DEBUG_PRINTADRINTREE((untypedValueType)container); 
    printf(", LE:");
    DEBUG_PRINTADRINTREE((untypedValueType)FILSLIST_GET_PRED(aDescriptor)); 
    printf(", RI:");
    DEBUG_PRINTADRINTREE((untypedValueType)FILSLIST_GET_SUCC(aDescriptor)); 
    printf(" ]\n"); fflush(stdout);
    
    if ( (depth > 0) && isTree ) {
      DEBUG_NODE(indentVal+2, depth, ANYFILS_GET_NODE_VALUE(aDescriptor)); }
    return true; }
  return false;
}

/*
  @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static boolean DEBUGMESS_SONINLISTNODE(int indentVal, int depth, const char * mess, ptfilslist aDescriptor)
{
  DEBUG_INDENTMESS_LN(indentVal, mess);
  return DEBUG_SONINLISTNODE(indentVal, depth, aDescriptor);
}

/*
  @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static void DEBUGLIGHTMESS_SONINLISTNODE(const char * mess, ptfilslist aDescriptor)
{
  printf(mess);
  (void)DEBUG_SONINLISTNODE(0,1, aDescriptor);
  printf("\n"); fflush(stdout);
}

/* print a fix size node descriptor
   @param indent is yhe left indentation
   @param depth: if depth is positive, recursively print a Descritor subtree
   @param aDescritor
   @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static boolean DEBUG_SONINFIXSIZENODE(int indentVal, int depth, ptfilsd aDescriptor, boolean gotoLine)
{
  ptnode container;
  int pos;
  boolean isTree;
  DEBUG_INDENT(indentVal);
  printf("[FILSD: ");
  if ( DEBUG_SONCOMMONCONTENT(aDescriptor, &pos, &isTree)  ) {
    if ( pos == 0 || pos == 1 )
      container = (ptnode)(aDescriptor) - 1;
    else if ( pos > 0 )
      container = (ptnode)(aDescriptor - (pos -1)) - 1;
    else /* SHOULD NOT OCCUR */
      container = FILSLIST_GET_NODE_CONTAINER((ptfilslist)aDescriptor);
    DEBUG_PRINTADRINTREE((untypedValueType)container);
    if ( pos == 0 )
      printf(", TREE ROOT ");
    printf("]");
    if ( gotoLine || (depth > 0))
      printf("\n"); fflush(stdout);
    if ( (depth > 0) && isTree ) {
      if ( pos < 0 ) {
	(void)DEBUG_SONINLISTNODE(indentVal+2, depth, ANYFILS_GET_FILSLIST_VALUE(aDescriptor)); }
      else {
	(void)DEBUG_NODE(indentVal+2, depth, ANYFILS_GET_NODE_VALUE(aDescriptor)); } }
    return true; }
  return false;
}

/* print a fix size node descriptor
   @param indent is yhe left indentation
   @param depth: if depth is positive, recursively print a Descritor subtree
   @param mess: initial message printed in left margin
   @param aDescritor
   @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static boolean DEBUGMESS_SONINFIXSIZENODE(int indentVal, int depth, const char * mess, ptfilsd aDescriptor)
{
  DEBUG_INDENTMESS_LN(indentVal, mess);
  return DEBUG_SONINFIXSIZENODE(indentVal, depth, aDescriptor, true);
}

/*
  @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static boolean DEBUGMESS_SONINANYNODE(int indentVal, int depth, const char * mess, ptfilsAny aDescriptor)
{
  if ( aDescriptor == NIL_filsAny ) {
    DEBUG_INDENTMESS_LN(indentVal, mess);
    printf("[???:NIL]\n"); fflush(stdout);
    return false; }
  if ( ANYFILS_ISIN_LIST(aDescriptor) )
    return DEBUGMESS_SONINLISTNODE(indentVal, depth, mess, (ptfilslist)aDescriptor);
  else
    return DEBUGMESS_SONINFIXSIZENODE(indentVal, depth, mess, (ptfilsd)aDescriptor);
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
static boolean DEBUG_REPERE(int indentVal, int depth, ptrepere aRepere)
{
  long int nodeMemoryPos;
  long int sonDescriptor;
  DEBUG_INDENT(indentVal);
  printf("[REPERE: ");
  if ( aRepere == NIL_repere ) {
    printf("- ]\n"); fflush(stdout);
    return false; }
  if ( ! TREEMEMORY_ISTREECOMPONENTADR(aRepere) ) {
    printf("#:%li ]\n", (long int)aRepere); fflush(stdout);
    return false; }
  nodeMemoryPos = TREEMEMORY_GETTREECOMPONENTPOS(aRepere);
  if ( REPERE_IS_DIRECT_ACCESS(aRepere) ) {
    sonDescriptor = (long int)REPERE_GET_FILSD_REFERED(aRepere);
    printf("%li, langage:%i, FD-->:", nodeMemoryPos, REPERE_GET_TMLANG(aRepere)); }
  else {
    sonDescriptor = (long int)REPERE_GET_FILSLIST_REFERED(aRepere);
    printf("%li, langage:%i, LD-->:", nodeMemoryPos, REPERE_GET_TMLANG(aRepere)); }
  DEBUG_PRINTADRINTREE((untypedValueType)sonDescriptor);
  printf(" ]\n"); fflush(stdout);
  return true;
}

/*
  @return true if aDescriptor is neither NIL nor an incorrect adress
*/
static void DEBUGLIGHTMESS_REPERE(const char * mess, ptrepere aRepere)
{
  printf(mess);
  (void)DEBUG_REPERE(0,1, aRepere);
  printf("\n"); fflush(stdout);
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
static boolean DEBUG_LISTSUBTREE(int indentVal, int depth, ptnode aNode)
{
  ptfilslist currentSonDescriptor;
  currentSonDescriptor = NODE_GET_FILSLIST_FIRST(aNode);
  while ( currentSonDescriptor != NIL_filslist ) {
    if ( DEBUG_SONINLISTNODE(indentVal, depth, currentSonDescriptor) )
      currentSonDescriptor = FILSLIST_GET_SUCC(currentSonDescriptor); 
    else 
      return false; }
  return true;
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
static boolean DEBUG_FIXSIZESUBTREE(int indentVal, int depth, ptnode aNode, int arity)
{
  ptfilsd currentSonDescriptor;
  int i;
  currentSonDescriptor = NODE_GET_FILSD_FIRST(aNode);
  for (i=0; i < arity; i++) {
    if ( DEBUG_SONINFIXSIZENODE(indentVal, depth, currentSonDescriptor, true) )
      currentSonDescriptor = FILSD_GET_SUCC(currentSonDescriptor);
    else
      return false; }
  return true;
}

/*
  @return true if anAttribute is neither NIL nor an incorrect adress
*/
static boolean DEBUG_LISTATTRIBUTE(int indentVal, int depth, ptattribute anAttribute)
{
  while ( anAttribute != NIL_attribute ) {
    if ( DEBUG_ATTRIBUTE(indentVal, depth, anAttribute) )
      anAttribute = ATTRIBUTE_GET_SUCC(anAttribute);
    else return false; }
  return true;
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
boolean DEBUG_NODE(int indentVal, int depth, ptnode aNode)
{
  long int nodeMemoryPos, attrPos;
  ptattribute anAttribute;
  long int fatherDescriptor;
  int arity;
  DEBUG_INDENT(indentVal);
  printf("[NODE: ");
  if ( aNode == NIL_node ) {
    printf("- ]\n"); fflush(stdout);
    return false; }
  if ( ! TREEMEMORY_ISTREECOMPONENTADR(aNode) ) {
    printf("#:%li ]\n", (long int)aNode); fflush(stdout);
   return false; }
  nodeMemoryPos = TREEMEMORY_GETTREECOMPONENTPOS(aNode);
  arity =  NODE_GET_ARITY(aNode);
  printf("%li, arity:%i, ", nodeMemoryPos, arity);
  if ( NODE_ISIN_FIXESIZE(aNode) ) {
    fatherDescriptor = (long int) NODE_GET_FILSD_DESCRIPTOR(aNode);
    printf("<--FD:"); }
  else {
    fatherDescriptor = (long int)NODE_GET_FILSLIST_DESCRIPTOR(aNode);
    printf("<--LD:"); }
  DEBUG_PRINTADRINTREE((untypedValueType)fatherDescriptor);
  printf("; ");
  if ( NODE_IS_LIST(aNode) )
    (void)DEBUG_SONINFIXSIZENODE(1, 0, NODE_GET_FILSD_FIRST(aNode), false);
  anAttribute = NODE_GET_ATTRIBUTE(aNode);
  attrPos = -1;
  printf("; ATT:");
  if ( anAttribute == NIL_attribute )
    printf("-");
  else if ( TREEMEMORY_ISTREECOMPONENTADR(anAttribute) ) {
    attrPos = TREEMEMORY_GETTREECOMPONENTPOS(anAttribute);
    printf("%li", attrPos ); }
  else {
    printf("#:%li", (long int)anAttribute );
    anAttribute = NIL_attribute; }
  printf(" ]\n"); fflush(stdout);

  if (depth >0) {
    if ( NODE_IS_LIST(aNode) )
      DEBUG_LISTSUBTREE(indentVal+5, depth-1, aNode);
    else
      DEBUG_FIXSIZESUBTREE(indentVal+5, depth-1, aNode, arity);
    if ( anAttribute != NIL_attribute ){
      DEBUG_LISTATTRIBUTE(indentVal+5, depth-1, anAttribute); } }
  return true;
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
static boolean DEBUGMESS_NODE(int indentVal, int depth, const char * mess, ptnode aNode)
{
  boolean res;
  DEBUG_OLINE(indentVal);
  DEBUG_INDENTMESS_LN(indentVal, mess);
  res = DEBUG_NODE(indentVal, depth, aNode);
  DEBUG_CLINE(indentVal);
  return res;
}

/*
*/
static void DEBUGLIGHTMESS_NODE(const char * mess, ptnode aNode)
{
  printf(mess);
  (void)DEBUG_NODE(0, 1, aNode);
  printf("\n"); fflush(stdout);
}

/*
   @param isTree is assigned true if the value is a tree, false otherwise
  @return true if anAttribute is neither NIL nor an incorrect adress
*/
static boolean DEBUG_ATTRIBUTE(int indentVal, int depth, ptattribute anAttribute)
{
  long int attributeMemoryPos;
  untypedValueType attrValue;
  ptnode ownerNode;
  int category, oper;
  boolean isTree;
  DEBUG_INDENT(indentVal);
  printf("[ATTR: ");
  if ( anAttribute == NIL_attribute ) {
    printf("- ]\n"); fflush(stdout);
    return false; }
  if ( ! TREEMEMORY_ISTREECOMPONENTADR(anAttribute) ) {
    printf("#:%li ]\n", (long int)anAttribute); fflush(stdout);
    return false; }
  attributeMemoryPos = TREEMEMORY_GETTREECOMPONENTPOS(anAttribute);
  category =  ATTRIBUTE_GET_TAG(anAttribute);
  oper =  ATTRIBUTE_GET_INNEROPER(anAttribute);
  ownerNode =  ATTRIBUTE_GET_NODE_OWNER(anAttribute);
  attrValue = ATTRIBUTE_GET_UNTYPED_VALUE(anAttribute);
  printf("%li, category:%i, ", attributeMemoryPos, category);
  isTree = DEBUG_OPER_VALUE(oper, attrValue);
  printf(", owner:");
  DEBUG_PRINTADRINTREE((untypedValueType)ownerNode); 
  printf(", LE:");
  DEBUG_PRINTADRINTREE((untypedValueType)(ATTRIBUTE_GET_PRED(anAttribute))); 
  printf(", RI:");
  DEBUG_PRINTADRINTREE((untypedValueType)(ATTRIBUTE_GET_SUCC(anAttribute))); 
  printf(" ]\n"); fflush(stdout);
  if ( (depth >0) && isTree )
    return DEBUG_NODE( indentVal+2, depth, FILSD_GET_NODE_CONTAINER_OF_FIRST(ATTRIBUTE_GET_FILSD_VALUE(anAttribute)) );
  return true;
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
static boolean DEBUGMESS_ATTRIBUTE(int indentVal, int depth, const char * mess, ptattribute anAttribute)
{
  boolean res;
  DEBUG_OLINE(indentVal);
  DEBUG_INDENTMESS_LN(indentVal, mess);
  res = DEBUG_ATTRIBUTE(indentVal, depth, anAttribute);
  DEBUG_CLINE(indentVal);
  return res;
}

/*
*/
static void DEBUGLIGHTMESS_ATTRIBUTE(const char * mess, ptattribute anAttribute)
{
  printf(mess);
  (void)DEBUG_ATTRIBUTE(0, 1, anAttribute);
  printf("\n"); fflush(stdout);
}

/*
  @return true if aNode is neither NIL nor an incorrect adress
*/
static boolean DEBUGMESS_NODEATTRIBUTELIST(int indentVal, int depth, const char * mess, ptnode aNode)
{
  boolean res;
  DEBUG_OLINE(indentVal);
  DEBUG_INDENTMESS_LN(indentVal, mess);
  res = DEBUG_LISTATTRIBUTE(indentVal, depth, NODE_GET_ATTRIBUTE(aNode));
  DEBUG_CLINE(indentVal);
  return res;
}

#endif
