/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __CONSTMEM__
#define __CONSTMEM__

/* ===========================================================================
   Initial values for the memory of the Tree Machine 
   see also Memory parameter declarations in gm_init.cpp
            set_memorySizeParameters in gm_init.cpp
   -------------------------------------------------
*/
// value of stringRegister_number: the maximal number of string buffers
#define STRINGREGISTER_NUMBER 20
// initial value of TreeStack_commonArea_BYTESIZE
// is changed to the value of the host variable "T_MEM_ARB" when it is set
#define TREESTACK_COMMONAREA_BYTESIZE  LONG_MAX/4
// initial value of TMStringCollision_commonArea_BYTESIZE
#define TMSTRINGCOLLISION_COMMONAREA_BYTESIZE  LONG_MAX/4
// value of LTACode_area_BYTESIZE
#define LTACODE_AREA_BYTESIZE 2048
// value of stringTable_entryNumber: prime number for hashing
#define TMSTRINGTABLE_ENTRYNUMBER  1021
// value of languageTable_entryNumber
#define LANGUAGETABLE_ENTRYNUMBER 1

#define TREESTACKSIZE_HOSTVARIABLE "T_MEM_ARB"
#define TMSTRINGCOLLISIONSIZE_HOSTVARIABLE "T_MEM_CHAIN"


// #define MAXLGCHAINE 256 moved, LBesnard Dec 2010, to 4096 (max size of the name of a Linux  file)
#define MAXLGCHAINE 4096


/* code du type d'acces a un noeud dans le repere */
#define DIRECTREPERE 512
// NOT USED #define LISTREPERE 513

/* code du type d'acces a un noeud dans les arbres */
#define DIRECTNODE 514
#define LISTNODE 515


#define NIL 0

/** Pour la definition des types voir le fichier typesbase.h */

/* definition des operateurs communs LTA et des types de base */
#define TMAX_phyl 511 /* to let free 512 used by DIRECTREPERE */
#define TMAX_op 2048
#define TMAX_codes 2048
#define TMAX_attribute 4095 /* represented by 12 bits in a short integer */

#define NBPHLTA 12    /* nombre de phylums LTA   */
#define NBOPLTA 30    /* nombre d'operateurs LTA */
#define NBTYPLTA 24   /* nombre de type  LTA     */

#define TCOMMUN 9   /* de 0 a 9 */
#define TCOMMUNVALEUR 7 /* commun "porteur" de valeur , de 0 a 7 */

#define OPUNDEF -1
#define OPTROU 0
#define OPNIL 1
#define OPENTIER 2
#define OPCAR 3
#define OPCHAINE 4
#define OPSTAR 5
#define OPREPERE 6
#define OPARBRE 7  /*codifie le type tree lta */
#define OPLISTE 8
#define OPFIXE 9  /* codifie l'operateur fixe */
#define OPIDENT 10
#define OPBINAIRE 11
#define OPDEF 12
#define OPAPPEL 13
#define OPFONCTION 14
#define OPMODULE 15
#define OPLANGAGE 16
#define OPATTRIBUT 17
#define OPIDELEMLANG 18
#define OPSEQ 19
#define OPCONS 20
#define OPBOUCLE 21
#define OPSORTIE 22
#define OPCAS 23
#define OPRENDRE 24
#define OPIDLANG 25
#define OPACCESATT 26
#define OPATTRIBUER 27
#define OPRETIRERATT 28
#define OPEXPCAS 29

#define OPBIDON 50
#define CODEVERY 0

#define TYPE 0   /* code de l'attribut type */
#define LIEN 1   /* code de l'attribut lien */
#define LANG 2   /* code de l'attribut nom du langage courant */
#define BRA  3   /* code de l'attribut branchement inconditionnel */
#define BRC  4   /* code de l'attribut branchement conditionnel */
#define LIENPRECOMP 5 /* code de l'attribut associe au ctrl a la precomp */
#define LIENLISTE 6 /*necessaire pou gerer une liste de reperes a la precomp*/

#define FBAS  0
#define FHAUT 1
#define FBRA  2
#define FBRC  3
#define POINT 4

#define TID   0
#define PI    1
#define SEG   2

#endif
