/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*********************************************************************************

  Copyright INRIA/IRISA Rennes. January 5th 1995

  Module:   error.h.... The error management file.
  Author:   Loic Besnard, David Joffrin
  Notes: Definition of the base class errorClass.
         Definition of the predefined errors (error, bug, fatal, msg, warn).
         Definition of manipulating macros :
           InitManager(), ClearManager(), ClearErrLog(),
           ErrorFound(), FlushError(), FATALEND(int).

 * Revision yy/mm/dd  hh:mm:  name
 * *** What done ... ***
 *

*********************************************************************************/

#ifndef __ERROR_H__
#define __ERROR_H__

#include <stdio.h>
#include "StdTypeDefs.h"

#define MAXERRORNB 20

typedef int t_errortype;
/*The following enumeration describes also the list of all the class identifiers for the class errorCalls. These class identifiers have all the type /I{t_errortype} defined in this header file.*/
enum {
  et_fatal /*is used to manage fatal errors. A fatal error stop the execution of a program, and is activated when the users does something forbidden.*/,
  et_error /*is used to manage errors.*/,
  et_warn /*is used to manage warnings.*/,
  et_bug /*is used to manage bugs. A bug is something wrong in the program.*/,
  et_msg /*is used to manage messages.*/
};

class errorClass
{
  public:
    errorClass(t_errortype et);

    t_errortype GetErrorType();

    void SetPosition(char* aFile, int line, int  col);
    char* GetFilename();
    int  GetLine();
    int  GetColumn();

    void Clean();

    virtual errorClass& operator << (const char* str);
    virtual errorClass& operator << (const char c);
    virtual errorClass& operator << (const int i);
    virtual errorClass& operator << (const unsigned int i);
    virtual errorClass& operator << (const short i);

    virtual void operator ++ (int );

    void Flush();

  protected:
    void Display();

  private:
    char* GetString();
    t_errortype _errortype;
    char _string[1024];
    int  _line;
    int  _column;
    char* _filename;
};

/*Returns the identifier of the errorClass instance.*/
inline t_errortype errorClass::GetErrorType()
{
  return _errortype;
}

void InitManager();
void ClearManager();
void ClearErrLog();
void FlushError();
int  ErrorFound();


extern errorClass bug_tree;
extern errorClass fatal_tree;
extern errorClass error_tree;
extern errorClass warn_tree;
extern errorClass msg_tree;


void FATALEND(int code);

#endif // __ERROR_H__


