/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __TREEFRONT__
#define __TREEFRONT__

EXTERN void DEBUG_treeRoot(int indentVal, int depth, const char * mess, trees aSubtree);
EXTERN void DEBUG_treeHyperRoot(int indentVal, int depth, const char * mess, trees aSubtree);
EXTERN void DEBUG_subtree(int indentVal, int depth, const char * mess, trees aSubtree);
EXTERN void DEBUG_attribute(int indentVal, int depth, const char * mess, ptattribute anAttribute);
EXTERN void DEBUG_aSubreeAttributes(int indentVal, int depth, const char * mess, trees aSubtree);

EXTERN trees getTM_father(trees aSubtree);
EXTERN trees getTM_root(trees aSubtree);
EXTERN trees getTM_hyperRoot(trees aSubtree);
EXTERN trees getTM_rightNeighbor(int i, trees aSubtree);
EXTERN trees getTM_leftNeighbor(int i, trees aSubtree);
EXTERN trees getTM_nChild(int n, trees aSubtree);
EXTERN short getTM_arity(trees aSubtree);
EXTERN void setTM_operator(trees aSubtree, int op);
EXTERN short getTM_operator(trees aSubtree);

EXTERN trees makeTM_hole(void);
EXTERN boolean isTM_hole(trees aSubtree);

EXTERN trees makeTM_nilAtom(short cop);
EXTERN boolean isTM_nilAtom(trees aSubtree);

EXTERN trees makeTM_intAtom(short cop, long int integerValue);
EXTERN boolean isTM_intAtom(trees aSubtree);
EXTERN void setTM_intValue(trees aSubtree, long int integerValue);
EXTERN long int getTM_intValue(trees aSubtree);

EXTERN trees makeTM_charAtom(short cop, char charValue);
EXTERN boolean isTM_charAtom(trees aSubtree);
EXTERN void setTM_charAtomValue(trees aSubtree, char val);
EXTERN char getTM_charAtomValue(trees aSubtree);

EXTERN trees makeTM_TMStringAtom(short cop, TMString ich);
EXTERN void assignTM_CStringValue2NodeSon(trees aSubtree, char * val, int pos );
EXTERN boolean isTM_TMStringAtom(trees aSubtree);
EXTERN void setTM_CStringAtomValue(trees aSubtree, char * val);
EXTERN boolean isTM_TMStringNodeSon(trees aSubtree, int pos);
EXTERN TMString getTM_TMStringAtomValue(trees aSubtree);
EXTERN TMString getTM_TMStringValueInNodeSon(trees aSubtree, int ind); 
EXTERN boolean hasTM_commonCategory(trees aSubtree);

EXTERN trees makeTM_unaryTree(short cop, trees assignedSubtree);
EXTERN trees makeTM_binaryTree(short cop, trees assignedSubtree1, trees assignedSubtree2);
EXTERN trees makeTM_ternaryTree(short cop, trees assignedSubtree1, trees assignedSubtree2, trees assignedSubtree3);
EXTERN trees makeTM_quaternaryTree(short cop, trees assignedSubtree1, trees assignedSubtree2, trees assignedSubtree3, trees assignedSubtree4);
EXTERN void resetTM_nchild(trees aTargetSubtree, trees treeValue, int n);

EXTERN trees makeTM_emptyListTree(short cop);
EXTERN trees makeTM_listTree(short cop, trees assignedSubtree);
EXTERN trees appendTM_lastInListTree(trees targetSubtreeList, trees insertedSubtree);
EXTERN trees insertTM_beforeSuccInListTree(trees insertedSubtree, trees nodeInList);
EXTERN trees insertTM_afterPredInListTree(trees insertedSubtree, trees nodeInList);
EXTERN trees insertTM_firstInListTree(trees insertedSubtree, trees targetSubtreeList);

EXTERN trees insertTM_afterIinListTree(int i, trees insertedSubtree, trees targetSubtreeList);
EXTERN trees insertTM_beforeIinListTree(int i, trees insertedSubtree, trees targetSubtreeList);
EXTERN boolean isTM_listSubtree(trees aSubtree);
EXTERN trees getTM_firstSubtreeInList(trees aListNode);
EXTERN short getTM_place(trees aSubtree);
EXTERN boolean isTM_emptyList(trees aSubtree);
EXTERN trees removeTM_inListSubtree(trees aSubtreeList, int i);

EXTERN void attachTM_newIntegerAttribute(trees aSubtree, short tagAttribute, longInteger valint);
EXTERN void attachTM_newTreeAttribute(trees aSubtree, short tagAttribute, trees treeValue);
EXTERN int getTM_numberOfAttributes(trees aSubtree);
EXTERN boolean hasTM_taggedAttribute(trees aSubtree, short tagAttribute);
EXTERN ptattribute getTM_firstAttribute(trees aSubtree);
EXTERN long int getTM_integerValueOfTaggedAttributeInSubtree(trees aSubtree, short tagAttribute);
EXTERN trees getTM_treeValueOfTaggedAttributeInSubtree(trees aSubtree, short tagAttribute);
EXTERN boolean removeTM_taggedAttribute(trees aSubtree, short tagAttribute);

EXTERN ptattribute getTM_succAttribute(ptattribute anAttribute);
EXTERN short getTM_tagAttribute(ptattribute anAttribute);
EXTERN boolean isTM_intAttribute(ptattribute anAttribute);
EXTERN long int getTM_intValueAttribute(ptattribute anAttribute);
EXTERN boolean isTM_treeAttribute(ptattribute anAttribute);
EXTERN trees getTM_treeValueAttribute(ptattribute anAttribute);
EXTERN trees getTM_treeAttributeOwner(trees aSubtree);

EXTERN trees copyTM_subtree(trees aSubtree, boolean copyAttributes);
EXTERN trees loadTM_tree(char * fileName);
EXTERN boolean saveTM_subTree(char * fileName, trees srcTree);

/** required initial action that allocates memory space
    the implicite parameter size of the memory can be change using the following environment variables
         "T_MEM_ARB" (total size in byte allocated to trees and stack)
         "T_MEM_CHAIN" (total size in byte allocated to strings)
 */
EXTERN void initTM_memory(void);
/** make static the tree memory used until its call
 */
EXTERN void resetTM_treeMemoryStart(void);
/** open a temporary area in tree memory
    openTM_temporaryTreeMemory and closeTM_temporaryTreeMemory must alternate; openTM_temporaryTreeMemory is the frist allowed action in this alternation
 */
EXTERN void openTM_temporaryTreeMemory(void);
/** close a temporary area in tree memory
    all subtrees built since previous openTM_temporaryTreeMemory occurrence are lost
 */
EXTERN void closeTM_temporaryTreeMemory(void);
/** clear the tree memory non atstic space (see resetTM_treeMemoryStart) and init other memory areas
*/
EXTERN void clearTM_userDataMemory(void) ;
/** @return the available space in tree memory
 */
EXTERN long int getTM_availableTreeSpace();
/** @return the total amount of non static tree memory used (including holes)
 */
EXTERN long int getTM_usedTreeSpace();
/** @return true if the total amount of available contiguous tree memory space 
    is smaller that the required space (minimalSpace)
 */
EXTERN boolean isNotTM_availableTreeSpace(int minimalSpace);
EXTERN void recup_mem(pttrees pt1, pttrees pt2, pttrees pt3, pttrees pt4, pttrees pt5);
EXTERN void printTreeMemoryState();

#endif
