/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __TYPEBAS__
#define __TYPEBAS__
/*****************************************************************************/
/*									     */
/*	typesbas.h							     */
/*									     */
/*  Ce module contient la definition des objets de la zone des arbres :      */ 
/*	- repere							     */
/*	- node							     */
/*	- filsd								     */
/*	- filslist							     */
/*	- attribut							     */
/*									     */
/*****************************************************************************/

// #include "basicTypes.h"
#include "constmem.h"
#include "typesmac.h"

typedef long int untypedValueType;
#define NIL_untypedValueType ((untypedValueType)NIL)

typedef struct s_repere repere;
typedef repere  *ptrepere;
#define NIL_repere ((ptrepere)NIL)

typedef struct s_node node;
typedef node  *ptnode;
#define NIL_node ((ptnode)NIL)

typedef struct s_filslist filslist;
typedef filslist  *ptfilslist;
#define NIL_filslist ((ptfilslist)NIL)

typedef struct s_filsd filsd;
typedef filsd   *ptfilsd;
#define NIL_filsd ((ptfilsd)NIL)

typedef long int * ptTreeAncestor;
#define NIL_treeAncestor ((ptTreeAncestor)NIL)

typedef ptfilsd ptfilsAny;
#define NIL_filsAny ((ptfilsAny)NIL)

typedef struct s_attrib attribut;
typedef attribut  *ptattribute;
#define NIL_attribute ((ptattribute)NIL)

typedef short typeAccess;

#ifdef _HAVE_64_BITS
#define ADDRESSSIZE 8      /* taille d'une adresse */
#else
#define ADDRESSSIZE 4      /* taille d'une adresse */
#endif

typedef short TMLang;
#define NIL_TMLang ((TMLang)NIL)

/* structure d'un repere sur le noeud d'un arbre*/
/* ===== REPERE AND ELEMENTARY FIELD ACCESS MACRO ==== */
typedef struct s_repere
{ 
  typeAccess filsAnyClass; /* (short) the 2 first bits are used for garbage marking */
  TMLang language; /* (short) */
  union
  { ptfilsd inFix;      /* when filsAnyClass is DIRECTREPERE */
    ptfilslist inList;  /* when filsAnyClass is LISTTREPERE */ 
  } link2filsAny;  /* pointer to a the descriptor of a shared node */
} repere;
#define REPERESIZE    sizeof(repere)
// #define REPERESIZE    (2*ADDRESSSIZE)             /* short+short+adress */

#define REPERE_GET_TACCES(r) ((r)->filsAnyClass)
#define REPERE_IS_DIRECT_ACCESS(AREPERE) ( REPERE_GET_TACCES(AREPERE) == DIRECTREPERE ) 
#define REPERE_GET_TMLANG(r) ((r)->language)
#define REPERE_GET_FILSD_REFERED(r) ((r)->link2filsAny.inFix)
#define REPERE_GET_FILSLIST_REFERED(r) ((r)->link2filsAny.inList)

#define REPERE_SET_TACCES(r,t) REPERE_GET_TACCES(r) = (t)
#define REPERE_SET_DIRECT_ACCESS(AREPERE) REPERE_SET_TACCES(AREPERE, DIRECTREPERE) 
#define REPERE_SET_LIST_ACCESS(AREPERE) REPERE_SET_TACCES(AREPERE, LISTREPERE) 
#define REPERE_SET_TMLANG(r,lang) REPERE_GET_TMLANG(r) = (lang)
#define REPERE_SET_FILSD_REFERED(r,fd) REPERE_GET_FILSD_REFERED(r) = (fd)
#define REPERE_SET_FILSLIST_REFERED(r,AFILSLIST) REPERE_GET_FILSLIST_REFERED(r) = (AFILSLIST)


/* ===== NODE AND ELEMENTARY FIELD ACCESS MACRO ==== */

/* a node N is immediately followed by a filsd descriptor D
   if N has a negative arity then N is a list node and D adfils is a ptfilslist towards the first element of N
   if N has a nul arity then D holds a constant value according to its operator
   if N has a fix strictly positive arity then N is a fix size node, and D is the frist filsd descriptor of the sons of N
   Pere is a pointer to the ptfilslist or the ptfilsd, according to tap, that has N as value, 
   A tree root is a unary node for which pos in D is 0, and pere is NIL_filsd
*/

/* structure d'un noeud: node; un lien externe est un pointeur sur un node */
typedef struct s_node
{
  typeAccess descriptorClass; /* the 2 first bits are used for garbage marking */
  short nodeArity;
  union
  { ptfilsd inFix;     /* when descriptorClass is DIRECTNODE: filsd or attribute */
    ptfilslist inList;    /* when descriptorClass is LISTNODE */
  } descriptor; /* acces au pere qui est un fils*/
  ptattribute attributes;      /* acces à la liste des atributs */
} node;
#define NODESIZE      sizeof(node)
// #define NODESIZE      (3*ADDRESSSIZE)             /* short+short+adress+adress */

#define NODE_GET_DESCRIPTORCLASS(ANODE) ((ANODE)->descriptorClass)
#define NODE_GET_ARITY(ANODE) ((ANODE)->nodeArity)
#define NODE_GET_FILSD_DESCRIPTOR(ANODE) ((ANODE)->descriptor.inFix)
#define NODE_GET_FILSLIST_DESCRIPTOR(ANODE) ((ANODE)->descriptor.inList)
#define NODE_GET_ANYFILS_DESCRIPTOR(ANODE) ((ptfilsAny)NODE_GET_FILSD_DESCRIPTOR(ANODE))
#define NODE_GET_ANCESTOR(ANODE) ((ptTreeAncestor)(NODE_GET_ANYFILS_DESCRIPTOR(ANODE)))
#define NODE_GET_ATTRIBUTE_TREECONTAINER(ANODE) ((ptattribute)NODE_GET_ANYFILS_DESCRIPTOR(ANODE))
#define NODE_GET_ATTRIBUTE(ANODE) ((ANODE)->attributes)
#define NODE_GET_FILSD_FIRST(e) ((ptfilsd)((ptnode)(e) + 1))

#define NODE_SET_DESCRIPTORCLASS(ANODE,ANACCESSTYPE) NODE_GET_DESCRIPTORCLASS(ANODE) = ANACCESSTYPE
#define NODE_SET_ARITY(ANODE,ANARITY) NODE_GET_ARITY(ANODE) = ANARITY
#define NODE_SET_FILSD_DESCRIPTOR(ANODE,AFILSD) NODE_GET_FILSD_DESCRIPTOR(ANODE) = AFILSD
#define NODE_LINK_FILSD_DESCRIPTOR(ANODE,AFILSD) { NODE_SET_FILSD_DESCRIPTOR(ANODE,AFILSD); NODE_SET_DESCRIPTORCLASS(ANODE, DIRECTNODE); }
#define NODE_SET_FILSLIST_DESCRIPTOR(ANODE,AFILSLIST) NODE_GET_FILSLIST_DESCRIPTOR(ANODE) = AFILSLIST
#define NODE_LINK_FILSLIST_DESCRIPTOR(ANODE,AFILSLIST) { NODE_SET_FILSLIST_DESCRIPTOR(ANODE, AFILSLIST); NODE_SET_DESCRIPTORCLASS(ANODE, LISTNODE); }
#define NODE_SET_ATTRIBUTE_CONTAINER(ANODE,ANATTR) NODE_SET_FILSD_DESCRIPTOR(ANODE, (ptfilsd)ANATTR )
#define NODE_SET_ATTRIBUTE(ANODE,ANATTRIBUTE) NODE_GET_ATTRIBUTE(ANODE) = ANATTRIBUTE
#define NODE_CLEAR_ATTRIBUTE(ANODE) NODE_SET_ATTRIBUTE(ANODE, NIL_attribute)

#define NODE_ISIN_LIST(ARG) ( NODE_GET_DESCRIPTORCLASS(ARG) == LISTNODE ) 
#define NODE_ISIN_FIXESIZE(ARG) ( NODE_GET_DESCRIPTORCLASS(ARG) == DIRECTNODE ) 
#define NODE_IS_LIST(ARG) ( NODE_GET_ARITY(ARG) <= -1 ) /* may be smaller than -1 when loaded */
#define NODE_IS_FIXESIZE(aNODE) ( NODE_GET_ARITY(aNODE) > 0  ) /* PLG NODE_GET_ARITY(ARG) is never 0 CORRECT ? */
#define NODE_MAYBE_ATOM(aNODE) ( NODE_GET_ARITY(aNODE) == 1  ) /* PLG NODE_GET_ARITY(ARG) is never 0 CORRECT ? */


/* ===== FILSD AND FILSANY/SPECIFIC FILSD ELEMENTARY FIELD ACCESS MACRO ==== */

/* definition d'un acces a un element de noeud fixe : filsd */
typedef struct s_filsd
{
  short phylum;    /* the 2 first bits are used for garbage marking */
  short position;     /* 0 for the unique initial son descriptor of a tree i for the ith son descriptor of the container node */
  short myOperator;   /* the operator of the described node */
  short bidon;
  untypedValueType value; /* the described actual value: it can be a char, a string pointer, a long integer, a node */
  /* the container node is before its first son in memory */
  /* the predecessor descriptor immediately precedes this descriptor in memory */
  /* the successor descriptor immediately follows this descriptor in memory */
} filsd;
#define FILSDSIZE     sizeof(filsd)
// #define FILSDSIZE     (2*ADDRESSSIZE)             /* short+short+short+short+adress: 8 char are lost with 64 bits */

#define CHECKOPER(OP) ( (0<=(OP)) && ((OP)<=TMAX_op) )
#define CHECKATTRIBUTENUMBER(CODE) ( (1<=(CODE)) && ((CODE)<=TMAX_attribute) )

#define INNEROPER(OP) ((OP) + TCOMMUNVALEUR)
#define USEROPER(OP) ((OP) - TCOMMUNVALEUR)
#define IS_USEROPERATOR(OP) (OP) > (TCOMMUNVALEUR)
#define ISCOMMONOPER(OP) ((OP) <= TCOMMUNVALEUR)

#define ANYFILS_GET_PHYL(ANANYFILS) ((ANANYFILS)->phylum)
#define ANYFILS_GET_POS(ANANYFILS) ((ANANYFILS)->position)
#define ANYFILS_GET_INNEROPER(ANANYFILS) ((ANANYFILS)->myOperator)
#define ANYFILS_GET_OPER(ANANYFILS) (USEROPER(ANYFILS_GET_INNEROPER(ANANYFILS)))
#define ANYFILS_GET_UNTYPED_VALUE(ANANYFILS) ((ANANYFILS)->value)
#define ANYFILS_GET_LONGINT_VALUE(ANANYFILS) ((long int)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))
#define ANYFILS_GET_CHAR_VALUE(ANANYFILS) ((char)(long int)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))
#define ANYFILS_GET_TMSTRING_VALUE(ANANYFILS) ((TMString)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))
#define ANYFILS_GET_FILSD_VALUE(ANANYFILS) ((ptfilsd)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))
#define ANYFILS_GET_FILSLIST_VALUE(ANANYFILS) ((ptfilslist)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))
#define ANYFILS_GET_NODE_VALUE(ANANYFILS) ((ptnode)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))
#define ANYFILS_GET_REPERE_VALUE(ANANYFILS) ((ptrepere)ANYFILS_GET_UNTYPED_VALUE(ANANYFILS))

#define ANYFILS_SET_PHYL(ANANYFILS,ph) (ANYFILS_GET_PHYL(ANANYFILS)) = (ph)
#define ANYFILS_SET_POS(ANANYFILS,p) (ANYFILS_GET_POS(ANANYFILS)) = (p)
#define ANYFILS_SET_INNEROPER(ANANYFILS,op) (ANYFILS_GET_INNEROPER(ANANYFILS)) = (op)
#define ANYFILS_SET_OPER(ANANYFILS,op) ANYFILS_SET_INNEROPER(ANANYFILS, INNEROPER(op))
#define ANYFILS_SET_VALUE(ANANYFILS,op) (ANYFILS_GET_UNTYPED_VALUE(ANANYFILS)) = (untypedValueType)(op)
#define ANYFILS_SET_LONGINT_VALUE(ANANYFILS,e) ANYFILS_SET_VALUE(ANANYFILS, e)
#define ANYFILS_SET_CHAR_VALUE(ANANYFILS,e) ANYFILS_SET_VALUE(ANANYFILS, e)
#define ANYFILS_SET_TMSTRING_VALUE(ANANYFILS,e) ANYFILS_SET_VALUE(ANANYFILS, e)
#define ANYFILS_SET_NODE_VALUE(ANANYFILS,e) ANYFILS_SET_VALUE(ANANYFILS, e)
#define ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, cop, VAL) {ANYFILS_SET_INNEROPER (ANANYFILS, cop); ANYFILS_SET_VALUE(ANANYFILS, VAL);} 

#define ANYFILS_ISIN_LIST(anANYFYLS) ( ANYFILS_GET_POS(anANYFYLS) <= -1 ) /* anANYFYLS should not be NIL */
#define ANYFILS_ISIN_NFIX(anANYFYLS) ( ANYFILS_GET_POS(anANYFYLS) > 0 ) /* anANYFYLS should not be NIL */
#define ANYFILS_IS_ROOT(anANYFYLS) ( ANYFILS_GET_POS(anANYFYLS) == 0 ) /* anANYFYLS should not be NIL */
#define ANYFILS_HAS_USEROPERATOR(SUBTREE) IS_USEROPERATOR(ANYFILS_GET_INNEROPER (SUBTREE))

#define FILSD_GET_NODE_CONTAINER_OF_FIRST(AFIRSTFILSD) ((ptnode)(AFIRSTFILSD) - 1)
#define FILSD_GET_LEFTMOST(AFILSD) ( (ptfilsd)(((ptfilsd)AFILSD) - ((int)(ANYFILS_GET_POS(AFILSD)) - 1)) )
#define FILSD_GET_NODE_CONTAINER(AFILSD) ( ((ANYFILS_GET_POS(AFILSD) == 0)||(ANYFILS_GET_POS(AFILSD) == 1)) ? FILSD_GET_NODE_CONTAINER_OF_FIRST(AFILSD) : FILSD_GET_NODE_CONTAINER_OF_FIRST(FILSD_GET_LEFTMOST(AFILSD)) )
#define FILSD_GET_SUCC(f) ((ptfilsd)((ptfilsd)(f) + (1)))
#define FILSD_GET_NSUCC(f,n) ((ptfilsd)((ptfilsd)(f) + (n)))
#define FILSD_GET_PRED(f) ((ptfilsd)((ptfilsd)(f) - (1)))
#define FILSD_GET_NPRED(f,n) ((ptfilsd)((ptfilsd)(f) - (n)))
#define FILSD_GET_FILSLIST_FIRST(fd) (ANYFILS_GET_FILSLIST_VALUE(fd))

#define FILSD_SET_FILSLIST_FIRST(fd,AFILSLIST) ANYFILS_SET_VALUE(fd, AFILSLIST)
#define FILSD_HAS_SUCC(f)    ( ANYFILS_GET_POS((ptfilsd)f) < NODE_GET_ARITY(FILSD_GET_NODE_CONTAINER((ptfilsd)f)) )
#define FILSD_HAS_nSUCC(f,n) ( ANYFILS_GET_POS((ptfilsd)f) <= (NODE_GET_ARITY(FILSD_GET_NODE_CONTAINER((ptfilsd)f))- n ) )


/* ===== FILSLIST AND SPECIFIC ELEMENTARY FIELD ACCESS MACRO ==== */

/* structure d'acces a un element de liste : filslist */
typedef struct s_filslist
{
  short phylum;    /* the 2 first bits are used for garbage marking */
  short position;     /* pos is -1 */
  short myOperator;   /* the operator of the described node */
  short bidon;
  untypedValueType value; /* the described actual value: it can be a char, a string pointer, a long integer, a node */
  ptnode container;     /* the container node (or father) */
  ptfilslist predecessor; /* the predecessor descriptor (or left brother) */
  ptfilslist successor; /* the successor descriptor (or right brother) */
} filslist;
#define FILSLISTSIZE sizeof(filslist)
// #define FILSLISTSIZE  (FILSDSIZE+(3*ADDRESSSIZE)) /* FILSD+adress+adress+adress: 8 char are lost with 64 bits */

#define FILSLIST_GET_NODE_CONTAINER(LINK2TREESINLIST) ((LINK2TREESINLIST)->container)
#define FILSLIST_GET_PRED(LINK2TREESINLIST) ((LINK2TREESINLIST)->predecessor)
#define FILSLIST_GET_SUCC(LINK2TREESINLIST) ((LINK2TREESINLIST)->successor)

#define FILSLIST_SET_NODE_CONTAINER(AFILSLIST,e) FILSLIST_GET_NODE_CONTAINER(AFILSLIST) = e
#define FILSLIST_SET_PRED(AFILSLIST,AFILSLIST1) FILSLIST_GET_PRED(AFILSLIST) = AFILSLIST1
#define FILSLIST_SET_SUCC(AFILSLIST,AFILSLIST1) FILSLIST_GET_SUCC(AFILSLIST) = AFILSLIST1

/* ===== ATTRIBUTE AND SPECIFIC ELEMENTARY FIELD ACCESS MACRO ==== */

/* structure d'acces a un attribut*/
typedef struct s_attrib
{
  short tag;  /* the 2 first bits are used for garbage marking */
  short myOperator; /* LTA operator giving the type of the value */
  untypedValueType value; /* the described actual value: it can be a char, a string pointer, a long integer, a tree (FILSD), aTMLang */
  ptnode owner; /* owner node */
  ptattribute predecessor; /* predecessor attribute list */
  ptattribute successor; /* tail of the list */
}attribut ;
#define ATTRIBUTESIZE sizeof(attribut)
// #define ATTRIBUTESIZE (5*ADDRESSSIZE)             /* short+short+adress+adress+adress+adress */

#define ATTRIBUTE_GET_INNERTAG(ANATTRIBUTE) ((ANATTRIBUTE)->tag)
#define ATTRIBUTE_GET_TAG(ANATTRIBUTE) (ATTRIBUTE_GET_INNERTAG(ANATTRIBUTE) - ATTRIBUTE_MIN)
#define ATTRIBUTE_GET_INNEROPER(ANATTRIBUTE) ((ANATTRIBUTE)->myOperator)
#define ATTRIBUTE_GET_OPER(ANATTRIBUTE) (USEROPER(ATTRIBUTE_GET_INNEROPER(ANATTRIBUTE)))
#define ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE) ((ANATTRIBUTE)->value)
#define ATTRIBUTE_GET_LONGINT_VALUE(ANATTRIBUTE) ((long int)ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE))
#define ATTRIBUTE_GET_CHAR_VALUE(ANATTRIBUTE) ((char)(long int)ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE))
#define ATTRIBUTE_GET_TMSTRING_VALUE(ANATTRIBUTE) ((TMString)(ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE)))
#define ATTRIBUTE_GET_FILSD_VALUE(ANATTRIBUTE) ((ptfilsd)ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE))
#define ATTRIBUTE_GET_TMLANG_VALUE(ANATTRIBUTE) ((TMLang)ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE))
#define ATTRIBUTE_GET_REPERE_VALUE(ANATTRIBUTE) ((ptrepere)ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE))
#define ATTRIBUTE_GET_NODE_OWNER(ANATTRIBUTE) ((ANATTRIBUTE)->owner)
#define ATTRIBUTE_GET_PRED(ANATTRIBUTE) ((ANATTRIBUTE)->predecessor)
#define ATTRIBUTE_GET_SUCC(ANATTRIBUTE) ((ANATTRIBUTE)->successor)

#define ATTRIBUTE_SET_INNERTAG(ANATTRIBUTE,n) ATTRIBUTE_GET_INNERTAG(ANATTRIBUTE) = (n)
#define ATTRIBUTE_SET_TAG(ANATTRIBUTE,n) ATTRIBUTE_SET_INNERTAG(ANATTRIBUTE, (n) + ATTRIBUTE_MIN)
#define ATTRIBUTE_SET_INNEROPER(ANATTRIBUTE,op) ATTRIBUTE_GET_INNEROPER(ANATTRIBUTE) = op
#define ATTRIBUTE_SET_OPER(ANATTRIBUTE,op) ATTRIBUTE_SET_INNEROPER(ANATTRIBUTE, INNEROPER(op))
#define ATTRIBUTE_SET_VALUE(ANATTRIBUTE,val) ATTRIBUTE_GET_UNTYPED_VALUE(ANATTRIBUTE) = (untypedValueType)(val)
#define ATTRIBUTE_SET_OWNER(ANATTRIBUTE,e) ATTRIBUTE_GET_NODE_OWNER(ANATTRIBUTE) = e
#define ATTRIBUTE_SET_PRED(ANATTRIBUTE,at1) ATTRIBUTE_GET_PRED(ANATTRIBUTE) = at1
#define ATTRIBUTE_SET_SUCC(ANATTRIBUTE,at1) ATTRIBUTE_GET_SUCC(ANATTRIBUTE) = at1

#define ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE,op,val) { ATTRIBUTE_SET_INNEROPER(ANATTRIBUTE,op); ATTRIBUTE_SET_VALUE(ANATTRIBUTE,val); }
#define ATTRIBUTE_MAKE_TROU(ANATTRIBUTE) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, OPTROU, NIL_untypedValueType) 
#define ATTRIBUTE_MAKE_INNEROPER_NIL(ANATTRIBUTE, op) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, op, NIL_untypedValueType) 
#define ATTRIBUTE_MAKE_INTEGER(ANATTRIBUTE, VALUE) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, OPENTIER, (long int)VALUE)
#define ATTRIBUTE_MAKE_CHAR(ANATTRIBUTE, VALUE) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, OPCAR, VALUE)
#define ATTRIBUTE_MAKE_TMSTRING(ANATTRIBUTE, VALUE) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, OPCHAINE, (TMString)VALUE)
#define ATTRIBUTE_MAKE_REPERE(ANATTRIBUTE, VALUE) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, OPREPERE, (ptchar)VALUE)
#define ATTRIBUTE_MAKE_TREE(ANATTRIBUTE, VALUE) ATTRIBUTE_SET_INNEROPER_VALUE(ANATTRIBUTE, OPARBRE, (ptnode)VALUE)
#define ATTRIBUTE_ISOLATE(ANATTRIBUTE) { ATTRIBUTE_SET_OWNER (ANATTRIBUTE, NIL_node); ATTRIBUTE_SET_PRED (ANATTRIBUTE, NIL_attribute); ATTRIBUTE_SET_SUCC (ANATTRIBUTE, NIL_attribute); }

/* ===== COMPOSITE MACROS ==== */
#define NODE_IS_ROOT(aNODE) ( ANYFILS_IS_ROOT(NODE_GET_ANYFILS_DESCRIPTOR(aNODE)) ) /* aNODE should not be NIL */
#define NODE_GET_FILSD_N(AFIXSIZENODE, NN) FILSD_GET_NSUCC(NODE_GET_FILSD_FIRST(AFIXSIZENODE), NN-1)
#define NODE_GET_FILSLIST_FIRST(ALISTNODE) (FILSD_GET_FILSLIST_FIRST(NODE_GET_FILSD_FIRST(ALISTNODE)))
#define NODE_LINK_ATTRIBUTE(ANODE, ANATTRIBUTE)  {NODE_SET_ATTRIBUTE (ANODE, ANATTRIBUTE); ATTRIBUTE_SET_OWNER (ANATTRIBUTE, ANODE); }

#define ANYFILS_IS_LIST(ARG) (NODE_IS_LIST(ANYFILS_GET_NODE_VALUE(ARG)))
#define ANYFILS_GET_FILSLIST_FIRST(ANANYFILS) NODE_GET_FILSLIST_FIRST(ANYFILS_GET_NODE_VALUE(ANANYFILS))
#define ANYFILS_GET_FILSD_FIRST(ANANYFILS) NODE_GET_FILSD_FIRST(ANYFILS_GET_NODE_VALUE(ANANYFILS)) 
#define ANYFILS_GET_NODE_CONTAINER(ARG) ( ((ANYFILS_ISIN_LIST(ARG)) ? FILSLIST_GET_NODE_CONTAINER((ptfilslist)(ARG)) : FILSD_GET_NODE_CONTAINER((ptfilsd)(ARG))) )

#define ANYFILS_MAKE_TROU(ANANYFILS)  ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, OPTROU, NIL) 
#define ANYFILS_MAKE_NIL(ANANYFILS)  ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, OPNIL, NIL) 
#define ANYFILS_MAKE_LONGINT(ANANYFILS, VAL) ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, OPENTIER, VAL)
#define ANYFILS_MAKE_CHAR(ANANYFILS, VAL) ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, OPCAR, VAL)
#define ANYFILS_MAKE_TMSTRING(ANANYFILS, VAL) ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, OPCHAINE, (TMString)VAL) 
#define ANYFILS_MAKE_TREE(ANANYFILS, VAL) ANYFILS_SET_INNEROPER_VALUE(ANANYFILS, OPARBRE, VAL) 

#define FILSD_GET_ANYFILS_PERE(LINK2TREESINFIX) (NODE_GET_ANYFILS_DESCRIPTOR(FILSD_GET_NODE_CONTAINER(LINK2TREESINFIX)))
#define FILSD_GET_ANCESTOR(LINK2TREESINFIX) (NODE_GET_ANCESTOR(FILSD_GET_NODE_CONTAINER(LINK2TREESINFIX)))
#define FILSD_GET_ATTRIBUTE_OF_CONTAINER(AFILSD) NODE_GET_ATTRIBUTE (FILSD_GET_NODE_CONTAINER_OF_FIRST(AFILSD))
#define FILSD_GET_ATTRIBUTE_TREECONTAINER(AFILSD) NODE_GET_ATTRIBUTE_TREECONTAINER(FILSD_GET_NODE_CONTAINER_OF_FIRST(AFILSD))

#define FILSLIST_GET_ANYFILS_PERE(LINK2TREESINLIST) (NODE_GET_ANYFILS_DESCRIPTOR(FILSLIST_GET_NODE_CONTAINER(LINK2TREESINLIST)))

#define TREEFILSD_GET_TMLANG(ATREEFILSD) ATTRIBUTE_GET_TMLANG_VALUE(FILSD_GET_ATTRIBUTE_OF_CONTAINER(ATREEFILSD))

#define ATTRIBUTE_GET_ANYFILS_OWNER(ATREEFILSD) NODE_GET_ANYFILS_DESCRIPTOR(ATTRIBUTE_GET_NODE_OWNER (ATREEFILSD))

/* Selecting a category of a memory cell
   - the first short cell of a filsd or a filslist contains a phylum value (<= TMAX_phyl)
   - the descriptorClass value of a node and the filsAnyClass of a repere have values that follow TMAX_phyl
   - attribute num values are greater than the above values 
 */
#define PHYLUM_MIN 0
#define PHYLUM_MAX TMAX_phyl
#define ATTRIBUTE_MIN (PHYLUM_MAX+5) /* TMAX_phyl+1, ... TMAX_phyl+4 are used by DIRECTREPERE,... */
#define ATTRIBUTE_MAX (ATTRIBUTE_MIN + TMAX_attribute) /* WARNING must be smaller than 16383 (14 bits) */

#define SHORT_VALUEFILTER 0x3fff
#define SHORT_GET_UNMARKEDVALUE(pt) (*pt & SHORT_VALUEFILTER)

// #define INT_VALUEFILTER 0x3fffffff
// #define INT_GET_UNMARKEDVALUE(pt) (*pt & INT_VALUEFILTER)

#define LONG_VALUEFILTER 0x3fffffffffffffff
#define LONG_GET_UNMARKEDVALUE(pt) (*pt & LONG_VALUEFILTER)

#define SHORT_IS_DESCRIPTORVALUETAG(SHORTVAL) ( (PHYLUM_MIN <= SHORTVAL) && (SHORTVAL <= PHYLUM_MAX) )
#define SHORT_IS_DESCRIPTOR(ACELL)  ( SHORT_IS_DESCRIPTORVALUETAG( SHORT_GET_UNMARKEDVALUE(ACELL)) )
#define SHORT_IS_ATTRIBUTEVALUETAG(SHORTVAL) ( (ATTRIBUTE_MIN <= SHORTVAL) && (SHORTVAL <= ATTRIBUTE_MAX) )
#define SHORT_IS_ATTRIBUTE(ACELL)  ( SHORT_IS_ATTRIBUTEVALUETAG( SHORT_GET_UNMARKEDVALUE(ACELL)) )
#define SHORT_IS_REPEREVALUETAG(SHORTVAL) ( SHORTVAL == DIRECTREPERE )
// LISTREPERE unused #define SHORT_IS_REPEREVALUETAG(SHORTVAL) ( (SHORTVAL == DIRECTREPERE) || (SHORTVAL == LISTREPERE) )
#define SHORT_IS_REPERE(ACELL) ( SHORT_IS_REPEREVALUETAG( SHORT_GET_UNMARKEDVALUE(ACELL)) )
#define SHORT_IS_NODEVALUETAG(SHORTVAL) ( (SHORTVAL == DIRECTNODE) || (SHORTVAL == LISTNODE) )
#define SHORT_IS_NODE(ACELL) ( SHORT_IS_NODEVALUETAG( SHORT_GET_UNMARKEDVALUE(ACELL)) )


typedef ptfilsd  *pointeuracces;

typedef ptnode trees;
#define NIL_trees ((trees)NIL)
#define IS_NILTREE(ATREE) ( (trees)(ATREE) == NIL_trees )

typedef trees  *pttrees;
typedef coll  *ptcoll;
typedef stringRegister  *ptStringRegister;
typedef TMStringTableEntry  TMStringAnyEntry;

/* definition d'une entree dans la table des langages */
typedef struct
           { TMString nom;
             ptfilsd arbrelang;
             ptfilsd dicolang;
             ptfilsd arbreatt;
             ptfilsd dicoatt;
           } languageTableEntry;

#define DESCLANG_GET_TMSTRINGNOM(lang) (lang)->nom
#define DESCLANG_SET_TMSTRINGNOM(lang,ch) (lang)->nom = (TMString)(ch)
#define DESCLANG_SET_MISSINGLANG(lang) DESCLANG_SET_TMSTRINGNOM(lang, (TMString)(-1))
#define DESCLANG_IS_MISSINGLANG(lang) ( DESCLANG_GET_TMSTRINGNOM(lang) == (TMString)(-1) )
#define DESCLANG_NAMEIS_TMSTRING(lang, ATMSTRING) ( DESCLANG_GET_TMSTRINGNOM(lang) == ATMSTRING )

#define LANGARBRELANG(lang) lang->arbrelang
#define LANGDICOLANG(lang) lang->dicolang
#define LANGARBREATT(lang) lang->arbreatt
#define LANGDICOATT(lang) lang->dicoatt

#define MOD_LANGARBRELANG(lang,fd) lang->arbrelang = fd
#define MOD_LANGDICOLANG(lang,fd) lang->dicolang = fd
#define MOD_LANGARBREATT(lang,fd) lang->arbreatt = fd
#define MOD_LANGDICOATT(lang,fd) lang->dicoatt = fd
typedef languageTableEntry  *ptlanguageTableEntry;

/********************************/
/* STRUCTURES NECESSAIRES AU    */
/* CONTROLE DE LA DESCRIPTION   */
/* DE LANGAGE ET A LA GENERATION*/
/* DE L'ARBRE LANGAGE           */
/********************************/
typedef struct
  { int arite, base;
  } aritetabint;
typedef aritetabint  *ptaritetabint;

#endif
