/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __TYPESMAC__
#define __TYPESMAC__
#include "basicTypes.h"

/* macros de manipulation de chaines de caracteres */

#define TMSTRINGCHARADR_GET_LENGTH(add) (*((ptSHORT)add))
#define TMSTRINGCHARADR_GET_CSTRING(add) ((char *)((char *)add + 2))


/* definition d'un registre de chaine */
typedef struct {
  short lg;
  short deb;   /* current translation for the string start point */
  char buf[MAXLGCHAINE];
} stringRegister;


#define stringRegister_GET_LG(REG) ((REG)->lg)
#define stringRegister_GET_NEXTFREE(REG) ((REG)->lg)
#define stringRegister_GET_USEDSIZE(REG) ((REG)->lg)
#define stringRegister_GET_STARTINDEX(REG) ((REG)->deb)
#define stringRegister_GET_BUF(REG) ((char*)((REG)->buf))

#define stringRegister_SET_NEXTFREE(REG, IND) (REG)->lg = (IND)
#define stringRegister_SET_USEDSIZE(REG, LG) (REG)->lg = (LG)
#define stringRegister_ADD_USEDSIZE(REG, LG) (REG)->lg = (REG)->lg + (LG)

#define stringRegister_INIT_STARTINDEX(REG) (REG)->deb = 0
#define stringRegister_INIT_STRINGBUF(REG) (REG)->buf[0] = '\0'
#define stringRegister_COMPLETE_STRING(REG) (REG)->buf[stringRegister_GET_LG(REG)] = '\0'

#define stringRegister_GET_FREECHARINDEX(REG) ( stringRegister_GET_STARTINDEX(REG) + stringRegister_GET_USEDSIZE(REG) )

#define stringRegister_GET_CSTRINGSTART(REG) ((char*)(stringRegister_GET_BUF(REG) + stringRegister_GET_STARTINDEX(REG)))

/* definition d'une entree dans la table de hasch des chaines */
// PLG WARNING THE SIZE OF collision HAS BEEN CHANGED
// PLG WARNING THE ORDERING OF collision AND lg HAVE BEEN CHANGED
typedef struct
           { TMString chaine; // sizeof(TMString)=8
             TMStringCollEntryIndice collision;   // sizeof(int)=4
             short lg;        // sizeof(short)=2
           } TMStringTableEntry;

/* definition d'une entree de collision provenant de la table des chaines */
// PLG WARNING THE SIZE OF collision HAS BEEN CHANGED
// PLG WARNING THE ORDERING OF collision AND lg HAVE BEEN CHANGED
typedef struct
           { TMString chaine; /* must remain common with same field in stringTableEntry*/
             TMStringCollEntryIndice collision; /* must remain common with same field in stringTableEntry*/
             short lg;        /* must remain common with same field in stringTableEntry*/
             short HASH_value;
           } TMStringCollEntry;

#define TMSTRINGANYENTRYADR_GET_TMSTRING(TMSTRINGANYENTRY_ARG) ((TMSTRINGANYENTRY_ARG)->chaine)
#define TMSTRINGANYENTRYADR_SET_TMSTRING(TMSTRINGANYENTRY_ARG,TMSTRING_ARG) TMSTRINGANYENTRYADR_GET_TMSTRING(TMSTRINGANYENTRY_ARG) = (TMSTRING_ARG)
#define TMSTRINGANYENTRYADR_GET_LG(TMSTRINGANYENTRY_ARG) ((TMSTRINGANYENTRY_ARG)->lg)
#define TMSTRINGANYENTRYADR_SET_LG(TMSTRINGANYENTRY_ARG,length) TMSTRINGANYENTRYADR_GET_LG(TMSTRINGANYENTRY_ARG) = (length)
#define TMSTRINGANYENTRYADR_GET_COLLISION(TMSTRINGANYENTRY_ARG) ((TMSTRINGANYENTRY_ARG)->collision)
#define TMSTRINGANYENTRYADR_SET_COLLISION(TMSTRINGANYENTRY_ARG,COLL) TMSTRINGANYENTRYADR_GET_COLLISION(TMSTRINGANYENTRY_ARG) = (COLL)

#define TMSTRINGCOLLENTRY_GET_H(cch) ((cch)->HASH_value)
#define TMSTRINGCOLLENTRY_SET_H(cch,h) TMSTRINGCOLLENTRY_GET_H(cch) = h

/* definition d'une entree de la table des collisions */
typedef union
           { TMStringCollEntry entreech;
           } coll;

#endif
