/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __VARGLOBA__
#define __VARGLOBA__

#include <boolean.h>
#include "typesbase.h"

/* Variables relatives aux registres de chaine*/
EXTERN ptStringRegister ptStringRegister_START;
EXTERN ptStringRegister ptStringRegister_FREE;
#define indexRegisterCAST_stringRegister(REG) ( ptStringRegister_START + REG)
#define stringRegisterCAST_indexRegister(STRINGREG) (STRINGREG - ptStringRegister_START)
#define getFreeRegisterIndex() (stringRegisterCAST_indexRegister(ptStringRegister_FREE))

/* Variables relatives a l'espace des arbres (ces variables se trouvaient avant dans gm_init)*/
EXTERN char * DEB_ARBRE; /* debut du tas */ /* transitively used in treefront */
#define GET_TREEMEMORY_FIRSTCELL() DEB_ARBRE
#define SET_TREEMEMORY_FIRSTCELL(PTCHARADR) DEB_ARBRE = PTCHARADR
#define TREEMEMORY_GETTREECOMPONENTPOS(ANADR) ( ((long int)(ANADR)) - ((long int)(DEB_ARBRE)) )
EXTERN char * LIB_ARBRE; /* adresse du 1er emplacement libre du tas */
#define GET_TREEMEMORY_FIRSTFREECELL() LIB_ARBRE
#define GET_TREEMEMORY_USEDSIZE() (LIB_ARBRE - DEB_ARBRE)
#define EMPTY_TREEMEMORY() LIB_ARBRE = DEB_ARBRE
#define RESET_TREEMEMORYSTART() DEB_ARBRE = LIB_ARBRE
#define IS_PTCHAR_INTREEMEMORY(ANADR) ((ANADR >= DEB_ARBRE) && (ANADR < LIB_ARBRE))
#define TREEMEMORY_ISTREECOMPONENTADR(ANADR) ( (((long int)(DEB_ARBRE)) <= ((long int)(ANADR))) && ( ((long int)(ANADR))< ((long int)(LIB_ARBRE))) )
EXTERN char * TEMPORARY_LIB_ARBRE; /* adresse du 1er emplacement libre du tas utilise pour demarrer une gestion en pile de la memoire d'arbre */
EXTERN boolean inTemporaryTreeMemory; /* true if nodes are created in the temporary tree memory */


/* variables relatives a la pile (ces variables se trouvaient avant dans gm_init)*/
EXTERN char * DEB_PILE; /* debut de la pile */
EXTERN long *  SOMMET_PILE; /* adresse du dernier element de la pile */ /* transitively used in treefront */

#define MARGEGARBAGE 32		/* 32 octets pour pouvoir gabarger et sauver */ /* transitively used in treefront */
#define AVAILABLE_TREEMEMORY() ( (char *) SOMMET_PILE - LIB_ARBRE ) /* transitively used in treefront */
#define ISNOTAVAILABLE_TREEMEMORY(SIZE) ( (AVAILABLE_TREEMEMORY() - SIZE) < MARGEGARBAGE ) /* used in treefront */

/* variables relatives au stockage des chaines */
EXTERN char * TMSTRINGAREA_START_ptchar; /* debut du tas */
EXTERN char * TMSTRINGAREA_FREE_ptchar; /*adresse du 1er emplacement libre du tas */
#define TMSTRING_CAST_TMSTRINGCHARADR(CHARINDICE) (TMSTRINGAREA_START_ptchar + (long int)(CHARINDICE))
#define TMSTRINGCHARADR_CAST_TMSTRING(CHARADR) ((TMString)((char *)CHARADR - TMSTRINGAREA_START_ptchar))

/* variables relatives à la zone de collisions */
EXTERN ptcoll COLLAREA_START_ptcoll; /* debut de la zone */
EXTERN ptcoll COLLAREA_FREE_ptcoll; /* 1er emplacement libre dans la zone */
#define COLLENTRYINDICE_CAST_COLLENTRYADR(COLLINDICE) (COLLAREA_START_ptcoll + COLLINDICE)
#define COLLENTRYADR_CAST_COLLENTRYINDICE(COLLADR) ((ptcoll)COLLADR - COLLAREA_START_ptcoll)
#define TMSTRINGCOLLINDICE_IS_TMSTRINGINDICE(INDICE) (INDICE >= 0)

/* Variables relatives a la table de H des chaines */
EXTERN TMStringTableEntry * ptTMStringTableEntry_START; /* debut de la table */
#define TMSTRINGTABLEENTRYINDICE_CAST_TMSTRINGTABLEENTRYADR(TMStringTableEntry_INDICE) (ptTMStringTableEntry_START + (TMStringTableEntry_INDICE))

/* Variables relatives a la table des langages */
EXTERN languageTableEntry * TAB_LANG; /* debut de la table */
#define DESCLANG_CAST_TMLANG(PTLANGUAGEENTRY) ((languageTableEntry *)(PTLANGUAGEENTRY) - TAB_LANG)
#define TMLANG_CAST_DESCLANG(INDEX) (TAB_LANG + (short)(INDEX))

EXTERN int NBRELANG; /* nombre de langages dans la table */
EXTERN TMLang ENVLANG; /* registre pointant arbre du langage courant */
EXTERN int SORTIELTA; /* caractere de sortie de l'interpreteur */

/*****************************************************************************/
/*                                                                           */
/* Le code associe a un operateur est l'indice dans TBL_lop, de 0 a TMAX_op  */
/* Le code associe a un phylum est l'indice dans TBL_lphyl, de 0 a TMAX_phyl */
/* La table de lien entre chaque operateur et la liste des phylums est       */
/* LIEN_opphyl. La table de lien entre chaque phylum et la liste des phylums */
/* LIEN_phylop                                                               */
/*                                                                           */
/*****************************************************************************/


EXTERN aritetabint LIEN_opphyl[TMAX_phyl ];
EXTERN aritetabint LIEN_phylop[TMAX_op ];
EXTERN int TABCODES[TMAX_codes];
EXTERN int IND_TABCODES;

EXTERN TMString TBL_lop[TMAX_op];
EXTERN TMString TBL_lphyl[TMAX_phyl];

EXTERN TMString TBL_oplta[NBOPLTA+1];
EXTERN TMString TBL_typlta[NBTYPLTA+1];

EXTERN int C_UNDEF,C_INTEGER,C_CHAR,C_STRING,C_REFERENCE,C_TREE;


#endif
