/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          allnta.c							     */
/*									     */
/*									     */
/*  Module contenant les primitives de 					     */
/*									     */
/*	  - lireentlnta							     */
/*        - lirecarlnta							     */
/*        - lirechainelnta						     */
/*	  - lireidoplnta				 	             */
/*									     */
/*****************************************************************************/

#include "boolean.h"
#include <ctype.h>
#include "system.h"
#include "typesbase.h"
#include "io_inter.h"
#include "aslnta.h"
#include "allnta.h"


int REGERR;			/* registre contenant le code d'erreur */
char BUFCAR;			/* registre contenant le caractere courant lu a l'analyse d'une chaine */
char BUFCHAINE[180];		/* registre contenant la chaine courant lue */


/*****************************************************************************/
/*									     */
/*	 LIRECAR							     */
/*									     */
/* entree : descripteur de fichier fd					     */
/* sortie : caractere lu dans le fichier de descripteur fd		     */
/*									     */
/*****************************************************************************/

/** return next char in pf 
    may return EOF
 */
extern char
lirecar (FILE * pf)
{
  char c;

  c = fgetc ((FILE *) pf);
  return (c);
}


/*****************************************************************************/
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/

/** put in BUFCAR the next char in pf, skipping spaces 
    may put EOF
 */
extern void
avalercar (FILE * fd)
{

  BUFCAR = lirecar (fd);
  while ((BUFCAR == '\n') || (BUFCAR == '\t') || (BUFCAR == '\r') || (BUFCAR == ' '))
    BUFCAR = lirecar (fd);
}

/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/

/** put in BUFCAR the next char in pf, skipping tabs
    may put EOF
 */
extern void
avalercarsaufblanc (FILE * fd)
{

  BUFCAR = lirecar (fd);
  while ((BUFCAR == '\t') || (BUFCAR == '\r'))
    BUFCAR = lirecar (fd);
}


/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/

/** read a number and return it in res
    return false if reading failed
 */
extern boolean
lireentlnta (FILE * fd, long int *res)
{
  char tab[20];
  int val;
  int i;

  i = 0;

  if (BUFCAR == EOF)
    return pk_false;
  if (BUFCAR != ')')
    {
      avalercar (fd);
      while ((BUFCAR != ')') && (BUFCAR != ' ') && (BUFCAR != EOF))
	{
	  if (isdigit (BUFCAR) == 0)
	    {
	      REGERR = 1;
	      erranalex ();
	      return pk_false;
	    }
	  tab[i] = BUFCAR;
	  i = i + 1;
	  avalercarsaufblanc (fd);
	}
    }
  if (i == 0)
    {
      REGERR = 1;
      erranalex ();
      return pk_false;
    }
  tab[i] = '\0';
  (*res) = atol (tab);
  return pk_true;
}


/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/


extern char
lirecarlnta (FILE * fd)
{
  char c;

  c = '\0';

  if ((BUFCAR == EOF) || (BUFCAR == ')'))
    {
      REGERR = 2;
      erranalex ();
      return '\0';
    }
  avalercar (fd);
  if (BUFCAR == '\'')
    {
      avalercarsaufblanc (fd);
      return BUFCAR;
    }
  if (BUFCAR != ')')
    {
      REGERR = 2;
      erranalex ();
    }
  return '\0';

}


/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/


extern void
lirechainelnta (FILE * fd)
{
  int i = 0;

  if ((BUFCAR != ')') && (BUFCAR != EOF))
    {
      avalercar (fd);
      if (BUFCAR == '"')
	{
	  BUFCAR = lirecar (fd);
	  while ((BUFCAR != '"') && (BUFCAR != EOF))
	    {
	      BUFCHAINE[i] = BUFCAR;
	      i = i + 1;
	      BUFCAR = lirecar (fd);
	    }
	}
      else if (BUFCAR != ')')
	{
	  REGERR = 3;
	  erranalex ();
	}
    }
  BUFCHAINE[i] = '\0';
}


/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/


extern void
lireidoplnta (FILE * fd)
{
  int i = 0;
  avalercar (fd);
  if (isalpha (BUFCAR) == 0)
    {
      REGERR = 7;
      erranalex ();
    }
  else
    {
      while ((BUFCAR != ' ') && (BUFCAR != ')') && (BUFCAR != '(') && (BUFCAR != '\n') && (BUFCAR != '\t') && (BUFCAR != EOF))
	{
	  BUFCHAINE[i] = BUFCAR;
	  BUFCAR = lirecar (fd);
	  i = i + 1;
	}
    }
  BUFCHAINE[i] = '\0';
}
