/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          aslnta.c							     */
/*									     */
/*  Module contenant les primitives de 					     */
/*									     */
/*        - traiteridop							     */
/*									     */
/* !!ATTENTION : la variable CHAINEVIDE declaree localement dans ce module   */
/* alors qu'elle etait avant definie dans vargloba.h                         */
/*****************************************************************************/
#include <string.h>
#include "system.h"
#include "varglobal.h"
#include "li_erreur.h"
#include "allnta.h"
#include "gm_pile.h"
#include "li_langa.h"
#include "io_inter.h"
#include "lta_ctree.h"
#include "lta_cbase.h"
#include "chaine.h"
#include "aslnta.h"

#define PO '('
#define PF ')'
#define BLANC ' '
#define DOLLAR '$'

static TMString CHAINEVIDE;

/* registre contenant le code courant du type de construction d'un sous_arbre*/
static short CODCONS;

/* variable contenant la fd d'un fichier en lecture  */
static FILE *fdlect;


/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/

char *
getCurrentString ()
{
  return BUFCHAINE;
}

void
erranalex (void)
{
  switch (REGERR)

    {
    case 1:
      setFatalTree ("Error on a reading of an 'lta' integer: a number is expected.");
      // "erreur a la lecture d'un entier lta: chiffre attendu.")++;
    case 2:
      setFatalTree ("Error on a reading of an 'lta' character: ' is expected.");
      // "erreur a la lecture d'un caractere lta: ' attendue.")++;
    case 3:
      setFatalTree ("Error on a reading of an 'lta' string: \" is expected.");
      // "erreur a la lecture d'une chaine lta : \" attendue.")++;
    case 4:
      setFatalTree ("Error on a reading of an 'lta' identifier: a letter is expected.");
      // "erreur a la lecture d'un ident lta : lettre attendue.")++;
    case 5:
      setFatalTree ("Error on a reading of an 'lta' identifier: an alphanum is expected.");
      // "erreur a la lecture d'un ident lta : alphanum attendu.")++;
    case 6:
      setFatalTree ("Error on a reading of an 'lta' identifier: ) is expected.");
      // "erreur a la lecture d'un ident lta : ) attendue.")++;
    case 7:
      setFatalTree ("Error on a reading of an 'lta' operator: a letter is expected.");
      // "erreur a la lecture d'un operateur lta :lettre attendue.")++;
    case 8:
      setFatalTree ("Error on a reading of an 'lta' operator: an alphanum is expected.");
      // "erreur a la lecture d'un operateur lta :un alphanum attendu.")++;
    case 9:
      setFatalTree ("Error on reand a new tree.");
      // "lire nouvel arbre.")++;
    case 10:
      setFatalTree ("Trying to get a child for an atomic node.");
      // "tentative de faire fils sur un noeud atomique.")++;
    case 11:
      setFatalTree ("Trying to get the brother a the end of a fixed node.");
      // "tentative de faire frere a la fin d'un noeud fixe.")++;
    case 12:
      setFatalTree ("Trying to get the father of the root.");
      // "tentative de faire pere sur la racine.")++;
    case 13:
      setFatalTree ("a name of an operator is expected at the end of an 'idop'.");
      //"nom d'operateur attendu a la lecture d'un idop.")++;
    case 14:
      setFatalTree ("a sub-tree cannot be attached.");
      //"sous_arbre non attachable.")++;
    default:
      setFatalTree ("'lta' lexical analysis: unexpected error number");
      //" cas d'erreur non prevu lors de l'analyse lexicale lta.")++;
    }
}

/*****************************************************************************/
/*                                                                           */
/*     CONSFIXEDEPLANG                                                       */
/*                                                                           */
/* entree : entier n, l'arite du noeud > 0,cod le cod operateur.             */
/* sortie :                                                                  */
/* fonction : allocation d'un noeud fixe d'arite n avec les champs phyl des  */
/* fils definis qui dependent du langage RLANG courant . Le resultat est     */
/* pointeur sur le node fixe cree.                 		     */
/*                                                                           */
/*****************************************************************************/

static void
consfixedeplang (short x, int y)
{
  printf ("##INTERNAL ERROR## consfixedeplang not implemented\n");
  fflush (stdout);
}

/*****************************************************************************/
/*                                                                           */
/*     CONSLISTDEPLANG                                                       */
/*                                                                           */
/* entree :                                                                  */
/* sortie :                                                                  */
/* fonction : allocation d'un noeud liste. Le resultat est en sommet de pile */
/* un pointeur sur le node liste cree dont le phylum des fils a ete*/
/* defini, dependant du langage courant RLANG.				     */
/*                                                                           */
/*****************************************************************************/
static void
conslistdeplang (short x)
{
  printf ("##INTERNAL ERROR## conslistdeplang not implemented\n");
  fflush (stdout);
}


short
estidoplnta (void)
{
  short typ;
  short rang;

  rang = -1;
  typ = 0;

  if (BUFCHAINE[0] != '\0')
    recherche_dicoC (&typ, &rang);

  if (typ != 1)
    {
      REGERR = 13;
      erranalex ();
      rang = -1;
    }

  return (rang);

}


/*****************************************************************************/
/*									     */
/*	OP_ESTDANS_PHYLFIXE						     */
/*									     */
/* entree : cop,code operateur du langage courant			     */
/* sortie : 	true ou false						     */
/* fonction : teste si le cop est compatible au phylum du filsdirect se      */
/* trouvant en sommet de pile					             */
/*									     */
/*****************************************************************************/

static int
op_estdans_phylfixe (short cop)
{
  ptfilsd f;

  f = (ptfilsd) sommet (0);
  return (appartientopphyl (cop, ANYFILS_GET_PHYL (f)));

}

/*****************************************************************************/
/*									     */
/*	OP_ESTDANS_PHYLLIST						     */
/*									     */
/* entree : cop,code operateur du langage courant			     */
/* sortie : 	true ou false						     */
/* fonction : teste si le cop est compatible au phylum du filslist se        */
/* trouvant en sommet de pile					             */
/*									     */
/*****************************************************************************/

static int
op_estdans_phyllist (short cop)
{
  ptfilslist fl;

  fl = (ptfilslist) sommet (0);

  return (appartientopphyl (cop, ANYFILS_GET_PHYL (fl)));
}


/*****************************************************************************/
/*									     */
/*									     */
/* entree : */
/* sortie : */
/* fonction : */
/*									     */
/*****************************************************************************/

boolean
traiteridop (void)
{
  ptfilsd tgtFilsAny;
  short rangop;
  ptfilsd filsdListDescriptor;
  ptfilslist newFilsList, filsListPred;
  char c;
  long int ent;
  TMString ichain;
  int n;

  ptnode aNode;

  CHAINEVIDE = (TMString) 0;
  lireidoplnta (fdlect);
  rangop = estidoplnta ();

  if (rangop >= 0)
    {
      switch (CODCONS)
	{
	case 1:
	  if (!op_estdans_phylfixe (rangop))
	    {
	      REGERR = 14;
	      erranalex ();
	    }
	  break;
	case 2:
	  if (!op_estdans_phylfixe (rangop))
	    {
	      REGERR = 14;
	      erranalex ();
	      break;
	    }
	  filsdListDescriptor = (ptfilsd) depiler ();
	  newFilsList = insertTM_newFilslist_first (filsdListDescriptor);
	  empiler ((ptchar) newFilsList);
	  break;

	case 3:
	  if (!op_estdans_phyllist (rangop))
	    {
	      REGERR = 14;
	      erranalex ();
	      break;
	    }
	  filsListPred = (ptfilslist) depiler ();
	  newFilsList = insertTM_newFilslist_after (filsListPred);
	  empiler ((ptchar) newFilsList);
	  break;
	default:
	  break;
	}

      if (REGERR != 0)
	{
	  tgtFilsAny = (ptfilsAny) depiler ();
	  CODCONS = getAccessTM_father (tgtFilsAny, &tgtFilsAny);
	  empiler ((ptchar) tgtFilsAny);
	}
      else
	{
	  switch (rangop)
	    {
	    case OPTROU:
	    case OPNIL:
	      ANYFILS_SET_INNEROPER_VALUE ((ptfilsAny) sommet (0), rangop, NIL_node);
	      break;
	    case OPENTIER:
	      if (!lireentlnta (fdlect, &ent))
		return pk_false;
	      ANYFILS_SET_INNEROPER_VALUE ((ptfilsAny) sommet (0), OPENTIER, ent);
	      break;
	    case OPCAR:
	      c = lirecarlnta (fdlect);
	      ANYFILS_SET_INNEROPER_VALUE ((ptfilsAny) sommet (0), OPCAR, c);
	      break;
	    case OPCHAINE:
	      lirechainelnta (fdlect);
	      if (BUFCHAINE[0] == '\0')
		ichain = CHAINEVIDE;
	      else
		{
		  ichain = TMSTRINGCHARADR_CAST_TMSTRING (newTM_CString2TMStringCharAdr (strlen (BUFCHAINE), BUFCHAINE));
		}
	      ANYFILS_SET_INNEROPER_VALUE ((ptfilsAny) sommet (0), OPCHAINE, ichain);
	      break;
	    case OPSTAR:
	    case OPREPERE:
	      ANYFILS_SET_INNEROPER_VALUE ((ptfilsAny) sommet (0), rangop, NIL_node);
	      break;
	      // default case OPARBRE
	    case OPFIXE:
	      while ((BUFCAR != '(') && (BUFCAR != ')') && (BUFCAR != EOF))
		{
		  BUFCAR = lirecar (fdlect);
		}
	      if (BUFCAR == '(')
		{
		  lireidoplnta (fdlect);
		  rangop = estidoplnta ();
		  if (rangop != OPENTIER)
		    {
		      REGERR = 20;
		      erranalex ();
		    }
		  else
		    {
		      if (!lireentlnta (fdlect, &ent))
			return pk_false;
		      if (ent < TMAX_op)
			aNode = newTM_fixSizeNode ((short) ent);
		      else
			{
			  REGERR = 24;
			  erranalex ();
			  aNode = NIL_node;
			}
		      linkTM_filsAny_describedNode ((ptfilsAny) sommet (0), aNode, OPLISTE);
		      while ((BUFCAR != ')') && (BUFCAR != EOF))
			BUFCAR = lirecar (fdlect);
		      BUFCAR = lirecar (fdlect);
		    }
		}
	      else
		{
		  consfixedeplang (rangop, 0);	// NOT IMPLEMENTED
		  aNode = (ptnode) depiler ();
		  linkTM_filsAny_describedNode ((ptfilsAny) sommet (0), aNode, rangop);
		}
	      break;
	    default:
	      n = ariteop (rangop);
	      if (n == -1)
		conslistdeplang (rangop);	// NOT IMPLEMENTED
	      else
		consfixedeplang (rangop, n);	// NOT IMPLEMENTED
	      aNode = (ptnode) depiler ();
	      linkTM_filsAny_describedNode ((ptfilsAny) sommet (0), aNode, rangop);
	      break;
	    }

	}
    }
  return pk_true;
}

/*****************************************************************************/

boolean
lirearbredepl (FILE * fd, ptfilsAny * ANYSON_saveRoot)
{
  ptfilsd tgtFilsAny, tgtFilsAny_res;
  char carpred;
  int fin;
  int nbparenth;

  fin = 0;
  fdlect = fd;
  //  fdecr = stdout;
  carpred = BLANC;
  nbparenth = 0;
  (*ANYSON_saveRoot) = newTM_fixSonRoot_emptyTree ();
  empiler ((ptchar) (*ANYSON_saveRoot));

  CODCONS = 1;

  BUFCAR = lirecar (fdlect);
  while (fin == 0)
    {
      if (REGERR != 0)
	{
	  REGERR = 9;
	  erranalex ();
	  (void) depiler ();
	  return pk_false;
	}
      switch (BUFCAR)
	{
	case EOF:
	  fin = 1;
	  break;
	case PO:
	  nbparenth++;
	  if (carpred == PO)
	    {
	      /* cas ( idop (.. */
	      tgtFilsAny = (ptfilsAny) depiler ();
	      CODCONS = getAccesTM_firstSon (tgtFilsAny, &tgtFilsAny);
	      empiler ((ptchar) tgtFilsAny);
	      if (CODCONS == 0)
		{
		  REGERR = 10;
		  erranalex ();
		  (void) depiler ();
		  return pk_false;
		}
	    }
	  else if (carpred == PF)
	    {
	      /* cas ...)(... */
	      tgtFilsAny = (ptfilsAny) depiler ();
	      CODCONS = getAccessTM_rightNeighbor (tgtFilsAny, &tgtFilsAny);	/* STACK = STACK0 . Y ==>  STACK = STACK0 . frere(Y) */
	      empiler ((ptchar) tgtFilsAny);
	      if (CODCONS == 0)
		{
		  REGERR = 11;
		  erranalex ();
		  (void) depiler ();
		  return pk_false;
		}
	    }
	  carpred = PO;
	  if (!traiteridop ())
	    fin = 1;
	  break;
	case PF:
	  nbparenth--;
	  if (carpred == PF)
	    {
	      /* cas )) */
	      tgtFilsAny = (ptfilsAny) depiler ();
	      CODCONS = getAccessTM_father (tgtFilsAny, &tgtFilsAny);
	      empiler ((ptchar) tgtFilsAny);
	      if (CODCONS == 0)
		{
		  REGERR = 12;
		  erranalex ();
		  (void) depiler ();
		  return pk_false;
		}
	    }
	  carpred = PF;
	  if (nbparenth == 0)
	    {
	      fin = 1;
	      break;
	    }
	  /* pour le cas (idop)  on fait une lecture de BUFCAR */
	  BUFCAR = lirecar (fdlect);
	  break;
	case DOLLAR:
	  (void) depiler ();
	  return pk_false;
	default:
	  BUFCAR = lirecar (fdlect);
	  break;
	}
    }
  (void) depiler ();
  // DEBUGMESS_SONINANYNODE(4,20,"exiting lirearbredepl:read son is ",*ANYSON_saveRoot);
  return pk_true;
}
