/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          chaine.c							     */
/*									     */
/*									     */
/*  Module contenant les primitives de recherche et de creation de chaine :  */
/*	ainsi que la fonction d'allocation d'entrees dans la table des       */
/*	collisions.							     */
/*****************************************************************************/

#include <string.h>
#include <stdio.h>
#include "StdTypeDefs.h"
#include "varglobal.h"
#include "f_coderr.h"
#include "li_erreur.h"
#include "gm_chaine.h"
#include "chaine.h"

short table_entryNumber;	/* nombre d'entrees dans la table des chaines */

/** @param(aCString) is a standard C string
    @param(lg) is the length of aCString
    @return an integer x with 0 <= x < table_entryNumber
 */
extern short
HASHCH (int lg, const char *aCString)
{
  int x, i;
  const char *cc;

  cc = aCString;
  x = 0;
  if (lg != 0)
    {
      for (i = 1; i < lg; i = i + 3)
	{
	  x = x + (int) (*cc);
	  cc = cc + 3;
	}
      cc = aCString + (lg - 1);
      x = x + (int) (*cc);
      if (x < 0)
	x = 0;			// ajout L. Besnard (caractres accentus)
    }
  return (short) (x % table_entryNumber);
}

/*****************************************************************************/
/*                                                                           */
/*  equiv_CString_TMStringCharAdr                                                             */
/*                                                                           */
/* Comparaison de chaines C et LTA					     */
/*   entree : pointeurs sur chaines CString1 et TMString2  	                     */
/*   sortie : 0 si chaines representees par *CString1 et *TMString2 sont egales         */
/*            1 si *CString1 est superieure (alphabetiquement) a *TMString2             */
/*            -1 si *CString1 est inferieure (alphabetiquement) a *TMString2            */
/*                                                                           */
/*****************************************************************************/

static int
equiv_CString_TMStringCharAdr (int lg1, const char *CString1, TMStringCharAdr TMString2)
{
  const char *CString2;

  // printf("DEBUG entering equiv_CString_TMStringCharAdr CString1 is <<%s>>, its length is %i\n", CString1, lg1);fflush(stdout);
  CString2 = TMSTRINGCHARADR_GET_CSTRING (TMString2);
  // printf("DEBUG      in  equiv_CString_TMStringCharAdr CString2 is <<%s>>\n", CString2);fflush(stdout);
  return strcmp (CString1, CString2);	/* returns an integer greater than, equal to, or less than zero,
					   accordingly as CString1 is greater than, equal to, or less than c2 */
  /*
     if (i1 == 0) {
     if (i2 == 0) 
     return (0); 
     return (-1); }
     if (i2 == 0)
     return (1);
     if (i1 == i2) {
     for (i = i1; i > 0; i = i - 1) {
     if (*c1 == *c2) {
     c1++;
     c2++; }
     else if (*c1 < *c2) {
     comp = -1;
     break; }
     else {
     comp = 1;
     break; } } }
     else {
     while ((i1 != 0) && (i2 != 0)) {
     if (*c1 == *c2) {
     i1 = i1 - 1;
     i2 = i2 - 1;
     c1++;
     c2++; }
     else if (*c1 < *c2) 
     i1 = 0;
     else
     i2 = 0;    }
     if (i1 == 0)
     comp = -1;
     else
     comp = 1;
     }
     return (comp);
   */
}

/** search for a registered TMString representing the C string aCString
    if aCString is yet registered return true and 
         assign to indiceRes the collision or table entry that contains the registered TMString entry
    if aCString is not yet registered return false and 
         assign to indiceRes the newly created collision entry or the table entry that will contain the registration
 */
static boolean
find_TMString (short lg, const char *aCString, TMStringAnyEntryIndice * indiceRes)
{
  TMStringAnyEntryIndice aCStringHashValue, aTMStringCollision;
  TMStringTableEntry *entree;
  TMString aTMString;
  TMStringCollEntry *currentColl, *newColl;

  // printf("DEBUG entering find_TMString aCString is <<%s>>\n", aCString);fflush(stdout);
  aCStringHashValue = (TMStringAnyEntryIndice) HASHCH (lg, aCString);
  entree = TMSTRINGTABLEENTRYINDICE_CAST_TMSTRINGTABLEENTRYADR (aCStringHashValue);
  aTMString = TMSTRINGANYENTRYADR_GET_TMSTRING (entree);
  if (aTMString == NIL_TMString)
    {				// the table entry is empty
      (*indiceRes) = aCStringHashValue;	// 0 <= aCStringHashValue
      return pk_false;
    }
  if (equiv_CString_TMStringCharAdr (lg, aCString, TMSTRING_CAST_TMSTRINGCHARADR (aTMString)) == 0)
    {				// the table entry contains a TMString equal to aCString
      (*indiceRes) = aCStringHashValue;
      return pk_true;
    }

  // the table entry contains a TMString not equal to aCString
  //         search in collisions
  aTMStringCollision = TMSTRINGANYENTRYADR_GET_COLLISION (entree);
  if (aTMStringCollision == 0)
    {				// no existing collision with aCStringHashValue: create one
      append_newColl ((TMStringAnyEntry *) entree, aCStringHashValue, lg, indiceRes);
      return pk_false;
    }
  // find aCString in existing collision with aCStringHashValue
  while (aTMStringCollision != 0)
    {
      currentColl = (TMStringCollEntry *) COLLENTRYINDICE_CAST_COLLENTRYADR (aTMStringCollision);
      if (equiv_CString_TMStringCharAdr
	  (lg, aCString, TMSTRING_CAST_TMSTRINGCHARADR (TMSTRINGANYENTRYADR_GET_TMSTRING ((TMStringAnyEntry *) currentColl))) == 0)
	{
	  (*indiceRes) = aTMStringCollision;
	  return pk_true;
	}
      aTMStringCollision = TMSTRINGANYENTRYADR_GET_COLLISION ((TMStringAnyEntry *) currentColl);
    }
  // none of the existing collisions with aCStringHashValue is equivalent to aCString: append a new one
  append_newColl ((TMStringAnyEntry *) currentColl, aCStringHashValue, lg, indiceRes);
  // printf("DEBUG  exiting find_TMString *\n");fflush(stdout);
  return pk_false;
}

/** find the registered TMString XX equal to C string cc 
    if XX is not found, it is created and registered
    @param cc is a correct C string (terminated by '\0')
    @param lg is the length of cc equal to strlen(cc)
    @return XX
 */
extern TMStringCharAdr
newTM_CString2TMStringCharAdr (short lg, const char *cc)
{
  TMStringCharAdr chain;
  TMStringAnyEntryIndice aTMStringEntryIndice;
  TMStringTableEntry *entree;
  TMStringCollEntry *aCollEntry;

  // printf("DEBUG entering newTM_CString2TMStringCharAdr length is %i, string is <<%s>>\n", lg, cc);fflush(stdout);
  if (find_TMString (lg, cc, &aTMStringEntryIndice))
    {				// cc is yet registered at aTMStringEntryIndice
      // printf("DEBUG      in  newTM_CString2TMStringCharAdr 4\n", cc);fflush(stdout);
      if (TMSTRINGCOLLINDICE_IS_TMSTRINGINDICE (aTMStringEntryIndice))
	{
	  entree = TMSTRINGTABLEENTRYINDICE_CAST_TMSTRINGTABLEENTRYADR (aTMStringEntryIndice);
	  chain = TMSTRING_CAST_TMSTRINGCHARADR (TMSTRINGANYENTRYADR_GET_TMSTRING (entree));
	}
      else
	{
	  aCollEntry = (TMStringCollEntry *) COLLENTRYINDICE_CAST_COLLENTRYADR (aTMStringEntryIndice);
	  chain = TMSTRING_CAST_TMSTRINGCHARADR (TMSTRINGANYENTRYADR_GET_TMSTRING ((TMStringAnyEntry *) aCollEntry));
	}
      // printf("DEBUG      in  newTM_CString2TMStringCharAdr 5\n", cc);fflush(stdout);
      return chain;
    }
  //  cc  is not yet registered
  //  aTMStringEntryIndice is the table entry or the newly created collision entry in which cc should be registered

  chain = new_TMString (lg, cc);
  if (TMSTRINGCOLLINDICE_IS_TMSTRINGINDICE (aTMStringEntryIndice))
    {
      entree = TMSTRINGTABLEENTRYINDICE_CAST_TMSTRINGTABLEENTRYADR (aTMStringEntryIndice);
      TMSTRINGANYENTRYADR_SET_LG ((TMStringAnyEntry *) entree, lg);
      TMSTRINGANYENTRYADR_SET_TMSTRING ((TMStringAnyEntry *) entree, TMSTRINGCHARADR_CAST_TMSTRING (chain));
    }
  else
    {
      aCollEntry = (TMStringCollEntry *) (COLLENTRYINDICE_CAST_COLLENTRYADR (aTMStringEntryIndice));
      TMSTRINGANYENTRYADR_SET_LG ((TMStringAnyEntry *) aCollEntry, lg);
      TMSTRINGANYENTRYADR_SET_TMSTRING ((TMStringAnyEntry *) aCollEntry, TMSTRINGCHARADR_CAST_TMSTRING (chain));
    }
  // printf("DEBUG exiting newTM_CString2TMStringCharAdr 2-1\n");fflush(stdout);
  return chain;
}

/** find the TMString XX equal to lg heading characters of cc in the string table 
    if XX is not found, it is created
    @param cc is a sequence of characters
    @param lg is greater than or equal to the length of C string starting at cc
    @return XX
 */
extern TMStringCharAdr
newTM_CSubstring2TMStringCharAdr (short lg, char *cc)
{
  char saveChar;
  TMStringCharAdr res;
  saveChar = cc[lg];
  cc[lg] = '\0';
  // printf("DEBUG      in  newTM_CSubstring2TMStringCharAdr length is %i, string is <<%s>>\n", lg, cc);fflush(stdout);
  res = newTM_CString2TMStringCharAdr (lg, cc);
  cc[lg] = saveChar;
  return res;;
}




/*****************************************************************************/
/*                                                                           */
/*  comparchaine                                                             */
/*                                                                           */
/*   entree : pointeurs sur chaines LTA ch1 et ch2  	                     */
/*   sortie : 0 si chaines representees par *ch1 et *ch2 sont egales         */
/*            1 si *ch1 est superieure (alphabetiquement) a *ch2             */
/*            -1 si *ch1 est inferieure (alphabetiquement) a *ch2            */
/*                                                                           */
/*****************************************************************************/

extern int
comparchaine (int ich1, int ich2)
{
  TMStringCharAdr ch1, ch2;
  ptchar c1, c2;
  int i1, i2, i, comp;

  ch1 = TMSTRING_CAST_TMSTRINGCHARADR (ich1);
  ch2 = TMSTRING_CAST_TMSTRINGCHARADR (ich2);
  comp = 0;
  i1 = TMSTRINGCHARADR_GET_LENGTH (ch1);
  c1 = TMSTRINGCHARADR_GET_CSTRING (ch1);
  i2 = TMSTRINGCHARADR_GET_LENGTH (ch2);
  c2 = TMSTRINGCHARADR_GET_CSTRING (ch2);


  if (i1 == 0)
    {
      if (i2 == 0)
	return (0);
      return (-1);
    }
  if (i2 == 0)
    return (1);
  if (i1 == i2)
    {
      for (i = i1; i > 0; i = i - 1)
	{
	  if (*c1 == *c2)
	    {
	      c1++;
	      c2++;
	    }
	  else if (*c1 < *c2)
	    {
	      comp = -1;
	      break;
	    }
	  else
	    {
	      comp = 1;
	      break;
	    }
	}
    }

  else
    {
      while ((i1 != 0) && (i2 != 0))
	{
	  if (*c1 == *c2)
	    {
	      i1 = i1 - 1;
	      i2 = i2 - 1;
	      c1++;
	      c2++;
	    }
	  else if (*c1 < *c2)
	    {
	      i1 = 0;
	    }
	  else
	    i2 = 0;
	}

      if (i1 == 0)
	{
	  comp = -1;
	}
      else
	comp = 1;
    }

  return (comp);
}

extern int
set_TMStringTableParameters (short entryNumber)
{
  table_entryNumber = entryNumber;
  return entryNumber * sizeof (TMStringTableEntry);	/* taille de la table des chaines */
}

/** build and enter in the table the empty TMString and the TMString that represents "LNTA"
    assume defined TMSTRINGAREA_START_ptchar, TMStringTableEntry *_START
 */
extern void
initTM_stringMemory ()
{
  TMStringTableEntry *curr_TMStringTableEntry;
  TMStringCharAdr aTMStringCharAdr;
  int i;
  ptSHORT lg_pos;
  ptchar c;

  TMSTRINGAREA_FREE_ptchar = TMSTRINGAREA_START_ptchar;

// init TMStringTable entries
  curr_TMStringTableEntry = ptTMStringTableEntry_START;
  for (i = 0; i < table_entryNumber; i++)
    {
      TMSTRINGANYENTRYADR_SET_TMSTRING (curr_TMStringTableEntry, NIL_TMString);
      TMSTRINGANYENTRYADR_SET_COLLISION (curr_TMStringTableEntry, NIL_TMStringCollEntryIndice);
      TMSTRINGANYENTRYADR_SET_LG (curr_TMStringTableEntry, (short) 0);
      curr_TMStringTableEntry++;
    }

// create empty TMString and its table entry
  aTMStringCharAdr = newTM_CString2TMStringCharAdr (0, "");
  // lg_pos = (ptSHORT)TMSTRINGAREA_FREE_ptchar; *lg_pos = (short)0;
  // c = TMSTRINGAREA_FREE_ptchar + 2; *c = '\0';
  // TMSTRINGAREA_FREE_ptchar = TMSTRINGAREA_FREE_ptchar +4;
  // TMSTRINGANYENTRYADR_SET_TMSTRING(CHCOUR, (TMString)0);
  // TMSTRINGANYENTRYADR_SET_COLLISION(CHCOUR, NIL_TMStringCollEntryIndice);
  // TMSTRINGANYENTRYADR_SET_LG(CHCOUR, (short)0);

// create "LNTA" TMString 
  aTMStringCharAdr = newTM_CString2TMStringCharAdr (4, "LNTA");

  // lg_pos = (ptSHORT)TMSTRINGAREA_FREE_ptchar ; *lg_pos = 4;
  // c = TMSTRINGAREA_FREE_ptchar + 2; *c = 'L';
  // c++; *c = 'N';
  // c++; *c = 'T';
  // c++; *c = 'A';
  // c++; *c = '\0';
  // TMSTRINGAREA_FREE_ptchar = TMSTRINGAREA_FREE_ptchar + 8;
  // PLG "LNTA" TMString" is not in the TMString table !!!

  // PLG WHY TMStringTableEntry *_START + 13 ???????
  // CHCOUR = TMStringTableEntry *_START + 17;  /* entree pour 'bidon' */
  // TMSTRINGANYENTRYADR_SET_TMSTRING(CHCOUR, (TMString)(4));
  // TMSTRINGANYENTRYADR_SET_LG(CHCOUR, 4);
}
