/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*         conslnta.c 							     */
/*									     */
/*									     */
/*  Module contenant les primitives permettant de generer le langage noyau   */
/*									     */
/*        - initoptyplta   						     */
/*	  - inittblgenlang						     */
/*        - genererlnta          					     */
/*									     */
/*****************************************************************************/

#include <string.h>
#include <stdio.h>

#include "StdTypeDefs.h"
#include "varglobal.h"
#include "chaine.h"
#include "gm_pile.h"
#include "li_anala.h"
#include "conslnta.h"

#define NONDEF -1

/* tableaux intermediaires utilises pour l'initialisation */
static TMStringCharAdr tblop[NBOPLTA + 1];
static TMStringCharAdr tbltyp[NBTYPLTA + 1];
static TMString CHAINELTA;
static TMString CHAINEVERY;



/*****************************************************************************/
/*	INITTYPESLTA 							     */
/* entree : 							             */
/* sortie : 								     */
/* fonction : initialise le tas des chaines avec les chaines lta qui denotent*/
/* les types de base lta                                                     */
/* LOCALE  A genererlnta						     */
/*****************************************************************************/


/*fonction qui initialise les chaines lta qui denotent les types de base lta*/

static void
inittypeslta (void)
{
  TMStringCharAdr cc;

  cc = newTM_CString2TMStringCharAdr (5, "undef");
  C_UNDEF = TMSTRINGCHARADR_CAST_TMSTRING (cc);
  cc = newTM_CString2TMStringCharAdr (7, "integer");
  C_INTEGER = TMSTRINGCHARADR_CAST_TMSTRING (cc);
  cc = newTM_CString2TMStringCharAdr (4, "char");
  C_CHAR = TMSTRINGCHARADR_CAST_TMSTRING (cc);
  cc = newTM_CString2TMStringCharAdr (6, "string");
  C_STRING = TMSTRINGCHARADR_CAST_TMSTRING (cc);
  cc = newTM_CString2TMStringCharAdr (9, "reference");
  C_REFERENCE = TMSTRINGCHARADR_CAST_TMSTRING (cc);
  cc = newTM_CString2TMStringCharAdr (4, "tree");
  C_TREE = TMSTRINGCHARADR_CAST_TMSTRING (cc);

}


/*****************************************************************************/
/*                                                                           */
/* 	INITOPTYPLTA	                          		             */
/*                                                                           */
/* entree:								     */
/* sortie:								     */
/* fonction: Initialise les tables TBL_oplta avec l'ensemble des chaines lta */
/* qui denotent les operateurs du langage lta et TBL_typlta avec les chaines */
/* qui denotent les types lta . Il y aura donc creation de ces chaines dans  */
/* la zone tas de chaines de notre machine lta                               */
/* LOCALE A genererlnta                                                      */
/*****************************************************************************/

static void
initoptyplta (void)
{
  int i;
  TMStringCharAdr ch;

/* creation de la chaine lta EVERY qui denote le nom du phylum commun*/
/* a tous les langages, il regroupe tous les operateurs d'un langage */

  // printf("DEBUG in initoptyplta every is %s\n", "EVERY"); fflush(stdout);
  ch = newTM_CString2TMStringCharAdr (5, "EVERY");
  CHAINEVERY = TMSTRINGCHARADR_CAST_TMSTRING (ch);

/* INITIALISATION du tableau avec les chaines C */


  tblop[0] = newTM_CString2TMStringCharAdr (4, "trou");
  tbltyp[0] = newTM_CString2TMStringCharAdr (5, "undef");

  tblop[1] = newTM_CString2TMStringCharAdr (3, "nil");
  tbltyp[1] = newTM_CString2TMStringCharAdr (6, "neutre");

  tblop[2] = newTM_CString2TMStringCharAdr (6, "entier");
  tbltyp[2] = newTM_CString2TMStringCharAdr (7, "integer");

  tblop[3] = newTM_CString2TMStringCharAdr (3, "car");
  tbltyp[3] = newTM_CString2TMStringCharAdr (4, "char");

  tblop[4] = newTM_CString2TMStringCharAdr (6, "chaine");
  tbltyp[4] = newTM_CString2TMStringCharAdr (6, "string");

  tblop[5] = newTM_CString2TMStringCharAdr (4, "star");
  tbltyp[5] = (TMStringCharAdr) 0;

  tblop[6] = newTM_CString2TMStringCharAdr (6, "repere");
  tbltyp[6] = newTM_CString2TMStringCharAdr (9, "reference");

  tblop[7] = newTM_CString2TMStringCharAdr (5, "arbre");
  tbltyp[7] = newTM_CString2TMStringCharAdr (4, "tree");

  tblop[8] = newTM_CString2TMStringCharAdr (5, "liste");
  tbltyp[8] = (TMStringCharAdr) 0;

  tblop[9] = newTM_CString2TMStringCharAdr (4, "fixe");
  tbltyp[9] = (TMStringCharAdr) 0;

  tblop[10] = newTM_CString2TMStringCharAdr (5, "ident");
  tbltyp[10] = (TMStringCharAdr) 0;

  tblop[11] = newTM_CString2TMStringCharAdr (7, "binaire");
  tbltyp[11] = (TMStringCharAdr) 0;

  tblop[12] = newTM_CString2TMStringCharAdr (3, "def");
  tbltyp[12] = (TMStringCharAdr) 0;

  tblop[13] = newTM_CString2TMStringCharAdr (5, "appel");
  tbltyp[13] = (TMStringCharAdr) 0;

  tblop[14] = newTM_CString2TMStringCharAdr (8, "fonction");
  tbltyp[14] = newTM_CString2TMStringCharAdr (8, "function");

  tblop[15] = newTM_CString2TMStringCharAdr (6, "module");
  tbltyp[15] = newTM_CString2TMStringCharAdr (6, "module");

  tblop[16] = newTM_CString2TMStringCharAdr (7, "langage");
  tbltyp[16] = newTM_CString2TMStringCharAdr (7, "langage");

  tblop[17] = newTM_CString2TMStringCharAdr (8, "attribut");
  tbltyp[17] = (TMStringCharAdr) 0;

  tblop[18] = newTM_CString2TMStringCharAdr (10, "idelemlang");
  tbltyp[18] = (TMStringCharAdr) 0;

  tblop[19] = newTM_CString2TMStringCharAdr (3, "seq");
  tbltyp[19] = (TMStringCharAdr) 0;

  tblop[20] = newTM_CString2TMStringCharAdr (4, "cons");
  tbltyp[20] = (TMStringCharAdr) 0;

  tblop[21] = newTM_CString2TMStringCharAdr (6, "boucle");
  tbltyp[21] = (TMStringCharAdr) 0;

  tblop[22] = newTM_CString2TMStringCharAdr (6, "sortie");
  tbltyp[22] = (TMStringCharAdr) 0;

  tblop[23] = newTM_CString2TMStringCharAdr (3, "cas");
  tbltyp[23] = (TMStringCharAdr) 0;

  tblop[24] = newTM_CString2TMStringCharAdr (6, "rendre");
  tbltyp[24] = (TMStringCharAdr) 0;

  tblop[25] = newTM_CString2TMStringCharAdr (6, "idlang");
  tblop[26] = newTM_CString2TMStringCharAdr (8, "accesatt");
  tblop[27] = newTM_CString2TMStringCharAdr (9, "attribuer");
  tblop[28] = newTM_CString2TMStringCharAdr (10, "retireratt");
  tblop[29] = newTM_CString2TMStringCharAdr (6, "expcas");
  tblop[30] = newTM_CString2TMStringCharAdr (7, "defmult");

  /* initialisation du tas chaines avec les chaines operateurs communs */


  for (i = 0; i <= NBOPLTA; i++)
    {
      ch = tblop[i];
      TBL_oplta[i] = TMSTRINGCHARADR_CAST_TMSTRING (ch);
    }
  for (i = 0; i <= NBTYPLTA; i++)
    {				/* Modifs Patricia 19 avril 96 */
      ch = tbltyp[i];
      if (ch != (TMStringCharAdr) 0)
	TBL_typlta[i] = TMSTRINGCHARADR_CAST_TMSTRING (ch);
    }


}				/* fin de initoptyplta */



static void
addtabcodes_1 (aritetabint * tab, int ind, int arg1)
{
  tab[ind].arite = 1;
  tab[ind].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = arg1;
  IND_TABCODES++;
}


static void
addtabcodes_l (aritetabint * tab, int ind, int arg1)
{
  tab[ind].arite = -1;
  tab[ind].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = arg1;
  IND_TABCODES++;
}


static void
addtabcodes_2 (aritetabint * tab, int ind, int arg1, int arg2)
{
  tab[ind].arite = 2;
  tab[ind].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = arg1;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg2;
  IND_TABCODES++;
}


static void
addtabcodes_3 (aritetabint * tab, int ind, int arg1, int arg2, int arg3)
{
  tab[ind].arite = 3;
  tab[ind].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = arg1;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg2;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg3;
  IND_TABCODES++;
}


static void
addtabcodes_5 (aritetabint * tab, int ind, int arg1, int arg2, int arg3, int arg4, int arg5)
{
  tab[ind].arite = 5;
  tab[ind].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = arg1;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg2;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg3;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg4;
  IND_TABCODES++;
  TABCODES[IND_TABCODES] = arg5;
  IND_TABCODES++;
}


/*****************************************************************************/
/*                                                                           */
/* 	INITTBLGENLANG	                          		             */
/*                                                                           */
/* entree:								     */
/* sortie:								     */
/* fonction: Initialise les tables tblop avec l'ensemble des chaines lta */
/* qui denotent les operateurs du langage lta et tbltyp avec les chaines */
/* qui denotent les types lta . Il y aura donc creation de ces chaines dans  */
/* la zone tas de chaines de notre machine lta                               */
/* LOCALE A genererlnta                                                      */
/*****************************************************************************/

static void
inittblgenlang (void)
{
  int i;
  ptchar ch;


  /* initialisation de TBL_lop */

  for (i = 0; i <= NBOPLTA; i++)
    TBL_lop[i] = TBL_oplta[i];

  /* initialisation de TBL_lphyl */

  TBL_lphyl[0] = CHAINEVERY;

  tblop[1] = newTM_CString2TMStringCharAdr (2, "ID");
  tblop[2] = newTM_CString2TMStringCharAdr (10, "IDELEMLANG");
  tblop[3] = newTM_CString2TMStringCharAdr (3, "EXP");
  tblop[4] = newTM_CString2TMStringCharAdr (6, "EXPLTA");
  tblop[5] = newTM_CString2TMStringCharAdr (5, "LISTE");
  tblop[6] = newTM_CString2TMStringCharAdr (3, "BIN");
  tblop[7] = newTM_CString2TMStringCharAdr (4, "INST");
  tblop[8] = newTM_CString2TMStringCharAdr (6, "CHAINE");
  tblop[9] = newTM_CString2TMStringCharAdr (6, "EXPCAS");
  tblop[10] = newTM_CString2TMStringCharAdr (6, "IDLANG");
  tblop[11] = newTM_CString2TMStringCharAdr (3, "SEQ");
  tblop[12] = newTM_CString2TMStringCharAdr (5, "CHBIN");
  tblop[13] = newTM_CString2TMStringCharAdr (5, "APPEL");

  for (i = 1; i <= NBPHLTA; i++)

    {
      ch = (ptchar) tblop[i];
      TBL_lphyl[i] = (TMString) TMSTRINGCHARADR_CAST_TMSTRING (ch);

    }

  /* initialisation de LIEN_opphyl */

  LIEN_opphyl[0].arite = 0;	/* trou  */
  LIEN_opphyl[1].arite = 0;	/* nil   */
  LIEN_opphyl[2].arite = 0;	/* entier */
  LIEN_opphyl[3].arite = 0;	/* car  */
  LIEN_opphyl[4].arite = 0;	/* chaine  */
  LIEN_opphyl[5].arite = 0;	/* star  */
  LIEN_opphyl[6].arite = 0;	/* repere  */
  LIEN_opphyl[7].arite = 0;	/* arbre */


  addtabcodes_l (LIEN_opphyl, 8, CODEVERY);	/* liste */

  LIEN_opphyl[9].arite = 0;	/* fixe */

  addtabcodes_1 (LIEN_opphyl, 10, 8);	/* ident */

  addtabcodes_2 (LIEN_opphyl, 11, CODEVERY, CODEVERY);	/* binaire */

  addtabcodes_2 (LIEN_opphyl, 12, 1, 4);	/* def */

  addtabcodes_2 (LIEN_opphyl, 13, 1, 5);	/* appel */

  addtabcodes_5 (LIEN_opphyl, 14, 5, 12, 5, 5, 11);	/* fonction */

  addtabcodes_5 (LIEN_opphyl, 15, 5, 5, 5, 5, 11);	/* module */

  addtabcodes_2 (LIEN_opphyl, 16, 8, 6);	/* langage */

  addtabcodes_2 (LIEN_opphyl, 17, 8, 5);	/* attribut */

  addtabcodes_1 (LIEN_opphyl, 18, 8);	/* idelemlang */

  addtabcodes_l (LIEN_opphyl, 19, 7);	/* seq */

  addtabcodes_2 (LIEN_opphyl, 20, 10, 5);	/* cons */

  addtabcodes_1 (LIEN_opphyl, 21, 11);	/* boucle */

  LIEN_opphyl[22].arite = 0;	/* sortie */

  addtabcodes_2 (LIEN_opphyl, 23, 9, 5);	/* cas */

  addtabcodes_l (LIEN_opphyl, 24, 3);	/* rendre */

  addtabcodes_1 (LIEN_opphyl, 25, 8);	/* idlang */

  addtabcodes_2 (LIEN_opphyl, 26, 3, 2);	/* accesatt */

  addtabcodes_3 (LIEN_opphyl, 27, 3, 2, 3);	/* attribuer */

  addtabcodes_2 (LIEN_opphyl, 28, 3, 2);	/* retireratt */

  addtabcodes_1 (LIEN_opphyl, 29, 3);	/* expcas */

  addtabcodes_2 (LIEN_opphyl, 30, 5, 13);	/* defmult */

  /* initialisation de LIEN_phylop */

  LIEN_phylop[0].arite = NBOPLTA + 1;	/* EVERY */
  LIEN_phylop[0].base = IND_TABCODES;	/* EVERY */
  for (i = 0; i <= NBOPLTA; i++)
    TABCODES[IND_TABCODES + i] = i;
  IND_TABCODES = IND_TABCODES + NBOPLTA + 1;

  addtabcodes_1 (LIEN_phylop, 1, 10);	/* ID */

  addtabcodes_1 (LIEN_phylop, 2, 18);	/* IDELEMLANG */

  LIEN_phylop[3].arite = 10;	/* EXP */
  LIEN_phylop[3].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = 1;
  TABCODES[IND_TABCODES + 1] = 2;
  TABCODES[IND_TABCODES + 2] = 3;
  TABCODES[IND_TABCODES + 3] = 4;
  TABCODES[IND_TABCODES + 4] = 10;
  TABCODES[IND_TABCODES + 5] = 13;
  TABCODES[IND_TABCODES + 6] = 20;
  TABCODES[IND_TABCODES + 7] = 26;
  TABCODES[IND_TABCODES + 8] = 27;
  TABCODES[IND_TABCODES + 9] = 28;
  IND_TABCODES = IND_TABCODES + 10;


  LIEN_phylop[4].arite = 11;	/* EXPLTA */
  LIEN_phylop[4].base = IND_TABCODES;
  TABCODES[IND_TABCODES] = 2;
  TABCODES[IND_TABCODES + 1] = 3;
  TABCODES[IND_TABCODES + 2] = 4;
  TABCODES[IND_TABCODES + 3] = 10;
  TABCODES[IND_TABCODES + 4] = 13;
  TABCODES[IND_TABCODES + 5] = 20;
  TABCODES[IND_TABCODES + 6] = 26;
  TABCODES[IND_TABCODES + 7] = 27;
  TABCODES[IND_TABCODES + 8] = 28;
  TABCODES[IND_TABCODES + 9] = 14;
  TABCODES[IND_TABCODES + 10] = 15;
  IND_TABCODES = IND_TABCODES + 11;

  addtabcodes_1 (LIEN_phylop, 5, 8);	/* LISTE */

  addtabcodes_1 (LIEN_phylop, 6, 11);	/* BIN */

  LIEN_phylop[7].arite = 14;	/* INST */
  LIEN_phylop[7].base = IND_TABCODES;
  TABCODES[IND_TABCODES + 0] = 2;
  TABCODES[IND_TABCODES + 1] = 3;
  TABCODES[IND_TABCODES + 2] = 4;
  TABCODES[IND_TABCODES + 3] = 10;
  TABCODES[IND_TABCODES + 4] = 12;
  TABCODES[IND_TABCODES + 5] = 13;
  TABCODES[IND_TABCODES + 6] = 20;
  TABCODES[IND_TABCODES + 7] = 21;
  TABCODES[IND_TABCODES + 8] = 22;
  TABCODES[IND_TABCODES + 9] = 23;
  TABCODES[IND_TABCODES + 10] = 24;
  TABCODES[IND_TABCODES + 11] = 26;
  TABCODES[IND_TABCODES + 12] = 27;
  TABCODES[IND_TABCODES + 13] = 28;
  IND_TABCODES = IND_TABCODES + 14;

  addtabcodes_1 (LIEN_phylop, 8, 4);	/* CHAINE */

  addtabcodes_1 (LIEN_phylop, 9, 29);	/* EXPCAS */

  addtabcodes_2 (LIEN_phylop, 10, 18, 25);	/* IDLANG */

  addtabcodes_1 (LIEN_phylop, 11, 19);	/* SEQ */

  addtabcodes_2 (LIEN_phylop, 12, 4, 11);	/* CHBIN */

  addtabcodes_1 (LIEN_phylop, 13, 13);	/* APPEL */

}





/*****************************************************************************/
/*                                                                           */
/* 	GENERERLNTA	                          		             */
/*                                                                           */
/* entree:								     */
/* sortie:								     */
/* fonction: Initialise les tables tblop avec l'ensemble des chaines lta     */
/* qui denotent les operateurs du langage lta et tbltyp avec les chaines     */
/* qui denotent les types lta . Il y aura donc creation de ces chaines dans  */
/* la zone tas de chaines de notre machine lta                               */
/* APPELEE DANS initsysteme                                                  */
/*****************************************************************************/

extern void
genererlnta (void)
{
  TMStringCharAdr TMString_lnta;
  ptfilsd f1, f2;

/* creation de la chaine de langage */
  // printf("DEBUG entering genererlnta\n"); fflush(stdout);
  TMString_lnta = newTM_CString2TMStringCharAdr (4, "LNTA");
  // printf("DEBUG      in  genererlnta 1\n"); fflush(stdout);
  CHAINELTA = TMSTRINGCHARADR_CAST_TMSTRING (TMString_lnta);


  inittypeslta ();
  // printf("DEBUG      in  genererlnta 2\n"); fflush(stdout);
  initoptyplta ();
  // printf("DEBUG      in  genererlnta 3\n"); fflush(stdout);
  inittblgenlang ();
  // printf("DEBUG      in  genererlnta 4\n"); fflush(stdout);

  /* creerentree languageTableEntry pour langage "LNTA" */

  ENVLANG = (TMLang) 0;

  nbetiq = NBOPLTA - TCOMMUN;
  nbphyl = NBPHLTA - 1;

  f1 = genarbrelang ();
  // printf("DEBUG      in  genererlnta 5\n"); fflush(stdout);

  gendicolang (0);
  f2 = (ptfilsd) depiler ();
  // printf("DEBUG      in  genererlnta 6\n"); fflush(stdout);

  DESCLANG_SET_TMSTRINGNOM (TAB_LANG, CHAINELTA);
  MOD_LANGARBRELANG (TAB_LANG, f1);
  MOD_LANGDICOLANG (TAB_LANG, f2);
  MOD_LANGARBREATT (TAB_LANG, NIL);
  MOD_LANGDICOATT (TAB_LANG, NIL);

  NBRELANG = 0;

}
