/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          GM_arbre.c						     	     */
/*									     */
/*									     */
/*  Module contenant la primitive d'allocation pour les arbres : 	     */
/*									     */
/*	  - allouerarbre						     */
/*									     */
/*****************************************************************************/

#include <stdio.h>
#include "StdTypeDefs.h"
#include "varglobal.h"
#include "f_coderr.h"
#include "li_erreur.h"
#include "gm_arbre.h"

/*****************************************************************************/
/*									     */
/*     ALLOUERARBRE							     */
/*									     */
/* entree : entier n							     */
/* sortie : pointeur sur octet						     */
/* fonction : allocation de n octets contigus dans la zone des arbres 	     */
/*									     */
/*****************************************************************************/

#define TREEMEMORY_ALLOCATE_NCHAR(SIZE) LIB_ARBRE = LIB_ARBRE + SIZE

/*
  Allocate n bytes in the tree memory
  If there is not enough space available, print an error message and return nil
 */
extern ptchar
allouerarbre (int n)
	/* nombre d'octets a allouer */
{
  ptchar a;

  if (ISNOTAVAILABLE_TREEMEMORY (n))
    {
      errmess (GM_SATARBRE);
      return (ptchar) NIL;
    }
  a = LIB_ARBRE;
  TREEMEMORY_ALLOCATE_NCHAR (n);
  // printf("DEBUG in gm_arbre allouerarbre, created node relative adr is %li\n", TREEMEMORY_GETTREECOMPONENTPOS(a)); fflush(stdout);
  return a;
}
