/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <stdio.h>
#include "StdTypeDefs.h"
#include "varglobal.h"
#include "f_coderr.h"
#include "li_erreur.h"
#include "gm_chaine.h"
#include "TM_DEBUG.h"

/* Variables relatives a la zone des registres de chaines */
short stringRegister_size;
ptStringRegister ptStringRegister_START;
ptStringRegister ptStringRegister_FREE;
#define SET_STRINGREGISTERBOUNDS(START_charAdr) { ptStringRegister_START = ((ptStringRegister)START_charAdr); ptStringRegister_FREE = ptStringRegister_START; }

int
allouerregchaine (void)
{
  int enteringFreeRegisterIndex, newFreeRegisterIndex;

  enteringFreeRegisterIndex = getFreeRegisterIndex ();
  if (enteringFreeRegisterIndex == -1)
    {
      setFatalTree ("No free register for string.");
      return -1;
    }
  newFreeRegisterIndex = stringRegister_GET_NEXTFREE (ptStringRegister_FREE);
  ptStringRegister_FREE = indexRegisterCAST_stringRegister (newFreeRegisterIndex);
  stringRegister_SET_USEDSIZE (indexRegisterCAST_stringRegister (enteringFreeRegisterIndex), 0);
  // printf("DEBUG AAA  exiting gm_chaine::allouerregchaine allocated index is %i,  new free index is %i\n", enteringFreeRegisterIndex, newFreeRegisterIndex); fflush(stdout);
  // DEBUG_regIndex(" exiting gm_chaine::allouerregchaine allocated register ", enteringFreeRegisterIndex);
  // DEBUG_regIndex(" exiting gm_chaine::allouerregchaine new free register ", newFreeRegisterIndex);

  return enteringFreeRegisterIndex;
}

static void
clear_freeStringRegister (ptStringRegister curFreeStringRegister, int nextFreeIndex)
{
  stringRegister_SET_NEXTFREE (curFreeStringRegister, nextFreeIndex);
  stringRegister_INIT_STARTINDEX (curFreeStringRegister);
  stringRegister_INIT_STRINGBUF (curFreeStringRegister);
}

void
libererregchaine (int newFreeIndex)
{
  int curFreeIndex;

  curFreeIndex = getFreeRegisterIndex ();
  ptStringRegister_FREE = indexRegisterCAST_stringRegister (newFreeIndex);
  if (curFreeIndex == -1)
    clear_freeStringRegister (ptStringRegister_FREE, -1);
  else
    clear_freeStringRegister (ptStringRegister_FREE, curFreeIndex);
  // printf("DEBUG AAA exiting gm_chaine::libererregchaine fried index is %i,  previous free index is %i\n", newFreeIndex, curFreeIndex); fflush(stdout);
  // DEBUG_regIndex(" exiting gm_chaine::libererregchaine fried register ", newFreeIndex);
}

extern void
clearTM_stringRegisters (void)	/* libere et reset tous les registres */
{
  int i;
  ptStringRegister curFreeStringRegister;
  // printf("DEBUG entering gm_chaine::clearTM_stringRegisters \n"); fflush(stdout);

  for (i = 0; i < stringRegister_size; i++)
    {
      curFreeStringRegister = indexRegisterCAST_stringRegister (i);
      clear_freeStringRegister (curFreeStringRegister, i + 1);
      // DEBUG_regIndex("entering gm_chaine:: clearTM_stringRegisters", i);
    }
  stringRegister_SET_NEXTFREE (curFreeStringRegister, -1);
  // DEBUG_regIndex("entering gm_chaine:: clearTM_stringRegisters", stringRegisterCAST_indexRegister(curFreeStringRegister));
  // printf("DEBUG  exiting gm_chaine::clearTM_stringRegisters \n"); fflush(stdout);
}

extern ptchar
init_stringRegisterArea (ptchar startAdr, short stringRegister_number)
{
  ptchar res;
  // printf("DEBUG entering gm_chaine:: init_stringRegisterArea \n"); fflush(stdout);
  stringRegister_size = stringRegister_number;
  SET_STRINGREGISTERBOUNDS (startAdr);
  clearTM_stringRegisters ();
  res = (ptchar) (indexRegisterCAST_stringRegister (stringRegister_size));
  // printf("DEBUG  exiting gm_chaine::init_stringRegisterArea res is %li\n", (long)res); fflush(stdout);
  return res;
}

static ptcoll
allouercollision (void)
{
  ptcoll c;

  if (COLLAREA_FREE_ptcoll - 1 < (ptcoll) TMSTRINGAREA_FREE_ptchar)
    {
      errmess (GM_SATCOLL);
      c = (ptcoll) NIL;
    }
  else
    {
      c = COLLAREA_FREE_ptcoll;
      COLLAREA_FREE_ptcoll--;
    }
  return (c);
}

/** create a new empty collision entry appent to prevAny 
 */
void
append_newColl (TMStringAnyEntry * prevAny, TMStringAnyEntryIndice aCStringHashValue, short lg, TMStringAnyEntryIndice * indiceRes)
{
  TMStringCollEntry *newColl;
  newColl = (TMStringCollEntry *) allouercollision ();
  *indiceRes = COLLENTRYADR_CAST_COLLENTRYINDICE (newColl);
  TMSTRINGANYENTRYADR_SET_COLLISION (prevAny, *indiceRes);
  TMSTRINGANYENTRYADR_SET_COLLISION ((TMStringAnyEntry *) newColl, NIL);
  TMSTRINGCOLLENTRY_SET_H (newColl, aCStringHashValue);
}

/** allocate a TMString that can contain length characters
    @param(length) is a positive integer
 */
static TMStringCharAdr
allocate_TMString (short length)
{
  short *TMStringLengthField;
  char *cc;

  TMStringLengthField = (short *) TMSTRINGAREA_FREE_ptchar;
  // allocate 2 bytes for lg field + length bytes for the string + 1 byte for terminal '\0' char in standard C string
  // allocate 1 byte if required to keep TMString alignment
  if ((length & 1) == 1)
    TMSTRINGAREA_FREE_ptchar = TMSTRINGAREA_FREE_ptchar + length + 3;
  else
    TMSTRINGAREA_FREE_ptchar = TMSTRINGAREA_FREE_ptchar + length + 4;
  if (TMSTRINGAREA_FREE_ptchar > (ptchar) COLLAREA_FREE_ptcoll)
    {
      errmess (GM_SATCHAINE);
      return NIL_TMStringCharAdr;
    }
  *TMStringLengthField = length;
  return (TMStringCharAdr) TMStringLengthField;
}

/** create a new TMString equivalent to aCString
 */
TMStringCharAdr
new_TMString (short aCStringLength, const char *aCString)
{
  TMStringCharAdr theTMString;
  char *cc;
  short i;

  theTMString = allocate_TMString (aCStringLength);
  cc = (char *) (theTMString) + 2;
  for (i = 0; i <= aCStringLength; i++)
    {				// copy aCString chracters including the null C string terminator
      *cc = *aCString;
      cc++;
      aCString++;
    }
  return theTMString;
}
