/* 
   Authors: Logic Bernardo, Patricia Burn, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          GM_init.c							     */
/*									     */
/*									     */
/*  Module contenant la primitive d'initialisation de la memoire :	     */
/*									     */
/*	  - initmem							     */
/*									     */
/*  Modules inclus : 							     */
/*									     */
/*	  - typesbase.h							     */
/*									     */
/*****************************************************************************/

#include <limits.h>
#include "system.h"
#include "varglobal.h"
#include "gm_chaine.h"
#include "chaine.h"
#include "sigmem.h"
#include "lta_ctree.h"
#include "gm_init.h"
#include "gm_pile.h"

#include "stringmemStr_type.h"
// #include "fileSystemEnv_g.h"
#include "systemFrontEnd_g.h"

/* Memory parameter declarations
 */

short stringRegister_number;	/* number of string registers; see set_memorySizeParameters() */
long int TreeStack_commonArea_BYTESIZE;	/* byte size of memory space shared by trees and stack; see set_memorySizeParameters() */
long int TMStringCollision_commonArea_BYTESIZE;	/* byte size of memory space shared by strings and hash table collision cells; see set_memorySizeParameters() */
long int LTACode_area_BYTESIZE;	/* byte size of memory allocated to LTA programs; see set_memorySizeParameters() */
short TMStringTable_entryNumber;	/* number of entries in the string hash table; see set_memorySizeParameters() */
short languageTable_entryNumber;	/* number of entries in the language table; see set_memorySizeParameters() */


int TMStringTable_area_BYTESIZE;	/* byte size of memory allocated to string hash table; see set_memorySizeParameters() */
int languageTable_area_BYTESIZE;	/* byte size of memory allocated to language table; see set_memorySizeParameters() */


/* Variables relatives a l'espace des arbres (ces variables se trouvaient avant dans gm_init)*/
ptchar DEB_ARBRE;		/* debut du tas */
ptchar LIB_ARBRE;		/* adresse du 1er emplacement libre du tas */
ptchar TEMPORARY_LIB_ARBRE;	/* adresse du 1er emplacement libre du tas */
boolean inTemporaryTreeMemory = pk_false;

#define SET_TREEMEMORYBOUNDS(LOWSTART) { DEB_ARBRE = (ptchar)LOWSTART; LIB_ARBRE = DEB_ARBRE; }

/* variables relatives a la pile (ces variables se trouvaient avant dans gm_init)*/
ptchar DEB_PILE;		/* debut de la pile */
ptLONG SOMMET_PILE;		/* adresse du dernier element de la pile */

#define SET_STACKMEMORYBOUNDS(HIGHSTART) { DEB_PILE = (ptchar) ((ptLONG)(HIGHSTART)-1); SOMMET_PILE = (ptLONG)DEB_PILE + 1;  }

/* variables relatives a l'espace des chaines */
ptchar TMSTRINGAREA_START_ptchar;	/* debut du tas */
ptchar TMSTRINGAREA_FREE_ptchar;	/* adresse du 1er emplacement libre du tas */
#define SET_TMSTRINGMEMORYBOUNDS(LOWSTART) { TMSTRINGAREA_START_ptchar = (ptchar)LOWSTART; TMSTRINGAREA_FREE_ptchar = TMSTRINGAREA_START_ptchar; }

/* Variables relatives a la zone des collisions */
ptcoll COLLAREA_START_ptcoll;	/* debut de la zone */
ptcoll COLLAREA_FREE_ptcoll;	/* 1er emplacement libre dans la zone */
#define SET_COLLISIONMEMORYBOUNDS(HIGHSTART) { COLLAREA_START_ptcoll = (ptcoll) (HIGHSTART);  COLLAREA_FREE_ptcoll = COLLAREA_START_ptcoll - 1; }

/* Variables relatives a la table de H des chaines */
TMStringTableEntry *ptTMStringTableEntry_START;	/* debut de la table */

/* Variables relatives a la table des langages */
languageTableEntry *TAB_LANG;	/* debut de la table */

int NBRELANG;			/* nombre de langages dans la table */
TMLang ENVLANG;			/* indice dans la table du registre pointant arbre du langage courant */
int SORTIELTA;			/* caractere de sortie de l'interpreteur */

/*****************************************************************************/
/*                                                                           */
/* Le code associe a un operateur est l'indice dans TBL_lop, de 0 a TMAX_op  */
/* Le code associe a un phylum est l'indice dans TBL_lphyl, de 0 a TMAX_phyl */
/* La table de lien entre chaque operateur et la liste des phylums est       */
/* LIEN_opphyl. La table de lien entre chaque phylum et la liste des phylums */
/* LIEN_phylop                                                               */
/*                                                                           */
/*****************************************************************************/

aritetabint LIEN_opphyl[TMAX_phyl];
aritetabint LIEN_phylop[TMAX_op];
int TABCODES[TMAX_codes];
int IND_TABCODES;

TMString TBL_lop[TMAX_op];
TMString TBL_lphyl[TMAX_phyl];

TMString TBL_oplta[NBOPLTA + 1];
TMString TBL_typlta[NBTYPLTA + 1];

int C_UNDEF, C_INTEGER, C_CHAR, C_STRING, C_REFERENCE, C_TREE;


static void
set_memorySizeParameters (void)
{
  char *taille;

  stringRegister_number = STRINGREGISTER_NUMBER;
  taille = getEnvDef (TREESTACKSIZE_HOSTVARIABLE);
  if (taille == NIL)
    TreeStack_commonArea_BYTESIZE = TREESTACK_COMMONAREA_BYTESIZE;
  else
    sscanf (taille, "%li", &TreeStack_commonArea_BYTESIZE);
  taille = getEnvDef (TMSTRINGCOLLISIONSIZE_HOSTVARIABLE);
  if (taille == NIL)
    TMStringCollision_commonArea_BYTESIZE = TMSTRINGCOLLISION_COMMONAREA_BYTESIZE;
  else
    sscanf (taille, "%li", &TMStringCollision_commonArea_BYTESIZE);
  LTACode_area_BYTESIZE = LTACODE_AREA_BYTESIZE;
  TMStringTable_entryNumber = TMSTRINGTABLE_ENTRYNUMBER;	// short prime number
  languageTable_entryNumber = LANGUAGETABLE_ENTRYNUMBER;	// short
}

extern long int
getTreeMemorySize ()
{
  return TreeStack_commonArea_BYTESIZE;
}


/*****************************************************************************/
/*									     */
/*     INITMEM								     */
/*									     */
/* entree : 								     */
/* sortie : 								     */
/* fonction : initialisation des espaces et zones memoire par allocation     */
/*	      d'octets 					     */
/*									     */
/*****************************************************************************/

extern void
initmem (void)
{
  long int fullMemorySize;
  languageTableEntry *RLANG;
  ptchar c, regLastAdr;
  // printf("DEBUG entering initmem TMString size is %i\n", sizeof(TMString)); fflush(stdout);
  // printf("DEBUG entering initmem TMStringTableEntry size is %i\n", sizeof(TMStringTableEntry)); fflush(stdout);
  // printf("DEBUG entering initmem TMStringCollEntry size is %i\n", sizeof(TMStringCollEntry)); fflush(stdout);
  set_memorySizeParameters ();
  TMStringTable_area_BYTESIZE = set_TMStringTableParameters (TMStringTable_entryNumber);	/* number of entries in the TMStringtable :short */
  languageTable_area_BYTESIZE = languageTable_entryNumber * sizeof (languageTableEntry);	/* taille de la table des langages */

  fullMemorySize = (stringRegister_number * sizeof (stringRegister))
    + TreeStack_commonArea_BYTESIZE + TMStringCollision_commonArea_BYTESIZE
    + LTACode_area_BYTESIZE + TMStringTable_area_BYTESIZE + languageTable_area_BYTESIZE;
  c = (ptchar) INITSIGMEM ((size_t) fullMemorySize);

  regLastAdr = init_stringRegisterArea (c, stringRegister_number);
  SET_TREEMEMORYBOUNDS (regLastAdr);
  SET_STACKMEMORYBOUNDS (LIB_ARBRE + TreeStack_commonArea_BYTESIZE);

  SET_TMSTRINGMEMORYBOUNDS (SOMMET_PILE);
  SET_COLLISIONMEMORYBOUNDS (TMSTRINGAREA_START_ptchar + TMStringCollision_commonArea_BYTESIZE);

  ptTMStringTableEntry_START = (TMStringTableEntry *) ((ptchar) (COLLAREA_START_ptcoll) + LTACode_area_BYTESIZE);
  TAB_LANG = (languageTableEntry *) ((ptchar) ptTMStringTableEntry_START + TMStringTable_area_BYTESIZE);
  // #endif
  initTM_stringMemory ();
  // printf("DEBUG     in  initmem 8\n"); fflush(stdout);

  RLANG = TAB_LANG;
  NBRELANG = 0;
  DESCLANG_SET_TMSTRINGNOM (RLANG, (TMString) 4);
  ENVLANG = (TMLang) 0;
  INITSTACKLEVEL ();
  // printf("DEBUG exiting initmem *\n"); fflush(stdout);
}
