/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          GM_pile.c							     */
/*									     */
/*									     */
/*  Module contenant les primitives de gestion de la pile statique :         */
/*									     */
/*         - sommet                                                          */
/*         - empiler                                                         */
/*         - depiler                                                         */
/*****************************************************************************/

#include <stdio.h>
#include "StdTypeDefs.h"
#include "varglobal.h"
#include "f_coderr.h"
#include "li_erreur.h"
#include "gm_pile.h"

int stackLevel;

#define STACKMEMORY_ALLOCATE() SOMMET_PILE = SOMMET_PILE - 1;
#define STACKMEMORY_FREE() SOMMET_PILE = SOMMET_PILE + 1;
#define PILEVIDE() (SOMMET_PILE > (ptLONG) DEB_PILE)
#define PILEPLEINE() (SOMMET_PILE - 4 <= (ptLONG) LIB_ARBRE)


/*****************************************************************************/
/*									     */
/*     SOMMET								     */
/*									     */
/* entree : 								     */
/* sortie : pointeur sur octet						     */
/* fonction : rend la derniere valeur empilee 				     */
/*									     */
/*****************************************************************************/
extern ptchar
sommet (int n)
{
  long int c;
  ptLONG pp;

  // printf("DEBUG sommet level is %i\n", stackLevel); fflush(stdout);
  pp = SOMMET_PILE + n;
  if (pp > (ptLONG) DEB_PILE)
    {
      printf ("###INTERNAL ERROR### in sommet(%i), not enough values in stack\n", n);
      fflush (stdout);
      errmess (GM_PILEVIDE);
    }
  c = *pp;
  return ((ptchar) c);
}

/*****************************************************************************/
/*									     */
/*     EMPILER								     */
/*									     */
/* entree : pointeur sur octet adr					     */
/* sortie : 								     */
/* fonction : range l'adresse adr en sommet de pile			     */
/*									     */
/*****************************************************************************/
extern void
empiler (ptchar adr)
{
  // printf("DEBUG empiler new level is %i\n", stackLevel); fflush(stdout);
  if (PILEPLEINE ())
    {
      errmess (GM_PILEPLEINE);
      return;
    }
  INCRSTACKLEVEL ();
  STACKMEMORY_ALLOCATE ();
  *SOMMET_PILE = (long int) adr;
}

/*****************************************************************************/
/*									     */
/*     DEPILER								     */
/*									     */
/* entree :    								     */
/* sortie : pointeur sur octet						     */
/* fonction : rend la derniere valeur empilee et la depile		     */
/*									     */
/*****************************************************************************/
extern ptchar
depiler (void)
{
  long int c;

  if (PILEVIDE ())
    {
      errmess (GM_PILEVIDE);
      return NIL_ptchar;
    }
  DECRSTACKLEVEL ();
  // printf("DEBUG depiler new level is %i\n", stackLevel); fflush(stdout);
  c = *SOMMET_PILE;
  STACKMEMORY_FREE ();
  return ((ptchar) c);
}
