/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


#include "OS_host.h"
#include "system.h"
#include "varglobal.h"
#include "f_coderr.h"
#include "li_erreur.h"
#include "chaine.h"
#include "gm_pile.h"
#include "langage.h"
#include "lta_cbase.h"
#include "lta_ctree.h"
#include "li_attri.h"
#include "marquer.h"
#include "io_inter.h"


#define MAXLGBUF 2048

typedef struct
{
  int deb;
  char buf[MAXLGBUF];
} registre_entree;

static registre_entree I_BUF;
static registre_entree O_BUF;


static ptfilsd ANYSON_current;	/*les 5 declarations qui suivent utilisees pour la sauvegarde */
static int TT;			/* taille de l'arbre */

/* ========== ========== ========== ========== ========== OBJECT MARK macros 
 */
#define FILSD_MARK 'D'
#define FILSLIST_MARK 'L'
#define ATTRIBUTE_MARK 'A'
#define MAINTREE_LOADCOMPLETION_MARK 'T'
#define SUBTREE_LOADCOMPLETION_MARK 'F'
#define MOD_PREFIXETYPE(adrent,ent)  *((ptchar )adrent) = (char)ent
#define IS_FILSD_MARK(ACHAR) ( ACHAR == FILSD_MARK )
#define IS_FILSLIST_MARK(ACHAR) ( ACHAR == FILSLIST_MARK )
#define IS_ATTRIBUTE_MARK(ACHAR) ( ACHAR == ATTRIBUTE_MARK )
#define IS_MAINTREE_LOADCOMPLETION(ACHAR) ( ACHAR == MAINTREE_LOADCOMPLETION_MARK )
#define IS_SUBTREE_LOADCOMPLETION(ACHAR) ( ACHAR == SUBTREE_LOADCOMPLETION_MARK )
#define IS_TREE_LOADCOMPLETION(ACHAR) ( IS_SUBTREE_LOADCOMPLETION(ACHAR) || IS_MAINTREE_LOADCOMPLETION(ACHAR) )
#define IS_MAINTREE_SAVECOMPLETION() (  )
#define IS_SUBTREE_SAVECOMPLETION() (  )
#define IS_TREE_SAVECOMPLETION(ACHAR) ( IS_SUBTREE_SAVECOMPLETION(ACHAR) || IS_MAINTREE_SAVECOMPLETION(ACHAR) )

/* ========== ========== ========== ========== ========== DEBUG functions 
 */

static int DEBUG_callNumber = 0;

static void
DEBUG_ln (const char *mess)
{
  printf ("DEBUG %s\n", mess);
  fflush (stdout);
}

static void
DEBUG_mess_int_ln (const char *mess, int intValue)
{
  printf ("DEBUG %s <<%i>>\n", mess, intValue);
  fflush (stdout);
}

static void
DEBUG_lit_ecr_int_ln (const char *lit_ecr, const char *mess, int intValue)
{
  printf ("DEBUG %s inte <<%i>> %s\n", lit_ecr, intValue, mess);
  fflush (stdout);
}

static void
DEBUG_lit_ecr_char_ln (const char *lit_ecr, char cc)
{
  printf ("DEBUG %s char <<%c>>\n", lit_ecr, cc);
  fflush (stdout);
}

static void
DEBUG_src (int margin, int maxLevel, const char *mess, ptfilsAny aFilsAny)
{
  DEBUG_srcORtgt (margin, maxLevel, mess, "srs", aFilsAny);
}

static void
DEBUG_tgt (int margin, int maxLevel, const char *mess, ptfilsAny aFilsAny)
{
  DEBUG_srcORtgt (margin, maxLevel, mess, "tgt", aFilsAny);
}


/* ========== ========== ========== ========== ========== read and write functions 
 */

static int
littleIndian_Or_big_indian (int val)
{
  int j;
  ptLONG pi;
  char cc[ADDRESSSIZE];		// ADDRESSSIZE is the size of an adress, defined in Treepack/incl/constmem.h
  char c;

  if (LITTLEENDIAN)
    return val;
  // big_indian
  pi = (ptLONG) cc;
  *pi = val;
  for (j = 0; j < 2; j++)
    {
      c = cc[j];
      cc[j] = cc[3 - j];
      cc[3 - j] = c;
    }
  return (*pi);
}

/* ========== ========== ========== ========== ========== read functions 
 */

static char
litcar (FILE * pf)
{
  static ptchar cc;

  if (I_BUF.deb == MAXLGBUF)
    {
      fread (I_BUF.buf, 1, MAXLGBUF, (FILE *) pf);
      I_BUF.deb = 0;
    }
  cc = I_BUF.buf + I_BUF.deb;
  I_BUF.deb = I_BUF.deb + 1;
  return (*cc);
}

extern int
read_binaryInteger (FILE * pf)
{
  static ptLONG pi;
  int i, j;
  ptchar cc;
  char ii[4];
  int res;

  if (I_BUF.deb + 4 >= MAXLGBUF)
    {
      j = 0;
      for (i = I_BUF.deb; i < MAXLGBUF; i++)
	{
	  I_BUF.buf[j] = I_BUF.buf[i];
	  j++;
	}
      cc = I_BUF.buf + j;
      fread ((char *) cc, 1, MAXLGBUF - j, (FILE *) pf);
      I_BUF.deb = 0;
    }
  for (i = 0; i < 4; i++)
    {
      ii[i] = I_BUF.buf[I_BUF.deb + i];
    }
  pi = (ptLONG) (ii);
  I_BUF.deb = I_BUF.deb + 4;
  res = littleIndian_Or_big_indian (*(ptLONG) (ii));
  return res;
  // return(*pi);
}

static TMStringCharAdr
litchaine (FILE * pf)
{
  static TMStringCharAdr TMString_res;
  char c;
  int LONGBUFF, i;
  char BUFFER[256];

  // printf("DEBUG lit stri <<");
  LONGBUFF = read_binaryInteger (pf);
  // printf("%i:", LONGBUFF); //DEBUG

  if (LONGBUFF > 255)
    {
      errmess (CHLONG);
      return ((TMStringCharAdr) (TMSTRINGAREA_START_ptchar - 1));
    }
  for (i = 0; i < LONGBUFF; i++)
    {
      c = litcar (pf);
      // printf("%c", c); //DEBUG
      BUFFER[i] = c;
    }
  // printf(">>\n"); fflush(stdout); //DEBUG
  TMString_res = newTM_CSubstring2TMStringCharAdr (LONGBUFF, BUFFER);
  return (TMString_res);
}


static ptfilsAny
loadTM_treeRoot (FILE * fde)
{
  ptfilsAny tgtFilsAny_res;
  TMStringCharAdr aTMString;
  TMString icc;
  int aPhylum;
  // printf("DEBUG entering io_inter::loadTM_treeRoot\n"); fflush(stdout);
  aTMString = litchaine (fde);
  // printf("DEBUG      in  io_inter::loadTM_treeRoot 1\n"); fflush(stdout);
  icc = TMSTRINGCHARADR_CAST_TMSTRING (aTMString);
  // printf("DEBUG      in  io_inter::loadTM_treeRoot 2 icc is %li\n", (long)icc); fflush(stdout);
  ENVLANG = chercheridlang (icc);
  // printf("DEBUG      in  io_inter::loadTM_treeRoot 3 ENVLANG is %li\n", ENVLANG); fflush(stdout);
  if (ENVLANG == (TMLang) (-1))
    {
      errmess (LGABSENT);
      return NIL_filsAny;
    }
  tgtFilsAny_res = (ptfilsAny) newTM_fixSonRoot_emptyTree ();
  aPhylum = read_binaryInteger (fde);
  // DEBUG_lit_ecr_int_ln("lit", "tree root phylum", aPhylum);
  ANYFILS_SET_PHYL (tgtFilsAny_res, aPhylum);
  return tgtFilsAny_res;
}

/** load atomic attributes from the current one until a tree attribute is met or until completion
    @return true if a tree attribute is met during tgtAttribute building
    @return false if the attribute loading is completed
    @param tgtFilsAny_res is assigned the first son of the newly created tree descriptor if a tree is met
    @param tgtFilsAny_res is assigned the attribute tgt list owner if the attribute loading is completed
 */

static boolean
load_attributesUntilTree (FILE * fde, ptattribute tgtAttribute, ptfilsAny * tgtFilsAny_res, char *nextTag)
{
  short attributeOperator;
  int anInt;
  char aChar;
  char *aString;
  TMString icc;
  short typatt;

  // DEBUG_tgt(4, 20,"entering load_attributesUntilTree, attribute container is", ATTRIBUTE_GET_ANYFILS_OWNER(tgtAttribute));
  do
    {
      attributeOperator = read_binaryInteger (fde);
      // DEBUG_lit_ecr_int_ln("lit", "attribute operator", attributeOperator);
      switch (attributeOperator)
	{
	case OPTROU:
	case OPNIL:
	  ATTRIBUTE_MAKE_INNEROPER_NIL (tgtAttribute, attributeOperator);
	  break;
	case OPENTIER:
	  anInt = read_binaryInteger (fde);
	  // DEBUG_lit_ecr_int_ln("lit", "attribute integer value", anInt);
	  // DEBUG_mess_int_ln("    in  load_attributesUntilTree OPENTIER read integer", (int)i);
	  ATTRIBUTE_MAKE_INTEGER (tgtAttribute, anInt);
	  break;
	case OPCAR:
	  {
	    aChar = litcar (fde);
	    // DEBUG_lit_ecr_char_ln("lit",car);
	    // DEBUG_ln("    in load_attributesUntilTree OPCAR read char car %c", car);
	    ATTRIBUTE_MAKE_CHAR (tgtAttribute, aChar);
	  }
	  break;
	case OPCHAINE:
	  aString = litchaine (fde);
	  // DEBUG_ln("    in load_attributesUntilTree OPCHAINE read string %s", aString);
	  icc = (TMString) TMSTRINGCHARADR_CAST_TMSTRING (aString);
	  ATTRIBUTE_MAKE_TMSTRING (tgtAttribute, icc);
	  break;
	case OPSTAR:
	  break;
	case OPREPERE:
	  anInt = read_binaryInteger (fde);
	  // DEBUG_lit_ecr_int_ln("lit", "attribute repere value", anInt);
	  // DEBUG_mess_int_ln("    in load_attributesUntilTree OPREPERE read integer ", (int)i);
	  break;
	case OPARBRE:
	  {
	    (*tgtFilsAny_res) = loadTM_treeRoot (fde);
	    typatt = OPARBRE;
	    MOD_PREFIXETYPE (&typatt, ENVLANG);	/* PLG ????? */
	    attachTM_treeAttribute (tgtAttribute, *tgtFilsAny_res);
	    (*nextTag) = litcar (fde);
	    // DEBUG_lit_ecr_char_ln("lit", *nextTag);
	    // DEBUG_tgt(4, 20," exiting load_attributesUntilTree, OPARBRE, attribute container is", ATTRIBUTE_GET_ANYFILS_OWNER(tgtAttribute));
	    // DEBUG_ln("exiting load_attributesUntilTree OPARBRE read char nextTag %c", (*nextTag));
	    return pk_true;
	  }
	}
      (*nextTag) = litcar (fde);
      // DEBUG_lit_ecr_char_ln("lit", *nextTag);
      // DEBUG_ln("    in load_attributesUntilTree !OPARBRE read char nextTag %c", (*nextTag));
      if (IS_ATTRIBUTE_MARK (*nextTag))
	{			/*5 */
	  int num;
	  num = read_binaryInteger (fde);
	  // DEBUG_lit_ecr_int_ln("lit", "attribute category", num);
	  // DEBUG_mess_int_ln("    in load_attributesUntilTree !OPARBRE char ATTRIBUTE_MARK, num is ", num);
	  tgtAttribute = insertTM_newAttributeAfter (tgtAttribute, num);
	}			/*5 */
      else
	{
	  (*tgtFilsAny_res) = ATTRIBUTE_GET_ANYFILS_OWNER (tgtAttribute);
	  // DEBUG_tgt(4, 20," exiting load_attributesUntilTree, completed, attribute container is", *tgtFilsAny_res);
	  return pk_false;
	}
    }
  while (pk_true);
}

/** determine the tgtFilsAny_res that follows tgtFilsAny in top down left to right walk
    @return 0 no container: tgtFilsAny_res is assigned tgtFilsAny
    @return 1 the container is in a fixe size subtree tgtFilsAny_res  is assigned this container
    @return 3 the container is in a list subtree tgtFilsAny_res  is assigned this container
 */
static short
suivantprefixecons (ptfilsAny tgtFilsAny, ptfilsAny * tgtFilsAny_res)
{
  short typattach;
  typattach = getAccesTM_firstSon (tgtFilsAny, &tgtFilsAny);
  /* 0 tgtFilsAny remains an atomic subtree or a OPARBRE tree
     1 tgtFilsAny is the first element of a fix size (non atomic) subtree, 
     2 tgtFilsAny is the filsd list descriptor of an empty list
     3 tgtFilsAny is the first element of a non empty list */
  if (typattach == 0)
    {
      typattach = getAccessTM_rightNeighbor (tgtFilsAny, &tgtFilsAny);
      /* 0 tgtFilsAny is the last element of a fixe size node and may be a root
         1 tgtFilsAny is the right element of its previous value in a fixe size subtree
         3 tgtFilsAny remains a list element */
    }
  if (ANYFILS_IS_ROOT (tgtFilsAny))
    {
      (*tgtFilsAny_res) = tgtFilsAny;
      return 0;
    }
  /* 0 tgtFilsAny is an atomic subtree in terminal position 
     1 tgtFilsAny is an element in a fix size (non atomic) subtree
     2 tgtFilsAny is the filsd list descriptor of an empty list
     3 tgtFilsAny is a list element */
  while (typattach == 0)
    {
      typattach = getAccessTM_father (tgtFilsAny, &tgtFilsAny);
      /* 0 tgtFilsAny is a tree root
         1 tgtFilsAny is an element in a fixe size subtree 
         3 tgtFilsAny is an element in a list subtree */
      if (typattach == 0)
	{
	  (*tgtFilsAny_res) = tgtFilsAny;
	  return 0;
	}
      typattach = getAccessTM_rightNeighbor (tgtFilsAny, &tgtFilsAny);
    }
  (*tgtFilsAny_res) = tgtFilsAny;
  return typattach;
}

/** find the next subtree to be built after tgtFilsAny
    when a list [node1,filsd1] is built its arity is set to -X-1
    each time a son is built this number is incremented by one 
    the list is completed when this number is equal to -1
 */
static void
suivant_a_cons (ptSHORT typattach, char suite, ptfilsAny tgtFilsAny, ptfilsAny * tgtFilsAny_res)
{
  ptnode tgtNodeContainer;
  ptfilslist newFilsList;

  // DEBUG_tgt(4, 20,"entering suivant_a_cons tgtFilsAny is", tgtFilsAny);
  (*typattach) = suivantprefixecons (tgtFilsAny, &tgtFilsAny);
  // printf("DEBUG     in  suivant_a_cons 1, typattach is %i\n", *typattach); fflush(stdout);
  // DEBUG_tgt(4, 20,"     in  suivant_a_cons 1 tgtFilsAny becomes", tgtFilsAny);
  if (*typattach >= 2)
    {				/* STACK = STACK0 */
      if (IS_SUBTREE_LOADCOMPLETION (suite))
	*typattach = 0;
      else
	{
	  if (IS_FILSD_MARK (suite))
	    {
	      while (*typattach >= 2)
		{
		  (*typattach) = getAccessTM_father (tgtFilsAny, &tgtFilsAny);
		  // printf("DEBUG     in  suivant_a_cons 2TFTW1, typattach is %i\n", (int)*typattach); fflush(stdout);
		  if (*typattach != 0)
		    {
		      (*typattach) = getAccessTM_rightNeighbor (tgtFilsAny, &tgtFilsAny);
		      // printf("     in  suivant_a_cons 2TFTW2T, typattach is %i\n", (int)*typattach); fflush(stdout);
		      if (*typattach == 0)	/* pas de frere, il faut continuer */
			*typattach = 2;
		    }
		}		/* while ( *typattach >=2 ) */
	    }
	  else if (IS_FILSLIST_MARK (suite))
	    {			/* suite == L ? */
	      // DEBUG_ln("    in  suivant_a_cons 2T1F1F, typattach = %i ", (int)*typattach);
	      if ((*typattach) == 2)
		tgtNodeContainer = FILSD_GET_NODE_CONTAINER ((ptfilsd) tgtFilsAny);
	      else if ((*typattach) == 3)
		tgtNodeContainer = FILSLIST_GET_NODE_CONTAINER ((ptfilslist) tgtFilsAny);
	      else
		{
		  printf ("##INTERNAL ERROR## in suivant_a_cons 1: unexpected attachment\n");
		  fflush (stdout);
		  return;
		}
	      while (NODE_GET_ARITY (tgtNodeContainer) >= -1)
		{		/* NEXT LIST TO BE COMPLETED */
		  // DEBUG_tgt(4, 4,"     in  suivant_a_cons 2TFF2W tgtFilsAny is", tgtFilsAny);
		  (*typattach) = getAccessTM_father (tgtFilsAny, &tgtFilsAny);
		  // printf("     in  suivant_a_cons 2TFF2W1, typattach is %i\n", (int)*typattach); fflush(stdout);
		  // DEBUG_tgt(4, 4,"     in  suivant_a_cons 2TFF2W1 tgtFilsAny becomes", tgtFilsAny);
		  if ((*typattach) == 1)
		    tgtNodeContainer = FILSD_GET_NODE_CONTAINER ((ptfilsd) tgtFilsAny);
		  else if ((*typattach) == 3)
		    tgtNodeContainer = FILSLIST_GET_NODE_CONTAINER ((ptfilslist) tgtFilsAny);
		  else
		    {
		      printf ("##INTERNAL ERROR## in suivant_a_cons 2: unexpected attachment\n");
		      fflush (stdout);
		      return;
		    }
		}		/* while(NODE_GET_ARITY(e) >= -1) */
	      NODE_SET_ARITY (tgtNodeContainer, (NODE_GET_ARITY (tgtNodeContainer) + 1));
	      if ((*typattach) == 2)
		{
		  newFilsList = insertTM_newFilslist_first (tgtFilsAny);
		}
	      else if ((*typattach) == 3)
		{
		  newFilsList = insertTM_newFilslist_after ((ptfilslist) tgtFilsAny);
		}
	      else
		{
		  newFilsList = newTM_filslist ();
		  empiler ((ptchar) tgtFilsAny);
		  printf ("##INTERNAL ERROR## in suivant_a_cons 3: unexpected attachment\n");
		  fflush (stdout);
		  return;
		}
	      tgtFilsAny = (ptfilsAny) newFilsList;
	    }			/* if ( IS_FILSLIST_MARK(suite) ) */
	  else
	    {
	      printf ("## ERROR ## in  suivant_a_cons, incorrect expected mark; input file is corrupted\n");
	      fflush (stdout);
	      return;
	    }
	}			/* if (  IS_SUBTREE_LOADCOMPLETION(suite) ) else */
    }				/* (*typattach >= 2)  */
  (*tgtFilsAny_res) = tgtFilsAny;
}

extern void
chargement (FILE * fde, ptfilsAny * filsAny_root)
{
  char *aString;
  int icc;
  long int aLongInt;
  char aChar;
  ptTreeAncestor tgtTreeAncestor;
  ptfilsAny tgtFilsAny, previous_tgtFilsAny;
  ptattribute tgtAttribute;
  char suite;
  int op, arite, aPhylum, num;
  boolean isReadyToLoad;	/* true if the current tgtFilsAny is ready to be built */
  short typattach;
  // ptSHORT pt;

  // printf("DEBUG entering io_inter::chargement\n"); fflush(stdout);
  tgtFilsAny = loadTM_treeRoot (fde);
  // printf("DEBUG      in  io_inter::chargement\n"); fflush(stdout);
  (*filsAny_root) = tgtFilsAny;
  // DEBUG_mess_int_ln("    in  chargement read integer phylum ", aPhylum);
  suite = litcar (fde);
  // DEBUG_lit_ecr_char_ln("lit", suite);
  // DEBUG_ln("    in chargement OPCAR read char suite %c", suite);
  // DEBUG_tgt(4, 20,"entering chargement full tree", *filsAny_root);
  while (!IS_MAINTREE_LOADCOMPLETION (suite))
    {				/*0 */
      /*WW*/
	// printf("DEBUG      in  io_inter::chargement while 0\n"); fflush(stdout);
	// DEBUG_tgt(8, 20,"      in  chargement W1 full tree", *filsAny_root);
	// DEBUG_tgt(8, 4,"      in  chargement W1", tgtFilsAny);
	while (!IS_TREE_LOADCOMPLETION (suite))
	{			/*1 */
	  // printf("DEBUG      in  io_inter::chargement while inner 0\n"); fflush(stdout);
	  // DEBUG_tgt(12, 20,"      in  chargement W1W1 full tree", *filsAny_root);
	  // DEBUG_tgt(12, 4,"      in  chargement W1W1 tgtFilsAny", tgtFilsAny);
	  op = read_binaryInteger (fde);
	  // DEBUG_lit_ecr_int_ln("lit", "operator", op);
	  // DEBUG_ln("    in  chargement W1W2 op is %i, suite is %c", op, suite);
	  isReadyToLoad = pk_false;
	  if (ISCOMMONOPER (op))
	    {			/*2 */
	      // DEBUG_tgt(16, 4,"      in  chargement W1W2T COMMON OPER", tgtFilsAny);
	      switch (op)
		{		/*3 */
		case OPTROU:
		case OPNIL:
		  ANYFILS_SET_INNEROPER_VALUE (tgtFilsAny, op, NIL_untypedValueType);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPTROU ... OPSTAR", tgtFilsAny);
		  break;
		case OPENTIER:
		  aLongInt = (long int) read_binaryInteger (fde);
		  // DEBUG_lit_ecr_int_ln("lit", "node integer value", (int)aLongInt);
		  // DEBUG_ln("    in  chargement W1W2T OPENTIER read integer value %li", aLongInt);
		  ANYFILS_MAKE_LONGINT (tgtFilsAny, aLongInt);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPENTIER", tgtFilsAny);
		  break;
		case OPCAR:
		  aChar = litcar (fde);
		  // DEBUG_lit_ecr_char_ln("lit", aChar);
		  // DEBUG_ln("    in chargement W1W2T OPCAR read char aChar %c", aChar);
		  ANYFILS_MAKE_CHAR (tgtFilsAny, aChar);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPCAR", tgtFilsAny);
		  break;
		case OPCHAINE:
		  aString = litchaine (fde);
		  // DEBUG_ln("    in  chargement W1W2T OPCHAINE read string %s", aString);
		  icc = TMSTRINGCHARADR_CAST_TMSTRING (aString);
		  ANYFILS_MAKE_TMSTRING (tgtFilsAny, icc);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPCHAINE", tgtFilsAny);
		  break;
		case OPSTAR:
		  ANYFILS_SET_INNEROPER_VALUE (tgtFilsAny, op, NIL_untypedValueType);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPTROU ... OPSTAR", tgtFilsAny);
		  break;
		case OPREPERE:
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPREPERE", tgtFilsAny);
		  break;
		case OPARBRE:
		  previous_tgtFilsAny = tgtFilsAny;
		  tgtFilsAny = loadTM_treeRoot (fde);
		  isReadyToLoad = pk_true;
		  ANYFILS_MAKE_TREE (previous_tgtFilsAny, tgtFilsAny);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2T OPARBRE", previous_tgtFilsAny);
		  break;
		}		/*3 */
	      // DEBUG_tgt(16, 4,"      in  chargement W1W2T*", tgtFilsAny);
	      suite = litcar (fde);
	      // DEBUG_lit_ecr_char_ln("lit", suite);
	      // DEBUG_ln("    in chargement W1W2T* OPCAR read char suite %c", suite);
	    }			/*2 */
	  else
	    {			/*2 *//* user defined operator */
	      ptnode tgtNode;
	      // DEBUG_tgt(16, 4,"      in  chargement W1W2F SPECIFIC OPER", tgtFilsAny);
	      arite = read_binaryInteger (fde);
	      // DEBUG_lit_ecr_int_ln("lit", "arity", arite);
	      // DEBUG_ln("     in  chargement W1W2F arite is %i ", arite);
	      if (arite < 0)
		{		/*3 *//* a list subtree */
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2F1T SPECIFIC OPER IS LIST", tgtFilsAny);
		  tgtNode = newTM_listNode ();
		  linkTM_filsAny_describedNode (tgtFilsAny, tgtNode, op);
		  NODE_SET_ARITY (tgtNode, arite);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2F1T-2 SPECIFIC OPER IS LIST", tgtFilsAny);
		  aPhylum = read_binaryInteger (fde);
		  // DEBUG_lit_ecr_int_ln("lit", "phylum", aPhylum);
		  ANYFILS_SET_PHYL (NODE_GET_FILSD_FIRST (tgtNode), aPhylum);	/* descriptor (chosen) or conatainer ? */
		}		/*3 */
	      else
		{		/*3 *//* a fixe size subtree */
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2F1F SPECIFIC OPER IS FIXE", tgtFilsAny);
		  tgtNode = newTM_fixSizeNode (arite);
		  linkTM_filsAny_describedNode (tgtFilsAny, tgtNode, op);
		  // DEBUG_tgt(16, 4,"      in  chargement W1W2F1F-2 SPECIFIC OPER IS FIXE", tgtFilsAny);
		  tgtFilsAny = NODE_GET_FILSD_FIRST (tgtNode);
		  if (arite == 0)
		    isReadyToLoad = pk_false;
		  else
		    {		/*4 */
		      int i;
		      for (i = 1; i <= arite; i++)
			{	/*5 */
			  aPhylum = read_binaryInteger (fde);
			  // DEBUG_lit_ecr_int_ln("lit", "phylum", aPhylum);
			  // DEBUG_mess_int_ln("    in  chargement W1W2F1F read integer phylum ", aPhylum);
			  ANYFILS_SET_PHYL (tgtFilsAny, aPhylum);
			  tgtFilsAny = FILSD_GET_SUCC (tgtFilsAny);
			}	/*5 */
		      tgtFilsAny = NODE_GET_FILSD_FIRST (tgtNode);
		      isReadyToLoad = pk_true;
		    }		/*4 */
		}		/*3 */
	      suite = litcar (fde);
	      // DEBUG_lit_ecr_char_ln("lit", suite);
	      // DEBUG_ln("    in chargement W1W3 OPCAR read char suite %c", suite);
	      if (IS_ATTRIBUTE_MARK (suite))
		{		/*3 */
		  num = read_binaryInteger (fde);
		  // DEBUG_lit_ecr_int_ln("lit", "attribute category", num);
		  // DEBUG_mess_int_ln("    in  chargement W1W3T read integer num ", num);
		  tgtAttribute = insertTM_newFirstAttribute (tgtNode, num);
		  isReadyToLoad = load_attributesUntilTree (fde, tgtAttribute, &tgtFilsAny, &suite);
		}		/*3 *//* if ( IS_ATTRIBUTE_MARK(suite) )  */
	      // DEBUG_tgt(16, 4,"      in  chargement W1W2F*", tgtFilsAny);
	    }			/*2 *//* end of user defined operator */
	  if (IS_TREE_LOADCOMPLETION (suite))
	    {			/*2 */
	      // DEBUG_tgt(16, 4,"      in  chargement W1W3T*", tgtFilsAny);
	      tgtFilsAny = getTM_filsd_root (tgtFilsAny);
	      typattach = 0;
	    }			/*2 */
	  else
	    {			/*2 */
	      // DEBUG_tgt(16, 4,"      in  chargement W1W3F", tgtFilsAny);
	      if (!isReadyToLoad)
		suivant_a_cons (&typattach, suite, tgtFilsAny, &tgtFilsAny);
	    }			/*2 */
	  // DEBUG_tgt(12, 4,"      in  chargement W1W*", tgtFilsAny);
	}			/*1 *//* while (! IS_TREE_LOADCOMPLETION(suite)){} */
      // printf("DEBUG      in  io_inter::chargement while 0-1\n"); fflush(stdout);
      // DEBUG_tgt(8, 4,"      in  chargement W2", tgtFilsAny);
      if (!IS_MAINTREE_LOADCOMPLETION (suite))
	{
	  /*TTT*/ tgtTreeAncestor = FILSD_GET_ANCESTOR (tgtFilsAny);
	  switch (cellCategory ((ptSHORT) tgtTreeAncestor))
	    {
	    case 0:		/* acces */
	      tgtFilsAny = FILSD_GET_ANYFILS_PERE (tgtFilsAny);
	      suite = litcar (fde);
	      // DEBUG_lit_ecr_char_ln("lit", suite);
	      // DEBUG_ln("    in chargement W2TSW0 OPCAR read char suite %c", suite);
	      if (IS_TREE_LOADCOMPLETION (suite))
		{
		  tgtFilsAny = getTM_filsd_root (tgtFilsAny);
		  typattach = 0;
		}
	      else
		suivant_a_cons (&typattach, suite, tgtFilsAny, &tgtFilsAny);
	      // DEBUG_tgt(8, 4,"      in  chargement W2T1SW0*", tgtFilsAny);
	      break;
	    case 1:		/* attribut */
	      tgtAttribute = FILSD_GET_ATTRIBUTE_TREECONTAINER (tgtFilsAny);
	      suite = litcar (fde);
	      // DEBUG_lit_ecr_char_ln("lit", suite);
	      // DEBUG_ln("    in chargement W2TSW1 OPCAR read char suite %c", suite);
	      if (IS_ATTRIBUTE_MARK (suite))
		{
		  num = read_binaryInteger (fde);
		  // DEBUG_lit_ecr_int_ln("lit", "attribute category", num);
		  // DEBUG_mess_int_ln("    in  chargement W2TSW1TD read integer num ", num);
		  tgtAttribute = insertTM_newAttributeAfter (tgtAttribute, num);
		  isReadyToLoad = load_attributesUntilTree (fde, tgtAttribute, &tgtFilsAny, &suite);
		  // DEBUG_ln("    in chargement W2TSW1TD* OPCAR read char suite %c", suite);
		  // DEBUGMESS_NODE(4,2,"after loop do 2  of chargement, tgtNode ", tgtNode);
		}
	      else
		{
		  tgtFilsAny = ATTRIBUTE_GET_ANYFILS_OWNER (tgtAttribute);
		  isReadyToLoad = pk_false;
		}
	      if (!isReadyToLoad)
		{
		  if (IS_TREE_LOADCOMPLETION (suite))
		    {
		      tgtFilsAny = getTM_filsd_root (tgtFilsAny);
		      typattach = 0;
		    }
		  else
		    suivant_a_cons (&typattach, suite, tgtFilsAny, &tgtFilsAny);
		}
	      // DEBUG_tgt(8, 4,"      in  chargement W2T1SW1*", tgtFilsAny);
	      break;		/* end attribut */
	    default:
	      printf ("##INTERNAL ERROR## in  chargement: unexpected cell category\n");
	      fflush (stdout);
	      break;
	    }			/* end switch */
	}			/* end !IS_MAINTREE_LOADCOMPLETION(suite) */
      else
	{			/* IS_MAINTREE_LOADCOMPLETION(suite) */
	  tgtFilsAny = getTM_filsd_root (tgtFilsAny);
	  typattach = 0;
	}
      // DEBUG_tgt(8, 4,"      in  chargement W1", tgtFilsAny);
    }				/* end outer while */
  // PLG REMOVED ENVLANG = lang1; langage courant devient le langage de l'arbre charge *
  // DEBUG_tgt(4, 20," exiting chargement full tree", *filsAny_root);
  // printf("DEBUG  exiting io_inter::chargement\n"); fflush(stdout);
}

extern boolean
TM_existFile (char *fileName)
{
  FILE *fd;

  fd = fopen (fileName, "r");
  if (fd == (FILE *) NULL)
    return pk_false;
  fclose (fd);
  return pk_true;
}

extern FILE *
ouvrelect (ptchar fileName)
{
  static FILE *fdl;
  fdl = fopen (fileName, "rb");	/* le "b" (binaire) est obligatoire pour PC */
  if (fdl == (FILE *) NULL)
    {
      printf ("###FILE ERROR### opening the file '%s' in reading mode failed\n", fileName);
      fflush (stdout);
      return (FILE *) NULL;
    }
  return fdl;
}

/** open the fileName file 
    @return the created file descriptor
    @return null if opening fileName failed 
    @return null if fileName is an empty file; fileName is closed
 */
extern FILE *
initload (ptchar fileName)
{
  FILE *fdl;
  fdl = ouvrelect (fileName);
  I_BUF.deb = 0;
  if (fread (I_BUF.buf, 1, MAXLGBUF, (FILE *) fdl) == 0)
    {
      printf ("###FILE ERROR### the file '%s' is empty\n", fileName);
      fflush (stdout);
      fclose (fdl);
      return (FILE *) NULL;
    }
// num = feof((FILE*)fdl);
// num = ferror((FILE*)fdl);
  return fdl;
}

/* ========== ========== ========== ========== ========== write functions 
 */

static void
emptyFullBuf (FILE * pf, short n)
{
  if (O_BUF.deb + n > MAXLGBUF)
    {
      fwrite (O_BUF.buf, 1, O_BUF.deb, (FILE *) pf);
      O_BUF.deb = 0;
    }
}

static void
putToBuf_char (char c)
{
  O_BUF.buf[O_BUF.deb] = c;
  O_BUF.deb = O_BUF.deb + 1;
}

#define ECRITCAR(pf,c) emptyFullBuf(pf, 1); putToBuf_char(c)

static void
ecrireentier (FILE * pf, int i)
{
  i = littleIndian_Or_big_indian (i);
#ifdef __PKWINDOWS__
  putwc (i, (FILE *) pf);
#else
  putw (i, (FILE *) pf);
#endif
}

static void
putToBuf_integer (int anInt)
{
  char cc[4];
  int j;
  ptLONG pi;

  pi = (ptLONG) cc;
  anInt = littleIndian_Or_big_indian (anInt);
  *pi = anInt;
  for (j = 0; j < 4; j++)
    {
      O_BUF.buf[O_BUF.deb + j] = cc[j];
    }
  O_BUF.deb = O_BUF.deb + 4;
}

#define ECRITENTIER(pf, i)  emptyFullBuf(pf, 4); putToBuf_integer(i)

static void
putToBuf_string (int lg, ptchar adchaine)
{
  int j;
  char c;
  // printf("DEBUG ecr stri <<%i:", lg);
  putToBuf_integer (lg);
  for (j = 0; j < lg; j++)
    {
      c = *adchaine;
      // printf("%c", c);
      putToBuf_char (c);
      adchaine = adchaine + 1;
    }
  // printf(">>\n"); fflush(stdout);   
}

static void
put_TMString (FILE * pf, TMString ichaine)
{
  int lg, i;
  ptchar adchaine;
  TMStringCharAdr chaine;

  char c;

  chaine = TMSTRING_CAST_TMSTRINGCHARADR (ichaine);
  // chaine = TMSTRING_CAST_TMSTRINGCHARADR(250);
  lg = TMSTRINGCHARADR_GET_LENGTH (chaine);
  adchaine = TMSTRINGCHARADR_GET_CSTRING (chaine);
  if (O_BUF.deb + lg + 4 > MAXLGBUF)
    {				      /** empty buf in ECRITENTIER, ecritcar??? */
      ECRITENTIER (pf, lg);
      // printf("DEBUG ecr stri <<%i:", lg);
      for (i = 1; i < lg + 1; i++)
	{
	  c = *adchaine;
	  // printf("%c", c);
	  ECRITCAR (pf, c);
	  adchaine = adchaine + 1;
	}
      // printf(">>\n"); fflush(stdout); 
    }
  else
    putToBuf_string (lg, adchaine);
}

/** save the node container (the tree root) of fixson_treeRoot and its language attribute
    @param fixson_treeRoot is assumed to be the alone son of the tree root node
    @return the tree root node
 */

static ptnode
saveTM_treeRoot (FILE * fds, ptfilsd fixson_treeRoot)
{
  ptattribute srcAttribute;
  ptnode NODE_treeRoot;
  TMString icc;
  NODE_treeRoot = FILSD_GET_NODE_CONTAINER_OF_FIRST (fixson_treeRoot);
  // DEBUGMESS_NODE(4, 1, "     in  saveTM_treeRoot, NODE_treeRoot is", NODE_treeRoot);
  srcAttribute = NODE_GET_ATTRIBUTE (NODE_treeRoot);
  icc = getTM_TMString_langName (ATTRIBUTE_GET_TMLANG_VALUE (srcAttribute));
  put_TMString (fds, icc);
  ECRITENTIER (fds, ANYFILS_GET_PHYL (fixson_treeRoot));
  // DEBUG_lit_ecr_int_ln("ecr", "tree root phylum", ANYFILS_GET_PHYL(fixson_treeRoot));
  TT = TT + NODESIZE + ATTRIBUTESIZE;
  return NODE_treeRoot;
}


/** save atomic attributes from srcAttribute until a tree attribute is met or until completion
    @return true if a tree attribute is met during saving
    @return false if the attribute saving is completed
    @param srcFilsAny_res is assigned the first son of the newly created tree descriptor if a tree is met
    @param srcFilsAny_res is assigned the attribute owner if the attribute saving is completed
 */
static boolean
save_attributesUntilTree (FILE * fds, ptattribute srcAttribute, ptfilsAny * srcFilsAny_res)
{
  short attributeOperator;
  TMString icc;

  (*srcFilsAny_res) = ATTRIBUTE_GET_ANYFILS_OWNER (srcAttribute);
  do
    {
      // DEBUGMESS_ATTRIBUTE(4,4,"in save_attributesUntilTree 1W",srcAttribute);
      TT = TT + ATTRIBUTESIZE;
      attributeOperator = ATTRIBUTE_GET_INNEROPER (srcAttribute);
      ECRITCAR (fds, ATTRIBUTE_MARK);
      // DEBUG_lit_ecr_char_ln("ecr", ATTRIBUTE_MARK);
      ECRITENTIER (fds, ATTRIBUTE_GET_TAG (srcAttribute));
      // DEBUG_lit_ecr_int_ln("ecr", "attribute category", ATTRIBUTE_GET_TAG(srcAttribute));
      ECRITENTIER (fds, attributeOperator);
      // DEBUG_lit_ecr_int_ln("ecr", "attribute operator", attributeOperator);
      switch (attributeOperator)
	{
	case OPTROU:
	case OPNIL:
	  break;
	case OPENTIER:
	  ECRITENTIER (fds, (int) (ATTRIBUTE_GET_LONGINT_VALUE (srcAttribute)));	/* SHOULD BE LONG INT */
	  // DEBUG_lit_ecr_int_ln("ecr", "attribute integer value", (int)(ATTRIBUTE_GET_LONGINT_VALUE(srcAttribute)));
	  break;
	case OPCAR:
	  ECRITCAR (fds, ATTRIBUTE_GET_CHAR_VALUE (srcAttribute));	/* MODIFS 6-2-96 */
	  // DEBUG_lit_ecr_char_ln("ecr", ATTRIBUTE_GET_CHAR_VALUE(srcAttribute));
	  break;
	case OPCHAINE:
	  icc = ATTRIBUTE_GET_TMSTRING_VALUE (srcAttribute);
	  put_TMString (fds, icc);
	  break;
	case OPSTAR:
	  break;
	case OPREPERE:
	  TT = TT + REPERESIZE;
	  ECRITENTIER (fds, 1111111111);	/* PLG BIDON to be consistant with lit repere */
	  /* nothing done with the referred to value ? */
	  break;
	case OPARBRE:
	  {
	    ptnode srcNode;
	    (*srcFilsAny_res) = ATTRIBUTE_GET_FILSD_VALUE (srcAttribute);
	    srcNode = saveTM_treeRoot (fds, *srcFilsAny_res);
	    NODE_SET_ATTRIBUTE_CONTAINER (srcNode, srcAttribute);
	    return pk_true;
	  }
	}
      srcAttribute = ATTRIBUTE_GET_SUCC (srcAttribute);
    }
  while (srcAttribute != NIL);
  return pk_false;
}

extern int
sauvegarde (short booleanatt, short booleanrep, FILE * fds, ptfilsAny srcFilsAnyStart)
{
  ptnode NODE_current;
  ptTreeAncestor srcTreeAncestor;
  ptattribute srcAttribute;
  ptfilsAny srcFilsAny, aFilsAny;
  short rootInitialPos, op;
  boolean booleanfin;
  short i, currentArity;

  // DEBUG_src(4, 40, "entering sauvegarde", srcFilsAnyStart);
  TT = 0;
  ECRITENTIER (fds, TT);	/* reservation de la place pour ecrire la taille a la fin */
  // DEBUG_lit_ecr_int_ln("ecr", "fictitious", TT);

  if ((ANYFILS_IS_ROOT (srcFilsAnyStart)))
    {
      srcFilsAny = srcFilsAnyStart;
      rootInitialPos = 0;
    }
  else
    {				/* srcFilsAnyStart is a strict sub tree */
      srcFilsAny = getTM_filsd_root (srcFilsAnyStart);
      rootInitialPos = ANYFILS_GET_POS (srcFilsAnyStart);
      ANYFILS_SET_POS (srcFilsAnyStart, 0);
    }
  // DEBUG_src(4, 40, "     in  sauvegarde 1", srcFilsAnyStart);
  (void) saveTM_treeRoot (fds, srcFilsAny);
  booleanfin = pk_true;
  srcFilsAny = srcFilsAnyStart;	// PLG ?? CHECK
  while ((srcFilsAny != srcFilsAnyStart) || booleanfin)
    {
      // DEBUG_src(4, 40, "     in  sauvegarde 1W", srcFilsAny);
      while ((!(ANYFILS_IS_ROOT (srcFilsAny))) || booleanfin)
	{
	  // DEBUG_src(4, 40, "     in  sauvegarde 1W.1W", srcFilsAny);
	  booleanfin = pk_false;
	  if (ANYFILS_ISIN_LIST (srcFilsAny))
	    {
	      ECRITCAR (fds, FILSLIST_MARK);
	      // DEBUG_lit_ecr_char_ln("ecr", FILSLIST_MARK);
	      TT = TT + FILSLISTSIZE;
	    }
	  else
	    {			/* ANYFILS_ISIN_FIX(srcFilsAny), including tree root */
	      ECRITCAR (fds, FILSD_MARK);
	      // DEBUG_lit_ecr_char_ln("ecr", FILSD_MARK);
	      TT = TT + FILSDSIZE;
	    }
	  op = ANYFILS_GET_INNEROPER (srcFilsAny);
	  ECRITENTIER (fds, op);
	  // DEBUG_lit_ecr_int_ln("ecr", "operator", op);
	  if (ISCOMMONOPER (op))
	    {
	      // DEBUG_mess_int_ln("     in  sauvegarde 1W.1W.1T,common  operator is %li", op);
	      switch (op)
		{
		case OPTROU:
		case OPNIL:
		  break;
		case OPENTIER:
		  ECRITENTIER (fds, (int) ANYFILS_GET_LONGINT_VALUE (srcFilsAny));
		  // DEBUG_lit_ecr_int_ln("ecr", "node integer value", (int)ANYFILS_GET_LONGINT_VALUE(srcFilsAny));
		  break;
		case OPCAR:
		  ECRITCAR (fds, ANYFILS_GET_CHAR_VALUE (srcFilsAny));	/* MODIFS 6-2-96 */
		  // DEBUG_lit_ecr_char_ln("ecr", ANYFILS_GET_CHAR_VALUE(srcFilsAny));
		  break;
		case OPCHAINE:
		  put_TMString (fds, (ANYFILS_GET_TMSTRING_VALUE (srcFilsAny)));
		  break;
		case OPSTAR:
		  break;
		case OPREPERE:
		  TT = TT + REPERESIZE;
		  break;
		case OPARBRE:
		  srcFilsAny = ANYFILS_GET_FILSD_VALUE (srcFilsAny);
		  // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2T-SW_ARBRE", srcFilsAny);
		  (void) saveTM_treeRoot (fds, srcFilsAny);
		  booleanfin = pk_true;
		  break;
		default:
		  printf ("##INTERNAL ERROR## in sauvegarde, common operator is %i\n", op);
		  fflush (stdout);
		  ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
		  return -1;
		}
	    }			/* end ISCOMMONOPER(op): TRUE */
	  else
	    {			/* user defined operator */
	      // DEBUG_mess_int_ln("     in  sauvegarde 1W.1W.1F, inner operator is ", op);
	      // DEBUG_mess_int_ln("     in  sauvegarde 1W.1W.1F, user defined operator is %i", USEROPER(op));
	      NODE_current = ANYFILS_GET_NODE_VALUE (srcFilsAny);
	      srcAttribute = NODE_GET_ATTRIBUTE (NODE_current);
	      // DEBUGMESS_NODE(4, 0, "     in  sauvegarde 1W.1W.1F, NODE_current is", NODE_current);
	      if (NODE_IS_LIST (NODE_current))
		{
		  // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2F-1T", srcFilsAny);
		  currentArity = getTM_listNode_sonNumber (NODE_current);
		  ECRITENTIER (fds, -(currentArity + 1));
		  // DEBUG_lit_ecr_int_ln("ecr", "arity", -(currentArity+1));
		  // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2F-1T-1", srcFilsAny);
		  ECRITENTIER (fds, ANYFILS_GET_PHYL (srcFilsAny));
		  // DEBUG_lit_ecr_int_ln("ecr", "phylum", ANYFILS_GET_PHYL(srcFilsAny));
		  if (currentArity > 0)
		    {
		      srcFilsAny = (ptfilsAny) NODE_GET_FILSLIST_FIRST (NODE_current);
		      booleanfin = pk_true;
		    }
		  TT = TT + NODESIZE + FILSDSIZE;
		}
	      else
		{
		  // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2F-1F", srcFilsAny);
		  currentArity = NODE_GET_ARITY (NODE_current);
		  ECRITENTIER (fds, currentArity);
		  // DEBUG_lit_ecr_int_ln("ecr", "arity", currentArity);
		  srcFilsAny = NODE_GET_FILSD_FIRST (NODE_current);
		  aFilsAny = srcFilsAny;
		  // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2F-1F-1", srcFilsAny);
		  for (i = 1; i <= currentArity; i++)
		    {
		      ECRITENTIER (fds, ANYFILS_GET_PHYL (aFilsAny));
		      // DEBUG_lit_ecr_int_ln("ecr", "phylum", ANYFILS_GET_PHYL(aFilsAny));
		      aFilsAny = FILSD_GET_SUCC (aFilsAny);
		    }
		  booleanfin = pk_true;
		  TT = TT + NODESIZE + currentArity * FILSDSIZE;
		}
	      // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2F-2", srcFilsAny);
	      if ((srcAttribute != NIL_attribute) && (booleanatt))
		{
		  booleanfin = save_attributesUntilTree (fds, srcAttribute, &srcFilsAny);
		  // DEBUG_src(4, 4, "     in  sauvegarde 1W.1W-2F-2T2", srcFilsAny);
		}
	    }			/* end ISCOMMONOPER(op): FALSE */
	  // DEBUG_src(4, 40, "     in  sauvegarde 1W.1W-3", srcFilsAny);
	  if (!booleanfin)
	    {
	      ptfilsd previous_tree = srcFilsAny;
	      if (getCondTM_nextSubtree (srcFilsAny, &srcFilsAny))
		{
		  // DEBUG_src(4, 40, "     in  sauvegarde 1W.1W-3T-1", srcFilsAny);
		  if (previous_tree == srcFilsAny)
		    {
		      printf ("##INTERNAL ERROR## in sauvegarde, there is a loop on subtree adr %li, with operator %i\n", (long int) srcFilsAny, op);
		      fflush (stdout);
		      ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
		      return -1;
		    }
		}
	    }
	}			/* while( (!(ANYFILS_IS_ROOT(srcFilsAny))) || booleanfin */
      // DEBUG_src(4, 4, "     in  sauvegarde 1W.2 SHOULD BE A ROOT", srcFilsAny);
      /* ANYFILS_IS_ROOT(srcFilsAny) */
      if (srcFilsAny != srcFilsAnyStart)
	{
	  srcTreeAncestor = FILSD_GET_ANCESTOR (srcFilsAny);
	  ECRITCAR (fds, SUBTREE_LOADCOMPLETION_MARK);
	  // DEBUG_lit_ecr_char_ln("ecr", SUBTREE_LOADCOMPLETION_MARK);
	  switch (cellCategory ((ptSHORT) srcTreeAncestor))
	    {
	    case 0:		/* acces */
	      // DEBUG_src(4, 4, "     in  sauvegarde 1W.2T2TSW0", srcFilsAny);
	      srcFilsAny = FILSD_GET_ANYFILS_PERE (srcFilsAny);
	      // DEBUG_src(4, 4, "     in  sauvegarde 1W.2T2TSW0-1", srcFilsAny);
	      (void) getCondTM_nextSubtree (srcFilsAny, &srcFilsAny);	/* called with ptfilsd ? */
	      break;
	    case 1:		/* attribut */
	      // DEBUG_src(4, 4, "     in  sauvegarde 1W.2T2TSW1", srcFilsAny);
	      srcAttribute = FILSD_GET_ATTRIBUTE_TREECONTAINER (srcFilsAny);
	      srcFilsAny = ATTRIBUTE_GET_ANYFILS_OWNER (srcAttribute);
	      // DEBUGMESS_ATTRIBUTE(8, 4, "     in  sauvegarde 1W.2T2TSW1-1", srcAttribute);
	      if ((ATTRIBUTE_GET_SUCC (srcAttribute) != NIL) && (booleanatt))
		{
		  srcAttribute = ATTRIBUTE_GET_SUCC (srcAttribute);
		  booleanfin = save_attributesUntilTree (fds, srcAttribute, &srcFilsAny);
		}
	      // DEBUG_src(4, 4, "     in  sauvegarde 1W.2T2TSW3", srcFilsAny);
	      if (!booleanfin)
		(void) getCondTM_nextSubtree (srcFilsAny, &srcFilsAny);
	      // DEBUG_src(4, 4, "     in  sauvegarde 1W.2T2TSW4", srcFilsAny);
	      // DEBUGMESS_ATTRIBUTE(8, 4, "     in  sauvegarde 1W.2T2TSW1-3", srcAttribute);
	      break;
	    default:
	      printf ("##INTERNAL ERROR## in sauvegarde: unexpected cell category\n");
	      fflush (stdout);
	      break;
	    }
	}
      // DEBUGMESS_SONINFIXSIZENODE(4, 0, "     in  sauvegarde 1W.3, srcFilsAny is", srcFilsAny);
    }
  ECRITCAR (fds, MAINTREE_LOADCOMPLETION_MARK);
  // DEBUG_lit_ecr_char_ln("ecr", MAINTREE_LOADCOMPLETION_MARK);
  // DEBUG_src(4, 2, " exiting sauvegarde", srcFilsAnyStart);
  ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
  return (TT);
}

static FILE *
ouvrecr (ptchar fileName)
{
  FILE *fdl;
  fdl = fopen (fileName, "wb");	/* le "b" (binaire) est obligatoire pour PC */
  if (fdl == (FILE *) NULL)
    {
      printf ("###FILE ERROR### opening the file '%s' in writing mode failed\n", fileName);
      fflush (stdout);
      return (FILE *) NULL;
    }
  return fdl;
}

extern FILE *
initsave (ptchar fileName)
{
  FILE *fdl;
  fdl = ouvrecr (fileName);
  O_BUF.deb = 0;
  return fdl;
}


extern void
finsave (FILE * pf, int TT)
{
  fwrite (O_BUF.buf, 1, O_BUF.deb, (FILE *) pf);
  fseek ((FILE *) pf, 0, 0);	/* positionnement au debut du fichier pour ecrire taille */
  // DEBUG_mess_int_ln("finsave ecrit file size = ", TT);
  ecrireentier (pf, TT);
  fclose (pf);
}
