/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          langage.c						             */
/*									     */
/*									     */
/*  Module contenant les primitives de manipulation de la table des langages */
/*									     */
/*	  - attribidlang						     */
/*	  - chercheridlang						     */
/*	  - fermerlang							     */
/*									     */
/*****************************************************************************/


#include "StdTypeDefs.h"
#include "varglobal.h"
#include "f_coderr.h"
#include "li_erreur.h"
#include "langage.h"

/*****************************************************************************/
/*									     */
/*	CHERCHERIDLANG							     */
/*									     */
/* entree : entier icc (chaine)						     */
/* sortie : pointeur sur entree de la table des langages		     */
/* fonction : recherche l'entree correspondant au nom icc dans la table des  */
/*	      langages							     */
/*									     */
/*****************************************************************************/

extern TMLang
chercheridlang (TMString icc)
{
  int i;
  short int vboolean;
  languageTableEntry *lang;

  lang = TAB_LANG;
  vboolean = 0;
  for (i = 0; i <= NBRELANG; i++)
    {
      if (DESCLANG_NAMEIS_TMSTRING (lang, icc))
	return DESCLANG_CAST_TMLANG (lang);
      lang++;
    }
  return (TMLang) (-1);
}


/*****************************************************************************/
/*									     */
/*	getTM_TMString_langName								     */
/*									     */
/* entree : entier num							     */
/* sortie : entier (chaine)						     */
/* fonction : rend la chaine de caracteres denotant la langage de numero num */
/*									     */
/*****************************************************************************/

extern TMString
getTM_TMString_langName (TMLang num)
{
  languageTableEntry *lang;

  // printf("DEBUG entering getTM_TMString_langName, parameter num is %li\n", (long)num); fflush(stdout);
  lang = TMLANG_CAST_DESCLANG (num);
  // printf("DEBUG      in  getTM_TMString_langName, value lang is %li\n", (long)lang); fflush(stdout);
  if (DESCLANG_IS_MISSINGLANG (lang))
    {
      // printf("DEBUG      in  getTM_TMString_langName, lang is missing\n"); fflush(stdout);
      errmess (LANGNONPRESENT);
    }
  return (DESCLANG_GET_TMSTRINGNOM (lang));
}

/*****************************************************************************/
/*									     */
/*	FERMERLANG							     */
/*									     */
/* entree : entier num							     */
/* sortie : 								     */
/* fonction : supprime l'entree dans la table des langages correspondant au  */
/*	      numero num 						     */
/*									     */
/*****************************************************************************/

extern void
fermerlang (TMLang num)
{
  languageTableEntry *lang;

  lang = TMLANG_CAST_DESCLANG (num);
  DESCLANG_SET_MISSINGLANG (lang);
  MOD_LANGARBRELANG (lang, NIL);
  MOD_LANGDICOLANG (lang, NIL);
  MOD_LANGARBREATT (lang, NIL);
  MOD_LANGDICOATT (lang, NIL);
}
