/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          LI_analang.c						     */
/*									     */
/*									     */
/*  Module contenant les primitives de compilation d'un langage :  	     */
/*									     */
/*	  - gendicolang							     */
/*	  - genarbrelang						     */
/*									     */
/*****************************************************************************/

#include <stdio.h>
#include "StdTypeDefs.h"
#include "varglobal.h"
#include "li_erreur.h"
#include "lta_cbase.h"
#include "gm_pile.h"
#include "lta_ctree.h"
#include "li_attri.h"
#include "li_anala.h"

#define NONDEF -2
#define OP 1
#define PH 2
#define NBOPCOMMUN 10		/*les communs sont de 0 a 9 */
#define NBPHCOMMUN 1

#define MAXACCES 500		/* taille max de la table des acces */
#define MAXELEM 250		/* taille max de la table des element */


int nbetiq;			/*nbre d'operateurs effectif comptes a partir de 0 */
int nbphyl;			/*nbre de phylums effectif comptes a partir de 0 */

/********************************/
/* STRUCTURES NECESSAIRES A LA  */
/* CONSTRUCTION DU DICTIONNAIRE */
/********************************/

typedef struct
{
  char car;
  short compt;
  short premier;
  short suivant;
} acces;


typedef struct
{
  short typ;
  short rang;
  long int adr;
  int longueur;
  TMString indchaine;
} elem;


typedef struct
{
  short pos;
  filsd *adresse;
} fcourant;			/* type pointeur sur l'acces au noeud courant de la construction du dico */

acces TACCES[MAXACCES];		/* table des acces aux idents */
elem TABELEM[MAXELEM];		/* table des "valeurs" des idents d'operateurs ou phylums */
short tete, libacces, libelem;	/* indices de gestion des tables */
TMString *table[2];		/* table regroupant TBL_lop et TBL_lphyl */
short type;			/* indice de table : 0 -> operateur, 1 -> phylum */


/*****************************************************************************/
/*                                                                           */
/*                     CONSTRUCTION DU DICTIONNAIRE                          */
/*                                                                           */
/*****************************************************************************/


/*************************************************/
/* creerboite                                    */
/* fonction qui creer un acces avec la lettre c  */
/*************************************************/

static short
creerboite (char c)
{
  TACCES[libacces].compt = 1;
  TACCES[libacces].car = c;
  libacces++;
  return (libacces - 1);
}

/*************************************************/
/* creerelem                                     */
/* fonction de creation d'1 elem de nom (c,l)    */
/*************************************************/

static short
creerelem (short i, ptchar c, int l)
{
  TABELEM[libelem].typ = type + 1;
  TABELEM[libelem].rang = i;
  TABELEM[libelem].adr = TMSTRINGCHARADR_CAST_TMSTRING (c) + 1;
  TABELEM[libelem].longueur = l - 1;
  TABELEM[libelem].indchaine = table[type][i];
  libelem++;
  return (libelem - 1);
}

/*************************************************/
/* creeracces2                                   */
/* fonction de creation, recherche du placement  */
/* d'1 acces secondaire                          */
/*************************************************/

static void
creeracces2 (int i, ptchar c, int l, short iacces)
{
  short j, k, a1, a2;
  ptchar cc;
  int ll;
  char c1, c2;
  short vboolean;

  j = creerelem (i, c, l);
  k = TACCES[iacces].premier;
  TACCES[iacces].compt++;
  ll = TABELEM[k].longueur;
  cc = TMSTRING_CAST_TMSTRINGCHARADR (TABELEM[k].adr);
  vboolean = 0;
  c++;
  l--;

  while (vboolean == 0)
    {
      if (ll <= 0)
	c1 = ' ';
      else
	{
	  c1 = *cc;
	}
      cc++;
      ll--;
      if (l <= 0)
	c2 = ' ';
      else
	{
	  c2 = *c;
	}
      c++;
      l--;

      if (c1 < c2)
	{
	  a1 = creerboite (c1);
	  a2 = creerboite (c2);
	  TABELEM[k].longueur = ll;
	  TABELEM[k].adr = TMSTRINGCHARADR_CAST_TMSTRING (cc);
	  TACCES[a1].premier = k;
	  TACCES[a1].suivant = a2;
	  TABELEM[j].longueur = l;
	  TABELEM[j].adr = TMSTRINGCHARADR_CAST_TMSTRING (c);
	  TACCES[a2].premier = j;
	  TACCES[a2].suivant = -1;
	  TACCES[iacces].premier = a1;
	  vboolean = 1;
	}
      else if (c1 > c2)
	{
	  a1 = creerboite (c1);
	  a2 = creerboite (c2);
	  TABELEM[k].longueur = ll;
	  TABELEM[k].adr = TMSTRINGCHARADR_CAST_TMSTRING (cc);
	  TACCES[a1].premier = k;
	  TACCES[a1].suivant = -1;
	  TABELEM[j].longueur = l;
	  TABELEM[j].adr = TMSTRINGCHARADR_CAST_TMSTRING (c);
	  TACCES[a2].premier = j;
	  TACCES[a2].suivant = a1;
	  TACCES[iacces].premier = a2;
	  vboolean = 1;
	}
      else
	{
	  a1 = creerboite (c1);
	  TACCES[iacces].premier = a1;
	  TACCES[a1].compt = 2;
	  TACCES[a1].suivant = -1;
	  iacces = a1;
	}
    }

}

/*************************************************/
/* creeracces                                    */
/* fonction de creation et placement d'un acces  */
/* secondaire                                    */
/*************************************************/

static void
creeracces (int i, ptchar c, int l, short pos, short iacces)
{
  short j;

  j = creerelem (i, c, l);
  TACCES[libacces].compt = 1;
  TACCES[libacces].car = *c;
  TACCES[libacces].premier = j;
  if (iacces == -1)
    {
      TACCES[libacces].suivant = TACCES[pos].premier;
      TACCES[pos].premier = libacces;
    }
  else
    {
      TACCES[libacces].suivant = TACCES[iacces].suivant;
      TACCES[iacces].suivant = libacces;
    }
  libacces++;
}

/*************************************************/
/* INSERERACCES                                  */
/* fonction de creation et d'insertion d'un acces*/
/* principale dans la liste deja existante       */
/*************************************************/

static void
insereracces (short i, short pos)
{
  TMStringCharAdr ch;
  ptchar c;

  int l;

  ch = TMSTRING_CAST_TMSTRINGCHARADR (table[type][i]);
  c = TMSTRINGCHARADR_GET_CSTRING (ch);
  l = TMSTRINGCHARADR_GET_LENGTH (ch);

  TACCES[libacces].car = *c;
  TACCES[libacces].compt = 1;
  TACCES[libacces].premier = creerelem (i, c, l);

  if (pos == -1)
    {
      TACCES[libacces].suivant = tete;
      tete = libacces;
    }
  else
    {
      TACCES[libacces].suivant = TACCES[pos].suivant;
      TACCES[pos].suivant = libacces;
    }
  libacces++;
}

/*************************************************/
/* RECHERCHEACCES2                               */
/* fonction de recherche de positionnement dans  */
/* une liste d'acces secondaires                 */
/*************************************************/

static int
rechercheacces2 (ptchar c, int l, ptSHORT iacces)
{
  short i;
  short vboolean;

  vboolean = 1;
  i = TACCES[*iacces].premier;
  if ((*c < TACCES[i].car) || (l <= 0))
    {
      *iacces = -1;
    }
  else
    {
      while (*c >= TACCES[i].car)
	{
	  if (TACCES[i].car == *c)
	    {
	      *iacces = i;
	      if (TACCES[i].compt == 1)
		{
		  vboolean = 2;
		  break;
		}
	      else
		{
		  vboolean = 0;
		  break;
		}
	    }
	  if (TACCES[i].suivant == -1)
	    {
	      *iacces = i;
	      break;
	    }
	  else
	    {
	      *iacces = i;
	      i = TACCES[i].suivant;
	    }
	}
    }
  return (vboolean);
}



/*************************************************/
/* RECHERCHEACCES                                */
/* fonction de recherche de positionnement d'un  */
/* acces principale                              */
/*************************************************/

static int
rechercheacces (char c, ptSHORT pos)
{
  short i;

  i = tete;

  if (c < TACCES[tete].car)
    {
      *pos = -1;
      return (0);
    }
  else
    {
      while (c > TACCES[i].car)
	{
	  if (TACCES[i].suivant == -1)
	    {
	      *pos = i;
	      break;
	    }
	  else
	    {
	      *pos = i;
	      i = TACCES[i].suivant;
	    }
	}
      if (c == TACCES[i].car)
	{
	  *pos = i;
	  return (1);
	}
      else
	return (0);
    }
}

/*************************************************/
/* PLACER                                        */
/* fonction de placement d'un ident operateur ou */
/* phylum d'indice i dans la liste ordonnee      */
/*************************************************/

static void
placer (short i)
{
  TMStringCharAdr ch;
  ptchar c, cc;
  short iacces, pos, j;
  short vboolean;
  int l, ll;

  ch = TMSTRING_CAST_TMSTRINGCHARADR (table[type][i]);
  c = TMSTRINGCHARADR_GET_CSTRING (ch);
  l = TMSTRINGCHARADR_GET_LENGTH (ch);
  vboolean = rechercheacces (*c, &iacces);
  if (vboolean == 0)
    insereracces (i, iacces);
  else
    {
      if (TACCES[iacces].compt == 1)
	{
	  pos = TACCES[iacces].premier;
	  ll = TABELEM[pos].longueur;
	  if (ll != 0)
	    {
	      TABELEM[pos].longueur--;
	      cc = TMSTRING_CAST_TMSTRINGCHARADR (TABELEM[pos].adr);
	      j = creerboite (*cc);
	      TABELEM[pos].adr++;
	    }
	  else
	    {
	      j = creerboite (' ');
	    }
	  TACCES[iacces].premier = j;
	  TACCES[j].premier = pos;
	  TACCES[j].suivant = -1;
	}
      vboolean = 0;
      while (vboolean == 0)
	{
	  pos = iacces;
	  TACCES[pos].compt = TACCES[pos].compt + 1;
	  c++;
	  l--;
	  vboolean = rechercheacces2 (c, l, &iacces);
	}
      if (vboolean == 1)
	{
	  creeracces (i, c, l, pos, iacces);
	}
      else
	{
	  creeracces2 (i, c, l, iacces);
	}
    }
}

/*************************************************/
/* CONSLISTEORDO                                 */
/* procedure de generation de la liste ordonnee  */
/* des "mots" du dico                            */
/* LOCALE gendicolang*/
/*************************************************/

static void
conslisteordo (int vboolean)
	  /*indique si dico langage ou attribut */
{
  TMStringCharAdr ch;
  short i;
  ptchar c;
  int n;

  ch = TMSTRING_CAST_TMSTRINGCHARADR (TBL_lop[0]);
  c = TMSTRINGCHARADR_GET_CSTRING (ch);
  TACCES[0].car = *c;
  TACCES[0].compt = 1;
  TACCES[0].premier = 0;
  TACCES[0].suivant = -1;
  c++;
  TABELEM[0].typ = OP;
  TABELEM[0].rang = 0;
  TABELEM[0].adr = TMSTRINGCHARADR_CAST_TMSTRING (c);
  TABELEM[0].longueur = TMSTRINGCHARADR_GET_LENGTH (ch) - 1;
  TABELEM[0].indchaine = TBL_lop[0];

  tete = 0;
  libacces = 1;
  libelem = 1;

  table[0] = TBL_lop;
  type = OP - 1;

  if (vboolean == 0)
    n = nbetiq + TCOMMUN + 1;
  else
    n = nbetiq;

  for (i = 1; i <= n; i++)
    {
      placer (i);
    }

  table[1] = TBL_lphyl;
  type = PH - 1;

  for (i = 0; i <= nbphyl + 1; i++)
    {
      placer (i);
    }

}

/*************************************************/
/* RECHERCHEPIVOT                                */
/* fonction de recherche de l'element pivot de la*/
/* liste de tete premier et de poids poids       */
/* LOCALE gendicolang*/
/*************************************************/

static short
recherchepivot (short premier, short poids, ptSHORT p1)
{
  int p;
  short i, j;

  *p1 = 0;
  i = premier;
  p = (poids >> 1) + 1;
  while (*p1 < p)
    {
      *p1 = *p1 + TACCES[i].compt;
      j = i;
      i = TACCES[i].suivant;
    }

  *p1 = *p1 - TACCES[j].compt;
  return (j);
}

/*************************************************/
/* CONSPIVOT                                     */
/* fonction de construction du node du pivot */
/* m dans le dictionnaire                        */
/* LOCALE gendicolang*/
/*************************************************/

static void
conspivot (short m, short i, short p, short p1)
{
  ptfilsd aFilsd0, aFilsd;
  ptnode aNode0;
  TMString indch;
  short p2;
  short j;

  aFilsd0 = (ptfilsd) sommet (0);
  switch (p)
    {
    case 1:
      j = TACCES[m].premier;
      ANYFILS_SET_PHYL (aFilsd0, TABELEM[j].typ);
      aNode0 = newTM_fixSizeNode (2);

      aFilsd = NODE_GET_FILSD_FIRST (aNode0);
      indch = TABELEM[j].indchaine;
      ANYFILS_MAKE_TMSTRING (aFilsd, indch);

      aFilsd = FILSD_GET_SUCC (aFilsd);
      ANYFILS_SET_PHYL (aFilsd, (TABELEM[j].rang));
      break;
    default:
      ANYFILS_SET_PHYL (aFilsd0, (short) (TACCES[m].car));
      aNode0 = newTM_fixSizeNode (3);

      aFilsd = NODE_GET_FILSD_FIRST (aNode0);
      ANYFILS_SET_INNEROPER (aFilsd, i);
      ANYFILS_SET_PHYL (aFilsd, p1);

      aFilsd = FILSD_GET_SUCC (aFilsd);
      ANYFILS_SET_INNEROPER (aFilsd, m);
      ANYFILS_SET_PHYL (aFilsd, (TACCES[m].compt));

      aFilsd = FILSD_GET_SUCC (aFilsd);
      ANYFILS_SET_INNEROPER (aFilsd, (TACCES[m].suivant));
      ANYFILS_SET_PHYL (aFilsd, (p - p1 - TACCES[m].compt));
      break;
    }
  linkTM_filsd_describedNode (aFilsd0, aNode0, OPFIXE);
}

/*************************************************/
/* dico_suivant                                  */
/* fonction de recherche de l'acces courant du   */
/* dico a construire                             */
/* LOCALE gendicolang*/
/*************************************************/

static void
dico_suivant (void)
{
  ptnode e;
  ptfilsd f;

  f = (ptfilsd) sommet (0);
  e = ANYFILS_GET_NODE_VALUE (f);
  if (e != NIL)
    {
      if (NODE_GET_ARITY (e) == 2)
	e = NIL;
    }
  if (e != NIL)
    f = NODE_GET_FILSD_FIRST (e);
  else
    {
      while (ANYFILS_GET_POS (f) == 3)
	{
	  e = FILSD_GET_NODE_CONTAINER (f);
	  f = NODE_GET_FILSD_DESCRIPTOR (e);
	}
      if (!(ANYFILS_IS_ROOT (f)))
	f = FILSD_GET_SUCC (f);
    }
  depiler ();
  empiler ((ptchar) f);
}

/*****************************************************************************/
/*									     */
/* creerfilslab                                                              */
/*									     */
/* entree : m l'arite d'un noeud decrivant un operateur ou un phylum,        */
/* tblien est le tableau qui associe a un op ou phyl sa description,         */
/* j l'entree dans tblien contenant la liste des codes associes a            */
/* l'operateur ou phylum                                                     */
/* sortie :                                                                  */
/* fonction : complete la description du noeudfixe e avec la liste des codes */
/* des labels associes dans tblien                                           */
/*									     */
/*****************************************************************************/

static ptnode
TM_creerfilslab (ptfilsd aFilsd, int m, ptaritetabint tblien, int i)
{
  int j;
  short phyla;

  if (m == -1)
    {
      phyla = TABCODES[tblien[i].base];
      FILSD_SET_FILSLIST_FIRST (aFilsd, NIL);
      ANYFILS_SET_PHYL (aFilsd, phyla);
    }
  else
    {
      for (j = 0; j <= m - 1; j++)
	{
	  phyla = TABCODES[tblien[i].base + j];
	  ANYFILS_SET_PHYL (aFilsd, phyla);
	  ANYFILS_MAKE_TROU (aFilsd);
	  if (j != (m - 1))
	    aFilsd = FILSD_GET_SUCC (aFilsd);
	}
    }
  return FILSD_GET_NODE_CONTAINER (aFilsd);
}				/*fin de creerfilslab */

static void
creerfilslab (int m, ptaritetabint tblien, int i)
{
  ptfilsd aFilsd;
  ptnode aNode;

  aFilsd = (ptfilsd) depiler ();
  aNode = TM_creerfilslab (aFilsd, m, tblien, i);
  empiler ((ptchar) aNode);
}				/*fin de creerfilslab */



/*****************************************************************************/
/*	GENLIST 							     */
/* entree : cod, peut etre OP ou PHYL selon la liste de label a traiter      */
/*          newSonSize is the size of the new created node                   */
/* Creates a node each son I of which describes the phylum or the operator I */
/* @return the newly created node                                            */
/*									     */
/* LOCALE genarbrelang*/
/*****************************************************************************/


static ptnode
genlist (int cod, int newSonSize)
{
  TMString *tbl;
  ptaritetabint tblien;
  TMString ch;
  int i, aTblienArity, tblMaxIndex;
  ptattribute att;
  ptnode aNode, theBigNode;
  ptfilsd aFilsd;
  boolean isnot_atom;
  // printf("DEBUG entering genlist\n"); fflush(stdout);
  // DEBUG_STACK_FULL("");
  if (cod == OP)
    {
      tbl = TBL_lop;
      tblien = LIEN_opphyl;
      tblMaxIndex = nbetiq + TCOMMUN;
    }
  else
    {
      tbl = TBL_lphyl;
      tblien = LIEN_phylop;
      tblMaxIndex = nbphyl;
    }
  tblMaxIndex = tblMaxIndex + 1;

  theBigNode = newTM_fixSizeNode (newSonSize);	/* moved */
  aFilsd = NODE_GET_FILSD_FIRST (theBigNode);
  /* STACK = STACK1.FIRSTSONINBIGNODE */
  for (i = 0; i <= tblMaxIndex; i++)
    {
    /*** creer la chaine qui denote l'operateur ou le phylum  ***/
      ch = tbl[i];
      aTblienArity = tblien[i].arite;
      isnot_atom = aTblienArity != 0;
      if (isnot_atom)
	{
	  if (aTblienArity == -1)
	    aNode = newTM_listNode ();
	  else
	    aNode = newTM_fixSizeNode (aTblienArity);
	}
      else
	aNode = newTM_fixSizeNode (0);
      att = attachTM_newAttribute0 (aNode, OPCHAINE, (untypedValueType) ch);
      if (isnot_atom)
	{
	  aFilsd = NODE_GET_FILSD_FIRST (aNode);
	  aNode = TM_creerfilslab (aFilsd, aTblienArity, tblien, i);
	}
      linkTM_filsd_describedNode (aFilsd, aNode, OPLISTE);
      if (i != tblMaxIndex)
	aFilsd = FILSD_GET_SUCC (aFilsd);
    }				/* inner loop for ... */
  // printf("DEBUG  exiting genlist\n"); fflush(stdout);
  return theBigNode;
}				/*fin de genlist */




/*****************************************************************************/
/*									     */
/*     GENDICOLANG							     */
/*									     */
/* entree : 								     */
/* sortie : pointeur sur l'acces au premier fils du dico                     */
/* fonction : construction de l'arbre dictionnaire d'1 langage a partir des  */
/*	      tables TBL_lop et TBL_lphyl				     */
/*									     */
/* APPELEE DANS CONSLNTA */
/*****************************************************************************/

extern void
gendicolang (int vboolean)
	  /* indique si dico de langage ou d'attribut */
{
  short i, m;
  short p1, p;
  ptfilsd f;

  conslisteordo (vboolean);
  empiler ((ptchar) newTM_fixSonRoot_emptyTree ());
  f = (ptfilsd) sommet (0);

  if (vboolean == 0)
    p = nbetiq + nbphyl + TCOMMUN + 4;
  else
    p = nbetiq + 1;

  m = recherchepivot (tete, p, &p1);
  conspivot (m, tete, p, p1);
  dico_suivant ();
  f = (ptfilsd) sommet (0);

  while (!(ANYFILS_IS_ROOT (f)))
    {
      i = ANYFILS_GET_INNEROPER (f);
      p = ANYFILS_GET_PHYL (f);
      if ((ANYFILS_GET_POS (f) == 2) && (p != 1))
	{
	  i = TACCES[i].premier;
	}
      if (p != 0)
	{
	  m = recherchepivot (i, p, &p1);
	  conspivot (m, i, p, p1);
	}
      else
	{
	  ANYFILS_MAKE_TROU (f);
	}
      dico_suivant ();
      f = (ptfilsd) sommet (0);
    }
}

/*****************************************************************************/
/*									     */
/*     GENARBRELANG							     */
/*									     */
/* entree : 								     */
/* sortie :                                                                  */
/* fonction : generation de l'arbre du langage a partir des tables TBL_lop,  */
/*	      TBL_lphyl, LIEN_opphyl et LIEN_phylop, le resultat est en      */
/* @return pointeur sur le 1er filsdirect du noeudfixe(2).	     */
/*									     */
/* APPELEE DANS CONSLNTA */
/*****************************************************************************/

ptfilsd
genarbrelang (void)
{
  ptnode aNode;
  ptfilsd aFilsdRoot, aFilsd1, aFilsd2;

  // printf("DEBUG entering genarbrelang\n"); fflush(stdout);
  aFilsdRoot = newTM_fixSonRoot_emptyTree ();
  aNode = newTM_fixSizeNode (2);
  linkTM_filsd_describedNode (aFilsdRoot, aNode, OPLISTE);

  aFilsd1 = NODE_GET_FILSD_FIRST (aNode);
  aNode = genlist (OP, TCOMMUN + nbetiq + 2);
  linkTM_filsd_describedNode (aFilsd1, aNode, OPLISTE);

  aFilsd2 = FILSD_GET_SUCC (aFilsd1);
  aNode = genlist (PH, nbphyl + 2);
  linkTM_filsd_describedNode (aFilsd2, aNode, OPLISTE);

  // DEBUGMESS_SONINFIXSIZENODE(4,6,"exiting genarbrelang", aFilsdRoot);

  return aFilsdRoot;
}				/*fin de genarbrelang */
