/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          LI_erreur.c							     */
/*									     */
/*									     */
/*  Module contenant les primitives d'impression des messages d'erreur :     */
/*									     */
/*	  - errmess							     */
/*	  - gm_err							     */
/*	  - cs_err							     */
/*	  - nv_err							     */
/*	  - lg_err							     */
/*	  - fl_err							     */
/*									     */
/*  Modules inclus : 							     */
/*									     */
/*	  - f_coderr.h							     */
/*									     */
/*  Externes :								     */
/*									     */
/*****************************************************************************/


#include "boolean.h"
#include "system.h"
#include "typesbase.h"
#include "f_coderr.h"
#include "error.h"
#include "li_erreur.h"
#include "lta_ctree.h"

int ERRBOOL;			/*variable globale contenant le code associe a une erreur */
	     /* cette variable est accessible par tous les modules     */

/*****************************************************************************/
/*                                                                           */
/*     gm_err                                                                */
/*                                                                           */
/* Fonction de traitement des cas d'erreur lies a la gestion memoire         */
/*                                                                           */
/*****************************************************************************/

static void
gm_err (void)
{
  DEBUG_MEMORY ("entering gm_err");
  fatal_tree << "String/Tree Memory error: ";

  switch (ERRBOOL)
    {
    case GM_SATARBRE:
      (fatal_tree << "the area for the heap of the trees is full.")++;

    case GM_SATCHAINE:
      (fatal_tree << "the area for the heap of the strings is full.")++;

    case GM_SATCOLL:
      (fatal_tree << "the area for the collisions is full.")++;

    case GM_PILEVIDE:
      (fatal_tree << "empty stack.")++;

    case GM_PILEPLEINE:
      (fatal_tree << "stack overflow.")++;

    case GM_LANG:
      (fatal_tree << "the languages table is full.")++;

    case LANGNONPRESENT:
      (fatal_tree << "a language is not present in the languages table.")++;

    case GM_REGPLEIN:
      (fatal_tree << "the area for the registers is full.")++;

    case GM_REGVIDE:
      (fatal_tree << "the area for the registers is empty.")++;

    default:
      (fatal_tree << "the error number=" << ERRBOOL << "is not managed.")++;
    }
}



/*****************************************************************************/
/*                                                                           */
/*     nv_err                                                                */
/*                                                                           */
/* Fonction de traitement des cas d'erreur lies a la navigation dans les     */
/* arbres                                                                    */
/*                                                                           */
/*****************************************************************************/

static void
nv_err (void)
{
  fatal_tree << "Error in the exploring of a tree:";	//erreur lie a la navigation dans l'arbre 

  switch (ERRBOOL)
    {

    case NV_DEC:
      (fatal_tree << "Attribute with a un ayant une valeur de type inimprimable).")++;	// it does not occur

    default:
      (fatal_tree << "the error number=" << ERRBOOL << "is not managed.")++;

    }
}



/*****************************************************************************/
/*                                                                           */
/*     cs_err                                                                */
/*                                                                           */
/* Fonction de traitement des cas d'erreur lies a la construction ou a la    */
/* modification d'arbre                                                      */
/*                                                                           */
/*****************************************************************************/

static void
cs_err (void)
{
  fatal_tree << "Error during the construction/modification of a tree: ";

  switch (ERRBOOL)
    {

    case CS_INCOMP:
      (fatal_tree << "the operator is incompatible with the declared phylum attached to a sub tree of a tree.")++;	//
      // operateur incompatible avec le phylum declare a l'attachement d'un sous_arbre a un arbre).")++;

    case CS_INCOMPATT:
      (fatal_tree << "the code of an attribute is incompatible with the operator of the node that must be attributed.")++;
      //code d'attribut incompatible avec l'operateur associe au noeud destine a etre attribue).")++;

    case CS_OPNONDEF:
      (fatal_tree << "undefined operator in the current language.")++;
      // "(operateur non defini dans le langage courant).")++;

    case CS_OPATTENDU:
      (fatal_tree << "waiting for a name of an operator during the reading of a tree.")++;
      // "(nom d'operateur attendu lors de la lecture d'un arbre).")++;

    case CS_CARINV:
      (fatal_tree << "unexpected character during the reading of a tree, '(' or `#' waited.")++;
      // "(caractere invalide lors de la lecture d'un arbre `(' ou `#' attendu).")++;

    case CS_POMANQ:
      (fatal_tree << "missing `(' during the reading of a tree.")++;
      // "`(' manquante lors de la lecture d'un arbre.")++;

    case CS_PFATT:
      (fatal_tree << "`(' waited.")++;

    default:
      (fatal_tree << "the error number=" << ERRBOOL << "is not managed.")++;
    }
}



/*****************************************************************************/
/*       lg_err                                                              */
/*                                                                           */
/* Fonction de traitement d'erreur , lors de la compilation d'un arbre       */
/* langage. Les cas d'erreurs correspondent au controle des definitions      */
/* des labels avant generation du code langage                               */
/*                                                                           */
/*****************************************************************************/

static void
lg_err (void)
{
  fatal_tree << "Error - compiling a laguage: ";

  switch (ERRBOOL)
    {
    case LG_VALNONDEF:
      (fatal_tree << "Null value for a label.")++;
      // "(valeur d'un label nulle)."
      break;

    case LG_DDEFLAB:
      (fatal_tree << "Operator or phylum already defined, ignored.")++;
      //operateur ou phylum deja defini, donc ignore)."

    case LG_LABNONDEF:
      (fatal_tree << "(a used label is not defined (control of the links).")++;
      // "(label utilise non defini au controle des liens)."

    default:
      (fatal_tree << "the error number=" << ERRBOOL << "is not managed.")++;
    }
}

/*****************************************************************************/
/*                                                                           */
/*    fl_err                                                                 */
/*                                                                           */
/* Fonction de traitement d'erreur relatif au filtrage d'un arbre            */
/* a un schema  quelconque                                                   */
/*                                                                           */
/*****************************************************************************/

static void
fl_err (void)
{
  fatal_tree << "Error connected to the filtering of a tree in a schema: ";
  //erreur liee au filtrage d'un arbre  un schema ";
  switch (ERRBOOL)
    {
    case FL_SSARBREVID:
      (fatal_tree << "the filtering fails with an empty sub-tree.")++;
      // "(le filtrage echoue a un sous_arbre vide qui rend l'arbre et le schema incomparable).")++;

    case FL_NOMOP:
      (fatal_tree << "the filtering fails on a name of an operator.")++;
      //le filtrage echoue a un nom d'operateur)."

    case FL_CASLIST:
      (fatal_tree << "the filtering fails on a comparison of two trees of list type without the same arity.")++;
      //       "(le filtrage echoue a une comparaison de deux arbres de type liste n'ayant pas la meme taille.")++;

    default:
      (fatal_tree << "the error number=" << ERRBOOL << "is not managed.")++;
      // erreur liee au filtrage d'arbre non prevu).")++;
      break;
    }
}




/*****************************************************************************/
/*                                                                           */
/*    div_err                                                                */
/*                                                                           */
/* Fonction de traitement d'erreurs diverses				     */
/*                                                                           */
/*****************************************************************************/

static void
div_err (void)
{
  switch (ERRBOOL)
    {
    case LGABSENT:
      (fatal_tree << "unknown language.")++;
      // langage non connu du systeme.")++;
      break;
    case CHLONG:
      (fatal_tree << "too big string.")++;
      // chaine trop longue.")++;
      break;
    }
}




/*****************************************************************************/
/*                                                                           */
/*     errmess                                                               */
/*                                                                           */
/* Fonction d'affichage des messages d'erreur. Cette fonction fait appel aux */
/* differents traitements d'erreur suivant la valeur du code erreur ERRBOOL  */
/*                                                                           */
/*****************************************************************************/

extern void
errmess (int errCode)
{
  ERRBOOL = errCode;
  if ((ERRBOOL >= 1) && (ERRBOOL <= 20))
    gm_err ();
  else if (ERRBOOL <= 40)
    cs_err ();
  else if (ERRBOOL <= 50)
    nv_err ();
  else if (ERRBOOL <= 60)
    lg_err ();
  else if (ERRBOOL <= 70)
    fl_err ();
  else if (ERRBOOL <= 80)
    div_err ();
  else
    (fatal_tree << "the error number=" << ERRBOOL << "is not managed.")++;
}

extern void
setFatalTree (const char *mess)
{
  (fatal_tree << mess)++;
}
