/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*                                                                           */
/*          LI_langage.c						     */
/*                                                                           */
/*                                                                           */
/*  Module contenant les primitives d'acces aux informations de l'arbre  du  */
/*  langage :                                                                */
/*	     - acceslab							     */
/*	     - aritelab							     */
/*           - ariteop                                                       */
/*   	     - appartientelemlab					     */
/*           - appartientopphyl                                              */
/*                                                                           */
/*  et les primitives d'acces aux informations de l'arbre langage a travers  */
/*  le dictionnaire :                                                        */
/*           - recherche_dicoC                                               */
/*                                                                           */
/*****************************************************************************/

#include <string.h>
#include "system.h"
#include "varglobal.h"
#include "aslnta.h"
#include "li_langa.h"

/*****************************************************************************/
/*      acceslab                                                             */
/*                                                                           */
/* entree : le code d'un label, et un registre R contenant le pointeur       */
/*          vers la representation d'une liste de labels                     */
/* sortie : pointeur sur l'acces au noeud description du label               */
/*                                                                           */
/*****************************************************************************/

ptfilsd
acceslab (short codlab, ptfilsd R)
{
  ptfilsd f;
  ptnode e;

  e = ANYFILS_GET_NODE_VALUE (R);
  f = NODE_GET_FILSD_FIRST (e);
  f = FILSD_GET_NSUCC (f, codlab);
  return (f);
}


/*****************************************************************************/
/*                                                                           */
/*     aritelab                                                              */
/*                                                                           */
/* entree : un code du label                                                 */
/*          et le registre R qui pointe vers la liste des labels             */
/* sortie : l'arite du label de code codlab                                  */
/*                                                                           */
/*****************************************************************************/

int
aritelab (short codlab, ptfilsd R)
{
  ptnode e;
  ptfilsd f;
  f = acceslab (codlab, R);
  e = ANYFILS_GET_NODE_VALUE (f);
  return (NODE_GET_ARITY (e));
}


/*****************************************************************************/
/*                                                                           */
/*     ariteop                                                               */
/*                                                                           */
/* entree : un code d'operateur                                              */
/* sortie : l'arite de l'operateur de code op                                */
/*                                                                           */
/*****************************************************************************/

int
ariteop (short op)
{
  short a;
  ptfilsd ROP;
  languageTableEntry *RLANG;

  RLANG = TMLANG_CAST_DESCLANG (ENVLANG);
  ROP = NODE_GET_FILSD_FIRST (ANYFILS_GET_NODE_VALUE (LANGARBRELANG (RLANG)));
  a = aritelab (op, ROP);
  return (a);
}


/*****************************************************************************/
/*                                                                           */
/*     appartientelemlab                                                     */
/*                                                                           */
/* entree : un code d'un element et un code du label , et R                  */
/* sortie : true si l'element de code codelem appartient au label codlab     */
/*          false sinon                                                       */
/*                                                                           */
/*****************************************************************************/

int
appartientelemlab (short codelem, short codlab, ptfilsd R)
{
  ptnode e;
  ptfilsd f;
  int i, T;

  T = pk_false;
  f = acceslab (codlab, R);
  e = ANYFILS_GET_NODE_VALUE (f);
  f = NODE_GET_FILSD_FIRST (e);
  for (i = 1; i <= NODE_GET_ARITY (e); i++)
    {
      if (ANYFILS_GET_PHYL (f) == codelem)
	{
	  T = pk_true;
	  break;
	}
      else
	f = FILSD_GET_SUCC (f);
    }
  return (T);
}


/*****************************************************************************/
/*                                                                           */
/*     appartientopphyl                                                      */
/*                                                                           */
/* entree : un code d'operateur et un code de phylum                         */
/* sortie : true si l'operateur de code op appartient au phylum de code phyl */
/*          false sinon                                                       */
/*                                                                           */
/*****************************************************************************/

int
appartientopphyl (short op, short phyl)
{
  int vboolean;
  ptfilsd RPHYL;
  languageTableEntry *RLANG;

  if (phyl == CODEVERY)
    return pk_true;

  RLANG = TMLANG_CAST_DESCLANG (ENVLANG);
  RPHYL = NODE_GET_FILSD_FIRST (ANYFILS_GET_NODE_VALUE (LANGARBRELANG (RLANG)));
  RPHYL = FILSD_GET_SUCC (RPHYL);
  vboolean = appartientelemlab (op, phyl, RPHYL);
  return (vboolean);
}


void
recherche_dicoC (ptSHORT t, ptSHORT r)
{
  ptnode e;
  ptfilsd dico, f;
  int i, vboolean, ich;
  int j;			/* compteur de caracteres deja reconnus dans le parcours du dico */
  ptchar p;
  char car;
  ptchar ch;
  int n;
  languageTableEntry *RLANG;

  p = getCurrentString ();
  n = strlen (p);

  RLANG = TMLANG_CAST_DESCLANG (ENVLANG);
  dico = LANGDICOLANG (RLANG);
  vboolean = 0;
  car = *p;
  j = 0;

  while (vboolean == 0)
    {
      if (car == ANYFILS_GET_PHYL (dico))
	{
	  i = 1;
	  n = n - 1;
	  j++;
	  if (n == 0)
	    car = ' ';
	  else
	    {
	      p++;
	      car = *p;
	    }
	}
      else if (car < ANYFILS_GET_PHYL (dico))
	i = 0;
      else
	{
	  i = 2;
	}

      e = ANYFILS_GET_NODE_VALUE (dico);
      dico = NODE_GET_FILSD_FIRST (e);
      dico = FILSD_GET_NSUCC (dico, i);
      e = ANYFILS_GET_NODE_VALUE (dico);
      if (e == NIL)
	vboolean = 1;
      else if (NODE_GET_ARITY (e) == 2)
	{
	  f = NODE_GET_FILSD_FIRST (e);
	  if (n == 0)
	    vboolean = 2;
	  else
	    {
	      ich = ANYFILS_GET_LONGINT_VALUE (f);
	      ch = TMSTRING_CAST_TMSTRINGCHARADR (ich);
	      ch = TMSTRINGCHARADR_GET_CSTRING (ch);
	      ch = ch + j;	/* adr du reste de la chaine a reconnaitre */

	      if (strcmp (ch, p) == 0)
		vboolean = 2;
	      else
		vboolean = 1;
	    }
	}
    }				/*while */

  switch (vboolean)
    {
    case 1:
      *t = 0;
      *r = 0;
      break;
    case 2:
      *t = ANYFILS_GET_PHYL (dico);
      f = FILSD_GET_SUCC (f);
      *r = (longInteger) (ANYFILS_GET_PHYL (f));
      break;
    }

}
