/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          LI_nav.c							     */
/*									     */
/*									     */
/*  Module contenant les primitives de navigation : 			     */
/*									     */
/*         - hasTM_filsd_nSons                                               */
/*         _ getTM_filsd_nchild                                              */

/*         - liste_vide     						     */
/*         - pere_sv        						     */
/*         - hasTM_successorInList					     */
/*         - pred_fixe_sv   						     */
/*         - a_pred_fixe    					 	     */
/*         - pred_liste_sv  						     */
/*         - a_pred_liste   						     */
/*									     */
/*****************************************************************************/
#include "boolean.h"
#include "StdTypeDefs.h"
#include "typesbase.h"
#include "gm_pile.h"
#include "lta_ctree.h"
#include "li_nav.h"

extern short
hasTM_filsd_nSons (ptfilsd aFilsd, int n)
{
  ptnode itsNode;

  if (ANYFILS_HAS_USEROPERATOR (aFilsd))
    {
      itsNode = ANYFILS_GET_NODE_VALUE (aFilsd);
      if (itsNode == NIL_node)
	return pk_false;
      return (NODE_GET_ARITY (itsNode) >= n);
    }
  return pk_false;
}

extern ptfilsd
getTM_filsd_nchild (ptfilsd aFilsd, int n)
{
  ptnode itsNodeValue;

  itsNodeValue = ANYFILS_GET_NODE_VALUE (aFilsd);
  aFilsd = NODE_GET_FILSD_FIRST (itsNodeValue);
  aFilsd = FILSD_GET_NSUCC (aFilsd, (n - 1));
  return aFilsd;
}


/*****************************************************************************/
/*									     */
/* 	PERE_SV								     */
/*									     */
/* entree : 								     */
/* fonction : delivre en sommet de pile le pere du noeud qui se trouvait en  */
/*	      sommet de pile						     */
/*									     */
/*****************************************************************************/

extern void
pere_sv (void)
{
  ptfilsAny filsAnyParam, filsAnyRes;

  filsAnyParam = (ptfilsAny) depiler ();
  if (ANYFILS_ISIN_LIST (filsAnyParam))
    filsAnyRes = FILSLIST_GET_ANYFILS_PERE ((ptfilslist) filsAnyParam);
  else
    filsAnyRes = FILSD_GET_ANYFILS_PERE ((ptfilsd) filsAnyParam);
  empiler ((ptchar) filsAnyRes);

}

/*****************************************************************************/
/*									     */
/*	A_SUCC_LISTE							     */
/*									     */
/* entree :								     */
/* sortie : booleaneen							     */
/* fonction : delivre true si fl a un suiv */
/*	      false sinon						     */
/*									     */
/*****************************************************************************/

extern boolean
hasTM_successorInList (ptfilslist fl)
{
  if (ANYFILS_ISIN_LIST (fl))
    return (FILSLIST_GET_SUCC (fl) != NIL_filslist);
  return pk_false;
}



/*****************************************************************************/
/*									     */
/*	PRED_FIXE_SV							     */
/*									     */
/* entree : entier n							     */
/* fonction : delivre en sommet de pile le nieme frere gauche du fils direct */
/*	      qui se trouvait en sommet de pile				     */
/*									     */
/*****************************************************************************/

extern void
pred_fixe_sv (int n)
{
  ptfilsd f;

  f = (ptfilsd) depiler ();
  f = FILSD_GET_NSUCC (f, (-n));
  empiler ((ptchar) f);
}



/*****************************************************************************/
/*									     */
/* 	A_PRED_FIXE							     */
/*									     */
/* entree : entier n							     */
/* sortie : booleaneen							     */
/* fonction: delivre vrai si le fils direct en sommet de pile a un nieme     */
/*	     frere gauche, false sinon					     */
/*									     */
/*****************************************************************************/

extern short
a_pred_fixe (int n)
{
  return (ANYFILS_GET_POS ((ptfilsd) sommet (0))) <= n;
}



/*****************************************************************************/
/*									     */
/*	PRED_LISTE_SV							     */
/*									     */
/* entree :								     */
/* fonction : delivre en sommet de pile le predecesseur de l'element de liste*/
/*	      qui se trouvait en sommet de pile				     */
/*									     */
/*****************************************************************************/

extern void
pred_liste_sv (void)
{
  ptfilslist fl;

  fl = (ptfilslist) depiler ();
  fl = FILSLIST_GET_PRED (fl);
  empiler ((ptchar) fl);
}



/*****************************************************************************/
/*									     */
/*	A_PRED_LISTE							     */
/*									     */
/* entree :								     */
/* sortie : booleaneen							     */
/* fonction : delivre true si l'element de liste en sommet de pile a un frere*/
/*	      gauche							     */
/*									     */
/*****************************************************************************/

extern short
a_pred_liste (void)
{
  ptfilslist fl;

  fl = (ptfilslist) sommet (0);

  if (ANYFILS_ISIN_LIST (fl))
    return (FILSLIST_GET_PRED (fl) != NIL);
  return pk_false;
}
