/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*****************************************************************************/
/*                                                                           */
/*          lta_cbase.c                                                      */
/*                                                                           */
/*  Module contenant les primitives de construction des objets de base :     */
/*                                                                           */
/*****************************************************************************/

#include "StdTypeDefs.h"
#include "varglobal.h"
#include "gm_pile.h"
#include "gm_arbre.h"
#include "li_attri.h"
#include "lta_cbase.h"

#define TREEMEMORY_ALLOCATE_UNARY() (ptnode)allouerarbre(NODESIZE + FILSDSIZE)
#define TREEMEMORY_ALLOCATE_NARY(n) (ptnode)allouerarbre(NODESIZE + ((n) * FILSDSIZE))
#define TREEMEMORY_ALLOCATE_FILSLIST() (ptfilslist)allouerarbre(FILSLISTSIZE)
#define TREEMEMORY_ALLOCATE_ATTRIBUT() (ptattribute)allouerarbre(ATTRIBUTESIZE)
#define TREEMEMORY_ALLOCATE_REPERE() (ptrepere)allouerarbre(REPERESIZE)

/* 
   ========= ================== =========  Elementary data structures
 */
ptrepere
newTM_repere (void)		/* TREEMEMORY_ALLOCATE_REPERE() */
{
  ptrepere newRepere;

  // printf("DEBUG entering newTM_repere: LIB_ARBRE is %li\n", (long int)(ptchar)LIB_ARBRE); fflush(stdout);
  newRepere = TREEMEMORY_ALLOCATE_REPERE ();
  REPERE_SET_DIRECT_ACCESS (newRepere);
  REPERE_SET_TMLANG (newRepere, ENVLANG);
  REPERE_SET_FILSD_REFERED (newRepere, NIL);
  // DEBUGLIGHTMESS_REPERE("\n created repere is ", newRepere);
  // printf("DEBUG  exiting newTM_repere: LIB_ARBRE is %li, SOMMET_PILE %li available space is %li\n", (long int)(ptchar)LIB_ARBRE, (long int)(ptchar)SOMMET_PILE, (long int)((ptchar) SOMMET_PILE-(ptchar)LIB_ARBRE)); fflush(stdout);
  return newRepere;
}

/* return a new allocated fix size node descriptor with initialized content
   see scheme in doc TreeDataStructure
 */
ptnode
newTM_fixSizeNode (short arity)	/* TREEMEMORY_ALLOCATE_NARY(n) */
{
  ptnode newNode;
  ptfilsd itsFilsd;
  int position;

  // printf("DEBUG entering newTM_fixSizeNode: LIB_ARBRE is %li, arity is %i\n", (long int)(ptchar)LIB_ARBRE, arity); fflush(stdout);
  newNode = TREEMEMORY_ALLOCATE_NARY (arity);
  NODE_SET_ARITY (newNode, arity);
  NODE_LINK_FILSD_DESCRIPTOR (newNode, NIL);
  NODE_CLEAR_ATTRIBUTE (newNode);
  itsFilsd = NODE_GET_FILSD_FIRST (newNode);
  for (position = 1; position <= arity; position++)
    {
      ANYFILS_SET_PHYL (itsFilsd, CODEVERY);
      ANYFILS_SET_POS (itsFilsd, position);
      ANYFILS_MAKE_TROU (itsFilsd);
      itsFilsd = FILSD_GET_SUCC (itsFilsd);
    }
  // DEBUGMESS_NODE(5, 1, "created fixSizeNode is", newNode);
  // printf("DEBUG  exiting newTM_fixSizeNode: LIB_ARBRE is %li, SOMMET_PILE %li available space is %li\n", (long int)(ptchar)LIB_ARBRE, (long int)(ptchar)SOMMET_PILE, (long int)((ptchar) SOMMET_PILE-(ptchar)LIB_ARBRE)); fflush(stdout);
  return newNode;
}

/* return a new allocated list node descriptor with initialized content
 */
ptnode
newTM_listNode (void)		/* TREEMEMORY_ALLOCATE_UNARY() */
{
  ptnode newNode;
  ptfilsd listHeaderDescriptor;

  // printf("DEBUG entering newTM_listNode: LIB_ARBRE is %li\n", (long int)(ptchar)LIB_ARBRE); fflush(stdout);
  newNode = TREEMEMORY_ALLOCATE_UNARY ();
  NODE_SET_ARITY (newNode, -1);
  NODE_LINK_FILSD_DESCRIPTOR (newNode, NIL);
  NODE_CLEAR_ATTRIBUTE (newNode);
  listHeaderDescriptor = NODE_GET_FILSD_FIRST (newNode);
  ANYFILS_SET_PHYL (listHeaderDescriptor, CODEVERY);
  ANYFILS_SET_POS (listHeaderDescriptor, 1);
  ANYFILS_MAKE_TROU (listHeaderDescriptor);
  // DEBUGMESS_NODE(5, 1, "created listNode is", newNode);
  // printf("DEBUG  exiting newTM_listNode: LIB_ARBRE is %li, SOMMET_PILE %li available space is %li\n", (long int)(ptchar)LIB_ARBRE, (long int)(ptchar)SOMMET_PILE, (long int)((ptchar) SOMMET_PILE-(ptchar)LIB_ARBRE)); fflush(stdout);
  return newNode;
}

/* return a new allocated intermediate descriptor for a node ina a list node with initialized content
 */
ptfilslist
newTM_filslist (void)		/* TREEMEMORY_ALLOCATE_FILSLIST() */
{
  ptfilslist newFilslist;

  // printf("DEBUG entering newTM_filslist: LIB_ARBRE is %li\n", (long int)(ptchar)LIB_ARBRE); fflush(stdout);
  newFilslist = TREEMEMORY_ALLOCATE_FILSLIST ();
  ANYFILS_SET_PHYL (newFilslist, CODEVERY);
  ANYFILS_SET_POS (newFilslist, -1);
  ANYFILS_MAKE_TROU (newFilslist);
  FILSLIST_SET_NODE_CONTAINER (newFilslist, NIL);
  FILSLIST_SET_PRED (newFilslist, NIL);
  FILSLIST_SET_SUCC (newFilslist, NIL);

  // DEBUGMESS_SONINLISTNODE(5, 1, "created filslist is", newFilslist);
  // printf("DEBUG  exiting newTM_filslist: LIB_ARBRE is %li, SOMMET_PILE %li available space is %li\n", (long int)(ptchar)LIB_ARBRE, (long int)(ptchar)SOMMET_PILE, (long int)((ptchar) SOMMET_PILE-(ptchar)LIB_ARBRE)); fflush(stdout);
  return newFilslist;
}

/** create and return a new OPTROU attribute, tagged by tagAttribute
    assume correct tagAttribute parameter (not checked)
 */
ptattribute
newTM_attribute (short tagAttribute)	/* TREEMEMORY_ALLOCATE_ATTRIBUT() */
{
  ptattribute newAttribute;

  // printf("DEBUG entering newTM_attribute: LIB_ARBRE is %li\n", (long int)(ptchar)LIB_ARBRE); fflush(stdout);
  newAttribute = TREEMEMORY_ALLOCATE_ATTRIBUT ();
  ATTRIBUTE_SET_TAG (newAttribute, tagAttribute);
  ATTRIBUTE_MAKE_TROU (newAttribute);
  ATTRIBUTE_ISOLATE (newAttribute);
  // printf("DEBUG  exiting newTM_attribute: LIB_ARBRE is %li, SOMMET_PILE %li available space is %li\n", (long int)(ptchar)LIB_ARBRE, (long int)(ptchar)SOMMET_PILE, (long int)((ptchar) SOMMET_PILE-(ptchar)LIB_ARBRE)); fflush(stdout);
  return newAttribute;
}

/* 
   ========= ================== =========  Tree initial data structure
 */
/* return a new allocated tree descriptor with initialized content and attribute
   see scheme in doc TreeDataStructure
 */
ptfilsd
newTM_fixSonRoot_emptyTree (void)	/* TREEMEMORY_ALLOCATE_{UNARY(), ATTRIBUT()} */
{
  ptnode theNewTreeNodeRoot;
  ptattribute languageAttribute;
  ptfilsd itsFixSingleSonRoot;

  // printf("DEBUG entering newTM_fixSonRoot_emptyTree: LIB_ARBRE is %li\n", (long int)(ptchar)LIB_ARBRE); fflush(stdout);
  theNewTreeNodeRoot = TREEMEMORY_ALLOCATE_UNARY ();

  languageAttribute = attachTM_newAttribute0 (theNewTreeNodeRoot, OPENTIER, (untypedValueType) ENVLANG);

  NODE_SET_ARITY (theNewTreeNodeRoot, 1);
  NODE_LINK_FILSD_DESCRIPTOR (theNewTreeNodeRoot, NIL);

  itsFixSingleSonRoot = NODE_GET_FILSD_FIRST (theNewTreeNodeRoot);
  ANYFILS_SET_PHYL (itsFixSingleSonRoot, CODEVERY);
  ANYFILS_SET_POS (itsFixSingleSonRoot, 0);
  ANYFILS_MAKE_TROU (itsFixSingleSonRoot);
  // DEBUGMESS_NODE(5, 4, "In newTM_fixSonRoot_emptyTree, ROOT IS ", theNewTreeNodeRoot);
  // printf("DEBUG  exiting newTM_fixSonRoot_emptyTree: LIB_ARBRE is %li, SOMMET_PILE %li available space is %li\n", (long int)(ptchar)LIB_ARBRE, (long int)(ptchar)SOMMET_PILE, (long int)((ptchar) SOMMET_PILE-(ptchar)LIB_ARBRE)); fflush(stdout);
  return itsFixSingleSonRoot;
}

/** link a node "aNode" to its filsd descriptor "aFilsd"
    @param cop is the operator of the node
    @param aFilsd is not NIL (not checked)
    @param aNode is not NIL (not checked)
*/
void
linkTM_filsd_describedNode (ptfilsd aFilsd, ptnode aNode, short innercop)
{
  ANYFILS_SET_INNEROPER_VALUE (aFilsd, innercop, aNode);
  NODE_LINK_FILSD_DESCRIPTOR (aNode, aFilsd);
}

/** link aNode "aNode" to its filslist descriptor "aFilslist"
    @param cop is the operator of the node
    @param aFilslist is not NIL (not checked)
    @param aNode is not NIL (not checked)
*/
void
linkTM_filslist_describedNode (ptfilslist aFilslist, ptnode aNode, short innercop)
{
  ANYFILS_SET_INNEROPER_VALUE (aFilslist, innercop, aNode);
  NODE_LINK_FILSLIST_DESCRIPTOR (aNode, aFilslist);
}

/** link aNode "aNode" to its filsAny descriptor "aFilslist"
    @param cop is the operator of the node
    @paramfilsAnyParam  is not NIL (not checked)
    @param aNode is not NIL (not checked)
*/
void
linkTM_filsAny_describedNode (ptfilsAny aFilsAny, ptnode aNode, short innercodop)
{
  if (ANYFILS_ISIN_LIST (aFilsAny))
    linkTM_filslist_describedNode ((ptfilslist) aFilsAny, aNode, innercodop);
  else
    linkTM_filsd_describedNode ((ptfilsd) aFilsAny, aNode, innercodop);
}

ptnode
newTM_userFixeTreeTop (short cop, short arity)
{
  ptnode newSubtree;
  newSubtree = newTM_fixSizeNode (arity);
  linkTM_filsd_describedNode (newTM_fixSonRoot_emptyTree (), newSubtree, INNEROPER (cop));
  return newSubtree;
}

/** clean destroyedFilsAny content
    @return the removed value node if the filsAny value is a node, NIL_node otherwise
    @param otherValue is assigned the value of the removed value when this value is not a node
    destroyedFilsAny is made an OPTROU but remains in its container, even if this container is a list
 */
ptnode
cleanTM_filsAny (ptfilsAny destroyedFilsAny, short *destroyedOpCode, untypedValueType * otherValue)
{
  ptnode theNodeValue;

  (*destroyedOpCode) = ANYFILS_GET_INNEROPER (destroyedFilsAny);
  if (IS_USEROPERATOR (*destroyedOpCode))
    theNodeValue = ANYFILS_GET_NODE_VALUE (destroyedFilsAny);
  else
    {
      (*otherValue) = ANYFILS_GET_UNTYPED_VALUE (destroyedFilsAny);
      theNodeValue = NIL_node;
    }

  ANYFILS_MAKE_TROU (destroyedFilsAny);
  return theNodeValue;
}

/** remove destroyedFilsList from its current list
    @return the removed value node if the filslist value is a node, NIL_node otherwise
    @param otherValue is assigned the value of the removed filslist
    destroyedFilsList is made an OPTROU and detached from its pere, succ and pred
 */
ptnode
unlinkTM_listElement (ptfilslist destroyedFilsList, short *destroyedOpCode, untypedValueType * otherValue)
{
  ptfilsd filsd_listDescriptor;
  ptnode aNodeContainer, theNodeValue;
  ptfilslist pred, succ;

  (*destroyedOpCode) = ANYFILS_GET_INNEROPER (destroyedFilsList);
  if (IS_USEROPERATOR (*destroyedOpCode))
    theNodeValue = ANYFILS_GET_NODE_VALUE (destroyedFilsList);
  else
    {
      (*otherValue) = ANYFILS_GET_UNTYPED_VALUE (destroyedFilsList);
      theNodeValue = NIL_node;
    }

  filsd_listDescriptor = NODE_GET_FILSD_FIRST (FILSLIST_GET_NODE_CONTAINER ((ptfilslist) destroyedFilsList));

  pred = FILSLIST_GET_PRED ((ptfilslist) destroyedFilsList);
  succ = FILSLIST_GET_SUCC ((ptfilslist) destroyedFilsList);
  if ((pred == NIL_filslist) || (pred == (ptfilslist) destroyedFilsList))
    {
      if ((succ == NIL_filslist) || (succ == (ptfilslist) destroyedFilsList))
	{
	  ANYFILS_MAKE_TROU (filsd_listDescriptor);
	}
      else
	{
	  FILSD_SET_FILSLIST_FIRST (filsd_listDescriptor, succ);
	  FILSLIST_SET_PRED (succ, NIL_filslist);
	}
    }
  else if ((succ == NIL_filslist) || (succ == (ptfilslist) destroyedFilsList))
    {
      FILSLIST_SET_SUCC (pred, NIL_filslist);
    }
  else
    {
      FILSLIST_SET_SUCC (pred, succ);
      FILSLIST_SET_PRED (succ, pred);
    }

  FILSLIST_SET_SUCC ((ptfilslist) destroyedFilsList, NIL_filslist);
  FILSLIST_SET_PRED ((ptfilslist) destroyedFilsList, NIL_filslist);
  FILSLIST_SET_NODE_CONTAINER ((ptfilslist) destroyedFilsList, NIL_node);
  ANYFILS_MAKE_TROU (destroyedFilsList);

  return theNodeValue;
}
