/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <stdio.h>
#include <string.h>
#include "StdTypeDefs.h"
#include "varglobal.h"
#include "gm_pile.h"
#include "lta_cbase.h"
#include "li_langa.h"
#include "li_nav.h"
#include "li_attri.h"
#include "lta_ctree.h"
#include "chaine.h"

#include "TM_DEBUG.h"

#define DEBUG_TREE_STACK_AVAILABLESPACE(MESS)   printf("DEBUG %s AVAILABLE SPACE FOR ARBRE and PILE:                         %li\n", MESS, (long int)(((ptchar)SOMMET_PILE)-LIB_ARBRE));
#define DEBUG_STRING_AVAILABLESPACE(MESS)       printf("DEBUG %s AVAILABLE SPACE FOR STRINGS:                                %li\n", MESS, (long int)(((ptchar)COLLAREA_FREE_ptcoll)-TMSTRINGAREA_FREE_ptchar));
void
DEBUG_MEMORY (const char *mess)
{
  printf ("DEBUG %s ADR VALUE OF DEB_ARBRE is                    %li\n", mess, (long int) (DEB_ARBRE));
  printf ("DEBUG %s ADR VALUE OF LIB_ARBRE is                    %li\n", mess, (long int) (LIB_ARBRE));
  printf ("DEBUG %s ALLOCATED SPACE FOR ARBRE:                                  %li\n", mess, (long int) (LIB_ARBRE) - (long int) (DEB_ARBRE));
  printf ("DEBUG %s ADR VALUE OF SOMMET_PILE is                  %li\n", mess, (long int) ((ptchar) SOMMET_PILE));
  printf ("DEBUG %s ADR VALUE OF DEB_PILE is                     %li\n", mess, (long int) ((ptchar) DEB_PILE));
  printf ("DEBUG %s ALLOCATED SPACE FOR PILE:                                   %li\n", mess,
	  (long int) (((ptchar) DEB_PILE) - ((ptchar) SOMMET_PILE) + sizeof (long int)));
  DEBUG_TREE_STACK_AVAILABLESPACE (mess);
  printf ("DEBUG %s ADR VALUE OF TMSTRINGAREA_START_ptchar is    %li\n", mess, (long int) (TMSTRINGAREA_START_ptchar));
  printf ("DEBUG %s ADR VALUE OF COLLAREA_START_ptcoll is        %li\n", mess, (long int) ((ptchar) COLLAREA_START_ptcoll));
  DEBUG_STRING_AVAILABLESPACE (mess);
  printf ("DEBUG %s ADR VALUE OF ptTMStringTableEntry_START is     %li\n", mess, (long int) ((ptchar) ptTMStringTableEntry_START));
  printf ("DEBUG %s ADR VALUE OF TAB_LANG is                     %li\n", mess, (long int) ((ptchar) TAB_LANG));
}

static void
DEBUG_STACK_ELEM (int level)
{
  char *stackElement = sommet (level);
  printf ("<%i:", level);
  if (TREEMEMORY_ISTREECOMPONENTADR (stackElement))
    printf ("T:%li>", TREEMEMORY_GETTREECOMPONENTPOS (stackElement));
  else
    printf ("O:%li>", stackElement);
}

static void
DEBUG_STACK_PARTIAL (int maxLevel, const char *mess)
{
  int level;
  printf ("DEBUG STACK %s stack level is %i, stack content from top to bottom: ", mess, STACKLEVEL ());
  for (level = 0; level < maxLevel; level++)
    DEBUG_STACK_ELEM (level);
  printf ("...\n");
  fflush (stdout);
}

void
DEBUG_STACK_FULL (const char *mess)
{
  DEBUG_STACK_PARTIAL (STACKLEVEL (), mess);
}

static int DEBUG_callNumber = 0;

void
DEBUG_srcORtgt (int margin, int maxLevel, const char *mess, const char *name, ptfilsAny aFilsAny)
{
  DEBUG_callNumber++;
  printf ("\n================================================================================ %i\n", DEBUG_callNumber);
  fflush (stdout);
  printf ("DEBUG %s", mess);
  printf ("---------------------------------------- %s is\n", name);
  fflush (stdout);
  DEBUGMESS_SONINANYNODE (margin, maxLevel, "", aFilsAny);
  printf ("--------------------------------------------------------------------------------\n");
  fflush (stdout);
}

void
DEBUG_src_tgt (int margin, int maxLevel, const char *mess, ptfilsAny srcFilsAny, ptfilsAny tgtFilsAny)
{
  DEBUG_callNumber++;
  printf ("\n================================================================================ %i\n", DEBUG_callNumber);
  fflush (stdout);
  printf ("DEBUG %s", mess);
  printf ("---------------------------------------- src is\n");
  fflush (stdout);
  DEBUGMESS_SONINANYNODE (margin, maxLevel, "", srcFilsAny);
  printf ("---------------------------------------- tgt is\n");
  fflush (stdout);
  DEBUGMESS_SONINANYNODE (margin, maxLevel, "", tgtFilsAny);
  printf ("--------------------------------------------------------------------------------\n");
  fflush (stdout);
}

/** @return the category of the cell memory pp
 *  0 if pp is a filsd or a filslist
 *  1 if pp is an attribute
 *  2 if pp is a repere
 *  3 if pp is a node
 *  -1 otherwise (unknown category)
 */
short
cellCategory (ptSHORT pp)
{
  short cellValue;
  cellValue = SHORT_GET_UNMARKEDVALUE (pp);
  if (SHORT_IS_DESCRIPTORVALUETAG (cellValue))
    return 0;
  if (SHORT_IS_ATTRIBUTEVALUETAG (cellValue))
    return 1;
  if (SHORT_IS_REPEREVALUETAG (cellValue))
    return 2;
  if (SHORT_IS_NODEVALUETAG (cellValue))
    return 3;
  return -1;
}

/** return true if aNode has a single son with operator ATOMOP
 */
extern boolean
isTM_anAtom (ptnode aNode, short ATOMOP)
{
  if (NODE_MAYBE_ATOM (aNode))
    return (ANYFILS_GET_INNEROPER (NODE_GET_FILSD_FIRST (aNode)) == ATOMOP);
  return pk_false;
}

/* ========= ========= ========= ========= 
 */
extern int
getTM_listNode_sonNumber (ptnode aNodeList)
{
  int i;
  ptfilslist aFilslistElem;

  i = 0;
  aFilslistElem = NODE_GET_FILSLIST_FIRST (aNodeList);
  while (aFilslistElem != NIL_filslist)
    {
      i++;
      aFilslistElem = FILSLIST_GET_SUCC (aFilslistElem);
    }
  return (i);
}

/** @return the filslist x_[pos] in the list aNode 
    @return NIL_filslist if such an element does not exit
    assume pos >= 1,  aNode is a not NIL list
 */
ptfilslist
getTM_node_nFilslist (ptnode aNode, int pos)
{
  ptfilslist aFilslist;
  aFilslist = NODE_GET_FILSLIST_FIRST (aNode);
  while (aFilslist != NIL_filslist)
    {
      if (pos == 1)
	return aFilslist;
      pos = pos - 1;
      aFilslist = FILSLIST_GET_SUCC (aFilslist);
    }
  return aFilslist;
}

/** @return the filsAny x_[n] in the list or fix size node aNode 
    @return NIL_filsAny if such an element does not exit
    assume pos >= 1,  aNode not equal to NIL
 */
ptfilsAny
getTM_node_nFilsAny (ptnode aNode, int pos)
{
  if (NODE_IS_LIST (aNode))
    return (ptfilsAny) getTM_node_nFilslist (aNode, pos);	// (pos >= 1,  aNode is a not NIL list) ==> (return aNode[pos], NIL if aNode[pos] does not exist)
  if (NODE_GET_ARITY (aNode) >= pos)
    return (ptfilsAny) NODE_GET_FILSD_N (aNode, pos);
  return NIL_filsAny;
}

/** @return the descriptor element x_[i+n] in the list that contains aFilsList at position i
    @return NIL_filslist if such an element does not exit
    assumes n >= 1
    aFilsList can be NIL
 */
ptfilslist
getTM_filslist_nNext (ptfilslist aFilsList, int n)
{
  int i = 0;
  while (i < n)
    {
      if (aFilsList == NIL_filslist)
	return NIL_filslist;
      aFilsList = FILSLIST_GET_SUCC (aFilsList);
      i++;
    }
  return aFilsList;
}

/** @return the descriptor element x_[i-n] in the list that contains aFilsList at position i
    @return NIL_filslist if such an element does not exit
    assumes n >= 1
 */
ptfilslist
getTM_filslist_nPrevious (ptfilslist aFilsList, int n)
{
  int i = 0;
  while (i < n)
    {
      if (aFilsList == NIL_filslist)
	return NIL_filslist;
      aFilsList = FILSLIST_GET_PRED (aFilsList);
      i++;
    }
  return aFilsList;
}

/** @return true if the anyFils element x_[n] exists, where x=anyFilsParam
    @param anyFilsParam is not NIL (not checked)
    @param n is strictly positive (not checked)
    @param anyFilsRes is assigned x_[i+n] when the function returns true, otherwise it remains unchanged
 */
boolean
getCondTM_filsAny_nSon (ptfilsAny anyFilsParam, int n, ptfilsAny * anyFilsRes)
{
  ptnode nodeOfFilsParam;
  ptfilsAny localAnyFilsRes;

  if (ANYFILS_HAS_USEROPERATOR (anyFilsParam))
    {
      nodeOfFilsParam = ANYFILS_GET_NODE_VALUE (anyFilsParam);
      if (nodeOfFilsParam == NIL_node)
	return pk_false;
      if (NODE_IS_LIST (nodeOfFilsParam))
	{
	  localAnyFilsRes = (ptfilsAny) getTM_node_nFilslist (nodeOfFilsParam, n);	// (n >= 1, nodeOfFilsParam is a not NIL list) ==> (return nodeOfFilsParam[n], NIL if it does not exist)
	  if (localAnyFilsRes == NIL_filsAny)
	    return pk_false;
	  (*anyFilsRes) = localAnyFilsRes;
	  return pk_true;
	}
      if (NODE_GET_ARITY (nodeOfFilsParam) >= n)
	{
	  (*anyFilsRes) = (ptfilsAny) NODE_GET_FILSD_N (nodeOfFilsParam, n);
	  return pk_true;
	}
    }
  return pk_false;
}

/** @return true if the anyFils element x_[i+n] exists, where x is the node such that x_[i]=anyFilsParam
    @param n can be negative
    @param anyFilsParam is not NIL and is not a root
    @param anyFilsRes is assigned x_[i+n] when the function returns true, otherwise it remains unchanged
 */
static boolean
getCondTM_filsAny_nNext (ptfilsAny anyFilsParam, int n, ptfilsAny * anyFilsRes)
{
  ptfilsAny localAnyFilsRes;
  if (n <= 0)
    return pk_false;
  if (ANYFILS_ISIN_LIST (anyFilsParam))
    {				/* anyFilsParam is a ptfilslist */
      localAnyFilsRes = (ptfilsAny) getTM_filslist_nNext ((ptfilslist) anyFilsParam, n);
    }
  else if (FILSD_HAS_nSUCC ((ptfilsd) anyFilsParam, n))
    localAnyFilsRes = FILSD_GET_NSUCC ((ptfilsd) anyFilsParam, n);
  else
    return pk_false;
  if (localAnyFilsRes == NIL_filsAny)
    return pk_false;
  (*anyFilsRes) = localAnyFilsRes;
  return pk_true;
}

/** if aFilsAny is linked to a subtree container PERE assign PERE to resFilsAny
 *  if aFilsAny is not linked to a subtree container (is nil or is a root) assign aFilsAny to resFilsAny
 *  @return false if resFilsAny is NIL or is a filsd root 
 *  @return true otherwise
 */
static boolean
getCondTM_filsAny_container (ptfilsAny aFilsAny, ptfilsAny * resFilsAny)
{
  ptnode nodeContainer;

  // DEBUG_STACK_FULL("entering getCondTM_filsAny_container()");
  (*resFilsAny) = aFilsAny;
  if (aFilsAny == NIL_filsAny)
    return pk_false;
  if (ANYFILS_IS_ROOT (aFilsAny))
    return pk_false;
  if (ANYFILS_ISIN_LIST (aFilsAny))
    {
      nodeContainer = FILSLIST_GET_NODE_CONTAINER ((ptfilslist) aFilsAny);
      if (nodeContainer == NIL_node)
	{
	  (*resFilsAny) = NIL_filsAny;
	  return pk_false;
	}
    }
  else
    nodeContainer = FILSD_GET_NODE_CONTAINER ((ptfilsd) aFilsAny);
  (*resFilsAny) = (ptfilsAny) NODE_GET_FILSD_DESCRIPTOR (nodeContainer);
  return ((*resFilsAny) != NIL_filsAny);
}

/** find the subtree TT following anyFilsParam in top down left to right walk
 *  @return true if TT exists, @param anyFilsRes is assigned TT
 *  @return false if anyFilsParam is a root or a root is reached
 *  @param anyFilsRes is assigned this root
 */
boolean
getCondTM_nextSubtree (ptfilsAny anyFilsParam, ptfilsAny * anyFilsRes)
{
  // DEBUGMESS_SONINANYNODE(4, 0, "entering getCondTM_nextSubtree anyFilsParam is", anyFilsParam); 
  (*anyFilsRes) = anyFilsParam;
  if (anyFilsParam == NIL_filsAny)
    return pk_false;
  if (getCondTM_filsAny_nSon (anyFilsParam, 1, anyFilsRes))
    return pk_true;
  if (ANYFILS_IS_ROOT (anyFilsParam))
    return pk_false;
  // DEBUGMESS_SONINANYNODE(4, 0, "     in  getCondTM_nextSubtree 1 anyFilsRes is", *anyFilsRes); 
  while (!getCondTM_filsAny_nNext (anyFilsParam, 1, anyFilsRes))
    {				/* STACK = STACK0 . ANYFILSRES */
      // DEBUGMESS_SONINANYNODE(4, 0, "     in  getCondTM_nextSubtree 1W anyFilsRes is", *anyFilsRes); 
      if (getCondTM_filsAny_container (anyFilsParam, anyFilsRes))
	{
	  if ((*anyFilsRes) == NIL_filsAny)
	    return pk_false;
	  if (ANYFILS_IS_ROOT (*anyFilsRes))
	    {
	      // DEBUGMESS_SONINANYNODE(4, 0, " exiting getCondTM_nextSubtree 1W2TT", *anyFilsRes); 
	      return pk_false;
	    }
	  anyFilsParam = *anyFilsRes;
	}
      else
	{
	  // DEBUGMESS_SONINANYNODE(4, 0, " exiting getCondTM_nextSubtree 1W2F", *anyFilsRes); 
	  return pk_false;
	}
    }
  // DEBUGMESS_SONINANYNODE(4, 0, " exiting getCondTM_nextSubtree * anyFilsRes is", *anyFilsRes); 
  return pk_true;
}

/** @return the filsd root of the tree that contains f it it exists
 *  @return NIL_fisd otherwise (INTERNAL ERROR)
 */
ptfilsd
getTM_filsd_root (ptfilsAny aFilsAny)
{
  ptfilsAny resFilsAny = aFilsAny;
  if (aFilsAny == NIL_filsAny)
    return NIL_filsd;
  while (getCondTM_filsAny_container (aFilsAny, &resFilsAny))
    {
      aFilsAny = resFilsAny;
    }
  if (ANYFILS_IS_ROOT (aFilsAny))
    return (ptfilsd) aFilsAny;
  return NIL_filsd;
}

/** @return the filsd hyper root of the tree that contains f it it exists
 *  @return NIL_fisd otherwise (INTERNAL ERROR)
 */
ptfilsd
getTM_filsd_hyperRoot (ptfilsAny aFilsAny)
{
  ptTreeAncestor treeAncestor;
  ptnode actualRoot, aNode;
  ptfilsd resFilsd = NIL_filsd;

  while (aFilsAny != NIL_filsAny)
    {
      resFilsd = getTM_filsd_root (aFilsAny);
      if (resFilsd == NIL_filsd)
	return NIL_filsd;
      /* resFilsd is a root */
      actualRoot = FILSD_GET_NODE_CONTAINER (resFilsd);	// NODE: 4448  3456
      treeAncestor = NODE_GET_ANCESTOR ((ptnode) actualRoot);	// [ATTR: 4408
      if (treeAncestor == NIL_treeAncestor)
	return resFilsd;
      if (SHORT_IS_ATTRIBUTEVALUETAG (SHORT_GET_UNMARKEDVALUE ((ptSHORT) treeAncestor)))
	aNode = ATTRIBUTE_GET_NODE_OWNER ((ptattribute) treeAncestor);	// [NODE: 4360
      if (aNode == NIL_node)
	return NIL_filsd;
      aFilsAny = NODE_GET_ANYFILS_DESCRIPTOR (aNode);
    }
  return NIL_filsd;
}

/**
 */
void
insertTM_filslist_first (ptfilsd filsdListDescriptor, ptfilslist aFilslist)
{
  ptfilslist firstFilslist;
  ptnode nodeContainer;

  nodeContainer = FILSD_GET_NODE_CONTAINER (filsdListDescriptor);	/* newNode, newNodeSon */
  firstFilslist = FILSD_GET_FILSLIST_FIRST (filsdListDescriptor);	/* firstFilslist = NIL */

  ANYFILS_SET_INNEROPER_VALUE (filsdListDescriptor, OPBIDON, aFilslist);
  FILSLIST_SET_NODE_CONTAINER (aFilslist, nodeContainer);
  ANYFILS_SET_PHYL (aFilslist, ANYFILS_GET_PHYL (filsdListDescriptor));

/* cas ou on rajoute un element en debut de liste */

  if (firstFilslist != NIL)
    {
      FILSLIST_SET_SUCC (aFilslist, firstFilslist);
      FILSLIST_SET_PRED (firstFilslist, aFilslist);
    }
}

/** insert a new filslist descriptor before in first position of filsdListDescriptor
    assume: "filsdListDescriptor" is a filsd list descriptor not equal to NIL
    return: the new filslist descriptor
 */
ptfilslist
insertTM_newFilslist_first (ptfilsd filsdListDescriptor)
{
  ptfilslist successor;
  ptfilslist filsList_new;
  ptnode listNode;

  filsList_new = newTM_filslist ();
  ANYFILS_SET_PHYL (filsList_new, ANYFILS_GET_PHYL (filsdListDescriptor));
  listNode = FILSD_GET_NODE_CONTAINER (filsdListDescriptor);	/* newNode, newNodeSon */
  FILSLIST_SET_NODE_CONTAINER (filsList_new, listNode);

  successor = FILSD_GET_FILSLIST_FIRST (filsdListDescriptor);	/* successor = NIL */
  FILSLIST_SET_PRED (filsList_new, NIL);
  FILSLIST_SET_SUCC (filsList_new, successor);
  ANYFILS_SET_INNEROPER_VALUE (filsdListDescriptor, OPBIDON, filsList_new);
  if (successor != NIL)
    {
      FILSLIST_SET_PRED (successor, filsList_new);
    }
  return filsList_new;
}

/** insert a new filslist descriptor before "successor"
    assume: "successor" is a correct filsList not equal to NIL
    return: the new filslist descriptor
 */
ptfilslist
insertTM_newFilslist_before (ptfilslist successor)
{
  ptfilslist predecessor;
  ptfilslist filsList_new;
  ptnode listNode;

  // DEBUGMESS_SONINLISTNODE(16, 2,"entering insertTM_newFilslist_before successor is", successor);
  filsList_new = newTM_filslist ();
  ANYFILS_SET_PHYL (filsList_new, ANYFILS_GET_PHYL (successor));
  listNode = FILSLIST_GET_NODE_CONTAINER (successor);
  FILSLIST_SET_NODE_CONTAINER (filsList_new, listNode);

  predecessor = FILSLIST_GET_PRED (successor);
  FILSLIST_SET_PRED (filsList_new, predecessor);
  FILSLIST_SET_SUCC (filsList_new, successor);
  FILSLIST_SET_PRED (successor, filsList_new);
  if (predecessor == NIL)
    {
      FILSD_SET_FILSLIST_FIRST (NODE_GET_FILSD_FIRST (listNode), filsList_new);
    }
  else
    {
      FILSLIST_SET_SUCC (predecessor, filsList_new);
    }
  return filsList_new;
}

/** insert a new filslist descriptor after "predecessor"
    assume: "predecessor" is a correct filsList not equal to NIL
    return: the new filslist descriptor
 */
ptfilslist
insertTM_newFilslist_after (ptfilslist predecessor)
{
  ptfilslist successor;
  ptfilslist filsList_new;
  ptnode listNode;

  filsList_new = newTM_filslist ();
  ANYFILS_SET_PHYL (filsList_new, ANYFILS_GET_PHYL (predecessor));
  listNode = FILSLIST_GET_NODE_CONTAINER (predecessor);
  FILSLIST_SET_NODE_CONTAINER (filsList_new, listNode);

  successor = FILSLIST_GET_SUCC (predecessor);
  FILSLIST_SET_PRED (filsList_new, predecessor);
  FILSLIST_SET_SUCC (filsList_new, successor);
  FILSLIST_SET_SUCC (predecessor, filsList_new);
  if (successor != NIL)
    {
      FILSLIST_SET_PRED (successor, filsList_new);
    }
  return filsList_new;

}

/** compute the first son of tgtFilsAny
    @param tgtFilsAny is a not null subtree descriptor
    @return 0 if X is an atomic subtree, tgtFilsAny_res is assigned tgtFilsAny
    @return 1 if X is a fix size (non atomic), tgtFilsAny_res is assigned tgtFilsAny
    @return 2 if X is an empty list, tgtFilsAny_res is assigned the filsd list descriptor
    @return 3 if X is a non empty list, tgtFilsAny_res is assigned the first list element
 */
short
getAccesTM_firstSon (ptfilsAny tgtFilsAny, ptfilsAny * tgtFilsAny_res)
{
  ptfilsd aFilsd;
  ptnode aNode;

  if (ANYFILS_HAS_USEROPERATOR (tgtFilsAny))
    {
      aNode = ANYFILS_GET_NODE_VALUE (tgtFilsAny);
      if (aNode == NIL_node)
	{
	  (*tgtFilsAny_res) = tgtFilsAny;
	  return 0;
	}
      aFilsd = NODE_GET_FILSD_FIRST (aNode);
      if (NODE_IS_LIST (aNode))
	{
	  (*tgtFilsAny_res) = (ptfilsAny) FILSD_GET_FILSLIST_FIRST (aFilsd);
	  if ((*tgtFilsAny_res) == NIL_filsAny)
	    {
	      (*tgtFilsAny_res) = (ptfilsAny) aFilsd;
	      return 2;
	    }
	  return 3;
	}
      if (NODE_IS_FIXESIZE (aNode))
	{			/* PLG: including root correct ? */
	  (*tgtFilsAny_res) = (ptfilsAny) aFilsd;
	  return 1;
	}
    }
  (*tgtFilsAny_res) = tgtFilsAny;
  return 0;
}




/** determine the right element of tgtFilsAny
    @return 0 tgtFilsAny is the last element of a fixe size node or is a root, tgtFilsAny_res is assigned tgtFilsAny
    @return 1 tgtFilsAny is in a fixe size subtree, and not the last element, tgtFilsAny_res is assigned the right element of tgtFilsAny
    @return 3 tgtFilsAny is a list element, tgtFilsAny_res is assigned tgtFilsAny
 */
short
getAccessTM_rightNeighbor (ptfilsAny tgtFilsAny, ptfilsAny * tgtFilsAny_res)
{
  ptnode tgtNode;

  // tgtFilsAny = (ptfilsAny) sommet (0);
  if (ANYFILS_IS_ROOT (tgtFilsAny))
    {
      (*tgtFilsAny_res) = tgtFilsAny;
      return 0;
    }
  if (ANYFILS_ISIN_LIST (tgtFilsAny))
    {
      (*tgtFilsAny_res) = tgtFilsAny;
      return 3;
    }
  if (FILSD_HAS_SUCC (tgtFilsAny))
    {
      (*tgtFilsAny_res) = FILSD_GET_SUCC ((ptfilsd) tgtFilsAny);
      return 1;
      /* (void)depiler();
         empiler((ptchar) tgtFilsAny); */ }
  (*tgtFilsAny_res) = tgtFilsAny;
  return 0;
}


/** determine the container of tgtFilsAny
    @return 0 no container: tgtFilsAny_res is assigned tgtFilsAny
    @return 1 the container is in a fixe size subtree tgtFilsAny_res  is assigned this container
    @return 3 the container is in a list subtree tgtFilsAny_res  is assigned this container
 */

short
getAccessTM_father (ptfilsAny tgtFilsAny, ptfilsAny * tgtFilsAny_res)
{
  ptnode aNode;

  if (ANYFILS_IS_ROOT (tgtFilsAny))
    {
      (*tgtFilsAny_res) = tgtFilsAny;
      return 0;
    }
  if (ANYFILS_ISIN_LIST (tgtFilsAny))
    aNode = FILSLIST_GET_NODE_CONTAINER ((ptfilslist) tgtFilsAny);
  else
    aNode = FILSD_GET_NODE_CONTAINER ((ptfilsd) tgtFilsAny);
  if (NODE_ISIN_LIST (aNode))
    {
      (*tgtFilsAny_res) = (ptfilsAny) NODE_GET_FILSLIST_DESCRIPTOR (aNode);
      return 3;
    }
  (*tgtFilsAny_res) = (ptfilsAny) NODE_GET_FILSD_DESCRIPTOR (aNode);
  return 1;
}

/** @return true if srcFilsAny has a son
    @return false otherwise
    @param srcFilsAny, tgtFilsAny are both not NIL filsd or not NIL filslist
    @param srcFilsAny_res is assigned the son of srcFilsAny if this son exits; it remains unchanged otherwise
    @param tgtFilsAny_res is assigned a newly created clone of srcFilsAny_res, linked to tgtFilsAny if the son of srcFilsAny exits; it remains unchanged otherwise
 */
static boolean
getCondTM_filsAny_cons (ptfilsAny srcFilsAny, ptfilsAny tgtFilsAny, ptfilsAny * srcFilsAny_res, ptfilsAny * tgtFilsAny_res)
{
  ptfilsAny srcContainerAny, tgtContainerAny;
  ptnode nodeDescriptor;

  /* STACK = STACK0 . src(i) . tgt(i) */

  // DEBUG_src_tgt(4, 4, "entering getCondTM_filsAny_cons", srcFilsAny, tgtFilsAny);
  srcContainerAny = srcFilsAny;
  if (getCondTM_filsAny_nSon (srcFilsAny, 1, &srcFilsAny))
    {				/* assumes NOT NIL srcFilsAny */
      // DEBUG_src_tgt(8, 4, "     in  getCondTM_filsAny_cons 1T", srcFilsAny, tgtFilsAny);
      if (ANYFILS_IS_LIST (srcContainerAny))
	{
	  // DEBUG_src_tgt(12, 4, "     in  getCondTM_filsAny_cons 1T-1T", srcFilsAny, tgtFilsAny);

	  tgtContainerAny = tgtFilsAny;
	  if (!getCondTM_filsAny_nSon (tgtFilsAny, 1, &tgtFilsAny))
	    {
	      tgtFilsAny = (ptfilsAny) newTM_filslist ();	/* LIST:... */
	      nodeDescriptor = newTM_listNode ();
	      linkTM_filsAny_describedNode (tgtContainerAny, nodeDescriptor, ANYFILS_GET_INNEROPER (srcContainerAny));
	      insertTM_filslist_first ((ptfilsd) NODE_GET_FILSD_FIRST (nodeDescriptor), (ptfilslist) tgtFilsAny);
	    }
	  // DEBUG_src_tgt(12, 4, "     in getCondTM_filsAny_cons 1T-1T-*", srcContainerAny, tgtContainerAny);;
	}
      else
	{
	  tgtFilsAny = (ptfilsAny) ANYFILS_GET_FILSD_FIRST (tgtFilsAny);
	  // DEBUG_src_tgt(12, 4, "     in  getCondTM_filsAny_cons 1T-1F", srcFilsAny, tgtFilsAny); 
	}
      (*srcFilsAny_res) = srcFilsAny;
      (*tgtFilsAny_res) = tgtFilsAny;
      // DEBUG_src_tgt(4, 10," exiting getCondTM_filsAny_cons 1T: new current filsAny is a SON, ", srcFilsAny, tgtFilsAny);
      return pk_true;
    }
  // DEBUG_src_tgt(4, 10," exiting getCondTM_filsAny_cons * new current filsAny is not a SON, ", srcFilsAny, tgtFilsAny);
  return pk_false;
}

/** copy atomic attributs from the current one until a tree attribute is met or until completion
    @return true if a tree attribute is met in srcAttr list
    @return false if the attribute copy is completed
    @param srcFilsAny_res is assigned the first son of the tree if a tree is met
    @param tgtFilsAny_res is assigned the first son of the newly created tree descriptor if a tree is met
    @param srcFilsAny_res is assigned the attribute src list owner if the attribute copy is completed
    @param tgtFilsAny_res is assigned the attribute tgt list owner if the attribute copy is completed
 */
static boolean
copyTM_attributesUntilTree (ptattribute srcAttr, ptattribute tgtAttribute, ptfilsAny * srcFilsAny_res, ptfilsAny * tgtFilsAny_res)
{
  ptfilsAny srcFilsAny, tgtFilsAny;

  do
    {
      switch (ATTRIBUTE_GET_INNEROPER (srcAttr))
	{
	case OPTROU:
	case OPNIL:
	case OPENTIER:
	case OPCAR:
	case OPCHAINE:
	  ATTRIBUTE_SET_INNEROPER_VALUE (tgtAttribute, ATTRIBUTE_GET_INNEROPER (srcAttr), ATTRIBUTE_GET_UNTYPED_VALUE (srcAttr));
	  break;
	  // default case OPSTAR
	case OPREPERE:
	  {
	    ptrepere srcReference, tgtReference;
	    srcReference = ATTRIBUTE_GET_REPERE_VALUE (srcAttr);
	    tgtReference = newTM_repere ();
	    *tgtReference = *srcReference;	/* PLG newly created repere is destroyed ??? */
	    ATTRIBUTE_MAKE_REPERE (tgtAttribute, tgtReference);
	    break;
	  }
	case OPARBRE:
	  {
	    ptfilsd srcFilsdAttr, tgtFilsdAttr;
	    srcFilsdAttr = ATTRIBUTE_GET_FILSD_VALUE (srcAttr);
	    ENVLANG = TREEFILSD_GET_TMLANG (srcFilsdAttr);
	    tgtFilsdAttr = newTM_fixSonRoot_emptyTree ();
	    attachTM_treeAttribute (tgtAttribute, tgtFilsdAttr);
	    (*srcFilsAny_res) = srcFilsdAttr;
	    (*tgtFilsAny_res) = tgtFilsdAttr;
	    return pk_true;
	  }
	default:
	  printf ("## INTERNAL ERROR ## in  copyTM_attributesUntilTree: illegal LTA operator %i \n", ATTRIBUTE_GET_INNEROPER (srcAttr));
	  fflush (stdout);
	}
      if (ATTRIBUTE_GET_SUCC (srcAttr) == NIL_attribute)
	{
	  (*srcFilsAny_res) = ATTRIBUTE_GET_ANYFILS_OWNER (srcAttr);
	  (*tgtFilsAny_res) = ATTRIBUTE_GET_ANYFILS_OWNER (tgtAttribute);
	  return pk_false;
	}
      srcAttr = ATTRIBUTE_GET_SUCC (srcAttr);
      tgtAttribute = insertTM_newAttributeAfter (tgtAttribute, ATTRIBUTE_GET_TAG (srcAttr));
    }
  while (pk_true);
}

/** find the next subtree to be copied after the parameter srcFilsAny
    @param srcFilsAny is the source subtree descriptor
    @param tgtFilsAny is the target subtree descriptor built as a copy of srcFilsAny
    @param srcFilsAny_res is assigned the next source subtree descriptor to be copied when such a subtree is found
    otherwise srcFilsAny_res is assigned the reached root (including srcFilsAny if srcFilsAny is a terminal root)
    @param tgtFilsAny_res is assigned the new target subtree descriptor built following srcFilsAny_res when srcFilsAny_res is not a root
    otherwise tgtFilsAny_res is assigned the reached root (including tgtFilsAny if srcFilsAny is a terminal root)
    @return true if the current tree copy is not completed
 */
static boolean
getCondTM_nextFilsAnyToCopy (ptfilsAny srcFilsAny, ptfilsAny tgtFilsAny, ptfilsAny * srcFilsAny_res, ptfilsAny * tgtFilsAny_res)
{
  ptfilsAny srcContainerAny, tgtContainerAny;
  ptnode nodeDescriptor;

  // DEBUG_src_tgt(4, 4, "entering getCondTM_nextFilsAnyToCopy", srcFilsAny, tgtFilsAny);
  /* src(i.1) does not exits: srcFilsAny unchanged */
  if (ANYFILS_IS_ROOT (tgtFilsAny))
    {
      (*srcFilsAny_res) = srcFilsAny;
      (*tgtFilsAny_res) = tgtFilsAny;
      return pk_false;
    }
  if (ANYFILS_ISIN_LIST (tgtFilsAny))
    tgtContainerAny = FILSLIST_GET_ANYFILS_PERE ((ptfilslist) tgtFilsAny);
  else
    tgtContainerAny = FILSD_GET_ANYFILS_PERE ((ptfilsd) tgtFilsAny);
  if (ANYFILS_ISIN_LIST (srcFilsAny))
    srcContainerAny = FILSLIST_GET_ANYFILS_PERE ((ptfilslist) srcFilsAny);
  else
    srcContainerAny = FILSD_GET_ANYFILS_PERE ((ptfilsd) srcFilsAny);
  if (ANYFILS_IS_ROOT (tgtContainerAny))
    {
      (*srcFilsAny_res) = srcContainerAny;
      (*tgtFilsAny_res) = tgtContainerAny;
      return pk_false;
    }
  if (getCondTM_filsAny_nNext (srcFilsAny, 1, &srcFilsAny))
    {
      if (ANYFILS_ISIN_LIST (srcFilsAny))
	{
	  tgtContainerAny = tgtFilsAny;
	  tgtFilsAny = (ptfilsAny) insertTM_newFilslist_after ((ptfilslist) tgtContainerAny);
	}
      else
	tgtFilsAny = (ptfilsAny) (FILSD_GET_SUCC (tgtFilsAny));
      (*srcFilsAny_res) = srcFilsAny;
      (*tgtFilsAny_res) = tgtFilsAny;
      // DEBUG_src_tgt(8, 10," exiting getCondTM_nextFilsAnyToCopy 4T", srcFilsAny, tgtFilsAny);
      return pk_true;
    }
  while (pk_true)
    {
      if (getCondTM_filsAny_container (srcFilsAny, &srcFilsAny))
	{
	  (void) getCondTM_filsAny_container (tgtFilsAny, &tgtFilsAny);
	  if (ANYFILS_IS_ROOT (tgtFilsAny))
	    {
	      (*srcFilsAny_res) = srcFilsAny;
	      (*tgtFilsAny_res) = tgtFilsAny;
	      // DEBUG_src_tgt(4, 10," exiting getCondTM_nextFilsAnyToCopy 5W-1T-1T", srcFilsAny, tgtFilsAny);
	      return pk_false;
	    }

	  if (getCondTM_filsAny_nNext (srcFilsAny, 1, &srcFilsAny))
	    {
	      if (ANYFILS_ISIN_LIST (srcFilsAny))
		{
		  tgtContainerAny = tgtFilsAny;
		  tgtFilsAny = (ptfilsAny) insertTM_newFilslist_after ((ptfilslist) tgtContainerAny);
		}
	      else
		tgtFilsAny = FILSD_GET_SUCC (tgtFilsAny);
	      (*srcFilsAny_res) = srcFilsAny;
	      (*tgtFilsAny_res) = tgtFilsAny;
	      // DEBUG_src_tgt(4, 10," exiting getCondTM_nextFilsAnyToCopy 5W-1T-3T", srcFilsAny, tgtFilsAny);
	      return pk_true;
	    }
	  // DEBUG_src_tgt(12, 3,"     in  getCondTM_nextFilsAnyToCopy 5W-1T-4 (unphased) ", srcFilsAny, tgtFilsAny);
	}
      else
	{
	  (*srcFilsAny_res) = srcFilsAny;
	  (*tgtFilsAny_res) = tgtFilsAny;
	  // DEBUG_src_tgt(12, 10," exiting getCondTM_nextFilsAnyToCopy 5W", srcFilsAny, tgtFilsAny);
	  return pk_false;
	}
      // DEBUG_src_tgt(8, 3,"     in  getCondTM_nextFilsAnyToCopy 5W*", srcFilsAny, tgtFilsAny);
    }				/* while ( true ) */
}


/** buid and return a new tree the value of which is a copy of srcFilsAny
    @param srcFilsAny is a direct or list son subtree (not checked) not equal to NIL
*/

ptfilsAny
copyTM_FilsAnySubtree (ptfilsAny srcFilsAnyStart, boolean copyAttributes)
{
  ptfilsAny tgtFilsAny, srcFilsAny, previous_tgtFilsAny, previous_srcFilsAny;
  ptattribute srcAttribute, tgtAttribute;
  ptTreeAncestor tgtTreeAncestor;
  short rootInitialPos, op;
  boolean completedSubtree;

  // DEBUG_srcORtgt(4, 100, "entering copyTM_FilsAnySubtree", "src", srcFilsAnyStart);
  if ((ANYFILS_IS_ROOT (srcFilsAnyStart)))
    rootInitialPos = 0;
  else
    {				/* srcFilsAnyStart is a strict sub tree */
      rootInitialPos = ANYFILS_GET_POS (srcFilsAnyStart);
      ANYFILS_SET_POS (srcFilsAnyStart, 0);
    }
  srcFilsAny = srcFilsAnyStart;
  completedSubtree = pk_false;
  tgtFilsAny = newTM_fixSonRoot_emptyTree ();
  // DEBUG_src_tgt(4, 4,"entering copyTM_FilsAnySubtree 1", srcFilsAny, tgtFilsAny);

  while (!((srcFilsAny == srcFilsAnyStart) && completedSubtree))
    {				/* inner tree loop: tgtFilsAny is a tree filsd son */
      // DEBUG_src_tgt(8, 4,"      in  copyTM_FilsAnySubtree W1", srcFilsAny, tgtFilsAny);
      while (!((ANYFILS_IS_ROOT (srcFilsAny)) && completedSubtree))
	{
	  // DEBUG_src_tgt(12, 4,"      in  copyTM_FilsAnySubtree W1W1", srcFilsAny, tgtFilsAny);
	  op = ANYFILS_GET_INNEROPER (srcFilsAny);
	  switch (op)
	    {
	    case OPTROU:
	    case OPNIL:
	    case OPENTIER:
	    case OPCAR:
	    case OPCHAINE:
	    case OPSTAR:
	      // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW1", srcFilsAny, tgtFilsAny);
	      ANYFILS_SET_PHYL (tgtFilsAny, ANYFILS_GET_PHYL (srcFilsAny));
	      ANYFILS_SET_INNEROPER_VALUE (tgtFilsAny, op, ANYFILS_GET_UNTYPED_VALUE (srcFilsAny));
	      completedSubtree = pk_true;
	      break;
	    case OPREPERE:
	      {
		ptrepere srcReference, tgtReference;
		// DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-repere", srcFilsAny, tgtFilsAny);
		srcReference = ANYFILS_GET_REPERE_VALUE (srcFilsAny);
		tgtReference = newTM_repere ();
		*tgtReference = *srcReference;
		ANYFILS_SET_PHYL (tgtFilsAny, ANYFILS_GET_PHYL (srcFilsAny));
		ANYFILS_SET_INNEROPER_VALUE (tgtFilsAny, OPREPERE, tgtReference);
		completedSubtree = pk_true;
		break;
	      }
	    case OPARBRE:
	      previous_srcFilsAny = srcFilsAny;
	      previous_tgtFilsAny = tgtFilsAny;
	      srcFilsAny = ANYFILS_GET_FILSD_VALUE (srcFilsAny);
	      if (srcFilsAny == NIL_filsAny)
		{
		  tgtFilsAny = (ptfilsAny) newTM_fixSonRoot_emptyTree ();
		  ANYFILS_SET_PHYL (previous_tgtFilsAny, ANYFILS_GET_PHYL (previous_srcFilsAny));
		  ANYFILS_MAKE_TREE (previous_tgtFilsAny, NIL_filsAny);
		  tgtFilsAny = previous_tgtFilsAny;
		  srcFilsAny = previous_srcFilsAny;
		  completedSubtree = pk_true;
		}
	      else
		{
		  // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-arbre", srcFilsAny, tgtFilsAny);
		  ENVLANG = TREEFILSD_GET_TMLANG (srcFilsAny);
		  tgtFilsAny = (ptfilsAny) newTM_fixSonRoot_emptyTree ();
		  ANYFILS_SET_PHYL (previous_tgtFilsAny, ANYFILS_GET_PHYL (previous_srcFilsAny));
		  ANYFILS_MAKE_TREE (previous_tgtFilsAny, tgtFilsAny);
		  completedSubtree = pk_false;
		}
	      break;
	    default:
	      {			/* language specific operators all: of them have one or more son */
		ptnode tgtNode, srcNode;
		// DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user", srcFilsAny, tgtFilsAny);
		srcNode = ANYFILS_GET_NODE_VALUE (srcFilsAny);
		// DEBUGMESS_NODE(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user NODE is", srcNode);
		if (srcNode == NIL_node)
		  completedSubtree = pk_true;
		else
		  {
		    if (NODE_IS_LIST (srcNode))
		      {		/* srcFilsAny is a list subtree */
			// DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user1-FT", srcFilsAny, tgtFilsAny);
			tgtNode = newTM_listNode ();
			ANYFILS_SET_PHYL (NODE_GET_FILSD_FIRST (tgtNode), ANYFILS_GET_PHYL (NODE_GET_FILSD_FIRST (srcNode)));
			linkTM_filsAny_describedNode (tgtFilsAny, tgtNode, op);
			previous_srcFilsAny = srcFilsAny;
			srcFilsAny = (ptfilsAny) NODE_GET_FILSLIST_FIRST (srcNode);
			if (srcFilsAny == NIL_filsAny)
			  {	/* empty list */
			    srcFilsAny = previous_srcFilsAny;
			    /* tgtFilaAny unchanged */
			    completedSubtree = pk_true;
			  }
			else
			  {	/* non empty list */
			    tgtFilsAny = (ptfilsAny) insertTM_newFilslist_first ((ptfilsd) NODE_GET_FILSD_FIRST (tgtNode));
			    completedSubtree = pk_false;
			  }
		      }
		    else if (NODE_IS_FIXESIZE (srcNode))
		      {		/* srcFilsAny is a fix size subtree */
			// DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user1-FF", srcFilsAny, tgtFilsAny);
			tgtNode = newTM_fixSizeNode (NODE_GET_ARITY (srcNode));
			ANYFILS_SET_PHYL (tgtFilsAny, ANYFILS_GET_PHYL (srcFilsAny));
			linkTM_filsAny_describedNode (tgtFilsAny, tgtNode, op);
			// DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-F -F1", srcFilsAny, tgtFilsAny);
			srcFilsAny = NODE_GET_FILSD_FIRST (srcNode);
			tgtFilsAny = NODE_GET_FILSD_FIRST (tgtNode);
			// DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-F -F2", srcFilsAny, tgtFilsAny);
			completedSubtree = pk_false;
		      }
		    else
		      {		/* node arity is 0  */
			printf ("##WARNING## in  copyTM_FilsAnySubtree W1W1SW-user-F arity is 0\n");
			fflush (stdout);
		      }
		    // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-1", srcFilsAny, tgtFilsAny);
		    // DEBUGMESS_NODE(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-1 src NODE is", srcNode);
		    // DEBUGMESS_NODE(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-1 tgt NODE is", tgtNode);
		    if (copyAttributes)
		      {
			srcAttribute = NODE_GET_ATTRIBUTE (srcNode);
			if (srcAttribute != NIL_attribute)
			  {
			    tgtAttribute = insertTM_newFirstAttribute (tgtNode, ATTRIBUTE_GET_TAG (srcAttribute));
			    // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-2-T", srcFilsAny, tgtFilsAny);
			    completedSubtree = !copyTM_attributesUntilTree (srcAttribute, tgtAttribute, &srcFilsAny, &tgtFilsAny);
			    // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-2-T1", srcFilsAny, tgtFilsAny);
			    /* completedSubtree is false if a tree attribute is met in srcAttribute list
			       srcFilsAny is assigned the first son of the tree if a tree is met
			       tgtFilsAny is assigned the first son of the newly created tree descriptor if a tree is met
			       completedSubtree is true if the attribute copy is completed
			       srcFilsAny is assigned the attribute src list owner if the attribute copy is completed
			       tgtFilsAny is assigned the attribute tgt list owner if the attribute copy is completed
			     */
			    if (completedSubtree)
			      completedSubtree = !getCondTM_filsAny_cons (srcFilsAny, tgtFilsAny, &srcFilsAny, &tgtFilsAny);
			    // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-2-T2", srcFilsAny, tgtFilsAny);
			    // DEBUGMESS_NODE(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-2-T2 src NODE is", srcNode);
			    // DEBUGMESS_NODE(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-2-T2 tgt NODE is", tgtNode);
			  }
		      }
		    // DEBUG_src_tgt(16, 4,"      in  copyTM_FilsAnySubtree W1W1SW-user-3", srcFilsAny, tgtFilsAny);
		  }
		break;
	      }
	    }			/* de switch */
	  // DEBUG_src_tgt(12, 4,"     in  copyTM_FilsAnySubtree W1W2", srcFilsAny, tgtFilsAny);
	  if (completedSubtree)
	    {
	      // DEBUG_src_tgt(12, 4,"     in  copyTM_FilsAnySubtree W1W2-2F", srcFilsAny, tgtFilsAny);
	      if (srcFilsAny == srcFilsAnyStart)
		{
		  ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
		  return tgtFilsAny;
		}
	      completedSubtree = !getCondTM_nextFilsAnyToCopy (srcFilsAny, tgtFilsAny, &srcFilsAny, &tgtFilsAny);
	    }
	  // DEBUG_src_tgt(8, 4,"     in  copyTM_FilsAnySubtree W1W3", srcFilsAny, tgtFilsAny);
	}			/*  while (!((ANYFILS_IS_ROOT(srcFilsAny)) && completedSubtree)) {...} */
      // DEBUG_src_tgt(8, 4,"     in  copyTM_FilsAnySubtree W2 tgt is expected to be tree root", srcFilsAny, tgtFilsAny);
      if (srcFilsAny != srcFilsAnyStart)
	{
	  tgtTreeAncestor = FILSD_GET_ANCESTOR (tgtFilsAny);

	  if (tgtTreeAncestor == NIL_treeAncestor)
	    {
	      ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
	      // DEBUG_src_tgt(4, 100,"  exiting copyTM_FilsAnySubtree 1", srcFilsAny, tgtFilsAny);
	      return tgtFilsAny;
	    }
	  // DEBUG_src_tgt(8, 4,"     in  copyTM_FilsAnySubtree W4", srcFilsAny, tgtFilsAny);

	  switch (cellCategory ((ptSHORT) tgtTreeAncestor))
	    {
	    case 0:		/* acces tgtFilsAny is assumed to be a fils direct. Not checked */
	      // DEBUG_src_tgt(12, 4,"     in  copyTM_FilsAnySubtree W5-SW0", srcFilsAny, tgtFilsAny);
	      tgtFilsAny = FILSD_GET_ANYFILS_PERE (tgtFilsAny);
	      srcFilsAny = FILSD_GET_ANYFILS_PERE (srcFilsAny);
	      // DEBUG_src_tgt(12, 4,"     in  copyTM_FilsAnySubtree W5-SW0-1", srcFilsAny, tgtFilsAny);
	      if (srcFilsAny == srcFilsAnyStart)
		{
		  ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
		  return tgtFilsAny;
		}
	      completedSubtree = !getCondTM_nextFilsAnyToCopy (srcFilsAny, tgtFilsAny, &srcFilsAny, &tgtFilsAny);
	      break;
	    case 1:		/* attribut ONLY WHEN copyAttributes is true */
	      // DEBUG_src_tgt(12, 4,"     in  copyTM_FilsAnySubtree W5-SW1", srcFilsAny, tgtFilsAny);
	      srcAttribute = FILSD_GET_ATTRIBUTE_TREECONTAINER (srcFilsAny);
	      tgtAttribute = FILSD_GET_ATTRIBUTE_TREECONTAINER (tgtFilsAny);
	      // DEBUG_src_tgt(12, 4,"     in  copyTM_FilsAnySubtree W5-SW1-1", srcFilsAny, tgtFilsAny);
	      if (ATTRIBUTE_GET_SUCC (srcAttribute) == NIL_attribute)
		{
		  // DEBUG_src_tgt(16, 4,"     in  copyTM_FilsAnySubtree W5-SW1-2T", srcFilsAny, tgtFilsAny);
		  srcFilsAny = ATTRIBUTE_GET_ANYFILS_OWNER (srcAttribute);
		  tgtFilsAny = ATTRIBUTE_GET_ANYFILS_OWNER (tgtAttribute);
		  // DEBUG_src_tgt(16, 4,"     in  copyTM_FilsAnySubtree W5-SW1-2T1", srcFilsAny, tgtFilsAny);
		  completedSubtree = pk_true;
		}
	      else
		{
		  srcAttribute = ATTRIBUTE_GET_SUCC (srcAttribute);
		  // DEBUG_src_tgt(16, 4,"     in  copyTM_FilsAnySubtree W5-SW1-2F", srcFilsAny, tgtFilsAny);
		  tgtAttribute = insertTM_newAttributeAfter (tgtAttribute, ATTRIBUTE_GET_TAG (srcAttribute));
		  // DEBUG_src_tgt(16, 4,"     in  copyTM_FilsAnySubtree W5-SW1-2F1", srcFilsAny, tgtFilsAny);
		  completedSubtree = !copyTM_attributesUntilTree (srcAttribute, tgtAttribute, &srcFilsAny, &tgtFilsAny);
		}
	      /* completedSubtree is false if a tree attribute is met in srcAttribute list
	         srcFilsAny is assigned the first son of the tree if a tree is met
	         tgtFilsAny is assigned the first son of the newly created tree descriptor if a tree is met
	         completedSubtree is true if the attribute copy is completed
	         srcFilsAny is assigned the attribute src list owner if the attribute copy is completed
	         tgtFilsAny is assigned the attribute tgt list owner if the attribute copy is completed
	       */
	      // DEBUG_src_tgt(16, 4,"     in  copyTM_FilsAnySubtree W5-SW1-2F2", srcFilsAny, tgtFilsAny);
	      if (completedSubtree)
		{		/* the list of attributes is completely copied */
		  if (getCondTM_filsAny_cons (srcFilsAny, tgtFilsAny, &srcFilsAny, &tgtFilsAny))
		    completedSubtree = pk_false;
		  else if (srcFilsAny == srcFilsAnyStart)
		    {
		      ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
		      return tgtFilsAny;
		    }
		  else
		    completedSubtree = !getCondTM_nextFilsAnyToCopy (srcFilsAny, tgtFilsAny, &srcFilsAny, &tgtFilsAny);
		}
	      break;
	    default:
	      printf ("##INTERNAL ERROR## in copyTM_FilsAnySubtree: unexpected cell category\n");
	      fflush (stdout);
	      ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
	      return NIL_filsAny;
	    }			/* switch(cellCategory(...)) */
	}
    }				/* while ( !((srcFilsAny == srcFilsAnyStart) && completedSubtree)) {} */
  ANYFILS_SET_POS (srcFilsAnyStart, rootInitialPos);
  // DEBUG_src_tgt(4, 100,"  exiting copyTM_FilsAnySubtree *", srcFilsAny, tgtFilsAny);
  return tgtFilsAny;
}

/** assign LTA representation of string val to aFilsAny
 */
extern void
assignTM_stringValue2FilsAny (ptfilsAny aFilsAny, char *val)
{
  TMString icc;

  // printf("DEBUG in assignTM_stringValue2FilsAny val is %s\n", val); fflush(stdout);
  icc = TMSTRINGCHARADR_CAST_TMSTRING (newTM_CString2TMStringCharAdr (strlen (val), val));
  ANYFILS_MAKE_TMSTRING (aFilsAny, icc);
}

/** @param link2DestinationSon is not NIL (not checked)
    if assignedSubtree is NIL, link2DestinationSon becomes a hole
    otherwise the subtree designed by assignedSubtree is moved to link2DestinationSon
    and detached from its current position
 */
extern void
assignTM_nodeValue2Filsd (ptfilsd targetFilsd, ptnode assignedNode)
{
  ptfilsd assignedFilsd;

  if (targetFilsd == (ptfilsd) NIL)
    return;
  // DEBUGMESS_SONINFIXSIZENODE(4,2,"entering assignTM_nodeValue2Filsd, target subtree is", targetFilsd); 
  // DEBUGMESS_NODE(4,2,"entering assignTM_nodeValue2Filsd, assignedNode is", assignedNode); 
  if (assignedNode == NIL_node)
    {
      ANYFILS_MAKE_TROU (targetFilsd);
    }
  else
    {
      assignedFilsd = NODE_GET_FILSD_DESCRIPTOR (assignedNode);
      // DEBUGMESS_SONINFIXSIZENODE(4,2,"     in  assignTM_nodeValue2Filsd 1, assigned filsd is", assignedFilsd); 
      if (assignedFilsd == NIL_filsd)
	{
	  printf ("##INTERNAL ERROR## in assignTM_nodeValue2Filsd node argument is not linked");
	  fflush (stdout);
	  ANYFILS_MAKE_TROU (targetFilsd);
	}
      else
	{
	  //  DEBUGMESS_SONINFIXSIZENODE(4,2,"     in  assignTM_nodeValue2Filsd 2, assigned filsd is", assignedFilsd); 
	  ANYFILS_SET_INNEROPER_VALUE (targetFilsd, ANYFILS_GET_INNEROPER (assignedFilsd), ANYFILS_GET_NODE_VALUE (assignedFilsd));
	  if (ANYFILS_HAS_USEROPERATOR (assignedFilsd))
	    {
	      assignedNode = ANYFILS_GET_NODE_VALUE (assignedFilsd);
	      if (!(IS_NILTREE (assignedNode)))
		{
		  NODE_LINK_FILSD_DESCRIPTOR (assignedNode, targetFilsd);
		}
	    }
	  // PLG FOLLOWING LINE REMOVED TO TRY TO KEEP SIDE EFFECT: BETTER BUT STILL INCORRECT BEHAVIOR ???? 
	  ANYFILS_MAKE_TROU (assignedFilsd);
	  // DEBUGMESS_SONINFIXSIZENODE(4,2,"     in assignTM_nodeValue2Filsd, assigned subtree becomes", assignedFilsd); 
	}
    }
}

/** @return the node X_[i+n] in the fix size node X that contains aNode at position i
    @return NIL_node if such an element does not exit
    @param(n) assumes n >= 1
    @param(aNode) is assumed to be a NOT NIL fixe size node
    assume: for all i X_[i] is not defined or is a node
 */
extern ptnode
getTM_node_nNextInFix (ptnode aNode, int n)	/* only used in treefront */
{
  ptfilsd aFilsd = NODE_GET_FILSD_DESCRIPTOR (aNode);
  if (aFilsd == NIL_filsd)
    {
      printf ("##INTERNAL ERROR## in getTM_node_nNextInFix, the ptnode argument is not well linked\n");
      fflush (stdout);
      return NIL_node;
    }
  if (FILSD_HAS_nSUCC (aFilsd, n))
    return ANYFILS_GET_NODE_VALUE ((ptfilsAny) FILSD_GET_NSUCC (aFilsd, n));
  return NIL_node;
}

/** @return the node X_[i+n] in the list X that contains aNode at position i
    @return NIL_node if such an element does not exit
    @param(n) assumes n >= 1
    @param(aNode) is assumed to be a NOT NIL list node
    assume: for all i X_[i] is not defined or is a node
 */
extern ptnode
getTM_node_nNextInlist (ptnode aNode, int n)	/* only used in treefront */
{
  ptfilslist aFilsList = NODE_GET_FILSLIST_DESCRIPTOR (aNode);
  if (aFilsList == NIL_filslist)
    {
      printf ("##INTERNAL ERROR## in getTM_node_nNextInList, the ptnode arguemnt is not well linked\n");
      fflush (stdout);
      return NIL_node;
    }
  aFilsList = getTM_filslist_nNext (aFilsList, n);
  if (aFilsList == NIL_filslist)
    return NIL_node;
  return ANYFILS_GET_NODE_VALUE ((ptfilsAny) aFilsList);
}

/** @return the node X_[i-n] in the fix size node X that contains aNode at position i
    @return NIL_node if such an element does not exit
    @param(n) assumes n >= 1
    @param(aNode) is assumed to be a NOT NIL fixe size node
    assume: for all i X_[i] is not defined or is a node
 */
extern ptnode
getTM_node_nPreviousInFix (ptnode aNode, int n)	/* only used in treefront */
{
  ptfilsd aFilsd = NODE_GET_FILSD_DESCRIPTOR (aNode);
  if (aFilsd == NIL_filsd)
    {
      printf ("##INTERNAL ERROR## in getTM_node_nNextInFix, the ptnode argument is not well linked\n");
      fflush (stdout);
      return NIL_node;
    }
  if (ANYFILS_GET_POS (aFilsd) > n)
    return ANYFILS_GET_NODE_VALUE ((ptfilsAny) FILSD_GET_NPRED (aFilsd, n));
  return NIL_node;
}

/** @return the node X_[i-n] in the list that contains aNode at position i
    @return NIL_node if such an element does not exit
    @param(n) assumes n >= 1
    @param(aNode) is assumed to be a NOT NIL list node
    assume: for all i X_[i] is not defined or is a node
 */
extern ptnode
getTM_node_nPreviousInlist (ptnode aNode, int n)	/* only used in treefront */
{
  ptfilslist aFilsList = NODE_GET_FILSLIST_DESCRIPTOR (aNode);
  if (aFilsList == NIL_filslist)
    {
      printf ("##INTERNAL ERROR## in getTM_node_nNextInList, the ptnode arguemnt is not well linked\n");
      fflush (stdout);
      return NIL_node;
    }
  aFilsList = getTM_filslist_nPrevious (aFilsList, n);
  if (aFilsList == NIL_filslist)
    return NIL_node;
  return ANYFILS_GET_NODE_VALUE ((ptfilsAny) aFilsList);
}
