/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/*****************************************************************************/
/*									     */
/*          marquer.c							     */
/*									     */
/*									     */
/*  Module contenant les primitives de marquage des objets a recuperer	     */
/*									     */
/*									     */
/*****************************************************************************/

#include "StdTypeDefs.h"
#include "varglobal.h"
#include "gm_pile.h"
#include "lta_ctree.h"
#include "marquer.h"

#define IS_PTPTSHORT_INSTACKMEMORY(PT) (PT >= (ptptSHORT) SOMMET_PILE)
#define GET_PTPTSHORT_FIRST_INSTACKMEMORY() (ptptSHORT) (DEB_PILE - 3)

/*****************************************************************************/
/*									     */
/*	ESTARBRE							     */
/*									     */
/* Fonction qui delivre true si le parametre est un pointeur vers la zone des*/
/* arbres.								     */
/*									     */
/*****************************************************************************/
boolean
estarbre (ptSHORT pp)
{
  if ((ptchar) pp == (ptchar) NIL)
    return pk_false;
  return TREEMEMORY_ISTREECOMPONENTADR ((long int) pp);
}

/*****************************************************************************/
/*									     */
/*	SUIVANT								     */
/*									     */
/* Delivre l'acces suivant de f en parcours descendant gauche droit	     */
/*									     */
/*****************************************************************************/

static ptfilsd
suivantnavig (ptfilsd f)
{
  ptfilsAny f_res;
  (void) getCondTM_nextSubtree (f, &f_res);	/* called with ptfilsd ? */
  return f_res;
}



/*****************************************************************************/
/*									     */
/*	MARQUERATTRIBUT							     */
/*									     */
/* Fonction de marquage d'1 attribut. Cette fonction gere les passages sur   */
/* attribut repere et arbre. Elle delivre un booleaneen qui indique :	     */
/*	0 -> la liste des attributs du noeud a ete entierement marquee	     */
/*	1 -> marquage d'1 attribut arbre ou repere en cours		     */
/*	2 -> la liste des attributs du noeud n'est pas entierement marquee   */
/*									     */
/*****************************************************************************/

static short
marquerattribut (ptattribute att, pointeuracces pf)
{
  short type, vboolean;
  ptnode e;
  ptfilsd f;
  ptrepere rep;

  vboolean = 2;
  SHORT_SET_INITIALMARK ((ptSHORT) att);
  type = ATTRIBUTE_GET_INNEROPER (att);
  switch (type)
    {
    case OPTROU:
    case OPNIL:
    case OPENTIER:
    case OPCAR:
    case OPCHAINE:
    case OPSTAR:
      break;

    case OPREPERE:
      rep = ATTRIBUTE_GET_REPERE_VALUE (att);
      if (rep != NIL)
	{
	  SHORT_SET_INITIALMARK ((ptSHORT) rep);
	  f = REPERE_GET_FILSD_REFERED (rep);
	  if (estarbre ((ptSHORT) f))
	    {
	      f = getTM_filsd_root (f);
	      if (SHORT_NONMARQUE ((ptSHORT) f))
		{
		  e = (ptnode) f - 1;
		  SHORT_SET_INITIALMARK ((ptSHORT) e);
		  NODE_SET_ATTRIBUTE_CONTAINER (e, att);
		  att = NODE_GET_ATTRIBUTE (e);
		  SHORT_SET_INITIALMARK ((ptSHORT) att);
		  vboolean = 1;
		  *pf = f;
		}
	    }
	}
      break;

    case OPARBRE:
      f = ATTRIBUTE_GET_FILSD_VALUE (att);
      if (SHORT_NONMARQUE ((ptSHORT) f))
	{
	  e = (ptnode) f - 1;
	  SHORT_SET_INITIALMARK ((ptSHORT) e);
	  att = NODE_GET_ATTRIBUTE (e);
	  SHORT_SET_INITIALMARK ((ptSHORT) att);
	  vboolean = 1;
	  *pf = f;
	}
      break;
    }

  if ((vboolean == 2) && (ATTRIBUTE_GET_SUCC (att) == NIL))
    vboolean = 0;
  return (vboolean);
}


/*****************************************************************************/
/*									     */
/*	MARQUERARBRE							     */
/*									     */
/* Fonction de marquage d'un arbre dont la racine se trouve en sommet de pile*/
/* Cette fonction necessite donc un emplacement de pile.		     */
/*									     */
/*****************************************************************************/
static void
marquerarbre (void)
{
  short vboolean, op;
  ptSHORT pt;
  ptnode e;
  ptattribute anAttribute;
  ptfilsd frac, f, f1;
  ptrepere rep;

  frac = (ptfilsd) depiler ();
  f = frac;
  vboolean = 1;
  e = (ptnode) f - 1;
  SHORT_SET_INITIALMARK ((ptSHORT) e);
  anAttribute = NODE_GET_ATTRIBUTE (e);
  SHORT_SET_INITIALMARK ((ptSHORT) anAttribute);

  while ((f != frac) || (vboolean == 1))
    {
      while ((!(ANYFILS_IS_ROOT (f))) || (vboolean == 1))
	{
	  vboolean = 0;
	  if (SHORT_NONMARQUE ((ptSHORT) f))
	    {
	      SHORT_SET_INITIALMARK ((ptSHORT) f);
	      op = ANYFILS_GET_INNEROPER (f);
	      if (op == OPREPERE)
		{		/* then 1 */
		  rep = ANYFILS_GET_REPERE_VALUE (f);
		  if (rep != NIL)
		    {
		      SHORT_SET_INITIALMARK ((ptSHORT) rep);
		      f1 = REPERE_GET_FILSD_REFERED (rep);
		      if (estarbre ((ptSHORT) f1))
			{
			  if (SHORT_NONMARQUE ((ptSHORT) f1))
			    {
			      f1 = getTM_filsd_root (f1);
			      e = (ptnode) f1 - 1;
			      SHORT_SET_INITIALMARK ((ptSHORT) e);
			      anAttribute = NODE_GET_ATTRIBUTE (e);
			      SHORT_SET_INITIALMARK ((ptSHORT) anAttribute);
			      NODE_SET_FILSD_DESCRIPTOR (e, f);	/* PLG not tap ? */
			      vboolean = 1;
			      f = f1;
			    }
			  else
			    f = suivantnavig (f);
			}
		      else
			f = suivantnavig (f);
		    }
		}		/* end of then 1 OPREPERE */
	      else if (op == OPARBRE)
		{		/* then 2 */
		  f1 = (ptfilsd) ANYFILS_GET_NODE_VALUE (f);
		  if (estarbre ((ptSHORT) f1))
		    {
		      if (SHORT_NONMARQUE ((ptSHORT) f1))
			{
			  e = (ptnode) f1 - 1;
			  SHORT_SET_INITIALMARK ((ptSHORT) e);
			  NODE_SET_FILSD_DESCRIPTOR (e, (ptfilsd) f);	/* PLG not tap ? */
			  anAttribute = NODE_GET_ATTRIBUTE (e);
			  SHORT_SET_INITIALMARK ((ptSHORT) anAttribute);
			  vboolean = 1;
			  f = f1;
			}
		      else
			f = suivantnavig (f);
		    }
		  else
		    f = suivantnavig (f);
		}		/* end of then 2 OPARBRE */
	      else
		{		/* else 2 op is neither OPREPERE nor OPARBRE */
		  if (!(ISCOMMONOPER (op)))
		    {		/* then 3 */
		      e = ANYFILS_GET_NODE_VALUE (f);
		      SHORT_SET_INITIALMARK ((ptSHORT) e);
		      if (NODE_GET_ARITY (e) == -1)
			{
			  f1 = NODE_GET_FILSD_FIRST (e);
			  SHORT_SET_INITIALMARK ((ptSHORT) f1);
			}
		      if (NODE_GET_ATTRIBUTE (e) != NIL)
			{
			  anAttribute = NODE_GET_ATTRIBUTE (e);
			  do
			    {
			      vboolean = marquerattribut (anAttribute, &f);
			      anAttribute = ATTRIBUTE_GET_SUCC (anAttribute);
			    }
			  while (vboolean == 2);
			}
		    }		/* end of then 3 user defined operator */
		  if (vboolean == 0)
		    f = suivantnavig (f);
		}		/* end of else 2 op is neither OPREPERE nor OPARBRE */
	    }			/* end of SHORT_NONMARQUE ((ptSHORT) f */
	  else			/* SHORT_MARQUE */
	    f = suivantnavig (f);
	}

      e = (ptnode) f - 1;
      if (NODE_GET_FILSD_DESCRIPTOR (e) != NIL)
	{
	  pt = (ptSHORT) NODE_GET_FILSD_DESCRIPTOR (e);
	  if (SHORT_IS_DESCRIPTOR (pt))
	    {
	      /* traiter retour sur repere ou sur acces a OPARBRE */
	      f = (ptfilsd) pt;
	      if (ANYFILS_GET_INNEROPER (f) == OPREPERE)
		NODE_SET_FILSD_DESCRIPTOR (e, (ptfilsd) NIL);
	    }			/* PLG not tap ? */
	  else
	    {
	      /* traiter retour sur attribut */
	      anAttribute = (ptattribute) pt;
	      if (ATTRIBUTE_GET_INNEROPER (anAttribute) == OPREPERE)
		NODE_SET_FILSD_DESCRIPTOR (e, NIL);	/* PLG not tap ? */
	      e = ATTRIBUTE_GET_NODE_OWNER (anAttribute);
	      f = NODE_GET_FILSD_DESCRIPTOR (e);
	      if (ATTRIBUTE_GET_SUCC (anAttribute) != NIL)
		{
		  anAttribute = ATTRIBUTE_GET_SUCC (anAttribute);
		  do
		    {
		      vboolean = marquerattribut (anAttribute, &f);
		      anAttribute = ATTRIBUTE_GET_SUCC (anAttribute);
		    }
		  while (vboolean == 2);
		}
	    }
	  if (vboolean == 0)
	    f = suivantnavig (f);
	}
    }
}

/*****************************************************************************/
/*									     */
/*	CHAINEINVERSE							     */
/*									     */
/*  inverse et chaine le pointeur vers la zone des arbres accessible par pt  */
/* et met a jour le marquage.						     */
/*									     */
/*****************************************************************************/
void
chaineinverse (ptptSHORT pt)
{
  ptLONG place1, place2;
//  ptLONG t1,t2;
  int i;

  place1 = (ptLONG) pt;
  place2 = (ptLONG) * place1;

//t1 = (ptLONG )*place1;
//t2 = (ptLONG )*place2;

  i = *place2;
  *place2 = (longInteger) place1;
  *place1 = i;
  if (SHORT_TESTMARQUE ((ptSHORT) place1) == SHORT_INITIALMARK)
    SHORT_SET_FINDECHAINE ((ptSHORT) place1);
  SHORT_SET_MOVEDMARK ((ptSHORT) place2);
//t1 = (ptLONG )*place1;
//t2 = (ptLONG )*place2;

}

/*****************************************************************************/
/*									     */
/*	MARQUERPILE							     */
/*									     */
/* Fonction qui marque les objets a recuperer a partir de la pile.	     */
/*									     */
/*****************************************************************************/
void
marquerpile (void)
{
  ptptSHORT pt;
  int i;
  ptattribute anAttribute;
  ptfilsd f, f1;
  ptnode e;
  ptrepere r;

  pt = GET_PTPTSHORT_FIRST_INSTACKMEMORY ();;

  while (IS_PTPTSHORT_INSTACKMEMORY (pt))
    {
      if (estarbre (*pt))
	{
	  switch (cellCategory (*pt))
	    {
	    case 0:		/* acces */
	      if (SHORT_NONMARQUE ((*pt)))
		{
		  f = (ptfilsd) * pt;
		  f1 = getTM_filsd_root (f);
		  empiler ((ptchar) f1);
		  marquerarbre ();
		}
	      chaineinverse (pt);
	      break;
	    case 1:		/* attribut */
	      if (SHORT_NONMARQUE ((*pt)))
		SHORT_SET_INITIALMARK ((*pt));
	      chaineinverse (pt);
	      break;
	    case 2:		/* ptrepere */
	      if (SHORT_NONMARQUE ((*pt)))
		{
		  SHORT_SET_INITIALMARK ((*pt));
		  r = (ptrepere) * pt;
		  f = REPERE_GET_FILSD_REFERED (r);
		  if (f != (ptfilsd) NIL)
		    {
		      if (SHORT_NONMARQUE ((ptSHORT) f))
			{
			  f = getTM_filsd_root (f);
			  empiler ((ptchar) f);
			  marquerarbre ();
			}
		    }
		}
	      chaineinverse (pt);
	      break;
	    case 3:		/* ptnode */
	      if (SHORT_NONMARQUE ((*pt)))
		{
		  e = (ptnode) * pt;
		  if (NODE_GET_FILSD_DESCRIPTOR (e) != (ptfilsd) NIL)
		    {
		      f = NODE_GET_FILSD_DESCRIPTOR (e);
		      f = getTM_filsd_root (f);
		      empiler ((ptchar) f);
		      marquerarbre ();
		    }
		  else
		    {
		      SHORT_SET_INITIALMARK ((*pt));
		      anAttribute = NODE_GET_ATTRIBUTE (e);
		      if (anAttribute != (ptattribute) NIL)
			SHORT_SET_INITIALMARK ((ptSHORT) anAttribute);
		      f = NODE_GET_FILSD_FIRST (e);
		      if (NODE_GET_ARITY (e) > 0)
			{
			  for (i = 1; i <= NODE_GET_ARITY (e); i++)
			    {
			      SHORT_SET_INITIALMARK ((ptSHORT) f);
			      f++;
			    }
			}
		      else
			{
			  if (NODE_GET_ARITY (e) == -1)
			    SHORT_SET_INITIALMARK ((ptSHORT) f);
			}
		    }
		}
	      chaineinverse (pt);
	      break;
	    default:		/* erreur */
	      break;
	    }
	}
      pt--;
    }
}
