/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include "boolean.h"
#include <stdlib.h>
#include <stdio.h>
#include "StdTypeDefs.h"

#include "error.h"
#include "stat.h"
#include "sigmem.h"

//#ifdef __PKWINDOWS__XX
// #include <windows.h>
//#endif

class t_sigmem
{
public:
  t_sigmem (size_t size, const char *title);
   ~t_sigmem ();
  size_t avail ();
  void *gettop ();
  void *_malloc (size_t size);

private:
    ptchar self_MemTop;
  size_t self_GlobalAllocated;
  size_t self_AvailMem;
  ptchar self_CurPtr;
  // booleans to prevent recursion with ERROR
  boolean self_ErrorMustFree;
  boolean self_ErrorMustUnlock;
  int self_sock;
  // #ifdef __WIN32__
  //#ifdef __PKWINDOWS__XX
  //  HGLOBAL self_handle;
  //#endif
};

typedef t_sigmem *ptt_sigmem;


// PLG WARNING 
// parameter of malloc has type size_t
// the limit of size_t is SIZE_MAX
// its value in ISO C99 is 65535 !!!
// moreover the compiler accepts 
//              fatal_tree << "not enough memory; can not allocate " << (int)size << "bytes.";
// but does not accept
//              fatal_tree << "not enough memory; can not allocate " << (long int)size << "bytes.";

t_sigmem::t_sigmem (size_t size, const char *title)
{
  self_ErrorMustFree = pk_false;
  //#ifdef __PKWINDOWS__XX
  // #ifdef __WIN32__
  //  self_ErrorMustUnlock = pk_false;
  //  self_handle = GlobalAlloc (GMEM_FIXED, size);
  //  if (self_handle == NULL)
  //    self_MemTop = NULL;
  //  else {
  // self_ErrorMustFree = pk_true;
  //    self_MemTop = GlobalLock (self_handle); }
  //#else
  self_MemTop = (ptchar) malloc (size);
  // #endif
  if (self_MemTop == NULL)
    {
      fatal_tree << "not enough memory; can not allocate " << (int) size << "bytes.";
      fatal_tree++;
    }
  self_CurPtr = self_MemTop;
  self_GlobalAllocated = size;
// #ifdef __PKWINDOWS__XX
// // #ifdef __WIN32__
//   self_ErrorMustUnlock = pk_true;
//   self_AvailMem = self_GlobalAllocated - 65536;
// #else
  self_AvailMem = self_GlobalAllocated;
  //#endif
  self_ErrorMustFree = pk_true;
  self_sock = beginstat (self_GlobalAllocated, title);
}

void *
t_sigmem::gettop ()
{
  return (void *) self_MemTop;
}

void *
t_sigmem::_malloc (size_t size)
{
  ptchar OldCurPtr;

  // alignment on words
  size = (((size - 1) / 4) + 1) * 4;

  // check if (size) bytes are still avaible
  if (self_AvailMem < size)
    (fatal_tree << "not enough memory.")++;
  OldCurPtr = self_CurPtr;
  // Update CurPtr and CanAllocate
  self_CurPtr += size;
  self_AvailMem -= size;
  sendstat (self_sock, self_AvailMem);
  // return the old value of CurPtr as the given memory emplacement
  return (void *) OldCurPtr;
}

t_sigmem::~t_sigmem ()
{
  endstat (self_sock);
// #ifdef __PKWINDOWS__XX
// // #ifdef __WIN32__
//   // check if memory should be unlocked
//   if (self_ErrorMustUnlock)
//     GlobalUnlock (self_handle);
//   // check if memory should be set free
//   if (self_ErrorMustFree)
//     GlobalFree (self_handle);
// #else
  // check if memory should be set free
  if (self_ErrorMustFree)
    free (self_MemTop);
  // #endif
}

static ptt_sigmem __FSCMEM__ = NULL;
static ptt_sigmem __PARSEMEM__ = NULL;
static ptt_sigmem __GOMEM__ = NULL;
static ptt_sigmem __SIGMEM__ = NULL;

void *
INITFSCMEM (size_t size)
{
  __FSCMEM__ = new t_sigmem (size, "GC memory");
  return __FSCMEM__->gettop ();
}

void *
ALLOCFSCMEM (size_t size)
{
  if (__FSCMEM__ == NULL)
    (bug_tree << "use INITFSCMEM before ALLOCFSCMEM.")++;
  return __FSCMEM__->_malloc (size);
}

void
FREEFSCMEM (void *)
{
}

void
DONEFSCMEM ()
{
  if (__FSCMEM__ == NULL)
    return;
  delete __FSCMEM__;
  __FSCMEM__ = NULL;
}

void *
INITPARSEMEM (size_t size)
{
  __PARSEMEM__ = new t_sigmem (size, "Parse memory");
  return __PARSEMEM__->gettop ();
}

void *
ALLOCPARSEMEM (size_t size)
{
  if (__PARSEMEM__ == NULL)
    (bug_tree << "use INITPARSEMEM before ALLOCPARSEMEM.")++;
  return __PARSEMEM__->_malloc (size);
}

void
FREEPARSEMEM (void *)
{
}

void
DONEPARSEMEM ()
{
  if (__PARSEMEM__ == NULL)
    return;
  delete __PARSEMEM__;
  __PARSEMEM__ = NULL;
}

/* PLG WARNING: in ISO limit of size_t: SIZE_MAX 65535 !!!!
"long unsigned int" in /usr/lib64/gcc/i686-w64-mingw32/4.7.2/include/stddef.h
 */
void *
INITSIGMEM (size_t size)
{
  __SIGMEM__ = new t_sigmem (size, "Signal memory");
  return __SIGMEM__->gettop ();
}

void *
ALLOCSIGMEM (size_t size)
{
  if (__SIGMEM__ == NULL)
    (bug_tree << "use INITSIGMEM before ALLOCSIGMEM.")++;
  return __SIGMEM__->_malloc (size);
}

void
FREESIGMEM (void *)
{
}

void
DONESIGMEM ()
{
  if (__SIGMEM__ == NULL)
    return;
  delete __SIGMEM__;
  __SIGMEM__ = NULL;
}

void *
INITGOMEM (size_t size)
{
  __GOMEM__ = new t_sigmem (size, "Hierarchy memory");
  return __GOMEM__->gettop ();
}

void *
ALLOCGOMEM (size_t size)
{
  if (__GOMEM__ == NULL)
    (bug_tree << "use INITGOMEM before ALLOCGOMEM.")++;
  return __GOMEM__->_malloc (size);
}

void
FREEGOMEM (void *)
{
}

void
DONEGOMEM ()
{
  if (__GOMEM__ == NULL)
    return;
  delete __GOMEM__;
  __GOMEM__ = NULL;
}

void
DONEMEM ()
{
  DONEFSCMEM ();
  DONEPARSEMEM ();
  DONESIGMEM ();
  DONEGOMEM ();
}
