/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <string.h>
#include "system.h"
#include "varglobal.h"
#include "gm_chaine.h"
#include "chaine.h"
#include "gm_init.h"
#include "gm_pile.h"
#include "li_nav.h"
#include "lta_ctree.h"
#include "lta_cbase.h"
#include "li_attri.h"
#include "io_inter.h"
#include "recuperer.h"
#include "treefront_g.h"

#include "TM_DEBUG.h"

#define _TREE_MEMORY_NOERROR 0
#define _TREE_MEMORY_FILENOTFOUND 1
#define _TREE_MEMORY_NOTENOUGHSPACE 2

#define _TREE_MEMORY_TOOBIG 3



/* ======> debug facilities 
 */
extern void
DEBUG_treeRoot (int indentVal, int depth, const char *mess, trees aSubtree)
{
  ptnode rootNode = getTM_root (aSubtree);
  if (IS_NILTREE (rootNode))
    DEBUGMESS_NODE (indentVal, depth, mess, NIL_node);
  else
    {				/* rootNode is the node X_1 in the tree:
				   [NODE X, arity:1, [FILSD F, PHY:0, POS:0, OP, value:X_1 UP:X] ]
				   with [NODE X_1, arity, <--FD:F; ; ATT ] */
      rootNode = FILSD_GET_NODE_CONTAINER_OF_FIRST (NODE_GET_FILSD_DESCRIPTOR (rootNode));
      DEBUGMESS_NODE (indentVal, 1, "this node is in the following tree", (ptnode) aSubtree);
      DEBUGMESS_NODE (indentVal, depth, mess, rootNode);
    }
}

extern void
DEBUG_treeHyperRoot (int indentVal, int depth, const char *mess, trees aSubtree)
{
  ptnode rootNode = getTM_hyperRoot (aSubtree);
  if (IS_NILTREE (rootNode))
    DEBUGMESS_NODE (indentVal, depth, mess, NIL_node);
  else
    {				/* rootNode is the node X_1 in the tree:
				   [NODE X, arity:1, [FILSD F, PHY:0, POS:0, OP, value:X_1 UP:X] ]
				   with [NODE X_1, arity, <--FD:F; ; ATT ] */
      rootNode = FILSD_GET_NODE_CONTAINER_OF_FIRST (NODE_GET_FILSD_DESCRIPTOR (rootNode));
      DEBUGMESS_NODE (indentVal, 1, "this node is in the following tree", (ptnode) aSubtree);
      DEBUGMESS_NODE (indentVal, depth, mess, rootNode);
    }
}

extern void
DEBUG_subtree (int indentVal, int depth, const char *mess, trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    DEBUGMESS_NODE (indentVal, depth, mess, NIL_node);
  else if (TREEMEMORY_ISTREECOMPONENTADR (aSubtree))
    {
      printf ("aSubtree is %li, the tree memory adress %li\n", (long int) aSubtree, TREEMEMORY_GETTREECOMPONENTPOS (aSubtree));
      fflush (stdout);
      if (NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree) == NIL_filsAny)
	DEBUGMESS_NODE (indentVal, depth, mess, (ptnode) aSubtree);
      else if (NODE_ISIN_LIST ((ptnode) aSubtree))
	DEBUGMESS_SONINLISTNODE (indentVal, depth, mess, NODE_GET_FILSLIST_DESCRIPTOR ((ptnode) aSubtree));
      else
	DEBUGMESS_SONINFIXSIZENODE (indentVal, depth, mess, NODE_GET_FILSD_DESCRIPTOR ((ptnode) aSubtree));
    }
  else
    printf ("###INTERNAL ERROR### in DEBUG_subtree %li is not a tree memory adress\n", (long int) aSubtree);
  fflush (stdout);
}

extern void
DEBUG_attribute (int indentVal, int depth, const char *mess, ptattribute anAttribute)
{
  DEBUGMESS_ATTRIBUTE (indentVal, depth, mess, anAttribute);
}

extern void
DEBUG_aSubreeAttributes (int indentVal, int depth, const char *mess, trees aSubtree)
{
  DEBUGMESS_NODEATTRIBUTELIST (indentVal, depth, mess, (ptnode) aSubtree);
}

/* ========= ========= ========= ========= node navigation functions
 */

/* ======> global property test  
 */
static char nom_fich_rename[4096];

/* ========= any subtree navigation =========*/
extern trees
getTM_father (trees aSubtree)
{
  ptfilsAny aSubtreeDescriptor;

  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  aSubtreeDescriptor = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree);
  if (aSubtreeDescriptor == NIL_filsAny)
    return NIL_trees;
  if (ANYFILS_IS_ROOT (aSubtreeDescriptor))
    return NIL_trees;
  return (trees) ANYFILS_GET_NODE_CONTAINER (aSubtreeDescriptor);
}

/** @return the root subtree that contains aSubtree
 *  @return aSubtree if aSubtree is a root 
 *  @return NIL_trees is aSubtree does not belong to a correct tree
 */
extern trees
getTM_root (trees aSubtree)
{
  ptfilsd rootDescriptor;

  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  rootDescriptor = getTM_filsd_root (NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree));
  if (rootDescriptor == NIL_filsd)
    return NIL_trees;
  return (trees) ANYFILS_GET_NODE_VALUE ((ptfilsAny) rootDescriptor);
}

/** @return the hyper root subtree that contains aSubtree; when a root R belongs to an attribute A then the hyper root of R is the hyper root of A
 *  @return aSubtree if aSubtree is an hyper root 
 *  @return the highest correct root if the largest tree is not correct
 */
extern trees
getTM_hyperRoot (trees aSubtree)
{
  ptfilsd rootDescriptor;

  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  rootDescriptor = getTM_filsd_hyperRoot (NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree));
  if (rootDescriptor == NIL_filsd)
    return NIL_trees;
  return (trees) ANYFILS_GET_NODE_VALUE ((ptfilsAny) rootDescriptor);
}

extern trees
getTM_rightNeighbor (int i, trees aSubtree)
{
  if (i <= 0)
    return NIL_trees;
  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  if (NODE_IS_ROOT ((ptnode) aSubtree))
    return NIL_trees;
  if (NODE_ISIN_LIST ((ptnode) aSubtree))
    return (trees) getTM_node_nNextInlist ((ptnode) aSubtree, i);
  return (trees) getTM_node_nNextInFix ((ptnode) aSubtree, i);
}

extern trees
getTM_leftNeighbor (int i, trees aSubtree)
{
  if (i <= 0)
    return NIL_trees;
  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  if (NODE_IS_ROOT ((ptnode) aSubtree))
    return NIL_trees;
  if (NODE_ISIN_LIST ((ptnode) aSubtree))
    return (trees) getTM_node_nPreviousInlist ((ptnode) aSubtree, i);
  return (trees) getTM_node_nPreviousInFix ((ptnode) aSubtree, i);
}

/** @return the nth son of subtree aSubtree
    @return nil if such a son does not exist or is not a subtree
 */
extern trees
getTM_nChild (int n, trees aSubtree)
{
  ptfilsAny aFilsAny;
  // DEBUGMESS_NODE(10, 4, "entering getTM_nChild aSubtree is", (ptnode)aSubtree);
  // printf("DEBUG treefront::getTM_nChild n is %i\n", n); fflush(stdout);
  if (IS_NILTREE (aSubtree) || (n <= 0))
    return NIL_trees;
  if (NODE_IS_LIST ((ptnode) aSubtree))
    {
      aFilsAny = (ptfilsAny) getTM_node_nFilslist ((ptnode) aSubtree, n);	// (n >= 1, aSubtree is a not NIL list) ==> (return aSubtree[n], NIL if it does not exist)
      // DEBUGMESS_SONINLISTNODE(10, 2, "     in  getTM_nChild list aFilsAny is", (ptfilslist)aFilsAny);
      if (aFilsAny == NIL_filsAny)
	return NIL_trees;
    }
  else if (NODE_GET_ARITY ((ptnode) aSubtree) >= n)
    {
      aFilsAny = (ptfilsAny) NODE_GET_FILSD_N ((ptnode) aSubtree, n);
      // DEBUGMESS_SONINFIXSIZENODE(10, 2, "     in  getTM_nChild fixe aFilsAny is", (ptfilsd)aFilsAny);
    }
  else
    return NIL_trees;
  if (ANYFILS_HAS_USEROPERATOR (aFilsAny))
    {
      // DEBUGMESS_NODE(10, 2, " exiting getTM_nCh<ild result is", ANYFILS_GET_NODE_VALUE(aFilsAny));
      return (trees) ANYFILS_GET_NODE_VALUE (aFilsAny);
    }
  return NIL_trees;
}

extern short
getTM_arity (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return (0);
  if (NODE_IS_LIST ((ptnode) aSubtree))
    return getTM_listNode_sonNumber (aSubtree);
  return (NODE_GET_ARITY (aSubtree));
}

extern void
setTM_operator (trees aSubtree, int op)	/* PLG return void */
{
  if (!(IS_NILTREE (aSubtree)))
    {
      ptfilsAny aSubtreeDescriptor;
      aSubtreeDescriptor = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree);
      if (aSubtreeDescriptor != NIL_filsAny)
	{
	  ANYFILS_SET_OPER (aSubtreeDescriptor, op);
	}
    }
}

/** @return the operator of NN if @param aSubtree is a correct node NN
 *  @return the LTA operator OPUNDEF (-8) if aSubtree is not a correct node
 *  may return a LTA operator (OPTROU -7, OPNIL -6, OPENTIER -5, OPCAR -4, OPCHAINE -3, OPSTAR -2, OPREPERE -1, OPARBRE 07 )
 *  or a user defined operator as a positive integer
 *  ERROR: an INTERNAL ERROR message is printed
 */
extern short
getTM_operator (trees aSubtree)
{
  // DEBUGMESS_NODE(4, 4, "entering getTM_operator", (ptnode)aSubtree);
  ptfilsAny aSubtreeDescriptor;
  short oper;
  if (IS_NILTREE (aSubtree))
    return USEROPER (OPUNDEF);
  aSubtreeDescriptor = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree);
  if (aSubtreeDescriptor == NIL_filsAny)
    {
      printf ("##INTERNAL ERROR## in getTM_operator aSubtree argument is not correctly linked\n");
      fflush (stdout);
      return USEROPER (OPUNDEF);
    }
  return ANYFILS_GET_OPER (aSubtreeDescriptor);
}

/* ========= atoms =========*/
/* ==== unvalued atoms ====*/

/** @return a newly created unvalued hole tree
 */
// PLG hhole ==> makeTM_hole
extern trees
makeTM_hole (void)		/* modif */
{
  ptnode newSubtree;

  newSubtree = newTM_fixSizeNode (1);
  linkTM_filsd_describedNode (newTM_fixSonRoot_emptyTree (), newSubtree, OPTROU);
  // printf("DEBUG treefront:: exiting makeTM_hole, new tree adr is %li, subtree e is %li\n", (ANYFILS_GET_LONGINT_VALUE(f)), (long int)e); fflush(stdout);
  return (trees) newSubtree;
}

/** 
   @return true if aSubtree is a hole
*/
// PLG hishole ==> isTM_hole
extern boolean
isTM_hole (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return isTM_anAtom ((ptnode) aSubtree, OPTROU);
}

/** @return a newly created unvalued atom tree
 *  @param cop is the tree operator 
 */
// PLG hatomnil ==> makeTM_nilAtom
extern trees
makeTM_nilAtom (short cop)	/* modif */
{
  ptnode newSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  newSubtree = newTM_userFixeTreeTop (cop, 1);

  ANYFILS_MAKE_NIL (NODE_GET_FILSD_FIRST (newSubtree));
  return (trees) newSubtree;
}

/** 
   @return true if aSubtree is a nil atom
*/
extern boolean
isTM_nilAtom (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return isTM_anAtom ((ptnode) aSubtree, OPNIL);
}

/* ==== integer atoms ====*/
/** @return a newly created integer valued atom tree
 *  @param cop is the tree operator 
 *  @param integerValue is the value of the atom
 */
// PLG hatomint ==> makeTM_intAtom
extern trees
makeTM_intAtom (short cop, long int integerValue)
{
  ptnode newSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  newSubtree = newTM_userFixeTreeTop (cop, 1);

  ANYFILS_MAKE_LONGINT (NODE_GET_FILSD_FIRST (newSubtree), integerValue);
  return (trees) newSubtree;
}

extern boolean
isTM_intAtom (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return isTM_anAtom ((ptnode) aSubtree, OPENTIER);
}

extern void
setTM_intValue (trees aSubtree, long int integerValue)
{
  if (isTM_intAtom (aSubtree))
    ANYFILS_SET_LONGINT_VALUE (NODE_GET_FILSD_FIRST ((ptnode) aSubtree), integerValue);
}

extern long int
getTM_intValue (trees aSubtree)
{
  if (isTM_intAtom (aSubtree))
    return (ANYFILS_GET_LONGINT_VALUE (NODE_GET_FILSD_FIRST ((ptnode) aSubtree)));
  return 0;
}

/* ==== char atoms ====*/
/** @return a newly created char valued atom tree
 *  @param cop is the tree operator 
 *  @param charValue is the value of the atom
 */
// PLG hatomchar ==> makeTM_charAtom
extern trees
makeTM_charAtom (short cop, char charValue)
{
  ptnode newSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  newSubtree = newTM_userFixeTreeTop (cop, 1);

  ANYFILS_MAKE_CHAR (NODE_GET_FILSD_FIRST (newSubtree), charValue);
  return (trees) newSubtree;
}

extern boolean
isTM_charAtom (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return isTM_anAtom ((ptnode) aSubtree, OPCAR);
}

extern void
setTM_charAtomValue (trees aSubtree, char val)
{
  if (isTM_charAtom (aSubtree))
    ANYFILS_SET_CHAR_VALUE (NODE_GET_FILSD_FIRST ((ptnode) aSubtree), val);
}

extern char
getTM_charAtomValue (trees aSubtree)
{
  if (isTM_charAtom (aSubtree))
    return ANYFILS_GET_CHAR_VALUE (NODE_GET_FILSD_FIRST ((ptnode) aSubtree));	/*modif 11-2-96 due a CC */
  return '\0';
}

/* ==== string atoms ====*/
/** @return a newly created string valued atom tree
 *  @param cop is the tree operator 
 *  @param ich is the value of the atom
 */

// PLG hatomstring ==> makeTM_TMStringAtom
extern trees
makeTM_TMStringAtom (short cop, TMString ich)	/* modif */
{
  ptnode newSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  newSubtree = newTM_userFixeTreeTop (cop, 1);

  ANYFILS_MAKE_TMSTRING (NODE_GET_FILSD_FIRST (newSubtree), ich);
  return (trees) newSubtree;
}

extern void
assignTM_CStringValue2NodeSon (trees aSubtree, char *val, int pos)
{
  ptfilsAny aFilsAny;
  if (IS_NILTREE (aSubtree))
    return;
  if (pos < 1)
    return;
  if (NODE_IS_LIST ((ptnode) aSubtree))
    {
      aFilsAny = (ptfilsAny) getTM_node_nFilslist ((ptnode) aSubtree, pos);	// (pos >= 1, aSubtree is a not NIL list) ==> (return aSubtree[pos], NIL if it does not exist)
      if (aFilsAny == NIL_filsAny)
	return;
    }
  else if (NODE_GET_ARITY ((ptnode) aSubtree) >= pos)
    aFilsAny = (ptfilsAny) NODE_GET_FILSD_N ((ptnode) aSubtree, pos);
  else
    return;
  assignTM_stringValue2FilsAny (NODE_GET_FILSD_FIRST ((ptnode) aSubtree), val);
}

extern boolean
isTM_TMStringAtom (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return isTM_anAtom ((ptnode) aSubtree, OPCHAINE);
}

extern void
setTM_CStringAtomValue (trees aSubtree, char *val)
{
  if (isTM_TMStringAtom (aSubtree))
    assignTM_stringValue2FilsAny (NODE_GET_FILSD_FIRST ((ptnode) aSubtree), val);
}

extern boolean
isTM_TMStringNodeSon (trees aSubtree, int pos)
{
  ptfilsAny aFilsAny;
  if (IS_NILTREE (aSubtree))
    return pk_false;
  aFilsAny = getTM_node_nFilsAny ((ptnode) aSubtree, pos);
  if (aFilsAny == NIL_filsAny)
    return pk_false;
  return (ANYFILS_GET_INNEROPER (aFilsAny) == OPCHAINE);
}

extern TMString
getTM_TMStringAtomValue (trees aSubtree)
{
  if (isTM_TMStringAtom (aSubtree))
    return (ANYFILS_GET_TMSTRING_VALUE (NODE_GET_FILSD_FIRST ((ptnode) aSubtree)));
  return (longInteger) 0;
}

extern TMString
getTM_TMStringValueInNodeSon (trees aSubtree, int ind)
{
  ptfilsAny aFilsAny;

  if (IS_NILTREE (aSubtree) || (ind <= 0))
    return 0;
  if (NODE_GET_ARITY ((ptnode) aSubtree) >= ind)
    {
      aFilsAny = (ptfilsAny) NODE_GET_FILSD_N ((ptnode) aSubtree, ind);
      return ANYFILS_GET_TMSTRING_VALUE (aFilsAny);
    }
  return 0;
}

/** @return true if aSubtree is NIL or has a single son with a common (a language-independent) operator 
 */
extern boolean
hasTM_commonCategory (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_true;
  if (NODE_MAYBE_ATOM ((ptnode) aSubtree))
    {
      ptfilsAny aFilsAny = (ptfilsAny) NODE_GET_FILSD_FIRST ((ptnode) aSubtree);
      return !(ANYFILS_HAS_USEROPERATOR (aFilsAny));
    }
  return pk_false;
}

/* ========= fix size subtree =========*/

// PLG static void
// PLG moveTM_subtree(ptfilsd link2DestinationSon, ptfilsd link2assignedSon)

/* ====  ====*/

/** @return a newly created unary tree
 *  @param cop is the tree operator 
 *  @param ssignedSubtree is the value of the unique component
 */
extern trees
makeTM_unaryTree (short cop, trees assignedSubtree)
{
  ptnode newSubtree;
  ptfilsd link2CurrentSonOfNewSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  // DEBUGMESS_NODE(4,1,"entering makeTM_unaryTree, assigned subtree is", (ptnode)assignedSubtree);

  newSubtree = newTM_userFixeTreeTop (cop, 1);

  link2CurrentSonOfNewSubtree = NODE_GET_FILSD_FIRST (newSubtree);

  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_unaryTree, assignedNode is", (ptnode)assignedSubtree); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree);

  // DEBUGMESS_NODE(4,4," exiting makeTM_unaryTree, resulting subtree is", newSubtree);
  return newSubtree;
}

/** @return a newly created binary tree
 *  @param cop is the tree operator 
 *  @param ssignedSubtree1 is the value of the first component
 *  @param ssignedSubtree2 is the value of the last component
 */
// PLG hmakebinary ==> makeTM_binaryTree
extern trees
makeTM_binaryTree (short cop, trees assignedSubtree1, trees assignedSubtree2)
{
  ptnode newSubtree;
  ptfilsd link2CurrentSonOfNewSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  // DEBUGMESS_NODE(4,1,"entering makeTM_binaryTree, assigned subtree 1 is", assignedSubtree1);
  // DEBUGMESS_NODE(4,1,"                            assigned subtree 2 is", assignedSubtree2);

  newSubtree = newTM_userFixeTreeTop (cop, 2);

  link2CurrentSonOfNewSubtree = NODE_GET_FILSD_FIRST (newSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_binaryTree, assignedNode 1 is", (ptnode)assignedSubtree1); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree1);

  link2CurrentSonOfNewSubtree = FILSD_GET_SUCC (link2CurrentSonOfNewSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_binaryTree, assignedNode 2 is", (ptnode)assignedSubtree2); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree2);

  // DEBUGMESS_NODE(4,4," exiting makeTM_binaryTree, resulting subtree is", newSubtree);
  return newSubtree;
}

/** @return a newly created ternary tree
 *  @param cop is the tree operator 
 *  @param ssignedSubtree1 is the value of the first component
 *  @param ssignedSubtree2 is the value of the second component
 *  @param ssignedSubtree3 is the value of the last component
 */
// PLG hmaketernary ==> makeTM_ternaryTree
extern trees
makeTM_ternaryTree (short cop, trees assignedSubtree1, trees assignedSubtree2, trees assignedSubtree3)
{
  ptnode newSubtree;
  ptfilsd link2CurrentSonOfNewSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  // DEBUGMESS_NODE(4,1,"entering makeTM_ternaryTree, assigned subtree 1 is", assignedSubtree1);
  // DEBUGMESS_NODE(4,1,"                             assigned subtree 2 is", assignedSubtree2);
  // DEBUGMESS_NODE(4,1,"                             assigned subtree 3 is", assignedSubtree3);

  newSubtree = newTM_userFixeTreeTop (cop, 3);

  link2CurrentSonOfNewSubtree = NODE_GET_FILSD_FIRST (newSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_ternaryTree, assignedNode 1 is", (ptnode)assignedSubtree1); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree1);

  link2CurrentSonOfNewSubtree = FILSD_GET_SUCC (link2CurrentSonOfNewSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_ternaryTree, assignedNode 2 is", (ptnode)assignedSubtree2); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree2);

  link2CurrentSonOfNewSubtree = FILSD_GET_SUCC (link2CurrentSonOfNewSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_ternaryTree, assignedNode 3 is", (ptnode)assignedSubtree3); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree3);

  // DEBUGMESS_NODE(4,4," exiting makeTM_ternaryTree, resulting subtree is", newSubtree);
  return newSubtree;
}


/** @return a newly created quaternary tree
 *  @param cop is the tree operator 
 *  @param ssignedSubtree1 is the value of the first component
 *  @param ssignedSubtree2 is the value of the second component
 *  @param ssignedSubtree3 is the value of the third component
 *  @param ssignedSubtree4 is the value of the last component
 */
// PLG hmakequaternary ==> makeTM_quaternaryTree
extern trees
makeTM_quaternaryTree (short cop, trees assignedSubtree1, trees assignedSubtree2, trees assignedSubtree3, trees assignedSubtree4)
{
  ptnode newSubtree;
  ptfilsd link2CurrentSonOfNewSubtree;
  ptfilsd link2CurrentAssignedSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  // DEBUGMESS_NODE(4,1,"entering makeTM_quaternaryTree, assigned subtree 1 is", assignedSubtree1);
  // DEBUGMESS_NODE(4,1,"                                assigned subtree 2 is", assignedSubtree2);
  // DEBUGMESS_NODE(4,1,"                                assigned subtree 3 is", assignedSubtree3);
  // DEBUGMESS_NODE(4,1,"                                assigned subtree 4 is", assignedSubtree4);

  newSubtree = newTM_userFixeTreeTop (cop, 4);

  link2CurrentSonOfNewSubtree = NODE_GET_FILSD_FIRST (newSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_quaternaryTree, assignedNode 1 is", (ptnode)assignedSubtree1); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree1);

  link2CurrentSonOfNewSubtree = FILSD_GET_SUCC (link2CurrentSonOfNewSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_quaternaryTree, assignedNode 2 is", (ptnode)assignedSubtree2); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree2);

  link2CurrentSonOfNewSubtree = FILSD_GET_SUCC (link2CurrentSonOfNewSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_quaternaryTree, assignedNode 3 is", (ptnode)assignedSubtree3); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree3);

  link2CurrentSonOfNewSubtree = FILSD_GET_SUCC (link2CurrentSonOfNewSubtree);
  // DEBUGMESS_NODE(4,2,"entering treefront::makeTM_quaternaryTree, assignedNode 4 is", (ptnode)assignedSubtree4); 
  assignTM_nodeValue2Filsd (link2CurrentSonOfNewSubtree, (ptnode) assignedSubtree4);

  // DEBUGMESS_NODE(4,4," exiting makeTM_quaternaryTree, resulting subtree is", newSubtree);
  return newSubtree;
}

/** Modifying. 
   Replace the n-th son of aTargetSubtree by treeValue. 
   treeValue is removed from its current tree container
   the previous content of this son may be lost
   If aTargetSubtree is NIL then treeValue remains unchanged in its current subtree
   If treeValue is NIL then resetTM_nchild is equivalent to destroy
*/
extern void
resetTM_nchild (trees aTargetSubtree, trees treeValue, int n)
{
  ptfilsAny detachedFilsAnyTree, aTargetFilsAny, treeValueAnyFilsDescriptor;
  ptnode theRemovedNodeValue, auxiNode;
  boolean targetIsList;
  short destroyedOpCode;
  untypedValueType otherValue;

  // printf("DEBUG treefront::entering resetTM_nchild, son number is %i\n", n);
  // DEBUGMESS_NODE(4,3,"entering resetTM_nchild aTargetSubtree is", aTargetSubtree);
  // DEBUGMESS_NODE(4,3,"entering resetTM_nchild aSubtreeValue is", treeValue);
  if (hasTM_commonCategory (aTargetSubtree) || (n <= 0))	// true if aTargetSubtree is NIL
    return;
  if (NODE_IS_LIST ((ptnode) aTargetSubtree))
    {
      aTargetFilsAny = (ptfilsAny) getTM_node_nFilslist ((ptnode) aTargetSubtree, n);	// (n >= 1, aTargetSubtree is a not NIL list) ==> (return aTargetSubtree[n], NIL if it does not exist)
      // DEBUGMESS_SONINLISTNODE(10, 2, "     in  getTM_nChild list aTargetFilsAny is", (ptfilslist)aTargetFilsAny);
      if (aTargetFilsAny == NIL_filsAny)
	return;
      targetIsList = pk_true;
    }
  else if (NODE_GET_ARITY ((ptnode) aTargetSubtree) >= n)
    {
      aTargetFilsAny = (ptfilsAny) NODE_GET_FILSD_N ((ptnode) aTargetSubtree, n);
      // DEBUGMESS_SONINFIXSIZENODE(10, 2, "     in  getTM_nChild fixe aTargetFilsAny is", (ptfilsd)aTargetFilsAny);
      targetIsList = pk_false;
    }
  else
    return;

  if ((ptnode) treeValue == NIL_node)
    if (targetIsList)
      theRemovedNodeValue = unlinkTM_listElement ((ptfilslist) aTargetFilsAny, &destroyedOpCode, &otherValue);	/* may return NIL */
    else
      theRemovedNodeValue = cleanTM_filsAny (aTargetFilsAny, &destroyedOpCode, &otherValue);	/* may return NIL */
  else
    {
      treeValueAnyFilsDescriptor = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) treeValue);
      if (treeValueAnyFilsDescriptor == NIL_filsd)
	{			/* INTERNAL ERROR */
	  printf ("### INTERNAL ERROR ### in resetTM_nchild: incorrectr tree\n");
	  fflush (stdout);
	  return;
	}
      if (aTargetFilsAny == treeValueAnyFilsDescriptor)
	return;
      if (!(ANYFILS_HAS_USEROPERATOR (treeValueAnyFilsDescriptor)))
	return;
      theRemovedNodeValue = cleanTM_filsAny (aTargetFilsAny, &destroyedOpCode, &otherValue);	/* may return NIL */
      ANYFILS_SET_INNEROPER_VALUE (aTargetFilsAny, ANYFILS_GET_INNEROPER (treeValueAnyFilsDescriptor), (ptnode) treeValue);
      if (targetIsList)
	{
	  NODE_LINK_FILSLIST_DESCRIPTOR ((ptnode) treeValue, (ptfilslist) aTargetFilsAny);
	}
      else
	{
	  NODE_LINK_FILSD_DESCRIPTOR ((ptnode) treeValue, (ptfilsd) aTargetFilsAny);
	}
    }

  detachedFilsAnyTree = newTM_fixSonRoot_emptyTree ();
  if (IS_USEROPERATOR (destroyedOpCode))
    {
      ANYFILS_SET_INNEROPER_VALUE (detachedFilsAnyTree, destroyedOpCode, theRemovedNodeValue);
      if (theRemovedNodeValue != NIL_node)
	NODE_LINK_FILSD_DESCRIPTOR (theRemovedNodeValue, detachedFilsAnyTree);
    }
  else
    {
      ANYFILS_SET_INNEROPER_VALUE (detachedFilsAnyTree, destroyedOpCode, otherValue);
    }
  // DEBUGMESS_NODE(4,4," exiting resetTM_nchild aTargetSubtree is", aTargetSubtree);
}

/* ========= list subtree =========*/
static void
linkTM_filslist2itsFilsAnyNode (ptfilslist aFilslist, ptfilsAny assignedFilsAny)
{
  ptnode aNode;;

  if (ANYFILS_HAS_USEROPERATOR (assignedFilsAny))
    {
      aNode = ANYFILS_GET_NODE_VALUE (assignedFilsAny);
      ANYFILS_SET_INNEROPER_VALUE (aFilslist, ANYFILS_GET_INNEROPER (assignedFilsAny), aNode);
      if (aNode != NIL_node)
	{
	  NODE_LINK_FILSLIST_DESCRIPTOR (aNode, aFilslist);
	}
    }
  else
    ANYFILS_SET_INNEROPER_VALUE (aFilslist, ANYFILS_GET_INNEROPER (assignedFilsAny), ANYFILS_GET_UNTYPED_VALUE (assignedFilsAny));
  ANYFILS_MAKE_TROU (assignedFilsAny);
}

/** @return a newly created empty list tree 
 *  @param cop is the tree operator 
 */
// PLG hmakemptylist ==> makeTM_emptyListTree
extern trees
makeTM_emptyListTree (short cop)
{
  ptnode newSubtree;

  if (!(CHECKOPER (cop)))
    return NIL_trees;
  newSubtree = newTM_listNode ();
  linkTM_filsd_describedNode (newTM_fixSonRoot_emptyTree (), newSubtree, INNEROPER (cop));
  return ((trees) newSubtree);
}

extern trees
makeTM_listTree (short cop, trees assignedSubtree)
{
  ptnode newNode;
  ptfilsAny assignedFilsAny;
  ptfilslist aFilslist;

  // printf("DEBUG treefront::entering makeTM_listTree operator is %i \n", cop); fflush(stdout);

  newNode = (ptnode) makeTM_emptyListTree (cop);
  if ((IS_NILTREE (assignedSubtree)) || (newNode == NIL_node))
    return (trees) newNode;
  // DEBUGMESS_NODE(10, 2, "entering  makeTM_listTree tree argument is ", assignedSubtree);

  aFilslist = insertTM_newFilslist_first (NODE_GET_FILSD_FIRST (newNode));

  assignedFilsAny = NODE_GET_ANYFILS_DESCRIPTOR (assignedSubtree);	/* f1 */
  linkTM_filslist2itsFilsAnyNode (aFilslist, assignedFilsAny);	/* attachlist_arbre */

  // DEBUG_treeRoot(10, 3, " exiting makeTM_listTree list tree is ", newNode);
  // DEBUGMESS_SONINANYNODE(10, 2, " exiting makeTM_listTree previous tree argument is ", assignedFilsAny);
  return (trees) newNode;
}

/** insert "insertedSubtree" as the first element of the list subtree "targetSubtreeList"
    insertion fails if "targetSubtreeList"is not a list subtree, or if "insertedSubtree" is nil
    remove "insertedSubtree" from its previous subtree container
    @return nil and arguments remain unchanged if insertion failed
    @return "targetSubtreeList" otherwise
 */
extern trees
insertTM_firstInListTree (trees insertedSubtree, trees targetSubtreeList)
{
  ptfilsAny insertedFilsAny;
  ptfilsd filsdListDescriptor;
  ptfilslist filsList_new;

  if ((IS_NILTREE (insertedSubtree)) || (IS_NILTREE (targetSubtreeList)))
    return NIL_trees;
  if (!(NODE_IS_LIST ((ptnode) targetSubtreeList)))
    return NIL_trees;

  filsdListDescriptor = NODE_GET_FILSD_FIRST ((ptnode) targetSubtreeList);
  filsList_new = insertTM_newFilslist_first (filsdListDescriptor);
  insertedFilsAny = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) insertedSubtree);
  linkTM_filslist2itsFilsAnyNode (filsList_new, insertedFilsAny);
  return (targetSubtreeList);
}

/** append "insertedSubtree" as the last element of the list subtree "targetSubtreeList"
    insertion fails if "targetSubtreeList"is not a list subtree, or if "insertedSubtree" is nil
    remove "insertedSubtree" from its previous subtree container
    @return nil and arguments remain unchanged if insertion failed
    @return "targetSubtreeList" otherwise
 */
extern trees
appendTM_lastInListTree (trees targetSubtreeList, trees insertedSubtree)
{
  ptfilsAny insertedFilsAny;
  ptfilslist filsList_pred;
  ptfilslist filsList_new;

  if ((IS_NILTREE (insertedSubtree)) || (IS_NILTREE (targetSubtreeList)))
    return NIL_trees;
  if (!(NODE_IS_LIST ((ptnode) targetSubtreeList)))
    return NIL_trees;

  filsList_pred = NODE_GET_FILSLIST_FIRST ((ptnode) targetSubtreeList);
  if (filsList_pred == NIL_filslist)
    return (insertTM_firstInListTree (insertedSubtree, targetSubtreeList));
  while (hasTM_successorInList (filsList_pred))
    filsList_pred = FILSLIST_GET_SUCC (filsList_pred);
  filsList_new = insertTM_newFilslist_after (filsList_pred);
  insertedFilsAny = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) insertedSubtree);
  linkTM_filslist2itsFilsAnyNode (filsList_new, insertedFilsAny);
  return (targetSubtreeList);
}

/** insert "insertedSubtree" after nodeInList
    insertion fails if nodeInList does not belong to a list subtree
    remove "insertedSubtree" from its previous subtree container
    @return nil and arguments remain unchanged if insertion failed
    @return "insertedSubtree" otherwise
 */
extern trees
insertTM_beforeSuccInListTree (trees insertedSubtree, trees nodeInList)
{
  ptfilslist filsListSucc, filsList_new;
  ptfilsAny insertedFilsAny;
  // DEBUG_treeRoot(5, 12, "\n entering insertTM_beforeSuccInListTree nodeInList root", nodeInList);
  // DEBUG_subtree(5, 2, "entering insertTM_beforeSuccInListTree nodeInList", nodeInList);
  // DEBUG_subtree(5, 2, "entering insertTM_beforeSuccInListTree insertedSubtree", insertedSubtree);
  if (IS_NILTREE (insertedSubtree))
    return NIL_trees;
  if (IS_NILTREE (nodeInList))
    return NIL_trees;
  if (!NODE_ISIN_LIST (nodeInList))
    return NIL_trees;
  filsListSucc = NODE_GET_FILSLIST_DESCRIPTOR (nodeInList);
  filsList_new = insertTM_newFilslist_before (filsListSucc);
  insertedFilsAny = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) insertedSubtree);
  linkTM_filslist2itsFilsAnyNode (filsList_new, insertedFilsAny);
  // DEBUG_subtree(5, 4, "     in  insertTM_beforeSuccInListTree 4 list", FILSLIST_GET_NODE_CONTAINER(filsListSucc));
  return (insertedSubtree);
}

/** insert "insertedSubtree" after nodeInList
    insertion fails if nodeInList is not a child in a list subtree
    insertion fails if insertedSubtree is NIL
    remove "insertedSubtree" from its previous subtree container
    @return nil and arguments remain unchanged if insertion failed
    @return "insertedSubtree" otherwise
 */
extern trees
insertTM_afterPredInListTree (trees insertedSubtree, trees nodeInList)
{
  ptfilslist filsListPred, filsList_new;
  ptfilsAny insertedFilsAny;

  if (IS_NILTREE (insertedSubtree))
    return NIL_trees;
  if (IS_NILTREE (nodeInList))
    return NIL_trees;
  if (!NODE_ISIN_LIST (nodeInList))
    return NIL_trees;
  filsListPred = NODE_GET_FILSLIST_DESCRIPTOR (nodeInList);
  filsList_new = insertTM_newFilslist_after (filsListPred);
  insertedFilsAny = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) insertedSubtree);
  linkTM_filslist2itsFilsAnyNode (filsList_new, insertedFilsAny);
  return (insertedSubtree);
}

/** insert "insertedSubtree" after position "i" in the list subtree "targetSubtreeList"
    insertion fails if i<1, or if "targetSubtreeList"is not a list subtree with at least "i"  sons, or if "insertedSubtree" is nil
    remove "insertedSubtree" from its previous subtree container
    @return nil and arguments remain unchanged if insertion failed
    @return "insertedSubtree" otherwise
 */
extern trees
insertTM_afterIinListTree (int i, trees insertedSubtree, trees targetSubtreeList)
{
  if (IS_NILTREE (targetSubtreeList))
    return NIL_trees;
  if (!(NODE_IS_LIST ((ptnode) targetSubtreeList)))
    return NIL_trees;
  return insertTM_afterPredInListTree (insertedSubtree, getTM_nChild (i, targetSubtreeList));
}

/** insert "insertedSubtree" before position "i" in the list subtree "targetSubtreeList"
    insertion fails if i<1, or if "targetSubtreeList"is not a list subtree with at least "i" sons, or if "insertedSubtree" is nil
    remove "insertedSubtree" from its previous subtree container
    @return nil and arguments remain unchanged if insertion failed
    @return "insertedSubtree" otherwise
 */
extern trees
insertTM_beforeIinListTree (int i, trees insertedSubtree, trees targetSubtreeList)
{
  if (IS_NILTREE (insertedSubtree))
    return NIL_trees;
  if (i == 1)
    {
      insertTM_firstInListTree (insertedSubtree, targetSubtreeList);
      return insertedSubtree;
    }
  return insertTM_afterIinListTree (i - 1, insertedSubtree, targetSubtreeList);	// remains unchanged if i<1 or targetSubtreeList is not a list subtree or insertedSubtree is NIL
}

extern boolean
isTM_listSubtree (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return NODE_IS_LIST ((ptnode) aSubtree);
}

/** @return the first element in the list aListNode
    aListNode is assumed to be a non nil list node
 */
extern trees
getTM_firstSubtreeInList (trees aListNode)
{
  ptfilslist link2firstNode;

  link2firstNode = (NODE_GET_FILSLIST_FIRST ((ptnode) aListNode));
  if (link2firstNode == NIL_filslist)
    return NIL_trees;
  return (ANYFILS_GET_NODE_VALUE (link2firstNode));
}

extern short
getTM_place (trees aSubtree)
{

  if (IS_NILTREE (aSubtree))
    return 0;
  if (NODE_IS_ROOT ((ptnode) aSubtree))
    return 0;
  if (NODE_ISIN_LIST ((ptnode) aSubtree))
    {
      ptfilslist filslistSubtree = NODE_GET_FILSLIST_DESCRIPTOR ((ptnode) aSubtree);
      short i = 1;
      while (FILSLIST_GET_PRED (filslistSubtree) != NIL_filslist)
	{
	  i++;
	  filslistSubtree = FILSLIST_GET_PRED (filslistSubtree);
	}
      return i;
    }
  return (ANYFILS_GET_POS (NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree)));
}

/** @return true if aSubtree is an empty list subtree
 */
extern boolean
isTM_emptyList (trees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  if (NODE_IS_LIST ((ptnode) aSubtree))
    return ((NODE_GET_FILSLIST_FIRST ((ptnode) aSubtree)) == NIL_filslist);
  return pk_false;
}

/** destroy X[i] in X, where X  is a non empty list (not checked)
    @return X if  X[i] does not exist 
    @return a new tree that contains  X[i] as unique node if  X[i] exists; NIL otherwise
 */
extern trees
removeTM_inListSubtree (trees aSubtreeList, int i)
{
  ptfilslist aFilslistContainer;
  ptfilsAny destroyedFilsList;

  // DEBUGMESS_NODE(4,4,"entering removeTM_inListSubtree aSubtreeList is", (ptnode)aSubtreeList);
  if (IS_NILTREE (aSubtreeList) || (i <= 0))
    return NIL_trees;
  aFilslistContainer = NODE_GET_FILSLIST_DESCRIPTOR ((ptnode) aSubtreeList);
  if (aFilslistContainer == NIL_filslist)
    return NIL_trees;
  if (getCondTM_filsAny_nSon ((ptfilsAny) aFilslistContainer, i, &destroyedFilsList))
    {				/* called with ptfilslist */
      ptnode theRemovedNodeValue;
      untypedValueType otherValue;
      short destroyedOpCode;
      ptfilsd detachedFilsListTree;

      theRemovedNodeValue = unlinkTM_listElement ((ptfilslist) destroyedFilsList, &destroyedOpCode, &otherValue);	/* may return NIL */
      if (theRemovedNodeValue == NIL_node)
	return NIL_trees;
      detachedFilsListTree = newTM_fixSonRoot_emptyTree ();
      ANYFILS_SET_INNEROPER_VALUE (detachedFilsListTree, destroyedOpCode, theRemovedNodeValue);
      NODE_LINK_FILSD_DESCRIPTOR (theRemovedNodeValue, detachedFilsListTree);
      // DEBUGMESS_NODE(4,4," exiting removeTM_inListSubtree aSubtreeList is", (ptnode)aSubtreeList);
      // DEBUGMESS_NODE(4,4," exiting removeTM_inListSubtree detached tree is", ANYFILS_GET_NODE_VALUE(detachedFilsListTree));
      return (trees) theRemovedNodeValue;
    }
  // DEBUGMESS_NODE(4,4," exiting removeTM_inListSubtree aSubtreeList is", (ptnode)aSubtreeList);
  return NIL_trees;
}

/* ========= attributes =========*/


/** @assign valint to the attribute of aSubtree tagged by tagAttribute after creating it if it does not exist
    @param aSubtree should not be nil (checked)
    @param tagAttribute is a correct attribute number (checked)
*/
extern void
attachTM_newIntegerAttribute (trees aSubtree, short tagAttribute, long int valint)
{
  ptattribute newAttr;

  newAttr = getOrCreateTM_attribute (aSubtree, tagAttribute);
  if (newAttr != NIL_attribute)
    {
      ATTRIBUTE_MAKE_INTEGER (newAttr, valint);
    }
}

/** @assign treeValue to the attribute of aSubtree tagged by tagAttribute after creating it if it does not exist
    @param aSubtree should not be nil (checked)
    @param tagAttribute is a correct attribute number (checked)
*/
extern void
attachTM_newTreeAttribute (trees aSubtree, short tagAttribute, trees treeValue)
{
  ptfilsd link2TreeValue;
  ptattribute newAttr;
  if (IS_NILTREE (treeValue))
    link2TreeValue = newTM_fixSonRoot_emptyTree ();
  else
    link2TreeValue = NODE_GET_FILSD_DESCRIPTOR ((ptnode) treeValue);

  newAttr = getOrCreateTM_attribute (aSubtree, tagAttribute);
  attachTM_treeAttribute (newAttr, link2TreeValue);
}

/** @return the number of attributes attached to aSubtree, 0 when aSubtree is nil*/
extern int
getTM_numberOfAttributes (trees aSubtree)
{
  int i;
  ptattribute att;

  if (aSubtree == NIL_trees)
    return 0;
  att = NODE_GET_ATTRIBUTE ((ptnode) aSubtree);
  if (att = (ptattribute) NIL)
    return 0;
  i = 1;
  while (ATTRIBUTE_GET_SUCC (att) != (ptattribute) NIL)
    {
      att = ATTRIBUTE_GET_SUCC (att);
      i = i + 1;
    }
  return i;
}

/** @return true if aSubtree has an attribute tagged by tagAttribute
    STACK free operation
*/
extern boolean			/* PLG TO BE CHECKED (entete ?? */
hasTM_taggedAttribute (trees aSubtree, short tagAttribute)
{
  ptattribute attrRes;
  if (aSubtree == NIL_trees)
    return pk_false;
  return getCondTM_taggedAttribute ((ptnode) aSubtree, tagAttribute, &attrRes);
}

extern ptattribute
getTM_firstAttribute (trees aSubtree)
{
  if (aSubtree == NIL_trees)
    return NIL_attribute;
  return NODE_GET_ATTRIBUTE ((ptnode) aSubtree);
}

/** @return the integer value of the attribute of aSubtree tagged by tagAttribute
    @param aSubtree should not be nil (checked)
    @return 0 if the operation fails
    STACK remains unchanged
*/
extern long int
getTM_integerValueOfTaggedAttributeInSubtree (trees aSubtree, short tagAttribute)
{
  ptattribute att;
  if (aSubtree == NIL_trees)
    return 0;
  if (getCondTM_typedAttribute ((ptnode) aSubtree, tagAttribute, OPENTIER, &att))
    {
      return ATTRIBUTE_GET_LONGINT_VALUE (att);
    }
  return 0;
}

/** @return the attribute of aSubtree tagged by tagAttribute 
    @param the tree aSubtree should not be nil (checked)
*/
extern trees
getTM_treeValueOfTaggedAttributeInSubtree (trees aSubtree, short tagAttribute)
{
  ptattribute anAttribute;
  ptfilsAny filsAnyRoot;
  ptfilsd filsdAttr;
  short t;

  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  if (getCondTM_taggedAttribute ((ptnode) aSubtree, tagAttribute, &anAttribute))
    {
      if ((ATTRIBUTE_GET_INNEROPER (anAttribute)) == OPARBRE)
	{
	  filsdAttr = ATTRIBUTE_GET_FILSD_VALUE (anAttribute);
	  return ANYFILS_GET_NODE_VALUE (filsdAttr);
	}
    }
  return NIL_trees;
}

/** detach from "aSubtree" the attribute tagged by "tagAttribute" 
    @return true: the attribute has been found and correctly removed
    @return false: the attribute has not been found or is not correctly removed
*/
extern boolean
removeTM_taggedAttribute (trees aSubtree, short tagAttribute)
{
  ptfilsd filsdAttributeValue;
  ptattribute anAttribute;
  ptnode aNode;

  aNode = (ptnode) aSubtree;
  if (getCondTM_taggedAttribute (aNode, tagAttribute, &anAttribute))
    {
      detachTM_attribute (aNode, anAttribute);
      if (ATTRIBUTE_GET_INNEROPER (anAttribute) == OPARBRE)
	{
	  filsdAttributeValue = ATTRIBUTE_GET_FILSD_VALUE (anAttribute);
	  aNode = FILSD_GET_NODE_CONTAINER_OF_FIRST (filsdAttributeValue);
	  NODE_SET_FILSD_DESCRIPTOR (aNode, (ptfilsd) NIL);
	}
      return pk_true;
    }
  return pk_false;
}

extern ptattribute
getTM_succAttribute (ptattribute anAttribute)
/*rend la liste d'attributs qui suit anAttribute.*/
{
  if (anAttribute == NIL_attribute)
    return NIL_attribute;
  return (ATTRIBUTE_GET_SUCC (anAttribute));
}

extern short
getTM_tagAttribute (ptattribute anAttribute)
/*rend le tag de anAttribute.*/
{
  if (anAttribute == NIL_attribute)
    return 0;
  return ATTRIBUTE_GET_TAG (anAttribute);
}

extern boolean
isTM_intAttribute (ptattribute anAttribute)
/*rend true si la valeur de anAttribute est un entier.*/
{
  if (anAttribute == NIL_attribute)
    return pk_false;
  return (ATTRIBUTE_GET_INNEROPER (anAttribute) == OPENTIER);
}

extern boolean
isTM_treeAttribute (ptattribute anAttribute)
/*rend true si la valeur de anAttribute  est un arbre.*/
{
  if (anAttribute == NIL_attribute)
    return pk_false;
  return (ATTRIBUTE_GET_INNEROPER (anAttribute) == OPARBRE);
}

extern long int
getTM_intValueAttribute (ptattribute anAttribute)
/*rend la valeur entière de anAttribute, 0 par défaut.*/
{
  if (anAttribute == NIL_attribute)
    return 0;
  return ATTRIBUTE_GET_LONGINT_VALUE (anAttribute);
}

extern trees
getTM_treeValueAttribute (ptattribute anAttribute)
/*rend la valeur arbre de anAttribute, NIL par défaut.*/
{
  ptfilsd filsdRoot;
  if (anAttribute == NIL_attribute)
    return NIL_trees;
  filsdRoot = ATTRIBUTE_GET_FILSD_VALUE (anAttribute);	/* 1 */
  if (filsdRoot == NIL_filsd)
    return NIL_trees;
  // DEBUGMESS_SONINFIXSIZENODE(4,4,"in getTM_treeValueAttribute, filsdRoot is",filsdRoot);
  if (ANYFILS_IS_ROOT (filsdRoot))
    {
      return (trees) ANYFILS_GET_NODE_VALUE (filsdRoot);
    }
  return NIL_trees;

  // return (trees)(filsdRoot);
}

/** @return the subtree to which "aSubtree" is a tree attribute
*/
extern trees
getTM_treeAttributeOwner (trees aSubtree)
{
  ptfilsAny filsdRoot;
  ptattribute att;

  // DEBUGMESS_NODE(10, 4, "entering getTM_treeAttributeOwner, aSubtree parameter is", (ptnode)aSubtree);
  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  filsdRoot = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree);
  if (ANYFILS_IS_ROOT (filsdRoot))
    {
      // DEBUGMESS_SONINFIXSIZENODE(10, 2, "     in  getTM_treeAttributeOwner 1, filsdRoot is", filsdRoot);
      att = FILSD_GET_ATTRIBUTE_TREECONTAINER (filsdRoot);
      if (att == NIL_attribute)
	{
	  printf ("##INTERNAL ERROR## in getTM_treeAttributeOwner 1, the subtree argument is not an attribute\n");
	  fflush (stdout);
	  return NIL_trees;
	}
      // DEBUGMESS_ATTRIBUTE(10, 2, " exiting getTM_treeAttributeOwner attribute is", att);
      // DEBUGMESS_NODE(10,6, " exiting getTM_treeAttributeOwner attribute is", ATTRIBUTE_GET_NODE_OWNER(att));
      return (trees) ATTRIBUTE_GET_NODE_OWNER (att);
    }
  printf ("##INTERNAL ERROR## in getTM_treeAttributeOwner 2, the subtree argument is not an attribute\n");
  fflush (stdout);
  return NIL_trees;
}

/** Copying. 
   @return a copy TT of aSubtree
   @return  nil if aSubtree is nil
   if copyAttributes is true TT contains a copy of attributes otherwise attributes are NIL in TT
*/
extern trees
copyTM_subtree (trees aSubtree, boolean copyAttributes)
{
  ptfilsAny aFilsAny;
  // DEBUG_MEMORY("entering copyTM_subTree");
  // DEBUG_STACK_FULL("entering copyTM_subTree");
  if (IS_NILTREE (aSubtree))
    return NIL_trees;
  // DEBUGMESS_NODE(0, 10, "entering copyTM_subtree source subtree is", (ptnode)aSubtree);
  aFilsAny = NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) aSubtree);
  // DEBUGMESS_SONINANYNODE(10, 4, "     in  copyTM_subtree srcTree pere is", aFilsAny);
  aFilsAny = copyTM_FilsAnySubtree (aFilsAny, copyAttributes);
  // DEBUGMESS_SONINANYNODE(10, 4, " exiting copyTM_subtree tgtTree pere is", aFilsAny);
  // DEBUGMESS_NODE(10, 4, " exiting copyTM_subtree tgtTree is", ANYFILS_GET_NODE_VALUE(aFilsAny));
  if (aFilsAny == NIL_filsAny)
    return NIL_trees;
  // DEBUGMESS_NODE(0, 10, " exiting copyTM_subtree source subtree is", (ptnode)(aSubtree));
  // DEBUGMESS_NODE(0, 10, " exiting copyTM_subtree resulting tree is", (ptnode)(ANYFILS_GET_NODE_VALUE(aFilsAny)));
  // DEBUG_MEMORY(" exiting copyTM_subTree");
  // DEBUG_STACK_FULL(" exiting copyTM_subTree");
  return (ptnode) (ANYFILS_GET_NODE_VALUE (aFilsAny));
}

/* ========= Saving-Loading =========*/

/** load the tree content of fileName
    @return the loaded tree
    @return NIL_tree when operation failed
 */
extern trees
loadTM_tree (char *fileName)
{
  FILE *fdl;
  ptfilsd aFilsAnyRoot;
  int treeSize;

  // printf("DEBUG treefront::entering loadTM_tree, file name is %s\n", fileName); fflush(stdout);
  // DEBUG_MEMORY("entering loadTM_tree");
  // DEBUG_STACK_FULL("entering loadTM_tree");
  fdl = initload (fileName);
  if (fdl == (FILE *) NULL)	/* file is closed */
    return NIL_trees;
  treeSize = read_binaryInteger (fdl);
  // printf("DEBUG treefront::     in loadTM_tree 1 size is <<%i>>\n", treeSize); fflush(stdout);
  if (treeSize == 0)
    {
      fclose (fdl);
      return NIL_trees;
    }
  if (ISNOTAVAILABLE_TREEMEMORY (treeSize))
    {
      fclose ((FILE *) fdl);
      if ((long int) treeSize > getTreeMemorySize ())
	{
	  printf ("###ERROR###: Attempt to load a too large file or a file that is not a graphic Signal object\n", treeSize);
	  fflush (stdout);
	  // DEBUG_subtree(4,2," exiting loadTM_tree", NIL_node);
	  return NIL_trees;
	}
      /* SHOULD TRY GARBAGE */
      // DEBUG_subtree(4,2," exiting loadTM_tree", NIL_node);
      printf ("###ERROR LIMIT### this Polychrony version cannot garbage trees during loading\n");
      fflush (stdout);
      return NIL_trees;
    }
  // printf("DEBUG treefront::     in loadTM_tree 2\n"); fflush(stdout);
  chargement (fdl, &aFilsAnyRoot);
  // printf("DEBUG treefront::     in loadTM_tree 3\n"); fflush(stdout);
  fclose ((FILE *) fdl);
  if (aFilsAnyRoot == (ptfilsd) NIL)
    return NIL_trees;
  // DEBUG_subtree(4,40," exiting loadTM_tree * resulting tree is", (trees)ANYFILS_GET_NODE_VALUE(aFilsAnyRoot));
  // DEBUG_MEMORY(" exiting * loadTM_tree");
  // DEBUG_STACK_FULL(" exiting loadTM_tree");
  return (trees) (ANYFILS_GET_NODE_VALUE (aFilsAnyRoot));
}

extern boolean
saveTM_subTree (char *fileName, trees srcTree)
{
  FILE *tgtFile;
  int treeSize;
  boolean b;

  // printf("DEBUG treefront::entering saveTM_subTree, file name is %s\n", fileName); fflush(stdout);
  // DEBUG_MEMORY("entering saveTM_subTree");
  // DEBUG_STACK_FULL("entering saveTM_subTree");
  if (TM_existFile (fileName))
    {
      strcpy ((char *) nom_fich_rename, fileName);
      strcat ((char *) nom_fich_rename, "~");
      rename ((char *) fileName, nom_fich_rename);
      b = pk_true;
    }
  else
    b = pk_false;
  tgtFile = initsave (fileName);
  if (tgtFile != 0)
    {
      // DEBUGMESS_SONINANYNODE(4, 40, "entering save, ", NODE_GET_ANYFILS_DESCRIPTOR((ptnode)srcTree));
      treeSize = sauvegarde  (pk_true, pk_false, tgtFile, NODE_GET_ANYFILS_DESCRIPTOR ((ptnode) srcTree));
      // printf("DEBUG treefront::saveTM_subTree size is <<%i>>\n", treeSize); fflush(stdout);
      finsave (tgtFile, treeSize);
      // DEBUG_MEMORY(" exiting saveTM_subTree");
      // DEBUG_STACK_FULL(" exiting saveTM_subTree");
      return pk_true;
    }
  if (b)
    rename (nom_fich_rename, (char *) fileName);
  return pk_false;
}

/** required initial action that allocates memory space
    the implicite parameter size of the memory can be change using the following environment variables
         "T_MEM_ARB" (total size in byte allocated to trees and stack)
         "T_MEM_CHAIN" (total size in byte allocated to strings)
 */
extern void
initTM_memory (void)
{
  initmem ();
  // DEBUG_MEMORY(" exiting initTM_memory");
}

/** make static the tree memory used until its call
 */
extern void
resetTM_treeMemoryStart (void)
{
  RESET_TREEMEMORYSTART ();	/* DEB_ARBRE = LIB_ARBRE */
}

/** open a temporary area in tree memory
    openTM_temporaryTreeMemory and closeTM_temporaryTreeMemory must alternate; openTM_temporaryTreeMemory is the frist allowed action in this alternation
 */
extern void
openTM_temporaryTreeMemory (void)
{
  if (inTemporaryTreeMemory)
    return;
  TEMPORARY_LIB_ARBRE = LIB_ARBRE;
  inTemporaryTreeMemory = pk_true;
}

/** close a temporary area in tree memory
    all subtrees built since previous openTM_temporaryTreeMemory occurrence are lost
 */
extern void
closeTM_temporaryTreeMemory (void)
{
  if (inTemporaryTreeMemory)
    {
      LIB_ARBRE = TEMPORARY_LIB_ARBRE;
      inTemporaryTreeMemory = pk_false;
    }
}

/** clear the tree memory non static space (see resetTM_treeMemoryStart) and init other memory areas
*/
extern void
clearTM_userDataMemory (void)
{
  EMPTY_TREEMEMORY ();		/* LIB_ARBRE = DEB_ARBRE */
  clearTM_stringRegisters ();
  initTM_stringMemory ();
}

/** @return the available space in tree memory
 */
extern long int
getTM_availableTreeSpace ()
{
  return AVAILABLE_TREEMEMORY ();
}

/** @return the total amount of non static tree memory used (including holes)
 */
extern long int
getTM_usedTreeSpace ()
{
  return GET_TREEMEMORY_USEDSIZE ();
}

/** @return true if the total amount of available contiguous tree memory space 
    is smaller that the required space (minimalSpace)
 */
extern boolean
isNotTM_availableTreeSpace (int minimalSpace)
{
  return (AVAILABLE_TREEMEMORY () < minimalSpace);
}

extern void
recup_mem (pttrees pt1, pttrees pt2, pttrees pt3, pttrees pt4, pttrees pt5)
{
  ptfilsd f;
  trees e1, e2, e3, e4, e5;

  if (pt1 == (pttrees) NIL)
    {
      e1 = NIL_trees;
      f = NIL_filsd;
    }
  else
    {
      e1 = *pt1;
      if (IS_NILTREE (e1))
	f = NIL_filsd;
      else
	f = (ptfilsd) NODE_GET_FILSD_DESCRIPTOR (e1);
    }
  empiler ((ptchar) f);
  if (pt2 == (pttrees) NIL)
    {
      e2 = NIL_trees;
      f = NIL_filsd;
    }
  else
    {
      e2 = *pt2;
      if (IS_NILTREE (e2))
	f = NIL_filsd;
      else
	f = (ptfilsd) NODE_GET_FILSD_DESCRIPTOR (e2);
    }
  empiler ((ptchar) f);
  if (pt3 == (pttrees) NIL)
    {
      e3 = NIL_trees;
      f = NIL_filsd;
    }
  else
    {
      e3 = *pt3;
      if (IS_NILTREE (e3))
	f = NIL_filsd;
      else
	f = (ptfilsd) NODE_GET_FILSD_DESCRIPTOR (e3);
    }
  empiler ((ptchar) f);
  if (pt4 == (pttrees) NIL)
    {
      e4 = NIL_trees;
      f = NIL_filsd;
    }
  else
    {
      e4 = *pt4;
      if (IS_NILTREE (e4))
	f = NIL_filsd;
      else
	f = (ptfilsd) NODE_GET_FILSD_DESCRIPTOR (e4);
    }
  empiler ((ptchar) f);
  if (pt5 == (pttrees) NIL)
    {
      e5 = NIL_trees;
      f = NIL_filsd;
    }
  else
    {
      e5 = *pt5;
      if (IS_NILTREE (e5))
	f = NIL_filsd;
      else
	f = (ptfilsd) NODE_GET_FILSD_DESCRIPTOR (e5);
    }
  empiler ((ptchar) f);

  recuperation ();

  f = (ptfilsd) depiler ();
  if (pt5 != (pttrees) NIL)
    {
      if (f != NIL)
	(*pt5) = (ANYFILS_GET_NODE_VALUE (f));
      else
	(*pt5) = NIL_trees;
    }
  f = (ptfilsd) depiler ();
  if (pt4 != (pttrees) NIL)
    {
      if (f != NIL)
	(*pt4) = (ANYFILS_GET_NODE_VALUE (f));
      else
	(*pt4) = NIL_trees;
    }
  f = (ptfilsd) depiler ();
  if (pt3 != (pttrees) NIL)
    {
      if (f != NIL)
	(*pt3) = (ANYFILS_GET_NODE_VALUE (f));
      else
	(*pt3) = NIL_trees;
    }
  f = (ptfilsd) depiler ();
  if (pt2 != (pttrees) NIL)
    {
      if (f != NIL)
	(*pt2) = (ANYFILS_GET_NODE_VALUE (f));
      else
	(*pt2) = NIL_trees;
    }
  f = (ptfilsd) depiler ();
  if (pt1 != (pttrees) NIL)
    {
      if (f != NIL)
	(*pt1) = (ANYFILS_GET_NODE_VALUE (f));
      else
	(*pt1) = NIL_trees;
    }
  clearTM_stringRegisters ();
}

/**
   It prints the state of the tree memory.
 */
extern void
printTreeMemoryState ()
{
  DEBUG_MEMORY (" Current state Treee memory");
}
