/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** Attribute managing: number of attributes. 
    @return the number of attributes assigned to an operand tree (aSubtree). 
            0 if aSubtree is nil
*/
extern int
nattrnumber (ntrees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return 0;
  return getTM_numberOfAttributes ((trees) aSubtree);
}

/** Attribute managing:
    @return the tag of anAttribute
           -1 if anAttribute is nil
*/
extern int
getTagAttribute (nattribute anAttribute)
{
  if (IS_NILATTRIBUTE (anAttribute))
    return -1;
  return getTM_tagAttribute ((ptattribute) anAttribute);
}

/** Attribute managing:
    @return true if the value of anAttribute is an integer
           false if anAttribute is nil
*/
extern boolean
isIntAttribute (nattribute anAttribute)
{
  if (IS_NILATTRIBUTE (anAttribute))
    return pk_false;
  return isTM_intAttribute ((ptattribute) anAttribute);
}

/** Attribute managing:
    @return true if the value of anAttribute is a tree
           false if anAttribute is nil
*/
extern boolean
isTreeAttribute (nattribute anAttribute)
{
  if (IS_NILATTRIBUTE (anAttribute))
    return pk_false;
  return isTM_treeAttribute ((ptattribute) anAttribute);
}

/** Attribute managing:
    @return the int value of anAttribute
           -1 if anAttribute is nil
*/
extern long int
getIntValueAttribute (nattribute anAttribute)
{
  long int val;
  // printf("DEBUG entering getIntValueAttribute\n"); fflush(stdout);
  if (IS_NILATTRIBUTE (anAttribute))
    {
      return -1;
    }
  val = getTM_intValueAttribute ((ptattribute) anAttribute);
  // printf("DEBUG  exiting getIntValueAttribute, val is %li\n", val); fflush(stdout);
  return val;
}

/** Attribute managing:
    @return the tree value of anAttribute
            nil  if anAttribute is nil
*/
extern ntrees
getTreeValueAttribute (nattribute anAttribute)
{
  if (IS_NILATTRIBUTE (anAttribute))
    return NILTREE;
  return (ntrees) getTM_treeValueAttribute ((ptattribute) anAttribute);
}

/**
  Attribute managing:builting. 
  Assign an integer attribute (valint) to a tree (aSubtree). 
  Null if aSubtree is nil
  @param aSubtree (in) a tree. 
  @param num (in) attribute number. 
  @param valint (in) int value of the attribute. 
*/
extern void
nmakeattrint (ntrees aSubtree, short num, long int valint)
{
  if (IS_NILTREE (aSubtree))
    return;
  // printf("DEBUG      in  nmakeattrint attributed tree adr is %li, attribute number is %i, integer value is %li\n", (long int)aSubtree, num, valint); fflush(stdout);
  attachTM_newIntegerAttribute ((trees) aSubtree, num, valint);
}


/**
  Attribute managing:builting. 
  Assign a pointer attribute (valptr) to the tree e. 
  Null if aSubtree is nil
  @param e (in) a tree. 
  @param num (in) attribute number. 
  @param valptr (in) pointer value of the attribute. 
*/
extern void
nmakeattrptr (ntrees aSubtree, short num, void *valptr)
{
  if (IS_NILTREE (aSubtree))
    return;
  // printf("DEBUG      in  nmakeattrptr attributed tree adr is %li, attribute number is %i, pointer value is %li\n", (long int)aSubtree, num, valptr); fflush(stdout);
  attachTM_newIntegerAttribute ((trees) aSubtree, num, (pointerAsInteger) valptr);
}

/**
  Attribute managing:builting. 
  Assigned a string attribute (valstr) to the tree e. 
  Null if aSubtree is nil
  @param e (in) a tree. 
  @param num (in) attribute number. 
  @param valstr (in) string value of the attribute. 
*/
extern void
nmakeattrstr (ntrees aSubtree, short num, ptrepstr valstr)
{
  if (IS_NILTREE (aSubtree))
    return;
  // printf("DEBUG      in  nmakeattrstr attributed tree adr is %li, attribute number is %i, string value is %s\n", (long int)aSubtree, num, repstr2string(valstr)); fflush(stdout);
  attachTM_newIntegerAttribute ((trees) aSubtree, num, (pointerAsInteger) valstr);
}

/**
  Attribute managing:builting. 
  Assigned a tree attribute (valtree) to the tree e. 
  Null if aSubtree is nil
  @param e (in) a tree. 
  @param num (in) attribute number. 
  @param valtree (in) tree value of the attribute. 
*/
extern void
nmakeattrntrees (ntrees aSubtree, short num, ntrees valtree)
{
  if (IS_NILTREE (aSubtree))
    return;
  // printf("DEBUG  \n ===================================================================================\n");
  // printf("     entering nmakeattrntrees attributed tree adr is %li, attribute number is %i, value tree adr is %li\n", (long int)aSubtree, (int)num, (long int)valtree); fflush(stdout);
  attachTM_newTreeAttribute ((trees) aSubtree, num, (trees) valtree);
  // printf("DEBUG      in  nmakeattrntrees the number of attributes is %i\n", nattrnumber(aSubtree)); fflush(stdout);
}


/**
  Attribute managing: testing if an attribute is assigned to a tree. 
  @param aSubtree (in) a tree. 
  @param num (in) attribute number. 
  @return true if an attribute with num as attribute number is assigned to the operand tree e. 
  false otherwise. 
*/
extern boolean
nattrexist (ntrees aSubtree, short num)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return hasTM_taggedAttribute ((trees) aSubtree, num);
}


/**
  Attribute managing: getting attribute value. 
  The attribute, numbered num, is assumed to be of int type. 
  @param aSubtree  (in) a tree. 
  @param num (in) attribute number. 
  @return the int value of the attribute, numbered num, assigned to the operand tree e. 
           -1 if aSubtree is nil
*/
extern long int
nattrint (ntrees aSubtree, short num)
{
  long int val;
  // printf("DEBUG entering nattrint\n"); fflush(stdout);
  if (IS_NILTREE (aSubtree))
    return -1;
  val = getTM_integerValueOfTaggedAttributeInSubtree ((trees) aSubtree, num);
  // printf("DEBUG  exiting nattrint, val is %li\n", val); fflush(stdout);
  return val;
}

/**
  Attribute managing: getting attribute value. 
  The attribute numbered num is assumed to be of pointer type. 
  @param aSubtree  (in) a tree. 
  @param num (in) attribute number. 
  @return the (pointer) value of the attribute, numbered num, assigned to the operand tree e. 
           nil if aSubtree is nil
*/
extern void *
nattrptr (ntrees aSubtree, short num)
{
  if (IS_NILTREE (aSubtree))
    return (void *) 0;
  return (void *) getTM_integerValueOfTaggedAttributeInSubtree ((trees) aSubtree, num);
}

/**
  Attribute managing: getting attribute value. 
  The attribute , numbered num, is assumed to be of string type. 
  @param e  (in) a tree. 
  @param num (in) attribute number. 
  @return the (string) value of the attribute, numbered num, assigned to the operand tree e. 
*/
extern ptrepstr
nattrstr (ntrees aSubtree, short num)
{
  if (IS_NILTREE (aSubtree))
    return nilstr;
  return (ptrepstr) getTM_integerValueOfTaggedAttributeInSubtree ((trees) aSubtree, num);
}

/**
  Attribute managing: getting attribute value. 
  The attribute, numbered num, is assumed to be of ntrees type. 
  @param e  (in) a tree. 
  @param num (in) attribute number. 
  @return the (ntrees) value of the attribute, numbered num, assigned to the operand tree e. 
*/
extern ntrees
nattrntrees (ntrees aSubtree, short num)
{
  if (IS_NILTREE (aSubtree))
    return NILTREE;
  return (ntrees) getTM_treeValueOfTaggedAttributeInSubtree ((trees) aSubtree, num);
}


/**
  Attribute managing: deleting an attribute assigned to a tree. 
  Removes, from the set of attributes assigned to the operand tree aSubtree, 
  the attribute with aTag as attribute number. 
  @param aSubtree  (in) a tree. 
  @param aTag (in) attribute number. 
  @return true if the attribute num was assigned to the operand tree e, false otherwise. 
*/
extern boolean
ndestroyattr (ntrees aSubtree, short aTag)
{
  if (IS_NILTREE (aSubtree))
    return pk_false;
  return removeTM_taggedAttribute ((trees) aSubtree, aTag);
}

/**
   Attribute managing: getting a link from a tree-attribute to an another tree. 
   @param aSubtree  (in) a tree. 
   @return the tree at which the tree-attribute e is assigned. 
*/
extern ntrees
nattrcarrier (ntrees aSubtree)
{
  if (IS_NILTREE (aSubtree))
    return NILTREE;
  return (ntrees) getTM_treeAttributeOwner ((trees) aSubtree);
}
