/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** garbage.
   Reseting tree memory (user tree area and registers).
*/
extern void
ntrees_memoryReset ()
{
  clearTM_userDataMemory ();
}

/** Data created in tree memory are made persistent
    Should only be used at the initialisation process
*/
extern void
ntrees_setExistingDataPersistent ()
{
  resetTM_treeMemoryStart ();
}

/** garbage. 
   It applies garbage and returns the new locations for the trees p1, p2, p3, p4, p5.
*/
extern void
ntrees_recup (ntrees * p1, ntrees * p2, ntrees * p3, ntrees * p4, ntrees * p5)
{
  recup_mem ((pttrees) p1, (pttrees) p2, (pttrees) p3, (pttrees) p4, (pttrees) p5);
}

/** garbage : Testing size of available area in tree memory.
   It returns pk_true if the size of free Tree memory is small (garbage seems necessary).
*/
extern boolean
ntrees_notEnoughSpace (int minimalSize)
{
  return (isNotTM_availableTreeSpace (minimalSize));
}

/** garbage misc.
   Store current index of the memory tree (free area). 
   This index should be used to reset the memory using the ntrees_memoryRestoreIndex function.
   It allows to build trees from this point and then to delete them after printing for example.
   \see{ntrees_memoryRestoreIndex}
*/
extern void
ntrees_memoryStoreIndex ()
{
  // printf("DEBUG entering ntrees_memoryStoreIndex: used tree memory size is %li, available tree memory size is %li\n", getTM_usedTreeSpace(), getTM_availableTreeSpace()); fflush(stdout);
  openTM_temporaryTreeMemory ();
}

/** garbage misc. 
   Sets the memory tree index (free area) to the index stored by the previous 
   call to ntrees_memoryStoreIndex function.
   \see{ntrees_memoryStoreIndex}
*/
extern void
ntrees_memoryRestoreIndex ()
{
  // printf("DEBUG entering ntrees_memoryRestoreIndex: used tree memory size is %li, available tree memory size is %li\n", getTM_usedTreeSpace(), getTM_availableTreeSpace()); fflush(stdout);
  closeTM_temporaryTreeMemory ();
  // printf("DEBUG                                     available tree memory size becomes %li\n", getTM_availableTreeSpace()); fflush(stdout);
}
