/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/// The followings don't reference ptrepstr type.
extern boolean
isDecimalDigit (char c)
{
  return ((boolean) (((unsigned) c >= (unsigned) ('0')) && ((unsigned) c <= (unsigned) ('9'))));
}

///
extern boolean
isupperletter (char c)
{
  return ((boolean) (((unsigned) c >= (unsigned) ('A')) && ((unsigned) c <= (unsigned) ('Z'))));
}

/** This function deletes the suffix of the name file filename using the separator ?sep.
    The sep separator is kept if keepSep.
*/
extern boolean
strKillSuffix (char *filename, char sep, boolean keepSep)
{
  char c;
  int lg, i;

  lg = strlen (filename);
  i = 0;
  while (i < lg)
    {
      c = *(filename + lg - i);
      if (c == sep)
	{
	  if (keepSep)
	    *(filename + lg - i + 1) = '\0';
	  else
	    *(filename + lg - i) = '\0';
	  return pk_true;
	}
      i++;
    }
  return pk_false;
}

/** @return string of nb characters, initialized with the null character.
    may return null pointer
*/
extern char *
allocate_string (int nb)
{
  char *vstring;
  int i;

  vstring = (char *) malloc (nb * sizeof (char));
  if (vstring != (char *) 0)
    *vstring = '\0';
  return vstring;
}


/** Suppression du dernier caractere de la chaine ?vstring si celui-ci est ?c */
extern char *
CondKillLastChar (char *vstring, char c)
{
  size_t lg;

  lg = strlen (vstring);
  if ((lg > 0) && (vstring[lg - 1] == c))
    vstring[lg - 1] = '\0';
  return vstring;
}

/** -->chaine ?vstring de laquelle le premier caractere est enleve si celui-ci est ?c */
extern char *
CondKillFirstChar (char *vstring, char c)
{
  size_t lg;
  char *vstr;
  int i;

  if (vstring[0] != c)
    return vstring;
  lg = strlen (vstring);
  if (lg == 0) return vstring;
  i = 1;
  vstr = allocate_string (lg);
  while (vstring[i] != '\0')
    {
      vstr[i - 1] = vstring[i];
      i = i + 1;
    }
  vstr[i - 1] = '\0';
  return vstr;
}

/**
   @return the substring s[start] to s[end].
*/
extern char * extractString(const char *s, int start, int end)
{
  char *vstr = NULL;
  int i;
  
  if (s != NULL && start < end)
    {
      vstr = allocate_string(end - start + 2);
      if (vstr != NULL)
	{
	  for (i = start; i <= end; i++) vstr[i-start] = s[i];
	  vstr[i-start] = '\0';
	}
      else; // fprintf (stderr, "Memoire insuffisante\n")
    }
  return vstr;
}

/** Replace the first occurrence of c1 by c2 in vstring  */
extern void
replaceChar (char *vstring, char c1, char c2)
{
  int i;
  i = 0;
  while (vstring[i] != '\0')
    {
      if (vstring[i] == c1)
	{
	  vstring[i] = c2;
	  return;
	}
      i = i + 1;
    }
}


/** @return the string obtained by the concatenation vstring and c1  */
extern char *
appendCharEnd (char *vstring, char c1)
{
  int n;
  char *r;

  n = strlen (vstring);
  r = allocate_string (n + 2);
  strcpy (r, vstring);
  r[n] = c1;
  r[n + 1] = '\0';
  return r;
}

/** @return a copy of the vstring string.  */
extern char *
copyString (char *vstring)
{
  char *r;
  if (vstring == NULL)
    return NULL;
  r = allocate_string (strlen (vstring) + 1);
  strcpy (r, vstring);
  return r;
}

extern boolean
isLowerString (char *vstring)
{
  int i = 0;
  if (vstring == NULL)
    return pk_false;
  while (vstring[i] != '\0')
    {
      if (!islower (vstring[i]))
	return pk_false;
      i++;
    }
  return pk_true;
}

extern char *
lowerString (char *vstring)
{
  int i = 0;
  if (vstring == NULL)
    return NULL;
  while (vstring[i] != '\0')
    {
      vstring[i] = tolower (vstring[i]);
      i++;
    }
  return vstring;
}

extern char *
safeLowerString (char *vstring)
{
  char *strRes;
  int i = 0;
  if (vstring == NULL)
    return vstring;
  if (isLowerString (vstring))
    return vstring;

  strRes = allocate_string (strlen (vstring) + 1);
  while (vstring[i] != '\0')
    {
      strRes[i] = tolower (vstring[i]);
      i++;
    }
  strRes[i] = '\0';
  return strRes;
}

/**
   @return  the substring str[ifrom] to str[ifrom+size].
 */
extern char *
subString (const char *str, int ifrom, int size)
{
  char *c_name;
  int i;

  c_name = (char *) malloc (size + 1);
  for (i = 0; i < size; i++)
    c_name[i] = str[ifrom + i];
  c_name[size] = '\0';
  return c_name;
}


/**
   Given a char* string (voptions) , it constructs an "int argc (nb) , char *argv[] (vopt)" style structure.
*/
extern void
string2ArgcArgv (char *voptions, int *nb, char *vopt[])
{
  int ibegin, iend, i;

  *nb = 0;
  ibegin = 0;
  iend = 0;
  i = 0;
  while (voptions[i] != '\0')
    {
      while (voptions[i] == ' ')
	i++;
      ibegin = i;
      while ((voptions[i] != ' ') && (voptions[i] != '\0'))
	i++;
      iend = i;
      vopt[*nb] = subString (voptions, ibegin, iend - ibegin);
      (*nb)++;
      while (voptions[i] == ' ')
	i++;
    }
}
