/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
  XString register : Allocating.  
  @return a new XString register and its associated space.
*/
extern boolean
strregnew (ptreg * reg)
{
  return new_reg (reg);
}

/** XString register : Freeing. 
   It frees the XString register pointed to by reg, which must have been 
   returned by a previous call to strregnew(). 
   @return none.
   \see{strregnew}.
*/
extern void
strregfree (ptreg reg)
{
  regFree (reg);
}

/** XString register:Reseting. 
   @param reg (in) a XString register.
   The buffer assigned to reg is emptied(xxx).
*/
extern void
strregreset (ptreg reg)
{
  regClear (reg);
}

/** XString register:Setting buffer index.  
   l becomes the size of the content of the register reg.
*/
extern void
strregsetlg (ptreg reg, int l)
{
  regSet_usedSize (reg, l);
}

/** XString register: register catenating.
   @return true if the operation succeeded, false otherwise, XString Memory is full, registers are unchanged.
   The content of the register reg2 is appended to the content of the register reg1.
*/
extern boolean
strregconcreg (ptreg reg1, ptreg reg2, boolean separate)
{
  return regConc_reg (reg1, reg2, separate);
}

/** XString register:appending a char.
   @return true if the operation succeeded, false otherwise, XString Memory is full, reg is unchanged.
   The character c is catenated to the content of the register reg.
 */
extern boolean
strregconccha (ptreg reg, char c)
{
  return regConc_char (reg, c);
}

/** XString register:appending l times a char.
   @return true if the operation succeeded, false otherwise, XString Memory is full, reg is unchanged.
   The character c is appended l times to the content of the register reg.
 */
extern boolean
strregconcmcha (ptreg reg, char c, int l)
{
  return regConcNtimes_char (reg, c, l);
}


/** XString register:appending l characters of a alfa-string.
   @return true if the operation succeeded, false otherwise, XString Memory is full, reg is unchanged.
   The l first characters of str are appended to the content of the register reg.
 */
extern boolean
strregconcalfatoreg (ptreg reg, alfa str, int l, boolean separate)
{
  // printf("\nDEBUG ======= strregconcalfatoreg ======= return regConcNchar_CStrin\n"); fflush(stdout);
  return regConcNchar_CString (reg, (ptchar) str, l, separate);
}

/** XString register: register/XString catenating.
   @return true if the operation succeeded, false otherwise, XString Memory is full, register is unchanged.
   The content of the XString str is appended to the content of the register reg.
*/
extern boolean
strregconcstring (ptreg reg, char *aString, boolean separate)
{
  return regConc_CString (reg, aString, separate);
}

/** XString register: register/XString catenating.
   @return true if the operation succeeded, false otherwise, XString Memory is full, register is unchanged.
   The content of the XString str is appended to the content of the register reg.
*/
extern boolean
strregconcstr (ptreg reg, ptrepstr str, boolean separate)
{
  return regConc_repstr (reg, str, separate);
}

/** split the register content by replacing each separator occurrence by '\0'
    @return the number of resulting strings
*/
extern int
strregsplit (ptreg reg, char separator, int lg)
{
  return regSplit (reg, separator, lg);
}

/** split the register content by replacing each separator occurrence by '\0'
    @return the number of resulting strings
*/
extern ptrepstr
strreggetsubstring (ptreg reg, int index)
{
  return regGet_repstrSubstring (reg, index);
}

/** XString register: Storing in XString Memory.
    @return true if the operation succeeded, false otherwise (i.e. not enought space in XString Memory).
    The content of the XString register is stored to the XString memory (unicity is guaranted).
    @assign nilstr to *res when allocation mamory fails
*/
extern boolean
strregstore (ptreg reg, ptrepstr * res)
{
  if (regStore_repstr (reg, res))
    return pk_true;
  (*res) = nilstr;
  return pk_false;
}

/** XString register:Getting buffer index.  
   @return the size of the content of the register reg.
*/
extern int
nregGet_usedSize (ptreg reg)
{
  return regGet_usedSize (reg);
}

/** XString register:Reducing buffer size.  
   @return none
*/
extern void
strregreducesize (ptreg reg, int n)
{
  regSet_usedSize (reg, regGet_usedSize (reg) - n);
}

/** XString register: Getting a char.
   @return the i-th character of the register reg.
   assume a correct rank i
*/
extern char
nregGet_char (ptreg reg, int i)
{
  return regGet_char (reg, i);
}

/** XString register : Testing the beginning of a register to a XString value. 
   @param reg (in) a string register.
   @param vstr (in) a string.
   @return true if the text in reg string register begins with the value vstr, "false" otherwise.
*/
extern boolean
strregTestHead (ptreg reg, ptrepstr vstr)
{
  int lvstr, lreg, i;
  boolean ok;
  char c, c1;

  lreg = nregGet_usedSize (reg);
  lvstr = strsize (vstr);
  if (lreg < lvstr)
    return pk_false;
  ok = pk_true;
  i = 1;
  while (ok && i <= lreg && i <= lvstr)
    {
      c = strgetchar (vstr, i);
      c1 = nregGet_char (reg, i);
      if (c == c1)
	i++;
      else
	ok = pk_false;
    }
  return (i > lvstr);
}

/** XString register : Testing the beginning of a register to a XString value. 
   @param reg (in) a string register.
   @param vstr (in) a string.
   @return true if the text in reg string register begins with the value vstr, "false" otherwise.
*/
extern void
strregremovelastchar (ptreg reg, char c)
{
  int nb, n;

  n = nregGet_usedSize (reg);
  nb = 0;
  while ((n > 0) && (nregGet_char (reg, n) == c))
    {
      n--;
      nb++;
    }
  if (nb > 0)
    strregreducesize (reg, nb);
}
