/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** XString : Testing undefined XString. 
   @return true if str is the undefined XString.
 */
extern boolean
isNilString (ptrepstr str)
{
  return (str == nilstr);
}

/** XString : int from. 
   @return the int value represented by the XString str (assumed to be an int)
*/
extern long long int
repstr2integer (ptrepstr str)
{
  return crepstr2integer (str);
}

/** XString : Getting lowercase form of a XString. 
   @return true if the operation succeeded, false when XString Memory is full..
   @param str (in) a String.
   @param strres (out) lowercase of the string str. 
*/
extern boolean
strGet_lowerCaseStr (ptrepstr str, ptrepstr * strres)
{
  boolean Rb;
  ptreg wreg;
  int i, l;
  char c;

  l = cstrsize (str);
  if (l == 0)
    {
      *strres = str;
      return pk_true;
    }
  if (new_reg (&wreg))
    {
      for (i = 1; i <= l; i++)
	{
	  c = cstrgetchar (str, i);
	  if (isupperletter (c))
	    c = (unsigned) (c) + ((unsigned) ('a') - (unsigned) ('A'));
	  if (!strregconccha (wreg, c))
	    {
	      regFree (wreg);
	      return pk_false;
	    }
	}
      Rb = regStore_repstr (wreg, strres);
      regFree (wreg);
      return Rb;
    }
  return pk_false;
}


/** SIGNAL. 
    It returns the string which represents a double precision constant for C language, 
    or nilstr when the operation failed(XString Memory is full).
    It rewrites the SIgnal xxxDyyy constant to C xxxeyyy constant.
*/
extern ptrepstr
strdoubletoCdouble (ptrepstr str)
{
  ptreg wreg;
  ptrepstr strres;
  int i, lg;
  char c;

  if (new_reg (&wreg))
    {
      lg = cstrsize (str);
      for (i = 1; i <= lg; i++)
	{
	  c = cstrgetchar (str, i);
	  if ((c == 'D') || (c == 'd'))
	    c = 'e';
	  if (!(strregconccha (wreg, c)))
	    {
	      regFree (wreg);
	      return nilstr;
	    }
	}
      if (!(regStore_repstr (wreg, &strres)))
	strres = nilstr;
      regFree (wreg);
      return strres;
    }
  return nilstr;
}

/** XString : builting a XString from a char. 
   @return the string that represents the character xb.
 */
extern ptrepstr
char2repstr (char xb)
{
  ptrepstr vstr;
  ptreg reg;

  if (new_reg_char (&reg, xb))
    {
      if (!(regStore_repstr (reg, &vstr)))
	vstr = nilstr;
      regFree (reg);
      return vstr;
    }
  return nilstr;
}

/** XString : builting a XString from an alfa-string.
   -->Chaine representant les ?L premiers caracteres de ?STR */
extern ptrepstr
strmakewithalfa (alfa str, int l)
{
  ptrepstr vstr;
  ptreg reg;

  // printf("\nDEBUG ======= entering strmakewithalfa ======= \n"); fflush(stdout);
  if (new_regNchar_CString (&reg, (char *) str, l))
    {
      if (!(regStore_repstr (reg, &vstr)))
	vstr = nilstr;
      regFree (reg);
      return vstr;
    }
  // printf("\nDEBUG =======  exiting strmakewithalfa ======= \n"); fflush(stdout);
  return nilstr;
}

/** XString : builting a XString from an int. 
    @return the String representing the int value n (n>0).
    @return nilstr if memry allocation failed
 */
extern ptrepstr
integer2repstr (long long int n)
{
  ptrepstr vres;
  ptreg reg;

  if (new_reg (&reg))
    {
      vres = nilstr;
      if (cstrinttostr (n, reg))
	if (regStore_repstr (reg, &vres));
      strregfree (reg);
      return vres;
    }
  return nilstr;
}


/** XString : return a new memory allocated copy of str prefixed by cc
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param cc  (in) a char.
  @param str  (in) a String.
  @param vstr (out) String obtained by the catenating of the character cc and the string str.
*/
extern boolean
strconcchastr (char cc, ptrepstr str, ptrepstr * vstr)
{
  boolean Rb;
  ptreg wreg;

  if (new_reg_char (&wreg, cc))
    {
      if (regConc_repstr (wreg, str, pk_false))
	Rb = regStore_repstr (wreg, vstr);
      else
	Rb = pk_false;
      regFree (wreg);
      return Rb;
    }
  return pk_false;
}

/** XString : return a new memory allocated copy of str prefixed by cc
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param cc  (in) a char.
  @param str  (in) a String.
  @param vstr (out) String obtained by the catenating of the character cc and the string str.
*/
extern boolean
strconcchacha (char cc1, char cc2, ptrepstr * vstr)
{
  boolean Rb;
  ptreg wreg;

  if (new_reg_char (&wreg, cc1))
    {
      if (regConc_char (wreg, cc2))
	Rb = regStore_repstr (wreg, vstr);
      else
	Rb = pk_false;
      regFree (wreg);
      return Rb;
    }
  return pk_false;
}

/** XString : return a new memory allocated copy of str followed by cc
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param cc  (in) a char.
  @param str  (in) a String.
  @param vstr (out) String obtained by the catenating of the character cc and the string str.
*/
extern boolean
strconcstrcha (ptrepstr str, char cc, ptrepstr * vstr)
{
  boolean Rb;
  ptreg wreg;

  if (new_reg_repstr (&wreg, str))
    {
      if (regConc_char (wreg, cc))
	Rb = regStore_repstr (wreg, vstr);
      else
	Rb = pk_false;
      regFree (wreg);
      return Rb;
    }
  return pk_false;
}

/** XString : Catenating two XString. No side effect
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param str1, str2 (in) two string.
  @param vstr (out) result of the catenating operation of str2 to str1.
*/
extern boolean
strconc2str (ptrepstr str1, ptrepstr str2, boolean separate, ptrepstr * vstr)
{
  boolean Rb;
  ptreg reg;

  if (new_reg_repstr (&reg, str1))
    {
      if (regConc_repstr (reg, str2, separate))
	Rb = regStore_repstr (reg, vstr);
      else
	Rb = pk_false;
      regFree (reg);
      return Rb;
    }
  return pk_false;
}


/** XString : return a new memory allocated copy of str1 followed by aString
  @return true if operation succeed, false when there is not free register, or XString Memory is full.
  @param str1, str2 (in) two string.
  @param vstr (out) result of the catenating operation of str2 to str1.
*/
extern boolean
strconcstrString (ptrepstr str1, char *aString, boolean separate, ptrepstr * vstr)
{
  return strconc2str (str1, string2repstr (aString), separate, vstr);
}



/** XString : Catenating XString, a char and l chars of an alfa-string. 
   @return true if operation succeed, false when there is not free register, or XString Memory is full.
   @param vstr (out) XString built by catenating of str (in) with c (in) and l first characters of cha (in).
*/
extern boolean
strconcstrcharalfa (ptrepstr str, char c, alfa cha, int l, ptrepstr * vstr)
{
  boolean Rb;
  ptreg wreg;

  // printf("\nDEBUG ======= entering strconcstrcharalfa  ======= \n"); fflush(stdout);
  if (new_reg_repstr (&wreg, str))
    {
      Rb = pk_false;
      if (regConc_char (wreg, c))
	if (regConcNchar_CString (wreg, (ptchar) cha, l, pk_false))
	  Rb = regStore_repstr (wreg, vstr);
      regFree (wreg);
      // printf("\nDEBUG =======  exiting strconcstrcharalfa  ======= \n"); fflush(stdout);
      return Rb;
    }
  return pk_false;
}



/** XString : size of a string. 
   @return the size of str.
*/
extern int
strsize (ptrepstr str)
{
  return cstrsize (str);
}



/** XString : Hash coding for a string. 
   @return the Tree Machine hash coding value of the string operand str
*/
extern short
strhash (ptrepstr str)
{
  return ptrepstr_get_short_hashValue (str);
}



/** XString : equality. 
   @return true if str1 is equal to str2. 
*/
extern boolean
streq (ptrepstr str1, ptrepstr str2)
{
  return cstreq (str1, str2);
}



/** XString : Comparing. 
   @param str1,str2 (in) two string.
   @return -1 if str1 < str2, 0 if str1 = str2, 1 if str1 > str2.
*/
extern int
strcompar (ptrepstr str1, ptrepstr str2)
{
  return cstrcompar (str1, str2);
}

/** XString : Comparing. 
   @param str1,str2 (in) two decimal strings that equal to 0 or with no heading 0 character (not checked)
   @return false if str1 > str2.
   if str1 == str2 assign false to strict and return pk_true
   if str1 < str2 assign pk_true to strict and return pk_true
*/
extern boolean
repstrInteger_leq (ptrepstr str1, ptrepstr str2, boolean * strict)
{
  int strSize1, strSize2, resCompar;
  if (str1 == str2)
    {
      (*strict) = pk_false;
      return pk_true;
    }
  strSize1 = strsize (str1);
  strSize2 = strsize (str2);
  if (strSize1 < strSize2)
    {
      (*strict) = pk_true;
      return pk_true;
    }
  if (strSize1 > strSize2)
    return pk_false;
  resCompar = cstrcompar (str1, str2);
  if (resCompar < 0)
    {
      (*strict) = pk_true;
      return pk_true;
    }
  if (resCompar > 0)
    return pk_false;
  /* the following is useless if unicity is correctly handled */
  (*strict) = pk_false;
  return pk_true;
}



/** XString : Getting a character.
   @return the i-th character of str.
*/
extern char
strgetchar (ptrepstr str, int i)
{
  return cstrgetchar (str, i);
}

/** XString :Getting the rank of a char in a sub-XString.
   @return the index of the first sep char from str[i], -1 when not found.
*/
extern int
strGetIndex (ptrepstr str, int i, char sep)
{
  int lg, k;
  char c;

  k = i;
  lg = strsize (str);
  while (k <= lg)
    {
      c = strgetchar (str, k);
      if (c == sep)
	return k;
      k++;
    }
  return -1;
}

/** XString : Extracting an int value from a string. 
   It extracts an int value from [str[i], str[k]] ( str[k+1] is not a Decimal digit).
   @param str (in) a String.
   @param i (in/out) the first index for the extraction  (in), the last index used for the extraction  (out).
   @param val (out) the extracted value.
   @return true if the operation succeed, false otherwise( XString Memory is full).
*/

/** XString : XString to C-String(conversion).
   @return C-string representation of a XString representation.
 */
extern char *
repstr2string (ptrepstr rep)
{
  int sizeofrep, i;
  char *strofrep;

  sizeofrep = strsize (rep);
  strofrep = (char *) malloc (sizeofrep * (sizeof (char)) + 1);
  for (i = 1; i <= sizeofrep; i++)
    strofrep[i - 1] = strgetchar (rep, i);
  strofrep[sizeofrep] = '\0';
  return strofrep;
}


/** XString : making a XString from a C-string (conversion).
   @return string of the C-string ch. It is truncated on an out of bounds exception.
   @return nilstr when the operation fails
*/
extern ptrepstr
string2repstr (const char *ch)
{
  ptreg wreg;
  ptrepstr no_ch;
  // printf("DEBUG entering stringmemStr_g::string2repstr ch is <<%s>>\n", ch); fflush(stdout);  
  no_ch = nilstr;
  if (new_reg_CString (&wreg, ch))
    {
      if (strregstore (wreg, &no_ch));
      strregfree (wreg);
    }
  // printf("DEBUG  exiting stringmemStr_g::string2repstr ch is <<%s>>\n", ch); fflush(stdout);  
  return (no_ch);
}

/** XString : making a XString from a C-string (conversion).
   @return string of the C-string ch. It is truncated on an out of bounds exception.
   @return nilstr when the operation fails
*/
extern ptrepstr
substring2repstr (const char *ch, int length)
{
  ptreg reg;
  ptrepstr no_ch;
  int i;

  no_ch = nilstr;
  i = 0;
  // printf("\nDEBUG< ======= substring2repstr ======= return regConcNchar_CString\n"); fflush(stdout);
  if (new_regNchar_CString (&reg, ch, length))
    {
      if (strregstore (reg, &no_ch));
      strregfree (reg);
    }
  return (no_ch);
}

/** XString : making a XString from a C-string (conversion).
    vstr string representation of the C-string ch.
    @return true if the operation succeeds, false otherwise (out of bounds exception).
*/
extern boolean
string2Srepstr (const char *ch, ptrepstr * vstr)
{
  ptreg wreg;
  int i;

  (*vstr) = nilstr;
  i = 0;
  if (new_reg_CString (&wreg, ch))
    {
      if (strregstore (wreg, vstr));
      strregfree (wreg);
      return pk_true;
    }
  return pk_false;
}

/** SIGNAL/DC+ Syntax. 
  Signal2DC+ conversion for a float/double constant (rep).
  \<int\> "." [\<int\>] [exponent]
*/
extern char *
repstrConstFloatDouble2DCplusFloatDouble (ptrepstr rep)
{
  int sizeofrep, i;
  char *strofrep;
  boolean ptfound, b, app;
  char c;

  sizeofrep = strsize (rep);
  strofrep = (char *) malloc (sizeofrep * (sizeof (char)) + 2);
  ptfound = pk_false;
  app = pk_false;
  for (i = 1; i <= sizeofrep; i++)
    {
      c = strgetchar (rep, i);
      if (!ptfound)
	{
	  b = !isDecimalDigit (c);
	  if (b)
	    {
	      if (c != '.')
		{
		  strofrep[i - 1] = '.';
		  app = pk_true;
		}
	      ptfound = pk_true;
	    }
	}
      if (c == 'e')
	c = 'E';
      else if (c == 'd')
	c = 'D';
      if (app)
	strofrep[i] = c;
      else
	strofrep[i - 1] = c;
    }
  if (app)
    sizeofrep = sizeofrep + 1;
  strofrep[sizeofrep] = 0;
  return strofrep;
}

/** PLG check: return false for path starting with sep or path empty
*/
extern boolean
extractStrDirname (const char *path, char sep, char lastCar, ptrepstr * dirName, int *pos)
{
  ptreg wreg;

  if (new_reg (&wreg))
    {
      boolean Rb;
      int i, lg;
      boolean stop, added;
      char c;
      i = *pos;
      stop = pk_false;
      added = pk_false;
      while (!stop)
	{
	  c = path[i];
	  if ((c == nullChar) || (c == sep))
	    stop = pk_true;
	  else
	    // Skiping space at the beginning of the string.
	  if ((c != spaceChar) || added)
	    {
	      if (!strregconccha (wreg, c))
		{
		  strregfree (wreg);
		  return pk_false;
		}
	      added = pk_true;
	    }
	  i++;
	}
      *pos = i;
      lg = nregGet_usedSize (wreg);
      Rb = pk_false;
      if (lg > 0)
	{
	  // removing space characters at the end of the string.
	  while (nregGet_char (wreg, lg) == spaceChar)
	    lg--;
	  strregsetlg (wreg, lg);
	  // adding a separator if needed
	  if (nregGet_char (wreg, lg) != lastCar)
	    if (!strregconccha (wreg, lastCar))
	      {
		strregfree (wreg);
		return pk_false;
	      }
	  Rb = strregstore (wreg, dirName);
	}
      strregfree (wreg);
      return Rb;
    }
  return pk_false;
}


/** XString : Adding double quote, at the begining and at the end of a XString.
   @param repstrEnonce (in) a XString.
   @param vnewrepstrEnonce (out) repstrEnonce with double quote at begining and end of the string.
   @return true if first and last character of repstrEnonce are double quote.
*/
extern boolean
getDoubleQuotedString (ptrepstr repstrEnonce, ptrepstr * vnewrepstrEnonce)
{
  boolean modif;
  ptrepstr vstr;

  modif = pk_false;
  vstr = char2repstr ('"');
  if (strgetchar (repstrEnonce, 1) != '"')
    {
      strconc2str (vstr, repstrEnonce, pk_false, vnewrepstrEnonce);
      modif = pk_true;
    }
  else
    *vnewrepstrEnonce = repstrEnonce;

  if (strgetchar (*vnewrepstrEnonce, strsize (*vnewrepstrEnonce)) != '"')
    {
      strconc2str (*vnewrepstrEnonce, vstr, pk_false, vnewrepstrEnonce);
      modif = pk_true;
    }
  return !modif;
}


/** XString : extracting XString.
   @return XString of EnonceValue without double quote at the begining and at the end.
*/
extern ptrepstr
ExtractString (ptrepstr EnonceValue)
{
  char *vstring;
  vstring = repstr2string (EnonceValue);
  vstring = CondKillFirstChar (vstring, '"');
  vstring = CondKillLastChar (vstring, '"');
  return string2repstr (vstring);
}


/** XString : extracting an int from a double quoted XString.
   @return the int value contained in EnonceValue that can be double quote string.
 */
extern long long int
ExtractInteger (ptrepstr EnonceValue)
{
  return repstr2integer (ExtractString (EnonceValue));
}
