/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** XString  Table managing : loadding a table from a file.
   @param filename (in) name of the file to load.
   @param res (out) entry point (base) of the loaded table.
   @return true if the operation succeeded, false otherwise (XString Memory is full).
*/
extern boolean
strTableLoad (ntrees * res, char *filename)
{
  // printf("DEBUG entering strTableLoad\n"); fflush(stdout);
  (*res) = (ntrees) loadTM_tree (filename);
  // printf("DEBUG  exiting strTableLoad\n"); fflush(stdout);
  return !(IS_NILTREE (*res));
}

/** XString  Table managing : getting sub-table.
   @return the n-th sub table of the table res.
   This function is used in Signal compiler for message tables (messignal.*.txt files in SIGNAL bin directory).
   These textual files are separated in sub-sets:internal error, fatal error, info, error, warning, comment. 
   The access to a specific message is done using this function.
*/
extern ptreptab
strDiaTableLoad (ptreptab aSubtree, int n)
{
  // printf("DEBUG entering strDiaTableLoad\n"); fflush(stdout);
  return (ptreptab) getTM_nChild (n + 1, (trees) aSubtree);
}

/** XString Table managing : Getting an element of a table.
   @return the n-th XString of the table Table. 
*/
extern ptrepstr
strTableGet (ptreptab Table, int n)
{
  // printf("DEBUG entering strTableGet\n"); fflush(stdout);
  return (ptrepstr) getTM_TMStringValueInNodeSon ((trees) Table, n);
}
