; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Emacs environment used by Signal to build C,C++ module from C,C++ elementary constituants.
; A C/C++ module is defined as a set of (include) constituant (see *.mkf file in Signal directory)
;             + types importation
;             + function/class importation.
;
; Header files (*.h) are automatically generated using "cextract" command.

; WARNING : the function :init_env (parameter) must be called first before using other functions.
; +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

(defun :replaceString(str1 str2)
;--------------------------------------------------------------------------------
; Replacing str1 string by str2 string
;--------------------------------------------------------------------------------
  (goto-char (point-min))
  (replace-string str1 str2)
  )

(defun :mkC_Comment()
;--------------------------------------------------------------------------------
; deleting c++ comments (// <COMMENT>)
;--------------------------------------------------------------------------------
  (goto-char (point-min))
  (while (re-search-forward "^//" nil t)
;    (replace-match C-Comment_begin nil nil)    (end-of-line)    (insert-string C-Comment_end)
    (beginning-of-line)
    (kill-line 1)
    )
)

(defun :mkHeaderLoc (fileOut fileIn vcommand)
;--------------------------------------------------------------------------------
; Making  C, C++ header file for fileIn code file
; specific internal symbols are substituted before applying "cextract" command.
;--------------------------------------------------------------------------------
  ( let ( (tmpfile (concat fileOut "_tmp")) )
    (find-file  fileIn)
    (:mkC_Comment ) ;; useful ???
    (write-file tmpfile)
    (kill-buffer (current-buffer))
    (shell-command (concat vcommand (concat fileOut (concat " " tmpfile))))
    (delete-file tmpfile)
    )
  )

(defun :mkHeader (fileOut fileIn)
;--------------------------------------------------------------------------------
; Making  C, C++ header file for fileIn code file
; specific internal symbols are substituted before applying "cextract" command.
;--------------------------------------------------------------------------------
  (:mkHeaderLoc  fileOut fileIn  "$pKPrototypeExtract -C\"*\" -o ") 
  )


(defun :recupbuffer (varbuf)
;--------------------------------------------------------------------------------
  (switch-to-buffer varbuf)
  (goto-char (point-min))
  (copy-region-as-kill (point-min) (point-max))
  (switch-to-buffer BUFFER)
)

(defun :insererincl (chaine)
;--------------------------------------------------------------------------------
 (insert-string (concat (concat "#include " "\"") (concat chaine "\"\n") ))
)

(defun :insert-incl-or-file (filename path)
;--------------------------------------------------------------------------------
; Adding into the global buffer BUFFER 
;          #include file/path.c when InitParamInsert (global parameter) is nil
;          the content of the file "file/path.c" otherwise
;--------------------------------------------------------------------------------

  (if (not InitParamInsert)
      (:insererincl filename)
    ( let ( ( srcfile (concat (concat path  file-name-separator) filename)))
      (if (file-exists-p srcfile)
	  (insert-file-contents srcfile nil nil nil nil)
	(:insererincl filename)
	)
      )
    )
  )

(defun :consext (fileg filee path pathi)
;--------------------------------------------------------------------------------------
; Building C header file in <pathi>/filee.h from path/fileg file
;     fileg constains a set of external functions with definitions (body)
;     filee constains a set of external interfaces (headers) of the fileg declarations.
;--------------------------------------------------------------------------------------
(let ( (filename_g (concat  path (concat  file-name-separator (concat fileg ".c") )))
       (filename_e (concat  pathi (concat  file-name-separator (concat filee ".h") ))) 
       )
  (:mkHeader filename_e filename_g)
))

(defun :mkmodul (file  path)
;--------------------------------------------------------------------------------
; Defining the global buffer BUFFER to build a module initialized by path/file ; 
;--------------------------------------------------------------------------------
(let ( (filename (concat  file ".c") ) ) 
         (setq BUFFER filename)
         (switch-to-buffer BUFFER)
         (:insert-incl-or-file filename path )
;	 (print file)
	 (goto-char (point-min))
	 (insert-string "/* \n")
	 (insert-string "   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.\n")
	 (insert-string "   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.\n")
	 (insert-string "\n")
	 (insert-string "   Contact: polychronycontact@listes.irisa.fr \n")
	 (insert-string "\n")
	 (insert-string "   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).\n")
	 (insert-string "\n")
	 (insert-string "   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under\n")
	 (insert-string "   the terms of the GNU General Public License Version 2 as published by the Free\n")
	 (insert-string "   Software Foundation.\n")
	 (insert-string "\n")
	 (insert-string "   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY\n")
	 (insert-string "   WARRANTY; without even the implied warranty of MERCHANTABILITY or\n")
	 (insert-string "   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2\n")
	 (insert-string "   for more details.\n")
	 (insert-string "\n")
	 (insert-string "   You should have received a copy of the GNU General Public License Version 2\n")
	 (insert-string "   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url \n")
	 (insert-string "    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free\n")
	 (insert-string "   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA\n")
	 (insert-string "   02110-1301, USA. \n")
	 (insert-string "*/\n")
	 (insert-string (concat @commentimporttype@ "\n") )
	 (insert-string (concat @commentbeginprog@ "\n") )
))

(defun :mkEmptyModule (filename)
;--------------------------------------------------------------------------------
;--------------------------------------------------------------------------------
  (setq BUFFER filename)
  (switch-to-buffer BUFFER)
  (goto-char (point-min))
  (copy-region-as-kill (point-min) (point-max))
  (insert-string (concat @commentimporttype@ "\n") )
  (insert-string (concat @commentbeginprog@ "\n") )
)

(defun :incvar (varbuf var)
;--------------------------------------------------------------------------------
  (switch-to-buffer varbuf)
  (goto-char (point-max))
  (:insert-incl-or-file (concat var ".h") nil )
  (switch-to-buffer BUFFER)
)

(defun :incvar_NEW (varbuf var)
;--------------------------------------------------------------------------------
  (switch-to-buffer var)
  ( let ( (x (buffer-size) ) )
    (switch-to-buffer varbuf)
    (goto-char (point-max))
    (if (= x 0)
	(:insert-incl-or-file (concat var ".h") nil )
      (insert-buffer var)
      )
    (switch-to-buffer BUFFER)
    )
  )

(defun :incvartype (varbuf var)
;--------------------------------------------------------------------------------
  (:incvar varbuf var)
)

(defun :incfile (varbuf filename)
;--------------------------------------------------------------------------------
  (switch-to-buffer varbuf)
  (goto-char (point-max))
  (:insert-incl-or-file filename nil)
  (switch-to-buffer BUFFER)
)

(defun :incloc (filename)
;--------------------------------------------------------------------------------
  (switch-to-buffer BUFFER)
  (:insert-incl-or-file filename nil )
)

(defun :inclocend (filename)
;--------------------------------------------------------------------------------
  (switch-to-buffer BUFFER)
  (goto-char (point-max)) 
  (:insert-incl-or-file filename nil )
)

(defun :inctype (varbuf)
;--------------------------------------------------------------------------------------
; Adding the buffer varbuf at the head of the gloabl buffer BUFFER  (type declarations)
;--------------------------------------------------------------------------------------
  (:recupbuffer varbuf)
  (goto-char (point-min))
  (if (not (search-forward @commentimporttype@ (point-max) t))
    ()
  (beginning-of-line)
  (yank)
  )
)

(defun :incext (varbuf)
;----------------------------------------------------------------------------------------
; Adding the buffer varbuf into the gloabl buffer BUFFER after the type declaration area
;----------------------------------------------------------------------------------------
(:recupbuffer varbuf)
(goto-char (point-min))
(if (not (search-forward @commentimporttype@ (point-max) t))
    (goto-char (point-min))
;;  (print (point))
  (end-of-line)
  (insert-string "\n")
  (end-of-line)
  )
(yank)
)

(defun :add (xBuffer varbuf)
;----------------------------------------------------------------------------------------
; Adding the buffer varbuf into the buffer xBuffer 
;----------------------------------------------------------------------------------------
(let (( xx)) 
  (setq xx BUFFER)
;  (:recupbuffer varbuf)
  (switch-to-buffer xBuffer)
;  (goto-char (point-max))
;  (yank)
  (insert-buffer varbuf)
  (setq BUFFER xx)
  )
)

(defun :incglo  (vide)
;--------------------------------------------------------------------------------
; Without effect 
)

(defun :mkglo (file  path)
;--------------------------------------------------------------------------------
; Adding into the global buffer BUFFER 
;          the content of the file "file/path.c" or
;          #include file/path.c
; according to initializing parameter
;--------------------------------------------------------------------------------
(let ( (filename (concat  file ".c")  ))
     (switch-to-buffer BUFFER)
     (goto-char (point-max))
     (:insert-incl-or-file filename path)
   )
)

(defun :lstocker (file  path lang)
;--------------------------------------------------------------------------------
; Storing the global buffer BUFFER in the file file/path.lang
;--------------------------------------------------------------------------------
(let ( (infile (concat path (concat  file-name-separator (concat file lang )))))
  (switch-to-buffer BUFFER)
  (write-file infile)
  (kill-buffer (current-buffer) ) 
  )
)

(defun :stocker_cpp (file  path)
;--------------------------------------------------------------------------------
; Storing the global buffer BUFFER in the file file/path.cpp
;--------------------------------------------------------------------------------
  (:lstocker file path ".cpp") 
)

(defun :stocker (file  path)
;--------------------------------------------------------------------------------
; Storing the global buffer BUFFER in the file file/path.c
;--------------------------------------------------------------------------------
  (:lstocker file path ".c") 
)

(defun :stockerbison(file  path)
;--------------------------------------------------------------------------------
; used for the parser builder
;--------------------------------------------------------------------------------
(let ( (infile (concat path (concat  file-name-separator file ))))
  (switch-to-buffer BUFFER)
  (goto-char (point-min))
  (insert-string  "%{\n")
  (goto-char (point-max))
  (insert-string  "%}\n")
  (write-file infile)
  (kill-buffer (current-buffer) ) 
))

(defun :init_env (p)
;---------------------------------------------------------------------------------
; Initializing Emacs environment : global variables.
; parameter p :
;    nil to have #include <filename> line for the definition of a c/C++ module.
;    otherwise the content of the constituant file are included in the code of the module.
;---------------------------------------------------------------------------------
  (setq BUFFER "xxxxx")
  (setq InitParamInsert p)
  (setq @commentimporttype@ "/* @End_import_type@ */")
  (setq @commentbeginprog@ "/* @Begin_prog@ */")
  (setq file-name-separator "/")
  (setq C-Comment_begin "/*")
  (setq C-Comment_end   "*/")
  )

