
_PUBLIC_
/* ==========  ==========  ==========  ========== TITLE: these attributes
 */
  _PUBLIC_ _VIRTUAL_ boolean AncestorFunctionElement::isAncestorFunctionElement ()
{
  return pk_true;
}

/* ==========  ==========  ==========  ========== TITLE: PRINTING
 */

/* prints object standard header: "{ <NAME>" or "{ <TYPE> <NAME>"
 */
_VIRTUAL_ void
AncestorFunctionElement::print_type_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_ancestorType_sp (outBuffer);
}

_VIRTUAL_ void
AncestorFunctionElement::print_declBody (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  // fprintf(stderr, "DEBUG AncestorFunctionElement::print_declBody\n");
  if (these_genericAncestors != NIL_GenericAncestorList)
    {
      outBuffer->line ();
      these_genericAncestors->print_formula (printingContext, outBuffer);
    }
}

/* ==========  ==========  ==========  ========== TITLE: DEBUG
 */
/* detailss in <outBuffer> file this context and local declarations
 * <outBuffer> is not NIL_IntePrinter
 */
_VIRTUAL_ void
AncestorFunctionElement::logDebug_declBody (int level)
{
  if (these_genericAncestors == NIL_GenericAncestorList)
    log_NILComponent_ln (level, "meta ancestor list");
  else
    these_genericAncestors->logDebug_subFormula_ln (level, "meta ancestor list");
}

/* ==========  ==========  ==========  ========== TITLE: READING
 */
/* 
 * reads body of a function declaration
 * success: cursor starts . [_ANCESTOR+] [eo] # '}:'
 * current char is the first not space following a node opening mark
 * success: terminates on not space started 
 * failure: let the buffer unchanged or on Eof()
 *          sets this node errored
 *          error message is printed
 */
_VIRTUAL_ boolean
AncestorFunctionElement::read_declBody_fSpStar_nSp ()
{
  // fprintf(stderr, "DEBUG AncestorFunctionElement::read_declBody_fSpStar_nSp\n");
  return this_innerContext->readOpt_genericAncestorList_eol_ (&these_genericAncestors);
}

/* ==========  ==========  ==========  ========== TITLE: GENERATOR
 */
/* expands actual parameter values, assign them to formal parameters and call this function
 * <callingContext> is not NIL (not checked)
 * these_genericAncestors is assumed to be correctly defined and not empty ( != NIL_GenericAncestorList )
 * success:
 *        <ancestorList_first> may be empty
 * failure: arguments do not fit or ancestor generation failed 
 *               error message is printed
 */
boolean
AncestorFunctionElement::single_call (ref2Context callingContext, ref2ArgumentList parameterValues, ref2AncestorList * ancestorList_first)
{
  // logDEBUG_in("single_call to element function", this_name);
  /* DEBUG parameterValues->log_formula_(pk_false); DEBUG */
  if ((parameterValues == NIL_ArgumentList) || parameterListLink_expand_ (parameterValues))
    return these_genericAncestors->build_ancestorList (callingContext, ancestorList_first);
  return pk_false;
}

/* appends a new ancestor list (NPL) to last
 * assumes: last is not NIL PLG TODO check
 *          iteratorList has a single identifier, this identifier does not have an empty list of values
 * returns NPL last element in last
 * iterator is a singleton associated with a string list type Context variable or parameter declaration D
 * if actual string list ISL associated with D is empty returns NIL_AncestorList
 * if string list ISL holds n(>0) actual values ISL[1]...ISL[n]
 *     ancestor list NPL has n elements 
 *     ancestor list element NPL[i] is the ancestor resulting of this function call
 *         when parameterValues elements are expanded when declaration D holds ISL[i]
 * if variable has no occurrence in parameterValues, the same ancestor is generated n times
 * success:
 *     iteratorList value is not empty
 *       <last> may remains unchanged
 * failure:
 *     an error occurred when calling the function
 *     <last> is set to NIL_ArgumentList
 *     returns pk_false;
 */
_PRIVATE_
  boolean AncestorFunctionElement::tail_singleCall (ref2Context callingContext,
						    ref2ArgumentList parameterValues, ref2IdentifierList iteratorList, ref2AncestorList * last)
{
  // logDEBUG_in("tail_singleCall to element function", this_name);
  ref2AncestorList
    current;
  while (single_call (callingContext, parameterValues, &current))
    {				/* correct linking of ancestors is provided by append, <current> may be empty OK */
      if (current != NIL_AncestorList)
	{
	  (*last)->append (current);
	  (*last) = current;
	}
      if (!(iteratorList->stepThisCurrentList ()))
	{
	  return pk_true;
	}
    }
  return pk_false;
}

/* generates a new ancestor list (NPL) in first
 * returns NPL last element in last
 * iterator is a singleton associated with a string list type Context variable or parameter declaration D
 * if actual string list ISL associated with D is empty returns NIL_AncestorList
 * if string list ISL holds n(>0) actual values ISL[1]...ISL[n]
 *     ancestor list NPL has n elements 
 *     ancestor list element NPL[i] is the ancestor resulting of this function call
 *         when parameterValues elements are expanded when declaration D holds ISL[i]
 * if variable has no occurrence in parameterValues, the same ancestor is generated n times
 * success:
 *        <last>, <first> may be empty
 * failure:
 *     some iterator has an empty list of values
 *         <correct> remains unchanged
 *         <last>, <first> remains unchanged
 *     parameterValues is NIL_ArgumentList or some of its elements does not have string type
 *         <correct> is set to false
 */
_PRIVATE_
  boolean AncestorFunctionElement::iterate_singleCall /*B*/ (ref2Context callingContext,
							     ref2ArgumentList parameterValues, ref2IdentifierList iteratorList,
							     boolean * correct, ref2AncestorList * first, ref2AncestorList * last)
{
  // logDEBUG_in("iterate_singleCall 2 to element function", this_name);
  if (iteratorList->downThisCurrentList (correct))
    {
      // logDEBUG_in("iterate_singleCall 2 to element function, downThisCurrentList OK", this_name);
      ref2AncestorList
	first_ancestor;
      while (single_call (callingContext, parameterValues, &first_ancestor))
	{			/* correct linking of ancestors is provided by append, <first_ancestor> may be empty */
	  if (first_ancestor != NIL_AncestorList)
	    {
	      (*first) = first_ancestor;
	      (*last) = first_ancestor;
	      if (iteratorList->stepThisCurrentList ())
		{
		  if (tail_singleCall (callingContext, parameterValues, iteratorList, last))
		    return pk_true;
		  (*correct) = pk_false;
		  return pk_false;
		}
	      return pk_true;
	    }
	  if (!(iteratorList->stepThisCurrentList ()))
	    return pk_true;
	}
      (*correct) = pk_false;
    }
  // logDEBUG_in("iterate_singleCall 2 to element function FAILED", this_name);
  return pk_false;
}

/* appends a new ancestor list (NPL) to last
 * assumes: last is not NIL PLG TODO check
 *          iteratorList has a single identifier, this identifier does not have an empty list of values
 * returns NPL last element in last
 * iterator is a singleton associated with a string list type Context variable or parameter declaration D
 * if actual string list ISL associated with D is empty returns NIL_AncestorList
 * if string list ISL holds n(>0) actual values ISL[1]...ISL[n]
 *     ancestor list NPL has n elements 
 *     ancestor list element NPL[i] is the ancestor resulting of this function call
 *         when parameterValues elements are expanded when declaration D holds ISL[i]
 * if variable has no occurrence in parameterValues, the same ancestor is generated n times
 * success:
 *     iteratorList value is not empty
 *       <last> may remains unchanged
 * failure:
 *     an error occurred when calling the function
 *     <last> is set to NIL_ArgumentList
 *     returns pk_false;
 */
_PRIVATE_
  boolean AncestorFunctionElement::coTail_singleCall (ref2Context callingContext,
						      ref2ArgumentList parameterValues, ref2IdentifierList iteratorList, ref2AncestorList * last)
{
  // logDEBUG_in("tail_singleCall to element function", this_name);
  ref2AncestorList
    current;
  while (single_call (callingContext, parameterValues, &current))
    {				/* correct linking of ancestors is provided by append, <current> may be empty OK */
      if (current != NIL_AncestorList)
	{
	  (*last)->append (current);
	  (*last) = current;
	}
      if (!(iteratorList->coStepThisCurrentList ()))
	{
	  return pk_true;
	}
    }
  return pk_false;
}

/* generates a new ancestor list (NPL) in first
 * returns NPL last element in last
 * iterator is a singleton associated with a string list type Context variable or parameter declaration D
 * if actual string list ISL associated with D is empty returns NIL_AncestorList
 * if string list ISL holds n(>0) actual values ISL[1]...ISL[n]
 *     ancestor list NPL has n elements 
 *     ancestor list element NPL[i] is the ancestor resulting of this function call
 *         when parameterValues elements are expanded when declaration D holds ISL[i]
 * if variable has no occurrence in parameterValues, the same ancestor is generated n times
 * success:
 *        <last>, <first> may be empty
 * failure:
 *     some iterator has an empty list of values
 *         <correct> remains unchanged
 *         <last>, <first> remains unchanged
 *     parameterValues is NIL_ArgumentList or some of its elements does not have string type
 *         <correct> is set to false
 */
_PRIVATE_
  boolean AncestorFunctionElement::coIterate_singleCall /*B*/ (ref2Context callingContext,
							       ref2ArgumentList parameterValues, ref2IdentifierList iteratorList,
							       boolean * correct, ref2AncestorList * first, ref2AncestorList * last)
{
  // logDEBUG_in("iterate_singleCall 2 to element function", this_name);
  if (iteratorList->coDownThisCurrentList (correct))
    {
      // logDEBUG_in("iterate_singleCall 2 to element function, downThisCurrentList OK", this_name);
      ref2AncestorList
	first_ancestor;
      while (single_call (callingContext, parameterValues, &first_ancestor))
	{			/* correct linking of ancestors is provided by append, <first_ancestor> may be empty */
	  if (first_ancestor != NIL_AncestorList)
	    {
	      (*first) = first_ancestor;
	      (*last) = first_ancestor;
	      if (iteratorList->coStepThisCurrentList ())
		{
		  if (coTail_singleCall (callingContext, parameterValues, iteratorList, last))
		    return pk_true;
		  (*correct) = pk_false;
		  return pk_false;
		}
	      return pk_true;
	    }
	  if (!(iteratorList->coStepThisCurrentList ()))
	    return pk_true;
	}
      (*correct) = pk_false;
    }
  // logDEBUG_in("iterate_singleCall 2 to element function FAILED", this_name);
  return pk_false;
}

/* expands actual parameter values and assign them to formal parameters
 * these_genericAncestors is assumed to be correctly defined and not empty( != NIL_GenericAncestorList )
 * returns in first the generated ancestor list
 * returns in last its last element
 * success:
 *     <last> may be empty 
 * failure:
 *     some iterator has an empty list of values
 *         <correct> remains unchanged
 *         <last>, is set to NIL_ArgumentList
 *     iterator list is empty or parameterValues is NIL_ArgumentList or an error occurred
 *         <correct> is set to false
 *     returns pk_false;
 */
_PUBLIC_
  boolean AncestorFunctionElement::call_iteration (ref2Context callingContext, ref2IteratorContext iteratorContext,
						   ref2ArgumentList parameterValues, boolean * correct, ref2AncestorList * first)
{
  // logDEBUG_in("call_iteration 3 to element function", this_name);
  /* DEBUG parameterValues->log_formula_(pk_false); DEBUG */
  ref2AncestorList
    last;
  ref2IdentifierList
    theseIterators = ((ref2DynamicContext) iteratorContext)->get_identifierList ();
  if (theseIterators != NIL_IdentifierList)
    {
      if ((theseIterators->next ()) == NIL_IdentifierList)
	return iterate_singleCall /*B*/ (callingContext, parameterValues, theseIterators, correct, first, &last);	/* <last> <first> may be empty */
      return coIterate_singleCall /*B*/ (callingContext, parameterValues, theseIterators, correct, first, &last);
    }				/* <last> <first> may be empty */
  (*correct) = pk_false;
  return pk_false;
}

/* ==========  ==========  ==========  ========== TITLE: BUILDER
 */
/* 2005-02-01 syntax
 * creates and returns a new ancestor function descriptor named <name>
 * <embeddingContext> is not NIL
 * <embeddingContext> is the context in which this declaration occurs
 * <comments> are comments associated with this declaration
 * <innerContext> is the new internal context
 * success: cursor starts .[_ANCESTOR+] #'}:' 
 *          [_ANCESTOR+] is read in <innerContext>
 *          <correct> remains unchanged
 *          cursor is moved to final '}:'
 * failure: an error occurs
 *          error message is printed
 *          <correct> is set to false
 */
AncestorFunctionElement::AncestorFunctionElement (ref2Context embeddingContext, char *name, ref2DynamicContext innerContext, ref2ArgumentList comments):FunctionElement (embeddingContext, name, innerContext,
		 comments)
{
  these_genericAncestors = NIL_GenericAncestorList;
}
