#define SET_GLOBAL_DECLARATION() ( this_declarationScope = 0 )
#define SET_FILE_DECLARATION() ( this_declarationScope = 1 )

#define IS_GLOBAL_DECLARATION() ( this_declarationScope == 0 )


_PUBLIC_ _VIRTUAL_ const char *
ArgumentElement::myClassName ()
{
  return "ArgumentElement";
}

/* ========== Attributes
 */

_PUBLIC_ _VIRTUAL_ boolean ArgumentElement::isArgumentElement ()
{
  return pk_true;
}

/* PLG 2005-12-12
 * this is supposed to be expanded
 * success: this is safe and has string type 
 *	    returns this expanded StringArgument in <stringArgument_res>
 * failure: this is not safe or this has list type
 *          error message is printed
 */
boolean
ArgumentElement::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  if (HAS_STRING_TYPE ())
    {
      if (IS_SAFE ())
	{
	  (*stringArgument_res) = (ref2StringArgument) this_argumentValue;
	  return pk_true;
	}
      unsafe_named_use (this_name, "get_expandedString");
      (*isFatal) = IS_FATAL ();
      return pk_false;
    }
  fatal_named_typeConflict_ln (this_name, "string");
  (*isFatal) = pk_true;
  return pk_false;
}

/* PLG 2005-12-12
 * this is supposed to be expanded
 * success: this is safe and this type is list
 *          returns this actual list in <argumentList_res>
 * failure: this is not safe or this type is not list
 *          error message is printed
 *    <argumentList_res> remains unchanged
 */
boolean
ArgumentElement::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  if (HAS_LIST_TYPE ())
    {
      if (IS_SAFE ())
	{
	  (*argumentList_res) = (ref2ArgumentList) this_argumentValue;
	  return pk_true;
	}
      unsafe_named_use (this_name, "get_expandedList");
      (*isFatal) = IS_FATAL ();
      return pk_false;
    }
  fatal_named_typeConflict_ln (this_name, "list");
  (*isFatal) = pk_true;
  return pk_false;
}

/* PLG 2004-10-19
 * this is supposed to be expanded
 * success: returns in <arg_res> this actual expanded argument 
 * failure: this  is in error
 *          error message is printed
 */
boolean
ArgumentElement::get_expanded (ref2Argument * arg_res)
{
  if (IS_SAFE ())
    {
      (*arg_res) = this_argumentValue;
      return pk_true;
    }
  else
    unsafe_named_use (this_name, "get_expanded");
  return pk_false;
}

/* ========== Writer
 */

/* prints string declaration header: "{ :string <NAME> "
 */
_VIRTUAL_ void
ArgumentElement::print_declHeader (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_argumentType_sp (outBuffer, IS_GLOBAL_DECLARATION (), HAS_LIST_TYPE ());
  outBuffer->space (this_name);
}

/* prints object standard body: empty
 */
_VIRTUAL_ void
ArgumentElement::print_declBody (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (this_argumentValue != NIL_Argument)
    this_argumentValue->Cprint_formula (printingContext, outBuffer, pk_true);
}

/* prints object standard body: empty
 */
_VIRTUAL_ void
ArgumentElement::logDebug_declBody (int level)
{
  if (this_argumentValue == NIL_Argument)
    log_NILComponent_ln (level, "argument");
  else
    {
      this_argumentValue->logDebug_subFormula_ln (level, "argument");
    }
}

/* ========== Log
 */

/* ========== Set
 */

/* ========== Printer
 */

/* ========== Reader
 */

/* ========== Grapher
 */

/* ========== Builder
 */

/* 2005-03-09
 * <embeddingContext> is not NIL
 * creates and returns a new static list argument element descriptor in <embeddingContext>
 * owned by MKFFile associated to <embeddingContext>
 * <hashi> is its name hashing code 
 * <CString_name> is its name
 * <comments> is the comment lines associated to this declaration
 * <list_value> is its expanded argument value
 */
ArgumentElement::ArgumentElement (ref2Context embeddingContext, boolean isLocal, boolean isString, char *CString_name,
				  ref2ArgumentList comments, ref2Argument value):
Element (embeddingContext, CString_name, comments)
{
  this_isUsed = pk_false;
  if (isLocal)
    {
      SET_FILE_DECLARATION ();
    }
  else
    {
      SET_GLOBAL_DECLARATION ();
    }
  if (isString)
    {
      SET_STRING_TYPE ();
    }
  else
    {
      SET_LIST_TYPE ();
    }
  this_argumentValue = value;
  this_declarationContext->printORlevelInfo_decl_ln (this, "argument element");
}
