/* an ArgumentList is always safe
 */
#define SET_SCOPE_CARSCOPE() SET_SCOPE(car->get_scope())
#define SET_SCOPE_CDRSCOPE() SET_SCOPE(cdr->get_scope())

#define IS_FLAT() this_isFlat

/* a SAFE list can be expanded (ie static) or not, an expanded list can be flat or not, 
 * a flat list is expanded
 * an (expanded) flat list a list in which 
 *    car is a not NIL (expanded) StringArgument, car CString value may be NIL
 *    cdr is an expanded flat ArgumentList or is NIL
 * an expanded not flat list is a list in which
 *    car is NIL_Argument, cdr is an expanded ArgumentList (including NIL)
 *    car is a not NIL (expanded) StringArgument, cdr is an expanded, not flat, not NIL ArgumentList
 *    car is an expanded, not NIL ArgumentList, cdr is an expanded ArgumentList
 * a (not flat) not expanded list is a list in which
 *    car is a not expanded Argument, cdr is an ArgumentList
 *    car is an (expanded) StringArgument, cdr is a not NIL, not expanded ArgumentList
 */

_VIRTUAL_ const char *
ArgumentList::myClassName ()
{
  return "ArgumentList";
}

void
ArgumentList::myUnsafe_use (const char *where)
{
  unsafe_argumentUse ("ArgumentList", where);
}

/* returns this value
 */
_PUBLIC_ ref2Argument ArgumentList::content ()
{
  return car;
}

/* returns this successor
 */
_PUBLIC_ ref2ArgumentList ArgumentList::next ()
{
  return cdr;
}

/* ========== Private Generator
 */
/* PLG 2005-12-09
 * success: car and cdr are NIL
 *          car is a static empty string type, cdr is NIL
 *          car is a static empty string type, cdr is NIL is recursively a static empty string
 *          car is recursively a static empty string list , cdr is NIL
 *          car and cdr are static empty string lists
 * failure: some element is unsafe or concatenation failed
 *          error message is printed
 */
_PRIVATE_ _VIRTUAL_ boolean ArgumentList::is_emptyString ()
{
  // logDEBUG_in("is_emptyString");
  if ((car == NIL_Argument) || car->is_emptyString ())
    return ((cdr == NIL_ArgumentList) || cdr->is_emptyString ());
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean ArgumentList::is_isomorphic (ref2Argument str_arg)
{
  if ((ref2Argument) this == str_arg)
    return pk_true;
  if (str_arg == NIL_Argument)
    return pk_false;
  return str_arg->is_isomorphic (this);
}

_PUBLIC_ boolean ArgumentList::is_singleton ()
{
  return cdr == NIL_ArgumentList;
}

_PUBLIC_ int
ArgumentList::get_length ()
{
  ref2ArgumentList str_argList = this;
  int length = 0;
  while (str_argList != NIL_ArgumentList)
    {
      length = length + 1;
      str_argList = str_argList->next ();
    }
  return length;
}

_PUBLIC_ _VIRTUAL_ boolean ArgumentList::is_isomorphic (ref2ArgumentList str_arg)
{
  if (this == str_arg)
    return pk_true;
  if (str_arg != NIL_ArgumentList)
    {
      ref2Argument
	str_arg_car = str_arg->content ();
      ref2ArgumentList
	str_arg_cdr;
      if (car == NIL_Argument)
	{
	  if (str_arg_car != NIL_Argument)
	    return pk_false;
	}
      else
	{
	  if (!car->is_isomorphic (str_arg_car))
	    return pk_false;
	}
      str_arg_cdr = str_arg->next ();
      if (cdr == NIL_ArgumentList)
	return str_arg_cdr == NIL_ArgumentList;
      return cdr->is_isomorphic (str_arg_cdr);
    }
  return pk_false;
}

/* ========== car protected Generators
 */
/* PLG 2005-12-12
 * should be private
 * this is supposed to be SAFE, expanded
 * car is supposed to not be a StringArgument or an ArgumentList
 * <tailList> is flat
 * <tailList> is not NIL (but it works even if <tailList> is NIL and flat)
 * returns a new m'allocated flat copy of these string values (including NIL_String) 
 *             followed by <tailList>
 */
ref2ArgumentList
ArgumentList::get_flatDuplicate_expanded_car (ref2ArgumentList tailList)
{
  // logDEBUG_in("get_flatDuplicate_expanded_car(ref2ArgumentList) ENTERRING");  
  if (car->has_string_type ())
    {
      /* car is a StringArgument */
      return new ArgumentList (car, tailList);
    }
  return ((ref2ArgumentList) car)->getIf_flatDuplicate_expandedList (tailList);
}

/* 2005-03-11
 * should be private
 * this is supposed to be SAFE, expanded
 * car is supposed to have not NIL reference
 * success: returns in <CStrList_res> a new m'allocated flat copy of these string values (including NIL_String) followed by CStrList_res
 * failure: this (or recursively one of these components) is not known or copy fails
 */
boolean
ArgumentList::getIf_CStringList_cast_flat_expanded_car (ref2CStringList tailList, ref2CStringList * CStrList_res)
{
  if (IS_SAFE ())
    {
      if (car->has_string_type ())
	{
	  (*CStrList_res) = new CStringList (car->get_CString (), tailList);
	  return pk_true;
	}
      if (((ref2ArgumentList) car)->getIf_CStringList_cast_flat_expandedList (tailList, CStrList_res))
	return pk_true;
      fatal_expansion_ln ("in ArgumentList::getIf_CStringList_cast_flat_expanded_car");
    }
  else
    myUnsafe_use ("getIf_CStringList_cast_flat_expanded_car");
  return pk_false;
}

/* 2005-03-11
 * should be private
 * this is supposed to be SAFE, expanded
 * car is supposed to have not NIL reference
 * success: returns in <CStrList_res> a new m'allocated flat copy of these string values (including NIL_String) followed by lineList_res
 * failure: some element is unsafe
 *          error message is printed
 */
boolean
ArgumentList::getIf_lineList_cast_flat_expanded_car (ref2LineList tailList, ref2LineList * lineList_res)
{
  if (IS_SAFE ())
    {
      if (car->has_string_type ())
	{
	  (*lineList_res) = new LineList (car->get_CString (), tailList);
	  return pk_true;
	}
      if (((ref2ArgumentList) car)->getIf_lineList_cast_flat_expandedList (tailList, lineList_res))
	return pk_true;
      fatal_expansion_ln ("in ArgumentList::getIf_lineList_cast_flat_expanded_car");
    }
  else
    myUnsafe_use ("getIf_lineList_cast_flat_expanded_car");
}

/* PLG 2005-12-09
 * should be private
 * this is supposed to be SAFE and expanded
 * car and cdr are not NIL (not checked)
 * success: returns in <CString_res> the concatenation of these element expanded values
 * failure: some element is unsafe or concatenation failed
 *          error message is printed
 */
boolean
ArgumentList::getIf_CString_cast_expanded_car_cdr (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expanded_car_cdr");
  int car_length = car->get_CStringLength_expanded ();
  int length = car_length + cdr->get_CStringLength_expanded ();
  if (length > 0)
    {
      char *lineContent;
      LGString wreg;
      if (new_sizedPKStringContainer (length, &wreg, &lineContent))
	{
	  if (car_length > 0)
	    {
	      car->cat_expanded (lineContent);
	      if (length > car_length)
		cdr->cat_expanded (lineContent);
	    }
	  else
	    cdr->cat_expanded (lineContent);
	  (*CString_res) = detachPKString (wreg);	/* NIL_String if <wreg> is empty or errored */
	  wreg = deletePKString (wreg);
	}
      else
	return pk_false;
    }
  else
    (*CString_res) = NIL_String;
  return pk_true;
}

/* PLG 2005-12-12
 * should be private
 * identifier reference occurring in this definition are supposed to be expanded
 * car should be expanded (not checked)
 * cdr is not empty (not checked)
 * success: this is a list
 *    this cdr is a constant ArgumentList
 *         returns this in <argumentList_res>
 *    this cdr is not a constant 
 *         returns a new m'allocated list XX
 *             with XX car equal to this car, XX cdr equal to this cdr expansion
 * failure: some element is unsafe or expansion failed
 *          error message is printed
 */
boolean
ArgumentList::getIf_expand_cdr (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_expand_cdr");
  ref2ArgumentList tailList;
  if (cdr->getIf_listExpand_ (&tailList))
    {
      if (tailList == cdr)
	(*argumentList_res) = this;
      else
	(*argumentList_res) = new ArgumentList (car, tailList);
      return pk_true;
    }
  return pk_false;
}

/* PLG 2005-12-12
 * should be private
 * identifier reference occurring in this definition are supposed to be expanded
 * car is not NIL (not checked)
 * cdr is not NIL (not checked)
 * success: this is a list
 *    this is a constant ArgumentList
 *         returns this in <argumentList_res>
 *    this is not a constant 
 *         returns a new m'allocated expansion of this in <argumentList_res>
 * failure: some element is unsafe or expansion failed
 *          error message is printed
 */
boolean
ArgumentList::getIf_listExpand_car_cdr (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_listExpand_car_cdr");
  ref2Argument car_arg;
  if (car->getIf_expand_ (&car_arg))
    {
      ref2ArgumentList tailList;
      if (car == car_arg)
	return getIf_expand_cdr (argumentList_res);
      if (cdr->getIf_listExpand_ (&tailList))
	{
	  (*argumentList_res) = new ArgumentList (car_arg, tailList);
	  return pk_true;
	}
    }
  return pk_false;
}

_PUBLIC_
/* ========== element Writer
 */
/* 2004-12-09
 * prints this actual value with <stringMark> embracing
 * appends a space
 */
  void
ArgumentList::printThis_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  outBuffer->space ();
  if (car == NIL_Argument)
    outBuffer->print ("()");
  else
    car->print_expandedValue (printingContext, outBuffer, pk_true);
  outBuffer->space ();
}

/* 2004-12-09
 * prints this actual formula without <stringMark> embracing and without space
 */
void
ArgumentList::glueThis_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  if (car != NIL_Argument)
    car->glue_actualFormula (printingContext, outBuffer, depth);
}

/* 2004-12-09
 * prints this actual formula
 * if this element is an actual string it is embraced with <stringMark>
 * appends a space
 */
void
ArgumentList::spaceThis_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  outBuffer->space ();
  if (car == NIL_Argument)
    outBuffer->print ("()");
  else
    car->print_actualFormula (printingContext, outBuffer, depth, pk_true);
  outBuffer->space ();
}

/* 2004-12-09
 * prints this actual Formula without <stringMark> embracing
 * append an end of line
 */
void
ArgumentList::lineThis_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  outBuffer->line ();
  if (car != NIL_Argument)
    {
      car->glue_actualFormula (printingContext, outBuffer, depth);
      outBuffer->line ();
    }
}

/* prints this formula
 * this is assumed to be either a StringArgument or a NamedArgument
 * if it is a named argument, the argument is printed as a name occurrence
 */
void
ArgumentList::glueThis_parameterizedFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)	/* DONE */
{
  if (car != NIL_Argument)
    car->glue_parameterized (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this formula without <stringMark> embracing and without space separation
 */
void
ArgumentList::glueThis_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (car != NIL_Argument)
    car->glue_formula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this actual formula without <stringMark> embracing
 * append an end of line
 */
void
ArgumentList::lineThis_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (car != NIL_Argument)
    car->line_formula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this formula without <stringMark>
 * appends a space
 */
void
ArgumentList::spaceThis_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  outBuffer->space ();
  if (car == NIL_Argument)
    outBuffer->print ("()");
  else
    car->print_formula (printingContext, outBuffer, pk_true);
  outBuffer->space ();
}

/* 2004-12-09
 * prints this formula 
 * StringArgument and MetaArgument in function call are embraced with <stringMark>
 * appends a space
 */
void
ArgumentList::CspaceThis_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  // logDEBUG_in("CspaceThis_formula");
  outBuffer->space ();
  if (car == NIL_Argument)
    outBuffer->print ("()");
  else
    car->Cprint_formula (printingContext, outBuffer, pk_true);
  outBuffer->space ();
}

/* PLG 2008-06-04
 * COND1: <outBuffer> == STRING, this = NIL, ...
 * POST1: <outBuffer> == STRING
 * COND2: <outBuffer> == STRING, this = <arg1>, ..., <arg1> is not NIL,, <actual> is pk_true
 *        <arg1> actual value is STR1
 * COND3: <outBuffer> == STRING, this = <arg1>, ..., <arg1> is not NIL,, <actual> is false
 *        <arg1> string conversion is STR1
 * POST2-3: <outBuffer> == STRING STR1 'eol'
 */
_PUBLIC_ void
ArgumentList::printThis_comment_ln (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean actual)
{
  if (car != NIL_Argument)
    {
      if (actual)
	car->print_string_value (outBuffer);
      else
	car->print_formula (printingContext, outBuffer, pk_false);
      outBuffer->line ();
    }
}

/* ========== Element Generator
 */

/* ========== Attributes
 */

/* returns pk_true iff this is a flat list without empty argument 
 *     car is not NIL and car has string type
 *     cdr is NIL_ArgumentList or cdr is a flat list
 */
_PUBLIC_ boolean ArgumentList::is_flat ()
{
  return this_isFlat;
}

/* <arg_carValue> is static, not NIL
 * <flatCdr> is pk_true iff cdr is flat
 * sets car equal to <arg_carValue> and cdr equal to NIL
 * sets this scope and this flat state
 */
void
ArgumentList::set_staticcar_ (ref2Argument arg_carValue, boolean flatCdr)
{
  // logDEBUG_in("set_staticcar_ ");
  boolean isFatal;
  if (arg_carValue->has_string_type ())
    {
      ref2StringArgument stringArgument;
      if (arg_carValue->get_expandedString (&stringArgument, &isFatal))
	{
	  car = stringArgument;
	  this_isFlat = flatCdr;
	  return;
	}
    }
  else
    {
      ref2ArgumentList argumentList_res;
      if (arg_carValue->get_expandedList (&argumentList_res, &isFatal))
	{
	  car = argumentList_res;
	  this_isFlat = pk_false;
	  return;
	}
      this_isFlat = pk_false;
    }
  set_fatal ();
}

/* sets car equal to NIL
 * sets cdr equal to NIL
 * sets this scope equal to NOTEMPTY_STATIC
 * sets this flat state equal to false
 */
void
ArgumentList::set_NILcar_NILcdr ()
{
  // logDEBUG_in("set_NILcar_NILcdr");
  car = NIL_Argument;
  cdr = NIL_ArgumentList;
  SET_STATIC ();
  this_isFlat = pk_false;
}

/* sets car equal to a new allocated StringArgument which value is <CString_carValue>
 * sets cdr equal to NIL
 * sets this scope equal to NOTEMPTY_STATIC
 * sets this flat state equal to pk_true
 */
void
ArgumentList::set_CStringcar_NILcdr (char *CString_carValue)
{
  car = new StringArgument (CString_carValue);
  cdr = NIL_ArgumentList;
  SET_STATIC ();
  this_isFlat = pk_true;
}

/* sets car equal to <arg_carValue> 
 * sets cdr equal to NIL
 * sets this scope and this flat state
 */
void
ArgumentList::set_car_NILcdr (ref2Argument arg_carValue)
{
  // logDEBUG_in("set_car_NILcdr");
  if (arg_carValue == NIL_Argument)
    set_NILcar_NILcdr ();	/* set static scope */
  else
    {
      cdr = NIL_ArgumentList;
      if (arg_carValue->is_static ())
	{
	  SET_STATIC ();
	  set_staticcar_ (arg_carValue, pk_true);
	}
      else
	{
	  car = arg_carValue;
	  SET_SCOPE_CARSCOPE ();
	  this_isFlat = pk_false;
	}
    }
}

/* should be protected
 * assumes <tail> not NIL
 * sets this successor
 * sets this flat state and this scope
 * WARNING side effect on this using cdr assignment
 */
void
ArgumentList::append_notEmptyFlat (ref2ArgumentList tail)
{
  // logDEBUG_in("append_notEmptyFlat");  
  if (cdr == NIL_ArgumentList)
    {
      cdr = tail;
      if (car == NIL_Argument)
	{
	  SET_STATIC ();
	  this_isFlat = pk_false;
	}
      else if (car->is_static ())
	{
	  /* car is StringArgument or ArgumentList */
	  SET_STATIC ();
	  this_isFlat = car->has_string_type ();
	}
      else
	{
	  SET_SCOPE_CARSCOPE ();
	  this_isFlat = pk_false;
	}
    }
  else
    {
      cdr->append_notEmptyFlat (tail);
      if (car == NIL_Argument)
	{
	  SET_SCOPE_CDRSCOPE ();
	  this_isFlat = pk_false;
	}
      else if (car->is_static ())
	{
	  /* car is StringArgument or ArgumentList */
	  SET_SCOPE_CDRSCOPE ();
	  this_isFlat = car->has_string_type () && cdr->is_flat ();
	}
      else
	{
	  if (car->is_iterated () || cdr->is_iterated ())
	    SET_ITERATED ();
	  else
	    SET_DYNAMIC ();
	  this_isFlat = pk_false;
	}
    }
}

/* should be protected
 * assumes <tail> not NIL
 * sets this successor
 * sets this flat state and this scope
 * WARNING side effect on this using cdr assignment
 */
void
ArgumentList::append_notEmptyNotFlat (ref2ArgumentList tail)
{
  // logDEBUG_in("append_notEmptyNotFlat");  
  if (cdr == NIL_ArgumentList)
    cdr = tail;
  else
    cdr->append_notEmptyNotFlat (tail);
  if ((car == NIL_Argument) || (car->is_static ()))
    SET_SCOPE_CDRSCOPE ();
  else if (car->is_iterated () || cdr->is_iterated ())
    SET_ITERATED ();
  else
    SET_DYNAMIC ();
  this_isFlat = pk_false;
}

/* sets this successor
 * sets this flat state and this scope
 * WARNING side effect on this using cdr assignment
 * returns this scope
 */
/*
void ArgumentList::append(ref2ArgumentList tail)
{
// logDEBUG_in("append");  
  if ( tail != NIL_ArgumentList ) 
    if ( tail->is_flat() )
      append_notEmptyFlat(tail);
    else
      append_notEmptyNotFlat(tail); 
}
*/

/* 2004-12-13
 *    this argument value has been previously expanded as a string
 *        return argument value CString value
 */
char *
ArgumentList::get_CStringCar ()
{
  if (car == NIL_Argument)
    return NIL_String;
  return car->get_CString ();
}

/* 2005-10-06 PLG
 * this is supposed to be safe and expanded NOT CHECKED
 * success: returns this in <argumentList_res>
 * failure: never
 */
_VIRTUAL_ boolean
ArgumentList::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  (*argumentList_res) = this;
  return pk_true;
}

/* 2005-10-06 PLG
 * this is supposed to be safe and expanded NOT CHECKED
 * success: returns the cumulated length of its component expanded values
 * failure: never
 */
_VIRTUAL_ int
ArgumentList::get_CStringLength_expanded ()
{
  // logDEBUG_in("get_CStringLength_expanded");
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	return 0;
      return cdr->get_CStringLength_expanded ();
    }
  if (cdr == NIL_ArgumentList)
    return car->get_CStringLength_expanded ();
  return car->get_CStringLength_expanded () + cdr->get_CStringLength_expanded ();
}

/* ========== this Writer
 */

/* 2004-12-09
 * prints these actual values separated by spaces 
 * each actual string is embraces with <stringMark>
 */
_VIRTUAL_ void
ArgumentList::print_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  ref2ArgumentList tailArgumentList = this;
  if (inAList)
    outBuffer->print ('(');
  do
    {
      tailArgumentList->printThis_expandedValue (printingContext, outBuffer);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
  if (inAList)
    outBuffer->print (')');
}

/* 2004-12-09
 * prints these actual formula without <stringMark> embracing and without space
 */
_VIRTUAL_ void
ArgumentList::glue_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  ref2ArgumentList tailArgumentList = this;
  do
    {
      tailArgumentList->glueThis_actualFormula (printingContext, outBuffer, depth);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

/* 2004-12-09
 * prints these actual elements separated by space
 * each actual string is embraced with <stringMark>
 */
_VIRTUAL_ void
ArgumentList::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth, boolean inAList)
{
  ref2ArgumentList tailArgumentList = this;
  if (inAList)
    outBuffer->print ('(');
  do
    {
      tailArgumentList->spaceThis_actualFormula (printingContext, outBuffer, depth);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
  if (inAList)
    outBuffer->print (')');
}

/* 2004-12-09
 * prints these actual elements separated by space
 * each actual string is embraced with <stringMark>
 */
_VIRTUAL_ void
ArgumentList::line_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  ref2ArgumentList tailArgumentList = this;
  do
    {
      tailArgumentList->lineThis_actualFormula (printingContext, outBuffer, depth);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

/* 2004-12-09
 * prints these elements without space separator and without <stringMark> embracing
 * these elements are assumed to be either StringArgument or NamedArgument
 * each named argument is printed as a name occurrence
 */
_PUBLIC_ _VIRTUAL_ void
ArgumentList::print_parameterizedFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)	/* DONE */
{
  ref2ArgumentList tailArgumentList = this;
  do
    {
      tailArgumentList->glueThis_parameterizedFormula (printingContext, outBuffer);	/* DONE */
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

/* 2004-12-09
 * prints these elements without space separation and without <stringMark> embracing
 */
_VIRTUAL_ void
ArgumentList::glue_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2ArgumentList tailArgumentList = this;
  do
    {
      tailArgumentList->glueThis_formula (printingContext, outBuffer);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

/* 2004-12-09
 * prints these elements separated by space
 * StringArgument and MetaArgument in function call are embraced with <stringMark>
 */
_VIRTUAL_ void
ArgumentList::line_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2ArgumentList tailArgumentList = this;
  do
    {
      tailArgumentList->lineThis_formula (printingContext, outBuffer);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

/* 2004-12-09
 * prints these elements separated by space
 * StringArgument and MetaArgument in function call are embraced with <stringMark>
 */
_VIRTUAL_ void
ArgumentList::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  ref2ArgumentList tailArgumentList = this;
  if (inAList)
    outBuffer->print ('(');
  do
    {
      tailArgumentList->spaceThis_formula (printingContext, outBuffer);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
  if (inAList)
    outBuffer->print (')');
}

/* 2004-12-09
 * prints these elements separated by space, embraced with string marks
 * StringArgument and MetaArgument in function call are embraced with <stringMark>
 */
_VIRTUAL_ void
ArgumentList::Cprint_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  // logDEBUG_in("Cprint_formula");
  ref2ArgumentList tailArgumentList = this;
  if (inAList)
    outBuffer->print ('(');
  do
    {
      tailArgumentList->CspaceThis_formula (printingContext, outBuffer);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
  if (inAList)
    outBuffer->print (')');
}

/* PLG 2008-06-04
 * LET: this = <arg_1>...<arg_n>
 * COND1: <outBuffer> == STRING, all elements in <arg_1>...<arg_n> are NIL
 *      LST is this subsequence <LST_1>...<LST_m> of not NIL elements in <arg_1>...<arg_n>
 * POST1: <outBuffer> == STRING 'eol'
 * COND2: <outBuffer> == STRING, LST is this subsequence <LST_1>...<LST_m> of not NIL elements in <arg_1>...<arg_n>
 *        LST is not empty, 
 * COND2-1: <actual> is pk_true, <LST_i> actual value is STR_i
 * COND2-2: <actual> is pk_false, <LST_i> string conversion is STR_i
 * POST2: <outBuffer> == STRING 'eol'';' STR_1 'eol' ... ';' STR_m 'eol'
 */
void
ArgumentList::print_commentList (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean actual)
{
  ref2ArgumentList tailArgumentList = this;
  outBuffer->put_ln_setCommentState ();
  do
    {
      tailArgumentList->printThis_comment_ln (printingContext, outBuffer, actual);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
  outBuffer->put_ln_unsetCommentState ();
}

/* ========== Log
 */
/* prints this formula
 * this is assumed to be either a StringArgument or a NamedArgument
 * if it is a named argument, the argument is printed as a name occurrence
 */
void
ArgumentList::logThis_formula_ (boolean isStringValue)	/* DONE */
{
  if (car == NIL_Argument)
    {
      log_nil ();
    }
  else
    {
      car->log_formula_ (isStringValue);
    }
}

/* prints this formula
 * this is assumed to be either a StringArgument or a NamedArgument
 * if it is a named argument, the argument is printed as a name occurrence
 */
void
ArgumentList::glueLogThis_parameterizedFormula ()	/* DONE */
{
  if (car != NIL_Argument)
    car->glueLog_parameterized ();
}

/* 2004-12-09
 * prints in log file these elements without space separator and without <stringMark> embracing
 * these elements are assumed to be either StringArgument or NamedArgument
 * each named argument is printed as a name occurrence
 */
_PUBLIC_ _VIRTUAL_ void
ArgumentList::log_parameterizedFormula ()	/* DONE */
{
  ref2ArgumentList tailArgumentList = this;
  do
    {
      tailArgumentList->glueLogThis_parameterizedFormula ();	/* DONE */
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

_PUBLIC_ _VIRTUAL_ void
ArgumentList::log_formula_ (boolean isStringValue)
{
  // logDEBUG_in("log_formula_");
  ref2ArgumentList tailArgumentList = this;
  log_listOpen ();
  do
    {
      tailArgumentList->logThis_formula_ (pk_false);
      tailArgumentList = tailArgumentList->cdr;
      if (tailArgumentList == NIL_ArgumentList)
	{
	  log_listClose ();
	  return;
	}
      log_listSeparator ();
    }
  while (pk_true);
}

/* prints this <arg_carValue> formula
 */
void
ArgumentList::logExtendedDebugThis_formula_ln (int level)
{
  // logDEBUG_in("logExtendedDebugThis_formula_ln");
  if (car == NIL_Argument)
    log_NILComponent_ln (level, "argument in a list");
  else
    {
      if (car->has_list_type ())
	car->logExtendedDebug_O_formula_C (level);
      else
	car->logExtendedDebug_formula_ln (level + 1);
    }
  log_ln ();
}

/* prints this <arg_carValue> formula followed by its successor when it is under a definition scope
 * separates by space if separate is pk_true
 */
_PUBLIC_ _VIRTUAL_ void
ArgumentList::logInnerExtendedDebug_formula_ln (int level)
{
  ref2ArgumentList tailArgumentList = this;
  // logDEBUG_in("logInnerExtendedDebug_formula_ln");
  log_subTitle_ln (level, "these elements are ");
  do
    {
      tailArgumentList->logExtendedDebugThis_formula_ln (level + 1);
      tailArgumentList = tailArgumentList->cdr;
    }
  while (tailArgumentList != NIL_ArgumentList);
}

/* ========== Reader
 */
/* ========== Private Generator
 */
/* ========== specific ListArgument protected Generators
/* PLG 2005-12-12
 * should be protected
 * this is supposed to be safe and expanded
 * <tailList> is flat
 * <tailList> is not NIL (but it works even if <tailList> is NIL and flat)
 * returns a new m'allocated flat copy of these string values (including NIL_String) 
 *        followed by <tailList>
 */
ref2ArgumentList
ArgumentList::getIf_flatDuplicate_expandedList (ref2ArgumentList tailList)
{
  // logDEBUG_in("getIf_flatDuplicate_expandedList() ENTERRING");  
  ref2ArgumentList cdr_res;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	return tailList;
      return cdr->getIf_flatDuplicate_expandedList (tailList);
    }
  if (cdr == NIL_ArgumentList)
    return get_flatDuplicate_expanded_car (tailList);
  cdr_res = cdr->getIf_flatDuplicate_expandedList (tailList);
  return get_flatDuplicate_expanded_car (cdr_res);
}

/* 2005-10-06 PLG
 * should be protected
 * this is supposed to be expanded
 * success:
 *    returns in <strList_res> a new m'allocated copy of this 
 * failure: some element is unsafe or concatenation failed
 *          error message is printed
 */
boolean
ArgumentList::getIf_CStringList_cast_flat_expandedList (ref2CStringList tailList, ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flat_expandedList ENTERRING");  
  ref2CStringList cdr_res;
  if (car == NIL_Argument)
    {				/* this may occur only when car is NIL_ArgumentList */
      if (cdr == NIL_ArgumentList)
	{
	  (*strList_res) = tailList;
	  return pk_true;
	}
      return cdr->getIf_CStringList_cast_flat_expandedList (tailList, strList_res);
    }
  if (cdr == NIL_ArgumentList)	/* this may occur only when car is NIL_ArgumentList */
    return getIf_CStringList_cast_flat_expanded_car (tailList, strList_res);
  if (cdr->getIf_CStringList_cast_flat_expandedList (tailList, &cdr_res))
    return getIf_CStringList_cast_flat_expanded_car (cdr_res, strList_res);
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be expanded
 * should be protected
 * success:
 *    returns in <strList_res> a new m'allocated copy of this 
 * failure: some element is unsafe
 *          error message is printed
 */
boolean
ArgumentList::getIf_lineList_cast_flat_expandedList (ref2LineList tailList, ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flat_expandedList ENTERRING");  
  ref2LineList cdr_res;
  if (car == NIL_Argument)
    {				/* this may occur only when car is NIL_ArgumentList */
      if (cdr == NIL_ArgumentList)
	{
	  (*lineList_res) = tailList;
	  return pk_true;
	}
      return cdr->getIf_lineList_cast_flat_expandedList (tailList, lineList_res);
    }
  if (cdr == NIL_ArgumentList)	/* this may occur only when car is NIL_ArgumentList */
    return getIf_lineList_cast_flat_expanded_car (tailList, lineList_res);
  if (cdr->getIf_lineList_cast_flat_expandedList (tailList, &cdr_res))
    return getIf_lineList_cast_flat_expanded_car (cdr_res, lineList_res);
  return pk_false;
}

/* PLG 2005-12-12
 * should be protected
 * this is supposed to be not SAFE and expanded
 * this is not flat 
 *     returns a new m'allocated flat copy of these string values (including NIL_ArgumentList)
 *        each element may be NIL_String
 * this is flat
 *     returns this
 */
ref2ArgumentList
ArgumentList::get_flatten_expandedList ()
{
  // logDEBUG_in("get_flatten_expandedList");  
  ref2ArgumentList cdr_res;
  if (IS_FLAT ())
    return this;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	return NIL_ArgumentList;
      return cdr->get_flatten_expandedList ();
    }
  if (cdr == NIL_ArgumentList)
    {
      if (car->has_string_type ())
	return this;
      return ((ref2ArgumentList) car)->get_flatten_expandedList ();
    }
  cdr_res = cdr->get_flatten_expandedList ();
  if (cdr_res == NIL_ArgumentList)
    {
      if (car->has_string_type ())
	return new ArgumentList (car);
      return ((ref2ArgumentList) car)->get_flatten_expandedList ();
    }
  return get_flatDuplicate_expanded_car (cdr_res);
}

/* PLG 2005-12-09
 * success: returns in <CString_res> the concatenation of these element expanded values
 * failure: some element is unsafe or concatenation failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_CString_cast_expanded (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expanded");
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*CString_res) = NIL_String;
	  return pk_true;
	}
      return cdr->getIf_CString_cast_expanded (CString_res);
    }
  if (cdr == NIL_ArgumentList)
    {
      return car->getIf_CString_cast_expanded (CString_res);
    }
  return getIf_CString_cast_expanded_car_cdr (CString_res);
}

/* PLG 2005-12-09
 * this is supposed to be not SAFE or expanded 
 * success:
 *    returns in <strList_res> a new m'allocated copy of this 
 *            <strList_res> does not have NIL element
 * failure: some element is unsafe or concatenation failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_CStringList_cast_flat_expanded (ref2CStringList * strList_res)
{
  return getIf_CStringList_cast_flat_expandedList (NIL_CStringList, strList_res);
}

/* 2005-10-06 PLG
 * this is supposed to be not SAFE or expanded 
 * success:
 *    returns in <lineList_res> a new m'allocated copy of this 
 *            <lineList_res> does not have NIL element
 * failure: some element is unsafe
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_lineList_cast_flat_expanded (ref2LineList * lineList_res)
{
  return getIf_lineList_cast_flat_expandedList (NIL_LineList, lineList_res);
}

/* 2005-10-06 PLG
 * this is supposed to be SAFE and expanded (no cheking)
 *    appends all not empty expanded string elements belonging to this
 * target is supposed to be large enough
 */
_VIRTUAL_ void
ArgumentList::cat_expanded (char *target)
{
  if (car != NIL_Argument)
    car->cat_expanded (target);
  if (cdr != NIL_ArgumentList)
    cdr->cat_expanded (target);
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: 
 *    this is a constant ArgumentList
 *         returns this in <argumentList_res>
 *    this is not a constant 
 *         returns a new m'allocated expansion of this in <argumentList_res>
 * failure: some element is unsafe or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_listExpand_ (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_listExpand_");
  if (IS_DYNAMIC ())
    {
      if (car == NIL_Argument)
	{
	  if (cdr == NIL_ArgumentList)
	    {
	      return pk_false;
	    }
	  return getIf_expand_cdr (argumentList_res);
	}
      if (cdr == NIL_ArgumentList)
	{
	  ref2Argument car_arg;
	  if (!car->getIf_expand_ (&car_arg))
	    return pk_false;
	  (*argumentList_res) = new ArgumentList (car_arg);
	  return pk_true;
	}
      return getIf_listExpand_car_cdr (argumentList_res);
    }
  (*argumentList_res) = this;
  return pk_true;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: 
 *    this is a constant ArgumentList
 *         returns this in <argument_res>
 *    this is not a constant 
 *         returns a new m'allocated expansion of this in <argumentList_res>
 * failure: some element is unsafe or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_expand_ (ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_expand_");
  ref2ArgumentList arg_res;
  if (getIf_listExpand_ (&arg_res))
    {
      (*argument_res) = arg_res;
      return pk_true;
    }
  return pk_false;
}

/* PLG 2005-12-07
 * identifier reference occurring in this definition are supposed to be expanded
 * <tailList> is a not NIL, flat ArgumentList
 * success: this is dynamic
 *              expands these string values
 *              returns in <argumentList_res> a flat copy of these expanded string values (including NIL_String)
 *                         followed by flat list <tailList>
 *          this is static 
 *              returns in <argumentList_res> a flat copy of these string values (including NIL_String)
 *                         followed by flat list <tailList>
 * failure: some element is unsafe or flat expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_flatten_listCast_expand_ (ref2ArgumentList tailList, ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listCast_expand_(ref2ArgumentList, ref2ArgumentList *)");
  ref2ArgumentList cdr_res;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*argumentList_res) = tailList;
	  return pk_true;
	}
      return cdr->getIf_flatten_listCast_expand_ (tailList, argumentList_res);
    }
  if (cdr == NIL_ArgumentList)
    {
      return car->getIf_flatten_listCast_expand_ (tailList, argumentList_res);
    }
  if (cdr->getIf_flatten_listCast_expand_ (tailList, &cdr_res))
    return car->getIf_flatten_listCast_expand_ (cdr_res, argumentList_res);
  return pk_false;
}

/* PLG 2005-12-12
 * success: this is dynamic
 *              expands these string values
 *              returns in <argumentList_res> a flat copy of these expanded string values (including NIL_String)
 *          this is static and not flat (flat meaning that all car are string elements)
 *              returns in <argumentList_res> a flat copy of these string values (including NIL_String)
 *          this is static and flat
 *              returns this in <argumentList_res> 
 * failure: some element is unsafe or flat expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_flatten_listExpand_ (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listExpand_(ref2ArgumentList *)");
  if (IS_DYNAMIC ())
    {
      ref2ArgumentList cdr_res;
      if (car == NIL_Argument)
	{
	  /* CDR IS NOT nil blindage */
	  if (cdr == NIL_ArgumentList)
	    {
	      (*argumentList_res) = NIL_ArgumentList;
	      return pk_true;
	    }
	  return cdr->getIf_flatten_listExpand_ (argumentList_res);
	}
      if (cdr == NIL_ArgumentList)
	return car->getIf_flatten_listCast_expand_ (NIL_ArgumentList, argumentList_res);
      if (cdr->getIf_flatten_listExpand_ (&cdr_res))
	return car->getIf_flatten_listCast_expand_ (cdr_res, argumentList_res);
      return pk_false;
    }
  (*argumentList_res) = get_flatten_expandedList ();
  return pk_true;
}

/* PLG 2005-12-09
 * should be private
 * this is supposed to be SAFE
 * car and cdr are not NIL (not checked)
 * success: expands  these element 
 *          returns in <CString_res> the concatenation of these element expanded values
 * failure: this is unsafe, or expansion failed, or concatenation failed
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_CString_cast_expand_ (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expand_");
  ref2ArgumentList arg_res;
  if (getIf_listExpand_ (&arg_res))
    {
      if (arg_res == NIL_ArgumentList)
	this_CString_expandedValue = NIL_String;
      else if (!arg_res->getIf_CString_cast_expanded (&this_CString_expandedValue))
	return pk_false;
      (*CString_res) = this_CString_expandedValue;
      return pk_true;
    }
  return pk_false;
}

/* PLG 2005-12-12
 *    expands this as a list (of list or string elements) 
 * success:
 *        concatenate its element actual values in this_CString_expandedValue
 *        returns this concatenation
 * failure:
 *          error message is printed
 */
boolean
ArgumentList::getIf_CString_cast_listExpand_ (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_listExpand_");
  ref2ArgumentList expandedList;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*CString_res) = NIL_String;
	  return pk_true;
	}
      return cdr->getIf_CString_cast_listExpand_ (CString_res);
    }
  if (cdr == NIL_ArgumentList)
    return car->getIf_CString_cast_expand_ (CString_res);
  if (getIf_listExpand_car_cdr (&expandedList))
    {
      if (expandedList == NIL_ArgumentList)
	{
	  (*CString_res) = NIL_String;
	  return pk_true;
	}
      return expandedList->getIf_CString_cast_expanded (CString_res);
    }
  return pk_false;
}

/* 
 * returns a flat string list
 * failure:
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_CStringList_cast_flatExpand_ (ref2CStringList tailList, ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flatExpand_ ENTERRING");  
  ref2CStringList cdr_res;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*strList_res) = tailList;
	  return pk_true;
	}
      return cdr->getIf_CStringList_cast_flatExpand_ (tailList, strList_res);
    }
  if (cdr == NIL_ArgumentList)
    return car->getIf_CStringList_cast_flatExpand_ (tailList, strList_res);
  if (cdr->getIf_CStringList_cast_flatExpand_ (tailList, &cdr_res))
    {
      return car->getIf_CStringList_cast_flatExpand_ (cdr_res, strList_res);
    }
  return pk_false;
}

/* 
 * returns a flat string list
 * failure: 
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_CStringList_cast_flatExpand_ (ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flatExpand_ ENTERRING");  
  ref2CStringList cdr_res;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*strList_res) = NIL_CStringList;
	  return pk_true;
	}
      return cdr->getIf_CStringList_cast_flatExpand_ (strList_res);
    }
  if (cdr == NIL_ArgumentList)
    return car->getIf_CStringList_cast_flatExpand_ (strList_res);
  if (cdr->getIf_CStringList_cast_flatExpand_ (&cdr_res))
    {
      if (cdr_res == NIL_CStringList)
	return car->getIf_CStringList_cast_flatExpand_ (strList_res);
      return car->getIf_CStringList_cast_flatExpand_ (cdr_res, strList_res);
    }
  return pk_false;
}

/* 
 * returns a flat string list
 * failure:
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_lineList_cast_flatExpand_ (ref2LineList tailList, ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_ 2(ref2LineList, ref2LineList *)");
  ref2LineList cdr_res;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*lineList_res) = tailList;
	  return pk_true;
	}
      return cdr->getIf_lineList_cast_flatExpand_ (tailList, lineList_res);
    }
  if (cdr == NIL_ArgumentList)
    {
      return car->getIf_lineList_cast_flatExpand_ (tailList, lineList_res);
    }
  if (cdr->getIf_lineList_cast_flatExpand_ (tailList, &cdr_res))
    {
      return car->getIf_lineList_cast_flatExpand_ (cdr_res, lineList_res);
    }
  return pk_false;
}

/* 
 * returns a flat line list
 * failure:
 *          error message is printed
 */
_VIRTUAL_ boolean
ArgumentList::getIf_lineList_cast_flatExpand_ (ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_ 1(ref2LineList *)");
  ref2LineList cdr_res;
  if (car == NIL_Argument)
    {
      if (cdr == NIL_ArgumentList)
	{
	  (*lineList_res) = NIL_LineList;
	  return pk_true;
	}
      return cdr->getIf_lineList_cast_flatExpand_ (lineList_res);
    }
  if (cdr == NIL_ArgumentList)
    {
      return car->getIf_lineList_cast_flatExpand_ (lineList_res);
    }
  if (cdr->getIf_lineList_cast_flatExpand_ (&cdr_res))
    {
      if (cdr_res == NIL_LineList)
	return car->getIf_lineList_cast_flatExpand_ (lineList_res);
      return car->getIf_lineList_cast_flatExpand_ (cdr_res, lineList_res);
    }
  return pk_false;
}

/* ========== Builder
 */

/* creates and returns a new string list valued by NIL_String
 */
ArgumentList::ArgumentList ():Argument ()
{
  // logDEBUG_in("ArgumentList()");
  SET_LIST_TYPE ();
  set_NILcar_NILcdr ();		/* set static scope */
}

/* creates and returns a new string list valued by <CString_carValue>
 */
ArgumentList::ArgumentList (char *CString_carValue):Argument ()
{
  // logDEBUG_in("ArgumentList(char *)");
  SET_LIST_TYPE ();
  set_CStringcar_NILcdr (CString_carValue);	/* set static scope */
}

/* creates and returns a new string list valued by <arg_carValue>
 */
ArgumentList::ArgumentList (ref2Argument arg_carValue):Argument ()
{
  // logDEBUG_in("ArgumentList(ref2Argument)");
  SET_LIST_TYPE ();
  set_car_NILcdr (arg_carValue);	/* set scope as car scope */
}

/* creates and returns a new string list valued by <CString_carValue>
 */
ArgumentList::ArgumentList (char *CString_carValue, ref2ArgumentList tail):
Argument ()
{
  // logDEBUG_in("ArgumentList(char *,ref2ArgumentList)");
  SET_LIST_TYPE ();
  if (tail == NIL_ArgumentList)
    set_CStringcar_NILcdr (CString_carValue);	/* set static scope */
  else
    {
      car = new StringArgument (CString_carValue);
      cdr = tail;
      if (cdr->is_static ())
	{
	  SET_STATIC ();
	  this_isFlat = cdr->is_flat ();
	}
      else
	{
	  SET_SCOPE_CDRSCOPE ();
	  this_isFlat = pk_false;
	}
    }
}

/* creates and returns a new string list valued by <arg_carValue>
 */
ArgumentList::ArgumentList (ref2Argument arg_carValue, ref2ArgumentList tail):Argument ()
{
  // logDEBUG_in("ArgumentList(ref2Argument, ref2ArgumentList)");
  SET_LIST_TYPE ();
  if (tail == NIL_ArgumentList)
    set_car_NILcdr (arg_carValue);	/* set scope as car scope */
  else
    {
      cdr = tail;
      if (arg_carValue == NIL_Argument)
	{
	  car = NIL_Argument;
	  SET_SCOPE_CDRSCOPE ();
	  this_isFlat = pk_false;
	}
      else if (arg_carValue->is_static ())
	{
	  SET_SCOPE_CDRSCOPE ();
	  set_staticcar_ (arg_carValue, cdr->is_flat ());
	}
      else
	{
	  car = arg_carValue;
	  if ((car->is_iterated ()) || (cdr->is_iterated ()))
	    SET_ITERATED ();
	  else
	    SET_DYNAMIC ();
	  this_isFlat = pk_false;
	}
    }
  // logDEBUG_in("ArgumentList 5 done");
}
