#define PRINT_STRING_VALUE(outBuffer)  outBuffer->print(this_CString_expandedValue)

/* Final type is
   string for a StringArgument
   string list for a ArgumentList
   final type of a Meta, Identifier, Parameter or Command argument may be unknown
   final type of a Meta, Identifier, Parameter or Command argument when it is known is string or list
   a SAFE argument can be expanded (ie static) or not, an expanded argument can be flat or not, 
   its type can be string or list
*/

_PUBLIC_
/* ==========  ==========  ==========  ========== TITLE: these attributes
 */
_VIRTUAL_ const char *
Argument::myClassName ()
{
  return "Argument";
}

/* returns pk_true iff this type is string
*/
boolean
Argument::has_string_type ()
{
  return HAS_STRING_TYPE ();
}

/* returns pk_true iff this type is string list
*/
boolean
Argument::has_list_type ()
{
  return HAS_LIST_TYPE ();
}

/* returns parameter sign 
   result is meaningless if stringtype is not string or string list type 
		       *      0 if this is empty
		       *      1 if this is not empty without parameter occurrences
		       *      2 if this has some parameter occurrence
*/
int
Argument::get_scope ()
{
  return GET_SCOPE ();
}

/* returns pk_true if this Argument is static (does not have parameter or iterator occurrence)
 * this type is assumed to be list type NOT CHECKED
 */
boolean
Argument::is_static ()
{
  return IS_STATIC ();
}

/* returns pk_true if this Argument is dynamic (has parameter occurrence but does not have iterator occurrence)
 */
boolean
Argument::is_dynamic ()
{
  return IS_DYNAMIC ();
}

/* returns pk_true if this Argument is iterated (has iterator occurrence)
 */
boolean
Argument::is_iterated ()
{
  return IS_ITERATED ();
}

/* PLG 2005-12-09
 * success: never
 * failure: always
 */
_VIRTUAL_ boolean
Argument::is_emptyString ()
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (char *str_val)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (ref2Argument str_arg)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (ref2StringArgument str_arg)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (ref2MetaArgument str_arg)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (ref2ArgumentList str_arg)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (ref2NamedArgument str_arg)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Argument::is_isomorphic (ref2CommandArgument str_arg)
{
  return pk_false;
}

/* sets this (temporary) actual string value equal to aString
 */
void
Argument::set_CString (char *aString)
{
  this_CString_expandedValue = aString;
}

/* ==========  ==========  ==========  ========== TITLE: PRINT
 */
/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::Cprint_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  //logDEBUG_in("Cprint_formula");
  printingContext->out_stringConstant (outBuffer, this_CString_expandedValue);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_PUBLIC_ _VIRTUAL_ void
Argument::print_parameterizedFormula (ref2Lexical lexicalContext, ref2IntePrinter outBuffer)	/* DONE */
{
  logUnsafe_parameterizedFormula ();
}

/* 2004-12-09
 * prints this actual string value 
 * this is supposed to be expanded
 */
void
Argument::print_string_value (ref2IntePrinter outBuffer)
{
  PRINT_STRING_VALUE (outBuffer);
}

_VIRTUAL_ void
Argument::print_value (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  PRINT_STRING_VALUE (outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::print_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  print_value (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::glue_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  print_value (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth, boolean inAList)
{
  print_value (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::line_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  print_actualFormula (printingContext, outBuffer, depth, pk_false);
  outBuffer->line ();
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::glue_parameterized (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_value (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::glue_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_value (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  glue_formula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this
 * StringArgument and MetaArgument are embraced with <stringMark>
 */
_VIRTUAL_ void
Argument::line_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_formula (printingContext, outBuffer, pk_false);
  outBuffer->line ();
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::print_externNameValue (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_externName (outBuffer, this_CString_expandedValue);
}

_VIRTUAL_ void
Argument::print_externNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_externNameValue (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::print_fileNameValue (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_fileNameValue (outBuffer, this_CString_expandedValue);
}

_VIRTUAL_ void
Argument::print_fileNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_fileNameValue (printingContext, outBuffer);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::print_varNameValue (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_expandedValue (printingContext, outBuffer, pk_false);
}

_VIRTUAL_ void
Argument::print_varNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_varNameValue (printingContext, outBuffer);
}

/* ==========  ==========  ==========  ========== TITLE: ACTIONS
 */
/* 2005-10-06 PLG
 * returns this linked actual string value 
 * this is supposed to be an argument previously casted to CString (not checked)
 *      is redefined in NamedArgument
 */
_VIRTUAL_ char *
Argument::get_CString ()
{
  return this_CString_expandedValue;
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  fatal_typeConflict_ln ("string");
  (*isFatal) = pk_true;
  return pk_false;
}

/* PLG 2005-12-08
 * success: never
 * failure: alwasy
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  fatal_typeConflict_ln ("list");
  (*isFatal) = pk_true;
  return pk_false;
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ int
Argument::get_CStringLength_expanded ()
{
  unsafe_use ("get_CStringLength_expanded()");
  return 0;
}

/* ========== Log
 */
/* ==========  ==========  ==========  ========== TITLE: printing in log
 */
/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::glueLog_parameterized ()
{
  glueLog_ (this_CString_expandedValue);
}

_PUBLIC_ _VIRTUAL_ void
Argument::log_formula_ (boolean isStringValue)
{
  // logDEBUG_in("log_formula_ in Argument");
  if (isStringValue)
    {
      if (this_CString_expandedValue == (char *) NULL)
	{
	  log_nil ();
	}
      else
	{
	  log_stringConstant (this_CString_expandedValue);
	}
    }
  else
    log_formula ();
}

_PUBLIC_ void
Argument::log_formula_ln ()
{
  log_formula ();
  log_ln ();
}

_PUBLIC_ _VIRTUAL_ void
Argument::logInnerExtendedDebug_formula_ln (int level)
{
}

_PUBLIC_ void
Argument::logExtendedDebug_formula_ln (int level)
{
  // logDEBUG_in("logExtendedDebug_formula_ln");
  logDebug_objAnnounce_ln (level);
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	log_hasStringType (level);
      else if (HAS_LIST_TYPE ())
	log_hasListType (level);
      else
	log_hasUnknowmType (level);
    }
  else
    log_isInError (level);
  if (IS_EMPTY ())
    log_NILComponent_ln (level, "parameter sign");
  else if (IS_NOTEMPTY ())
    {
      if (ISKNOWN_STATIC ())
	log_spn_str_ln (level, "this parameter sign is dynamic identifier free");
      else if (IS_NOTITERATED ())
	log_spn_str_ln (level, "this parameter sign is iterator free");
      else
	log_spn_str_ln (level, "this parameter sign is iterated");
    }
  else
    log_spn_str_ln (level, "this parameter sign is unknown");
  if (HAS_STRING_TYPE ())
    if (this_CString_expandedValue == NIL_String)
      log_NILComponent_ln (level, "string actual value");
    else
      log_extendedDebug_stringConstant (level, this_CString_expandedValue);
  else
    log_spn_str_ln (level, "this string actual value is meaningless");
  logInnerExtendedDebug_formula_ln (level);
  logDebug_objClosing_ln (level);
}

void
Argument::logExtendedDebug_O_formula_C (int level)
{
  //logDEBUG_in("logExtendedDebug_O_formula_C");
  log_subFormulaOpen ();
  logExtendedDebug_formula_ln (level + 1);
  log_subFormulaClose ();
}

/* ========== Reader
 */
/* ========== Generator
 */

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_CString_cast_expanded (char **CString_res)
{
  unsafe_use ("getIf_CString_cast_expanded(char **)");
  return pk_false;
}

/* assumes expanded CString in this_CString_expandedValue
 * returns this expanded CString converted to CString list
 */
ref2CStringList
Argument::new_CStringList ()
{
  return new CStringList (this_CString_expandedValue);
}

/* assumes expanded CString in this_CString_expandedValue
 * returns this expanded CString converted to CString list
 */
ref2CStringList
Argument::new_CStringList (ref2CStringList tailList)
{
  return new CStringList (this_CString_expandedValue, tailList);
}

/* assumes expanded CString in this_CString_expandedValue
 * returns this expanded CString converted to Line list
 */
ref2LineList
Argument::new_LineList ()
{
  return new LineList (this_CString_expandedValue);
}

/* assumes expanded CString in this_CString_expandedValue
 * returns this expanded CString converted to Line list
 */
ref2LineList
Argument::new_LineList (ref2LineList tailList)
{
  return new LineList (this_CString_expandedValue, tailList);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_CStringList_cast_flat_expanded (ref2CStringList * strList_res)
{
  unsafe_use ("getIf_CStringList_cast_flat_expanded(ref2CStringList *)");
  return pk_false;
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_lineList_cast_flat_expanded (ref2LineList * lineList_res)
{
  unsafe_use ("getIf_lineList_cast_flat_expanded(ref2LineList *)");
  return pk_false;
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ void
Argument::cat_expanded (char *target)
{
  unsafe_use ("cat_expanded(char *)");
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_stringExpand_ (ref2StringArgument * stringArgument_res)
{
  unsafe_use ("getIf_stringExpand_(ref2StringArgument *)");
  logExtendedDebug_formula_ln (10);
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_listExpand_ (ref2ArgumentList * argumentList_res)
{
  unsafe_use ("getIf_listExpand_(ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_expand_ (ref2Argument * argument_res)
{
  unsafe_use ("getIf_expand_(ref2Argument *)");
  return pk_false;
}


/* PLG 2005-12-07
 * this is supposed to be a string (not checked)
 * expand this string value
 * success: returns in <argumentList_res> a new m'allocated string list equal to this_CString_expandedValue followed by <tailList>
 * failure: this is in error or flat expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_flatten_listCast_expand_ (ref2ArgumentList tailList, ref2ArgumentList * argumentList_res)
{
  unsafe_use ("getIf_flatten_listCast_expand_(ref2ArgumentList, ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_flatten_listExpand_ (ref2ArgumentList * argumentList_res)
{
  unsafe_use ("getIf_flatten_listExpand_(ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_CString_cast_expand_ (char **CString_res)
{
  unsafe_use ("getIf_CString_cast_expand_(char **)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_CStringList_cast_flatExpand_ (ref2CStringList tailList, ref2CStringList * strList_res)
{
  unsafe_use ("getIf_CStringList_cast_flatExpand_(ref2CStringList, ref2CStringList *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_CStringList_cast_flatExpand_ (ref2CStringList * strList_res)
{
  unsafe_use ("getIf_CStringList_cast_flatExpand_(ref2CStringList *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_lineList_cast_flatExpand_ (ref2LineList tailList, ref2LineList * lineList_res)
{
  unsafe_use ("getIf_lineList_cast_flatExpand_(ref2LineList, ref2LineList *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
Argument::getIf_lineList_cast_flatExpand_ (ref2LineList * lineList_res)
{
  unsafe_use ("getIf_lineList_cast_flatExpand_(ref2LineList *)");
  return pk_false;
}

/* ========== Builder
 */

/* creates and returns a new generic argument 
 */
Argument::Argument ():ArgumentReporting ()
{
  // logDEBUG_in("Argument()");
  SET_UNKNOWN_SCOPE ();
  SET_UNKNOWN_TYPE ();
  this_CString_expandedValue = NIL_String;
}

/* creates and returns a new generic argument
 */
Argument::Argument (char *aString):ArgumentReporting ()
{
  // logDEBUG_in("Argument(char *)");
  SET_UNKNOWN_SCOPE ();
  SET_STRING_TYPE ();
  if (aString == NIL_String)
    SET_EMPTY ();
  else
    SET_STATIC ();
  this_CString_expandedValue = aString;
}
