#define HAS_PRELUDE() ( this_prelude != NIL_LineList )
#define HAS_ANCESTOR() ( this_ancestorList != NIL_AncestorList )
#define HASEMPTY_DEFAULT() ( this_default == NIL_LineList )
#define HAS_DEFAULT() ( this_default != NIL_LineList )

_PUBLIC_
/* ==========  ==========  Attributes
 */
  boolean BuildingData::has_ancestor ()
{
  return HAS_ANCESTOR ();
}

/* returns anObject default action 
 */
ref2AncestorList
BuildingData::get_ancestors ()
{
  return this_ancestorList;
}

/* returns pk_true default action is not empty
 */
boolean
BuildingData::has_mainAction ()
{
  if (HAS_PRELUDE ())
    return pk_true;
  return HAS_ANCESTOR ();
}

/* returns pk_true if default action must be used when components are up to date
 */
boolean
BuildingData::is_defaultActivated ()
{
  return thisHas_conditionalBuilding;
}

boolean
BuildingData::set_defaultActivated ()
{
  thisHas_conditionalBuilding = pk_true;
  return pk_true;
}

/* returns pk_true default action is not empty
 */
boolean
BuildingData::has_defaultAction ()
{
  return HAS_DEFAULT ();
}

/* returns pk_true default action is not empty
 */
boolean
BuildingData::is_empty ()
{
  if (HAS_PRELUDE ())
    return pk_false;
  if (HAS_ANCESTOR ())
    return pk_false;
  return HASEMPTY_DEFAULT ();
}

/*
 * creates a building data descriptor in which 
 * thisHas_conditionalBuilding is initiated with isCond value
 * this_visitationLevel,  has been initiated with -1
 * prelude, this_ancestorList, postludeAction are intiated with NIL
 * this_default is intiated with default_lines
 */
void
BuildingData::set_attributes (ref2LineList prelude_lines, ref2AncestorList ancestors, ref2LineList default_lines)
{
  this_prelude = prelude_lines;
  this_ancestorList = ancestors;
  this_default = default_lines;
}

/* ==========  ==========  Grapher
 */

/* PLG 2005-08-30
 */
void
BuildingData::set_recursiveEverVisited ()
{
  this_visitationLevel = 2;
}

/* PLG 2005-08-30
 */
void
BuildingData::set_everVisited ()
{
  if (this_visitationLevel <= 0)
    this_visitationLevel = 1;
}

/* PLG 2006-01-03
 * recursive ever build visitation
 * success: a previous recursive visitation has occurred or
 * failure: this does not have been previously recursively visited
 *               this is set as recursively visited
 */
boolean
BuildingData::time_checking_recursiveEver_done ()
{
  // fprintf(stderr, "DEBUG  BuildingData::time_checking_recursiveEver_done, recursive level is %i\n", this_visitationLevel);
  if (this_visitationLevel >= 2)	/* has been recursively visited */
    return pk_true;
  this_visitationLevel = 2;
  return !(HAS_ANCESTOR ());
}

/* ==========  ==========  Writer
 */
_PRIVATE_ void
BuildingData::print_prelude_ln (ref2IntePrinter outBuffer)
{
  if (HAS_PRELUDE ())
    {
      outBuffer->put_ln_preludeOpening_ln ();
      this_prelude->line (outBuffer);
    }
}

_PRIVATE_ void
BuildingData::print_AncestorList_ln (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (HAS_ANCESTOR ())
    this_ancestorList->print_formula (printingContext, outBuffer);
}

_PRIVATE_ void
BuildingData::print_postlude_ln (ref2IntePrinter outBuffer)
{
  /* PLG not implemented 
   */
}

_PRIVATE_ void
BuildingData::print_default_ln (ref2IntePrinter outBuffer)
{
  if (HAS_DEFAULT ())
    {
      outBuffer->put_ln_defaultOpening_ln ();
      this_default->line (outBuffer);
    }
}

_PUBLIC_ void
BuildingData::print (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_prelude_ln (outBuffer);
  print_AncestorList_ln (printingContext, outBuffer);
  print_postlude_ln (outBuffer);
  print_default_ln (outBuffer);
}

/* ==========  ==========  Log
 */

void
BuildingData::logDebug_prelude_ln (int level)
{
  if (HAS_PRELUDE ())
    this_prelude->logExtendedDebug_formula_ln (level);
  else
    log_NILComponent_ln (level, "prelude generator");
}

void
BuildingData::logDebug_AncestorList_ln (int level)
{
  if (HAS_ANCESTOR ())
    {
      this_ancestorList->logDebug_subFormula_ln (level, "ancestor list");
    }
  else
    log_NILComponent_ln (level, "ancestor list");
}

void
BuildingData::logDebug_postlude_ln (int level)
{
  /* PLG not implemented 
   */
}

void
BuildingData::logDebug_default_ln (int level)
{
  if (HAS_DEFAULT ())
    this_default->logDebug_subFormula_ln (level, "default generator");
  else
    log_NILComponent_ln (level, "default generator");
}

_PUBLIC_ _VIRTUAL_ void
BuildingData::logExtendedDebug_formula_ln (int level)
{
  logDebug_prelude_ln (level);
  logDebug_AncestorList_ln (level);
  logDebug_postlude_ln (level);
  logDebug_default_ln (level);
}

/* ==========  ========== Reader
 */

/* PLG 2008_07-02
 * reads and sets this BuildingData content
 * success: FILE = PREFIX.[_NODE_ACTION]#'}'[Any*][EOF]
 *          where [_NODE_ACTION] ::= opt{[_PRELUDE]} [_ANCESTOR*] opt{[_DEFAULT]}
 * failure: some error occurred during component reading
 *          error message is printed
 *          cursor is moved to  errored position
 * PLG SOURCE OF CONSTRAINTS ON SPACE READING readOpt_buildingData_eol_
 */
boolean
BuildingData::readOpt_buildingData_eol_ (ref2Context readingContext)
{
  // readingContext->log_ASDEBUG("BuildingData::readOpt_buildingData_eol_");
  // fprintf(stderr, "DEBUG BuildingData::readOpt_buildingData_eol_, readingContext type is %s\n", readingContext->myClassName());
  if (!readingContext->readOpt_nodePrelude_eol_ (&this_prelude))
    return pk_false;
  if (!readingContext->readOpt_ancestorList_eol_ (&this_ancestorList))
    return pk_false;
  return readingContext->readOpt_nodeDefault_eol_ (&this_default);
}

/* ==========  ==========  Generator
 */
/*
 *     curBuiltFile is the nearest file that depends on node which this is a components
 *     
 */
boolean
BuildingData::executeDefault ()
{
  if (HAS_DEFAULT ())
    return this_default->execute ();
  return pk_true;
}

/*
 *     curBuiltFile is the nearest file that depends on node which this is a components
 *     
 */
boolean
BuildingData::execute (ref2NodeElement curExecutedNode)
{
  // fprintf(stderr, "DEBUG BuildingData::execute\n");
  if (HAS_PRELUDE ())
    {
      if (!(this_prelude->execute ()))
	return pk_false;
    }
  if (HAS_ANCESTOR ())
    {
      return (this_ancestorList->execute (curExecutedNode));
    }
  return pk_true;
}

/*
 *     curBuiltFile is the nearest file that depends on node which this is a components
 *     
 */
void
BuildingData::generateDefault (ref2Buffer fileToPrint)
{
  if (HAS_DEFAULT ())
    this_default->generate (fileToPrint);
}

/*
 *     curBuiltFile is the nearest file that depends on node which this is a components
 *     
 */
boolean
BuildingData::generate (ref2NodeElement curBuiltFile, ref2Buffer fileToPrint)
{
  if (HAS_PRELUDE ())
    this_prelude->generate (fileToPrint);
  if (HAS_ANCESTOR ())
    return (this_ancestorList->generate (curBuiltFile, fileToPrint));
  return pk_true;
}

/*
 * proceeds all inner constructive nodes
 * is not safe if some cycle exists
 *     
 */
boolean
BuildingData::do_innerConstructive (ref2Context logContext, boolean activeDependence)
{
  if (HAS_ANCESTOR ())
    return this_ancestorList->do_innerConstructive (logContext, activeDependence);
  return pk_true;
}

/* ========== Builders ==========
 */

/*
 * used in Node
 */
BuildingData::BuildingData ()
{
  this_prelude = NIL_LineList;
  this_ancestorList = NIL_AncestorList;
  postludeAction = NIL_LineList;
  thisHas_conditionalBuilding = pk_false;
  this_default = NIL_LineList;
  this_visitationLevel = -1;
}
