#define IS_SAFE_STATIC() ( IS_SAFE() && (this_scope <= 1) )

#define STRING_CALL(TARGET) ( this_functionCall->string_call(TARGET) )

/* provided that creation requirements are satisfied
 *     this_functionCall is never NIL
 */
_PUBLIC_ _VIRTUAL_ const char *
CommandArgument::myClassName ()
{
  return "CommandArgument";
}

/* ========== Attributes
 */

void
CommandArgument::myFatal_typeConflict (const char *where, const char *typeStr)
{
  fatal_typeConflict_ln (typeStr);
}

_PUBLIC_ ref2ArgumentFunctionCall CommandArgument::get_functionCall ()
{
  return this_functionCall;
}

/* 2005-10-06 PLG
 * this is supposed to be expanded  (not checked)
 * success: returns in <stringArgument_res> this evaluated actual string value 
 * failure: this does not have string type   
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  if (HAS_STRING_TYPE ())
    {
      if (IS_SAFE ())
	{
	  (*stringArgument_res) = (ref2StringArgument) this_expandedArgument;
	  return pk_true;
	}
      unsafe_use ("CommandArgument::get_expandedString");
      (*isFatal) = IS_FATAL ();
      return pk_false;
    }
  myFatal_typeConflict ("get_expandedString", "string");
  (*isFatal) = pk_true;
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be expanded  (not checked)
 * success: returns in <argumentList_res> this evaluated actual list value 
 * failure: this is not safe or does not have list type   
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  // logDEBUG_in("get_expandedList");
  if (HAS_LIST_TYPE ())
    {
      if (IS_SAFE ())
	{
	  (*argumentList_res) = (ref2ArgumentList) this_expandedArgument;
	  return pk_true;
	}
      unsafe_use ("CommandArgument::get_expandedList");
      (*isFatal) = IS_FATAL ();
      return pk_false;
    }
  myFatal_typeConflict ("get_expandedList", "list");
  (*isFatal) = pk_true;
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be safe and expanded NOT CHECKED
 * this has string type
 *    returns length of this expanded string
 * this has list type
 *    returns the cumulated length of these expanded element strings
 */
_VIRTUAL_ int
CommandArgument::get_CStringLength_expanded ()
{
  // logDEBUG_in("get_CStringLength_expanded");
  if (this_expandedArgument == NIL_Argument)
    return 0;
  return this_expandedArgument->get_CStringLength_expanded ();
}

_PUBLIC_ _VIRTUAL_ boolean CommandArgument::is_isomorphic (ref2Argument str_arg)
{
  if ((ref2Argument) this == str_arg)
    return pk_true;
  if (str_arg != NIL_Argument)
    {
      return str_arg->is_isomorphic (this);
    }
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean CommandArgument::is_isomorphic (ref2CommandArgument str_arg)
{
  if (this == str_arg)
    return pk_true;
  if (str_arg != NIL_CommandArgument)
    {
      ref2ArgumentFunctionCall
	str_arg_functionCall = str_arg->get_functionCall ();
      if (this_functionCall == str_arg_functionCall)
	return pk_true;
      return str_arg_functionCall->is_isomorphic (this_functionCall);
    }
  return pk_false;
}

/* ========== Writer
 */

/* 2004-12-09
 * print this actual argument value embraced with <stringMark>
 */
_VIRTUAL_ void
CommandArgument::print_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  if (this_expandedArgument != NIL_Argument)
    this_expandedArgument->print_expandedValue (printingContext, outBuffer, inAList);
}

/* 2004-12-09
 * prints this partially expanded value (expansion is only done for static values)
 */
_VIRTUAL_ void
CommandArgument::glue_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  if (depth > 0)
    this_functionCall->print_actualFormula (printingContext, outBuffer, depth);
  else if (this_expandedArgument != NIL_Argument)
    this_expandedArgument->print_actualFormula (printingContext, outBuffer, 0, pk_false);
}

/* 2004-12-09
 * print this actual value as an extern reference
 */
_VIRTUAL_ void
CommandArgument::print_externNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_externNameCommand (outBuffer, this_functionCall->get_name ());
  this_functionCall->print_actualParametersformula (printingContext, outBuffer);
}

/* 2004-12-09
 * print this actual value as a file reference
 */
_VIRTUAL_ void
CommandArgument::print_fileNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_fileNameCommand (outBuffer, this_functionCall->get_name ());
  this_functionCall->print_actualParametersformula (printingContext, outBuffer);
}

/* 2004-12-09
 * print this actual value as a var reference
 */
_VIRTUAL_ void
CommandArgument::print_varNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_varNameCommand (outBuffer, this_functionCall->get_name ());
  this_functionCall->print_actualParametersformula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this partially expanded value (expansion is only done for static values)
 */
_VIRTUAL_ void
CommandArgument::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth, boolean inAList)
{
  CommandArgument::glue_actualFormula (printingContext, outBuffer, depth);
}

/* 2004-12-09
 * prints this without <stringMark>
 */
_VIRTUAL_ void
CommandArgument::glue_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  this_functionCall->print_formula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this without <stringMark>
 */
_VIRTUAL_ void
CommandArgument::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  this_functionCall->print_formula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this without <stringMark>
 */
_VIRTUAL_ void
CommandArgument::Cprint_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  // logDEBUG_in("Cprint_formula");
  this_functionCall->print_formula (printingContext, outBuffer);
}

/* ========== Log
 */
/* 
 */
_PUBLIC_ _VIRTUAL_ void
CommandArgument::log_formula_ (boolean isStringValue)
{
  // logDEBUG_in("log_formula_");
  log_innerOpen ();
  if (this_functionCall == NIL_ArgumentFunctionCall)
    {
      log_nil ();
    }
  else
    {
      this_functionCall->log_formula_ ();
    }
  log_valueOpen ();
  if (this_expandedArgument == NIL_Argument)
    log_nil ();
  else
    this_expandedArgument->log_formula_ (isStringValue);
  log_valueClose ();
  log_innerClose ();
}

_PUBLIC_ _VIRTUAL_ void
CommandArgument::logInnerExtendedDebug_formula_ln (int level)
{
  if (HAS_STRING_TYPE ())
    logDebug_thisIs_ln (level, "string function call");
  else if (HAS_LIST_TYPE ())
    logDebug_thisIs_ln (level, "list function call");
  else
    logDebug_thisIs_ln (level, "an errored function call");
  this_functionCall->logDebug_subFormula_ln (level, "function call");
  if (this_expandedArgument == NIL_Argument)
    log_NILComponent_ln (level + 1, "actual argument");
  else
    this_expandedArgument->logDebug_subFormula_ln (level + 1, "actual argument formula");
}

/* ========== Reader
 */
/* ========== Generator
 */
/* PLG 2005-12-12
 * this is supposed to be expanded (not checked)
 * success: this has string type
 *             returns in <CString_res> this expanded CString value 
 *          this has list type
 *             returns in <CString_res> the CString concatenation of strings belonging to this expanded list value 
 * failure: this is errored   
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_CString_cast_expanded (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expanded");
  if (IS_SAFE ())
    {
      if (this_expandedArgument == NIL_Argument)
	{
	  (*CString_res) = NIL_String;
	  return pk_true;
	}
      return this_expandedArgument->getIf_CString_cast_expanded (CString_res);
    }
  else
    unsafe_use ("getIf_CString_cast_expanded(char **)");
  return pk_false;
}

/* 2005-10-06 PLG
 *    this has been previously expanded as a string or as a list (of list or string elements) 
 * success:
 *    this is a string
 *        returns in <strList_res> a new m'allocated singleton containing this expanded CString value 
 *    this is a list
 *        returns in <strList_res> a new m'allocated copy of this 
 *            <strList_res> does not have NIL element
 * failure: this is in error
 */
_VIRTUAL_ boolean
CommandArgument::getIf_CStringList_cast_flat_expanded (ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flat_expanded");
  if (IS_SAFE ())
    {
      if (this_expandedArgument == NIL_Argument)
	{
	  (*strList_res) = NIL_CStringList;
	  return pk_true;
	}
      return this_expandedArgument->getIf_CStringList_cast_flat_expanded (strList_res);
    }
  else
    unsafe_use ("getIf_CStringList_cast_flat_expanded(ref2CStringList *)");
  return pk_false;
}

/* 2005-10-06 PLG
 *    this has been previously expanded as a string or as a list (of list or string elements) 
 * success:
 *    this is a string
 *        returns in <lineList_res> a new m'allocated singleton containing this expanded CString value 
 *    this is a list
 *        returns in <lineList_res> a new m'allocated copy of this 
 *            <lineList_res> does not have NIL element
 * failure: this is in error
 */
_VIRTUAL_ boolean
CommandArgument::getIf_lineList_cast_flat_expanded (ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flat_expanded");
  if (IS_SAFE ())
    {
      if (this_expandedArgument == NIL_Argument)
	{
	  (*lineList_res) = NIL_LineList;
	  return pk_true;
	}
      return this_expandedArgument->getIf_lineList_cast_flat_expanded (lineList_res);
    }
  else
    unsafe_use ("getIf_lineList_cast_flat_expanded(ref2LineList *)");
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be expanded
 * this is a string
 *    if this actual value is not NIL_String
 *      appends, without C string marks, this actual value to target
 * this is a list
 *    appends all not empty expanded string elements belonging to this
 * target is supposed to be large enough
 */
_VIRTUAL_ void
CommandArgument::cat_expanded (char *target)
{
  // logDEBUG_in("cat_expanded");
  if (this_expandedArgument != NIL_Argument)
    this_expandedArgument->cat_expanded (target);
}

/* 2005-01-17
 * this_functionCall resulting type is string (not checked)
 * success: expands this string value by processing this_functionCall
 * failure: call failed
 *          error message is printed
 */
boolean
CommandArgument::string_call ()
{
  // logDEBUG_in("string_call");
  ref2StringArgument stringArgument;
  if (STRING_CALL (&stringArgument))
    {
      this_expandedArgument = (ref2Argument) stringArgument;
      if (stringArgument == NIL_StringArgument)
	this_CString_expandedValue = NIL_String;
      else
	this_CString_expandedValue = stringArgument->get_CString ();
      return pk_true;
    }
  this_expandedArgument = NIL_Argument;
  this_CString_expandedValue = NIL_String;
  return pk_false;
}

/* PLG 2005-12-07
 * this_functionCall result is a list (not checked)
 * success: expands and flattens this list value by processing this_functionCall
 * failure: call failed
 *          error message is printed
 */
boolean
CommandArgument::list_call ()
{
  // logDEBUG_in("list_call()");
  ref2ArgumentList argumentList_res;
  if (this_functionCall->list_call (&argumentList_res))
    {
      this_expandedArgument = (ref2Argument) argumentList_res;
      this_CString_expandedValue = NIL_String;
      return pk_true;
    }
  this_expandedArgument = NIL_Argument;
  this_CString_expandedValue = NIL_String;
  return pk_false;
}

/* PLG 2005-12-12
 * success: expands this value by calling this_functionCall
 * failure: this_functionCall failed
 *          error message is printed
 */
boolean
CommandArgument::call (ref2Argument * argument_res)
{
  // logDEBUG_in("call(ref2Argument *)");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (!string_call ())
	    return pk_false;
	}
      else
	{			/* has list type */
	  if (!list_call ())
	    return pk_false;
	}
      (*argument_res) = this_expandedArgument;
      return pk_true;
    }
  else
    unsafe_use ("call(ref2Argument *)");
  return pk_false;
}

/* PLG 2005-12-07
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is a string
 *    this is a constant (StringArgument) returns this in <stringArgument_res>
 *    this is not a constant (StringArgument) returns a new m'allocated expansion of this in <stringArgument_res>
 * failure: this is not safe or does not have list type, or expansion failed 
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_stringExpand_ (ref2StringArgument * stringArgument_res)
{
  // logDEBUG_in("getIf_stringExpand_()");
  if (IS_SAFE () && HAS_STRING_TYPE ())
    {
      if (string_call ())
	{
	  (*stringArgument_res) = (ref2StringArgument) this_expandedArgument;
	  return pk_true;
	}
      return pk_false;
    }
  unsafe_use ("getIf_stringExpand_(ref2StringArgument *)");
  return pk_false;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is safe, this type is list
 *    this is a constant returns this expanded value in <stringArgument_res>
 *    this is not a constant (StringArgument) returns a new m'allocated expansion of this in <argumentList_res>
 * failure: this is not safe, or this is a string
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_listExpand_ (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_listExpand_()");
  if (IS_SAFE () && HAS_LIST_TYPE ())
    {
      if (list_call ())
	{
	  (*argumentList_res) = (ref2ArgumentList) this_expandedArgument;
	  return pk_true;
	}
      return pk_false;
    }
  unsafe_use ("getIf_listExpand_(ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is a string
 *             returns this in <argument_res> the value returned by getIf_stringExpand_(argument_res)
 *          this is list
 *             returns this in <argument_res> the value returned by getIf_listExpand_(argument_res)
 * failure: this is in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_expand_ (ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_expand_");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (!string_call ())
	    return pk_false;
	}
      else
	{			/* has list type */
	  if (!list_call ())
	    return pk_false;
	}
      (*argument_res) = this_expandedArgument;
      return pk_true;
    }
  unsafe_use ("getIf_expand_(ref2Argument *)");
  return pk_false;
}

/* PLG 2005-12-07
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is dynamic
 *              expands these string values
 *              returns in <argumentList_res> a flat copy of these expanded string values (including NIL_String)
 *                         followed by flat list <tailList>
 *          this is static 
 *              returns in <argumentList_res> a flat copy of these string values (including NIL_String)
 *                         followed by flat list <tailList>
 * failure: this is a string or in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_flatten_listCast_expand_ (ref2ArgumentList tailList, ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listCast_expand_");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (!string_call ())
	    return pk_false;
	  (*argumentList_res) = new ArgumentList (this_expandedArgument, tailList);
	  return pk_true;
	}
      else
	{			/* has list type */
	  if (!list_call ())
	    return pk_false;
	  if (this_expandedArgument == NIL_Argument)
	    {
	      (*argumentList_res) = tailList;
	      return pk_true;
	    }
	  return this_expandedArgument->getIf_flatten_listCast_expand_ (tailList, argumentList_res);
	}
    }
  else
    unsafe_use ("getIf_flatten_listCast_expand_(ref2ArgumentList, ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is a list
 *    this is a flat constant 
 *        returns this expanded value in <argumentList_res>
 *    this is not a flat constant list
 *        returns a new m'allocated expansion of this in <argumentList_res>
 * failure: this is a string or in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_flatten_listExpand_ (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listExpand_(ref2ArgumentList *) to", this_functionCall->get_name());
  if (IS_SAFE () && HAS_LIST_TYPE ())
    {
      if (!list_call ())
	return pk_false;
      if (this_expandedArgument == NIL_Argument)
	(*argumentList_res) = NIL_ArgumentList;
      else
	(*argumentList_res) = ((ref2ArgumentList) this_expandedArgument)->get_flatten_expandedList ();
      return pk_true;
    }
  else
    unsafe_use ("getIf_flatten_listExpand_(ref2ArgumentList *)");
  return pk_false;
}

/* 2004-12-13
 * success: this has string type 
 *             sets this_CString_expandedValue equal to this expansion
 *             returns this expanded value
 *          this has list type 
 *             returns a new StringArgument equal to the concatenation of these expanded elements
 * failure: this is errored or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_CString_cast_expand_ (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expand_ this_functionCall is", this_functionCall->myClassName());
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (string_call ())
	    {
	      (*CString_res) = this_CString_expandedValue;
	      return pk_true;
	    }
	  return pk_false;
	}
      if (list_call ())
	{
	  ref2ArgumentList arg_res = (ref2ArgumentList) this_expandedArgument;
	  if (arg_res == NIL_ArgumentList)
	    this_CString_expandedValue = NIL_String;
	  else if (!arg_res->getIf_CString_cast_expanded (&this_CString_expandedValue))
	    return pk_false;
	  (*CString_res) = this_CString_expandedValue;
	  return pk_true;
	}
      return pk_false;
    }
  else
    unsafe_use ("getIf_CString_cast_expand_(char **)");
  return pk_false;
}

/* 2005-03-11
 * success: returns in <CString_res> a singleton CString list which single value is this expanded value list
 * failure: this is in error or this linked identifier returns NIL_Argument (is assumed to be in error)
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList tailList, ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flatExpand_() 1");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (string_call ())
	    {
	      (*strList_res) = new_CStringList (tailList);
	      return pk_true;
	    }
	  return pk_false;
	}
      if (!list_call ())
	return pk_false;
      if (this_expandedArgument == NIL_Argument)
	{
	  (*strList_res) = tailList;
	  return pk_true;
	}
      return this_expandedArgument->getIf_CStringList_cast_flatExpand_ (tailList, strList_res);
    }
  else
    unsafe_use ("getIf_CStringList_cast_flatExpand_(ref2CStringList, ref2CStringList *)");
  return pk_false;
}

/* 2005-03-11
 * success: returns in <strList_res> a singleton CString list which single value is this expanded value list
 * failure: this is in error or this linked identifier returns NIL_Argument (is assumed to be in error)
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flatExpand_() 2");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (string_call ())
	    {
	      (*strList_res) = new_CStringList ();
	      return pk_true;
	    }
	  return pk_false;
	}
      else
	{			/* has list type */
	  ref2ArgumentList argumentList_res;
	  if (!list_call ())
	    return pk_false;
	  argumentList_res = (ref2ArgumentList) this_expandedArgument;
	  if (argumentList_res != NIL_ArgumentList)
	    return argumentList_res->getIf_CStringList_cast_flat_expanded (strList_res);
	}
      this_CString_expandedValue = NIL_String;
      (*strList_res) = NIL_CStringList;
    }
  else
    unsafe_use ("getIf_CStringList_cast_flatExpand_(ref2CStringList *)");
  return pk_false;
}

/* 2005-03-11
 * success: returns in <lineList_res> a singleton Line list which single value is this expanded value list
 * failure: this is in error or this linked identifier returns NIL_Argument (is assumed to be in error)
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_lineList_cast_flatExpand_ (ref2LineList tailList, ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_ 2(ref2LineList , ref2LineList *)");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (string_call ())
	    {
	      (*lineList_res) = new_LineList (tailList);
	      return pk_true;
	    }
	  return pk_false;
	}
      else
	{			/* has list type */
	  if (!list_call ())
	    return pk_false;
	  if (this_expandedArgument == NIL_Argument)
	    {
	      (*lineList_res) = tailList;
	      return pk_true;
	    }
	  return ((ref2ArgumentList) this_expandedArgument)->getIf_lineList_cast_flatExpand_ (tailList, lineList_res);
	}
    }
  else
    unsafe_use ("getIf_lineList_cast_flatExpand_(ref2LineList, ref2LineList *)");
  return pk_false;
}

/* 2005-03-11
 * success: returns in <lineList_res> the flat expansion of this actual value
 * failure: this is in error or this linked identifier returns NIL_Argument (is assumed to be in error)
 *          error message is printed
 */
_VIRTUAL_ boolean
CommandArgument::getIf_lineList_cast_flatExpand_ (ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_ 1(ref2LineList *)");
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (string_call ())
	    {
	      (*lineList_res) = new_LineList ();
	      return pk_true;
	    }
	  return pk_false;
	}
      else
	{			/* has list type */
	  if (!list_call ())
	    return pk_false;
	  if (this_expandedArgument == NIL_Argument)
	    {
	      (*lineList_res) = NIL_LineList;
	      return pk_true;
	    }
	  return ((ref2ArgumentList) this_expandedArgument)->getIf_lineList_cast_flat_expanded (lineList_res);
	}
    }
  else
    unsafe_use ("getIf_lineList_cast_flatExpand_(ref2LineList *)");
  return pk_false;
}


/* ========== Builder
 */

/* <functionCall> is not NIL (not checked)
 * creates a list <functionCall> argument or an iterated string <functionCall>
 */

_PUBLIC_ CommandArgument::CommandArgument (boolean isList, ref2ArgumentFunctionCall functionCall, boolean isIterated):Argument ()
{
  // logDEBUG_in("CommandArgument(ref2ArgumentFunctionCall)");
  this_expandedArgument = NIL_Argument;
  this_functionCall = functionCall;
  if (isIterated)
    {
      SET_LIST_TYPE ();
      SET_ITERATED ();
    }
  else
    {
      SET_SCOPE (this_functionCall->get_scope ());
      if (isList)
	{
	  SET_LIST_TYPE ();
	  if (IS_SAFE_STATIC () && (!list_call ()))
	    set_error ();
	}
      else
	{
	  SET_STRING_TYPE ();
	  if (IS_SAFE_STATIC () && (!string_call ()))
	    set_error ();
	}
    }
}
