_PUBLIC_ _VIRTUAL_ const char *
DynamicContext::myClassName ()
{
  return "DynamicContext";
}

/* ==========  ==========  ==========  ========== TITLE: these attributes
 */
/* returns formal parameter list
 */
ref2IdentifierList
DynamicContext::get_identifierList ()
{
  return this_localIdentifierList;
}

_VIRTUAL_ ref2IdentifierList
DynamicContext::get_parameterList ()
{
  return this_localIdentifierList;
}

/* ==========  ==========  ==========  ========== TITLE: DEBUG
 */

void
DynamicContext::logDebug_identifierList_ln (const char *mess)
{
  if (this_localIdentifierList == NIL_IdentifierList)
    logDEBUG_emptyFormula_ln (mess, "identifier list");
  else
    {
      LOG_STR_LN (mess);
      this_localIdentifierList->logDEBUG_formula_ln (this, 0);
    }
}

_VIRTUAL_ void
DynamicContext::logDebug (int level)
{
  if (this_localIdentifierList == NIL_IdentifierList)
    logDEBUG_emptyFormula_ln ("in dynamic context", "identifier list");
  else
    {
      LOG_STR_LN ("in dynamic context");
      this_localIdentifierList->logDEBUG_formula_ln (this, level + 1);
    }
}

/* ==========  ==========  ==========  ========== TITLE: PRINTING
 */

/* prints object standard header: "<NAME> ( PARAMETER list )" 
 */
_VIRTUAL_ void
DynamicContext::print_funcProfile_ln (ref2Lexical printingContext, ref2IntePrinter outBuffer, const char *funcName)
{
  // fprintf(stderr, "DEBUG DynamicContext::print_funcProfile_ln, context adr is %i\n", (int)this);
  outBuffer->put_nameValue_openTag_sp (funcName);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      this_localIdentifierList->print_typedFormula (printingContext, outBuffer);
      outBuffer->space ();
    }
  outBuffer->put_closeTag_ln ();
}

_PUBLIC_ _VIRTUAL_ void
DynamicContext::print (ref2IntePrinter outBuffer)
{
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      outBuffer->put_openTag_sp ();
      this_localIdentifierList->print_typedFormula (this_lexicalContext, outBuffer);
      outBuffer->put_closeTag_ln ();
    }
}

/* ==========  ==========  ==========  ========== TITLE: FATAL
 */
/* this is set to fatal or error as indicated by this lexical context
 */
_PROTECTED_ void
DynamicContext::error_typeIdentifier_ (const char *CString_name)
{
  LOG_LN_ERRO_STR_SP (CString_name);
  LOG_STR_LN (": unexpected type");
  set_error ();
}

/* ==========  ==========  ==========  ========== TITLE: SPECIFIC ACTIONS
 */

/* ========== Declaration linking
*/
_PUBLIC_ _VIRTUAL_ boolean DynamicContext::parameterListLink_listExpand_ (ref2ArgumentList parameterValues)
{
  // logDEBUG_in("parameterListLink_listExpand_");
  if (this_localIdentifierList == NIL_IdentifierList)
    {
      if (parameterValues == NIL_ArgumentList)
	return pk_true;
      OUPS ("too much actual parameters in DynamicContext::parameterListLink_listExpand_");
    }
  return this_localIdentifierList->listLink_listExpand_ (parameterValues);
}

/*
 * this_nestingContext not NIL (not checked)
 * success: <name> is a parameter list identifier
 *              a reference to this identifier is assigned to a newly created <iterator>
 *          <name> is a static list identifier
 *              a new iterator is assigned to <iterator>
 *              its value is the expanded value of the static identifier
 * failure: 
 *     <name> is found, but is not a node function in this_MKFFileContext context
 *          or a cyclic search has been met
 *          error message is printed
 *          <isFatal> is set to pk_true
 *     <name> is not found
 *          <isFatal> remains unchanged
 */
_VIRTUAL_ boolean
DynamicContext::build_iteratorRefTo_ (char *name, ref2Iterator * iterator, boolean * isFatal)
{
  // fprintf(stderr, "DEBUG DynamicContext::build_iteratorRefTo_ name is %s\n", name);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier anIdentifier;
      if (this_localIdentifierList->find (name, &anIdentifier))
	{
	  if (anIdentifier->has_list_type ())
	    {
	      (*iterator) = new Iterator (name, anIdentifier);
	      return pk_true;
	    }
	  fatal_named_typeConflict_ln (name, "string list");
	  (*isFatal) = pk_true;
	  return pk_false;
	}
    }
  return this_nestingContext->build_iteratorRefTo_ (name, iterator, isFatal);
}

/* PLG 2007-02-05
 * this_nestingContext not NIL (not checked)
 * search for a string identifier declaration named <CString_name>
 * required: not empty <CString_name>
 * success: a local or global identifier with expected type is found
 *          returns its actual value in <stringArgument_res>
 *          <correct> remains unchanged
 * failure: <CString_name> has not been found in this context and in its embedding context
 *             <correct> remains unchanged
 *          <CString_name> has been found but has unexpected type, or a cyclic search has been met
 *             (fatal) error message is printed
 *             <correct> is set to false
 *          <stringArgument_res> remains unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean DynamicContext::getIf_expandedString_idVal_ (char *CString_name, boolean * correct, ref2StringArgument * stringArgument_res)
{
  /* search in local declarations */
  // fprintf(stderr, "DEBUG DynamicContext::getIf_expandedString_idVal_ name is %s, this_contextTree adr is %i\n", CString_name, (int)this_contextTree);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (CString_name, &id_res))
	{
	  boolean
	    isFatal;
	  if (id_res->get_expandedString (stringArgument_res, &isFatal))
	    return pk_true;
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  /* search in context declarations */
  return this_nestingContext->getIf_expandedString_idVal_ (CString_name, correct, stringArgument_res);
}

/* PLG 2007-02-05
 * this_nestingContext not NIL (not checked)
 * search for a list identifier declaration named <CString_name>
 * required: not empty <CString_name>
 * success: a local or global identifier with expected type is found
 *          returns its actual value in <argumentList_res>
 *             <correct> remains unchanged
 * failure: <CString_name> has not been found in this context
 *             <found> is set to false
 *          <CString_name> has been found but has unexpected type
 *             (fatal)error message is printed
 *             <correct> is set to false
 *             <isFatal> is set to pk_true
 *          <CString_name> has been found with expected type, expansion failed
 *             error message is printed
 *             <correct> is set to false
 *             <isFatal> is set to pk_true
 *          <argumentList_res> remains unchanged
 */
_VIRTUAL_ boolean
DynamicContext::getIf_expandedList_idVal_ (char *CString_name, boolean * correct, ref2ArgumentList * argumentList_res)
{
  /* search in local declarations */
  // fprintf(stderr, "DEBUG DynamicContext::getIf_expandedList_idVal_, name is %s, this_contextTree adr is %i\n", CString_name, (int)this_contextTree);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier id_res;
      if (this_localIdentifierList->find (CString_name, &id_res))
	{
	  boolean isFatal;
	  if (id_res->get_expandedList (argumentList_res, &isFatal))
	    return pk_true;
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  /* search in context declarations */
  return this_nestingContext->getIf_expandedList_idVal_ (CString_name, correct, argumentList_res);
}

/* 2005-01-17
 * this_nestingContext not NIL (not checked)
 * <CString_name> is a (not empty) identifier (not checked)
 * success: <CString_name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <CString_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) StringArgument value of this static identifier; 
 *              <argument_res> scope is static
 *          <argument_res> type is string
 *          <correct> remains unchanged
 * failure: <CString_name> declaration has not been (recursively) found
 *              <correct> remains unchanged
 *          <CString_name> declaration has unexpected type
 *              <correct> is set to false
 *              error message is printed
 *          <argument_res> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean DynamicContext::getIf_string_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_string_id_, context adr is %i, ident is %s\n", (int)this, CString_name);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (CString_name, &id_res))
	{
	  if (id_res->has_string_type ())
	    {
	      (*argument_res) = new NamedArgument (CString_name, id_res);
	      return pk_true;
	    }
	  error_typeIdentifier_ (CString_name);
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_string_id_ (CString_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean DynamicContext::getIf_string_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_string_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  return this_nestingContext->getIf_string_id_ (id_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean DynamicContext::getIf_string_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_string_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  return this_nestingContext->getIf_string_id_ (id_name, correct, argument_res);
}

/* 2005-12-14
 * this_nestingContext not NIL (not checked)
 * <CString_name> is a (not empty) identifier (not checked)
 * success: <CString_name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <CString_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) ArgumentList value of this static identifier; 
 *              <argument_res> scope is static
 *          <correct> remains unchanged
 *          <argument_res> type is list
 * failure: <CString_name> declaration is not (recursively) found
 *              <correct> remains unchanged
 *          <CString_name> declaration has unexpected type
 *              error message is printed
 *             <correct> is set to false
 *          <argument_res> remains unchanged
 */
_VIRTUAL_ boolean
DynamicContext::getIf_list_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_list_id_, context adr is %i, ident is %s\n", (int)this, CString_name);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier id_res;
      if (this_localIdentifierList->find (CString_name, &id_res))
	{
	  if (id_res->has_list_type ())
	    {
	      (*argument_res) = new NamedArgument (CString_name, id_res, pk_false);
	      return pk_true;
	    }
	  error_typeIdentifier_ (CString_name);
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_list_id_ (CString_name, correct, argument_res);
}

_VIRTUAL_ boolean
DynamicContext::getIf_list_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_list_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  return this_nestingContext->getIf_list_id_ (id_name, correct, argument_res);
}

_VIRTUAL_ boolean
DynamicContext::getIf_list_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_list_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  return this_nestingContext->getIf_list_id_ (id_name, correct, argument_res);
}

/* 2005-12-14
 * this_nestingContext not NIL (not checked)
 * <CString_name> is a (not empty) identifier (not checked)
 * success: <CString_name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <CString_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) ArgumentList value of this static identifier; 
 *              <argument_res> scope is static
 *          <correct> remains unchanged
 * failure: <CString_name> declaration is not (recursively) found
 *              <correct> remains unchanged
 *          <CString_name> declaration has unexpected type
 *              error message is printed
 *             <correct> is set to false
 *          <argument_res> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean DynamicContext::getIf_argument_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG DynamicContext::getIf_argument_id_ name is %s\n", CString_name);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (CString_name, &id_res))
	{
	  if (id_res->has_list_type ())
	    {
	      (*argument_res) = new NamedArgument (CString_name, id_res, pk_false);
	      return pk_true;
	    }
	  if (id_res->has_string_type ())
	    {
	      (*argument_res) = new NamedArgument (CString_name, id_res);
	      return pk_true;
	    }
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_argument_id_ (CString_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean DynamicContext::getIf_argument_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  return this_nestingContext->getIf_argument_id_ (id_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean DynamicContext::getIf_argument_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  return this_nestingContext->getIf_argument_id_ (id_name, correct, argument_res);
}

_PUBLIC_ boolean DynamicContext::append_uniqueNewIdentifier (char *identName, boolean isAList, ref2Identifier * createdIdent)
{
  // log_ASDEBUG("DynamicContext::append_uniqueNewIdentifier");
  if (this_localIdentifierList == NIL_IdentifierList)
    {
      // fprintf(stderr, "DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList is empty\n");
      (*createdIdent) = new Identifier (identName, isAList);
      this_localIdentifierList = new IdentifierList (*createdIdent);
      // logDebug_identifierList_ln("DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  // logDebug_identifierList_ln("DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList is ----------------");
  if (this_localIdentifierList->append_uniqueNewIdentifier (identName, isAList, createdIdent))
    {
      // logDebug_identifierList_ln("DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  ASFatal_doubleDecl (identName);
  return pk_false;
}

_PUBLIC_ boolean DynamicContext::append_uniqueNewIdentifier (char *identName, boolean isAList, ref2Argument formula, ref2Identifier * createdIdent)
{
  // fprintf(stderr, "DEBUG DynamicContext::append_uniqueNewIdentifier, context adr is %i, ident is %s\n", (int)this, identName);
  // log_ASDEBUG("DynamicContext::append_uniqueNewIdentifier");
  if (this_localIdentifierList == NIL_IdentifierList)
    {
      // fprintf(stderr, "DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList is empty\n");
      (*createdIdent) = new Identifier (identName, isAList, formula);
      this_localIdentifierList = new IdentifierList (*createdIdent);
      // logDebug_identifierList_ln("DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  // logDebug_identifierList_ln("DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList is ----------------");
  if (this_localIdentifierList->append_uniqueNewIdentifier (identName, isAList, formula, createdIdent))
    {
      // logDebug_identifierList_ln("DEBUG DynamicContext::append_uniqueNewIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  ASFatal_doubleDecl (identName);
  return pk_false;
}

/* ==========  ==========  ==========  ========== TITLE: READING
 */

/* PLG 2007-02-08 readIf_formalIdentifier_fSpStar_nSp:
 * reads an identifier
 * success: FILE = PREFIX . opt{'*'[fileSpace*]}[Identifier][fileSpace+] # [Character_but_Space][Any*][EOF]
 *          <correct> remains unchanged
 * failure: FILE = PREFIX .# [Any_but_IdentCharacter][Any*][EOF]
 *             <correct> remains unchanged
 *          FILE = PREFIX . '*'[fileSpace*] # [Any_but_IdentCharacter][Any*][EOF]
 *            or   PREFIX . opt{'*'[fileSpace*]}[Identifier][fileSpace*] # [EOF]
 *            or   PREFIX . opt{'*'[fileSpace*]}[Identifier][fileSpace+] # [Any*][EOF]
 *            or   PREFIX . opt{'*'[fileSpace*]} ... # [Any*][EOF], reading failed
 *          a (fatal) syntax error occurred
 *             error message is printed
 *             <correct> is set to false
 *          <curIdent> remains unchanged
 */
_PRIVATE_ boolean DynamicContext::readIf_formalIdentifier_fSpStar_nSp (boolean * correct)
{
  char *
    CString_name;
  ref2Identifier
    curIdent;
  boolean
    isSpaced,
    isAList;
  boolean
    lCorrect = pk_true;
  if (this_lexicalContext->stepIf_listMark_fSpStar_nSp (&lCorrect))
    isAList = pk_true;
  else if (lCorrect)
    isAList = pk_false;
  else
    {
      (*correct) = pk_false;
      return pk_false;
    }
  /* FILE = PREFIX . opt{'*'[fileSpace*]} # [Character_but_Space][Any*][EOF] */
  if (this_lexicalContext->stepIf_userIdentifier_fSpStar_nSp (correct, &CString_name, &isSpaced))
    {
      /* success: [Identifier] is not a reserved word
       *          FILE = PREFIX . opt{'*'[fileSpace*]}[Identifier] # [Character_but_Space_but_IdentCharacter][Any*][EOF]
       *                <isSpaced> is set FALSE
       *          FILE = PREFIX . opt{'*'[fileSpace*]}[Identifier][fileSpace+] # [Character_but_Space][Any*][EOF]
       *                <isSpaced> is set TRUE
       *          <CString_name> is loaded with a new m'allocated copy of [Identifier]
       *          <correct> remains unchanged */
      if (append_uniqueNewIdentifier (CString_name, isAList, &curIdent))
	return pk_true;
      (*correct) = pk_false;
    }
  /* FILE = PREFIX . opt{'*'[fileSpace*]} # [Any_but_IdentCharacter][Any*][EOF]
   *                 <correct> remains unchanged
   * FILE = PREFIX . opt{'*'[fileSpace*]}[Identifier][fileSpace*] # [EOF]
   *                <CString_name> is loaded with a new m'allocated copy of [Identifier]
   *                 error message is printed
   *                 <correct> is set FALSE
   * FILE = PREFIX . opt{'*'[fileSpace*]}[Identifier][fileSpace+] # [Any*][EOF]
   *                 [Identifier] is a reserved word
   *                 error message is printed
   *                 <correct> is set FALSE
   * FILE = PREFIX . opt{'*'[fileSpace*]} ... # [Any*][EOF], reading failed
   *                 <CString_name>  remains unchanged
   *                 error message is printed
   *                 <correct> is set FALSE */
  if (isAList)
    {
      /* FILE = PREFIX . '*'[fileSpace*] # [Any_but_IdentCharacter][Any*][EOF] */
      ASFatal_expected ("Identifier");
      (*correct) = pk_false;
    }
  return pk_false;
}

/* 2005-01-10
 * reads a possibly empty identifier list 
 * is assumed to be UNKNOWN (not checked)
 * success: cursor starts  ')'
 *          cursor starts [TypedIdentifier_fSp+*] ')' 
 *          cursor is moved to ')'
 * failure: syntax error
 *          error message is printed
 */
_VIRTUAL_ boolean
DynamicContext::read_local_identifierList_fSpStar_nSp ()
{
  // log_ASDEBUG("entering DynamicContext::read_local_identifierList_fSpStar_nSp");
  boolean lCorrect = pk_true;
  this_localIdentifierList = NIL_IdentifierList;
  do
    {
      if (!readIf_formalIdentifier_fSpStar_nSp (&lCorrect))
	{
	  return lCorrect;
	}
    }
  while (pk_true);
}

/* 2005-01-12
 * reads an identifier list 
 * is assumed to be UNKNOWN (not checked)
 * success: cursor starts '(' [fileSpace*] list_fSp{[VAR_Identifier]} ')' [fileSpace*] [FUNCTION_BODY]
 *          or cursor starts '(' [fileSpace*] [TypedIdentifier_fSp+*] ')' [fileSpace*] [FUNCTION_BODY]
 *          cursor is moved to first [Character] in [FUNCTION_BODY]
 * failure: a syntax error occurred
 *          error message is printed
 */
_PUBLIC_ boolean DynamicContext::read_dynamicContext_fSpStar_nSp ()
{
  // log_ASDEBUG("entering DynamicContext::read_dynamicContext_fSpStar_nSp()");
  boolean
    isEof = pk_false;
  if (this_lexicalContext->req_listO_fSpStar_nSp ())
    {
      /* FILE = PREFIX . '(' [fileSpace*] # [Character_but_Space][Any*][EOF] */
      if (read_local_identifierList_fSpStar_nSp ())
	{
	  if (this_lexicalContext->req_listC_fSpStar_ (&isEof))
	    {			/* success: FILE = PREFIX . XXX ')'[fileSpace*] # opt{[Character_but_Space][Any*]}[EOF], '}' is not followed by ':'
				 * failure: WELLBALANCED is [Any_but_'}:'_but_'}']* in which '(', ')' are well balanced
				 *          FILE = PREFIX . XXX ... [EOF], reading failed
				 *          FILE = PREFIX . XXX WELLBALANCED # [EOF]
				 *          FILE = PREFIX . XXX WELLBALANCED '('[Any_'}:'_but_'}']* # [EOF], 
				 *             [Any_'}:'_but_'}']* does not match <<WELLBALANCED')'[Any*]>>
				 *          FILE = PREFIX . XXX WELLBALANCED ')'[fileSpace*] # opt{[Character_but_Space][Any*]}[EOF]
				 *          FILE = PREFIX . XXX [Any_but_'}:'_but_'}']*  # '}'[Any*][EOF]
				 *             [Any_but_'}:'_but_'}']* does not match <<WELLBALANCED')'[Any*]>> 
				 *          error message is printed
				 * <isEof> is set according to # position */
	      return (!isEof);
	    }
	  return pk_false;
	}
      this_lexicalContext->recover_second_fSpStar_ (&isEof);
      /* WELLBALANCED is [Any_but_'}:'_but_'}']* that is well balanced wrt to '(', ')'
       * FILE = PREFIX . XXX ... # [EOF], reading failed
       * FILE = PREFIX . XXX WELLBALANCED opt{'('[Any_'}:'_but_'}']*} # [EOF], 
       *              [Any_'}:'_but_'}']* does not match <<WELLBALANCED')'[Any*]>>
       * 
       * FILE = PREFIX . XXX [Any_but_'}:'_but_'}']*  # '}:'[Any*][EOF]
       *   or   PREFIX . XXX [Any_but_'}:'_but_'}']*  # '}'[Any*][EOF]
       * FILE = PREFIX . XXX WELLBALANCED ')'[fileSpace*] # opt{[Character_but_Space][Any*]}[EOF]
       *             [Any_but_'}:'_but_'}']* does not match <<WELLBALANCED')'[Any*]>> 
       * error message is printed
       * <isEof> is set according to # position
       */
      return pk_false;
    }
  /*  FILE = PREFIX . ... # [EOF], reading failed
   *   or   PREFIX . '(' [fileSpace*] # [EOF]
   * FILE = PREFIX .# [Character_but_Space_but_'('][Any*][EOF] 
   * error message is printed */
  return pk_false;
}

/* ============================================================ node GenericIdentification
 */
/* 2004-11-22
 * reads a generic extern identification
 * success: cursor starts [EXTERN_IDENTIFICATION] [lineSpaces]
 *            where
 *            [EXTERN_IDENTIFICATION] ::= '~' [Character+] '~'
 *            		'~' does not occur in [Character+] 
 *            		<commandChar> does not start [Character+]
 *            [EXTERN_IDENTIFICATION] ::= '~' [parameterized_string] '~'
 *            		'~' does not occur in [parameterized_string]
 *            		<commandChar> does not start [Character*]
 *            [EXTERN_IDENTIFICATION] ::= '~' [StringFunction_CALL] [lineSpaces]'~'
 *            		'~' does not occur in [StringFunction_CALL]
 *          cursor is moved to terminal position of  [lineSpaces]
 *          returns the node name in <argument_res> (new m'allocated)
 * failure: cursor starts a not correct [EXTERN_IDENTIFICATION]
 *              error message is printed
 *          current character is not '~' 
 * PLG THIS MAY NOT BE FREE
 */
_PRIVATE_ boolean DynamicContext::readIf_externIdentification_spStar_ (boolean * correct, ref2Argument * argument_res, boolean * isEol)
{
  // log_ASDEBUG("DynamicContext::readIf_externIdentification_spStar_");
  boolean
    lCorrect = pk_true;
  if (readIf_extIdentCommand_spStar_ (&lCorrect, argument_res, isEol))
    return pk_true;
  if (lCorrect)
    {
      char *
	externName;
      if (this_lexicalContext->stepIf_externIdentifier_spStar_ (&lCorrect, &externName, isEol))
	{
	  /* FILE = PREFIX . '~'[Character*]'~'opt{[lineSpaces]} # opt{[eol][Any*]}[EOF] 
	   *          FILE = PREFIX . '~'[Character*]'~'[White*]#[Character_but_Space][Character+] [NULL_char] 
	   *          a new fresh copy of [Character*] is returned in <CString_name>
	   *          <lCorrect> remains unchanged, <isEol> is set  */
	  (*argument_res) = new StringArgument (externName);
	  return pk_true;
	}
      if (lCorrect)
	/* PREFIX .# opt{[Character_but_'~'][Character*]}opt{[eol][Any*]}[EOF], isEol> is set */
	return pk_false;
      /* FILE = PREFIX . '~''~' opt{[lineSpaces]} # opt{[eol][Any*]}[EOF]
       *   or   PREFIX . '~''~'[White*] # [Character_but_Space][Any+][EOF]
       * some memory error occurred
       *               FILE = PREFIX . '~'[Character*]'~' opt{[lineSpaces]} # opt{[eol][Any*]}[EOF]
       *                 or   PREFIX . '~'[Character*]'~'[White*] # [Character_but_Space][Character*] opt{[eol][Any*]}[EOF]
       * FILE = PREFIX . '~' # [Character*]opt{[eol][Any*]}[EOF], '~' does not occur in [Character*]
       * error message is printed, <isEol> is set */
      (*correct) = pk_false;
      return pk_false;
    }
  (*correct) = pk_false;
  return pk_false;
}

/* 2005-02-28
 * reads a generic file identification
 * success: cursor starts [FILE_IDENTIFICATION] [lineSpaces]
 *            where
 *            [FILE_IDENTIFICATION] ::= '!' [Character+] '!'
 *            		'!' does not occur in [Character+] 
 *            		<commandChar> does not start [Character+]
 *            [FILE_IDENTIFICATION] ::= '!' [parameterized_string] '!'
 *            		'!' does not occur in [parameterized_string]
 *            		<commandChar> does not start [Character*]
 *            [FILE_IDENTIFICATION] ::= '!' [StringFunction_CALL] [lineSpaces]'!'
 *            		'!' does not occur in [StringFunction_CALL]
 *          cursor is moved to final character in [lineSpaces]
 *          <correct> remains unchanged
 *          returns in <argument_res> a new m'allocated copy of [FILE_IDENTIFICATION]
 * failure: cursor starts a not correct [FILE_IDENTIFICATION]
 *              <correct> is set to false
 *              error message is printed
 *          current character is not '!' 
 *              <correct> remains unchanged
 */
_PRIVATE_ boolean DynamicContext::readIf_fileIdentification_spStar_ (boolean * correct, ref2Argument * argument_res, boolean * isEol)
{
  // fprintf(stderr, "DEBUG DynamicContext::readIf_fileIdentification_spStar_\n");
  boolean
    lCorrect = pk_true;
  if (readIf_fileIdentCommand_spStar_ (&lCorrect, argument_res, isEol))
    return pk_true;
  if (lCorrect)
    {
      char *
	fileName;
      if (this_lexicalContext->stepIf_fileIdentifier_spStar_ (&lCorrect, &fileName, isEol))
	{
	  /* FILE = PREFIX . '!'[Character*]'!'opt{[lineSpaces]} # opt{[eol][Any*]}[EOF] 
	   *          FILE = PREFIX . '!'[Character*]'!'[White*]#[Character_but_Space][Character+] [NULL_char] 
	   *          a new fresh copy of [Character*] is returned in <CString_name>
	   *          <lCorrect> remains unchanged, <isEol> is set  */
	  (*argument_res) = new StringArgument (fileName);
	  return pk_true;
	}
      if (lCorrect)
	/* PREFIX .# opt{[Character_but_'!'][Character*]}opt{[eol][Any*]}[EOF], <isEol> is set */
	return pk_false;
      /* FILE = PREFIX . '!''!' opt{[lineSpaces]} # opt{[eol][Any*]}[EOF]
       *   or   PREFIX . '!''!'[White*] # [Character_but_Space][Any+][EOF]
       * some memory error occurred
       *               FILE = PREFIX . '!'[Character*]'!' opt{[lineSpaces]} # opt{[eol][Any*]}[EOF]
       *                 or   PREFIX . '!'[Character*]'!'[White*] # [Character_but_Space][Character*] opt{[eol][Any*]}[EOF]
       * FILE = PREFIX . '!' # [Character*]opt{[eol][Any*]}[EOF], '!' does not occur in [Character*]
       * error message is printed, <isEol> is set */
      (*correct) = pk_false;
      return pk_false;
    }
  (*correct) = pk_false;
  return pk_false;
}

/* 2004-12-24
 * reads a generic node name beginning at current position
 * success: cursor starts [NODE_IDENTIFICATION] [lineSpaces]
 *          where
 *            [NODE_IDENTIFICATION] ::= [VAR_IDENTIFICATION]
 *               [VAR_IDENTIFICATION] ::= [Identifier]
 *               [VAR_IDENTIFICATION] ::= [Parameterized_ident] 
 *               [VAR_IDENTIFICATION] ::= [referenceTAG] [StringFunction_CALL]
 *               set <isVar> to pk_true
 *            [NODE_IDENTIFICATION] ::= [FILE_IDENTIFICATION]
 *               [FILE_IDENTIFICATION] ::= '!' [Character+] '!' 
 *               [FILE_IDENTIFICATION] ::= '!' [Parameterized_string] '!'
 *               [FILE_IDENTIFICATION] ::= '!' [StringFunction_CALL]
 *               set <isVar> to pk_false, <isFile> to pk_true 
 *            [NODE_IDENTIFICATION] ::= [EXTERN_Identifier]
 *               [EXTERN_IDENTIFICATION] ::= '~' [Character+] '~' 
 *               [EXTERN_IDENTIFICATION] ::= '~' [Parameterized_string] '~' 
 *               [EXTERN_IDENTIFICATION] ::= '~' [StringFunction_CALL] '~' 
 *               set <isVar> to pk_false, <isFile> to false 
 *          cursor is moved to end of [lineSpaces]
 *          returns the node name generator in <nameArgument>
 * failure: on syntax error log message is printed in the_logBuffer
 *          cursor is moved to errored position
 *          <nameArgument> remains unchanged
 */
_PROTECTED_ boolean DynamicContext::read_nodeIdentification_spStar_ (ref2Argument * nameArgument, boolean * isVar, boolean * isFile, boolean * isEol)
{
  // log_ASDEBUG("DEBUG DynamicContext::read_nodeIdentification_spStar_");
  boolean
    lCorrect = pk_true;
  if (readIf_externIdentification_spStar_ (&lCorrect, nameArgument, isEol))
    {
      (*isVar) = pk_false;
      (*isFile) = pk_false;
      return pk_true;
    }
  if (lCorrect)
    {
      if (readIf_fileIdentification_spStar_ (&lCorrect, nameArgument, isEol))
	{
	  (*isVar) = pk_false;
	  (*isFile) = pk_true;
	  return pk_true;
	}
      if (lCorrect)
	{
	  if (readIf_varIdentCommand_spStar_ (&lCorrect, nameArgument, isEol))
	    {
	      (*isVar) = pk_true;
	      return pk_true;
	    }
	  if (lCorrect)
	    {
	      char *
		CString_namePrefix;
	      boolean
		isSpaced;
	      if (this_lexicalContext->stepIf_userIdentifier_spStar_ (&lCorrect, &CString_namePrefix, &isSpaced, isEol))
		{
		  ref2ArgumentList
		    tailList;
		  if (isSpaced)
		    {
		      (*nameArgument) = new StringArgument (CString_namePrefix);
		      (*isVar) = pk_true;
		      return pk_true;
		    }
		  if (stepIf_referenceHeadedIdentifier_spStar_ (&lCorrect, &tailList, isEol))
		    {
		      (*nameArgument) = new MetaArgument (new ArgumentList (CString_namePrefix, tailList));
		      (*isVar) = pk_true;
		      return pk_true;
		    }
		  if (lCorrect)
		    {
		      (*nameArgument) = new StringArgument (CString_namePrefix);
		      (*isVar) = pk_true;
		      return pk_true;
		    }
		  return pk_false;
		}
	      /* <isEol> is set according to final cursor position */
	      else if (lCorrect)
		{
		  ref2ArgumentList
		    listRes;
		  if (stepIf_referenceHeadedIdentifier_spStar_ (&lCorrect, &listRes, isEol))
		    {
		      (*nameArgument) = new MetaArgument (listRes);
		      (*isVar) = pk_true;
		      return pk_true;
		    }
		  if (lCorrect)
		    ASFatal_unexpected  (pk_true, "node reference");
		}
	    }
	}
    }
  return pk_false;
}

/* 2004-12-31
 * reads a generic node name beginning at current position
 * success: cursor starts [NODE_IDENTIFICATION] [lineSpaces]
 *          where
 *            [NODE_IDENTIFICATION] ::= [VAR_IDENTIFICATION]
 *               [VAR_IDENTIFICATION] ::= [Identifier]
 *               [VAR_IDENTIFICATION] ::= [Parameterized_ident] 
 *               [VAR_IDENTIFICATION] ::= [referenceTAG] [StringFunction_CALL]
 *               set <isVar> to pk_true
 *            [NODE_IDENTIFICATION] ::= [FILE_IDENTIFICATION]
 *               [FILE_IDENTIFICATION] ::= '!' [Character+] '!' 
 *               [FILE_IDENTIFICATION] ::= '!' [Parameterized_string] '!'
 *               [FILE_IDENTIFICATION] ::= '!' [StringFunction_CALL]
 *               set <isVar> to pk_false, <isFile> to pk_true 
 *            [NODE_IDENTIFICATION] ::= [EXTERN_Identifier]
 *               [EXTERN_Identifier] ::= '~' [Character+] '~' 
 *               set <isVar> to pk_false, <isFile> to false 
 *          cursor is moved to end of [lineSpaces]
 *          returns the node name generator in <nameArgument>
 * failure: on syntax error log message is printed in the_logBuffer
 *          cursor is moved to errored position
 *          <nameArgument> remains unchanged
 */
boolean
DynamicContext::readIf_nodeIdentification_fSpStar_nSp (boolean * correct, ref2Argument * nameArgument, boolean * isVar, boolean * isFile)
{
  // fprintf(stderr, "DEBUG DynamicContext::readIf_nodeIdentification_fSpStar_nSp, this is %i\n", (int)this);
  return pk_true;
}

/* 2004-12-30
 * reads a list of meta lines beginning at current line and terminating with
 *                      close tag as line first character
 * success: FILE = PREFIX[eol].[_ACTION_eol]#[_ACTION_eol*]'}'
 *             where  
 *                [_ACTION_eol] ::= [line_CALL] [lineSpaces_eol]
 *                   [line_CALL] ::= [StringFunction_CALL]
 *                   [line_CALL] ::= [ListFunction_CALL]
 *                [_ACTION_eol] ::= [ReferenceFree_string] [eoX] where  
 *                	         '}', <commandChar>  do not start 
 *                [_ACTION_eol] ::= [Parameterized_string] [eoX] where  
 *                	         '}', <commandChar>  do not start 
 *                [_ACTION_eol] ::= [ITERATOR] [fileSpace*] [line_CALL] [lineSpaces_eol] PLG NOT IMPLEMENTED
 *          returns in <argument_res> a new Argument containing this line 
 *          cursor is moved to final [eoX]
 * failure: syntax error
 *          error message is printed
 */
_PRIVATE_ boolean DynamicContext::read_argument_genericLine_eol_ (ref2Argument * argument_res)
{
  boolean
    lCorrect = pk_true;
  boolean
    isEol;
  // log_ASDEBUG("DynamicContext::read_argument_genericLine_eol_");
  if (this_lexicalContext->loadIf_eol_ (&lCorrect))
    {
      /* FILE = PREFIX . [eol] # [Any*][EOF] */
      (*argument_res) = new StringArgument ();
      return pk_true;
    }
  /* FILE = PREFIX . # [Character][Any*][EOF]
   *              <lCorrect> remains unchanged
   * FILE = PREFIX . # [EOF] or reading failed
   *              <lCorrect> is set FALSE
   *              error message is printed
   *              this is set ERRORCOMPLETED or ERRORCLOSED */
  if (lCorrect)
    {
      boolean
	isList;
      if (readIf_argument_call_argList_spStar_ (&lCorrect, argument_res, &isEol, &isList))
	{
	  if (isEol)
	    return this_lexicalContext->load_nl ();	/* success: FILE = PREFIX . [eol] # [Any*][EOF]
							 * failure: PREFIX . # [EOF] or reading failed
							 *              error message is printed */
	  /* PLG error message is missing or comment is wrong */
	  return pk_false;
	}
      if (lCorrect)
	{
	  (*argument_res) = new StringArgument (this_lexicalContext->strdupRemainingString ());
	  if (this_lexicalContext->load_nl ())	/* success: FILE = PREFIX . [eol] # [Any*][EOF]
						 * failure: PREFIX . # [EOF] or reading failed
						 *              error message is printed */
	    return pk_true;
	}
    }
  (*argument_res) = NIL_Argument;
  return pk_false;
}

/* 2004-12-30
 * reads a list of meta lines beginning at current line and terminating with
 *                      close tag as line first character
 * success: FILE = PREFIX[eol].[_ACTION_eol*]#'}'
 *          returns in <argumentList_res> a new ArgumentList containing those lines but the stopping one
 *          cursor is moved to final '}'
 * failure: cursor starts  [_ACTION_eol*] [EOF]
 *          cursor is moved to [EOF]
 *          erro message is printed
 * PLG SOURCE OF CONSTRAINTS ON SPACE READING readOpt_genericLineList_eol_
 */
_PRIVATE_ boolean DynamicContext::readOpt_genericLineList_eol_ (ref2ArgumentList * argumentList_res)
{
  // log_ASDEBUG("DynamicContext::readOpt_genericLineList_eol_");
  if (this_lexicalContext->is_actionClosing ())
    {
      /* FILE = opt{PREFIX [eol]}. # [declClosingTAG][Any*][EOF]
       */
      (*argumentList_res) = NIL_ArgumentList;
      return pk_true;
    }
  else
    {
      /* FILE = opt{PREFIX [eol]}[Character+] . # [declClosingTAG][Any*][EOF]
       *          FILE = opt{PREFIX [eol]}. # [Any_but_declClosingTAG][Any*][EOF]
       *          FILE is not READY, error message is printed; this is set ERRORXXX 
       */
      ref2Argument
	argument;
      if (read_argument_genericLine_eol_ (&argument))
	{
	  ref2ArgumentList
	    lineTail;
	  if (readOpt_genericLineList_eol_ (&lineTail))
	    {
	      (*argumentList_res) = new ArgumentList (argument, lineTail);
	      return pk_true;
	    }
	}
    }
  ASFatal_unexpected (pk_false, "meta action lines");
  return pk_false;
}

/* 2004-11-16
 * reads and appends this ancestor and its following ancestors 
 *    when this header line is a node reference
 * success: FILE = PREFIX.[NODE_Identifier] opt{[Antecedence_attribute]} [lineSpaces_eol] [ANCESTOR_BODY][eol]#[Any*][EOF]
 *          appends this ancestor
 *          sets [ANCESTOR_BODY] as this ancestor actions,
 *          appends [ANCESTOR*] to this ancestor 
 * failure: FILE = PREFIX.#[Any*][EOF], [Any*] does not start with a [NODE_Identifier]
 *          syntax error occurred
 *              error message is printed
 */
_PRIVATE_
  boolean DynamicContext::read_nodeMetaRef_nl (ref2Argument * argumentName,
					       boolean * isVar, boolean * isFile, int *dependenceNature, boolean * dependenceLevel)
{
  boolean
    isEol;

  // log_ASDEBUG("DynamicContext::read_nodeMetaRef_nl");
  if (read_nodeIdentification_spStar_ (argumentName, isVar, isFile, &isEol))
    return this_lexicalContext->stepOpt_antecedenceModality_eol_ (dependenceNature, dependenceLevel);
  /* stepOpt_antecedenceModality_eol_
   * success: FILE = PREFIX . opt{ '~'opt{[lineSpaces]} }[eol] # [Any*][EOF] 
   *              <dependenceNature> is set to newerBuildCODE
   *              <highDependenceLevel> is set  FALSE if '~' is present, TRUE otherwise
   *          FILE = PREFIX . '!'opt{[White*]'*'}opt{[lineSpaces]}[eol] # [Any*][EOF]
   *              <dependenceNature> is set to everBuildCODE
   *              <highDependenceLevel> is set TRUE if '*' is present, FALSE otherwise
   *          FILE = PREFIX . '-'opt{[White*]'~'}opt{[lineSpaces]}[eol] # [Any*][EOF]
   *              <dependenceNature> is set to missingBuildCODE
   *              <highDependenceLevel> is set FALSE if '~' is present, TRUE otherwise
   * failure: none of the above case occurs 
   *             error message is printed 
   */

  return pk_false;
}

/* 2005-01-03
 * reads and appends this ancestor and its following ancestors 
 * success: cursor starts [ANCESTOR*] [NOT_ANCESTOR]
 *             where 
 *                [ANCESTOR] ::= '}?' [fileSpace*] [ANCESTOR_HEADER] [lineSpaces_eol] [_ACTION_eol*]
 *                [_ANCESTOR_HEADER] ::= [NODE_IDENTIFICATION] [White*] [Antecedence_attribute]
 *                [_ANCESTOR_HEADER] ::= [NODE_IDENTIFICATION]
 *                [_ANCESTOR_HEADER] ::= [ITERATOR_fSp] [ANCESTOR_CALL]  
 *                [_ANCESTOR_HEADER] ::= [ANCESTOR_CALL]  
 *                [Antecedence_attribute] ::= [neverTAG] + [everTAG] [White*] <recursiveTAG>
 *                [NOT_ANCESTOR] ::= [Character*], '}?' is not a prefix of [Character*]
 *          the list of meta ancestors that have been read is assigned to <genericAncestorList_res>
 *          cursor is moved to head of [NOT_ANCESTOR]
 * failure: cursor starts inaccurate term
 *          cursor is moved to errored position in file
 * PLG SOURCE OF CONSTRAINTS ON SPACE READING readOpt_genericAncestorList_eol_
 */
_PUBLIC_ boolean DynamicContext::readOpt_genericAncestorList_eol_ (ref2GenericAncestorList * genericAncestorList_res)
{
  // log_ASDEBUG("DynamicContext::readOpt_genericAncestorList_eol_");
  boolean
    lCorrect = pk_true;
  if (this_lexicalContext->stepIf_ancestorMark_fSpStar_nSp (&lCorrect))
    {
      /* FILE = PREFIX . '}?'[fileSpace*] # [Character_but_Space][Any*][EOF] */
      ref2GenericAncestor
	genericAncestor;
      ref2AncestorFunctionCall
	ancestorFunctionCall;
      ref2ArgumentList
	ancestorAction;
      if (readIf_ancestorCall_eol_ (&lCorrect, &ancestorFunctionCall))
	{			/* lCorrect remains unchanged */
	  if (!(readOpt_genericLineList_eol_ (&ancestorAction)))	/* FILE = PREFIX [eol] . .. ? */
	    return pk_false;
	  genericAncestor = new CommandAncestor (ancestorFunctionCall, ancestorAction);
	}
      else if (lCorrect)
	{
	  ref2Argument
	    argumentName;
	  boolean
	    isVar,
	    isFile,
	    dependenceLevel;
	  int
	    dependenceNature;
	  if (!(read_nodeMetaRef_nl (&argumentName, &isVar, &isFile, &dependenceNature, &dependenceLevel)))
	    return pk_false;
	  if (!(readOpt_genericLineList_eol_ (&ancestorAction)))
	    return pk_false;
	  genericAncestor = new MetaAncestor (argumentName, isVar, isFile, dependenceNature, dependenceLevel, ancestorAction);
	}
      if (lCorrect)
	{
	  ref2GenericAncestorList
	    genericAncestorTail;
	  if (readOpt_genericAncestorList_eol_ (&genericAncestorTail))
	    {
	      (*genericAncestorList_res) = new GenericAncestorList (genericAncestor, genericAncestorTail);
	      return pk_true;
	    }
	}
      return pk_false;
    }
  /* FILE = PREFIX .# [Any_but_'}?'][Any*][EOF]
   *            <correct> remains unchanged
   * FILE = PREFIX . '}?'[fileSpace*] # [EOF]
   * or FILE = PREFIX . ... [EOF], reading failed
   *            error message is printed
   *            <correct> is set FALSE
   *            this is set  ERRORCOMPLETED or ERRORCLOSED
   */
  (*genericAncestorList_res) = NIL_GenericAncestorList;
  return lCorrect;
}

/* PLG 2008_07-02
 * reads and sets this_metaPrelude, these_genericAncestors and this_defaultMetaAction contents
 * success: FILE = PREFIX.opt{[_PRELUDE]}#'}'[Any*][EOF]
 *          where
 *             [PRELUDE] ::= '{' [lineSpaces_eol] [_line*], none of the [_line*] begins with '}'
 * failure: some error occurred during component reading
 *          error message is printed
 */
_PUBLIC_ boolean DynamicContext::readOpt_genericNodePrelude_eol_ (ref2ArgumentList * lineList_res)
{
  boolean
    lCorrect = pk_true;
  if (this_lexicalContext->stepIf_bodyPreludeO_spStar_eol_ (&lCorrect))
    /* FILE = PREFIX . '{' opt{[lineSpaces]}[eol] # [Any*][EOF], <correct> remains unchanged */
    return readOpt_genericLineList_eol_ (lineList_res);
  /* FILE = PREFIX .# [Any_but_'{'][Any*][EOF]
   *             <correct> remains unchanged
   *          FILE = PREFIX . '{' [White*] # [Character_but_Space][Any*] 
   *          or FILE = PREFIX . '{' opt{[lineSpaces]} # [EOF]
   *          or FILE = PREFIX . ... [EOF], reading failed
   *             <lCorrect> is set FALSE
   *             error message is printed
   *             this is set ERRORREADY or ERRORCOMPLETED */
  (*lineList_res) = NIL_ArgumentList;
  return lCorrect;
}

/* PLG 2008_07-02
 * reads and sets this_metaPrelude, these_genericAncestors and this_defaultMetaAction contents
 * success: FILE = PREFIX.opt{[_DEFAULT]}#'}'[Any*][EOF]
 *          where
 *             [DEFAULT] ::= <DEFAULT_mark> [lineSpaces_eol] [_line*], none of the [_line*] begins with '}'
 *                  <DEFAULT_mark> ::= '}/'
 * failure: some error occurred during component reading
 *          error message is printed
 */
_PUBLIC_ boolean DynamicContext::readOpt_genericNodeDefault_eol_ (ref2ArgumentList * default_res)
{
  boolean
    lCorrect = pk_true;
  if (this_lexicalContext->stepIf_defaultOpeningLine_eol_ (&lCorrect))
    /* FILE = PREFIX . '}/'opt{[lineSpaces]}[eol] # [Any*][EOF] */
    return readOpt_genericLineList_eol_ (default_res);
  /* FILE = PREFIX .# [Character_but_Space][Any*], '}/' is not a prefix of [Character_but_Space][Any*]
   *             <correct> remains unchanged
   *          FILE = PREFIX . '}/' [White*] # [Character_but_Space][Any*] 
   *          or FILE = PREFIX . '}/' opt{[lineSpaces]} # [EOF]
   *          or FILE = PREFIX . ... [EOF], reading failed
   *             <correct> is set FALSE
   *             error message is printed
   *            this is set  ERRORCOMPLETED or ERRORCLOSED */
  (*default_res) = NIL_ArgumentList;
  return lCorrect;
}

_PUBLIC_ _VIRTUAL_ boolean DynamicContext::build_instance_in (ref2Context callingContext, ref2ArgumentList parameterValues)
{
  // logDEBUG_in("build_instance_in in DynamicContext");
  return parameterListLink_listExpand_ (parameterValues);
}

/* ==========  ==========  ==========  ========== TITLE: BUILDER
 */

/* creates the dynamic context for a function declaration embedded in <owner_context> with same substitution marks
 * <owner_context> is not NIL (not checked)
 *          this scope is set to DYNAMIC
 */
DynamicContext::DynamicContext (ref2Context owner_context):Context (owner_context)
{
  this_nestingContext = owner_context;
  this_localIdentifierList = NIL_IdentifierList;
  // fprintf(stderr, "DEBUG new DynamicContext_2 this context adr is %i, embedded in %i\n", (int)this, (int)this_nestingContext);
}
