_PUBLIC_ ref2ElementList ElementList::next ()
{
  return cdr;
}

void
ElementList::setNext (ref2ElementList succ)
{
  cdr = succ;
}

ref2Element
ElementList::content ()
{
  return car;
}

/* PLG 2006-01-17
 * the element in ElementList (car) is supposed to be not NIL
 * search for an existing node named <name> 
 * success: a node named <name> is found in this list
 *          returns this node in <refFoundElement>, <refFoundElement> is not NIL
 * failure: this does not contain node named <name>
 *          this is empty
 *              <refFoundElement> remains unchanged
 *              <refFoundElement> holds this last node 
 */
boolean
ElementList::find (char *name, ref2Element * refFoundElement)
{
  // fprintf(stderr, "DEBUG  ElementList::find name is %s\n", name);
  ref2ElementList theElementList = this;
  while (theElementList != NIL_ElementList)
    {
      (*refFoundElement) = theElementList->car;
      if ((*refFoundElement)->isNamed (name))
	return pk_true;
      theElementList = theElementList->cdr;
    }
  return pk_false;
}

/* PLG 2006-01-17
 * <pt2Element> is supposed to be not NIL
 * if this does not already contains <pt2Element> inserts before this a new element containing <pt2Element>
 * returns the new element
 */
ref2ElementList
ElementList::insertInSet (ref2Element pt2Element)
{
  ref2ElementList prevElementList = this;

  while (prevElementList != NIL_ElementList)
    {
      if ((prevElementList->car) == pt2Element)
	return this;
      prevElementList = prevElementList->cdr;
    }
  prevElementList = new ElementList (pt2Element);
  prevElementList->cdr = this;
  return prevElementList;
}


/* PLG 2006-01-17
 * <pt2Element> is supposed to be not NIL
 * success: pt2Element is added
 * failure: pt2Element is yet registered
 * if this does not already contains <pt2Element> appends a new element containing <pt2Element> at the end of this 
 */
boolean
ElementList::appendInSet (ref2Element pt2Element)
{
  ref2ElementList prevElementList = this;

  if (car == pt2Element)
    return pk_false;
  while (prevElementList->cdr != NIL_ElementList)
    {
      prevElementList = prevElementList->cdr;
      if ((prevElementList->car) == pt2Element)
	return pk_false;
    }
  prevElementList->cdr = new ElementList (pt2Element);
  return pk_true;
}

/* PLG 2006-01-17
 * <pt2Element> is supposed to be not NIL
 * appends a new element containing <pt2Element> at the end of this 
 * a list may contain several occurrences of the same object
 */
void
ElementList::append (ref2Element pt2Element)
{
  ref2ElementList prevElementList = this;
  while (prevElementList->cdr != NIL_ElementList)
    prevElementList = prevElementList->cdr;
  prevElementList->cdr = new ElementList (pt2Element);
}

void
ElementList::replace (ref2Element prevElement, ref2Element newElement)
{
  // fprintf(stderr, "DEBUG ElementList::replace\n");
  // fprintf(stderr, "         prevElement: adr is %i, name is %s\n", (int)prevElement, prevElement->get_name());
  // fprintf(stderr, "         newElement: adr is %i, name is %s\n", (int)newElement, newElement->get_name());
  ref2ElementList prevElementList = this;
  while (prevElementList->car != prevElement)
    {
      prevElementList = prevElementList->cdr;
    }
  prevElementList->car = newElement;
}

int
ElementList::size ()
{
  ref2ElementList lat = this;
  int count = 0;

  while (lat != NIL_ElementList)
    {
      count = count + 1;
      lat = lat->cdr;
    }
  return count;
}

void
ElementList::log_list ()
{
  ref2ElementList lat = this;

  while (lat != NIL_ElementList)
    {
      log_ln ();
      (lat->car)->log_name_sp ();
      lat = lat->cdr;
    }
}

void
ElementList::logUnreferred ()
{
  ref2ElementList lat = this;
  ref2Element anElement;

  while (lat != NIL_ElementList)
    {
      anElement = lat->content ();
      if (!anElement->isReferredTo ())
	{
	  anElement->log_comment_Reference_ln ();
	}
      lat = lat->next ();
    }
}

void
ElementList::print (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2ElementList lat = this;

  while (lat != NIL_ElementList)
    {
      outBuffer->line ();
      (lat->car)->print (printingContext, outBuffer, pk_true);
      lat = lat->cdr;
    }
}

void
ElementList::print (ref2Lexical printingContext, ref2IntePrinter outBuffer, ref2Context myOwner)
/* PLG 2007-03-26 previous parameter was 
									 * MKFFileNode containing this list; TO BE CHECKED */
{
  ref2ElementList lat = this;

  while (lat != NIL_ElementList)
    {
      if ((lat->car)->has_owner (myOwner))
	{
	  outBuffer->line ();
	  (lat->car)->print (printingContext, outBuffer, pk_true);
	}
      lat = lat->cdr;
    }
}

void
ElementList::printRef (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2ElementList lat = this;

  while (lat != NIL_ElementList)
    {
      (lat->car)->print_nameValue_sp (printingContext, outBuffer);
      outBuffer->line ();
      lat = lat->cdr;
    }
}

/* PLG 2006-01-17
 * <pt2Element> is supposed to be not NIL
 */
ElementList::ElementList (ref2Element pt2Element, ref2ElementList succ)
{
  cdr = succ;
  car = pt2Element;
}

/* PLG 2006-01-17
 * <pt2Element> is supposed to be not NIL
 */
ElementList::ElementList (ref2Element pt2Element)
{
  cdr = NIL_ElementList;
  car = pt2Element;
}
