#define IS_REDIRECTED() ( this_aliasedNode != NIL_FileNode )
#define ISNOT_REDIRECTED() ( this_aliasedNode == NIL_FileNode )
#define SET_REDIRECTED_TO(referredTo) this_aliasedNode = referredTo

#define IS_TOBEBUILT() this_buildingRequired

#define LOG_NAME_LN() LOGFILE()->put_fileNameValue_ln(this_name)
#define LOG_FILENAME_LN() LOGFILE()->put_fileNameValue_ln(this_filePathName)

#define GENERATE_(fileToPrint) ( this_buildingData->generate(this, fileToPrint) )
#define DO_INNER_() ( this_buildingData->do_innerConstructive(this_declarationContext, pk_false) )
#define EXECUTE_() ( this_buildingData->execute(this) )

_PUBLIC_ _VIRTUAL_ const char *
FileNode::myClassName ()
{
  return "FileNode";
}

/* ========== forward references and aliases 
 */

/* assumes acyclic references
 * returns ultimate dereferenced object
 */
ref2FileNode
FileNode::get_redirection ()
{
  if (ISNOT_REDIRECTED ())
    return this;
  return this_aliasedNode->get_redirection ();
}

/* PLG 2005-09-07
 * referredTo is a terminal file descriptor (its_aliasedNode is NIL) not in error and not equal to this
 * assumes acyclic references
 * success: redirection does not introduce cycle
 *          sets this referring to referredTo
 * failure: redirection would introduce cycle
 *          this and its successors are set in error
 *              this is set in error
 */
boolean
FileNode::set_redirectedToNew (ref2FileNode referredTo)
{
  // fprintf(stderr, "DEBUG FileNode::set_redirectedToNew\n");
  // fprintf(stderr, "           redirected: adr %i, name %s, file name %s\n", (int)this, this_name,this_filePathName);
  // fprintf(stderr, "           target: adr %i, name %s, file name %s\n", (int)referredTo, referredTo->get_name(), referredTo->get_fileName());
  if (IS_REDIRECTED ())
    {
      if (!(this_aliasedNode->set_redirectedToNew (referredTo)))
	{
	  set_fatal ();
	  return pk_false;
	}
    }
  SET_REDIRECTED_TO (referredTo);
  return pk_true;
}


/* referredTo is a terminal file descriptor (its_aliasedNode is NIL) not in error
 * assumes acyclic references
 * success: redirection does not introduce cycle
 *          sets this referring to referredTo
 * failure: redirection would introduce cycle
 *          this and its successors are set in error
 */
boolean
FileNode::set_redirectedTo (ref2FileNode referredTo)
{
  // fprintf(stderr, "DEBUG FileNode::set_redirectedTo\n");
  // fprintf(stderr, "           redirected: adr %i, name %s, file name %s\n", (int)this, this_name,this_filePathName);
  // fprintf(stderr, "           target: adr %i, name %s, file name %s\n", (int)referredTo, referredTo->get_name(), referredTo->get_fileName());
  if (this == referredTo)
    {
      set_fatal ();
      return pk_false;
    }
  if (IS_REDIRECTED ())
    {
      if (!(this_aliasedNode->set_redirectedTo (referredTo)))
	{
	  set_fatal ();
	  return pk_false;
	}
    }
  SET_REDIRECTED_TO (referredTo);
  return pk_true;
}

/* ==========  ==========  global actions
 */

/* ==========  ==========  Attributes
 */
/* returns this file system name
 */
boolean
FileNode::is_writable ()
{
  return this_isWritable;
}

boolean
FileNode::is_present ()
{
  return this_isPresent;
}

/* returns this file system name
 */
const char *
FileNode::get_fileName ()
{
  return this_filePathName;
}

/* strcmp this (indirect accessible) system file name and <name>
 * returns an int 
 *         < 0 if this system file name is lexicographically less than <name> or is NIL
 *         0 if this system file name is equal to <name>
 *         > 0 if this system file name is lexicographically greater than <name>
 */
int
FileNode::cmp_fileName (const char *name)
{
  if (this_filePathName == NIL_String)
    return -1;
  return strcmp (this_filePathName, name);
}

/* ==========  ==========  Writer
 */
_VIRTUAL_ void
FileNode::print_nameValue_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_fileNameValue (outBuffer, this_name);
  outBuffer->space ();
}

/* ==========  ==========  Log
 */

_VIRTUAL_ void
FileNode::log_outerName_ln ()
{
  if (this_filePathName == NIL_String)
    LOG_NAME_LN ();
  else
    LOG_FILENAME_LN ();
}

_VIRTUAL_ void
FileNode::log_date_sp ()
{
  if (this_isPresent)
    {
      NodeElement::log_date_sp ();
    }
  else
    LOG_STR_SP ("missing file");
}

/* PLG 2005-08-31
 * this is supposed to be safe
 * prints commented ancestor reference
 */
void
FileNode::levelInfo_comment_missing_ln ()
{
  levelInfo_comment_ref_sp__ln ("+++ missing");
}

void
FileNode::levelInfo_comment_presentAncestor_ln (time_t contextTime)
{
  if (levelInfo_comment_ref_sp ())
    if (IS_strictlyYOUNGER_THAN (contextTime))
      {
	NodeElement::log_date_sp ();
	LOG_STR_LN ("+++ younger");
      }
    else
      {
	NodeElement::log_date_sp ();
	LOG_LN ();
      }
}

/* PLG 2005-08-31
 * this is supposed to be safe
 * prints commented ancestor reference
 */
void
FileNode::levelInfo_comment_present_notDefined_ln ()
{
  if (levelInfo_comment_ref_sp ())
    {
      NodeElement::log_date_sp ();
      LOG_STR_LN ("+++ cannot be rebuilt");
    }
}

/* PLG 2005-08-31
 * this is supposed to be safe
 * prints commented ancestor reference
 */
void
FileNode::levelInfo_comment_present_notDefined_ln (time_t contextTime)
{
  if (levelInfo_comment_ref_sp ())
    {
      if (IS_strictlyYOUNGER_THAN (contextTime))
	{
	  NodeElement::log_date_sp ();
	  LOG_STR_LN ("+++ younger, definition missing: cannot be rebuilt");
	}
      else
	{
	  NodeElement::log_date_sp ();
	  LOG_STR_LN ("--- definition missing");
	}
    }
}

_PUBLIC_ void
FileNode::aliasWarning_ (const char *name)
{
  if (ISDEBUG ())
    {
      log_fullReference_sp ();
      LOG_STR_SP_STR_LN ("is aliased to", name);
    }
}

_PUBLIC_ void
FileNode::erroredAliasWarning_ ()
{
  if (ISDEBUG ())
    {
      log_fullReference_sp ();
      LOG_STR_LN ("is aliased to errored file");
    }
}

/* PLG 2005-08-30
 * assumes this not redirected and safe,
 * this is missing 
 * this is not defined and required
 * set this stated to pk_true
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *                <cycleRoot> remains unchanged
 * failure: this is not dclared
 *          or some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 */
_PROTECTED_ void
FileNode::fatal_missingRequired ()
{
  LOG_LN_FATA_STR_SP ("required file missing without definition");
  LOG_STR_LN (this_filePathName);
  set_fatal ();
}

_PROTECTED_ void
FileNode::warning_notARootFile ()
{
  LOG_LN_WARN_SP ();
  log_fullReference_sp ();
  log_str_ln ("FILE IS NOT A ROOT FILE (neither write nor exec)");
  set_error ();
}

_PROTECTED_ void
FileNode::errored_ln_file_ (const char *mess)
{
  LOG_LN_ERRO_STR_SP ("file");
  log_fullReference_sp ();
  log_str_ln (mess);
  set_error ();
}

/* ==========  ========== Reader
 */

/* ==========  ==========  Grapher
 */
/* PLG 2006-01-02
 * assumes this not redirected and safe,
 * this is missing
 * this_isDefined is pk_true (=>this_buildingData not NIL ) 
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <comingFrom> is the node immediately preceded by this
 * success: these ancestors are safe, cycle free
 *               a message is printed
 *               this is set to be built
 *               this is registered as a passive node
 *               this_date is assigned FUTURE
 *          not NIL <creator> is assigned to this_creator
 * failure: this is set in error 
 *            this is cycle free
 *               <cycleRoot> remains unchanged
 *            this participates to a cycle
 *               <cycleRoot> is set to the root of this cycle
 * 
 * this_stepIndex is set to the_stepIndex
 */
_PRIVATE_ boolean FileNode::_visitMessIf_missing_defined_toBeBuilt (ref2NodeElement creator, ref2NodeElement comingFrom, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::_visitMessIf_missing_defined_toBeBuilt node is  %s \n", this_name);
  levelInfo_comment_missing_ln ();
  /* this is defined (=>this_buildingData is not NIL */
  if (_innerExplorationToDo_ln (cycleRoot))
    {
      time_t
	updatedDate;
      if (reset_toBeBuilt (creator))
	return pk_true;
      SET_FUTURE ();
      if (_visit_standard_toBeBuilt (creator, comingFrom, &updatedDate, cycleRoot))
	{
	  register_passiveNode ();
	  return pk_true;
	}
      return pk_false;
    }
  if (IS_SAFE ())
    {
      /* this is ancestor free */
      if (reset_toBeBuilt (creator))
	return pk_true;
      set_toBeBuilt (creator);
      SET_FUTURE ();
      register_passiveNode ();
      return pk_true;
    }
  return pk_false;
}

/* PLG 2006-01-17
 * this is supposed to be safe, not an alias
 * this_isDefined is pk_true (=>this_buildingData not NIL ) 
 * this building data is not NIL
 * <contextTime> is the time of the file which building can be decided when looking at these ancestors
 * <comingFrom> is the most inner node preceded by this
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *          updated date is computed wrt dependence modality
 *             0-a reference YY to an ancestor XX can be 
 *                 never or standard modality: XX is updated with intrinsic modality
 *                     XX updated date is set
 *                     YY updated date is XX updated date
 *                 ever or recursive ever modality: XX is updated with intrinsic modality
 *                     XX updated date is set
 *                     YY dependence updated date is FUTURE
 *                 missing modality: XX is not updated
 *                     XX updated date is not set
 *                     XX is a missing file or a variable referring to some missing file
 *                        YY dependence updated date is FUTURE
 *                     other
 *                        YY dependence updated date is OLDEST
 *              1- all YY dependence updated date are older than this initial date
 *                         this updated date is this_date
 *              2- some YY dependence updated date is newer than this initial date
 *                         this updated date is FUTURE
 *             this is present, this updated date is assigned to this_date and <updatedDate>
 *             this is missing, FUTURE is assigned to this_date and <updatedDate>
 *          <cycleRoot> remains unchanged
 * failure: some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 * this is not set to be built
 * this constructor remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_
  boolean FileNode::_visitMessIf_intrinsicUpdate (time_t contextTime, ref2NodeElement comingFrom, time_t * updatedDate, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::_visitMessIf_intrinsicUpdate name is <%s>, pathname is <%s>\n", this_name, this_filePathName);
  if (this_isPresent)
    {
      levelInfo_comment_presentAncestor_ln (contextTime);
      (*updatedDate) = GET_DATE ();
      if (_innerExplorationToDo_ln (cycleRoot))
	{
	  if (IS_UPDATED ())
	    return pk_true;
	  if (_visit_intrinsicUpdate (GET_DATE (), comingFrom, updatedDate, cycleRoot))
	    {
	      if (IS_strictlyOLDER_THAN (*updatedDate))
		{
		  SET_FUTURE ();
		  levelInfo_ObjectToBuild ();
		}
	      return pk_true;
	    }
	}
      return IS_SAFE ();
    }
  levelInfo_comment_missing_ln ();
  SET_FUTURE ();
  (*updatedDate) = GET_DATE ();
  if (_innerExplorationToDo_ln (cycleRoot))
    {
      if (IS_UPDATED ())
	return pk_true;
      if (_visit_intrinsicUpdate (GET_DATE (), comingFrom, updatedDate, cycleRoot))
	{
	  levelInfo_ObjectToBuild ();
	  return pk_true;
	}
    }
  return IS_SAFE ();
}

/* PLG 2006-01-06
 * should be protected, 
 * this is supposed to be safe, defined, not a MKFFileNode, not an ExternNode, 
 * this_isDefined is pk_true (=>this_buildingData not NIL ) 
 * time computation for these ancestors, when these ancestor updating dates may affect this updating decision
 * this may have been previoudly visited
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <contextTime> is the time of the file which building can be decided when looking at these ancestors
 * <comingFrom> is the most inner node preceded by this
 * this is the current node which ancestors are to be checked
 *          an ancestor XX is referred to with 
 *              a passive standard dependence
 *              a standard dependence
 *              a building requirement dependence:
 *              a recursive building requirement dependence
 *              a missing building requirement, XX updated date does not affect this updated date
 *              a passive missing requirement, XX updated date does not affect this updated date
 * success: some ancestor updated date is newer than this_date initial value
 *              this_date is assigned newest intrinsic updated date of these ancestors
 *          all ancestor updated dates are older than this_date initial value
 *               this_date remains unchanged
 *          this is set to be built
 *          updatedDate> is assigned updated date
 * failure: some of these ancestors is in error 
 *              <cycleRoot> remains unchanged
 *          this participates to a cycle
 *              <cycleRoot> is set to this cycle root
 *          set this in error
 *          <updatedDate> is assigned initial date
 * this_date remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_
  boolean FileNode::_visitMessIf_buildable_recursiveEverUpdate (ref2NodeElement creator, ref2NodeElement comingFrom,
								time_t * updatedDate, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::_visitMessIf_buildable_recursiveEverUpdate, name is <%s>\n", this_name);
  boolean
    hasYoungerAncestor = pk_false;
  if (this_isPresent)
    {
      levelInfo_comment_ref_date_ln ();
      if (_visitIf_recursiveEverUpdate (creator, comingFrom, updatedDate, &hasYoungerAncestor, cycleRoot))
	{
	  if (hasYoungerAncestor)
	    SET_FUTURE ();
	  if (!reset_toBeBuilt (creator))
	    set_toBeBuilt (creator);
	  (*updatedDate) = GET_DATE ();
	  return pk_true;
	}
      if (IS_SAFE ())
	{
	  if (!reset_toBeBuilt (creator))
	    set_toBeBuilt (creator);
	  return pk_true;
	}
      return pk_false;
    }
  levelInfo_comment_missing_ln ();
  SET_FUTURE ();
  if (_visitIf_recursiveEverUpdate (creator, comingFrom, updatedDate, &hasYoungerAncestor, cycleRoot))
    {
      (*updatedDate) = GET_DATE ();
      if (!reset_toBeBuilt (creator))
	set_toBeBuilt (creator);
      return pk_true;
    }
  if (IS_SAFE ())
    {
      if (!reset_toBeBuilt (creator))
	set_toBeBuilt (creator);
      return pk_true;
    }
  return pk_false;
}

/* ========== Entry points, this may not be defined 
*/

/* PLG 2006-01-04
 * this remains unchanged
 * success: this is safe and missing in host
 *              <correct> remains unchanged
 * failure: this is safe and present in host
 *              <correct> remains unchanged
 *          this is not safe
 *              <correct> is set to false
 * <cycleRoot> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean FileNode::has_missingAncestorFile (ref2NodeElement comingFrom, boolean * correct, ref2NodeElement * cycleRoot)
{
  if (IS_REDIRECTED ())
    return (get_redirection ())->has_missingAncestorFile (comingFrom, correct, cycleRoot);
  if (IS_SAFE ())
    {
      if (this_isPresent)
	{
	  levelInfo_comment_ref_date_ln ();
	  return pk_false;
	}
      levelInfo_comment_missing_ln ();
      return pk_true;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  (*correct) = pk_false;
  return pk_false;
}

/* PLG 2006-01-09
 * assumes this_buildingData not NIL when this is defined and safe
 * this node analysis is required by a presence inquiring dependence following an active link
 * <creator> is the most inner constructive node that precedes this, creator is not NIL
 * <comingFrom> is the most inner file (transitively) preceded by this
 * <cycleRoot> is set to a cycle root if this is in a cycle not beginning with this
 * success: this is safe and host present
 *               <isMissing> remains unchanged
 *          this is safe, not host present, defined, not in a cyclic dependence, all referred to nodes are properly defined 
 *               <isMissing> is set to pk_true
 *               this date is set to FUTURE
 *               this is set to be built
 *               this is registered as a passive node if needed
 *               this_stepIndex is set to the_stepIndex
 * failure: this is not safe, or this is host missing or this is not properly defined
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean FileNode::set_missingUpdate (ref2NodeElement creator, ref2NodeElement comingFrom, boolean * isMissing,
						 ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::set_missingUpdate node is  %s, this is %i \n", this_name, (int)this);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_missingUpdate (creator, comingFrom, isMissing, cycleRoot);
  if (IS_SAFE ())
    {
      if (this_isPresent)
	{
	  levelInfo_comment_ref_date_ln ();
	  return pk_true;
	}
      if (this_isDefined)
	{
	  (*isMissing) = pk_true;
	  /* PLG 2008-08-14, following condition added */
	  if (IS_REGISTERED ())
	    {
	      levelInfo_comment_missing_ln ();
	      return pk_true;
	    }
	  return _visitMessIf_missing_defined_toBeBuilt (creator, comingFrom, cycleRoot);
	  /* TRUE: registered as a passive node */ }
      fatal_missingRequired ();
      return pk_false;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* PLG 2006-01-11
 * assumes this_buildingData not NIL when this is defined and safe
 * <context_time> is the time of the file which building can be decided when looking at this
 * <comingFrom> is the node immediately preceded by this
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *          this_stepIndex is not smaller than the_stepIndex
 *             this_date is assigned to <updatedDate>
 *          this_stepIndex is smaller than the_stepIndex
 *             this is defined, present
 *                this does not have newer ancestor file
 *                     this updated date is this_date
 *                this has newer ancestor file
 *                     this updated date is FUTURE
 *                     this is to be built is set to pk_true
 *             this is defined, missing
 *                 this updated date is FUTURE
 *                 this is to be built is set to pk_true
 *             this is not defined, present in host
 *                 this updated date is this_date
 *             this updated date is assigned to this_date, <updatedDate>
 *          this_stepIndex is set to the_stepIndex
 *          <cycleRoot> remains unchanged
 * failure: this is absent in host, this is not defined or
 *          some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 *          <updatedDate> remains unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean FileNode::set_intrinsicUpdate (time_t contextTime, ref2NodeElement comingFrom, time_t * updatedDate, ref2NodeElement * cycleRoot)
{
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_intrinsicUpdate (contextTime, comingFrom, updatedDate, cycleRoot);
  // fprintf(stderr, "DEBUG FileNode::set_intrinsicUpdate name is <%s>, pathname is <%s>\n", this_name, this_filePathName);
  if (IS_SAFE ())
    {
      if (this_isDefined)
	return _visitMessIf_intrinsicUpdate (contextTime, comingFrom, updatedDate, cycleRoot);
      if (this_isPresent)
	{
	  levelInfo_comment_presentAncestor_ln (contextTime);
	  (*updatedDate) = GET_DATE ();
	  return pk_true;
	}
      fatal_missingRequired ();
      return pk_false;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* PLG 2005-08-30 
 * assumes this_buildingData not NIL when this is defined and safe
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <contextTime> is the time of the file which building can be decided when looking at this
 * <comingFrom> is the node immediately preceded by this
 * success: updating dates have been computed for this node and its ancestors
 *          this is registered as a passive node if needed
 * failure: this is set in error 
 *            this is cycle free
 *               <cycleRoot> remains unchanged
 *            this participates to a cycle
 *               <cycleRoot> is set to the root of this cycle
 * 
 * refers to comments in overlaped Node::set_standardUpdate 
 */
_PUBLIC_
  _VIRTUAL_ boolean FileNode::set_standardUpdate (ref2NodeElement creator, time_t contextTime, ref2NodeElement comingFrom,
						  time_t * updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::set_standardUpdate name is <%s>, pathname is <%s>\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_standardUpdate (creator, contextTime, comingFrom, updatedDate, isToBeBuilt, cycleRoot);
  if (IS_SAFE ())
    {
      /* PLG 2008-08-14, following condition added */
      if (IS_REGISTERED ())
	{
	  levelInfo_comment_presentAncestor_ln (contextTime);
	  (*updatedDate) = GET_DATE ();
	  return pk_true;
	}
      if (this_isDefined)
	{
	  if (this_isPresent)
	    {
	      levelInfo_comment_presentAncestor_ln (contextTime);
	      (*updatedDate) = GET_DATE ();
	      if (_innerExplorationToDo_ln (cycleRoot))
		{
		  /* this is safe and present, this_isDefined, this has ancestors, standard visit */
		  if (reset_buildable (creator, updatedDate, isToBeBuilt))
		    return pk_true;
		  if (_visit_standardUpdate (creator, GET_DATE (), comingFrom, updatedDate, isToBeBuilt, cycleRoot))
		    {
		      if (!(*isToBeBuilt))
			{
			  if (IS_strictlyOLDER_THAN (*updatedDate))
			    {
			      set_toBeBuilt ();
			      (*isToBeBuilt) = pk_true;
			    }
			  else
			    return pk_true;
			}
		      SET_FUTURE ();
		      (*updatedDate) = GET_DATE ();
		      register_passiveNode ();
		      return pk_true;
		    }
		  return pk_false;
		}
	      if (IS_SAFE ())
		{
		  if (reset_buildable (creator, updatedDate, isToBeBuilt))
		    return pk_true;
		  set_actualCreator (creator);
		  return pk_true;
		}
	      return pk_false;
	    }
	  (*updatedDate) = GET_FUTURE ();
	  (*isToBeBuilt) = pk_true;
	  return _visitMessIf_missing_defined_toBeBuilt (creator, comingFrom, cycleRoot);
	  /* TRUE: registered as a passive node */ }
      if (this_isPresent)
	{
	  levelInfo_comment_present_notDefined_ln (contextTime);
	  (*updatedDate) = GET_DATE ();
	  return pk_true;
	}
      fatal_missingRequired ();
      return pk_false;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* 
 * assumes this_buildingData not NIL when this is defined and safe
 * <comingFrom> is the most inner file (transitively) preceded by this
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <recursive> is pk_true if building is required for this and recursively its ancestors
 * assumes this in error when provided with NIL building data
 * set this stated to pk_true
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *          this is registered as a passive node if needed
 *                <cycleRoot> remains unchanged
 * failure: some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean FileNode::set_everUpdate (ref2NodeElement creator, ref2NodeElement comingFrom, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::set_everUpdate name is <%s>, pathname is <%s>\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_everUpdate (creator, comingFrom, cycleRoot);
  if (IS_SAFE ())
    {
      /* PLG 2008-08-14, following condition added */
      if (IS_REGISTERED ())
	{
	  levelInfo_comment_ref_date_ln ();
	  return pk_true;
	}
      if (this_isDefined)
	{
	  if (this_isPresent)
	    {
	      levelInfo_comment_ref_date_ln ();
	      /* this is defined (=>this_buildingData is not NIL OK */
	      if (_innerExplorationToDo_ln (cycleRoot))
		{
		  time_t
		    updatedDate;
		  if (reset_toBeBuilt (creator))
		    return pk_true;
		  if (_visit_standard_toBeBuilt (creator, comingFrom, &updatedDate, cycleRoot))
		    {
		      if (IS_strictlyOLDER_THAN (updatedDate))
			SET_FUTURE ();
		      register_passiveNode ();
		      return pk_true;
		    }
		  return pk_false;
		}
	      if (IS_SAFE ())
		{
		  /* this is ancestor free */
		  set_toBeBuilt (creator);
		  register_passiveNode ();
		  return pk_true;
		}
	      return pk_false;
	    }
	  return _visitMessIf_missing_defined_toBeBuilt (creator, comingFrom, cycleRoot);
	  /* TRUE: registered as a passive node */ }
      if (this_isPresent)
	{
	  levelInfo_comment_present_notDefined_ln ();
	  return pk_true;
	}
      fatal_missingRequired ();
      return pk_false;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}


/* 
 * assumes this_buildingData not NIL when this is defined and safe
 * <comingFrom> is the most inner file (transitively) preceded by this
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <recursive> is pk_true if building is required for this and recursively its ancestors
 * assumes this in error when provided with NIL building data
 * set this stated to pk_true
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *          this is registered as a passive node if needed
 *                <cycleRoot> remains unchanged
 * failure: some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean FileNode::set_recursiveEverUpdate (ref2NodeElement creator, ref2NodeElement comingFrom,
						       time_t * updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG FileNode::set_recursiveEverUpdate name is <%s>, pathname is <%s>\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_recursiveEverUpdate (creator, comingFrom, updatedDate, isToBeBuilt, cycleRoot);
  if (IS_SAFE ())
    {
      if (this_isDefined)
	{
	  (*isToBeBuilt) = pk_true;
	  if (_visitMessIf_buildable_recursiveEverUpdate (creator, comingFrom, updatedDate, cycleRoot))
	    {
	      if (!IS_REGISTERED ())
		register_passiveNode ();
	      return pk_true;
	    }
	  return pk_false;
	}
      if (this_isPresent)
	{
	  levelInfo_comment_present_notDefined_ln ();
	  (*updatedDate) = GET_DATE ();
	  return pk_true;
	}
      fatal_missingRequired ();
      return pk_false;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* failure: this is not a direct file
 */
_PUBLIC_ _VIRTUAL_ boolean FileNode::set_rootUpdate (ref2MKFFileNode comingFrom, int force)
{
  // fprintf(stderr, "DEBUG FileNode::set_rootUpdate name is <%s>, pathname is <%s>\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_rootUpdate (comingFrom, force);
  if (IS_SAFE ())
    error_constructiveFile (get_name ());
  else
    levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* ==========  ==========  Generator
 */

/* assumes this not redirected 
 * success: this is safe, 
 *          this has not to be built
 *              <somethingToDo> remains unchanged
 *          this has to be built, this has a creator
 *              <somethingToDo> is set to pk_true
 * failure: this is not safe
 *          or this is safe, this has to be built, this does not have creator
 *          <somethingToDo> remains unchanged
 *          error message is printed
 */
boolean
FileNode::_checkGenerator (boolean * _remainsToBoBuilt)
{
  // fprintf(stderr, "DEBUG FileNode::_checkGenerator this_name is <%s>\n", this_name);
  if (IS_SAFE ())
    {
      if (IS_TOBEBUILT ())
	{
	  if (IS_PROCESSED ())
	    {
	      levelInfo_DefaultGeneration (pk_false, "already done");
	      (*_remainsToBoBuilt) = pk_false;
	    }
	  else if (has_creator_err ())
	    (*_remainsToBoBuilt) = pk_true;
	  else
	    return pk_false;
	  return pk_true;
	}
      (*_remainsToBoBuilt) = pk_false;
      return pk_true;
    }
  else
    OUPS_name ("unsafe file node", this_name);
  return pk_false;
}

/* assumes this not redirected 
 * success: this is safe, 
 *          this has not to be built
 *              <somethingToDo> remains unchanged
 *          this has to be built, this has a creator
 *              <somethingToDo> is set to pk_true
 * failure: this is not safe
 *          or this is safe, this has to be built, this does not have creator
 *          <somethingToDo> remains unchanged
 *          error message is printed
 */
boolean
FileNode::checkGenerator (boolean * somethingToDo)
{
  boolean remainsToBoBuilt;
  // fprintf(stderr, "DEBUG FileNode::checkGenerator this_name is <%s>\n", this_name);
  if (_checkGenerator (&remainsToBoBuilt))
    {
      if (remainsToBoBuilt)
	{
	  SET_PROCESSED ();
	  (*somethingToDo) = mainToGenerate ();
	}
      return pk_true;
    }
  return pk_false;
}

/* 2005-08-26
 * this method is called when this is an ancestor of a constructive node with never modality
 * success: this is safe
 *              this construction is no more required 
 *              this construction is required and this has a creator
 *                 a aliasWarning is printed
 * failure: this is in error
 *          or this execution is required and
 *                this does not have a creator
 */
_VIRTUAL_ boolean
FileNode::nested_requiredConstructive ()
{
  // fprintf(stderr, "DEBUG FileNode::nested_requiredConstructive this_name is <%s>\n", this_name);
  boolean remainsToBoBuilt;
  if (IS_REDIRECTED ())
    return (get_redirection ())->nested_requiredConstructive ();
  if (_checkGenerator (&remainsToBoBuilt))
    {
      if (remainsToBoBuilt)
	levelInfo_shouldBeDone ();
      return pk_true;
    }
  return pk_false;
}

/* If this is not an executable file, its building requires execution of its content
 * PLG introduce exec modality to execute an executable file that is not declared
 * <curExecutedNode> is the nearest node that depends on this
 * executes commands line by line
 */
_VIRTUAL_ boolean
FileNode::execute (ref2NodeElement curExecutedNode)
{
  // fprintf(stderr, "DEBUG FileNode::execute this_name is <%s>\n", this_name);
  boolean somethingToDo = pk_false;
  if (IS_REDIRECTED ())
    return (get_redirection ())->execute (curExecutedNode);
  if (checkGenerator (&somethingToDo))
    {
      if (somethingToDo)
	{
	  return EXECUTE_ ();
	}
      return pk_true;
    }
  error_line_reference ("Trying to use");
  return pk_false;
}

/*
 * on the fly generation of a not constructive file construction commands when building curBuiltFile
 * a not constructive file cannot be a generation root
 * curBuiltFile is the nearest file that depends on this
 */
_VIRTUAL_ boolean
FileNode::generate (ref2NodeElement curBuiltFile, ref2Buffer fileToPrint)
{
  // fprintf(stderr, "DEBUG FileNode::generate this_name is <%s>\n", this_name);
  boolean somethingToDo = pk_false;
  if (IS_REDIRECTED ())
    return (get_redirection ())->generate (curBuiltFile, fileToPrint);
  if (checkGenerator (&somethingToDo))
    {
      if (somethingToDo)
	{
	  return GENERATE_ (fileToPrint);
	}
      return pk_true;
    }
  return pk_false;
}

/* 
 * success: this is safe
 *          <activeDependence> is pk_true
 *          <activeDependence> is false
 *              this has not to be built
 *              this has to be built, this has a constructor
 *                   these ancestors are processed
 * failure: this is not safe
 *          or this is safe, this has to be built, this does not have constructor
 */
_VIRTUAL_ boolean
FileNode::do_innerConstructive (boolean activeDependence)
{
  // fprintf(stderr, "DEBUG FileNode::do_innerConstructive this_name is <%s>\n", this_name);
  boolean remainsToBeBuilt;
  if (IS_REDIRECTED ())
    return (get_redirection ())->do_innerConstructive (activeDependence);
  if (activeDependence)
    return IS_SAFE ();
  if (_checkGenerator (&remainsToBeBuilt))
    {
      if (IS_TOBEBUILT ())
	{
	  if (remainsToBeBuilt)
	    levelInfo_shouldBeDone ();
	  return DO_INNER_ ();
	}
      return pk_true;
    }
  return pk_false;
}

/*
 * root generation of an executable variable
 *     
 */
_VIRTUAL_ boolean
FileNode::root_generate ()
{
  // fprintf(stderr, "DEBUG FileNode::root_generate this_name is <%s>\n", this_name);
  if (IS_REDIRECTED ())
    return (get_redirection ())->root_generate ();
  OUPS_name ("illegal FileNode::root_generate, VIRTUAL expected", this_name);
  return pk_false;
}

/* ========== Builder
 */

/* PLG 2006-01-17
 * <readingContext> is not NIL (not checked)
 * aliasing a reference to get a declaration with the same name and the same host name
 * success: this is not defined (this has been created as an ancestor reference)
 *             creates a new mkf file node assigned to <mkfRes>
 *             this is redirected to <mkfRes> content
 * failure: this is a defined file node
 *             error message is printed
 *             this is set in error
 */
_PUBLIC_ _VIRTUAL_ boolean FileNode::aliasToMkf (ref2Context readingContext, char *name, boolean * isMkf)
{
  // readingContext->log_ASDEBUG("FileNode::aliasToMkf");
  if (IS_REDIRECTED ())
    return (get_redirection ())->aliasToMkf (readingContext, name, isMkf);
  if (this_isDefined)
    {
      readingContext->ASFatal_doubleDecl (name);
      set_fatal ();
      return pk_false;
    }
  return pk_true;
}

/* PLG 2005-09-07
 * this is not <prevNode>
 * <prevNode> is not NIL
 * success: 
 *          sets <prevNode> redirected to this
 *          sets this filename, this host presence, this date to <prevNode> corresponding values
 *          sets this referred to
 * failure: this filename is NIL or <prevNode> redirection to this failed
 *              this is set in error
 */
_PRIVATE_ void
FileNode::initAsRedirectedDefinition (ref2FileNode prevNode)
{
  this_isUsed = pk_true;
  if (prevNode->isInError ())
    {
      erroredAliasWarning_ ();
      set_fatal ();
    }
  else
    {
      ref2FileNode redirectedNode = prevNode->get_redirection ();
      if (redirectedNode->isInError ())
	{
	  erroredAliasWarning_ ();
	  set_fatal ();
	}
      else
	{
	  this_filePathName = redirectedNode->get_fileName ();
	  if (this_filePathName != NIL_String)
	    if (redirectedNode->set_redirectedToNew (this))
	      {
		SET_DATE (redirectedNode->get_date ());
		this_isPresent = redirectedNode->is_present ();
		this_isWritable = pk_false;
		return;
	      }
	  set_fatal ();
	}
    }
}

/* PLG 2005-09-07
 * initializes a new file descriptor
 * <fullFileName> is the system file name of this node
 * success: <fullFileName> is a legal file name in host
 *              sets its time attribute to
 *                 file date if file exists
 *                 APOCALYPSE if file doesn't exist
 *          registers this host file
 * failure: <fullFileName> is a not legal file name in host
 *          set this in error 
 *          sets its time attribute to NIL_time if file search is errored
 *          <fullFileName> is not NIL
 *               registers this host file
 */
_PRIVATE_ void
FileNode::init (char *fullFileName)
{
  // fprintf(stderr, "DEBUG FileNode::init filePath is %s\n", fullFileName);
  this_filePathName = fullFileName;
  this_aliasedNode = NIL_FileNode;
  if (fullFileName != NIL_String)
    {
      const char *curErrMess;
      time_t fileDate = getFDate (this_filePathName, &curErrMess);
      if (IS_UNKNOWN_TIME (fileDate))
	{
	  LOG_LN_ERRO_STR_SP_STR_LN ("file properties", curErrMess);
	  LOG_STR_SP_STR_LN ("file name is ", this_filePathName);
	  set_fatal ();
	}
      else
	{
	  this_isWritable = pk_false;
	  SET_DATE (fileDate);
	  this_isPresent = !(IS_MISSING ());
	}
    }
  else
    set_fatal ();
}

/* ========= Reference descriptor creation
 */

/* PLG 2005-09-07
 * <prevNode> is not NIL
 * this and <prevNode> have distinct name but refer to the same host file
 *
 * (Node :: sets this not defined)
 * creates and returns a file node reference named <name>, aliased to <prevNode> 
 *        <prevNode> may be defined file
 *        this is redirected to <prevNode>
 * initial date is the universal OLDEST date (GENESIA) but should not be used
 * sNets this context as the virtual context
 * FN1: USED IN ContextTree::get_fileReference
 */
_PUBLIC_ FileNode::FileNode (ref2FileNode prevNode, char *name)	/* FN1 */
  :
NodeElement (name)
/* initial date is the universal OLDEST date (GENESIA) */
{
  // fprintf(stderr, "DEBUG FileNode::FileNode(ref2FileNode prevNode, char * name)\n");
  // fprintf(stderr, "           this: adr %i, name %s, fileName in prevNode\n", (int)this, name);
  // fprintf(stderr, "           prevNode: adr %i, name, fileName %s\n", (int)prevNode, prevNode->get_name(), prevNode->get_fileName());
  this_isUsed = pk_true;
  if (prevNode->isInError ())
    {
      erroredAliasWarning_ ();
      set_fatal ();
    }
  else
    {
      ref2FileNode
	referredTo = prevNode->get_redirection ();
      if (referredTo->isInError ())
	{
	  erroredAliasWarning_ ();
	  set_fatal ();
	}
      else
	{
	  aliasWarning_ (referredTo->get_name ());
	  this_filePathName = referredTo->get_fileName ();
	  if (this_filePathName == NIL_String)
	    set_fatal ();
	  else
	    {
	      this_isWritable = referredTo->is_writable ();
	      if (set_redirectedTo (referredTo));
	      referredTo->setReferredTo ();
	    }
	}
    }
}

/* PLG 2005-09-07
 * <name> is not NIL
 * 
 * creates and returns a new object descriptor named <name>
 * this is a file node reference 
 * this is set not defined 
 * initial date is host file modification date if host file exists
 * initial date is APOCALYPSE if host file doesn't exist
 * registers this as file
 * sets this context as the virtual context
 * M6: USED TO INIT MKFFileNode::MKFFileNode(char * name, char * fullPathName, boolean * correct)
 * FN2: USED IN ContextTree::get_fileReference
 */
_PUBLIC_ FileNode::FileNode (char *name, char *fullFileName)	/* FN2 */
  :
NodeElement (name)
/* initial date is the universal OLDEST date (GENESIA) */
{
  // fprintf(stderr, "DEBUG FileNode::FileNode(char * name, char * fullFileName)\n");
  // fprintf(stderr, "           this: adr %i, name %s, file name %s\n", (int)this, name, fullFileName);
  this_isUsed = pk_true;
  init (fullFileName);
}

/* ========= Declaration descriptor creation
 */

/* PLG 2005-09-07
 * <prevNode> is not NIL
 * <readingContext> is not NIL (not checked)
 * this and <prevNode> have the same name and the same host name
 * <prevNode> is a reference to
 * this is a declaration
 * <execIt> is pk_true if this is executable
 * 
 * (Node :: sets this declared in readingContext)
 * success: sets <prevNode> redirected to this
 *          sets this filename, this host presence, this date to <prevNode> corresponding values
 *          sets this referred to
 *          replaces <prevNode> by this in adressing actions
 * failure: this filename is NIL or <prevNode> redirection to this failed
 *              this is set in error
 * M1:USED TO INIT MKFFileNode::MKFFileNode(ref2Context readingContext, ref2FileNode prevNode, boolean * correct)
 * D1: USED TO INIT DirectFileNode(ref2Context readingContext, ref2FileNode prevNode, boolean execIt) 
 */
_PROTECTED_ FileNode::FileNode (ref2Context readingContext, ref2FileNode prevNode)	/* FN3 */
:NodeElement (readingContext, prevNode->get_name ())
/* initial date is the universal OLDEST date (GENESIA) */
{
  // fprintf(stderr, "DEBUG FileNode::FileNode(ref2Context readingContext, ref2FileNode prevNode)\n");
  // fprintf(stderr, "           this: adr %i, name in prevNode, fileName in prevNode\n", (int)this);
  // fprintf(stderr, "           prevNode: adr %i, name %s, fileName %s\n", (int)prevNode, prevNode->get_name(), prevNode->get_fileName());
  initAsRedirectedDefinition (prevNode);
}

/* PLG 2005-09-07
 * <readingContext> is not NIL
 * <prevNode> is not NIL
 * this and <prevNode> have distinct name but refer to the same host file
 * <prevNode> is a reference to, or is a mkffile
 *     this or <prevNode> can be used as definition
 * in DirectFileNode::DirectFileNode
 *     this is used as declaration
 *
 * (Node :: sets this declared in readingContext)
 * success: sets <prevNode> redirected to this
 *          sets this filename, this host presence, this date to <prevNode> corresponding values
 *          sets this referred to
 * failure: this filename is NIL or <prevNode> redirection to this failed
 *              this is set in error
 * registers this static element declaration
 * M2: USED TO INIT MKFFileNode::MKFFileNode(ref2Context readingContext, ref2FileNode prevNode, boolean * correct)
 * M3: USED TO INIT MKFFileNode::MKFFileNode(ref2MKFFileNode prevNode, int hashi, char * name, boolean * correct), with readingContext set as prevNode reading context
 * D2: USED TO INIT DirectFileNode(ref2Context readingContext, ref2FileNode prevNode, char * name,
 * FN4: USED IN ContextTree::getIf_fileDeclaration(char * name, ref2FileNode *fileNode_res)
 */
_PUBLIC_ FileNode::FileNode (ref2Context readingContext, char *name, ref2FileNode prevNode)	/* FN4 */
  :
NodeElement (readingContext, name)
/* initial date is the universal OLDEST date (GENESIA) */
{
  // fprintf(stderr, "DEBUG FileNode::FileNode(ref2Context readingContext, char * name, ref2FileNode prevNode)\n");
  // fprintf(stderr, "           this: adr %i, name %s, fileName in prevNode\n", (int)this, name);
  // fprintf(stderr, "           prevNode: adr %i, name, fileName %s\n", (int)prevNode, prevNode->get_name(), prevNode->get_fileName());
  initAsRedirectedDefinition (prevNode);
}

/* PLG 2005-09-07
 * <readingContext> is not NIL
 * <name> is not NIL
 * <fullFileName> is a legal host file path
 * this is the first occurrence of host file <fullFileName>)
 * this is the unique reference and declaration descriptor node named <name>
 *
 * in DirectFileNode::DirectFileNode
 *
 * (Node :: sets this declared in readingContext)
 * success: sets this filename, this host presence, 
 *          sets this date to 
 *             host file modification date if <fullFileName> exists in host
 *             APOCALYPSE if <fullFileName> doesn't exist in host
 * set to not include file (???)
 * registers this as file
 * M4: USED TO INIT MKFFileNode::MKFFileNode(ref2Context readingContext, char * name, char * fullFileName, boolean * correct)
 * D3: USED TO INIT DirectFileNode(ref2Context readingContext, char * name, char * filePathName, boolean execIt)
 * FN5: USED IN ContextTree::getIf_fileDeclaration(char * name, ref2FileNode *fileNode_res)
 */
_PUBLIC_ FileNode::FileNode (ref2Context readingContext, char *name, char *fullFileName)	/* FN5 */
  :
NodeElement (readingContext, name)
/* initial date is the universal OLDEST date (GENESIA) */
{
  // fprintf(stderr, "DEBUG FileNode::FileNode(ref2Context readingContext, char * name, char * fullFileName)\n");
  // fprintf(stderr, "           this: adr %i, name %s, file name %s\n", (int)this, name, fullFileName);
  this_isUsed = pk_false;
  init (fullFileName);
}

/* ======================== stdin creation
 */

/* creates and returns the file descriptor of "stdin"
 * set to  include file
 * initial date is the universal OLDEST date (GENESIA)
   PLG used in MKFFileNode
 * sets this context as the virtual context
 * M5:USED TO INIT MKFFileNode::MKFFileNode()
 */
_PROTECTED_ FileNode::FileNode ()	/* FN6 */
:NodeElement ("stdin")
/* initial date is the universal OLDEST date (GENESIA) */
{
  // fprintf(stderr, "DEBUG FileNode::FileNode()\n");
  // fprintf(stderr, "           this: adr %i, name stdin\n", (int)this);
  this_filePathName = "stdin";
  this_isWritable = pk_false;
  this_aliasedNode = NIL_FileNode;
  this_isPresent = pk_true;
}
