_PUBLIC_ _VIRTUAL_ const char *
FormattedFileNode::myClassName ()
{
  return "FormattedFileNode";
}

/* ========== Attributes
 */
/* ========== Log
 */
/* ========== Printer
 */

_VIRTUAL_ void
FormattedFileNode::print_attributes_ln (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_fileWriteContext_sp (outBuffer, this_isExecutable, this_breakCharList, this_continuationCharacter);
  NodeElement::print_attributes_ln (printingContext, outBuffer);
}

/* ========== Reader
 */
/* ========== Grapher
 */
/* ========== Generator
 */
/* open a direct file that is being generated
 */
_VIRTUAL_ boolean
FormattedFileNode::openGeneration ()
{
  boolean lCorrect = pk_true;
  if (IS_SAFE ())
    {
      this_fileToPrint = new NBuffer (this_maxLineLength, this_breakCharList, this_continuationCharacter, &lCorrect);
      if (lCorrect && this_fileToPrint->open (this_filePathName))
	return pk_true;
      delete this_fileToPrint;
      set_fatal ();
    }
  else
    OUPS_name ("unsafe formatted file node", this_name);
  return pk_false;
}

/* ========== Generator
 */

/* ========== Builder
 */

_PRIVATE_ void
FormattedFileNode::init (char *bkCharList, char continuationChar)
{
  this_breakCharList = bkCharList;
  this_continuationCharacter = continuationChar;
  this_maxLineLength = 127;
}

/* PLG 2005-09-07
 * <readingContext> is not NIL (not checked)
 *     this and prevNode have the same name and the same host name
 *     <prevNode> is a forward reference, that has been set declared, and provided with not NIL building data
 *     this is used as declaration
 * 
 * sets this defined, initializes corresponding attributes
 * sets <prevNode> redirected to this
 * initial date is <prevNode> initial date
 * replaces <prevNode> by this in adressing actions
 */
_PUBLIC_ FormattedFileNode::FormattedFileNode (ref2Context readingContext, ref2FileNode prevNode,	/* F1 */
					       boolean execIt, char *bkCharList, char continuationChar):
DirectFileNode (readingContext, prevNode, execIt)
{
  init (bkCharList, continuationChar);
}

/* PLG 2005-09-07
 * <readingContext> is not NIL (not checked)
 *     this and prevNode have distinct name but refer to the same host file
 *     <prevNode> is a forward reference, that has been set declared, and provided with not NIL building data
 *     this is used as definition
 * 
 * sets this defined, initializes corresponding attributes
 * sets <prevNode> redirected to this
 * initial date is <prevNode> initial date
 */
_PUBLIC_ FormattedFileNode::FormattedFileNode (ref2Context readingContext, ref2FileNode prevNode, char *name,	/* F2 */
					       boolean execIt, char *bkCharList, char continuationChar):
DirectFileNode (readingContext, prevNode, name, execIt)
{
  init (bkCharList, continuationChar);
}


/* PLG 2005-09-07
 * <readingContext> is not NIL (not checked)
 *     this is the first occurrence of host file <fullFileName>)
 *     this is the unique reference and declaration descriptor node named <name>
 *
 * sets this defined, initializes corresponding attributes
 * initial date is host file modification date if host file exists
 * initial date is APOCALYPSE if host file doesn't exist
 */
_PUBLIC_ FormattedFileNode::FormattedFileNode (ref2Context readingContext, char *name, char *filePathName,	/* F3 */
					       boolean execIt, char *bkCharList, char continuationChar):
DirectFileNode (readingContext, name, filePathName, execIt)
{
  init (bkCharList, continuationChar);
}
