_PUBLIC_ _VIRTUAL_ const char *
FunctionCall::myClassName ()
{
  return "FunctionCall";
}

/* ========== Attributes
 */

/* command level printing
 */
boolean
FunctionCall::is_static ()
{
  if (IS_SAFE ())
    return IS_STATIC ();
  return pk_false;
}

/* command level printing
 */
int
FunctionCall::get_scope ()
{
  return GET_SCOPE ();
}

_VIRTUAL_ const char *
FunctionCall::get_name ()
{
  return this_functionName;
}

_VIRTUAL_ ref2ArgumentList
FunctionCall::get_actualParameters ()
{
  return these_actualParameters;
}

/* 
 */
_PUBLIC_ boolean FunctionCall::have_isomorphicParameters (ref2ArgumentList arg_actualParameters)
{
  if (these_actualParameters == arg_actualParameters)
    return pk_true;
  if (these_actualParameters == NIL_ArgumentList)
    return pk_false;
  return these_actualParameters->is_isomorphic (arg_actualParameters);
}

/* _funcCall may be NIL
 */
_PUBLIC_ boolean FunctionCall::have_isomorphicParameters (ref2FunctionCall _funcCall)
{
  if (_funcCall != NIL_FunctionCall)
    {
      ref2ArgumentList
	arg_actualParameters = _funcCall->get_actualParameters ();
      if (these_actualParameters == arg_actualParameters)
	return pk_true;
      if (these_actualParameters == NIL_ArgumentList)
	return pk_false;
      return these_actualParameters->is_isomorphic (arg_actualParameters);
    }
  return pk_false;
}

/* ========== Writer
 */

/* ==========  ==========  ==========  ========== TITLE: PRINT
 */
_PRIVATE_ void
FunctionCall::printIf_iteration_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  // fprintf(stderr, "DEBUG FunctionCall::printIf_iteration_sp\n");
  if (this_isIterated)
    ((ref2IteratorContext) this_readingContext)->print_iteration_sp (printingContext, outBuffer);
}

/* 2004-12-09
 * prints this 
 * actual arguments are printed with depth -1
 */
_PUBLIC_ _VIRTUAL_ void
FunctionCall::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  printIf_iteration_sp (printingContext, outBuffer);
  printingContext->out_callOpening_sp (outBuffer, this_functionName);
  if (these_actualParameters != NIL_ArgumentList)
    these_actualParameters->print_actualFormula (printingContext, outBuffer, depth - 1, pk_false);
  printingContext->out_callClosing (outBuffer);
}

/* command level printing
 */
_PUBLIC_ _VIRTUAL_ void
FunctionCall::print_actualParametersformula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  // fprintf(stderr, "DEBUG FunctionCall::print_formula these_actualParameters adr is %i\n", (int)these_actualParameters);
  if (these_actualParameters == NIL_ArgumentList)
    outBuffer->put_openCloseTag_sp ();
  else
    these_actualParameters->Cprint_formula (printingContext, outBuffer, pk_true);
}

/* command level printing
 */
_PUBLIC_ _VIRTUAL_ void
FunctionCall::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  // logDEBUG_in("print_formula_");
  printIf_iteration_sp (printingContext, outBuffer);
  printingContext->out_callHeader (outBuffer, this_functionName);
  print_actualParametersformula (printingContext, outBuffer);
  // logDEBUG_in("print_formula_ DONE");
}

_PROTECTED_ void
FunctionCall::logDEBUG_actualParameters ()
{
  // logDEBUG_in(FunctionCall::"======================================================================");
  // logDEBUG_in("these_actualParameters formula is ");
  /* DEBUG */
  if (these_actualParameters == NIL_ArgumentList)
    log_NILComponent_ln (10, "these_actualParameters");
  else
    these_actualParameters->log_formula_ (pk_true);	/* DEBUG */
}

_PUBLIC_ void
FunctionCall::log_formula ()
{
  // logDEBUG_in("log_formula_");
  print_formula (LOGMARKS (), LOGFILE ());
}

/* ========== Log
 */

void
FunctionCall::logSimpleDebug_formula_ln ()
{
  print_formula (LOGMARKS (), LOGFILE ());
  log_ln ();
}

_PUBLIC_ _VIRTUAL_ void
FunctionCall::logExtendedDebug_formula_ln (int level)
{
  logDebug_objAnnounce_ln (level);
  logSimpleDebug_formula_ln ();
  if (these_actualParameters == NIL_ArgumentList)
    log_NILComponent_ln (level, "argument list");
  else
    {
      log_subTitle_ln (level, "these arguments are ");
      these_actualParameters->logExtendedDebug_formula_ln (level + 1);
    }
  logDebug_objClosing_ln (level);
}

/* ========== Set
 */

/* PLG 2005-12-07
 * success: this is iterated, some iterator list is not empty
 *              those referred to iterators list values are expanded
 *          <hasEmptyList> is set to pk_true if some identifier is linked to an empty list
 *          <hasEmptyList> is set to false if all identifiers are linked to not empty lists
 *          <correct> remains unchanged
 * failure: this is not iterated 
 *               <correct> remains unchanged
 *          this is iterated, iterator list is empty or some iterator expansion failed
 *               <correct> is set to false
 *               system error message is printed
 */
_PROTECTED_ boolean FunctionCall::setIterate (boolean * correct, boolean * hasEmptyList)
{
  // fprintf(stderr, "DEBUG FunctionCall::setIterate\n");
  if (this_isIterated)
    {
      if (((ref2IteratorContext) this_readingContext)->link_iteratorList_ (hasEmptyList))
	return pk_true;
      (*correct) = pk_false;
    }
  // fprintf(stderr, "DEBUG FunctionCall::setIterate return on error\n");
  return pk_false;
}

/* 2007-06-26
 */
_PRIVATE_ void
FunctionCall::set_actualParameters (ref2ArgumentList actualParameters)
{
  these_actualParameters = actualParameters;
  if (these_actualParameters == NIL_ArgumentList)
    SET_EMPTY_PARAMETER ();
  else
    {
      SET_PARAMETER_SCOPE (these_actualParameters->get_scope ());
      if (IS_ITERATED_PARAMETER ())
	SET_ITERATED ();
    }
}

/* ========== Printer
 */

/* ========== Reader
 */

/* ========== Generator
 */

/* ========== Builder
 */

/* PLG 2007-02-08 FunctionCall::FunctionCall(ref2Context readingContext, const char * name, ref2ArgumentList actualParameterList)
 * creates a new call to function named <name>
 * <readingContext> is the reading context of this function call
 * this scope remains unknown
 */
_PROTECTED_ FunctionCall::FunctionCall (ref2Context readingContext, const char *name, ref2ArgumentList actualParameterList):
Soundness ()
{
  // fprintf(stderr, "DEBUG FunctionCall::FunctionCall ref2Context\n");
  this_readingContext = readingContext;
  SET_DYNAMIC ();
  this_isIterated = pk_false;
  this_functionName = name;
  set_actualParameters (actualParameterList);
}

/* PLG 2007-02-08 FunctionCall::FunctionCall(ref2IteratorContext readingContext, const char * name, ref2ArgumentList actualParameterList)
 * creates a new iterated call to function named <name>
 * <readingContext> is the reading context of this function call
 * this scope is set to iterated
 */
_PROTECTED_ FunctionCall::FunctionCall (ref2IteratorContext readingContext, const char *name, ref2ArgumentList actualParameterList):
Soundness ()
{
  // fprintf(stderr, "DEBUG FunctionCall::FunctionCall ref2IteratorContext\n");
  this_readingContext = readingContext;
  SET_ITERATED ();
  this_isIterated = pk_true;
  this_functionName = name;
  set_actualParameters (actualParameterList);
}
