/* TODO 
 */
_PUBLIC_
/* ========== Attributes
 */
  ref2GenericAncestorList GenericAncestorList::next ()
{
  return cdr;
}

void
GenericAncestorList::append (ref2GenericAncestorList succ)
{
  // fprintf(stderr,"DEBUG GenericAncestorList::append, this is %i, succ = %i\n", (int)this, (int)succ);
  if (cdr == NIL_GenericAncestorList)
    cdr = succ;
  else
    cdr->append (succ);
}

/* ========== Log
 */
void
GenericAncestorList::logDebugThis_formula (int level)
{
  if (car != NIL_GenericAncestor)
    car->logExtendedDebug_formula_ln (level);
}

_PUBLIC_ _VIRTUAL_ void
GenericAncestorList::logExtendedDebug_formula_ln (int level)
{
  ref2GenericAncestorList predActList = this;
  logDebug_objAnnounce_ln (level);
  do
    {
      predActList->logDebugThis_formula (level + 1);
      predActList = predActList->cdr;
    }
  while (predActList != NIL_GenericAncestorList);
  logDebug_objClosing_ln (level);
}

/* ========== Printer
 */

void
GenericAncestorList::printThis_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  if (car == NIL_GenericAncestor)
    return;
  outBuffer->put_ln_dependenceOpening_spN ();
  car->print_actualFormula (printingContext, outBuffer, depth);
}

void
GenericAncestorList::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  ref2GenericAncestorList predActList = this;
  do
    {
      predActList->printThis_actualFormula (printingContext, outBuffer, depth);
      predActList = predActList->cdr;
    }
  while (predActList != NIL_GenericAncestorList);
}

void
GenericAncestorList::printThis_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (car == NIL_GenericAncestor)
    return;
  outBuffer->put_ln_dependenceOpening_spN ();
  car->print_formula (printingContext, outBuffer);
}

void
GenericAncestorList::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2GenericAncestorList predActList = this;
  do
    {
      predActList->printThis_formula (printingContext, outBuffer);
      predActList = predActList->cdr;
    }
  while (predActList != NIL_GenericAncestorList);
}

/* ========== Grapher
 */

/* ========== Generator
 */

/* generates a ancestor list from this car
 * <callingContext> is not NIL (not checked)
 * if this list is empty, following meta actions are ignored
 * success: generated ancestors are returned in <ancestorList_res>
 *              <ancestorList_res> may be empty
 * failure: this is not a correct ancestor generator
 *               error message is printed
 */
boolean
GenericAncestorList::build_ancestorList_car (ref2Context callingContext, ref2AncestorList * ancestorList_res)
{
  if (car == NIL_GenericAncestor)
    return pk_true;
  return car->build_ancestorList (callingContext, ancestorList_res);	/* <ancestorList_res> may be empty */
}

/* generates a ancestor list from this
 * <callingContext> is not NIL (not checked)
 * if this list is empty, following meta actions are 
ignored
 * success: generated ancestors are returned in <ancestorList_res>
 *              <ancestorList_res> may be empty
 * failure: this is not a correct ancestor generator
 *               error message is printed
 */
boolean
GenericAncestorList::build_ancestorList (ref2Context callingContext, ref2AncestorList * ancestorList_res)
{
  // fprintf(stderr, "DEBUG GenericAncestorList::build_ancestorList, this is %i \n", (int)this);
  ref2GenericAncestorList predActList = this;
  ref2AncestorList ancestorList_tail;
  do
    {
      if (!(predActList->build_ancestorList_car (callingContext, &ancestorList_tail)))
	{			/* <ancestorList_tail> may be empty OK */
	  return pk_false;
	}
      predActList = predActList->cdr;
      if (predActList == NIL_GenericAncestorList)
	{
	  (*ancestorList_res) = ancestorList_tail;	/* <ancestorList_res> may be empty  OK */
	  return pk_true;
	}
    }
  while (ancestorList_tail == NIL_AncestorList);
  (*ancestorList_res) = ancestorList_tail;	/* <ancestorList_res> is not empty OK */
  if (predActList->build_ancestorList (callingContext, &ancestorList_tail))
    {				/* <ancestorList_tail> may be empty */
      if (ancestorList_tail != NIL_AncestorList)
	{
	  (*ancestorList_res)->append (ancestorList_tail);
	}
      return pk_true;
    }
  return pk_false;
}

/* ========== Reader
 */

/* ========== Builder
 */
/* 
 * Empty ancestor list creation
 */
GenericAncestorList::GenericAncestorList (ref2GenericAncestor genericAncestor, ref2GenericAncestorList genericAncestorTail)
{
  car = genericAncestor;
  cdr = genericAncestorTail;
}
