
/* ========== Attributes
 */

_PUBLIC_ _VIRTUAL_ const char *
HostValFunctionCall::myClassName ()
{
  return "HostValFunctionCall";
}

_PUBLIC_ _VIRTUAL_ boolean HostValFunctionCall::is_isomorphic (ref2ArgumentFunctionCall arg_funcCall)
{
  if ((ref2ArgumentFunctionCall) this == arg_funcCall)
    return pk_true;
  if (arg_funcCall != NIL_ArgumentFunctionCall)
    {
      return arg_funcCall->is_isomorphic (this);
    }
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean HostValFunctionCall::is_isomorphic (ref2HostValFunctionCall arg_funcCall)
{
  if (this == arg_funcCall)
    return pk_true;
  return have_isomorphicParameters (arg_funcCall);
}

/* ========== Log
 */

/* ========== Writer
 */

/* ========== Set
 */

/* ========== Printer
 */

/* ========== Reader
 */
/* ========== Generator
 */

/* PLG 2005-12-14
 * should be private
 * this is supposed to be SAFE
 * this_CString_varName is supposed to be a legal host variable name
 * success: this_CString_res is set to a new m'allocated copy of 
 *               this host identifier string value
 * failure: this_CString_varName is not defined in host
 *          error message is printed
 */
_PRIVATE_ boolean HostValFunctionCall::setIf_string_HostValue_ ()
{
  if (getIf_string_HostValue_ (this_CString_varName, &this_CString_expanded))
    {
      this_expandedArgument = new StringArgument (this_CString_expanded);
      return pk_true;
    }
  set_error ();
  return pk_false;
}

/* PLG 2005-12-14
 * should be protected
 * this is supposed to be SAFE
 * these_actualParameters is supposed to have one and only one (string) element (not checked)
 * success: this string element is correctly expanded to a host identifier 
 *          <CString_res> is set to 
 *               the string value of this host identifier 
 * failure: this string element cannot be correctly expanded to a host identifier 
 *          error message is printed
 */
_PUBLIC_ _VIRTUAL_ boolean HostValFunctionCall::single_call ()
{
  // fprintf(stderr, "DEBUG HostValFunctionCall::single_call()\n");
  if (IS_SAFE ())
    {
      if (IS_STATIC ())
	return pk_true;
      if (HAS_DYNAMIC_PARAMETER ())
	if (!get_identifier_expand_singleton_ (&this_CString_varName))
	  {			/*(fatal)error message is printed, fatal is set */
	    return pk_false;
	  }
      return setIf_string_HostValue_ ();
    }
  return pk_false;			/* false: error message is printed, this is set in error */
}

/* ========== Grapher
 */

/* ========== Builder
 */

/* PLG 2005-10-13
 * should be protected
 * reads in <inBuffer>, the string argument of a call to string or list predefined functions
 * identifier occurrences in generic string arguments are preceded by '$' and embraced with <leftChar>, <rightChar>
 * those identifiers are searched for in this_readingContext declarations
 * success: FILE = PREFIX. '(' [fileSpace*] [string_ARGUMENT_fSp] ')' [lineSpaces]#[Character_but_Space][Any*][EOF]
 *            <isEol> is set to pk_true
 *          FILE = PREFIX. '(' [fileSpace*] [string_ARGUMENT_fSp] ')' [lineSpaces]#[eol][Any*][EOF]
 *            <isEol> is set to false
 *                       these_actualParameters is the Argument containing [string_ARGUMENT_fSp]
 *          links context variable, parameter and function declarations in ??? TODO
 *          this scope is set
 * failure: syntax error 
 *          error message is printed
 *          this is set in error
 */
_PRIVATE_ boolean HostValFunctionCall::set_identifierArg (ref2Argument actualParameter)
{
  if (set_stringIdentifierArg (actualParameter, &this_CString_varName))
    {				/*(fatal)error message is printed, fatal is set */
      if (HAS_STATIC_PARAMETER ())
	{
	  if (setIf_string_HostValue_ ())
	    {			/* false: error message is printed, this is set in error */
	      SET_STATIC ();
	      return pk_true;
	    }
	  return pk_false;
	}
      return pk_true;
    }
  return pk_false;
}

/* 2004-10-14
 * creates a new :Hostval function call 
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
_PUBLIC_ HostValFunctionCall::HostValFunctionCall (ref2Context readingContext, ref2Argument actualParameter, boolean * correct):StringFunctionCall (readingContext, hostvalWORD,
		    actualParameter)
{
  // fprintf(stderr, "DEBUG HostValFunctionCall::HostValFunctionCall(ref2Context, ...)\n");
  if (!set_identifierArg (actualParameter))	/*(fatal)error message is printed, fatal is set */
    (*correct) = pk_false;
}

/* 2004-10-14
 * creates a new iterated :Hostval function call 
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 * PLG check if parameter is a singleton !!!!!
 */
_PUBLIC_ HostValFunctionCall::HostValFunctionCall (ref2IteratorContext readingContext, ref2Argument actualParameter, boolean * correct):StringFunctionCall (readingContext, hostvalWORD,
		    actualParameter)
{
  //  ((ref2Context)readingContext)->log_ASDEBUG("HostValFunctionCall::HostValFunctionCall iterated");
  if (!set_identifierArg (actualParameter))	/*(fatal)error message is printed, fatal is set */
    (*correct) = pk_false;
}
