/* ================================== */
_PUBLIC_ _VIRTUAL_ const char *
Identifier::myClassName ()
{
  return "Identifier";
}

/* ========== Attributes
 */

/* returns pk_true iff this refers to a parameter
 */
boolean
Identifier::isParameter ()
{
  return IS_NESTED ();
}

/* returns pk_true iff this refers to an iterator
 */
boolean
Identifier::isIterator ()
{
  return IS_ITERATED ();
}

/* returns pk_true iff this has string type
 */
boolean
Identifier::has_string_type ()
{
  return HAS_STRING_TYPE ();
}

/* returns pk_true iff this has string list type
 */
boolean
Identifier::has_list_type ()
{
  return HAS_LIST_TYPE ();
}

/* returns this char * name
 */
char *
Identifier::get_name ()
{
  return this_name;
}

_PUBLIC_ _VIRTUAL_ boolean Identifier::is_isomorphic (char *str_name)
{
  if (this_name == NIL_String)
    return (str_name == NIL_String);
  if (str_name == NIL_String)
    return pk_false;
  return (strcmp (this_name, str_name) == 0);
}

_PUBLIC_ _VIRTUAL_ boolean Identifier::is_isomorphic (ref2MetaArgument metaIdent)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Identifier::is_isomorphic (ref2CommandArgument commandIdent)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean Identifier::is_isomorphic (ref2Identifier ident_arg)
{
  if (ident_arg == NIL_Identifier)
    return pk_false;
  return ident_arg->is_innerIsomorphic (this);
}

/* ident_arg is not NIL (not checked)
 */
_PUBLIC_ _VIRTUAL_ boolean Identifier::is_innerIsomorphic (ref2Identifier ident_arg)
{
  return is_isomorphic (ident_arg->get_name ());
}

/* ident_arg is not NIL (not checked)
 */
_PUBLIC_ _VIRTUAL_ boolean Identifier::is_innerIsomorphic (ref2MetaIdentifier ident_arg)
{
  return pk_false;
}

/* ========== Log
 */
/* ========== Writer
 */

/* prints this name separated by spaces
 */
_VIRTUAL_ void
Identifier::print_nameValue_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (this_name != NIL_String)
    outBuffer->space (this_name);
}

/* prints this name separated by spaces
 */
_VIRTUAL_ void
Identifier::print_typedName_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  outBuffer->put_typedName_sp (HAS_LIST_TYPE (), this_name);
}

/* prints this actual value with CString tags
 */
_VIRTUAL_ void
Identifier::print_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  if (this_argumentValue != NIL_Argument)
    this_argumentValue->print_expandedValue (printingContext, outBuffer, inAList);
}

/* prints this actual value with CString tags
 */
_VIRTUAL_ void
Identifier::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth, boolean inAList)
{
  if (this_argumentValue != NIL_Argument)
    this_argumentValue->print_actualFormula (printingContext, outBuffer, depth, inAList);
  else
    outBuffer->print (this_name);
}

/* prints this actual value 
 */
_VIRTUAL_ void
Identifier::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_nameValue_sp (printingContext, outBuffer);
}

/* prints this actual value 
 */
_VIRTUAL_ void
Identifier::print_value (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (this_argumentValue != NIL_Argument)
    this_argumentValue->print_formula (printingContext, outBuffer, pk_false);
}

_VIRTUAL_ void
Identifier::log_formula (int fdepth, int vdepth)
{
  log_strOrNil (this_name);
}

/* ========== Log
 */
_PROTECTED_ _VIRTUAL_ void
Identifier::log_name_sp ()
{
  if (this_name == NIL_String)
    {
      LOG_STR_SP ("#");
    }
  else
    {
      LOG_STR_SP (this_name);
    }
}

_PRIVATE_ void
Identifier::log_value ()
{
  // logDEBUG_in("log_value");
  if (this_argumentValue == NIL_Argument)
    LOG_STR_ ("#");
  else
    {
      this_argumentValue->log_formula_ (HAS_STRING_TYPE ());
    }
}

_PUBLIC_ _VIRTUAL_ void
Identifier::log_formula_ ()
{
  // logDEBUG_in("log_formula_()");
  if (IS_SAFE ())
    {
      log_name_sp ();
      log_valueOpen ();
      log_value ();
      log_valueClose ();
    }
  else
    LOG_STR_SP ("##");
}

/* prints this actual value with CString tags
 */
void
Identifier::logDebug_expandedFormula_ln (int level)
{
  if (this_argumentValue == NIL_Argument)
    log_NILComponent_ln (level, "argument");
  else
    {
      log_subTitle_ln (level, "this is linked to the following argument");
      this_argumentValue->logExtendedDebug_formula_ln (level + 1);
    }
}

_PROTECTED_ void
Identifier::logDebug_formulaHead_ln (int level)
{
  logDebug_objAnnounce_ln (level);
  if (IS_NESTED ())
    log_spn_str_ln (level, "this is a parameter");
  else
    log_spn_str_ln (level, "this is static");
  if (HAS_STRING_TYPE ())
    log_spn_str_ln (level, "this type is string");
  else
    log_spn_str_ln (level, "this type is list");
  LOG_SPN_STR_SP (level, "this name is");
}

/* prints this actual value with CString tags
 */
_PUBLIC_ _VIRTUAL_ void
Identifier::logExtendedDebug_formula_ln (int level)
{
  if (IS_SAFE ())
    {
      logDebug_formulaHead_ln (level);
      LOG_STR_LN (this_name);
      logDebug_expandedFormula_ln (level);
    }
  else
    log_spn_str_ln (level, "this is ## in error ##");
}

/* 
 */
_PUBLIC_ _VIRTUAL_ void
Identifier::logDebug_formula_sp (int depth)
{
  if (IS_SAFE ())
    {
      LOG_STR_ ("<");
      log_name_sp ();
      log_value ();
      LOG_STR_SP (">");
    }
  else
    LOG_STR_SP ("##");
}

/* ========== Generator
 */
_VIRTUAL_ boolean
Identifier::link_expandedList_IdVal_ (boolean * isEmptyList)
{
  OUPS ("illegal call to Identifier::link_expandedList_IdVal_(), unsafe");
  return pk_false;
}

_VIRTUAL_ boolean
Identifier::downCurrentList (boolean * correct)
{
  OUPS ("illegal call to Identifier::downCurrentList()");
  return pk_false;
}

_VIRTUAL_ boolean
Identifier::stepCurrentList ()
{
  OUPS ("illegal call to Identifier::stepCurrentList(), this is not safe");
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be an expanded string (not checked)
 * returns this linked actual string value 
 */
_VIRTUAL_ char *
Identifier::get_CString ()
{
  if (this_argumentValue == NIL_Argument)
    return NIL_String;
  return this_argumentValue->get_CString ();
}

/* PLG 2005-12-07
 * this is supposed to be linked to an expanded value (not checked)
 * success: this is safe, this has string type
 *          returns this Argument value in <stringArgument_res>
 * failure: this is not safe or this does not have string type
 *          error message is printed
 */
_VIRTUAL_ boolean
Identifier::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  // logDEBUG_in("get_expandedString");
  if (HAS_STRING_TYPE ())
    {
      if (IS_SAFE ())
	{
	  (*stringArgument_res) = (ref2StringArgument) this_argumentValue;
	  return pk_true;
	}
      unsafe_use ("Identifier::get_expandedString");
      (*isFatal) = IS_FATAL ();
      return pk_false;
    }
  fatal_named_typeConflict_ln (this_name, "string");
  (*isFatal) = pk_true;
  return pk_false;
}

/* PLG 2005-12-07
 * assumes this being linked to an expanded list value when this is a safe list
 * success: this is safe and has list type
 *          returns this expanded argument list value in <argumentList_res>
 * failure: this is not safe or does not have list type
 *          system error message is printed
 */
_VIRTUAL_ boolean
Identifier::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  // logDEBUG_in("get_expandedList of", this_name);
  if (HAS_LIST_TYPE ())
    {
      if (IS_SAFE ())
	{
	  (*argumentList_res) = (ref2ArgumentList) this_argumentValue;
	  return pk_true;
	}
      unsafe_named_use (this_name, "Identifier::get_expandedList");
      (*isFatal) = IS_FATAL ();
      return pk_false;
    }
  fatal_named_typeConflict_ln (this_name, "list");
  (*isFatal) = pk_true;
  return pk_false;
}

/* this is supposed to be expanded
 * success: returns this Argument value in <argument_res>
 * failure: this is not safe
 *          system error message is printed
 */
_VIRTUAL_ boolean
Identifier::get_expanded (ref2Argument * argument_res)
{
  if (IS_SAFE ())
    {
      (*argument_res) = this_argumentValue;
      return pk_true;
    }
  unsafe_use ("Identifier::get_expanded");
  return pk_false;
}

/* 2004-12-13
 * this is supposed to be expanded
 * returns this actual string length
 */
_VIRTUAL_ int
Identifier::get_CStringLength_expanded ()
{
  if (this_argumentValue == NIL_Argument)
    return 0;
  return this_argumentValue->get_CStringLength_expanded ();
}

/* if this actual value is not NIL_String
 *    appends, without C string marks, this actual value to target
   target is supposed to be large enough
 * this is supposed to be expanded
 */
_VIRTUAL_ void
Identifier::cat_expanded (char *target)
{
  if (IS_SAFE ())
    if (this_argumentValue != NIL_Argument)
      this_argumentValue->cat_expanded (target);
}

/* PLG 2005-12-08
 * assumes: <anArgument> is not NIL (not checked)
 * types are assumed to be correct
 * success: sets expansion of <anArgument> as this parameter value
 * failure: this is not safe
 *          error message is printed
 */
boolean
Identifier::link_ (ref2Argument anArgument)
{
  this_argumentValue = anArgument;
  return pk_true;
}

/* PLG 2005-12-08
 * assumes: <anArgument> is not NIL (not checked)
 * types are assumed to be correct
 * success: sets expansion of <anArgument> as this parameter value
 * failure: this is not safe
 *          error message is printed
 */
boolean
Identifier::expand_ ()
{
  // logDEBUG_in("expand_ identifier", this_name);
  if (IS_SAFE ())
    {
      if (this_argumentValue == NIL_Argument)
	return pk_true;
      if (HAS_STRING_TYPE ())
	{
	  ref2StringArgument CString_res;
	  if (this_argumentValue->getIf_stringExpand_ (&CString_res))
	    {
	      this_argumentValue = CString_res;
	      return pk_true;
	    }
	  set_fatal ();
	}
      else
	{			/* has list type */
	  ref2ArgumentList list_res;
	  if (this_argumentValue->getIf_flatten_listExpand_ (&list_res))
	    {
	      this_argumentValue = (ref2Argument) list_res;
	      return pk_true;
	    }
	  set_fatal ();
	}
    }
  else
    OUPS ("illegal call to Identifier::link_expand_(), unsafe");
  return pk_false;
}

/* PLG 2005-12-08
 * assumes: <anArgument> is not NIL (not checked)
 * types are assumed to be correct
 * success: sets expansion of <anArgument> as this parameter value
 * failure: this is not safe
 *          error message is printed
 */
boolean
Identifier::link_expand_ (ref2Argument anArgument)
{
  /* 
     if ( this_type == 0 )
     logDEBUG_in("link_expand_ string identifier", this_name);
     else
     logDEBUG_in("link_expand_ list identifier", this_name);
     if ( this_scope  == 0 )
     logDEBUG_inner("link_expand_","scope is static");
     else if ( this_scope  == 1 )
     logDEBUG_inner("link_expand_","scope is parameter");
     else if ( this_scope  == 2 )
     logDEBUG_inner("link_expand_","scope is iterator");
     else 
     logDEBUG_inner("link_expand_","scope is unknown");
   */
  if (IS_SAFE ())
    {
      link_ (anArgument);
      return expand_ ();
    }
  else
    OUPS ("illegal call to Identifier::link_expand_(), unsafe");
  return pk_false;
}

/* ========== Builder
 */

/* creates and returns a list identifier named <aName>, reference to a context variable 
 * linked to <aValue>
 */
Identifier::Identifier (char *aName, ref2ArgumentList listValue):Soundness ()
{
  // logDEBUG_in("Identifier 1(char *, ref2ArgumentList) EMPTY or STATIC identifier named", aName);
  SET_LIST_TYPE ();
  if (listValue == NIL_ArgumentList)
    SET_EMPTY ();
  else
    SET_STATIC ();
  this_name = aName;
  this_argumentValue = (ref2Argument) listValue;
}

/* creates and returns a parameter identifier named <aName>
 * if isAListIdentifier this type is string list else it is string
 */
Identifier::Identifier (char *aName, boolean isAListIdentifier, ref2Argument formula):
Soundness ()
{
  /*
     if ( isAListIdentifier )
     logDEBUG_in("Identifier 2(char *, boolean, ref2Argument) DYNAMIC identifier named", aName);
     else
     logDEBUG_in("Identifier 2(char *, boolean, ref2Argument) DYNAMIC identifier named", aName);
   */
  if (isAListIdentifier)
    SET_LIST_TYPE ();
  else
    SET_STRING_TYPE ();
  SET_DYNAMIC ();
  this_name = aName;
  this_argumentValue = formula;
}

/* creates and returns a parameter identifier named <aName>
 * if isAListIdentifier this type is string list else it is string
 */
Identifier::Identifier (char *aName, boolean isAListIdentifier):
Soundness ()
{
  /*
     if ( isAListIdentifier )
     logDEBUG_in("Identifier 3(char *, boolean) DYNAMIC identifier named", aName);
     else
     logDEBUG_in("Identifier 3(char *, boolean) DYNAMIC identifier named", aName);
   */
  if (isAListIdentifier)
    SET_LIST_TYPE ();
  else
    SET_STRING_TYPE ();
  SET_DYNAMIC ();
  this_name = aName;
  this_argumentValue = NIL_Argument;
}

/* creates and returns an iterator named <aName>
 */
Identifier::Identifier (char *aName):
Soundness ()
{
  // logDEBUG_in("Identifier 4(char *) ITERATED identifier named", aName);
  SET_ITERATED ();
  SET_STRING_TYPE ();
  this_name = aName;
}
