_PUBLIC_
/* ==========  ==========  ==========  ========== TITLE: these attributes
 */
_VIRTUAL_ const char *
IteratorContext::myClassName ()
{
  return "IteratorContext";
}

boolean
IteratorContext::append_uniqueNewIterator (char *identName, ref2Iterator iterator)
{
  // fprintf(stderr, "DEBUG IteratorContext::append_uniqueNewIterator, context adr is %i, ident is %s\n", (int)this, identName);
  // log_ASDEBUG("IteratorContext::append_uniqueNewIterator");
  if (this_localIdentifierList == NIL_IdentifierList)
    {
      this_localIdentifierList = new IdentifierList (iterator);
      return pk_true;
    }
  if (this_localIdentifierList->append_uniqueNewIterator (identName, iterator))
    return pk_true;
  ASFatal_doubleDecl (identName);
  return pk_false;
}

/* ==========  ==========  ==========  ========== TITLE: PRINTING
 */

/* command level printing
 */
void
IteratorContext::print_iteration_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_iterate_sp (outBuffer);
  if (IS_SAFE ())
    this_localIdentifierList->print_sp (printingContext, outBuffer);
  outBuffer->put_closing_sp ();
}

/* ==========  ==========  ==========  ========== TITLE: FATAL
 */
_PRIVATE_ void
IteratorContext::ASFatal_emptyIteratorList ()
{
  (this_lexicalContext->ASFatal_emptyIteratorList) ();
}

/* ==========  ==========  ==========  ========== TITLE: SPECIFIC ACTIONS
 */
/*
 * this_nestingContext not NIL (not checked)
 * <CString_name> is a (not empty) identifier (not checked)
 * success: <CString_name> is an iterator identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is iterated
 *          <CString_name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <CString_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) StringArgument value of this static identifier; 
 *              <argument_res> scope is static
 *          <argument_res> type is string
 * failure: <CString_name> declaration is not (recursively) found
 *          <CString_name> declaration has unexpected type
 *              error message is printed
 */
_PUBLIC_ _VIRTUAL_ boolean IteratorContext::getIf_string_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG IteratorContext::getIf_string_id_, context adr is %i, ident is %s\n", (int)this, CString_name);
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (CString_name, &id_res))
	{
	  (*argument_res) = new NamedArgument (CString_name, id_res, pk_true);
	  return pk_true;
	}
    }
  return this_nestingContext->getIf_string_id_ (CString_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean IteratorContext::getIf_string_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG IteratorContext::getIf_string_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  (*argument_res) = new NamedArgument (id_res, pk_false);
	  return pk_true;
	}
    }
  return this_nestingContext->getIf_string_id_ (id_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean IteratorContext::getIf_string_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG IteratorContext::getIf_string_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  (*argument_res) = new NamedArgument (id_res, pk_false);
	  return pk_true;
	}
    }
  return this_nestingContext->getIf_string_id_ (id_name, correct, argument_res);
}

/* 2005-12-14
 * this_nestingContext not NIL (not checked)
 * <CString_name> is a (not empty) identifier (not checked)
 * success: <CString_name> is not an iterator identifier
 *          <CString_name> is a dynamic (not iterator) identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <CString_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) ArgumentList value of this static identifier; 
 *              <argument_res> scope is static
 *          <correct> remains unchanged
 *          <argument_res> type is list
 * failure: <CString_name> declaration is not (recursively) found
 *              <correct> remains unchanged
 *          <CString_name> is an iterator identifier which declaration is (recursively) found in a dynamic context
 *          or <CString_name> declaration has unexpected type
 *              error message is printed
 *             <correct> is set to false
 *          <argument_res> remains unchanged
 */
_VIRTUAL_ boolean
IteratorContext::getIf_list_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG IteratorContext::getIf_list_id_, context adr is %i, ident is %s\n", (int)this, CString_name);
  ref2Identifier id_res;
  if (this_localIdentifierList->find (CString_name, &id_res))
    {
      error_typeIdentifier_ (CString_name);
      (*correct) = pk_false;
      return pk_false;
    }
  return this_nestingContext->getIf_list_id_ (CString_name, correct, argument_res);
}

/* 2005-12-14
 * this_nestingContext not NIL (not checked)
 * <id_name> is a (not empty) identifier (not checked)
 * success: <id_name> is not an iterator identifier
 *          <id_name> is a dynamic (not iterator) identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <id_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) ArgumentList value of this static identifier; 
 *              <argument_res> scope is static
 *          <correct> remains unchanged
 *          <argument_res> type is list
 * failure: <id_name> declaration is not (recursively) found
 *              <correct> remains unchanged
 *          <id_name> is an iterator identifier which declaration is (recursively) found in a dynamic context
 *          or <id_name> declaration has unexpected type
 *              error message is printed
 *             <correct> is set to false
 *          <argument_res> remains unchanged
 */
_VIRTUAL_ boolean
IteratorContext::getIf_list_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG IteratorContext::getIf_list_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  ref2Identifier id_res;
  if (this_localIdentifierList->find (id_name, &id_res))
    {
      error_typeIdentifier_ ("generic name");
      (*correct) = pk_false;
      return pk_false;
    }
  return this_nestingContext->getIf_list_id_ (id_name, correct, argument_res);
}

_VIRTUAL_ boolean
IteratorContext::getIf_list_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // fprintf(stderr, "DEBUG IteratorContext::getIf_list_id_, context adr is %i, ident is\n", (int)this);
  /* DEBUG id_name->log_formula_ln(); */
  ref2Identifier id_res;
  if (this_localIdentifierList->find (id_name, &id_res))
    {
      error_typeIdentifier_ ("generic name");
      (*correct) = pk_false;
      return pk_false;
    }
  return this_nestingContext->getIf_list_id_ (id_name, correct, argument_res);
}

/*
 * this_nestingContext not NIL (not checked)
 * <name> is a (not empty) identifier (not checked)
 * success: <name> is an iterator identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is iterated
 *              <argument_res> type is string 
 *          <name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *              <argument_res> type is string or list
 *          <name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the static (StringArgument or ArgumentList) value of this static identifier; 
 *              <argument_res> scope is static
 *              <argument_res> type is string or list
 * failure: <name> declaration is not (recursively) found
 *          <name> declaration has unexpected type
 *              error message is printed
 *          <argument_res> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean IteratorContext::getIf_argument_id_ (char *name, boolean * correct, ref2Argument * argument_res)
{
  ref2Identifier
    id_res;
  if (this_localIdentifierList->find (name, &id_res))
    {
      (*argument_res) = new NamedArgument (name, id_res, pk_true);
      return pk_true;
    }
  return this_nestingContext->getIf_argument_id_ (name, correct, argument_res);
}

/*
 * this_nestingContext not NIL (not checked)
 * <id_name> is a (not empty) identifier (not checked)
 * success: <id_name> is an iterator identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is iterated
 *              <argument_res> type is string 
 *          <id_name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *              <argument_res> type is string or list
 *          <id_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the static (StringArgument or ArgumentList) value of this static identifier; 
 *              <argument_res> scope is static
 *              <argument_res> type is string or list
 * failure: <id_name> declaration is not (recursively) found
 *          <id_name> declaration has unexpected type
 *              error message is printed
 *          <argument_res> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean IteratorContext::getIf_argument_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  ref2Identifier
    id_res;
  if (this_localIdentifierList->find (id_name, &id_res))
    {
      (*argument_res) = new NamedArgument (id_res, pk_false);	/* PLG DO NOT ACCEPT GENRIC IDENT IN ITERATOR ? */
      return pk_true;
    }
  return this_nestingContext->getIf_argument_id_ (id_name, correct, argument_res);
}

/* PLG 2005-12-07
 * success: this is safe, all referred to identifiers are parameters or static variables
 *              those referred to identifier list values are expanded
 *          <hasEmptyList> is set to pk_true if some identifier is linked to an empty list
 *          <hasEmptyList> is set to false if all identifiers are linked to not empty lists
 * failure: this is not safe or
 *          this list is NIL or 
 *          some identifier is NIL, in error or has wrong type
 *          system error message is printed
 */
_PUBLIC_ boolean IteratorContext::link_iteratorList_ (boolean * hasEmptyList)
{
  // fprintf(stderr, "DEBUG IteratorContext::link_iteratorList_\n");
  if (IS_SAFE ())
    return this_localIdentifierList->link_iteratorList_ (hasEmptyList);
  OUPS ("illegal call to IteratorContext::link_iteratorList_, this iterator context is not safe");
  return pk_false;
}


_PROTECTED_ _VIRTUAL_ boolean IteratorContext::new_StringFunctionCall (ref2ArgumentList actualParameters, ref2StringFunctionCall * functionCall_res)
{
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new StringFunctionCall (this, actualParameters, &lCorrect);
  return lCorrect;
}

_PROTECTED_ _VIRTUAL_ boolean IteratorContext::new_StringValFunctionCall (ref2Argument actualParameter, ref2StringValFunctionCall * functionCall_res)
{
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new StringValFunctionCall (this, actualParameter, &lCorrect);
  return lCorrect;
}

_PROTECTED_ _VIRTUAL_ boolean IteratorContext::new_HostValFunctionCall (ref2Argument actualParameter, ref2HostValFunctionCall * functionCall_res)
{
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new HostValFunctionCall (this, actualParameter, &lCorrect);
  return lCorrect;
}

_PROTECTED_
  _VIRTUAL_ boolean IteratorContext::new_UserStringFunctionCall (char *functionName, ref2UserStringFunctionElement linkedFunction,
								 ref2ArgumentList actualParameters, ref2UserStringFunctionCall * functionCall_res)
{
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new UserStringFunctionCall (this, functionName, linkedFunction, actualParameters, &lCorrect);
  return lCorrect;
}

_PROTECTED_ _VIRTUAL_ boolean IteratorContext::new_ListFunctionCall (ref2ArgumentList actualParameters, ref2ListFunctionCall * functionCall_res)
{
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new ListFunctionCall (this, actualParameters, &lCorrect);
  return lCorrect;
}

_PROTECTED_ _VIRTUAL_ boolean IteratorContext::new_ListValFunctionCall (ref2Argument actualParameter, ref2ListValFunctionCall * functionCall_res)
{
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new ListValFunctionCall (this, actualParameter, &lCorrect);
  return lCorrect;
}

_PROTECTED_
  _VIRTUAL_ boolean IteratorContext::new_UserListFunctionCall (char *functionName, ref2UserListFunctionElement linkedFunction,
							       ref2ArgumentList actualParameters, ref2UserListFunctionCall * functionCall_res)
{
  // logDEBUG_in("new_UserListFunctionCall");
  boolean
    lCorrect = pk_true;
  (*functionCall_res) = new UserListFunctionCall (this, functionName, linkedFunction, actualParameters, &lCorrect);
  return lCorrect;
}

/* ==========  ==========  ==========  ========== TITLE: READING
 */

/* 2005_01_12
 * reads an iterator list 
 * is assumed to be UNKNOWN (not checked)
 * success: cursor starts list_fSp{[VAR_Identifier]} ')'
 *          read identifiers are declared as list elements in reading context 
 *          cursor is moved to ')'
 * failure: syntax error or some read identifier are not declared neither in this_MKFFileContext nor in contextIdentifiers
 *          some read identifier is not a string list variable
 *          read identifier list may be empty
 *          error message is printed
 */
_VIRTUAL_ boolean
IteratorContext::read_local_identifierList_fSpStar_nSp ()
{
  // log_ASDEBUG("IteratorContext::read_local_identifierList_fSpStar_nSp");
  // fprintf(stderr, "DEBUG IteratorContext::read_local_identifierList_fSpStar_nSp()\n");
  boolean lCorrect = pk_true;
  boolean fullyCorrect = pk_true;
  char *identName;
  ref2Iterator iterator;
  this_localIdentifierList = NIL_IdentifierList;
  do
    {
      if (this_nestingContext->readIf_varIdentification_fSpStar_nSp (&lCorrect, &identName))
	{
	  boolean isFatal;
	  if (this_nestingContext->build_iteratorRefTo_ (identName, &iterator, &isFatal))
	    {
	      if (!append_uniqueNewIterator (identName, iterator))
		fullyCorrect = pk_false;
	    }
	  else
	    fullyCorrect = pk_false;
	}
      else
	{
	  if (lCorrect)
	    {
	      if (this_localIdentifierList == NIL_IdentifierList)
		{
		  ASFatal_emptyIteratorList ();
		  return pk_false;
		}
	      return fullyCorrect;
	    }
	  ASFatal_expected ("list identifier");
	  return pk_false;
	}
    }
  while (pk_true);
}

/* 2004-12-17, 2007-02-07
 * creates a call to a list function which name is in atom
 * success: cursor starts '(' [fileSpace*] [ARGUMENTLIST*] ')' [lineSpaces]
 *          cursor is moved to end of [lineSpaces]
 *          returns in <argument_res> a new command argument containing this call
 *          <isEol> is set according to new cursor position
 *          this  <argument_res> is expanded if these ARGUMENTList are static
 * failure: (fatal) context or syntax error has occurred
 *             error message is printed
 * USED IN read_iterated_argumentCall_nSp(ref2Argument * argument_res, boolean *isEol)
 */
_PRIVATE_
  boolean IteratorContext::create_list_iteratedUserCall (char *functionName, ref2Context readingContext, ref2ArgumentList actualParameters,
							 ref2Argument * argument_res)
{
  // log_ASDEBUG("entering IteratorContext::create_list_iteratedUserCall");
  ref2FunctionElement
    foundFunction;
  boolean
    isList;
  if (find_argumentFunctionElement (functionName, &foundFunction, &isList))
    {
      /* PLG CHECK foundFunction safe */
      if (isList)
	{
	  ref2ListFunctionCall
	    functionCall;
	  if (create_userListFunctionCall
	      (functionName, (ref2UserListFunctionElement) foundFunction, readingContext, actualParameters, &functionCall))
	    {
	      (*argument_res) = new CommandArgument  (pk_true, (ref2ListFunctionCall) functionCall, pk_true);
	      return pk_true;
	    }
	  (*argument_res) = NIL_Argument;
	  return pk_false;
	}
      else
	{
	  ref2StringFunctionCall
	    functionCall;
	  if (create_userStringFunctionCall
	      (functionName, (ref2UserStringFunctionElement) foundFunction, readingContext, actualParameters, &functionCall))
	    {
	      (*argument_res) = new CommandArgument (pk_false, (ref2StringFunctionCall) functionCall, pk_true);
	      return pk_true;
	    }
	  (*argument_res) = NIL_Argument;
	  return pk_false;
	}
    }
  /* failure: error message is printed */
  (*argument_res) = NIL_Argument;
  return pk_false;
}

/* 2005-02-11 PLG TERMINAISON TO BE SET
 * creates an iterated call to a string or list function 
 *         PLG list function iteration should be allowed
 * correct <commandCode> in  { stringCODE, stringvalCODE, hostvalCODE, userCODE, listCODE, listvalCODE }
 * success: cursor starts [Iterator_LIST_fSp] [StringFunction_CALL] [lineSpaces]
 *          where
 *              [StringFunction_CALL] ::= <commandChar> 'string' [fileSpace*] [ARGUMENT_LIST]
 *              [StringFunction_CALL] ::= <commandChar> 'stringval' [fileSpace*] [ARGUMENT_LIST]
 *              [StringFunction_CALL] ::= <commandChar> 'hostval' [fileSpace*] [ARGUMENT_LIST]
 *              [StringFunction_CALL] ::= <commandChar> [userIdentifier] [fileSpace*] [ARGUMENT_LIST]
 *              		       where [userIdentifier] denotes a string or list function,
 *                                     actual argument types and formal argument types must fit
 *          cursor is moved to end of [lineSpaces]
 *          <isEol> is set according to new cursor position
 *          returns in <argument_res> a new command argument containing this call
 * failure: syntax error has occurred
 *             error message is printed
 * USED IN readIf_argument_call_nSp(boolean *correct, ref2Argument * argument_res, boolean *isEol)
 */
/* PLG 2008-06-06 functionName added*/
_PUBLIC_
  boolean IteratorContext::read_list_iteratedCall_argList_spStar_ (int commandCode, char *functionName, ref2Argument * argument_res, boolean * isEol)
{
  // log_ASDEBUG("entering IteratorContext::read_list_iteratedCall_argList_spStar_");
  if (commandCode == iterateCODE)
    {
      this_lexicalContext->ASFatal_named_typeConflict_ln ("iterated iteration not implemented");
      (*argument_res) = NIL_Argument;
      return pk_false;
    }
  else
    {
      ref2ArgumentList
	actualParameters;
      ref2Context
	readingContext = new DynamicContext ((ref2Context) this);	/* to avoid parameter inner call being iterated */
      if (readingContext->ASRequire_O_freeARGList_C_spStar_ (&actualParameters, isEol))
	{
	  boolean
	    isList;
	  if (commandCode == userCODE)
	    return create_list_iteratedUserCall (functionName, readingContext, actualParameters, argument_res);
	  return create_stdArgumentFunctionCall  (pk_true, commandCode, actualParameters, argument_res, &isList);
	}
    }
  (*argument_res) = NIL_Argument;
  return pk_false;
}

/* ===== READ FUNCTION CALL ANCESTOR 1-1: new_ancestorFunctionCall =====
 */
_PROTECTED_
  _VIRTUAL_ boolean IteratorContext::new_AncestorFunctionCall (char *functionName, ref2AncestorFunctionElement linkedFunction,
							       ref2ArgumentList actualParameters, ref2AncestorFunctionCall * functionCall)
{
  boolean
    lCorrect = pk_true;
  (*functionCall) = new AncestorFunctionCall (this, functionName, linkedFunction, actualParameters, &lCorrect);
  return lCorrect;
}

/* ===== READ FUNCTION CALL NODE 1-1: new_elementFunctionCall =====
 */
_PROTECTED_
  _VIRTUAL_ boolean IteratorContext::new_elementFunctionCall (char *functionName, ref2ArgumentList comments,
							      ref2ElementFunctionElement linkedFunction, ref2ArgumentList actualParameters,
							      ref2ElementFunctionCall * functionCall)
{
  boolean
    lCorrect = pk_true;
  (*functionCall) = new ElementFunctionCall (this, functionName, comments, linkedFunction, actualParameters, &lCorrect);
  return lCorrect;
}

/* ==========  ==========  ==========  ========== TITLE: BUILDER
 */

/* 2005-01-10
 * creates an iterator context
 * this iterator context has the same substitution marks as <readingContext>
 * <readingContext> is not NIL  (not checked)
 * reads this context identifier list
 * success: cursor starts '(' [fileSpace*] list_fSp{[VAR_Identifier]} ')' [fileSpace*] [Character_but_Space]
 *          cursor is moved to final [Character_but_Space]
 *          <correct> remains unchanged
 *          this is set as DYNAMIC
 * failure: a syntax error occurred
 *          <correct> is set to false
 *          error message is printed
 */
_PUBLIC_ IteratorContext::IteratorContext (ref2Context readingContext):DynamicContext (readingContext)
{
  // logDEBUG_in("IteratorContext");
}
