_PUBLIC_ _VIRTUAL_ const char *
Iterator::myClassName ()
{
  return "Iterator";
}

/* 2004-12-13
 * returns this iterated actual string value
 * this is supposed to be expanded in current iteration
 */
_VIRTUAL_ char *
Iterator::get_CString ()
{
  return this_iteratedCString;
}

/* ========== Writer
 */

/* ========== Log
 */

/* if this is safe, this_linkedIdentifier is not NIL
 */
_PUBLIC_ _VIRTUAL_ void
Iterator::logExtendedDebug_formula_ln (int level)
{
  if (IS_SAFE ())
    {
      logDebug_objAnnounce_ln (level);
      logDebug_expandedFormula_ln (level);
      log_subTitle_ln (level, "this iterator is linked to the following identifier");
      this_linkedIdentifier->logExtendedDebug_formula_ln (level + 1);
      if (this_iteratedCString == NIL_String)
	log_spn_str_ln (level, "this iterator does not have a string value");
      else
	log_subTitle_ln (level, "this iterator has the following string value", this_iteratedCString);
      if (this_iteratedTail == NIL_ArgumentList)
	log_spn_str_ln (level, "this iterator does not have a current argument tail");
      else
	{
	  log_spn_str_ln (level, "this iterator has the following  argument tail");
	  this_iteratedTail->logExtendedDebug_formula_ln (level + 1);
	}
      logDebug_objClosing_ln (level);
    }
  else
    log_spn_str_ln (level, "this iterator is ## in error ##");
}

_PRIVATE_ void
Iterator::log_iteratorLinkedIdentifier ()
{
  if (this_linkedIdentifier == NIL_Identifier)
    {
      LOG_STR_ ("--{#}--");
    }
  else
    {
      LOG_STR_ ("--{");
      this_linkedIdentifier->log_formula_ ();
      LOG_STR_ ("}--");
    }
}

_PRIVATE_ void
Iterator::log_iteratorCurrentCStringValue ()
{
  if (this_iteratedCString == NIL_String)
    LOG_STR_ ("<string value:#>");
  else
    {
      LOG_STR_SP_STR_SP ("<string value:", this_iteratedCString);
      LOG_STR_ (">");
    }
}

_PRIVATE_ void
Iterator::log_iteratorCurrentTail ()
{
  if (this_iteratedTail == NIL_ArgumentList)
    LOG_STR_ ("<iterator tail:#>");
  else
    {
      LOG_STR_ ("<iterator tail:  ");
      this_iteratedTail->log_formula_ (pk_false);
      LOG_STR_ (">");
    }
}

_PUBLIC_ _VIRTUAL_ void
Iterator::log_formula_ ()
{
  if (IS_SAFE ())
    {
      LOG_STR_ (":loop<...");
      Identifier::log_formula_ ();
      log_iteratorLinkedIdentifier ();
      log_iteratorCurrentCStringValue ();
      log_iteratorCurrentTail ();
      LOG_STR_ ("...>pool:");
    }
  else
    LOG_STR_LN ("this iterator is ## in error ##");
}

/* PLG 2005-12-07
 * this is supposed to be linked to a current expanded string value (not checked)
 * success: this is safe
 *          returns a new m'allocated copy of this current expanded string value in <stringArgument_res>
 * failure: this is not safe
 *          error message is printed
 */
_VIRTUAL_ boolean
Iterator::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  // logDEBUG_in("get_expandedString");
  if (IS_SAFE ())
    {
      (*stringArgument_res) = new StringArgument (this_iteratedCString);
      return pk_true;
    }
  unsafe_named_use (this_name, "Iterator::get_expandedString");
  (*isFatal) = IS_FATAL ();
  return pk_false;
}

/* PLG 2007-02-05
 * success: never
 * failure: always
 *          (fatal)system error message is printed
 */
_VIRTUAL_ boolean
Iterator::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  fatal_named_typeConflict_ln (this_name, "list");
  (*isFatal) = pk_true;
  return pk_false;
}

/* this is supposed to be expanded
 * success: returns this Argument value in <argument_res>
 * failure: this is not safe
 */
_VIRTUAL_ boolean
Iterator::get_expanded (ref2Argument * argument_res)
{
  if (IS_SAFE ())
    {
      (*argument_res) = new StringArgument (this_iteratedCString);
      return pk_true;
    }
  unsafe_use ("Iterator::get_expanded");
  return pk_false;
}

/* 2004-12-13
 * returns this iterated actual string length
 * this is supposed to be expanded in current iteration
 */
_VIRTUAL_ int
Iterator::get_CStringLength_expanded ()
{
  if (this_iteratedCString == NIL_String)
    return 0;
  return strlen (this_iteratedCString);
}

/* if this actual value is not NIL_String
 *    appends, without C string marks, this actual value to target
   target is supposed to be large enough
 * this is supposed to be expanded
 */
_VIRTUAL_ void
Iterator::cat_expanded (char *target)
{
  if (IS_SAFE ())
    if (this_iteratedCString != NIL_String)
      target = strcat (target, this_iteratedCString);
}

/* PLG 2005-12-07
 * if this is safe, this_linkedIdentifier is not NIL
 * success: this referred to identifier is a parameter or a static variable
 *              this referred to identifier is a parameter 
 *                   sets this_argumentValue equal to this_linkedIdentifier actual string list value
 *          <isEmptyList> is set to pk_true if this_argumentValue is an empty list
 *          <isEmptyList> is set to false if this_argumentValue is a not empty list
 * failure: this is in error, or has wrong type
 *          system error message is printed
 */
_VIRTUAL_ boolean
Iterator::link_expandedList_IdVal_ (boolean * isEmptyList)
{
  // logDEBUG_in("link_expandedList_IdVal_");
  if (IS_SAFE ())
    {
      if (this_linkedIdentifier->isParameter ())
	{
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    this_argumentValue = argumentList;
	  else
	    {			/* (fatal)system error message is printed */
	      set_error ();
	      return pk_false;
	    }
	}
      (*isEmptyList) = (((ref2ArgumentList) this_argumentValue) == NIL_ArgumentList);
      return pk_true;
    }
  unsafe_use ("Iterator::link_expandedList_IdVal_");
  return pk_false;
}

/* PLG 2005-12-08
 * this_argumentValue is assumed to be linked to a list argument
 * success: sets this current (sub-)list as this_argumentValue 
 * failure: this is safe, this expanded list value is empty
 *               <correct> remains unchanged;
 *          this is not safe
 *               <correct> becomes pk_false;
 *               system error message is printed
 */
_VIRTUAL_ boolean
Iterator::downCurrentList (boolean * correct)
{
  // logDEBUG_in("downCurrentList", this_name);
  // logDEBUG_in("----------------------------- list value is");
  /* DEBUG this_argumentValue->log_formula_(pk_false); DEBUG */
  // logDEBUG_in("-----------------------------");
  if (IS_SAFE ())
    {
      boolean isFatal;
      if (this_argumentValue == NIL_Argument)
	{
	  (*correct) = pk_false;
	  return pk_false;
	}
      if (this_argumentValue->get_expandedList (&this_iteratedTail, &isFatal))
	{
	  // logDEBUG_in("downCurrentList", this_name);
	  if (this_iteratedTail == NIL_ArgumentList)
	    return pk_false;
	  this_iteratedCString = this_iteratedTail->get_CStringCar ();
	  return pk_true;
	}
    }
  else
    OUPS ("illegal call to Iterator::downCurrentList()");
  (*correct) = pk_false;
  return pk_false;
}

/* PLG 2005-12-08
 * this_argumentValue is assumed to be linked to a list argument
 * success: this current sublist is <e,L>, <L> is not empty
 *          sets this current sublist as <L>
 * failure: 
 *          this is safe, this current sublist has at most 1 element
 *          this is not safe, or this identifier is NIL
 *               system error message is printed
 */
_VIRTUAL_ boolean
Iterator::stepCurrentList ()
{
  // fprintf(stderr,"DEBUG Iterator::stepCurrentList iterator is %s\n", this_name);
  if (IS_SAFE ())
    {
      if (this_iteratedTail == NIL_ArgumentList)
	return pk_false;
      this_iteratedTail = this_iteratedTail->next ();
      if (this_iteratedTail == NIL_ArgumentList)
	return pk_false;
      this_iteratedCString = this_iteratedTail->get_CStringCar ();
      return pk_true;
    }
  OUPS ("illegal call to Iterator::stepCurrentList(), this is not safe");
  return pk_false;
}

/* ========== Generator
 */
/* ========== Builder
 */

/* 2004-12-13
 * creates and returns a new iterator referring to a static named <aName> argument
 * <anArgumentList> is a static list 
 * this expanded list is static <anArgumentList>, refered to by a newly created identifier
 * this expanded CString value is dynamically set in this_iteratedCString  
 *    by stepCurrentList(), downCurrentList() applied to this_argumentValue
 */
Iterator::Iterator (char *aName, ref2ArgumentList anArgumentList):Identifier (aName)
{
  // logDEBUG_in("Iterator(char *, ref2ArgumentList) named", aName);
  this_argumentValue = anArgumentList;
  this_linkedIdentifier = new Identifier (aName, anArgumentList);
  this_iteratedCString = NIL_String;
  this_iteratedTail = NIL_ArgumentList;
}

/* creates and returns a new iterator referring to a dynamic identifier <anIdentifier>
 * <anIdentifier> has string list type (not checked)
 * this this_argumentValue is dynamicallly <anIdentifier>  expanded list
 */
Iterator::Iterator (char *aName, ref2Identifier anIdentifier):
Identifier (aName)
{
  // logDEBUG_in("Iterator(char *, ref2Identifier) named", aName);
  if (anIdentifier == NIL_Identifier)
    set_fatal ();
  else
    {
      this_argumentValue = NIL_ArgumentList;
      this_linkedIdentifier = anIdentifier;
      this_iteratedCString = NIL_String;
      this_iteratedTail = NIL_ArgumentList;
    }
}
