_PUBLIC_ _VIRTUAL_ const char *
ListFunctionCall::myClassName ()
{
  return "ListFunctionCall";
}

/* ========== Attributes
 */

_PUBLIC_ _VIRTUAL_ boolean ListFunctionCall::is_isomorphic (ref2ArgumentFunctionCall arg_funcCall)
{
  if ((ref2ArgumentFunctionCall) this == arg_funcCall)
    return pk_true;
  if (arg_funcCall != NIL_ArgumentFunctionCall)
    {
      return arg_funcCall->is_isomorphic (this);
    }
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean ListFunctionCall::is_isomorphic (ref2ListFunctionCall arg_funcCall)
{
  if (this == arg_funcCall)
    return pk_true;
  return have_isomorphicParameters (arg_funcCall);
}

/* ========== Log
 */

/* ========== Writer
 */

/* ========== Set
 */

/* ========== Printer
 */

/* ========== Reader
 */

/* ========== Generator
 */
/* PLG 2005-12-08
 * should be protected
 * this is supposed to be SAFE (not checked)
 * actual argument list and formal argument list fit
 * generates in this_expandedArgument a new flat list which elements are expanded value of these_actualParameters element values
 * success: may return NIL_ArgumentList (correction not checked)
 * failure: some argument expansion failed, or parameters do not fit
 *          error message is printed
 */
_VIRTUAL_ boolean
ListFunctionCall::single_call ()
{
  // logDEBUG_in("single_call to list");
  // logDEBUG_actualParameters();
  ref2ArgumentList argumentList_res;
  if (IS_EMPTY () || IS_STATIC ())
    return pk_true;
  if (these_actualParameters->getIf_flatten_listExpand_ (&argumentList_res))
    {
      this_expandedArgument = argumentList_res;
      return pk_true;
    }
  return pk_false;
}

/* ========== Grapher
 */

/* ========== Builder
 */

boolean
ListFunctionCall::init_parameterStatus ()
{
  if (HAS_EMPTY_PARAMETER ())
    {
      SET_EMPTY ();
      this_expandedArgument = NIL_Argument;
    }
  else if (HAS_STATIC_PARAMETER ())
    {
      ref2ArgumentList argumentList_res;
      SET_STATIC ();
      if (these_actualParameters->getIf_flatten_listExpand_ (&argumentList_res))
	{
	  this_expandedArgument = argumentList_res;
	  return pk_true;
	}
      set_fatal ();
      return pk_false;
    }
  return pk_true;
}

/* PLG 2005-12-07
 * External entry point
 * if this is safe
 *         actual argument list and formal argument list fit
 * if this is static
 *          this single call value is expanded in this_expandedArgument
 * success: this is safe, runs this list (possibly iterated) function call
 *          returns in <argumentList_res> the computed argument list
 * failure: this is not safe, or function call failed, or result type is not list
 */
_VIRTUAL_ boolean
ListFunctionCall::list_call (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("list_call(ref2ArgumentList *) call to", this_functionName);
  // logDEBUG_actualParameters();
  if (IS_SAFE ())
    {
      boolean lCorrect = pk_true;
      boolean hasEmptyIteratorList;
      if (IS_EMPTY () || IS_STATIC ())
	{
	  (*argumentList_res) = (ref2ArgumentList) this_expandedArgument;
	  return pk_true;
	}
      if (setIterate (&lCorrect, &hasEmptyIteratorList))
	{
	  // logDEBUG_in("list_call(ref2ArgumentList *) Iterated ");
	  if (hasEmptyIteratorList)
	    {
	      (*argumentList_res) = NIL_ArgumentList;
	      return pk_true;
	    }
	  if (call_iteration (&lCorrect, argumentList_res))
	    return pk_true;
	  (*argumentList_res) = NIL_ArgumentList;
	  return lCorrect;
	}
      /* this is not iterated */
      // logDEBUG_in("list_call(ref2ArgumentList *) NOT Iterated ");
      if (lCorrect)
	{
	  if (single_call ())
	    {
	      (*argumentList_res) = (ref2ArgumentList) this_expandedArgument;
	      // logDEBUG_in("list_call(ref2ArgumentList *) NOT Iterated single_call done");
	      return pk_true;
	    }
	  set_fatal ();
	  return pk_false;
	}
    }
  else
    OUPS_name ("illegal ListFunctionCall::list_call(ref2ArgumentList * argumentList_res), unsafe", this_functionName);
  set_fatal ();
  return pk_false;
}

/* creates a new list function call 
 * success: remains correct unchanged
 * failure: set correct to false
 *          error message is printed
 */
/* referred to 
 *    in ListValFunctionCall::
 *             ListValFunctionCall(ref2Context readingContext, ref2Argument actualParameter, boolean *correct)
 *          as ListFunctionCall(readingContext, "listval", actualParameter)
 */
ListFunctionCall::ListFunctionCall (ref2Context readingContext, const char *name, ref2Argument actualParameter):ArgumentFunctionCall (readingContext, pk_true, name,
		      actualParameter)
{
  // logDEBUG_in("ListFunctionCall(ref2Context, const char *, ref2Argument) ");
}

/* referred to 
 *    in UserListFunctionCall::
 *             UserListFunctionCall(ref2Context readingContext, const char * name, ref2UserListFunctionElement aFunction, ref2ArgumentList actualParameters, boolean *correct)
 *          as ListFunctionCall(readingContext, name, actualParameters)
 */
ListFunctionCall::ListFunctionCall (ref2Context readingContext, const char *name, ref2ArgumentList actualParameters):
ArgumentFunctionCall (readingContext, pk_true, name, actualParameters)
{
  // logDEBUG_in("ListFunctionCall(ref2Context, const char *, ref2ArgumentList) ");
}

/* referred to 
 *    in Context::new_ListFunctionCall(ref2ArgumentList actualParameters, ref2ListFunctionCall * functionCall_res)
 *          as new ListFunctionCall(this, actualParameters, &lCorrect)
 */
ListFunctionCall::ListFunctionCall (ref2Context readingContext, ref2ArgumentList actualParameters, boolean * correct):ArgumentFunctionCall (readingContext, pk_true, listWORD,
		      actualParameters)
{
  // logDEBUG_in("ListFunctionCall(ref2Context, ref2ArgumentList, boolean *) ");
  if (!init_parameterStatus ())	/*(fatal)error message is printed, fatal is set */
    (*correct) = pk_false;
}

/* creates a new iterated list function call 
 * success: remains correct unchanged
 *          terminates on line completion following function call closing mark
 * failure: set correct to false
 *          error message is printed
 */
/* referred to 
 *    in ListValFunctionCall::
 *             ListValFunctionCall(ref2IteratorContext readingContext, ref2Argument actualParameter, boolean *correct)
 *          as ListFunctionCall(readingContext, "listval", actualParameter)
 */
ListFunctionCall::ListFunctionCall (ref2IteratorContext readingContext, const char *name, ref2Argument actualParameter):
ArgumentFunctionCall (readingContext, name, actualParameter)
{
  // logDEBUG_in("ListFunctionCall(ref2IteratorContext, const char *, ref2Argument) ");
}

/* referred to 
 *    in UserListFunctionCall::
 *             UserListFunctionCall(ref2IteratorContext readingContext, const char * name, ref2UserListFunctionElement aFunction, ref2ArgumentList actualParameters, boolean *correct)
 *          as ListFunctionCall(readingContext, name, actualParameters)
 */
ListFunctionCall::ListFunctionCall (ref2IteratorContext readingContext, const char *name, ref2ArgumentList actualParameters):
ArgumentFunctionCall (readingContext, name, actualParameters)
{
  // logDEBUG_in("ListFunctionCall(ref2IteratoContext, const char *, ref2ArgumentList ) ");
}

/* creates a new :list function call 
 * success: remains correct unchanged
 *          terminates on line completion following function call closing mark
 * failure: set correct to false
 *          error message is printed
 */
/* referred to 
 *    in IteratorContext::new_ListFunctionCall(ref2ArgumentList actualParameters, ref2ListFunctionCall * functionCall_res)
 *          as new ListFunctionCall(this, actualParameters, &lCorrect)
 */
ListFunctionCall::ListFunctionCall (ref2IteratorContext readingContext, ref2ArgumentList actualParameters, boolean * correct):ArgumentFunctionCall (readingContext, listWORD,
		      actualParameters)
{
  // logDEBUG_in("ListFunctionCall(ref2IteratorContext, ref2ArgumentList, boolean *) ");
  if (!init_parameterStatus ())	/*(fatal)error message is printed, fatal is set */
    (*correct) = pk_false;
}
