#define IS_REDIRECTED() ( this_aliasedNode != NIL_FileNode )

/* A missing context file is not safe
*/

/* ==========  ==========  ==========  ========== TITLE: these attributes
 */
_PUBLIC_ _VIRTUAL_ const char *
MKFFileNode::myClassName ()
{
  return "MKFFileNode";
}

/* =========== THIS NODE ===========
/* ========== Attributes
 */
/* returns pk_true if obj date has been stated
 */
boolean
MKFFileNode::isLoaded ()
{
  return this_isLoaded;
}

/* set obj date as stated
 */
void
MKFFileNode::setLoaded (boolean status)
{
  this_isLoaded = status;
}

_PUBLIC_ ref2Context MKFFileNode::get_innerContext ()
{
  return this_innerContext;
}

/* ==========  ==========  ==========  ========== TITLE: printing
 */
/* prints in <outBuffer> file this context and local declarations
 * <outBuffer> is not NIL_IntePrinter
 */
_VIRTUAL_ void
MKFFileNode::print_declBody (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  this_innerContext->print (outBuffer);
}

/* prints in <outBuffer> file this boot command
 * <outBuffer> is not NIL_IntePrinter
 */
void
MKFFileNode::print_bootCommand (ref2IntePrinter outBuffer, ref2LineList commands)
{
}

/* ==========  ==========  ==========  ========== TITLE: printing info
 */
/* prints object standard body: empty
 */
_VIRTUAL_ void
MKFFileNode::logDebug_declBody (int level)
{
  log_spn_str_ln (level, "out of purpose");
}

/* ==========  ==========  ==========  ========== TITLE: printing info in log
 */
_PRIVATE_ void
MKFFileNode::log_prof_name_sp (int prof)
{
  int ii;
  for (ii = 0; ii < prof; ii = ii + 1)
    LOG_STR_SP ("--");
  print_nameValue_sp (LOGMARKS (), LOGFILE ());
}

/* ==========  ==========  ==========  ========== TITLE: WARNING
 */
/* ==========  ==========  ==========  ========== TITLE: ERROR
 */

_PROTECTED_ void
MKFFileNode::warning_missingFile ()
{
  LOG_LN_WARN_SP ();
  log_fullReference_sp ();
  log_str_ln ("MISSING MKF FILE NOT LOADED");
}

/* ==========  ==========  ==========  ========== TITLE: 
 */
/*  ================================= Constructive nodes
 */

/* =================================
 */

/* ========== AS Errors
 */
_PRIVATE_ int
MKFFileNode::traceContextCycle (ref2MKFFileNode topMKFFile)
{
  int prof = 0;
  if (topMKFFile != this)
    {
      if (this_traveller == NIL_MKFFileNode)
	return -1;
      prof = this_traveller->traceContextCycle (topMKFFile);
    }
  log_prof_name_sp (prof);
  LOG_LN ();
  return prof + 1;
}

_PRIVATE_ void
MKFFileNode::ASError_contextCycle (ref2MKFFileNode loader)
{
  int prof;
  // fprintf(stderr,"ASError_contextCycle in %s, line %d\n",__FILE__,__LINE__);
  /* PLG WARNING this_innerContext may be NIL if the file is errored */
  (this_innerContext->get_lexicalContext ())->ASFatal_contextCycle ();
  // fprintf(stderr,"ASError_contextCycle in %s, line %d\n",__FILE__,__LINE__);
  prof = loader->traceContextCycle (this);
  // fprintf(stderr,"ASError_contextCycle in %s, line %d\n",__FILE__,__LINE__);
  log_prof_name_sp (prof);
  LOG_LN ();
  // fprintf(stderr,"ASError_contextCycle in %s, line %d\n",__FILE__,__LINE__);
  set_fatal ();
}

/* ========== Grapher
 */
/*
 * No ancestor for a make file
 *     
 */

/* ========== Generator
 */

/*
 * Nothing to generate for a MKF file
 *     
 */
_VIRTUAL_ boolean
MKFFileNode::generate (ref2NodeElement curBuiltFile, ref2IntePrinter outBuffer)
{
  if (IS_SAFE ())
    {
      levelInfo_comment_ref_sp__ln ("generate access to a make file");
      return pk_true;
    }
  errored_ln_file_ ("Try to generate access to an errored file");
  return pk_false;
}

/* =========== FILE CONTENT ===========
 */

/* ========== Standard object
 */

/* PLG 2009-07-08
 * success: FILE = [fileWhite*].{[Comment_string][eol][fileWhite*]}* opt{[SyntaxAdaptor_fSpStar]} [ELEMENT_fSpStar]* [EOF]#
 * failure: FILE = [fileWhite*].[Any+][EOF]#, [Any+][EOF]# in error
 *               this object is set errored HERE
 *               error messages have been printed
 */
_PRIVATE_ boolean MKFFileNode::close_read_eof (boolean isEof, ref2MKFFileNode loader)
{
  // fprintf(stderr, "DEBUG MKFFileNode::close_read_eof for  <<<%s>>> \n", get_name()); 
  setLoaded (pk_true);
  /* FILE = .[fileWhite*]#{[Comment_string][eol][fileWhite*]}* opt{[SyntaxAdaptor_fSpStar]}[ELEMENT_fSpStar]*[EOF]
     a character but space starts {[Comment_string][eol][fileWhite*]}* opt{[SyntaxAdaptor_fSpStar]}[ELEMENT_fSpStar]*[EOF] */
  if (isEof)
    {
      this_innerContext->closeFileReading ();
      return pk_true;
    }
  this_traveller = loader;
  if (this_innerContext->close_read_eof (&these_commentLines))
    {
      this_traveller = NIL_MKFFileNode;
      return pk_true;
    }
  reset_safetyLevel (this_innerContext->get_safetyLevel ());
  return pk_false;
}

/* PLG 2009-07-08 OK
 * creates a new buffer in this_context,
 *     opens, reads, analyzes, closes this and deletes associated buffers
 * <loader> is the context associated with this (if this is the root)
 * success: FILE = .[fileSpace*] opt{[SyntaxAdaptor_fSpStar]} [ELEMENT_fSpStar]* [EOF]#
 * failure: error messages have been printed
 */
_PRIVATE_ boolean MKFFileNode::close_read_open_MKFFile (ref2MKFFileNode loader)
{
  // fprintf(stderr, "DEBUG MKFFileNode:: close_read_open_MKFFile for  <<<%s>>> \n", get_name()); 
  this_isUsed = pk_true;
  boolean
    isEof = pk_false;
  if (this_innerContext->fileOpen_fWhStar_ (this_filePathName, &isEof))	/* HERE 1.2 */
    /*  FILE = .[fileWhite*]#{[Comment_string] [eol][fileWhite*]}* opt{[SyntaxAdaptor_fSpStar]}[ELEMENT_fSpStar]*[EOF] */
    return close_read_eof (isEof, loader);
  return pk_false;
}

/* PLG 2009-07-08 OK
 * ENTRY POINT from Context inner context loading
 *      <loader> is the MKF file context loading this file
 *      <contextTree> is the context in which this context tree has to be put
 *     opens, reads, analyzes, closes this, deletes associated buffers
 * success: FILE = .[fileSpace*] opt{[SyntaxAdaptor_fSpStar]} [ELEMENT_fSpStar]* [EOF]#
 *             [EOF] is [eof](system end of file) or {[eol]+[sol]}'^'opt{[eol][Any*]}[eof], [eol][Any*] is ignored
 *               where [sol] is file start
 * failure: FILE is in error, error message is printed
 */
_PUBLIC_ boolean MKFFileNode::loadIf_innerFileContext (ref2ContextTree contextTree, ref2MKFFileNode loader, boolean * newFile)
{
  // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
  if (IS_SAFE ())
    {
      // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
      if (IS_REDIRECTED ())
	return ((ref2MKFFileNode) (get_redirection ()))->loadIf_innerFileContext (contextTree, loader, newFile);
      /* created context should not be deleted; it is used in error messages */
      // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
      if (this_traveller == NIL_MKFFileNode)
	{
	  // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
	  contextTree->register_context (this_innerContext->get_contextTree ());
	  // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
	  if (isLoaded ())
	    {
	      // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
	      this_innerContext->levelInfo_loadedAnnounce_ln ();
	      return pk_true;
	    }
	  (*newFile) = pk_true;
	  if (close_read_open_MKFFile (loader))
	    return pk_true;
	  set_fatal ();
	  return pk_false;
	}
      // fprintf(stderr,"loadIf_innerFileContext in %s, line %d\n",__FILE__,__LINE__);
      ASError_contextCycle (loader);
      return pk_false;
    }
  errored_ln_file_ ("Try to load errored file");	//this_name, this_filePathName
  return pk_false;
}

/* PLG 2009-07-07 OK
 * creates a new context
 * sets this_context as a new static context associated with this
 *     sets this_context printing Buffer as NIL
 *     creates a new reading buffer in this_context,
 *     opens, reads, analyzes, closes this, deletes associated buffers
 *     remove created context
 * <loader> is the context associated with this (if this is the root)
 * success: FILE = .[fileSpace*] opt{[SyntaxAdaptor_fSpStar]} [ELEMENT_fSpStar]* [EOF]#
 *             [EOF] is [eof](system end of file) or {[eol]+[sol]}'^'opt{[eol][Any*]}[eof], [eol][Any*] is ignored
 *               where [sol] is file start
 * failure: FILE is in error
 */
_PUBLIC_ boolean MKFFileNode::load_mainFileContext (ref2ContextTree * rootContextTree)
{
  (*rootContextTree) = this_innerContext->get_contextTree ();
  return close_read_open_MKFFile (this);
}

/* PLG 2009-07-07 OK
 * creates a new buffer in this_context reading from stdin
 *     reads, analyzes this and deletes associated buffers
 * success: FILE = .[fileSpace*] opt{[SyntaxAdaptor_fSpStar]} [ELEMENT_fSpStar]* {[eol]+[sol]}'^'[eol]#
 * failure: FILE is in error
 */
_PUBLIC_ boolean MKFFileNode::load_stdinContext (ref2ContextTree * rootContextTree, char *stdinName)
{
  boolean
    isEof = pk_false;
  (*rootContextTree) = this_innerContext->get_contextTree ();
  if (this_innerContext->stdinOpen_fWhStar_ (stdinName, &isEof))
    /*  FILE = .[fileWhite*]# {[Comment_string][eol][fileWhite*]}* opt{[SyntaxAdaptor_fSpStar]}[ELEMENT_fSpStar]*[eol'^'[eol] */
    return close_read_eof (isEof, this);
  return pk_false;
}

/* ========== File content grapher
 */
/* ========== Grapher
 */

/* PLG 2006-01-09
 * <context_time> is the time of the file which building can be decided when looking at this
 * <comingFrom> ignored
 * success: this is safe
 *          this_date is assigned to <updatedDate>
 * failure: this is not safe
 * <cycleRoot> remain unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean MKFFileNode::set_intrinsicUpdate (time_t contextTime, ref2NodeElement comingFrom,
						      time_t * updatedDate, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG MKFFileNode::set_intrinsicUpdate name is %s, pathname is %s\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_intrinsicUpdate (contextTime, comingFrom, updatedDate, cycleRoot);
  if (IS_SAFE ())
    {
      levelInfo_comment_presentAncestor_ln (contextTime);
      (*updatedDate) = GET_DATE ();
      return pk_true;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* PLG 2006-01-11
 * assumes this_buildingData not NIL when this is defined and safe
 * <creator> ignored
 * <contextTime> is the time of the file which building can be decided when looking at this
 * <comingFrom> ignored
 * success: this is safe
 *          <updatedDate> is assigned OLDEST
 *          <isToBeBuilt> remains unchanged
 * failure: this is not safe
 * <cycleRoot> remians unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean MKFFileNode::set_standardUpdate (ref2NodeElement creator, time_t contextTime, ref2NodeElement comingFrom,
						     time_t * updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG MKFFileNode::set_standardUpdate name is %s, pathname is %s\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_standardUpdate (creator, contextTime, comingFrom, updatedDate, isToBeBuilt, cycleRoot);
  if (IS_SAFE ())
    {
      levelInfo_comment_presentAncestor_ln (contextTime);
      (*updatedDate) = GET_DATE ();
      return pk_true;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* 
 * <creator> ignored
 * <context_time> is the time of the file which building can be decided when looking at this
 * <recursive> ignored
 * <comingFrom> ignored
 * <cycleRoot> ignored
 * success: always
 */
_PUBLIC_ _VIRTUAL_ boolean MKFFileNode::set_everUpdate (ref2NodeElement creator, ref2NodeElement comingFrom, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG MKFFFileNode::set_everUpdate name is %s, pathname is %s\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_everUpdate (creator, comingFrom, cycleRoot);
  return NodeElement::set_everUpdate (creator, comingFrom, cycleRoot);
}

/* 
 * <creator> ignored
 * <context_time> is the time of the file which building can be decided when looking at this
 * <recursive> ignored
 * <comingFrom> ignored
 * <cycleRoot> ignored
 * success: always
 */
_PUBLIC_
  _VIRTUAL_ boolean MKFFileNode::set_recursiveEverUpdate (ref2NodeElement creator, ref2NodeElement comingFrom,
							  time_t * updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG MKFFFileNode::set_recursiveEverUpdate name is %s, pathname is %s\n", this_name, this_filePathName);
  if (IS_REDIRECTED ())
    return (get_redirection ())->set_recursiveEverUpdate (creator, comingFrom, updatedDate, isToBeBuilt, cycleRoot);
  return NodeElement::set_recursiveEverUpdate (creator, comingFrom, updatedDate, isToBeBuilt, cycleRoot);
}

/* ========== File content generator
 */
/* ========== File content builder
 */

/* initiates this root file descriptor
   associates a new context
 */
_PUBLIC_ _VIRTUAL_ boolean MKFFileNode::aliasToMkf (ref2Context readingContext, char *name, boolean * isMkf)
{
  (*isMkf) = pk_true;
  return pk_true;
}

/* ======================== on the fly creation
 */
_PRIVATE_ void
MKFFileNode::std_init ()
{
  this_innerContext = new Context (this);
  this_traveller = NIL_MKFFileNode;
  setLoaded (pk_false);
}

/* PLG 2005-09-07
 * <readingContext> is not NIL (not checked)
 *     this and prevNode have the same name and the same host name
 *     <prevNode> is not defined
 *     this is used as declaration
 * 
 * sets this defined, initializes corresponding attributes
 * sets <prevNode> redirected to this
 * initial date is <prevNode> initial date
 * replaces <prevNode> by this in adressing actions
 * success: <correct> remains unchanged, <missing> is set to pk_true if file does not exist
 * failure: set <correct> to pk_false, <missing> remains unchanged
 * USED in ContextTree::getIf_mkfFileDeclaration, prevNode is an existing reference to this
 */
_PUBLIC_ MKFFileNode::MKFFileNode (ref2Context readingContext, ref2FileNode prevNode, boolean * correct, boolean * missing)	/* 1 */
:FileNode (readingContext, prevNode)	/* FN3 */
{
  if (IS_SAFE ())
    if ((IS_MISSING ()))
      {
	warning_missingFile ();
	(*missing) = pk_true;
      }
    else
      std_init ();
  else
    (*correct) = pk_false;
}

/* creates an object and redirects <prevNode> to this object
 * <readingContext> is not NIL (not checked)
 * an existing <prevNode> with a different name refers to the same file
 * <prevNode> is a reference, and is not a MKFFileNode
 * initial date is <prevNode> initial date
 * success: <correct> remains unchanged, <missing> is set to pk_true if file does not exist
 * failure: set <correct> to pk_false, <missing> remains unchanged
 * USED in ContextTree::getIf_mkfFileDeclaration, prevNode is an existing alias reference to this
 */
_PUBLIC_ MKFFileNode::MKFFileNode (ref2Context readingContext, ref2FileNode prevNode, char *name, boolean * correct, boolean * missing)	/* 2 */
  :
FileNode (readingContext, name, prevNode)	/* FN4 */
{
  prevNode->aliasWarning_ (name);
  if (IS_SAFE ())
    if (IS_MISSING ())
      {
	warning_missingFile ();
	(*missing) = pk_true;
      }
    else
      std_init ();
  else
    (*correct) = pk_false;
}

/* creates a context file and redirects <prevNode> to this context file
 * an existing <prevNode> with a different name refers to the same file
 * initial date is <prevNode> initial date
 * success: <correct> remains unchanged, <missing> is set to pk_true if file does not exist
 * failure: set <correct> to pk_false, <missing> remains unchanged
   PLG: used in MKFFileNode
 */
_PUBLIC_ MKFFileNode::MKFFileNode (ref2MKFFileNode prevNode, char *name, boolean * correct, boolean * missing)	/* 3 */
  :
FileNode (prevNode->get_context (), name, (ref2FileNode) prevNode)	/* FN4 */
{
  prevNode->aliasWarning_ (name);
  if (IS_SAFE ())
    if (IS_MISSING ())
      {
	warning_missingFile ();
	(*missing) = pk_true;
      }
    else
      {
	this_innerContext = prevNode->get_innerContext ();
	this_innerContext->reset_ownerFile (this);
	this_traveller = NIL_MKFFileNode;
	setLoaded (prevNode->isLoaded ());
      }
  else
    (*correct) = pk_false;
}

/* creates and returns a new object descriptor named <name>
 * <readingContext> is not NIL (not checked)
 * neither its name nor its full name are shared with existing file
 * initial date is host file modification date if host file exists
 * initial date is APOCALYPSE if host file doesn't exist
 * success: <correct> remains unchanged, <missing> is set to pk_true if file does not exist
 * failure: set <correct> to pk_false, <missing> remains unchanged
   PLG: used in MKFFileNode
 */
_PUBLIC_ MKFFileNode::MKFFileNode (ref2Context readingContext, char *name, char *fullFileName, boolean * correct, boolean * missing)	/* 4 */
  :
FileNode (readingContext, name, fullFileName)	/* FN5 */
{
  if (IS_SAFE ())
    if (IS_MISSING ())
      {
	warning_missingFile ();
	(*missing) = pk_true;
      }
    else
      std_init ();
  else
    (*correct) = pk_false;
}

/* ======================== Root creation
 */

/* creates a context 
 * initial date is GENESIA
 * USED in pKMake (registered)
*/
_PUBLIC_ MKFFileNode::MKFFileNode ()	/* 5 */
:FileNode ()			/* FN6 */
{
  this_isUsed = pk_true;
  std_init ();
}

/* creates a context and recursively reads and returns make files from this root named <name>
 * initial date is host file modification date if host file exists
 * initial date is APOCALYPSE if host file doesn't exist
   success: <correct> remains unchanged
  failure: reading failed
           set <correct> to false
 * USED in pKMake (registered)
 */
_PUBLIC_ MKFFileNode::MKFFileNode (char *name, char *fullPathName, boolean * correct)	/* 6 */
  :
FileNode (name, fullPathName)	/* FN2 */
{
  // fprintf(stderr, "DEBUG MKFFileNode::MKFFileNode(char * name, char * fullPathName, boolean * correct)\n");
  if (IS_SAFE ())
    {
      if (IS_MISSING ())
	{
	  fatal_missingRequired ();
	  (*correct) = pk_false;
	}
      else
	{
	  this_isUsed = pk_true;
	  std_init ();
	}
    }
  else
    (*correct) = pk_false;
}
