#define SET_GLOBAL_DECLARATION() ( this_declarationScope = 0 )
#define SET_FILE_DECLARATION() ( this_declarationScope = 1 )
#define SET_LOCAL_DECLARATION() ( this_declarationScope = 1 )
#define IS_GLOBAL_DECLARATION() ( this_declarationScope == 0 )

_PUBLIC_ _VIRTUAL_ const char *
MetaArgumentElement::myClassName ()
{
  return "MetaArgumentElement";
}

/* ========== Attributes
 */

/* ========== Log
 */

_PROTECTED_ _VIRTUAL_ void
MetaArgumentElement::logDebug_metaCategory (int level)
{
  if (HAS_STRING_TYPE ())
    if (IS_GLOBAL_DECLARATION ())
      {
	log_spn_str_ln (level, "this is a static string meta element");
      }
    else
      {
	log_spn_str_ln (level, "this is a local string meta element");
      }
  else if (IS_GLOBAL_DECLARATION ())
    {
      log_spn_str_ln (level, "this is a static list meta element");
    }
  else
    {
      log_spn_str_ln (level, "this is a local list meta element");
    }
}

_PROTECTED_ _VIRTUAL_ void
MetaArgumentElement::logDebug_body (int level)
{
  if (this_argumentValue == NIL_Argument)
    log_NILComponent_ln (level, "element value");
  else
    {
      log_subTitle_ln (level, "this element value is");
      this_argumentValue->logExtendedDebug_formula_ln (level + 1);
    }
}

/* ========== Writer
 */

_PROTECTED_ _VIRTUAL_ void
MetaArgumentElement::print_genericHeader (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  printingContext->out_argumentType_sp (outBuffer, IS_GLOBAL_DECLARATION (), HAS_LIST_TYPE ());
  this_name->print_varNameFormula (printingContext, outBuffer);
  outBuffer->space ();
}

_VIRTUAL_ void
MetaArgumentElement::print_genericValue (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (this_argumentValue != NIL_Argument)
    this_argumentValue->Cprint_formula (printingContext, outBuffer, pk_true);
}

/* ========== Set
 */

/* ========== Reader
 */

/* ========== Generator
 */

/* builds an argument declaration from this
 * <callingContext> is not NIL (not checked)
 * parameterValues has been assigned to formal parameters
 * remains callingContext unchanged
 * failure: this is not a correct argument element generator
 *               error message is printed
 */
_PROTECTED_ _VIRTUAL_ boolean MetaArgumentElement::build_elementInnerDecl (ref2DynamicContext callingContext, char *name, ref2ArgumentList comments)
{
  if (IS_GLOBAL_DECLARATION ())
    {
      if (HAS_STRING_TYPE ())
	return callingContext->appendNew_staticStringElement (comments, name, this_argumentValue);
      return callingContext->appendNew_staticListElement (comments, name, this_argumentValue);
    }
  else
    {
      ref2Identifier
	identifier;
      if (callingContext->append_uniqueNewIdentifier (name, HAS_LIST_TYPE (), this_argumentValue, &identifier))
	{
	  return identifier->expand_ ();
	}
    }
  return pk_false;
}

/* ========== Builder
 */

/* creates and returns a new object descriptor named <name>
 */
_PUBLIC_ MetaArgumentElement::MetaArgumentElement (ref2DynamicContext readingContext, ref2ArgumentList comments, boolean isString, boolean isGlobal, ref2Argument nameArgument, ref2Argument formula):MetaElement (readingContext, comments,
	     nameArgument)
{
  if (isString)
    {
      SET_STRING_TYPE ();
    }
  else
    {
      SET_LIST_TYPE ();
    }
  if (isGlobal)
    SET_GLOBAL_DECLARATION ();
  else
    SET_LOCAL_DECLARATION ();
  this_argumentValue = formula;
}
