#define PRINT_STRING_VALUE(outBuffer)  outBuffer->print(this_CString_expandedValue)

_PUBLIC_ _VIRTUAL_ const char *
MetaArgument::myClassName ()
{
  return "MetaArgument";
}

/* ========== Attributes
 */

_PUBLIC_ ref2ArgumentList MetaArgument::get_formula ()
{
  return this_formula;
}

_PUBLIC_ _VIRTUAL_ boolean MetaArgument::is_isomorphic (char *str_val)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean MetaArgument::is_isomorphic (ref2Argument str_arg)
{
  if ((ref2Argument) this == str_arg)
    return pk_true;
  if (str_arg == NIL_Argument)
    return pk_false;
  return str_arg->is_isomorphic (this);
}

_PUBLIC_ _VIRTUAL_ boolean MetaArgument::is_isomorphic (ref2StringArgument str_arg)
{
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean MetaArgument::is_isomorphic (ref2MetaArgument str_arg)
{
  if (this == str_arg)
    return pk_true;
  if (str_arg == NIL_MetaArgument)
    return pk_false;
  if (this_formula == NIL_ArgumentList)
    return ((str_arg->get_formula ()) == NIL_ArgumentList);
  return this_formula->is_isomorphic (str_arg->get_formula ());
}

/* ========== Writer
 */

/* 2004-12-09
 * if this is static, prints this actual string value without <stringMark> embracing
 * if this is dynamic prints this actual formula without <stringMark> embracing
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::glue_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  if (IS_NOTEMPTY ())
    if (IS_NOTEMPTY_STATIC ())
      {
	PRINT_STRING_VALUE (outBuffer);
      }
    else if (this_formula != NIL_ArgumentList)
      this_formula->print_parameterizedFormula (printingContext, outBuffer);
}

/* 2004-12-09
 * if this is static prints this actual value embraced without <stringMark> 
 * if this is not static prints this value list embraced without <stringMark> 
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth, boolean inAList)
{
  if (IS_NOTEMPTY ())
    if (IS_NOTEMPTY_STATIC ())
      PRINT_STRING_VALUE (outBuffer);
    else if (this_formula != NIL_ArgumentList)
      this_formula->print_parameterizedFormula (printingContext, outBuffer);
}

/* 2004-12-09
 * prints without <stringMark> not expanded value of this actual generic string 
 * printingContext is a lexical Context
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::glue_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (this_formula != NIL_ArgumentList)
    this_formula->print_parameterizedFormula (printingContext, outBuffer);
}

/* prints not expanded value of this actual generic string;
 * printingContext is a lexical Context
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  if (this_formula != NIL_ArgumentList)
    this_formula->print_parameterizedFormula (printingContext, outBuffer);
}

/* prints not expanded value of this actual generic string embraced with <stringMark>
 * printingContext is a lexical Context
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::Cprint_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)	/* TODO */
{
  // fprintf(stderr, "DEBUG MetaArgument::Cprint_formula\n");
  printingContext->out_stringMark (outBuffer);
  if (this_formula != NIL_ArgumentList)
    this_formula->print_parameterizedFormula (printingContext, outBuffer);
  printingContext->out_stringMark (outBuffer);
}

_VIRTUAL_ void
MetaArgument::print_varNameFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_formula (printingContext, outBuffer, pk_false);
}

/* ========== Log
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::log_formula_ (boolean isStringValue)
{
  // logDEBUG_in("log_formula_");
  if (this_formula == NIL_ArgumentList)
    log_nil ();
  else
    {
      if (isStringValue)
	{
	  log_stringMark ();
	  this_formula->log_parameterizedFormula ();
	  log_stringMark ();
	}
      else
	this_formula->log_parameterizedFormula ();
    }
}

/* prints not expanded value of this actual generic string;
 * printingContext is a lexical Context
 */
_PUBLIC_ _VIRTUAL_ void
MetaArgument::logInnerExtendedDebug_formula_ln (int level)
{
  if (this_formula == NIL_ArgumentList)
    log_NILComponent_ln (level, "argument list");
  else
    this_formula->logDebug_subFormula_ln (level, "argument list");
}

/* ========== Generator
 */

/* 2005-10-06 PLG
 * shoud be private
 * this is supposed to be safe
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this has some identifier occurrence
 *        expands this
 * failure: this expansion failed
 *          error message is printed
 */

_PRIVATE_ boolean MetaArgument::stringExpand_ ()
{
  // fprintf(stderr, "DEBUG MetaArgument::stringExpand_\n");
  if (IS_DYNAMIC ())
    {
      if (this_formula->getIf_CString_cast_listExpand_ (&this_CString_expandedValue))
	return pk_true;
      set_fatal ();
      return pk_false;
    }
  return pk_true;
}

/* PLG 2005-12-07
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is a constant meta string returns this in <stringArgument_res>
 *    this has some identifier occurrence
 *        returns a new m'allocated expansion of this in <stringArgument_res>
 * failure: this expansion failed
 *          error message is printed
 */

_PUBLIC_ _VIRTUAL_ boolean MetaArgument::getIf_stringExpand_ (ref2StringArgument * stringArgument_res)
{
  // fprintf(stderr, "DEBUG MetaArgument::getIf_stringExpand_\n");
  if (IS_SAFE ())
    {
      if (IS_DYNAMIC ())
	{
	  if (this_formula->getIf_CString_cast_listExpand_ (&this_CString_expandedValue))
	    {
	      (*stringArgument_res) = new StringArgument (this_CString_expandedValue);
	      return pk_true;
	    }
	  set_fatal ();
	  return pk_false;
	}
      (*stringArgument_res) = this;
      return pk_true;
    }
  else
    unsafe_use ("getIf_stringExpand_(ref2StringArgument *)");
  return pk_false;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: returns in <argument_res> the value returned by getIf_stringExpand_(argument_res)
 * failure: expansion failed
 *          error message is printed
 */
_PUBLIC_ _VIRTUAL_ boolean MetaArgument::getIf_expand_ (ref2Argument * argument_res)
{
  fprintf (stderr, "DEBUG MetaArgument::getIf_expand_(ref2Argument *)\n");
  if (IS_SAFE ())
    {
      ref2StringArgument
	arg_res;
      if (getIf_stringExpand_ (&arg_res))
	{
	  (*argument_res) = arg_res;
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_expand_(ref2Argument *)");
  return pk_false;
}

/* PLG 2005-12-07
 * success: expands this value
 *         returns in <argumentList_res> a new m'allocated string list equal to this expanded value followed by <tailList>
 * failure: this is in error or flat expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
MetaArgument::getIf_flatten_listCast_expand_ (ref2ArgumentList tailList, ref2ArgumentList * argumentList_res)
{
  // fprintf(stderr, "DEBUG MetaArgument::getIf_flatten_listCast_expand_(ref2ArgumentList, ref2ArgumentList *)\n");
  if (IS_SAFE ())
    {
      if (stringExpand_ ())
	{
	  (*argumentList_res) = new ArgumentList (this_CString_expandedValue, tailList);
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_flatten_listCast_expand_(ref2ArgumentList, ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-12
 * success: this is safe, 
 *          sets this_CString_expandedValue equal to this expansion
 *          returns this expanded value
 * failure: this is not safe, or string expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
MetaArgument::getIf_CString_cast_expand_ (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expand_");
  if (IS_SAFE ())
    {
      if (stringExpand_ ())
	{
	  (*CString_res) = this_CString_expandedValue;
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_CString_cast_expand_(char **)");
  return pk_false;
}

/* PLG 2005-12-08
 * success: expands this
 *          returns in <strList_res>  a new m'allocated CStringList containing this expanded value followed by <tailList> 
 * failure: this is in error or string expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
MetaArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList tailList, ref2CStringList * strList_res)
{
  if (IS_SAFE ())
    {
      if (stringExpand_ ())
	{
	  (*strList_res) = new_CStringList (tailList);
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_CStringList_cast_flatExpand_(ref2CStringList, ref2CStringList *)");
  return pk_false;
}

/* PLG 2005-12-08
 * success: a new m'allocated CStringList containing this expanded value is returned in <strList_res>
 * failure: this is in error or string expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
MetaArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList * strList_res)
{
  if (IS_SAFE ())
    {
      if (stringExpand_ ())
	{
	  (*strList_res) = new_CStringList ();
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_CStringList_cast_flatExpand_(ref2CStringList *)");
  return pk_false;
}

/* PLG 2005-12-08
 * success: expands this value
 *          a new m'allocated LineList containing this expanded value followed by <tailList> is returned in <lineList_res>
 * failure: this is in error or string expansion failed
 *          error message is printed
 */

_VIRTUAL_ boolean
MetaArgument::getIf_lineList_cast_flatExpand_ (ref2LineList tailList, ref2LineList * lineList_res)
{
  // fprintf(stderr, "DEBUG MetaArgument::getIf_lineList_cast_flatExpand_(ref2LineList , ref2LineList *)\n");
  if (IS_SAFE ())
    {
      if (stringExpand_ ())
	{
	  (*lineList_res) = new_LineList (tailList);
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_lineList_cast_flatExpand_(ref2LineList, ref2LineList *)");
  return pk_false;
}

/* PLG 2005-12-08
 * success: a new m'allocated LineList containing this expanded value is returned in <lineList_res>
 * failure: this is in error or string expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
MetaArgument::getIf_lineList_cast_flatExpand_ (ref2LineList * lineList_res)
{
  // fprintf(stderr, "DEBUG MetaArgument::getIf_lineList_cast_flatExpand_(ref2LineList *)\n");
  if (IS_SAFE ())
    {
      if (stringExpand_ ())
	{
	  (*lineList_res) = new_LineList ();
	  return pk_true;
	}
    }
  else
    unsafe_use ("getIf_lineList_cast_flatExpand_(ref2LineList *)");
  return pk_false;
}

/* ========== Reader
 */
/* ========== Builder
 */

/* PLG 2004-12-07
 * creates and returns a new meta string cell 
 * if this is static and not empty, this expanded value is assigned to this_CString_expandedValue
 */
_PUBLIC_ MetaArgument::MetaArgument (ref2ArgumentList argument):StringArgument ()
{
  // fprintf(stderr, "DEBUG MetaArgument::MetaArgument(ref2ArgumentList argument)\n");
  this_formula = argument;
  if (this_formula == NIL_Argument)
    SET_EMPTY ();
  else
    {
      SET_SCOPE (this_formula->get_scope ());
      if (IS_NOTEMPTY_STATIC ())
	{
	  if (!this_formula->getIf_CString_cast_expanded (&this_CString_expandedValue))
	    set_fatal ();
	}
      else
	this_CString_expandedValue = NIL_String;
    }
}

/* destroy this and its successors
 */
_PUBLIC_ MetaArgument::~MetaArgument ()
{
  if (this_formula != NIL_ArgumentList)
    delete
      this_formula;
}
