_PUBLIC_
/* provided that creation requirements are satisfied
 *     this_linkedIdentifier is never NIL
 *     this_name is never NIL
 * a parameter occurrence met during a function declaration is linked on the fly to its declaration
 * thus, when a parameter occurrence is met during a node declaration expansion, 
 *        this_linkedIdentifier holds parameter actual value
 */
_VIRTUAL_ const char *
NamedArgument::myClassName ()
{
  return "NamedArgument";
}

/* ========== Attributes
 */

/* sets this in error
 */
void
NamedArgument::myFatal_expansion (const char *where)
{
  fatal_named_expansion_ln (this_name, "in NamedArgument");
}

void
NamedArgument::myFatal_typeConflict (const char *where, const char *typeStr)
{
  fatal_named_typeConflict_ln (this_name, typeStr);
}

void
NamedArgument::myUnsafe_use (const char *where)
{
  unsafe_named_argumentUse (this_name, "NamedArgument", where);
}

/* returns linked identifier
 * this is supposed to be linked to an identifier
 */
ref2Identifier
NamedArgument::get_identifier ()
{
  return this_linkedIdentifier;
}


/* 2004-12-13
 * returns this linked identifier actual string value 
 * this linked identifier is supposed to be an expanded string (not checked)
 */
_VIRTUAL_ char *
NamedArgument::get_CString ()
{
  return this_linkedIdentifier->get_CString ();
}

/* 2005-10-06 PLG
 * this is supposed to be linked to a string identifier which value is expanded  (not checked)
 * success: returns in <stringArgument_res> this linked identifier actual string value 
 * failure: this does not have string type   
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  // logDEBUG_in("get_expandedString, name is", this_name);
  return this_linkedIdentifier->get_expandedString (stringArgument_res, isFatal);
}

/* 2005-10-06 PLG
 * this is supposed to be linked to a list identifier which value is expanded  (not checked)
 * success: returns in <argumentList_res> this linked identifier actual list value 
 * failure: this does not have list type   
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::get_expandedList (ref2ArgumentList * argumentList_res, boolean * isFatal)
{
  // logDEBUG_in("get_expandedList, name is", this_name);
  return this_linkedIdentifier->get_expandedList (argumentList_res, isFatal);
}

/* 2005-10-06 PLG
 * this is supposed to be safe and linked to a list or string identifier which value is expanded  (not checked)
 * this has string type
 *    returns length of this expanded string
 * this has list type
 *    returns the cumulated length of these element expanded strings
 */
_VIRTUAL_ int
NamedArgument::get_CStringLength_expanded ()
{
  // logDEBUG_in("get_CStringLength_expanded, name is", this_name);
  return this_linkedIdentifier->get_CStringLength_expanded ();
}

_PUBLIC_ _VIRTUAL_ boolean NamedArgument::is_isomorphic (ref2Argument str_arg)
{
  if ((ref2Argument) this == str_arg)
    return pk_true;
  if (str_arg == NIL_Argument)
    return pk_false;
  return str_arg->is_isomorphic (this);
}

_PUBLIC_ _VIRTUAL_ boolean NamedArgument::is_isomorphic (ref2NamedArgument str_arg)
{
  if (this == str_arg)
    return pk_true;
  if (str_arg != NIL_NamedArgument)
    {
      ref2Identifier
	str_arg_ident = str_arg->get_identifier ();
      if (this_linkedIdentifier == str_arg_ident)
	return pk_true;
      return this_linkedIdentifier->is_isomorphic (str_arg_ident);
    }
  return pk_false;
}

/* ========== Writer
 */

/* 2004-12-09
 * prints this actual argument value
 */
_VIRTUAL_ void
NamedArgument::print_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  this_linkedIdentifier->print_expandedValue (printingContext, outBuffer, inAList);
}

/* 2004-12-09
 * if this is static, prints this actual string value without <stringMark> embracing
 * if this is not static prints this actual formula without <stringMark> embracing
 */
_VIRTUAL_ void
NamedArgument::glue_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  this_linkedIdentifier->print_actualFormula (printingContext, outBuffer, depth, pk_false);
}

/* 2004-12-09
 * if this is static, prints this actual string value without <stringMark> embracing
 * if this is not static prints this name
 */
_VIRTUAL_ void
NamedArgument::print_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth, boolean inAList)
{
  this_linkedIdentifier->print_actualFormula (printingContext, outBuffer, depth, inAList);
}

/* prints this name bounded by leftChar and rightChar 
 * outBuffer is supposed to be a correct printer; not checked
 */
_VIRTUAL_ void
NamedArgument::glue_parameterized (ref2Lexical printingContext, ref2IntePrinter outBuffer)	/* DONE */
{
  printingContext->out_parameter (outBuffer, this_name);
}

/* 2004-12-09
 * print this name
 */
_VIRTUAL_ void
NamedArgument::glue_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  outBuffer->print (this_name);
}

/* 2004-12-09
 * print this name
 */
_VIRTUAL_ void
NamedArgument::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  outBuffer->space (this_name);
}


/* 2004-12-09
 * print this name
 */
_VIRTUAL_ void
NamedArgument::Cprint_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer, boolean inAList)
{
  // logDEBUG_in("Cprint_formula, name is", this_name);
  outBuffer->space (this_name);
}

/* ========== Log
 */
/* prints this name bounded by leftChar and rightChar 
 * outBuffer is supposed to be a correct printer; not checked
 */
_VIRTUAL_ void
NamedArgument::glueLog_parameterized ()	/* DONE */
{
  log_parameter (this_name);
}

_PUBLIC_ _VIRTUAL_ void
NamedArgument::log_formula_ (boolean isStringValue)
{
  // logDEBUG_in("log_formula_");
  if (this_linkedIdentifier == NIL_Identifier)
    {
      log_nil ();
    }
  else
    {
      log_openParameter ();
      this_linkedIdentifier->log_formula_ ();
      log_closeParameter ();
    }
}

/* prints this name 
 * outBuffer is supposed to be a correct printer; not checked
 */
_PUBLIC_ _VIRTUAL_ void
NamedArgument::logInnerExtendedDebug_formula_ln (int level)
{
  logDebug_thisNameIs_ln (level, this_name);
  log_subTitle_ln (level, "this argument is linked to the following identifier ");
  this_linkedIdentifier->logExtendedDebug_formula_ln (level + 1);
}

/* ========== Reader
 */
/* ========== Generator
 */

/* PLG 2005-12-12
 * this is supposed to be linked to a list or string identifier which value is expanded  (not checked)
 * success: this has string type
 *             returns in <CString_res> this expanded CString value 
 * success: this has list type
 *             returns in <CString_res> the CString concatenation of strings belonging to this expanded list value 
 * failure: this is errored   
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_CString_cast_expanded (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expanded, name is", this_name);
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  (*CString_res) = this_linkedIdentifier->get_CString ();
	  return pk_true;
	}
      else
	{			/* has list type */
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  (*CString_res) = NIL_String;
		  return pk_true;
		}
	      if (argumentList->getIf_CString_cast_expanded (CString_res))
		return pk_true;
	    }
	}
      set_error ();
      return pk_false;
    }
  myUnsafe_use ("getIf_CString_cast_expanded");
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be linked to a list or string identifier which value is expanded  (not checked)
 * success:
 *    this linked identifier is a string identifier
 *        returns in <strList_res> a new m'allocated singleton containing its expanded CString value 
 *    this linked identifier is a list identifier
 *        returns in <strList_res> a new m'allocated copy of its expanded list
 *            <strList_res> does not have NIL element
 * failure: this is in error
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_CStringList_cast_flat_expanded (ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flat_expanded, name is", this_name);
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  this_CString_expandedValue = this_linkedIdentifier->get_CString ();
	  (*strList_res) = new_CStringList ();
	  return pk_true;
	}
      else
	{			/* has list type */
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  (*strList_res) = NIL_CStringList;
		  return pk_true;
		}
	      if (argumentList->getIf_CStringList_cast_flat_expanded (strList_res))
		return pk_true;
	    }
	  set_error ();
	}
    }
  else
    myUnsafe_use ("getIf_CStringList_cast_flat_expanded");
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be linked to a list or string identifier which value is expanded  (not checked)
 * success:
 *    this linked identifier is a string identifier
 *        returns in <lineList_res> a new m'allocated singleton containing its expanded CString value 
 *    this linked identifier is a list identifier
 *        returns in <lineList_res> a new m'allocated copy of its expanded list
 *            <lineList_res> does not have NIL element
 * failure: this is in error
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_lineList_cast_flat_expanded (ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flat_expanded, name is", this_name);
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  this_CString_expandedValue = this_linkedIdentifier->get_CString ();
	  (*lineList_res) = new_LineList ();
	  return pk_true;
	}
      else
	{			/* has list type */
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  (*lineList_res) = NIL_LineList;
		  return pk_true;
		}
	      if (argumentList->getIf_lineList_cast_flat_expanded (lineList_res))
		return pk_true;
	    }
	  set_error ();
	}
    }
  else
    myUnsafe_use ("getIf_lineList_cast_flat_expanded");
  return pk_false;
}

/* 2005-10-06 PLG
 * this is supposed to be linked to a list or string identifier which value is expanded  (not checked)
 * this value is a string
 *    if this actual value is not NIL_String
 *      appends, without C string marks, this actual value to target
 * this value is a list
 *    appends all not empty expanded string elements belonging to this
 * target is supposed to be large enough
 */
_VIRTUAL_ void
NamedArgument::cat_expanded (char *target)
{
  this_linkedIdentifier->cat_expanded (target);
}

/* PLG 2005-12-07
 * this is supposed to be linked to a list or string identifier which value is expanded  (not checked)
 * if this linked identifier is an iterator, its actual value is its current string argument
 * success: this is a string parameter or an iterator
 *          if this linked identifier is an iterator, its expanded value is its current string argument
 *             returns identifier expanded value in <stringArgument_res>
 * failure: this is not safe or does not have list type, or expansion failed 
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_stringExpand_ (ref2StringArgument * stringArgument_res)
{
  // logDEBUG_in("getIf_stringExpand_, name is", this_name);
  if (HAS_STRING_TYPE ())
    {
      if (IS_SAFE ())
	{
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedString (stringArgument_res, &isFatal))
	    return pk_true;
	  if (isFatal)
	    myFatal_expansion ("getIf_stringExpand_");
	  return pk_false;
	}
      myUnsafe_use ("getIf_stringExpand_ 1");
      return pk_false;
    }
  myFatal_typeConflict ("getIf_stringExpand_", "string");
  return pk_false;
}

/* PLG 2005-12-12
 * this is supposed to be linked to a list or string identifier which value is expanded  (not checked)
 * success: this is safe, this type is list 
 *            returns identifier expanded value in <stringArgument_res>
 * failure: this is not safe, or this is a string 
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_listExpand_ (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_listExpand_, name is", this_name);
  if (HAS_LIST_TYPE ())
    {
      if (IS_SAFE ())
	{
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (argumentList_res, &isFatal))
	    return pk_true;
	  if (isFatal)
	    myFatal_expansion ("getIf_listExpand_");
	  return pk_false;
	}
      myUnsafe_use ("getIf_listExpand_");
      return pk_false;
    }
  myFatal_typeConflict ("getIf_listExpand_", "list");
  return pk_false;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is a string
 *             returns this in <argument_res> the value returned by getIf_stringExpand_(argument_res)
 *          this is list
 *             returns this in <argument_res> the value returned by getIf_listExpand_(argument_res)
 * failure: this is in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_expand_ (ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_expand, name is", this_name);
  if (IS_SAFE ())
    {
      boolean isFatal;
      if (HAS_STRING_TYPE ())
	{
	  ref2StringArgument stringArgument;
	  if (this_linkedIdentifier->get_expandedString (&stringArgument, &isFatal))
	    {
	      (*argument_res) = stringArgument;
	      return pk_true;
	    }
	}
      else
	{			/* has list type */
	  ref2ArgumentList argumentList;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      (*argument_res) = argumentList;
	      return pk_true;
	    }
	}
      if (isFatal)
	myFatal_expansion ("getIf_expand_");
      return pk_false;
    }
  myUnsafe_use ("getIf_expand_");
  return pk_false;
}

/* 2005-10-06 PLG
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is dynamic
 *              expands these string values
 *              returns in <argumentList_res> a flat copy of these expanded string values (including NIL_String)
 *                         followed by flat list <tailList>
 *          this is static 
 *              returns in <argumentList_res> a flat copy of these string values (including NIL_String)
 *                         followed by flat list <tailList>
 * failure: this is a string or in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_flatten_listCast_expand_ (ref2ArgumentList tailList, ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listCast_expand_ 1, name is", this_name);
  if (IS_SAFE ())
    {
      boolean isFatal;
      if (HAS_STRING_TYPE ())
	{
	  ref2StringArgument stringArgument;
	  if (this_linkedIdentifier->get_expandedString (&stringArgument, &isFatal))
	    {
	      (*argumentList_res) = new ArgumentList (stringArgument, tailList);
	      return pk_true;
	    }
	}
      else
	{			/* has list type */
	  ref2ArgumentList argumentList;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  (*argumentList_res) = tailList;
		  return pk_true;
		}
	      if (argumentList->getIf_flatten_listCast_expand_ (tailList, argumentList_res))
		return pk_true;
	    }
	}
      if (isFatal)
	myFatal_expansion ("getIf_flatten_listCast_expand_");
      return pk_false;
    }
  myUnsafe_use ("getIf_flatten_listCast_expand_");
  return pk_false;
}

/* PLG 2005-12-12
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is a list parameter
 *    this is a flat constant ArgumentList
 *         returns this in <stringArgument_res>
 *    this is not a flat constant 
 *         returns a new m'allocated flat expansion of this in <argumentList_res>
 * failure: this is a string or in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_flatten_listExpand_ (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listExpand_ 2, name is", this_name);
  if (HAS_LIST_TYPE ())
    {
      if (IS_SAFE ())
	{
	  boolean isFatal;
	  ref2ArgumentList argumentList;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		(*argumentList_res) = NIL_ArgumentList;
	      else
		(*argumentList_res) = argumentList->get_flatten_expandedList ();
	      return pk_true;
	    }
	  if (isFatal)
	    myFatal_expansion ("getIf_flatten_listExpand_");
	  return pk_false;
	}
      myUnsafe_use ("getIf_flatten_listExpand_");
      return pk_false;
    }
  myFatal_typeConflict ("getIf_flatten_listExpand_", "list");
  return pk_false;
}

/* PLG 2005-12-12
 * success: this has string type 
 *             sets this_CString_expandedValue equal to this expansion
 *             returns this expanded value
 *          this has list type 
 *             returns a new StringArgument equal to the concatenation of these expanded elements
 * failure: this is errored or has unknown type
 *          error message is printed
 *          return NIL_Argument
 */
_VIRTUAL_ boolean
NamedArgument::getIf_CString_cast_expand_ (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expand_, name is", this_name);
  if (IS_SAFE ())
    {
      if (HAS_LIST_TYPE ())
	{
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  this_CString_expandedValue = NIL_String;
		  (*CString_res) = this_CString_expandedValue;
		  return pk_true;
		}
	      if (argumentList->getIf_CString_cast_expanded (&this_CString_expandedValue))
		{
		  (*CString_res) = this_CString_expandedValue;
		  return pk_true;
		}
	      return pk_false;
	    }
	  if (isFatal)
	    myFatal_expansion ("getIf_CString_cast_expand_");
	  return pk_false;
	}
      /* has string type */
      this_CString_expandedValue = this_linkedIdentifier->get_CString ();
      (*CString_res) = this_CString_expandedValue;
      return pk_true;
    }
  myUnsafe_use ("getIf_CString_cast_expand_");
  return pk_false;
}

/* 2005-10-06 PLG
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is dynamic
 *              expands these string values
 *              returns in <argumentList_res> a flat copy of these expanded string values (including NIL_String)
 *                         followed by flat list <tailList>
 *          this is static 
 *              returns in <argumentList_res> a flat copy of these string values (including NIL_String)
 *                         followed by flat list <tailList>
 * failure: this is a string or in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList tailList, ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flatExpand_ 1, name is", this_name);
  if (IS_SAFE ())
    {
      ref2Argument argument;
      if (this_linkedIdentifier->get_expanded (&argument))
	{
	  if (argument == NIL_Argument)
	    {
	      (*strList_res) = tailList;
	      return pk_true;
	    }
	  if (argument->getIf_CStringList_cast_flatExpand_ (tailList, strList_res))
	    return pk_true;
	}
      return pk_false;
    }
  myUnsafe_use ("getIf_CStringList_cast_flatExpand_");
  return pk_false;
}

/* 2005-03-11
 * success: returns in <strList_res> a singleton CString list which single value is this expanded value list
 * failure: this is in error or this linked identifier returns NIL_Argument (is assumed to be in error)
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList * strList_res)
{
  // logDEBUG_in("getIf_CStringList_cast_flatExpand_ 2, name is", this_name);
  if (IS_SAFE ())
    {
      if (HAS_LIST_TYPE ())
	{
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  (*strList_res) = NIL_CStringList;
		  return pk_true;
		}
	      return argumentList->getIf_CStringList_cast_flat_expanded (strList_res);
	    }
	  if (isFatal)
	    myFatal_expansion ("getIf_CStringList_cast_flatExpand_");
	  return pk_false;
	}
      /* has string type */
      this_CString_expandedValue = this_linkedIdentifier->get_CString ();
      (*strList_res) = new_CStringList ();
      return pk_true;
    }
  myUnsafe_use ("getIf_CStringList_cast_flatExpand_");
  return pk_false;
}

/* 2005-10-06 PLG
 * identifier reference occurring in this definition are supposed to be expanded
 * success: this is dynamic
 *              expands these string values
 *              returns in <argumentList_res> a flat copy of these expanded string values (including NIL_String)
 *                         followed by flat list <tailList>
 *          this is static 
 *              returns in <argumentList_res> a flat copy of these string values (including NIL_String)
 *                         followed by flat list <tailList>
 * failure: this is a string or in error or expansion failed
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_lineList_cast_flatExpand_ (ref2LineList tailList, ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_ 1, name is", this_name);
  if (IS_SAFE ())
    {
      ref2Argument argument;
      if (this_linkedIdentifier->get_expanded (&argument))
	{
	  if (argument == NIL_Argument)
	    {
	      (*lineList_res) = tailList;
	      return pk_true;
	    }
	  if (argument->getIf_lineList_cast_flatExpand_ (tailList, lineList_res))
	    return pk_true;
	}
      return pk_false;
    }
  myUnsafe_use ("getIf_lineList_cast_flatExpand_");
  return pk_false;
}

/* 2005-03-11
 * linked identifier is assumed to be (flat) expanded
 * success: returns in <lineList_res> the line list conversion of this linked identifier expanded value
 * failure: this is in error 
 *          error message is printed
 */
_VIRTUAL_ boolean
NamedArgument::getIf_lineList_cast_flatExpand_ (ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_ 2, name is", this_name);
  if (IS_SAFE ())
    {
      if (HAS_LIST_TYPE ())
	{
	  ref2ArgumentList argumentList;
	  boolean isFatal;
	  if (this_linkedIdentifier->get_expandedList (&argumentList, &isFatal))
	    {
	      if (argumentList == NIL_ArgumentList)
		{
		  (*lineList_res) = NIL_LineList;
		  return pk_true;
		}
	      return argumentList->getIf_lineList_cast_flat_expanded (lineList_res);
	    }
	  if (isFatal)
	    myFatal_expansion ("getIf_lineList_cast_flatExpand_");
	  return pk_false;
	}
      /* has string type */
      this_CString_expandedValue = this_linkedIdentifier->get_CString ();
      (*lineList_res) = new_LineList ();
      return pk_true;
    }
  myUnsafe_use ("getIf_lineList_cast_flatExpand_");
  return pk_false;
}

/* ========== Builder
 */

/* 2004-12-13
 * <name> is not NIL (not checked)
 * <ident> is an existing (not NIL) string parameter identifier (not checked)
 * creates a reference to <ident>
 */
NamedArgument::NamedArgument (ref2Identifier ident, boolean isList):Argument ()
{
  this_name = "generic name";
  this_linkedIdentifier = ident;
  SET_DYNAMIC ();
  if (isList)
    SET_LIST_TYPE ();
  else
    SET_STRING_TYPE ();
}

/* 2004-12-13
 * <name> is not NIL (not checked)
 * <ident> is an existing (not NIL) string parameter identifier (not checked)
 * creates a reference to <ident>
 */
NamedArgument::NamedArgument (char *name, ref2Identifier ident):
Argument ()
{
  this_name = name;
  this_linkedIdentifier = ident;
  SET_DYNAMIC ();
  SET_STRING_TYPE ();
}

/* 2004-12-13
 * <ident> is an existing (not NIL) list identifier (not checked)
 * <name> is not NIL (not checked)
 * creates a reference to <ident>, which name is <name> 
 * if <isIterator> is pk_true, this argument type is string 
 * if <isIterator> is pk_false, this argument type is list 
 */
NamedArgument::NamedArgument (char *name, ref2Identifier ident, boolean isIterator):
Argument ()
{
  this_name = name;
  this_linkedIdentifier = ident;
  if (isIterator)
    {
      SET_ITERATED ();
      SET_STRING_TYPE ();
    }
  else
    {
      SET_DYNAMIC ();
      SET_LIST_TYPE ();
    }
}
