/* CONSTRUCTOR CONTEXT
 *   YY is a direct file or an executable variable in CC constructor context
 *       YY-CC constructor context is YY constructor context
 *   YY is neither a direct file nor an executable variable
 *       YY-CC constructor context is CC constructor context
 * DATE ORDERING
 *   GENESIA < actual dates <= now < FUTURE < APOCALYPSE
 * node YY INITIAL DATE
 *    YY is a file present in host
 *       YY initial date is its host date
 *    YY is a file missing in host
 *       YY initial date is the universal newest date (APOCALYPSE)
 *    YY is an extern or a variable node 
 *       YY initial date is the universal oldest date
 * node YY INTRINSIC UPDATE 
 *    YY is a file missing in host
 *       YY updated date is FUTURE
 *       YY is to be built
 *    YY is a file present in host that does not have ancestor, 
 *    or YY is a variable that does not have ancestor
 *    or YY is an extern node 
 *       YY updated date is its initial date
 *       YY is not to be built
 *    YY is a file present in host, YY has ancestors
 *       YY is strictly older than some YY ancestor updated date, wrt to modal dependence
 *          YY updated date is FUTURE
 *          YY is to be built
 *       YY is newer than all ancestor updated dates, wrt to modal dependence
 *          YY updated date is YY initial date
 *          YY is not to be built
 *    YY is a variable, YY has ancestors 
 *       some YY ancestor is to be built wrt to modal dependence
 *          YY updated date is FUTURE
 *          YY is to be built
 *       all ancestors are no to be built wrt to modal dependence
 *          YY updated date is the newest date of its ancestor updated dates
 *          YY is not to be built
 * node XX MODAL UPDATE in CC constructor context
 *    when modality is
 *       active missing (set_missingUpdate)
 *          XX is a file missing in host, or a variable that (recursively) has file missing ancestor(has_missingAncestorFile)
 *             XX is set with ever modality in XX-CC constructor context
 *          other
 *             XX is not set
 *       recursive active missing to be built !!
 *       never (set_intrinsicUpdate)
 *          XX MODAL UPDATE date is XX INTRINSIC UPDATE date in NIL constructor context
 *       standard (set_standardUpdate)
 *          XX MODAL UPDATE is XX INTRINSIC UPDATE in XX-CC constructor context 
 *       ever (set_everUpdate)
 *          XX ancestors are set with standard modality in XX-CC constructor context
 *          XX MODAL UPDATE date is XX INTRINSIC UPDATE date in XX-CC constructor context
 *          XX is to be built
 *       recursive ever (set_recursiveEverUpdate)
 *          XX ancestors are set with recursive ever modality in XX-CC constructor context
 *          XX MODAL UPDATE date is XX INTRINSIC UPDATE date
 *          XX is to be built
 * DEPENDENCE UPDATE and BUILDING RULES
 *    when XX refers to an ancestor YY in CC constructor context, with following modality
 *    passive missing
 *       YY is a missing file or a variable referring to some missing file (has_missingAncestorFile)
 *           XX_ancestor updated date is FUTURE
 *       other
 *           XX_ancestor updated date is not affected
 *    recursive active missing to be built !!
 *    active missing
 *       YY is (conditionally) set with active missing modality in YY-CC constructor context
 *       YY is a missing file or a variable referring to some missing file
 *           XX_ancestor updated date is FUTURE
 *           YY-CC is not NIL
 *              YY-CC is the root constructor for YY
 *       other
 *           XX_ancestor updated date, XX_ancestor status are not affected
 *    never OK
 *       YY INTRINSIC UPDATE date is strictly newer than current XX_ancestor updated date(set_intrinsicUpdate)
 *          current XX_ancestor updated date becomes YY INTRINSIC UPDATE date 
 *       YY INTRINSIC UPDATE date is older than current XX updated date
 *          current XX_ancestor updated date is not affected
 *    standard
 *       YY is set with standard modality in YY-CC constructor context
 *       YY updated date is newer than current XX_ancestor updated date
 *           current XX_ancestor updated date is set to YY updated date
 *       YY updated date is older than current XX updated date
 *           current XX_ancestor updated date is not affected
 *       YY-CC is not NIL
 *           YY-CC is the root constructor for YY
 *    ever
 *       YY is set with ever modality in YY-CC constructor context
 *       YY status is set to be built
 *       XX_ancestor updated date is set to FUTURE
 *       YY-CC is not NIL
 *           YY-CC is the root constructor for YY
 *    recursive ever
 *       YY is set with recursive ever modality in YY-CC constructor context
 *       YY status is (recursively) set to be built
 *       XX updated date is set to FUTURE
 *       YY-CC is not NIL
 *           YY-CC is the root constructor for YY

 * PLG 2006-01-24 
 * constructive flow 
 *      constructive policies are --> standard, constructive missing, ever and recursive ever 
 *      XX is a never variable or an extern node
 *         XX never is in any constructive flow 
 *      XX is a direct file, or XX is an executable variable, selected as root generators
 *         XX is in XX-constructive flow 
 *      if YY is in XX-constructive flow with ancestor ZZ
 *         if ancestor policy is  a constructive policy
 *            if ZZ is a direct file, or ZZ is an executable variable
 *               ZZ is in ZZ-constructive flow 
 *            if ZZ is a buildable variable or a standard file
 *               ZZ is in XX-constructive flow 
 *         if ancestor policy is never 
 *               ZZ is in XX-update flow 
 * update flow
 *      update policies are --> standard, never, ever and recursive ever 
 *      XX is a never variable or an extern node
 *         XX never is in any update flow 
 *      XX is a direct file, or XX is an executable variable, selected as root generators
 *         XX is in XX-update flow 
 *      if YY is in XX-update flow with ancestor ZZ
 *         if ancestor policy is an update policy
 *            ZZ is in XX-update flow 
 *        
 * a recursive flow is a constructive flow
 *      recursive policies are --> standard, ever and recursive ever 
 *      XX is a never variable or an extern node
 *         XX never is in any recursive flow 
 *      XX is a constructive generator, selected as recursive
 *         XX is in XX-recursive flow 
 *      if YY is in XX-recursive flow with ancestor ZZ
 *         if ancestor policy is a recursive policy
 *            if ZZ is a direct file, or ZZ is an executable variable
 *               ZZ is in ZZ-recursive flow 
 *            if ZZ is a buildable variable or a standard file
 *               ZZ is in XX-recursive flow 
 *         if ancestor policy is never
 *            ZZ is in XX-update flow 
 *      if YY is in XX-constructive flow with ancestor ZZ
 *         if ancestor policy is recursive
 *            if ZZ is a direct file, or ZZ is an executable variable
 *               ZZ is in ZZ-recursive flow 
 *            if ZZ is a buildable variable or a standard file
 *               ZZ is in XX-recursive flow 
 *             
 *        
 * node status
 *      XX does not belong to any YY-constructive flow 
 *         XX constructor is NIL
 *      XX belongs to some YY-constructive flow 
 *         XX constructor is one of those YY such that XX belongs to YY-constructive flow 
 *
 *      FUTURE is younger that all other dates
 *      XX is a missing file in an updated or a constructive flow
 *         => XX updated date is FUTURE
 *      XX is a present file in an constructive flow, XX has YY ancestor, YY updated date is FUTURE
 *         => XX updated date is FUTURE
 *      XX is a present file in an updated flow, XX has YY ancestor, YY updated date is younger than XX host date
 *         => XX updated date is FUTURE
 *      XX is a variable in a constructive flow, XX has YY ancestor, YY updated date is FUTURE
 *         => XX updated date is FUTURE
 *      XX is a present file in an updated flow
 *         => XX updated date is the yungest of its ancestor updated dates 
 *
 *      XX is in a recursive flow
 *         XX is to be built
 *      XX is in a constructive flow
 *         XX updated date is FUTURE, XX is to be built
 *         XX has an ancestor YY, 
 *            ancestor policy is ever 
 *              YY is to be built
 *              XX is to be built
 *            ancestor policy is never missing, 
 *              YY is a missing file or YY is a variable which some ancestor has a missing file
 *                 XX is to be built
 *         
 */

/* if this is to be built, then this has been visited
 */

/* PLG the creator may not be well managed; currently, it is a potential one
 */

#define HAS_ANCESTOR() ( this_buildingData->has_ancestor() )	/* PRE: this_isDefined is pk_true */
#define GET_ANCESTORS() this_buildingData->get_ancestors()	/* PRE: this_isDefined is pk_true */

#define UNLOCK() made = NIL_NodeElement
#define LOCK_BY(locker) made = locker
#define GET_LOCKER() made	/* PRE: this_isDefined is pk_true */
#define IS_LOCKED() ( made != NIL_NodeElement )	/* PRE: this_isDefined is pk_true */
#define IS_LOCKED_BY_THIS() ( made == this )	/* PRE: this_isDefined is pk_true */

#define SET_NULLCREATOR()  SET_UPDATED(); this_creator = NIL_NodeElement	/* PRE: this_isDefined is pk_true, cycle detection is achieved */
#define SET_ACTUALCREATOR(CREATOR) SET_UPDATED(); this_creator = CREATOR	/* PRE: this_isDefined is pk_true, cycles detection is achieved */
#define HAS_CREATOR() ( this_creator != NIL_NodeElement )	/* PRE: this_stepIndex is not smaller than the_stepIndex */
#define HASNOT_CREATOR() ( this_creator == NIL_NodeElement )	/* PRE: this_stepIndex is not smaller than the_stepIndex */


#define SET_TOBEBUILT() this_buildingRequired = pk_true;	/* PRE: initialized when this is created */
#define IS_TOBEBUILT() this_buildingRequired

/* PRE: this_isDefined is pk_true, cycles detection is achieved */
_PUBLIC_ _VIRTUAL_ const char *
NodeElement::myClassName ()
{
  return "Node";
}

/* ==========  ==========  (FATAL) ERROR
 */

/* ==========  ==========  Attributes
 */

/* returns obj creation or updated date
 */
_PUBLIC_ ref2BuildingData NodeElement::get_buildingData ()
{
  return this_buildingData;
}

/* returns obj creation or updated date
 */
_PUBLIC_ time_t NodeElement::get_date ()
{
  return GET_DATE ();
}

/* if this building data is not NIL, creates a new allocated descriptor
 * sets this building data attributes
 * <isCond> is pk_true if this node is conditionnally executed
 */
_PUBLIC_ boolean NodeElement::set_builder (ref2LineList preludeLineList, ref2AncestorList ancestors, ref2LineList defaultLineList, boolean isCond)
{
  if (isCond)
    {
      if (defaultLineList == NIL_LineList)
	{
	  fatal_line_reference ("required default missing");
	  return pk_false;
	}
      this_buildingData->set_defaultActivated ();
    }
  this_buildingData->set_attributes (preludeLineList, ancestors, defaultLineList);
  return pk_true;
}

_PRIVATE_ void
NodeElement::set_definedImplicitAttributes ()
{
  this_isDefined = pk_true;
  this_isExecutable = pk_false;
  UNSET_UPDATED ();
  UNSET_REGISTERED ();
  UNLOCK ();			/* PLG 2008-06-10 */
}

/*
 * sets this as defined, not executable
 * creates a new building data
 * sets this creator as NIL
 */
_PUBLIC_ void
NodeElement::set_defined ()
{
  set_definedImplicitAttributes ();
  this_buildingData = new BuildingData ();
}

/*
 * sets this as defined, not executable
 * sets this_buildingData equal to prevNode::this_buildingData
 * sets this creator as NIL
 */
_PUBLIC_ void
NodeElement::set_defined (ref2NodeElement prevNode)
{
  set_definedImplicitAttributes ();
  this_buildingData = prevNode->get_buildingData ();
}

/* PRE: this_isDefined is pk_true, cycle detection is achieved */
_PROTECTED_ void
NodeElement::set_actualCreator (ref2NodeElement creator)
{
  SET_ACTUALCREATOR (creator);
}

_PROTECTED_ boolean NodeElement::has_creator_err ()
{
  // fprintf(stderr, "DEBUG NodeElement::has_creator_err this_name is <%s>, this is %i, this registered creator is %i\n", this_name, (int)this, (int)this_creator);
  if (HAS_CREATOR ())
    return pk_true;
  fatal_line_reference ("does not have constructive parent");
  return pk_false;
}

_PROTECTED_ boolean NodeElement::has_creator ()
{
  if (HAS_CREATOR ())
    return pk_true;
  levelInfo_comment_ref_sp__ln ("does not have constructive parent");
  return pk_false;
}

_PROTECTED_ void
NodeElement::set_toBeBuilt ()
{
  levelInfo_ObjectToBuild ();
  SET_TOBEBUILT ();
  SET_ACTUALCREATOR (this);
}

/* PLG 2006-01-18
 * this is safe, neither a MKFFileNode, nor an ExternNode, 
 * this is defined, this may have ancestor
 * this has not been previously visited
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * success: this is safe
 *          a message is printed
 *          this is set to be built
 *          this_creator is assigned not NIL <creator>
 * failure: this is not safe
 * this_date remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_ void
NodeElement::set_toBeBuilt (ref2NodeElement creator)
{
  // fprintf(stderr, "DEBUG NodeElement::set_toBeBuilt with creator node is <%s>\n", this_name);
  levelInfo_ObjectToBuild ();
  SET_TOBEBUILT ();
  SET_ACTUALCREATOR (creator);
}

/* PLG 2006-01-17
 * this is defined 
 * this is buildable
 * this may have ancestor
 * <creator> is not NIL, it is the most inner constructive node that precedes this
 * -->
 * this has been previously set to be built, this_creator is NIL
 *      this_creator is assigned <creator>
 * this has not been previously set to be built
 *      this is set to be built
 *      this_creator is assigned <creator>
 *      log message is printed
 * this_date remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_ boolean NodeElement::reset_toBeBuilt (ref2NodeElement creator)
{
  // fprintf(stderr, "DEBUG NodeElement::reset_toBeBuilt node is <%s>\n", this_name);
  /* is buildable, stated, consistent and has been visited */
  if (IS_UPDATED ())
    {
      if (IS_TOBEBUILT ())
	{
	  if (HAS_CREATOR ())
	    return pk_true;
	  SET_ACTUALCREATOR (creator);
	}			/* ancestor free is cycle free */
      else
	set_toBeBuilt (creator);
      return pk_true;
    }
  return pk_false;
}

/* ==========  ==========  Writer
 */
_PUBLIC_ _VIRTUAL_ void
NodeElement::print_declBody (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (this_buildingData != NIL_BuildingData)
    this_buildingData->print (printingContext, outBuffer);
}

/* ==========  ==========  Log
 */

/* prints object standard body: empty
 */
_PUBLIC_ _VIRTUAL_ void
NodeElement::logDebug_declBody (int level)
{
  if (this_buildingData == NIL_BuildingData)
    log_spn_str_ln (level, "this body is not defined");
  else
    {
      log_subTitle_ln (level, "this body is");
      this_buildingData->logExtendedDebug_formula_ln (level + 1);
    }
}

_PROTECTED_ _VIRTUAL_ void
NodeElement::log_date_sp ()
{
  if (IS_SAFE ())
    {
      if (IS_MISSING ())
	log_str_sp ("to be created");
      else if (IS_TOBEBUILT ())
	log_str_sp ("to be updated");
      else if (!(IS_OLDEST ()))
	log_releaseDate_sp (GET_DATE ());
    }
  else
    log_str_sp ("*** in ERROR ***");
}


/* 
 * prints a message
 * sets node in error
 */
_PRIVATE_ void
NodeElement::logInCycle (ref2NodeElement * cycleRoot)
{
  if ((*cycleRoot) == this)
    *cycleRoot = NIL_NodeElement;
  else
    fatal_line_reference ("    referred to in ");
}

/* PLG 2005-08-31
 * this is supposed to be safe
 * prints commented ancestor reference
 */
_PROTECTED_ void
NodeElement::levelInfo_comment_ancestor_unsafe_ln ()
{
  levelInfo_comment_ref_sp__ln ("is unsafe");
}

_PRIVATE_ void
NodeElement::UNLOCK_inError (ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG NodeElement::UNLOCK_inError\n");
  UNLOCK ();			/* releases cycle detection */
  if ((*cycleRoot) != NIL_NodeElement)
    logInCycle (cycleRoot);
  set_fatal ();
  levelInfo_comment_ancestor_unsafe_ln ();
}

/* PLG 2005-08-31
 * <embeddingContext> is not NIL (not checked)
 * success this is not yet declared
 * success: sets this defined, this_buildingData to new buildingData 
 *          sets this_declarationContext as <embeddingContext>
 *          sets this as defined, not executable
 *          sets this creator as NIL
 * failure: conflicting definition
 *          error message is printed
 *          this is set in error
 */
_PUBLIC_ boolean NodeElement::set_declared_in (ref2Context embeddingContext, char *name)
{
  // fprintf(stderr, "DEBUG NodeElement::set_declared_in, name is %s\n", name);
  // embeddingContext->log_ASDEBUG("NodeElement::set_declared_in");
  if (this_isDefined)
    {
      embeddingContext->ASFatal_doubleDecl (name);
      set_fatal ();
      return pk_false;
    }
  set_defined ();		/* SET this_buildingData to new buildingData  */
  setContext (embeddingContext);
  return pk_true;
}

/* ==========  ========== Reader
 */
/* 2005-01-21
 * reads (possibly empty) declaration content
 * PRE::this_buildingData is not NIL
 * success: FILE = PREFIX. [_NODE_BODY] #'}' [Any*][EOF]
 *          where 
 *             [_NODE_BODY] ::= opt{[conditionTAG_fSpStar]} opt{[_PRELUDE]} [_ANCESTOR*] opt{[_DEFAULT]}
 * failure: cursor is moved to errored character
 *          sets this node errored
 *          error message is printed
 */
_PUBLIC_ boolean NodeElement::read_dataDecl_fSpStar_nSp (ref2ArgumentList comments, boolean isCond, const char *mess)
{
  // fprintf(stderr, "DEBUG NodeElement::read_dataDecl_fSpStar_nSp, this_declarationContext type is %s\n", this_declarationContext->myClassName());
  debug_declHeader_ln (mess);
  setComments (comments);
  if (isCond)
    this_buildingData->set_defaultActivated ();
  if (this_buildingData->readOpt_buildingData_eol_ (this_declarationContext))
    {
      if (this_buildingData->is_empty ())
	(this_declarationContext->get_lexicalContext ())->ASInfo_emptyBuilder_ln ();
      else if (isCond)
	{
	  if (!(this_buildingData->has_defaultAction ()))
	    {
	      (this_declarationContext->get_lexicalContext ())->ASFatal_defaultMissing (this_name);
	      set_fatal ();
	      return pk_false;
	    }
	}
      return pk_true;
    }
  set_fatal ();
  return pk_false;
}

/* ==========  ==========  Grapher
 */
/* PLG 2005-08-31
 * this is supposed to be safe, and not MKFFileNode
 *         this is defined (this_isDefined is pk_true) ( ==> this_buildingData is not NIL) 
 * checks that system file dependences may be properly stated
 * success: this has ancestor and this is not among its own ancestor
 *          <cycleRoot> remains unchanged
 * failure: this does not have ancestor
 *             <cycleRoot>, safety remain unchanged
 *          this completes a cycle
 *             error message is printed
 *             if cycle has more than one node
 *                <cycleRoot> is set to cycle root
 *             if cycle has only one node
 *                <cycleRoot> remains unchanged
 *             this is set in fatal error
 * this_date remains unchanged
 * this_stepIndex remains unchanged
 */
_PROTECTED_ boolean NodeElement::_innerExplorationToDo_ln (ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG NodeElement::_innerExplorationToDo_ln, name is %s\n", this_name);
  if (IS_LOCKED ())
    {
      if (IS_LOCKED_BY_THIS ())
	fatal_reflexive ();
      else
	{
	  (*cycleRoot) = GET_LOCKER ();
	  fatal_cycle ();
	}
      return pk_false;
    }
  return (HAS_ANCESTOR ());
}

/* ========== buildable nodes: buildable variables or standard files */

/* PRE: this_isDefined is pk_true, cycle detection is achieved */
_PROTECTED_ boolean NodeElement::set_nullCreator ()
{
  if (IS_SAFE ())
    {
      if (!IS_UPDATED ())
	SET_NULLCREATOR ();
      return pk_true;
    }
  return pk_false;
}

_PROTECTED_ boolean NodeElement::set_nullCreator (time_t * updatedDate)
{
  if (IS_SAFE ())
    {
      if (IS_UPDATED ())
	{
	  (*updatedDate) = GET_DATE ();
	}
      else
	{
	  SET_NULLCREATOR ();
	}
      return pk_true;
    }
  return pk_false;
}


/* PLG 2006-01-17
 * this is buildable
 * this has not been previously set to be built
 * and then this has not been previously registered
 * -->
 * this is set to be built
 * this_creator is assigned this
 * log message is printed
 * this is registered as a constructive node
 * -->
 * this_date remains unchanged
 * this_stepIndex remains unchanged
 */
_PROTECTED_ void
NodeElement::register_constructiveNode ()
{
  /* is buildable, stated, consistent and has been visited or is being visited */
  if (!this_isRegistered)
    {
      ContextTree::append_constructiveNode (this);
      SET_REGISTERED ();
    }
}

_PROTECTED_ void
NodeElement::register_passiveNode ()
{
  /* is buildable, stated, consistent and has been visited or is being visited */
  if (!this_isRegistered)
    {
      ContextTree::append_passiveNode (this);
      SET_REGISTERED ();
    }
}

/* PLG 2006-01-17
 * this is defined
 * this is ancestor free or locking has been excluded TODO check
 * this_stepIndex is not smaller than the_stepIndex
 * this is buildable
 * this may have been previously visited
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * success: this is to be built
 *             this_creator is NIL: this_creator is assigned <creator>
 *          this is not to be built, this_date is FUTURE: 
 *             this is set to be built
 *             this_creator is assigned <creator>
 * failure: this is not to be built, this_date is not FUTURE
 * this_date remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_ boolean NodeElement::reset_buildable (ref2NodeElement creator, time_t * updatedDate, boolean * isToBeBuilt)
{
  // fprintf(stderr, "DEBUG NodeElement::reset_buildable node is  %s \n", this_name);
  /* is buildable, stated, consistent and has been visited or is being visited */
  if (IS_UPDATED ())
    {
      (*updatedDate) = GET_DATE ();
      if (IS_TOBEBUILT ())
	{
	  if (HASNOT_CREATOR ())
	    SET_ACTUALCREATOR (creator);
	  (*isToBeBuilt) = pk_true;
	  return pk_true;
	}
      if (IS_FUTURE ())
	set_toBeBuilt (creator);
      return pk_true;
    }
  return pk_false;
}

/* ========== constructive nodes: executable variables or direct files */
/* PLG 2006-01-17
 * this is buildable
 * this is defined 
 * this is ancestor free or locking has been excluded TODO check
 * this has been previously registered
 * -->
 * this is set to be built
 * this_creator is assigned this
 * log message is printed
 * this is registered as a constructive node
 * -->
 * this_date remains unchanged
 * this_stepIndex remains unchanged
 */
_PROTECTED_ boolean NodeElement::reset_ancestor_constructive (time_t contextTime, boolean * isToBeBuilt)
{
  // fprintf(stderr, "DEBUG NodeElement::reset_ancestor_constructive node is  %s \n", this_name);
  if (IS_UPDATED ())
    {
      if (IS_FUTURE () || IS_TOBEBUILT ())
	{
	  (*isToBeBuilt) = pk_true;
	  set_toBeBuilt ();
	  register_constructiveNode ();
	}
      return pk_true;
    }
  return pk_false;
}

_PROTECTED_ boolean NodeElement::reset_ancestorFree_constructive (time_t contextTime, boolean * isToBeBuilt)
{
  // fprintf(stderr, "DEBUG NodeElement::reset_ancestorFree_constructive node is  %s \n", this_name);
  if (IS_UPDATED ())
    {
      if (IS_FUTURE () || IS_TOBEBUILT ())
	{
	  (*isToBeBuilt) = pk_true;
	  set_toBeBuilt ();
	  register_constructiveNode ();
	}
      return pk_true;
    }
  return pk_false;
}

/* PLG 2008-06-13
 * this is supposed to be safe, unlocked
 * this_isDefined is pk_true (=> this building data is not NIL)  
 * success: these ancestors are safe, some ancestor file is (recursively) missing
 *            <correct> remains unchanged
 * failure: 1-these ancestors are safe, all ancestor files are present
 *            <correct> remains unchanged
 *          2-some of these ancestors is in error, this is cycle free
 *            <correct> is set to false
 *               <cycleRoot> remains unchanged
 *               this is set in error 
 *          2-this participates to a cycle
 *            <correct> is set to false
 *               <cycleRoot> is set to the root of this cycle
 *               this is set in error 
 * this_stepIndex remains unchanged
 * this is set unlocked
 */
_PROTECTED_ boolean NodeElement::_visit_missingAncestorFile (ref2NodeElement comingFrom, boolean * correct, ref2NodeElement * cycleRoot)
{
  boolean
    lCorrect = pk_true;
  LOCK_BY (comingFrom);		/* activates cycle detection */
  if ((GET_ANCESTORS ())->has_missingAncestorFile (this, this_declarationContext, &lCorrect, cycleRoot))
    {
      UNLOCK ();		/* releases cycle detection */
      return pk_true;
    }
  if (lCorrect)
    {
      UNLOCK ();
    }				/* releases cycle detection */
  else
    {
      (*correct) = pk_false;
      UNLOCK_inError (cycleRoot);
    }
  return pk_false;
}

/* PLG 2006-01-13
 * should be protected
 * this is supposed to be safe, this may be locked
 * this_isDefined is pk_true (=> this building data is not NIL)
 * this has not been previously visited (this_stepIndex is smaller than the_stepIndex)
 * <contextTime> is the time of the file which building can be decided when looking at these ancestors
 * <comingFrom> is the most inner node preceded by this
 * does not set this creator
 * success: this is safe, this has ancestors
 *          <updatedDate> is assigned the youngest (including FUTURE) updated date of its ancestors
 * failure: 1-this is unsafe
 *          2-this was safe, this has ancestors
 *             some of these ancestors is in error 
 *                 <cycleRoot> remains unchanged
 *             this participates to a cycle
 *                 <cycleRoot> is set to this cycle root
 *             set this in fatal error
 * this_date remains unchanged
 */
_PROTECTED_
  boolean NodeElement::_visit_intrinsicUpdate (time_t contextTime, ref2NodeElement comingFrom, time_t * updatedDate, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG NodeElement::_visit_intrinsicUpdate name is %s\n", this_name);
  LOCK_BY (comingFrom);		/* activates cycle detection */
  if ((GET_ANCESTORS ())->set_intrinsicUpdate (contextTime, this, this_declarationContext, updatedDate, cycleRoot))
    {
      UNLOCK ();		/* releases cycle detection */
      SET_NULLCREATOR ();
      return pk_true;
    }
  UNLOCK_inError (cycleRoot);
  return pk_false;
}

/* PLG 2006-01-13
 * this is supposed to be safe, this is not locked TODO check
 * this_isDefined is pk_true (=> this building data is not NIL) 
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <contextTime> is the time of the file which building can be decided when looking at these ancestors
 * <comingFrom> is the most inner node preceded by this
 * <updatedDate> is intiated with this date
 * success: this has ancestors
 *          this has some missing file ancestor or this has some dependence date inversion 
 *                 wrt to creation date and dependence modalities
 *          or some dependence modality force some ancestor to be built
 *              <isToBeBuilt> is set to pk_true
 *              this is set to be built
 *          <updatedDate> is assigned the youngest (including FUTURE) updated date of its ancestors
 *          this constructor is assigned <creator>
 * failure: 1-this does not have ancestors
 *          2-this has ancestors
 *            some of these ancestors is in error 
 *              <cycleRoot> remains unchanged
 *            this participates to a cycle
 *              <cycleRoot> is set to this cycle root
 *            set this in error
 * this_date remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_
  boolean NodeElement::_visit_standardUpdate (ref2NodeElement creator, time_t contextTime, ref2NodeElement comingFrom,
					      time_t * updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG NodeElement::_visit_standardUpdate name is %s\n", this_name);
  LOCK_BY (comingFrom);		/* activates cycle detection */
  if ((GET_ANCESTORS ())->set_standardUpdate (creator, contextTime, this, this_declarationContext, updatedDate, isToBeBuilt, cycleRoot))
    {
      UNLOCK ();		/* releases cycle detection */
      if (*isToBeBuilt)
	set_toBeBuilt (creator);
      else
	{
	  SET_ACTUALCREATOR (creator);
	}
      return pk_true;
    }
  UNLOCK_inError (cycleRoot);
  return pk_false;
}

/* PLG 2006-01-17
 * should be protected
 * this is supposed to be safe
 * this_isDefined is pk_true (=> this building data is not NIL)
 * this ancestor list is not empty
 * locking has been excluded TODO check
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * success: a message is printed
 *          this is set to be built
 *          not NIL <creator> is assigned to this_creator
 *          this_date is assigned the youngest (including FUTURE) updated date (wrt dependence policy) of its ancestors
 * failure: some of these ancestors is in error 
 *              <cycleRoot> remains unchanged
 *          this participates to a cycle
 *              <cycleRoot> is set to this cycle root
 *          set this in error
 *          this_date remains unchanged
 * this_creator is assigned not NIL <creator>
 * this_stepIndex is set to the_stepIndex
 */
/* called in VariableNode::set_missingUpdate, when this is buildable and has defined missing ancestor file
          in VariableNode::set_everUpdate when this is buildable */
_PROTECTED_
  boolean NodeElement::_visit_standard_toBeBuilt (ref2NodeElement creator, ref2NodeElement comingFrom, time_t * updatedDate,
						  ref2NodeElement * cycleRoot)
{
  // fprintf(stderr, "DEBUG NodeElement::_visit_standard_toBeBuilt name is %s\n", this_name);
  boolean
    isToBeBuilt = pk_false;

  (*updatedDate) = GET_DATE ();
  LOCK_BY (comingFrom);		/* activates cycle detection */
  if ((GET_ANCESTORS ())->set_standardUpdate (creator, GET_FUTURE (), this, this_declarationContext, updatedDate, &isToBeBuilt, cycleRoot))
    {
      UNLOCK ();		/* releases cycle detection */
      set_toBeBuilt (creator);
      return pk_true;
    }
  UNLOCK_inError (cycleRoot);
  return pk_false;
}

/* PLG 2006-01-17
 * should be protected
 * this is supposed to be safe, defined
 *         this_isDefined is pk_true (=> this building data is not NIL) 
 * this may have been previoudly visited
 * <creator> is the most inner constructive node that precedes this, <creator> is not NIL
 * <comingFrom> is the most inner node preceded by this
 * success: this has ancestors
 *          1-this has yet been recursively set
 *               <updatedDate> is assigned this updated date
 *              <hasYoungerAncestor> remains unchanged
 *          2-this has not yet been recursively set, 
 *                 this has some missing file ancestor or this has some dependence date inversion 
 *                 wrt to creation date and dependence modalities
 *              <updatedDate> is assigned the youngest (including FUTURE) updated date of its ancestors
 *              <hasYoungerAncestor> is set to pk_true
 *              this is set to be built
 *          3-this has not yet been recursively set, 
 *                 all ancestor files are present, none of them has a dependence date inversion
 *              <updatedDate> is assigned the youngest (including FUTURE) updated date of its ancestors
 *              <hasYoungerAncestor> is set to pk_true iff some of these ancestors is younger than this
 *              this is set to be built
 *          this constructor is assigned <creator>
 * failure: 1 this does not have ancestors
 *          2 this has ancestors
 *            some of these ancestors is in error 
 *              <cycleRoot> remains unchanged
 *            this participates to a cycle
 *              <cycleRoot> is set to this cycle root
 *            set this in error
 * this_date remains unchanged
 * this_stepIndex is set to the_stepIndex
 */
_PROTECTED_
  boolean NodeElement::_visitIf_recursiveEverUpdate (ref2NodeElement creator, ref2NodeElement comingFrom,
						     time_t * updatedDate, boolean * hasYoungerAncestor, ref2NodeElement * cycleRoot)
{
  boolean
    isToBeBuilt;
  // fprintf(stderr, "DEBUG NodeElement::_visitIf_recursiveEverUpdate, name is %s\n", this_name);
  (*updatedDate) = GET_DATE ();
  /* this is defined (=>this_buildingData is not NIL */
  if (_innerExplorationToDo_ln (cycleRoot))
    {
      if (IS_UPDATED ())
	{
	  /* PLG 2008-06-10 TODO check recursivity safety */
	  if (this_buildingData->time_checking_recursiveEver_done ())
	    {
	      if (HASNOT_CREATOR ())
		{
		  SET_ACTUALCREATOR (creator);
		}
	      return pk_true;
	    }
	}
      else
	this_buildingData->set_recursiveEverVisited ();	/* PLG 2008-06-10 */
      LOCK_BY (comingFrom);	/* activates cycle detection */
      if ((GET_ANCESTORS ())->set_recursiveEverUpdate (creator, this, this_declarationContext, updatedDate, &isToBeBuilt, cycleRoot))
	{
	  UNLOCK ();		/* releases cycle detection */
	  if (isToBeBuilt && !IS_TOBEBUILT ())	/* how may isToBeBuilt being false ? */
	    set_toBeBuilt (creator);
	  else if (IS_UPDATED ())
	    {
	      if (HASNOT_CREATOR ())
		{
		  SET_ACTUALCREATOR (creator);
		}
	    }
	  else
	    {
	      SET_ACTUALCREATOR (creator);
	    }
	  (*hasYoungerAncestor) = IS_strictlyOLDER_THAN (*updatedDate);
	  return pk_true;
	}
      UNLOCK_inError (cycleRoot);
      return pk_false;
    }
  if (IS_SAFE ())
    {
      /* this is ancestor free */
      if (IS_UPDATED ())
	{
	  if (HASNOT_CREATOR ())
	    {
	      SET_ACTUALCREATOR (creator);
	    }
	}
      else
	{
	  SET_ACTUALCREATOR (creator);
	}
      this_buildingData->set_recursiveEverVisited ();
    }				/* PLG 2008-06-10 */
  return pk_false;
}

/* PLG 2006-01-04
 * success: never
 * failure: this is safe
 *              <correct> remains unchanged
 *          this is not safe
 *              <correct> is set to false
 *          <cycleRoot> remains unchanged
 * this_stepIndex remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::has_missingAncestorFile (ref2NodeElement comingFrom, boolean * correct, ref2NodeElement * cycleRoot)
{
  if (IS_SAFE ())
    return pk_false;
  OUPS_name ("unsafe node", this_name);
  (*correct) = pk_false;
  return pk_false;
}

/* PLG 2006-01-09
 * this node analysis is required by a presence inquiring dependence following an active link
 * <creator> unused
 * <comingFrom> unused
 * <cycleRoot>, <isMissing> remain unchanged
 * this_stepIndex is set to the_stepIndex
 * success: this is safe
 * failure: this is not safe
 */
/* is defined in VariableNode, FileNode, FileNode::DirectFileNode */
_PUBLIC_
  _VIRTUAL_ boolean NodeElement::set_missingUpdate (ref2NodeElement creator, ref2NodeElement comingFrom, boolean * isMissing,
						    ref2NodeElement * cycleRoot)
{
  if (IS_SAFE ())
    return pk_true;
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* PLG 2006-01-04
 * this node analysis is required by a time inquiring dependence
 * <context_time> ignored
 * <comingFrom> ignored
 * success: this is safe
 *                <cycleRoot> remains unchanged
 *                the oldest date is assigned to <updatedDate>
 * failure: this is not safe
 */
_PUBLIC_
  _VIRTUAL_ boolean NodeElement::set_intrinsicUpdate (time_t contextTime, ref2NodeElement comingFrom,
						      time_t * updatedDate, ref2NodeElement * cycleRoot)
{
  if (IS_SAFE ())
    {
      (*updatedDate) = GET_OLDEST ();
      return pk_true;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* 2006-01-11
 * this node dependence analysis is required by a standard dependence
 * this method has to be defined for those class nodes that may have dependence
 * <creator> ignored
 * <contextTime> ignored
 * <comingFrom> ignored
 * success: this is safe
 *          <updatedDate> is assigned OLDEST
 *          <isToBeBuilt> remains unchanged
 * failure: this is unsafe
 * <cycleRoot> remains unchanged
 */
_PUBLIC_
  _VIRTUAL_ boolean NodeElement::set_standardUpdate (ref2NodeElement creator, time_t contextTime, ref2NodeElement comingFrom,
						     time_t * updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  if (IS_SAFE ())
    {
      (*updatedDate) = GET_OLDEST ();
      return pk_true;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* this node analysis is required by a dependence that requires this building
 * <comingFrom> is the node immediately preceded by this
 * <creator> is the most inner constructive node that precedes this
 * <recursive> is pk_true if building is required for this and recursively its ancestors
 * <cycleRoot> is set to a cycle root if this is in a cycle not beginning with this
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *                <cycleRoot> remains unchanged
 * failure: some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 * Node behaviour
 * --------------
 * success: this is safe
 * failure: this is in error
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::set_everUpdate (ref2NodeElement creator, ref2NodeElement comingFrom, ref2NodeElement * cycleRoot)
{
  if (IS_SAFE ())
    return pk_true;
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

/* this node analysis is required by a dependence that requires this building
 * <comingFrom> is the most inner file (transitively) preceded by this
 * <creator> is the most inner constructive node that precedes this
 * <recursive> is pk_true if building is required for this and recursively its ancestors
 * <cycleRoot> is set to a cycle root if this is in a cycle not beginning with this
 * success: this is not in a cyclic dependence, all referred to nodes are properly defined 
 *                <cycleRoot> remains unchanged
 * failure: some referred to node is not properly defined or some missing file does not have definition
 *          a cyclic dependence including this is found, this is not the cycle root
 *                sets <cycleRoot> to cycle root
 *          no cyclic dependence includes this, or this is the cycle root
 *                <cycleRoot> remains unchanged
 * Node behaviour
 * --------------
 * success: this is safe
 * failure: this is in error
 */
_PUBLIC_
  _VIRTUAL_ boolean NodeElement::set_recursiveEverUpdate (ref2NodeElement creator, ref2NodeElement comingFrom,
							  time_t * inner_updatedDate, boolean * isToBeBuilt, ref2NodeElement * cycleRoot)
{
  if (IS_SAFE ())
    {
      (*inner_updatedDate) = GET_OLDEST ();
      return pk_true;
    }
  levelInfo_comment_ancestor_unsafe_ln ();
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean NodeElement::set_rootUpdate (ref2MKFFileNode comingFrom, int force)
{
  OUPS_name ("illegal NodeElement::set_rootUpdate, VIRTUAL expected", this_name);
  return pk_false;
}

/* ==========  ==========  Generator
 */

/* should be protected
 * this is supposed to be safe, defined and cycle free, not a MKFFileNode, not an ExternNode, 
 *         this_buildingData is not NIL
 * success: main part is not empty;
 * failure: main part is empty;
 */
_PROTECTED_ boolean NodeElement::mainToGenerate ()
{
  if (this_buildingData->has_mainAction ())
    {
      levelInfo_MainGeneration (pk_true);
      return pk_true;
    }
  levelInfo_MainGeneration (pk_false);
  return pk_false;
}

/* 2005-08-26
 * success: this is not in error 
 * failure: this in error 
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::nested_requiredConstructive ()
{
  OUPS_name ("illegal NodeElement::nested_requiredConstructive, VIRTUAL expected", this_name);
  return pk_false;
}

/* 
 * curBuiltNode is the nearest node that depends on this
 * executes commands line by line
 * 
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::execute (ref2NodeElement curExecutedNode)
{
  OUPS_name ("illegal NodeElement::execute, VIRTUAL expected", this_name);
  return pk_false;
}

/* 
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::do_innerConstructive (boolean activeDependence)
{
  OUPS_name ("illegal NodeElement::do_innerConstructive, VIRTUAL expected", this_name);
  return pk_false;
}

/* 
 * this is a (not executable) variable or a( not direct, not MKF) file
 * curBuiltFile is the nearest file that depends on this
 * PLG: there is a problem when a node N depends on a node P in a never generate dependence
 * and P is not a direct file. N nesting direct file should wait for P nesting direct file
 * A solution would be to manage list of suspended files 
 * 
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::generate (ref2NodeElement curBuiltFile, ref2Buffer fileToPrint)
{
  OUPS_name ("illegal NodeElement::generate, VIRTUAL expected", this_name);
  return pk_false;
}

/*
 * root generation of a direct file
 *     
 */
_PUBLIC_ _VIRTUAL_ boolean NodeElement::root_generate ()
{
  OUPS_name ("illegal NodeElement::root_generate, writable file or executable node expected", this_name);
  return pk_false;
}

/* ==========  ==========  Builder
 */

/* ==========  Reference to not defined node
 */

_PRIVATE_ void
NodeElement::init ()
{
  this_isDefined = pk_false;
  SET_OLDEST ();
  this_buildingRequired = pk_false;
}

/* PLG 2005-09-07
 * creates and returns a new node reference descriptor named <name>
 * sets this_isDefined to false
 * sets initial date as the universal OLDEST date (GENESIA)
 * sets this context as the static context
 */
_PROTECTED_ NodeElement::NodeElement (const char *name):
Element (name)
{
  init ();
}

/* ==========  Declaration in embeddingContext
 */

/* PLG 2005-09-07
 * <embeddingContext> is not NIL (not checked)
 * creates and returns a new node descriptor named <name> in <embeddingContext>, as not used
 * owned by MKFFile associated to <embeddingContext>
 * sets initial date as the universal OLDEST date (GENESIA)
 * sets this as not executable
 * sets this creator as NIL
 */
_PROTECTED_ NodeElement::NodeElement (ref2Context embeddingContext, const char *name):
Element (embeddingContext, name, NIL_ArgumentList)
{
  init ();
}

_PUBLIC_
/*
 *
 * Static Elements
 *
 */
  time_t NodeElement::NOW;
time_t
  NodeElement::FUTURE;
time_t
  NodeElement::APOCALYPSE;
int
  NodeElement::the_stepIndex;

_STATIC_ void
NodeElement::setStatic ()
{
  NOW = getStartTime ();
  FUTURE = NOW + one_hour ();
  APOCALYPSE = getApocalypse ();
  the_stepIndex = 0;
}

_STATIC_ void
NodeElement::incrStepIndex ()
{
  the_stepIndex = the_stepIndex + 1;
}

_STATIC_ int
NodeElement::getStepIndex ()
{
  return the_stepIndex;
}
