#define SET_SAFE() SET_SAFELEVEL(_safe_)
#define SET_ERROR() SET_SAFELEVEL(_error_)
#define SET_FATAL() SET_SAFELEVEL(_fatal_)
#define ARG_IS_FATAL(_ARG) ( _ARG == _fatal_ )


/* ==========  ==========  ==========  ========== TITLE: these attributes
 */

_PUBLIC_ _VIRTUAL_ const char *
Soundness::myClassName ()
{
  return "Soundness";
}

/* set anObject in error
 */
_PUBLIC_ safeLevel Soundness::get_safetyLevel ()	/* called in MKFFileNode */
{
  return GET_SAFELEVEL ();
}

/* set anObject in error
 */
_PROTECTED_ void
Soundness::reset_safetyLevel (safeLevel _newLevel)	/* called in MKFFileNode */
{
  if (IS_SAFE ())
    {
      SET_SAFELEVEL (_newLevel);
    }
  else if (ARG_IS_FATAL (_newLevel))
    {
      SET_FATAL ();
    }
}

_PUBLIC_ void
Soundness::set_error ()
{
  if (IS_SAFE ())
    {
      SET_ERROR ();
    }
}

_PUBLIC_ boolean Soundness::is_fatal ()
{
  return IS_FATAL ();
}

_PUBLIC_ void
Soundness::set_fatal ()
{
  SET_FATAL ();
}

/* ==========  ==========  ==========  ========== TITLE: printing
 */

/** called in log_value
*/
_PUBLIC_ _VIRTUAL_ void
Soundness::print_value (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
}

/** called in log_formula
*/
_PUBLIC_ _VIRTUAL_ void
Soundness::print_formula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  print_value (printingContext, outBuffer);
}

/* ==========  ==========  ==========  ========== TITLE: printing in log
 */

void
Soundness::log_sp ()
{
  LOG_SP ();
}

void
Soundness::log_ln ()
{
  LOG_LN ();
}

_PROTECTED_ void
Soundness::log_commentItem_sp ()
{
  LOG_COMMENTITEM_SP ();
}

_PROTECTED_ void
Soundness::log_nil ()
{
  LOG_STR_ ("#");
}

_PROTECTED_ void
Soundness::log_str (const char *str)
{
  LOG_STR_ (str);
}

_PROTECTED_ void
Soundness::log_str_sp (const char *str)
{
  LOG_STR_SP (str);
}

_PROTECTED_ void
Soundness::log_str_ln (const char *str)
{
  LOG_STR_LN (str);
}

_PROTECTED_ void
Soundness::log_str_sp_str_ln (const char *str1, const char *str2)
{
  LOG_STR_SP_STR_LN (str1, str2);
}

_PROTECTED_ void
Soundness::log_strOrNil (const char *str)
{
  if (str == NIL_String)
    log_nil ();
  else
    LOG_STR_ (str);
}

_PROTECTED_ void
Soundness::log_spn_str_ln (int level, const char *mess)
{
  LOG_SPN_STR_LN (level, mess);
}

_PROTECTED_ void
Soundness::log_innerOpen ()	/* in Soundness::FunctionCall::..., Soundness::Argument */
{
  LOG_STR_ ("<[[");
}

_PROTECTED_ void
Soundness::log_innerClose ()	/* in Soundness::FunctionCall::..., Soundness::Argument */
{
  LOG_STR_ ("]]>");
}

_PROTECTED_ void
Soundness::log_valueOpen ()	/* in Soundness::FunctionCall::..., Soundness::Argument */
{
  LOG_STR_ ("==");
}

_PROTECTED_ void
Soundness::log_valueClose ()	/* in Soundness::FunctionCall::..., Soundness::Argument */
{
}

/** called in log_object when (fdepth <= 0) and (vdepth > 0)
 */
_PUBLIC_ _VIRTUAL_ void
Soundness::log_value ()
{
  print_value (LOGMARKS (), LOGFILE ());
}

/** called in log_object via log_innerObject when (fdepth > 0) and (vdepth > 0), with fdepth-1, vdepth-1
 */
_PUBLIC_ _VIRTUAL_ void
Soundness::log_value (int fdepth, int vdepth)
{
  log_value ();
}

/** called in log_object when (fdepth > 0) and (vdepth <= 0)
 */
_PUBLIC_ _VIRTUAL_ void
Soundness::log_formula ()
{
  log_value ();
}

/** called in log_object via log_innerObject when (fdepth > 0) and (vdepth > 0), with fdepth-1, vdepth-1
 */
_PUBLIC_ _VIRTUAL_ void
Soundness::log_formula (int fdepth, int vdepth)
{
  log_value (fdepth, vdepth);
}

/** called in log_object when (fdepth > 0) and (vdepth > 0), with fdepth-1, vdepth-1
 */
_PUBLIC_ _VIRTUAL_ void
Soundness::log_innerObject (int fdepth, int vdepth)
{
  log_formula (fdepth, vdepth);
  log_valueOpen ();
  log_value (fdepth, vdepth);
  log_valueClose ();
}

_PUBLIC_ void
Soundness::log_object (int fdepth, int vdepth)
{
  if (fdepth > 0)
    {
      if (vdepth > 0)
	{
	  log_innerOpen ();
	  log_innerObject (fdepth - 1, vdepth - 1);
	  log_innerClose ();
	}
      else
	log_formula ();
    }
  else if (vdepth > 0)
    log_value ();
}

/* ==========  ==========  printing in log::MKF commented elementary functions
 */
/* ==========  ==========  printing in log::MKF element attributes
 */

/* 2004-09-20
 * prints release <date> in the log buffer
 */

/* ==========  ==========  ==========  ========== TITLE: INFO
 */

void
Soundness::log_ln_INFO_sp ()
{
  LOG_LN_INFO_SP ();
}

void
Soundness::log_ln_INFO__ln (const char *mess)
{
  LOG_LN_INFO_STR_LN (mess);
}


/* ==========  ==========  ==========  ========== TITLE: LEVEL INFO
 */

void
Soundness::fullVerboseInfo_opening (const char *mess)
{
  if (ISLVERBOSE ())
    if (ISDEBUG ())
      LOG_STEPTITLE_STR_LN (mess);
    else
      LOG_LN_INFO_STR_LN (mess);
}

void
Soundness::fullVerboseInfo_closing ()
{
  if (ISLVERBOSE ())
    {
      LOG_LN ();
      if (ISDEBUG ())
	LOG_BACKTAB ();
    }
}

void
Soundness::debugLine ()
{
  if (ISDEBUG ())
    LOG_LN ();
}

void
Soundness::debugLineTabulation (int n)
{
  if (ISDEBUG ())
    {
      LOG_LN ();
      LOG_TAB (n);
    }
}

void
Soundness::debugTabulation (int n)
{
  if (ISDEBUG ())
    LOG_TAB (n);
}

void
Soundness::debugBackTabulation ()
{
  if (ISDEBUG ())
    LOG_BACKTAB ();
}

void
Soundness::lVerboseBackTabulation ()
{
  if (ISLVERBOSE ())
    LOG_BACKTAB ();
}

/* ==========  ==========  ==========  ========== TITLE: WARNING
 */
/* 2004-12-08
 */
void
Soundness::log_ln_WARN__ln (const char *mess)
{
  LOG_LN_WARN_STR_LN (mess);
}

boolean
Soundness::levelInfo_ln_WARN_sp ()
{
  if (ISLVERBOSE ())
    {
      LOG_LN_WARN_SP ();
      return pk_true;
    }
  return pk_false;
}

void
Soundness::levelInfo_ln_WARN__sp (const char *mess)
{
  if (ISDEBUG ())
    LOG_LN_WARN_STR_SP (mess);
}

/* ==========  ==========  ==========  ========== TITLE: TOOL ERROR
 */
void
Soundness::unsafe_use (const char *where)
{
  LOG_LN_ERRO_STR_SP_STR_SP (myClassName (), "unsafe call to");
  LOG_STR_LN (where);
}

void
Soundness::unsafe_named_use (const char *name, const char *where)
{
  LOG_LN_ERRO_STR_SP_STR_SP (myClassName (), name);
  LOG_STR_SP_STR_LN ("unsafe call to", where);
}

void
Soundness::OUPS (const char *mess)
{
  LOG_LN_ERRO_SP ();
  LOG_STR_LN ("#################### sorry the following error occurred in PKMAKE compiler");
  LOG_STR_LN (mess);
}

void
Soundness::OUPS_name (const char *mess, const char *name)
{
  LOG_LN_ERRO_SP ();
  LOG_STR_SP_STR_LN ("#################### sorry the following error occurred in PKMAKE compiler, name entity is", name);
  LOG_STR_LN (mess);
}

void
Soundness::OUPS_memory (const char *whereIn)
{
  LOG_LN_ERRO_SP ();
  LOG_STR_SP_STR_LN ("#################### sorry a memory error occurred in", whereIn);
}

/* ==========  ==========  ==========  ========== TITLE: ERROR
 */
void
Soundness::error_ln_ (const char *mess)
{
  LOG_LN_ERRO_STR_SP (mess);
  LOG_LN ();
  set_error ();
}

void
Soundness::error_ln_ (const char *mess1, const char *mess2)
{
  LOG_LN_ERRO_STR_SP (mess1);
  LOG_STR_LN (mess2);
  set_error ();
}

/* ==========  ==========  ==========  ========== TITLE: FATAL
 */

void
Soundness::fatal_missingIdentifier (const char *category, const char *CString_name)	/* in  Context, FunctionCall::...::ListValFunctionCall, 
											   FunctionCall::...::StringValFunctionCall */
{
  LOG_LN_FATA_STR_SP ("missing declaration for ");
  LOG_STR_SP_STR_LN (category, CString_name);
  set_fatal ();
}

void
Soundness::fatal_expansion_ln (const char *postMess)	/* in  FunctionCall::ArgumentFunctionCall, Argument::ArgumentList */
{
  LOG_LN_FATA_STR_SP ("expansion failed");
  LOG_STR_LN (postMess);
  set_fatal ();
}

void
Soundness::fatal_named_expansion_ln (const char *name, const char *postMess)	/* in  FunctionCall::ArgumentFunctionCall, Argument::NamedArgument */
{
  LOG_LN_FATA_STR_SP (name);
  LOG_STR_SP_STR_LN ("expansion failed", postMess);
  set_fatal ();
}

void
Soundness::fatal_named_typeConflict_ln (const char *name, const char *typeStr)	/* in  Context, Context::DynamicContext,
										   Element::ArgumentElement, Argument::NamedArgument, 
										   Identifier, Identifier::Iterator */
{
  LOG_LN_FATA_STR_SP ("type conflict for");
  LOG_STR_SP (name);
  LOG_STR_SP_STR_LN (": expected type is", typeStr);
  set_fatal ();
}

/* ==========  ==========  ==========  ========== TITLE: DEBUG
 */

_PRIVATE_ void
Soundness::incrDEBUGMargin ()
{
  the_debugLevel = the_debugLevel + 1;
  LOG_TAB (4);
}

_PRIVATE_ void
Soundness::decrDEBUGMargin ()
{
  LOG_BACKTAB ();
  the_debugLevel = the_debugLevel + 1;
}

_PRIVATE_ void
Soundness::_logDEBUG_in_sp (const char *methodName)
{
  LOG_LN_DEBU_STR_SP (myClassName ());
  LOG_STR_SP_STR_SP ("::", methodName);
}

_PROTECTED_ void
Soundness::logDEBUG_in (const char *methodName)
{
  _logDEBUG_in_sp (methodName);
  LOG_LN ();
}

_PROTECTED_ void
Soundness::logDEBUG_in (const char *methodName, const char *mess)
{
  _logDEBUG_in_sp (methodName);
  LOG_STR_LN (mess);
}

_PROTECTED_ void
Soundness::logDEBUG_in (const char *methodName, const char *mess1, const char *mess2)
{
  _logDEBUG_in_sp (methodName);
  LOG_STR_SP_STR_LN (mess1, mess2);
}

_PROTECTED_ void
Soundness::logDEBUG_inner (const char *methodName)
{
  LOG_TAB (4);
  logDEBUG_in (methodName);
  LOG_BACKTAB ();
}

_PROTECTED_ void
Soundness::logDEBUG_inner (const char *methodName, const char *mess)
{
  LOG_TAB (4);
  logDEBUG_in (methodName, mess);
  LOG_BACKTAB ();
}

_PROTECTED_ void
Soundness::logDEBUG_inner (const char *methodName, const char *mess1, const char *mess2)
{
  LOG_TAB (4);
  logDEBUG_in (methodName, mess1, mess2);
  LOG_BACKTAB ();
}

_PROTECTED_ void
Soundness::openDEBUG_in (const char *methodName)
{
  incrDEBUGMargin ();
  logDEBUG_in (methodName, "enter");
}

_PROTECTED_ void
Soundness::openDEBUG_in (const char *methodName, const char *mess)
{
  incrDEBUGMargin ();
  logDEBUG_in (methodName, mess);
}

_PROTECTED_ void
Soundness::closeDEBUG_in (const char *methodName)
{
  logDEBUG_in (methodName, "exit");
  decrDEBUGMargin ();
}

_PROTECTED_ void
Soundness::closeDEBUG_in (const char *methodName, boolean success)
{
  if (success)
    logDEBUG_in (methodName, "exit on SUCCESS");
  else
    logDEBUG_in (methodName, "exit on FAILURE");
  decrDEBUGMargin ();
}

_PROTECTED_ void
Soundness::closeDEBUG_in (const char *methodName, const char *mess)
{
  logDEBUG_in (methodName, mess);
  decrDEBUGMargin ();
}

void
Soundness::log_NILComponent_ln (int level, const char *component)
{
  LOG_SPN_STR_SP (level, "this");
  LOG_STR_SP_STR_LN (component, "is empty");
}

void
Soundness::logDebug_thisNameIs_ln (int level, const char *name)
{
  LOG_SPN_STR_SP (level, "this name is");
  LOG_STR_LN (name);
}

void
Soundness::log_subTitle_ln (int level, const char *mess)
{
  LOG_PREFIX2_STR_LN (level, mess);
}

void
Soundness::log_subTitle_ln (int level, const char *mess1, const char *mess2)
{
  LOG_PREFIX2_STR_SP (level, mess1);
  LOG_STR_LN (mess2);
}


_PUBLIC_ _VIRTUAL_ void
Soundness::logDebug_objAnnounce_ln (int level)
{
  LOG_PREFIX1_STR_SP (level, "this class is");
  LOG_STR_SP (myClassName ());
  LOG_STR_SP (" at adr");
  LOG_ADR_LN (this);
}

_PUBLIC_ _VIRTUAL_ void
Soundness::logDebug_objClosing_ln (int level)
{
  LOG_PREFIX1_STR_SP (level, "end of class");
  LOG_STR_SP (myClassName ());
  LOG_STR_SP ("at adr");
  LOG_ADR_LN (this);
}

_PUBLIC_ _VIRTUAL_ void
Soundness::logExtendedDebug_formula_ln (int level)
{
  LOG_PREFIX1_STR_LN (level, myClassName ());
}

_PUBLIC_ void
Soundness::logDEBUG_formula_ln (const char *mess)
{
  LOG_LN_DEBU_STR_LN (mess);
  logExtendedDebug_formula_ln (0);
}

void
Soundness::logDebug_subFormula_ln (int level, const char *attribute)
{
  LOG_PREFIX2_STR_SP (level, "this");
  LOG_STR_SP_STR_LN (attribute, "is <--<");
  logExtendedDebug_formula_ln (level + 1);
  LOG_STR_LN (">-->");
}

/* PLG 2007-02-23
 */
_PUBLIC_ Soundness::Soundness ()
{
  // logDEBUG_in("Soundness");
  SET_SAFE ();
}

_VIRTUAL_
Soundness::~
Soundness ()
{
  // logDEBUG_in("~Soundness");
}

_PUBLIC_
/*
 *
 * Static Elements
 *
 */
/* global list of nodes to be built or processed, sorted during dependence analysis
 */
  ref2IntePrinter Soundness::the_logBuffer;
int
  Soundness::the_verboseLevel;
int
  Soundness::the_errorLevel;
int
  Soundness::the_debugLevel;

_STATIC_ boolean
Soundness::setStatic ()
{
  boolean lCorrect = pk_true;
  SET_VERBOSELEVEL (0);
  SET_LOGFILE (new IntePrinter (0, 127, " ", '\\', &lCorrect));
  the_errorLevel = 0;
  the_debugLevel = 0;
  return lCorrect;
}
