/* a StringArgument is always safe
 */

_PUBLIC_ _VIRTUAL_ const char *
StringArgument::myClassName ()
{
  return "StringArgument";
}

/* ========== Attributes
 */
/* PLG 2005-12-09
 * success: never
 * failure: always
 */
_VIRTUAL_ boolean
StringArgument::is_emptyString ()
{
  return (this_CString_expandedValue == NIL_String);
}

_PUBLIC_ _VIRTUAL_ boolean StringArgument::is_isomorphic (char *str_val)
{

  if (this_CString_expandedValue == NIL_String)
    return (str_val == NIL_String);
  if (str_val == NIL_String)
    return pk_false;
  return (strcmp (this_CString_expandedValue, str_val) == 0);
}

_PUBLIC_ _VIRTUAL_ boolean StringArgument::is_isomorphic (ref2Argument str_arg)
{
  if ((ref2Argument) this == str_arg)
    return pk_true;
  if (str_arg == NIL_Argument)
    return pk_false;
  return str_arg->is_isomorphic (this);
}

_PUBLIC_ _VIRTUAL_ boolean StringArgument::is_isomorphic (ref2StringArgument str_arg)
{
  if (this == str_arg)
    return pk_true;
  if (str_arg == NIL_StringArgument)
    return pk_false;
  return str_arg->is_isomorphic (this_CString_expandedValue);
}

/* 2005-10-06 PLG
 * this is supposed to be expanded
 * this has string type
 *    returns length of this expanded string
 * this has list type
 *    returns the cumulated length of these expanded element strings
 */
_VIRTUAL_ int
StringArgument::get_CStringLength_expanded ()
{
  // logDEBUG_in("get_CStringLength_expanded");
  if (this_CString_expandedValue == NIL_String)
    return 0;
  return strlen (this_CString_expandedValue);
}

/* PLG 2005-12-08
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
StringArgument::get_expandedString (ref2StringArgument * stringArgument_res, boolean * isFatal)
{
  (*stringArgument_res) = this;
  return pk_true;
}

/* ========== Writer
 */

/* 2004-12-09
 * prints this partially expanded value (expansion is only done for static values)
 */
_VIRTUAL_ void
StringArgument::line_actualFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer, int depth)
{
  print_actualFormula (printingContext, outBuffer, depth, pk_false);
  outBuffer->line ();
}

/* ========== Log is Argument writer log
 */
_PUBLIC_ _VIRTUAL_ void
StringArgument::log_formula_ (boolean isStringValue)
{
  if (isStringValue)
    {
      log_stringConstant (this_CString_expandedValue);
    }
  else
    {
      log_formula ();
    }
}

/* ==========  ==========  ==========  ========== TITLE: DEBUG
 */

/* ========== Reader
 */
/* ========== Generator
 */

/* 2005-10-06 PLG
 * this is supposed to be expanded (not checked)
 * success: returns in <CString_res> this expanded CString value 
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_CString_cast_expanded (char **CString_res)
{
  (*CString_res) = this_CString_expandedValue;
  return pk_true;
}

/* 2005-10-06 PLG
 *    this has been previously expanded 
 * success:
 *        returns in <strList_res> a new m'allocated singleton containing this expanded CString value 
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_CStringList_cast_flat_expanded (ref2CStringList * strList_res)
{
  (*strList_res) = new_CStringList ();
  return pk_true;
}

/* 2005-10-06 PLG
 *    this has been previously expanded 
 * success:
 *        returns in <lineList_res> a new m'allocated singleton containing this expanded CString value 
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_lineList_cast_flat_expanded (ref2LineList * lineList_res)
{
  (*lineList_res) = new_LineList ();
  return pk_true;
}

/* 2005-10-06 PLG
 * this is supposed to be expanded
 * this is a string
 *    if this actual value is not NIL_String
 *      appends, without C string marks, this actual value to target
 * target is supposed to be large enough
 */
_VIRTUAL_ void
StringArgument::cat_expanded (char *target)
{
  if (this_CString_expandedValue != NIL_String)
    target = strcat (target, this_CString_expandedValue);
}

/* PLG 2005-12-07
 * success: this is a string constant 
 *          returns this in <stringArgument_res>
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_stringExpand_ (ref2StringArgument * stringArgument_res)
{
  // logDEBUG_in("getIf_stringExpand_");
  (*stringArgument_res) = this;
  return pk_true;
}

/* PLG 2005-12-07
 * success: never
 * failure: 
 *          system error message is printed
 */
_VIRTUAL_ boolean
StringArgument::getIf_listExpand_ (ref2ArgumentList * argumentList_res)
{
  unsafe_use ("getIf_listExpand_(ref2ArgumentList *)");
  return pk_false;
}

/* 2005-10-06 PLG
 * success: returns this in <argument_res> 
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_expand_ (ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_expand_");
  (*argument_res) = this;
  return pk_true;
}

/* PLG 2005-12-07
 * success: returns in <argumentList_res> a new m'allocated string list equal to this_CString_expandedValue followed by <tailList>
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_flatten_listCast_expand_ (ref2ArgumentList tailList, ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("getIf_flatten_listCast_expand_(ref2ArgumentList, ref2ArgumentList *)");
  (*argumentList_res) = new ArgumentList (this, tailList);
  return pk_true;
}

/* PLG 2005-12-07
 * success: never
 * failure: always
 *          system error message is printed
 */
_VIRTUAL_ boolean
StringArgument::getIf_flatten_listExpand_ (ref2ArgumentList * argumentList_res)
{
  unsafe_use ("getIf_flatten_listExpand_(ref2ArgumentList *)");
  return pk_false;
}

/* PLG 2005-12-12
 * success: returns this expanded string value
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_CString_cast_expand_ (char **CString_res)
{
  // logDEBUG_in("getIf_CString_cast_expand_");
  (*CString_res) = this_CString_expandedValue;
  return pk_true;
}

/* PLG 2005-12-08
 * success: returns in <strList_res>  a new m'allocated CStringList containing this expanded value followed by <tailList> 
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList tailList, ref2CStringList * strList_res)
{
  (*strList_res) = new_CStringList (tailList);
  return pk_true;
}

/* PLG 2005-12-08
 * success: a new m'allocated CStringList containing this expanded value is returned in <strList_res>
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_CStringList_cast_flatExpand_ (ref2CStringList * strList_res)
{
  (*strList_res) = new_CStringList ();
  return pk_true;
}

/* PLG 2005-12-08
 * success: a new m'allocated LineList containing this expanded value followed by <tailList> is returned in <lineList_res>
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_lineList_cast_flatExpand_ (ref2LineList tailList, ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_(ref2LineList , ref2LineList *)");
  (*lineList_res) = new_LineList (tailList);
  return pk_true;
}

/* PLG 2005-12-08
 * success: a new m'allocated LineList containing this expanded value is returned in <lineList_res>
 * failure: never
 */
_VIRTUAL_ boolean
StringArgument::getIf_lineList_cast_flatExpand_ (ref2LineList * lineList_res)
{
  // logDEBUG_in("getIf_lineList_cast_flatExpand_(ref2LineList *), value is ", this_CString_expandedValue);
  (*lineList_res) = new_LineList ();
  return pk_true;
}

/* ========== Builder
 */

/* creates and returns a new pk_true string (excluding Meta) list cell valued by <aString>
 */
StringArgument::StringArgument (char *aString):Argument (aString)
{
  SET_STATIC ();
}

/* creates and returns a new NIL string argument
 */
StringArgument::StringArgument ():Argument (NIL_String)
{
  SET_EMPTY ();
}
