#define HAS_STRING_TYPE() ( this_type == 0 )
#define HAS_LIST_TYPE() ( this_type == 1 )

_PUBLIC_
/* ========== Attributes
 */
_VIRTUAL_ const char *
StringFunctionCall::myClassName ()
{
  return "StringFunctionCall";
}

_PUBLIC_ _VIRTUAL_ boolean StringFunctionCall::is_isomorphic (ref2ArgumentFunctionCall arg_funcCall)
{
  if ((ref2ArgumentFunctionCall) this == arg_funcCall)
    return pk_true;
  if (arg_funcCall != NIL_ArgumentFunctionCall)
    {
      return arg_funcCall->is_isomorphic (this);
    }
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean StringFunctionCall::is_isomorphic (ref2StringFunctionCall arg_funcCall)
{
  if (this == arg_funcCall)
    return pk_true;
  return have_isomorphicParameters (arg_funcCall);
}

/* ========== Log

 */

/* ========== Writer
 */

/* ========== Set
 */

/* ========== Printer
 */

/* ========== Reader
 */
/* ========== Generator
 */
/* PLG 2005-12-07
 * should be protected
 * this is supposed to be SAFE, 
 * this_linkedFunction is not NIL
 * actual argument list and formal argument list fit
 * success: these_actualParameters are correctly expanded
 *          function call is correctly processed
 *          the result of this_linkedFunction call is assigned to this_expandedArgument
 * failure: these_actualParameters expansion failed or function call failed
 *          error message is printed
 */
_VIRTUAL_ boolean
StringFunctionCall::single_call ()
{
  // logDEBUG_in("single_call() call to string");
  // logDEBUG_actualParameters();
  if (IS_EMPTY () || IS_STATIC ())
    return pk_true;
  if (these_actualParameters->getIf_CString_cast_listExpand_ (&this_CString_expanded))
    {
      this_expandedArgument = new StringArgument (this_CString_expanded);
      return pk_true;
    }
  return pk_false;
}

/* PLG 2005-12-07
 * this_linkedFunction is not NIL
 * success: this is safe, these_actualParameters are correctly expanded
 *          function call is correctly processed
 *          the result of this_linkedFunction call is returned in <CString_res>
 * failure: this is not safe, or these_actualParameters expansion failed or function call failed
 *          error message is printed
 */
_VIRTUAL_ boolean
StringFunctionCall::CString_call (char **CString_res)
{
  // logDEBUG_in("CString_call(char **) call to", this_functionName);
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (IS_EMPTY () || IS_STATIC ())
	    {
	      (*CString_res) = this_CString_expanded;
	      return pk_true;
	    }
	  /*      if ( these_actualParameters->getIf_CString_cast_listExpand_(&this_CString_expanded) ) { */
	  if (single_call ())
	    {
	      this_expandedArgument = new StringArgument (this_CString_expanded);
	      (*CString_res) = this_CString_expanded;
	      return pk_true;
	    }
	  set_fatal ();
	  return pk_false;
	}
      OUPS_name ("illegal StringFunctionCall::CString_call(char **) to an iterated string function", this_functionName);
    }
  else
    OUPS_name ("illegal StringFunctionCall::CString_call(char **), unsafe", this_functionName);
  set_fatal ();
  return pk_false;
}

/* PLG 2005-12-07
 * External entry point
 * success: this is safe, this type result is string, this evaluation succeeds
 *          returns in <stringArgument_res> the result of this evaluation
 * failure: this is not safe
 *          this is an iterated string function call
 *              system error message is printed
 */
_VIRTUAL_ boolean
StringFunctionCall::string_call (ref2StringArgument * stringArgument_res)
{
  // logDEBUG_in("string_call(ref2StringArgument *) call to", this_functionName);
  if (IS_SAFE ())
    {
      if (HAS_STRING_TYPE ())
	{
	  if (IS_EMPTY () || IS_STATIC ())
	    {
	      (*stringArgument_res) = (ref2StringArgument) this_expandedArgument;
	      return pk_true;
	    }
	  if (single_call ())
	    {
	      this_expandedArgument = new StringArgument (this_CString_expanded);
	      (*stringArgument_res) = (ref2StringArgument) this_expandedArgument;
	      return pk_true;
	    }
	  set_fatal ();
	  return pk_false;
	}
      OUPS_name ("illegal StringFunctionCall::string_call(ref2StringArgument *) to an iterated string function", this_functionName);
    }
  else
    OUPS_name ("illegal StringFunctionCall::string_call(ref2StringArgument **), unsafe", this_functionName);
  set_fatal ();
  return pk_false;
}

/* PLG 2005-12-07
 * External entry point
 * this is supposed to be an iterated string function call
 * success: this is safe, this type result is list, this evaluation succeeds
 *          returns in <argumentList_res> the result of this evaluation
 * failure: this is not safe
 *          this evaluation failed
 *          error message is printed
 */
_VIRTUAL_ boolean
StringFunctionCall::list_call (ref2ArgumentList * argumentList_res)
{
  // logDEBUG_in("list_call() call to", this_functionName);
  // logDEBUG_actualParameters();
  if (IS_SAFE ())
    {
      if (HAS_LIST_TYPE ())
	{
	  boolean lCorrect = pk_true;
	  boolean hasEmptyIteratorList;
	  if (setIterate (&lCorrect, &hasEmptyIteratorList))
	    {
	      // logDEBUG_in("list_call(),setIterate OK call to", this_functionName);
	      if (hasEmptyIteratorList)
		{
		  (*argumentList_res) = NIL_ArgumentList;
		  return pk_true;
		}
	      if (call_iteration (&lCorrect, argumentList_res))
		return pk_true;
	      (*argumentList_res) = NIL_ArgumentList;
	      return lCorrect;
	    }
	}
      else
	OUPS_name ("illegal StringFunctionCall::list_call(ref2ArgumentList *) to a string function", this_functionName);
    }
  else
    OUPS_name ("illegal StringFunctionCall::list_call(ref2StringArgument **), unsafe", this_functionName);
  return pk_false;
}

/* ========== Grapher
 */

/* ========== Builder
 */

boolean
StringFunctionCall::init_parameterStatus ()
{
  // logDEBUG_in("init_parameterStatus");
  if (HAS_EMPTY_PARAMETER ())
    {
      SET_EMPTY ();
      this_CString_expanded = NIL_String;
      this_expandedArgument = new StringArgument (this_CString_expanded);
      return pk_true;
    }
  if (HAS_STATIC_PARAMETER ())
    {
      if (these_actualParameters->getIf_CString_cast_listExpand_ (&this_CString_expanded))
	{
	  SET_STATIC ();
	  this_expandedArgument = new StringArgument (this_CString_expanded);
	  return pk_true;
	}
      set_fatal ();
      return pk_false;
    }
  return pk_true;
}

/* 2004-10-14
 * creates a new string function call (stringval or hostval)
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
/* referred to 
 *    in StringValFunctionCall::
 *             StringValFunctionCall(ref2Context readingContext, ref2Argument actualParameter)
 *          as StringFunctionCall(readingContext, "stringval", actualParameter)
 *    in HostValFunctionCall::
 *             HostValFunctionCall(ref2Context readingContext, ref2Argument actualParameter)
 *          as StringFunctionCall(readingContext, "hostval", actualParameter)
 */
StringFunctionCall::StringFunctionCall (ref2Context readingContext, const char *name, ref2Argument actualParameter):ArgumentFunctionCall (readingContext, pk_false, name,
		      actualParameter)
{
  // logDEBUG_in("StringFunctionCall(ref2Context, const char *) create call to", name);
}

/* 2004-10-14
 * creates a new string function call (stringval or hostval)
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
/* referred to 
 *    in UserStringFunctionCall::
 *                 UserStringFunctionCall(ref2Context readingContext, const char * name, ref2UserStringFunctionElement aFunction,
 *					  ref2ArgumentList actualParameters, boolean * correct))
 *          as StringFunctionCall(readingContext, name, actualParameters)
 */
StringFunctionCall::StringFunctionCall (ref2Context readingContext, const char *name, ref2ArgumentList actualParameters):
ArgumentFunctionCall (readingContext, pk_false, name, actualParameters)
{
  // logDEBUG_in("StringFunctionCall(ref2Context, const char *) create call to", name);
}

/* 2004-10-14
 * creates a new :string function call 
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
/* referred to 
 *    in Context::new_StringFunctionCall(ref2ArgumentList actualParameters, ref2StringFunctionCall * functionCall_res)
 *          as new StringFunctionCall(this, actualParameters, &lCorrect)
 */
StringFunctionCall::StringFunctionCall (ref2Context readingContext, ref2ArgumentList actualParameters, boolean * correct):ArgumentFunctionCall (readingContext, pk_false, stringWORD,
		      actualParameters)
{
  // logDEBUG_in("StringFunctionCall(ref2Context), create call to string");
  if (!init_parameterStatus ())	/*(fatal)error message is printed, fatal is set */
    (*correct) = pk_false;
}

/* 2004-10-14
 * creates a new iterated string function call (stringval or hostval)
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
/* referred to in StringValFunctionCall::StringValFunctionCall(ref2IteratorContext readingContext, ref2Argument actualParameter)
 *          as StringFunctionCall(readingContext, "stringval", actualParameter)
 * referred to in HostValFunctionCall::HostValFunctionCall(ref2IteratorContext readingContext, ref2Argument actualParameter)
 *          as StringFunctionCall(readingContext, "hostval", actualParameter)
 */
StringFunctionCall::StringFunctionCall (ref2IteratorContext readingContext, const char *name, ref2Argument actualParameter):
ArgumentFunctionCall (readingContext, name, actualParameter)
{
  // logDEBUG_in("StringFunctionCall(ref2IteratorContext, const char *), create iterated call to", name);
}

/* 2004-10-14
 * creates a new iterated string function call (stringval or hostval)
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
/* referred to in UserStringFunctionCall::
 *                    UserStringFunctionCall(ref2teratorContext readingContext, const char * name, ref2UserStringFunctionElement aFunction,
 *					     ref2ArgumentList actualParameters, boolean * correct))
 *          as StringFunctionCall(readingContext, name, actualParameters)
 */
StringFunctionCall::StringFunctionCall (ref2IteratorContext readingContext, const char *name, ref2ArgumentList actualParameters):
ArgumentFunctionCall (readingContext, name, actualParameters)
{
  // logDEBUG_in("StringFunctionCall(ref2IteratorContext, const char *), create iterated call to", name);
}

/* 2004-10-14
 * creates a new iterated :string function call 
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
                             not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
                     followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 *          <correct> remains unchanged
 * failure: set <correct> to false
 *          error message is printed
 */
/* referred to in IteratorContext::new_StringFunctionCall(ref2ArgumentList actualParameters, ref2StringFunctionCall * functionCall_res)
 *          as new StringFunctionCall(this, actualParameters, &lCorrect)
 */
StringFunctionCall::StringFunctionCall (ref2IteratorContext readingContext, ref2ArgumentList actualParameters, boolean * correct):ArgumentFunctionCall (readingContext, stringWORD,
		      actualParameters)
{
  // logDEBUG_in("StringFunctionCall(ref2IteratorContext) create iterated call to string");
  if (!init_parameterStatus ())	/*(fatal)error message is printed, fatal is set */
    (*correct) = pk_false;
}
