_PUBLIC_ _VIRTUAL_ const char *
UserStringFunctionCall::myClassName ()
{
  return "UserStringFunctionCall";
}

/* ========== Attributes
 */

_PUBLIC_ ref2UserStringFunctionElement UserStringFunctionCall::get_linkedFunction ()
{
  return this_linkedFunction;
}

_PUBLIC_ _VIRTUAL_ boolean UserStringFunctionCall::is_isomorphic (ref2ArgumentFunctionCall arg_funcCall)
{
  if ((ref2ArgumentFunctionCall) this == arg_funcCall)
    return pk_true;
  if (arg_funcCall != NIL_ArgumentFunctionCall)
    {
      return arg_funcCall->is_isomorphic (this);
    }
  return pk_false;
}

_PUBLIC_ _VIRTUAL_ boolean UserStringFunctionCall::is_isomorphic (ref2UserStringFunctionCall arg_funcCall)
{
  if (this == arg_funcCall)
    return pk_true;
  if (arg_funcCall == NIL_UserStringFunctionCall)
    return pk_false;
  if (this_linkedFunction == arg_funcCall->get_linkedFunction ())
    return have_isomorphicParameters (arg_funcCall->get_actualParameters ());
  return pk_false;
}

/* ========== Writer
 */

/* ========== Set
 */

/* ========== Printer
 */

/* ========== Log
 */
/* ========== Reader
 */

/* ========== Generator
 */

/* PLG 2005-10-13
 * should be protected
 * this is supposed to be SAFE, 
 * this_linkedFunction is not NIL
 * actual argument list and formal argument list fit
 * success: these_actualParameters are correctly expanded
 *          function call is correctly processed
 *          the result of this_linkedFunction call is assigned to this_expandedArgument
 * failure: these_actualParameters expansion failed or function call failed
 *          error message is printed
 */
_PUBLIC_ _VIRTUAL_ boolean UserStringFunctionCall::single_call ()
{
  // logDEBUG_in("ENTER single_call() to function", this_functionName);
  // logDEBUG_actualParameters();
  if (this_linkedFunction->CString_call (these_actualParameters, &this_CString_expanded))
    {
      this_expandedArgument = new StringArgument (this_CString_expanded);
      return pk_true;
    }
  set_error ();
  return pk_false;
}

/* ========== Grapher
 */

/* ========== Grapher
 */

/* ========== Builder
 */

/* 2004-11-23
 * creates a not iterated call to a user defined string function 
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
 *                            not followed by not space and is completed
 * success: <name> is the name of a string function
 *          cursor starts [ARGUMENT_LIST] ::= '(' [fileSpace*] [ARGUMENTS*] ')' [lineSpaces_eol]
 *                 <isEol> is set to pk_true
 *                 cursor is moved to end of line following ')'
 *          cursor starts [ARGUMENT_LIST] ::= '(' [fileSpace*] [ARGUMENTS*] ')' [Character_but_Space]
 *                 cursor is moved to [Character_but_Space] following ')'
 *                 <isEol> is set to false
 */
UserStringFunctionCall::UserStringFunctionCall (ref2Context readingContext, char *name, ref2UserStringFunctionElement aFunction, ref2ArgumentList actualParameters, boolean * correct):StringFunctionCall (readingContext, name,
		    actualParameters)
{
  // logDEBUG_in("UserStringFunctionCall(ref2Context, char *, ref2UserStringFunctionElement, ref2ArgumentList, boolean *) for ", name);
  this_linkedFunction = aFunction;
  (*correct) = set_actualParameters_scope ();
}

/* 2004-10-14
 * <aFunction> is not NIL (not checked)
 * creates an iterated call to a user defined string function 
 * this <readingContext> current character is a not space character following the function name
 *                       or this <readingContext> current line contains the function name
 *                       not followed by not space and is completed
 * success: current character is a parameter list opening mark followed by actual parameters
 *             followed by a parameter list closing mark
 *          this <readingContext> cursor is set after function call closing mark
 */
UserStringFunctionCall::UserStringFunctionCall (ref2IteratorContext readingContext, char *name,
						ref2UserStringFunctionElement aFunction, ref2ArgumentList actualParameters, boolean * correct):
StringFunctionCall (readingContext, name, actualParameters)
{
  // logDEBUG_in("UserStringFunctionCall(ref2IteratorContext, char *, ref2UserStringFunctionElement, ref2ArgumentList, boolean *) for ", name);
  this_linkedFunction = aFunction;
  (*correct) = set_actualParameters_scope ();
}
