/* PLG: rootFileNode becomes rootContextTree
   add ContextTree as an ancestor for new_pKMake
   move  to ContextTree
	 -generate_constructiveNodeList
	 -has_constructiveNode
*/

/* =============== Diagnostics
*/

ref2new_pKMakeReporting the_mainReporter;

static void
closeOnError ()
{
  the_mainReporter->log_mainErrorClosing (NodeElement::getStepIndex ());
  exit (-1);
}

static void
exitErr ()
{
  the_mainReporter->log_errorClosing ();
  exit (-1);
}

static void
exitArgErrOn (const char *arg, int argErrNum)
{
  the_mainReporter->log_argumentErrorClosing (arg, argErrNum);
  exit (-1);
}

/* gets node selection argument
 * <strOption> is the current option string
 * <argc> is the number of arguments
 * <argIndex> is the value of next argument index
 * <argv> is the list of arguments
 * success: <rootNodeIsDefined> is false
 *             <strOption> is "-node", <argIndex> <= <argc>
 *                <rootVariableNode> gets <argv>[<argIndex>] value
 *                <argIndex> is incremented by 1
 *                <allRoot> remains unchanged
 *             <strOption> is "-all"
 *                <rootVariableNode>, <argIndex> remain unchanged
 *                <allRoot> is set to false
 *             <strOption> is "-ALL"
 *                <rootVariableNode>, <argIndex> remain unchanged
 *                <allRoot> is set to pk_true
 *          <rootNodeIsDefined> is set to pk_true
 * failure: syntax error
 *             error message is printed
 *             error exit
 *          <strOption> does not belong to "-node", "-all", "-ALL"
 */
static boolean
getRoot_nodeOption (char *strOption, int argc, int *argIndex, char *argv[], boolean * rootNodeIsDefined, char **rootVariableNode, boolean * allRoot)
{
  if (strcmp (strOption, "-all") == 0)
    {
      if (*rootNodeIsDefined)
	exitArgErrOn (strOption, 0);
      (*rootNodeIsDefined) = pk_true;
      (*allRoot) = pk_false;
      return pk_true;
    }
  if (strcmp (strOption, "-ALL") == 0)
    {
      if (*rootNodeIsDefined)
	exitArgErrOn (strOption, 0);
      (*rootNodeIsDefined) = pk_true;
      (*allRoot) = pk_true;
      return pk_true;
    }
  if (strcmp (strOption, "-node") == 0)
    {
      if (*rootNodeIsDefined)
	exitArgErrOn (strOption, 0);
      if (argc < (*argIndex))
	exitArgErrOn (strOption, 1);
      (*rootNodeIsDefined) = pk_true;
      (*rootVariableNode) = argv[(*argIndex)];
      (*argIndex) = (*argIndex) + 1;
      return pk_true;
    }
  return pk_false;
}

/* gets building mode argument
 * <strOption> is the current option string
 * success: <force> is 0
 *             <strOption> is "-r"
 *                <force> is set to 1
 *             <strOption> is "-R"
 *                <force> is set to 2
 *          <force> is not 0
 *             warning message is printed
 *             <strOption> is "-r"
 *                 <force> remains unchanged
 *             <strOption> is "-R"
 *                <force> is set to 2
 * failure: <strOption> does not belong to "-r", "-R"
 */
static boolean
get_buildingModeOption (char *strOption, int *force)
{
  if (strcmp (strOption, "-r") == 0)
    {
      if ((*force) == 0)
	*force = 1;
      else
	the_mainReporter->log_resetRecursivityArg ();
      return pk_true;
    }
  if (strcmp (strOption, "-R") == 0)
    {
      if ((*force) != 0)
	the_mainReporter->log_resetRecursivityArg ();
      *force = 2;
      return pk_true;
    }
  return pk_false;
}

/* gets loging mode argument
 * <strOption> is the current option string
 * success: <verboseLevel> is 0
 *             <strOption> is "-v"
 *                <verboseLevel> is set to 1
 *             <strOption> is "-V"
 *                <verboseLevel> is set to 2
 *             <strOption> is "-d"
 *                <verboseLevel> is set to 3
 *             <strOption> is "-D"
 *                <verboseLevel> is set to 4
 *          <verboseLevel> is not 0
 *             warning message is printed
 *             <strOption> is "-v"
 *                <verboseLevel> remains unchanged
 *             <strOption> is "-V"
 *                <verboseLevel> is set to max(2, <verboseLevel>)
 *             <strOption> is "-d"
 *                <verboseLevel> is set to max(3, <verboseLevel>)
 *             <strOption> is "-D"
 *                <verboseLevel> is set to 4
 * failure: <strOption> does not belong to "-v", "-V", "-d", "-D" 
 */
static boolean
get_logingModeOption (char *strOption)
{
  if (strcmp (strOption, "-v") == 0)
    {
      the_mainReporter->reset_verbose (1);
      return pk_true;
    }
  if (strcmp (strOption, "-V") == 0)
    {
      the_mainReporter->reset_verbose (2);
      return pk_true;
    }
  if (strcmp (strOption, "-d") == 0)
    {
      the_mainReporter->reset_verbose (3);
      return pk_true;
    }
  if (strcmp (strOption, "-D") == 0)
    {
      the_mainReporter->reset_verbose (4);
      return pk_true;
    }
  return pk_false;
}

/* gets statistics option
 * <strOption> is the current option string
 * success: <logStat> is 0
 *             <strOption> is "-stat"
 *                <logStat> is set to 1
 *             <strOption> is "-STAT"
 *                <logStat> is set to 2
 *          <logStat> is > 0
 *             warning message is printed
 *             <strOption> is "-stat"
 *                <logStat> remains unchanged
 *             <strOption> is "-STAT"
 *                <logStat> is set to 2
 * failure: <strOption> does not belong to "-stat", "-STAT"
 */
static boolean
get_statisticsOption (char *strOption, int *logStat)
{
  if (strcmp (strOption, "-stat") == 0)
    {
      if (*logStat > 0)
	the_mainReporter->log_resetStatArg ();
      else
	*logStat = 1;
      return pk_true;
    }
  if (strcmp (strOption, "-STAT") == 0)
    {
      if (*logStat > 0)
	the_mainReporter->log_resetStatArg ();
      *logStat = 2;
      return pk_true;
    }
  return pk_false;
}

/** check if <varName> is a mkf user variable name
    on success retruns a new mallocated copy of varNam
    exit on error
 */
static char *
get_varName (char *varName)
{
  char *CString_name;
  if (InteLineAnalyzer::is_userVarName (varName))
    {
      CString_name = newCstr_String (varName);
      if (CString_name == NIL_String)
	exitArgErrOn ("memory allocation failed", 5);
      return CString_name;
    }
  exitArgErrOn ("incorrect variable name", 5);
  return NIL_String;
}

/* reads a (composite) argument
 * <argc> is the number of arguments
 * <argIndex> is the value of next argument index, <argc> < <argc>
 * <argv> is the list of arguments
 * success: <argv>[*<argIndex>] is "-D<ident>",
 *                    where <ident> is a C identifier
 *              1-<argv>[*<argIndex>+1] is "<markChar><STRING>",
 *                    where <markChar> is a printable character and 
 *                          <STRING>=[<character>*] is a string for the shell language that does not start with <markChar>
 *                 A static string identifier named <ident> is created; 
 *                     the identifier value is <STRING>
 *              2-<argv>[*<argIndex>+1] is "<markChar0><markChar0>",
 *                <argv>[*<argIndex>+n] is "<markChar0><markChar0>", for n>1
 *                <argv>[*<argIndex>+i] is "<markChar(i-1)><STRING(i-1)>, for 1<i<n
 *                    where <markCharj> is a printable character for j in {0..n}
 *                          <STRING(j)>=[<character>*] is a string for the shell language, such that <markChar(j)><STRING(j)> is not <markChar0><markChar0>
 *                 A static list identifier named <ident> is created; 
 *                     if n=2 the value of <ident> is the empty string,
 *                     if n>2 its value is (<STRING(2)>,...<STRING(n-1)>>
 * failure: syntax error
 *             error message is printed
 *             error exit
 */
static boolean
get_varDefinition (char *strArg, int argc, char *argv[], int *argIndex)
{
  if (strArg[1] == '=')
    {
      if (strlen (strArg) > 2)
	{
	  char *identifierName = get_varName ((char *) (strArg + 2));
	  if ((*argIndex) <= argc)
	    {
	      strArg = argv[*argIndex];
	      (*argIndex) = (*argIndex) + 1;
	      if (Context::the_staticContext->appendNew_staticStringElement (NIL_ArgumentList, identifierName, new StringArgument (strArg)))
		return pk_true;
	      exitErr ();
	    }
	  else
	    exitArgErrOn ("argument missing ", 5);
	}
      else
	exitArgErrOn ("variable name missing", 5);
    }
  return pk_false;
}

/* reads a (composite) argument
 * <argc> is the number of arguments
 * <argIndex> is the value of next argument index, <argc> < <argc>
 * <argv> is the list of arguments
 * success: node root argument, <rootNodeIsDefined> is false
 *             - <argv>[*<argIndex>] is "-node", <argIndex> < <argc>-1 
 *                <rootVariableNode> gets <argv>[*<argIndex>] value
 *                <allRoot> remains unchanged
 *             - <argv>[*<argIndex>] is "-all", <argIndex> < <argc>
 *                <rootVariableNode>, <argIndex> remain unchanged
 *                <allRoot> is set to false
 *             - <argv>[*<argIndex>] is "-ALL", <argIndex> < <argc>
 *                <rootVariableNode>, <argIndex> remain unchanged
 *                <allRoot> is set to pk_true
 *             <rootNodeIsDefined> is set to pk_true
 *          building mode 
 *             warning message is printed if building mode option has several occurrences
 *             - <argv>[*<argIndex>] is "-r"
 *                <force> is set to max(1, <force>)
 *             - <argv>[*<argIndex>] is "-R"
 *                <force> is set to 2
 *          loging mode 
 *             warning message is printed if loging mode option has several occurrences
 *             <strOption> is "-v"
 *                <verboseLevel> is set to max(1, <verboseLevel>)
 *             <strOption> is "-V"
 *                <verboseLevel> is set to max(2, <verboseLevel>)
 *             <strOption> is "-d"
 *                <verboseLevel> is set to max(3, <verboseLevel>)
 *             <strOption> is "-D"
 *                <verboseLevel> is set to 4
 *          <argv>[*<argIndex>] is "-ignore"
 *             warning message is printed if "-ignore" option has several occurrences
 *             <ignoreError> is set to pk_true
 *          statistics otions
 *             warning message is printed if statictics mode option has several occurrences
 *             <strOption> is "-stat"
 *                <logStat> is set to max(1, <logStat>)
 *             <strOption> is "-STAT"
 *                <logStat> is set to 2
 *          <argIndex> is incremented by the size of current (composite) argument
 * failure: syntax error
 *             error message is printed
 *             error exit
 */
static void
get_otherOption (char *strOption, int *force, boolean * ignoreError, int *logStat)
{
  if (get_buildingModeOption (strOption, force))
    return;
  if (get_logingModeOption (strOption))
    return;
  if (strcmp (strOption, "-ignore") == 0)
    {
      if (*ignoreError)
	the_mainReporter->log_ignoreArg ();
      else
	*ignoreError = pk_true;
      return;
    }
  if (get_statisticsOption (strOption, logStat))
    return;
  exitArgErrOn (strOption, 2);
}

/* ============== Command arguments
 * see infoCorrectArgs function
 */
static void
getArguments (int argc, char *argv[],
	      boolean * rootNodeIsDefined, char **rootVariableNode, boolean * allRoot,
	      int *force, boolean * ignoreError, int *logStat, char **rootPath)
{
  int argIndex = 1;
  char *strOption;
  /* add exec tag and execution of exec tagged files */
  while (argIndex <= argc)
    {
      strOption = argv[argIndex];
      argIndex = argIndex + 1;
      if (strOption[0] == '-')
	{
	  if (!getRoot_nodeOption (strOption, argc, &argIndex, argv, rootNodeIsDefined, rootVariableNode, allRoot))
	    if (!get_varDefinition (strOption, argc, argv, &argIndex))
	      get_otherOption (strOption, force, ignoreError, logStat);
	}
      else if ((*rootPath) == NIL_String)
	(*rootPath) = strOption;
      else
	exitArgErrOn (strOption, 3);
    }
}

static boolean
readingStep (char *rootPath, ref2ContextTree * rootContextTree)
{
  ref2MKFFileNode rootFileNode;
  boolean lCorrect = pk_true;
  the_mainReporter->openStep (1);
  if (rootPath == NIL_String)
    {
      char *stdinName = newCstr_String ("stdin");
      rootFileNode = new MKFFileNode ();
      (*rootContextTree)->register_staticFile (rootFileNode, stdinName);
      lCorrect = rootFileNode->load_stdinContext (rootContextTree, stdinName);
    }
  else
    {
      char *rootedPathName = getEnv_RootedPathName (rootPath);	/* in systemFrontEnd_e.h */
      if (rootedPathName == NIL_String)
	{
	  the_mainReporter->log_fileArgumentErrorClosing (rootPath);
	  exit (-1);
	}
      rootFileNode = new MKFFileNode (rootPath, rootedPathName, &lCorrect);
      (*rootContextTree)->register_rootFile (rootFileNode, rootPath);
      if (lCorrect)
	{
	  lCorrect = rootFileNode->load_mainFileContext (rootContextTree);
	}
    }
  the_mainReporter->closeStep (1);
  return lCorrect;
}

static boolean
dependenceCheckingStep (boolean rootNodeIsDefined, char *rootVariableNode, ref2ContextTree rootContextTree, int force, boolean allRoot)
{
  // fprintf(stderr, "DEBUG dependenceCheckingStep\n");
  boolean lCorrect = pk_true;
  the_mainReporter->openStep (2);
  if (rootNodeIsDefined)
    {
      if (rootVariableNode == NIL_String)
	{
	  if (allRoot)
	    {
	      lCorrect = rootContextTree->allRoots_dependenceChecking (force);
	    }
	  else
	    {
	      lCorrect = rootContextTree->allLocalRoots_dependenceChecking (force);
	    }
	}
      else
	{
	  lCorrect = rootContextTree->root_dependenceChecking (rootVariableNode, force);
	}
    }
  else
    {
      lCorrect = rootContextTree->allLocalRoots_dependenceChecking (force);
    }
  the_mainReporter->closeStep (2);
  return lCorrect;
}

static void
generationStep (ref2ContextTree rootContextTree)
{
  rootContextTree->debug_constructiveNodeList ();
  the_mainReporter->openStep (3);
  if (!(rootContextTree->generate_constructiveNodeList ()))
    closeOnError ();
  the_mainReporter->closeStep (3);
}


int
main (int argc, char *argv[])
{
  char *rootPath = NIL_String;	/* "" for stdin or a file path */
  boolean rootNodeIsDefined = pk_false;	/* pk_true if a root label is provided or if -all or -ALL options has been met */
  boolean allRoot;		/* meanningfull when rootNodeIsDefined is pk_true and rootVariableNode is NIL_String
				 *      pk_true if all constructive nodes in all files must be considered as root */
  char *rootVariableNode = NIL_String;	/* a root label if rootNodeIsDefined is pk_true, -all or -ALL */
  int force = 0;		/* -1 for -R, 0 , 1 for -r */
  boolean ignoreError = pk_false;
  int logStat = 0;		/* 0, 1, 2 */

  ref2ContextTree rootContextTree;
  boolean lCorrect = pk_true, fullyCorrect = pk_true;

  int constructiveNodeNumber;
  int passiveNodeNumber;

  initFileSystemEnv ();		/* in systemFrontEnd_e.h */

  if (!(Soundness::setStatic ()))
    exit (-1);			/* opening stderr should not fail */
  the_mainReporter = new new_pKMakeReporting ();
  rootContextTree = Context::setStatic ();
  NodeElement::setStatic ();

  /* PARAMETERS
   */
  if (argc >= 2)
    {
      if (strcmp ("-h", argv[1]) == 0)
	{
	  the_mainReporter->infoCorrectArgs ();
	  exit (0);
	}
      getArguments (argc - 1, argv, &rootNodeIsDefined, &rootVariableNode, &allRoot, &force, &ignoreError, &logStat, &rootPath);
    }


  /* ================= Reading step
   */
  fullyCorrect = readingStep (rootPath, &rootContextTree);
  if (!fullyCorrect)
    {
      if (!ignoreError)
	closeOnError ();
      the_mainReporter->log_mainErrorContinue ();
    }
  /* ================= Dependence analysis step
   */
  lCorrect = dependenceCheckingStep (rootNodeIsDefined, rootVariableNode, rootContextTree, force, allRoot);
  if (!lCorrect)
    {
      if (!ignoreError)
	closeOnError ();
      the_mainReporter->log_mainErrorContinue ();
      fullyCorrect = pk_false;
    }
  rootContextTree->log_unreferedNodes ();

  /* ================= File generation step
   */
  constructiveNodeNumber = rootContextTree->get_constructiveNodeNumber ();
  passiveNodeNumber = rootContextTree->get_passiveNodeNumber ();
  if (constructiveNodeNumber > 0)
    {
      int newConstructiveNodeNumber, newPassiveNodeNumber;
      fprintf (stderr, "DEBUG maximal index is %i\n", ContextTree::the_elementNumber);
      lCorrect = dependenceCheckingStep (rootNodeIsDefined, rootVariableNode, rootContextTree, force, allRoot);
      newConstructiveNodeNumber = rootContextTree->get_constructiveNodeNumber ();
      newPassiveNodeNumber = rootContextTree->get_passiveNodeNumber ();
      generationStep (rootContextTree);	/* exit on error */
      if (fullyCorrect)
	{
	  the_mainReporter->log_mainSuccessClosing (pk_true);
	  exit (0);
	}
    }
  else if (fullyCorrect)
    {
      the_mainReporter->log_mainSuccessClosing (pk_false);
      exit (0);
    }
  the_mainReporter->log_mainErrorClosing ();
  exit (-1);
}
