/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*
 * This file is strongly host system dependent
 */

/**
   @returns a new malloc'd string containing the symbolic pathname translation of 
   NULL terminated string source; source can be empty string
   returns NIL_String if some environnement variable is missing
*/
extern char *
convertParemeterizedName (char *source)
{
  ref2VariableType refVariable;
  int sourceIndex = 0, fullLength = 0;
  size_t sourceLength = strlen (source);
  size_t targetIndex = 0;
  int lg;
  boolean symbolic = pk_false;

  while ((sourceIndex < (int) sourceLength) && (fullLength < FILENAME_MAX))
    {
      if ((source[sourceIndex]) == envNameTag)
	{
	  if (!(symbolic))
	    {
	      targetIndex = resetPathBuffer (source, sourceIndex);
	      symbolic = pk_true;
	    }
	  sourceIndex = sourceIndex + 1;
	  refVariable = findVariable ((char *) (source + sourceIndex));
	  if (refVariable->value == NIL_String)
	    {
	      fprintf (stderr, "## ERROR ## %s: Undefined variable\n", refVariable->envName);
	      return NIL_String;
	    }
	  sourceIndex = sourceIndex + strlen (refVariable->envName);
	  lg = append2Target (refVariable->value, targetIndex);
	  targetIndex = targetIndex + lg;
	}
      if (symbolic)
	{
	  absoluteFilePathName[targetIndex] = source[sourceIndex];
	  targetIndex = targetIndex + 1;
	}
      sourceIndex = sourceIndex + 1;
    }
  absoluteFilePathName[targetIndex] = (char) 0;
  if (symbolic)
    return strduplicate (absoluteFilePathName);
  return source;
}

/**
  varName is a 0 terminated string denoting a variable environnement
    - success returns the environnement variable value
      the resulting string must not be modified,
    - failure returns NIL_String
 */
extern char *
getEnvDef (char *varName)
{
  return (findVariable (varName))->value;
}

/**
   It defines an environment variable (eVar) by an associated data (vVar).
   It can be used to avoid the access to the execution environment (shell variables).
*/
extern void
setEnvDef (char *eVar, char *vVar)
{
  ref2VariableType refVariable;

  refVariable = newVariable (strduplicate (eVar), strduplicate (vVar));
  registerVariable (refVariable);
}

/**
   @return true if an environment variable (eVar) is already defined, false otherwise.
*/
extern boolean
isEnvDef (char *eVar)
{
  ref2VariableType refVariable;
  return isDefinedVariable (eVar, &refVariable);
}

/**
   container is a non symbolic directory pathname.
   @returns the full path name of fileName in container in a new malloc'd string
   the result may be an invalid filename
 */
extern char *
getFilePathName (char *container, char *fileName)
{
  char *filePathName;
  int lgs = strlen (container);
  if (container[lgs] == fileSeparator)
    {
      filePathName = (char *) calloc (lgs + strlen (fileName) + 1, 1);
      filePathName = strcpy (filePathName, container);
    }
  else
    {
      filePathName = (char *) calloc (lgs + strlen (fileName) + 2, 1);
      filePathName = strcpy (filePathName, container);
      filePathName[lgs] = fileSeparator;
    }
  filePathName = strcat (filePathName, fileName);
  return filePathName;
}

/**
   Initialization of the module.
*/
extern void
initFileSystemEnv ()
{
  absoluteFilePathName = (char *) calloc (FILENAME_MAX + 1, 1);
  curErrMess = (char **) calloc (sizeof (char *), 1);
}
