/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
 * This file is strongly host system dependent
 */

ref2LstRefVarType variableList = NIL_LstRefVarType;

static char *absoluteFilePathName;
static char **curErrMess;


static char *
strduplicate (char *absoluteFilePathName)
{
  char *x;
  int n;

  n = (int) strlen (absoluteFilePathName);
  x = (char *) malloc (n + 1);
  memset (x, 0, n);
  x = strcat (x, absoluteFilePathName);
  return x;
}

/** ===================================================================== 
 * Memory level variable management
 *
 */

/**
   @return a new, created, aliasing from varName to pathName.
 */
static ref2VariableType
newVariable (char *varName, char *pathName)
{
  ref2VariableType refVariable = (ref2VariableType) malloc ((unsigned) (sizeof (*refVariable)));
  refVariable->envName = varName;
  refVariable->value = pathName;
  refVariable->inError = (pathName == NULL);
  return refVariable;
}

/**
  @return  a new, created, aliasing entry in association table.
 */
static void
registerVariable (ref2VariableType newVariable)
{
  ref2LstRefVarType newatom = (ref2LstRefVarType) malloc ((unsigned) (sizeof (*newatom)));
  newatom->car = newVariable;
  newatom->cdr = variableList;
  variableList = newatom;
}


/**
   @return a created string that is the copy of the length first characters of source
   to conform to strings, it is terminated by the C string terminator (0)
   requires source to be a string with at least length characters.
 */
static char *
clonePrefix (char *source, size_t length)
{
  char *shvar = (char *) malloc ((unsigned) (length + 1));
  memcpy (shvar, source, length);
  shvar[length] = 0;
  return shvar;
}

/**
   source begins with a variable character.
   @return the length of the longest identifier prefix in source.
 */
static size_t
getVariableLength (char *source)
{
  return strcspn (source, " /!\n");
}

static boolean
isstrnprefix (char *ref, char *source, int length)
{
  // if ( (int) strlen(ref) != length) return pk_false;
  return (strcmp (ref, source) == 0);
  // return (strncmp(ref, source, length) == 0 );
}

/**
   varName is a 0 terminated string denoting a variable environnement.
   @return 
   - (success) a copy of the environnement variable value
   - (failure) returns NIL_String
 */
static char *
getEnvStrDef (char *varName)
{
  char *shvar;

  shvar = getenv (varName);
  if (shvar == NULL)
    return NIL_String;
  return strduplicate (shvar);
}

/**
   source contains a symbolic name.
   @return a pointer to the pair (var name, var value).
   If var name cannot be translated set var value to NIL_String
 */
static ref2VariableType
findVariable (char *source)
{
  ref2VariableType refVariable;
  char *varName;
  char *pathName;
  ref2LstRefVarType searchedList = variableList;
  size_t length = getVariableLength (source);

  while (searchedList != NIL_LstRefVarType)
    {
      refVariable = searchedList->car;
      if (isstrnprefix ((refVariable)->envName, source, length))
	return refVariable;
      searchedList = searchedList->cdr;
    }

  varName = clonePrefix (source, length);
  pathName = getEnvStrDef (varName);
  refVariable = newVariable (varName, pathName);
  registerVariable (refVariable);
  return refVariable;
}


/** 
    source contains a symbolic name.
    @return "true" if source is already defined and refVariable is this variable, "false" otherwise.
 */
static boolean
isDefinedVariable (char *source, ref2VariableType * refVariable)
{
  ref2LstRefVarType searchedList = variableList;
  size_t length = getVariableLength (source);

  while (searchedList != NIL_LstRefVarType)
    {
      (*refVariable) = searchedList->car;
      if (isstrnprefix ((*refVariable)->envName, source, length))
	return pk_true;
      searchedList = searchedList->cdr;
    }
  return pk_false;
}


/** 
    @return 
    - false if source doesn't contain symbolic part, target remains unchanged,
    - pk_true if source contains symbolic parts target contains full 
           name with converted symbolic parts.
 */
static size_t
resetPathBuffer (char *source, size_t length)
{
  memcpy (absoluteFilePathName, source, length);
  absoluteFilePathName[length] = '\0';
  return length;
}

/**
  append the string source to absoluteFilePathName from pos targetIndex.
  @return the size of the source string.
 */
static int
append2Target (char *source, int targetIndex)
{
  int k = 0;
  int lg = strlen (source);
  while (k < lg)
    {
      absoluteFilePathName[targetIndex] = source[k];
      targetIndex = targetIndex + 1;
      k = k + 1;
    }
  return lg;
}
