/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/**
 * This file is strongly host system dependent
 */

/**
   a GNU variant of lineget.

   Read up to (and including) a newline from STREAM into *LINEPTR
   (and null-terminate it). *LINEPTR is a pointer to lSize+1 characters of space. 
 
 @return 
   - success: the number of characters read (not including the null terminator)
   - failure: -1 on EOF, -2 on error
 */
extern int
getpKLine (char *lineptr, int lSize, FILE * thisFile)
{
  char *res = fgets (lineptr, lSize, thisFile);
  if (res == lineptr)
    return (strlen (res));
  if (feof (thisFile))
    return (-1);
  return -2;
}

/**
   close a file descriptor
   host system direct dependency: function fclose.
   @param file (in) file to close
 */
extern void
rwFClose (ref2File file)
{
  fclose (file);
}

/**
   open a file in writing mode and returns its new system file descriptor
   - host system direct dependency: function fopen.
   - PK direct dependency: constant NIL_String, NIL_File
   
   @param filePathName (in) string terminated by a standard 0 refering to a file full identification
   filePathName contains only authorised characters in the host system.
   @param errMess (out) an error message in case of failure or NIL_String.

   @return 
     - success: open file system descriptor
     - failure: NIL_File
*/
extern ref2File
wFOpen (char *filePathName, ref2String errMess)
{
  ref2File writeFile;
  if (filePathName == NIL_String)
    {
      *errMess = (char *) "empty path name";
      return (NIL_File);
    }
  writeFile = fopen (filePathName, "w");
  if (writeFile == NULL)
    {
      *errMess = (char *) "cannot be opened in writing mode";
      return NIL_File;
    }
  *errMess = NIL_String;
  return writeFile;
}

/**
   open a file in reading mode and returns its new system file descriptor
   - host system direct dependency: function fopen
   - PK direct dependency: constant NIL_String, NIL_File

   @param filePathName (in) string terminated by a standard 0 refering to a file full identification
   filePathName contains only authorised characters in the host system
   @param errMess (out) an error message in case of failure or NIL_String.

   @return 
   - success: open file system descriptor
   - failure: NIL_File
 */
extern ref2File
rFOpen (char *filePathName, ref2String errMess)
{
  ref2File readFile;
  if (filePathName == (NIL_String))
    {
      *errMess = (char *) "empty path name";
      return NIL_File;
    }
  readFile = fopen (filePathName, "r");
  if (readFile == NULL)
    {
      *errMess = (char *) "cannot be opened in reading mode";
      return NIL_File;
    }
  *errMess = NIL_String;
  return readFile;
}
