/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/*
 * This file is strongly host system dependent
 */

/**
   @returns a time_t value formated as YYYY-MM-JJ HH:MM:SS
 */
extern char *
strcvtTime (time_t value)
{
  static char buffer[timeStrSize + 1];
  const char *format = "%Y-%m-%d %H:%M:%S";
  struct tm *tm = localtime (&value);
  size_t length = strftime (buffer, timeStrSize + 1, format, tm);
  return buffer;
}

/**
   host system direct dependency:
   - function time 

   @returns current date
 */
extern time_t
getNow ()
{
  return (time_t) (time ((ref2Time) 0));
}

/**
   get filePathName last modification date; file may be missing.\n
   host system direct dependency: 
      - types struct stat
      - constant ENOENT
      - variable errno
      - function stat

   @param filePathName (in) string terminated by a standard 0 refering to a file full identification.
           filePathName contains only authorised characters in the host system.

   @param errMess (out) an error message in case of failure or NIL_String.
   @param  newf (out):
         - sucess : false if filePathName denotes an existing file, pk_true otherwise.
	 - failure: false

   @return 
     - success: local last modification date in system units if file exists apocalypse otherwise
     - failure: NIL_time
 */
extern time_t
getFDate (char *filePathName, boolean * newf, ref2String errMess)
{

#if (xKHOST != xKwinNT)
  struct stat buf;

  if (NIL_time == genesia)
    {
      *newf = pk_false;
      *errMess = (char *) "Damned: systemFrontEnd has not been initiated";
      return NIL_time;
    }
  if (filePathName == NIL_String)
    {
      *errMess = (char *) "empty file path name";
      return (NIL_time);
    }
  if ((stat (filePathName, &buf)) == 0)
    {
      *newf = pk_false;
      *errMess = NIL_String;
      return ((time_t) (buf.st_mtime));
    }
  if (errno == ENOENT)
    {
      errno = 0;
      *newf = pk_true;
      *errMess = NIL_String;
      return apocalypse;
    }
#endif
  errno = 0;
  *newf = pk_false;
  *errMess = (char *) "file access error";
  return NIL_time;
}

extern time_t
initDate (ref2Time apoca)
{
  if (now == NIL_time)
    now = (time_t) (time ((ref2Time) 0));
  apocalypse = (time_t) (now + oneFYear);
  *apoca = apocalypse;
  return now;
}
