/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* ---------------------------------------------------------------------
   Documentation : file logPrinter 
   issued from automatic Pascal to C translation 
   -------------------------------------------------------------------- */

/*
 * direct printing
 */
static void logPrintLine()
{
  if (wcc > wccMin) logPrintEol(); 
}

static void logSkipLine()
{
  if (wcc > wccMin) logPrintEol(); 
  logPrintEol();
}

static void logSkipnLine(int n)
{
  int i;
  if (wcc > wccMin) logPrintEol();
  for ( i = 0; i <= n; i++ )
    logPrintEol();
}

/*
 * pretty printing
 */
static void getspace (int lg)
{
  if (wcc + lg > lineSize) logPrintLine();
  if (wcc == 0) {
    int i,n;

    if (imbric > 0) {
      logOutChr(commentMKFTag);
      if (imbric <= maximbric) n = imbric; else  n = maximbric;
      for (i = 1; i<= n; i++) 
	logOutStr("   |");
      wccMin = 4 * n;
      if (imbric > n) {
	wccMin = wccMin +4;
	logOutStr("... "); } }
    else 
      wccMin = 0; }  
  wcc = wcc + lg;
}

static void logPrintString(char *str)
{
  int n = strlen(str);
  getspace(n);
  logPrintStr(str,n);
}

static void logPrintStringLine(char *str)
{
  logPrintString(str);
  logPrintLine();
}

static void logPrintReleaseDate(time_t date)
{
  getspace (10+timeStrSize);
  logPrintnForm(" release: %s", strcvtTime(date), 10+timeStrSize);
}

/* 
 *
 */
static void logPrintCommentMark()
{
  getspace (2);
  logPrintnForm("%c ", commentMKFTag, 2); 
}

static void logPrintOpenMark()
{
  getspace (2);
  logPrintnForm("%c ", openTag, 2); 
}

static void logPrintCloseTag()
{
  getspace (2);
  logPrintnForm("%c", closeTag, 2); 
}

static void logPrintDirectMark()
{
  getspace (2);
  logPrintnForm(" %c", writeFileTag, 2); 
}

static void logPrintDonotMark()
{
  getspace (2);
  logPrintnForm(" %c", donotTag, 2); 
}

static void logPrintHandMadeMark()
{
  getspace (2);
  logPrintnForm(" %c", handMadeTag, 2); 
}

static void logPrintObjectClosure()
{
  getspace(1);
  logPrintChr(closeTag); 
  logPrintLine(); 
}

static void logPrintPreludeOpening()
{
  logPrintOpenMark();
  logPrintLine();
}

static void logPrintDefaultOpening()
{
  getspace(2);
  logPrintChr(closeTag);
  logPrintChr(defaultTag);
  logPrintLine ();
}

static void logPrintVarName(ref2ObjectType obj)
{
  logPrintString( (char*) obj->objectLabel); 
}

static void logPrintEnvVarName(char *name)
{
  int n = strlen(name) + 1;
  getspace(n);
  logPrintnForm("$%s", name, n); 
}

static void logPrintExtName (ref2ObjectType obj)
{
  int n = strlen(obj->objectLabel);
  getspace(n+2);
  logPrintChr(startFileTag);
  logPrintStr(obj->objectLabel, n);
  logPrintChr(endFileTag);
}

static void logPrintEnvName (ref2ObjectType obj)
{
  int n = strlen(obj->objectLabel);
  getspace(n+2);
  logPrintChr(startEnvTag);
  logPrintString( (char*) obj->objectLabel);
  logPrintChr(endEnvTag);
}

static void logPrintObjectDate (ref2ObjectType obj)
{
  if ( obj->created == NIL_time ) {
    return;}
  if ( obj->created == genesia ) {
    logPrintString( (char*) " not existing file");
    return; }
  if ( obj->created == apocalypse ) {
    logPrintString( (char*) " to be created");
    return; }
  logPrintReleaseDate( obj->created );
}

static void logPrintObjectName (ref2ObjectType obj)
{
  if (obj->internal) logPrintVarName(obj);
  else if (obj->isFile) logPrintExtName(obj);
  else logPrintEnvName(obj);
}

static void logPrintObjectOwning(ref2ObjectType obj)
{
  int n;
  getspace(12);
  if ((obj-> owner) != NIL_ObjectType) 
    logPrintForm(" (in %i)", obj->owner->ordinal, n );
}

static void logPrintObjectNumber(ref2ObjectType obj)
{
  int n;
  getspace(7);
  if ((obj-> owner) == NIL_ObjectType) 
    logPrintForm("(%i)", obj->ordinal,n );
  else {
    getspace(7);
    logPrintForm("(%i::",  obj->owner->ordinal,n);
    logPrintForm("%i)",  obj->ordinal,n); }
}

static void logPrintObjectRef(ref2ObjectType obj)
{
  logPrintObjectName(obj);
  logPrintCommentMark();
  logPrintObjectNumber(obj);
  if (obj->inError) logPrintString( (char*) " ***"); 
}

static void logPrintObjectRefDate(ref2ObjectType object)
{
  logPrintCommentMark();
  logPrintObjectRef(object);
  if ( (object-> ordinal) == 0 ) return;
  logPrintObjectDate(object);
  if (object->inError) logPrintString( (char*) " erroneous"); 
}

static void logPrintObjectRefDateLine(ref2ObjectType object)
{
  logPrintObjectRefDate(object);
  logPrintLine();
}

static void logPrintFileFullRef(ref2ObjectType object)
{
  int n;
  logPrintExtName(object);
  logPrintObjectOwning(object);
  if ( (object-> ordinal) == 0 ) return;
  if (object->inError) {
    logPrintString( (char*) " *** erroneous"); 
    return; }
  if (object->objectAlias == NIL_String ) return;
  if (object->objectLabel == object->objectAlias ) return;
  logPrintLine();
  logPrintCommentMark();
  logPrintForm("    alias: %s", object->objectAlias, n);
}

static void logPrintBuildingObj(ref2ObjectType obj)
{
  logPrintLine ();
  logPrintCommentMark();
  logPrintString( (char*) "+++ build object ");
  logPrintObjectRef(obj);
  logPrintLine ();
}

static void logPrintPage (void)
{
  int logLineLength = 80;
  int i;
  if ( wcl > 1 ) {
    logSkipnLine(5);
    wcl = 0; }
  logPrintCommentMark();
  for (i= 1; i <= logLineLength; i++) logPrintChr('=');
  wcc = logLineLength;
  logPrintLine();
}

static void logPrintObjectHeading(ref2ObjectType object)
{
  logSkipLine();
  logPrintCommentMark();
  logPrintnForm("----(%4d)----> declared object: ", object->ordinal, 33);
  if (object->internal) {
    logPrintVarName(object);
    if (object->inError) logPrintString( (char*) " *** erroneous"); }
  else if (object->isFile)
    logPrintFileFullRef(object);
  else logPrintEnvName(object);
  logPrintLine();
}

static void logPrintPredecessorLine(ref2ObjectType object)
{
  int lg = strlen(object->objectLabel);
  getspace(lg+7);
  logPrintChr(closeTag);
  logPrintChr(includeTag);
  logPrintString( (char*) "     ");
  logPrintObjectRef(object);
  logPrintLine();
}

static void logPrintOpenObjectMark(ref2ObjectType object)
{
  logPrintOpenMark();
  logPrintObjectName(object);
}

static void logPrintIncludeDecl(ref2ObjectType object)
{
  logPrintCommentMark();
  logPrintOpenObjectMark(object);
  logPrintnForm("%c ", includeTag, 2); 
}

static void logPrintReadMKF (ref2ObjectType object)
{
  logSkipnLine(2);
  logPrintCommentMark();
  logPrintnForm("========(%4d)========> read file: ", object->ordinal, 35);
  logPrintFileFullRef(object);
  logSkipLine();
}

/*
 * INFORMATION messages
 */
static void logPrintInfo(char *mess)
{
  logPrintCommentMark(); 
  logPrintString( (char*) "MKF Info: "); 
  logPrintStringLine(mess);
}

static void logPrintInfoTitle(boolean verbose, char *mess)
{
  if (verbose) logPrintPage();
  else logPrintLine();
  logPrintInfo(mess);
  logSkipLine();
}

static void logPrintInfoObjTitle(boolean verbose, char *mess, ref2ObjectType fileObject)
{
  if (verbose) logPrintPage();
  else logPrintLine();
  logPrintInfo(mess);
  logPrintCommentMark();
  logPrintObjectName(fileObject);
  logSkipLine();
}

/*
 * WARNING  messages
 */
static void logPrintWar(boolean line, char *mess)
{
  int n;
  if (line) logPrintLine ();
  logPrintCommentMark();
  logPrintForm("-- MKF WARNING --:: %s", mess,n); 
}

static void logPrintWarTitle(char *mess)
{
  logPrintPage();
  logPrintWar(true, mess);
  logSkipLine();
}

/*
 * ERROR  messages
 */
static void logPrintErr(boolean line, char *mess)
{
  int n;
  if (line) logPrintLine ();
  logPrintCommentMark();
  logPrintForm("## MKF ERROR ##: %s", mess,n); 
}

static void logPrintErrStringMess(char *str, char *mess)
{
  int n;
  logPrintErr(true, str); 
  logPrintForm(" %s", mess, n); 
  logPrintLine();
}

static void logPrintErrObj (ref2ObjectType obj)
{
  if (obj->internal) logPrintErr(false, (char*)"variable ");
  else logPrintErr(false, (char*) "file ");
  logPrintObjectRef(obj);
  if (obj->inError) logPrintString( (char*) " is in error");
  else if (obj->internal) logPrintString( (char*) " is not defined");
  else logPrintString( (char*) " cannot be opened");
  logPrintLine();
}

static void logPrintErrCycle(ref2ObjectType obj)
{
    logPrintErr(true, (char*)" A cycle contains ");
    logPrintObjectRef(obj);
    logPrintLine();
}

static void logPrintErrObjMess (ref2ObjectType obj, char *mess)
{
  int n;

  logPrintErr(true, (char*)"file "); 
  logPrintObjectName(obj);
  logPrintForm("%s", mess,n);
  logPrintLine ();
}

static void logPrintHeadSyntax(int cl, int cc, boolean isError)
{
  int n;
  if (isError) logPrintErr(true, (char*)" in file ");
  else logPrintWar(true, (char*)" in file ");
  logPrintObjectName(curFileObject);
  logPrintForm(", line %i", cl, n);
  logPrintForm(", column %i", cc, n);
  logPrintLine ();
  logPrintCommentMark();
  logPrintString( (char*) "                 ");
}

static void logPrintConflict(int cl, int cc, boolean isError, ref2ObjectType newob, char *mess)
{
  int n;
  logPrintHeadSyntax(lineCounter, cc, isError);
  logPrintObjectName(newob);
  logPrintForm(", object %s", mess, n);
  logPrintLine();
}

static void logPrintSyntax(int cl, int cc, char *mess)
{
  int n;
  logPrintHeadSyntax(lineCounter, cc, true);
  logPrintString( (char*) mess);
  logPrintLine();
}

static void initPrinter()
{
  wcc = 0;
  wcl = 1;
  imbric = 0;
}

