/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/* Definitions for case-statements */

//#define Line __LINE__
static void
Caseerror (int n)
{
  fprintf (stderr, "Missing case limb: line %d\n", n);
  fflush (stderr);
  exit (1);
}

#define doneTitle (char*) " done"
#define failTitle (char*) " failed"
#define readTitle (char*) " MKF make file loading"
#define fileContentTitle (char*) " object(s) defined in "
#define nodefTitle (char*)" undefined object(s) "
#define graphTitle (char*)" dependence analysis"
#define buildTitle (char*)" file construction "
#define directFileTitle (char*)" files marked for direct building"
#define constructionFileTitle (char*)" files to build"
#define norefTitle (char*)" unreferred to object(s) "
#define missingTitle (char*)" missing file(s) "

#define conflictErrMess (char*)"Type conflict"
#define doubleErrMess (char*)"previously defined, ignored"

typedef struct ObjectType *ref2ObjectType;

typedef struct ActionLstType *ref2ActionLstType;
typedef struct ActionLstType
{
  char *car;
  ref2ActionLstType cdr;
} t1;
#define NIL_ActionLstType (ref2ActionLstType) NULL

typedef struct PredecessorLstType *ref2PredecessorLstType;
typedef struct PredecessorLstType
{
  ref2ObjectType car;
  boolean neverBuild;
  ref2ActionLstType postConsAction;
  ref2PredecessorLstType cdr;
} t2;
#define NIL_PredecessorLstType (ref2PredecessorLstType) NULL

typedef struct ActionType *ref2ActionType;
typedef struct ActionType
{
  boolean handmade, direct, conditional;
  ref2ActionLstType prelConsAction;
  ref2PredecessorLstType innerLst;
  ref2ActionLstType noconsAction;
} t3;
#define NIL_ActionType  (ref2ActionType) NULL

typedef struct LstRefObjType *ref2LstRefObjType;
typedef struct LstRefObjType
{
  ref2ObjectType car;
  ref2LstRefObjType cdr;
} t4;
#define NIL_LstRefObjType  (ref2LstRefObjType)NULL

typedef struct ObjectType
{
  boolean inError, internal;
  boolean isFile;		/* is pk_true when object is a defined file, false otherwise */
  boolean defined;		/* pk_true when object has been defined, or set as a mkf file to be read */
  ref2ObjectType owner;		/* mkf containing its definition */
  boolean include;		/* pk_true when object is a mkf file */
  int ordinal;

  char *objectLabel;
  char *objectAlias;
  time_t created;

  boolean visited;		/* pk_true when a file has been visited during graph exploration */
  ref2ObjectType made;		/* used during building: file for which object has been built */
  boolean stated;
  ref2ActionType actions;
  ref2LstRefObjType firstPred, lastPred;
} t5;
#define NIL_ObjectType (ref2ObjectType) NULL

#define hashsize 41		/* size oh hash table; must be > 0 */

static ref2LstRefObjType hashEntries[hashsize];
static int imbric;

static time_t now;		/* denotes objects that are being built;
				 * all predecessor dates have been correctly stated
				 */
static time_t apocalypse;	/* denotes missing objects
				 */
static FILE *currentWriteFile;


/* ============================================= Management of mkf files */
/* ========== all file objects */
static ref2LstRefObjType firstRefFileObject, curRefFileObject, lastRefFileObject;
static ref2ObjectType curFileObject;	/* pointer to object addressed by curRefFileObject */

/* ========== inner objects during file loading */
static ref2ObjectType curBuiltObject, curRefObject;

/* ========== source file objects for filegeneration */
static ref2LstRefObjType firstDirectFile, lastDirectFile;

/* ========== list of files to be generated*/
static ref2LstRefObjType firstFileConstruction, lastFileConstruction;	/* sorted during dependence analysis */

/* ========== object count */
static int objectNumber;

// static boolean verbose = pk_false;
static boolean verbose = pk_true;

static ref2String curErrMess;
