/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/* =============================================================================
 * standard basic functions
 */

/* */
static boolean isEof()
{ return ( pKEof() );
}

/* */
static boolean isEol()
{ return ( pKEoln() );
}

/*
 * set current character at the beginning of next line
 * success: returns true
 *          get a new line from file, assumed to be terminated by Eoln or pKEof
 *          increments line counter
 * failure: returns false
 *          
 */
static boolean fileNextLine()
{
  if ( lineLength == -2 ) {
    logPrintErr(true, (char*)" Damned: no current file to read");
    return false; }
  if ( pKEof() ) return false;
  lineLength = getpKLine (fileLine, lineSize, currentFile);
  //  fprintf(stderr, "==========fileNextLine lineLength=%i, fileLine=%s\n", lineLength, fileLine);
  if ( lineLength<=0 ) return false;
  lineCounter = lineCounter + 1;
  cc = 0;
  return true;
}

/*
 * set current character to its next position in  the current line
 * success: returns true
 * failure: returns false if either pKEoln() or pKEof() is reached
 *          
 */
static boolean lineNextChar()
{
  if (pKEof() || pKEoln()) return false;
  cc = cc + 1;
  return true; 
}

/*
 * set current character to the last position in  the current line
 * success: returns true
 * failure: returns false if pKEof() is reached
 *          
 */
static boolean skipTopKEoln()
{
  if (pKEof()) return false;
  cc = lineLength-1;
  return true; 
}

/*
 * set current character to its next position, ignoring the end of line
 * success: returns true
 * failure: returns false
 *          
 */
static boolean fileNextChar()
{
  if ( pKEof() ) return false;
  if ( pKEoln() ) return  fileNextLine();
  cc = cc + 1;
  return true; 
}

/*
 * set current character immediately after the first one equal to *t in the current line
 *             current character is examined
 *             does not change current line !!
 * success: returns true
 * failure: returns false  if either pKEoln() or pKEof() is reached
 */
static boolean lineNextCharAfter(char t)
{
  do {
    if ( (currentChr()== t) ) return (lineNextChar());}
  while ( lineNextChar() );
  return false;
}

/*
 * set current character immediately after the first one equal to *t in the current line
 *             current character is examined
 *             does not change current line !!
 * success: returns true
 * failure: returns false  if either pKEoln() or pKEof() is reached
 */
static boolean lineNextCharAfterClosing()
{
  return (lineNextCharAfter(closeTag));
}

/*
 * set current character immediately after the first one equal to t in the current file
 *             current character is examined
 * success: returns true
 * failure: returns false
 */
static boolean fileNextCharAfter(char t)
{
  do {
    if ( (currentChr()== t) ) return (fileNextChar());}
  while ( fileNextChar() );
  return false;
}

/*
 * set current character immediately after the last tab or ' ' character in the current line
 *             current character is examined
 *             does not change current line !!
 * success: returns true
 * failure: returns false if either pKEoln() or pKEof() is reached
 */
static boolean lineNextNonSpace()
{
  while ( !(pKEoln() || pKEof()) ) {
    if ( (currentChr() != ' ') && (currentChr() != ctab) ) return true; 
    if (lineNextChar()); }
  return false;
}

/*
 * set current character immediately after the last tab or ' ' character in the current file
 *             current character is examined
 * success: returns true
 * failure: returns false
 */
static boolean fileNextNonSpace()
{
  if ( pKEoln() ) 
    if ( !(fileNextLine()) ) return false;
  do {
    if ( (currentChr() != ' ') && (currentChr() != ctab)  && (!(pKEoln()))) return true;}
  while ( fileNextChar() );
  return false;
}

/*
 * set current character immediately after the last tab or ' ' character in the current file
 *             ends of line are ignored
 *             current character is not examined
 * success: returns true
 *          set t value to this new current character
 * failure: returns false
 */
static boolean fileNextNonSpaceAfter()
{
  if ( !(fileNextChar()) ) return false;
  if ( !(fileNextNonSpace()) ) return false;
  return true; 
}

/* =============================================================================
 * specifc basic functions
 */

/*
 * source begins with a variable character 
 * returns the length of the longest identifier prefix in source
 */
static size_t getVariableLength(char *source)
{
  return strcspn(source, " /!\n");
}

/* =============================================================================
 * syntactic get functions
 */

/*
 *  Reserved MKF line  
 */
static boolean isMKFTag()
{
  return ( (fileLine[0] == openTag) || (fileLine[0] == closeTag) ); 
}

/*
 */
static boolean isCommentTag()
{ 
  return ( currentChr() == commentMKFTag );
}

static boolean isIncludeMark()
{
  return ( currentChr() == includeTag ); 
}

static boolean isLoadMark()
{
  return ( currentChr() == handMadeTag ); 
}

static boolean isDirectMark()
{
  return ( currentChr() == writeFileTag ); 
}

static boolean isForcedMark()
{
  return ( currentChr() == forceTag ); 
}

static boolean isDonotMark()
{
  return ( currentChr() == donotTag ); 
}

/*
 *          
 */
static boolean isOpenPrelude()
{ 
  return ( (Boln()) && (fileLine[0] == openTag) );
}

/*
 *          
 */
static boolean stepOpenTag()
{ 
  if ( (Boln()) && (fileLine[0] == openTag) )
    { cc = 1; return true; };
  return false;
}

/*
 *          
 */
static boolean isCloseTag()
{ 
  return ( (Boln()) && (fileLine[0] == closeTag) && lineLength <= 2 ); 
}

/*
 * tests if current character is the first character of an atom denoting a reference to object
 * success set currentcharcter immediately after this atom
 */
static boolean lineNextPredecessorTag()
{ 
  if ( (Boln()) && (fileLine[0] == closeTag) && (fileLine[1] == includeTag) ) {
    cc = 2;
    return true; }
  return false;
}

/*
 * tests if current character is the first character of an atom denoting a reference to object
 * success set currentcharcter immediately after this atom
 */
static boolean isOpenDefault()
{ 
  return ( (Boln()) && (fileLine[0] == closeTag) && (fileLine[1] == defaultTag) );
}

/*
 * skip after the next opening tag from current position
 * success set currentcharcter immediately after this atom
 */
static boolean skipafterNewDefTag()
{
  while ( !(pKEof()) ) {
    if (stepOpenTag() )
      return true;
    fileNextLine(); }  
  return false;
}

/*
 * set current character to the first character of the line following object closing
 *             current character is ignored
 * success: returns true
 * failure: returns false on pKEof
 */
static boolean skip2SafeState()
{
  while ( fileNextCharAfter(closeTag) )
    if ( (pKEoln()) )
      return ( fileNextChar() );
  return false;
}

static boolean isupperletter (char c)
{
  return ( (((unsigned) currentChr() >= (unsigned) ('A'))
	   && ((unsigned) currentChr() <= (unsigned) ('Z'))));
}

static boolean islowerletter (char c)
{
  return ( (((unsigned) currentChr() >= (unsigned) ('a'))
	   && ((unsigned) currentChr() <= (unsigned) ('z'))) );
}

static boolean isDecimalDigit (char c)
{
  return ( (((unsigned) currentChr() >= (unsigned) ('0'))
	   && ((unsigned) currentChr() <= (unsigned) ('9'))) );
}

/*
 * set current character to the first character of the line following object closing
 *             current character is ignored
 * success: returns true
 * failure: returns false on pKEof
 */
static boolean syntaxError(char *mess)
{
  logPrintSyntax(lineCounter, cc, mess);
  return (skip2SafeState());
}

/*
 * set current character to the first character of the line following object closing
 *             current character is ignored
 * success: returns true
 * failure: returns false on pKEof
 */
static boolean conflictError(char *mess, ref2ObjectType newob)
{
  logPrintConflict(lineCounter, cc, true, newob, mess);
  return (skip2SafeState()); 
}

/*
 * set current character to the first character of the line following object closing
 *             current character is ignored
 * success: returns true
 * failure: returns false on pKEof
 */
static void conflictWarning(char *mess, ref2ObjectType newob)
{
  logPrintConflict(lineCounter, cc, false, newob, mess);
}

/*
 * reads a MKF variable name beginning at current character
 *             the first character that doesn't belong to name becomes current character
 *             does not change current line !!
 * success: returns name length
 * failure: returns false  if either pKEoln() or pKEof() is reached
 */
static int readVarName(char *atomBuffer)
{
  int length = 0;
  while ( (currentChr() == '_') || (currentChr() == '.') || (isalnum(currentChr()) != 0) ) {
    atomBuffer[length] = currentChr();
    length = length + 1;
    if (!lineNextChar()) {
      atomBuffer[length]='\0';
      return length; } }
  atomBuffer[length]='\0';
  if (length == 0) {
    if (syntaxError( (char*) "empty string not allowed for an identifier"));}
  return length;
}

/*
 * try to read a pair (variable name, file name) the file name has to be searched in directories
 * which file names are in variable name
 * success: returns true
 * failure: returns false
 */
static int readVarFileName(char finalTag)
{
  int length = 0;
  while ( lineNextChar() ) {
    if ( currentChr() == finalTag) {
      if ( lineNextChar () ) {
	if ( currentChr() != finalTag) {
	  atomBuffer[length]='\0';
	  return length; } }
      else {
	atomBuffer[length]='\0';
	return length; } }
    atomBuffer[length] = currentChr();
    length = length + 1; }
  atomBuffer[0]='\0';
  if (syntaxError( (char*) "invalid character in a file identifier"));
  return 0 ;
}

/*
 * try to read a file name or a file var name or a variable name
 *          after skipping spaces in the current line,
 *          current character is examined
 * success: returns true
 *          set atomLength to object name length
 *          set isvar to true if object is a variable, 
 *          set isvar to false if object is a file
 *                    set isFile to true if this path identifies a single file !name!
 *                    set isFile to false otherwise <name>
 *          set current character immediately after the atom
 * failure: returns false
 *          set atomLength to 0
 *          set isvar_ispKeof to true if pKeof has been reached 
 */
static boolean readObjectName(boolean *isvar, boolean *isFile)
{
  atomLength = 0;
  if ( !(lineNextNonSpace()))
    return false; 
  if ( pKEoln() ) return false; /* the end of current line has been reached */
  if ( currentChr() == startFileTag) {
    *isvar = false;
    *isFile = true;
    atomLength = readVarFileName(endFileTag); }
  else if ( currentChr() == startEnvTag) {
    *isvar = false;
    *isFile = false;
    atomLength = readVarFileName(endEnvTag); }
  else { 
    *isvar = true;
    *isFile = false;
    atomLength = readVarName(atomBuffer); }
  return ( atomLength > 0 );
}

static boolean lineNextAtom (void)
{
  /* not implemented */
  return true; 
}


static boolean isstrnprefix(char *ref, char *source, int length)
{
  if (strlen(ref) != length) return false;
  return ( strncmp(ref, source, length) == 0 );
}

static char *appendLine ()
{
  char *line;
  if ( pKEof() ) return (char *)0;
  //line = (char *) malloc ( (unsigned)(strlen(fileLine)) );
  line = strdup(fileLine);
  if ( fileNextLine() );
  return line;
}

static int getAtomHash(void)
{
  int i, hashi = atomLength;
  for (i = 0; i < atomLength; i=i+4) 
    hashi = hashi + (unsigned) (atomBuffer[i]);
  hashi = (hashi % hashsize) + 1;
  return hashi;
}

static char *strdupAtom()
{
  return(strdup(atomBuffer));
}

static boolean isAtomEq(char *name)
{
  return ( strcmp(name, atomBuffer) == 0 );
}

static boolean openStdin()
{
  currentFile = stdin;
  lineLength = 0;
  lineCounter = 0;
  return (fileNextLine());
}

static boolean closeStdin()
{
  lineLength = -2;
  lineCounter = 0;
  cc = -1;
}

static boolean openReadMKF(char *fileName)
{
  if (isFileOpen()) {
    logPrintErr(true,  (char *) "Damned: cannot open twice without closing");
    return false; }
  if (fileName == NULL) {
    logPrintErr(true, (char *) "Damned: cannot open NULL file name");
    return false; }
  currentFile = rFOpen (fileName, curErrMess);
  if ( currentFile == NIL_File) {
    logPrintErrStringMess(fileName, *curErrMess);
    return false; }
  lineLength = 0;
  lineCounter = 0;
  if ( fileNextLine() ) return true;
  if ( pKEof() ) {
    logPrintWar(true, (char *) "Empty file ");
    logPrintString(fileName); } 
  return false;
}

static boolean closeMKF()
{
  rwFClose(currentFile);
  lineLength = -2;
  lineCounter = 0;
  cc = -1;
}

static void initGlobalReader()
{
  fileLine = (char *) calloc (lineSize+1,1);
  atomBuffer = (char *) calloc (lineSize+1,1);
  lineLength = -2;
}

