/*
 * main parsing routines for the program pKPrototypeExtract
 * It is derived from Aam Bryant parse (see below)
 * Copyright (c) 2010, 2011 by Loc Besnard, Paul Le Guernic for modifications
 *
 */

#define is_idHeader(x)	(isalpha(x) || (x == '_'))
#define is_idChar(x)	(isalnum(x) || (x == '_'))

/* commands to C processor and Switch on invocation line */
#ifndef DEF_LEADER
#define DEF_LEADER	"-D"
#endif /* DEF_LEADER */

typedef enum
{ V_public, V_protected, V_private, V_none } accesMode;	/* V_private is the default value */
accesMode printed_accesMode = V_none;	/* this is the curent acces mode in the file */
accesMode new_accesMode = V_none;	/* is the last read access mode */

static boolean was_ident = pk_false;
static boolean was_comment = pk_false;
#define MAX_SIZE 16384

static char code_info[MAX_SIZE + 1];
static int codeInfo_index;	/* index of the last pk_true character in code_info (-1 when code_info is empty) */
static boolean is_spaced = pk_true;	/* spaced is set to pk_true when 
					 * one of the following space characters is registered: space (' '), form feed ('\f'), new-line('\n'), 
					 *       carriage return ('\r'), horizontal tab ('\t'), and vertical tab ('\v').
					 * it is set to false when isalnum() becomes pk_true
					 */
static char virtualFile[MAX_SIZE + 1];
static int virtualFile_index = -1;
#define FREE_CODE_INFO (codeInfo_index == -1)

static boolean
is_accesMode_label (char *ident)
{
  if ((strcmp (ident, "public:")) == 0)
    {
      new_accesMode = V_public;
      return pk_true;
    }
  if ((strcmp (ident, "protected:")) == 0)
    {
      new_accesMode = V_protected;
      return pk_true;
    }
  if ((strcmp (ident, "private:")) == 0)
    {
      new_accesMode = V_private;
      return pk_true;
    }
  return pk_false;
}

static boolean
is_accesMode_macro (char *ident)
{
  if ((strcmp (ident, "_PUBLIC_")) == 0)
    {
      new_accesMode = V_public;
      return pk_true;
    }
  if ((strcmp (ident, "_PROTECTED_")) == 0)
    {
      new_accesMode = V_protected;
      return pk_true;
    }
  if ((strcmp (ident, "_PRIVATE_")) == 0)
    {
      new_accesMode = V_private;
      return pk_true;
    }
  return pk_false;
}


/* storage elements */

typedef struct s_index
{
  int index;			/* the name of the variable */
  struct s_index *pred, *next;	/* pointer to the next element */
} S_INDEX, *P_INDEX;
P_INDEX first_index, free_index;

static int num_func = 0;
/* ========================= comments */
static char first_comment[MAX_SIZE];
static char last_comment[MAX_SIZE];
static boolean free_line = pk_true;
static int comment_len = 0;
static int num_comment = -1;
static boolean fileComment_printable = pk_true;
static boolean fileComment_printed = pk_false;

static void
empty_last_comment ()
{
  last_comment[0] = '\0';
  comment_len = 0;
}

/* PLG WARNING CHECK FOR MULTIBYTE CHARACTERS !!! specially escape sequences ('\' with "'" '"' '?' '\') */


/* ========================= DEBUG FUNCTIONS ========================= */
#define append_char2code_info(CHAR) code_info[++codeInfo_index] = CHAR
#define completeString_in_code_info() code_info[codeInfo_index+1] = '\0'

static void
reset_codeInfo ()
{
  code_info[0] = '\0';
  codeInfo_index = -1;
  is_spaced = pk_true;
}

static void
confirm_space ()
{
  if (codeInfo_index >= 0)
    if ((code_info[codeInfo_index] != ' ') && (code_info[codeInfo_index] != '\0'))
      /* combine any white space into a single space character */
      append_char2code_info (' ');
  is_spaced = pk_true;
}

static void
append_NOTspace (char ch)
{
  append_char2code_info (ch);
  is_spaced = pk_false;
}

static void
append_char (char ch)
{
  if (isspace (ch))
    confirm_space ();
  else
    append_NOTspace (ch);
}

/* =========================================================================================================
 * fpin Parsing
 * =========================================================================================================
 */

/* read and store all characters until end comment indicator included 
 * when entering, last read character is "*" following '/' 
 * when leaving, last read character is '/' following '*'
 */
static void
keepComment ()
{
  if (was_comment)
    {
      if (warnComment)
	warningCommentLost ();
    }
  else
    was_comment = pk_true;
  if (num_comment < 0)
    comment_len = get_comment (first_comment);
  else
    {
      empty_last_comment ();
      comment_len = get_comment (last_comment);
    }
  num_comment++;
}

/** 
 * FPIN = [Character*]['/']~[Character*][EOF]
 * success: FPIN = [Character*]['/'] # ['*'][Character*]['*''/'] ~ [Character*][EOF]
 *          FPIN = [Character*]['/'] # ['/'][Character_but_eol*][eol] ~ [Character*][EOF]
 * failure: FPIN = [Character*]['/'] #~ [Character*][EOF]
 */
static boolean
parseIf_comment_completion ()
{
  if (getc_if_logicalNext_char_is ('*'))
    {
      if (free_line && FREE_CODE_INFO)
	{
	  if (commentMode_option == ANY_COMMENTS)
	    {
	      keepComment ();
	      return pk_true;
	    }
	  if (commentMode_option == SELECTED_COMMENTS)
	    {
	      if (logicalNext_string_is (commentPrefix))
		{
		  keepComment ();
		  return pk_true;
		}
	    }
	}
      /* not a kept comment */
      skip_C_fullComment_completion ();	/* exit on error; otherwise, '/' is the last read character in fpin */
      free_line = pk_false;
      confirm_space ();
      return pk_true;
    }

  if (getc_if_logicalNext_char_is ('/'))
    {
      /* just yank out any C++ comments */
      skip_C_lineComment_completion ();	/* EOF is reached or [eol] is the last read character */
      free_line = pk_true;
      confirm_space ();
      return pk_true;
    }
  append_NOTspace ('/');
  free_line = pk_false;
  return pk_false;
}

/* ACTUAL PRINTING */

static void
put_CPlusPlus_accesMode ()
{
  switch (printed_accesMode)
    {
    case V_public:
      put_publicAccess ();
      break;
    case V_protected:
      put_protectedAccess ();
      break;
    case V_private:
      put_privateAccess ();
      break;
    default:
      break;
    }
  printed_accesMode = new_accesMode;
}

static void
actualPrint_fileComment ()
{
  if (commentMode_option != NO_COMMENTS)
    {
      if (num_comment < 0)
	{			/* file_comment not printed: num_func <= 0 */
	  if (forceComment)
	    put_forced_comment_eol ("AND");
	}
      else if (num_comment == 0)
	{			/* file_comment not printed: num_func <= 0 */
	  if (forceComment)
	    put_forced_comment_eol ("OR");
	}
      else
	{			/* num_comment >= 1 */
	  put_comment_eol (first_comment);
	  put_eol ();
	}
    }
  fileComment_printed = pk_true;
}

static void
actualPrint_headerFileComment ()
{
  if (commentMode_option != NO_COMMENTS)
    {
      if (num_comment < 0)
	{			/* file_comment not printed: num_func <= 0 */
	  if (forceComment)
	    {
	      /* removed LBesnard . put_headingCommentMissing(); put_eol();  */
	    }
	}
      else
	{			/* num_comment >= 0 */
	  put_comment_eol (first_comment);
	  put_eol ();
	  if (num_comment == 0)
	    num_comment = -1;
	}
    }
  fileComment_printed = pk_true;
}

static boolean
actualPrint_protoComment ()
{
  if (commentMode_option != NO_COMMENTS)
    {
      if (num_comment == 0)
	put_comment_eol (first_comment);
      else
	if (num_comment > num_func)
	  {
	    num_comment = num_func;
	    put_comment_eol (last_comment);
	  }
	else 
	  if (forceComment)
	    {
	      if (exportOnlyCommentedFunctions) return pk_true;
	      put_forced_functionComment ();
	    }
    }
  return pk_false;
}

static void
actualPrint_CPlusPlusAccessMode ()
{
  if (printed_accesMode == V_none)
    {
      if (new_accesMode == V_none)
	{
	  printed_accesMode = V_private;
	  new_accesMode = V_private;
	}
      else
	printed_accesMode = new_accesMode;
      put_CPlusPlus_accesMode ();
    }
  else if (printed_accesMode != new_accesMode)
    {
      printed_accesMode = new_accesMode;
      put_CPlusPlus_accesMode ();
    }
}

/**
   @return "true" if the function/method is without any comment.
   (the comment \todo{ MISSING FUNCTION API DOCUMENTATION } is added),
   "false" otherwise.
 */
static boolean
actualPrint_proto (boolean CPlusPlus)
{
  int iii = 0;
  boolean bNoComment;

  if (fileComment_printable) if (!fileComment_printed) actualPrint_fileComment ();
  if (CPlusPlus) actualPrint_CPlusPlusAccessMode ();
  bNoComment = actualPrint_protoComment ();
  if (!bNoComment)
    {
      while (iii <= virtualFile_index)
	{
	  if (virtualFile[iii] == '\0')
	    {
	      iii++;
	      put_space ();
	    }
	  else if (is_idHeader (virtualFile[iii]))
	    {
	      put_string ((char *) (virtualFile + iii));
	      iii = iii + strlen ((char *) (virtualFile + iii)) + 1;
	      if (iii <= virtualFile_index)
		{
		  if ((virtualFile[iii] == '*') || (virtualFile[iii] == '('))
		    {
		      put_char (virtualFile[iii]);
		      iii++;
		    }
		  else
		    put_space ();
		}
	    }
	  else
	    {
	      put_string ((char *) (virtualFile + iii));
	      iii = iii + strlen ((char *) (virtualFile + iii)) + 1;
	    }
	}
      put_abstractProtoEnd ();
    }
  return bNoComment;
}

/* VIRTUAL PRINTING */

static boolean requiresSpace = pk_false;

static void
clean_virtualFile ()
{
  virtualFile_index = -1;
}

static void
vput_char (char cc)
{
  if (virtualFile_index < MAX_SIZE - 1)
    {
      virtualFile_index++;
      virtualFile[virtualFile_index] = cc;
    }
  else
    exitERR ("OUPS: virtual file buffer overloaded");
}

#define VPUT_SEPARATOR() vput_char('\0')
#define IS_SEPARATOR() ( virtualFile[virtualFile_index] != '\0' )

static void
vput_uniqueSeparator ()
{
  if (virtualFile_index >= 0)
    if (IS_SEPARATOR ())
      VPUT_SEPARATOR ();
}

static void
vput_string (const char *str)
{
  int lngth = strlen (str), iii;
  for (iii = 0; iii < lngth; iii++)
    vput_char (str[iii]);
}

/**
 * append str to the virtual file, preceded and followed by a unique separator
 */
static void
vput_ident (char *str)
{
  vput_uniqueSeparator ();
  vput_string (str);
  VPUT_SEPARATOR ();
}

/** append str[0..(max_size-1)] to the virtual file, preceded and followed by a unique separator
 */
static void
vput_stringsInArray (const char str[], int first_index, int last_index)
{
  int jjj;
  vput_uniqueSeparator ();
  for (jjj = first_index; jjj <= last_index; jjj++)
    if ((str[jjj] == '\0') || (str[jjj] == '\n') || isspace (str[jjj]))
      vput_uniqueSeparator ();
    else
      vput_char (str[jjj]);
  vput_uniqueSeparator ();
}

/** append commented str followed by suff, preceded and followed by a unique separator
 */
static void
vput_commentStr (const char *str, const char *suff)
{
  vput_uniqueSeparator ();
  vput_char ('/');
  vput_char ('*');
  vput_string (str);
  vput_string (suff);
  vput_char ('*');
  vput_char ('/');
  VPUT_SEPARATOR ();
}

/**
 * append commented header_string,  preceded and followed by a unique separator
 */
static void
vput_headerString ()
{
  if (header_string != NULL)
    {
      vput_uniqueSeparator ();
      vput_char ('/');
      vput_char ('*');
      vput_string (header_string);
      vput_char ('*');
      vput_char ('/');
      VPUT_SEPARATOR ();
    }
}

static void
vput_CDeclarationPrefix (boolean isExtern, int start_index, int postLastIndex)
{
  if (isExtern)
    vput_ident ("EXTERN");	/* preceded and followed by a unique separator */
  else
    vput_ident ("static");	/* preceded and followed by a unique separator */
  if (postLastIndex <= start_index)
    vput_ident ("void");	/* preceded and followed by a unique separator */
  else
    vput_stringsInArray (code_info, start_index, postLastIndex - 1);	/* preceded and followed by a unique separator */
}

static void
vputCplusPlus_declarationPrefix (int postLastIndex)
{
  int jjj = 0;
  char *curr_word;
  int curr_length;
  while (jjj < postLastIndex)
    {
      curr_word = (char *) (code_info + jjj);
      curr_length = strlen (curr_word);
      if (curr_length > 0)
	{
	  if (strcmp (curr_word, "_VIRTUAL_") == 0)
	    vput_ident ("virtual");	/* preceded and followed by a unique separator */
	  else if (strcmp (curr_word, "_STATIC_") == 0)
	    vput_ident ("static");	/* preceded and followed by a unique separator */
	  else
	    vput_ident (curr_word);
	}			/* preceded and followed by a unique separator */
      jjj = jjj + strlen (curr_word) + 1;
    }
}

static void
reset_comments ()
{
  if (fileComment_printable)
    {
      fileComment_printable = pk_false;
      if (!fileComment_printed)
	actualPrint_headerFileComment ();
    }
  if (num_func > 0)
    num_comment = num_func;
  else
    num_comment = -1;
}

static void
reset_fileUnit ()
{
  reset_codeInfo ();
  reset_comments ();
}

/** in:  FPIN = PREFIX <curr_char>~ [Character*][EOF]
 *       <codeInfo_index> is an integer that satisfies -1 <= <codeInfo_index> < sizeof(code_info)
 *       <code_info> is large enough to contain an identifier after <codeInfo_index> position
 *  out: FPIN = PREFIX <curr_char># IDENT ~ [Character*][EOF] 
 *             the first character in [Character*][EOF] is not an identifier character
 *       <curr_char>IDENT is stored in <code_info>[(codeInfo_index+1)..(codeInfo_index+strlen(IDENT)+1)] if it is not an access mode
 *       <code_info>[<codeInfo_index>] == '\0'
 * returns the index in code_info of the first character of the new identifier
 *         this index is greater than codeInfo_index when the identifier is not register in code_info
 *
 */
static int
parse_identifierCompletion (char curr_char)
{
  int save_codeInfo_index;
  if (codeInfo_index >= 0)
    if (code_info[codeInfo_index] != '\0')
      {
	codeInfo_index = codeInfo_index + 1;
	code_info[codeInfo_index] = '\0';
      }
  save_codeInfo_index = codeInfo_index;
  codeInfo_index = get_identifier_completion (curr_char, code_info, save_codeInfo_index);	/* FPIN = PREFIX <curr_char> # IDENT ~ [Character*][EOF]
												 * code_info[(save_codeInfo_index+1)..codeInfo_index] = IDENT'\0' */
  if (getc_if_logicalNext_char_is (':'))
    {
      code_info[codeInfo_index] = ':';
      code_info[codeInfo_index + 1] = '\0';
      if (is_accesMode_label ((char *) (code_info + save_codeInfo_index + 1)))
	{
	  reset_comments ();
	  codeInfo_index = save_codeInfo_index;
	}
      else
	{
	  code_info[codeInfo_index] = '\0';
	  char_backward (':');
	}
    }
  return save_codeInfo_index + 1;
}

/** in:  FPIN = PREFIX '_'~ [Character*][EOF]
 *       <codeInfo_index> is an integer that satisfies -1 <= <codeInfo_index> < sizeof(code_info)
 *       <code_info> is large enough to contain an identifier after <codeInfo_index> position
 *  out: FPIN = PREFIX '_'# IDENT ~ [Character*][EOF] 
 *             the first character in [Character*][EOF] is not an identifier character
 *       '_'IDENT is stored in <code_info>[(codeInfo_index+1)..(codeInfo_index+strlen(IDENT)+1)] if it is not an access mode
 *       <code_info>[<codeInfo_index>] == '\0'
 * returns the index in code_info of the first character of the new identifier
 *         this index is greater than codeInfo_index when the identifier is not register in code_info
 *
 */
static int
parse_subIdentifierCompletion ()
{
  int save_codeInfo_index;
  if (codeInfo_index >= 0)
    if (code_info[codeInfo_index] != '\0')
      {
	codeInfo_index = codeInfo_index + 1;
	code_info[codeInfo_index] = '\0';
      }
  save_codeInfo_index = codeInfo_index;
  codeInfo_index = get_identifier_completion ('_', code_info, save_codeInfo_index);	/* FPIN = PREFIX '_' # IDENT ~ [Character*][EOF]
											 * code_info[(save_codeInfo_index+1)..codeInfo_index] = IDENT'\0' */
  if (is_accesMode_macro ((char *) (code_info + save_codeInfo_index + 1)))
    {
      reset_comments ();
      codeInfo_index = save_codeInfo_index;
    }
  return save_codeInfo_index + 1;
}

/** in:  FPIN = PREFIX ~ [Character*][EOF]
 *       <codeInfo_index> is an integer that satisfies -1 <= <codeInfo_index> < sizeof(code_info)
 *       <code_info> is large enough to contain an identifier after <codeInfo_index> position
 *  out: FPIN = PREFIX # IDENT ~ [Character*][EOF] 
 *             the first character in [Character*][EOF] is not an identifier character
 *       IDENT is stored in <code_info>[(codeInfo_index+1)..(codeInfo_index+strlen(IDENT))] if it is not an access mode
 * return: pk_true an IDENT has been read
 *              this ident is not an acces mode: it is append to <code_info>
 *                   <start_index> is the index in code_info of the first character of the new identifier
 *              this ident is an acces mode: <code_info> remains unchanged
 *                   <start_index> is the index in code_info of the first character following codeInfo_index
 *         false next char is not an identifier character
 *         <code_info>[<codeInfo_index>] == '\0'
 */
static boolean
parse_identifier (int *start_index)
{
  char curr_char;
  if ((curr_char = get_logicalChar ()) == EOF)
    return pk_false;
  if (isalpha (curr_char))
    {
      (*start_index) = parse_identifierCompletion (curr_char);
      return pk_true;
    }
  if ((curr_char == '_'))
    {
      (*start_index) = parse_subIdentifierCompletion ();
      return pk_true;
    }
  char_backward (curr_char);
  return pk_false;
}

static void
parse_CplusPlusAttributes ()
{
  // const {, =0;, : xxxx {
  char curr_char;
  boolean first_char = pk_true;
  while ((curr_char = get_logicalChar ()) != EOF)
    {
      if (curr_char == ';')
	{
	  char_backward (curr_char);
	  return;
	}
      if (curr_char == '{')
	{
	  char_backward (curr_char);
	  return;
	}
      if (curr_char == ':' && first_char)
	return;			// C1( x) : C(x) {  }  init upper class 
      if (curr_char == '\n')
	{
	  if (skip_spacesAndDirectives_inNewLogiLine (pk_true));
	  else if (isspace (curr_char))
	    {
	      if (skip_logicalSpaces_but_eol ());
	    }
	}
      else
	{
	  vput_char (curr_char);
	  if (isalpha (curr_char))
	    first_char = pk_false;
	}
    }
}

/* INPUT: FPIN = PREFIX '(' ~ [Character*]EOF
 * OUTPUT: FPIN = PREFIX '(' # IDENT [PARAMETER]* ')' ~ [Character*]EOF
 */
static void
parse_parameterList ()
{
  char curr_char;
  int depth = 1;
  int identifier_index;
  boolean emptyParameters = pk_true;

  vput_char ('(');
  requiresSpace = pk_false;
  while ((curr_char = get_logicalChar ()) != EOF)
    {
      if (curr_char == '\n')
	register_logicalEol ();
      else if (isspace (curr_char))
	skip_logicalSpaces_but_eol ();
      else if (curr_char == '/')
	{
	  if (skipIf_comment_completion ())	/* fail: FPIN = [Character*]['/'] #~ [Character*][EOF] ; exit on error;  */
	    vput_uniqueSeparator ();
	  else
	    {
	      vput_char ('/');
	      requiresSpace = pk_false;
	    }
	}
      else if (curr_char == '"')
	{
	  /* --------    --- string mark should not be present here */
	  /* nuke string quotes -- must be on same line */
	  skip_C_string_completion ();	/* exit on failure, '"' is the last read character  */
	  vput_uniqueSeparator ();
	}			/* end reading string */
      else if (curr_char == '\'')
	{
	  /* --------    --- character mark should not be present here */
	  /* nuke character quotes -- must be on same line */
	  skip_C_characterConstant_completion ();	/* exit on error ''' is the last read character  */
	  vput_uniqueSeparator ();
	}			/* exit on error */
      else if (curr_char == ')')
	{
	  /* --------    --- block closing */
	  --depth;
	  if (depth == 0)
	    {
	      if (emptyParameters) vput_ident ("void");	/* preceded and followed by a unique separator */
	      vput_char (')');
	      parse_CplusPlusAttributes ();	// Added LBesnard
	      VPUT_SEPARATOR ();
	      return;
	    }
	  vput_char (')');
	  requiresSpace = pk_false;
	}
      else if (curr_char == '(')
	{
	  vput_char ('(');
	  depth++;
	  requiresSpace = pk_false;
	}
      else if (is_idHeader (curr_char))
	{
	  vput_uniqueSeparator ();
	  emptyParameters = pk_false;
	  identifier_index = codeInfo_index + 1;
	  codeInfo_index = get_identifier_completion (curr_char, code_info, codeInfo_index);	/* * returns <res_index>+strlen(<curr_char>IDENT): position is that of '\0' */
	  vput_ident ((char *) (code_info + identifier_index));	/* preceded and followed by a unique separator */
	  requiresSpace = pk_true;
	}
      else if ((curr_char == ',') || (curr_char == '*'))
	{
	  requiresSpace = pk_true;
	  vput_char (curr_char);
	  vput_uniqueSeparator ();
	}
      else
	{
	  vput_char (curr_char);
	  requiresSpace = pk_false;
	}
    }
  vput_char (')');
}

/** test if this function prototype is to be extracted wrt to staticMode_option
 * set <isExtern> to pk_true (the function is extern) or to false (the function is local)
 * set <new_index> to the first position that follows the scope indication
 */
static boolean
test_C_accesMode (int postLastIndex, boolean * isExtern, int *new_index)
{
  if (postLastIndex <= 0)
    {
      (*isExtern) = pk_true;
      (*new_index) = 0;
    }
  else if ((strcmp (code_info, "static")) == 0)
    {
      (*isExtern) = pk_false;
      (*new_index) = 7;
    }
  else if ((strcmp (code_info, "_STATIC_")) == 0)
    {
      (*isExtern) = pk_false;
      (*new_index) = 9;
    }
  else if ((strcmp (code_info, "extern")) == 0)
    {
      (*isExtern) = pk_true;
      (*new_index) = 7;
    }
  else if ((strcmp (code_info, "EXTERN")) == 0)
    {
      (*isExtern) = pk_true;
      (*new_index) = 7;
    }
  else
    {
      (*isExtern) = pk_true;
      (*new_index) = 0;
    }
  if (staticMode_option == ONLY_STATICS)
    return !(*isExtern);
  if (staticMode_option == NO_STATICS)
    return (*isExtern);
  return pk_true;
}

/*
 */
static void
complete_prototypeExtraction (boolean CPlusPlus, int functionIdentifier_index)
{
  boolean bNoComment;

  /* ===== function name */
  vput_ident ((char *) (code_info + functionIdentifier_index));	/* preceded and followed by a unique separator */
  /* ===== parameters */
  if (skipIf_comments_and_spaces (pk_false));	/* FPIN = PREFIX # [SPACE]* ~ [Character*] [EOF] */
  parse_parameterList ();	/*  FPIN = PREFIX '(' # IDENT [PARAMETER]* ')' ~ [Character*]EOF */
  if (reach_blockStart (pk_false))
    {
      // fprintf (stderr, "LBesnard--WARNING: declaration of <<%s>>\n", (char *) (code_info + functionIdentifier_index));
      bNoComment = actualPrint_proto (CPlusPlus);
      if ( ! bNoComment || ! exportOnlyCommentedFunctions) num_func++;
      clean_virtualFile ();
      skip_braceInner ();
    }
  else 
    if (CPlusPlus) // added lbesnard.
      {				/* PLG TEST OF ABSTRACT DECLARATION ??? */
	/* FPIN = PREFIX '(' IDENT [PARAMETER]* ')' [Character_[';'+'{']]* ';' ~ [Character*]EOF  or 
	 * FPIN = PREFIX '(' IDENT [PARAMETER]* ')' [Character_[';'+'{']]* EOF ~
	 */
	// fprintf (stderr, "LBesnard--WARNING: ABSTRACT declaration of <<%s>>\n", (char *) (code_info + functionIdentifier_index));
	bNoComment = actualPrint_proto (CPlusPlus);
	if ( !bNoComment || !exportOnlyCommentedFunctions) num_func++;
	clean_virtualFile ();
      }
    else
      {
	clean_virtualFile ();
	skipTo_endOfUnit (pk_false);
      }
}

/* INPUT: FPIN = PREFIX '(' ~ [Character*]EOF
 * OUTPUT: FPIN = PREFIX '(' # IDENT [PARAMETER]* ')' ~ [Character*]EOF
 */
static void
parse_CDeclaration (int identifier_index)
{
  int functionIdentifier_index = -1;
  int type_index;
  boolean isExtern;

  /* IS EXTRACTION REQUIRED ? */

  if (strcmp ((char *) (code_info + identifier_index), "main") == 0)
    {
      skip_parenthesisInner ();
    }
  else
    {
      if (test_C_accesMode (identifier_index, &isExtern, &type_index))
	{			/* the function prototype is to be extracted */
	  /* ===== comment */
	  /* ===== declaration heading  */
	  functionIdentifier_index = identifier_index;
	  vput_CDeclarationPrefix (isExtern, type_index, identifier_index);
	  vput_headerString ();	/* preceded and followed by a unique separator */
	  complete_prototypeExtraction (pk_false, functionIdentifier_index);
	  return;
	}
      else			/* don't extract the prototype */
	skip_parenthesisInner ();
    }
  if (reach_blockStart (pk_false))
    {
      // build_proto_method();
      skip_braceInner ();
    }
  else				/* PLG TEST OF ABSTRACT DECLARATION ??? */
    skipTo_endOfUnit (pk_false);
}


/* 
 * codeInfo_index >= 1 and code_info[codeInfo_index] = '\0
 * INPUT: FPIN = PREFIX ':' ':' ~ [Character*]EOF
 * OUTPUT: FPIN = PREFIX ':' ':' # IDENT '(' [PARAMETER]* ')' ~ [Character*]EOF
 * OUTPUT: FPIN = PREFIX ':' ':' # ~ IDENT '(' [PARAMETER]* ')' ~ [Character*]EOF
 */
static void
parse_CPlusPlusDeclaration (int identifier_index)
{
  char curr_char;
  int className_index;
  char *class_name;
  boolean is_destructor;
  int functionIdentifier_index;

  /* ===== declaration heading including class name */
  className_index = identifier_index;
  class_name = (char *) (code_info + className_index);
  if (skipIf_comments_and_spaces (pk_false));	/* FPIN = PREFIX # [SPACE]* ~ [Character*] [EOF] */

  /* ===== method name */
  if (getc_if_logicalNext_char_is ('~'))
    {
      is_destructor = pk_true;
      if (skipIf_comments_and_spaces (pk_false));
    }
  else
    is_destructor = pk_false;

  if (parse_identifier (&functionIdentifier_index))
    was_ident = pk_true;	/* if code_info <res_index>+strlen(IDENT): position is that of '\0' */
  if (codeInfo_index == identifier_index)
    exitERR_line ("missing expected identifier");
  if (skipIf_comments_and_spaces (pk_false));	/* FPIN = PREFIX # [SPACE]* ~ [Character*] [EOF] */

  /* ===== parameters */
  curr_char = get_logicalChar ();
  if (curr_char == '(')
    {
      if (className_index > 0)	/* code_info contains strings preceding className_index */
	vputCplusPlus_declarationPrefix (className_index);
      /* printing class name as a comment */
      vput_commentStr (class_name, "::");	/* preceded and followed by a unique separator */
      if (is_destructor)
	vput_char ('~');
      complete_prototypeExtraction (pk_true, functionIdentifier_index);
    }
  else
    {
      char_backward (curr_char);
      skipTo_endOfUnit (pk_false);
    }
}

/* go through the file extracting functions */
static void
parse_file ()
{
  char curr_char;
  int identifier_index = -1;

  first_comment[0] = '\0';
  last_comment[0] = '\0';

  init_reading ();
  reset_codeInfo ();

  while ((curr_char = get_logicalChar ()) != EOF)
    {
      if (curr_char == '\n')
	{
	  if (skip_spacesAndDirectives_inNewLogiLine (pk_true))	/* FPIN  = [Character*][eol] # { SPACES DIRECTIVE }+ SPACES ~ 
								   {[Character_but_'#'][Character*]}[EOF]  
								   line count is incremented for [eol] occurrence preceding initial position */
	    reset_fileUnit ();	/* a directive occured */
	  confirm_space ();
	  free_line = pk_true;
	}
      else if (isspace (curr_char))
	{
	  if (skip_logicalSpaces_but_eol ());
	  confirm_space ();
	}
      else if (curr_char == '/')
	{
	  if (!parseIf_comment_completion ())
	    {			/* FPIN = [Character*]['/'] #~ [Character*][EOF]; free_line is set */
	      was_ident = pk_false;
	    }
	}
      else if (curr_char == '"')
	{
	  /* --------    --- string mark */
	  /* nuke string quotes -- must be on same line */
	  skip_C_string_completion ();	/* exit on failure, '"' is the last read character  */
	  free_line = pk_false;
	  was_ident = pk_false;
	  confirm_space ();
	}			/* end reading string */
      else if (curr_char == '\'')
	{
	  /* --------    --- character mark */
	  /* nuke character quotes -- must be on same line */
	  skip_C_characterConstant_completion ();	/* exit on error ''' is the last read character  */
	  free_line = pk_false;
	  was_ident = pk_false;
	}			/* exit on error */
      else if (curr_char == ';')
	{
	  reset_fileUnit ();
	  free_line = pk_false;
	  was_ident = pk_false;
	}
      else if (curr_char == '{')
	{
	  /* --------    --- block opening */
	  /* reset recording process */
	  skip_braceInner ();
	  free_line = pk_false;
	  was_ident = pk_false;
	}
      else if (curr_char == '}')
	{
	  /* --------    --- block closing */
	  exitERR_line ("extra right brace encountered");
	}
      else if (curr_char == '(')
	{
	  if (was_ident)
	    {
	      parse_CDeclaration (identifier_index);
	      reset_fileUnit ();
	      free_line = pk_true;
	      was_ident = pk_false;
	    }
	  else
	    skip_parenthesisInner ();
	}
      else if (curr_char == ')')
	exitERR_line ("unexpected ')' character");
      else if (curr_char == ']')
	exitERR_line ("unexpected ']' character");
      else if (curr_char == '[')
	{
	  skip_squareBracketInner ();
	  free_line = pk_false;
	  was_ident = pk_false;
	}
      else if (curr_char == ':')
	{
	  if (getc_if_logicalNext_char_is (':'))
	    {
	      if (codeInfo_index > 0)
		{
		  if (code_info[codeInfo_index] != '\0')
		    exitERR ("expected class identifier before double ':'");
		}		/* exit */
	      else
		exitERR ("expected class identifier before double ':'");	/* exit */
	      parse_CPlusPlusDeclaration (identifier_index);
	      reset_fileUnit ();
	    }
	  else			/* perhaps a C++ call to ancestor creator that has to be removed */
	    append_NOTspace (':');
	  free_line = pk_false;
	  was_ident = pk_false;
	}
      else if (curr_char == '_')
	{
	  // DEBUG_code_info("before ident reading");  
	  identifier_index = parse_subIdentifierCompletion (curr_char);
	  free_line = pk_false;
	  was_ident = pk_true;
	}			/* * returns <res_index>+strlen(<curr_char>IDENT): position is that of '\0' */
      else if (isalpha (curr_char))
	{
	  // DEBUG_code_info("before ident reading");  
	  identifier_index = parse_identifierCompletion (curr_char);
	  free_line = pk_false;
	  was_ident = pk_true;
	}			/* * returns <res_index>+strlen(<curr_char>IDENT): position is that of '\0' */
      else
	{
	  append_char (curr_char);
	  free_line = pk_false;
	  was_ident = pk_false;
	}
    }
}

// #define toto tutu
/** extract prototypes of C/C++ functions defined in in_filePath
 */
/* bla bla */

void
file_proceed (char *in_filePath, char *out_filePath)
{
  open_inputFile (in_filePath);	/* exit on error */
  open_outputFile (out_filePath);	/* exit on error */
  if (! exportOnlyCommentedFunctions) put_filePrelude ();		/* if header_string == NULL or is empty nothing is done */
  parse_file ();
  if (fileComment_printable)
    if (!fileComment_printed)
      actualPrint_headerFileComment ();
  put_filePostlude ();		/* if header_string == NULL or is empty nothing is done */
  close_outputFile ();
  close_inputFile ();
}
