/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Contructor.
   Exclusive clocks operation using matrix calculus technic for a graph (g).
*/
  exclusiveClocks::exclusiveClocks (ptDCG g)
{
  CKM = g->getClockManager ();
  AGRAPH = g;
}

/**
   Destructor.
*/
exclusiveClocks::~exclusiveClocks ()
{
}

/**
   It computes the exclusive clocks applied to the clocks marked by a mark (m), 
   It applies 
   - the selectClocks() operation, 
   - followed by the lowerBoundPairOfClocksCalculus() one, in which the hierarchy is exploited to avoid to compute all the cases. 
   - then, it assigns the result (setAttribute()) as an attribute for each clock. This attribute is a predefined 
   one called EXCLUSIVE_CLOCKS. It is defined in the file $GraphVM_export/CompilerAttributes.h.

   \note when m is equal to UNDEFINED_MARK, all the clocks are selected.
*/
void
exclusiveClocks::apply (typeMark m)
{
  selectClocks (m);
  lowerBoundPairOfClocksCalculus ();
  setAttribute ();
}
